#!/bin/bash

show_help()
{
    cat <<EOF

    Usage:
    DeclareFcl.sh [options] file1 file2 ....
    
    Declares listed fcl files to SAM, copies them to a desired location, and adds their location to SAM.
    
    Options:
    --dir <output_dir> Sets the output directory for the fcl files [default: /nova/data/mc/fcl]
    -v                 Verbose
    --test             Test run (do not issue SAM commands)
    --dev              Use the development SAM db
EOF
}

get_metadata()
{
    metadataFile=/tmp/meta$$
    size=`cat $1 | wc -c`
    chkSum=`samweb file-checksum $1`
    metadata=`parseFCLMetadata $1`

    jsonData=`cat <<-EOF
    {
	"file_name": "$1"
	,"file_type": "mc_config"
	,"file_size": $size
	,"crc":        $chkSum
	,"file_format": "fcl"                       
	,$metadata
	}`

    echo "$jsonData" > $metadataFile

    if [ -n "$VERBOSE" ]; then
	cat $metadataFile
    fi
}

declare_metadata()
{
    if [ $DEVEL ]; then
	devOpt="--dev"
    else
	devOpt=""
    fi
    
    if [ -n "$VERBOSE" ]; then
	echo samweb declare-file $devOpt -e nova $metadataFile
    fi

    if [ -z "$TESTING" ]; then 
	samweb declare-file $devOpt -e nova $metadataFile
    fi

    if [ -n "$VERBOSE" ]; then
	echo samweb add-file-location $devOpt -e nova $file novadata:$DIR
    fi

    if [ -z "$TESTING" ]; then 
	samweb add-file-location $devOpt -e nova $file novadata:$DIR
    fi

    cp $file $DIR/$file

    rm $metadataFile
}

fileList=()

DIR=/nova/data/mc/fcl

while [ $# -gt 0 ]
  do
  case $1 in
    -h|--help|-\?)
         show_help
         exit 0
         ;;
    -v)
         VERBOSE=1
	 shift
         ;;	 
    --test)
         TESTING=1
         shift
         ;;	 
    --dev)
         DEVEL=1
         shift
         ;;
    --dir)
         shift
         if [ $# -gt 0 ]
	     then
	     DIR=$1
	     shift
	 else
	     echo "No directory set!"
	     exit 1
	 fi
	 ;;
    -*)
         echo "Unknown option: $1"
         exit 1
         ;;
    *)                 	 
         fileList+=($1)
         shift
         ;;	 
  esac
done

if [ $VERBOSE ]; then
    echo Output directory: "$DIR"
    if [ -n "$DEVEL" ]; then
	echo Using development db
    fi
    if [ -n "$TESTING" ]; then
	echo Test run only: No changes to SAM db
    fi
fi

if [ ! -e $DIR ]; then 
    if [ -n "$VERBOSE" ]; then
	echo Output directory does not exist: creating
    fi
    mkdir $DIR
else
    if [ ! -d $DIR ]; then
	echo Chosen output directory already exists and is not a directory!
	exit 1
    fi
fi

for file in "${fileList[@]}"
  do
  if [ -n "$VERBOSE" ]; then
      echo Parsing file: "$file"
  fi
  get_metadata "$file"
  declare_metadata
done

exit 0
