#! /usr/bin/env python 

import sys, os, stat
import argparse
from collections import OrderedDict
import code

PERMISSION_RW_R = stat.S_IRUSR | stat.S_IWUSR | stat.S_IRGRP # Permission mask: user r,w; group r 
PERMISSION_RW_R_R = stat.S_IRUSR | stat.S_IWUSR | stat.S_IRGRP | stat.S_IROTH  # Permission mask: user r,w; group r; others r 

def checkIsFile(path):
  if os.path.isfile(path):
    return 1  
  print "Warning, not a file: ", path
  return 0 

def setPermission(path, permission=PERMISSION_RW_R_R):
  if checkIsFile(path):
    os.chmod(path, permission)  

def tail(path, window=20):
    f = open(path, 'r')
    BUFSIZ = 1024
    f.seek(0, 2)
    bytes = f.tell()
    size = window
    block = -1
    data = []
    while size > 0 and bytes > 0:
        if (bytes - BUFSIZ > 0):
            # Seek back one whole BUFSIZ
            f.seek(block*BUFSIZ, 2)
            # read BUFFER
            data.append(f.read(BUFSIZ))
        else:
            # file too small, start from begining
            f.seek(0,0)
            # only read what was not read
            data.append(f.read(bytes))
        linesFound = data[-1].count('\n')
        size -= linesFound
        bytes -= BUFSIZ
        block -= 1
    return '\n'.join(''.join(data).splitlines()[-window:])



class Process:
  # Constructor 
  def __init__(self, num, proj):
    self.num     = num 
    self.msgs    = []
    self.proj    = proj
  
  # Implementation of str()
  def __str__(self):
    return "Process " + str(self.num) + ", last message: \n" + self.lastMsg()
  
  # How the thing represents it self, notably for tostring and the "in" operator 
  def __repr__(self):
    return str(self)
  
  # How the thing represents it self, notably for tostring and the "in" operator 
  def __contains__(self, test):
    return test in self.lastMsg()
  
  # Implementation of == 
  def __eq__(self,other):
    return self.num == other.num and self.proj.logPath == other.proj.logPath

  # Get the number of messages in the .log file 
  def nMsgs(self):
    return len(self.msgs)
  
  # Get the last message, usually the most interesting one
  def lastMsg(self):
    return self.msgs[len(self.msgs) - 1]
  
  # Add a message to the process, intended for use in the Project class when parsing .log files
  def addMsg(self, msg):
    self.msgs.append(msg)

  # Print all messages for this process 
  def printMsgs(self):
    for msg in self.msgs:
      print msg
  
  ## Get the path to the std err file 
  def errPath(self):
    path = self.proj.logPath.replace(".log", "_" + str(self.num) + ".err")
    checkIsFile(path)
    return path 

  ## Get the path to the std out file 
  def outPath(self):
    path = self.proj.logPath.replace(".log", "_" + str(self.num) + ".out")
    checkIsFile(path)
    return path

  ## Tail the std err file 
  def errTail(self, window=20): 
    path = self.errPath()
    if not checkIsFile(path):
      return 
    return tail(path, window)
    
  ## Tail the std out file 
  def outTail(self, window=20): 
    path = self.outPath()
    if not checkIsFile(path):
      return 
    return tail(path, window)
     
  ## Open the std out file with less
  def errLess(self): 
    path = self.errPath()
    if not checkIsFile(path):
      return 
    print os.system("less " + path)
    return  
  
  ## Open the std err file with less
  def outLess(self): 
    path = self.outPath()
    if not checkIsFile(path):
      return 
    print os.system("less " + path)
    return  

  ## Find the return code from the last message, -1 if last message is not normal termination 
  def retCode(self):
    if not "return value" in self.lastMsg():
      return -1 
    start = self.lastMsg().find("return value")
    
    end  = self.lastMsg().find(")", start)
    valStr = self.lastMsg()[start:end]
    values = valStr.split()
    try: 
      return int(values[2])
    except:
      raise Exception("Can't find return code, improperly fomatted lastMessage: " + self.lastMsg())


class Project:
  def __init__(self, logPath):
    self.logPath = logPath        # Path to log file
    self.msgs    = []             # List of all messages from .log file, not sorted by process
    self.procs   = OrderedDict()  # Main dictionary of processes, indexed by condor job number.  
    self.decomp  = dict()         # 
    self.nNormal = 0              # Number of processes that ended normally  (return value == 0 )
    self.normal = []              # List process objects that ended normally  (return value == 0 )
    self.nAbnormal = 0            # Number of processes that ended abnormally  (return value != 0 or any other last message)
    self.abnormal = []            # List process objects that ended abnormally (return value != 0 or any other last message)

    if self.logPath.endswith(".cmd"):
      print "Condor .cmd file supplied, swapping extension for .log"
      self.logPath = logPath.replace(".cmd", ".log")
    self.separateMessages()
    self.processRecon()
    self.inspectProcs()
    self.decomposeModes()

  ## separateMessages() turns all of the separate messages in the file into entries in a list. 
  ### Need to break the log into chunks.  Many messages from many jobs
  ### Example: 
  ###  ...
  ###  009 (17365926.880.000) 08/06 21:16:09 Job was aborted by the user.
  ###        The system macro SYSTEM_PERIODIC_REMOVE expression '((NumJobStarts > 3) || (ImageSize>=4000000) ||   (JobStatus == 2 && JobUniverse == 5 && ((CurrentTime - EnteredCurrentSt
  ###         atus) > 86400*3)))' evaluated to TRUE
  ###  ...
  ### Seems we can rely on those dots, here we go
  def separateMessages(self):
    log = open(self.logPath, 'r')
    msg = ""
    for line in log:
      if line.strip() == "...": 
        self.msgs.append(msg) # append the current message to the list of messages 
        msg = "" # Reset the current message
      else:
        msg += line
    

  ### processRecon() loops over all the messages to make Process objects, fills a dict
  ### Relies on separateMessages() to have been called first, but that happens properly in the constructor 
  def processRecon(self):
    for msg in self.msgs:
      smsg = msg.split()

      if len(smsg) > 2:
        if "028" in smsg[0]:
          continue
        if smsg[1][0] == "(":
          numStr = smsg[1]
          num = int(numStr.split(".")[1])
          if num in self.procs.keys():
            self.procs[num].addMsg(msg)
          else:
            self.procs[num] = Process(num, self)  
            self.procs[num].addMsg(msg)
 
  ### Inspect the processes, make sure the numbers jive and decompose them into "normal" and "abnormal"
  def inspectProcs(self):
    print "Found",  self.nProcs(), "processes." 
    lastKey = None
    sequential = True 
    firstKey = None
    for  key in sorted(self.procs.keys()):
      if lastKey == None:
        lastKey = key
        firstKey = key 
        continue
      diff = key - lastKey
      if not diff == 1:
        sequential = False
        break
      lastKey = key 

    if sequential and firstKey == 0: 
      print "Process numbers are sequential starting with",str(firstKey)+", things seem ok."
    elif sequential: 
      print "Process numbers are sequential, but do not start with 0.  Be careful with your indexing."  
      print "First process number is", str(firstKey) + "."
    else:
      print "Process numbers are not sequential, be careful with your indexing."


    for proc in self.procs.values():
      if "return value 0" in str(proc):
        self.nNormal += 1
        self.normal.append(proc)
      else: 
        self.nAbnormal += 1  
        self.abnormal.append(proc) 
    print self.nNormal, "of those processes ended normally with status 0"

  ##  
  def decomposeModes(self):
    for proc in self.procs.values():
      splitMsg = proc.lastMsg().split()
      try:
        mode =  int(splitMsg[0])
      except:
        raise Exception("Could not find mode (three digits) at the beginning of this message: " + str(proc))
      
      if not mode in self.decomp.keys():
        self.decomp[mode] = []
      
      self.decomp[mode].append(proc)
    sum = 0
    for mode in self.decomp.values(): 
      sum += len(mode)
    if not sum == self.nProcs():
      raise Exception("Sum of decomposed processes not equal to total number of processes.")
    
  def printModes(self):
    print "Jobs have", len(self.decomp), "different modes.  Here is an example of each."
    for mode in self.decomp.keys():
      print "* Mode", mode, "-- there are", len(self.decomp[mode]), "of these, one example is:"
      for line in str(self.decomp[mode][0]).splitlines():
        print "\t", line
      print ""

  ## Number of process objects in dict 
  def nProcs(self):
    return len(self.procs)  

  ## Implementation of [] operator for indexing
  def __getitem__(self, n):
    return self.procs[n]
  
  # Implementation of len(), returns number of items in dictionary 
  def __len__(self):
    return len(self.procs)
  
  # Implementation of iter, use as "for thing in proj"
  def __iter__(self):
    for key in self.procs.keys():
      yield self.procs[key]
 
  ## Make .log and all .err and .out group readable and user read/writable 
  def makeReadable(self, permission=PERMISSION_RW_R_R):
    setPermission(self.logPath, permission)
    for proc in self.procs.values(): 
      setPermission(proc.outPath(), permission)
      setPermission(proc.errPath(), permission)


def main(argv):
  parser = argparse.ArgumentParser(description="""Loop over the ".log" file and find the final state of each 
                                                 process, then find the .err and .out""")
  parser.add_argument('logFile', help='The input file to run over', type=str)
  args = parser.parse_args()
  logPath = args.logFile
  proj = Project(logPath)
  return proj 



if __name__ == "__main__":
  proj = main(sys.argv)
  code.interact(local=locals(), banner="")

