#!/usr/bin/env python
"""
Compute operations statistics (uptime, POT recorded, delivered, etc.)
for NOvA.

Based on original work of Jon Paley.

messier@indiana.edu
"""
import sys
import getopt
import shlex, subprocess
import re, urllib2, csv
from datetime import date, datetime, timedelta
import psycopg2
import psycopg2.extras

#.......................................................................
DETECTOR = 'fardet'

#.......................................................................
#
# Do we want the downtime report? Uptime report?
#
DO_DOWNTIME_REPORT = False
DO_UPTIME_REPORT   = True
#
# The shortest downtime worth reporting
#
DOWNTIME_MIN = timedelta(minutes=2)

#.......................................................................
#
# If we are asked for just one run summary, set this to the value of
# the run number
#
DO_RUN_POT_SUMMARY = 0
    
#.......................................................................
#
# Real spills will have more than this number (time 1e12) POT recorded
#
GOOD_SPILL_POT = 0.05

#.......................................................................
#
# Some database variables
#
DB_CURSOR     = None
DB_QUERY_TIME = timedelta(0)

def connect_to_database():
    """ Establish connection to runs database  
        JMP 2015-04-06: This function seems to not be used for anything.  
        Might want to remove it.  But just in case, changed dB connection info
        to use streaming replica of FarDet DAQ dB.  Use port 5434 for the 
        NearDet, otherwise this is for FarDet
    """
    try:
        conn = psycopg2.connect(\
            "dbname=nova_prod "\
            "host=ifdbrep.fnal.gov "\
            "user=nova_reader "\
            "password=nova2pass port=5436"\
            )
        global DB_CURSOR
        DB_CURSOR = conn.cursor(cursor_factory=psycopg2.extras.DictCursor)
    except:
        print "I am unable to connect to the database"
                                                
#.......................................................................

def in_unix(input):
    """ Convert a datetime to UNIX time

    Args:
    input: A datetime object

    Returns:
    A floating points number of seconds since Jan. 1, 1970 (UNIX standard)
    
    """
    start = datetime(year=1970,month=1,day=1,hour=0,minute=0,second=0)
    diff = input - start
    return diff.total_seconds()

#.......................................................................

def nactive_chans(run):
    """ Return how many active channels there we at start of run

    Args:
    run: Run number

    Returns:
    Number of active channels (APD pixels)
    """
    if (run==0):
        return 0
    sql = "select nactivechan from "+DETECTOR+".runs where run="+str(run)
    DB_CURSOR.execute(sql)
    n = DB_CURSOR.fetchall()
    return n[0]['nactivechan']
    
#.......................................................................

def runs_for_period(day0, day1):
    """ Get all runs that start or stop within a particular date range

    Args:
    day0 : A datetime object for first day
    day1 : A datetime object for the last day
    """
    sql0 = "select run,tstart,tstop from "+DETECTOR+".runs";
    cut1 = " where partition=1";
    cut2 = " tstop is not null"
    cut3 = "tstart >= '{}'".format(day0.strftime("%Y-%m-%dT%H:%M:%S"))
    cut4 = "tstart < '{}'".format(day1.strftime("%Y-%m-%dT%H:%M:%S"))
    
    cut5 = "tstop >= '{}'".format(day0.strftime("%Y-%m-%dT%H:%M:%S"))
    cut6 = "tstop < '{}'".format(day1.strftime("%Y-%m-%dT%H:%M%:S"))

    sql1 = "order by tstart asc"
    
    SQL = \
        sql0 + \
        cut1 + " and " + \
        cut2 + " and " + \
        "((" + cut3+" and "+cut4+") or ("+cut5+" and "+cut6+")) " + \
        sql1

    DB_CURSOR.execute(SQL);
    runs = DB_CURSOR.fetchall();

    return runs

#.......................................................................
    
def runs_by_day(day0):
    """ Get a complete list of run numbers for a specified day

    Any run that starts or stops on the specified date is contaned in
    the list

    Args:
    day0: A datetime object for the requested day

    Returns:
    An list of SQL results for run,tstart,tstop ordered by start time
    """
    day1 = day0+timedelta(days=1)
    return runs_for_period(day0, day1)

#.......................................................................

def run_minmax(runs):
    """ Find the highest and lowest run number in a list

    Args:
    runs: A list of SQL results (run,tstart,tstop)

    Returns:
    (runlo,runhi), the lowest and highest run numbers in the list
    """
    runmin = runs[0]['run']
    runmax = runs[0]['run']
    for r in runs:
        if (r['run']<runmin):
            runmin = r['run']
        if (r['run']>runmax):
            runmax = r['run']
    return (runmin,runmax)

#.......................................................................

def compute_uptime(day,runs):
    """ Compute the number of live hours in a particular day.

    Also computes the fractional up time (by live seconds) for the day

    Args:
    day (datetime): The date in question
    runs (list,int): List of runs selected to be on that day

    Returns:
    (livetime, fraction): The uptime in hours (timedelta) and the
    fractional livetime
    """

    daystart = day
    dayend   = day + timedelta(days=1)
    sum = timedelta(0)

    for r in runs:
        t1  = r['tstart']
        t2  = r['tstop']
        #
        # Handle end points. Runs that carried into this day and runs
        # that extedned beyond the end of this day should be pinnedto
        # the start and end of the day.
        #
        if (t1<daystart):
            t1 = daystart
        if (t2>dayend):
            t2 = dayend

        sum += (t2-t1)
    #
    # Return the total number of hours of livetime and the fractional
    # livetime for the day
    #
    return (sum,sum.seconds/(24.0*60.0*60.0))

#.......................................................................

def spills_between(t1, t2):
    """
    Return spill information (unix time and POT) for a datetime range

    Args:
    t1 (datetime): Start time
    t2 (datetime): End time

    Returns:
    List of (time,pot) tuples
    """
    global DB_QUERY_TIME

    spills = [];    

    tf1 = in_unix(t1)
    tf2 = in_unix(t2)
    webaddress = 'http://ifb-data.fnal.gov:8099'\
                 '/ifbeam/data/data?'\
                 'v=E:TRTGTD&e=e,a9&t0={}&t1={}&&f=csv'.\
                 format(tf1,tf2)

    dd = None
    #
    # Make three tries and then bail...
    #
    t0 =  datetime.utcnow()
    try:
        dd = urllib2.urlopen(webaddress,None,600)
    except:
        try:
            dd = urllib2.urlopen(webaddress,None,600)
        except:
            try:
                dd = urllib2.urlopen(webaddress,None,600)
            except:
                return spills
    t1 =  datetime.utcnow()

    DB_QUERY_TIME += (t1-t0)
    
    if (dd==None):
        return spills
    
    csvdata = csv.reader(dd)

    for row in csvdata:
        if (row.count('E:TRTGTD')>0):
            if (row[4]!='null'):
                spills.append( (float(row[2])/1000., float(row[4])) )

    return spills

#.......................................................................

def pot_between(t1,t2,spills):
    """ Compute the POT recorded between two time stamps.

    Args:
    t1 (datetime)              : Start time
    t2 (datetime)              : End time
    spills ( list[(time,pot)]) : Complete list of spills for time period

    Returns:
    pot (float) : Protons in units of 1e18 POT
    """

    tf1 = in_unix(t1)
    tf2 = in_unix(t2)

    sum = 0.0
    for (t,q) in spills:
        if ((t>=tf1)&(t<=tf2)&(q>GOOD_SPILL_POT)):
            sum += q
    #
    # Database is in units of 1e12, I want 1e18.
    #
    return sum*(1.0E12/1.0E18)

def nspills_between(t1,t2,spills):
    """ Count the number of good spills between t1 and t2

    Args:
    t1 (datetime)              : Start time
    t2 (datetime)              : End time
    spills ( list[(time,pot)]) : Complete list of spills for time period

    Returns:
    nt (int) : Number of spills between t1 and t2
    """

    tf1 = in_unix(t1)
    tf2 = in_unix(t2)

    sum = 0.0
    for (t,q) in spills:
        if ((t>=tf1)&(t<=tf2)&(q>GOOD_SPILL_POT)):
            sum += 1

    return sum

#.......................................................................

def compute_pot(day,runs,spills):
    """ Tally the protons on target delivered and recorded in a day

    Args:
    day                       : datetime of day of interest
    runs   (sql)              : SQL set of runs (run, tstart, tend)
    spills (list[(time,pot)]) : Complete list of spills for time period

    Returns:
    (pot_recorded,pot_delivered) (float,float)
    """
    daystart = day
    dayend   = day + timedelta(days=1)

    pot_delivered = pot_between(daystart,dayend,spills)

    pot_recorded = 0.0
    for r in runs:
        t1  = r['tstart']
        t2  = r['tstop']
        #
        # Handle end points. Runs that carried into this day and runs
        # that extedned beyond the end of this day should be pinned to
        # the start and end of the day.
        #
        if (t1<daystart):
            t1 = daystart
        if (t2>dayend):
            t2 = dayend

        pot_recorded += pot_between(t1,t2,spills)

    return (pot_recorded,pot_delivered)

#.......................................................................

def run_pot_summary(run):
    """ Print a single line summary of a particular run number
    """
    #
    # Querry runs database to find start and end time of run
    #
    sql0 = "select tstart,tstop from "+DETECTOR+".runs";
    cut1 = " where partition=1";
    cut2 = " tstop is not null"
    cut3 = " run=%d"%run
    
    SQL = \
        sql0 + \
        cut1 + " and " + \
        cut2 + " and " + \
        cut3;

    DB_CURSOR.execute(SQL);
    run_data = DB_CURSOR.fetchall();

    #
    # Collect the POT inside the time interval
    #
    t1 = run_data[0]['tstart']
    t2 = run_data[0]['tstop']
    spills = spills_between(t1,t2)
    npot = pot_between(t1,t2,spills)
    
    print run,t1,t2,(t2-t1),npot
    

#.......................................................................


def downtime_report(day,runs,spills):
    """
    """
    daystart = day
    dayend   = day + timedelta(days=1)

    lastrun = 0
    tstop   = daystart
    for r in runs:
        #
        # Compute the time difference between the last run stop and
        # the current run start
        #
        run    = r['run']
        tstart = r['tstart']
        dt = (tstart-tstop)

        if (dt>DOWNTIME_MIN):
            pot = pot_between(tstop,tstart,spills)
            print lastrun,'\t',run,'\t',tstop,'\t',tstart,'\t',dt,'\t',pot
        #
        # Save the stop time of this run
        #
        lastrun = run
        tstop   = r['tstop']
    #
    # Check for any downtime between last run and midnight 
    #
    run = 0
    tstart = dayend    
    dt = (tstart-tstop)
    if (dt>DOWNTIME_MIN):
        pot = pot_between(tstop,tstart,spills)
        print lastrun,'\t',run,'\t',tstop,'\t',tstart,'\t',dt,'\t',pot

#.......................................................................

def usage():
    """ Print how to use the program
    """
    print "Usage:";
    print "% python nova-uptime.py -e mm/dd/YY -s /mm/dd/YY -n [#]"
    print "Where:"
    print "--neardet : Make summary for neardet (default is fardet)"
    print "-d : Do the downtime report instead of uptime report"
    print "-e : Last date to report"
    print "-s : First date to report"
    print "-n : Number of days to report"
    print "-m : Minimum number of minutes to report in downtime mode"
    print ""
    print "Any combinated of s, e, and n may be used:"
    print " -e 9/7/2013 -n 7  : Seven days ending on Sept. 7th, 2013"
    print " -s 10/1/2013 -n 10 : Ten days starting on Oct. 1st, 2013"
    print " -s 12/1/2013 -e 12/31/2013 : The entire month of December"
    print ""
    print " 'yesterday' and 'today' are a valid dates"
    print ""
    print "With nothing specified, default report is previous 7 days from yesterday."
    print "";
    print "In 'uptime' mode the data reported are:"
    print " date,day,run start,run end,uptime in hrs,uptime frac.,1e12 POT recorded,1e12 POT delivered,POT frac."
    print "";
    print "In 'downtime' mode the data reported are:"
    print " run last,run next,start of downtime,end of downtime,duration,1e12 POT lost"
    
#.......................................................................

def options():
    """ Parse the command line options
    Returns:
    t1 (datetime) : First day to report
    t2 (datetime) : Last day to report
    n  (int)      : Number of days to report
    """
    global DO_DOWNTIME_REPORT
    global DO_UPTIME_REPORT
    global DO_RUN_POT_SUMMARY
    global DOWNTIME_MIN
    global DETECTOR
    
    date_start = None
    date_end   = None
    ndays      = None
    try:
        (opts, args) = getopt.getopt(sys.argv[1:],\
                                     "hds:e:n:m:r:a",\
                                     ["help",\
                                      "downtime",\
                                      "start=",\
                                      "end=",\
                                      "ndays=",\
                                      "downtime_min",
                                      "run=",
                                      "neardet"])
    except getopt.GetoptError as err:
        print str(err)
        sys.exit(2)
    for opt, arg in opts:
        if opt in ("-h", "--help"):
            usage()
            exit(0)
        elif opt in ("-d", "--downtime"):
            DO_DOWNTIME_REPORT = True
            DO_UPTIME_REPORT   = False            
        elif opt in ("-s", "--start"):
            date_start = arg
        elif opt in ("-e", "--end"):
            date_end = arg
        elif opt in ("-n", "--ndays"):
            ndays = int(arg)
        elif opt in ("-r", "--run"):
            DO_RUN_POT_SUMMARY = int(arg)
            return (0,0,0)
        elif opt in ("-m", "--downtime_min"):
            DOWNTIME_MIN = timedelta(minutes=int(arg))
        elif opt in ("-a", "--neardet"):
            DETECTOR = 'neardet'
        else:
            assert False, "unknown option"
    #
    # If no options are given, default to most recent 7 days ending on
    #"yesterday"
    #
    if ((date_start==None) & (date_end==None) & (ndays==None)):
        ndays = 7
        t0 =  datetime.utcnow()
        t0 -= timedelta(days=ndays)
        t0 =  t0.replace(hour=0, minute=0, second=0, microsecond=0)
        t1 =  t0
        t1 += timedelta(days=ndays)
        return (t0, t1, ndays)
    #
    # Convert start and end date text to datetime object
    #
    if (date_start!=None):
        if (date_start=='yesterday'):
            t0 =  datetime.utcnow()
            t0 -= timedelta(days=1)
            t0 =  t0.replace(hour=0, minute=0, second=0, microsecond=0)
        elif (date_start=='today'):
            t0 =  datetime.utcnow()
            t0 =  t0.replace(hour=0, minute=0, second=0, microsecond=0)
        else:
            t0 = datetime.strptime(date_start,'%m/%d/%Y')
    if (date_end!=None):
        if (date_end=='yesterday'):
            t1 =  datetime.utcnow()
            t1 -= timedelta(days=1)
            t1 =  t1.replace(hour=0, minute=0, second=0, microsecond=0)
        elif (date_end=='today'):
            t1 =  datetime.utcnow()
            t1 =  t1.replace(hour=0, minute=0, second=0, microsecond=0)
        else:
            t1 = datetime.strptime(date_end,'%m/%d/%Y')

    #...................................................................
    #
    # Check the information I have, fill in blanks or check for
    # internal consistency
    #
    if ((date_start==None)&(date_end!=None)&(ndays!=None)):
        #
        # Need to supply a start time
        #
        t0 =  t1
        t0 -= timedelta(days=(ndays-1))
    if ((date_start==None)&(date_end!=None)&(ndays==None)):
        #
        # Need to supply a start time and a number of days
        #
        ndays = 7
        t0 =  t1
        t0 -= timedelta(days=(ndays-1))
    if ((date_start!=None)&(date_end==None)&(ndays!=None)):
        #
        # Need to supply an end time
        #
        t1 =  t0
        t1 += timedelta(days=ndays)
    if ((date_start!=None)&(date_end!=None)&(ndays==None)):
        #
        # Need to supply a number of days
        #
        n = (t1-t0)+timedelta(days=1)
        ndays = int(n.days)
    if ((date_start!=None)&(date_end!=None)&(ndays!=None)):
        #
        # Have everything we need, but are they consistent?
        #
        d = (t1-t0)+timedelta(days=1)
        n = int(d.days)
        if (ndays != n):
            print 'Inconsistent time range'
            exit(2)
            
    return (t0, t1, ndays)
    
#.......................................................................
#
# "Main" function
#
def main():
    (t0, t1, ndays) = options()
    
    if (DETECTOR=='fardet'):
        try:
            conn = psycopg2.connect(\
                "dbname=nova_prod "\
                    "host=ifdbrep.fnal.gov "\
                    "user=nova_reader "\
                    "password=nova2pass port=5436"\
                    )
        except:
            print "I am unable to connect to the database"
    else:
        try:
            conn = psycopg2.connect(\
                "dbname=nova_prod "\
                    "host=ifdbrep.fnal.gov "\
                    "user=nova_reader "\
                    "password=nova2pass port=5434"\
                    )
        except:
            print "I am unable to connect to the database"
            
    global DB_CURSOR
    DB_CURSOR = conn.cursor(cursor_factory=psycopg2.extras.DictCursor)

    if (DO_RUN_POT_SUMMARY!=0):
        run_pot_summary(DO_RUN_POT_SUMMARY)
        return

    date0     = date(    year=t0.year, month=t0.month, day=t0.day)
    datetime0 = datetime(year=t0.year, month=t0.month, day=t0.day)

    weekdays = ['Mo','Tu','We','Th','Fr','Sa','Su']

    for i in range(0,ndays):
        datei     = date0     + timedelta(days=i)
        datetimei = datetime0 + timedelta(days=i)
        datetimef = datetimei + timedelta(days=1)

        #
        # Get the complete list of runs for the day
        #
        runs = runs_by_day(datei)

        #
        # Get the complete list of spills for the day
        #
        spills = spills_between(datetimei, datetimef)
        
        #
        # Compute the fraction of the day a run was going
        #
        r1 = 0
        r2 = 0
        uptime = 0
        upfrac = 0
        if (runs):
            (r1,r2) = run_minmax(runs);
            (uptime,upfrac) = compute_uptime(datetimei, runs)
        #
        # Compute how many protons were delivered and recorded
        #
        (pot_rec,pot_del) = compute_pot(datetimei, runs, spills)

        if (DO_DOWNTIME_REPORT):
            downtime_report(datetimei, runs, spills)
        #
        # Compute the fraction of POT recorded. In cases where the POT
        # delivered is zero, record the fraction recorded as the up time
        # computed above.
        #
        pot_frac = 0.0
        if (pot_del==0.0):
            pot_frac = upfrac
        else:
            pot_frac = pot_rec/pot_del

        if (DO_UPTIME_REPORT):
            print datei,",",\
                  weekdays[datei.weekday()],",",\
                  r1,",",\
                  r2,",",\
                  uptime,",",\
                  upfrac,",",\
                  pot_rec,",",\
                  pot_del,",",\
                  pot_frac;
    print 'Completed summary for',DETECTOR
    print 'Total time in IFBEAMs queries:',DB_QUERY_TIME
#.......................................................................

if __name__ == "__main__":
    main()
########################################################################
    
