///
/// \file dEdxCalculator.h 
/// \brief Calculates dE/dx in cells passed through by a track.
///
#ifndef BPFIT_DEDXCALCULATOR_H
#define BPFIT_DEDXCALCULATOR_H

#include <vector>
#include "TVector3.h"

#include "art/Persistency/Common/Ptr.h"
#include "art/Persistency/Common/PtrVector.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"

#include "Calibrator/Calibrator.h"
#include "Geometry/Geometry.h"
#include "GeometryObjects/OfflineChan.h"
#include "GeometryObjects/GeometryBase.h"

namespace rb{ class Track; }
namespace rb{ class CellHit; }

namespace bpfit
{

  class dEdxCalculator {
  public:
    ///
    /// \brief Constructor...
    ///
    dEdxCalculator();
    
    ///
    /// \brief Compute dE/dx for all cells along this track and the fitsum object that went with it.
    ///
    void computeDEDX(art::Ptr<rb::Track> &track, int PDG);

    ///
    /// \brief Return the results of the dE/dx calculation.
    ///
    /// param: dE    - empty vector to be filled with dE values
    /// param: dx    - empty vector to be filled with dx values
    /// param: W     - empty vector to be filled with track W values
    /// param: BG    - empty vector to be filled with corresponding values of beta*gamma
    /// param: dxTol - the path length through the cell must be greater than this,
    ///                to trust the computed dE/dx as good
    ///
    void getResults(std::vector<double> &dEtemp, std::vector<double> &dxtemp,
		    std::vector<double> &Wtemp, std::vector<double> &bgtemp,
		    double dxTol);

    ///
    /// \brief Return the list of rb::CellHits used in the dE/dx calculation.
    ///
    /// param: dxTol - the path length through the cell must be greater than this,
    ///                to trust the computed dE/dx as good
    art::PtrVector<rb::CellHit> getCellHits(double dxTol);

  private:
    ///
    /// \brief Clear all internally used vectors.
    ///
    void clearVectors();

    // Vectors used internally to store information.
    std::vector<geo::OfflineChan> chan;  ///< offline channel associated with a cell hit on a track
    art::PtrVector<rb::CellHit>   cells; ///< list of cell hits on the track
    std::vector<double>           dE;    ///< computed value of energy for the cell hit
    std::vector<double>           dx;    ///< computed value of the track path length through the cell hit
    std::vector<double>           W;     ///< value of track W for the cell hit
    std::vector<double>           s;     ///< total path length along the track to the middle of the cell hit
    std::vector<double>           bg;    ///< value of beta*gamma for the tracked particle at s

    art::ServiceHandle<geo::Geometry>     fGeom;
    art::ServiceHandle<calib::Calibrator> fCal;
  };
}

#endif
////////////////////////////////////////////////////////////////////////
