///
/// \file Path.h 
/// \brief Information about materials along a path in the detector
///
#ifndef BPFIT_PATH_H
#define BPFIT_PATH_H
#include <vector>
#include "TVector3.h"
class TGeoMaterial;
namespace geo { class GeometryBase; }

namespace bpfit
{
  ///
  /// \brief Track parameters (s, X0, KE, beta, ...) along a
  /// particular path in the detector
  ///
  class Path {
  public:
    ///
    /// \brief Construct a simple straight-line path from p0 to p1
    ///
    /// \param geo - pointer to the detector geometry
    /// \param p0  - x/y/z coordinates of start point
    /// \param p1  - x/y/z coordinates of end point
    /// \param s0  - value of s at location p0
    ///
    Path(const geo::GeometryBase& geo,
	 const double*            p0,
	 const double*            p1,
	 double                   s0=0);
    
    ///
    /// \brief Construct a path which will have n points
    ///
    /// \param geo - pointer to the detector geometry
    /// \param n   - number of (xyz) points along path
    ///
    Path(const geo::GeometryBase& geo, 
	 unsigned int             n,
	 double                   s0=0);

    ///
    /// \brief Construct a path using a vector of TVector3
    ///
    /// \param geo - pointer to the detector geometry
    /// \param n   - number of (xyz) points along path
    ///
    Path(const geo::GeometryBase& geo, 
	 std::vector<TVector3> const& trajectory,
	 double                   s0=0);
    
    ///
    /// \brief Add a space point to the path
    /// \param i   - point index [0...n)
    /// \parm  xyz - x/y/z location triplet
    ///
    void SetPoint(unsigned int i, const double* xyz);

    ///
    /// \brief Remove points from the trajectory that have negligible
    /// distances between them
    ///
    /// \param eps - tolerance in cm^2 for deciding if points are
    ///              close
    /// \returns The number of points removed from the lists.
    ///
    /// Scrub is guranteed to always leave the end points in the list.
    ///
    unsigned int Scrub(double eps=0.01);
    
    ///
    /// \brief Return the total path length
    ///
    double Length() const;

    ///
    /// \brief Return the number of radiation lengths crossed by the path
    /// \param s - position along the path (cm)
    ///
    double X0(double s) const;

    ///
    /// \brief Return the direction of the track after it has
    /// travelled a distance s
    ///
    /// \param s - distance (cm) along the path
    ///
    TVector3 Dir(double s=0.0);

    ///
    /// \brief Return the estimated muon parameters at path length s
    ///
    /// \param s     - position on path in cm
    /// \param T     - on return, kinetic energy in GeV
    /// \param p     - on return, momentum in GeV/c
    /// \param beta  - on return, beta=v/c=p/E
    /// \param gamma - on return, gamma=1/sqrt(1-b^2)=E/m
    ///
    void MuonParams(double s, 
		    double* T=0, 
		    double* p=0,
		    double* beta=0,
		    double* gamma=0) const;
    void PionParams(double s, 
		    double* T=0, 
		    double* p=0,
		    double* beta=0,
		    double* gamma=0) const;
    void ProtonParams(double s, 
		      double* T=0, 
		      double* p=0,
		      double* beta=0,
		      double* gamma=0) const;
  private:
    unsigned int ScrubOne(double eps);
    double IntegrateT(double ds, const char* mat, double m, double Tin) const;
    void   IntegrateLeg(const double* p1,
			const double* p2,
			double* stot,
			double* x0,
			double* Tmuon,
			double* Tpion,
			double* Tproton) const;
    void   Integrate() const;

  private:
    static void TPBG(double  t,
		     double  m,
		     double* T,
		     double* p,
		     double* beta,
		     double* gamma);
    static double Interp(double x,
			 const std::vector<double>& X,
			 const std::vector<double>& Y);
  public:
    const geo::GeometryBase& fGeo; ///< Pointer to the detector geometry
    std::vector<TVector3>    fXYZ; ///< x/y/z points along the path
    double                   fS0;  ///< s value at path start
  public:
    //
    // The following are labeled mutable as the "state" of the class is
    // encapsulated by the information above (space point
    // collections). The quantities below are calculated in a lazy way
    // when required.
    //
    mutable bool                fTablesOK; ///< Are the tables OK?
    mutable std::vector<double> fS;        ///< Accumulated path length
    mutable std::vector<double> fX0;       ///< Path length / radiation length
    mutable std::vector<double> fMuonKE;   ///< Muon kinetic energy
    mutable std::vector<double> fPionKE;   ///< Pion kinetic energy
    mutable std::vector<double> fProtonKE; ///< Proton kinetic energy
  };
}

#endif
////////////////////////////////////////////////////////////////////////
