///////////////////////////////////////////////////////////////////////////
/// \brief   Histograms used by attenuation calibration
/// \author  Christopher Backhouse - bckhouse@caltech.edu
/// \date    Apr 2013
///////////////////////////////////////////////////////////////////////////
#ifndef CALDP_ATTENPROFILES_H
#define CALDP_ATTENPROFILES_H

#include "TH2.h"

#include "GeometryObjects/OfflineChan.h"

#include <cassert>
#include <cmath>
#include <map>

namespace caldp
{
  /// Lightweight TProfile equivalent
  class LiteProfile
  {
  public:

    // No-one in Calibration wants default constructors, they just lead to
    // things accidentally getting the wrong binning, much better to be
    // explicit. But generating reflex dictionaries relies on them, so control
    // them with this #define
#ifdef ATTENPROFILES_DEFAULT_CONSTRUCTORS
    LiteProfile() 
      : fMinW(-1)
      , fMaxW(-1)
      {}
#endif

#ifndef __GCCXML__
    LiteProfile(float w0, float w1);

    void Fill(float x, float y);
    LiteProfile& operator+=(const LiteProfile& rhs);
    TH1F* ToTH1F() const;
    int TotalHits() const;
    int TotalEntries();

#endif
  protected:
    float fMinW, fMaxW;

    static const size_t kNumBins = 100;

    std::vector<int>   fCount;
    std::vector<float> fSum;
    std::vector<float> fSqSum;
  };

  /// Profiles used by attenuation calibration
  class AttenProfiles
  {
  public:
#ifdef ATTENPROFILES_DEFAULT_CONSTRUCTORS
    AttenProfiles() {}
#endif

#ifndef __GCCXML__
    AttenProfiles(float minW, float maxW);

    AttenProfiles& operator+=(const AttenProfiles& rhs);
#endif

    LiteProfile WPE, WPE_corr, WPE_corr_xy, WPE_corr_z, WPE_corr_xy_truth;
  };

  /// AttenProfiles for many channels
  class AttenProfilesMap
  {
  public:

#ifdef ATTENPROFILES_DEFAULT_CONSTRUCTORS
    AttenProfilesMap() : fMinW(-1), fMaxW(-1) {}
#endif

#ifndef __GCCXML__
    AttenProfilesMap(float w0, float w1);

    AttenProfilesMap(const AttenProfiles&) = delete; // uncopyable

    bool HasProfiles(const geo::OfflineChan& c);
    bool HasProfiles(int plane, int cell);

    AttenProfiles& GetProfiles(const geo::OfflineChan& c);
    AttenProfiles& GetProfiles(int plane, int cell);

    /// List of channels with profiles on
    std::vector<geo::OfflineChan> Channels() const;

    /// Intended for writing out/drawing all the histograms
    std::map<int, std::map<int, std::vector<TH1F*>>> GetAllProfilesByPlaneAndCell();

    AttenProfilesMap& operator+=(const AttenProfilesMap& rhs);

    void Clear();

    double MinW() const {return fMinW;}
    double MaxW() const {return fMaxW;}
#endif
  protected:
    float fMinW, fMaxW;

    std::map<geo::OfflineChan, AttenProfiles> fMap;
  };

}

#endif
