/// 
/// \brief   Interface to the run-by-run list of bad channels
/// \author  betancourt@physics.umn.edu, jpaley@anl.gov
/// \date
///
#ifndef CHANINFO_BADCHANLIST_H
#define CHANINFO_BADCHANLIST_H

#include <vector>
#include <map>
#include <string>
#include <ctime>

// NOvA includes
#include "RawData/RawDigit.h"
#include "GeometryObjects/OfflineChan.h"
#include "NovaDAQConventions/DAQConventions.h"

//Framework includes
#include "art/Framework/Services/Registry/ActivityRegistry.h"
#include "art/Framework/Services/Registry/ServiceMacros.h"
#include "art/Persistency/Common/PtrVector.h"
#include "art/Persistency/Common/Ptr.h"
//#include "art/Persistency/Provenance/Timestamp.h"
#include "fhiclcpp/ParameterSet.h"
#include "CLHEP/Random/RandFlat.h"

namespace nova{namespace dbi{class Table;}}

namespace chaninfo {

  ///
  /// Interface to the run-by-run list of bad channels
  ///
  class BadChanList {
  public:
    // Get a BadChanList instance here
    BadChanList(fhicl::ParameterSet const& pset, art::ActivityRegistry &reg);
    ~BadChanList();

    void postBeginRun(art::Run const& run);
    void postBeginSubRun(art::SubRun const& subrun);
    void preBeginEvent ( art::Event const & evt);
    
    int  Apply(art::PtrVector<rawdata::RawDigit>& rd);
    int  Apply(std::vector< art::Ptr<rawdata::RawDigit> >& rd);
    int  Apply(std::vector< const rawdata::RawDigit*>&);

    bool IsBad(int plane, int cell);
    bool IsLowEfficiency(int plane, int cell);
    void reconfigure(fhicl::ParameterSet p);

    void LoadNewData();

    int NBadInSubRun(int i);

  private:
    void SetRandomBadChannels();

    void LoadCosmicEff(int newrun, int newsubrun);

    // private data members for BadChanList class
    int                                  fApplyBadChan;           ///< Option to activate the bad channels   
    int                                  fMCRunNumber;            ///< Run number for the MC
    int                                  fMCSubrunNumber;         ///< Subrun number for the MC
    int                                  fStateIdx;  ///< index of "state" column in chan_rate_state dB table
    int                                  fFEBStateIdx;  ///< index of "state" column in febdropout dB table

    int                                  fNBadSR;

    int                                  fVldRunGrouping;

    bool                                 fAbortIfNoneFound; ///< Option to throw an exception if no channel rate state rows are returned from dB.

    bool                                 fNeedNewData;

    bool                                 fIsPerfect;

    bool                                 fAbortOnSmallRunNum;

    std::map<geo::OfflineChan, bool >    fLowEfficiency;

    std::map<geo::OfflineChan, bool >    fRandomBCMap;   ///< random bad channel map

    std::string                          fBadChanCSVFile;

    bool                                 fEventCheck;    ///< do we check each event for dropped DCMs?

    std::string                          fRawDigitLabel;  ///< what is the name of the raw digits ART object

    bool fNewCosmicEffTable; ///< Use new-style table?

    float fRandomBadChannelFraction; ///< If >0, ignore DB and set this fraction bad at random

    std::unique_ptr<nova::dbi::Table> fBCTable;
    std::unique_ptr<nova::dbi::Table> fFEBDOTable;

    int fRun;
    int fSubRun;

    time_t fMinVldTime;                 ///<  Minimum time used in database request

    time_t fMaxVldTime;                 ///<  Maximum time used in database request



    std::string fValidityTag;
    std::string fFEBDropoutTag;

    bool fApplyFEBDropout;

    std::auto_ptr<CLHEP::RandFlat> fFlatRand;

    int fDCMHits[12][14]; ///< count hits in each DCM each event to look for dropped DCMs
  };
}
DECLARE_ART_SERVICE(chaninfo::BadChanList, LEGACY)
#endif
////////////////////////////////////////////////////////////////////////
