#ifndef DAQ2RAWDIGIT_NOVASOCKETINPUTDRIVER_H
#define DAQ2RAWDIGIT_NOVASOCKETINPUTDRIVER_H

////////////////////////////////////////////////////////////////////////
/// \file  NOvASocketInputDriver.h
/// \brief Source to read raw Events from a socket, per specifications
///        in NOvA-DocDB-3947 and NOvA-DocDB-4390.
///
/// \version $Id: NOvASocketInputDriver.h,v 1.5 2012-11-18 18:52:46 gsdavies Exp $
/// \author  paterno@fnal.gov
////////////////////////////////////////////////////////////////////////
#include "DAQ2RawDigit/NOvAInputDriverBase.h"

#include "art/Framework/IO/Sources/SourceHelper.h"
#include "art/Persistency/Provenance/SubRunID.h"
#include "art/Framework/IO/Sources/SourceTraits.h"

#include "boost/asio.hpp" // socket support

#include <string>
#include <vector>

// forward declarations
namespace fhicl{
    class ParameterSet;
}
namespace art{
    class ProductRegistryHelper;
    class FileBlock;
    class EventPrincipal;
    class RunPrincipal;
    class SubRunPrincipal;
}


namespace daq2raw {
  class NOvASocketInputDriver;
}

namespace art {
  // Specialize this trait from SourceTraits.h
  template <>
    struct Source_wantFileServices<daq2raw::NOvASocketInputDriver> {
      static constexpr bool value = false;
    };
}


class daq2raw::NOvASocketInputDriver : public daq2raw::NOvAInputDriverBase{
  /// Class to fill the constraints on a template argument to the class,
  /// art::Source
public:
  // Required constructor
  NOvASocketInputDriver(fhicl::ParameterSet const &pset,
                        art::ProductRegistryHelper &help,
                        art::SourceHelper const &pm);

  // Required by art::Source:
  void closeCurrentFile();
  void readFile(std::string const &name,
                art::FileBlock* &fb);
  bool readNext(art::RunPrincipal* const &inR,
                art::SubRunPrincipal* const &inSR,
                art::RunPrincipal* &outR,
                art::SubRunPrincipal* &outSR,
                art::EventPrincipal* &outE);

private:
  typedef std::vector<uint32_t> eb_t;      ///< Event buffer type.
  typedef eb_t::value_type ebe_t;          ///< Event buffer element type.
  typedef std::vector<uint8_t> cb_t; ///< Command buffer type.

  static size_t in_bytes(size_t nElements) {
    return nElements * sizeof(ebe_t);
  }

  static size_t in_elements(size_t nBytes) {
    size_t result = nBytes / sizeof(ebe_t);
    return (in_bytes(result) == nBytes)?
      result:
      result+1;
  }

  art::SourceHelper          fSourceHelper;           ///< Class to help with {Run, SubRun, Event}Principal construction.
  std::string                  fCurrentFilename;          ///< Hang on to the current filename.
  art::SubRunID                fCurrentSubRunID;          ///< Keep track of the current subRun details.
  boost::asio::io_service      fIoService;                ///< Object representing IO services of the OS.
  boost::asio::ip::tcp::socket fSocket;                   ///< Object representing the socket from which we read.
  size_t                       fCurrentEventCapacity;     ///< Current event capacity in buffer elements (current buffer size is 1 more than this);
  size_t                       fMaxEventSizeBytes;        ///< Will not try to read an event bigger than this.
  std::vector<uint32_t>        fSocketBuffer;             ///< Raw data buffer for socket communication.

  bool                         fFilterCorruptedNanoslices;///< Filter the corrupted nanaslices in the unpacking?
  bool                         fDoFillFlatDAQData;        ///< Do we need to fill FlatDAQData object?

  int                          fMaxReadAttempts;               ///< Maximum number of read attempts before we throw

  int                          fCorruptEventCount;             ///< Counter for the total number of corrupt events encountered
  int                          fConsecutiveCorruptEventCount;  ///< Counter for the number of consecutive corrupt events encountered
  int                          fBadReadCount;                  ///< Counter for the total number of corrupt events encountered
  int                          fConsecutiveBadReadCount;       ///< Counter for the number of consecutive corrupt events encountered

  uint32_t trigmask_word1; ///< Lowest word in the trigger mask only the lower 24 bit are used
  uint32_t trigmask_word2; ///< Middle word (bits 31-63) in the trigger mask
  uint32_t trigmask_word3; ///< High word (bits 64-96) in the trigger mask

  int                          fTransmitDelay;           ///< Delay in microseconds to wait between command transmission

};  // NOvASocketInputDriver

#endif /* DAQ2RAWDIGIT_NOVASOCKETINPUTDRIVER_H */

