///////////////////////////////////////////////////////////////
// \file    OnlineUnpack.h
// \author  denis@fnal.gov
//
// \brief   This class does the unpack from the flat (binary) 
//          DAQ data to the online format
// \version $Id: OnlineUnpack.h,v 1.22 2012-11-09 17:22:59 jpaley Exp $
////////////////////////////////////////////////////////////////
#ifndef DAQ2RAW_ONLINEUNPACK_H
#define	DAQ2RAW_ONLINEUNPACK_H
#include "DAQChannelMap/DAQChannelMapConstants.h"
#include "DAQQualityCheck/QualityCheck.h"
#include "RawData/RawDigit.h"

#include <stdint.h> // int64_t

/// Do we need to subtract event time from the TDC?
#define SUBTRACT_EVENT_TIME_FROM_HIT

/// Define this only if you want to use checks of the DAQDataFormat
//#define RUNCHECKS

// forward dectarations
class TH1F;

namespace rawfileparser {
    class RawFileParser;
}

namespace rawdata{
    class DAQHeader;
    class RawTrigger;
    class FlatDAQData;
}

namespace daq2raw {

  typedef struct {
    // See DocDB 4390
    // using unsigned long long rather than uint64_t to ensure consistent behavior on OSX and LINUX
    uint32_t             fTriggerHeader_Source;             ///! Trigger source (see DAQDataFormats/TriggerDefines.h)   
    uint32_t   		 fTriggerHeader_SourceID;           ///! Identifier of the source system		     
    uint32_t   		 fTriggerHeader_SourceSubID;        ///! Identifier of the subsystem or channel on source device
    unsigned long long   fTriggerHeader_TriggerNumber;      ///! Universal trigger number
    unsigned long long   fTriggerHeader_MasterTriggerNumber;
    unsigned long long   fTriggerTime_GenTime;
    uint32_t             fTriggerTime_ClockSource;
    unsigned long long   fTriggerTimingMarker_TimeStart;
    unsigned long long   fTriggerTimingMarker_ExtractionStart;
    uint32_t             fTriggerRange_InfBit;	       
    uint32_t   		 fTriggerRange_TriggerLength;    
    uint32_t   		 fTriggerMask_MCBit;	       
    uint32_t   		 fTriggerMask_TMod;	       
    uint32_t   		 fTriggerMask_TriggerTypesLow;   
    uint32_t   		 fTriggerMask_TriggerTypesHigh;  
    uint32_t   		 fTriggerMask_TriggerTypesHigh2; 
    uint32_t   		 fTriggerMask_Prescale;          

    int                  fDetectorType;   ///! Detector type	
    int        		 fNumberOfDigits; ///! Number of RawDigits

  } UnpackedEvent;


  class OnlineUnpack{

  public:
    /// Constructor
    OnlineUnpack (TH1F *pulseHeightVals, TH1F *timeStampVals, TH1F *triggerVals, const std::string* file_name=nullptr);
    /// Destructor
    ~OnlineUnpack ();

    /// Unpack the RawEvent
    bool ProcessRawEvent (void* rawevent_buffer, std::vector<rawdata::RawDigit>*, rawdata::RawTrigger*, rawdata::FlatDAQData*, rawdata::DAQHeader*);
    bool ProcessRawEvent (rawfileparser::RawFileParser &parser, std::vector<rawdata::RawDigit>*, rawdata::RawTrigger*, rawdata::FlatDAQData*, rawdata::DAQHeader*);

    uint32_t                                    getRunNumber()                   const { return fRawEvent.getRunNumber();     }
    uint32_t 			 		getSubrunNumber()                const { return fRawEvent.getSubrun();        }
    unsigned long long 			 	getEventNumber()                 const { return fRawEvent.getEventNumber();   }
    int                                         getDetId()                       const { return fDetId;                       }
    uint32_t                                    getNumberOfNonemptyMicroblocks() const { return fNumberOfNonemptyMicroblocks; }


  private:
    // Common code to unpack the event
    bool ProcessRawEvent_(std::vector<rawdata::RawDigit>*, rawdata::RawTrigger*, rawdata::FlatDAQData*, rawdata::DAQHeader*);

    /// Try to get Detector ID from the file name
    bool getDetIDFromFileName(const std::string*);



  public:
    bool                     fFilterCorruptedNanoslices;  ///< Do we need to filter out Corrupted NanoSlices?

  private:

    int                      fDetId;                      ///< Detector ID reading from Microblocks
    uint32_t                 fNumberOfNonemptyMicroblocks;

    daqdataformats::RawEvent fRawEvent;                   ///< Flat DAQ RawEvent

    rawdata::RawDigit        fDigit;                      ///< Instance of RawDigit class

    UnpackedEvent            fUnpackedEventParameters;

    // Unpacking Histograms
    TH1F*                    fPulseHeight_values;
    TH1F*                    fTimeStamp_values;
    TH1F*                    fTrigger_values;
    TH1F*                    fTDC_values;
    uint32_t                 fHitNumber;

    const std::string*       fCurrentFileName;            ///< Current file name

#ifdef RUNCHECKS
    daqqc::QualityCheck      qc;                          ///< Runs checks on raw event
#endif
  };
 
}// end of namespace rawp

#endif	/* DAQ2RAW_ONLINEUNPACK_H */

