#ifndef GEO_CELLGEO_H
#define GEO_CELLGEO_H

#include "GeometryObjects/CellUniqueId.h"
#include "GeometryObjects/GeometryBase.h"
#include "TGeoMatrix.h"

#include <vector>
class TGeoNode;
class TGeoHMatrix;
class TVector3;

namespace geo {
  /// \brief Encapsulate the cell geometry
  //
  /// A note on the cell geometry: Cells are constructed such that, in
  /// their local frame, their profile occupies the x-y plane with
  /// their long dimension running along the z-axis. The dimension
  /// transverse to the beam runs along the local x and is called the
  /// width. The dimension parallel to the beam runs along the local y
  /// and is called the depth. The length of the cell runs along the
  /// local z-axis with z increasing as one moves toward the readout
  /// end of the cell.
  //
  class CellGeo {
  public:
    CellGeo(std::vector<const TGeoNode*>& n, int depth);
    ~CellGeo();

    void   GetCenter(double* xyz, double localz=0.0) const;
    void   GetCenter(TVector3& xyz, double localz=0.0) const;
    double DistToReadOut(double localz) const;
    double HalfW() const;
    double HalfL() const;
    double HalfD() const;

    void LocalToWorld(const double* local, double* world)     const;
    void LocalToWorldVect(const double* local, double* world) const;
    void WorldToLocal(const double* local, double* world)     const;
    void WorldToLocalVect(const double* local, double* world) const;


    /// Translate plane and all cells in plane by dx, dy and dz
    /// New transform will be stored in a separate TGeoHMatrix
    void TranslateCell(double dx, double dy = 0, double dz = 0);
                
    /// Translate plane using a previously defined TGeoCombiTrans
    /// New transform will be stored in a separate TGeoHMatrix
    void TranslateCell(TGeoCombiTrans* comb);
                            
    /// Restore geometry to original
    void RestoreOriginal();

    const TGeoNode*     Node() const { return fCellNode; }
    const CellUniqueId& Id()   const { return fId; }

  private:
    CellUniqueId    fId;                ///< Node ID number
    const TGeoNode* fCellNode;          ///< Pointer to the cell node
    TGeoHMatrix*    fGeoMatrix;         ///< Transformation matrix to world frame
    TGeoHMatrix*    fGeoMatrixOriginal; ///< Original transformation matrix to world frame
    Bool_t          fIsOriginal;        ///< Is fGeoMatrix the original transform?
  };
}


#endif
