/*
 * File:   CoordinateTransformation.h
 * Author: denis
 *
 * This class handles transformation from the beam coordinates
 * to the NOvA detector coordinates and back.
 * Also has some supplementary functions.
 *
 * Currently support only NEAR, FAR detectors AND NuMI beamline
 *
 * Created on April 10, 2011, 10:21 PM
 */

#ifndef _COORDIATETRANSFORMATION_H
#define	_COORDIATETRANSFORMATION_H
#include <CLHEP/Units/SystemOfUnits.h> // to define m, cm, degrees, radians, GeV, etc.

#include "DAQChannelMap/DAQChannelMapConstants.h" // to get the Detector ID enum

// Standard boost linear algebra library - matrix and vector.
#include <boost/numeric/ublas/matrix.hpp>
#include <boost/numeric/ublas/vector.hpp>

namespace geo {

    /// Z position of the point S (mean hadron decay) in beam coordinates.
    /// Taken from NOvA-doc-5485.
    const double pointS_Z = 184.25 * CLHEP::m;

    /// Angle between NuMI beamline and the earth's surface.
    /// Equals to 3.3 degrees
    const double NuMiBeamLineAngle    = 0.057715289 * CLHEP::radian;

    /// Angle between NuMi beamline and the NOvA near detectors
    const double NOvAneardetNuMiAngle = 0.0145766 * CLHEP::radian; // From NOvA-doc-5485

    /// Beam type. BOONE or NuMI
    enum BeamType {
        BEAMTYPE_NUMI  = 1,
        BEAMTYPE_BOONE = 2
    };


    class CoordinateTransformation {

        public:
            CoordinateTransformation();
            CoordinateTransformation(uint32_t detector_id, uint32_t beam_type);

            ~CoordinateTransformation();


            /// Try to set the detector and beam
            bool setDetectorAndBeam(uint32_t detector_id, uint32_t beam_type);

            /// Transformation of Beam coordinates to NOvA detector coordinates
            /// detector_id is as defined in DAQChannelMap
            /// IMPORTANT! Input coordinates must be CLHEP-dimensional.
            /// For example if x is in cm, make it x = x * CLHEP::cm
            /// IMPORTANT! Output coordinates are also CLHEP-dimensional.
            /// Fox example to get x in meters, do length_x = output_detector_coords[0] / CLHEP::m
            bool transformBeamToDetectorCoordinates(double* input_beam_coords,
                                                    double* output_detector_coords) const;

            /// Transformation of NOvA Detector coordinates to Beam coordinates
            /// detector_id is as defined in DAQChannelMap
            /// IMPORTANT! Input coordinates must be CLHEP-dimensional.
            /// For example if x is in cm, make it x = x * CLHEP::cm
            /// IMPORTANT! Output coordinates are also CLHEP-dimensional.
            /// Fox example to get x in meters, do length_x = output_detector_coords[0] / CLHEP::m
            bool transformDetectorToBeamCoordinates(double* input_detector_coords,
                                                    double* output_beam_coords) const;


            /// Print Matrices
            bool print();

        private:
            /// Initializing matrices
            bool init();

            /// Setting the coordinate parameters
            /// If returns tru, everything is set fine
            /// If false, then some configuration was not supported and the parameters were not set
            bool setCoordinateParameters(uint32_t detector_id, uint32_t beam_type);

        private:
            uint32_t _detectorID; ///< Current Detector ID
            uint32_t _beamType;   ///< Current Beam Type

            boost::numeric::ublas::matrix <double> _rotationMatrix;        ///< Rotational matrix
            boost::numeric::ublas::matrix <double> _rotationMatrixInverse; ///< Rotational matrix inverse

            /// coordinates of the detector zero in the beam coordinates
            boost::numeric::ublas::vector <double> _simpleTranslation;



            double _alpha; ///<Angle between NuMI beamline and the earth's surface.
            double _beta;  ///<Angle between NuMi beamline and the detector looking from top


    };// end of class CoordinateTransformation

}// end of namespace geo

#endif	/* _COORDINATETRANSFORMATION_H */

