/////////////////////////////////////////////////////////////////////////
/// \brief   Perform a "2 point" Hough transform on a collection of hits
/// \authors messier@indiana.edu, mibaird@indiana.edu
/// \date
/////////////////////////////////////////////////////////////////////////
#ifndef MULTIHOUGH_MULTIHOUGH2P_H
#define MULTIHOUGH_MULTIHOUGH2P_H
#include <cmath>
#include <vector>

#include "RecoBase/HoughResult.h"

namespace rb { class HitList; }
class TH2F;

namespace hough {

  class LiteTH2;
  
  /// Hough transform based on 2-point combinations
  class MultiHough2P {
  public:
    ///
    /// \brief Construct the 2-point Hough transform
    ///
    /// The transform proceeds by considering all paris of points that
    /// occur within a cut off distance (rsqr) of each other. For each
    /// pair an ellipse is filled in Hough space weighted by the
    /// quality of the line fit to the two spatial points. The
    /// algorithm is largely derived from
    ///
    /// "Real-time line detection through an improved Hough transform
    /// voting scheme", L.A.F. Fernandes, M.M. Oliveira / Pattern
    /// Recognition 41 (2008) 299-314
    ///
    /// \param v       - Which detector view? geo::kX or geo::kY
    /// \param rhosz   - Size of rho bins (cm)
    /// \param thetasz - Size of theta bins (degrees)
    /// \param rhosm   - Number of bins to use in smoothing rho
    /// \param thetasm - Number of bins to use in smoothing theta
    /// \param rsqr    - Cut on distance between two points
    /// \param pco     - Number of sigma above average peak height for threshold (peak cut-off)
    /// \param loops   - Max number of multi-Hough loops
    ///
    MultiHough2P(geo::View_t v, 
		 double      rhosz,
		 double      thetasz,
		 int         rhosm,
		 int         thetasm,
		 double      rsqr,
		 double      pco,
		 int         loops,
		 double      rmdist);
    ~MultiHough2P();

    TH2F* MapToTH2F() const;
    
    /// Make the Hough map for the hit list h.
    /// \param h - The list of hits (can be mixed X and Y view hits)
    
    void Map(const rb::HitList& h);
    
    // make the Hough map through multiple steps
    void MultiMap(rb::HitList h);
    
  private:
    void BuildMap(double xlo, double xhi,
		  double ylo, double yhi);
    void RhoTheta(double x1,        double y1,
		  double x2,        double y2,
		  double* rho,      double* theta,
		  double* sigmarho, double* sigmatheta);
    void SmoothMap();
    void RefinePeak(double &rho, double &theta, int xp, int yp);
    void ReweightHits(double rho, double theta, rb::HitList &h);
  public:
    // Configuration data
    geo::View_t fView;    ///< Which detector view?
    double      fRhoSz;   ///< Size of rho bins (cm)
    double      fThetaSz; ///< Size of theta bins (degrees)
    int         fXwinSz;  ///< Smoothing size (bins) in x
    int         fYwinSz;  ///< Smoothing size (bins) in y
    double      fSigma;   ///< Assumed location error on points (cm)
    double      fRsqr;    ///< Distance^2 cut on points (cm^2)
    double      fPco;     ///< Number of sigma abouve average peak height to use as cutoff
    int         fLoops;   ///< Max number of Multi-Hough loops
    double      fRmDst;   ///< Distance cut for removing points from a Hough line
  public:
    // Data used in / resulting from calculations
    LiteTH2*          fHspaceW; ///< Hough transform
    rb::HoughResult   fH;       ///< Results of Hough transformation;
  };
}

#endif
////////////////////////////////////////////////////////////////////////
