/* 
 * File:   DAQSimulator.h
 * Author: denis
 *
 * This class simulates a Data Concentrator Module(DCM) in NOvA DAQ
 *
 * Created on January 31, 2012, 4:56 PM
 */

#ifndef _MOCKDAATADAQ_DAQSIMULATOR_H
#define	_MOCKDAATADAQ_DAQSIMULATOR_H

#include "DAQChannelMap/DAQChannelMapConstants.h"
#include <vector>
#include <stdint.h>

// forward declarations
namespace mcdatadaq{
	class DCMSimulator;
        class BufferNodeSimulator;
        class GlobalTriggerSimulator;
        class DataLoggerSimulator;
}

namespace daqdataformats{
    class RawMilliSlice;
    class RawTrigger;
    class RawDataBlock;
}

namespace rawdata{
    class RawDigit;
}


namespace mcdatadaq {

    /// Version of the DAQ simulation
    static const uint32_t DAQSIMULATION_VERSION = 33;

    /// Order of the nanoslices. i-th nanoslice element represents the number in RawDigits vector
    typedef std::vector<uint32_t> nanoslice_order_t;

    class DAQSimulator {
        public:
            DAQSimulator(); ///< Constructor
            ~DAQSimulator();///< Destructor

            /// Initialize DAQ Simulator. Need to know the Geometry at this point
            bool init();

            /// Take RawDigits for the event and simulate them in the DAQ.
            /// Writing results into output binary files
            bool processEvent(const std::vector<rawdata::RawDigit>&);

            /// Produce a nanoslice order from an input vector of RawDigits
            bool getNanoSliceOrder(const std::vector<rawdata::RawDigit>&, nanoslice_order_t&) const;

            /// Some print method to see what order was assigned
            bool print(const std::vector<rawdata::RawDigit>& digits, const nanoslice_order_t& digits_order) const;

            /// Print the introductory version of the DAQ Simulator
            bool print() const;

        private:
            /// Get an instance of DCMSilulator for a given diblock and a local DCM id for that diblock
            DCMSimulator* getDCMSimulator(const daqchannelmap::diblock_t, const daqchannelmap::dcm_id_t) const;

            /// Produce MicroSlices in all DCMs with input RawDigits
            bool fillMicroSlices  (const std::vector<rawdata::RawDigit>&);
            /// Given all MicroSlices on DCMs which were already produced, now produce MilliSlices on each DCM
            bool fillMilliSlices(std::vector<daqdataformats::RawMilliSlice*>&);
            /// Produce Trigger. Also, given all MilliSlices on DCMs which were already produced, now produce DataBlocks in BufferNodes
            bool fillDataBlocks(const std::vector<daqdataformats::RawMilliSlice*>&, daqdataformats::RawTrigger*, std::vector<daqdataformats::RawDataBlock*>&);
            /// Given all DataBlocks in BufferNodes and Trigger in GlobalTrigger which were already produced, 
            /// now produce Events and add to Run in DataLogger
            bool fillEventAndRun(const std::vector<daqdataformats::RawDataBlock*>&, daqdataformats::RawTrigger*);

            /// Prepare DAQ for the next event
            bool nextEvent();

        private:

            std::vector<std::vector<DCMSimulator*> > fDCMs;                ///< All DCMs in the detector
            std::vector<BufferNodeSimulator*>        fBufferNodes;         ///< All Buffer Nodes
            GlobalTriggerSimulator*                  fGlobalTrigger;       ///< Global Trigger
            DataLoggerSimulator*                     fDataLoggerSimulator; ///< DataLogger

    };// end of class DAQSimulator
}// end of namespace mcdatadaq

#endif	/* _MOCKDAATADAQ_DAQSIMULATOR_H */

