////////////////////////////////////////////////////////////////////////
// \file    Cluster.h
// \brief   Cluster of CellHits. Base of other reconstruction classes.
// \version $Id: Cluster.h,v 1.22 2012-10-26 22:28:40 bckhouse Exp $
// \author  Author: Christopher Backhouse - bckhouse@caltech.edu
// \date    $Date: 2012-10-26 22:28:40 $
////////////////////////////////////////////////////////////////////////
#ifndef RBCLUSTER_H
#define RBCLUSTER_H

#include <iosfwd>
#include <vector>

#include "art/Persistency/Common/PtrVector.h"

#include "GeometryObjects/OfflineChan.h"
#include "GeometryObjects/PlaneGeo.h"

#include "TVector3.h"

namespace rb
{
  class CellHit;
  class RecoHit;
  class WeightedHit;

  enum AveragingScheme
  {
    kByClusterWeightOnly,
    kByEnergy ///< In addition to the intrinsic weights
  };
  const AveragingScheme kDefaultScheme = kByEnergy;

  /// \brief A collection of associated CellHits
  ///
  /// A Cluster is simply a "bag" of CellHits which have been reconstructed to
  /// be associated with each other in some way.
  ///
  /// A cluster may be created as either 2D or 3D (ie containing hits from one or
  /// both views).
  ///
  /// This class also contains various utility functions for obtaining the bulk
  /// properties of these hits.
  ///
  /// It forms the basis of the reconstruction inheritance hierachy, with the
  /// derived classes adding additional information about the hits.
  class Cluster
  {
  public:
    /// \name 3D cluster constructors
    //@{
    Cluster();
#ifndef __GCCXML__
    explicit Cluster(const art::PtrVector<rb::CellHit>& cells, int id=0);
    //@}
    /// \name 2D cluster constructors
    //@{
    Cluster(geo::View_t view, int id=0);
    Cluster(geo::View_t view, const art::PtrVector<rb::CellHit>& cells, int id=0);
    Cluster(geo::View_t view, const std::vector<art::Ptr<rb::CellHit> >& cells, int id=0);
#endif // __GCCXML__
    //@}

    virtual ~Cluster();

#ifndef __GCCXML__

    // Note: the redundant 'rb::' is not required by C++ but by
    // CINT. Best to just keep it until CINT catches up.
    virtual void Add(const art::Ptr<rb::CellHit>& cell, double weight = 1);
    virtual void Add(const art::PtrVector<rb::CellHit>& cells,
                     const std::vector<double>& weights = std::vector<double>());

    void      SetID(int id)       { fID = id;   }
    const int ID()          const { return fID; }

    /// \brief Estimate the unmeasured coordinate of \a chit
    ///
    /// The nomenclature for the coordinates is that XYZ are the physical
    /// coordinates. Whichever of XY is measured by the cell is "V", the
    /// direction along the cell is "W". The origin of VW space is the same as
    /// the origin of XY space, the detector centreline.
    ///
    /// This method should be overridden in all descendants of \ref Cluster
    /// which have additional information to make this determination.
    ///
    /// For \ref Cluster, this is simply the average coordinate of all hits in
    /// the view in question.
    ///
    /// \param  chit Any \ref CellHit, not necessarily owned by this container
    /// \return      Estimated unmeasured coordinate of this hit (cm) assuming
    ///              it to be associated with this reconstructed object.
    virtual double W(const rb::CellHit* chit) const;

    bool Is3D() const {return fView == geo::kXorY;}
    bool Is2D() const {return !Is3D();}

    /// kXorY for 3D clusters.
    virtual geo::View_t View() const {return fView;}

    /// \name Cell counts
    //@{
    /// Number of cells in view \a view
    unsigned int NCell(geo::View_t view) const;
    /// Number of cells in the x-view
    unsigned int NXCell() const {return fXCell.size();}
    /// Number of cells in the y-view
    unsigned int NYCell() const {return fYCell.size();}
    /// Number of cells in either view
    unsigned int NCell() const {return NXCell()+NYCell();}
    //@}

    /// \name Cell Accessors
    //@{
    /// Get the ith cell from view \a view
    art::Ptr<rb::CellHit>  Cell(geo::View_t view, unsigned int viewIdx) const;
    /// Get the ith cell in the x-view
    art::Ptr<rb::CellHit> XCell(unsigned int xIdx) const;
    /// Get the ith cell in the y-view
    art::Ptr<rb::CellHit> YCell(unsigned int yIdx) const;
    /// Get the ith cell from either view
    art::Ptr<rb::CellHit>  Cell(unsigned int globalIdx) const;
    /// Get all cells from the x-view
    const art::PtrVector<rb::CellHit>& XCells() const { return fXCell; }
    /// Get all cells from the x-view
    const art::PtrVector<rb::CellHit>& YCells() const { return fYCell; }
    /// Get all cells from both views
    art::PtrVector<rb::CellHit> AllCells() const;
    /// Positions of all the CellHits
    std::vector<geo::OfflineChan> OfflineChans() const;
    /// Get all hits from both views, with weights attached
    std::vector<rb::WeightedHit> WeightedHits() const;
    /// Weight assigned to the cell
    double Weight(unsigned int globalIdx) const;
    /// Weight assigned to the cell
    double Weight(geo::View_t view, unsigned int viewIdx) const;
    //@}

    /// Create a cluster from this one, but with the hits of \a excl removed
    rb::Cluster Exclude(const rb::Cluster* excl) const;

    /// \brief Return calibrated hit based on assumed W coordinate
    ///
    /// To calibrate a hit you must know its distance from the readout. This is
    /// a convenience function which uses the \ref W function to estimate
    /// the position of a \ref CellHit and then uses \ref Calibrator to
    /// produce a calibrated hit from that information.
    rb::RecoHit RecoHit(const art::Ptr<rb::CellHit>& chit) const;
    rb::RecoHit RecoHit(geo::View_t view, unsigned int viewIdx) const;
    rb::RecoHit RecoHit(unsigned int globalIdx) const;

    /// \brief Forget about all owned cell hits
    ///
    /// Virtual function because derived classes may have more stuff to clear
    virtual void Clear();
    /// Remove \a hit from current cluster
    void RemoveHit(const art::Ptr<rb::CellHit> hit);
    /// Remove the ith hit from current cluster
    void RemoveHit(unsigned int globalIdx);
    /// Declare the cluster to consist of noise hits or not
    void SetNoise(bool noise) {fNoiseCluster = noise;}
    /// Is the noise flag set?
    bool IsNoise() const {return fNoiseCluster;}

    /// Set weight of the cell at this index
    void SetWeight(unsigned int globalIdx, double weight);
    /// Set weight of the cell at this index in this view
    void SetWeight(geo::View_t view, unsigned int viewIdx, double weight);

    /// \name Totals
    //@{
    /// Sum of the ADC of all the contained hits
    double TotalADC() const;
    /// Sum of the PE value of all the contained hits
    double TotalPE() const;
    /// \brief Simple sum of the estimated GeV of all the hits
    ///
    /// WARNING: This is not a sensible energy estimator by itself. If you just
    /// want a simple calorimetric estimate look at \ref
    /// CalorimetricEnergy. Otherwise, a range of more sophisticated/elaborate
    /// estimators are available elsewhere.
    double TotalGeV() const;
    /// Sum of all the weights. The effective number of hits
    double TotalWeight() const;
    /// \brief Simple estimate of neutrino energy
    ///
    /// Uses a simple scale factor to correct for dead material and thresholds,
    /// could be more advanced, but this is simple enough to use for
    /// preselection etc.
    double CalorimetricEnergy() const;
    //@}

    /// \name Minima
    /// The "lower" corner of a box containing all the hits
    //@{
    TVector3 MinXYZ() const;
    double MinV(geo::View_t view) const;
    double MinX() const {return MinXYZ().X();}
    double MinY() const {return MinXYZ().Y();}
    double MinZ() const {return MinXYZ().Z();}
    double MinTNS() const;
    unsigned int MinPlane(geo::View_t view = geo::kXorY) const;
    unsigned int MinCell(geo::View_t view) const;
    //@}

    /// \name Maxima
    /// The "upper" corner of a box containing all the hits
    //@{
    TVector3 MaxXYZ() const;
    double MaxV(geo::View_t view) const;
    double MaxX() const {return MaxXYZ().X();}
    double MaxY() const {return MaxXYZ().Y();}
    double MaxZ() const {return MaxXYZ().Z();}
    double MaxTNS() const;
    unsigned int MaxPlane(geo::View_t view = geo::kXorY) const;
    unsigned int MaxCell(geo::View_t view) const;
    //@}

    /// \name Means
    /// The unweighted mean positrion and time of all the hits in the cluster
    //@{
    TVector3 MeanXYZ(rb::AveragingScheme = kDefaultScheme) const;
    double MeanV(geo::View_t view,
                 rb::AveragingScheme scheme = kDefaultScheme) const;
    double MeanX(rb::AveragingScheme scheme = kDefaultScheme) const {return MeanXYZ(scheme).X();}
    double MeanY(rb::AveragingScheme scheme = kDefaultScheme) const {return MeanXYZ(scheme).Y();}
    double MeanZ(rb::AveragingScheme scheme = kDefaultScheme) const {return MeanXYZ(scheme).Z();}
    double MeanTNS(rb::AveragingScheme scheme = kDefaultScheme) const;

    /// Gets the min/max/mean all at once, called by the functions above
    void MinMaxMeanXYZ(TVector3& lo, TVector3& hi, TVector3& mean,
                       rb::AveragingScheme scheme = kDefaultScheme) const;
    
    //@}

    /// \name Extents
    /// The size of a box containing all the hits
    //@{
    TVector3 ExtentXYZ() const {return MaxXYZ()-MinXYZ();}
    double ExtentV(geo::View_t view) const {return MaxV(view)-MinV(view);}
    double ExtentX() const {return ExtentXYZ().X();}
    double ExtentY() const {return ExtentXYZ().Y();}
    double ExtentZ() const {return ExtentXYZ().Z();}
    unsigned int ExtentPlane(geo::View_t view = geo::kXorY) const {return MaxPlane(view)-MinPlane(view)+1;}
    unsigned int ExtentCell(geo::View_t view) const;
    double ExtentTNS() const {return MaxTNS()-MinTNS();}
    //@}

    /// \name Connectedness measures
    //@{
    /// \brief Longest run of adjacent planes with hits
    ///
    /// \a view may be kXorY to require hits on every plane no matter the view
    int MostContiguousPlanes(geo::View_t view) const;

    /// \brief Longest run of adjacent planes with no hits
    ///
    /// \a view may be kXorY to allow a hit in either view to end the run
    int MostMissingPlanes(geo::View_t view) const;

    /// \brief Total number of missing planes in cluster.  
    ///
    /// Put simply, number of planes spanned minus number of hit planes. 
    /// If track is 2D, only planes in that view are counted. 
    int NMissingPlanes(geo::View_t view) const;
    //@}


    bool operator< (const Cluster& other) const;

    friend std::ostream& operator << (std::ostream& o, const Cluster& c);

  protected:
    /// \name Internal helpers
    //@{


    /// Helper. Resizes weights vectors to match cell vectors
    void EnsureWeightAlloc();

  private:
    /// Optimized form of TotalGeV, only valid for actual Clusters
    double TotalGeVFastClusterOnly() const;
    //@}
#endif // __GCCXML__
  protected:

    geo::View_t                 fView;         ///< view this cluster is in
    art::PtrVector<rb::CellHit> fXCell;        ///< collection of x-view cells in cluster
    art::PtrVector<rb::CellHit> fYCell;        ///< collection of y-view cells in cluster
    std::vector<double>         fXWeights;     ///< Weights, matching cell indexing
    std::vector<double>         fYWeights;     ///< May be empty, means all weights are 1
    int                         fID;           ///< ID for cluster
    bool                        fNoiseCluster; ///< flag for whether this is a noise cluster
  };
}

#endif // RBCLUSTER_H
/////////////////////////////////////////////////////////////////////////////////////////////
