////////////////////////////////////////////////////////////////////////
// \file    FitSum.h
// \brief   Container class for basic kinematic information.
// \version $Id: FitSum.h,v 1.0 2014-09-25 00:00:00 mbaird42 Exp $
// \author  $Author: mbaird42 $
// \date    $Date: 2014-09-25 00:00:00 $
////////////////////////////////////////////////////////////////////////
#ifndef FITSUM_H
#define FITSUM_H

#include "TVector3.h"
#include "TLorentzVector.h"

namespace rb
{
  /// \brief A container for kinematic information.  
  ///
  /// The rb::FitSum object is a simple container that
  /// holds simple kinematic information.  It is intended to be an 
  /// object produced in association with either a prong or a track.
  ///
  class FitSum 
  {
  public:

    FitSum();
    ~FitSum() {};

#ifndef __GCCXML__

    FitSum(TVector3 dir, double momentum, double mass, int pdg = 0,
	   double chi2_hit1 = 0.0, double chi2_hit2 = 0.0, 
	   double chi2_angle1 = 0.0, double chi2_angle2 = 0.0, 
	   int Ndof_hit1 = 1, int Ndof_hit2 = 1,
	   int Ndof_angle1 = 1, int Ndof_angle2 = 1);

    void SetKin(TVector3 dir, double momentum, double mass);
    void SetPDG(int pdg) { fPDG = pdg; }

    //
    // Setting chi2 and ndof for each fit parameter and view.
    //
    void SetChi2(double chi2_hit1, double chi2_hit2,
		 double chi2_angle1, double chi2_angle2,
		 int Ndof_hit1, int Ndof_hit2,
		 int Ndof_angle1, int Ndof_angle2)
    { 
      fChi2[0]    = chi2_hit1;   // chi squared for hits in XZ view.
      fChi2[1]    = chi2_hit2;   // chi squared for hits in YZ view.
      fChi2[2]    = chi2_angle1; // chi squared for kink angle in XZ view.
      fChi2[3]    = chi2_angle2; // chi squared for kink angle in YZ view.
      fNdof[0]    = Ndof_hit1;   // Degrees of freedom (nhits) in XZ view.
      fNdof[1]    = Ndof_hit2;   // Degrees of freedom (nhits) in YZ view.
      fNdof[2]    = Ndof_angle1; // Degrees of freedom for fit angle (1) in XZ view.
      fNdof[3]    = Ndof_angle2; // Degrees of freedom for fit angle (1) in YZ view.
    }

    //
    // Function returning chi2 values for specific parameters and views.
    //
    const double Chi2(int index) const{ 
      return fChi2[index];
    }

    //
    // Function returning degrees of freedom for specific parameters and views.
    //
    const int Ndof(int index) const{ 
      return fNdof[index];
    }
    
    const TLorentzVector FourMom()       const { return fFourMom;    }
    const int            PDG()           const { return fPDG;        }

#endif // __GCCXML__

  protected:
    TLorentzVector fFourMom;       ///< four momentum
    int            fPDG;           ///< pdg code for particle type
    double         fChi2[8];       ///< provide user with a chi-squared for track fit parameters 
    int            fNdof[8];       ///< provide user with N degrees of freedom for track fit parameters 
  };
}

#endif // FITSUM_H
////////////////////////////////////////////////////////////////////////
