////////////////////////////////////////////////////////////////////////////////////
/// \file    HoughResult.h
/// \brief   Data resulting from a Hough transform on the cell hit positions
/// \version $Id: HoughResult.h,v 1.7 2012-10-26 21:31:55 bckhouse Exp $
/// \author  messier@indiana.edu
/////////////////////////////////////////////////////////////////////////////////////
#ifndef RB_HOUGHRESULT_H
#define RB_HOUGHRESULT_H

#include <vector>

#ifndef __GCCXML__
#include "GeometryObjects/PlaneGeo.h"
#endif

namespace rb {
  /// \brief Data for a single peak in Hough space
  class HoughPeak 
  {
  public:
    HoughPeak();
    HoughPeak(double rho, double theta, double height, double thresh, double sr, double st);
  public:
    double fRho;   ///< Distance to line
    double fTheta; ///< Angle to line (radians)
    double fH;     ///< Height of peak (arbirary units)
    double fThr;   ///< Calculated threshold for hough map
    double fSR;    ///< Uncertainty in Rho   (999 indicates peak was below threshold)
    double fST;    ///< Uncertainty in Theta (999 indicates peak was below threshold)
  };
  
  //......................................................................
  
  /// \brief Summary from a Hough transform applied to a group of cell hits
  class HoughResult
  {
  public:
    HoughResult();
#ifndef __GCCXML__
    HoughResult(geo::View_t v,
		double      tlo,
		double      thi,
		double      xy0,
		double      z0);
    
    /// \brief Slope and intercepts of Hough lines
    /// \param i - Which line?
    /// \param m - Slope of line
    /// \param b - Intercept of line (detector coordinates)
    void SlopeIcept(unsigned int i, double* m, double* b) const;
    
    /// \brief Sort Hough peaks from best (highest) to worst
    void SortPeaks();
#endif // __GCCXML__
  public:
    geo::View_t            fView;  ///< Transform of x or y view?
    double                 fTNSlo; ///< Low edge of time slice transformed
    double                 fTNShi; ///< Upper edge of time slice transformed
    double                 fXY0;   ///< X/Y offest applied to hits
    double                 fZ0;    ///< Z offset applied to hits
    std::vector<HoughPeak> fPeak;  ///< List of peaks found in Hough space
  };
} // end namespace rb

#endif
////////////////////////////////////////////////////////////////////////
