////////////////////////////////////////////////////////////////////////
// \file    Prong.h
// \brief   A Prong is a Cluster with defined start position and direction.
// \version $Id: Prong.h,v 1.11 2012-10-26 21:31:55 bckhouse Exp $
// \author  $Author: bckhouse $
// \date    $Date: 2012-10-26 21:31:55 $
////////////////////////////////////////////////////////////////////////
#ifndef RBPRONG_H
#define RBPRONG_H

#include "RecoBase/Cluster.h"
#include "RecoBase/CellHit.h"

#include "TVector3.h"

namespace rb
{
  /// \brief A \ref Cluster with defined start position and direction
  class Prong : public Cluster
  {
  public:
    Prong(){}

#ifndef __GCCXML__

    /// \name 3D prong constructors
    //@{
    /// \param cells CellHits associated with this prong
    /// \param start 3D position of the prong vertex
    /// \param dir   Reconstructed prong direction. Automatically normalized
    Prong(const art::PtrVector<rb::CellHit>& cells,
    TVector3 start, TVector3 dir, int id=0);
    /// \param clust Cluster holding CellHits associated with this prong
    /// \param start 3D position of the prong vertex
    /// \param dir   Reconstructed prong direction. Automatically normalized
    Prong(const rb::Cluster& clust,
    TVector3 start, TVector3 dir, int id=0);
    //@}
    /// \name 2D prong constructors
    //@{
    /// \param cells CellHits associated with this prong
    /// \param view  View this prong is defined in
    /// \param v0    %Vertex position in coordinate matching \a view
    /// \param z0    %Vertex z-position
    /// \param dv    Component of prong direction in \a view coordinate
    /// \param dz    z-component of prong direction
    Prong(const art::PtrVector<rb::CellHit>& cells,
	  geo::View_t view, double v0, double z0, double dv, double dz, int id=0);
    //@}
    /// \name 2D prong constructors
    //@{
    /// \param cells CellHits associated with this prong
    /// \param view  View this prong is defined in
    /// \param v0    %Vertex position in coordinate matching \a view
    /// \param z0    %Vertex z-position
    /// \param dv    Component of prong direction in \a view coordinate
    /// \param dz    z-component of prong direction
    Prong(const std::vector< art::Ptr<rb::CellHit> >& cells,
	  geo::View_t view, double v0, double z0, double dv, double dz, int id=0);
    /// \param clust Cluster holding CellHits associated with this prong
    /// \param v0    %Vertex position in coordinate matching Cluster's view
    /// \param z0    %Vertex z-position
    /// \param dv    Component of prong direction in the Cluster's view
    /// \param dz    z-component of prong direction
    Prong(const rb::Cluster& clust,
	  double v0, double z0, double dv, double dz, int id=0);
    //@}

    virtual ~Prong();

    virtual TVector3 Start() const {return fStart;}
    /// \brief Unit vector describing prong direction
    ///
    /// For 2D prongs the unmeasured component is zero
    virtual TVector3 Dir() const {return fDir;}

    virtual void SetStart(TVector3 start);
    virtual void SetStart(double v0, double z0);

    virtual void SetDir(TVector3 dir);
    /// \param dv Component of prong direction in \ref View coordinate
    /// \param dz z-component of prong direction
    virtual void SetDir(double dv, double dz);

    /// \brief Estimate the unmeasured coordinate of \a chit
    ///
    /// For \ref Prong, this is calculated by taking the position of the
    /// straight line defined by \ref Start and \ref Dir at the z-coordinate
    /// of the cell.
    ///
    /// \param  chit Any \ref CellHit, not necessarily owned by this container
    /// \return      Estimated unmeasured coordinate of this hit (cm) assuming
    ///              it to be associated with this reconstructed object.
    virtual double W(const rb::CellHit* chit) const;

    /// \brief Distance along prong to reach last cell hit
    ///
    /// Overridden by Track and Shower
    virtual double TotalLength() const;

    /// \param  z z-coordinate to evaluate trajectory at
    /// \return   Distance from \ref Start measured along the trajectory
    virtual double DistanceFromStart(double z) const;

    bool operator< (const Prong& other) const;

#endif // __GCCXML__
  protected:
    TVector3 fStart; ///< Start location (xyz, cm)
    TVector3 fDir;   ///< Direction at starting point
  };

} // end namespace

#endif // RBPRONG_H
//////////////////////////////////////////////////////////////////////////////
