////////////////////////////////////////////////////////////////////////
// \file    Track.h
// \brief   A track is a \ref rb::Prong with full reconstructed trajectory.
// \version $Id: Track.h,v 1.15 2012-10-26 21:31:55 bckhouse Exp $
// \author  Christopher Backhouse - bckhouse@caltech.edu
// \date    $Date: 2012-10-26 21:31:55 $
////////////////////////////////////////////////////////////////////////
#ifndef RBTRACK_H
#define RBTRACK_H

#include "RecoBase/Prong.h"

#include <vector>
#include <utility>
#include <map>

namespace rb
{
  /// \brief A \ref rb::Prong with full reconstructed trajectory
  class Track : public Prong
  {
  public:
    /// \name Constructors
    ///
    /// See \ref rb::Prong documentation
    //@{
    Track(){}
#ifndef __GCCXML__
    Track(const art::PtrVector<rb::CellHit>& cells,
	  TVector3 start, 
	  TVector3 dir, 
	  int      id = 0);
    Track(const art::PtrVector<rb::CellHit>& cells, 
	  geo::View_t view,
	  double      v0,   
	  double      z0,   
	  double      dv,   
	  double      dz,   
	  int         id = 0);
    Track(const std::vector< art::Ptr<rb::CellHit> >& cells, 
	  geo::View_t view,
	  double      v0,   
	  double      z0,   
	  double      dv,   
	  double      dz,   
	  int         id = 0);
    Track(const rb::Cluster& clust,
	  TVector3 start, 
	  TVector3 dir, 
	  int      id = 0);
    Track(const rb::Cluster& clust,
	  double v0, 
	  double z0, 
	  double dv, 
	  double dz, 
	  int    id = 0);
    Track(const rb::Prong& prong, 
	  int id = 0);
    //@}
#endif // __GCCXML__
    virtual ~Track();
#ifndef __GCCXML__

    void AppendTrajectoryPoint(TVector3 pt);
    void AppendTrajectoryPoint(double pv, double pz);

    /// \brief Support constructing tracks backwards
    ///
    /// Note: This function also sets the track start to \a pt
    void PrependTrajectoryPoint(TVector3 pt);
    void PrependTrajectoryPoint(double pv, double pz);

    size_t NTrajectoryPoints() const {return fTraj.size();}

    /// \brief return a constant reference to the track trajectory points
    std::vector<TVector3> const& Trajectory() const { return fTraj; }

    /// The ith point on the trajectory, a 3-vector in cm
    TVector3 TrajectoryPoint(unsigned int i) const;

    // Need to override these to set the first trajectory point
    virtual void SetStart(TVector3 start);
    virtual void SetStart(double v0, double z0);

    /// Position of the final trajectory point
    TVector3 Stop() const;

    /// Get direction at the stop point
    virtual TVector3 StopDir() const;

    /// \brief Length (cm) of all the track segments
    ///
    /// ie This measure follows any curvature in the track
    virtual double TotalLength() const;

    virtual double DistanceFromStart(double z) const;
    virtual double DistanceFromEnd  (double z) const;

    /// \brief Estimate the unmeasured coordinate of \a chit
    ///
    /// For \ref Track, this is calculated by taking the position of the 
    /// straight line defined by the two neighbouring trajectory points at 
    /// the z-coordinate of the cell. In the case of a hit beyond the end of 
    /// the track, the line defined by the two hits at the end is extrapolated.
    ///
    /// \param  chit Any \ref CellHit, not necessarily owned by this container
    /// \return      Estimated unmeasured coordinate of this hit (cm) assuming 
    ///              it to be associated with this reconstructed object.
    virtual double W(const rb::CellHit* chit) const;

    /// \param       z Coordinate at which trajectory is to be evaluated
    /// \param [out] x x-position of trajectory at this z
    /// \param [out] y y-position of trajectory at this z
    virtual void InterpolateXY(double z, double& x, double& y) const;

    /// \param  z Coordinate at which trajectory is to be evaluated
    /// \return   The track direction estimated at that point
    virtual TVector3 InterpolateDir(double z) const;

    /* Commented out because the current implementation is overengineered and slow
    /// Pathlength of the trajectory through cell \a chit, in cm.
    double L(const rb::CellHit* chit) const;
    */

    /// Forget about all trajectory points
    void ClearTrajectoryPoints();

    /// Remove the ith trajectory point from the track
    void RemoveTrajectoryPoint(unsigned int i);

    /// Combine with a 2D track from the other view to make a 3D track
    rb::Track ZipWith(const rb::Track& trk) const;

    /// map of the direction cosines at each plane (ie z)
    std::map<unsigned int, TVector3> PlaneDirMap() const;

#endif // __GCCXML__
  protected:

#ifndef __GCCXML__ 

    /// Helper function for \ref W and \ref L
    void FindNeighbouringPoints(double z, TVector3& pt0, TVector3& pt1) const;

    void FindNeighbouringPointIndices(double z, int& i0, int& i1) const;

    /// Helper function for \ref InterpolateXY and \ref ZipWith
    void InterpolatePts(double z,
                        const TVector3& a, const TVector3& b,
                        double& x, double& y) const;

#endif // __GCCXML__

    std::vector<TVector3> fTraj;
 };
}

#endif // RBTRACK_H
