////////////////////////////////////////////////////////////////////////
/// \brief   An  SRNeutrino represents a true neutrino, adding neutrino 
///          information on top of the common to all particles SRTruth info.
/// \author  gsdavies@iastate.edu
/// \date    February 2013
////////////////////////////////////////////////////////////////////////
#ifndef SRNEUTRINO_H
#define SRNEUTRINO_H

#include "StandardRecord/SRBeam.h"
#include "StandardRecord/SRTrueMichelE.h"
#include "StandardRecord/SRTrueParticle.h"

#include "TVector3.h"
#include "TLorentzVector.h"

namespace caf
{
  
  /// Neutrino interaction categories
  enum int_type_{
    kQE                        = 0,
    kRes                       = 1,
    kDIS		       = 2,
    kCoh		       = 3,
    kNuanceOffset              = 1000,                ///< offset to account for adding in Nuance codes to this enum
    kCCQE                      = kNuanceOffset + 1,   ///< charged current quasi-elastic	      	
    kNCQE                      = kNuanceOffset + 2,   ///< neutral current quasi-elastic	      	
    kResCCNuProtonPiPlus       = kNuanceOffset + 3,   ///< resonant charged current, nu p -> l- p pi+
    kResCCNuNeutronPi0         = kNuanceOffset + 4,   ///< resonant charged current, nu n -> l- n pi0
    kResCCNuNeutronPiPlus      = kNuanceOffset + 5,   ///< resonant charged current, nu n -> l- n pi+
    kResNCNuProtonPi0          = kNuanceOffset + 6,   ///< resonant neutral current, nu p -> nu p pi0
    kResNCNuProtonPiPlus       = kNuanceOffset + 7,   ///< resonant neutral current, nu p -> nu p pi+
    kResNCNuNeutronPi0         = kNuanceOffset + 8,   ///< resonant neutral current, nu n -> nu n pi0
    kResNCNuNeutronPiMinus     = kNuanceOffset + 9,   ///< resonant neutral current, nu n -> nu p pi-
    kResCCNuBarNeutronPiMinus  = kNuanceOffset + 10,  ///< resonant charged current, nubar n -> l+ n pi-
    kResCCNuBarProtonPi0       = kNuanceOffset + 11,  ///< resonant charged current, nubar p -> l+ n pi0
    kResCCNuBarProtonPiMinus   = kNuanceOffset + 12,  ///< resonant charged current, nubar p -> l+ p pi-
    kResNCNuBarProtonPi0       = kNuanceOffset + 13,  ///< resonant charged current, nubar p -> nubar p pi0
    kResNCNuBarProtonPiPlus    = kNuanceOffset + 14,  ///< resonant charged current, nubar p -> nubar n pi+
    kResNCNuBarNeutronPi0      = kNuanceOffset + 15,  ///< resonant charged current, nubar n -> nubar n pi0
    kResNCNuBarNeutronPiMinus  = kNuanceOffset + 16,  ///< resonant charged current, nubar n -> nubar p pi-  
    kResCCNuDeltaPlusPiPlus    = kNuanceOffset + 17,
    kResCCNuDelta2PlusPiMinus  = kNuanceOffset + 21,
    kResCCNuBarDelta0PiMinus   = kNuanceOffset + 28,
    kResCCNuBarDeltaMinusPiPlus= kNuanceOffset + 32,
    kResCCNuProtonRhoPlus      = kNuanceOffset + 39,
    kResCCNuNeutronRhoPlus     = kNuanceOffset + 41,
    kResCCNuBarNeutronRhoMinus = kNuanceOffset + 46,
    kResCCNuBarNeutronRho0     = kNuanceOffset + 48,
    kResCCNuSigmaPlusKaonPlus  = kNuanceOffset + 53,
    kResCCNuSigmaPlusKaon0     = kNuanceOffset + 55,
    kResCCNuBarSigmaMinusKaon0 = kNuanceOffset + 60,
    kResCCNuBarSigma0Kaon0     = kNuanceOffset + 62,
    kResCCNuProtonEta          = kNuanceOffset + 67,
    kResCCNuBarNeutronEta      = kNuanceOffset + 70,
    kResCCNuKaonPlusLambda0    = kNuanceOffset + 73,
    kResCCNuBarKaon0Lambda0    = kNuanceOffset + 76,
    kResCCNuProtonPiPlusPiMinus= kNuanceOffset + 79,
    kResCCNuProtonPi0Pi0       = kNuanceOffset + 80,
    kResCCNuBarNeutronPiPlusPiMinus = kNuanceOffset + 85,
    kResCCNuBarNeutronPi0Pi0   = kNuanceOffset + 86,
    kResCCNuBarProtonPi0Pi0    = kNuanceOffset + 90,
    kCCDIS                     = kNuanceOffset + 91,  ///< charged current deep inelastic scatter
    kNCDIS                     = kNuanceOffset + 92,  ///< charged current deep inelastic scatter
    kUnUsed1                   = kNuanceOffset + 93,
    kUnUsed2                   = kNuanceOffset + 94,
    kCCQEHyperon               = kNuanceOffset + 95,
    kNCCOH                     = kNuanceOffset + 96,
    kCCCOH                     = kNuanceOffset + 97,  ///< charged current coherent pion         
    kNuElectronElastic         = kNuanceOffset + 98,  ///< neutrino electron elastic scatter	
    kInverseMuDecay            = kNuanceOffset + 99   ///< inverse muon decay			  
  };
  
  
  /// The SRNeutrino is a representation of neutrino interaction information.  
  /// It inherits information from SRTruth but adds more information pertinent 
  /// only to the neutrino.
  class SRNeutrino
    {
    public:
      ///  Default constructor for SRNeutrino. 
      SRNeutrino();

      /// Default destuctor.
      ~SRNeutrino() {  };

      short          pdg;           ///< pdg code
      float          E;             ///< True energy [GeV]
      float          visE;          ///< Sum of FLS hits that made CellHits from this neutrino [GeV]
      float          visEinslc;     ///< Sum of FLS hits that made CellHits  from this neutrino in this subevent [GeV]
      bool           isvtxcont;     ///< Checks if neutrino true vertex is within detector

      float          eff;           ///< Slicer efficiency for this truth interaction
      float          pur;           ///< Slicer purity for this truth interaction
      unsigned int   nhitslc;       ///< Number of hits recorded in this slice by this truth interaction
      unsigned int   nhittot;       ///< Total number of hits recorded for this truth interaction

      float          time;          ///< interaction time.
      TLorentzVector p;             ///< True momentum [GeV]
      TVector3       vtx;           ///< Vertex postion in detector coordinates [cm]
      std::vector<SRTrueMichelE> michel;  ///< Vector of true Michel electrons
      
      short          pdgorig;       ///< Unoscillated (unswapped) pdg code
      short          hitnuc;        ///< Struck nucleus
      float          woscdumb;      ///< Simplest possible oscillation weight 
      
      int            mode;          ///< interaction mode from enum int_type[QE, RES, COH, ...]
      bool           iscc;          ///< true if charged-current interaction, false if not.
      int            inttype;       ///< Interaction type enum int_type::[...]

      float          q2;            ///< Squared momentum transfer [GeV^2]
      float          x;             ///< Bjorken x = (k-k')^2 / (2*p.q), [Dimensionless]
      float          y;             ///< Bjorken y =  (p.q) / (k.p), fractional energy loss of incoming particle [Dimensionless]
      float          W2;            ///< Invariant mass of final state squared. [GeV^2]

      
      std::vector<SRTrueParticle> prim;  ///< Primary daughters, lepton comes first in vector.

      SRBeam beam; ///< Information about neutrino production
    };
  
} // end namespace

#endif // SRNEUTRINO_H
//////////////////////////////////////////////////////////////////////////////
