//////////////////////////////////////////////////////////////////////////
/// \brief  Cosmic track utility code
/// \author Brian Rebel - brebel@fnal.gov
/// \date 
//////////////////////////////////////////////////////////////////////////
#ifndef TRACKFIT_COSMICTRACKUTILITIES_H
#define TRACKFIT_COSMICTRACKUTILITIES_H

#include "fhiclcpp/ParameterSet.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"

#include "GeometryObjects/CellUniqueId.h"
#include "Geometry/LiveGeometry.h"
#include "Geometry/Geometry.h"

class TVector3;

namespace rb{
  class Track;
  class RecoHit;
}
namespace simb{
  class MCParticle;
}
namespace sim{
  class FLSHit;
}

namespace trk{

  typedef std::pair<TVector3, TVector3> zBounds;
  typedef std::map<double, zBounds>     zBoundMap;

  // A class to select cosmic ray muons for analysis
  class CosmicTrackUtilities {

  public:
    
    explicit CosmicTrackUtilities();
    virtual ~CosmicTrackUtilities();

    void reconfigure(fhicl::ParameterSet const& pset);

    float RangeMomentum   (rb::Track const& track);
    float LengthCorrection(float     const& length);
    float CosTheta        (float     const& dcosY, 
			   float     const& magnitude);
    float Azimuth         (float     const& dcosX, 
			   float     const& dcosZ, 
			   float     const& magnitude);

    std::pair<std::pair<float, float>, 
      std::pair<TVector3, TVector3> > TrueLengthEndPoints(simb::MCParticle const& part); 

    std::map<geo::CellUniqueId, float> TruePathLengthInCell(simb::MCParticle         const& part,
							    std::vector<sim::FLSHit> const& flsHits);

    void  FindTriCells(rb::Track                                    const& track,
		       std::map<std::pair<uint32_t, uint32_t>,bool>      & planeCellMap);
    float TriCellPathLength(rb::Track   const& track,
			    uint32_t    const& plane,
			    double      const& cellWidth,
			    geo::View_t const& view);
    
    float PathLengthInCell(zBoundMap                     const& zBounds,
			   TVector3                      const& recoHitLoc,
			   std::pair<uint32_t, uint32_t> const& pc);

    zBoundMap           MakeZBoundaryMap(std::vector<TVector3> const& trajectory);
    zBounds             ZBounds(zBoundMap const& bounds,
				double    const& hitZ);    
    std::vector<float>  TrackResiduals(rb::Track const& track);
    
    bool                PointInDetector(TVector3 const& point);

  private:

    void                FindZBoundaries();

    double        	fEffectivedEdx;           ///< effective dE/dx in MeV/cm  
    std::vector<double> fPolCoeff;                ///< the coefficients in an 8th degree polynomial  
    std::set<double>    fPlaneZBounds;            ///< boundaries of each plane in the z direction
    double              fPlaneZDepth;             ///< extent of one plane (cell) in the z direction
    
    art::ServiceHandle<geo::LiveGeometry> fLiveGeom;
    art::ServiceHandle<geo::Geometry>     fGeom;

  };

}

#endif // TRACKFIT_COSMICTRACKSELECTION_H
