///
/// \file    RLFit.h
/// \brief   Make a linear fit to a collection of raw hits
/// \version $Id: RLFit.h,v 1.4 2011-04-13 04:15:02 messier Exp $
/// \author  messier@indiana.edu
///
#ifndef TRACKFIT_RLFIT_H
#define TRACKFIT_RLFIT_H

#include <vector>

#include "art/Persistency/Common/Ptr.h"

namespace rawdata { class RawDigit; }

namespace trk {
  ///
  /// Perform a straight line fit to a collection of raw hits
  ///
  /// RLFit = Raw Line Fit
  ///
  class RLFit {
  public:
    ///
    /// Construct the RLFit object
    ///
    /// \param d    - list of raw digits to fit
    /// \param i1   - first hit in list to fit
    /// \param i2   - last hit in list to fit (inclusive)
    /// \param cm   - the experiment connection map
    /// \param geom - the experiment geometry
    ///
    RLFit(const std::vector< art::Ptr<rawdata::RawDigit> >& d,
	  unsigned int i1,
	  unsigned int i2);
    ///
    /// Perform the fit. Results stored into member data.
    ///
    void Fit();

    ///
    /// Flip the fit results to force a down-going cosmic ray
    /// assumption on the results
    ///
    void AssumeDownGoing();

    ///
    /// Flip the fit results to force a beam-related assumption on the
    /// results
    ///
    void AssumeNorthGoing();

    ///
    /// Angle of track wrt detector z-axis
    ///
    double CosThetaBeam();

    ///
    /// Angle of track in detector x-y plane
    ///
    double PhiBeam();

    ///
    /// Angle of track wrt to vertical
    ///
    double CosThetaCosmic();
    
    ///
    /// Angle of track in east-west plane
    ///
    double PhiCosmic();
    
    ///
    /// Update the schedule of the boundaries and widths used in
    /// calculating weights
    ///
    void SetRTSchedule(const std::vector<double>& r, 
		       const std::vector<double>& t);
    
  private:
    void UpdateSlopes();
    void SeedWeights(const std::vector<double>& x, 
		     const std::vector<double>& zx,
		     const std::vector<double>& y,
		     const std::vector<double>& zy,
		     std::vector<double>& w);

  public:
    std::vector<double> fR; ///< Range of weight function
    std::vector<double> fT; ///< Transition width of weight function

    //..................................................................
    std::vector<const rawdata::RawDigit*> fXhit; ///< X view hits
    std::vector<const rawdata::RawDigit*> fYhit; ///< Y view hits
    std::vector<double>                   fXx;   ///< x positions (cm)
    std::vector<double>                   fZx;   ///< z positions (cm)
    std::vector<double>                   fWx;   ///< x view weights
    std::vector<double>                   fYy;   ///< y positions (cm)
    std::vector<double>                   fZy;   ///< z positions (cm)
    std::vector<double>                   fWy;   ///< y view weights

    double fMzx; ///< Slope in x-z view
    double fBzx; ///< Intercept in x-z view
    double fMzy; ///< Slope in y-z view
    double fBzy; ///< Intercept in y-z view

    double fX1; ///< Start x position (cm)
    double fY1; ///< Start y position (cm)
    double fZ1; ///< Start z position (cm)
    double fX2; ///< End x position (cm)
    double fY2; ///< End y positon (cm)
    double fZ2; ///< End z position (cm)
    
    double fBoxXlo; ///< Detector box X edge
    double fBoxXhi; ///< Detector box X edge
    double fBoxYlo; ///< Detector box Y edge
    double fBoxYhi; ///< Detector box Y edge
    double fBoxZlo; ///< Detector box Z edge
    double fBoxZhi; ///< Detector box Z edge
    
    double fChi2; /// Sum of weights * deltas
    double fDOF;  /// Sum of weights
  };
}

#endif // RAWUTIL_LINEFIT_H
////////////////////////////////////////////////////////////////////////
