////////////////////////////////////////////////////////////////////////
/// \brief   Apply mis-alignment with TranslatePlane() method and then 
///          Alignment code calculates residuals for tracks for each cell.
///          Populates a TH2 using D2M method that can be used to calculate
///          alignment correction inter-cell.
/// \author  gsdavies@iastate.edu
////////////////////////////////////////////////////////////////////////

// C/C++ includes
#include <cmath>
#include <iostream>
#include <map>
#include <vector>

// NOvASoft includes
#include "Geometry/Geometry.h"
#include "GeometryObjects/Geo.h"
#include "GeometryObjects/OfflineChan.h"
#include "GeometryObjects/CellGeo.h"
#include "RecoBase/CellHit.h"
#include "RecoBase/HitMap.h"
#include "RecoBase/Track.h"
#include "Utilities/func/MathUtil.h"

// ROOT includes
#include "TH1.h"
#include "TH2.h"

// ART includes
#include "art/Framework/Core/EDAnalyzer.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Services/Optional/TFileService.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "art/Framework/Principal/Handle.h"
#include "messagefacility/MessageLogger/MessageLogger.h"


/// Relative alignment of detector cells and blocks
namespace align {
  
  class Alignment : public art::EDAnalyzer {
    public:
      explicit Alignment(fhicl::ParameterSet const& pset);
      virtual ~Alignment();                        
      
      void analyze(art::Event const& evt);
      void reconfigure(const fhicl::ParameterSet& pset);
      
      void beginJob();
      void endJob();
      
    private:
      
      bool IsGoodTrack(art::Ptr<rb::Track> t, double x, double y, double z); ///< yes/no is "track" entering within a buffer window
      std::string   fCellHitsListLabel; ///< Where to find CellHit
      std::string   fTrackLabel;        ///< Where to find Tracks
            
      unsigned int  fMuonNhit;          ///< Required # hits on muon track
      unsigned int  fMuonNhitX;         ///< # required in x view
      unsigned int  fMuonNhitY;         ///< # required in y view
      double        fDetEdgeBuffer;     ///< Detector edge buffer zone for entering cosmic ray track decision [cm]
      
      //......................................................................
      
      struct Val_t{
        TH1 *ResidV, *ResidZ;
        TH2 *HCellPos;
      };
      Val_t GetChannelHists(int plane, int cell);

      /// This is where \ref GetChannelHist stores its histograms
      std::map<geo::OfflineChan, Val_t> fChannelMap; 
    };
};


////////////////////////////////////////////////////////////////////////
namespace align{

  Alignment::Alignment(fhicl::ParameterSet const& pset): 
    EDAnalyzer(pset),
    fCellHitsListLabel(pset.get< std::string  >("CellHitsListLabel")             ),
    fTrackLabel       (pset.get< std::string  >("TrackLabel")                    ),
    fMuonNhit         (pset.get< unsigned int >("MuonNhit")                      ),
    fMuonNhitX        (pset.get< unsigned int >("MuonNhitX")                     ),
    fMuonNhitY        (pset.get< unsigned int >("MuonNhitY")                     ),
    fDetEdgeBuffer    (pset.get< double       >("DetEdgeBuffer")                 )
  {
    reconfigure(pset); 
  }

  //......................................................................
  Alignment::~Alignment()
  {
  }
      
  //......................................................................
  void Alignment::beginJob()
  {
  }

  //......................................................................
  
  void Alignment::analyze(art::Event const& evt)
  {
    art::ServiceHandle<geo::Geometry>         geom;

    // Get the tracks that were made by CosmicTrack
    // and test that there are tracks in the event
    art::Handle< std::vector<rb::Track> > trks;
    evt.getByLabel(fTrackLabel,trks);
    if (trks->size()==0) {
      mf::LogWarning("AlignWarn") << "There are " << trks->size() 
 	                          << " Tracks in this event, skip";
      return;
    }
    LOG_DEBUG("Alignment") << "There are " << trks->size() << " Tracks in this event";
    
    
    const unsigned int trkMax = trks->size();
    
    for(unsigned int trkIdx = 0; trkIdx < trkMax; ++trkIdx){
      art::Ptr<rb::Track> track(trks, trkIdx);
      
      TVector3 start(track->Start());
      TVector3 end(track->Stop());
      if(end.Y() > start.Y()) std::swap(start, end);
      
      // Is this a good track?? i.e. entering and extends across at least 8 planes
      if (this->IsGoodTrack(track,start.X(),start.Y(),start.Z()) == false) return;
      assert(track->ExtentPlane() >= 8); 

      const unsigned int hitMax = track->NCell();  // maximum number of CellHits on track
      const unsigned int hitX   = track->NXCell(); // maximum number of CellHits on track in XZ
      const unsigned int hitY   = track->NYCell(); // maximum number of CellHits on track in YZ

      // Do NOT select tracks if max number of hits or XZ- YZ-hits less than configurable values
      if ( (hitMax < fMuonNhit) || (hitX < fMuonNhitX) || (hitY < fMuonNhitY) ) return;
      assert(hitMax >= fMuonNhit);
      
      rb::HitMap hmap;
      // Loop through both views and make a list of every CellHit in this track
      for(int view = geo::kX; view <= geo::kY; ++view){	
        // C++ is a bit strict about the conversions. Save typing.	
        const geo::View_t geoview = geo::View_t(view);
        // Make a list of every hit in this track
        const unsigned int hitMaxView = track->NCell(geoview);
        for(unsigned int hitIdx = 0; hitIdx < hitMaxView; ++hitIdx){
          art::Ptr<rb::CellHit> hit = track->Cell(geoview, hitIdx);
	  hmap.Add(hit);
	} // end for hitIdx
      } // view
      
      const unsigned int pMax = geom->NPlanes();
    
      for( unsigned int pIdx = 0; pIdx < pMax; pIdx++){ // loop through planes

        geo::PlaneGeo* p = new geo::PlaneGeo(*geom->Plane(pIdx));
	const unsigned int cMax = p->Ncells();
	
	// This command is what actually applies the Translation/Rotation.
	// In this case it applies a 50cm translation in the +z-axis
	// For now this is the default. Will re-work code so that this is a
	// fcl parameter input.
	p->TranslatePlane(0,0,50);
       

        for( unsigned int cIdx = 0; cIdx < cMax; cIdx++){ // loop through cells
	  if(hmap.CellExists(pIdx,cIdx)) {           
            double xyz[3];
	    geo::CellGeo* c = new geo::CellGeo(*p->Cell(cIdx));
	  
            c->GetCenter(xyz, 0.0);

            const geo::View_t view = geom->Plane(pIdx)->View();
      
            double residV[2];
            double residZ[2];
            TVector3 closest; 
            double vz,zz;
        
            // New format (correct) whereby residual is V(Z0) - V0 so that V0' = V0 + Residual[V]
            if(view == geo::kX){
              TVector3 start_vec(start.X(),0.0,start.Z());
              TVector3 end_vec(end.X(),0.0, end.Z());
              TVector3 xyz_vec(xyz[0],0.0,xyz[2]);
              geo::ClosestApproach(xyz_vec,start_vec,(end_vec - start_vec),closest);
              vz = closest[0];
              zz = closest[2];
              residV[view] = vz - xyz[0];
              residZ[view] = zz - xyz[2];
            }
            if(view == geo::kY){
              TVector3 start_vec(0.0,start.Y(),start.Z());
              TVector3 end_vec(0.0,end.Y(),end.Z());
              TVector3 xyz_vec(0.0,xyz[1],xyz[2]);
              geo::ClosestApproach(xyz_vec,start_vec,(end_vec - start_vec),closest);
              vz = closest[1];
              zz = closest[2];
              residV[view] = vz - xyz[1];
              residZ[view] = zz - xyz[2];
            }
            Val_t hists = GetChannelHists(pIdx, cIdx);
            hists.ResidV->Fill(residV[view]);
            hists.ResidZ->Fill(residZ[view]);
            hists.HCellPos->SetEntries(1);
            //loop over 2D bins of a histo
            double bs,bt,d,moda;
            for(int ibin = 1; ibin <= hists.HCellPos->GetNbinsX(); ibin++){
              for(int jbin = 1; jbin <= hists.HCellPos->GetNbinsY(); jbin++){
                bs = hists.HCellPos->GetXaxis()->GetBinCenter(ibin);
                bt = hists.HCellPos->GetYaxis()->GetBinCenter(jbin);
                
                moda = util::pythag(residV[view],residZ[view]);
                d = moda - ((residV[view]*bs + residZ[view]*bt)/moda);
                int bin = hists.HCellPos->GetBin(ibin,jbin);
                hists.HCellPos->AddBinContent(bin,util::sqr(d));
              } // jbin 
            } // ibin
	  } // if cell exists 
        } // cIdx
      } // pIdx      	 
    } // end for trkIdx
    
    
    return;    
  }
  
  //......................................................................
    
  void Alignment::reconfigure(const fhicl::ParameterSet& pset)
  {
  }  
  //......................................................................
  
  void Alignment::endJob() 
  {
  }
    
  //......................................................................

  Alignment::Val_t Alignment::GetChannelHists(int plane, int cell)
  {
    geo::OfflineChan key(plane, cell);	
    std::map<geo::OfflineChan, Val_t>::iterator it = fChannelMap.find(key);
    if(it != fChannelMap.end()) return it->second;
    art::ServiceHandle<art::TFileService> tfs;
    TH1* residv =
      tfs->make<TH1I>(TString::Format("residualmcV_%03d_%03d", plane, cell),
                             ";Residual in V (cm);count", 1000, -50, 50);
    TH1* residz =
      tfs->make<TH1I>(TString::Format("residualmcZ_%03d_%03d", plane, cell),
                             ";Residual in Z (cm);count", 1000, -50, 50);                         
    TH2* hCellPos =
      tfs->make<TH2D>(TString::Format("hCellPos_%03d_%03d", plane, cell),
                             ";Position in Z (cm); Position in Y (cm)", 40, -10, 10, 40, -10, 10);                         

    Val_t v;	
    v.ResidV    = residv;
    v.ResidZ    = residz;
    v.HCellPos  = hCellPos;
    fChannelMap[key] = v;
    return v;
  }
  
  //.....................................................................
  
  bool Alignment::IsGoodTrack(art::Ptr<rb::Track> t, double x, double y, double z)
  {        
    art::ServiceHandle<geo::Geometry> geom;
    if (t->ExtentPlane() >= 8)                         return true;
    else return false;         
    if (y > geom->DetHalfHeight()-fDetEdgeBuffer)      return true;
    if (fabs(x) > geom->DetHalfWidth()-fDetEdgeBuffer) return true;
    if (z < fDetEdgeBuffer)                            return true;        
    if (z > geom->DetLength()-fDetEdgeBuffer)          return true;
    return false;        	
  } 
} //namespace align
////////////////////////////////////////////////////////////////////////

namespace align
{
  DEFINE_ART_MODULE(Alignment);
}

