////////////////////////////////////////////////////////////////////////
//
// \file    BPFEnergyEstimator_module.cc
// \brief   Module to produce the BPF Energy object.
// \version 
// \author  Michael Baird - mbaird42@fnal.gov
//
////////////////////////////////////////////////////////////////////////

// C/C++ includes
#include <cmath>
#include <iostream>
#include <vector>

// ROOT includes
#include "TFile.h"
#include "TH2F.h"
#include "TMVA/Factory.h"
#include "TMVA/Tools.h"
#include "TMVA/Reader.h"
#include "TMVA/MethodKNN.h"

// Framework includes
#include "art/Framework/Core/EDProducer.h"
#include "art/Framework/Core/FindManyP.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Services/Optional/TFileDirectory.h"
#include "art/Framework/Services/Optional/TFileService.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "art/Persistency/Common/Ptr.h"
#include "art/Persistency/Common/PtrVector.h"
#include "fhiclcpp/ParameterSet.h"
#include "messagefacility/MessageLogger/MessageLogger.h"

// NOvASoft includes
#include "RecoBase/CellHit.h"
#include "RecoBase/RecoHit.h"
#include "RecoBase/Cluster.h"
#include "RecoBase/Prong.h"
#include "RecoBase/Track.h"
#include "RecoBase/FitSum.h"

#include "Utilities/func/EnvExpand.h"
#include "Utilities/AssociationUtil.h"
#include "Geometry/Geometry.h"
#include "Calibrator/Calibrator.h"

#include "BreakPointFitter/dEdxCalculator.h"
#include "BreakPointFitter/BPFEnergy.h"
#include "MuonRemove/art/TrackCleanUpAlg.h"



// BPFEnergyEstimator header
namespace bpfit {
  class BPFEnergyEstimator : public art::EDProducer {
  public:
    explicit BPFEnergyEstimator(fhicl::ParameterSet const& pset);    
    ~BPFEnergyEstimator();                        
    
    void produce(art::Event& evt); 
    void reconfigure(const fhicl::ParameterSet& pset);
    void beginRun(art::Run& run);
    void resetVars();

  private:
    std::string   fSlicerLabel;   ///< label for module that made the slices
    std::string   fProngLabel;    ///< label for module that made the prongs
    std::string   fProngInstance; ///< instance label for the prongs to be used
    std::string   fTrackLabel;    ///< label for module that made the tracks
    std::string   fLibPath;       ///< location of all BPF library files listed below
    std::string   fHistoFile;     ///< name of dE/dx log-likelihood histogram file
    std::string   fPIDXMLFile;    ///< name of the XML file for the muon PID
    std::string   fEnXMLFile1;    ///< name of the XML file for the first energy estimator
    std::string   fEnXMLFile2;    ///< name of the XML file for the second energy estimator
    std::string   fEnXMLFile3;    ///< name of the XML file for the third energy estimator
    double        fdxTOL;         ///< lower limit for dx values to be used in dE/dx calculation

    TH2F  *fdEdxVSbg[2][2];          ///< histo of computed dE/dx for cell hits (vs. beta*gamma)
    bpfit::dEdxCalculator fdEdxCalc; ///< helper class for computing dEdx values

    murem::TrackCleanUpAlg fTrackCleanUpAlg; ///< object to determine hadronic energy on the muon track

    bool fBookedKNNs;                ///< Have we yet booked the kNNs?

    TMVA::Reader fMuonPID;           ///< TMVA reader object for the muon PID
    TMVA::Reader fEnergyEstm1;       ///< TMVA reader object for the first energy estimator
    TMVA::Reader fEnergyEstm2;       ///< TMVA reader object for the second energy estimator
    TMVA::Reader fEnergyEstm3;       ///< TMVA reader object for the third energy estimator


    
    //
    // Variables used with the kNNs
    //

    // muon PID variables:
    float fLength;    ///< track length
    float fChi2T;     ///< total chi^2
    float fdEdxLL;    ///< dEdx LL value
    float fHitRatio;  ///< ratio of track hits to prong hits (used as a track/shower discriminator)

    // energy estimator variables:
    //   * target variables:
    float fE1;        ///< first estimator true energy
    float fEsq1;      ///< first estimator true energy squared (used to estiamte st. dev.)
    float fE2;        ///< second estimator true energy
    float fEsq2;      ///< second estimator true energy squared (used to estiamte st. dev.)
    float fE3;        ///< third estimator true energy
    float fEsq3;      ///< third estimator true energy squared (used to estiamte st. dev.)
    //   * input variables:
    float fPmu;       ///< reconstructed muon momentum from the track selected to be the muon
    float fDirZmu;    ///< muon track dirZ from the track selected to be the muon
    float fN3Dprongs; ///< total number of 3D prongs
    float fE3Dprongs; ///< sum of hit energies in 3D prongs (excluding the "muon" prong) - this does NOT account for dead material
    float fEremain;   ///< sum of remaining slice energy not in 3D prongs - this does NOT account for dead material
    float fSumPE;     ///< sum of total PE for all hits not on the muon track

  };
}



namespace bpfit
{
  //.......................................................................
  BPFEnergyEstimator::BPFEnergyEstimator(fhicl::ParameterSet const& pset) :
      fdEdxCalc(),
      fTrackCleanUpAlg(pset.get<fhicl::ParameterSet>("TrackCleanUpAlgPSet")),
      fMuonPID(),
      fEnergyEstm1(),
      fEnergyEstm2(),
      fEnergyEstm3()
  {
    produces< std::vector<BPFEnergy> >();
    produces< art::Assns<BPFEnergy, rb::Cluster> >();

    this->reconfigure(pset);

    fBookedKNNs = false;
  }

  //......................................................................
  BPFEnergyEstimator::~BPFEnergyEstimator()
  {
    //======================================================================
    // Clean up any memory allocated by your module...   ...if you dare...
    //======================================================================
  }

  //......................................................................
  void BPFEnergyEstimator::reconfigure(const fhicl::ParameterSet& pset)
  {
    fSlicerLabel   = pset.get<std::string> ("SlicerLabel");
    fProngLabel    = pset.get<std::string> ("ProngLabel");
    fProngInstance = pset.get<std::string> ("ProngInstance");
    fTrackLabel    = pset.get<std::string> ("TrackLabel");
    fLibPath       = pset.get<std::string> ("LibPath");
    fHistoFile     = pset.get<std::string> ("HistoFile");
    fPIDXMLFile    = pset.get<std::string> ("PIDXMLFile");
    fEnXMLFile1    = pset.get<std::string> ("EnXMLFile1");
    fEnXMLFile2    = pset.get<std::string> ("EnXMLFile2");
    fEnXMLFile3    = pset.get<std::string> ("EnXMLFile3");
    fdxTOL         = pset.get<double>      ("dxTOL");
  }
      
  //......................................................................
  void BPFEnergyEstimator::beginRun(art::Run& run)
  {

    // Interpret the env variable LibPath
    std::string libpath = util::EnvExpansion(fLibPath);

    //
    // Read in the log-likelihood histos from the input file
    //
    const char *filename = (libpath+fHistoFile).c_str();
    TFile infile(filename);

    fdEdxVSbg[0][0] = (TH2F*)infile.FindObjectAny("fdEdxVSbg_W0_X0");
    fdEdxVSbg[0][1] = (TH2F*)infile.FindObjectAny("fdEdxVSbg_W0_X1");
    fdEdxVSbg[1][0] = (TH2F*)infile.FindObjectAny("fdEdxVSbg_W1_X0");
    fdEdxVSbg[1][1] = (TH2F*)infile.FindObjectAny("fdEdxVSbg_W1_X1");
    
    infile.Close();



    if(!fBookedKNNs) {
      //
      // Set up the TMVA reader for the muon PID.
      //
      fMuonPID.AddVariable("length",&fLength);
      fMuonPID.AddVariable("chi2T",&fChi2T);
      fMuonPID.AddVariable("dEdxLL",&fdEdxLL);
      fMuonPID.AddVariable("hitRatio",&fHitRatio);
      fMuonPID.BookMVA("KNN Method",libpath+fPIDXMLFile);
      
      //
      // Set up the TMVA reader for the energy estimators.
      //
      fEnergyEstm1.AddVariable("Pmu",&fPmu);
      fEnergyEstm1.AddVariable("DirZmu",&fDirZmu);
      fEnergyEstm1.AddVariable("N3Dprongs",&fN3Dprongs);
      fEnergyEstm1.AddVariable("E3Dprongs",&fE3Dprongs);
      fEnergyEstm1.AddVariable("Eremain",&fEremain);
      fEnergyEstm1.AddVariable("SumPE",&fSumPE);
      fEnergyEstm1.BookMVA("KNN Method",libpath+fEnXMLFile1);
      
      fEnergyEstm2.AddVariable("Pmu",&fPmu);
      fEnergyEstm2.AddVariable("DirZmu",&fDirZmu);
      fEnergyEstm2.AddVariable("N3Dprongs",&fN3Dprongs);
      fEnergyEstm2.AddVariable("E3Dprongs",&fE3Dprongs);
      fEnergyEstm2.AddVariable("Eremain",&fEremain);
      fEnergyEstm2.AddVariable("SumPE",&fSumPE);
      fEnergyEstm2.BookMVA("KNN Method",libpath+fEnXMLFile2);
      
      fEnergyEstm3.AddVariable("Pmu",&fPmu);
      fEnergyEstm3.AddVariable("DirZmu",&fDirZmu);
      fEnergyEstm3.AddVariable("N3Dprongs",&fN3Dprongs);
      fEnergyEstm3.AddVariable("E3Dprongs",&fE3Dprongs);
      fEnergyEstm3.AddVariable("Eremain",&fEremain);
      fEnergyEstm3.AddVariable("SumPE",&fSumPE);
      fEnergyEstm3.BookMVA("KNN Method",libpath+fEnXMLFile3);
      
      fBookedKNNs = true;
    }

  }

  //......................................................................
  void BPFEnergyEstimator::resetVars()
  {
    //
    // PID variables:
    //
    fLength      = -5.0;
    fChi2T       = 1.0e9;
    fdEdxLL      = -1.0e9;
    fHitRatio    = -5.0;

    //
    // energy estimator variables:
    //
    fE1        = -5.0;
    fEsq1      = -5.0;
    fE2        = -5.0;
    fEsq2      = -5.0;
    fE3        = -5.0;
    fEsq3      = -5.0;
    fPmu       = -5.0;
    fDirZmu    = -5.0;
    fN3Dprongs = -5.0;
    fE3Dprongs = -5.0;
    fEremain   = -5.0;
    fSumPE     = -5.0;
  }

  //......................................................................
  void BPFEnergyEstimator::produce(art::Event& evt)
  {

    // Get the slices and put them into a more convenient container.
    art::Handle< std::vector< rb::Cluster > > slhandle;
    evt.getByLabel(fSlicerLabel, slhandle);
    art::PtrVector<rb::Cluster> slices;
    for(unsigned int i=0; i < slhandle->size(); ++i) {
      slices.push_back(art::Ptr<rb::Cluster>(slhandle,i));
    }

    // Create the FindMany object for getting prongs.
    art::FindManyP<rb::Prong> prongFMP(slhandle, evt, art::InputTag(fProngLabel,fProngInstance));

    // Make the required service handles...
    art::ServiceHandle<geo::Geometry> geom;
    art::ServiceHandle<calib::Calibrator> cal;

    // Make the containers for the BPFEnergy objects and their associations
    std::unique_ptr< std::vector<BPFEnergy> > 
      energies(new std::vector<BPFEnergy>);
    std::unique_ptr< art::Assns<BPFEnergy,rb::Cluster> > 
      EnSlAssns(new art::Assns<BPFEnergy,rb::Cluster>);

    // loop over all slices and get the fuzzyk 3D prongs
    for(unsigned int islice = 0; islice < slhandle->size(); ++islice) {

      // As usual, skip the noise slice...
      if((*slhandle)[islice].IsNoise()) continue;
      
      // NOTE: Currently, the fundamental object of analysis is the slice.
      // If in the future we change things so that the fundamental object
      // of analysis becomes the vertex, then we will have to get prongs
      // associated with the vertex instead.

      // Reset all variables.
      this->resetVars();

      // get the 3D prongs associated with this slice
      std::vector< art::Ptr< rb::Prong > > prongs;
      if(prongFMP.isValid()) prongs = prongFMP.at(islice);

      // create the FindMany object for getting tracks
      art::FindManyP<rb::Track> trackFMP(prongs, evt, fTrackLabel);



      // Variables used to keep track with the best muon PID.
      float bestPID   = -5.0; // best muonPID value
      int   prongPick = -1;   // Prong picked as having the longest track.



      //////////
      //
      // Step 1:  Determine which prong is the most muon-like.
      //
      //////////

      // loop over all prongs and get the tracks
      for(unsigned int iprong = 0; iprong < prongs.size(); ++iprong) {

	// get the tracks associated with this prong
	std::vector< art::Ptr< rb::Track > > tracks;
	if(trackFMP.isValid()) tracks = trackFMP.at(iprong);

	// create the FindMany object for getting FitSum objects
	art::FindManyP<rb::FitSum> fitsumFMP(tracks, evt, fTrackLabel);

	// Loop over all tracks...
	for(unsigned int itrack = 0; itrack < tracks.size(); ++itrack) {

	  // get the FitSum object associated with this track
	  std::vector< art::Ptr< rb::FitSum > > fitsums;
	  if(fitsumFMP.isValid()) fitsums = fitsumFMP.at(itrack);

	  // There can be only one!!!
	  // (something went really wrong with BPF if there's not one and
	  // only one of these per track.)
	  if(fitsums.size() != 1) abort();

	  // For now, this energy estimator only uses information from the muon track fit.
	  if(fitsums[0]->PDG() != 13) continue;

	  float LLtemp = 0.0;

	  // fill in chi2 and NDOF values
	  float tempH = 0.0;
	  float tempA = 0.0;
	  float tempN = 0.0;
	  tempH = fitsums[0]->Chi2(0)+fitsums[0]->Chi2(1);   // chi2 from the hits
	  tempA = fitsums[0]->Chi2(2)+fitsums[0]->Chi2(3);   // chi2 from the angles
	  tempN = fitsums[0]->Ndof(0)+fitsums[0]->Ndof(1)-4; // NDOF (NOTE: currently 4 is NOT subtracted from NDOF when the fitsum object is made by BPF)

	  if(tempN <= 0) continue;

	  // Keep track of the relevant info...
	  fChi2T  = tempH/tempN + tempA/tempN;
	  fLength = tracks[itrack]->TotalLength();

	  //
	  // Calculate dE/dx for each cell the track passes through using
	  // the helper class.
	  //
	  fdEdxCalc.computeDEDX(tracks[itrack],fitsums[0]->PDG());
	  
	  std::vector<double> dE;
	  std::vector<double> dx;
	  std::vector<double> W;
	  std::vector<double> BG;

	  fdEdxCalc.getResults(dE,dx,W,BG,fdxTOL);

	  //
	  // Compute the de/dx likelihood values for all cells.
	  //
	  for(unsigned int a = 0; a < dE.size(); ++a) {

	    unsigned int w = 0;
	    unsigned int x = 0;

	    if(W[a]  > 0.0) w = 1;
	    if(dx[a] > 8.0) x = 1;

	    // For the NearDet, we will just the "close to the readout end" FarDet histos.
	    if(geom->DetId() == novadaq::cnv::kNEARDET) w = 1;

	    double content = fdEdxVSbg[w][x]->GetBinContent(fdEdxVSbg[w][x]->FindBin(log(BG[a]),dE[a]/dx[a]));
	    // We can't take the log of zero, so put a minimum on it.
	    if(content < 1.0e-9) content = 1.0e-9;
	    LLtemp += log(content);

	  }

	  if(dE.size() > 0) {
	    // Keep track of the relevant info...
	    fdEdxLL = LLtemp/(float)dE.size();
	    // NOTE: For this ratio, the number of hits on the track is taken to be the
	    //       number of hits for which a value of dE/dx was successfully calculated.
	    fHitRatio = (float)dE.size()/(float)prongs[iprong]->NCell();
	  }
	  // end computing dE/dx LL info...
	  


	  // Set the variables for the track with the best muonPID
	    if(fLength > 0.0 && fdEdxLL > -1.0e9 && fChi2T < 1.0e9) {
	      // Get the PID value...
	      float pid = fMuonPID.EvaluateMVA("KNN Method");
	      if(pid > bestPID) {
		bestPID   = pid;
		prongPick = (int)iprong;
		fPmu      = fitsums[0]->FourMom().P();
		fDirZmu   = tracks[itrack]->Dir().Z();
	      }
	    }
	  
	} // end loop over tracks (itrack)

      } // end loop over prongs (iprong)



      //////////
      //
      // Step 2:  Compute the values to be used for the energy estimator.
      //
      //////////

      //
      // Add up the energy of the non-muon 3D prongs.
      //

      float N3Dpr = 0.0;
      float E3Dpr = 0.0;
      float sumPE = 0.0;
      rb::Cluster  hits3Dprongs; // a container for all hits on 3D prongs

      for(unsigned int p = 0; p < prongs.size(); ++p) {
	
	// Skip 2D prongs.
	if(prongs[p]->Is2D()) continue;
	N3Dpr++;

	// Keep track of all hits in 3D prongs (to be used later when adding up
	// the remaining slice energy.)
	// NOTE: FuzzyK prongs can sometimes share hits, so we have to be careful
	//       not to double count.
	for(unsigned int h = 0; h < prongs[p]->NCell(); ++h) {

	  const rb::CellHit *pronghit = prongs[p]->Cell(h).get();

	  // Check to see if this hit is already in the list.
	  bool onlist = false;
	  for(unsigned int hpr = 0; hpr < hits3Dprongs.NCell(); ++hpr) {
	    const rb::CellHit *pronghit3D = hits3Dprongs.Cell(hpr).get();
	    if((*pronghit) == (*pronghit3D)) {
	      onlist = true;
	      break;
	    }
	  } // end loop over hits in 3D prongs (hpr)

	  // Add the hit if its not already on the list.
	  if(!onlist) hits3Dprongs.Add(prongs[p]->Cell(h));

	} // end loop over hits (h)

	// Skip this prong if it is the identified muon prong.
	if((int)p == prongPick) continue;

	// Loop over all hits in this 3D, non-muon prong and add up their energies.
	//
	// NOTE: For the hit W position, I am currently using the prong W instead of
	//       the track W. Since I haven't determined which of the three BPF fits
	//       is the best one for this prong, this saves me from having to try
	//       to guess at which track to use. My assumption is that prong->W will
	//       be good enough since I am lumping all hits from 3D prongs together
	//       into one 3D prong energy number...
	//
	// WARNING: Since hits can be shared by fuzzyK prongs, there is the potential
	//          below to double (or triple) count energy on shared cell hits. For
	//          now, I am assuming that this is a small effect but something
	//          smarter could/should be put into place in the future.
	for(unsigned int h = 0; h < prongs[p]->NCell(); ++h) {
	  
	  const rb::CellHit *chit = prongs[p]->Cell(h).get();
	  float W = prongs[p]->W(chit);
	  rb::RecoHit rhit(cal->MakeRecoHit(*chit,W));

	  sumPE += chit->PE();
	  if(rhit.IsCalibrated()) E3Dpr += rhit.GeV();

	} // end loop over hits (h)

      } // end loop over prongs (p)

      // Set the prong variables.
      if(bestPID >= 0.0) {
	fN3Dprongs = N3Dpr;
	fE3Dprongs = E3Dpr;
      }



      //
      // Add up the remaining energy in the slice.
      //

      rb::Cluster hitsRemain; // a container for all hits NOT on 3D prongs

      // Loop over all hits in this slice.
      for(unsigned int hsl = 0; hsl < (*slhandle)[islice].NCell(); ++hsl) {

	const rb::CellHit *slicehit = (*slhandle)[islice].Cell(hsl).get();

	// Check to see if this hit is in the list of hits on 3D prongs.
	bool onlist = false;
	for(unsigned int hpr = 0; hpr < hits3Dprongs.NCell(); ++hpr) {
	  const rb::CellHit *pronghit = hits3Dprongs.Cell(hpr).get();
	  if((*pronghit) == (*slicehit)) {
	    onlist = true;
	    break;
	  }
	} // end loop over hits in 3D prongs (hpr)

	// If this hit is not in a 3D prong, add it to the list of remaining slice hits.
	if(!onlist) {
	  hitsRemain.Add((*slhandle)[islice].Cell(hsl));
	  sumPE += slicehit->PE();
	}

      } // end loop over hits in this slice (hsl)

      // Set the remaining slice energy variable.
      if(bestPID >= 0.0) {
	if(hitsRemain.NCell() > 0) fEremain = hitsRemain.TotalGeV();
	else                       fEremain = 0.0;
      }

      if(bestPID >= 0.0) {
	fSumPE = sumPE;
      }



      //
      // Compute the hadronic contamination on the muon track and add that back in.
      //
      float muonTrackHadE = 0.0;

      // Only do this if we actually picked a track to be the muon.
      if(prongPick >= 0) {

	std::vector< art::Ptr< rb::Track > > tracks;
	if(trackFMP.isValid()) tracks = trackFMP.at(prongPick);
	art::FindManyP<rb::FitSum> fitsumFMP(tracks, evt, fTrackLabel);

	// Figure out which BPF track was made with the muon assumption.
	for(unsigned int itrack = 0; itrack < tracks.size(); ++itrack) {

	  std::vector< art::Ptr< rb::FitSum > > fitsums;
	  if(fitsumFMP.isValid()) fitsums = fitsumFMP.at(itrack);
	  if(fitsums.size() != 1) abort();

	  // Skip if not the muon track...
	  if(fitsums[0]->PDG() != 13) continue;

	  // If it is the muon track, do the extra Ehad calculation
	  muonTrackHadE = fTrackCleanUpAlg.ExtraEOnTrack(*tracks[itrack],*slices[islice]);

	} // end loop over tracks (itrack)

      } // end if(prongPick >= 0)

      fEremain += muonTrackHadE;



      //
      // Set the BPFEnergy object values and make the associations.
      //
      float Eres1 = -5.0;
      float Eres2 = -5.0;
      float Eres3 = -5.0;
      if(bestPID >= 0.0) {
	fE1   = (fEnergyEstm1.EvaluateRegression("KNN Method"))[0];
	fEsq1 = (fEnergyEstm1.EvaluateRegression("KNN Method"))[1];
	Eres1 = sqrt(fabs(fEsq1 - fE1*fE1));

	fE2   = (fEnergyEstm2.EvaluateRegression("KNN Method"))[0];
	fEsq2 = (fEnergyEstm2.EvaluateRegression("KNN Method"))[1];
	Eres2 = sqrt(fabs(fEsq2 - fE2*fE2));

	fE3   = (fEnergyEstm3.EvaluateRegression("KNN Method"))[0];
	fEsq3 = (fEnergyEstm3.EvaluateRegression("KNN Method"))[1];
	Eres3 = sqrt(fabs(fEsq3 - fE3*fE3));
      }

      BPFEnergy energy(fE1,Eres1,fE2,Eres2,fE3,Eres3,
		       bestPID,fPmu,fDirZmu,
		       fN3Dprongs,fE3Dprongs,fEremain,fSumPE);      
      energies->push_back(energy);
      util::CreateAssn(*this,evt,*(energies.get()),slices[islice],*(EnSlAssns.get()));

    } // end loop over slhandle (islice)
    
    evt.put(std::move(energies));
    evt.put(std::move(EnSlAssns));

  }

} // end namespace bpfit
////////////////////////////////////////////////////////////////////////



#include "art/Framework/Core/ModuleMacros.h"

namespace bpfit
{
  DEFINE_ART_MODULE(BPFEnergyEstimator);
}
