//
/// \file      HitList3D.h
/// \brief     Contain 3D hit information by extrapolating ortho view 
///            coordinate from surrounding points in the ortho view
/// \version   $Id: HitList3D.h,v 1.0 2014-07-30 9:49:00 merritth $
/// \author    hdmerritt@indiana.edu
///
#ifndef BPFIT_HITLIST3D_H
#define BPGIT_HITLIST3D_H
#include <vector>
#include "GeometryObjects/PlaneGeo.h"
namespace rb { class Cluster; }

namespace bpfit 
{
  class Hit3D
  {
  public:
    // Constructor
    Hit3D(geo::View_t v = geo::kXorY,
	  double x      = 0,
	  double y      = 0,
	  double z      = 0,
	  double t      = 0,
	  double sig_x  = 0,
	  double sig_y  = 0,
	  double sig_z  = 0,
	  double sig_t  = 0,
	  double w      = 1) :
      fView(v), fX(x), fY(y), fZ(z), fT(t), 
      fSigX(sig_x), fSigY(sig_y), fSigZ(sig_z), fSigT(sig_t), fW(w)
      { }
  public:
    geo::View_t fView; ///< View hit belongs in (x or y)
    double      fX;    ///< X coordinate of hit
    double      fY;    ///< Y coordinate of hit
    double      fZ;    ///< Z coordinate of hit
    double      fT;    ///< time coordinate of hit
    double      fSigX; ///< Error on x
    double      fSigY; ///< Error on y
    double      fSigZ; ///< Error on z
    double      fSigT; ///< Error on t
    double      fW;    ///< Weight of hit
  };
  
  ///
  /// Forming the 3D hit list requires completeing information for a
  /// hit using the other view. We have one way of doing that, but
  /// we've left a hook for others to invent their own way.
  ///
  /// @param h - Hit that needs information from the orthogonal view
  /// @param hl - The complete track hit list
  ///
  class HitList3D;
  typedef void (*Interp3D)(Hit3D& h, HitList3D& hl);
    
  /// Essential 3D information for tracking
  class HitList3D : public std::vector<Hit3D>
  {   
  public:
    HitList3D();
    HitList3D(unsigned int n);
    HitList3D(const rb::Cluster& clust, bool wbyq=false);
    void SetHit(unsigned int i,
		geo::View_t, 
		double x, 
		double y,
		double z,
		double t,
		double sig_x,
		double sig_y,
		double sig_z,
		double sig_t,
		double w=1.0); 
    void AddHit(geo::View_t, 
		double x, 
		double y,
		double z,
		double t,
		double sig_x,
		double sig_y,
		double sig_z,
		double sig_t,
		double w=1.0);
    ///
    /// CompleteOrtho extrapolates the coordinate in the orthogonal
    /// view corresponding to the same z value in the current view.
    ///
    /// @param interp - optional alternate interpolation function
    ///
    void CompleteOrtho(Interp3D interp=0);
    
    ///
    /// Sort the hits from south to north
    ///
    void SortByZ();

    ///
    /// Sort the hits from top to bottom
    ///
    void SortByY();
  };
}

#endif // BPGIT_HITLIST3D_H
////////////////////////////////////////////////////////////////////////

