///
/// \file Path.cxx 
/// \brief Information about materials along a path in the detector
///
#include "BreakPointFitter/func/Path.h"
#include <algorithm>
#include <cmath>
#include "TGeoMaterial.h"
#include "GeometryObjects/GeometryBase.h"
#include "BreakPointFitter/func/dEdxTable.h"
#include "BreakPointFitter/func/BPException.h"
using namespace bpfit;

static const double gsMmuon   = 105.658E-3;
static const double gsMpion   = 139.570E-3;
static const double gsMproton = 938.272E-3;

//......................................................................

Path::Path(const geo::GeometryBase& geo,
	   const double*            p0,
	   const double*            p1,
	   double                   s0) :
  fGeo(geo),
  fXYZ(2),
  fS0(s0),
  fTablesOK(false)
{
  this->SetPoint(0,p0);
  this->SetPoint(1,p1);
}

//......................................................................

Path::Path(const geo::GeometryBase& geo, 
	   unsigned int               n,
	   double                    s0) :
  fGeo(geo),
  fXYZ(n),
  fS0(s0),
  fTablesOK(false)
{ }

//......................................................................

Path::Path(const geo::GeometryBase& geo, 
	   std::vector<TVector3> const& trajectory,
	   double                    s0) :
  fGeo(geo),
  fXYZ(trajectory),
  fS0(s0),
  fTablesOK(false)
{ 
}

//......................................................................    

void Path::SetPoint(unsigned int i, const double* xyz) 
{
  fXYZ[i].SetXYZ(xyz[0],xyz[1],xyz[2]);
  fTablesOK = false;
}

//......................................................................

unsigned int Path::ScrubOne(double eps) 
{
  if (fXYZ.size()<3) return 0;

  std::vector<TVector3>::iterator p0(fXYZ.begin());
  std::vector<TVector3>::iterator p1(p0+1);
  std::vector<TVector3>::iterator pEnd(fXYZ.end());
  
  TVector3 d;
  for (; p1!=pEnd; ++p0, ++p1) {
    d.SetXYZ(p0->X()-p1->X(), p0->Y()-p1->Y(), p0->Z()-p1->Z());
    if (d.Mag2()<eps) {
      fXYZ.erase(p1);
      fTablesOK = false;
      return 1;
    }
  }
  return 0;
}

//......................................................................

unsigned int Path::Scrub(double eps) 
{
  unsigned int n;
  unsigned int ntot=0;
  for (;;) {
    n = this->ScrubOne(eps);
    if (n==0) return ntot;
    ntot += n;
  }
  return ntot;
}

//......................................................................

double Path::Length() const 
{
  if (!fTablesOK) this->Integrate();
  return fS[fS.size()-1];
}

//......................................................................

double Path::X0(double s) const 
{
  if (!fTablesOK) this->Integrate();
  return this->Interp(s-fS0, fS, fX0);
}

//......................................................................

TVector3 Path::Dir(double s)
{
  if (!fTablesOK) this->Integrate();
  
  s -= fS0;
  TVector3 d(0,0,1);
  for (unsigned int i=0; i+1<fS.size(); ++i) {
    if (fS[i]<=s && s<=fS[i+1]) {
      d.SetXYZ(fXYZ[i+1].X()-fXYZ[i].X(),
	       fXYZ[i+1].Y()-fXYZ[i].Y(),
	       fXYZ[i+1].Z()-fXYZ[i].Z());
      if(d.Mag() == 0.0) {
	// Something terrible has happened if the direction vector is
	// zero. Throw an exception to prevent EVERYTHING from crashing
	// down...
	throw BPException(__FILE__, __LINE__, kBAD_TRACK, 0);
      }
      d.SetMag(1.0);
      return d;
    }
  }
  return d;
}

//......................................................................
///
/// Given kinetic energy (GeV) and mass (GeV) compute other parameters
/// relavent for relativistic kinematics
///
void Path::TPBG(double  t,
		double  m,
		double* T,
		double* p,
		double* beta,
		double* gamma)
{
  //
  // Integration sometimes overshoots 0 -- clamp it here to 0.0
  //
  static double eps = 1.0E-5; // eps is allowable error
  if (t<0.0) {
    if (t>-eps) t = 0.0;
    else
      throw BPException(__FILE__, __LINE__, kBAD_KE, t);
  }
  double E = t+m;
  double g = m/E;
  double P = E*sqrt(1.0-g*g);
  
  if (T)     *T = t;
  if (p)     *p = P;
  if (beta)  *beta  = P/E;
  if (gamma) *gamma = E/m;
}

//......................................................................

void Path::MuonParams(double s,
		      double* T, 
		      double* p,
		      double* beta,
		      double* gamma) const
{
  if (!fTablesOK) this->Integrate();
  double t = this->Interp(s-fS0, fS, fMuonKE);
  this->TPBG(t, gsMmuon, T, p, beta, gamma);
}

//......................................................................

void Path::PionParams(double s, 
		      double* T, 
		      double* p,
		      double* beta,
		      double* gamma) const
{
  if (!fTablesOK) this->Integrate();
  double t = this->Interp(s-fS0, fS, fPionKE);
  this->TPBG(t, gsMpion, T, p, beta, gamma);
}

//......................................................................

void Path::ProtonParams(double s, 
			double* T, 
			double* p,
			double* beta,
			double* gamma) const 
{
  if (!fTablesOK) this->Integrate();
  double t = this->Interp(s-fS0, fS, fProtonKE);
  this->TPBG(t, gsMproton, T, p, beta, gamma);
}

//......................................................................
///
/// Integrate the dT = int[ dEdx(T), dx=0...ds ]
///
/// \param ds  - step size [cm]
/// \param mat - material name
/// \param m   - mass of particle [GeV]
/// \param Tin - kinetic energy at start of step [GeV]
///
double Path::IntegrateT(double ds, const char* mat, double m, double Tin) const
{
  unsigned int i;
  //
  // Control the step size
  //
  unsigned int nstep = 1;
  double dx = ds;
  static const double dxmax = 10.0;
  while (dx>dxmax) {
    nstep *= 2;
    dx = ds/(float)nstep;
  }
  
  //
  // Take the steps leap-frog style
  //
  dEdxTable dedx(mat);
  double T    = Tin;
  double Tmid = 0.0;
  for (i=0; i<nstep; ++i) {
    Tmid = T + 0.5*dx*dedx(T,m);
    T += dx*dedx(Tmid,m);
  }

  return T;
}

//......................................................................
///
/// Integrate along one straight-line segment of a path
///
/// \param p1      - start x/y/z point [cm]
/// \param p2      - end   x/y/z point [cm]
/// \param stot    - path length, updated on return [cm]
/// \param x0      - path length in units of radiation lengths, updated on return.
/// \param Tmuon   - muon KE, updated on return [GeV]
/// \param Tpion   - pion KE, updated on return [GeV]
/// \param Tproton - proton KE, updated on return [GeV]
///
void Path::IntegrateLeg(const double* p1,
			const double* p2,
			double* stot,
			double* x0,
			double* Tmuon,
			double* Tpion,
			double* Tproton) const
{
  unsigned int i;
  std::vector<double>              ds;
  std::vector<const TGeoMaterial*> mat;
  fGeo.MaterialsBetweenPoints(p1, p2, ds, mat);

  for (i=0; i<ds.size(); ++i) {
    *stot += ds[i];

    double rlen = mat[i]->GetRadLen();

    // Something changed somewhere in ??? between 14-09-29 and 14-10-08 and now
    // TGeoMaterial believes that the radiation length in Vacuum is zero.  So for
    // now I'm going to intercept and correct this number.  I hope that a commit will
    // follow this sometime later removing this hard coded case.
    if(mat[i]->GetName() == std::string("Vacuum")) rlen = 1.0e30;

    if(rlen <= 0.0)
      throw BPException(__FILE__, __LINE__, kBAD_RADL, rlen);

    *x0 += ds[i]/rlen;
    *Tmuon   = this->IntegrateT(ds[i], mat[i]->GetName(), gsMmuon,   *Tmuon);
    *Tpion   = this->IntegrateT(ds[i], mat[i]->GetName(), gsMpion,   *Tpion);
    *Tproton = this->IntegrateT(ds[i], mat[i]->GetName(), gsMproton, *Tproton);

    fS.       push_back(*stot);
    fX0.      push_back(*x0);
    fMuonKE.  push_back(*Tmuon);
    fPionKE.  push_back(*Tpion);
    fProtonKE.push_back(*Tproton);
  }
}

//......................................................................
///
/// Update all the look-up tables
///
void Path::Integrate() const
{
  fS.       clear();
  fX0.      clear();
  fMuonKE.  clear();
  fPionKE.  clear();
  fProtonKE.clear();
 
  //
  // Initialize path data. I've arbitrarily decided that a particle
  // stops when its kinetic energy is 1% of its rest mass.
  //
  double stot = 0.0;
  double x0   = 0.0;
  double Tmu  = 0.01*gsMmuon;
  double Tpi  = 0.01*gsMpion;
  double Tp   = 0.01*gsMproton;
  fS.       push_back(stot);
  fX0.      push_back(x0);
  fMuonKE.  push_back(Tmu);
  fPionKE.  push_back(Tpi);
  fProtonKE.push_back(Tp);

  //
  // Walk the path backwards since we are integrating up the energy
  // loss curve. Reverse the tables when done.
  //
  unsigned int i1 = fXYZ.size()-1;
  unsigned int i2 = i1-1;
  double p1[3], p2[3];

  for (; i1!=0; --i1, --i2) {
    p1[0] = fXYZ[i1].X(); p1[1] = fXYZ[i1].Y(); p1[2] = fXYZ[i1].Z();
    p2[0] = fXYZ[i2].X(); p2[1] = fXYZ[i2].Y(); p2[2] = fXYZ[i2].Z();
 
    this->IntegrateLeg(p1, p2, &stot, &x0, &Tmu, &Tpi, &Tp);
  }

  //
  // Put all the tables in time order
  //
  unsigned int i;
  std::reverse(fS.begin(), fS.end());
  for (i=0; i<fS.size(); ++i) fS[i] = stot-fS[i];

  std::reverse(fX0.begin(), fX0.end());
  for (i=0; i<fX0.size(); ++i) fX0[i] = x0-fX0[i];

  std::reverse(fMuonKE.  begin(), fMuonKE.  end());
  std::reverse(fPionKE.  begin(), fPionKE.  end());
  std::reverse(fProtonKE.begin(), fProtonKE.end());

  fTablesOK = true;
}

//......................................................................

double Path::Interp(double x, 
		    const std::vector<double>& X,
		    const std::vector<double>& Y)
{
  unsigned int i1=0, i2=1;
  if      (x< X[0])          { i1 = 0;          i2 = 1; }
  else if (x>=X[X.size()-1]) { i1 = X.size()-2; i2 = X.size()-1; }
  else {
    for (;;++i1,++i2) {
      if (X[i1]<=x && x<X[i2]) break;
    }
  }
  double rtn = Y[i1]+(Y[i2]-Y[i1])*(x-X[i1])/(X[i2]-X[i1]);
 
  //
  // Trap illegal values here. Tables are all kinetic energies so
  // nan's and negative values are illegal. Note that tracking errors
  // of ~1 cell will result is uncertainties of up to about 12 MeV, so
  // I've set the tolerance (eps) to 50 MeV to be safe.
  //
  static const double eps = 50.0E-3;
  if (rtn<0.0) {
    if (rtn>-eps) rtn = 0.0;
    else throw BPException(__FILE__, __LINE__, kBAD_KE, rtn);
  }

  return rtn;
}

////////////////////////////////////////////////////////////////////////
