///
/// \file   ScatteringSurfaces.cxx
/// \brief  Construct a set of scattering surfaces for Lutz.
/// \author messier@indiana.edu
///
#include "BreakPointFitter/func/ScatteringSurfaces.h"
#include <iostream>
#include <cmath>
#include "BreakPointFitter/func/Path.h"
#include "BreakPointFitter/func/BPException.h"
using namespace bpfit;

ScatteringSurfaces::ScatteringSurfaces(int PDG,
				       double dstol,
				       double dxtol,
				       double dttol) :
  fPDG(abs(PDG)),
  fdStol(dstol),
  fdXtol(dxtol),
  fdTtol(dttol)
{ }

//......................................................................

void ScatteringSurfaces::MakeSurfaces(const geo::GeometryBase& geo,
				      const double* p0,
				      const double* p1,
				      const std::vector<double>& s)
{
  fS.       clear();
  fSigAlpha.clear();

  Path path(geo, p0, p1, s[0]);
  this->MakeSteps(path, s);

  if (fS.size()==0) this->ForceOnePlane(path, s);
}

//......................................................................
///
/// \brief Make one scattering surface between first and last
/// measurements
///
/// \param path - The track data
/// \param s    - Measurement plane locations
///
void ScatteringSurfaces::ForceOnePlane(const Path& path,
				       const std::vector<double>& s) 
{
  unsigned int i1=0;
  unsigned int i2=s.size()-1;
  double ds, smid, dx, psi, dt;
  this->GetStepData(path, s, i1, i2, &ds, &smid, &dx, &psi, &dt);
  fS.       push_back(smid);
  fSigAlpha.push_back(psi);
}

//......................................................................
///
/// \brief Check the step data to see if its within tolerance
///
/// \param ds - step size in cm
/// \param dx - step size in radiation lengths
/// \param dt - RMS transverse scattering distance in cm
/// \return true=good step, false=step out of tolerance.
///
bool ScatteringSurfaces::GoodStep(double ds, double dx, double dt) 
{
  if (ds>fdStol) return false;
  if (dx>fdXtol) return false;
  if (dt>fdTtol) return false;
  return true;
}

//......................................................................
///
/// \brief Round off near track end can cause the momentum to go to
/// exactly zero or even a little negative. Protect against that.
/// 
/// beta - Input beta value. Possibly modified on return
/// p    - Input momentum value. Possibly modified on return
///
void ScatteringSurfaces::CheckLimits(double* beta, double* p)
{
  static const double betatol = 1.0E-6;
  static const double ptol    = 1.0E-5;
  if (*p<=ptol) {
    if (fabs(*p)<ptol) *p = ptol;
    else
      throw BPException(__FILE__, __LINE__, kBAD_P, *p);
  }
  if (*beta<betatol) {
    if (fabs(*beta)<betatol) *beta = betatol;
    else
      throw BPException(__FILE__, __LINE__, kBAD_P, *beta);
  }
}

//......................................................................
///
/// \brief Get relevant data about one interval between measurements
///
/// \param path - The track parameters
/// \param s    - Locations of measurement planes (cm)
/// \param i1   - step starts at s[i1]
/// \param i2   - step ends at s[i2]
/// \param ds   - on return, path length between i2 and i2 in cm
/// \param dx   - on return, ds in units of radiation lengths
/// \param psi  - RMS in-plane scattering angle (radians)
/// \param dt   - RMS in-place scattering distance (cm)
///
void ScatteringSurfaces::GetStepData(const Path& path,
				     const std::vector<double>& s,
				     unsigned int i1,
				     unsigned int i2,
				     double* ds,
				     double* smid,
				     double* dx,
				     double* psi,
				     double* dt)
{
  static const double psi0 = 13.6E-3/sqrt(3.0);
  double s1 = s[i1];
  double s2 = s[i2];

  *smid = 0.5*(s1+s2);
  
  double beta=0;
  double p   =0;
  if      (fPDG==13)   path.MuonParams  (*smid, 0, &p, &beta, 0);
  else if (fPDG==211)  path.PionParams  (*smid, 0, &p, &beta, 0);
  else if (fPDG==2212) path.ProtonParams(*smid, 0, &p, &beta, 0);
  else {
    std::cout << "\n\n\nERROR: BreakPointFitter is unable to fit under the assumption PDG = "
	      << fPDG
	      << ".  Aborting...\n\n\n";
    abort();
  }
  this->CheckLimits(&beta, &p);
  
  *ds  = s2-s1;
  *dx  = path.X0(s2)-path.X0(s1);

  if (*dx>1.0E-6) *psi = (psi0/beta/p)*sqrt(*dx)*(1.0+0.038*log(*dx));
  else{         
    *psi = 1.0E-6;
  }
  *dt = (*psi)*(*ds);

  if(*psi < 0.0){
    std::cout<<"\n beta: "<<beta
	     <<"\n    p: "<<p
	     <<"\n   dx: "<<*dx
	     <<std::endl;
  }

}

//......................................................................
///
/// \brief Make the largest step that is within tolerance
///
/// \param path - Track parameters
/// \param s    - Measurement plane locations
/// \param i1   - step starts at s[i1]
/// \return i2 - end point of step if s[i2]
///
unsigned int ScatteringSurfaces::MakeStep(const Path& path,
					  const std::vector<double>& s,
					  unsigned int i1) 
{
  unsigned int i2; // Index of end of step
  double ds;       // Step size in cm
  double smid;     // Midpoint of step
  double psi;      // RMS in-plane scattering angle
  double dx;       // Step size in radiation lengths
  double dt;       // Predicted RMS transverse scatter for step
  //
  // Increase i2 until the step data goes out of tolerance
  //
  for (i2=i1+1;i2+1<s.size();++i2) {
    this->GetStepData(path, s, i1, i2, &ds, &smid, &dx, &psi, &dt);
    if (this->GoodStep(ds, dx, dt)==false) {
      //
      // Last step took us out of tolerance, back up if we can
      //
      if (i2-i1>1) --i2;
      break;
    }
  }
  return i2;
}

//......................................................................
///
/// \brief Find good locations for the scattering surfaces
///  
/// \param path - Track parameters
/// \param s    - Measurement plane locations (cm)
///
void ScatteringSurfaces::MakeSteps(const Path& path,
				   const std::vector<double>& s)
{
  unsigned int i1;
  unsigned int i2;
  double ds;   // Step size in cm
  double smid; // Path length at middle of step
  double dx;   // Step size in radiation lengths
  double psi;  // RMS in-plane scattering angle
  double dt;   // RMS transvers scattering at end of step in cm
  for (i1=0;(i1+1)<s.size();i1=i2) {
    i2 = this->MakeStep(path, s, i1);
    this->GetStepData(path, s, i1, i2, &ds, &smid, &dx, &psi, &dt);
    if(ds==0) continue;

    fS.       push_back(smid);
    fSigAlpha.push_back(psi);
    
  }
}

//......................................................................

unsigned int ScatteringSurfaces::N() const { return fS.size(); }

//......................................................................

double ScatteringSurfaces::S(unsigned int i) const { return fS[i]; }

//......................................................................

double ScatteringSurfaces::SigAlpha(unsigned int i) const
{
  return fSigAlpha[i];
}

////////////////////////////////////////////////////////////////////////
