///
/// \file   ScatteringSurfaces.h
/// \brief  Construct a set of scattering surfaces for Lutz.
/// \author messier@indiana.edu
///
#include <vector>
namespace geo   { class GeometryBase; }
namespace bpfit { class Path;         }

namespace bpfit {
  ///
  /// \brief Construct scattering surfaces for Lutz
  ///
  class ScatteringSurfaces {
  public:
    ///
    /// \brief Construct tool for building scattering surfaces
    ///
    /// \param PDG - Assumed track particle ID (PDG code)
    /// \param dstol - Largest step size allowed (cm)
    /// \param dxtol - Largest step size allowed (radiation lengths)
    /// \param dttol - Largest RMS scatter allowd (cm)
    ///
    /// This class will attempt to place scattering surfaces between a
    /// set of measurement planes. The distance between scattering
    /// planes will be as large as possible subject to the tolerances
    /// above:
    ///
    /// - No two planes will be allowed to be more than dstol apart
    /// from each other in units of cm. Typical value dstol = 100 cm
    ///
    /// - No two planes will be allowed to be more than dxtol apart
    /// from each other in units of radiation lengths. Typical value
    /// dxtol = 2.0
    ///
    /// - The predicted RMS scatter between two planes will be kept
    /// within dttol. That is, the produce of the estimated RMS
    /// scattering angle of the particle times the step size must be
    /// smaller than dttol. Typical value dttol = 0.3 cm
    ///
    /// The class will try to meet all the criteria above, but it will
    /// not insert more than one scattering plane between
    /// measurements.
    ///
    /// The class will ensure that at least one scattering plane is
    /// inserted among the measurements.
    ///
    ScatteringSurfaces(int PDG,
		       double dstol,
		       double dxtol,
		       double dttol);
    ///
    /// \brief Build scattering surfaces
    /// \param geo - Detector geometry
    /// \param p0  - Track start point (x/y/z cm)
    /// \param p1  - Track end point (x/y/z cm)
    /// \param s   - list of measurement plane locations
    ///
    /// All of the measurement planes s are candidates for locations
    /// of scattering surfaces. In the interests of optimization, some
    /// surfaces will be dropped if the expected scattering is rather
    /// small.
    ///
    void MakeSurfaces(const geo::GeometryBase& geo,
		      const double* p0,
		      const double* p1,
		      const std::vector<double>& s);
    
    ///
    /// \brief The number of scattering planes constructed
    ///
    unsigned int N() const;
    
    ///
    /// \brief The location of the ith scattering plane
    ///
    double S(unsigned int i) const;

    ///
    /// \brief The predicted RMS scattering angle at plane i
    ///
    double SigAlpha(unsigned int i) const;
    
  private:
    void CheckLimits(double* beta, double* p);
    void ForceOnePlane(const Path& path,
		       const std::vector<double>& s);
    bool GoodStep(double ds, double dx, double dt);
    void GetStepData(const Path& path,
		     const std::vector<double>& s,
		     unsigned int i1,
		     unsigned int i2,
		     double* ds,
		     double* smid,
		     double* dx,
		     double* psi,
		     double* dt);
    unsigned int MakeStep(const Path& path,
			  const std::vector<double>& s,
			  unsigned int i1);
    void MakeSteps(const Path& path,
		   const std::vector<double>& s);
  private:
    unsigned int        fPDG;      ///< Assumed track PDG code
    double              fdStol;    ///< Largest allowed step (cm)
    double              fdXtol;    ///< Largest allowed step (rad. len.)
    double              fdTtol;    ///< Largest allowed RMS scatter (cm)
    std::vector<double> fS;        ///< Locations of scattering surfaces
    std::vector<double> fSigAlpha; ///< RMS scattering uncertainties
  };
}
////////////////////////////////////////////////////////////////////////
