
// #include "BreakPointFitter/func/Lutz.h"
#include "Lutz.h"

#include <iostream>
#include <cmath>
#include <getopt.h>
#include "TFile.h"
#include "TRandom3.h"
#include "TH1F.h"
#include "TGraph.h"

static void help() 
{
  std::cerr 
    << "Test program for the Lutz.cxx fitting routines.\n"
    << "This program will generate and fit random tracks.\n"
    << "Summary results will be stored in lutz.root.\n"
    << "Then run 'root -l lutz.C'.\n"
    << "\n"
    << "Usage:\n" 
    << "\n"
    << "% testLutz --seed=# --nitr=# --n=# --N=# --pitch=# --sigx=# --sigSJ=#\n"
    << "\n"
    << "Where\n"
    << "  seed is the random seed\n"
    << "  nitr is the number of track fits to make\n"
    << "  n is the number of track measurements\n"
    << "  N is the number of scattering planes\n"
    << "  pitch is the distance between measurement planes\n"
    << "  sigx is the measurement resolution\n"
    << "  sigSJ is the RMS scattering at each scatter plane\n"
    << std::endl;
  exit(0);
}

//......................................................................

int main(int argc, char** argv) 
{
  unsigned int i, j, itr;
  TFile* f = new TFile("lutz.root","RECREATE");
  
  TH1F* chi2Histo       = new TH1F("chi2Histo",       ";#chi^{2};tracks",           200,0.0,100.0);
  TH1F* aHisto          = new TH1F("aHisto",          ";a_{fit}-a_{true}",          200,-10.0,10.0);
  TH1F* bHisto          = new TH1F("bHisto",          ";b_{fit}-b_{true}",          200,-1.0,1.0);
  TH1F* xResidHisto     = new TH1F("xResidHisto",     ";x_{fit}-x_{true}",          200,-10.0,10.0);
  TH1F* alphaResidHisto = new TH1F("alphaResidHisto",";#alpha_{fit}-#alpha_{true}",200,-1.0,1.0);

  unsigned int seed   = 11121970;
  unsigned int maxItr = 1000000;
  unsigned int n      = 20;
  unsigned int N      = 10;
  double       pitch  = 12.0;
  double       sigx   = 2.0/sqrt(12.0);
  double       sigSJ  = 0.15;

  static struct option longopts[] = {
    {"seed", required_argument, NULL, 's'},
    {"nitr", required_argument, NULL, 'i'},
    {"n",    required_argument, NULL, 'n'},
    {"N",    required_argument, NULL, 'N'},
    {"pitch",required_argument, NULL, 'p'},
    {"sigx", required_argument, NULL, 'x'},
    {"sigSJ",required_argument, NULL, 'a'},
    {NULL,   0,                 NULL, 0}
  };
  int ch;
  while ((ch=getopt_long(argc, argv, "s:i:n:N:p:a:S:", longopts, NULL))!=-1) {
    switch (ch) {
    case 's': seed   = atoi(optarg); break;
    case 'i': maxItr = atoi(optarg); break;
    case 'n': n      = atoi(optarg); break;
    case 'N': N      = atoi(optarg); break;
    case 'p': pitch  = atof(optarg); break;
    case 'x': sigx   = atof(optarg); break;
    case 'a': sigSJ  = atof(optarg); break;
    default:  help();
    }
  }

  double           zmax = n*pitch;
  TVectorT<double> z(n);
  TVectorT<double> Z(N);
  double           aTRUE;
  double           bTRUE;
  TVectorT<double> alphaTRUE(N);
  TVectorT<double> xTRUE(n);
  TVectorT<double> x(n);
  TVectorT<double> xi(n);
  double           zeta;
  TRandom3         rand(seed);
  
  //
  // Make z locations of measurement planes
  //
  for (i=0; i<n; ++i) z[i] = i*zmax/(n-1);
  
  //
  // Make z locations of scattering planes
  //
  for (j=0; j<N; ++j) Z[j] = (j+1)*zmax/(N+1);
  
  //
  // Let's make lots of trials
  //
  for (itr=0; itr<maxItr; ++itr) {
    //
    // Choose some random track parameters
    //
    aTRUE = rand.Uniform(-10.0,10.0);
    bTRUE = rand.Uniform(-1.0,1.0);
    for (j=0; j<N; ++j) alphaTRUE[j] = rand.Gaus(0.0,sigSJ);
    
    //
    // Construct some track parameters
    //
    for (i=0; i<n; ++i) {
      xTRUE[i] = aTRUE + bTRUE*z[i];
      for (j=0; j<N; ++j) {
	zeta = z[i]-Z[j];
	if (zeta>0.0) xTRUE[i] += alphaTRUE[j]*zeta;
      }
    }
    for (i=0; i<n; ++i) {
      xi[i] = rand.Gaus(xTRUE[i],sigx);
    }
  
    //
    // Build and make the fit
    //
    double a, b, alpha[N], chi2;
    bpfit::Lutz lutz(n,N);
    for (i=0; i<n; ++i) lutz.SetMeasurement(i, z[i], xi[i], sigx);
    for (j=0; j<N; ++j) lutz.SetScatteringPlane(j, Z[j], sigSJ);
    lutz.Fit(&a, &b, alpha, &chi2);
    
    //
    // Log results to histograms
    //
    chi2Histo->Fill(chi2);
    aHisto->Fill(a-aTRUE);
    bHisto->Fill(b-bTRUE);
    for (i=0; i<n; ++i) xResidHisto->    Fill(lutz.X(z[i])-xTRUE[i]);
    for (j=0; j<N; ++j) alphaResidHisto->Fill(alpha[j]-alphaTRUE[j]);

    if (itr<10) {
      char buff[256];
      sprintf(buff,"trkT_%.2d",itr);
      TGraph* g1 = new TGraph(n, z.GetMatrixArray(), xTRUE.GetMatrixArray());
      g1->SetName(buff);

      sprintf(buff,"trkM_%.2d",itr);
      TGraph* g2 = new TGraph(n, z.GetMatrixArray(), xi.GetMatrixArray());
      g2->SetName(buff);

      sprintf(buff,"trkF_%.2d",itr);
      double x[n];
      for (i=0; i<n; ++i) x[i] = lutz.X(z[i]);
      TGraph* g3 = new TGraph(n, z.GetMatrixArray(), x);
      g3->SetName(buff);
      
      g1->Write();
      g2->Write();
      g3->Write();

    }
    if ( (itr%1000)==0) {
      std::cout << itr << " fits of " << maxItr << std::endl;
    }
  }
  
  chi2Histo->      Write();
  aHisto->         Write();
  bHisto->         Write();
  xResidHisto->    Write();
  alphaResidHisto->Write();
  f->Close();

  return 0;
}
////////////////////////////////////////////////////////////////////////
