#include "CAFMaker/Blinding.h"
#include <iostream>

// This helper handles the blinding of CAF records.  It defines the (currently simple)
// cuts deciding if a record needs to be blinded.  When required, it goes through
// the record punching holes in it.
// 
// Energy, PID, and tracklenth variables are reset.  Other variables too closely
// related to energy, PID, and tracklength are also reset.

namespace caf
{

  // Using standard dummy value (-5), except for unsigned vals (0).
  // Let casting take care of the type cconversions.
  static const int blind_val = -5;
  static const unsigned int blind_val_unsigned = 0;

  //......................................................................
  void BlindThisSlice(SRSlice& slc) {
    // Small numbers of ncontplanes are useful for cosmic rejection, large
    // numbers can leak the event length (~ energy).
    slc.ncontplanes = std::max(10u, slc.ncontplanes);
    slc.lastplane   = blind_val_unsigned;
    slc.calE        = blind_val;
    slc.boxmax. SetZ( blind_val );
    slc.meanpos.SetZ( blind_val );
  }

  //......................................................................
  void BlindThisTrack(SRTrack& trk) {
    trk.nhit         = blind_val_unsigned;
    trk.nhitx        = blind_val_unsigned;
    trk.nhity        = blind_val_unsigned;
    trk.nplane       = blind_val_unsigned;
    trk.maxplanecont = blind_val_unsigned;
    trk.calE         = blind_val;
    trk.len          = blind_val;
    trk.lenE         = blind_val;
    trk.stop.    SetZ( blind_val );
  }

  //......................................................................
  void BlindThisJMShower(SRJMShower& jm) {
    jm.nhit          = blind_val_unsigned;
    jm.nhitx         = blind_val_unsigned;
    jm.nhity         = blind_val_unsigned;
    jm.nplane        = blind_val_unsigned;
    jm.maxplanecont  = blind_val_unsigned;
    jm.calE          = blind_val;
    jm.len           = blind_val;
    jm.shwE          = blind_val;
    jm.vtxE          = blind_val;
    jm.stop.     SetZ( blind_val );
  }

  //......................................................................
  void BlindThisShower(SRShower& shw) {
    shw.nhit          = blind_val_unsigned;
    shw.nhitx         = blind_val_unsigned;
    shw.nhity         = blind_val_unsigned;
    shw.nplane        = blind_val_unsigned;
    shw.nplanex       = blind_val_unsigned;
    shw.nplaney       = blind_val_unsigned;
    shw.maxplanecont  = blind_val_unsigned;
    shw.calE          = blind_val;
    shw.len           = blind_val;
    shw.stop.     SetZ( blind_val );

  }

  //......................................................................
  void BlindThisElasticVtx(SRElastic& vtx) {
    for (unsigned int i=0; i<vtx.png3d.size(); i++) {
      vtx.png3d[i].nhit         = blind_val_unsigned;
      vtx.png3d[i].nhitx        = blind_val_unsigned;
      vtx.png3d[i].nhity        = blind_val_unsigned;
      vtx.png3d[i].nplane       = blind_val_unsigned;
      vtx.png3d[i].maxplanecont = blind_val_unsigned;
      vtx.png3d[i].calE         = blind_val;
      vtx.png3d[i].len          = blind_val;
    }
    for (unsigned int i=0; i<vtx.png2d.size(); i++) {
      vtx.png2d[i].nhit         = blind_val_unsigned;
      vtx.png2d[i].nhitx        = blind_val_unsigned;
      vtx.png2d[i].nhity        = blind_val_unsigned;
      vtx.png2d[i].nplane       = blind_val_unsigned;
      vtx.png2d[i].maxplanecont = blind_val_unsigned;
      vtx.png2d[i].calE         = blind_val;
      vtx.png2d[i].len          = blind_val;
    }
  }
 
  //......................................................................
  void BlindThisTrkME(SRTrkME& me){
    me.meanpos.SetZ( blind_val );
  }

  //......................................................................
  void BlindThisSlcME(SRSlcME& me){
    me.meanpos.SetZ( blind_val );
  }      

  //......................................................................
  void BlindThisCosRej(SRCosRej& cosrej) {
    cosrej.eratio = blind_val;
    cosrej.hadE   = blind_val;
  }

  //......................................................................
  void BlindThisNueSandbox(SRNueSandbox& nue) {
    nue.slicemeanz          = blind_val;
    nue.nplanesfilled       = blind_val;
    nue.nxplanes            = blind_val_unsigned;
    nue.nyplanes            = blind_val_unsigned;
    nue.mipmax.         SetZ( blind_val );
    nue.evtsump             = blind_val;
    nue.evtsumpt            = blind_val;
    nue.evtsump0             = blind_val;
    nue.evtsumpt0            = blind_val;
    nue.evtetot             = blind_val;
    nue.sh1energy           = blind_val;
    nue.sh1exclenergy       = blind_val;
    nue.sh1totalL           = blind_val;
    nue.sh1stop.        SetZ( blind_val );
    nue.sh1vtxdoca          = blind_val;
    nue.sh1gap              = blind_val;
    nue.sh1nplane           = blind_val_unsigned;
    nue.sh1xnplane          = blind_val_unsigned;
    nue.sh1ynplane          = blind_val_unsigned;
    nue.sh1ncell            = blind_val_unsigned;
    nue.sh1xncell           = blind_val_unsigned;
    nue.sh1yncell           = blind_val_unsigned;
    nue.sh1pid              = blind_val;
    nue.sh2energy           = blind_val;
    nue.sh2exclenergy       = blind_val;
    nue.sh2totalL           = blind_val;
    nue.sh2stop.        SetZ( blind_val );
    nue.sh2vtxdoca          = blind_val;
    nue.sh2gap              = blind_val;
    nue.sh2nplane           = blind_val_unsigned;
    nue.sh2xnplane          = blind_val_unsigned;
    nue.sh2ynplane          = blind_val_unsigned;
    nue.sh2ncell            = blind_val_unsigned;
    nue.sh2xncell           = blind_val_unsigned;
    nue.sh2yncell           = blind_val_unsigned;
    nue.sh2pid              = blind_val;
    nue.elll                = blind_val;
    nue.eedge2cell          = blind_val;
    nue.eedge5cell          = blind_val;
    nue.eedge10cell         = blind_val;
    nue.E                   = blind_val;
    nue.depE                = blind_val;
    nue.shwE                = blind_val;
    nue.fracangchanges      = blind_val;
    nue.fracangchangesloose = blind_val;
    nue.fracmodalhits       = blind_val;
    nue.fracnoncontig       = blind_val;
    nue.fracnoncontigloose  = blind_val;
    nue.maxgap              = blind_val;

    nue.jmid.setDefault();

  }

  //......................................................................
  void BlindThisNumuSandbox(SRNumuSandbox& numu) {
  }

  //......................................................................
  void BlindThisRecord(StandardRecord* rec) {
    /// If required, scrub the record of any quantities that shouldn't be viewed

    if(rec->hdr.det != kFARDET){
      std::cerr << "Should only apply blinding to the Far Detector, aborting."
                << std::endl;
      abort();
    }

    if(rec->hdr.ismc){
      std::cerr << "Shouldn't apply blinding to Monte Carlo, aborting."
                << std::endl;
      abort();
    }

    // Is this record in the box?

    // Blinding 0.8 to 3 GeV events within 30 microsecond window
    // around center of beam spill (t=222000 microseconds).
    if (rec->slc.calE > 0.8 &&
	rec->slc.calE < 3.0 &&
	rec->slc.meantime > 207000 && 
	rec->slc.meantime < 237000 ) {

      // Walk through each rec branch.

      // -- start of branches --

      // -- hdr
      // Tag the record as blind
      rec->hdr.blind = 1;
    
      // -- spill

      // -- slc
      BlindThisSlice(rec->slc);

      // -- trk
      for (unsigned int i=0; i<rec->trk.kalman.size(); i++) {
	BlindThisTrack(rec->trk.kalman[i]);
      }
      for (unsigned int i=0; i<rec->trk.discrete.size(); i++) {
	BlindThisTrack(rec->trk.discrete[i]);
      }
      for (unsigned int i=0; i<rec->trk.discrete2d.size(); i++) {
	BlindThisTrack(rec->trk.discrete2d[i]);
      }
      for (unsigned int i=0; i<rec->trk.cosmic.size(); i++) {
	BlindThisTrack(rec->trk.cosmic[i]);
      }

      // -- jm shw
      for (unsigned int i=0; i<rec->sand.nue.jm.size(); i++) {
	BlindThisJMShower(rec->sand.nue.jm[i]);
      }

      // -- lid shw
      for (unsigned int i=0; i<rec->shw.shwlid.size(); i++) {
	BlindThisShower(rec->shw.shwlid[i]);
      }

      // -- vtx
      for (unsigned int i=0; i<rec->vtx.elastic.size(); i++) {
	BlindThisElasticVtx(rec->vtx.elastic[i]);
      }

      // -- michels
      for (unsigned int i=0; i< rec->me.trkkalman.size(); i++){
	BlindThisTrkME(rec->me.trkkalman[i]);
      }
      for (unsigned int i=0; i< rec->me.trkdiscrete.size(); i++){
	BlindThisTrkME(rec->me.trkdiscrete[i]);
      }
      for (unsigned int i=0; i< rec->me.trkcosmic.size(); i++){
	BlindThisTrkME(rec->me.trkcosmic[i]);
      }
      for (unsigned int i=0; i< rec->me.slc.size(); i++){
	BlindThisSlcME(rec->me.slc[i]);
      }

      // -- energy
      // for now, everything goes, so use setDefault() methods
      rec->energy.numusimp.setDefault();
      rec->energy.elecid  .setDefault();
      for (unsigned int i=0; i<rec->energy.mutrkE.size(); i++) {
	rec->energy.mutrkE[i].setDefault();
      }
      for (unsigned int i=0; i<rec->energy.shwlid.size(); i++) {
        rec->energy.shwlid[i].setDefault();
      }

      // -- sel
      // for now, everything goes, so use setDefault() methods
      rec->sel.remid .setDefault();
      rec->sel.qepid .setDefault();
      rec->sel.elecid.setDefault();
      for (unsigned int i=0; i<rec->sel.elecid.shwlid.size(); i++){
        rec->sel.elecid.shwlid[i].setDefault();
      }
      rec->sel.lem   .setDefault();
      rec->sel.rvp   .setDefault();
      // some stuff survives here:
      BlindThisCosRej   (rec->sel.cosrej);

      // -- sand
      BlindThisNueSandbox (rec->sand.nue );
      BlindThisNumuSandbox(rec->sand.numu);

      // -- mc

      // -- parent

      // -- end of branches --

    } // end if (rec needs blinding)

  } 

} // namespace caf
