////////////////////////////////////////////////////////////////////////
/// \file  CAFReweight.cxx
/// \brief Wrapper for reweight neutrino interactions with GENIE
///
/// \author  nathan.mayer@tufts.edu
////////////////////////////////////////////////////////////////////////

// C/C++ includes
#include <math.h>
#include <map>
#include <fstream>
#include <memory>
#include <vector>

//ROOT includes
#include "TVector3.h"
#include "TLorentzVector.h"

//GENIE includes
#include "Conventions/Units.h"
#include "EVGCore/EventRecord.h"
#include "GHEP/GHepUtils.h"
#include "Geo/ROOTGeomAnalyzer.h"
#include "Geo/GeomVolSelectorFiducial.h"
#include "Geo/GeomVolSelectorRockBox.h"
#include "Utils/StringUtils.h"
#include "Utils/XmlParserUtils.h"
#include "Interaction/InitialState.h"
#include "Interaction/Interaction.h"
#include "Interaction/Kinematics.h"
#include "Interaction/KPhaseSpace.h"
#include "Interaction/ProcessInfo.h"
#include "Interaction/XclsTag.h"
#include "Messenger/Messenger.h"
#include "GHEP/GHepParticle.h"
#include "PDG/PDGCodeList.h"

// Necessary because the GENIE LOG_* macros don't fully qualify Messenger
using genie::Messenger;

//Novasoft includes
#include "StandardRecord/SRGenerator.h"
#include "StandardRecord/SRGenParticle.h"
#include "CAFReweight/CAFReweight.h"


// Framework includes
//#include "art/Framework/Services/Registry/ServiceHandle.h"
//#include "fhiclcpp/ParameterSet.h"
//#include "messagefacility/MessageLogger/MessageLogger.h"

namespace cafrwgt {

  CAFReweight::CAFReweight():
    fWgtMultiNus(false) {
  }

  CAFReweight::~CAFReweight() {
    for(unsigned int i = 0; i < fGRwgt.size(); ++i) delete fGRwgt[i];
  }

  rwgt::GENIEReweight* CAFReweight::GenieReweight(int i) {
    return fGRwgt[i];
  }
  
  void CAFReweight::AddReweightValue(rwgt::ReweightLabel_t rLabel, double value, unsigned int i) {
    if(fGRwgt.size() <= i) {
      fGRwgt.push_back(new rwgt::GENIEReweight);
    }
    
    fGRwgt[i]->AddReweightValue(rLabel, value);
  }
  
  void CAFReweight::Configure(int i) {
    if(i >= 0){
      fGRwgt[i]->Configure();
    }
    else{
      for(unsigned int j = 0; j < fGRwgt.size(); j++) {
        fGRwgt[j]->Configure();
      }
    }
  }

  std::vector<double> CAFReweight::CalcWeightVector(const std::vector<caf::SRGenerator>& genvec) {
    
    std::vector<double> wgts;
    
    //Initialize wgts vector.
    for(unsigned int i = 0; i < fGRwgt.size(); i++) {
      wgts.push_back(1.0);
    }
    
    //Catch the case where the generator vector in the caf is empty
    if(genvec.size()==0) {
      return wgts;
    }
    else { //If the genvec is not empty calcuate actual weights
      //This is the case where you are not doing multi nu weighting
      if(fWgtMultiNus==false) {
	//Store the genie event record for the circumstances that
	//the multiple weights with different configurations are being
	//requested.
	genie::EventRecord evr = this->RetrieveGHEP(genvec[0]);
	for(unsigned int i = 0; i < fGRwgt.size(); i++) { //loop over the vector of configurations.  But use the same event record
	  double wgt = fGRwgt[i]->CalculateWeight(evr);
	  wgts[i] = wgt;
	}
	return wgts;
      }
      else { //Now for the multi nu case
	std::vector<genie::EventRecord> evr_vec; //make vector of event records.
	for(unsigned int i = 0; i < genvec.size(); i++) {
	  genie::EventRecord evr = this->RetrieveGHEP(genvec[i]);
	  evr_vec.push_back(evr);
	}
	//Loop over the configurations
	for(unsigned int i = 0; i < fGRwgt.size(); i++) {
	  double wgt = 0.0;
	  //Loop over the event records
	  for(unsigned int j = 0; j < evr_vec.size(); j++) {
	    double nu_wgt = genvec[j].fVisEfraction; //This should have the same size as evr_vec
	     wgt += nu_wgt*(fGRwgt[i]->CalculateWeight(evr_vec[j]));
	  }
	  wgts[i]=wgt;
	}
	return wgts; //Return the vector of weights.
      }
    }
  }
  

  double CAFReweight::CalcWeight(const std::vector<caf::SRGenerator>& genvec, unsigned int i) {
    double wgt = 1.0;
    if(genvec.size()==0) {
      return wgt;
    }
    else {
      if(fWgtMultiNus==false) {
	genie::EventRecord evr = this->RetrieveGHEP(genvec[0]);
	return fGRwgt[i]->CalculateWeight(evr);
      }
      else {
	wgt = 0;
	for(unsigned int i = 0; i < genvec.size(); i++) {
	  genie::EventRecord evr = this->RetrieveGHEP(genvec[i]);
	  double nu_wgt = genvec[i].fVisEfraction;
	  wgt += nu_wgt*(fGRwgt[i]->CalculateWeight(evr));
	}
	return wgt;
      }
    }
  }

  genie::EventRecord CAFReweight::RetrieveGHEP(const caf::SRGenerator& gen) {

    genie::EventRecord newEvent;
    newEvent.SetWeight(gen.fweight);
    newEvent.SetProbability(gen.fprobability);
    newEvent.SetXSec(gen.fXsec);

#ifndef SETDIFFXSEC_1ARG
    genie::KinePhaseSpace_t space = genie::kPSNull; // kPSQ2fE; // ????                                                                                                                          
    // dsig/dQ2, dsig/dQ2dW, dsig/dxdy ...                                                                                                                                                       

    newEvent.SetDiffXSec(gen.fDiffXsec,space);

    // TODO:  we don't know currently know what to use ...                                                                                                                                       
    // for now just to get things to compile ... I (Nate) needs to look at this                                                                                                                      
    static int nmsg = 10;
    if ( nmsg > 0 ) {
      LOG_WARN("CAFReweight") << "RetrieveGHEP(simb::MCTruth,simb::GTruth) is not correctly setting KinePhaseSpace_t in SetDiffXSec()\n"
                              << "At the time of the conversion to R-2_8_0 (2013-05-01) this is not critical\n"
                              << "But it should be fixed";
      --nmsg;
      if ( nmsg == 0 ) LOG_WARN("CAFReweight") << "... last of such messages";
    }
    //assert(0);                                                                                                                                                                                 
#else
    newEvent.SetDiffXSec(gen.fDiffXsec);
#endif
    
    TLorentzVector vtx = gen.fVertex;
    newEvent.SetVertex(vtx);

    int nParticles = gen.fParticleList.size();
    // Reserve space for the particles we're about to add
    newEvent.Expand(nParticles);
    for(int i = 0; i < nParticles; i++) {
      int gmid = gen.fParticleList[i].fPdgCode;
      int gmmo = gen.fParticleList[i].fMother;
      int gmfd = gen.fParticleList[i].fFirstDaughter;
      int gmld = gen.fParticleList[i].fLastDaughter;
      
      int gmri = gen.fParticleList[i].fRescatterCode;
      
      double gmvx = gen.fParticleList[i].fNuclVtx.X();
      double gmvy = gen.fParticleList[i].fNuclVtx.Y();
      double gmvz = gen.fParticleList[i].fNuclVtx.Z();
      double gmvt = gen.fParticleList[i].fNuclVtx.T();
      
      double gmpx = gen.fParticleList[i].fP4.Px();
      double gmpy = gen.fParticleList[i].fP4.Py();
      double gmpz = gen.fParticleList[i].fP4.Pz();
      double gme = gen.fParticleList[i].fP4.E();
      
      genie::GHepStatus_t gmst = (genie::GHepStatus_t)gen.fParticleList[i].fStatus;

      TVector3 polz = gen.fParticleList[i].fPolz;

      // Normally one would create a GHepParticle and add it to the event with
      // AddParticle().
      //
      // But AddParticle() calls UpdateDaughterLists() every time, which is
      // slow. And unnecessary to be called even once, because we set the
      // daughters by ourselves above.
      //
      // Instead, create the particle in-place in the EventRecord by hand. This
      // is significantly faster in tests.

      genie::GHepParticle* gpart = new (newEvent[i]) genie::GHepParticle(gmid, gmst, gmmo, -1, gmfd, gmld, gmpx, gmpy, gmpz, gme, gmvx, gmvy, gmvz, gmvt);
      gpart->SetRescatterCode(gmri);

      if(polz.x() !=0 || polz.y() !=0 || polz.z() !=0) {
	gpart->SetPolarization(polz);
      }
    }

    genie::ProcessInfo proc_info;
    genie::ScatteringType_t gscty = (genie::ScatteringType_t)gen.fGscatter;
    genie::InteractionType_t ginty = (genie::InteractionType_t)gen.fGint;

    proc_info.Set(gscty, ginty);

    genie::XclsTag gxt;
    
    //Set Exclusive Final State particle numbers
    genie::Resonance_t gres = (genie::Resonance_t)gen.fResNum;
    gxt.SetResonance(gres);
    gxt.SetNPions(gen.fNumPiPlus, gen.fNumPi0, gen.fNumPiMinus);
    gxt.SetNNucleons(gen.fNumProton, gen.fNumNeutron);
    
    if(gen.fIsCharm) {
      gxt.SetCharm(0);
    }
     else {
       gxt.UnsetCharm();
     }
   
    //Set the GENIE kinematic info
    genie::Kinematics gkin;
    gkin.Setx(gen.fgX, true);
    gkin.Sety(gen.fgY, true);
    gkin.Sett(gen.fgT, true);
    gkin.SetW(gen.fgW, true);
    gkin.SetQ2(gen.fgQ2, true);
    gkin.Setq2(gen.fgq2, true);
    TLorentzVector lep_p4 = gen.fParticleList[gen.fFSLeptonIndx].fP4;
    
    gkin.SetFSLeptonP4(lep_p4.Px(), lep_p4.Py(), lep_p4.Pz(), lep_p4.E());
    gkin.SetHadSystP4(gen.fFShadSystP4.Px(), gen.fFShadSystP4.Py(), gen.fFShadSystP4.Pz(), gen.fFShadSystP4.E());

    //Set the GENIE final state interaction info
    genie::InitialState ginstate(gen.ftgtPDG, gen.fProbePDG);
    
    // Set up the target, except for PDG set above
    genie::Target* target123 = ginstate.TgtPtr();

    //int Z = gtruth.ftgtZ;
    //int A = gtruth.ftgtA;
    
    //target123->SetId(Z,A);
    
    //int pdg_code = pdg::IonPdgCode(A, Z);
    //TParticlePDG * p = PDGLibrary::Instance()->Find(pdg_code);

    //mf::LogWarning("GENIEReweight") << "Setting Target Z and A";
    //mf::LogWarning("GENIEReweight") << "Saved PDG: " << gtruth.ftgtPDG;
    //mf::LogWarning("GENIEReweight") << "Target PDG: " << target123->Pdg();
    target123->SetHitNucPdg(gen.fHitNucleon);
    target123->SetHitQrkPdg(gen.fHitQuark);
    target123->SetHitSeaQrk(gen.fIsSeaQuark);

    if(newEvent.HitNucleonPosition()> 0) {
      target123->SetHitNucP4(*newEvent.HitNucleon()->P4());
    }  
    else {
      target123->SetHitNucP4(TLorentzVector(0.,0.,0.,0.));
    }

    // Set probe information
    ginstate.SetProbeP4(*newEvent.Probe()->P4());
    if(newEvent.TargetNucleusPosition()> 0) {
      ginstate.SetTgtP4(*newEvent.TargetNucleus()->P4());
    }
    else {
      ginstate.SetTgtP4(TLorentzVector(0.,0.,0.,0.));
    }

    // Build the final Interaction object
    genie::Interaction* p_gint = new genie::Interaction(ginstate, proc_info);
    p_gint->SetKine(gkin);
    p_gint->SetExclTag(gxt);
    newEvent.AttachSummary(p_gint);

    return newEvent;

  }
  
}
