////////////////////////////////////////////////////////////////////////
// \file    AssessCalib_module.cc
// \brief   Make TTree with reconstructed and true energies
// \author  Christopher Backhouse - bckhouse@caltech.edu
////////////////////////////////////////////////////////////////////////

// Framework includes
#include "art/Framework/Core/EDAnalyzer.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Services/Optional/TFileService.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "art/Persistency/Common/Ptr.h"
#include "fhiclcpp/ParameterSet.h"

#include "Calibrator/Calibrator.h"
#include "Geometry/Geometry.h"
#include "GeometryObjects/CellGeo.h"
#include "MCCheater/BackTracker.h"
#include "RecoBase/CellHit.h"
#include "RecoBase/RecoHit.h"

#include "TTree.h"

namespace{
  // Variables for the tree
  int gPlane, gCell, gView;
  double gPE;
  double gTrueE;
  double gW;
  double gPECorr;
  double gGeV;
}

namespace calib
{
  /// Make TTree with reconstructed and true energies
  class AssessCalib: public art::EDAnalyzer
  {
  public:
    explicit AssessCalib(const fhicl::ParameterSet& pset);
    ~AssessCalib();

    void beginJob();

    void reconfigure(const fhicl::ParameterSet& pset);

    void analyze(const art::Event& evt);

  protected:
    std::string fCellHitLabel;

    TTree* fTree;
  };

  //.......................................................................
  AssessCalib::AssessCalib(const fhicl::ParameterSet& pset)
    : EDAnalyzer(pset)
  {
    reconfigure(pset);
  }

  //......................................................................
  AssessCalib::~AssessCalib()
  {
  }

  //......................................................................
  void AssessCalib::reconfigure(const fhicl::ParameterSet& pset)
  {
    fCellHitLabel = pset.get<std::string>("CellHitLabel");
  }

  //......................................................................
  void AssessCalib::beginJob()
  {
    art::ServiceHandle<art::TFileService> tfs;

    fTree = tfs->make<TTree>("tree", "tree");

    fTree->Branch("plane", &gPlane);
    fTree->Branch("cell", &gCell);
    fTree->Branch("view", &gView);
    fTree->Branch("pe", &gPE);
    fTree->Branch("trueE", &gTrueE);
    fTree->Branch("W", &gW);
    fTree->Branch("pecorr", &gPECorr);
    fTree->Branch("gev", &gGeV);
  }

  //......................................................................
  void AssessCalib::analyze(const art::Event& evt)
  {
    art::ServiceHandle<cheat::BackTracker> bt;
    art::ServiceHandle<calib::Calibrator> cal;
    art::ServiceHandle<geo::Geometry> geom;

    art::Handle<std::vector<rb::CellHit>> chits;
    evt.getByLabel(fCellHitLabel, chits);

    const unsigned int N = chits->size();
    for(unsigned int chitIdx = 0; chitIdx < N; ++chitIdx){
      art::Ptr<rb::CellHit> chit(chits, chitIdx);
      if(bt->IsNoise(chit)) continue;

      const std::vector<sim::FLSHit> flss = bt->HitToFLSHit(chit);
      if(flss.empty()) continue;

      gPlane = chit->Plane();
      gCell = chit->Cell();
      gView = int(chit->View());
      gPE = chit->PE();

      gTrueE = 0;
      for(const sim::FLSHit& fls: flss) gTrueE += fls.GetEdep();

      double xyz[3];
      bt->HitToXYZ(chit).GetXYZ(xyz);

      double world[3];
      geom->Plane(gPlane)->Cell(gCell)->LocalToWorld(xyz, world);
      const double W = world[1-gView];

      gW = W;

      const rb::RecoHit rhit(cal->MakeRecoHit(*chit, W));

      if(!rhit.IsCalibrated()) continue;

      gPECorr = rhit.PECorr();
      gGeV = rhit.GeV();

      fTree->Fill();
    } // end for chitIdx
  }

  DEFINE_ART_MODULE(AssessCalib);

} // end namespace calib
////////////////////////////////////////////////////////////////////////

