////////////////////////////////////////////////////////////////////////
/// \brief   Service to store calibration data products (CDP) in the
///          SQLite3 metadatabase of a file                         
/// \author  brebel@fnal.gov
////////////////////////////////////////////////////////////////////////
#include "Calibration/CDPStorage.h"

#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "messagefacility/MessageLogger/MessageLogger.h"
#include "art/Utilities/OutputFileInfo.h"
#include "art/Persistency/RootDB/SQLite3Wrapper.h"
#include "art/Persistency/RootDB/SQLErrMsg.h"
#include "TFile.h"

namespace calib 
{
  //----------------------------------------------------------------------------
  CDPStorage::CDPStorage(fhicl::ParameterSet const& pset,
			 art::ActivityRegistry& reg)
  {
    this->reconfigure(pset);
    reg.sPostCloseOutputFile.watch(this, &CDPStorage::postCloseOutputFile);
    reg.sPostOpenFile.watch(this, &CDPStorage::postOpenFile);
  }

  //----------------------------------------------------------------------------
  CDPStorage::~CDPStorage()
  {
  }

  //----------------------------------------------------------------------------
  void CDPStorage::reconfigure(const fhicl::ParameterSet& pset)
  {
    fRetrieveDriftCalFromFile = pset.get<bool>("RetrieveDriftCalFromFile");
    
    return;
  }

  //----------------------------------------------------------------------------
  void CDPStorage::postOpenFile(std::string const& fn)
  {
    // test the input flags to see if we are to get the 
    // stored calibrations from the file - currently only
    // using this functionality for the drift calibration, but 
    // could use it for others in the future by including ||
    // with each flag in the following line
    if(fRetrieveDriftCalFromFile){

      std::shared_ptr<TFile> file(TFile::Open(fn.c_str(), "READ"));

      // get the wrapper to the SQLite3 DB
      art::SQLite3Wrapper sql3w(file.get(), 
				"NOvACalibrations", 
				SQLITE_OPEN_READONLY);

      if(fRetrieveDriftCalFromFile) this->RetrieveDriftCalFromFile(sql3w);
    }

    return;
  }

  //----------------------------------------------------------------------------
  void CDPStorage::postCloseOutputFile(art::OutputFileInfo const& ofi)
  {
    std::shared_ptr<TFile> file(TFile::Open(ofi.fileName().c_str(), "UPDATE"));

    // get the wrapper to the SQLite3 DB
    art::SQLite3Wrapper sql3w(file.get(), 
			      "NOvACalibrations", 
			      SQLITE_OPEN_CREATE | SQLITE_OPEN_READWRITE);

    this->WriteDriftResponse(sql3w);

    // write the DB
    sql3w.reset();

    // close the ROOT file
    file->Close();

    return;
  }

  //----------------------------------------------------------------------------
  void CDPStorage::WriteDriftResponse(art::SQLite3Wrapper& sql3w)
  {
    art::SQLErrMsg errMsg;

    // make the table for the drift response, one column for each of the important
    // quantities
    sqlite3_exec(sql3w,
		 "BEGIN TRANSACTION; "                          // +
		 "DROP TABLE IF EXISTS DriftResponse; "         // +
		 "CREATE TABLE DriftResponse(ID PRIMARY KEY, "  // +
		 "                           Channel   INT,   " // +
		 "                           Mean      REAL,  " // +
		 "                           MeanErr   REAL,  " // +
		 "                           RMS       REAL,  " // +
		 "                           NHits     INT,   " // +
		 "                           StartTime INT,   " // +
		 "                           EndTime   INT );"  // +
		 "COMMIT;",
		 0,
		 0,
		 errMsg);    

    // make sure there were no problems
    errMsg.throwIfError();

    sqlite3_exec(sql3w, "BEGIN TRANSACTION;", 0, 0, errMsg);
    sqlite3_stmt *stmt = 0;
    sqlite3_prepare_v2(sql3w,
		       "INSERT INTO DriftResponse(Channel, Mean, "
		       "MeanErr, RMS, NHits, StartTime, "
		       "EndTime) "
		       "VALUES(?, ?, ?, ?, ?, ?, ?);",
		       -1, 
		       &stmt, 
		       NULL);

    LOG_DEBUG("CDPStorage") << " vector size: " << fDriftResponseVec.size();


    for( auto const& dr : fDriftResponseVec){

      LOG_DEBUG("CDPStorage") << dr;
      
      // get the map of means for each channel number for each
      // dr
      auto const& meanMap = dr.ReturnMap();

      LOG_DEBUG("CDPStorage") << "map size: " << meanMap.size()
				 << " ReturnMap size: " << dr.ReturnMap().size();

      for(auto const& mitr : meanMap){ 
		
	int channel = mitr.first;
	caldp::MeanStruct mean = mitr.second;

	sqlite3_bind_int   (stmt, 1, channel        );
	sqlite3_bind_double(stmt, 2, mean.Mean()    );
	sqlite3_bind_double(stmt, 3, mean.MeanErr() );
	sqlite3_bind_double(stmt, 4, mean.RMS()     );
	sqlite3_bind_int   (stmt, 5, mean.NHits()   );
	sqlite3_bind_int   (stmt, 6, dr.StartTime() );
	sqlite3_bind_int   (stmt, 7, dr.EndTime()   );
	if( sqlite3_step(stmt) != SQLITE_DONE) 
	  throw cet::exception("CDPStorage") << "problem writing to SQLite3 DB";
	sqlite3_reset(stmt);
	sqlite3_clear_bindings(stmt);

      } // end loop over channel to MeanStruct map

    } // end loop over DriftResponse data products

    sqlite3_finalize(stmt);

    sqlite3_exec(sql3w, 
		 "END TRANSACTION;", 
		 0, 
		 0, 
		 art::SQLErrMsg());
    
    return;
  }

  //----------------------------------------------------------------------------
  void CDPStorage::DriftResponseToStore(std::vector<caldp::DriftResponse>& drvec)
  {
    fDriftResponseVec.swap(drvec);

    return;
  }

  //----------------------------------------------------------------------------
  void CDPStorage::RetrieveDriftCalFromFile(art::SQLite3Wrapper& sql3w)
  {

    // Read the ParameterSets into memory.
    sqlite3_stmt *stmt = 0;

    std::string query("SELECT ID, Channel, Mean, MeanErr, RMS, NHits, StartTime, EndTime from DriftResponse;");

    LOG_DEBUG("CDPStorage") << query;

    sqlite3_prepare_v2(sql3w, 
		       query.c_str(), 
		       -1, 
		       &stmt, 
		       NULL);

    
    while(sqlite3_step(stmt) == SQLITE_ROW){
      LOG_DEBUG("CDPStorage") << "row returned with " << sqlite3_column_count(stmt) << " columns";
      for(int c = 0; c < sqlite3_column_count(stmt); ++c)
	LOG_DEBUG("CDPStorage") << c << " " << sqlite3_column_text(stmt, c);
    }

  }

  DEFINE_ART_SERVICE(CDPStorage)

}// namespace
