////////////////////////////////////////////////////////////////////////
// \file    CosmicCalib_module.cc
// \brief   Write out PE vs W plots from the input PCHits files
// \author  Christopher Backhouse, Dan Pershey - bckhouse@caltech.edu, pershey@caltech.edu
////////////////////////////////////////////////////////////////////////

// ROOT includes
#include "TH2.h"

// Framework includes
#include "art/Framework/Core/EDFilter.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Services/Optional/TFileService.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Principal/SubRun.h"

// NOvASoft includes
#include "GeometryObjects/CellGeo.h"
#include "Geometry/Geometry.h"
#include "GeometryObjects/PlaneGeo.h"
#include "CalibrationDataProducts/AttenHists.h"
#include "CalibrationDataProducts/AttenProfiles.h"
#include "CalibrationDataProducts/PCHit.h"
#include "NovaDAQConventions/DAQConventions.h"

namespace calib
{
  class CosmicCalib: public art::EDFilter
  {
  public:
    explicit CosmicCalib(const fhicl::ParameterSet& pset);
    ~CosmicCalib();

    virtual bool filter(art::Event& evt);

    void reconfigure(const fhicl::ParameterSet& pset);

    virtual void beginJob();
    virtual bool beginRun(art::Run& run);
    virtual bool endSubRun(art::SubRun& sr);
    virtual void endJob();

  protected:
    caldp::AttenHists GetChannelHists(bool isData, int plane, int cell);
    caldp::AttenProfiles& GetChannelProfiles(novadaq::cnv::DetId det,
					     bool isData, int plane, int cell);
    int GetView(int plane, art::ServiceHandle<geo::Geometry> geom);
    bool fIncludeBelowThresholdZeros;

    // General histograms for investigating variables.
    // Some of these could be derived from fChannelMap, but it's easier
    // just to fill them as we go.

    caldp::AttenHists fSumChannels[4];

    double fInitEdgeX, fInitEdgeY;
    std::string fPCHitsLabel;

    bool fProfileMode;
    bool fConsolidateViewsMode;
    bool fHistoOutput;

    TH2* fCellPE[4];
    TH2* fPlanePE[4];

    // Direction cosines
    TH1 *fdx, *fdy, *fdz;
    TH2 *fdxdy, *fdxdz, *fdydz;

    // Track vertices
    TH2 *fStartXY, *fStartZX, *fStartZY;
    TH2 *fStopXY,  *fStopZX,  *fStopZY;

    TH1* fPathLength[4];

    /// This is where \ref GetChannelHist stores its histograms. Map is from block #
    caldp::AttenHistsMap fChannelMap;
    std::map<int, std::unique_ptr<caldp::AttenProfilesMap>> fChannelMapProf;
    /// This is what we put in the histogram file
    caldp::AttenHistsMap fChannelMapTotal;
    std::map<int, caldp::AttenProfilesMap> fChannelMapProfTotal;
  };

  //......................................................................
  CosmicCalib::CosmicCalib(const fhicl::ParameterSet& pset)
    : fdx(0) // This one is the signal that no histograms are created yet
  {
    reconfigure(pset);

    if(fProfileMode){
      // TODO: I don't think we can determine the detector in time...
      for(int block = 0; block < 28; ++block){
        produces<caldp::AttenProfilesMap, art::InSubRun>(TString::Format("block%d", block).Data());
      }
    }
    else{
      produces<caldp::AttenHistsMap, art::InRun>();
    }
  }

  //......................................................................
  CosmicCalib::~CosmicCalib()
  {
  }

  //......................................................................

  void CosmicCalib::reconfigure(const fhicl::ParameterSet& pset)
  {
    fPCHitsLabel = pset.get<std::string>("PCHitsLabel");
    fIncludeBelowThresholdZeros = pset.get<bool>("IncludeBelowThresholdZeros");
    fProfileMode = pset.get<bool>("ProfileMode");
    fConsolidateViewsMode = pset.get<bool>("ConsolidateViewsMode");
    fHistoOutput = pset.get<bool>("HistoOutput");
  }

  //......................................................................
  void CosmicCalib::beginJob()
  {
  }

  //......................................................................
  bool CosmicCalib::beginRun(art::Run& /*run*/)
  {
    art::ServiceHandle<geo::Geometry> geom;
    fInitEdgeX = geom->DetHalfWidth();
    fInitEdgeY = geom->DetHalfHeight();

    // The rest is making histograms. Don't do it again on a subsequent run
    if(fdx) return true;
    
    const TString viewStrArr[4]={"X","Y","muX","muY"};
    art::ServiceHandle<art::TFileService> tfs;
    bool isND=geom->DetId() == novadaq::cnv::kNEARDET;
    for(int view = geo::kX; view <= (geo::kY+2*isND); ++view){
      const TString viewStr = viewStrArr[view];
      
      double maxW;
      if(geom->DetId() == novadaq::cnv::kFARDET){
	maxW = 900;
      }
      else if(geom->DetId() == novadaq::cnv::kNEARDET){
	maxW = 250;
      }
      else{
	// NDOS (hopefully)
	maxW = (view == geo::kX) ? 250 : 150;
      }

      caldp::AttenHists v;

      v.WPE = tfs->make<TH2F>("h_WPE_in"+viewStr,
                              ";W;PE",
                              2*maxW, -maxW, +maxW,
                              1000, 0, 1000);

      v.WPE_corr = tfs->make<TH2F>("h_WPE_corr_avg_in"+viewStr,
                                   ";W;PE/cm",
                                   2*maxW, -maxW, +maxW,
                                   1000, 0, 200);

      v.WPE_corr_xy = tfs->make<TH2F>("h_WPE_corr_xy_in"+viewStr,
                                      ";W;PE/cm",
                                      2*maxW, -maxW, +maxW,
                                      1000, 0, 200);

      v.WPE_corr_xy_truth = tfs->make<TH2F>("h_WPE_corr_xy_truth_in"+viewStr,
                                            ";W;MeV/cm",
                                            2*maxW, -maxW, +maxW,
                                            1000, 0, 10);

      v.WPE_corr_z = tfs->make<TH2F>("h_WPE_corr_z_in"+viewStr,
                                     ";W;PE/cm",
                                     2*maxW, -maxW, +maxW,
                                     1000, 0, 200);

      fSumChannels[view] = v;

      fCellPE[view] = tfs->make<TH2F>("h_CellPE_in"+viewStr, ";Cell;PE",
				       105, -5, 100, 1000, 0, 1000);

      fPlanePE[view] = tfs->make<TH2F>("h_PlanePE_in"+viewStr, ";Plane;PE",
					103, -6, 200, 1000, 0, 1000);

      fPathLength[view] = tfs->make<TH1F>("h_PathLength"+viewStr,
					  ";Path length (cm)",
					  80, 0, 20);
    } // end for view

    fdx = tfs->make<TH1F>("h_dx", ";|dx|", 100, 0, 1);
    fdy = tfs->make<TH1F>("h_dy", ";|dy|", 100, 0, 1);
    fdz = tfs->make<TH1F>("h_dz", ";|dz|", 100, 0, 1);

    fdxdy = tfs->make<TH2F>("h_dxdy", ";|dx|;|dy|", 100, 0, 1, 100, 0, 1);
    fdxdz = tfs->make<TH2F>("h_dxdz", ";|dx|;|dz|", 100, 0, 1, 100, 0, 1);
    fdydz = tfs->make<TH2F>("h_dydz", ";|dy|;|dz|", 100, 0, 1, 100, 0, 1);

    fStartXY = tfs->make<TH2F>("h_StartXY", ";x;y",
			       100, -200, +200, 100, -250, +250);
    fStartZX = tfs->make<TH2F>("h_StartZX", ";z;x",
			       100, -50, +900, 100, -200, +200);
    fStartZY = tfs->make<TH2F>("h_StartZY", ";z;y",
			       100, -50, +900, 100, -250, +250);

    fStopXY = tfs->make<TH2F>("h_StopXY", ";x;y",
			      100, -200, +200, 100, -250, +250);
    fStopZX = tfs->make<TH2F>("h_StopZX", ";z;x",
			      100, -50, +900, 100, -200, +200);
    fStopZY = tfs->make<TH2F>("h_StopZY", ";z;y",
			      100, -50, +900, 100, -250, +250);

    return true;
  }

  //......................................................................
  bool CosmicCalib::endSubRun(art::SubRun& sr)
  {
    if(fProfileMode){
      for(int block = 0; block < 28; ++block){
	auto it = fChannelMapProf.find(block);
	if(it == fChannelMapProf.end()) continue;

	if(fHistoOutput){
	  // Accumulate a running total for the histogram file
	  auto ittot = fChannelMapProfTotal.find(block);
	  if(ittot == fChannelMapProfTotal.end())
	    fChannelMapProfTotal.insert(std::make_pair(block, *it->second));
	    else
	      ittot->second += *it->second;
	}

	sr.put(std::move(it->second), TString::Format("block%d", block).Data());
      }

      fChannelMapProf.clear();
   }
   else{
      std::unique_ptr<caldp::AttenHistsMap> cm(new caldp::AttenHistsMap(fChannelMap));
      sr.put(std::move(cm));

      // Accumulate a running total for the histogram file
      fChannelMapTotal += fChannelMap;
      // Don't double count for the output products
      fChannelMap.Clear();
    }

    return true;
  }

  //......................................................................
  bool CosmicCalib::filter(art::Event& evt)
  {
    art::ServiceHandle<geo::Geometry> geom;

    art::Handle< std::vector< caldp::PCHit > > pchit_avg;
    evt.getByLabel(fPCHitsLabel, "pathQualAvg",pchit_avg);

    art::Handle< std::vector< caldp::PCHit > > pchit_xy;
    evt.getByLabel(fPCHitsLabel, "pathQualXY",pchit_xy);

    art::Handle< std::vector< caldp::PCHit > > pchit_xy_thresh;
    if(fIncludeBelowThresholdZeros)
      evt.getByLabel(fPCHitsLabel, "pathQualXYThresh",pchit_xy_thresh);

    art::Handle< std::vector< caldp::PCHit > > pchit_z;
    evt.getByLabel(fPCHitsLabel, "pathQualZ",pchit_z);

    for (unsigned int idx = 0; idx < pchit_avg->size(); idx++){

      caldp::PCHit current_hit = pchit_avg->at(idx);
      double w = current_hit.W();
      double path = current_hit.Path();
      double pe = current_hit.PE();
      // Unused
      //      double mev = current_hit.TrueMeV();
      int plane = current_hit.Plane();
      int cell = current_hit.Cell();
      int firstPlane=(geom->FirstPlaneInMuonCatcher());
      int isMuonCatcher=firstPlane<=plane;
      const int view = geom->Plane(plane)->View()+(2*isMuonCatcher);
      //if(isMuonCatcher){
      //std::cout<<"First Plane: "<<firstPlane<<std::endl;
      //std::cout<<"Plane: "<<plane<<std::endl;
      //}
      if(path > 10) continue;

      fCellPE[view]->Fill(cell, pe);
      fPlanePE[view]->Fill(plane, pe);

      //      fSumChannels[view].WPE->Fill(w, pe);
      //      fSumChannels[view].WPE_corr->Fill(w, pe/path);

      if(fProfileMode){
	//        caldp::AttenProfiles& profs = GetChannelProfiles(evt.isRealData(), plane, cell);
	//        profs.WPE.Fill(w, pe);
	//        profs.WPE_corr.Fill(w, pe/path);
      }
      else{
        caldp::AttenHists hists = GetChannelHists(evt.isRealData(), plane, cell);
	hists.WPE->Fill(w, pe);
	hists.WPE_corr->Fill(w, pe/path);
      }

      fPathLength[view]->Fill(path);

    } // loop over pc_hit_avg

    for (unsigned int idx = 0; idx < pchit_xy->size(); idx++){

      caldp::PCHit current_hit = pchit_xy->at(idx);
      double w = current_hit.W();
      double path = current_hit.Path();
      double pe = current_hit.PE();
      double mev = current_hit.TrueMeV();
      int plane = current_hit.Plane();
      int cell = current_hit.Cell();
      const int view = GetView(plane,geom);
      if(path > 10) continue;

      fSumChannels[view].WPE_corr_xy->Fill(w, pe/path);
      fSumChannels[view].WPE_corr_xy_truth->Fill(w, mev/path);

      if(fProfileMode){
        caldp::AttenProfiles& profs = GetChannelProfiles(geom->DetId(), evt.isRealData(), plane, cell);
        profs.WPE_corr_xy.Fill(w, pe/path);
        profs.WPE_corr_xy_truth.Fill(w, mev/path);
      }
      else{
        caldp::AttenHists hists = GetChannelHists(evt.isRealData(), plane, cell);
        hists.WPE_corr_xy->Fill(w, pe/path);
        hists.WPE_corr_xy_truth->Fill(w, mev/path);
      }
    }// loop over pc_hit_xy

    if(fIncludeBelowThresholdZeros){
      for (unsigned int idx = 0; idx < pchit_xy_thresh->size(); idx++){

        caldp::PCHit current_hit = pchit_xy_thresh->at(idx);
        double w = current_hit.W();
        double path = current_hit.Path();
        double pe = current_hit.PE();
        assert(pe == 0);
        int plane = current_hit.Plane();
        int cell = current_hit.Cell();
	const int view = GetView(plane,geom);

        if(path > 10) continue;

        fSumChannels[view].WPE_corr_xy->Fill(w, 0);

        if(fProfileMode){
          caldp::AttenProfiles& profs = GetChannelProfiles(geom->DetId(), evt.isRealData(), plane, cell);
          profs.WPE_corr_xy.Fill(w, 0);
        }
        else{
          caldp::AttenHists hists = GetChannelHists(evt.isRealData(), plane, cell);
          hists.WPE_corr_xy->Fill(w, 0);
        }
        // TODO: think about the other hists and profs
      }// loop over pc_hit_xy
    }

    for (unsigned int idx = 0; idx < pchit_z->size(); idx++){

      caldp::PCHit current_hit = pchit_z->at(idx);
      double w = current_hit.W();
      double path = current_hit.Path();
      double pe = current_hit.PE();
      // Unused
      //      double mev = current_hit.TrueMeV();
      int plane = current_hit.Plane();
      int cell = current_hit.Cell();
      const int view = GetView(plane,geom);

      if(path > 10) continue;

      fSumChannels[view].WPE_corr_z->Fill(w, pe/path);

      if(fProfileMode){
        caldp::AttenProfiles& profs = GetChannelProfiles(geom->DetId(), evt.isRealData(), plane, cell);
        profs.WPE_corr_z.Fill(w, pe/path);
      }
      else{
        caldp::AttenHists hists = GetChannelHists(evt.isRealData(), plane, cell);
        hists.WPE_corr_z->Fill(w, pe/path);
      }

    }// loop over pc_hit_z

    // For automatic output file naming to work, need to allow at least one
    // event per subrun for now. ART issue 5427
    static unsigned int prevR = -1, prevSR = -1;
    if(evt.run() != prevR || evt.subRun() != prevSR){
      prevR = evt.run();
      prevSR = evt.subRun();
      return true;
    }
    return false;
  }

  //......................................................................
  caldp::AttenHists CosmicCalib::GetChannelHists(bool isData, int plane, int cell)
  {
    art::ServiceHandle<geo::Geometry> geom;

    const double maxW = geom->Plane(plane)->Cell(0)->HalfL()*1.15;

    if(!fConsolidateViewsMode){
      int view = GetView(plane,geom);
      plane = view;
    }

    caldp::AttenHists& hists = fChannelMap.GetHists(plane, cell);

    // If the histograms are created for this channel, just return them
    if(hists.WPE) return hists;

    // Otherwise have to construct them

    TH2F* wadc = new TH2F(TString::Format("h_WPE_%03d_%03d", plane, cell),
			  ";W;PE", 100, -maxW, +maxW, 100, 0, 1000);

    TH2F* wadc_corr = new TH2F(TString::Format("h_WPE_corr_avg_%03d_%03d",
                                               plane, cell),
                               ";W;PE/cm", 100, -maxW, +maxW, 100, 0, 200);

    TH2F* wadc_corr_xy = new TH2F(TString::Format("h_WPE_corr_xy_%03d_%03d",
                                                  plane, cell),
                                  ";W;PE/cm", 100, -maxW, +maxW, 100, 0, 200);

    TH2F* wadc_corr_xy_truth = new TH2F(TString::Format("h_WPE_corr_xy_truth_%03d_%03d", plane, cell),
                                        ";W;MeV/cm", 100, -maxW, +maxW, 100, 0, 10);


    TH2F* wadc_corr_z = new TH2F(TString::Format("h_WPE_corr_z_%03d_%03d", plane, cell),
                                 ";W;PE/cm", 100, -maxW, +maxW, 100, 0, 200);

    hists.WPE = wadc;
    hists.WPE_corr = wadc_corr;
    hists.WPE_corr_xy = wadc_corr_xy;
    hists.WPE_corr_xy_truth = wadc_corr_xy_truth;
    hists.WPE_corr_z = wadc_corr_z;

    return hists;
  }

  //......................................................................
  caldp::AttenProfiles& CosmicCalib::
  GetChannelProfiles(novadaq::cnv::DetId det,
		     bool isData, int plane, int cell)
  {
    // MC only varies for these reasons. Consolidate due to low stats.
    if(fConsolidateViewsMode){
      art::ServiceHandle<geo::Geometry> geom;
      int view = GetView(plane,geom);
      plane = view;
    }

    double maxW;
    if(det == novadaq::cnv::kFARDET) maxW = 800;
    else if(det == novadaq::cnv::kNEARDET) maxW = 250;
    else maxW = 250; // Don't bother to make non-square NDOS profiles

    const int block = plane/32;
    if(fChannelMapProf.find(block) == fChannelMapProf.end()) 
      fChannelMapProf.emplace(std::make_pair(block, std::unique_ptr<caldp::AttenProfilesMap>(new caldp::AttenProfilesMap(-maxW, +maxW))));
    return fChannelMapProf.find(block)->second->GetProfiles(plane, cell);
  }
  //......................................................................
  int CosmicCalib::GetView(int plane, art::ServiceHandle<geo::Geometry> geom){//get the view for the current combination of view and plane (x,y,mux, or muy)
    int view=0;
    if(geom->DetId() == novadaq::cnv::kNEARDET){
      int firstPlane=geom->FirstPlaneInMuonCatcher();
      int isMuonCatcher=firstPlane <= plane;
      view = geom->Plane(plane)->View()+(2*isMuonCatcher);
    }
    else{
      view = geom->Plane(plane)->View();
    } 
    return view;
  }
  //......................................................................
  void CosmicCalib::endJob()
  {
    if(!fHistoOutput) return;

    art::ServiceHandle<art::TFileService> tfs;

    for(auto& it: fChannelMapTotal.GetAllHistsByPlane()){
      const int plane = it.first;
      const std::vector<TH2F*>& hs = it.second;

      art::TFileDirectory dir = tfs->mkdir(Form("plane_%03d", plane));

      for(TH2F* h: hs) dir.make<TH2F>(*h);
    }

    for(int block = 0; block < 28; ++block){
      auto itmap = fChannelMapProfTotal.find(block);
      if(itmap == fChannelMapProfTotal.end()) continue;
      for(auto& it: itmap->second.GetAllProfilesByPlaneAndCell()){
        const int plane = it.first;
        std::cout << plane << std::endl;
        if(it.second.empty()) continue;

        art::TFileDirectory dirBlock = tfs->mkdir(Form("block_%02d", plane/32));
        art::TFileDirectory dirPlane = dirBlock.mkdir(Form("plane_%03d", plane));

        for(auto& it2: it.second){
          const int cell = it2.first;
          const std::vector<TH1F*>& hs = it2.second;
          if(hs.empty()) continue;

          art::TFileDirectory dirMod = dirPlane.mkdir(Form("module_%02d", cell/32));
          art::TFileDirectory dirCell = dirMod.mkdir(Form("cell_%03d", cell));

          for(TH1F* h: hs) dirCell.make<TH1F>(*h);
        }
      }
    }
  }

  DEFINE_ART_MODULE(CosmicCalib);

} // end namespace
////////////////////////////////////////////////////////////////////////

