////////////////////////////////////////////////////////////////////////
// \file    FindOutliers_module.cc
// \brief   Simple module to find modules with unusual calibration constants
// \author  Christopher Backhouse - bckhouse@caltech.edu
////////////////////////////////////////////////////////////////////////

// Framework includes
#include "art/Framework/Core/EDFilter.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Principal/Run.h"
#include "art/Framework/Principal/SubRun.h"
#include "art/Framework/Services/Optional/TFileService.h"

// NOvAsoft includes
#include "Calibrator/art/Calibrator.h"
#include "Calibrator/func/AttenCurve.h"

#include "RunHistory/service/RunHistoryService.h"

#include "CMap/CMap.h"

#include "TH2.h"

namespace calib
{
  /// \brief Simple module to find modules with unusual calibration constants
  ///
  /// Run me on any file from the right detector and time period
  class FindOutliers: public art::EDFilter
  {
  public:
    explicit FindOutliers(const fhicl::ParameterSet& pset);
    ~FindOutliers();

    void reconfigure(const fhicl::ParameterSet& pset);
    virtual void beginJob();
    virtual bool filter(art::Event& evt) {return false;}
    virtual bool endSubRun(art::SubRun& sr);
    virtual bool endRun(art::Run& run);

  protected:
    TH1* fChisq;
    TH1* fPE;

    TH2* fEmptyChans[2];
    TH2* fBadFit[2];
    TH2* fLowPE[2];
    TH2* fHighPE[2];
  };

  //......................................................................
  FindOutliers::FindOutliers(const fhicl::ParameterSet& pset)
  {
    reconfigure(pset);
  }

  //......................................................................
  FindOutliers::~FindOutliers()
  {
  }

  //......................................................................
  void FindOutliers::reconfigure(const fhicl::ParameterSet& pset)
  {
  }

  //......................................................................
  void FindOutliers::beginJob()
  {
    art::ServiceHandle<art::TFileService> tfs;
    fChisq = tfs->make<TH1F>("qual", ";Fit quality;Number of channels", 1000, 0, 1);
    fPE = tfs->make<TH1F>("pe", ";Mean PE/cm at far end;Number of channels", 1000, 0, 20);

    for(int view = 0; view <= 1; ++view){
      const TString viewStr = view ? "Y" : "X";
      fEmptyChans[view] = tfs->make<TH2F>("empty"+viewStr, ";Plane;Cell", 11*32, 0, 11*64, 32*12, 0, 32*12);
      fBadFit[view] = tfs->make<TH2F>("badfit"+viewStr, ";Plane;Cell", 11*32, 0, 11*64, 32*12, 0, 32*12);
      fLowPE[view] = tfs->make<TH2F>("lowpe"+viewStr, ";Plane;Cell", 11*32, 0, 11*64, 32*12, 0, 32*12);
      fHighPE[view] = tfs->make<TH2F>("highpe"+viewStr, ";Plane;Cell", 11*32, 0, 11*64, 32*12, 0, 32*12);
    }
  }

  //......................................................................
  bool FindOutliers::endSubRun(art::SubRun& sr)
  {
    return false;
  }

  //......................................................................
  bool FindOutliers::endRun(art::Run& run)
  {
    art::ServiceHandle<calib::Calibrator> cal;
    art::ServiceHandle<geo::Geometry> geom;
    art::ServiceHandle<nova::dbi::RunHistoryService> rh;
    //    rh->LoadAPDSettings();

    std::set<geo::OfflineChan> bad;

    for(int plane = 0; plane < 11*64; ++plane){
      const geo::View_t view = geom->Plane(plane)->View();
      for(int cell = 0; cell < 32*12; ++cell){
        const geo::OfflineChan chan(plane, cell);

        const calib::AttenCurve* curve = 0;
        double maxW;
	double minW;
        cal->GetAttenCurve(plane, cell, true, curve, minW,maxW);
        if(!curve || curve->chisq == -1){
          fEmptyChans[view]->Fill(plane, cell);
          bad.insert(chan);
          continue;
        }

        fChisq->Fill(curve->chisq);

        if(curve->chisq > 0.2){
          fBadFit[view]->Fill(plane, cell);
          bad.insert(chan);
        }

        const double pe = curve->MeanPEPerCmAt(-700);

        fPE->Fill(pe);

        if(pe < 4.75){
          fLowPE[view]->Fill(plane, cell);
          bad.insert(chan);
        }
        if(pe > 11){
          fHighPE[view]->Fill(plane, cell);
          bad.insert(chan);
        }

        //        bool ok;
        //        std::cout << rh->GetGain(chan, ok) << std::endl;
      } // end for cell
    } // end for plane

    art::ServiceHandle<cmap::CMap> cmap;

    struct Key
    {
      int block, plane, mod;
      bool operator<(const Key& rhs) const
      {
        return std::make_tuple(block, plane, mod) < std::make_tuple(rhs.block, rhs.plane, rhs.mod);
      }
    };

    std::map<Key, int> badmods;

    for(geo::OfflineChan chan: bad){
      const int block = chan.Plane()/32;
      const int mod = chan.Cell()/32;

      ++badmods[{block, chan.Plane()%32, mod}];
    }

    std::map<int, std::vector<Key>> trans;
    for(auto it: badmods) trans[it.second].push_back(it.first);

    std::cout << "# bad\tblock\tplane\tmod" << std::endl;
    for(int nbad = 32; nbad >= 8; --nbad){
      const std::vector<Key>& mods = trans[nbad];
      if(mods.empty()) continue;

      for(Key mod: mods){
        std::cout << nbad << ",\t" << mod.block << ",\t" << mod.plane << ",\t" << mod.mod << std::endl;
      }
    }

    return true;
  }

  DEFINE_ART_MODULE(FindOutliers);

} // namespace
