////////////////////////////////////////////////////////////////////////
// \file    ParticleCorrections_module.cc
// \brief   Make plots of threshold effect from PCHits files
// \author  Christopher Backhouse - bckhouse@caltech.edu
////////////////////////////////////////////////////////////////////////

//C++ and C:
#include <cmath>
#include <vector>

//Framework:
#include "art/Framework/Core/EDAnalyzer.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Services/Optional/TFileService.h"

//NovaSoft:
#include "CalibrationUtils/CalibUtil.h"
#include "CalibrationDataProducts/PCHit.h"
#include "Geometry/Geometry.h"
#include "GeometryObjects/CellGeo.h"
#include "MCCheater/BackTracker.h"
#include "Simulation/FLSHitList.h"
#include "Simulation/Particle.h"

// ROOT
#include "TH2.h"

namespace calib
{
  /// Make plots of threshold effect from PCHits files
  class ParticleCorrections: public art::EDAnalyzer
  {
  public:
    explicit ParticleCorrections(const fhicl::ParameterSet& pset);
    virtual ~ParticleCorrections();

    virtual void beginJob() override;

    virtual void analyze(const art::Event& evt) override;

    virtual void reconfigure(const fhicl::ParameterSet& pset) override;

  protected:
    TH1 *fDepE, *fDepPos, *fDepMum, *fDepMup, *fDepProt, *fDepPip, *fDepPim;
    TH1 *fDepE2, *fDepPos2, *fDepMum2, *fDepMup2, *fDepProt2, *fDepPip2, *fDepPim2;

    TH1* fT;
  };

  //---------------------------------------------------------------------
  ParticleCorrections::ParticleCorrections(const fhicl::ParameterSet& pset)
    : EDAnalyzer(pset)
  {
    reconfigure(pset);
  }

  //---------------------------------------------------------------------
  ParticleCorrections::~ParticleCorrections()
  {
  }

  //---------------------------------------------------------------------
  void ParticleCorrections::reconfigure(const fhicl::ParameterSet& pset)
  {
  }

  //---------------------------------------------------------------------
  void ParticleCorrections::beginJob()
  {
    art::ServiceHandle<art::TFileService> tfs;

    fDepE = tfs->make<TH1F>("depe", "True e^{-};True energy deposit (GeV);Cells", 1000, 0, .1);
    fDepPos = tfs->make<TH1F>("deppos", "True e^{+};True energy deposit (GeV);Cells", 1000, 0, .1);
    fDepMum = tfs->make<TH1F>("depmum", "True #mu^{-};True energy deposit (GeV);Cells", 1000, 0, .1);
    fDepMup = tfs->make<TH1F>("depmup", "True #mu^{+};True energy deposit (GeV);Cells", 1000, 0, .1);
    fDepProt = tfs->make<TH1F>("depprot", "True p;True energy deposit (GeV);Cells", 1000, 0, .1);
    fDepPip = tfs->make<TH1F>("deppip", "True #pi^{+};True energy deposit (GeV);Cells", 1000, 0, .1);
    fDepPim = tfs->make<TH1F>("deppim", "True #pi^{-};True energy deposit (GeV);Cells", 1000, 0, .1);

    fDepE2 = tfs->make<TH1F>("depe2", "True e^{-};True energy deposit (GeV);Cells", 1000, 0, .1);
    fDepPos2 = tfs->make<TH1F>("deppos2", "True e^{+};True energy deposit (GeV);Cells", 1000, 0, .1);
    fDepMum2 = tfs->make<TH1F>("depmum2", "True #mu^{-};True energy deposit (GeV);Cells", 1000, 0, .1);
    fDepMup2 = tfs->make<TH1F>("depmup2", "True #mu^{+};True energy deposit (GeV);Cells", 1000, 0, .1);
    fDepProt2 = tfs->make<TH1F>("depprot2", "True p;True energy deposit (GeV);Cells", 1000, 0, .1);
    fDepPip2 = tfs->make<TH1F>("deppip2", "True #pi^{+};True energy deposit (GeV);Cells", 1000, 0, .1);
    fDepPim2 = tfs->make<TH1F>("deppim2", "True #pi^{-};True energy deposit (GeV);Cells", 1000, 0, .1);

    fT = tfs->make<TH1F>("t", "", 500, 0, 500e3);
  }

  //---------------------------------------------------------------------
  void ParticleCorrections::analyze(const art::Event& evt)
  {
    art::ServiceHandle<cheat::BackTracker> bt;
    const sim::ParticleNavigator& pnav = bt->ParticleNavigator();

    art::Handle<std::vector<sim::FLSHitList>> flsls;
    evt.getByLabel("geantgen", flsls);

    assert(flsls->size() == 1);

    const std::vector<sim::FLSHit>& flss = (*flsls)[0].fHits;

    std::map<int, std::map<geo::OfflineChan, double>> deps;

    for(const sim::FLSHit& fls: flss){
      // Worried about Michels and pion decay
      /*
      if(!pnav.IsPrimary(fls.GetTrackID())){
	std::cout << "Rejecting non-primary " << fls.GetTrackID() << " " << fls.GetPDG() << std::endl;
	continue;
      }
      */

      const sim::Particle* part = bt->TrackIDToParticle(fls.GetTrackID());
      // Only want the main particle
      if(part->Process() == "Decay") continue;

      if(!part){
	std::cout << "No particle" << std::endl;
	continue;
      }
      //      if(part->PdgCode() != fls.GetPDG()) continue;

      const int evePdg = bt->TrackIDToParticle(pnav.EveId(fls.GetTrackID()))->PdgCode();

      // Don't worry about Birks for now
      deps[evePdg/*fls.GetPDG()*/][geo::OfflineChan(fls.GetPlaneID(),
						    fls.GetCellID())] += fls.GetEdep();

      // Make sure we don't have Michels etc
      if(evePdg/*fls.GetPDG()*/ == 11) fT->Fill(fls.GetEntryT());
    }

    for(auto it: deps[ +11]) fDepE   ->Fill(it.second);
    for(auto it: deps[ -11]) fDepPos ->Fill(it.second);
    for(auto it: deps[ +13]) fDepMum ->Fill(it.second);
    for(auto it: deps[ -13]) fDepMup ->Fill(it.second);
    for(auto it: deps[2212]) fDepProt->Fill(it.second);
    for(auto it: deps[+211]) fDepPip ->Fill(it.second);
    for(auto it: deps[-211]) fDepPim ->Fill(it.second);

    for(auto it: deps[ +11]) fDepE2   ->Fill(it.second, it.second);
    for(auto it: deps[ -11]) fDepPos2 ->Fill(it.second, it.second);
    for(auto it: deps[ +13]) fDepMum2 ->Fill(it.second, it.second);
    for(auto it: deps[ -13]) fDepMup2 ->Fill(it.second, it.second);
    for(auto it: deps[2212]) fDepProt2->Fill(it.second, it.second);
    for(auto it: deps[+211]) fDepPip2 ->Fill(it.second, it.second);
    for(auto it: deps[-211]) fDepPim2 ->Fill(it.second, it.second);
  }

  DEFINE_ART_MODULE(calib::ParticleCorrections);

} //end namespace
