///////////////////////////////////////////////////////////////////////
// \file    StopperSelection.cxx
// \brief   Selects stopping cosmics
// \author  bckhouse@caltech.edu, bays@caltech.edu
////////////////////////////////////////////////////////////////////////

#include "art/Framework/Core/EDFilter.h"
#include "art/Framework/Core/FindManyP.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"

// NOvASoft includes
#include "MCCheater/BackTracker.h"
#include "MichelEFilter/MichelECluster.h"
#include "RecoBase/CellHit.h"
#include "RecoBase/Track.h"
#include "Geometry/LiveGeometry.h"
#include "NovaDAQConventions/DAQConventions.h"
#include "Utilities/AssociationUtil.h"

namespace calib
{
  /// Selects stopping cosmics
  class StopperSelection: public art::EDFilter
  {
  public:
    explicit StopperSelection(const fhicl::ParameterSet& pset);
    ~StopperSelection();

    virtual bool filter(art::Event& evt);

    void reconfigure(const fhicl::ParameterSet& pset);

    void beginJob();
    void endJob();

  protected:
    bool IsContainedStopper(const rb::Track* trk) const;
    bool IsTrulyContainedStopper(const rb::Track* trk) const;

    std::string fMichelLabel;
    std::string fTrackLabel;
    bool        fApplyFilter;
    double      fMinDistToEdge;     ///< minimum distance in cm to edge of detector
    double      fMaxElectronEnergy; ///< maximum energy for the electron
    int         fTable[2][2];

    art::ServiceHandle<geo::LiveGeometry> fLiveGeom;

  };

  //......................................................................
  StopperSelection::StopperSelection(const fhicl::ParameterSet& pset)
  {
    produces< std::vector<rb::Track> >();
    produces< art::Assns<rb::Track, rb::Cluster> >();

    reconfigure(pset);

    for(int i = 0; i < 2; ++i) for(int j = 0; j < 2; ++j) fTable[i][j] = 0;
  }

  //......................................................................
  StopperSelection::~StopperSelection()
  {
  }

  //......................................................................
  void StopperSelection::reconfigure(const fhicl::ParameterSet& pset)
  {
    fMichelLabel       = pset.get< std::string >("MichelLabel"              );
    fTrackLabel        = pset.get< std::string >("TrackLabel"               );
    fApplyFilter       = pset.get< bool        >("ApplyFilter"              );
    fMinDistToEdge     = pset.get< double      >("MinDistToEdge",     50.   );
    fMaxElectronEnergy = pset.get< double      >("MaxElectronEnergy", 0.035);

    return;
  }

  //......................................................................
  void StopperSelection::beginJob()
  {
  }

  //......................................................................
  bool StopperSelection::filter(art::Event& evt)
  {
    std::unique_ptr<std::vector<rb::Track> > outTracks(new std::vector<rb::Track>);
    std::unique_ptr<art::Assns<rb::Track, rb::Cluster> > assns(new art::Assns<rb::Track, rb::Cluster>);

    art::Handle< std::vector<rb::Track> > trkInfo;
    evt.getByLabel(fTrackLabel,  trkInfo);

    art::FindManyP<rb::Cluster> fmpslc(trkInfo, evt, fTrackLabel);
    art::FindManyP<me::MichelECluster> fmpMEs(trkInfo, evt, fMichelLabel);

    for (unsigned int trkIdx = 0; trkIdx < trkInfo->size(); ++trkIdx){
      const rb::Track* track = &(trkInfo->at(trkIdx));
      bool isStopper = IsContainedStopper(track);
      bool hasME = (fmpMEs.at(trkIdx).size() == 0) ? false : true;
      if (!isStopper) continue;
      if (hasME)
        outTracks->push_back(*track);
      if (!evt.isRealData()){
        const bool trueStopper = IsTrulyContainedStopper(track);
	++fTable[hasME][trueStopper];
      }
      // Find the slice the original track was associated to
      if (hasME){
        const std::vector<art::Ptr<rb::Cluster> > slices = fmpslc.at(trkIdx);
        assert(slices.size() == 1);
        util::CreateAssn(*this, evt, *outTracks, slices[0], *assns);
      }
    }

    const bool ret = !outTracks->empty();
    evt.put(std::move(outTracks));
    evt.put(std::move(assns));
    if(fApplyFilter) return ret; else return true;
  }

  //......................................................................
  void StopperSelection::endJob()
  {
    if(fTable[0][0]+fTable[1][0] == 0 || fTable[0][1]+fTable[1][1] == 0) return;
    std::cout << "Stopper efficiency table. x-axis selection, y-axis truth:" << std::endl
    << fTable[0][0] << "\t" << fTable[1][0] << "\t" << .1*(1000*fTable[0][0]/(fTable[0][0]+fTable[1][0])) << "%" << std::endl
    << fTable[0][1] << "\t" << fTable[1][1] << "\t" << .1*(1000*fTable[1][1]/(fTable[0][1]+fTable[1][1])) << "%" << std::endl
    << .1*(1000*fTable[0][0]/(fTable[0][0]+fTable[0][1])) << "%\t" << .1*(1000*fTable[1][1]/(fTable[1][0]+fTable[1][1])) << "%" << std::endl
    << "Purity = " << .1*(1000*fTable[1][1]/(fTable[1][0]+fTable[1][1])) << "%" << std::endl
    << "Efficieny = " << .1*(1000*fTable[1][1]/(fTable[0][1]+fTable[1][1])) << "%" << std::endl;
  }

  //......................................................................
  bool StopperSelection::
  IsContainedStopper(const rb::Track* trk) const
  {
    // Start and end don't mean much, pick whichever is lowest in y.
    const TVector3 v1 = trk->Start();
    const TVector3 v2 = trk->Stop();
    const TVector3 stop = (v1.Y() < v2.Y()) ? v1 : v2;

    if(fLiveGeom->DistToEdgeXY(stop) < fMinDistToEdge ||
       fLiveGeom->DistToEdgeZ(stop)  < fMinDistToEdge ) return false;
    if(v1.X() == v2.X() || v1.Y() == v2.Y())            return false; // dx or dy = 0 means bad fit

    return true;
  }

  //......................................................................
  bool StopperSelection::IsTrulyContainedStopper(const rb::Track* trk) const
  {
    art::ServiceHandle<cheat::BackTracker> bt;

    std::vector<const sim::Particle*> parts = bt->HitsToParticle(trk->AllCells());
    if(parts.empty()) return false;

    const sim::Particle* part = parts.front();

    const TVector3 stop = part->EndPosition().Vect();

    if(fLiveGeom->DistToEdgeXY(stop) < 0.1 ||
       fLiveGeom->DistToEdgeZ(stop)  < 0.1 ) return false;
    if(std::abs(part->PdgCode())!=13)             return false;

    return true;
  }

  DEFINE_ART_MODULE(StopperSelection);

} // end namespace
////////////////////////////////////////////////////////////////////////

