////////////////////////////////////////////////////////////////////////
// Class:       CosmicTrends
// Module Type: analyzer
// File:        CosmicTrends_module.cc
//
// Generated at Tue Aug 21 13:38:36 2012 by Brian_Rebel using artmod
// from art v1_00_11.
////////////////////////////////////////////////////////////////////////
#ifndef CosmicTrends_h
#define CosmicTrends_h

#include <map>

#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Services/Optional/TFileService.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Core/EDAnalyzer.h"
#include "fhiclcpp/ParameterSet.h"
#include "messagefacility/MessageLogger/MessageLogger.h"

#include "Geometry/Geometry.h"
#include "GeometryObjects/PlaneGeo.h"
#include "GeometryObjects/CellGeo.h"
#include "RecoBase/Track.h"
#include "MCCheater/BackTracker.h"
#include "RunHistory/service/RunHistoryService.h"

#include "TStopwatch.h"
#include "TH1.h"
#include "TH2.h"
#include "TVector3.h"

namespace calib {
  class StopperThreshold;  

  enum _oneDHists{
  };

  enum _twoDHists{
  };

}

class calib::StopperThreshold : public art::EDAnalyzer {
public:
  explicit StopperThreshold(fhicl::ParameterSet const & p);
  virtual ~StopperThreshold();

  virtual void analyze(art::Event const & e);

  virtual void beginRun(art::Run const&);
  virtual void reconfigure(fhicl::ParameterSet const & p);

private:

  void FindTrajectoryPoints(art::Ptr<rb::Track>                                 const& track,
			    std::set<TVector3, 
				     bool(*)(TVector3 const&, TVector3 const&)>      & trajPoints);

  void CellsInWindow(art::Ptr<rb::Track>                                  const& track,
		     std::set<TVector3, 
			      bool(*)(TVector3 const&, TVector3 const&) > const& trajPoints,
		     std::vector< art::Ptr<rb::CellHit> >                      & xCells, 
		     std::vector< art::Ptr<rb::CellHit> >      		       & yCells);

  art::ServiceHandle<geo::Geometry> fGeo;             ///< handle to the geometry
  std::string                       fTrackLabel;      ///< label of module creating the tracks we use       
  double       	     		    fMinDistFromEnd;  ///< minimum distance from end of track to use a point
  double       	     		    fMaxDistFromEnd;  ///< maximum distance from end of track to use a point
  std::vector<TH2D*> 		    fXModuleADCVsW;   ///< ADC vs W for X modules
  std::vector<TH2D*> 		    fYModuleADCVsW;   ///< ADC vs W for Y modules

};

bool cellSortZ(art::Ptr<rb::CellHit> const& a, art::Ptr<rb::CellHit> const& b)
{
  return a->Plane() < b->Plane();
}

bool ltTVec(TVector3 const& a, TVector3 const& b)
{
  return a.Z() < b.Z();
}

//---------------------------------------------------------------------
calib::StopperThreshold::StopperThreshold(fhicl::ParameterSet const & p)
  : EDAnalyzer(p)
{
  this->reconfigure(p);
}

//---------------------------------------------------------------------
calib::StopperThreshold::~StopperThreshold()
{
}

//---------------------------------------------------------------------
void calib::StopperThreshold::beginRun(art::Run const&)
{
  if(fXModuleADCVsW.size() > 0) return;

  art::ServiceHandle<art::TFileService> tfs;
  // use the geometry to get the number of cells for planes in each view
  // also get the total number of planes

  unsigned int planeInView = *(fGeo->GetPlanesByView(geo::kX).begin());  
  double xCells = fGeo->Plane(planeInView)->Ncells();

  planeInView = *(fGeo->GetPlanesByView(geo::kY).begin());  
  double yCells = fGeo->Plane(planeInView)->Ncells();

  double maxCells    = std::max(yCells, xCells);
  int    modules     = maxCells/32.;
  
  // the near and far detectors are basically square, so just use the larger
  double detHalfWidth  = std::max(fGeo->DetHalfWidth(), fGeo->DetHalfHeight());
  int    wBins         = detHalfWidth/5;

  TString name = "";
  for(int i = 0; i < modules; ++i){
    name = "YModule" + i;
    fYModuleADCVsW.push_back(tfs->make<TH2D>("ADCVsW"+name, ";Module;W (cm)", 500, 0, 5000, wBins, -detHalfWidth, detHalfWidth));
    name = "XModule" + i;
    fXModuleADCVsW.push_back(tfs->make<TH2D>("ADCVsW"+name, ";Module;W (cm)", 500, 0, 5000, wBins, -detHalfWidth, detHalfWidth));
  } // end loop to make module histograms

  return;
}

//---------------------------------------------------------------------
void calib::StopperThreshold::reconfigure(fhicl::ParameterSet const & p)
{
  fTrackLabel     = p.get< std::string  >("TrackLabel"          );
  fMinDistFromEnd = p.get< double       >("MinDistFromEnd", 100.);
  fMaxDistFromEnd = p.get< double       >("MaxDistFromEnd", 200.);

  return;
}

//---------------------------------------------------------------------
void calib::StopperThreshold::analyze(art::Event const & e)
{
 
  // get the tracks for this event
  art::Handle< art::PtrVector<rb::Track> > th;
  e.getByLabel(fTrackLabel, th);

  // loop over all the tracks
  // make a vector of those trajectory points that are between
  // fMinDistFromEnd and fMaxDistFromEnd cm from the end of the track
  for(size_t t = 0; t < th->size(); ++t){

    art::Ptr<rb::Track> const& track = (*th)[t];

    // the set of TVector3 will be ordered with smallest Z position first
    // that is OK, because the vectors of CellHits will be too
    std::set<TVector3, bool(*)(TVector3 const&, TVector3 const&)> trajPoints(ltTVec);
    std::vector<art::Ptr<rb::CellHit> > xCells;
    std::vector<art::Ptr<rb::CellHit> > yCells;

    // get the trajectory points within the window
    // from the end of the track that we are interested in
    this->FindTrajectoryPoints(track, trajPoints);

    // get the X and Y view cells that are within the bounds
    // of the window based on their z position
    this->CellsInWindow(track, trajPoints, xCells, yCells);

    // now connect the cells to the trajectory points
    // using a set of TVector3 trajectory points is 
    // the way to go as you can use the lower_bound and 
    // upper_bound methods to find the bounds in z around 
    // the center z position of each cell

    // loop over the cells in each view and figure out
    // - ADC deposited for a normally incident muon, given by CellHit::ADC()/cos(theta_z)
    // - which module in a plane the cell is in, given by cell # mod 32
    // - w position of the hit, given by the trajectory

    // Fill histograms with this track, one histogram for each module in 
    // each view
    
    
  }// end loop over tracks in this event

  return;
}

//------------------------------------------------------------------------------
void calib::StopperThreshold::FindTrajectoryPoints(art::Ptr<rb::Track>                              const& track,
						   std::set<TVector3, 
							    bool(*)(TVector3 const&, TVector3 const&)>   & trajPoints)
{

  double totalDist = 0.;

  // loop over the trajectory points in reverse so that we know
  // when we are between fMinDistFromEnd and fMaxDistFromEnd
  for(size_t tp = track->NTrajectoryPoints()-1; tp > 1; --tp){
    
    totalDist += (track->TrajectoryPoint(tp) - track->TrajectoryPoint(tp - 1)).Mag();
    
    if(totalDist > fMinDistFromEnd && totalDist < fMaxDistFromEnd)
      trajPoints.insert(track->TrajectoryPoint(tp));
    
    // no need to keep going over the trajectory points if we pass fMaxDistFromEnd
    if(totalDist > fMaxDistFromEnd) break;
    
  }// end loop over track trajectory points
  
  return;
}

//______________________________________________________________________________
void calib::StopperThreshold::CellsInWindow(art::Ptr<rb::Track>                                 const& track,
					    std::set<TVector3, 
						     bool(*)(TVector3 const&, TVector3 const&)> const& trajPoints,
					    std::vector< art::Ptr<rb::CellHit> >                     & xCells,
					    std::vector< art::Ptr<rb::CellHit> >      		     & yCells)
{  
  // loop over the cells in the track and select those that are within the window of the 
  // trajectory points
  float minZ = trajPoints.begin()->Z();
  float maxZ = trajPoints.rbegin()->Z();
  float cellHalfDepth = fGeo->Plane(0)->Cell(0)->HalfD();
  if(maxZ < minZ) std::swap(minZ, maxZ);

  // let the range extend half a plane on either side
  // of the trajectory, as most trajectory points are 
  // taken in the center of the plane
  minZ -= cellHalfDepth;
  maxZ += cellHalfDepth;

  double xyz[3]= {0.};
  for(auto ch : track->AllCells()){
    fGeo->Plane(ch->Plane())->Cell(ch->Cell())->GetCenter(xyz);

    if(xyz[2] > minZ && xyz[2] < maxZ){
      if     (ch->View() == geo::kX) xCells.push_back(ch);
      else if(ch->View() == geo::kY) yCells.push_back(ch);
    } // end if the cell is within the allowed range in z
  }// end loop over cells in the detector

  // sort the cells in z position with cells coming 
  // from lower Z going at the front of the vector
  std::sort(xCells.begin(), xCells.end(), cellSortZ);
  std::sort(yCells.begin(), yCells.end(), cellSortZ);
  
  return;
}

DEFINE_ART_MODULE(calib::StopperThreshold)

#endif /* StopperThreshold_h */
