// \file    DriftResponse.h
// \brief   A DriftResponse entry contains Drift Calibration information for an individual subrun and offline channel
// \version $Id: DriftResponse.h,v 1.0 2013/02/1 22:06:10 rtoner Exp $  
// \author  $Author: rtoner $
// \date    $Date: 2013/02/01 22:06:10 $

#ifndef DRIFTRESPONSE_H
#define DRIFTRESPONSE_H

#include "art/Persistency/Provenance/Timestamp.h"
#include <map>
#include <stdint.h>
#include <iosfwd>

namespace caldp
{

  struct MeanStruct
    {
      
      double  mean;     ///Mean response
      double  meanerr;  ///Error on mean response
      double  rms;      ///Error on mean response
      int     nhits;    ///Number of hits in mean calculation
      
      double  Mean    () const {return mean;    }; ///Return mean	      
      double  MeanErr () const {return meanerr; }; ///Return error on mean
      double  RMS     () const {return rms;     }; ///Return rms of distribution
      int     NHits   () const {return nhits;   }; ///Return number of hits in distribution
   
      void SetResponse    ( double fMean, double fMeanErr, double fRMS, int fNHits  ); ///Set a single channel

      
  };///Structure to hold drift response in a single channel at a single time

  class DriftResponse
  {
  public:

    DriftResponse();		  
    DriftResponse(int fRun, uint32_t fStartTime, uint32_t fEndTime);
    
    ~DriftResponse();
    
#ifndef __GCCXML__
    int        Run      () const {return run;       }; ///Return run number  
    uint32_t   StartTime() const {return startTime; }; ///Return subrun start time
    uint32_t   EndTime  () const {return endTime;   }; ///Return subrun end time

    void AddChannelResponse  ( int fOffChan, double fMean, double fMeanErr, double fRMS, int fNHits); ///Function to add response value for single channel;

    double     Mean         (int fOffChan) const; ///Return mean for given offline channel
    double     MeanErr      (int fOffChan) const; ///Return mean error for given offline channel
    double     RMS          (int fOffChan) const; ///Return rms for given offline channel
    double     NHits        (int fOffChan) const; ///Return number of hits for given offline channel

    //RBTNOTE: ADD THREE OF THESE?
    bool       CheckForChannel ( int fOffChan ) const; ///Check for existence of channel in DriftResponse 
    bool       GetChannelResponse  ( int fOffChan, double &fMean, double &fMeanErr, double &fRMS, int &fNHits) const; 

    std::map<int, MeanStruct > const& ReturnMap () const { return meanMap; };

    friend std::ostream& operator << (std::ostream& o, const DriftResponse& dr);

#endif

  private:    

    int      run      ;///Run number
    uint32_t startTime;///Start time of the run period 
    uint32_t endTime  ;///End time of the run period 

    std::map<int, MeanStruct > meanMap  ;///Map of offline channel number to MeanInfo struct information

  };
  
} // end namespace

#endif // DRIFTRESPONSE_H
