////////////////////////////////////////////////////////////////////////
// \file    CalibUtil.cxx
// \brief   Functions of general use in calibration
// \author  Christopher Backhouse - bckhouse@caltech.edu
////////////////////////////////////////////////////////////////////////

#include "CalibrationUtils/CalibUtil.h"

// Framework includes
#include "art/Framework/Services/Registry/ServiceHandle.h"

// NOvASoft includes
#include "ChannelInfo/BadChanList.h"
#include "GeometryObjects/CellGeo.h"
#include "Geometry/Geometry.h"
#include "GeometryObjects/Geo.h"

namespace calib
{
  /// Provide a simple utility function for our sets
  struct ChannelSet: public std::set<geo::OfflineChan>
  {
    bool CellExists(int plane, int cell) const
    {
      return count(geo::OfflineChan(plane, cell));
    }
  };

  //......................................................................
  /// Internal helper for \ref BestPathEstimates. Takes account of bad channels
  bool HasXYAdjacents(int plane, int cell, const ChannelSet& hmap)
  {
    art::ServiceHandle<geo::Geometry> geom;
    art::ServiceHandle<chaninfo::BadChanList> bcl;

    // Look first below and then above this cell
    for(int step = -1; step <= +1; step += 2){
      int curCell = cell;
      while(true){
        curCell += step;

        // Ran out of detector
        if(curCell < 0 ||
           curCell >= int(geom->Plane(plane)->Ncells())) return false;

        // If this channel is bad we can keep on looking in this direction
        if(bcl->IsBad(plane, curCell)) continue;

        // If there's a hit this side move on to the other side. If there's
        // isn't then this cell fails the test.
        if(hmap.CellExists(plane, curCell))
          break;
        else
          return false;

      } // end while
    } // end for step
    return true;
  }

  //......................................................................
  // Internal helper for \ref BestPathEstimates. Needs to take account of bad
  // channels.
  bool HasZAdjacents(int plane, int cell,
                     const ChannelSet& hmap,
                     const art::ServiceHandle<geo::Geometry> geom)
  {
    return (!hmap.CellExists(plane, cell-1) &&
            !hmap.CellExists(plane, cell+1) &&
            hmap.CellExists(geom->NextPlaneInView(plane, +1), cell) &&
            hmap.CellExists(geom->NextPlaneInView(plane, -1), cell));
  }

  //......................................................................
  std::vector<double> BestPathEstimates(const std::vector<geo::OfflineChan>& chans,
					const std::map<unsigned int, TVector3>& directions,
                                        std::vector<EPathQuality>& quals,
                                        const std::vector<geo::OfflineChan>& extras)
  {
    const int trkN = chans.size();
    const unsigned int extrasN = extras.size();
    const int hitMax = trkN+extrasN;

    std::vector<double> paths;
    paths.resize(hitMax);

    quals.clear();
    quals.resize(hitMax);

    art::ServiceHandle<geo::Geometry> geom;

    ChannelSet hmap;
    for(int i = 0; i < trkN; ++i) hmap.insert(chans[i]);
    hmap.insert(extras.begin(), extras.end());


    // Look for ones that have perpendicular hits either side
    for(int hitIdx = 0; hitIdx < hitMax; ++hitIdx){
      int plane, cell;
      geo::View_t view;
      if(hitIdx < trkN){
        plane = chans[hitIdx].Plane();
        cell = chans[hitIdx].Cell();
        view = geom->Plane(plane)->View();
      }
      else{
        plane = extras[hitIdx-trkN].Plane();
        cell  = extras[hitIdx-trkN].Cell();
        view  = geom->Plane(plane)->View();
      }

      EPathQuality qual = kAverage;

      // Mark as xy if it satisfies the criteria. Or if not then maybe as
      // z. Otherwise falls back to the default kAverage.
      if(HasXYAdjacents(plane, cell, hmap)){
        qual = kXYAdjacents;
      }
      else{
        if(HasZAdjacents(plane, cell, hmap, geom)){
          qual = kZAdjacents;
        }
      }

      quals[hitIdx] = qual;

      // get the direction for this plane
      TVector3 dr = calib::DirectionAtPlane(directions, plane);

      const double dx = std::abs(dr.X());
      const double dy = std::abs(dr.Y());
      const double dz = std::abs(dr.Z());
      if(std::isnan(dx) || std::isnan(dy) || std::isnan(dz) ||
         (view == geo::kX && dx == 0) ||
         (view == geo::kY && dy == 0) ||
         (qual == kZAdjacents && dz == 0)){
        paths[hitIdx] = 0;
        quals[hitIdx] = kBadTrack;
        continue;
      }

      const geo::CellGeo* cellgeo = geom->Plane(plane)->Cell(cell);
      const double width = 2*cellgeo->HalfW();
      const double depth = 2*cellgeo->HalfD();

      switch(qual){
      case kXYAdjacents:
        paths[hitIdx] = (view == geo::kX) ? width/dx : width/dy;
        break;
      case kZAdjacents:
        paths[hitIdx] = depth/dz;
        break;
      case kAverage:
        paths[hitIdx] = geo::AverageCellPathLength(view, dx, dy, dz);
        break;
      default:
        assert(0 && "Shouldn't happen");
      } // end switch
    } // end for hitIdx

    return paths;
  }

  //......................................................................
  TVector3 const DirectionAtPlane(const std::map<unsigned int, TVector3>& directions,
				  unsigned int plane)
  {
    TVector3 dr;

    if(directions.count(plane) > 0){
      dr = directions.find(plane)->second;
    }
    else{

      // the plane is not in the map, so now need to 
      // determine the direction of the track in z, 
      // and then use that result to determine whether
      // to use the upper_bound or lower_bound methods
      // remember that maps are ordered by the < operator
      // of the key, in this case an unsigned int
      // if the desired plane is outside the bounds of
      // the map, just take the end point vector of the
      // map in the direction of travel.
      
      auto ubitr = directions.upper_bound(plane);
      auto lbitr = directions.lower_bound(plane);
    
      // the direction in z better not change or
      // this track has problems
      float dZ = directions.begin()->second.Z();
      if( dZ > 0.){
	if( ubitr != directions.end() )
	  dr = ubitr->second;
	else
	  dr = directions.rbegin()->second;
      }
      if( dZ < 0. ){
	if( lbitr != directions.end() )
	  dr = lbitr->second;
	else
	  dr = directions.begin()->second;
      }
    } // end no key corresponding to plane in the map
    
    return dr;
  }
    
  //......................................................................
  void FindBelowThresholdCalibCandidates(const std::vector<geo::OfflineChan>& trkchans,
					 const std::map<unsigned int, TVector3>& directions,
                                         std::vector<EPathQuality>& quals,
                                         std::vector<double>& paths,
                                         std::vector<geo::OfflineChan>& chans)
  {
    // In case user passed in junk
    quals.clear();
    paths.clear();
    chans.clear();

    const int hitMax = trkchans.size();

    art::ServiceHandle<geo::Geometry> geom;

    ChannelSet hmap;
    for(int n = 0; n < hitMax; ++n) hmap.insert(trkchans[n]);

    for(int hitIdx = 0; hitIdx < hitMax; ++hitIdx){
      const int plane = trkchans[hitIdx].Plane();
      const geo::View_t view = geom->Plane(plane)->View();

      auto const dr = calib::DirectionAtPlane(directions, plane);
      
      const double dx = std::abs(dr.X());
      const double dy = std::abs(dr.Y());
      const double dz = std::abs(dr.Z());
      
      if(std::isnan(dx) || std::isnan(dy) || std::isnan(dz)) continue;
      
      if((view == geo::kX && dx == 0) || (view == geo::kY && dy == 0)){
        continue;
      }

      // Need to search between each cell and the next hit cell in the same
      // plane
      int cell = trkchans[hitIdx].Cell();
      while(true){
        ++cell;

        // If it's hit, then it isn't thresholded. And that ends our search
        // because we'll pick it up and start there next.
        if(hmap.CellExists(plane, cell)) break;

        if(HasXYAdjacents(plane, cell, hmap)){
          const geo::CellGeo* cellgeo = geom->Plane(plane)->Cell(cell);
          const double width = 2*cellgeo->HalfW();

          quals.push_back(kXYAdjacents);
          paths.push_back((view == geo::kX) ? width/dx : width/dy);
          chans.push_back(geo::OfflineChan(plane, cell));
        }
        else{
          // Definition of the XY criteria mean that nothing else in this range
          // can work either
          break;
        }
      } // end while
    }
  }

  //-----
  void GetXYZD(geo::OfflineChan chan, double w, double *xyzd) {
    // Return (via 'xyzd' argument) position of deposition in world
    // coordinates and the distance to the readout, assuming deposition
    // has cell-local-z position = w.

    art::ServiceHandle<geo::Geometry> geom;

    UInt_t plane = chan.Plane();
    UInt_t cell  = chan.Cell();
    if ( plane >= geom->NPlanes() ) {
      throw cet::exception("BAD_PLANE_NUMBER")
        << "plane " << plane << " of " << geom->NPlanes() << " planes\n"
        << __FILE__ << ":" << __LINE__ << "\n";
    }
    const geo::PlaneGeo* p = geom->Plane(plane);
    if ( cell >= p->Ncells() ) {
      throw cet::exception("BAD_CELL_NUMBER")
        << "cell " << cell << " of " << p->Ncells()
        << " in plane " << plane << "\n"
        << __FILE__ << ":" << __LINE__ << "\n";
    }
    const geo::CellGeo* c = p->Cell(cell);

    // Should be replaced with alignment tables or whatnot.  Okay for now.

    // w is not localZ, which makes this a lot messier than just
    // c.GetCenter(xyzd,w) and c.DistToReadOut(w)
    c->GetCenter(xyzd);
    if (geom->Plane(plane)->View()==geo::kX)
      xyzd[1] = w;
    else
      xyzd[0] = w;
    Double_t localxyz[3];
    c->WorldToLocal(xyzd,localxyz);
    xyzd[3] = geom->DistToElectronics(localxyz[2], *c);
  }

} // end namespace
////////////////////////////////////////////////////////////////////////

