////////////////////////////////////////////////////////////////////////
// \file    ADCShapeFitAna_module.cc
// \brief   TODO
// \author  Christopher Backhouse - bckhouse@caltech.edu
////////////////////////////////////////////////////////////////////////

// Framework includes
#include "art/Framework/Core/EDAnalyzer.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Services/Optional/TFileService.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/Handle.h"

#include "Calibrator/func/ADCShapeFit.h"
#include "Geometry/Geometry.h"
#include "RawData/RawDigit.h"

#include "TCanvas.h"
#include "TGraph.h"
#include "TH2.h"

namespace calib
{
  /// \brief TODO
  class ADCShapeFitAna: public art::EDAnalyzer
  {
  public:
    explicit ADCShapeFitAna(const fhicl::ParameterSet& pset);
    ~ADCShapeFitAna();

    virtual void analyze(const art::Event& evt);

    void reconfigure(const fhicl::ParameterSet& pset);

    virtual void beginJob();
  protected:
    calib::ADCShapeFitTable* fShapeFit;
    TH2* fAvg;

    int fNCanvs;
  };

  //......................................................................
  ADCShapeFitAna::ADCShapeFitAna(const fhicl::ParameterSet& pset)
    : EDAnalyzer(pset), fShapeFit(0), fAvg(0), fNCanvs(0)
  {
    reconfigure(pset);
  }

  //......................................................................
  ADCShapeFitAna::~ADCShapeFitAna()
  {
  }

  //......................................................................

  void ADCShapeFitAna::reconfigure(const fhicl::ParameterSet& pset)
  {
  }

  //......................................................................
  void ADCShapeFitAna::beginJob()
  {
  }

  //......................................................................
  void ADCShapeFitAna::analyze(const art::Event& evt)
  {
    art::ServiceHandle<geo::Geometry> geom;

    const bool isND = geom->DetId() == novadaq::cnv::kNEARDET;

    if(!fShapeFit){
      std::string fname;
      cet::search_path sp("FW_SEARCH_PATH");

      if(isND){
	sp.find_file("Calibrator/adc_shape_fit_table_nd_140_4500.root", fname);
      }
      else{
	sp.find_file("Calibrator/adc_shape_fit_table_fd_380_7000.root", fname);
      }
      assert(!fname.empty());

      fShapeFit = new ADCShapeFitTable(fname, isND, !evt.isRealData());
    }

    const double sampleTime = isND ? 125 : 500; // ns
          double riseTime = (isND ?  140 : 380); // ns
    const double fallTime = (isND ? 4500 : 7000); // ns

    // ND data and MC are different, the ND data seems to have a ~140ns rise
    // time rather than 80ns. This solution isn't very scaleable, but it'll do
    // for now.
    if(isND && evt.isRealData()) riseTime = 140;

    static const double peak =
      pow(riseTime/(riseTime+fallTime),   riseTime/fallTime)-
      pow(riseTime/(riseTime+fallTime), 1+riseTime/fallTime);


    if(!fAvg){
      art::ServiceHandle<art::TFileService> tfs;
      fAvg = tfs->make<TH2F>("avg", ";TNS", 64*9, -sampleTime, 8*sampleTime, 140, -.2, 1.2);

      TGraph* nominal = tfs->make<TGraph>();

      nominal->SetPoint(0, -sampleTime, 0);
      for(double t = 0; t < 8.1*sampleTime; t += 10){
	nominal->SetPoint(nominal->GetN(), t, exp(-t/fallTime)*(1-exp(-t/riseTime)));
      }
      nominal->SetLineWidth(2);
      nominal->SetLineColor(kRed);

      nominal->Write("nominal");
    }

    art::Handle<std::vector<rawdata::RawDigit>> digcol;
    evt.getByLabel("daq", digcol);

    for(unsigned int digIdx = 0; digIdx < digcol->size(); ++digIdx){
      const art::Ptr<rawdata::RawDigit> dig(digcol, digIdx);

      if(dig->NADC() == 1) continue;

      double adcpeak, base;
      bool goodTime;
      const double tns = fShapeFit->TNS(0,
					dig->ADC(1) - dig->ADC(0),
					dig->ADC(2) - dig->ADC(0),
					dig->ADC(3) - dig->ADC(0),
					adcpeak, goodTime, base);

      base += dig->ADC(0);

      if(!goodTime) continue;

      for(unsigned int i = 0; i < dig->NADC(); ++i){
	fAvg->Fill(sampleTime*i-tns, (dig->ADC(i)-base)/adcpeak*peak);
      }

      if(fNCanvs < 100){
	++fNCanvs;

	art::ServiceHandle<art::TFileService> tfs;
	TCanvas* c = tfs->make<TCanvas>(TString::Format("trace%d", fNCanvs),
					TString::Format("trace%d", fNCanvs));
	c->cd();

	TH2* axes = new TH2F("", ";TNS;ADC",
			     100, -sampleTime, sampleTime*dig->NADC(),
			     100, dig->ADC(0)-.2*adcpeak, dig->ADC(3)+.2*adcpeak);
	axes->Draw();

	TGraph* g = new TGraph;
	for(unsigned int i = 0; i < dig->NADC(); ++i){
	  g->SetPoint(i, i*sampleTime, dig->ADC(i));
	}

	g->SetMarkerStyle(kFullCircle);
	g->Draw("p same");

	TGraph* fit = new TGraph;

	fit->SetPoint(0, -sampleTime, base);

	for(double t = tns; t < dig->NADC()*sampleTime; t += 10){
	  fit->SetPoint(fit->GetN(), t, base+adcpeak/peak*exp(-(t-tns)/fallTime)*(1-exp(-(t-tns)/riseTime)));
	}

	fit->SetLineWidth(2);
	fit->SetLineColor(kRed);
	fit->Draw("l same");

	c->Write();
      }
    } // end for digIdx
  }

  DEFINE_ART_MODULE(ADCShapeFitAna);

} // end namespace
////////////////////////////////////////////////////////////////////////

