////////////////////////////////////////////////////////////////////////
/// \file    Calibrator_service.cc
/// \brief   Service that handles calibrations for a particular
///          validity context (eg: detector+timestamp).
/// \author  Christopher Backhouse - bckhouse@caltech.edu
////////////////////////////////////////////////////////////////////////
#include "Calibrator/Calibrator.h"
#include "RecoBaseHit/CellHit.h"
#include "RecoBaseHit/RecoHit.h"

#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/SubRun.h"
#include "messagefacility/MessageLogger/MessageLogger.h"

#include "Calibrator/func/AbsCache.h"
#include "Calibrator/func/ADCShapeFit.h"
#include "Calibrator/func/AttenCurve.h"
#include "Calibrator/func/DriftCache.h"
#include "Calibrator/func/TimingCache.h"

#include "CalibrationUtils/CalibUtil.h"

#include "DAQDataFormats/NanoSliceVersionConvention.h"

#include "GeometryObjects/CellGeo.h"

#include "RunHistory/service/RunHistoryService.h"

#include "TRandom3.h"

#include <cmath>

namespace calib 
{
  // Number of points we expect to receive, to find the best peak match with
  const unsigned int kNumFineTimingADCPoints = 4; // Including the DCS origin

  //------------------------------------------------------------
  Calibrator::Calibrator(fhicl::ParameterSet const& pset, 
			 art::ActivityRegistry& reg)
    : fCurrTimeStamp(0)
    , fShapeTable(0)
    , fAttenCache(new AttenCache)
    , fDriftCache(new DriftCache)
    , fAbsCache(new AbsCache)
    , fTimingCache(new TimingCache)
    , fTimingCacheVldTimeSet(false)
    , fStatsNumCalibrated(0)
    , fStatsNumUncalibrated(0)
    , fStatsNumAveraged(0)
  {
    this->reconfigure(pset);
    reg.sPreProcessEvent.watch(this, &Calibrator::preEvent);
    reg.sPostBeginSubRun.watch(this, &Calibrator::postBeginSubRun);
    reg.sPostBeginRun.watch(this, &Calibrator::postBeginRun);
  }
  
  //------------------------------------------------------------
  Calibrator::~Calibrator()  
  {
    if(fShapeTable){
      // Expecting about 95%
      mf::LogInfo("Calibrator") << "ADCShapeFitTable hit rate was "
                                << int(100*fShapeTable->TableHitFraction())
                                << "%";
      delete fShapeTable;
    }

    // NumCalibrated includes NumAveraged
    const unsigned int numRet = fStatsNumCalibrated+fStatsNumUncalibrated;
    if(numRet > 0){
      // Hopefully around 1%
      mf::LogInfo("Calibrator")
        << "Provided " << numRet << " RecoHits, of which "
        << ((1000*fStatsNumUncalibrated)/numRet)*.1 << "% were uncalibrated and "
        << ((1000*fStatsNumAveraged)/numRet)*.1 << "% were average constants.";
    }

    delete fAttenCache;
    delete fDriftCache;
    delete fAbsCache;
    delete fTimingCache;
  }
  
  //------------------------------------------------------------
  void Calibrator::reconfigure(const fhicl::ParameterSet& pset)
  {
    // Fix up the filename by looking in the search path
    cet::search_path sp("FW_SEARCH_PATH");
    sp.find_file(pset.get< std::string >("ShapeTableFilenameFDMC"), fShapeTableFilenameFDMC);
    assert(!fShapeTableFilenameFDMC.empty());
    sp.find_file(pset.get< std::string >("ShapeTableFilenameFDData"), fShapeTableFilenameFDData);
    assert(!fShapeTableFilenameFDData.empty());
    sp.find_file(pset.get< std::string >("ShapeTableFilenameNDMC"), fShapeTableFilenameNDMC);
    assert(!fShapeTableFilenameNDMC.empty());

    sp.find_file(pset.get< std::string >("ShapeTableFilenameNDData"), fShapeTableFilenameNDData);
    assert(!fShapeTableFilenameNDData.empty());

    fAttenCache->SetConstsCSVSource(pset.get<std::string>("AttenConstsCSVPath"),
				    pset.get<std::string>("AttenPointsCSVPath"));

    fAttenCache->UseCSVsFromUPS(pset.get<bool>("UseCSVsFromUPS"));

    fAttenCache->SetTag(pset.get<std::string>("Tag"));

    // Chunked loading is only effective for non-CSV sources
    if(pset.get<std::string>("AttenPointsCSVPath").empty() && !pset.get<bool>("UseCSVsFromUPS"))
      fAttenCache->SetMaxDiblock(pset.get<int>("ChunkedLoadingMaxDiblock"));

    fAbsCache->SetCSVSource(pset.get<std::string>("AbsConstsCSVPath"));
    fAbsCache->UseCSVsFromUPS(pset.get<bool>("UseCSVsFromUPS"));
    fAbsCache->SetTag(pset.get<std::string>("Tag"));
    
    fTimingCache->SetCSVSource(pset.get<std::string>("TimingConstsCSVPath"));
    fTimingCache->UseCSVsFromUPS(pset.get<bool>("UseCSVsFromUPS"));
    fTimingCache->SetTag(pset.get<std::string>("TimingTag"));

    fUseFlatCalibOffset = pset.get<bool>("UseFlatCalibOffset");
    fFlatCalibOffset = pset.get<double>("FlatCalibOffset");

    fUseXSlopeOffset = pset.get<bool>("UseXSlopeOffset");
    fXSlopeOffset = pset.get<double>("XSlopeOffset");

    fUseYSlopeOffset = pset.get<bool>("UseYSlopeOffset");
    fYSlopeOffset = pset.get<double>("YSlopeOffset");

    fUseXSlopeFit  = pset.get<bool>("UseXSlopeFit");
    fXSlopeFitForm = pset.get<std::string>("XSlopeFitForm","");
    fXSlopeFitParams = pset.get<std::vector<float> >("XSlopeFitParams");

    fUseYSlopeFit  = pset.get<bool>("UseYSlopeFit");
    fYSlopeFitForm = pset.get<std::string>("YSlopeFitForm","");
    fYSlopeFitParams = pset.get<std::vector<float> >("YSlopeFitParams");

    fUseCellByCellCalibRandomOffset = pset.get<bool>("UseCellByCellCalibRandomOffset");
    fCellByCellCalibRandomOffsetHighResolution = pset.get<bool>("CellByCellCalibRandomOffsetHighResolution");
    fCellByCellCalibRandomOffset = pset.get<double>("CellByCellCalibRandomOffset");
    fCellByCellCalibRandomOffsetSeed = pset.get<int>("CellByCellCalibRandomOffsetSeed",24680);

    fMaskUncalibratedChannelsInMC = pset.get<bool>("MaskUncalibratedChannelsInMC");

    fMakeSinglePointCellHits = pset.get<bool>("MakeSinglePointCellHits");
    fUseTimingOffsets = pset.get<bool>("UseTimingOffsets");
    fUseDrift = pset.get<bool>("UseDrift");
    fDriftCSV = pset.get<std::string>("DriftCSV");

    if (fUseXSlopeFit){
      fXSlopeFit = new TF1("x_slope_fit", fXSlopeFitForm.c_str(), -900, 900);
      assert(fXSlopeFit);
      for (unsigned int i=0;i<fXSlopeFitParams.size();i++){
        fXSlopeFit->SetParameter(i,fXSlopeFitParams[i]);
      }
    }

    if (fUseYSlopeFit){
      fYSlopeFit = new TF1("y_slope_fit", fYSlopeFitForm.c_str(), -900, 900);
      assert(fYSlopeFit);
      for (unsigned int i=0;i<fYSlopeFitParams.size();i++){
        fYSlopeFit->SetParameter(i,fYSlopeFitParams[i]);
      }
    }
  }
  
  //------------------------------------------------------------
  void Calibrator::preEvent(art::Event const& evt)
  {
    // use the art::Timestamp to create the fCurrTrimeStamp value
    unsigned long long int tsval = evt.time().value();
    
    // taking it apart
    // the masking isn't strictly necessary *if* "long" is truly 32bits
    // but this can vary w/ compiler/platform
    const unsigned long int mask32 = 0xFFFFFFFFUL;
    time_t lup = ( tsval >> 32 ) & mask32;
    int llo = tsval & mask32;
    TTimeStamp tts(lup, llo);
    fCurrTimeStamp = tts;
  }

  //------------------------------------------------------------
  void Calibrator::postBeginSubRun(const art::SubRun& subrun)
  {

  }

  //------------------------------------------------------------
  void Calibrator::postBeginRun(art::Run const& run)
  {
    // New run. We need to update the TimingCache VldTime. But don't do it
    // until someone actually accesses TimingCache, just make a note. Avoids
    // accessing RunHistory when we don't have to, especially in case of
    // generation jobs where it will actually fail.
    fTimingCacheVldTimeSet = false;
    fCurrentRun=run.run();

    if (fUseDrift) {
      // using RunHistory to set validity time range for drift correction
      fDriftCache->SetCSVFile(fDriftCSV);
      art::ServiceHandle<nova::dbi::RunHistoryService> rh;      
      // load from either CSV or DB
      fDriftCache->SetVld(fGeom->DetId(), rh->TStart(),rh->TStop());
    }
  }


  //------------------------------------------------------------
  //   RawDigit-to-CellHit methods
  //------------------------------------------------------------
  
  rb::CellHit Calibrator::MakeCellHit(const rawdata::RawDigit *rawdigit)
  {
    // Copy of the hit that's modified to look like single-point data
    rawdata::RawDigit digSingle = *rawdigit;
    if(fMakeSinglePointCellHits && rawdigit->Version() != 0){
      digSingle.SetVersion(0);
      digSingle.fADC.clear();
      digSingle.SetADC(0, rawdigit->ADC());

      // Rest of function uses modified hit in place of argument provided
      rawdigit = &digSingle;
    }

    rb::CellHit chit(*rawdigit);
    
    // new info
    chit.SetCell  (fCMap->GetCell(rawdigit));
    chit.SetPlane (fCMap->GetPlane(rawdigit));
    chit.SetView  (fGeom->Plane(chit.Plane())->View());
    bool goodTime;
    const double tns = GetTNS(rawdigit, goodTime);
 
    const daqchannelmap::lchan lchan = fCMap->Map()->encodeLChan(fGeom->DetId(),
							  chit.Plane(),
							  chit.Cell());
    const daqchannelmap::dchan dchan = fCMap->Map()->encodeDChan(lchan);

    //check if timing offsets should be added in, 
    if (fUseTimingOffsets){
      EnsureTimingCacheVldTime();
      const double offset = fTimingCache->GetOffset(fCMap->Map()->getDCM(dchan), fCMap->Map()->getDiBlock(dchan), fCurrTimeStamp.GetSec(), chit.IsRealData());
      chit.SetTNS   (tns+offset, goodTime);
    }
    else{
      chit.SetTNS   (tns, goodTime);
    }

    chit.SetPE    (GetPE(rawdigit));

    return chit;
  }

  //-----
  float Calibrator::GetPE(rb::CellHit const& ch)
  {
    return this->GetPE(&ch);
  }

  //-----
  float Calibrator::GetPE(const rawdata::RawDigit* dig)
  {
    // Look up gain from DB.  For now, must hard code the
    // number.  (This should not be put in any XML file.
    // That'll just confuse the issue.)
    const int maxadc = (1<<12)-1; // 4095
    // 1800 is the value of ADCMaxPE in ReadoutSim
    double gain = maxadc/1800.0;
    // That was the NDOS value. In FD, and ND, it's these values.
    // FEB 4.1
    if(fGeom->DetId() == novadaq::cnv::kFARDET ) gain = maxadc/2170.0;
    // FEB 5.2
    if(fGeom->DetId() == novadaq::cnv::kNEARDET) gain = maxadc/2048.0;

    // PE comes from the value determined as a side-effect of the timing fit,
    // the best-fit height of the curve when t0 is at its best value.

    // Doesn't matter whether or not timing fit suceeds, good fallbacks are in
    // place.
    bool junk;
    double peakadc;
    GetTNS(dig, junk, &peakadc);

    return peakadc/gain;
  }
  
  //-----
  float Calibrator::GetTNS(const rawdata::RawDigit* dig, bool& goodTime,
			   double* maxadc)
  {
    // Allows the rest of the function to assume maxadc is writeable
    double junk;
    if(!maxadc) maxadc = &junk;

    const int det = fGeom->DetId();

    // This is how the old MC fast-timing was encoded. Three
    // baseline-subtracted points, in non-chronological order. Treat specially
    // below.
    const bool isOldMC = (dig->IsMC() &&
			  dig->Version() == 0 &&
			  dig->NADC() == 3);

    static daqdataformats::NanoSliceVersionConvention conv;

    const unsigned int nSamples = conv.getNSamples(dig->Version());
    assert(isOldMC || dig->NADC() == nSamples);

    const double tns0 = dig->TDC()*15.625;

    // Version zero means this is a single DCS point
    if(dig->Version() == 0 && !isOldMC){
      goodTime = false;
      *maxadc = dig->ADC(); // Make sure this gets set reasonably
      // Simply convert the time stored in TDC to nanoseconds
      return tns0;
    }

    // Otherwise we can do the timing fit
    const unsigned int nPretrig = conv.getNPretriggeredSamples(dig->Version());

    // Need to have enough points to be able to see back to the baseline
    assert(isOldMC || nPretrig >= 3);
    // Also need enough points to see up to the peak
    assert(isOldMC || nSamples >= 4);

    // Create lookup table if required
    if(!fShapeTable){
      const bool isND = (det == novadaq::cnv::kNEARDET);
      std::string fname;
      if(isND){
        if(dig->IsMC()){
          fname = fShapeTableFilenameNDMC;
        }
        else{
          fname = fShapeTableFilenameNDData;
        }
      }
      else{
        if(dig->IsMC()){
          fname = fShapeTableFilenameFDMC;
        }
        else{
          fname = fShapeTableFilenameFDData;
        }
      }
      fShapeTable = new ADCShapeFitTable(fname, isND, dig->IsMC());
    }

    double tns;

    if(isOldMC){
      double junk;
      // ShapeTable wants to see the points in chronological order, but old MC
      // had ADC0 as the peak for backwards compatibility.
      tns = fShapeTable->TNS(tns0,
			     dig->ADC(1),
			     dig->ADC(2),
			     dig->ADC(0),
			     *maxadc, goodTime, junk);
    }
    else{
      double junk;
      // Figure out which sample is the baseline sample (ie the one combined
      // with the peak in the DCS calculation).
      const int baseIdx = nPretrig-3;
      // ShapeTable expects points in order, baseline subtracted
      tns = fShapeTable->TNS(tns0,
			     dig->ADC(baseIdx+1)-dig->ADC(baseIdx),
			     dig->ADC(baseIdx+2)-dig->ADC(baseIdx),
			     dig->ADC(baseIdx+3)-dig->ADC(baseIdx),
			     *maxadc, goodTime, junk);
    }

    return tns;
  }


  //------------------------------------------------------------
  //   CellHit-to-RecoHit methods
  //------------------------------------------------------------

  // For these two, calculate the misssing piece of information, then forward
  // to the protected version.
  rb::RecoHit Calibrator::MakeRecoHit(rb::CellHit const& cellhit, double w)
  {
    Double_t xyzd[4];
    GetXYZD(cellhit.OfflineChan(),w,xyzd);
    return MakeRecoHit(cellhit, w, xyzd);
  }

  //----
  rb::RecoHit Calibrator::MakeRecoHit(rb::CellHit const& cellhit, double *xyz)
  {
    double w = xyz[0];
    if(cellhit.View() == geo::kX) w = xyz[1];

    return MakeRecoHit(cellhit, w, xyz);
  }


  //----
  rb::RecoHit Calibrator::MakeRecoHit(rb::CellHit const& chit,
					double w, double *xyz)
  {
    rb::RecoHit rhit;

    rhit.SetX(xyz[0]);
    rhit.SetY(xyz[1]);
    rhit.SetZ(xyz[2]);

    rhit.SetT(GetT(chit, xyz[3]));

    // Set hit up uncalibrated in case we don't fill it below
    rhit.SetPECorr(-1);
    rhit.SetMIP   (-1);
    rhit.SetGeV   (-1);

    if(std::isnan(w)){
      mf::LogError("Calibrator")
        << "W position provided is NaN. This hit will be uncalibrated.";
    }

    const double peCorr = std::isnan(w) ? -1 : GetPECorr(chit, w);

    assert(!std::isnan(peCorr));
    assert(!std::isinf(peCorr));

    if(peCorr > 0){
      rhit.SetPECorr(peCorr);
      const double gevPerPECorr = GetPECorrToGeVScale(chit);
      if(gevPerPECorr > 0){
        rhit.SetGeV(peCorr     * gevPerPECorr);
        rhit.SetMIP(rhit.GeV() * GetGeVToMIPScale(chit));
        ++fStatsNumCalibrated;
      }
      else{
        // Uncalibrated
        rhit.SetPECorr(-1);
        ++fStatsNumUncalibrated;
      }
    }
    else{
      ++fStatsNumUncalibrated;
    }

    return rhit;
  }

  //-----
  double Calibrator::GetT(rb::CellHit const& cellhit, double dist_to_readout) {
    // This needs to be calibrated and put in a DB table
    Double_t speedOfFiberTransport = 15.3; // cm/ns, "first principles" calib.
    // Differs from c/n due to zigzag
    // paths down fiber.  But, this
    // value may be way off (10%? 30%?).
    return (cellhit.TNS()-dist_to_readout/speedOfFiberTransport);
  }

  //-----
  double Calibrator::GetPECorr(rb::CellHit const& cellhit, double w)
  {
    const double attenFactor = GetAttenScale(cellhit, w);
    if(attenFactor <= 0) return -1;

    const double driftFactor = GetDriftScale(cellhit);
    if(driftFactor <= 0) return -1;
    
    //set X or Y view sloped scaling with max/min scales at each end of the cell equal to fXSlopeOffset or
    //fYSlopeOffset
    double scale=1.0;
    if(cellhit.View() == geo::kX && fUseXSlopeOffset){
      double slopescaleperw=fXSlopeOffset/(fGeom->Plane(cellhit.Plane())->Cell(cellhit.Cell())->HalfL());
      scale=1+(w*slopescaleperw);
    }

    if(cellhit.View() == geo::kY && fUseYSlopeOffset){
      double slopescaleperw=fYSlopeOffset/(fGeom->Plane(cellhit.Plane())->Cell(cellhit.Cell())->HalfL());
      scale=1+(w*slopescaleperw);
    }

    if(cellhit.View() == geo::kX && fUseXSlopeFit){
      double fit_value = fXSlopeFit->Eval(w);
      scale = scale * fit_value;
    }

    if(cellhit.View() == geo::kY && fUseYSlopeFit){
      double fit_value = fYSlopeFit->Eval(w);
      scale = scale * fit_value;
    }

    //set a flat scaling factor to the absolute calibration
    if(fUseFlatCalibOffset){
      scale *= fFlatCalibOffset;
    }

    return cellhit.PE()*attenFactor*driftFactor*scale;
  }

  //----
  double Calibrator::GetTimeRes(rb::CellHit const& cellhit)
  {
    const int det = fGeom->DetId();
    double pe = cellhit.PE();
    //This is a fit for the timing resolution as a function of PE.  
    //For FD single point data this is derived by plotting the time difference between 
    //pairs of hits within a DCM on a cosmic track.  More information is in docdb-11571

    double stdev = 144.0; //simple assumption of 500ns/sqrt(12)
    if (det == novadaq::cnv::kFARDET) {
      if (cellhit.GoodTiming()){
        stdev = 231594.0/(2267.16+pow(pe,2.01011)) + 9.55689;   //2014-10-13 fit, multipoint readout from data, docdb-12122, with 460 ns rise time
      }
      else {
        stdev = 848357.0/(5734.75+pow(pe,2.20060)) + 142.221; //2014-06-13 fit, from data
      }
    }
    else if (det == novadaq::cnv::kNEARDET) {
      if (cellhit.GoodTiming()){
        stdev = 2229.53/(77.0043+pow(pe,1.22743)) + 4.89355; //2014-09-15 fit from data with 140 ns rise time, 4.5 us fall time
      }
      //TODO:  ND single point timing resolution needs to be updated.  Currently simulation and data are all in
      //multipoint.  For now take the FD singlpe point resolution and dived by 4 since the sampling is 4 times as fast
      else {
        stdev = 6136470.0/(185518.0+pow(pe,2.92360)) + 31.6071; //2014-09-15 singe point fit of data
      }
    }

    return stdev;

  }

  //-----
  double Calibrator::GetAttenScale(rb::CellHit const& cellhit, double w)
  {

    const calib::AttenCurve* curve = 0;
    double                   maxW  = 0.0;
    double                   minW  = 0.0;

    if(!GetAttenCurve(cellhit.Plane(), cellhit.Cell(), cellhit.IsRealData(), curve, minW,maxW)) return -1;

    if(curve->chan.Plane() >= 1000) ++fStatsNumAveraged;

    double scale=1.0;
    
    //set random scaling factor for systematic studies at either the resolution of the cell or per cm in the cell around
    //a gaussian with a sigma set at the fCellByCellCalibRandomOffset
    if(fUseCellByCellCalibRandomOffset){
      if(fCellByCellCalibRandomOffsetHighResolution){
	int wInt= (int) w;
	if(wInt==0){
	  wInt=9999;
	}
	TRandom3 *noiseGenerator = new TRandom3(fCellByCellCalibRandomOffsetSeed*(cellhit.Plane()+1)*(cellhit.Cell()+1)*wInt);
	scale=noiseGenerator->Gaus(1,fCellByCellCalibRandomOffset);
	delete noiseGenerator;
      }
      else{
	TRandom3 *noiseGenerator = new TRandom3(fCellByCellCalibRandomOffsetSeed*(cellhit.Plane()+1)*(cellhit.Cell()+1));
        scale=noiseGenerator->Gaus(1,fCellByCellCalibRandomOffset);
        delete noiseGenerator;
      }
    }

    return scale*GetAttenScale(curve, minW, maxW,w);
  }

  //-----
  double Calibrator::GetAttenScale(const calib::AttenCurve* curve, const double minW, const double maxW, double w)
  {

    if(w < minW) w = minW;
    if(w > maxW) w = maxW;

    const double mean = curve->MeanPEPerCmAt(w,fAttenCache->GetTag()); 

    if(mean <= 0) return -1;

    // The goal of this factor is to make PECorrs approximately the same size
    // as PEs. But the exact values are terrible historical accident. If we
    // decide to clean this up in future calibration versions we can do so and
    // then condition this on the tag version.
    double pecorrscale = 0;
    switch(fGeom->DetId()){
    case novadaq::cnv::kNDOS:    pecorrscale = 75.00; break;
    case novadaq::cnv::kNEARDET: pecorrscale = 37.51; break;
    case novadaq::cnv::kFARDET:  pecorrscale = 39.91; break;
    default:
      assert(0 && "Unknown detector");
    }

    const double factor = pecorrscale/mean;

    return factor;
  }

  //-----
  bool Calibrator::GetAttenCurve(const int plane, const int cell, const bool is_realdata
				 ,const calib::AttenCurve*& curve, double& minW, double& maxW) const
  {
    const geo::OfflineChan offline_channel(plane, cell);

    //make sure not to mask uncalibrated mc in perfect mc
    bool maskUncalibratedChannels = fMaskUncalibratedChannelsInMC;
    if(fCurrentRun < 1e4 || fCurrentRun >= 1e6){
      maskUncalibratedChannels = false;
    }

    bool isOldTag=fAttenCache->GetTag() == "v1" ||
      fAttenCache->GetTag() == "v2" ||
      fAttenCache->GetTag() == "v3" ||
      fAttenCache->GetTag() == "v3.1" ||
      fAttenCache->GetTag() == "v4" ||
      fAttenCache->GetTag() == "v5" ||
      fAttenCache->GetTag() == "v5.1" ||
      fAttenCache->GetTag() == "v6";
    bool isNDMCX=fGeom->DetId() == novadaq::cnv::kNEARDET && plane >= int(fGeom->FirstPlaneInMuonCatcher()) && (fGeom->Plane(plane)->View() == geo::kX);
    
    //make sure not to mask the ND muon catcher x planes after v6 where we have forced the fit to converge
    if(!(isOldTag)&&isNDMCX){
      maskUncalibratedChannels = false;
    }

    if(is_realdata)
      curve = fAttenCache->ForData(&(*fGeom), offline_channel);
    else
      curve = fAttenCache->ForMC(&(*fGeom), offline_channel, maskUncalibratedChannels);
    if(!curve) return false;
    //allow poor calibration quality in the ND muon catcher x planes after v6 where we have used LOWESS to force fits to work in spite of the attenuation fit
    if(!curve->IsCalibrated() && (isOldTag || !isNDMCX)) return false;

    // Clamp given w-coordinate into the detector to avoid bad behaviour of the
    // attenuation function.
    maxW = fGeom->Plane(plane)->Cell(cell)->HalfL();
    minW= -fGeom->Plane(plane)->Cell(cell)->HalfL();
    
    //check to see if we're in the nd muon catcher and therefore have an asymmetric cell
    if(fGeom->DetId() == novadaq::cnv::kNEARDET &&
       plane >= int(fGeom->FirstPlaneInMuonCatcher())){
      int i = 0;
        const geo::PlaneGeo* regularPlane = fGeom->Plane(i);
        while(regularPlane->View() != fGeom->Plane(plane)->View())
          regularPlane = fGeom->Plane(++i);
        const geo::CellGeo* regularCell = regularPlane->Cell(cell);
        // Assume that the cells are aligned on the floor
	minW=-regularCell->HalfL();
	maxW=(2*maxW)-regularCell->HalfL();
    }

    return true;
  }

  //-----
  double Calibrator::GetDriftScale(rb::CellHit const& cellhit)
  {
    if (!fUseDrift) return 1;

    // MC doesn't drift
    if(!cellhit.IsRealData()) return 1;

    return fDriftCache->GetFactor(cellhit.Plane(), cellhit.Cell(),
				  fCurrTimeStamp.GetSec());
    
  }

  //-----
  double Calibrator::GetGeVToMIPScale(rb::CellHit const& /* cellhit */ )
  {
    // MIPs for one GeV times one cell in z in NOvA scintillator
    return 94.62;
  }

  //-----
  double Calibrator::GetPECorrToGeVScale(rb::CellHit const& chit)
  {
    const daqchannelmap::lchan lchan = fCMap->Map()->encodeLChan(fGeom->DetId(),
							  chit.Plane(),
							  chit.Cell());
    const daqchannelmap::dchan dchan = fCMap->Map()->encodeDChan(lchan);

    return fAbsCache->GetPECorrToGeVScale(fGeom->DetId(),
					  fCMap->Map()->getDiBlock(dchan),
					  chit.IsRealData());
  }

  //-----
  double Calibrator::GetGeVToCalorimetricScale() const
  {
    // From Kanika's study: docdb 11528
    //
    // Could vary by detector or calibration version etc. We have all that
    // information available here. If it gets complicated enough may need to
    // come from database. For now it's a single number, and there's a
    // reasonable chance it might never need to change.
    return 1.78;
  }

  //-----
  double Calibrator::GetShowerEnergy(TVector3 const& start,
				     TVector3 const& stop,
				     double   const& totalGeV)
  {//now only for electron
    double ecor =  5.66050e-01;//for dead material correction

    double avgx = (start.X()+stop.X())/2.;
    double avgy = (start.Y()+stop.Y())/2.;
    double poscor = 1.0+7.00279e-05*avgx+6.91714e-05*avgy;//for position correction

    return totalGeV/(ecor*poscor);
  }

  //-------
  double Calibrator::GetTimingOffset(unsigned int plane, unsigned int cell, bool isData)
  {
    const daqchannelmap::lchan lchan = fCMap->Map()->encodeLChan(fGeom->DetId(),
							  plane,
							  cell);
    const daqchannelmap::dchan dchan = fCMap->Map()->encodeDChan(lchan);

    double offset = 0.0;
    //check if timing offsets should be added in, 
    if (fUseTimingOffsets){
      EnsureTimingCacheVldTime();
      offset = fTimingCache->GetOffset(fCMap->Map()->getDCM(dchan), fCMap->Map()->getDiBlock(dchan), fCurrTimeStamp.GetSec(), isData);
    }
  
    return offset;
  }

  //-------
  void Calibrator::EnsureTimingCacheVldTime()
  {
    if(fTimingCacheVldTimeSet) return;

    //use run history to improve database performance.  Apparently asking the run object start and end time was not producing the same result for each subrun
    //and so the cacheing performance of the timing calibration table was poor, clogging the system.
    art::ServiceHandle<nova::dbi::RunHistoryService> rh;
    fTimingCache->SetVldTime(fGeom->DetId(), rh->TStart(), rh->TStop());

    fTimingCacheVldTimeSet = true;
  }

  DEFINE_ART_SERVICE(Calibrator);

} // end namespace calib
///////////////////////////////////////////////////////////////////////
