////////////////////////////////////////////////////////////////////////
/// \brief  Look up absolute attenuation constants
/// \author Christopher Backhouse - bckhouse@caltech.edu
////////////////////////////////////////////////////////////////////////

#include "Calibrator/func/AbsCache.h"

#ifdef NOVACMAKE
#include "Database/cxx/include/Table.h"
#else
#include "Database/Table.h"
#endif

#include "messagefacility/MessageLogger/MessageLogger.h"

#include "TString.h"

namespace calib
{
  //......................................................................
  AbsCache::AbsCache() : fInitialized(false)
  {
  }

  //......................................................................
  AbsCache::~AbsCache()
  {
  }

  //......................................................................
  double AbsCache::GetPECorrToGeVScale(novadaq::cnv::DetId det,
				       int diblock,
				       bool isData)
  {
    if(!fInitialized) LoadFromVldDatabase(det, isData);

    auto it = fRecords.find(diblock);
    if(it == fRecords.end()) return -1;
    const Record& rec = it->second;
    return rec.meuGeV / rec.meu;
  }

  //......................................................................
  void AbsCache::LoadFromVldDatabase(novadaq::cnv::DetId det, bool isData)
  {
    // Even if it goes wrong there's no point trying again
    fInitialized = true;

    mf::LogInfo("AbsCache")
      << "Loading absolute energy constants for " << novadaq::cnv::DetInfo::GetName(det) << " from database...\n";

    nova::dbi::Table tblConsts("Calibration/tables/CalibAbsConsts.xml");
    tblConsts.SetDetector(det);
    tblConsts.SetDataTypeMask(isData ? nova::dbi::kDataOnly : nova::dbi::kMCOnly);
    if(!fTag.empty()) tblConsts.SetTag(fTag);
    tblConsts.SetVerbosity(100);

    tblConsts.SetMinTSVld(1);
    tblConsts.SetMaxTSVld(100);

    bool ok = true;

    if(fUseCSVsFromUPS || !fCSVSource.empty()){
      std::string csvSource = fCSVSource;
      if(fUseCSVsFromUPS){
        const char* ups_path = getenv("CALIBCSVS_CSV_PATH");
        assert(ups_path);
        const char* detStr = (det == novadaq::cnv::kFARDET) ? "fd" : "nd";
        const char* mcStr = isData ? "data" : "mc";
        csvSource = TString::Format("%s/%s/calib_abs_consts.%s.%s.%s.csv",
                                    ups_path, fTag.c_str(),
                                    detStr, mcStr, fTag.c_str()).Data();
      }

      mf::LogInfo("AbsCache") << "Loading absolute cache from "
			      << csvSource << std::endl;
      ok = ok && tblConsts.LoadFromCSV(csvSource);
    }
    else{
      ok = ok && tblConsts.Load();
    }

    if(!ok){
      // No safe way to handle this
      mf::LogError("AbsCache") << "Unable to load absolute tables. Aborting"
                                 << std::endl;
      // I've definitely seen mf lose races with shutdown
      std::cerr << "Unable to load absolute tables. Aborting"
                << std::endl;
      abort();
    }

    const int meuCol    = tblConsts.GetColIndex("meu");
    const int meuErrCol = tblConsts.GetColIndex("meu_err");
    const int meuGeVCol = tblConsts.GetColIndex("meu_gev");

    const int Nconsts = tblConsts.NRow();
    for(int nconsts = 0; nconsts < Nconsts; ++nconsts){
      nova::dbi::Row* row = tblConsts.GetRow(nconsts);

      const int diblock = row->Channel();

      Record rec;

      ok = ok && row->Col(meuCol).Get(rec.meu);
      ok = ok && row->Col(meuErrCol).Get(rec.meuErr);
      ok = ok && row->Col(meuGeVCol).Get(rec.meuGeV);

      assert(ok);

      fRecords[diblock] = rec;
    } // end for nconsts
  }

} // end namespace calib
////////////////////////////////////////////////////////////////////////
