////////////////////////////////////////////////////////////////////////
/// \brief   Helper for AttenCurve
/// \author bckhouse@caltech.edu
/// \date
////////////////////////////////////////////////////////////////////////
#ifndef CALIB_ATTENCACHE_H
#define CALIB_ATTENCACHE_H

#include <map>
#include <string>

#include "NovaDAQConventions/DAQConventions.h"
#include "GeometryObjects/OfflineChan.h"

namespace geo{class GeometryBase; class PlaneGeo;}

namespace calib
{
  class AttenCurve;

  class AttenCache
  {
  public:
    /// \param maxdiblock Largest diblock in DB when loading points table
    ///                   in chunks. Pass -1 not to do chunking.
    AttenCache(int maxdiblock = -1);
    virtual ~AttenCache();

    /// Return an AttenCurve object matching your parameters, or NULL
    /// if there isn't one in the database. Do not delete the resulting object,
    /// they are cached internally.
    const AttenCurve* ForData(const geo::GeometryBase* geom,
			      geo::OfflineChan chan);
    const AttenCurve* ForMC(const geo::GeometryBase* geom,
			    geo::OfflineChan chan, bool maskUncalibratedChannels);

    void SetConstsCSVSource(const std::string& s, const std::string& p)
    {
      fConstsCSVSource = s;
      fPointsCSVSource = p;
    }

    void UseCSVsFromUPS(bool use = true) {fUseCSVsFromUPS = use;}

    void SetTag(const std::string& tag) {fTag = tag;}

    std::string GetTag() {return fTag;}
    
    void SetMaxDiblock(int db) {fMaxDiblock = db;}

  protected:
    // Loading curves from the database as required is too slow (latency).
    // Instead, just read the whole detector in and cache it.

    void Add(AttenCurve* res);

    class Key_t
    {
    public:
      Key_t(int d, geo::OfflineChan chan) : fDet(d), fChan(chan) {}
      bool operator<(const Key_t& rhs) const;
    protected:
      int fDet;
      geo::OfflineChan fChan;
    };

    /// NULL if entry isn't in the cache. Do not delete the returned pointer.
    const AttenCurve* Get(Key_t key) const;

    std::string GetCSVPath(novadaq::cnv::DetId det,
                           bool isData, bool points) const;

    void LoadFromDatabase(const geo::GeometryBase* geom, bool isData);
    void LoadFromDatabaseUncalibratedMask(const geo::GeometryBase* geom);

    /// Deal with averaged constants etc
    Key_t TranslateChannelData(geo::OfflineChan chan,
			       const geo::GeometryBase* geom) const;
    /// Deal with merged views, muon catcher, etc
    Key_t TranslateChannelMC(geo::OfflineChan chan,
			     const geo::GeometryBase* geom) const;

    int fMaxDiblock; ///< -1 if not chunking loading by diblock
    /// Get the view for the current combination of view and plane (x,y,mux, or muy)
    int GetView(int plane, const geo::GeometryBase* geom) const;

    const geo::PlaneGeo* RepresentativePlane(const geo::GeometryBase* geom,
                                             int view,
                                             bool& inMuonCatcher) const;

    std::string fConstsCSVSource, fPointsCSVSource;
    bool fUseCSVsFromUPS;
    std::string fTag;

    // Whether we have (attempted to) read from db or csv yet
    bool fInitialized;
    bool fInitializedUncalibratedMask;
    bool fIsRealData; ///< Only set after fInitialized

    std::map<Key_t, AttenCurve*> fMap;
    std::set<Key_t> fMapUncalibratedMask;

    std::set<int> fDiblocks; ///< Diblocks with calibrations
  };

} // end namespace calib

#endif
////////////////////////////////////////////////////////////////////////
