////////////////////////////////////////////////////////////////////////
/// \brief   Hold drift constants from current run
/// \author  bckhouse@caltech.edu
/// \date
/////////////////////////////////////////////////////////////////////////
#include "Calibrator/func/DriftCache.h"

#include "messagefacility/MessageLogger/MessageLogger.h"

#ifdef NOVACMAKE
#include "Database/cxx/include/Table.h"
#else
#include "Database/Table.h"
#endif

#include <cassert>
#include "cetlib/exception.h"

namespace calib
{
  //......................................................................
  DriftCache::DriftCache()
    : fDet(0),
      fFactorIdx(-1),
      fTStart(0),
      fTStop(0),
      fEvtTime(0),
      fCSVFile(""),
      fTbl(0)
  {
  }

  //......................................................................
  DriftCache::~DriftCache()
  {
    if (fTbl) delete fTbl;
  }

  //......................................................................
  double DriftCache::GetFactor(int plane, int cell, time_t tevt)
  {
    
    if (fEvtTime == 0)
      fEvtTime = tevt;
    
    //**********
    // The map from (plane,cell) to FEBId needs to be corrected here
    // by Ranjan
    //**********

    geo::OfflineChan chan(plane,cell);
    int tchan = chan.ToDBValidityChan();
    
    //**********
    nova::dbi::Row* row = fTbl->GetVldRow(tchan,fEvtTime);
    if (!row) {
      throw cet::exception("DriftCacheGetFactor") << "DriftCache no factor found for channel " << chan;
    }
    double factor;
    assert(row->Col(fFactorIdx).Get(factor));
    return factor;
  }

  //......................................................................
  void DriftCache::SetVld(int det, time_t tstart, time_t tstop)
  {
    
    fDet = det;
    
    fTStart = tstart;
    fTStop = tstop;

    if (fTbl) delete fTbl;
    
    fTbl = new nova::dbi::Table("Calibration/tables/DriftCorrVld.xml");
    fTbl->SetDetector(fDet);
    fTbl->SetVerbosity(100);
    fFactorIdx = fTbl->GetColIndex("corr");
    
    this->Load();
  }

  //......................................................................
  void DriftCache::Load()
  {
    mf::LogInfo("DriftCache")
      << "Initializing drift cache ...\n";
    
    if (!fCSVFile.empty()) {
      if (!fTbl->LoadFromCSV(fCSVFile))
	throw cet::exception("DriftCacheLoadCSV") << "DriftCache Load failed to read from CSV file " << fCSVFile;      
      if (fTbl->NRow() == 0) 
	throw cet::exception("DriftCacheLoadCSV") << "DriftCache Load found 0 rows in CSV file " << fCSVFile;      
    }
    else {
      fTbl->SetMinTSVld(fTStart);
      fTbl->SetMaxTSVld(fTStop);
      if (!fTbl->Load())
	throw cet::exception("DriftCacheLoadDB") << "DriftCache Load from dB failed!";
      if (fTbl->NRow() == 0) {
	throw cet::exception("DriftCacheLoadDB") << "DriftCache Load found 0 rows from database!";
      }
    }
    fTbl->FillChanRowMap();
  }
  
} // end namespace calib
////////////////////////////////////////////////////////////////////////
