////////////////////////////////////////////////////////////////////////
/// \brief   Hold drift constants from current run
/// \author  bckhouse@caltech.edu
/// \date
/////////////////////////////////////////////////////////////////////////
#include "Calibrator/func/TimingCache.h"

#include "messagefacility/MessageLogger/MessageLogger.h"

#ifdef NOVACMAKE
#include "Database/cxx/include/Table.h"
#else
#include "Database/Table.h"
#endif

#include <cassert>

#include "TString.h"

namespace calib
{
  //......................................................................
  TimingCache::TimingCache()
    : fInitialized(false),
      fCurDet(novadaq::cnv::kUNKNOWN_DET),
      fMinTime(std::numeric_limits<uint32_t>::max()-1),
      fMaxTime(std::numeric_limits<uint32_t>::min())
  {
  }

  //......................................................................
  TimingCache::~TimingCache()
  {
  }

  //......................................................................
  double TimingCache::GetOffset(int dcm, int diblock, std::uint32_t time, bool isData)
  {

    if(!fInitialized)
      LoadFromDatabase(isData);

    int key = (diblock*1000)+dcm;
    std::map<int, DelayRec>::iterator it;
    it = fMap.find(key);
    if(it == fMap.end()){
      mf::LogDebug("TimingCache") << "No timing offset for dcm "<<dcm;
      abort();
    }


    const DelayRec& rec = it->second;
    return rec.dcmoffset;
  }

  //......................................................................
  double TimingCache::GetSigma(int dcm, int diblock, std::uint32_t time, bool isData)
  {

    if(!fInitialized)
      LoadFromDatabase(isData);

    std::map<int, DelayRec>::iterator it;
    int key = (diblock*1000)+dcm;
    it = fMap.find(key);
    if(it == fMap.end()){
      mf::LogDebug("TimingCache") << "No timing offset sigma for dcm "<<dcm;
      abort();
    }

    const DelayRec& rec = it->second;
    return rec.sigma;
  }

  //......................................................................
  void TimingCache::SetVldTime(novadaq::cnv::DetId det, std::uint32_t minTime, std::uint32_t maxTime)
  {
    if(det != fCurDet || minTime != fMinTime || maxTime != fMaxTime){
      fInitialized = false;
      fMap.clear();

      fCurDet = det;
      fMinTime= minTime;
      fMaxTime = maxTime;
    }
  }

  //......................................................................
  void TimingCache::LoadFromDatabase(bool isData)
  {
    // Even if it goes wrong there's no point trying again
    fInitialized = true;

    //if this is MC, move the mintime away from 0
    if (!isData && fMinTime == 0){
      fMinTime = 1;
      fMaxTime = 2;
    }
    mf::LogInfo("TimingCache")
      << "Initializing timing delay cache for "
      << novadaq::cnv::DetInfo::GetName(fCurDet) << " from database...\n";


    nova::dbi::Table tblConsts("Calibration/tables/CalibTimingConsts.xml");
    tblConsts.SetDetector(fCurDet);
    tblConsts.SetDataTypeMask(isData ? nova::dbi::kDataOnly : nova::dbi::kMCOnly);
    if(!fTag.empty()) tblConsts.SetTag(fTag);
    tblConsts.SetVerbosity(100);

    tblConsts.SetMinTSVld(fMinTime);
    tblConsts.SetMaxTSVld(fMaxTime);

    bool ok = true;

    if(fUseCSVsFromUPS || !fCSVSource.empty()){
      std::string csvSource = fCSVSource;
      if(fUseCSVsFromUPS){
        const char* ups_path = getenv("CALIBCSVS_CSV_PATH");
        assert(ups_path);
        const char* detStr = (fCurDet == novadaq::cnv::kFARDET) ? "fd" : "nd";
        const char* mcStr = isData ? "data" : "mc";
        csvSource = TString::Format("%s/%s/calib_dcm_delay_consts.%s.%s.%s.csv",
                                    ups_path, fTag.c_str(),
                                    detStr, mcStr, fTag.c_str()).Data();
      }

      mf::LogInfo("TimingCache") << "Loading timing cache from "
                                 << csvSource << std::endl;
      ok = ok && tblConsts.LoadFromCSV(csvSource);
    }
    else{
      ok = ok && tblConsts.Load();
    }


    if(!ok){
      // No safe way to handle this
      mf::LogError("TimingCache") << "Unable to load timing delay tables. No timing delays will be set."
                                 << std::endl;

      abort();
    }
    else{
      const int dcmoffsetCol    = tblConsts.GetColIndex("dcmoffset");
      const int sigmaCol = tblConsts.GetColIndex("sigma");
      const int statsCol = tblConsts.GetColIndex("stats");

      const int Nconsts = tblConsts.NRow();
      for(int nconsts = 0; nconsts < Nconsts; ++nconsts){
        nova::dbi::Row* row = tblConsts.GetRow(nconsts);

        const int dcmkey = row->Channel();

        DelayRec rec;

        ok = ok && row->Col(dcmoffsetCol).Get(rec.dcmoffset);
        ok = ok && row->Col(sigmaCol).Get(rec.sigma);
        ok = ok && row->Col(statsCol).Get(rec.stats);

        //For data, abort if the value is bad,
        if (isData) assert(ok);

        if (ok) fMap[dcmkey] = rec;
      } // end for nconsts
    }

  }

} // end namespace calib
////////////////////////////////////////////////////////////////////////
