//////////////////////////////////////////////////////////////////////////
/// \brief   Interface to the run-by-run list of bad channels
/// \author  betancourt@physics.umn.edu
//////////////////////////////////////////////////////////////////////////
#include "ChannelInfo/BadChanList.h"

#include <sys/stat.h>
#include <sstream>

// NOvA includes
#include "CMap/CMap.h"
#include "Geometry/Geometry.h"
#include "SummaryData/DataRunNumInMC.h"
#include "SummaryData/RunData.h"
#include "NovaDAQConventions/DAQConventions.h"
#include "DAQChannelMap/DAQChannelMap.h"

//#include "NovaDatabase/Table.h"
#ifdef NOVACMAKE
#include "Database/cxx/include/Table.h"
#else
#include "Database/Table.h"
#endif
#include "RunHistory/service/RunHistoryService.h"
#include "Simulation/Simulation.h"

// Framework includes
#include "art/Framework/Principal/SubRun.h"
#include "art/Framework/Principal/Run.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "art/Framework/Services/Optional/RandomNumberGenerator.h"
#include "cetlib/search_path.h"
#include "cetlib/exception.h"
#include "messagefacility/MessageLogger/MessageLogger.h"

#include "CLHEP/Random/RanluxEngine.h"

#include "TString.h"

namespace chaninfo {

  //------------------------------------------------------------
  BadChanList::BadChanList(fhicl::ParameterSet const& pset,
                           art::ActivityRegistry &reg) :
    fNewCosmicEffTable(false)
  {
    fRandomBCMap.clear();
    fNeedNewData = true;
    fValidityTag = "";
    fMCRunNumber = -1;
    fMCSubrunNumber = -1;
    fNBadSR = -1; // unknown yet
    fIsPerfect = false;
    CLHEP::RanluxEngine rle(sim::GetRandomNumberSeed());
    fFlatRand = std::auto_ptr<CLHEP::RandFlat>(new CLHEP::RandFlat(rle));

    this->reconfigure(pset);
    reg.sPostBeginRun.watch(this, &BadChanList::postBeginRun);
    reg.sPostBeginSubRun.watch(this, &BadChanList::postBeginSubRun);
    reg.sPreProcessEvent.watch(this, &BadChanList::preBeginEvent);    
  }

  //----------------------------------------------------------

  BadChanList::~BadChanList()
  {
    fRandomBCMap.clear();
    fBCTable.reset();

  }

  //----------------------------------------------------------
  void BadChanList::postBeginRun(art::Run const & run)
  {
    if (run.run() >= 1000000)
      fIsPerfect=true;
    else if (run.run() < 10000) {
      if (fAbortOnSmallRunNum)
	throw cet::exception("BadChannelInvalidRunNumber") << "Bad Channels found a run number below 10000.  This should never happen and is dangerous.";
      else
	fIsPerfect=true;
    }
    else
      fIsPerfect=false;

    fNeedNewData = true;

    fNBadSR = -1;
  }

  //----------------------------------------------------------
  void BadChanList::postBeginSubRun(art::SubRun const & subrun)
  {
    art::ServiceHandle<nova::dbi::RunHistoryService> rh;
    int run = subrun.run();
    int sr = subrun.subRun();

    if (fMCRunNumber > 0) {
      run = fMCRunNumber;
      sr = fMCSubrunNumber;
    }

    fRun = run;
    fSubRun = sr;

    fNBadSR = -1;
  }

  //----------------------------------------------------------
  void BadChanList::preBeginEvent(art::Event const & event)
  {
    if(fEventCheck) {  // check for dropped DCMs every event
      art::Handle< std::vector<rawdata::RawDigit> > digitcol;
      event.getByLabel(fRawDigitLabel, digitcol);

      if(digitcol.failedToGet()) return;  // for MC generation etc when we don't have data
      for(int i = 0; i < 12; i++) { // re-initialize every event
        for(int j = 0; j < 14; j++) { // just keep array sized 12 x 14; for ND we just won't use it all
          fDCMHits[i][j] = 0;
        }
      }
      int dib, dcm;
      art::ServiceHandle<geo::Geometry> geom;
      daqchannelmap::DAQChannelMap* daqmap = daqchannelmap::DAQChannelMap::getInstance(geom->DetId());
      for(unsigned int i = 0; i < digitcol->size(); i++) { // count the number of raw digit hits in each DCM
        const rawdata::RawDigit* digit = &(*digitcol)[i];
        dcm = daqmap->getDCM(digit->DaqChannel());
        dib = daqmap->getDiBlock(digit->DaqChannel());
        fDCMHits[dcm-1][dib-1]++;
      }
    } // end looking for dropped DCMs
  }

  //-----------------------------------------------------------
  int BadChanList::NBadInSubRun(int isr)
  {
    art::ServiceHandle<nova::dbi::RunHistoryService> rh;

    if (isr > rh->NSubruns()) return 0;
    
    if (fNeedNewData) LoadNewData();

    art::ServiceHandle<geo::Geometry> geom;
    int detId = geom->DetId();
    daqchannelmap::DAQChannelMap* daqmap = daqchannelmap::DAQChannelMap::getInstance(detId);

    if(fNBadSR >= 0) return fNBadSR;

    // Otherwise, look up the answer
    int nbad=0;
    int fRun_save = fRun;
    int fSubRun_save = fSubRun;
    fRun = rh->RunNumber();
    fSubRun = isr;

    // turn off event checking here, turn it back on when we're done
    // counting bad channels
    bool fEventCheck_save = fEventCheck;
    fEventCheck = false;

    for (int idb=0; idb<rh->NDiBlocks(); ++idb) {
      nova::dbi::RunHistory::DiBlock dblk = rh->GetDiBlock(idb,false);

      if (!rh->IsGoodDiBlock(dblk.num)) continue;

      for (unsigned int idcm=0; idcm<dblk.dcm.size(); ++idcm) {
        for (unsigned int ifeb=0; ifeb<dblk.dcm[idcm].feb.size(); ++ifeb) {
          if (dblk.dcm[idcm].feb[ifeb].hasAPD) {
            for (int ipix=0; ipix<=31; ++ipix) {
            //for (int ipix=1; ipix<=32; ++ipix) {
              daqchannelmap::dchan daqChan = daqmap->encodeDChan(detId,
                                                                 dblk.num,
                                                                 idcm+1,
                                                                 ifeb,
                                                                 ipix);
              daqchannelmap::lchan logChan = daqmap->encodeLChan(daqChan);

              int plane = daqmap->getPlane(logChan);
              int cell = daqmap->getCell(logChan);
              if (IsBad(plane,cell)) ++nbad;
            } // end loop over pixels
          } // end if hasAPD
        } // end loop over FEBs
      } // end loop over DCMs
    } // end loop over DiBlocks

    fEventCheck = fEventCheck_save;
    fNBadSR = nbad;

    fRun = fRun_save;
    fSubRun = fSubRun_save;

    return fNBadSR;
  }

  //-----------------------------------------------------------
  void BadChanList::SetRandomBadChannels()
  {
    if (!fRandomBCMap.empty()) return;

    std::cerr << " RANDOM BAD CHANNELS at rate "
              << fRandomBadChannelFraction << std::endl;

    fLowEfficiency.clear();

    art::ServiceHandle<geo::Geometry> geom;
    for(unsigned int plane: geom->GetPlanesByView()){
      const unsigned int nCells = geom->Plane(plane)->Ncells();

      for(unsigned int cell = 0; cell < nCells; ++cell){
	geo::OfflineChan chan(plane, cell);
	fRandomBCMap[chan] = (fFlatRand->fire() > fRandomBadChannelFraction);
      }
    }
  }

  //-----------------------------------------------------------
  void BadChanList::LoadCosmicEff(int newrun, int newsubrun)
  {
    fLowEfficiency.clear();

    nova::dbi::Table tblEff(fNewCosmicEffTable ?
			    "/ChannelInfo/CosmicEffNew.xml" :
			    "/ChannelInfo/CosmicEff.xml");
    tblEff.SetValidityRange("detector", "ndos");
    tblEff.SetValiditySQL(TString::Format("(beginrun<%d or (beginrun=%d and beginsubrun<=%d)) and "
					  "(endrun>%d or (endrun=%d and endsubrun>=%d))",
					  newrun, newrun, newsubrun, newrun, newrun, newsubrun).Data());
    //    }

    // Only ask for the bad channels. We won't get any "good" ones back, so
    // they won't go in the map, which is fine.
    tblEff.SetValidityRange("good", false);

    tblEff.SetVerbosity(100);
    const bool ok = tblEff.Load();
    if(!ok){
      // No safe way to handle this
      std::cerr << "Couldn't load efficiencies channels from DB. Aborting" << std::endl;
      abort();
    }

    int planeIdx = tblEff.GetColIndex("plane");
    int cellIdx = tblEff.GetColIndex("cell");
    int goodIdx = tblEff.GetColIndex("good");

    for(int i = 0; i < tblEff.NRow(); ++i){
      int plane, cell;
      bool good;
      bool ok = tblEff.GetRow(i)->Col(planeIdx).Get(plane);
      ok = ok && tblEff.GetRow(i)->Col(cellIdx).Get(cell);
      ok = ok && tblEff.GetRow(i)->Col(goodIdx).Get(good);
      if(!ok) throw;
      geo::OfflineChan chan(plane, cell);
      // DB stores "goodness", we store "badness"
      fLowEfficiency[chan] = !good;
    }
  }

  //-----------------------------------------------------------
  void BadChanList::reconfigure(fhicl::ParameterSet pset)
  {
    fApplyBadChan            = pset.get<int>("ApplyBadChan");
    fMCRunNumber             = pset.get<int>("MCRunNumber");
    fMCSubrunNumber             = pset.get<int>("MCSubrunNumber");
    fNewCosmicEffTable       = pset.get<bool>("NewCosmicEffTable");
    fRandomBadChannelFraction= pset.get<float>("RandomBadChannelFraction");
    fBadChanCSVFile          = pset.get<std::string>("BadChanCSVFile");
    fValidityTag             = pset.get<std::string>("ValidityTag");
    fAbortIfNoneFound        = pset.get<bool>("AbortIfNoneFound");
    fAbortOnSmallRunNum      = pset.get<bool>("AbortOnSmallRunNum");
    fEventCheck              = pset.get<bool>("EventCheck");
    fRawDigitLabel           = pset.get<std::string>("RawDigitLabel");
    fApplyFEBDropout         = pset.get<bool>("ApplyFEBDropout");
    fFEBDropoutTag           = pset.get<std::string>("FEBDropoutTag");
    fVldRunGrouping          = pset.get<int>("VldRunGrouping");

    if (!fValidityTag.empty()) {
      float tagNum = atof(fValidityTag.substr(1).c_str());
      if (tagNum < 9.0)
        throw cet::exception("BadChannelTagFail") << "This version of Bad Channels requires tag >= 9.0";
    }

  }

  //----------------------------------------------------------
  void BadChanList::LoadNewData()
  {
    if (fIsPerfect) return;

    art::ServiceHandle<geo::Geometry> geom;

    // reset bad channel table; set vld time to 0, so that EnsureRun 
    // will correctly set validity times
    fBCTable.reset(new nova::dbi::Table("/ChannelInfo/ChanRateState.xml"));
    fBCTable->SetVerbosity(100);
    fBCTable->SetDetector(geom->DetId());
    if (fValidityTag != "")
      fBCTable->SetTag(fValidityTag);
    fStateIdx = fBCTable->GetColIndex("state");

    fFEBDOTable.reset(new nova::dbi::Table("/ChannelInfo/FEBDropout.xml"));
    fFEBDOTable->SetVerbosity(100);
    fFEBDOTable->SetDetector(geom->DetId());
    if (fFEBDropoutTag != "")
      fFEBDOTable->SetTag(fFEBDropoutTag);
    fFEBStateIdx = fFEBDOTable->GetColIndex("state");

    if (!fBadChanCSVFile.empty()) 
    {
      fBCTable->LoadFromCSV(fBadChanCSVFile);
      std::cerr << "Read " << fBCTable->NRow() << " BadChannel rows from "
      << fBadChanCSVFile << std::endl;
    }
    else 
    {
      if (fMCRunNumber > 0) 
      {
        nova::dbi::RunHistory rh(fMCRunNumber,geom->DetId());
        fMinVldTime = rh.RunNumber()*10000 + 1;
        fMaxVldTime = fMinVldTime + rh.NSubruns()*100;
      }
      else if (geom->DetId() == novadaq::cnv::kFARDET ||
               geom->DetId() == novadaq::cnv::kNEARDET)
      {
        art::ServiceHandle<nova::dbi::RunHistoryService> rh;
	int minrun = int(rh->RunNumber()/fVldRunGrouping)*fVldRunGrouping;
	fMinVldTime = minrun*10000 + 1;
        fMaxVldTime = fMinVldTime - 1 + fVldRunGrouping*10000;
      }

      fBCTable->SetDataTypeMask(nova::dbi::kMCOnly|nova::dbi::kDataOnly);
      fBCTable->SetMinTSVld(fMinVldTime);
      fBCTable->SetMaxTSVld(fMaxVldTime);	  	
      
      if (!fBCTable->Load()) 
        throw cet::exception("BadChannelTableLoadFail") << "Bad Channels table failed to load data!";
      if (fBCTable->NRow() == 0) {
        std::cerr << "******** WARNING: BadChanList_service found no bad " 
                  << "channels!" << std::endl; 
        if (fAbortIfNoneFound) {
          throw cet::exception("BadChannelTableEmpty") << "Bad Channels table is empty, this should never happen!";
        }
      }
      
      if (fApplyFEBDropout && (geom->DetId() == novadaq::cnv::kNEARDET)) {
	fFEBDOTable->SetMinTSVld(fMinVldTime-480);
	fFEBDOTable->SetMaxTSVld(fMaxVldTime+480);
	fFEBDOTable->Load();
	if (fFEBDOTable->NRow() == 0) {
	  std::cerr << "******** WARNING: BadChanList_service found no FEB "
		    << "dropout data!" << std::endl;
	}
      }
    }

    fBCTable->FillChanRowMap();
    fFEBDOTable->FillChanRowMap();
    fNeedNewData = false;
  }

  //----------------------------------------------------------
  bool BadChanList::IsBad(int plane, int cell)
  {
    if(!fApplyBadChan) return false;
    if (fIsPerfect) return false;

    geo::OfflineChan chan(plane,cell);

    if(fRandomBadChannelFraction > 0){
      SetRandomBadChannels();
      return fRandomBCMap[chan];
    }

    if (fNeedNewData) LoadNewData();

    art::ServiceHandle<nova::dbi::RunHistoryService> rh;
    art::ServiceHandle<geo::Geometry> geom;

    const int sr = fSubRun%rh->NSubruns();
    const time_t evtTime = fRun*10000 + sr*100 + 1;

    // check if event time within validity range, i.e. sane
    if ( fBadChanCSVFile.empty() && (evtTime < fMinVldTime || evtTime > fMaxVldTime))
      throw cet::exception("BadChannelInvalidValidityTime") 
	<< "Event time is outside the validity range of the database request.  "
	<< "Database request does not contain complete information, "
	<< "which could lead to unpredictable behavior." << std::endl 
        << "Event time: " << evtTime << std::endl
        << "Request start: " << fMinVldTime << std::endl
        << "Request end: " << fMaxVldTime << std::endl; 
    
    // check for being in a dropped DCM
    if(fEventCheck) {
      daqchannelmap::DAQChannelMap* daqmap = daqchannelmap::DAQChannelMap::getInstance(geom->DetId());
      uint32_t dchan = daqmap->encodeDChan(daqmap->encodeLChan(geom->DetId(), plane, cell));
      if(fDCMHits[daqmap->getDCM(dchan)-1][daqmap->getDiBlock(dchan)-1]==0) return true;
    }
    
    // check with RH whether this channel was instrumented, part of the partition, or masked
    if (!rh->HasAPD(chan)) return true;
      
    // for now, just deal with dB tables for the FarDet and NearDet
    if (geom->DetId() == novadaq::cnv::kFARDET ||
	geom->DetId() == novadaq::cnv::kNEARDET) {
      
      int tchan = chan.ToDBValidityChan();
      
      // get validity row for this timestamp
      nova::dbi::Row* row = fBCTable->GetVldRow(tchan,evtTime);
      int state;

      if (row) {
	// there is an entry for this timestamp (expected behavior), so
	// make sure that the state is "ok" (==0)
        const bool ok = row->Col(fStateIdx).Get(state);
        assert(ok);
	if (state != 0) return true;
      }
      
      if (fApplyFEBDropout) {
	daqchannelmap::DAQChannelMap* daqcmap = daqchannelmap::DAQChannelMap::getInstance(geom->DetId());
	daqchannelmap::lchan logChan = daqcmap->encodeLChan(geom->DetId(),plane,cell);
	daqchannelmap::dchan daqChan = daqcmap->encodeDChan(logChan);
	
	int db = daqcmap->getDiBlock(daqChan);
	int dcm = daqcmap->getDCM(daqChan);
	int feb = daqcmap->getFEB(daqChan);	
	
	tchan = db*1000+(dcm-1)*64+feb;
	row = fFEBDOTable->GetVldRow(tchan,evtTime);

	if (row) {
	  const bool ok = row->Col(fFEBStateIdx).Get(state);
          assert(ok);
	  // a state of "1" means the FEB is live at this point
	  if (state != 1) return true;
	}
      }
    }
    
    // If we can't find anything wrong with it then it must not be bad
    return false;
  }

  //----------------------------------------------------------
  bool BadChanList::IsLowEfficiency(int plane, int cell)
  {
    if(!fApplyBadChan) return false;

    const geo::OfflineChan chan(plane, cell);
    return fLowEfficiency[chan];
  }

  //--------------------------------------------------------
  
  int BadChanList::Apply(std::vector< art::Ptr<rawdata::RawDigit> >& rd)
  {
    art::ServiceHandle<cmap::CMap> cmap;
    int ch;
    std::vector< art::Ptr<rawdata::RawDigit> > good;
    for(unsigned int i=0; i<rd.size(); ++i) {
      int plane = cmap->GetPlane(rd[i].get());
      int cell =  cmap->GetCell(rd[i].get());
      if(!IsBad(plane,cell)) good.push_back(rd[i]);
    }
    rd.swap(good);
    ch=rd.size();
    return ch;
  }
  
  //--------------------------------------------------------
  
  int BadChanList::Apply(art::PtrVector<rawdata::RawDigit>& rd)
  {
    art::ServiceHandle<cmap::CMap> cmap;
    int ch;
    art::PtrVector<rawdata::RawDigit> good;
    for(unsigned int i=0; i<rd.size(); ++i) {
      int plane = cmap->GetPlane(rd[i].get());
      int cell =  cmap->GetCell(rd[i].get());
      if(!IsBad(plane,cell)) good.push_back(rd[i]);
    }
    rd.swap(good);
    ch=rd.size();
    return ch;
  }

  //--------------------------------------------------------

  int BadChanList::Apply( std::vector<const rawdata::RawDigit*>& rd)
  {
    art::ServiceHandle<cmap::CMap> cmap;
    int ch;
    std::vector<const rawdata::RawDigit*> good;
    for(unsigned int i=0; i<rd.size(); ++i) {
      int plane = cmap->GetPlane(rd[i]);
      int cell =  cmap->GetCell(rd[i]);
      if(!IsBad(plane,cell)) good.push_back(rd[i]);
    }
    rd.swap(good);
    ch=rd.size();
    return ch;
  }



  DEFINE_ART_SERVICE(BadChanList);

}
////////////////////////////////////////////////////////////////////////
