#include <iostream>
#include <iomanip>
#include <vector>
#include <string>
#include <cstdlib>
#include <ctime>
#include <getopt.h>

#include <boost/filesystem.hpp>

#ifdef NOVACMAKE
#include "Database/cxx/include/Table.h"
#else
#include "Database/Table.h"
#endif
#include "RunHistory/RunHistory.h"
#include "NovaDAQConventions/DAQConventions.h"

#include "TFile.h"
#include "TH1F.h"
#include "TH2F.h"
#include "TTree.h"
#include "TMath.h"

int run=0;
std::string det="";
std::string rel="";
bool write2DB=false;
//------------------------------------------------------------

void PrintUsage()
{
  std::cout << "Usage: getMissingRunHistoryInfo [options] [detector] [run]" 
	    << std::endl;
  std::cout << "options:" << std::endl;
  std::cout << "\t -d (--write2DB)" << std::endl;
}

//------------------------------------------------------------

bool ParseCLArgs(int argc, char* argv[])
{
  if (argc == 1) {
    PrintUsage();
    exit(0);
  }

  struct option long_options[] = {
    {"write2DB", no_argument, 0, 'd'},
    //    {"nohistos", no_argument, 0, 'h'},
    {0,0,0,0}
  };
  
  while (1) {
    int optindx;
    int c = getopt_long(argc,argv,"d",long_options,&optindx);
        
    if (c==-1) break;
    
    switch(c) {
      /*
	case 'h':
	saveHistos=false;
	break;
      */
    case 'd':
      write2DB=true;
      break;
    default:
      break;
    }
  }
  
  if ((argc-optind) < 2) {
    std::cerr << "Missing arguments." << std::endl;
    PrintUsage();
    exit(0);
  }
  
  if ((argc-optind) > 2) {
    std::cerr << "Unrecognized arguments." << std::endl;
    PrintUsage();
    exit(0);
  }
  
  det = argv[optind++];
  run = atoi(argv[optind]);
  
  return true;
}

//------------------------------------------------------------

int main(int argc, char *argv[])
{
  ParseCLArgs(argc, argv);
  
  std::ifstream relList;
  relList.open("releases.txt");

  std::vector<std::string> rel;
  std::vector<std::string> prefix;
  std::vector<std::string> suffix1;
  std::vector<std::string> suffix2;

  std::string str1;
  std::string str2;
  std::string str3;
  std::string str4;

  while(relList >> str1 >> str2 >> str3 >> str4) {
    rel.push_back(str1);
    prefix.push_back(str2);
    suffix1.push_back(str3);
    suffix2.push_back(str4);
  }
  relList.close();
  
  uint32_t fEndYear;  ///< Header: actual years since 0 A.D.
  uint16_t fEndMonth; ///< Header: month of first processed event (1-12)
  uint16_t fEndDay;   ///< Header: day of first processed event (01-31)
  double   fEndHour;  ///< Header: hour of first processed event (0.0 - 23.99999)
  
  char fname[256];
  const char* dir1 = getenv("NOVA_DATA");
  if (!dir1) {
    std::cerr << "$NOVA_DATA undefined!" << std::endl;
    exit(1);
  }

  int irel=(int)rel.size()-1;
  bool foundFile = false;

  for ( ; irel>=0; --irel) {
    sprintf(fname,"%s/nearline-OnMon/%s/%s/%06d/%08d/%s%08d%s000%s",dir1,det.c_str(),rel[irel].c_str(),int(run/100),run,prefix[irel].c_str(),run,suffix1[irel].c_str(),suffix2[irel].c_str());

    if (boost::filesystem::exists(fname)) {
      foundFile = true;
      break;
    }
    
  }

  if (!foundFile) {
    std::cerr << "File not found!" << std::endl;
    exit(1);
  }

  TFile* fin = new TFile(fname);
  if (!fin) {
    std::cerr << "File not found!" << std::endl;
    exit(1);
  }
  fin->Close();

  for (int isr=63; isr>=0; --isr) {

    sprintf(fname,"%s/nearline-OnMon/%s/%s/%06d/%08d/%s%08d%s%03d%s",dir1,det.c_str(),rel[irel].c_str(),int(run/100),run,prefix[irel].c_str(),run,suffix1[irel].c_str(),isr,suffix2[irel].c_str());

    if (!boost::filesystem::exists(fname)) continue;

    std::cout << "Found file for run " << run << ", subrun " << isr << std::endl;

    fin = new TFile(fname);
    if (!fin) {
      continue;
    }

    TTree* hdr = (TTree*)fin->Get("Header");
    hdr->SetBranchAddress("EndYear",&fEndYear);
    hdr->SetBranchAddress("EndMonth",&fEndMonth);
    hdr->SetBranchAddress("EndDay",&fEndDay);
    hdr->SetBranchAddress("EndHour",&fEndHour);
    hdr->GetEvent(0);

    if (fEndYear < 110) {
      std::cerr << "WARNING: subrun " << isr << " appears to have an invalid "
		<< "start time recorded in OnMon.  Skipping this subrun."
		<< std::endl;
      fin->Close();
      continue;
    }

    if (fEndYear > 1900) {
      fEndYear -= 1900;
      --fEndMonth;
    }

    int hour = int(fEndHour);
    int min = int((fEndHour-hour)*60.);
    int sec = int((fEndHour-hour-(float)min/60.)*3600.);
 
    struct tm t1;
    t1.tm_sec = sec;
    t1.tm_min = min;
    t1.tm_hour = hour;
    t1.tm_mday = fEndDay;
    t1.tm_mon = fEndMonth-1;
    t1.tm_year = fEndYear-1900;

    char timeStr[64];
    strftime(timeStr,64,"%Y-%m-%d %H:%M:%S",&t1);
    std::cout << timeStr << std::endl;

    //    time_t tstop = timegm(&t1);

    nova::dbi::Table* table = new nova::dbi::Table("NovaDatabase/tables/RunHistory/Runs.xml");
    table->SetDetector(det);
    table->SetValidityRange("run",run,run);
    int tstopIdx = table->GetColIndex("tstop");
    int nsubrunsIdx = table->GetColIndex("nsubruns");

    table->LoadFromDB();
    std::cout << "runs table got " << table->NRow() << " rows" << std::endl;
    nova::dbi::Row* row = table->GetRow(0);
    row->Update(tstopIdx,timeStr);
    row->Update(nsubrunsIdx,isr+1);

    fin->Close();
    
    if (write2DB)
      if (! table->WriteToDB()) 
	std::cerr << "Failed to write to database.  Aborting." << std::endl;
    
    break;


  }
    
  return 0;
}
