#include <iostream>
#include <iomanip>
#include <vector>
#include <string>
#include <map>
#include <cstdlib>
#include <ctime>
#include <getopt.h>
#include <boost/filesystem.hpp>

#ifdef NOVACMAKE
#include "Database/cxx/include/Table.h"
#else
#include "Database/Table.h"
#endif
#include "RunHistory/RunHistory.h"
#include "NovaDAQConventions/DAQConventions.h"
#include "DAQChannelMap/HardwareDisplay.h"

int run=0;
int subrun=-1;
bool write2CSV=false;
bool nbadOnly=false;
bool fracOnly=false;
bool includeAll=false;
bool febOnly=false;
std::string csvFileName="";
std::string validityTag="";
std::string febN="";
std::string det="";

//------------------------------------------------------------

void PrintUsage()
{
  std::cout << "Usage: showBadChannels [options] [detector] [run]" << std::endl;
  std::cout << "options:\n";
  std::cout << "\t -s (--subrun) [subrun]: optional subrun" << std::endl;
  std::cout << "\t -t (--tag) [tag]: set validity tag" << std::endl;
  std::cout << "\t -c (--tocsv) [filename]: write results to csv file" << std::endl;
  std::cout << "\t -n (--nbadonly): only print the number of bad channels" << std::endl;
  std::cout << "\t -f (--fraction): only print the fraction of bad channels" << std::endl;
  std::cout << "\t -a (--all): include all channels, including those excluded by analysis cuts" << std::endl;
  std::cout << "\t -F (--feb): only print the number of bad channels for a given FEB (i.e. 2-02-08:41)" << std::endl;
}

//------------------------------------------------------------

bool ParseCLArgs(int argc, char* argv[])
{
  if (argc == 1) {
    PrintUsage();
    exit(0);
  }
  
  struct option long_options[] = {
    {"tocsv", 0, 0, 'c'},
    {"tag", 0, 0, 't'},
    {"subrun", 0, 0, 's'},
    {"nbadonly", 0, 0, 'n'},
    {"fraction", 0, 0, 'f'},
    {"all", 0, 0, 'a'},
    {"feb", 0, 0, 'F'},
    {0,0,0,0}
  };
  
  while (1) {
    int optindx;
    //If a character is followed by a colon, the option requires an argument
    int c = getopt_long(argc,argv,"c:t:s:nfaF:",long_options,&optindx);
        
    if (c==-1) break;
    
    switch(c) {
    case 'n':
      nbadOnly=true;
      break;
    case 'f':
      fracOnly=true;
      break;
    case 'c':
      write2CSV=true;
      csvFileName=optarg;
      break;
    case 't':
      validityTag=optarg;
      break;
    case 's':
      subrun = atoi(optarg);
      break;
    case 'a':
      includeAll = true;
      break;
    case 'F':
      febOnly=true;
      febN=optarg;
      break;
    default:
      break;
    }
  }
  
  if ((argc-optind) < 2) {
    std::cerr << "Missing arguments." << std::endl;
    PrintUsage();
    exit(0);
  }
  
  if ((argc-optind) > 2) {
    std::cerr << "Unrecognized arguments." << std::endl;
    PrintUsage();
    exit(0);
  }
  
  det = argv[optind++];
  run = atoi(argv[optind++]);

  return true;
}

//------------------------------------------------------------

int main(int argc, char *argv[])
{
  ParseCLArgs(argc, argv);
  
  int detId = novadaq::cnv::DetInfo::GetId(det);
  if (detId != novadaq::cnv::kFARDET && detId != novadaq::cnv::kNEARDET) {
    std::cout << "No bad channels available for " << det << " at this time"
	      << std::endl;
    exit(0);
  }

  nova::dbi::Table* t = new nova::dbi::Table("ChannelInfo/ChanRateState.xml");
  t->SetDetector(det);
  t->SetDataTypeMask(nova::dbi::kMCOnly|nova::dbi::kDataOnly);
  if (validityTag != "") 
    t->SetTag(validityTag);

  if (run > 1000000) run -= 1000000;
  nova::dbi::RunHistory rh(run,det);

  if (!rh.LoadDAQRunHistory()) {
    std::cerr << "RunHistory::LoadDAQRunHistory failed, aborting." << std::endl;
    exit(1);
  }

  rh.LoadHardwareInfo();

  if (!includeAll) {
    rh.SetIfAnyMasking(true);
    rh.SetAnalysisDiBlockMasking(true);
    rh.SetIfCustomMasking(false);
  }

  if (subrun >= 0 && subrun < rh.NSubruns()) {
    t->SetMinTSVld(rh.RunNumber()*10000 + subrun*100 + 1);
    t->SetMaxTSVld(rh.RunNumber()*10000 + (subrun+1)*100 - 1);	  
  }
  else {
    t->SetMinTSVld(rh.RunNumber()*10000 + 1);
    t->SetMaxTSVld(rh.RunNumber()*10000 + rh.NSubruns()*100);
  }
  
  t->SetVerbosity(100);

  if (!t->Load()) {
    std::cerr << "Failed to load bad channels data!" << std::endl;
    exit(1);
  }
  if (t->NRow() == 0) 
    std::cerr << "No bad channels found!" << std::endl; 

  if (!csvFileName.empty()) 
    t->WriteToCSV(csvFileName);

  t->FillChanRowMap();

  daqchannelmap::DAQChannelMap* daqcmap = daqchannelmap::DAQChannelMap::getInstance(detId);
  
  int ipix=0;
  int state;
  int nbad=0;
  int stateIdx=t->GetColIndex("state");
  
  int nchan=0;
  
  
  if((febN != "") && (febOnly)){
    
    std::string sdib = febN.substr(2,2);
    int idb= atoi(sdib.c_str());
    std::string sdcm = febN.substr(5,2);
    int idcm= atoi(sdcm.c_str());
    std::string sfeb = febN.substr(8,2);
    int ifeb= atoi(sfeb.c_str());
    //std::cout<<"+++ "<<idb<<","<<idcm<<","<<ifeb<<std::endl;
     
    //idb-1 since starts form 0
    nova::dbi::RunHistory::DiBlock dblk = rh.GetDiBlock(idb-1,false);
    
    if (!includeAll)
      if (!rh.IsGoodDiBlock(dblk.num)) return -1; 
    
    if (dblk.dcm[idcm].feb[ifeb].hasAPD) {
      for (ipix=0; ipix<32; ++ipix) {
	daqchannelmap::dchan daqChan = daqcmap->encodeDChan(detId,dblk.num,
							    idcm,ifeb,
							    ipix);
	daqchannelmap::lchan logChan = daqcmap->encodeLChan(daqChan);
	
	int plane = daqcmap->getPlane(logChan);
	int cell = daqcmap->getCell(logChan);
	//std::cout<<"plane: "<<plane<<", cell: "<<cell<<std::endl;
	geo::OfflineChan chan(plane, cell);
	int tchan = chan.ToDBValidityChan();	    
	++nchan;
	
	std::vector<nova::dbi::Row*> rowVec = t->GetVldRows(tchan);
	if (rowVec.empty()) continue;
	assert(rowVec[0]->Col(stateIdx).Get(state));
	//std::cout<<"pixel: "<<ipix<<", state: "<<state<<std::endl;
	if (rowVec.size() == 1 && state == 0) continue; // channel is ok
	  std::cout << detId << "-" << dblk.num << "-" << idcm << ":"
		    << ifeb << "-" << ipix << ": " << rowVec.size() 
		    << " bad channel entries" << std::endl;
	++nbad;
      }
      
    }
  }
  else{
    for (int idb=0; idb<rh.NDiBlocks(); ++idb) {
      nova::dbi::RunHistory::DiBlock dblk = rh.GetDiBlock(idb,false);
      if (!includeAll)
	if (!rh.IsGoodDiBlock(dblk.num)) continue;
      for (unsigned int idcm=0; idcm<dblk.dcm.size(); ++idcm) {
	for (unsigned int ifeb=0; ifeb<dblk.dcm[idcm].feb.size(); ++ifeb) {
	  if (dblk.dcm[idcm].feb[ifeb].hasAPD) {
	    for (ipix=0; ipix<32; ++ipix) {
	      daqchannelmap::dchan daqChan = daqcmap->encodeDChan(detId,dblk.num,
								  idcm+1,ifeb,
								  ipix);
	      daqchannelmap::lchan logChan = daqcmap->encodeLChan(daqChan);
	      
	      int plane = daqcmap->getPlane(logChan);
	      int cell = daqcmap->getCell(logChan);
	      geo::OfflineChan chan(plane, cell);
		int tchan = chan.ToDBValidityChan();	    
		++nchan;
		
		std::vector<nova::dbi::Row*> rowVec = t->GetVldRows(tchan);
		if (rowVec.empty()) continue;
		assert(rowVec[0]->Col(stateIdx).Get(state));
		if (rowVec.size() == 1 && state == 0) continue; // channel is ok
		
		if (!nbadOnly && !fracOnly){
		  std::cout << detId << "-" << dblk.num << "-" << idcm+1 << ":"
			    << ifeb << "-" << ipix << ": " << rowVec.size() 
			    << " bad channel entries" << std::endl;
		}
		
		++nbad;  
	    }
	  }
	}
      }
    }
    
  }//end else
  
  if (!fracOnly && !nbadOnly && !febOnly) {
    std::cout << nbad << " bad channels found for run " << run;
    if (subrun >=0 )
      std::cout << ", subrun " << subrun;
    std::cout << std::endl;
    std::cout << nchan << " total channels" << std::endl;
  }
  
  if (fracOnly)
    std::cout <<"Fraction of bad channels: "<< (float(nbad)/float(nchan)) << std::endl;
  else if (nbadOnly)
    std::cout <<"Number of bad channels: "<< nbad << std::endl;
  else if (febOnly)
    std::cout <<"Number of bad channels: "<< nbad <<" for FEB "<< febN <<std::endl;
  
  return 0;
}
