/////////////////////////////////////////////////////////////////////////
/// \file    CanaFilter.cxx
/// \brief   Filter events based on the Cana analysis
/// \author  messier@indiana.edu
/// \version $Id: CanaFilter.cxx,v 1.12 2012-09-25 16:06:22 gsdavies Exp $
/////////////////////////////////////////////////////////////////////////
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Core/EDFilter.h"
#include "art/Framework/Core/ModuleMacros.h"

#include "Commissioning/CanaNt.h"
#include "DAQDataFormats/TriggerDefines.h"

#include <string>
#include <vector>
#include <iostream>

namespace comi {
  class CanaFilter : public art::EDFilter 
  {
  public:
    explicit CanaFilter(fhicl::ParameterSet const& p);
    virtual ~CanaFilter();

    virtual void reconfigure(fhicl::ParameterSet const& p);
    
    virtual bool filter(art::Event& e);

  private:
    std::string fCanaLabel;///< Label of the Cana module instance              
    float       fT1CutNuMI;///< [usec] Low  edge of NuMI time window		      
    float 	fT2CutNuMI;///< [usec] High edge of NuMI time window		      
    float 	fT1CutBNB; ///< [usec] Low  edge of BNB  time window		      
    float 	fT2CutBNB; ///< [usec] High edge of BNB  time window		      
    float 	fT1CutCAL; ///< [usec] Accept events inside calibration time window 
    float 	fT2CutCAL; ///< [usec] Accept events inside calibration time window 
    int   	fNhitXcut; ///< Require at least this many hits in the window	      
    int   	fNhitYcut; ///< Require at least this many hits in the window	      
    int   	fNplaneX;  ///< Require size of event to be at least this long      
    int   	fNplaneY;  ///< Require size of event to be at least this long      
    float 	fXlo;      ///< Low  edge of "fiducial volume" to select	      
    float 	fXhi;      ///< High edge of "fiducial volume" to select	      
    float 	fYhi;      ///< High edge of "fiducial volume" to select	      
    float 	fZlo;      ///< Low  edge of "fiducial volume" to select	      
    float 	fZhi;      ///< High edge of "fiducial volume" to select            
  };

  //......................................................................

  CanaFilter::CanaFilter(fhicl::ParameterSet const& p) 
  {
    this->reconfigure(p);
  }
  
  //......................................................................
  
  void CanaFilter::reconfigure(fhicl::ParameterSet const& p)
  {
    fT1CutNuMI = p.get<float>("T1CutNuMI");
    fT2CutNuMI = p.get<float>("T2CutNuMI");
    
    fT1CutBNB = p.get<float>("T1CutBNB");
    fT2CutBNB = p.get<float>("T2CutBNB");
    
    fT1CutCAL = p.get<float>("T1CutCAL");
    fT2CutCAL = p.get<float>("T2CutCAL");
    
    fNhitXcut = p.get<int>("NhitXcut");
    fNhitYcut = p.get<int>("NhitYcut");
    
    fNplaneX = p.get<int>("NplaneX");
    fNplaneY = p.get<int>("NplaneY");
    
    fXlo = p.get<float>("Xlo");
    fXhi = p.get<float>("Xhi");
    
    fYhi = p.get<float>("Yhi");
    
    fZlo = p.get<float>("Zlo");
    fZhi = p.get<float>("Zhi");
    
    fCanaLabel = p.get<std::string>("CanaLabel", "cana");
  }
  
  //......................................................................
  
  CanaFilter::~CanaFilter() { }
  
  //......................................................................
  
  bool CanaFilter::filter(art::Event& evt) 
  {
    art::Handle< std::vector<comi::CanaNt> > nth;
    evt.getByLabel(fCanaLabel, nth);
    
    bool isgoodtrk = false;
    bool isgoodevt = false;
    
    for (unsigned register int i=0; i<nth->size(); ++i) {
      const CanaNt& nt = (*nth)[i];
      //
      // Check if this is a good track. Good events will have at least
      // one good track in them
      //
      isgoodtrk = true;
      
      //
      // Apply correct timing cuts for this trigger source. The magic
      // numbers are 0 for NuMI, 1 for BNB and 2 for calibration
      // triggers
      // Eventually try to use TRIGBITs from TriggerDefines.h
      // but the numbers don't match??
      //
      if (nt.trgsrc==0) {
	if (nt.tave<fT1CutNuMI) isgoodtrk = false;
	if (nt.tave>fT2CutNuMI) isgoodtrk = false;
      }
      else if (nt.trgsrc==1) {
	if (nt.tave<fT1CutBNB)  isgoodtrk = false;
	if (nt.tave>fT2CutBNB)  isgoodtrk = false;
      }
      else if (nt.trgsrc==2) {
	if (nt.tave<fT1CutCAL)  isgoodtrk = false;
	if (nt.tave>fT2CutCAL)  isgoodtrk = false;
      }
      else {
	std::cerr << __FILE__ << ":" << __LINE__ 
		  << " Unknown trigger source id=" << nt.trgsrc
		  << std::endl;
	return false;
      }
      
      //
      // Ask for required number of hits in each view
      //
      if (nt.nhitx<fNhitXcut) isgoodtrk = false;
      if (nt.nhity<fNhitYcut) isgoodtrk = false;
      
      //
      // Ask for required number of planes crossed in each view
      //
      if ((abs(nt.plane2x-nt.plane1x))<fNplaneX) isgoodtrk = false;
      if ((abs(nt.plane2y-nt.plane1y))<fNplaneY) isgoodtrk = false;
      
      //
      // Require track start point (assumed down-going) to be inside the
      // defined fiducial volume
      //
      if (nt.x1<fXlo) isgoodtrk = false;
      if (nt.x1>fXhi) isgoodtrk = false;
      if (nt.y1>fYhi) isgoodtrk = false;
      if (nt.z1<fZlo) isgoodtrk = false;
      if (nt.z1>fZhi) isgoodtrk = false;
      
      isgoodevt |= isgoodtrk;
    }
    return isgoodevt;
  }
} // end namespace comi  
////////////////////////////////////////////////////////////////////////
DEFINE_ART_MODULE(comi::CanaFilter);
////////////////////////////////////////////////////////////////////////
