///
/// \file    FEBFlash_module.cc
/// \brief   Module to study FEB flash occurances
/// \author  messier@indiana.edu
///
#include <string>
#include <vector>
#include <map>
#include <iostream>

#include "TH1F.h"
#include "TH2F.h"
#include "TTree.h"

// ART includes
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/SubRun.h"
#include "art/Framework/Services/Optional/TFileService.h"
#include "art/Framework/Core/EDAnalyzer.h"
#include "art/Framework/Core/ModuleMacros.h"

// NOvA includes
#include "RawData/RawDigit.h"
#include "Utilities/RawUtil.h"

namespace comi {
  
  /// Analysis module to hunt for neutrino candidates
  class FEBFlash : public art::EDAnalyzer
  {
  public:
    // NOvA uses a 64 Mhz clock
    const int    kTDC_PER_USEC = 64;
    const double kUSEC_PER_TDC = 1.0/kTDC_PER_USEC;
  public:
    unsigned int fNhitFEBSlice = 2;
    unsigned int fdTFEBSlice   = kTDC_PER_USEC/2;
    unsigned int fNhit0  = 5;
    unsigned int fNsat0  = 1;
    unsigned int fNafter = 0;
    unsigned int fADC0   = 0;
    int          fADCSat = 3400;
  public:
    TH1F* fFilterFrac;   ///< Fraction of hits filtered as FEB flash
    TH1F* fADC;          ///< ADC distribution of hits
    TH1F* fNcand;        ///< Number of candidates found per event
    TH1F* fNFEBSlice;    ///< Time slices found per FEB per event
    TH1F* fNhit;         ///< Number of hits per FEB slice
    TH1F* fNsat;         ///< Number of hits near saturation
    TH1F* fSumADC;       ///< Total ADC in a slice
    TH1F* fNafterCand;   ///< Number of slices following a flash candidate
    TH1F* fNafterNoCand; ///< Number of slices following a non-candidate
    TH1F* fADC1;         ///< Summed ADC for after slices
    TH1F* fDeltaT;       ///< Time differences between flash and after slices
    TH2F* fADC0ADC1;     ///< ADC for slice pairs
    TH2F* fDeltaTADC0;   ///< Time / ADC for slice pairs
    TH2F* fDeltaTADC1;   ///< Time / ADC for slice pairs
    
    //..................................................................
    
    unsigned int
    FilterFEBFlash(std::vector< art::Ptr<rawdata::RawDigit> >& rd)
    {
      unsigned int i, j;
      int deltaTDC      = 30*kTDC_PER_USEC;
      int ADCSaturation = 3400;

      //
      // For efficiency, time sort the hits in advance
      //
      util::TimeSort(rd);
      
      //
      // Flag all hits as good, look for ones that might be bad
      //
      std::vector<bool> isgood(rd.size());
      for (i=0; i<rd.size(); ++i) isgood[i] = true;
      for (i=0; (i+1)<rd.size(); ++i) {
	//
	// If the hit i is not in saturation, we can skip to the next
	// hit
	//
	if (rd[i]->ADC()<ADCSaturation) continue;
	//
	// Get ID number for the FEB ID that has the hit in saturation
	//
	unsigned int chi      = rd[i]->DaqChannel();
	unsigned int diblocki = (chi&0x0FC00000)>>22;
	unsigned int dcmi     = (chi&0x003F0000)>>16;
	unsigned int febi     = (chi&0x0000FF00)>>8;
	for (j=i+1; j<rd.size(); ++j) {
	  //
	  // Don't need to look beyond the time limit
	  //
	  if ((rd[j]->TDC()-rd[i]->TDC())>deltaTDC) break;
	  //
	  // Hit j is in time range of another hit which is in
	  // saturation. Check if hit j comes from the same FEB as hit
	  // i
	  //
	  unsigned int chj      = rd[j]->DaqChannel();
	  unsigned int diblockj = (chj&0x0FC00000)>>22;
	  unsigned int dcmj     = (chj&0x003F0000)>>16;
	  unsigned int febj     = (chj&0x0000FF00)>>8;
	  bool sameFEB = 
	    (diblocki==diblockj) && (dcmi==dcmj) && (febi==febj);
	  if (sameFEB) isgood[j] = false;
	}
      }
      //
      // Push all the good hits into the final list
      //
      std::vector< art::Ptr<rawdata::RawDigit> > good;
      good.reserve(rd.size());
      for (i=0; i<rd.size(); ++i) {
	if (isgood[i]) good.push_back(rd[i]);
      }
      rd.swap(good);
      return rd.size();
    }
    
    //..................................................................
    
    explicit FEBFlash(fhicl::ParameterSet const& p) : EDAnalyzer(p)
    {
      this->reconfigure(p);
    }

    //..................................................................
    
    virtual ~FEBFlash() { }
    
    //..................................................................
    
    void reconfigure(const fhicl::ParameterSet& p) { }
    
    //..................................................................
    
    void beginJob() 
    {
      art::ServiceHandle<art::TFileService> f;

      fFilterFrac = f->make<TH1F>("fFilterFrac",
				  ";fraction of hits filtered;events",
				  1001, -0.0005, 1.0005);
      
      fADC = f->make<TH1F>("fADC",
			   ";ADC counts;number of hits",
			   4096, -0.5, 4095.5);
      
      fNcand = f->make<TH1F>("fNcand",
			     ";number of candidates;events",
			     101, -0.5, 100.5);
      
      fNFEBSlice = f->make<TH1F>("fNFEBSlice",
				 ";number of slices;counts/FEB/events",
				 101, -0.5, 100.5);
      
      fNhit = f->make<TH1F>("fNhit",
			    ";hits per slice;slices/FEB/event",
			    51, -0.5, 50.5);
      
      fNsat = f->make<TH1F>("fNsat",
			    ";saturated hits per slice;slices/FEB/event",
			    33, -0.5, 32.5);
      
      fSumADC = f->make<TH1F>("fSumADC",
			      ";summed ADC;slices/FEB/event",
			      120, 0, 60000);

      fNafterCand = f->make<TH1F>("fNafterCand",
				  ";N after slices : candidate; counts",
				  101, -0.5, 100.5);
      
      fNafterNoCand = f->make<TH1F>("fNafterNoCand",
				    ";N after slices : non-candidate; counts",
				    101, -0.5, 100.5);
      
      fADC1 = f->make<TH1F>("fADC1",
			    ";summed ADC for after slices;counts/FEB/event",
			    120, 0, 60000);
      
      fDeltaT = f->make<TH1F>("fDeltaT",
			      "; time difference (usec); counts/FEB/event",
			      500, 0, 500);
      
      fADC0ADC1 = f->make<TH2F>("fADC0ADC1",
				";summed ADC 0;summed ADC1",
				120, 0, 60000,
				120, 0, 60000);

      fDeltaTADC0 = f->make<TH2F>("fDeltaTADC0",
				  ";deltaT (usec);ADC0",
				  500, 0, 500,
				  120, 0, 60000);
      
      fDeltaTADC1 = f->make<TH2F>("fDeltaTADC1",
				  ";deltaT (usec);ADC1",
				  500, 0, 500,
				  120, 0, 60000);
    }
    
    //......................................................................
    ///
    /// Produce a list of time slices for hits from a single FEB
    ///
    /// d - Input list of hits selected by FEB. Time sorted on return.
    /// s - slices, index pairs into d where significant activity occurs
    ///
    void FEBTimeSlice(std::vector< art::Ptr<rawdata::RawDigit> >& d,
		      std::vector<util::RawSlice>&                s)
    {
      util::TimeSort(d);
      util::TimeSlice(d, fdTFEBSlice, fNhitFEBSlice, 0, 0, s);
    }

    //..................................................................
    ///
    /// Compile some statistics on each of the slices
    ///
    void SliceStatistics(std::vector< art::Ptr<rawdata::RawDigit> >& d,
			 std::vector<util::RawSlice>& s,
			 std::vector<unsigned int>&   nhit,
			 std::vector<unsigned int>&   nsat,
			 std::vector<unsigned int>&   adc,
			 std::vector<double>&         avet,
			 std::vector<unsigned int>&   nafter,
			 std::vector<bool>&           iscandidate)
    {
      unsigned int i, j;
      for (i=0; i<s.size(); ++i) {
	unsigned int n = s[i].second-s[i].first+1;
	nhit.push_back(n);
	
	unsigned int ns     = 0;
	unsigned int na     = 0;
	unsigned int sumadc = 0;
	unsigned int sumtdc = 0;
	for (j=s[i].first; j<=s[i].second; ++j) {
	  if (d[j]->ADC()>=fADCSat) ++ns;
	  sumadc += d[j]->ADC();
	  sumtdc += d[j]->TDC();
	}
	nsat.push_back(ns);
	adc. push_back(sumadc);
	avet.push_back(kUSEC_PER_TDC*sumtdc/(float)n);

	na = s.size()-(i+1);
	nafter.push_back(na);

	//
	// Check if this slice is a candidate for an initial "pow"
	// that triggers afterflashes
	//
	bool isc = (n      >= fNhit0 &&
		    ns     >= fNsat0 &&
		    sumadc >= fADC0  &&
		    na     >= fNafter);
	iscandidate.push_back(isc);
      }
    }

    //..................................................................
    
    void analyze(art::Event const& evt)
    {
      unsigned register int i, j;
      //
      // Pull out the raw digits from the event and shuffle into ART
      // pointers
      //
      art::Handle< std::vector<rawdata::RawDigit> > digidummy;
      evt.getByLabel("daq", digidummy);
      std::vector< art::Ptr<rawdata::RawDigit> > d(digidummy->size());
      for (i=0; i<digidummy->size(); ++i) {
	d[i] = art::Ptr<rawdata::RawDigit>(digidummy,i);
      }
      
      //
      // If requested, filter out flash hits
      //
      bool fFilterFEBFlash = true;
      unsigned int n1 = 0; // Number of hits before filter
      unsigned int n2 = 0; // Number of hits after filter
      n1 = d.size();
      if (fFilterFEBFlash) {
	std::cerr << "Filtering FEB flash hits...";
	n2 = this->FilterFEBFlash(d);
	std::cerr << "done." << std::endl;
      }
      fFilterFrac->Fill((float)(n1-n2)/(float)n1);
      
      //
      // Shuffle the hits around so that we can examine the time
      // structure of the hits one FEB at a time
      //
      typedef std::map<
	unsigned int,
	std::vector< art::Ptr<rawdata::RawDigit> >
	> HitsByFEB_t;
      
      HitsByFEB_t digi;
      for (i=0; i<d.size(); ++i) {
	unsigned int ch      = d[i]->DaqChannel();
	unsigned int diblock = (ch&0x0FC00000)>>22;
	unsigned int dcm     = (ch&0x003F0000)>>16;
	unsigned int feb     = (ch&0x0000FF00)>>8;
	unsigned int FEBID   = feb+100*(dcm+100*diblock);

	fADC->Fill(d[i]->ADC());

	digi[FEBID].push_back(d[i]);
      }
      
      unsigned int ncand = 0;
      HitsByFEB_t::iterator itr   (digi.begin());
      HitsByFEB_t::iterator itrEnd(digi.end());
      for (; itr!=itrEnd; ++itr) {
	//
	// Give myself easy access to the hits for just one FEB
	//
	std::vector<art::Ptr<rawdata::RawDigit>>& d(itr->second);
	
	//
	// Get list of time slices with significant activity
	//
	std::vector<util::RawSlice> slice;
	this->FEBTimeSlice(d, slice);
	fNFEBSlice->Fill(slice.size());
	
	//
	// Compile some statistics on each slice; nhit, ADC, time, ...
	//
	std::vector<unsigned int> nhit; // # of hits in slice
	std::vector<unsigned int> nsat; // # of hits near saturation
	std::vector<unsigned int> adc;  // Summed ADC for slice
	std::vector<double>       avet; // Slice time
	std::vector<bool>         iscandidate; // "pow!" candidate?
	std::vector<unsigned int> nafter;
	this->SliceStatistics(d,
			      slice,
			      nhit,
			      nsat,
			      adc,
			      avet,
			      nafter,
			      iscandidate);
	//
	// Log statistics to histograms for analysis
	//
	for (i=0; i<slice.size(); ++i) {
	  fNhit->Fill(nhit[i]);
	  fNsat->Fill(nsat[i]);
	  fSumADC->Fill(adc[i]);
	  if (iscandidate[i]) fNafterCand->  Fill(nafter[i]);
	  else                fNafterNoCand->Fill(nafter[i]);
	  if (iscandidate[i]) {
	    ++ncand;
	    for (j=i+1; j<slice.size(); ++j) {
	      fADC1->Fill(adc[j]);
	      fDeltaT->Fill(avet[j]-avet[i]);
	      fADC0ADC1->Fill(adc[i], adc[j]);
	      fDeltaTADC0->Fill(avet[j]-avet[i],adc[i]);
	      fDeltaTADC1->Fill(avet[j]-avet[i],adc[j]);
	      std::cout << "FLASH-CANDIDATE:\t"
			<< evt.run() << "\t"
			<< evt.subRun() << "\t"
			<< evt.id().event() << "\t"
			<< itr->first << "\t"
			<< nhit[i] << "\t"
			<< nsat[i] << "\t"
			<< adc[i] << "\t"
			<< avet[i] << "\t"
			<< nafter[i] << "\t"
			<< nhit[j] << "\t"
			<< nsat[j] << "\t"
			<< adc[j] << "\t"
			<< avet[j] << "\t"
			<< nafter[j] << std::endl;
	    } // loop on slice j
	  } // if is candidate
	} // loop on slice i
      } // loop over FEBs
      fNcand->Fill(ncand);
      
      //
      // Make sure log file has at least one entry per event
      //
      if (ncand==0) {
	std::cout << "FLASH-CANDIDATE:\t"
		  << evt.run() << "\t"
		  << evt.subRun() << "\t"
		  << evt.id().event() << "\t"
		  << "-" << "\t"
		  << "-" << "\t"
		  << "-" << "\t"
		  << "-" << "\t"
		  << "-" << "\t"
		  << "-" << "\t"
		  << "-" << "\t"
		  << "-" << "\t"
		  << "-" << "\t"
		  << "-" << "\t"
		  << "-" << std::endl;
      }
    } // end analyze function
  }; // end FEBFlash class
} // end namespace comi
////////////////////////////////////////////////////////////////////////
DEFINE_ART_MODULE(comi::FEBFlash);
////////////////////////////////////////////////////////////////////////
