///////////////////////////////////////////////////////////////////////
// \file    NearlineAna_module.cc
// \brief   Analysis module to make the Nearline Plots
// \version $Id: NearlineAna_module.cc,v 1.0 2013-10-09 14:51:38 mbaird42 Exp $
// \author  $Author: mbaird42 $
////////////////////////////////////////////////////////////////////////

// C/C++ includes
#include <cmath>
#include <iostream>

// ROOT includes
#include "TH1F.h"
#include "TH1D.h"
#include "TH2F.h"
#include "TTree.h"
#include "TTimeStamp.h"   // needed for printing the date/time
#include "TVector3.h"

// Framework includes
#include "art/Framework/Core/EDAnalyzer.h"
#include "art/Framework/Core/FindManyP.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Services/Optional/TFileDirectory.h"
#include "art/Framework/Services/Optional/TFileService.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "art/Persistency/Common/Ptr.h"
#include "art/Persistency/Common/PtrVector.h"
#include "fhiclcpp/ParameterSet.h"
#include "messagefacility/MessageLogger/MessageLogger.h"

// NOvASoft includes
#include "Geometry/Geometry.h"
#include "GeometryObjects/CellGeo.h"
#include "NovaDAQConventions/DAQConventions.h"
#include "RecoBase/CellHit.h"
#include "RecoBase/Cluster.h"
#include "RecoBase/Track.h"
#include "SummaryData/SpillData.h"          // needed for getting spill data

#include "Utilities/AssociationUtil.h"
#include "Utilities/func/MathUtil.h"
#include "DAQDataFormats/RawEvent.h"
#include "DAQDataFormats/RawTrigger.h"
#include "RawData/FlatDAQData.h"

// needed for DCM sync status plots
#include "CMap/CMap.h"
#include "DAQChannelMap/DAQChannelMap.h"
#include "NovaDAQConventions/DAQConventions.h"
#include "RunHistory/service/RunHistoryService.h"      //Map of Instrumented Detector


static const double ns = 1.0E-9;

// NearlineAna header
namespace comi {
  class NearlineAna : public art::EDAnalyzer {
  public:
    
    bool IsContained(double x, double y, double z);
    
    explicit NearlineAna(fhicl::ParameterSet const& pset);    
    ~NearlineAna();                        
    
    void analyze(const art::Event& evt); 
    void beginJob();
    void endJob();
    void beginRun(const art::Run& run);
    
  private:
    
    int           fMinTrackHits;      ///< minimum # of hits from a track within a DCM for the track to count as having passed through that DCM
    std::string   fSlicerLabel;       ///< Where to find Clusters
    std::string   fCellHitsListLabel; ///< Where to find CellHits    
    
    fhicl::ParameterSet fPSetFD, fPSetND, fPSetNDOS;  ///< Parameter sets for each detector
    
    std::string   fTrackLabel;        ///< Where to find Tracks
    double fMinX;   ///< containment cut in X
    double fMaxX;   ///< containment cut in X
    double fMinY;   ///< containment cut in Y
    double fMaxY;   ///< containment cut in Y
    double fMinZ;   ///< containment cut in Z
    double fMaxZ;   ///< containment cut in Z
    double fP1MaxZ; ///< containment cut in Z for partition 1
    double fMinT;   ///< for cutting in T
    double fMaxT;   ///< for cutting in T
    int    fMinFEBs;  ///< NEW Minimum number of FEBs in a DCM for allow diblock instrumented
    
    double fMinPOTCut;    ///< min POT cut for beam quality
    double fMinHornICut;  ///< min horn current cut for beam quality
    double fMaxHornICut;  ///< max horn current cut for beam quality
    double fMinPosXCut;   ///< min x position cut for beam quality
    double fMaxPosXCut;   ///< max x position cut for beam quality
    double fMinPosYCut;   ///< min y position cut for beam quality
    double fMaxPosYCut;   ///< max y position cut for beam quality
    double fMinWidthXCut; ///< min x width cut for beam quality
    double fMaxWidthXCut; ///< max x width cut for beam quality
    double fMinWidthYCut; ///< min y width cut for beam quality
    double fMaxWidthYCut; ///< max y width cut for beam quality
    double fMaxDeltaTCut; ///< mqax delta t cut for beam quality
    
    // Header TTree variables etc.
    TTree* fHeader;           ///< Header object to be written to disk with the histograms
    unsigned int fLastRun;    ///< Header: last run number analyzed
    unsigned int fLastSubrun; ///< Header: last subrun number analyzed
    unsigned int fPartition;  ///< Header: partition number from last event analyzed
    int          fFirstEvent; ///< Header: first event processed
    int          fLastEvent;  ///< Header: last event processed
    int          fNevents;    ///< Header: number of events processed
    unsigned long long int fStartTime; ///< Header: time of first processed event - initalized to -1 to make it a HUGE positive number
    unsigned long long int fEndTime;   ///< Header: time of last processed event
    double       fLiveTime;   ///< Header: total live time in seconds
    unsigned int fStartYear;  ///< Header: year of first processed event (actual year since 0 A.D.)
    unsigned int fStartMonth; ///< Header: month of first processed event (actual month 1-12)
    unsigned int fStartDay;   ///< Header: day of first processed event (day of the month 1-31)
    double       fStartHour;  ///< Header: hour of first processed event (since midnight 0.0 - 23.99999)
    unsigned int fEndYear;    ///< Header: year of last processed event (actual year since 0 A.D.)
    unsigned int fEndMonth;   ///< Header: month of last processed event (actual month 1-12)
    unsigned int fEndDay;     ///< Header: day of last processed event (day of the month 1-31)
    double       fEndHour;    ///< Header: hour of last processed event (since midnight 0.0 - 23.99999)
    
    //
    // Histograms
    //
    // beam histos
    
    //NEW CODE
    TH1D * fTimeDiffNanoSec;          ///< time difference between the spill times recorded as event timestamp and by IFDB
    TH1D * fPOTSum;                   ///< sum of POT
    TH1D * fspillPOT;                 ///< spill POT
    TH1D * fHornCurrent;              ///< value of horn current
    TH1D * fXPosition;                ///< horizontal position of beam at target
    TH1D * fYPosition;                ///< vertical position of beam at target
    TH1D * fXWidth;                   ///< horizontal width of beam at TRTGT
    TH1D * fYWidth;                   ///< vertical width of beam at TRTGT
    TH1D * fGoodBeam;                 ///< value of goodbeam parameter, ie whether spill passed beam quality cuts
    TH1I * fBadSpills;                ///< failure modes for those spills that don't pass the beam quality cuts
    
    // Slice histos
    TH1D *fNumSlices;                   ///< number of non-noise slices
    TH1D *fNumSlicesGoodBeam;                   ///< number of non-noise slices for good beam
    TH1D *fNumHits;                     ///< number of hits per non noise slice
    TH1D *fNumHitsNoise;                ///< number hits in the noise slice
    TH1D *fNumHitsX;                    ///< number of X hits in non-noise slices
    TH1D *fNumHitsXNoise;               ///< number of X hits in the noise slice
    TH1D *fNumHitsY;                    ///< number of Y hits in non-noise slices
    TH1D *fNumHitsYNoise;               ///< number of Y hits in the noise slice
    
    TH1D *fADCAllHits;                  ///< ADC for all hits
    TH1D *fADCNoiseHits;                ///< ADC for hits in the noise slice
    TH1D *fADCNonNoiseHits;             ///< ADC for hits in non-noise slices
    TH1D *fPEAllHits;                   ///< PE for all hits
    TH1D *fPENoiseHits;                 ///< PE for hits in the noise slice
    TH1D *fPENonNoiseHits;              ///< PE for hits in non-noise slices
    
    TH2D *fPCXave;                      ///< for each non-noise slice, average plane/cell of hits
    TH2D *fPCYave;                      ///< for each non-noise slice, average plane/cell of hits
    TH2D *fPCXaveCR;                    ///< for each non-noise slice, average plane/cell of hits (control room view - meaning X is flipped and Z is flipped for both views)
    TH2D *fPCYaveCR;                    ///< for each non-noise slice, average plane/cell of hits (control room view)
    TH1D *fTave;                        ///< for each non-noise slice, average time of hits
    TH2D *fPCXdiff;                     ///< for each non-noise slice, difference btwn highest and lowest plane/cell
    TH2D *fPCYdiff;                     ///< for each non-noise slice, difference btwn highest and lowest plane/cell
    TH1D *fTsd;                         ///< for each non-noise slice, standard dev. of hit times
    
    // Histos comparing slices and tracks
    TH1D *fDeltaSliceTrack;             ///< number of slices minus number of 3D tracks 
    TH1D *fSliceTrackRatio;             ///< ratio of number of 3D tracks to number of slices
    TH1D *fSliceTrackNHitRatio;         ///< ratio of number of hits on 3D track to number of hits in the slice (0 means no track or 2D track)
    TH2D *fSliceTrackNHitRatioVSnhit;   ///< ratio of number of hits on 3D track to number of hits in the slice, 2D plot vs nhit in slice (0 means no track or 2D track)
    
    // Tracks histos
    TH1D *fTrackLengthAll2D;            ///< length of track [cm] (all 2D tracks)
    TH1D *fTrackCosZenithAll2D;         ///< cosine theta of track with respect to zenith direction (all 2D tracks)
    TH1D *fTrackCosAzimuthAll2D;        ///< cosine theta of track with respect to azimuth direction (all 2D tracks)
    TH2D *fTrackCosZenAziAll2D;         ///< 2D plot of cos zentih and cos azimuth (all 2D tracks)
    TH1D *fTrackCosNumiAll2D;           ///< cosine theta of track with respect to NuMI beam (all 2D tracks)
    TH1D *fTrackZenithAll2D;            ///< track zenith angle (all 2D tracks)
    TH1D *fTrackAzimuthAll2D;           ///< track azimuthal angle (all 2D tracks)
    TH2D *fTrackZenAziAll2D;            ///< 2D plot of zentih and azimuth (all 2D tracks)
    TH1D *fTrackNumiAll2D;              ///< track numi angle (all 2D tracks)
    TH1D *fNumTracksAll2D;              ///< counter for number of tracks (all 2D tracks)
    TH1D *fTrackFractionAll2D;          ///< fraction of total tracks that are 2D
    TH2D *fStartPointTrackXZAll2D;      ///< 2D histogram showing starting point of reco track in XZ view (all 2D tracks)
    TH2D *fStartPointTrackYZAll2D;      ///< 2D histogram showing starting point of reco track in YZ view (all 2D tracks)
    TH2D *fStopPointTrackXZAll2D;       ///< 2D histogram showing stopping point of reco track in XZ view (all 2D tracks)
    TH2D *fStopPointTrackYZAll2D;       ///< 2D histogram showing stopping point of reco track in YZ view (all 2D tracks)
    TH2D *fStartPointTrackXZCRAll2D;    ///< 2D histogram showing starting point of reco track in XZ view (control room view) (all 2D tracks)
    TH2D *fStartPointTrackYZCRAll2D;    ///< 2D histogram showing starting point of reco track in YZ view (control room view) (all 2D tracks)
    TH2D *fStopPointTrackXZCRAll2D;     ///< 2D histogram showing stopping point of reco track in XZ view (control room view) (all 2D tracks)
    TH2D *fStopPointTrackYZCRAll2D;     ///< 2D histogram showing stopping point of reco track in YZ view (control room view) (all 2D tracks)
    TH1D *fPETrackHitsAll2D;            ///< PE for hits on tracks (all 2D tracks)
    TH1D *fTtrkAll2D;                   ///< track times (all 2D tracks)
    
    TH1D *fTrackLengthAll3D;            ///< length of track [cm] (all 3D tracks)
    TH1D *fTrackCosZenithAll3D;         ///< cosine theta of track with respect to zenith direction (all 3D tracks)
    TH1D *fTrackCosAzimuthAll3D;        ///< cosine theta of track with respect to azimuth direction (all 3D tracks)
    TH2D *fTrackCosZenAziAll3D;         ///< 2D plot of cos zentih and cos azimuth (all 3D tracks)
    TH1D *fTrackCosNumiAll3D;           ///< cosine theta of track with respect to NuMI beam (all 3D tracks)
    TH1D *fTrackZenithAll3D;            ///< track zenith angle (all 3D tracks)
    TH1D *fTrackAzimuthAll3D;           ///< track azimuthal angle (all 3D tracks)
    TH2D *fTrackZenAziAll3D;            ///< 2D plot of zentih and azimuth (all 3D tracks)
    TH1D *fTrackNumiAll3D;              ///< track numi angle (all 3D tracks)
    TH1D *fNumTracksAll3D;              ///< counter for number of tracks (all 3D tracks)
    TH1D *fTrackFractionAll3D;          ///< fraction of total tracks that are 3D
    TH2D *fStartPointTrackXZAll3D;      ///< 2D histogram showing starting point of reco track in XZ view (all 3D tracks)
    TH2D *fStartPointTrackYZAll3D;      ///< 2D histogram showing starting point of reco track in YZ view (all 3D tracks)
    TH2D *fStopPointTrackXZAll3D;       ///< 2D histogram showing stopping point of reco track in XZ view (all 3D tracks)
    TH2D *fStopPointTrackYZAll3D;       ///< 2D histogram showing stopping point of reco track in YZ view (all 3D tracks)
    TH2D *fStartPointTrackXZCRAll3D;    ///< 2D histogram showing starting point of reco track in XZ view (control room view) (all 3D tracks)
    TH2D *fStartPointTrackYZCRAll3D;    ///< 2D histogram showing starting point of reco track in YZ view (control room view) (all 3D tracks)
    TH2D *fStopPointTrackXZCRAll3D;     ///< 2D histogram showing stopping point of reco track in XZ view (control room view) (all 3D tracks)
    TH2D *fStopPointTrackYZCRAll3D;     ///< 2D histogram showing stopping point of reco track in YZ view (control room view) (all 3D tracks)
    TH1D *fPETrackHitsAll3D;            ///< PE for hits on tracks (all 3D tracks)
    TH1D *fTtrkAll3D;                   ///< track times (all 3D tracks)
    
    TH1D *fTrackLengthCont3D;           ///< length of track [cm] (all contained 3D tracks)
    TH1D *fTrackCosZenithCont3D;        ///< cosine theta of track with respect to zenith direction (all contained 3D tracks)
    TH1D *fTrackCosAzimuthCont3D;       ///< cosine theta of track with respect to azimuth direction (all contained 3D tracks)
    TH2D *fTrackCosZenAziCont3D;        ///< 2D plot of cos zentih and cos azimuth (all contained 3D tracks)
    TH1D *fTrackCosNumiCont3D;          ///< cosine theta of track with respect to NuMI beam (all contained 3D tracks)
    TH1D *fTrackZenithCont3D;           ///< track zenith angle (all contained 3D tracks)
    TH1D *fTrackAzimuthCont3D;          ///< track azimuthal angle (all contained 3D tracks)
    TH2D *fTrackZenAziCont3D;           ///< 2D plot of zentih and azimuth (all contained 3D tracks)
    TH1D *fTrackNumiCont3D;             ///< track numi angle (all contained 3D tracks)
    TH1D *fNumTracksCont3D;             ///< counter for number of tracks (all contained 3D tracks)
    TH1D *fTrackFractionCont3D;         ///< fraction of total tracks that are contained 3D
    TH2D *fStartPointTrackXZCont3D;     ///< 2D histogram showing starting point of reco track in XZ view (all contained 3D tracks)
    TH2D *fStartPointTrackYZCont3D;     ///< 2D histogram showing starting point of reco track in YZ view (all contained 3D tracks)
    TH2D *fStopPointTrackXZCont3D;      ///< 2D histogram showing stopping point of reco track in XZ view (all contained 3D tracks)
    TH2D *fStopPointTrackYZCont3D;      ///< 2D histogram showing stopping point of reco track in YZ view (all contained 3D tracks)
    TH2D *fStartPointTrackXZCRCont3D;   ///< 2D histogram showing starting point of reco track in XZ view (control room view) (all contained 3D tracks)
    TH2D *fStartPointTrackYZCRCont3D;   ///< 2D histogram showing starting point of reco track in YZ view (control room view) (all contained 3D tracks)
    TH2D *fStopPointTrackXZCRCont3D;    ///< 2D histogram showing stopping point of reco track in XZ view (control room view) (all contained 3D tracks)
    TH2D *fStopPointTrackYZCRCont3D;    ///< 2D histogram showing stopping point of reco track in YZ view (control room view) (all contained 3D tracks)
    TH1D *fPETrackHitsCont3D;           ///< PE for hits on tracks (all contained 3D tracks)
    TH1D *fTtrkCont3D;                  ///< track times (all contained 3D tracks)
    
    // DCM sync status stuff
    int  counter2D[14][12];   ///< counter array for the number of hits/tracks per dcm
    int  counter3D[14][12];   ///< counter array for the number of hits/tracks per dcm
    TH2D *f2DHitNum;          ///< Number of hits on 2D tracks per DCM
    TH2D *f3DHitNum;          ///< Number of hits on 3D tracks per DCM
    
    TH2D *f2DTrackNum;        ///< Number of 3D tracks per DCM
    TH2D *f3DTrackNum;        ///< Number of 2D tracks per DCM
    
    TH2F *Instr;              ///< Map view of the instrumented detector (DCM w/ > 50APDs)
    TH2F *fullyinstr;         ///< Counter of fully instrumented diblocks
    
  };
}



namespace comi
{
  //.......................................................................
  NearlineAna::NearlineAna(fhicl::ParameterSet const& pset) :
    EDAnalyzer(pset),
    fMinTrackHits     (pset.get< int >("MinTrackHits")              ),
    fSlicerLabel      (pset.get< std::string  >("SlicerLabel")      ),
    fCellHitsListLabel(pset.get< std::string  >("CellHitsListLabel")),    
    fPSetFD           (pset.get< fhicl::ParameterSet >("fd")        ),
    fPSetND           (pset.get< fhicl::ParameterSet >("nd")        ),
    fPSetNDOS         (pset.get< fhicl::ParameterSet >("ndos")      ),
    fTrackLabel("none"),
    fMinX(0),
    fMaxX(0),
    fMinY(0),
    fMaxY(0),
    fMinZ(0),
    fMaxZ(0),
    fP1MaxZ(0),
    fMinT(0),
    fMaxT(0),
    fMinFEBs(0),
    fMinPOTCut(0),
    fMinHornICut(0),
    fMaxHornICut(0),
    fMinPosXCut(0),
    fMaxPosXCut(0),
    fMinPosYCut(0),
    fMaxPosYCut(0),
    fMinWidthXCut(0),
    fMaxWidthXCut(0),
    fMinWidthYCut(0),
    fMaxWidthYCut(0),
    fMaxDeltaTCut(0),
    fHeader(0),
    fLastRun(0),
    fLastSubrun(0),
    fPartition(4),
    fFirstEvent(1e9),
    fLastEvent(-1),
    fNevents(0),
    fStartTime(-1),
    fEndTime(0),
    fLiveTime(0.0)
  { }
  
  //......................................................................
  NearlineAna::~NearlineAna()
  { }
  
  //......................................................................
  void NearlineAna::beginJob()
  {
    
    art::ServiceHandle<art::TFileService> tfs;
    
    // make header
    fHeader = tfs->make<TTree>("Header","Subrun Information");
    
    fHeader->Branch("Run",&fLastRun);
    fHeader->Branch("Subrun",&fLastSubrun);
    fHeader->Branch("Partition",&fPartition);
    fHeader->Branch("FirstEvent",&fFirstEvent);
    fHeader->Branch("LastEvent",&fLastEvent);
    fHeader->Branch("Nevents",&fNevents);
    fHeader->Branch("LiveTime",&fLiveTime);
    fHeader->Branch("StartYear",&fStartYear);
    fHeader->Branch("StartMonth",&fStartMonth);
    fHeader->Branch("StartDay",&fStartDay);
    fHeader->Branch("StartHour",&fStartHour);
    fHeader->Branch("EndYear",&fEndYear);
    fHeader->Branch("EndMonth",&fEndMonth);
    fHeader->Branch("EndDay",&fEndDay);
    fHeader->Branch("EndHour",&fEndHour);
    // I would like to add these two variables to the header object but
    // root complains that an unsigned long long int is an unacceptable type.
    //
    // fHeader->Branch("StartTime", &fStartTime);
    // fHeader->Branch("EndTime", &fEndTime);
    
    // book the histograms
    
    
    //NEW CODE
    fTimeDiffNanoSec      = tfs->make<TH1D>("fTimeDiffNanoSec",
					    "Time Difference;Time difference (sec);count",
					    500,-0.1,0.6);
    
    fPOTSum               = tfs->make<TH1D>("fPOTSum",
					    "Sum POT;Sum POT;count",
					    2,0,1);
    
    fspillPOT             = tfs->make<TH1D>("fspillPOT",
					    "Spill POT;Spill POT;count",
					    200,0,50e12);
    
    fHornCurrent          = tfs->make<TH1D>("fHornCurrent",
					    "Horn Current (kA);Current(kA);count",
					    100,-203,-197);
    
    fXPosition            = tfs->make<TH1D>("fXPosition",
					    "Horizontal Beam Position (mm);Position (mm);count",
					    240,0.0,2.1);
      
    fYPosition            = tfs->make<TH1D>("fYPosition",
					    "Vertical Beam Position (mm);Position (mm);count",
					    240,0.0,2.1);
    
    fXWidth               = tfs->make<TH1D>("fXWidth",
					    "Horizontal Beam Width;Width (mm);count",
					    120,0.5,1.7);
    
    fYWidth               = tfs->make<TH1D>("fYWidth",
					    "Vertical Beam Width;Width (mm); count",
					    120,0.5,1.7);
    
    fGoodBeam             = tfs->make<TH1D>("fGoodBeam",
					    "Good Beam;Good Beam;count",
					    2,0,2);
    
    fBadSpills            =tfs->make<TH1I>("fBadSpills",
					   "Failure Mode for Bad Spills;;count",
					   8,1,9);
    
    fBadSpills->SetStats(kFALSE);
    
    
    TAxis *badspillsxaxis = fBadSpills->GetXaxis();
    badspillsxaxis->SetBinLabel(1,"delta t");
    badspillsxaxis->SetBinLabel(2,"POT");
    badspillsxaxis->SetBinLabel(3,"horn I");
    badspillsxaxis->SetBinLabel(4,"x position");
    badspillsxaxis->SetBinLabel(5,"y position");
    badspillsxaxis->SetBinLabel(6,"x width");
    badspillsxaxis->SetBinLabel(7,"y width");
    badspillsxaxis->SetBinLabel(8,"Total Bad Spills");
    
      
    fNumSlices      = tfs->make<TH1D>("fNumSlices",
				      "# of non-noise slices;# of non-noise slices;count",
				      501,-0.5,500.5);
    fNumSlicesGoodBeam      = tfs->make<TH1D>("fNumSlicesGoodBeam",
				      "# of non-noise slices Good Beam;# of non-noise slices good beam;count",
				      501,-0.5,500.5);
    fNumHits        = tfs->make<TH1D>("fNumHits",
				      "# of hits per non-noise slice;# of hits;count",
				      1000,0.0,5000.0);
    fNumHitsNoise   = tfs->make<TH1D>("fNumHitsNoise",
				      "# of hits in the noise slice;# of hits;count",
				      8000,0.0,40000.0);
    fNumHitsX       = tfs->make<TH1D>("fNumHitsX",
				      "# of X hits per non-noise slice;# of hits;count",
				      1000,0.0,5000.0);
    fNumHitsXNoise  = tfs->make<TH1D>("fNumHitsXNoise",
				      "# of X hits in the noise slice;# of hits;count",
				      4000,0.0,20000.0);
    fNumHitsY       = tfs->make<TH1D>("fNumYHits",
				      "# of Y hits per non-noise slice;# of hits;count",
				      1000,0.0,5000.0);
    fNumHitsYNoise  = tfs->make<TH1D>("fNumHitsYNoise",
				      "# of Y hits in the noise slice;# of hits;count",
				      4000,0.0,20000.0);   

    fADCAllHits      = tfs->make<TH1D>("fADCAllHits",
				       "ADC for All Hits;ADC;count",
				       4096,0,4096);
    fADCNoiseHits    = tfs->make<TH1D>("fADCNoiseHits",
				       "ADC for Hits in the Noise Slice;ADC;count",
				       4096,0,4096);
    fADCNonNoiseHits = tfs->make<TH1D>("fADCNonNoiseHits",
				       "ADC for Hits in Non-Noise Slices;ADC;count",
				       4096,0,4096);
    fPEAllHits       = tfs->make<TH1D>("fPEAllHits",
				       "PE for All Hits;PE;count",
				       1862,0,4096);
    fPENoiseHits     = tfs->make<TH1D>("fPENoiseHits",
				       "PE for Hits in the Noise Slice;PE;count",
				       1862,0,4096);
    fPENonNoiseHits  = tfs->make<TH1D>("fPENonNoiseHits",
				       "PE for Hits in Non-Noise Slices;PE;count",
				       1862,0,4096);

    fPCXave   = tfs->make<TH2D>("fPCXave",
				"ave hit plane vs. ave hit X cell;plane;X cell",
				660,-20.0,1300.0,255,-10.0,500.0);
    fPCYave   = tfs->make<TH2D>("fPCYave",
				"ave hit plane vs. ave hit Y cell;plane;Y cell",
				660,-20.0,1300.0,255,-10.0,500.0);
    fPCXaveCR = tfs->make<TH2D>("fPCXaveCR",
				"ave hit plane vs. ave hit X cell (control room view);plane;X cell",
				660,-1300.0,20.0,255,-500.0,10.0);
    fPCYaveCR = tfs->make<TH2D>("fPCYaveCR",
				"ave hit plane vs. ave hit Y cell (control room view);plane;Y cell",
				660,-1300.0,20.0,255,-10.0,500.0);

    fTave    = tfs->make<TH1D>("fTave",
			       "ave hit time per slice; time [ns]; count",
			       1200,-50000.0,550000.0);
    fPCXdiff = tfs->make<TH2D>("fPCXdiff",
			       "diff btwn hi and lo for plane vs X cell per slice;delta plane;delta X cell",
			       500,0.0,1000.0,250,0.0,500.0);
    fPCYdiff = tfs->make<TH2D>("fPCYdiff",
			       "diff btwn hi and lo for plane vs Y cell per slice;delta plane;delta Y cell",
			       500,0.0,1000.0,250,0.0,500.0);
    fTsd     = tfs->make<TH1D>("fTsd",
			       "st. dev. of  hit time per slice; sigma time [ns]; count",
			       500000,0.0,500000.0);


    fDeltaSliceTrack           = tfs->make<TH1D>("fDeltaSliceTrack",
						 "N-slices minus N-3D-tracks per Trigger;(N-sl) - (N-3D-trk);count",
						 201,-100.5,100.5);
    fSliceTrackRatio           = tfs->make<TH1D>("fSliceTrackRatio",
						 "Ratio of N-3D-tracks to N-slices per Trigger;(N-3D-tracks)/(N-slices);count",
						 2000,0.0,20.0);
    fSliceTrackNHitRatio       = tfs->make<TH1D>("fSliceTrackNHitRatio",
						 "NHit Ratio from 3D-tracks to Slices per Slice;(NHit-3D-tracks)/(NHit-slices);count",
						 101,0.0,1.01);
    fSliceTrackNHitRatioVSnhit = tfs->make<TH2D>("fSliceTrackNHitRatioVSnhit",
						 "NHit Ratio from 3D-tracks to Slices per Slice VS. NHit-Slices;NHit-Slice;(NHit-3D-tracks)/(NHit-slices)",
						 200,0.0,2000.0,101,0.0,1.01);



    fTrackLengthAll3D     = tfs->make<TH1D>("fTrackLengthAll3D",
					    "Reconstructed Track Length distribution;Track Length[cm]; count",
					    650,0.0,6500.0);
    fTrackCosZenithAll3D  = tfs->make<TH1D>("fTrackCosZenithAll3D",
					    "Cos #theta with respect to zenith; Cos #theta_{Ze};count",
					    202,-1.01,1.01);
    fTrackCosAzimuthAll3D = tfs->make<TH1D>("fTrackCosAzimuthAll3D",
					    "Cos #theta with respect to azimuth (in detector coordinates); Cos #theta_{Az};count",
					    202,-1.01,1.01);
    fTrackCosZenAziAll3D  = tfs->make<TH2D>("fTrackCosZenAziAll3D",
					    "Cos #theta_{Zenith} vs. Cos #theta_{Azimuth} (in detector coordinates);Cos #theta_{Ze};Cos #theta_{Az}",
					    202,-1.01,1.01,202,-1.01,1.01);
    fTrackCosNumiAll3D    = tfs->make<TH1D>("fTrackCosNumiAll3D",
					    "Cos #theta with respect to NuMI beam; Cos #theta_{NuMI};count",
					    202,-1.01,1.01);
    fTrackZenithAll3D     = tfs->make<TH1D>("fTrackZenithAll3D",
					    "Zenith Angle;#theta_{Ze};count",
					    91,-0.5,180.5);
    fTrackAzimuthAll3D    = tfs->make<TH1D>("fTrackAzimuthAll3D",
					    "Azimuth Angle (in detector coordinates);#theta_{Az};count",
					    181,-0.5,360.5);
    fTrackZenAziAll3D     = tfs->make<TH2D>("fTrackZenAziAll3D",
					    "#theta_{Zenith} vs. #theta_{Azimuth} (in detector coordinates);#theta_{Ze};#theta_{Az}",
					    91,-0.5,180.5,181,-0.5,360.5);
    fTrackNumiAll3D       = tfs->make<TH1D>("fTrackNumiAll3D",
					    "NuMI Angle;#theta_{NuMI};count",
					    91,-0.5,180.5);

    fNumTracksAll3D       = tfs->make<TH1D>("fNumTracksAll3D",
					    ";No. Tracks;Events",
					    301,-0.5,300.5);
    fTrackFractionAll3D   = tfs->make<TH1D>("fTrackFractionAll3D",
					    ";Track Fraction;Events",
					    101,0.0,1.01);

    fStartPointTrackXZAll3D  = tfs->make<TH2D>("fStartPointTrackXZAll3D",
					       "Starting Point of Reco Track in XZ view; Start Z [cm]; Start X [cm]",
					       630,0.,6300,160,-800.,800.);
    fStartPointTrackYZAll3D  = tfs->make<TH2D>("fStartPointTrackYZAll3D",
					       "Starting Point of Reco Track in YZ view; Start Z [cm]; Start Y [cm]",
					       630,0.,6300,160,-800.,800.);
    fStopPointTrackXZAll3D   = tfs->make<TH2D>("fStopPointTrackXZAll3D",
					       "Stopping Point of Reco Track in XZ view; Start Z [cm]; Start X [cm]",
					       630,0.,6300,160,-800.,800.);
    fStopPointTrackYZAll3D   = tfs->make<TH2D>("fStopPointTrackYZAll3D",
					       "Stopping Point of Reco Track in YZ view; Start Z [cm]; Start Y [cm]",
					       630,0.,6300,160,-800.,800.);

    fStartPointTrackXZCRAll3D = tfs->make<TH2D>("fStartPointTrackXZCRAll3D",
						"Starting Point of Reco Track in XZ view (control room view); Start Z [cm]; Start X [cm]",
						630,-6300,0.0,160,-800.,800.);
    fStartPointTrackYZCRAll3D = tfs->make<TH2D>("fStartPointTrackYZCRAll3D",
						"Starting Point of Reco Track in YZ view (control room view); Start Z [cm]; Start Y [cm]",
						630,-6300,0.0,160,-800.,800.);
    fStopPointTrackXZCRAll3D  = tfs->make<TH2D>("fStopPointTrackXZCRAll3D",
						"Stopping Point of Reco Track in XZ view (control room view); Start Z [cm]; Start X [cm]",
						630,-6300,0.0,160,-800.,800.);
    fStopPointTrackYZCRAll3D  = tfs->make<TH2D>("fStopPointTrackYZCRAll3D",
						"Stopping Point of Reco Track in YZ view (control room view); Start Z [cm]; Start Y [cm]",
						630,-6300,0.0,160,-800.,800.);

    fPETrackHitsAll3D        = tfs->make<TH1D>("fPETrackHitsAll3D",
					       "PE for Hits on Tracks;PE;count",
					       1862,0,4096);
    fTtrkAll3D               = tfs->make<TH1D>("fTtrkAll3D",
					       "ave hit time per track; time [ns]; count",
					       1200,-50000.0,550000.0);


    fTrackLengthAll2D     = tfs->make<TH1D>("fTrackLengthAll2D",
					    "Reconstructed Track Length distribution;Track Length[cm]; count",
					    650,0.0,6500.0);
    fTrackCosZenithAll2D  = tfs->make<TH1D>("fTrackCosZenithAll2D",
					    "Cos #theta with respect to zenith; Cos #theta_{Ze};count",
					    202,-1.01,1.01);
    fTrackCosAzimuthAll2D = tfs->make<TH1D>("fTrackCosAzimuthAll2D",
					    "Cos #theta with respect to azimuth (in detector coordinates); Cos #theta_{Az};count",
					    202,-1.01,1.01);
    fTrackCosZenAziAll2D  = tfs->make<TH2D>("fTrackCosZenAziAll2D",
					    "Cos #theta_{Zenith} vs. Cos #theta_{Azimuth} (in detector coordinates);Cos #theta_{Ze};Cos #theta_{Az}",
					    202,-1.01,1.01,202,-1.01,1.01);
    fTrackCosNumiAll2D    = tfs->make<TH1D>("fTrackCosNumiAll2D",
					    "Cos #theta with respect to NuMI beam; Cos #theta_{NuMI};count",
					    202,-1.01,1.01);
    fTrackZenithAll2D     = tfs->make<TH1D>("fTrackZenithAll2D",
					    "Zenith Angle;#theta_{Ze};count",
					    91,-0.5,180.5);
    fTrackAzimuthAll2D    = tfs->make<TH1D>("fTrackAzimuthAll2D",
					    "Azimuth Angle (in detector coordinates);#theta_{Az};count",
					    181,-0.5,360.5);
    fTrackZenAziAll2D     = tfs->make<TH2D>("fTrackZenAziAll2D",
					    "#theta_{Zenith} vs. #theta_{Azimuth} (in detector coordinates);#theta_{Ze};#theta_{Az}",
					    91,-0.5,180.5,181,-0.5,360.5);
    fTrackNumiAll2D       = tfs->make<TH1D>("fTrackNumiAll2D",
					    "NuMI Angle;#theta_{NuMI};count",
					    91,-0.5,180.5);

    fNumTracksAll2D       = tfs->make<TH1D>("fNumTracksAll2D",
					    ";No. Tracks;Events",
					    301,-0.5,300.5);
    fTrackFractionAll2D   = tfs->make<TH1D>("fTrackFractionAll2D",
					    ";Track Fraction;Events",
					    101,0.0,1.01);

    fStartPointTrackXZAll2D  = tfs->make<TH2D>("fStartPointTrackXZAll2D",
					       "Starting Point of reco Track in XZ view; Start Z [cm]; Start X [cm]",
					       630,0.,6300,160,-800.,800.);
    fStartPointTrackYZAll2D  = tfs->make<TH2D>("fStartPointTrackYZAll2D",
					       "Starting Point of reco Track in YZ view; Start Z [cm]; Start Y [cm]",
					       630,0.,6300,160,-800.,800.);
    fStopPointTrackXZAll2D   = tfs->make<TH2D>("fStopPointTrackXZAll2D",
					       "Stopping Point of reco Track in XZ view; Start Z [cm]; Start X [cm]",
					       630,0.,6300,160,-800.,800.);
    fStopPointTrackYZAll2D   = tfs->make<TH2D>("fStopPointTrackYZAll2D",
					       "Stopping Point of reco Track in YZ view; Start Z [cm]; Start Y [cm]",
					       630,0.,6300,160,-800.,800.);

    fStartPointTrackXZCRAll2D = tfs->make<TH2D>("fStartPointTrackXZCRAll2D",
						"Starting Point of Reco Track in XZ view (control room view); Start Z [cm]; Start X [cm]",
						630,-6300,0.0,160,-800.,800.);
    fStartPointTrackYZCRAll2D = tfs->make<TH2D>("fStartPointTrackYZCRAll2D",
						"Starting Point of Reco Track in YZ view (control room view); Start Z [cm]; Start Y [cm]",
						630,-6300,0.0,160,-800.,800.);
    fStopPointTrackXZCRAll2D  = tfs->make<TH2D>("fStopPointTrackXZCRAll2D",
						"Stopping Point of Reco Track in XZ view (control room view); Start Z [cm]; Start X [cm]",
						630,-6300,0.0,160,-800.,800.);
    fStopPointTrackYZCRAll2D  = tfs->make<TH2D>("fStopPointTrackYZCRAll2D",
						"Stopping Point of Reco Track in YZ view (control room view); Start Z [cm]; Start Y [cm]",
						630,-6300,0.0,160,-800.,800.);

    fPETrackHitsAll2D        = tfs->make<TH1D>("fPETrackHitsAll2D",
					       "PE for Hits on Tracks;PE;count",
					       1862,0,4096);
    fTtrkAll2D               = tfs->make<TH1D>("fTtrkAll2D",
					       "ave hit time per track; time [ns]; count",
					       1200,-50000.0,550000.0);

    fTrackLengthCont3D     = tfs->make<TH1D>("fTrackLengthCont3D",
					     "Reconstructed Track Length distribution;Track Length[cm]; count",
					     650,0.0,6500.0);
    fTrackCosZenithCont3D  = tfs->make<TH1D>("fTrackCosZenithCont3D",
					     "Cos #theta with respect to zenith; Cos #theta_{Ze};count",
					     202,-1.01,1.01);
    fTrackCosAzimuthCont3D = tfs->make<TH1D>("fTrackCosAzimuthCont3D",
					     "Cos #theta with respect to azimuth (in detector coordinates); Cos #theta_{Az};count",
					     202,-1.01,1.01);
    fTrackCosZenAziCont3D  = tfs->make<TH2D>("fTrackCosZenAziCont3D",
					     "Cos #theta_{Zenith} vs. Cos #theta_{Azimuth} (in detector coordinates);Cos #theta_{Ze};Cos #theta_{Az}",
					     202,-1.01,1.01,202,-1.01,1.01);
    fTrackCosNumiCont3D    = tfs->make<TH1D>("fTrackCosNumiCont3D",
					     "Cos #theta with respect to NuMI beam; Cos #theta_{NuMI};count",
					     202,-1.01,1.01);
    fTrackZenithCont3D     = tfs->make<TH1D>("fTrackZenithCont3D",
					     "Zenith Angle;#theta_{Ze};count",
					     91,-0.5,180.5);
    fTrackAzimuthCont3D    = tfs->make<TH1D>("fTrackAzimuthCont3D",
					     "Azimuth Angle (in detector coordinates);#theta_{Az};count",
					     181,-0.5,360.5);
    fTrackZenAziCont3D     = tfs->make<TH2D>("fTrackZenAziCont3D",
					     "#theta_{Zenith} vs. #theta_{Azimuth} (in detector coordinates);#theta_{Ze};#theta_{Az}",
					     91,-0.5,180.5,181,-0.5,360.5);
    fTrackNumiCont3D       = tfs->make<TH1D>("fTrackNumiCont3D",
					     "NuMI Angle;#theta_{NuMI};count",
					     91,-0.5,180.5);



    fNumTracksCont3D       = tfs->make<TH1D>("fNumTracksCont3D",
					     ";No. Tracks;Events",
					     301,-0.5,300.5);
    fTrackFractionCont3D   = tfs->make<TH1D>("fTrackFractionCont3D",
					     ";Track Fraction;Events",
					     101,0.0,1.01);

    fStartPointTrackXZCont3D  = tfs->make<TH2D>("fStartPointTrackXZCont3D",
						"Starting Point of Reco Track in XZ view; Start Z [cm]; Start X [cm]",
						630,0.,6300,160,-800.,800.);
    fStartPointTrackYZCont3D  = tfs->make<TH2D>("fStartPointTrackYZCont3D",
						"Starting Point of Reco Track in YZ view; Start Z [cm]; Start Y [cm]",
						630,0.,6300,160,-800.,800.);
    fStopPointTrackXZCont3D   = tfs->make<TH2D>("fStopPointTrackXZCont3D",
						"Stopping Point of Reco Track in XZ view; Start Z [cm]; Start X [cm]",
						630,0.,6300,160,-800.,800.);
    fStopPointTrackYZCont3D   = tfs->make<TH2D>("fStopPointTrackYZCont3D",
						"Stopping Point of Reco Track in YZ view; Start Z [cm]; Start Y [cm]",
						630,0.,6300,160,-800.,800.);

    fStartPointTrackXZCRCont3D = tfs->make<TH2D>("fStartPointTrackXZCRCont3D",
						 "Starting Point of Reco Track in XZ view (control room view); Start Z [cm]; Start X [cm]",
						 630,-6300,0.0,160,-800.,800.);
    fStartPointTrackYZCRCont3D = tfs->make<TH2D>("fStartPointTrackYZCRCont3D",
						 "Starting Point of Reco Track in YZ view (control room view); Start Z [cm]; Start Y [cm]",
						 630,-6300,0.0,160,-800.,800.);
    fStopPointTrackXZCRCont3D  = tfs->make<TH2D>("fStopPointTrackXZCRCont3D",
						 "Stopping Point of Reco Track in XZ view (control room view); Start Z [cm]; Start X [cm]",
						 630,-6300,0.0,160,-800.,800.);
    fStopPointTrackYZCRCont3D  = tfs->make<TH2D>("fStopPointTrackYZCRCont3D",
						 "Stopping Point of Reco Track in YZ view (control room view); Start Z [cm]; Start Y [cm]",
						 630,-6300,0.0,160,-800.,800.);

    fPETrackHitsCont3D        = tfs->make<TH1D>("fPETrackHitsCont3D",
						"PE for Hits on Tracks;PE;count",
						1862,0,4096);
    fTtrkCont3D               = tfs->make<TH1D>("fTtrkCont3D",
						"ave hit time per track; time [ns]; count",
						1200,-50000.0,550000.0);

    // DCM sync status stuff
    int minDiBlock=1;
    int maxDiBlock=14;
    int nDiBlock = maxDiBlock-minDiBlock+1;

    //ENTIRE DETECTOR
    f2DHitNum = tfs->make<TH2D>("H2DHitOnTrack","Number of Hits from 2D Tracks;Diblock;DCM",
				nDiBlock,(float)minDiBlock-0.5,
				(float)maxDiBlock+0.5,12,0.5,12.5);
    f3DHitNum = tfs->make<TH2D>("H3DHitOnTrack","Number of Hits from 3D Tracks;Diblock;DCM",
				nDiBlock,(float)minDiBlock-0.5,
				(float)maxDiBlock+0.5,12,0.5,12.5);

    f2DTrackNum = tfs->make<TH2D>("H2DTrackNum","Number of 2D Tracks;Diblock;DCM",
				  nDiBlock,(float)minDiBlock-0.5,
				  (float)maxDiBlock+0.5,12,0.5,12.5);
    f3DTrackNum = tfs->make<TH2D>("H3DTrackNum","Number of 3D Tracks;Diblock;DCM",
				  nDiBlock,(float)minDiBlock-0.5,
				  (float)maxDiBlock+0.5,12,0.5,12.5);

    Instr = tfs->make<TH2F>("instrumented","Map of Instrumented Detector (DCMs with more than fMinFEBs APDs);Diblock;DCM",
                            nDiBlock,(float)minDiBlock-0.5,
                            (float)maxDiBlock+0.5,12,0.5,12.5);

    fullyinstr = tfs->make<TH2F>("fully_instrumented","Map of Fully Instrumented Detector;Diblock;DCM",
                                 nDiBlock,(float)minDiBlock-0.5,
                                 (float)maxDiBlock+0.5,12,0.5,12.5);

  }

  //......................................................................
  void NearlineAna::endJob()
  {
    //
    // Compute header info.
    //

    //
    // DISECTING the time from evt.time().value() into "human readable" format to display the date & time
    // taken from EventDisplay/HeaderDrawer.cxx
    //
    unsigned int hour, minute, second;
    int          nano;

    // get the time stamp.  art::Timestamp::value() returns a TimeValue_t which is a typedef to unsigned long long.
    // The conventional use is for the upper 32 bits to have the seconds since 1970 epoch and the lower 32 bits to be
    // the number of nanoseconds with the current second.

    // taking start time apart
    const unsigned long int mask32 = 0xFFFFFFFFUL;

    unsigned long int lup = ( fStartTime >> 32 ) & mask32;
    unsigned long int llo = fStartTime & mask32;
    TTimeStamp ts1(lup, (int)llo);
    ts1.GetDate(kTRUE,0,&fStartYear,&fStartMonth,&fStartDay);
    ts1.GetTime(kTRUE,0,&hour,&minute,&second);
    nano = ts1.GetNanoSec();
    double sec = ((double)second + (double)nano/1.0e9);
    fStartHour = (double)hour + (double)minute/60.0 + sec/3600.0;

    // taking end time apart
    lup = ( fEndTime >> 32 ) & mask32;
    llo = fEndTime & mask32;
    TTimeStamp ts2(lup, (int)llo);
    ts2.GetDate(kTRUE,0,&fEndYear,&fEndMonth,&fEndDay);
    ts2.GetTime(kTRUE,0,&hour,&minute,&second);
    nano = ts2.GetNanoSec();
    sec = ((double)second + (double)nano/1.0e9);
    fEndHour = (double)hour + (double)minute/60.0 + sec/3600.0;

    fHeader->Fill();
  }

  //......................................................................
  void NearlineAna::beginRun(const art::Run& run)
  {

    art::ServiceHandle<geo::Geometry> geom;
    fhicl::ParameterSet DETpset;

    switch(geom->DetId()){
    case novadaq::cnv::kNDOS:
      DETpset = fPSetNDOS;
      break;
    case novadaq::cnv::kNEARDET:
      DETpset = fPSetND;
      break;
    case novadaq::cnv::kFARDET:
      DETpset = fPSetFD;
      break;
    default:
      std::cout << "\nUnknown detector type!\n";
      assert(0);
    }

    fTrackLabel = DETpset.get<std::string>("TrackLabel");
    fMinX       = DETpset.get< double >   ("MinX");
    fMaxX       = DETpset.get< double >   ("MaxX");
    fMinY       = DETpset.get< double >   ("MinY");
    fMaxY       = DETpset.get< double >   ("MaxY");
    fMinZ       = DETpset.get< double >   ("MinZ");
    fMaxZ       = DETpset.get< double >   ("MaxZ");
    fP1MaxZ     = DETpset.get< double >   ("P1MaxZ");
    fMinT       = DETpset.get< double >   ("MinT");
    fMaxT       = DETpset.get< double >   ("MaxT");
    fMinFEBs    = DETpset.get< int    >   ("MinFEBs");    // NEW
      
    fMinPOTCut      = DETpset.get< double >("MinPOTCut"     );
    fMinHornICut    = DETpset.get< double >("MinHornICut"   );
    fMaxHornICut    = DETpset.get< double >("MaxHornICut"   );
    fMinPosXCut     = DETpset.get< double >("MinPosXCut"    );
    fMaxPosXCut     = DETpset.get< double >("MaxPosXCut"    );
    fMinPosYCut     = DETpset.get< double >("MinPosYCut"    );
    fMaxPosYCut     = DETpset.get< double >("MaxPosYCut"    );
    fMinWidthXCut   = DETpset.get< double >("MinWidthXCut"  );
    fMaxWidthXCut   = DETpset.get< double >("MaxWidthXCut"  );
    fMinWidthYCut   = DETpset.get< double >("MinWidthYCut"  );
    fMaxWidthYCut   = DETpset.get< double >("MaxWidthYCut"  );
    fMaxDeltaTCut   = DETpset.get< double >("MaxDeltaTCut"  );


    ////////////////////////////////////////////////////////////////////////////////////////////////////
    //INSTRUMENTATION HISTOGRAMS

    art::ServiceHandle<nova::dbi::RunHistoryService> fRH;    
    bool  RunHistoryConnectedOK = fRH->LoadDAQRunHistory();     //TESTING FAILURE TO CONNECT

    if(DETpset == fPSetFD){ 
      
      if(RunHistoryConnectedOK){   //IF IT CAN REACH THE CONNECTIONS, THEN FILL THE HISTOGRAMS	
	int ndb = fRH->NDiBlocks();
	for (int idb=1; idb<=ndb; ++idb) { 
	  
	  //std::cout<<"NDiBlocks idb: "<< idb <<std::endl;      // number of diblocks
	  
	  nova::dbi::RunHistory::DiBlock db = fRH->GetDiBlock(idb-1,false);
	  
	  for (int idcm=1; idcm<=12; ++idcm) {
	    
	    // number of FEBs (and APDs by extension!) in DCM	  
	    //std::cout<<"No. of instrumentedFEBs ("<<idcm<<") : "<<db.dcm[idcm-1].nInstrumentedFEBs<<std::endl;  
	    	    	    
	    if ((db.dcm[idcm-1].nInstrumentedFEBs >= fMinFEBs) &&   // LOOK AT THE # OF APDs IN A DCM PER DIBLOCK
		(db.dcm[idcm-1].nInstrumentedFEBs > -1) &&      
		(db.dcm[idcm-1].nInstrumentedFEBs <= 64) ){      
	      Instr->Fill(db.num,idcm);                // change idb to db.num
	      //std::cout<<"FILLING No. of instrumentedFEBs ("<<idcm<<") : "<<db.dcm[idcm-1].nInstrumentedFEBs<<std::endl;  
	    }
	    
	    if(fRH->IsDiBlockFullyInstrumented(idb)){
	      fullyinstr->Fill(idb,idcm);
	    }
	    
	  }
	}
      }
      else{                                   // IF FAILS REACHING CONNECTION, FILL EMPTY HISTOGRAMS AND LET US KNOW
	
	LOG_WARNING("NearlineAna") << " LOG_WAR: Error: Failed to LoadDAQRunHistory"
				   << " LOG_WAR: Histograms from Instrumented Detector and DCM Sync plots will be empty! " << std::endl;
	
	std::cerr << " STD CERR: Error: Failed to LoadDAQRunHistory" << std::endl;
	
	std::cerr << " STD CERR: Histograms from Instrumented Detector and DCM Sync plots will be empty! " << std::endl;
	
      }
      
    }
    else if(DETpset == fPSetND){ 

      if(RunHistoryConnectedOK){  	
	int ndb = fRH->NDiBlocks();
	for (int idb=1; idb<=ndb; ++idb) { 
	  
	  std::cout<<"NDiBlocks idb: "<< idb <<std::endl;      // number of diblocks
	  
	  nova::dbi::RunHistory::DiBlock db = fRH->GetDiBlock(idb-1,false);	  
	  for (int idcm=1; idcm<=4; ++idcm) {
	    
	    // number of FEBs (and APDs by extension!) in DCM	  
	    std::cout<<"No. of instrumentedFEBs ("<<idcm<<") : "<<db.dcm[idcm-1].nInstrumentedFEBs<<std::endl;  
	    	    	    
	    if ((db.dcm[idcm-1].nInstrumentedFEBs >= fMinFEBs) &&      //fMinFEBs  // LOOK AT THE # OF APDs IN A DCM PER DIBLOCK
		(db.dcm[idcm-1].nInstrumentedFEBs > -1) &&      
		(db.dcm[idcm-1].nInstrumentedFEBs <= 64) ){      
	      Instr->Fill(db.num,idcm);                // change idb to db.num
	      std::cout<<"FILLING No. of instrumentedFEBs ("<<idcm<<") : "<<db.dcm[idcm-1].nInstrumentedFEBs<<std::endl;  
	    }
	    
	    // DOES IT MAKE SENSE FOR ND??
	    if(fRH->IsDiBlockFullyInstrumented(idb)){
	      fullyinstr->Fill(idb,idcm);
	    }
	    
	  }
	}
      }
      else{                                   
	
	LOG_WARNING("NearlineAna") << " LOG_WAR: Error: Failed to LoadDAQRunHistory"
				   << " LOG_WAR: Histograms from Instrumented Detector and DCM Sync plots will be empty! " << std::endl;
	
	std::cerr << " STD CERR: Error: Failed to LoadDAQRunHistory" << std::endl;
	
	std::cerr << " STD CERR: Histograms from Instrumented Detector and DCM Sync plots will be empty! " << std::endl;
	
      }
      
    }
    else{
      LOG_WARNING("NearlineAna") << " LOG_WAR: Error: unknown detector"
				 << " LOG_WAR: Histograms from Instrumented Detector and DCM Sync plots will be empty! " << std::endl;	
      
      std::cerr << " STD CERR: Error: unknown detector" << std::endl;
      
      std::cerr << " STD CERR: Histograms from Instrumented Detector and DCM Sync plots will be empty! " << std::endl;
            
    }      
    ////////////////////////////////////////////////////////////////////////////////////////////////////
    
  }    
  //......................................................................
  void NearlineAna::analyze(const art::Event& evt)
  {

    // needed for DCM sync status plots
    art::ServiceHandle<cmap::CMap> cmap;

    art::Handle< sumdata::SpillData > spill;
      
      
    //NEW CODE
    double timediffnsec = -9999999.;
    Double_t spillPoT = 0.;
    double horncurrent = 0.;
    double xposition = -9999.;
    double yposition = -9999.;
    double xwidth = -9999.;
    double ywidth = -9999.;
    bool goodbeam = false;
    bool goodevent = false; //don't anything in the histogram if there is no good beam

      
    try {
      evt.getByLabel("ifdbspillinfo", spill);
      
      //NEW CODE
      timediffnsec = spill->deltaspilltimensec;
      spillPoT = spill->spillpot;
      horncurrent = spill->hornI;
      xposition = spill->posx;
      yposition = spill->posy;
      xwidth = spill->widthx;
      ywidth = spill->widthy;
      goodbeam = spill->goodbeam;
      goodevent = true;
      std::cout << " spill " << spillPoT << " " << goodbeam << std::endl;
      if(!goodbeam ) std::cout << " !!!!!!!!!!!!!! bad beam !!!!!!!!! " << std::endl;
    }
    catch(...) { 
      std::cerr << "No Spill Info!" << std::endl;
    }
    
    
    //
    // Determine header info.
    //
    
    unsigned int           run    = evt.run();
    unsigned int           subrun = evt.subRun();
    unsigned int           event  = evt.id().event();
    unsigned long long int time   = evt.time().value();

    fNevents++;
    fLastRun    = run;
    fLastSubrun = subrun;

    // Get the partition and the trigger length from the raw data.
    static daqdataformats::RawEvent rawevt;
    try {
      art::Handle<std::vector<rawdata::FlatDAQData> > flatdaq_handle;
      evt.getByLabel("daq",flatdaq_handle);
      if (flatdaq_handle->size()!=1) {
	std::cerr << "Error:   More than one flat daq data object!" 
		  << std::endl;
      }
      else {
	rawevt.readData((*flatdaq_handle)[0].getRawBufferPointer());
	fPartition = rawevt.getPartition();

	// Compute the live time for this event.  Note: The tigger length is in units of 500 ns
	// and is always one microblock short of the the actual recorded time.
	double range = (double)rawevt.getTrigger()->getTriggerRange()->getTriggerLength();
	range = 100.0*floor(range/100.0)+100.0;
	range = range*5.0E-7;

	fLiveTime += range;
      }
    }
    catch (...) {
      std::cerr << "Error:   Flat DAQ data object missing!" << std::endl;
    }
    
    // first/last event are not necessarily coorelated with earliest/latest time
    if(time < fStartTime)
      fStartTime = time;
    if((int)event < fFirstEvent)
      fFirstEvent = event;
    if(time > fEndTime)
      fEndTime = time;
    if((int)event > fLastEvent)
      fLastEvent = event;



    // counters...
    int NSlice     = 0;
    int N2DTrack   = 0;
    int N3DTrack   = 0;
    int N3DContTrk = 0;
    int NTrackTot  = 0;

    //
    // Compute information about slices.
    //

    // get the slices
    art::Handle< std::vector< rb::Cluster > > slices;
    evt.getByLabel(fSlicerLabel, slices);

    // get associations between slices and tracks
    art::FindManyP<rb::Track> fmSlTrk(slices, evt, fTrackLabel);



    // Loop over all slices.
    for(unsigned int i = 0; i < slices->size(); ++i) {
     
      // define default values for the slice info
      int          NumHitsX = -1;
      int          NumHitsY = -1;

      double Pave  = 0.0;
      double CXave = 0.0;
      double CYave = 0.0;
      double Tave  = 0.0;

      double Psd   = 0.0;
      double CXsd  = 0.0;
      double CYsd  = 0.0;
      double Tsd   = 0.0;

      double Phi  = -1.0e6;
      double Plo  = 1.0e6;
      double Thi  = -1.0e6;
      double Tlo  = 1.0e6;
      double CXhi = -1.0e6;
      double CXlo = 1.0e6;
      double CYhi = -1.0e6;
      double CYlo = 1.0e6;
      
      NumHitsX = (*slices)[i].NXCell();
      NumHitsY = (*slices)[i].NYCell();

      // count the non-noise slices (to verify that there is only 1 noise slice)
      if(!(*slices)[i].IsNoise()) NSlice++;



      // If the slice is noise, only keep track of the number of hits/PE and skip everything else.
      if((*slices)[i].IsNoise()) {
	fNumHitsNoise ->Fill(NumHitsX+NumHitsY);
	fNumHitsXNoise->Fill(NumHitsX);
	fNumHitsYNoise->Fill(NumHitsY);
	// loop over all hits and add up PE
	for(unsigned int j = 0; j < (*slices)[i].NCell(); ++j) { 
	  const rb::CellHit* h = (*slices)[i].Cell(j).get();
	  fADCAllHits->Fill(h->ADC());
	  fADCNoiseHits->Fill(h->ADC());
	  fPEAllHits->Fill(h->PE());
	  fPENoiseHits->Fill(h->PE());
	}
	continue;
      }



      // Calculate average and st. dev. for plane #, cell #, and time for
      // hits in the slice
      // art::PtrVector< rb::CellHit > sliceHits = (*slices)[i].AllCells();
      
      double nx = 0.0, ny = 0.0;

      // loop over all hits
      for(unsigned int j = 0; j < (*slices)[i].NCell(); ++j) {

	const rb::CellHit* h = (*slices)[i].Cell(j).get();

	// If we've reached this point, this is a non-noise slice.
	fADCAllHits->Fill(h->ADC());
	fADCNonNoiseHits->Fill(h->ADC());
	fPEAllHits->Fill(h->PE());
	fPENonNoiseHits->Fill(h->PE());

	// Problems occur if you do the time calculations in ns since the numbers
	// are huge when you square them (even if you use doubles.)  So I will
	// convert to microseconds to do my sums for Tave and Tsd.
	double Tus = (h->TNS())/1000.0;

	Pave += h->Plane();
	Tave += Tus;
	Psd  += (h->Plane())*(h->Plane());
	Tsd  += Tus*Tus;

	if(h->Plane() > Phi) Phi = h->Plane();
	if(h->Plane() < Plo) Plo = h->Plane();
	if(h->TNS() > Thi) Thi = h->TNS();
	if(h->TNS() < Tlo) Tlo = h->TNS();

	if(h->View() == geo::kX) {
	  CXave += h->Cell();
	  CXsd  += (h->Cell())*(h->Cell());
	  nx += 1.0;
	  if(h->Cell() > CXhi) CXhi = h->Cell();
	  if(h->Cell() < CXlo) CXlo = h->Cell();
	}
	else if(h->View() == geo::kY) {
	  CYave += h->Cell();
	  CYsd  += (h->Cell())*(h->Cell());
	  ny += 1.0;
	  if(h->Cell() > CYhi) CYhi = h->Cell();
	  if(h->Cell() < CYlo) CYlo = h->Cell();
	} 
	else {
	  // Trigger some kind of warning that an invalid view occured.
	  CXave = 1.0e9;
	  CYave = 1.0e9;
	}
      } // end loop over j

      if(nx == 0.0 && ny == 0.0) {
	// assign a negative number to attract attention
	Pave = -100.0;
	Tave = -100.0;
	Psd  = -100.0;
	Tsd  = -100.0;
      }
      else {
	Pave = Pave/(nx+ny);
	Tave = Tave/(nx+ny);
	Psd  = sqrt(Psd/(nx+ny) - Pave*Pave);
	Tsd  = sqrt(Tsd/(nx+ny) - Tave*Tave);

	// convert back to ns
	Tave = Tave*1000.0;
	Tsd  = Tsd*1000.0;
      }

      if(nx != 0.0) {
	CXave = CXave/nx;
	CXsd  = sqrt(CXsd/nx - CXave*CXave);
      }
      else {
	CXave = -100.0;
	CXsd  = -100.0;
      }

      if(ny != 0.0) {
	CYave = CYave/ny;
	CYsd  = sqrt(CYsd/ny - CYave*CYave);
      }
      else {
	CYave = -100.0;    
	CYsd  = -100.0;
      }


      
      // fill slice histos
      fNumHits ->Fill(NumHitsX+NumHitsY);
      fNumHitsX->Fill(NumHitsX);
      fNumHitsY->Fill(NumHitsY);

      fPCXave  ->Fill(Pave,CXave);
      fPCYave  ->Fill(Pave,CYave);
      fPCXaveCR->Fill(-Pave,-CXave);
      fPCYaveCR->Fill(-Pave, CYave);

      fPCXdiff ->Fill(Phi-Plo,CXhi-CXlo);
      fPCYdiff ->Fill(Phi-Plo,CYhi-CYlo);
      fTave    ->Fill(Tave);       
      fTsd     ->Fill(Tsd);



      //
      // Get info about the tracks that go with this slice.
      //
      
      std::vector < art::Ptr< rb::Track > > tracks;
      if(fmSlTrk.isValid()) {
	tracks = fmSlTrk.at(i);
      }

      // set some counters
      int NHit3Dtracks = 0;

      // initializing counters for DCM sync status plots
      for (int idb=0; idb<14; ++idb) {
	for (int idcm=0; idcm<12; ++idcm) {
          counter3D[idb][idcm] = 0;
	  counter2D[idb][idcm] = 0;
	}
      }

      for(unsigned int k = 0; k < tracks.size(); k++) {

	// Fill histos for 3D tracks
	if(tracks[k]->Is3D()) {
	
	  NTrackTot    += tracks.size();
	  N3DTrack++;
	  NHit3Dtracks += tracks[k]->NCell();

	  // Get vectors for the end points of the track
	  TVector3 start(tracks[k]->Start());
	  TVector3 end(tracks[k]->Stop());
	  TVector3 dir(tracks[k]->Dir());
	  TVector3 length = end - start;

	  // loop over all hits and count PE
	  for(unsigned int j = 0; j < tracks[k]->NCell(); ++j) {
	    const rb::CellHit* h = tracks[k]->Cell(j).get();
	    fPETrackHitsAll3D->Fill(h->PE());

	    // count hits by diblock/DCM
	    const unsigned int chan = h->DaqChannel();
	    const unsigned int dib  = cmap->Map()->getDiBlock(chan);
	    const unsigned int ldc  = cmap->Map()->getDCM(chan);
	    
	    f3DHitNum->Fill(dib,ldc);                  // Filling the Hit rate histogram
	    ++counter3D[dib-1][ldc-1];
	  }

	  // count tracks by diblock/DCM
	  for(int i = 0; i < 14; i++){
	    for(int j = 0; j < 12; j++){
	      // count a track as contributing to a DCM if it has > fMinTrackHits 
	      if(counter3D[i][j] > fMinTrackHits){
		f3DTrackNum->Fill(i+1,j+1);
	      }
	    }
	  }

	  fTtrkAll3D->Fill(tracks[k]->MeanTNS());	

	  fStartPointTrackXZAll3D->Fill(start.Z(), start.X());
	  fStartPointTrackYZAll3D->Fill(start.Z(), start.Y());
	  fStopPointTrackXZAll3D->Fill(end.Z(), end.X());
	  fStopPointTrackYZAll3D->Fill(end.Z(), end.Y());
	  
	  fStartPointTrackXZCRAll3D->Fill(-start.Z(), -start.X());
	  fStartPointTrackYZCRAll3D->Fill(-start.Z(), start.Y());
	  fStopPointTrackXZCRAll3D->Fill(-end.Z(), -end.X());
	  fStopPointTrackYZCRAll3D->Fill(-end.Z(), end.Y());
	  
	  fTrackLengthAll3D->Fill(length.Mag());

	  // track angles
	  fTrackCosZenithAll3D->Fill(-dir.Y());
	  fTrackCosAzimuthAll3D->Fill(-dir.Z()/util::pythag(dir.X(),dir.Z()));
	  fTrackCosZenAziAll3D->Fill(-dir.Y(),-dir.Z()/util::pythag(dir.X(),dir.Z()));
	  art::ServiceHandle<geo::Geometry> geom;
	  fTrackCosNumiAll3D->Fill(tracks[k]->Dir().Dot(geom->NuMIBeamDirection()));
	  
	  double angle = 0.0;
	  fTrackZenithAll3D->Fill(acos(-dir.Y())*180.0/3.1415926536);
	  angle = (atan2(dir.X(),-dir.Z()))*180.0/3.1415926536;
	  if(angle < 0.0) angle += 360.0;
	  fTrackAzimuthAll3D->Fill(angle);
	  fTrackZenAziAll3D->Fill(acos(-dir.Y())*180.0/3.1415926536,angle);
	  fTrackNumiAll3D->Fill(acos(tracks[k]->Dir().Dot(geom->NuMIBeamDirection()))*180.0/3.1415926536);

	  // If track is contained, also fill the histos for contained 3D tracks
	  if(this->IsContained(start.X(), start.Y(), start.Z()) &&
	     this->IsContained(end.X(),   end.Y(),   end.Z()) ) {

	    N3DContTrk++;

	    // loop over all hits and count PE
	    for(unsigned int j = 0; j < tracks[k]->NCell(); ++j) {
	      const rb::CellHit* h = tracks[k]->Cell(j).get();
	      fPETrackHitsCont3D->Fill(h->PE());
	    }

	    fTtrkCont3D->Fill(tracks[k]->MeanTNS());	

	    fStartPointTrackXZCont3D->Fill(start.Z(), start.X());
	    fStartPointTrackYZCont3D->Fill(start.Z(), start.Y());
	    fStopPointTrackXZCont3D->Fill(end.Z(), end.X());
	    fStopPointTrackYZCont3D->Fill(end.Z(), end.Y());
	    
	    fStartPointTrackXZCRCont3D->Fill(-start.Z(), -start.X());
	    fStartPointTrackYZCRCont3D->Fill(-start.Z(), start.Y());
	    fStopPointTrackXZCRCont3D->Fill(-end.Z(), -end.X());
	    fStopPointTrackYZCRCont3D->Fill(-end.Z(), end.Y());
	    
	    fTrackLengthCont3D->Fill(length.Mag());
	    
	    // track angles
	    fTrackCosZenithCont3D->Fill(-dir.Y());
	    fTrackCosAzimuthCont3D->Fill(-dir.Z()/util::pythag(dir.X(),dir.Z()));
	    fTrackCosZenAziCont3D->Fill(-dir.Y(),-dir.Z()/util::pythag(dir.X(),dir.Z()));
	    fTrackCosNumiCont3D->Fill(tracks[k]->Dir().Dot(geom->NuMIBeamDirection()));
	    
	    fTrackZenithCont3D->Fill(acos(-dir.Y())*180.0/3.1415926536);
	    fTrackAzimuthCont3D->Fill(angle);
	    fTrackZenAziCont3D->Fill(acos(-dir.Y())*180.0/3.1415926536,angle);
	    fTrackNumiCont3D->Fill(acos(tracks[k]->Dir().Dot(geom->NuMIBeamDirection()))*180.0/3.1415926536);
	    
	  } // end IsContained

	} // end is 3D track

	// otherwise this must be a 2D track
	else {
	  
	  NTrackTot += tracks.size();
	  N2DTrack++;
	  
	  // Get vectors for the end points of the track
	  TVector3 start(tracks[k]->Start());
	  TVector3 end(tracks[k]->Stop());
	  TVector3 dir(tracks[k]->Dir());
	  TVector3 length = end - start;
	  
	  // loop over all hits and count PE
	  for(unsigned int j = 0; j < tracks[k]->NCell(); ++j) {
	    const rb::CellHit* h = tracks[k]->Cell(j).get();
	    fPETrackHitsAll2D->Fill(h->PE());

	    // count hits by diblock/DCM
	    const unsigned int chan = h->DaqChannel();
	    const unsigned int dib  = cmap->Map()->getDiBlock(chan);
	    const unsigned int ldc  = cmap->Map()->getDCM(chan);

	    f2DHitNum->Fill(dib,ldc);
	    ++counter2D[dib-1][ldc-1]; 
	  }
	  
	  // count tracks by diblock/DCM
	  for(int i = 0; i < 14; i++){
	    for(int j = 0; j < 12; j++){
	      // count a track as contributing to a DCM if it has > fMinTrackHits 
	      if(counter2D[i][j] > fMinTrackHits){
		f2DTrackNum->Fill(i+1,j+1);
	      }
	    }
	  }

	  fTtrkAll2D->Fill(tracks[k]->MeanTNS());	
	  
	  fStartPointTrackXZAll2D->Fill(start.Z(), start.X());
	  fStartPointTrackYZAll2D->Fill(start.Z(), start.Y());
	  fStopPointTrackXZAll2D->Fill(end.Z(), end.X());
	  fStopPointTrackYZAll2D->Fill(end.Z(), end.Y());
	  
	  fStartPointTrackXZCRAll2D->Fill(-start.Z(), -start.X());
	  fStartPointTrackYZCRAll2D->Fill(-start.Z(), start.Y());
	  fStopPointTrackXZCRAll2D->Fill(-end.Z(), -end.X());
	  fStopPointTrackYZCRAll2D->Fill(-end.Z(), end.Y());
	  
	  fTrackLengthAll2D->Fill(length.Mag());

	  // track angles
	  fTrackCosZenithAll2D->Fill(-dir.Y());
	  fTrackCosAzimuthAll2D->Fill(-dir.Z()/util::pythag(dir.X(),dir.Z()));
	  fTrackCosZenAziAll2D->Fill(-dir.Y(),-dir.Z()/util::pythag(dir.X(),dir.Z()));
	  art::ServiceHandle<geo::Geometry> geom;
	  fTrackCosNumiAll2D->Fill(tracks[k]->Dir().Dot(geom->NuMIBeamDirection()));
	  
	  double angle = 0.0;
	  fTrackZenithAll2D->Fill(acos(-dir.Y())*180.0/3.1415926536);
	  angle = (atan2(dir.X(),-dir.Z()))*180.0/3.1415926536;
	  if(angle < 0.0) angle += 360.0;
	  fTrackAzimuthAll2D->Fill(angle);
	  fTrackZenAziAll2D->Fill(acos(-dir.Y())*180.0/3.1415926536,angle);
	  fTrackNumiAll2D->Fill(acos(tracks[k]->Dir().Dot(geom->NuMIBeamDirection()))*180.0/3.1415926536);

	} // end is 2D track

      } // end loop over tracks (k)

      // fill histos with per slice info
      fSliceTrackNHitRatio->Fill((double)NHit3Dtracks/(double)(NumHitsX+NumHitsY));
      fSliceTrackNHitRatioVSnhit->Fill((double)(NumHitsX+NumHitsY),(double)NHit3Dtracks/(double)(NumHitsX+NumHitsY));

    } // end loop over slices (i)



    // fill histos with per trigger info
    fNumSlices          ->Fill(NSlice);
    if(goodbeam) fNumSlicesGoodBeam->Fill(NSlice);
      
    //NEW CODE
    if(goodevent==true){ //if there was no beam, don't fill the histograms 
      fTimeDiffNanoSec->Fill(1e-9*timediffnsec);
      fHornCurrent->Fill(horncurrent);
      fXPosition->Fill(xposition);
      fYPosition->Fill(yposition);
      fXWidth->Fill(xwidth);
      fYWidth->Fill(ywidth);
    }
    fPOTSum->Fill(0.0,spillPoT*1e12);
    fspillPOT->Fill(spillPoT*1e12); //we still want to know that there were no POTs
    fGoodBeam->Fill(goodbeam);
    
    // 0-1:delta t cut, 1-2:POT cut, 2-3:horn I cut, 3-4:x position cut, 4-5:y position cut, 5-6: x width cut, 6-7:y width cut
      if (!goodbeam){
          fBadSpills->Fill(8,1);//count the total # of bad spills
          
          // if delta t condition is not met, record the spill as bad and don't worry about other quality cuts
          if (fMaxDeltaTCut<timediffnsec) {
              fBadSpills->Fill(1,1);
          }
          
          // if the delta t condition IS met, check the POT.
          // if the POT condition is not met, record the spill as bad and don't worry about other cuts
          else {
              if (spillPoT<fMinPOTCut) {
                  fBadSpills->Fill(2,1);
              }
              // record the reason for the spill failure in all other cases.  note, there may be overlap.
              else {
                  if(horncurrent<fMinHornICut || fMaxHornICut<horncurrent) fBadSpills->Fill(3,1);
                  if(xposition<fMinPosXCut || fMaxPosXCut<xposition) fBadSpills->Fill(4,1);
                  if(yposition<fMinPosYCut || fMaxPosYCut<yposition) fBadSpills->Fill(5,1);
                  if(xwidth<fMinWidthXCut || fMaxWidthXCut<xwidth) fBadSpills->Fill(6,1);
                  if(ywidth<fMinWidthYCut || fMaxWidthYCut<ywidth) fBadSpills->Fill(7,1);

              }
          }
      }
      
    
      
    fNumTracksAll2D     ->Fill(N2DTrack);
    fNumTracksAll3D     ->Fill(N3DTrack);
    fNumTracksCont3D    ->Fill(N3DContTrk);
    fTrackFractionAll2D ->Fill((double)N2DTrack/(double)NTrackTot);
    fTrackFractionAll3D ->Fill((double)N3DTrack/(double)NTrackTot);
    fTrackFractionCont3D->Fill((double)N3DContTrk/(double)NTrackTot);
    fDeltaSliceTrack    ->Fill(NSlice - N3DTrack);
    fSliceTrackRatio    ->Fill((double)N3DTrack/(double)NSlice);


    return;

  } // end analyze module

  //......................................................................
  bool NearlineAna::IsContained(double x, double y, double z)
  {
    if(x > fMinX && x < fMaxX && y > fMinY && y < fMaxY) {
      // If this is partition 1, then use the specific Z cut for partition 1
      if(fPartition == 1 && z > fMinZ && z < fP1MaxZ) return true;
      if(fPartition != 1 && z > fMinZ && z < fMaxZ)   return true;
    }
    return false;
  }

} // end namespace comi
////////////////////////////////////////////////////////////////////////



#include "art/Framework/Core/ModuleMacros.h"

namespace comi
{
  DEFINE_ART_MODULE(NearlineAna);
}
