////////////////////////////////////////////////////////////////////////
/// \brief  Sum up livetimes from individual cosmic triggers
/// \author Christopher Backhouse - bckhouse@caltech.edu
////////////////////////////////////////////////////////////////////////

// NOvASoft includes
#include "DAQDataFormats/TriggerDefines.h"
#include "RawData/RawTrigger.h"
#include "SummaryData/CosmicExposure.h"

// Framework includes
#include "art/Framework/Core/EDProducer.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/SubRun.h"
#include "art/Framework/Principal/Handle.h"
#include "fhiclcpp/ParameterSet.h"

namespace cosmic
{
  /// \brief Sum up livetimes from individual cosmic triggers
  class CosmicExposureInfo : public art::EDProducer
  {
  public:
    explicit CosmicExposureInfo(const fhicl::ParameterSet& pset);
    virtual ~CosmicExposureInfo();

    void produce(art::Event& evt);

    void beginSubRun(art::SubRun& sr);
    void endSubRun(art::SubRun& sr);

  protected:
    double fExposure; //< Livetime so far this subrun in seconds
  };

  //......................................................................
  CosmicExposureInfo::CosmicExposureInfo(const fhicl::ParameterSet& /*pset*/)
    : fExposure(0)
  {
    produces<sumdata::CosmicExposure, art::InSubRun>();
  }

  //......................................................................
  CosmicExposureInfo::~CosmicExposureInfo()
  {
  }

  //......................................................................
  void CosmicExposureInfo::beginSubRun(art::SubRun& sr)
  {
    fExposure = 0;
  }

  //......................................................................
  void CosmicExposureInfo::endSubRun(art::SubRun& sr)
  {
    if(fExposure >= 0){
      std::unique_ptr<sumdata::CosmicExposure> ce(new sumdata::CosmicExposure);
      ce->totlivetime = fExposure;
      sr.put(std::move(ce));
    }
  }

  //......................................................................
  void CosmicExposureInfo::produce(art::Event& evt)
  {
    art::Handle<std::vector<rawdata::RawTrigger> > trigs;
    evt.getByLabel("daq", trigs);

    for(unsigned int n = 0; n < trigs->size(); ++n){
      const rawdata::RawTrigger& trig = (*trigs)[n];

      // No: beam spills are just as valid cosmic live time
//       // Should be a cosmic trigger
//       const uint8_t trigtype = trig.fTriggerMask_TriggerType;
//       assert(trigtype & daqdataformats::TRIGBIT_ID_CAL_PER ||
// 	     trigtype & daqdataformats::TRIGBIT_ID_CAL_RND);

      // Units of trigger field are 500ns
      fExposure += trig.fTriggerRange_TriggerLength*500e-9;
    }
  }

  DEFINE_ART_MODULE(CosmicExposureInfo);

} // end namespace

