////////////////////////////////////////////////////////////////////////
// \file    Chunk.cxx
// \version $Id: Chunk.cxx,v 1.15 2012-10-30 23:33:08 bckhouse Exp $
// \author  Christopher Backhouse - bckhouse@caltech.edu
////////////////////////////////////////////////////////////////////////

#include "DiscreteTracker/Chunk.h"

#include "Geometry/Geometry.h"
#include "GeometryObjects/CellGeo.h"
#include "RecoBase/CellHit.h"

#include "art/Framework/Services/Registry/ServiceHandle.h"

namespace dt
{
  // From gdml file. Not exposed by the Geometry service
  const double kCornerR = .94; // cm

  //......................................................................
  Chunk::Chunk(int plane, int cellInLo, int cellInHi, int cellOutLo, int cellOutHi, bool up,
               art::PtrVector<rb::CellHit> hits)
    : fPlane(plane),
      fEntryCellLo(cellInLo), fEntryCellHi(cellInHi),
      fExitCellLo(cellOutLo), fExitCellHi(cellOutHi),
      fUp(up), fShowerChunk(false),
      fSegsDirty(true),
      fHits(hits)
  {
    assert(cellInLo <= cellInHi);
    assert(cellOutLo <= cellOutHi);
  }

  //......................................................................
  bool Chunk::Matches(const Chunk& ch) const
  {
    if(fPlane != ch.fPlane) return false;

    if(BottomExtremeCell() != ch.BottomExtremeCell()) return false;
    if(TopExtremeCell() != ch.TopExtremeCell()) return false;

    if(BottomHitCell() != ch.BottomHitCell()) return false;
    if(TopHitCell() != ch.TopHitCell()) return false;

    return true;
  }

  //......................................................................
  bool Chunk::Up() const
  {
    return fUp;
  }

  //......................................................................
  void Chunk::SetUp(bool up)
  {
    fSegsDirty = true;

    if(up != fUp){
      std::swap(fEntryCellLo, fExitCellLo);
      std::swap(fEntryCellHi, fExitCellHi);
    }
    fUp = up;
  }

  //......................................................................
  art::PtrVector<rb::CellHit> Chunk::HitsOnLine(double grad,
                                                double icept) const
  {
    art::ServiceHandle<geo::Geometry> geom;

    const geo::PlaneGeo* geoplane = geom->Plane(fPlane);

    art::PtrVector<rb::CellHit> ret;

    for(unsigned int n = 0; n < fHits.size(); ++n){
      const art::Ptr<rb::CellHit>& hit = fHits[n];

      double xyz[3];
      const geo::CellGeo* geocell = geoplane->Cell(hit->Cell());
      geocell->GetCenter(xyz);
      const double z0 = xyz[2];
      const double dz = geocell->HalfD();
      const double z1 = z0-dz;
      const double z2 = z0+dz;
      const double v0 = xyz[hit->View()];
      const double dv = geocell->HalfW();
      const double vlo = v0-dv;
      const double vhi = v0+dv;

      double v1 = icept+grad*z1;
      double v2 = icept+grad*z2;
      if(v1 > v2) std::swap(v1, v2);

      if(v1 <= vhi && v2 >= vlo) ret.push_back(hit);
    } // end for n

    return ret;
  }

  //......................................................................
  std::pair<Segment, Segment> Chunk::GetSegs() const
  {
    if(!fSegsDirty) return fSegs;
    fSegsDirty = false;

    art::ServiceHandle<geo::Geometry> geom;

    const geo::PlaneGeo* plane = geom->Plane(fPlane);

    double v0lo, v1lo, v2lo, v3lo, z1lo, z2lo;
    CellPoints(plane, BottomExtremeCell(),
               v0lo, v1lo, v2lo, v3lo,
               z1lo, z2lo);

    double v0hi, v1hi, v2hi, v3hi, z1hi, z2hi;
    CellPoints(plane, TopExtremeCell(),
               v0hi, v1hi, v2hi, v3hi,
               z1hi, z2hi);

    double junk;
    if(BottomHitCell() != BottomExtremeCell()){
      CellPoints(plane, BottomHitCell(),
                 junk, junk, v2lo, v3lo,
                 junk, junk);
    }
    if(TopHitCell() != TopExtremeCell()){
      CellPoints(plane, TopHitCell(),
                 v0hi, v1hi, junk, junk,
                 junk, junk);
    }

    // If this is a completely dead chunk, it's OK to enter in the crack part
    // on one side and exit in that same part without ever traversing any real
    // cell.
    if(fHits.empty()){
      v1lo = v0lo;
      v1hi = v0hi;
      v2lo = v3lo;
      v2hi = v3hi;
    }

    // If the track is upgoing it can enter the plane in plastic below the
    // lowest cell, and exit in the plastic above the highest cell. But it
    // mustn't enter higher than the scintillator of the first cell, or lower
    // than the scintillator of the last cell. Vice versa for downgoing.

    if(fShowerChunk){
      if(fUp){
        fSegs.first = Segment(z1lo, v0lo-kCornerR, v2hi,
                              z1lo+kCornerR, v0lo, v3hi,
                              true, fPlane, fEntryCellLo, fEntryCellHi);
        fSegs.second = Segment(z2hi-kCornerR, v0lo, v3hi,
                               z2hi, v1lo, v3hi+kCornerR,
                               false, fPlane, fExitCellLo, fExitCellHi);
      }
      else{
        fSegs.first = Segment(z1hi, v1lo, v3hi+kCornerR,
                              z1hi+kCornerR, v0lo, v3hi,
                              true, fPlane, fEntryCellLo, fEntryCellHi);
        fSegs.second = Segment(z2lo-kCornerR, v0lo, v3hi,
                               z2lo, v0lo-kCornerR, v2hi,
                               false, fPlane, fExitCellLo, fExitCellHi);
      }

      return fSegs;
    }

    if(fUp){
      // bottom left
      fSegs.first = Segment(z1lo, v0lo-kCornerR, v2lo,
                            z1lo+kCornerR, v0lo, v3hi,
                            true, fPlane, fEntryCellLo, fEntryCellHi);
      // top right
      fSegs.second = Segment(z2hi-kCornerR, v0lo, v3hi,
                             z2hi, v1hi, v3hi+kCornerR,
                             false, fPlane, fExitCellLo, fExitCellHi);
    }
    else{
      // top left
      fSegs.first = Segment(z1hi, v1hi, v3hi+kCornerR,
                            z1hi+kCornerR, v0lo, v3hi,
                            true, fPlane, fEntryCellLo, fEntryCellHi);
      // bottom right
      fSegs.second = Segment(z2lo-kCornerR, v0lo, v3hi,
                             z2lo, v0lo-kCornerR, v2lo,
                             false, fPlane, fExitCellLo, fExitCellHi);
    }

    return fSegs;
  }

  //......................................................................
  bool Chunk::IsWorthShowering() const
  {
    // Single hit
    if(fEntryCellLo == fExitCellHi &&
       fEntryCellHi == fExitCellLo) return false;
    return true;
  }

  //......................................................................
  void Chunk::CellPoints(const geo::PlaneGeo* plane, int cell,
                         double& v0, double& v1, double& v2, double& v3,
                         double& z0, double& z1) const
  {
    const geo::View_t view = plane->View();

    // Measurements of the cell itself
    double vc, dv, zc, dz;
    CellMeasurements(view, plane->Cell(cell), vc, dv, zc, dz);
    v1 = vc-dv;
    v2 = vc+dv;
    z0 = zc-dz;
    z1 = zc+dz;

    // Would handle bad channels by walking these until the channel they're on
    // is good.
    const geo::CellGeo* cellabove = plane->Cell(cell+1);
    const geo::CellGeo* cellbelow = plane->Cell(cell-1);

    double junk;
    if(cellbelow){
      // Lowest point not in another cell
      CellMeasurements(view, cellbelow, v0, dv, junk, junk);
      v0 += dv;
    }
    else{
      // No cell below, must be edge of the detector, could be anywhere low
      v0 = -1e3;
    }

    if(cellabove){
      // Highest point not in another cell
      CellMeasurements(view, cellabove, v3, dv, junk, junk);
      v3 -= dv;
    }
    else{
      // No cell above, must be edge of the detector, could be anywhere high
      v3 = +1e3;
    }
  }

  //......................................................................
  void Chunk::CellMeasurements(geo::View_t view, const geo::CellGeo* cell,
                               double& v0, double& dv,
                               double& z0, double& dz) const
  {
    double xyz[3];
    cell->GetCenter(xyz);
    v0 = xyz[view];
    dv = cell->HalfW();
    z0 = xyz[2];
    dz = cell->HalfD();
  }

} // namespace
