////////////////////////////////////////////////////////////////////////
// \file    DiscreteTracker.cxx
// \author  Christopher Backhouse - bckhouse@caltech.edu
////////////////////////////////////////////////////////////////////////

// ROOT includes
#include "TH1.h"

// NOvASoft includes
#include "GeometryObjects/Geo.h"

#include "RecoBase/Track.h"
#include "RecoBase/Vertex.h"
#include "RecoBase/Cluster.h"
#include "RecoBase/CellHit.h"
#include "RecoBase/HitMap.h"
#include "RecoBase/FilterList.h"


#include "ChannelInfo/BadChanList.h"
#include "Utilities/AssociationUtil.h"
#include "Utilities/func/MathUtil.h"

// Package includes
#include "DiscreteTracker/Cand.h"
#include "DiscreteTracker/Chain.h"
#include "DiscreteTracker/Chunk.h"
#include "DiscreteTracker/Types.h"
#include "DiscreteTracker/View.h"

// Framework includes
#include "art/Framework/Core/EDProducer.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Services/Optional/TFileService.h"
#include "fhiclcpp/ParameterSet.h"

namespace dt
{
  /// Job module performing "discrete" track reconstruction
  class DiscreteTracker : public art::EDProducer
  {
  public:
    explicit DiscreteTracker(const fhicl::ParameterSet& pset);
    ~DiscreteTracker();

    /// Splits hits according to slice and view for feeding to \ref
    /// RecoDiscrete. Writes the returned tracks into the event.
    void produce(art::Event& evt);

    void reconfigure(const fhicl::ParameterSet& pset);

    void beginJob();
    void endJob();

  protected:
    /// \brief Cut to prevent choking on large air shower events
    ///
    /// Adapted from the version in KalmanTrack
    bool HighActivity(const art::Ptr<rb::Cluster>& clust) const;

    bool TryExtendChainOne(Cand& cur,
                           const Cand& prev,
                           const View& v,
                           const dt::View::ChunkMap& chunk_map,
                           Direction dir,
                           bool exactHits) const;

    void ExtendChain(const View& v,
                     const dt::View::ChunkMap& chunk_map,
                     Chain& chain, Direction dir) const;

    Chain RecoOneTrack(const View& v,
                       const dt::View::ChunkMap& chunk_map,
                       std::vector<rb::Track>* seeds_out,
                       bool& ok) const;

    void SpliceChains(std::vector<Chain>& chains) const;

    bool SpliceIsSane(const Chain& a, const Chain& b) const;

    std::vector<rb::Track>
    RecoDiscrete(const art::PtrVector<rb::CellHit>& cells,
                 std::vector<rb::Track>* seeds_out,
                 std::vector<rb::Track>* unspliced_out,
                 std::vector<rb::Track>* chains_out) const;

    std::vector<Segment>
    PossibleNextSegs(const View& v,
                     const dt::View::ChunkMap& chunk_map,
                     const Cand& cand, Direction dir,
                     std::vector<Chunk>& chunks_out,
                     bool shw, bool exactHits) const;

    double DotScoreHelper(double m1, double m2) const;

    double DotScore(const Cand& c1, const Cand& c2,
                    bool allowFlip = false) const;

    bool TryExtendCandOne(Cand& cand,
                          const View& v,
                          const dt::View::ChunkMap& chunk_map,
                          Direction dir,
                          const Cand& prev_dir,
                          bool shw,
                          bool exactHits) const;

    bool TryExtendCand(Cand& cand,
                       const Cand& old_cand,
                       const View& v,
                       const dt::View::ChunkMap& chunk_map,
                       Direction dir,
                       bool exactHits) const;

    Cand BestSeed(const dt::View& v, bool& ok) const;

    // Modifies \a chain in-place, returns bits clipped off, if any
    std::vector<Chain> CheckAndFixSparseChain(Chain& chain) const;

    /// \brief Toy vertex finder
    ///
    /// Uses the intersection of the two tracks in each view whose intersection
    /// is proportionally nearest their ends.
    rb::Vertex* FindVertex(std::vector<rb::Track> xTrks,
                           std::vector<rb::Track> yTrks) const;

    void MergeParasiteTracks(std::vector<rb::Track>& tracks) const;

    std::string fClusterInput; ///< Module name to look for Clusters under
    bool fConsiderBadChans;    ///< Passed to \ref View
    bool fObeyPreselection;    ///< Check rb::IsFiltered?

    /// \brief Joined Cands that only have one chunk in common must have a
    /// dot-product better than this.
    double fMinDotProduct;
    /// \brief The number of adjacent cells in a single plane, otherwise not
    /// fit into any track, required to declare them a vertical track
    int fVerticalTrackMinCells;

    bool fOnlyOneTrack; ///< Stop after the first track in each view

    TH1* fAngle; ///< Opening angle of all cands considered for joining
    TH1* fSinglesAngle; ///< For all cands with only one overlapping chunk

    FILE* fDebugFile; ///< File that track details are written to
  };


  //......................................................................
  DiscreteTracker::DiscreteTracker(const fhicl::ParameterSet& pset)
    : fDebugFile(0)
  {
    produces<std::vector<rb::Track> >();
    produces<std::vector<rb::Track> >("seeds");
    produces<std::vector<rb::Track> >("unspliced");
    produces<std::vector<rb::Track> >("chains");
    produces<std::vector<rb::Vertex> >();
    // Tracks are associated with the slice they came from
    produces<art::Assns<rb::Track, rb::Cluster> >();

    reconfigure(pset);
  }

  //......................................................................
  DiscreteTracker::~DiscreteTracker()
  {
  }

  //......................................................................
  void DiscreteTracker::reconfigure(const fhicl::ParameterSet& pset)
  {
    fClusterInput     = pset.get<std::string>("ClusterInput");
    fConsiderBadChans = pset.get<bool>("ConsiderBadChans");
    fMinDotProduct    = pset.get<double>("MinDotProduct");
    fVerticalTrackMinCells = pset.get<int>("VerticalTrackMinCells");
    fOnlyOneTrack     = pset.get<bool>("OnlyOneTrack");
    fObeyPreselection = pset.get<bool>("ObeyPreselection");

    if(fDebugFile) fclose(fDebugFile);
    fDebugFile = fopen(pset.get<std::string>("DebugFile").c_str(), "w");
  }

  //......................................................................
  void DiscreteTracker::beginJob()
  {
    art::ServiceHandle<art::TFileService> tfs;
    fAngle = tfs->make<TH1F>("angle", ";#Delta#theta", 100, 0, 1);
    fSinglesAngle = tfs->make<TH1F>("singles_angle", ";#Delta#theta",
                                    100, 0, 1);
  }

  //......................................................................
  void DiscreteTracker::endJob()
  {
    if(fDebugFile) fclose(fDebugFile);
  }

  //......................................................................
  void DiscreteTracker::produce(art::Event& evt)
  {
    std::unique_ptr<std::vector<rb::Track> > debug_seeds_col(new std::vector<rb::Track>);
    std::unique_ptr<std::vector<rb::Track> > debug_unspliced_col(new std::vector<rb::Track>);
    std::unique_ptr<std::vector<rb::Track> > debug_chains_col(new std::vector<rb::Track>);
    std::unique_ptr<std::vector<rb::Track> > trackcol(new std::vector<rb::Track>);
    std::unique_ptr<std::vector<rb::Vertex> > vtxcol(new std::vector<rb::Vertex>);
    // Associate tracks with the slice they came from
    std::unique_ptr<art::Assns<rb::Track, rb::Cluster> > assns(new art::Assns<rb::Track, rb::Cluster>);

    art::Handle<std::vector<rb::Cluster> > clusts;
    evt.getByLabel(fClusterInput, clusts);

    const int clustMax = clusts->size();
    for(int clustIdx = 0; clustIdx < clustMax; ++clustIdx){
      art::Ptr<rb::Cluster> clust(clusts, clustIdx);
      if(fObeyPreselection && rb::IsFiltered(evt, clust)) continue;

      if(clust->IsNoise()) continue;

      if(HighActivity(clust)){
        mf::LogWarning("DiscreteTracker") << "Skipping event due to very high activity";
        continue;
      }

      std::vector<rb::Track> tracks[2];
      for(geo::View_t view: {geo::kX, geo::kY}){
        const char viewC = (view == geo::kX) ? 'x' : 'y';

        art::PtrVector<rb::CellHit> cells;
        const int Nx = clust->NCell(view);
        for(int n = 0; n < Nx; ++n) cells.push_back(clust->Cell(view, n));

        fprintf(fDebugFile, "%d: %c: ", evt.event(), viewC);

        tracks[view] = RecoDiscrete(cells,
                                    debug_seeds_col.get(),
                                    debug_unspliced_col.get(),
                                    debug_chains_col.get());

        for(unsigned int n = 0; n < tracks[view].size(); ++n)
          fprintf(fDebugFile, "%u ", tracks[view][n].NCell());
        fprintf(fDebugFile, "\n");

        for(std::vector<rb::Track>::const_iterator it = tracks[view].begin();
            it != tracks[view].end(); ++it){
          trackcol->push_back(*it);
          // Associate most recently added track with the slice
          util::CreateAssn(*this, evt, *trackcol, clust, *assns);
        } // end for it
      } // end for view

      const rb::Vertex* vtx = FindVertex(tracks[geo::kX], tracks[geo::kY]);
      if(vtx) vtxcol->push_back(*vtx);
      delete vtx;
    } // end for clustIdx

    evt.put(std::move(debug_seeds_col), "seeds");
    evt.put(std::move(debug_unspliced_col), "unspliced");
    evt.put(std::move(debug_chains_col), "chains");
    evt.put(std::move(trackcol));
    evt.put(std::move(vtxcol));
    evt.put(std::move(assns));
  }

  //......................................................................
  bool DiscreteTracker::HighActivity(const art::Ptr<rb::Cluster>& clust) const
  {
    art::ServiceHandle<chaninfo::BadChanList> bcl;
    art::ServiceHandle<geo::Geometry> geom;

    // get the total number of non masked off channels by looping over all
    // planes/cells in the detector
    int numGoodChannels = 0;

    for(unsigned int iPlane = 0; iPlane < geom->NPlanes(); ++iPlane){
      const geo::PlaneGeo* p = geom->Plane(iPlane);
      for(unsigned int iCell = 0; iCell < p->Ncells(); ++iCell){
        if(!bcl->IsBad(iPlane, iCell)) ++numGoodChannels;
      }
    }

    return clust->NCell() > 0.15*numGoodChannels;
  }

  //......................................................................
  std::vector<Segment>
  DiscreteTracker::PossibleNextSegs(const View& v,
                                    const dt::View::ChunkMap& chunk_map,
                                    const Cand& cand, Direction dir,
                                    std::vector<Chunk>& chunks_out,
                                    bool shw, bool exactHits) const
  {
    std::vector<Segment> ret;

    // If the current chunk is only partially taken then the next thing to do
    // must be to try and complete it.
    if(dir == kUpstream && !cand.FirstSeg().left){
      ret.push_back(cand.FirstChunk().GetSegs().first);
      return ret;
    }

    if(dir == kDownstream && cand.LastSeg().left){
      ret.push_back(cand.LastChunk().GetSegs().second);
      return ret;
    }

    int nextPlane = v.AdjacentPlane(cand.ExtremalPlane(dir), dir);
    if(nextPlane == -1){
      if(!exactHits) return ret;
      if(dir == kDownstream){
        nextPlane = chunk_map.begin()->first;
        if(nextPlane <= int(cand.LastPlane())) return ret;
      }
      else{
        nextPlane = chunk_map.rbegin()->first;
        if(nextPlane >= int(cand.FirstPlane())) return ret;
      }
    }

    dt::View::ChunkMap::const_iterator it = chunk_map.find(nextPlane);
    if(exactHits){
      while(it == chunk_map.end()){
        nextPlane = v.AdjacentPlane(nextPlane, dir);
        if(nextPlane == -1) return ret;
        it = chunk_map.find(nextPlane);
      }
    }
    if(it == chunk_map.end()) return ret;

    const std::vector<Chunk>& chunks = it->second;
    const unsigned int chunkMax = chunks.size();

    for(unsigned int chunkIdx = 0; chunkIdx < chunkMax; ++chunkIdx){
      Chunk chunk = chunks[chunkIdx];
      if(exactHits) chunk.SetUp(cand.Up());

      if(chunk.Up() != cand.Up()) continue;
      if(shw && !chunk.IsWorthShowering()) continue;
      if(shw) chunk.SetShowerChunk();

      chunks_out.push_back(chunk);
      if(dir == kUpstream)
        ret.push_back(chunk.GetSegs().second);
      else
        ret.push_back(chunk.GetSegs().first);
    }

    return ret;
  }

  //......................................................................
  double DiscreteTracker::DotScoreHelper(double m1, double m2) const
  {
    return fabs(1+m1*m2)/sqrt((1+m1*m1)*(1+m2*m2));
  }

  //......................................................................
  double DiscreteTracker::DotScore(const Cand& c1, const Cand& c2,
                                   bool allowFlip) const
  {
    // Can't even determine a direction
    if(c1.NSegs() < 2 || c2.NSegs() < 2) return -1;

    double bestDot = -2;

    for(bool flip1: {false, true}){
      if(flip1 && !allowFlip) continue;

      bool ok = true;
      const Cand c1f = flip1 ? c1.MaybeFlip(ok) : c1;
      if(!ok) continue;

      for(bool flip2: {false, true}){
        if(flip2 && !allowFlip) continue;

        bool ok = true;
        const Cand c2f = flip2 ? c2.MaybeFlip(ok) : c2;
        if(!ok) continue;

        const double m1lo = c1f.MinGradient();
        const double m1hi = c1f.MaxGradient();
        const double m2lo = c2f.MinGradient();
        const double m2hi = c2f.MaxGradient();

        assert(m1lo <= m1hi);
        assert(m2lo <= m2hi);

        // Overlapping ideas of direction, best is perfect
        if((m1hi > m2lo && m1lo < m2hi) ||
           (m1lo < m2hi && m1hi > m2lo)) return 1;

        bestDot = std::max(bestDot, DotScoreHelper(m1hi, m2lo));
        bestDot = std::max(bestDot, DotScoreHelper(m1lo, m2hi));
      } // end for flip2
    } // end for flip1

    return bestDot;
  }

  //......................................................................
  bool DiscreteTracker::TryExtendCandOne(Cand& cand,
                                         const View& v,
                                         const dt::View::ChunkMap& chunk_map,
                                         Direction dir,
                                         const Cand& prev_dir,
                                         bool shw, bool exactHits) const
  {
    // TODO: hoist this out of while loop below
    std::vector<Chunk> chunks;
    const std::vector<Segment> segs = PossibleNextSegs(v, chunk_map,
                                                       cand, dir, chunks, shw,
                                                       exactHits);
    if(segs.empty()) return false;

    // Backup of cand, so we can restore it in case TryAddSeg succeeds
    Cand bak = cand;

    std::vector<Cand> extended;

    const unsigned int segMax = segs.size();
    for(unsigned int segIdx = 0; segIdx < segMax; ++segIdx){
      const Segment& seg = segs[segIdx];

      if(cand.TryAddSeg(seg, dir)){

        // May need to add the chunk to match
        if((dir == kUpstream && !seg.left) ||
           (dir == kDownstream && seg.left)){
          cand.AddChunk(chunks[segIdx], dir);
        }

        const bool smallOverlap = (bak.NSegs() <= 2);
        if(true){//smallOverlap){
          // If there's only a small overlap, ensure dot product is sensible
          const double dotscore = DotScore(prev_dir, cand);
          fAngle->Fill(acos(dotscore));
          if(smallOverlap) fSinglesAngle->Fill(acos(dotscore));
          if(smallOverlap && dotscore < fMinDotProduct){
            cand = bak;
            continue;
          }
        }

        // Otherwise it's good
        extended.push_back(cand);
        cand = bak;
      } // end if added seg
    } // end for segIdx

    if(extended.empty()) return false;

    // TODO: Arbitrary for now
    cand = extended.front();
    if(extended.size() > 1){
      //      std::cout << "***WARNING: MORE THAN ONE EXTENSION CANDIDATE (" << extended.size() << "). SHOULD HAVE TRIED TO MAKE A SMARTER DECISION***" << std::endl;
    }

    return true;
  }

  //......................................................................
  bool DiscreteTracker::TryExtendCand(Cand& cand,
                                      const Cand& old_cand,
                                      const View& v,
                                      const dt::View::ChunkMap& chunk_map,
                                      Direction dir,
                                      bool exactHits) const
  {
    const Cand bak = cand;

    for(bool shwNext: {false, true}){
      if(shwNext && !cand.IsExtremalChunkComplete(dir)) continue;
      for(bool shwLast: {false, true}){
        if(shwLast && cand.ExtremalChunk(dir).IsShowerChunk()) continue;
        if(shwLast && !cand.ExtremalChunk(dir).IsWorthShowering()) continue;

        if(shwLast && cand.NSegs() <= 2) continue;

        if(shwLast) cand.MarkExtremalChunkShower(dir);
        // Reset back by hand at end of loop. Careful about early exits

        bool any = false;
        while(TryExtendCandOne(cand, v, chunk_map, dir, old_cand, shwNext, exactHits)) any = true;

        if(any) return true;

        // TODO: can probably hoist this higher with care
        bool ok;
        Cand flipped = cand.MaybeFlip(ok);
        if(!ok){
          if(shwLast) cand = bak;
          continue;
        }

        while(TryExtendCandOne(flipped, v, chunk_map, dir, old_cand, shwNext, exactHits)) any = true;

        if(any){
          cand = flipped;
          return true;
        }

        if(shwLast) cand = bak;
      } // end for shwLast
    } // end for shwNext

    cand = bak;
    return false;
  }

  //......................................................................
  bool DiscreteTracker::TryExtendChainOne(Cand& cur,
                                          const Cand& prev,
                                          const View& v,
                                          const dt::View::ChunkMap& chunk_map,
                                          Direction dir,
                                          bool exactHits) const
  {
    while(true){
      if(cur.NSegs() < 2) return false;
      if(TryExtendCand(cur, prev, v, chunk_map, dir, exactHits)) return true;

      cur.ShortenOne(dir);

      // Too much freedom
      if(cur.IsAllShower()) return false;
      if(cur.NChunks() == 0) return false; // all dead
      if(cur.AllHits().empty()) return false;
    }
  }

  //......................................................................
  void DiscreteTracker::ExtendChain(const View& v,
                                    const dt::View::ChunkMap& chunk_map,
                                    Chain& chain, Direction dir) const
  {
    Cand cur = chain.ExtremalCand(dir);

    while(TryExtendChainOne(cur, chain.ExtremalCand(dir), v, chunk_map, dir, false)){
      chain.Add(cur, dir);
    }
  }

  //......................................................................
  Chain DiscreteTracker::RecoOneTrack(const View& v,
                                      const dt::View::ChunkMap& chunk_map,
                                      std::vector<rb::Track>* seeds_out,
                                      bool& ok) const
  {
    const Cand& seed = BestSeed(v, ok);
    if(!ok) return Chain();
    seeds_out->push_back(seed.ToTrack());

    Chain chain(seed);
    ExtendChain(v, chunk_map, chain, kUpstream);
    ExtendChain(v, chunk_map, chain, kDownstream);
    ok = true;
    return chain;
  }

  //......................................................................
  std::vector<rb::Track>
  DiscreteTracker::RecoDiscrete(const art::PtrVector<rb::CellHit>& cells,
                                std::vector<rb::Track>* seeds_out,
                                std::vector<rb::Track>* unspliced_out,
                                std::vector<rb::Track>* chains_out) const
  {
    std::vector<rb::Track> ret;
    if(cells.empty()) return ret;

    std::vector<Chain> chains;

    dt::View v(cells, fConsiderBadChans);

    while(true){
      // TODO, this should be shared
      const dt::View::ChunkMap chunk_map = v.MakeChunks();

      bool ok;
      Chain chain = RecoOneTrack(v, chunk_map, seeds_out, ok);

      if(!ok) break;

      chain.TrimEnds();
      // TODO: how can this happen?
      if(chain.FirstCand().NSegs() < 2) break;
      if(chain.LastCand().NSegs() < 2) break;

      chains.push_back(chain);
      unspliced_out->push_back(chain.ToTrack());

      // We're going to try and find another track, but first remove all the
      // hits we've already explained.
      art::PtrVector<rb::CellHit> hits = chain.AllHits();
      for(unsigned int n = 0; n < hits.size(); ++n){
        v.Channel(hits[n]->Plane(), hits[n]->Cell()).MarkHitFound();
      }

      if(fOnlyOneTrack) break;
    } // end while

    SpliceChains(chains);

    unsigned int N = chains.size();
    for(unsigned int n = 0; n < N; ++n){
      const std::vector<Chain> remnants = CheckAndFixSparseChain(chains[n]);
      chains.insert(chains.end(), remnants.begin(), remnants.end());
    }

    N = chains.size();
    for(unsigned int n = 0; n < N; ++n){
      // Comes out of Truncate() at least sometimes
      if(chains[n].AllHits().empty()){
        mf::LogWarning("DiscreteTracker") << "Track with no hits. Skipping.";
        continue;
      }

      if(chains[n].FirstSeg().zL > chains[n].LastSeg().zL){
        mf::LogWarning("DiscreteTracker") << "Reversed track. Skipping.";
        continue;
      }

      ret.push_back(chains[n].ToTrack());
      const std::vector<rb::Track> debug = chains[n].ToDebugTracks();
      chains_out->insert(chains_out->end(), debug.begin(), debug.end());
    }

    MergeParasiteTracks(ret);

    return ret;
  }

  //......................................................................
  bool DiscreteTracker::SpliceIsSane(const Chain& a, const Chain& b) const
  {
    // Must not overlap. Can't jump more than half the whole total length

    if(a.LastPlane() < b.FirstPlane()){
      if(DotScore(a.LastCand(), b.FirstCand(), true) < fMinDotProduct) return false;

      return (2*(b.FirstPlane()-a.LastPlane()) < b.LastPlane()-a.FirstPlane());
    }
    if(b.LastPlane() < a.FirstPlane()){
      if(DotScore(b.LastCand(), a.FirstCand(), true) < fMinDotProduct) return false;

      return (2*(a.FirstPlane()-b.LastPlane()) < a.LastPlane()-b.FirstPlane());
    }

    return false;
  }

  struct SpliceDesc
  {
    SpliceDesc(int cIdx, int pIdx, Chain& ch, int g, int t)
      : childIdx(cIdx), parentIdx(pIdx), spliced(ch), gap(g), total(t) {}

    bool operator<(const SpliceDesc& sd) const
    {
      // Smaller gaps are better. Break ties with total length of spliced
      // pieces.
      if(gap == sd.gap) return total > sd.total;
      return gap < sd.gap;
    }

    int childIdx, parentIdx;
    Chain spliced;
    int gap;
    int total;
  };

  //......................................................................
  void DiscreteTracker::SpliceChains(std::vector<Chain>& chains) const
  {
    while(true){
      std::vector<int> parentIdxs(chains.size());
      for(unsigned int n = 0; n < chains.size(); ++n) parentIdxs[n] = n;

      std::vector<int> childIdxs(chains.size());
      for(unsigned int n = 0; n < chains.size(); ++n) childIdxs[n] = n;

      // Keep the best splice first
      std::multiset<SpliceDesc> splices;

      for(unsigned int i = 0; i < chains.size(); ++i){
        const unsigned int parentIdx = parentIdxs[i];
        const Chain& parent = chains[parentIdx];

        for(Direction dir: {kDownstream, kUpstream}){
          for(unsigned int j = 0; j < chains.size(); ++j){
            const unsigned int childIdx = childIdxs[j];

            if(childIdx == parentIdx) continue;

            const Chain& child = chains[childIdx];

            int gap;
            if(dir == kDownstream)
              gap = child.FirstPlane() - parent.LastPlane();
            else
              gap = parent.FirstPlane() - child.LastPlane();

            // Have to get to correct side
            if(gap < 0) continue;

            // Bail out early if we know we can't beat the best splice
            if(!splices.empty() && gap > splices.begin()->gap) continue;

            if(!SpliceIsSane(parent, child)) continue;

            View v(child.AllHits(), fConsiderBadChans);
            // Doing it this way preserves the "shower" flag on chunks
            const dt::View::ChunkMap chunk_map = child.AsChunkMap();

            Cand join = parent.ExtremalCand(dir);
            if(!TryExtendChainOne(join, parent.ExtremalCand(dir),
                                  v, chunk_map,
                                  dir, true)) continue;

            // Also need to check that the join is willing to be made in the
            // other direction. Otherwise you can get a one-hit cand on one
            // side, that's willing to match almost anything, in crazy ways.
            View pview(parent.AllHits(), fConsiderBadChans);
            const dt::View::ChunkMap pmap = parent.AsChunkMap();
            Cand join2 = child.ExtremalCand(OtherDir(dir));
            if(!TryExtendChainOne(join2, child.ExtremalCand(OtherDir(dir)),
                                  pview, pmap, OtherDir(dir), true)) continue;

            Chain spliced = parent;
            spliced.Add(join, dir);
            // The splice can often subsume the whole child chain
            if(child.ExtremalPlane(dir) != join.ExtremalPlane(dir)){
              spliced.Add(child, dir);
            }

            splices.insert(SpliceDesc(childIdx, parentIdx, spliced, gap,
                                      child.NHitPlanes()+parent.NHitPlanes()));
          } // end for j (childIdx)
        } // end for dir
      } // end for i (parentIdx)

      if(splices.empty()) return;

      const SpliceDesc sd = *splices.begin();
      chains[sd.parentIdx] = sd.spliced;
      chains.erase(chains.begin()+sd.childIdx);
    }
  }

  //......................................................................
  Cand DiscreteTracker::BestSeed(const dt::View& v, bool& ok) const
  {
    ok = true;

    // TODO, we could iterate over these instead
    //    const std::set<unsigned int> planes = v.GetPlanes();
    const dt::View::ChunkMap chunks_map = v.MakeChunks();

    Cand best(v.GeoView(), true);

    if(chunks_map.empty()){
      ok = false;
      return best;
    }

    std::list<Cand> live;

    // Go through planes
    for(dt::View::ChunkMap::const_iterator it = chunks_map.begin(); it != chunks_map.end(); ++it){
      const int plane = it->first;

      const std::vector<Chunk>& chunks = it->second;

      std::list<Cand> newLive;

      for(std::list<Cand>::iterator live_it = live.begin(); live_it != live.end(); ++live_it){

        if(plane != v.NextPlane(live_it->LastChunk().Plane())){
          if(live_it->NSegs() > best.NSegs()) best = *live_it;
          continue;
        }

        // TODO: make a less wasteful version of TryAddSeg? Try to keep just
        // the one list.
        Cand curCand = *live_it; // Take a copy

        // Were any chunks any use?
        bool any = false;

        for(unsigned int chunkIdx = 0; chunkIdx < chunks.size(); ++chunkIdx){
          const Chunk& chunk = chunks[chunkIdx];
          // TODO: potentially real data could have seeds that never had two
          // adjacent hits?
          if(chunk.AllHits().empty()) continue;

          // TODO: have chunks given back in two lists?
          if(chunk.Up() != curCand.Up()) continue;

          const std::pair<Segment, Segment>& segs = chunk.GetSegs();

          if(curCand.TryAddSeg(segs.first, kDownstream)){
            curCand.AddChunk(chunk, kDownstream);
            if(curCand.TryAddSeg(segs.second, kDownstream)){
              // Good, keep going
              newLive.insert(newLive.end(), curCand);
              curCand = *live_it; // New one to play with
            }
            else{
              // Failed second half, first might be useful
              if(curCand.NSegs() > best.NSegs()) best = curCand;
              curCand = *live_it;
            }
            // Got at least the first seg in
            any = true;
          }
        } // end for chunkIdx

        if(!any){
          // Got stuck, but it might still be the best
          if(curCand.NSegs() > best.NSegs()) best = curCand;
          // No need to make a new curCand anymore
        }
      } // end for live_it

      live.swap(newLive);
      newLive.clear();

      // Seed new cands from this chunk
      for(unsigned int chunkIdx = 0; chunkIdx < chunks.size(); ++chunkIdx){
        const Chunk& chunk = chunks[chunkIdx];

        const std::pair<Segment, Segment>& segs = chunk.GetSegs();

        Cand newcand(v.GeoView(), chunk.Up());
        newcand.AddChunk(chunk, kDownstream);
        bool startok = newcand.TryAddSeg(segs.first, kDownstream);
        assert(startok);
        startok = startok && newcand.TryAddSeg(segs.second, kDownstream);
        assert(startok);
        live.insert(live.end(), newcand);

        // Only the second seg
        Cand newcand2(v.GeoView(), chunk.Up());
        newcand2.AddChunk(chunk, kDownstream);
        startok = startok && newcand2.TryAddSeg(segs.second, kDownstream);
        live.insert(live.end(), newcand2);

        assert(startok);
      } // end for chunkIdx
    } // end for it

    // Maybe one that made it all the way to the end is the best?
    for(std::list<Cand>::iterator live_it = live.begin(); live_it != live.end(); ++live_it){
      if(live_it->NSegs() > best.NSegs()) best = *live_it;
    }

    if(best.NSegs() == 0) ok = false;
    // TODO: how can this even happen? But if it does we'll get stuck in an
    // endless loop, so bail out of the rest of the event.
    if(best.AllHits().empty()) ok = false;
    return best;
  }

  //......................................................................
  std::vector<Chain> DiscreteTracker::CheckAndFixSparseChain(Chain& chain) const
  {
    std::vector<Chain> remnants;

    const std::vector<Chunk> chunks = chain.AllChunks();
    assert(!chunks.empty());

    std::set<int> planes;
    for(unsigned int n = 0; n < chunks.size(); ++n)
      planes.insert(chunks[n].Plane());

    bool progress;
    do{
      progress = false;
      for(Direction end: {kUpstream, kDownstream}){
        const Direction dir = OtherDir(end);

        int nTot = 0;
        int nMiss = 0;

        int curPlane = chain.ExtremalPlane(end);

        if(chain.AllHits().empty()) break;
        View v(chain.AllHits(), fConsiderBadChans);

        do{
          ++nTot;

          const bool hitHere = planes.count(curPlane);
          if(!hitHere) ++nMiss;

          // If we've accumulated more than 1/2 gaps, and have now come back
          // into hits
          if(hitHere && nMiss > 0){
            // Want to allow HGH... and ban GH... and HGH
            if(nTot == nMiss || (nMiss > 1 && nMiss*2 >= nTot-1)){
              Chain remnant = chain;
              remnant.Truncate(curPlane+((dir == kUpstream) ? +1 : -1), dir);
              remnants.push_back(remnant);

              chain.Truncate(curPlane, end);
              progress = true;
              // Shouldn't look any further from this end, but try the other one
              break;
            }
          }

          curPlane = v.AdjacentPlane(curPlane, dir);
        } while(curPlane != -1);
      } // end for end/dir
    } while(progress);

    return remnants;
  }

  bool CompareByLength(const rb::Track& a, const rb::Track& b)
  {
    return a.TotalLength() < b.TotalLength();
  }

  //......................................................................
  rb::Vertex* DiscreteTracker::FindVertex(std::vector<rb::Track> xTrks,
                                          std::vector<rb::Track> yTrks) const
  {
    if(xTrks.empty() || yTrks.empty()) return 0;

    // Longest tracks first. Only used for fallback at the end
    std::sort(xTrks.rbegin(), xTrks.rend(), CompareByLength);
    std::sort(yTrks.rbegin(), yTrks.rend(), CompareByLength);

    double bestVx, bestVy, bestVz1, bestVz2;
    // I don't believe there's a path that doesn't set these, but need this to
    // shut the compiler up in optimized builds.
    bestVx = bestVy = bestVz1 = bestVz2 = 0;

    for(geo::View_t view: {geo::kX, geo::kY}){
      double bestScore = 1e10;
      const std::vector<rb::Track>& trks = (view == geo::kX) ? xTrks : yTrks;
      for(unsigned int idx1 = 0; idx1 < trks.size()-1; ++idx1){
        if(trks[idx1].ExtentPlane() < 3) continue;
        for(bool end1: {false, true}){
          for(unsigned int idx2 = idx1+1; idx2 < trks.size(); ++idx2){
            if(trks[idx2].ExtentPlane() < 3) continue;
            for(bool end2: {false, true}){

              // Start and direction defining point of these two lines
              const TVector3 s1 = end1 ? trks[idx1].Stop() : trks[idx1].Start();
              const TVector3 s2 = end2 ? trks[idx2].Stop() : trks[idx2].Start();
              const TVector3 d1 = trks[idx1].TrajectoryPoint(end1 ? trks[idx1].NTrajectoryPoints()-2 : 1);
              const TVector3 d2 = trks[idx2].TrajectoryPoint(end2 ? trks[idx2].NTrajectoryPoints()-2 : 1);

              double vxy, vz;
              const bool ok = geo::LineIntersection(s1[view], s1.Z(), d1[view], d1.Z(),
                                                    s2[view], s2.Z(), d2[view], d2.Z(),
                                                    vxy, vz);
              if(!ok) continue;

              const double dist1 = util::pythag(s1[view]-vxy, s1.Z()-vz);
              const double dist2 = util::pythag(s2[view]-vxy, s2.Z()-vz);

              // Favour both long tracks and hits close to the ends
              const double score = dist1/trks[idx1].TotalLength()+dist2/trks[idx2].TotalLength();
              if(score < bestScore){
                bestScore = score;
                if(view == geo::kX){
                  bestVx = vxy;
                  bestVz1 = vz;
                }
                else{
                  bestVy = vxy;
                  bestVz2 = vz;
                }
              } // end if better
            } // end for end2
          } // end for idx2
        } // end for end1
      } // end for idx1
      // Too far from the ends, fallback
      if(bestScore > 6){
        if(view == geo::kX){
          bestVx = trks[0].Start().X();
          bestVz1 = trks[0].Start().Z();
        }
        else{
          bestVy = trks[0].Start().Y();
          bestVz2 = trks[0].Start().Z();
        }
      }
    } // end for view

    // Time is an afterthought
    return new rb::Vertex(bestVx, bestVy, (bestVz1+bestVz2)/2, xTrks[0].MinTNS());
  }

  //......................................................................
  void DiscreteTracker::MergeParasiteTracks(std::vector<rb::Track>& tracks) const
  {
    // Worried this could be too aggressive around the vertex

    const unsigned int N = tracks.size();
    if(N < 2) return;

    // Longest first
    std::sort(tracks.begin(), tracks.end(),
              [](const rb::Track& a, const rb::Track& b){
                return a.ExtentPlane() > b.ExtentPlane();
              });

    for(unsigned int hostIdx = 0; hostIdx < N-1; ++hostIdx){
      const rb::HitMap hmap(&tracks[hostIdx]);
      for(unsigned int paraIdx = hostIdx+1; paraIdx < N; ++paraIdx){
        const rb::Track& para = tracks[paraIdx];
        bool match = true;
        for(unsigned int i = 0; i < para.NCell(); ++i){
          const art::Ptr<rb::CellHit> chit = para.Cell(i);
          if(!hmap.CellExists(chit->Plane(), chit->Cell()-1) &&
             !hmap.CellExists(chit->Plane(), chit->Cell()+1)){
            match = false;
            break;
          }
        } // end for i
        if(match){
	  // Make sure there are no duplicates in the final host track
	  rb::HitMap hmap(tracks[hostIdx].AllCells());
	  for(unsigned int n = 0; n < para.NCell(); ++n){
	    const art::Ptr<rb::CellHit> chit = para.Cell(n);
	    if(!hmap.CellExists(chit->Plane(), chit->Cell()))
	      tracks[hostIdx].Add(chit);
	  }
          tracks.erase(tracks.begin()+paraIdx);
          MergeParasiteTracks(tracks);
          return;
        }
      } // end for paraIdx
    } // end for hostIdx
  }

  DEFINE_ART_MODULE(DiscreteTracker);

} //namespace dt
////////////////////////////////////////////////////////////////////////

