////////////////////////////////////////////////////////////////////////
// \file    ViewMerger.cxx
// \author  Christopher Backhouse - bckhouse@caltech.edu
////////////////////////////////////////////////////////////////////////

#include <algorithm>

// NOvASoft includes
#include "ChannelInfo/BadChanList.h"
#include "Geometry/Geometry.h"
#include "GeometryObjects/Geo.h"

#include "RecoBase/Track.h"
#include "RecoBase/Vertex.h"
#include "RecoBase/Cluster.h"
#include "RecoBase/HitMap.h"
#include "RecoBase/CellHit.h"
#include "RecoBase/RecoHit.h"
#include "RecoBase/FilterList.h"

#include "Utilities/AssociationUtil.h"
#include "Utilities/func/MathUtil.h"

// Framework includes
#include "art/Framework/Core/EDProducer.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h"
#include "fhiclcpp/ParameterSet.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Services/Optional/TFileService.h"
#include "art/Framework/Core/FindManyP.h"
#include "art/Persistency/Common/PtrVector.h"
#include "art/Persistency/Common/Ptr.h"

namespace dt
{
  /// Match 2D tracks into 3D tracks
  class ViewMerger : public art::EDProducer
  {
  public:
    explicit ViewMerger(const fhicl::ParameterSet& pset);
    ~ViewMerger();

    virtual void reconfigure(const fhicl::ParameterSet& pset);

    virtual void beginJob();
    virtual void endJob();

    virtual void produce(art::Event& evt);

  protected:
    /// \brief Do most of the work. Called from \ref produce
    ///
    /// \param sliceMap All the hits in this slice
    /// \param xtracks The 2D tracks from the x-view to merge. Those used will
    /// be removed from the list. Any left unmerged will remain in the list.
    /// \param ytracks Ditto
    /// \return The 3D tracks produced
    std::vector<rb::Track>
    MatchTracks(const rb::HitMap& sliceMap,
                std::list<rb::Track>& xtracks,
                std::list<rb::Track>& ytracks) const;

    /// Find vertical tracks seperated by one plane. Return results of merge
    std::vector<rb::Track>
    MatchVerticals(std::list<rb::Track>& xtracks,
                   std::list<rb::Track>& ytracks) const;

    /// Tilt tracks arbitrarily and merge them into a 3D track
    rb::Track ZipVerticalTracks(const rb::Track& xtrk,
                                const rb::Track& ytrk) const;

    /// Corrected for direction cosines, so if you divide this by extent in Z
    /// you'd get dE/dx
    void TotalChargePerView(const rb::Track& zip,
                            double& totx, double& toty) const;

    /// If we extend tracks to match in length, how many cells do they pass
    /// over but don't hit? Use the mean energy deposit in these tracks to
    /// estimate how much charge has gone missing, in GeV.
    double MissingChargeByExtension(const rb::HitMap& sliceMap,
                                    const rb::Track& aTrk,
                                    const rb::Track& bTrk) const;

    double ScoreForExtension(const rb::HitMap& sliceMap,
                             const rb::Track& xTrk,
                             const rb::Track& yTrk) const;

    double ScoreForJoinPlusExtension(const rb::HitMap& sliceMap,
                                     const rb::Track& single,
                                     const rb::Track& parta,
                                     const rb::Track& partb) const;

    /// Append two tracks in the same view (crudely).
    rb::Track JoinTracks(const rb::Track& parta, const rb::Track& partb) const;

    /// \a hmap should contain cells that are hit and so OK to pass through
    int CountUnexplainedOnLine(const rb::HitMap& hmap,
                               TVector3 start, TVector3 end,
                               geo::View_t view) const;

    double ChargeDifferenceBetweenViews(const rb::Track& zip) const;

    /// Extend each track to the vertex point, only if it's completely
    /// uncontroversial.
    void MaybeExtendToVertex(const rb::Vertex* vtx,
                             std::vector<rb::Track>& tracks) const;

    /// Find number of overlapping planes, return unexp/overlap
    double MatchScore(double unexp,
                      const rb::Track& aTrk,
                      const rb::Track& bTrk) const;

    /// Make a 3D track based on two 2D tracks
    rb::Track ZipTracks(const rb::Track& aTrk, const rb::Track& bTrk) const;

    /// Helper for \ref Adjacent. Fills \a mins and \a maxs with minimum and
    /// maximum cell number in each plane of \a trk
    void Extremes(const rb::Track& trk,
                  std::map<int, int>& mins, std::map<int, int>& maxs) const;

    /// Helper for \ref MergeAdjacentTracks. Are all hits on one track adjacent
    /// to the other one?
    bool Adjacent(const rb::Track& a, const rb::Track& b) const;

    /// Any tracks that are adjacent along the whole length of the shorter one
    /// will be merged
    void MergeAdjacentTracks(std::list<rb::Track>& trks) const;

    /// \brief Toy vertex finder
    ///
    /// Uses the intersection of the 3D tracks whose intersection is
    /// proportionally nearest their ends.
    rb::Vertex* FindVertex(std::vector<rb::Track> trks) const;

    /// \brief Turn a 2D track into a 3D track
    ///
    /// Uses \a merged to guess a W position
    rb::Track FakeThirdDimension(const rb::Track& trk,
                                 const std::vector<rb::Track>& merged) const;

    std::string fSliceInput; ///< Module name to look for slices under
    std::string fTrackInput; ///< Module name to look for 2D tracks under

    bool fObeyPreselection;  ///< Check rb::IsFiltered?

    FILE* fDebugFile;
  };


  ViewMerger::ViewMerger(const fhicl::ParameterSet& pset)
    : fDebugFile(0)
  {
    produces<std::vector<rb::Track> >();
    produces<std::vector<rb::Vertex> >();
    // Tracks are associated with the slice they came from
    produces<art::Assns<rb::Track, rb::Cluster> >();

    reconfigure(pset);
  }

  //......................................................................
  ViewMerger::~ViewMerger()
  {
  }

  //......................................................................
  void ViewMerger::reconfigure(const fhicl::ParameterSet& pset)
  {
    fSliceInput = pset.get<std::string>("SliceInput");
    fTrackInput = pset.get<std::string>("TrackInput");

    if(fDebugFile) fclose(fDebugFile);
    fDebugFile = fopen(pset.get<std::string>("DebugFile").c_str(), "w");

    fObeyPreselection = pset.get<bool>("ObeyPreselection");
  }

  //......................................................................
  void ViewMerger::beginJob()
  {
  }

  //......................................................................
  void ViewMerger::endJob()
  {
    if(fDebugFile) fclose(fDebugFile);
  }

  //......................................................................
  void ViewMerger::produce(art::Event& evt)
  {
    std::unique_ptr<std::vector<rb::Track> > trackcol(new std::vector<rb::Track>);
    std::unique_ptr<std::vector<rb::Vertex> > vtxcol(new std::vector<rb::Vertex>);
    // Associate tracks with the slice they came from
    std::unique_ptr<art::Assns<rb::Track, rb::Cluster> > assns(new art::Assns<rb::Track, rb::Cluster>);

    art::Handle<std::vector<rb::Cluster> > hSlices;
    evt.getByLabel(fSliceInput, hSlices);

    art::PtrVector<rb::Cluster> slices;
    for(unsigned int i = 0; i < hSlices->size(); ++i)
      slices.push_back(art::Ptr<rb::Cluster>(hSlices, i));

    const unsigned int sliceMax = slices.size();
    art::FindManyP<rb::Track> fmt(hSlices, evt, fTrackInput);
    for(unsigned int sliceIdx = 0; sliceIdx < sliceMax; ++sliceIdx){
      if(slices[sliceIdx]->IsNoise()) continue;

      if(fObeyPreselection && rb::IsFiltered(evt, slices[sliceIdx])) continue;

      // Get all the tracks associated with this slice
      std::vector< art::Ptr<rb::Track> > pTracks = fmt.at(sliceIdx);

      if(pTracks.empty()) continue;

      const rb::HitMap sliceMap(slices[sliceIdx]->AllCells());

      // In a nicer form
      std::list<rb::Track> xtracks, ytracks;
      for(unsigned int trkIdx = 0; trkIdx < pTracks.size(); ++trkIdx){
        // No sensible way to merge these, and they make a mess
        if(pTracks[trkIdx]->NCell() == 1) continue;
        if(pTracks[trkIdx]->View() == geo::kX)
          xtracks.push_back(*pTracks[trkIdx]);
        else
          ytracks.push_back(*pTracks[trkIdx]);
      } // end for trkIdx

      MergeAdjacentTracks(xtracks);
      MergeAdjacentTracks(ytracks);

      // Actually do all the work
      std::vector<rb::Track> matchedTracks = MatchTracks(sliceMap,
                                                         xtracks, ytracks);
      std::vector<rb::Track> verts = MatchVerticals(xtracks, ytracks);
      matchedTracks.insert(matchedTracks.end(), verts.begin(), verts.end());

      // Write summary to debug file
      fprintf(fDebugFile, "%d: ", evt.event());

      if(!matchedTracks.empty()){
        fprintf(fDebugFile, "3d: ");
        for(unsigned int n = 0; n < matchedTracks.size(); ++n){
          fprintf(fDebugFile, "%u ", matchedTracks[n].NCell());
        }
      }

      if(!xtracks.empty() || !ytracks.empty()){
        fprintf(fDebugFile, "2d: ");
        typedef std::list<rb::Track>::iterator it_t;
        for(it_t it = xtracks.begin(); it != xtracks.end(); ++it)
          fprintf(fDebugFile, "%u(x) ", it->NCell());
        for(it_t it = ytracks.begin(); it != ytracks.end(); ++it)
          fprintf(fDebugFile, "%u(y) ", it->NCell());
      }

      fprintf(fDebugFile, "\n");

      const rb::Vertex* vtx = FindVertex(matchedTracks);
      if(vtx){
        MaybeExtendToVertex(vtx, matchedTracks);
        vtxcol->push_back(*vtx);
      }
      delete vtx;

      // This can look pretty silly and I'm not sure it adds anything.
      /*
      // Invent a third dimension for the unmatched 2D tracks and include them
      // in the list.
      std::vector<rb::Track> fakes;

      for(std::list<rb::Track>::iterator it = xtracks.begin(); it != xtracks.end(); ++it){
        fakes.push_back(FakeThirdDimension(*it, matchedTracks));
      }
      for(std::list<rb::Track>::iterator it = ytracks.begin(); it != ytracks.end(); ++it){
        fakes.push_back(FakeThirdDimension(*it, matchedTracks));
      }

      matchedTracks.insert(matchedTracks.end(), fakes.begin(), fakes.end());
      */

      for(unsigned int n = 0; n < matchedTracks.size(); ++n){
        trackcol->push_back(matchedTracks[n]);
        util::CreateAssn(*this, evt, *trackcol, slices[sliceIdx], *assns);
      }
    } // end for sliceIdx

    evt.put(std::move(trackcol));
    evt.put(std::move(vtxcol));
    evt.put(std::move(assns));
  }

  //......................................................................
  std::vector<rb::Track> ViewMerger::
  MatchTracks(const rb::HitMap& sliceMap,
              std::list<rb::Track>& xtracks,
              std::list<rb::Track>& ytracks) const
  {
    // This is what we return
    std::vector<rb::Track> matchedTracks;

    typedef std::list<rb::Track>::iterator it_t;

    // Keep going until we stop being able to match anything
    while(true){
      double bestScore = 1e10;
      it_t bestX = xtracks.end();
      it_t bestY = ytracks.end();
      it_t bestJoin = xtracks.end(); // Picking this end() is just convention

      for(it_t xTrk = xtracks.begin(); xTrk != xtracks.end(); ++xTrk){
        assert(xTrk->View() == geo::kX);
        // No way to sensibly extend the trajectory for these right now
        if(xTrk->ExtentPlane() == 1) continue;

        for(it_t yTrk = ytracks.begin(); yTrk != ytracks.end(); ++yTrk){
          assert(yTrk->View() == geo::kY);
          if(yTrk->ExtentPlane() == 1) continue;

          // Simple merge, join in x, join in y
          for(int mode = 0; mode < 3; ++mode){
            // Don't try to join tracks within a view. DiscreteTracker got
            // better at this, and right now we're just making a mess when we
            // try.
            if(mode == 1 || mode == 2) continue;

            it_t joinTrk;
            if(mode == 0) joinTrk = xtracks.end();
            if(mode == 1){joinTrk = xTrk; ++joinTrk;}
            if(mode == 2){joinTrk = yTrk; ++joinTrk;}

            // For simple mode, go once with joinTrk invalid. For other modes,
            // loop through all joinable tracks in the right view.
            for(; mode == 0 ||
                  (mode == 1 && joinTrk != xtracks.end()) ||
                  (mode == 2 && joinTrk != ytracks.end()); ++joinTrk){
              // This could be done smarter if we interfaced with
              // DiscreteTracker properly. It has a lot more information about
              // these tracks that we could use.
              double score;
              if(mode == 0) score = ScoreForExtension(sliceMap, *xTrk, *yTrk);
              if(mode == 1) score = ScoreForJoinPlusExtension(sliceMap, *yTrk,
                                                              *xTrk, *joinTrk);
              if(mode == 2) score = ScoreForJoinPlusExtension(sliceMap, *xTrk,
                                                              *yTrk, *joinTrk);

              if(score < bestScore){
                bestScore = score;
                bestX = xTrk;
                bestY = yTrk;
                bestJoin = joinTrk;
              }

              if(mode == 0) break;
            } // end for joinTrk
          } // end for mode
        } // end for yTrk
      } // end for xTrk

      // No matches left
      if(bestX == xtracks.end()) break;

      LOG_DEBUG("ViewMerger") << "Best score: " << bestScore << std::endl;

      if(bestJoin != xtracks.end()){
        if(bestJoin->View() == geo::kX)
          matchedTracks.push_back(ZipTracks(JoinTracks(*bestX, *bestJoin),
                                            *bestY));
        else
          matchedTracks.push_back(ZipTracks(*bestX,
                                            JoinTracks(*bestY, *bestJoin)));
      }
      else{
        matchedTracks.push_back(ZipTracks(*bestX, *bestY));
      }

      // Can't use either 2D track again
      xtracks.erase(bestX);
      ytracks.erase(bestY);
      if(bestJoin != xtracks.end()){
        if(bestJoin->View() == geo::kX)
          xtracks.erase(bestJoin);
        else
          ytracks.erase(bestJoin);
      }
    } // end while

    return matchedTracks;
  }

  //......................................................................
  std::vector<rb::Track> ViewMerger::
  MatchVerticals(std::list<rb::Track>& xtracks,
                 std::list<rb::Track>& ytracks) const
  {
    std::vector<rb::Track> ret;

    typedef std::list<rb::Track>::iterator it_t;

    bool progress = true;
    while(progress){
      progress = false;

      for(it_t xTrk = xtracks.begin(); xTrk != xtracks.end(); ++xTrk){
        assert(xTrk->View() == geo::kX);
        if(xTrk->ExtentPlane() != 1) continue;

        for(it_t yTrk = ytracks.begin(); yTrk != ytracks.end(); ++yTrk){
          assert(yTrk->View() == geo::kY);
          if(yTrk->ExtentPlane() != 1) continue;

          if(abs(xTrk->MinPlane()-yTrk->MinPlane()) != 1) continue;

          ret.push_back(ZipVerticalTracks(*xTrk, *yTrk));
          xtracks.erase(xTrk);
          ytracks.erase(yTrk);
          progress = true;
          break;
        } // end for yTrk
        if(progress) break;
      } // end for xTrk
    } // end while

    return ret;
  }

  //......................................................................
  rb::Track ViewMerger::ZipVerticalTracks(const rb::Track& xtrk,
                                          const rb::Track& ytrk) const
  {
    art::ServiceHandle<geo::Geometry> geom;

    double xyz[3], dxyz[3];
    geom->CellInfo(xtrk.MinPlane(), xtrk.MinCell(geo::kX), 0, xyz, dxyz);
    const double x0 = xyz[0];
    const double xz0 = xyz[2]-dxyz[2];
    geom->CellInfo(xtrk.MinPlane(), xtrk.MaxCell(geo::kX), 0, xyz, dxyz);
    const double x1 = xyz[0];
    const double xz1 = xyz[2]+dxyz[2];

    geom->CellInfo(ytrk.MinPlane(), ytrk.MinCell(geo::kY), 0, xyz, dxyz);
    const double y0 = xyz[1];
    const double yz0 = xyz[2]-dxyz[2];
    geom->CellInfo(ytrk.MinPlane(), ytrk.MaxCell(geo::kY), 0, xyz, dxyz);
    const double y1 = xyz[1];
    const double yz1 = xyz[2]+dxyz[2];

    // If we had any idea where the vertex was we could potentially swap x0, x1
    // or y0, y1.

    const double dxdz = (x1-x0)/(xz1-xz0);
    const double dydz = (y1-y0)/(yz1-yz0);

    TVector3 start, stop;

    if(xz0+xz1 < yz0+yz1){
      // x is upstream of y
      start.SetX(x0);
      start.SetY(y0+dydz*(xz0-yz0));
      start.SetZ(xz0);

      stop.SetX(x0+dxdz*(yz1-xz0));
      stop.SetY(y1);
      stop.SetZ(yz1);
    }
    else{
      // y is upstream of x
      start.SetX(x0+dxdz*(yz0-xz0));
      start.SetY(y0);
      start.SetZ(yz0);

      stop.SetX(x1);
      stop.SetY(y1+dydz*(xz1-yz1));
      stop.SetZ(xz1);
    }

    rb::Track ret(xtrk.AllCells(), start, stop-start);
    ret.Add(ytrk.AllCells());
    ret.AppendTrajectoryPoint(stop);

    return ret;
  }

  //......................................................................
  void ViewMerger::TotalChargePerView(const rb::Track& zip,
                                      double& totx, double& toty) const
  {
    // TODO: accounting for uncalibrated hits and dead channels

    totx = 0;
    toty = 0;

    for(unsigned int n = 0; n < zip.NCell(); ++n){
      const art::Ptr<rb::CellHit> chit = zip.Cell(n);
      const rb::RecoHit rhit = zip.RecoHit(chit);
      if(rhit.IsCalibrated()){
        // Estimate track direction here
        const double eps = 1e-3;
        double x0, y0, x1, y1;
        zip.InterpolateXY(rhit.Z(), x0, y0);
        zip.InterpolateXY(rhit.Z()+eps, x1, y1);
        const double dz = eps/util::pythag(eps, x1-x0, y1-y0);

        if(chit->View() == geo::kX)
          totx += rhit.GeV()*dz;
        else
          toty += rhit.GeV()*dz;
      }
    }
  }

  //......................................................................
  double ViewMerger::MissingChargeByExtension(const rb::HitMap& slice,
                                              const rb::Track& xTrk,
                                              const rb::Track& yTrk) const
  {
    // UnexplainedByJoinPlusExtension doesn't know any better. Easier to fix it
    // here.
    if(xTrk.View() == geo::kY){
      assert(yTrk.View() == geo::kX);
      return MissingChargeByExtension(slice, yTrk, xTrk);
    }

    const rb::Track& zip = ZipTracks(xTrk, yTrk);

    const int unexpX =
      CountUnexplainedOnLine(slice, zip.Start(), xTrk.Start(), geo::kX)+
      CountUnexplainedOnLine(slice, zip.Stop(),  xTrk.Stop(),  geo::kX);

    const int unexpY =
      CountUnexplainedOnLine(slice, zip.Start(), yTrk.Start(), geo::kY)+
      CountUnexplainedOnLine(slice, zip.Stop(),  yTrk.Stop(),  geo::kY);

    // End up passing through too many live cells that were unhit
    if(unexpX+unexpY > 4) return 1e12;

    double xdens, ydens;
    TotalChargePerView(zip, xdens, ydens);
    xdens /= xTrk.NCell();
    ydens /= yTrk.NCell();

    return unexpX*xdens+unexpY*ydens;
  }

  //......................................................................
  double ViewMerger::ScoreForExtension(const rb::HitMap& sliceMap,
                                       const rb::Track& xTrk,
                                       const rb::Track& yTrk) const
  {
    // No overlap: MatchScore requires that anyway
    if(xTrk.MinPlane() > yTrk.MaxPlane()) return 1e12;
    if(yTrk.MinPlane() > xTrk.MaxPlane()) return 1e12;

    const double miss = MissingChargeByExtension(sliceMap, xTrk, yTrk);

    // End up passing through too many live cells that were unhit
    if(miss > 5e11) return 1e12;

    // Disable charge difference component for now.
    //    const double diff = ChargeDifferenceBetweenViews(ZipTracks(xTrk, yTrk));

    const double score = MatchScore(miss/*+diff*/, xTrk, yTrk);

    LOG_DEBUG("ViewMerger") << "Missing: " << miss
                            << ", score = " << score;
    //                            << ", charge difference " << diff << std::endl;

    return score;
  }

  //......................................................................
  double ViewMerger::ScoreForJoinPlusExtension(const rb::HitMap& sliceMap,
                                               const rb::Track& single,
                                               const rb::Track& parta,
                                               const rb::Track& partb) const
  {
    // Must be in this order, must not overlap
    if(partb.MinPlane() < parta.MaxPlane()) return 1e12;

    // No overlap: MatchScore requires that anyway
    if(parta.MinPlane() > single.MaxPlane()) return 1e12;
    if(single.MinPlane() > partb.MaxPlane()) return 1e12;

    const rb::Track& join = JoinTracks(parta, partb);

    const rb::Track& zip = ZipTracks(single, join);
    double totx, toty;
    TotalChargePerView(zip, totx, toty);
    const double avgDens = (totx+toty)/zip.NCell();

    // Don't allow join to use existing hits, only extension
    const int unexpJoin = CountUnexplainedOnLine(rb::HitMap(),
                                                 parta.Stop(), partb.Start(),
                                                 join.View());
    // Join shouldn't go through any unhit cells or others' cells. This
    // provides some protection against crazy merges.
    if(unexpJoin > 0) return 1e12;

    // Charge missing at the ends
    const double miss = MissingChargeByExtension(sliceMap, single, join);

    // End up passing through too many live cells that were unhit
    if(miss > 4*avgDens) return 1e12;

    const double diff = ChargeDifferenceBetweenViews(ZipTracks(single, join));

    // Small penalty for having a join
    const double score = MatchScore(miss+avgDens+diff, single, join);

    LOG_DEBUG("ViewMerger") << "Join: missing charge " << miss
                            << ", score = " << score
                            << ", charge difference: " << diff << std::endl;

    return score;
  }

  //......................................................................
  rb::Track ViewMerger::JoinTracks(const rb::Track& parta,
                                   const rb::Track& partb) const
  {
    assert(parta.View() == partb.View());

    // Must be in this order, must not overlap
    assert(partb.MinPlane() >= parta.MaxPlane());

    // This is pretty crude. Could do a neater job with tighter interface into
    // the 2D tracker.

    rb::Track join(parta);
    // Right now, no two tracks can share hits, so this is safe. Might not be
    // so in future, would have to remove duplicates.
    join.Add(partb.AllCells());
    for(unsigned int n = 0; n < partb.NTrajectoryPoints(); ++n){
      join.AppendTrajectoryPoint(partb.TrajectoryPoint(n));
    }

    return join;
  }

  //......................................................................
  int ViewMerger::CountUnexplainedOnLine(const rb::HitMap& hmap,
                                         TVector3 start, TVector3 end,
                                         geo::View_t view) const
  {
    // TODO: we need to penalize the case where the track now runs outside the
    // detector for some of its length. This is actually leading to
    // mismerges...
    art::ServiceHandle<geo::Geometry> geom;
    art::ServiceHandle<chaninfo::BadChanList> bcl;

    std::vector<geo::OfflineChan> xcells, ycells;
    geom->CountCellsOnLineFast(start.X(), start.Y(), start.Z(),
                               end.X(), end.Y(), end.Z(),
                               xcells, ycells);

    int ret = 0;
    if(view != geo::kY){ // X or 3D
      for(unsigned int n = 0; n < xcells.size(); ++n){
        const int plane = xcells[n].Plane();
        const int cell = xcells[n].Cell();
        if(!hmap.CellExists(plane, cell) &&
           !bcl->IsBad(plane, cell) &&
           !bcl->IsLowEfficiency(plane, cell)) ++ret;
      }
    }
    if(view != geo::kX){ // Y or 3D
      for(unsigned int n = 0; n < ycells.size(); ++n){
        const int plane = ycells[n].Plane();
        const int cell = ycells[n].Cell();
        if(!hmap.CellExists(plane, cell) &&
           !bcl->IsBad(plane, cell) &&
           !bcl->IsLowEfficiency(plane, cell)) ++ret;
      }
    }

    return ret;
  }

  //......................................................................
  double ViewMerger::ChargeDifferenceBetweenViews(const rb::Track& zip) const
  {
    double totx, toty;
    TotalChargePerView(zip, totx, toty);

    return fabs(totx-toty);
  }

  //......................................................................
  void ViewMerger::MaybeExtendToVertex(const rb::Vertex* vtx,
                                       std::vector<rb::Track>& tracks) const
  {
    rb::HitMap hmap;
    for(unsigned int n = 0; n < tracks.size(); ++n) hmap.Add(&tracks[n]);

    for(unsigned int n = 0; n < tracks.size(); ++n){
      // Only if we don't pass through a single non-dead cell
      if(CountUnexplainedOnLine(hmap, tracks[n].Start(), vtx->GetXYZ(),
                                tracks[n].View()) == 0){

        TVector3 xyz = vtx->GetXYZ();
        if(!tracks[n].Is3D()) xyz[1-tracks[n].View()] = 0;

        // And only if the dot-product is reasonable
        if((tracks[n].Start()-xyz).Unit().Dot(tracks[n].Dir()) < .94) continue;

        tracks[n].PrependTrajectoryPoint(xyz);
        if(tracks[n].Is3D()){
          // Overwrites the point we just prepended
          tracks[n].SetStart(vtx->GetXYZ());
        }
        else{
          // Overwrites the point we just prepended
          tracks[n].SetStart(xyz[tracks[n].View()], xyz.Z());
        }
      }
    } // end for n
  }

  //......................................................................
  double ViewMerger::MatchScore(double miss,
                                const rb::Track& aTrk,
                                const rb::Track& bTrk) const
  {
    // These casts to int are absolutely necessary :-@
    const int dStart = abs(int(aTrk.MinPlane())-int(bTrk.MinPlane()));
    assert(dStart > 0);
    const int dEnd = abs(int(aTrk.MaxPlane())-int(bTrk.MaxPlane()));
    assert(dEnd > 0);
    const int totL = aTrk.ExtentPlane()+bTrk.ExtentPlane();
    const int overlap = totL-dStart-dEnd;
    if(overlap <= 0) return 1e12;

    // Add a small amount to miss so we can distinguish different zeroes
    return (miss+1e-3)/overlap;
  }

  //......................................................................
  rb::Track ViewMerger::ZipTracks(const rb::Track& xTrk,
                                  const rb::Track& yTrk) const
  {
    return xTrk.ZipWith(yTrk);

    // TODO: restore protection against going up to a plane too far. Can just
    // truncate that returned track.

    /*
    if(xTrk.View() == geo::kY){
      assert(yTrk.View() == geo::kX);
      return ZipTracks(yTrk, xTrk);
    }

    art::ServiceHandle<geo::Geometry> geom;

    std::vector<TVector3> trajPts;

    double minZ = std::min(xTrk.Start().Z(), yTrk.Start().Z());
    double maxZ = std::max(xTrk.Stop().Z(), yTrk.Stop().Z());

    const int minPlane = std::min(xTrk.MinPlane(), yTrk.MinPlane());
    const int maxPlane = std::max(xTrk.MaxPlane(), yTrk.MaxPlane());

    // 2D tracks often stick halfway into the next plane. Don't allow the final
    // track to go further than we have any evidence of hits. In that case put
    // it to within 1mm of the end of the last plane with hits.
    double xyz[3], dxyz[3];
    geom->CellInfo(minPlane, 0, 0, xyz, dxyz);
    minZ = std::max(minZ, xyz[2]-dxyz[2]+.1);

    geom->CellInfo(maxPlane, 0, 0, xyz, dxyz);
    maxZ = std::min(maxZ, xyz[2]+dxyz[2]-.1);

    const double kZStep = 5; // cm
    const int stepMax = int((maxZ-minZ)/kZStep);
    for(int step = 0; step <= stepMax; ++step){
      const double z = minZ+(double(step)/stepMax)*(maxZ-minZ);

      double xyz[3];
      xyz[2] = z;

      double junk;
      xTrk.InterpolateXY(z, xyz[0], junk);
      yTrk.InterpolateXY(z, junk, xyz[1]);

      trajPts.push_back(xyz);
    } // end for step

    rb::Track ret(art::PtrVector<rb::CellHit>(),
                  trajPts[0], trajPts[1]-trajPts[0]);
    ret.Add(xTrk.AllCells());
    ret.Add(yTrk.AllCells());

    for(unsigned int n = 1; n < trajPts.size(); ++n)
      ret.AppendTrajectoryPoint(trajPts[n]);

    return ret;
    */
  }

  //......................................................................
  void ViewMerger::Extremes(const rb::Track& trk,
                            std::map<int, int>& mins,
                            std::map<int, int>& maxs) const
  {
    for(unsigned int n = 0; n < trk.NCell(); ++n){
      const int p = trk.Cell(n)->Plane();
      const int c = trk.Cell(n)->Cell();

      if(c > maxs[p]) maxs[p] = c;
      // Careful with zero default value
      if(mins.find(p) == mins.end() || c < mins[p]) mins[p] = c;
    }
  }

  //......................................................................
  bool ViewMerger::Adjacent(const rb::Track& a, const rb::Track& b) const
  {
    assert(a.View() == b.View());
    assert(a.Is2D());

    if(a.ExtentPlane() < b.ExtentPlane()) return Adjacent(b, a);
    // Guaranteed b is the shorter one

    // Don't attach verticals
    if(b.ExtentPlane() == 1) return false;

    std::map<int, int> amins, amaxs, bmins, bmaxs;
    Extremes(a, amins, amaxs);
    Extremes(b, bmins, bmaxs);

    // Maybe a is all below b
    bool all = true;
    for(std::map<int, int>::iterator it = bmins.begin(); it != bmins.end(); ++it){
      if(amaxs[it->first] != it->second-1){
        all = false;
        break;
      }
    }
    if(all) return true;

    // Or maybe a is all above b
    for(std::map<int, int>::iterator it = bmaxs.begin(); it != bmaxs.end(); ++it){
      if(amins[it->first] != it->second+1) return false;
    }
    return true;
  }

  //......................................................................
  void ViewMerger::MergeAdjacentTracks(std::list<rb::Track>& trks) const
  {
    for(std::list<rb::Track>::iterator iti = trks.begin(); iti != trks.end(); ++iti){
      for(std::list<rb::Track>::iterator itj = trks.begin(); itj != trks.end(); ++itj){
        if(iti == itj) continue;

        if(iti->ExtentPlane() < itj->ExtentPlane()) continue;

        if(Adjacent(*iti, *itj)){
          LOG_DEBUG("ViewMerger") << "Adjacent: " << *iti << " and " << *itj << std::endl;

          iti->Add(itj->AllCells());
          trks.erase(itj);
          MergeAdjacentTracks(trks); // Start again from the top
          return;
        }
      } // end for itj
    } // end for iti
  }

  // Links against the version in DiscreteTracker
  bool CompareByLengthVM(const rb::Track& a, const rb::Track& b)
  {
    return a.TotalLength() < b.TotalLength();
  }

  //......................................................................
  rb::Vertex* ViewMerger::FindVertex(std::vector<rb::Track> trks) const
  {
    if(trks.empty()) return 0;

    // Longest tracks first. Only used for fallback at the end
    std::sort(trks.rbegin(), trks.rend(), CompareByLengthVM);

    double bestVx, bestVy, bestVz;
    // I don't believe there's a path that doesn't set these, but need this to
    // shut the compiler up in optimized builds.
    bestVx = bestVy = bestVz = 0;

    double bestScore = 1e10;
    for(unsigned int idx1 = 0; idx1 < trks.size()-1; ++idx1){
      if(trks[idx1].View() != geo::kXorY) continue;
      if(trks[idx1].ExtentPlane() < 3) continue;
      for(bool end1: {false, true}){
        for(unsigned int idx2 = idx1+1; idx2 < trks.size(); ++idx2){
          if(trks[idx2].View() != geo::kXorY) continue;
          if(trks[idx2].ExtentPlane() < 3) continue;
          for(bool end2: {false, true}){

            // Start and direction defining point of these two lines
            const TVector3 s1 = end1 ? trks[idx1].Stop() : trks[idx1].Start();
            const TVector3 s2 = end2 ? trks[idx2].Stop() : trks[idx2].Start();
            const TVector3 d1 = trks[idx1].TrajectoryPoint(end1 ? trks[idx1].NTrajectoryPoints()-2 : 1);
            const TVector3 d2 = trks[idx2].TrajectoryPoint(end2 ? trks[idx2].NTrajectoryPoints()-2 : 1);

            double vx, vy, vz1, vz2;
            bool ok = geo::LineIntersection(s1.X(), s1.Z(), d1.X(), d1.Z(),
                                            s2.X(), s2.Z(), d2.X(), d2.Z(),
                                            vx, vz1);
            if(!ok) continue;
            ok = geo::LineIntersection(s1.Y(), s1.Z(), d1.Y(), d1.Z(),
                                       s2.Y(), s2.Z(), d2.Y(), d2.Z(),
                                       vy, vz2);
            if(!ok) continue;
            const double vz = (vz1+vz2)/2;

            const double dist1 = util::pythag(s1.X()-vx, s1.Y()-vy, s1.Z()-vz);
            const double dist2 = util::pythag(s2.X()-vx, s2.Y()-vy, s2.Z()-vz);

            // Favour both long tracks and hits close to the ends
            const double score = dist1/trks[idx1].TotalLength()+dist2/trks[idx2].TotalLength();
            if(score < bestScore){
              bestScore = score;
              bestVx = vx;
              bestVy = vy;
              bestVz = vz;
            } // end if better
          } // end for end2
        } // end for idx2
      } // end for end1
    } // end for idx1
    // Too far from the ends, fallback
    if(bestScore > 6){
      bestVx = trks[0].Start().X();
      bestVy = trks[0].Start().Y();
      bestVz = trks[0].Start().Z();
    }

    // Time is an afterthought
    return new rb::Vertex(bestVx, bestVy, bestVz, trks[0].MinTNS());
  }

  //......................................................................
  rb::Track ViewMerger::FakeThirdDimension(const rb::Track& trk,
                                           const std::vector<rb::Track>& merged) const
  {
    assert(!trk.Is3D());

    const double z = trk.Start().Z();

    const rb::Track* best = 0;

    // The best source for W information is the longest overlapping track
    for(unsigned int n = 0; n < merged.size(); ++n){
      if(merged[n].Start().Z()-5 < z &&
         merged[n].Stop().Z()+5 > z){
        if(!best || merged[n].ExtentPlane() > best->ExtentPlane())
          best = &merged[n];
      }
    }

    // Failing that, just the longest track
    if(!best){
      for(unsigned int n = 0; n < merged.size(); ++n){
        if(!best || merged[n].ExtentPlane() > best->ExtentPlane())
          best = &merged[n];
      }
    }

    const geo::View_t view = trk.View();

    double w = 0;
    if(best){
      double x, y;
      best->InterpolateXY(z, x, y);
      w = (view == geo::kX) ? y : x;
    }

    // Fill in W information everywhere

    TVector3 start = trk.Start();
    start[1-view] = w;

    rb::Track ret(trk.AllCells(), start, trk.Dir());

    for(unsigned int n = 1; n < trk.NTrajectoryPoints(); ++n){
      TVector3 pt = trk.TrajectoryPoint(n);
      pt[1-view] = w;
      ret.AppendTrajectoryPoint(pt);
    }

    return ret;
  }

  DEFINE_ART_MODULE(ViewMerger);

} //namespace dt
////////////////////////////////////////////////////////////////////////

