///
/// \file    GridSearch.cxx
/// \brief   Produce a list of vertex and prong seeds for ElasticArms
/// \version $Id: GridSearch.cxx,v 1.7 2012-10-04 21:23:23 messier Exp $
/// \author  messier@indiana.edu
///
#include "ElasticArms/art/GridSearch.h"
#include "ElasticArms/func/ElasticArms.h"

#include <iostream>
#include <cmath>
#include <algorithm>

#include "RecoBase/HoughResult.h"

namespace earms {

  //......................................................................

  GridSearch::GridSearch(double xmn, double xmx,
			 double ymn, double ymx,
			 double zmn, double zmx,
			 double colinear,
			 unsigned int nseeddir) :
    fXmin(xmn),
    fXmax(xmx),
    fYmin(ymn),
    fYmax(ymx),
    fZmin(zmn),
    fZmax(zmx),
    fColinear(colinear),
    fNseedDir(nseeddir)
  {  }

  //......................................................................

  void GridSearch::AddVertex(double x, double y, double z) 
  {
    if (x<fXmin) return;
    if (y<fYmin) return;
    if (z<fZmin) return;
    if (x>fXmax) return;
    if (y>fYmax) return;
    if (z>fZmax) return;
    fX.push_back(x);
    fY.push_back(y);
    fZ.push_back(z);
  }

  //......................................................................

  void GridSearch::AddVtxPoints(const ElasticArms& arms,
				const std::vector<double>& mx,
				const std::vector<double>& bx,
				const std::vector<double>& my,
				const std::vector<double>& by,
				const std::vector<double>& f)
  {
    unsigned register int i, j;
    std::vector< std::pair<double, double> > xz;
    std::vector< std::pair<double, double> > yz;
  
    //
    // Sort the xz and yz hits by z
    //
    for (i=0; i<arms.fN; ++i) {
      if (arms.fView[i]==0) {
	xz.push_back(std::pair<double,double>(arms.fZ[i], arms.fXorY[i]));
      }
      else if (arms.fView[i]==1) {
	yz.push_back(std::pair<double,double>(arms.fZ[i], arms.fXorY[i]));
      }
      else abort();
    }
    sort(xz.begin(), xz.end());
    sort(yz.begin(), yz.end());
  
    //
    // Grab points from the specified locations in the hit lists
    //
    unsigned int ix;
    unsigned int iy;
    double xseed, zxseed;
    double yseed, zyseed;
    for (i=0; i<f.size(); ++i) {
      ix = fabs(f[i])*xz.size()/100;
      iy = fabs(f[i])*yz.size()/100;

      
      //
      // In forward direction, pick the hits
      //
      if (f[i]>0.0) {
	zxseed = xz[ix].first;
	xseed  = xz[ix].second;
	zyseed = yz[iy].first;
	yseed  = yz[iy].second;
      }
      else {
	//
	// In backward direction, pick mirror images of the hits
	// reflected across the boundary of the first hit
	//
	zxseed = 2.0*xz[0].first -xz[ix].first;
	xseed  = 2.0*xz[0].second-xz[ix].second;
	zyseed = 2.0*yz[0].first -yz[iy].first;
	yseed  = 2.0*yz[0].second-yz[iy].second;
      }
      
      //
      // A point on the line in the yz view
      //
      for (j=0; j<my.size(); ++j) {
	this->AddVertex(xseed, my[j]*zxseed+by[j], zxseed);
      }
      //
      // A point on the line in the xz view
      //
      for (j=0; j<mx.size(); ++j) {
	this->AddVertex(mx[j]*zyseed+bx[j], yseed, zyseed);
      }
      //
      // A point that takes information just from the hits
      //
      this->AddVertex(xseed, yseed, 0.5*(zxseed+zyseed));
    }
  }

  //......................................................................

  void GridSearch::AddDirection(double theta, double phi)
  {
    fTheta.push_back(theta);
    fPhi.  push_back(phi);
  }

  //......................................................................

  void GridSearch::AddHoughIntersections(const rb::HoughResult& hrx,
					 const rb::HoughResult& hry,
					 unsigned int nmx,
					 unsigned int nmatch,
					 bool flipped) 
  {
    //
    // For there to be any intersections we need at minimum of two lines
    // in x and one in y
    //
    if ((hrx.fPeak.size()<2) && (hry.fPeak.size()<1)){
      //try flipping before we exit
      if (flipped==false) {
        this->AddHoughIntersections(hry,hrx,nmx,nmatch,true);
      }
      return;
    }
  
    //
    // Only use the top set of results
    //
    unsigned int nx = hrx.fPeak.size();
    if (nx>nmx) nx = nmx;
  
    unsigned register int i, j, k;
    //
    // Loop over all the intersections in the x view. Grab y location
    // from the dominant line in that view.
    //
    double x, y, z;
    double my, by;
    double dot;

    unsigned int ny = nmatch;
    if (hry.fPeak.size()<ny) ny = hry.fPeak.size();
  
    for (i=0; i<nx; ++i) {
      double mi, bi;
      hrx.SlopeIcept(i, &mi, &bi);
      for (j=i+1; j<nx; ++j) {
	double mj, bj;
	hrx.SlopeIcept(j, &mj, &bj);
      
	// Skip lines paired with themselves
	if (mi==mj) continue;
      
	// Also skip lines that are nearly colinear
	dot = fabs((mi*mj+1.0)/sqrt(mi*mi+1.0)/sqrt(mj*mj+1.0));
	if (dot>fColinear) continue;
      
	z = (bj-bi)/(mi-mj);
	x = 0.5*((mi+mj)*z+(bi+bj));

      
	//
	// What y value to assign to this x-z pair? Check the top
	// nmatch hough lines in the other view.
	//
	for (k=0; k<ny; ++k) {
	  hry.SlopeIcept(k, &my, &by);
	
	  y = my*z + by;

	
	  if (flipped) this->AddVertex(y,x,z);
	  else         this->AddVertex(x,y,z);
	}
      
      } // loop on j hough lines in x view
    } // loop on i hough lines x view

    //
    // Repeat exchanging the meaning of x and y
    //
    if (flipped==false) {
      this->AddHoughIntersections(hry,hrx,nmx,nmatch,true);
    }
  }

  //......................................................................

  ///
  /// Set a standard set of directions uniformly distributed on a
  /// sphere using the "Golden Section spiral"
  /// (http://www.xsi-blog.com/archives/115)
  ///
  void GridSearch::SetStandardDirections() 
  {
    static const double nd  = fNseedDir;
    static const double inc = M_PI*(3.0-sqrt(5.0));
    static const double off = 2.0/nd;
    
    unsigned int i;
    double x, y, z;
    double r, phi;
    for (i=0; i<fNseedDir; ++i) {
      y = (i+0.5)*off-1.0;
      r = sqrt(1.0-y*y);
      phi = i*inc;
      x = r*cos(phi);
      z = r*sin(phi);

      fTheta.push_back(acos(z/sqrt(x*x+y*y+z*z)));
      fPhi.  push_back(atan2(y,x));
    }
  }

  //......................................................................

  double GridSearch::ChooseBestProngDir(ElasticArms& arms, unsigned int a) 
  {
    unsigned register int b;
    unsigned int j, jbest=0;
    double E = 99E99;
    double Ebest = 99E99;

    //
    // Loop over all directions in the table
    //
    for (j=0; j<fTheta.size(); ++j) {
      arms.SetArm(a, fTheta[j], fPhi[j]);
      //
      // Check this direction against arms which have already been set.
      // If it's too close to an existing arm, skip it.
      //
      double dot;
      double maxdot = -1.0;
      double mindot =  1.0;
      for (b=0; b<a; ++b) {
	dot =
	  arms.fdXds[a]*arms.fdXds[b]+
	  arms.fdYds[a]*arms.fdYds[b]+
	  arms.fdZds[a]*arms.fdZds[b];
	if (dot>maxdot) maxdot = dot;
	if (dot<mindot) mindot = dot;
      }
      if (maxdot> fColinear) continue;
      if (mindot<-fColinear) continue;
    
      //
      // Test the goodness of this prong. Remember it if its the best so
      // far.
      //
      arms.UpdateMia();
      arms.UpdateVia();
      E = arms.E();
      if (E<Ebest) {
	jbest = j;
	Ebest = E;
      }
    }
    arms.SetArm(a, fTheta[jbest], fPhi[jbest]);
    return Ebest;
  }

  //......................................................................

  double GridSearch::FindBestVtx(ElasticArms& arms)
  {
    unsigned register int i, j;
    double E = 99E99;
    double Ebest = 99E99;
    double xbest=0;
    double ybest=0;
    double zbest=0;
    for (i=0; i<fX.size(); ++i) {
      arms.SetVertex(fX[i], fY[i], fZ[i]);
      for (j=0; j<arms.fM; ++j) {
	arms.SetArm(j, 0.5*M_PI, 2.0*M_PI*(float)j/(float)arms.fM);
      }
      for (j=0; j<arms.fM; ++j){
        E = this->ChooseBestProngDir(arms, j);
      }
      if (E<Ebest) {
	Ebest = E;
	xbest = fX[i];
	ybest = fY[i];
	zbest = fZ[i];
      }
    }
    //
    // Leave the vertex and arm directions in their best configuration
    // 
    arms.SetVertex(xbest, ybest, zbest);
    for (j=0; j<arms.fM; ++j) {
      arms.SetArm(j, 0.5*M_PI, 2.0*M_PI*(float)j/(float)arms.fM);
    }
    for (j=0; j<arms.fM; ++j) E = this->ChooseBestProngDir(arms, j);
    return E;
  }
} // end namespace earms
////////////////////////////////////////////////////////////////////////
