///
/// \file    ElasticArms.cxx
/// \brief   The basic elastic arms energy function
/// \author  messier@indiana.edu
/// \version $Id: ElasticArms.cxx,v 1.5 2012-05-31 01:08:49 messier Exp $
///
#include "ElasticArms/func/ElasticArms.h"
#include "GeometryObjects/PlaneGeo.h"
#include <iostream>
#include <cmath>
#include <cstdlib>
#include <cassert>
using namespace earms;

//
// These math utils are included in the NOvA software but they are so
// small its easier to just define them then link to all the libraries
// required to link to the existing versions.
//
static double sqr(double x) { return x*x; };
static double dsqrtoline(double x0, double y0,
			 double x1, double y1,
			 double x2, double y2) 
{
  assert(x1!=x2 || y1!=y2);
  double dx = x2-x1;
  double dy = y2-y1;
  return sqr(dx*(y1-y0)-dy*(x1-x0)) / (dx*dx+dy*dy);
}

//......................................................................

ElasticArms::ElasticArms(unsigned int n, 
			 unsigned int m,
			 double       l,
			 double       lv) :
  fProngFit(1),
  fT(200.0),
  fBeta(1.0/fT),
  fLambda(l),
  fLambdaV(lv),
  fN(n),
  fXorY(n),
  fView(n),
  fZ(n),
  fWsqr(n),
  fDhi(n),
  fDaa(m),
  fMia(n),
  fExpBetaMia(n),
  fVia(n),
  fX0(0),
  fY0(0),
  fZ0(0),
  fM(m),
  fdXds(m),
  fdYds(m),
  fdZds(m)
{
  unsigned register int i;
  for (i=0; i<n; ++i) {
    fMia[i].       resize(m);
    fExpBetaMia[i].resize(m);
    fVia[i].       resize(m);
  }
}

//......................................................................

void ElasticArms::SetHit(unsigned int i,
			 double       z,
			 double       xory, 
			 double       sigma,
			 int          view) 
{
  assert(i<fN);
  fZ[i]    = z;
  fXorY[i] = xory;
  fWsqr[i] = 1.0/(sigma*sigma);
  fView[i] = view;
}

//......................................................................

void ElasticArms::SetT(double t) 
{
  fT    = t;
  fBeta = 1.0/fT;
  fExpBetaLambda = exp(-fBeta*fLambda);
}

//......................................................................

void ElasticArms::SetLambda(double lam) 
{
  fLambda = lam;
  fExpBetaLambda = exp(-fBeta*fLambda);
}

//......................................................................

void ElasticArms::UpdateMia() 
{
  unsigned register int i; // hit index
  unsigned register int a; // track index
  //
  // Loop for each track a of the set of M track templates
  //
  for (a=0; a<fM; ++a) {
  
    // Update track end point
    double trkX1 = fX0 + fdXds[a];
    double trkY1 = fY0 + fdYds[a];
    double trkZ1 = fZ0 + fdZds[a];
    
    //
    // Loop for each hit i of the set of N hits
    //
    double trkXorY0;
    double trkXorY1;
    for (i=0; i<fN; ++i) {
      double dot;
      //
      // Set coordinates based on which view the hit is in
      //
      switch (fView[i]) {
      case geo::kX:
	trkXorY0 = fX0;
	trkXorY1 = trkX1;
	dot = (trkZ1-fZ0)*(fZ[i]-fZ0) + (trkXorY1-fX0)*(fXorY[i]-fX0);
	break;
      case geo::kY:
	trkXorY0 = fY0;
	trkXorY1 = trkY1;
	dot = (trkZ1-fZ0)*(fZ[i]-fZ0) + (trkXorY1-fY0)*(fXorY[i]-fY0);
	break;
      default: abort();
      }
      //
      // If we're doing a line fit or it the hit is in the forward
      // hemisphere use the distance from the hit to the arm template
      //
      if (fProngFit==0 || dot>0.0) {
	fMia[i][a] = fWsqr[i]*dsqrtoline(fZ[i], fXorY[i],
					 fZ0,   trkXorY0,
					 trkZ1, trkXorY1);
      }
      else {
	//
	// In the backwards hemisphere take the distance to the
	// vertex. Make it get worse rapidly for differences larger
	// than one sigma
	//
	fMia[i][a] = fWsqr[i]*(sqr(fZ[i]-fZ0)+sqr(fXorY[i]-trkXorY0));
	if (fMia[i][a]>1.0) fMia[i][a] *= fMia[i][a];
      }
      fExpBetaMia[i][a] = exp(-fBeta*fMia[i][a]);
    } // loop on i hits
  } // loop on a tracks
}

//......................................................................

void ElasticArms::UpdateVia() 
{
  static const double TINY = 1e-16;
  unsigned register int i;
  unsigned register int a;
  unsigned register int b;

  double d1, d2, denom;

  d1 = fExpBetaLambda+TINY;
  for (i=0; i<fN; ++i) {
    d2 = TINY;
    for (b=0; b<fM; ++b) d2 += fExpBetaMia[i][b];
    
    denom = d1+d2;
    for (a=0; a<fM; ++a) {
      fVia[i][a] = fExpBetaMia[i][a]/denom;
    }
  } // loop on i=0...N hits
}

//......................................................................

void ElasticArms::UpdateDaa() 
{
  unsigned register int i, a;
  double dminx, dminy;
  double mavex, mavey;
  double nx,    ny;

  for (a=0; a<fM; ++a) {
    //
    // Compute the average weight for hits on this track
    //
    nx = 0.0;
    ny = 0.0;
    mavex = 0.0;
    mavey = 0.0;
    for (i=0; i<fN; ++i) {
      switch (fView[i]) {
      case 0:  mavex += fVia[i][a]; nx += 1.0; break;
      case 1:  mavey += fVia[i][a]; ny += 1.0; break;
      default: abort();
      }
    }
    mavex /= nx;
    mavey /= ny;
  
    dminx = 99E99;
    dminy = 99E99;
    for (i=0; i<fN; ++i) {
      switch (fView[i]) {
      case 0:
	if (fVia[i][a]>=mavex && fDhi[i]<dminx) dminx = fDhi[i];
	break;
      case 1:
	if (fVia[i][a]>=mavey && fDhi[i]<dminy) dminy = fDhi[i];
	break;
      default: abort();
      }
    }
    fDaa[a] = sqrt(dminx)+sqrt(dminy);
  }
}

//......................................................................

double ElasticArms::E() 
{
  unsigned register int i;
  unsigned register int a;
  double sViaMia = 0.0; // First term in the energy
  double sVia    = 0.0; // The inside  sum in the second term
  double ssVia   = 0.0; // The outside sum in the second term
  
  //
  // Update all the pieces that go into the calculation
  //
  this->UpdateMia();
  this->UpdateVia();
  this->UpdateDaa();

  //
  // Loop for each hit i of N
  //
  for (i=0; i<fN; ++i) {
    //
    // Loop for each track a of M
    //
    sVia = 0.0;
    for (a=0; a<fM; ++a) {
      sViaMia += fVia[i][a]*fMia[i][a];
      sVia    += fVia[i][a];
    }
    ssVia += sqr(sVia-1.0);
  }

  //
  // Loop for each prong a of M
  //
  double sDaa = 0.0;
  for (a=0; a<fM; ++a) sDaa += fDaa[a];
  sDaa *= 2.0/fLambdaV;
  
  //
  // Three terms:
  //
  // 1st accounts for distance from hits to prongs
  // 2nd is a penalty term for hits labeled as noise
  // 3rd is a penalty term for arms that are detached from the vertex
  //
  return sViaMia + fLambda*ssVia + sDaa;
}

//......................................................................

void ElasticArms::SetArm(unsigned int a, double theta, double phi)
{
  assert(a<fM);
  fdXds[a] = sin(theta)*cos(phi);
  fdYds[a] = sin(theta)*sin(phi);
  fdZds[a] = cos(theta);
}

//......................................................................

void ElasticArms::SetVertex(double x, double y, double z)
{
  fX0 = x;
  fY0 = y;
  fZ0 = z;

  //
  // Update the table of distances from the vertex to the hits
  //
  unsigned register int i;
  for (i=0; i<fN; ++i) {
    switch (fView[i]) {
    case 0:
      fDhi[i] = sqr(fXorY[i]-fX0) + sqr(fZ[i]-fZ0);
      break;
    case 1:
      fDhi[i] = sqr(fXorY[i]-fY0) + sqr(fZ[i]-fZ0);
      break;
    default: abort();
    }
  }
}

//......................................................................

void ElasticArms::GetVertex(double& x, double& y, double& z) const
{
  x = fX0;
  y = fY0;
  z = fZ0;
}

////////////////////////////////////////////////////////////////////////
