///
/// \file    Minimize.cxx
/// \brief   A helper function to optimize the elastic arms
/// \author  messier@indiana.edu
/// \version $Id: Minimizer.cxx,v 1.2 2012-05-25 06:01:32 messier Exp $ 
///
///
#include "ElasticArms/func/Minimizer.h"
#include "ElasticArms/func/ElasticArms.h"
#include <cstdlib>
#include <cstdio>
#include <cmath>
#include <iostream>
#include "TMinuit.h"
using namespace earms;

//......................................................................

Minimizer::Minimizer(ElasticArms* arms) : 
  fArms(arms), 
  fNdim(2*arms->fM+3),
  fM(arms->fM),
  fVtxStep(10.0),
  fAngleStep(5.0*M_PI/180.0),
  fAnnealMax(50),
  fT0(0.1*0.1),
  fT1(6.0*6.0),
  fAnnealFrac(0.8)
{
  unsigned int i;
  double theta, phi;
  
  //
  // Copy the seed parameters to the fit parameters
  //
  this->SeedVertex(fArms->fX0,fArms->fY0,fArms->fZ0);
  for (i=0; i<fArms->fM; ++i) {
    theta = acos(fArms->fdZds[i]);
    phi   = atan2(fArms->fdYds[i], fArms->fdXds[i]);
    this->SeedArm(i,theta,phi);
  }
}

//......................................................................

void Minimizer::SetT0(double T0) { fT0 = T0; }

//......................................................................

void Minimizer::SetT1(double T1) { fT1 = T1; }

//......................................................................

void Minimizer::SetAnnealFrac(double f) { fAnnealFrac = f; }

//......................................................................

void Minimizer::SetAnnealMaxItr(unsigned int i) { fAnnealMax = i; }

//......................................................................

void Minimizer::SeedArm(unsigned int a, double th, double ph) 
{
  fP[2*a+3] = th;
  fP[2*a+4] = ph;
  fArms->SetArm(a, th, ph);
}

//......................................................................

void Minimizer::SeedVertex(double x, double y, double z)
{
  fP[0] = x;
  fP[1] = y;
  fP[2] = z;
  fArms->SetVertex(x, y, z);
}

//......................................................................

static ElasticArms* gsArms=0;
static void gsFCN(int& n, double*, double& f, double* p, int) 
{
  unsigned register int i;
  gsArms->SetVertex(p[0], p[1], p[2]);
  for (i=0; i<gsArms->fM; ++i) {
    gsArms->SetArm(i, p[2*i+3], p[2*i+4]);
  }
  f = gsArms->E();
}

//......................................................................

double Minimizer::Fit() 
{
  unsigned register int i;
  int ierflg;

  TMinuit mn;
  mn.SetPrintLevel(-99);
  
  //
  // Set the minimization function. Have to talk to it through global
  // static pointers...
  //
  gsArms = fArms;
  mn.SetFCN(gsFCN);
  
  //
  // Setup the fit parameters
  //
  mn.mnparm(0, "x0", fP[0], fVtxStep, 0, 0, ierflg);
  mn.mnparm(1, "y0", fP[1], fVtxStep, 0, 0, ierflg);
  mn.mnparm(2, "z0", fP[2], fVtxStep, 0, 0, ierflg);
  for (i=0; i<fArms->fM; ++i) {
    char theta[256];
    char phi[256];
    sprintf(theta,"th_%d",i);
    sprintf(phi,  "ph_%d",i);
    //
    // The limits for the theta range are straight forward. The limits
    // for the phi range are trickier since 2pi connects to 0 and we
    // would like the minimizer to be able to explore that wrap
    // around. I've given phi an extra pi radians on both sides.
    //
    unsigned int ith = 2*i+3;
    unsigned int iph = ith+1;
    mn.mnparm(ith, theta, fP[ith], fAngleStep,  0,      M_PI, ierflg);
    mn.mnparm(iph, phi,   fP[iph], fAngleStep, -M_PI, 3*M_PI, ierflg);
  }

  //
  // Initialize the elastic arms for 1st iteration
  //
  fArms->SetVertex(fP[0],fP[1],fP[2]);
  for (i=0; i<fM; ++i) fArms->SetArm(i, fP[2*i+3], fP[2*i+4]);
  fArms->UpdateMia();
  fArms->UpdateVia();
  fArms->UpdateDaa();
  
  //
  // Begin the annealing iterations
  //
  unsigned int itr;
  double t = fT1;
  for (itr=0; itr<fAnnealMax; ++itr) {
    if (t<fT0) t = fT0;
    fArms->SetT(t);

    //if (itr==0) mn.mnseek();
    mn.mnmigr();
    double th, ph, error;
    mn.GetParameter(3, th, error);
    mn.GetParameter(4, ph, error);
    /*if (itr > 0){
      if((fabs(fArms->fX0 - xold) > 50.0) || (fabs(fArms->fY0 - yold) > 50.0) || (fabs(fArms->fZ0 - zold) > 50.0)){
        fArms->SetVertex(xold,yold,zold);
        return fArms->E();
      }
    }*/

    t *= fAnnealFrac;
    if (t<=fT0) break;
  }
  return fArms->E();
}

////////////////////////////////////////////////////////////////////////
