#include "ElasticArms/func/ElasticArms.h"
#include "ElasticArms/func/Minimizer.h"
#include "TFile.h"
#include "TH2F.h"
#include "TRandom3.h"
#include <cstdlib>
#include <iostream>
#include <cmath>

static TRandom* gsRand = 0;
static double   gsSig  = 0.5*4.0/sqrt(12.0);

//......................................................................

static void make_track(unsigned int n, 
		       double  x0,
		       double  y0,
		       double  z0,
		       double  theta,
		       double  phi, 
		       double* z,
		       double* xory,
		       int*    view) 
{
  unsigned int i;
  for (i=0; i<n; ++i) {
    view[i] = i%2;
    if (view[i]==0) {
      xory[i] = x0 + sin(theta)*cos(phi)*5.0*(float)i;
    }
    else {
      xory[i] = y0 + sin(theta)*sin(phi)*5.0*(float)i;
    }
    xory[i] += gsRand->Gaus(0.0,gsSig);
    z[i] = z0 + cos(theta)*5.0*(float)i;
  }
}

//......................................................................

static void two_track_tests(double theta1,
			    double phi1,
			    double theta2,
			    double phi2,
			    unsigned int n1,
			    unsigned int n2,
			    unsigned int nnoise,
			    double lam,
			    double t,
			    const char* angle1,
			    const char* angle2,
			    const char* vtx1,
			    const char* vtx2,
			    const char* vtx3)
{
  unsigned register int i, j;
  unsigned int n = n1+n2+nnoise;
  static const unsigned int m = 2;
  
  double x0 = 0.0;
  double y0 = 0.0;
  double z0 = 0.0;
  double th, ph;
  double vx, vy, vz;

  double xory[n];
  double z[n];
  int    view[n];
  
  //
  // Make the tracks
  //
  make_track(n1, x0, y0, z0, theta1, phi1, z,    xory,    view);
  make_track(n2, x0, y0, z0, theta2, phi2, z+n1, xory+n1, view+n1);
  //
  // Add in some random noise hits
  //
  for (i=n1+n2; i<n; ++i) {
    z[i]    = gsRand->Uniform(-100.0,100.0);
    xory[i] = gsRand->Uniform(-100.0,100.0);
    if (gsRand->Uniform()>0.5) view[i] = 0;
    else                       view[i] = 1;
  }

  double lamv = 30.0;
  earms::ElasticArms arms(n, m, lam, lamv);
  for (i=0; i<n; ++i) {
    arms.SetHit(i,z[i],xory[i],gsSig,view[i]);
  }
  arms.SetT(t);
  arms.SetVertex(x0,y0,z0);
  arms.SetArm(0,theta1,phi1);
  arms.SetArm(1,theta2,phi2);
  TH2F* thetaphi1h = new TH2F(angle1,
			      ";#theta_{1} [deg.];#phi_{1} [deg.]",
			      50,  0.0,  180.0,
			      50,-180.0, 180.0);
  for (i=0; i<50; ++i) {
    th = M_PI*((float)i+0.5)/50.0;
    for (j=0; j<50; ++j) {
      ph = -M_PI+2.0*M_PI*((float)j+0.5)/50.0;

      arms.SetArm(0,th,ph);

      thetaphi1h->Fill(th*180.0/M_PI, ph*180.0/M_PI, arms.E());
    }
  }
  arms.SetArm(0,theta1,phi1);

  TH2F* thetaphi2h = new TH2F(angle2,
			      ";#theta_{2} [deg.];#phi_{1} [deg.]",
			      50,  0.0,  180.0,
			      50,-180.0, 180.0);
  for (i=0; i<50; ++i) {
    th =  M_PI*((float)i+0.5)/50.0;
    for (j=0; j<50; ++j) {
      ph = -M_PI+2.0*M_PI*((float)j+0.5)/50.0;
      arms.SetArm(1,th,ph);
      thetaphi2h->Fill(th*180.0/M_PI, ph*180.0/M_PI, arms.E());
    }
  }
  arms.SetArm(1,theta2,phi2);

  TH2F* xyh = new TH2F(vtx1,
		       ";x [cm]; y [cm]",
		       50, -50.0, 50.0,
		       50, -50.0, 50.0);
  for (i=0; i<50; ++i) {
    vx = -50.0+100.0*((float)i+0.5)/50.0;
    for (j=0; j<50; ++j) {
      vy = -50.0+100.0*((float)j+0.5)/50.0;
      arms.SetVertex(vx,vy,z0);
      xyh->Fill(vx, vy, arms.E());
    }
  }
  arms.SetVertex(x0, y0, z0);

  TH2F* zxh = new TH2F(vtx2,
		       ";z [cm]; x [cm]",
		       50, -100.0,100.0,
		       50, -50.0, 50.0);
  for (i=0; i<50; ++i) {
    vz = -100.0+200.0*((float)i+0.5)/50.0;
    for (j=0; j<50; ++j) {
      vx = -50.0+100.0*((float)j+0.5)/50.0;
      arms.SetVertex(vx, y0, vz);
      zxh->Fill(vz, vx, arms.E());
    }
  }
  arms.SetVertex(x0, y0, z0);

  TH2F* zyh = new TH2F(vtx3,
		       ";z [cm]; y [cm]",
		       50, -100.0,100.0,
		       50, -50.0, 50.0);
  for (i=0; i<50; ++i) {
    vz = -100.0+200.0*((float)i+0.5)/50.0;
    for (j=0; j<50; ++j) {
      vy = -50.0+100.0*((float)j+0.5)/50.0;
      arms.SetVertex(x0, vy, vz);
      zyh->Fill(vz, vy, arms.E());
    }
  }
}

//......................................................................

static void minimizer_tests(double theta1,
			    double phi1,
			    double theta2,
			    double phi2,
			    unsigned int n1,
			    unsigned int n2,
			    unsigned int nnoise,
			    double lam)
{
  unsigned register int i;
  unsigned int n = n1+n2+nnoise;
  static const unsigned int m = 2;
  
  double x0 = 0.0;
  double y0 = 0.0;
  double z0 = 0.0;

  double xory[n];
  double z[n];
  int    view[n];
  
  //
  // Make the tracks
  //
  make_track(n1, x0, y0, z0, theta1, phi1, z,    xory,    view);
  make_track(n2, x0, y0, z0, theta2, phi2, z+n1, xory+n1, view+n1);
  //
  // Add in some random noise hits
  //
  for (i=n1+n2; i<n; ++i) {
    z[i]    = gsRand->Uniform(-100.0,100.0);
    xory[i] = gsRand->Uniform(-100.0,100.0);
    if (gsRand->Uniform()>0.5) view[i] = 0;
    else                       view[i] = 1;
  }

  double lamv = 30.0;
  earms::ElasticArms arms(n, m, lam, lamv);
  for (i=0; i<n; ++i) arms.SetHit(i,z[i],xory[i],gsSig,view[i]);

  arms.SetVertex(gsRand->Gaus(x0,10.), 
		 gsRand->Gaus(y0,10.),
		 gsRand->Gaus(z0,10.));
  arms.SetArm(0,
	      gsRand->Gaus(theta1,M_PI/9.),
	      gsRand->Gaus(phi1,  M_PI/9.));
  arms.SetArm(1,
	      gsRand->Gaus(theta2,M_PI/9.),
	      gsRand->Gaus(phi2,  M_PI/9.));
  
  earms::Minimizer mn(&arms);
  double E = mn.Fit();
  
  std::cout << E << " ["
	    << mn.fP[3]*180.0/M_PI << ","
	    << mn.fP[4]*180.0/M_PI << "] ["
	    << mn.fP[5]*180.0/M_PI << ","
	    << mn.fP[6]*180.0/M_PI << "] ("
	    << mn.fP[0] << ","
	    << mn.fP[1] << ","
	    << mn.fP[2] << ")"
	    << std::endl;
}

//......................................................................

static void beta_tests(double theta1,
		       double phi1,
		       double theta2,
		       double phi2,
		       unsigned int n1,
		       unsigned int n2,
		       unsigned int nnoise,
		       double lam)
{
  unsigned register int i;
  unsigned int n = n1+n2+nnoise;
  static const unsigned int m = 2;
  
  double x0 = 0.0;
  double y0 = 0.0;
  double z0 = 0.0;
  
  double xory[n];
  double z[n];
  int    view[n];
  
  //
  // Make the tracks
  //
  make_track(n1, x0, y0, z0, theta1, phi1, z,    xory,    view);
  make_track(n2, x0, y0, z0, theta2, phi2, z+n1, xory+n1, view+n1);
  //
  // Add in some random noise hits
  //
  for (i=n1+n2; i<n; ++i) {
    z[i]    = gsRand->Uniform(-100.0,100.0);
    xory[i] = gsRand->Uniform(-100.0,100.0);
    if (gsRand->Uniform()>0.5) view[i] = 0;
    else                       view[i] = 1;
  }

  double lamv = 30.0;
  earms::ElasticArms arms(n, m, lam, lamv);
  for (i=0; i<n; ++i) {
    arms.SetHit(i,z[i],xory[i],gsSig,view[i]);
  }
  arms.SetArm(0,theta1,phi1);
  arms.SetArm(1,theta1,phi1);
  arms.SetVertex(x0,y0,z0);
  TH1F* h = new TH1F("evsw",";t;E",100,0.0,100.0);
  for (i=0; i<100; ++i) {
    arms.SetT((float)i+0.5);
    h->Fill((float)i+0.5,arms.E());
  }
}

//......................................................................

int main(void) 
{
  gsRand = new TRandom3(time(0));

  TFile* f = new TFile("f.root","RECREATE");

  std::cerr << "Two track test 1...";
  two_track_tests( 30*M_PI/180, // theta 1
		   60*M_PI/180, // phi 1
		   60*M_PI/180, // theta 2
		  -30*M_PI/180, // phi 2
		   100,         // nhit 1
		   20,          // nhit 2
		   0,           // n_noise
		   36.0,        // lambda
		   1.0,         // t
		   "test1_angles1",
		   "test1_angles2",
		   "test1_vtx1",
		   "test1_vtx2",
		   "test1_vtx3");
  std::cerr << "done." << std::endl;
  
  std::cerr << "Two track test 2...";
  two_track_tests( 30*M_PI/180, // theta 1
		   60*M_PI/180, // phi 1
		   60*M_PI/180, // theta 2
		  -30*M_PI/180, // phi 2
		   100,         // nhit 1
		   20,          // nhit 2
		   10,          // n_noise
		   36.0,        // lambda
		   1.0,         // temp
		   "test2_angles1",
		   "test2_angles2",
		   "test2_vtx1",
		   "test2_vtx2",
		   "test2_vtx3");
  std::cerr << "done." << std::endl;
  
  std::cerr << "Two track test 3...";
  two_track_tests( 30*M_PI/180, // theta 1
		   60*M_PI/180, // phi 1
		   60*M_PI/180, // theta 2
		  -30*M_PI/180, // phi 2
		   100,         // nhit 1
		   20,          // nhit 2
		   100,         // n_noise
		   36.0,        // lambda
		   1.0,         // temp
		   "test3_angles1",
		   "test3_angles2",
		   "test3_vtx1",
		   "test3_vtx2",
		   "test3_vtx3");
  std::cerr << "done." << std::endl;

  std::cerr << "Two track test 4...";
  two_track_tests( 30*M_PI/180, // theta 1
		   60*M_PI/180, // phi 1
		   60*M_PI/180, // theta 2
		  -30*M_PI/180, // phi 2
		   100,         // nhit 1
		   20,          // nhit 2
		   100,         // n_noise
		   36.0,        // lambda
		   100,         // temp
		   "test4_angles1",
		   "test4_angles2",
		   "test4_vtx1",
		   "test4_vtx2",
		   "test4_vtx3");
  std::cerr << "done." << std::endl;

  std::cerr << "Minimizer test...";
  minimizer_tests( 30.0*M_PI/180., // theta1
		   60.0*M_PI/180., // phi1
		   60.0*M_PI/180., // theta2
                  -30.0*M_PI/180., // phi2
		   100,            // n1
		   20,             // n2
		   5,              // nnoise
		   36.0);          // lambda
  std::cerr << "done." << std::endl;

  std::cerr << "Beta test...";
  beta_tests( 30.0*M_PI/180.,
	      60.0*M_PI/180.,
	      60.0*M_PI/180.,
	     -30.0*M_PI/180.,
	      100,
	      20,
	      10,
	      6.0);
  std::cerr << "done." << std::endl;

  f->Write();
}
////////////////////////////////////////////////////////////////////////
