////////////////////////////////////////////////////////////////////////
/// \file    ClusterEva.cxx
/// \brief   TODO
/// \version $Id: ClusterEva.cxx,v 1.7 2012-09-25 06:30:12 ricken Exp $
/// \author  ricken@fnal.gov
////////////////////////////////////////////////////////////////////////
#include <cmath>
#include <vector>
#include <string>

// ROOT includes
#include "TH1I.h"
#include "TNtuple.h"
#include "TBenchmark.h"

// NOvA includes
#include "Calibrator/Calibrator.h"
#include "GeometryObjects/CellGeo.h"
#include "Geometry/Geometry.h"
#include "GeometryObjects/PlaneGeo.h"
#include "RecoBase/CellHit.h"
#include "RecoBase/Cluster.h"
#include "RecoBase/RecoHit.h"
#include "MCCheater/BackTracker.h"
#include "Simulation/ParticleNavigator.h"

// Framework includes
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "art/Framework/Services/Optional/TFileService.h"
#include "art/Framework/Core/EDAnalyzer.h"
#include "art/Framework/Core/ModuleMacros.h"

class TTree;
class TVector3;

namespace geo { class Geometry; }
namespace rb  { class CellHit; class Cluster; }

///tracking algorithms
namespace pa {
  class ClusterEva : public art::EDAnalyzer {
  public:
    explicit ClusterEva(fhicl::ParameterSet const& pset);
    virtual ~ClusterEva();

    void beginJob();
    void analyze(const art::Event& evt);
    void reconfigure(const fhicl::ParameterSet& p);
    
  private:
    std::string fCellHitInput;
    std::string fClusterInput;    ///< Input folder from cluster reco
    unsigned fNpart;
    TTree* ClusterTruth;
    bool _view;
    double _tmin, _tmax;
    int _nhits, _eventID;
    int _ADC;
    std::vector<float> _nTH;
    std::vector<float> _nTHtot;
    std::vector<float> _TADC;
    std::vector<float> _TADCtot;
    std::vector<float> _Ttmin;
    std::vector<float> _Ttmax;
    std::vector<float> _Tenergy;
  };
}

namespace pa{  
  //......................................................................
  ClusterEva::ClusterEva(fhicl::ParameterSet const& pset)
    : EDAnalyzer(pset)
  {
    reconfigure(pset);
  }
  
  //......................................................................
  ClusterEva::~ClusterEva()
  {
  }
  
  //......................................................................
  void ClusterEva::reconfigure(const fhicl::ParameterSet& pset)
  {
    fCellHitInput    = pset.get< std::string >("CellHitInput");
    fClusterInput    = pset.get< std::string >("ClusterInput") ;
 
  }

  //......................................................................
  void ClusterEva::beginJob()
  {
    art::ServiceHandle<art::TFileService> tfs;
    ClusterTruth = tfs->make<TTree>("ClusterTruth","Hough Phase Space");

    ClusterTruth -> Branch("evtID", &_eventID, "evtID/I");
    ClusterTruth -> Branch("view",  &_view,    "VIEW/O");      //0: X; 1: Y
    ClusterTruth -> Branch("Tmin",  &_tmin,    "Tmin/D");
    ClusterTruth -> Branch("Tmax",  &_tmax,    "Tmax/D");
    ClusterTruth -> Branch("Nhits", &_nhits,   "Nhits/I");
    ClusterTruth -> Branch("totADC",&_ADC,     "totADC/I");

    ClusterTruth -> Branch("TrueHits",        &_nTH);
    ClusterTruth -> Branch("totTrueHits",     &_nTHtot);
    ClusterTruth -> Branch("TrueADC",         &_TADC);
    ClusterTruth -> Branch("totTrueADC",      &_TADCtot);
    ClusterTruth -> Branch("TrueTmin",        &_Ttmin);
    ClusterTruth -> Branch("TrueTmax",        &_Ttmax);
    ClusterTruth -> Branch("TrueEnergy",      &_Tenergy);
  }

  void ClusterEva::analyze(const art::Event& evt) 
  {
    _eventID = evt.id().event();

    //make a backtracker
    art::ServiceHandle<cheat::BackTracker> bt;
    // define vector holding the clusters
    art::Handle< std::vector<rb::Cluster> > slices;
    evt.getByLabel(fClusterInput, slices);
    art::Handle< std::vector<rb::CellHit> > hitcol;
    evt.getByLabel(fCellHitInput, hitcol);

    art::ServiceHandle<geo::Geometry> geom;

    for(unsigned sliceIdx = 0; sliceIdx != slices->size(); ++sliceIdx){
      //Now you are inside a new cluster, you have to clear the vector: 
      _nTH.clear();
      _nTHtot.clear();
      _TADC.clear();
      _TADCtot.clear();
      _Ttmin.clear();
      _Ttmax.clear();
      _Tenergy.clear();

      const rb::Cluster& slice = (*slices)[sliceIdx];
      _ADC   = 0;  
      _tmin  = (slice.Cell(0))->TNS();
      _tmax  =_tmin;
      _nhits = slice.NCell();
      if ((slice.Cell(0))-> View()==geo::kX) _view = 0;
      else _view = 1;

      std::vector<int> track_ides;

      for(int hitIdx = 0; hitIdx != _nhits; ++hitIdx){

	const art::Ptr<rb::CellHit> chit = slice.Cell(hitIdx);
	_ADC += chit->ADC();
 
	if (_tmin > chit->TNS() ) _tmin=chit->TNS();
	if (_tmax < chit->TNS() ) _tmax=chit->TNS();

	//Now work on MC truth:
	try {
	  if (!bt->IsNoise(chit)) {
	    const sim::Particle *part = bt->HitToParticle(chit);

	    if ( find(track_ides.begin(),track_ides.end(),part->TrackId()) == track_ides.end()) 
	      {

		//		std::cout<<"new particle track id:"<<part->TrackId()<<std::endl;
		track_ides.push_back(part->TrackId() );
		//Initialize items:
		int   cheater_nhits  = 0;
		int   cheater_ADC    = 0;
		float cheater_tmin   = 9.e19;
		float cheater_tmax   = -9.e10;

		//Now you have to loop over all the hits to check the true hits mainly caused by this particle:
		for (unsigned i=0; i != hitcol->size(); ++i) {

		  art::Ptr<rb::CellHit> hit_ptr(hitcol, i);
		  /*
		  if (!_view && hit_ptr->View()==geo::kY) continue;
		  else if (_view && hit_ptr->View()==geo::kX) continue;
		  */ 
		  //	  std::cout<<"pass view check"<<std::endl;
		  if (bt->HitToParticle(hit_ptr) == part) {
		    // std::cout<<"Get the particle ID:"<<(bt->HitToParticle(hit_ptr) )->TrackId()<<std::endl;
		    ++cheater_nhits;
		    cheater_ADC+=hit_ptr->ADC();

		    // std::cout<<cheater_nhits<<"th hit with ADC "<<hit_ptr->ADC()<<std::endl;
		    if (cheater_tmin > hit_ptr->TNS() ) cheater_tmin = hit_ptr->TNS();  
		    if (cheater_tmax < hit_ptr->TNS() ) cheater_tmax = hit_ptr->TNS();  
		  }
		}

		_nTH.push_back(1);
		_TADC.push_back(chit->ADC());

		_Tenergy.push_back(part->E() );
		_nTHtot.push_back(cheater_nhits);
		_TADCtot.push_back(cheater_ADC);
		_Ttmin.push_back(cheater_tmin);
		_Ttmax.push_back(cheater_tmax);

		//		std::cout<<"real energy of this particle: "<<part->E()<<" GeV"<<std::endl;


	      }
	    //Or this particle has been left trace in previous hits:
	    else {
	      int particle_idx = find(track_ides.begin(),track_ides.end(),part->TrackId()) - track_ides.begin();

	      ++_nTH.at(particle_idx);
	      _TADC.at(particle_idx) += chit->ADC();  

	    }
	  }
	}
	catch (...) {
	  continue;
	}
      }//End looping over this slice

      //Fill the tree:
      /*
      std::cout<<"clustered hits Particle size: "<<_nTH.size()<<std::endl;
      std::cout<<"clustered ADC Particle size: " <<_TADC.size()<<std::endl;
      std::cout<<"total hits Particle size: "    <<_nTHtot.size()<<std::endl;
      std::cout<<"total ADC Particle size: "     <<_TADC.size()<<std::endl;
      std::cout<<"size of truth energy:  "       <<_Tenergy.size()<<std::endl;
      */
      ClusterTruth->Fill();


    }// end looping over all slices

    return;

  }   //end analyze

  DEFINE_ART_MODULE(ClusterEva)

}
