////////////////////////////////////////////////////////////////////////
// \file    DataSliceEva.cxx
//
// \brief   First official DDT analyzer in offline!
//
// \author  ricken@fnal.gov
////////////////////////////////////////////////////////////////////////
#include "art/Framework/Core/EDAnalyzer.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Core/FindOneP.h"
#include "art/Framework/Core/FindManyP.h"
#include "art/Framework/Services/Optional/TFileService.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "MCCheater/BackTracker.h"
#include "SimulationBase/MCTruth.h"

#include "DDTBaseDataProducts/Track3D.h"
#include "DDTBaseDataProducts/HitList.h"
#include "DDTBaseDataProducts/DAQHit.h"
#include "DDTBaseDataProducts/CompareDAQHit.h"
#include "DDTBaseDataProducts/TriggerDecision.h"
#include "DDTBaseDataProducts/BaseProducts.h"
#include "DDTBaseDataProducts/GroupedHitList.h"
#include "DDTBaseDataProducts/Track.h"

#include "DAQChannelMap/DAQChannelMap.h"
#include "RecoBase/Track.h"
#include "RecoBase/Cluster.h"



#include "TTree.h"


#include <string>
#include <iostream>
#include <cmath>
#include <algorithm>
#include <vector>
#include <fstream>

class TTree;

namespace ddt {
  class DataSliceEva;
}


class ddt::DataSliceEva : public art::EDAnalyzer
{
public:
  explicit DataSliceEva(fhicl::ParameterSet const &pset);
  virtual ~DataSliceEva();
                        
  void analyze(const art::Event &evt);
  void beginJob();

      
private:
  std::string _slicelabel;
  std::string _sliceinstance;
  int _ADCthr;
  int _xMin;  
  int _xMax;
  int _yMin;
  int _yMax;
  int _zMin;
  int _zMax;
  int _deltx;
  int _delty;
  int _deltz;

  TTree* _slicetree;
 
  float ADC_mean;
  unsigned long long TDC_min, TDC_max;
  int X_min, X_max;
  int Y_min, Y_max;
  int P_min, P_max;
  int N_hits, N_sh;
  int eventID;
  bool PassThrough;
};



ddt::DataSliceEva::DataSliceEva(fhicl::ParameterSet const& p)
  : EDAnalyzer(p),
  _slicelabel(p.get<std::string>("slice_label")),
  _sliceinstance(p.get<std::string>("slice_instance")),
  _ADCthr(p.get<float>("ADC_thr")),
  _xMin(p.get<float>("x_min")),
  _xMax(p.get<float>("x_max")),
  _yMin(p.get<float>("y_min")),
  _yMax(p.get<float>("y_max")),
  _deltx(p.get<float>("x_delt")),
  _delty(p.get<float>("y_delt")) 
{
}

//......................................................................
ddt::DataSliceEva::~DataSliceEva(){}
  
//......................................................................

void ddt::DataSliceEva::beginJob()
{
  art::ServiceHandle<art::TFileService> tfs;
  _slicetree = tfs->make<TTree>("slicetree","hit information");
  _slicetree -> Branch("ADC_mean",&ADC_mean,"ADC_mean/F");
  _slicetree -> Branch("TDC_min",&TDC_min,"TDC_min/l");
  _slicetree -> Branch("TDC_max",&TDC_max,"TDC_max/l");
  _slicetree -> Branch("X_min",&X_min,"X_min/I");
  _slicetree -> Branch("X_max",&X_max,"X_max/I");
  _slicetree -> Branch("Y_min",&Y_min,"Y_min/I");
  _slicetree -> Branch("Y_max",&Y_max,"Y_max/I");
  _slicetree -> Branch("P_min",&P_min,"P_min/I");
  _slicetree -> Branch("P_max",&P_max,"P_max/I");
  _slicetree -> Branch("N_hits",&N_hits,"N_hits/I");           //Number of hits in the slice
  _slicetree -> Branch("N_sh",&N_sh,"N_sh/I");
  _slicetree -> Branch("though",&PassThrough,"through/O");

  return;
}

void ddt::DataSliceEva::analyze(const art::Event& e)
{
  eventID = e.id().event();

  //Get the Slice Info:
  art::Handle<std::vector<novaddt::HitList> > slices;
  e.getByLabel(_slicelabel, _sliceinstance, slices);

  for (unsigned i = 0; i!= slices->size(); ++i) {
    novaddt::HitList const &hits = slices->at(i);
    TDC_min = hits.begin()->TDC().val;
    TDC_max = TDC_min;
    P_min = 9999;
    P_max = -99;
    X_min = 9999;      
    X_max = -99;
    Y_min = 9999;      
    Y_max = -99;

    int ADC_sum = 0;
    N_hits = hits.size();
    N_sh = 0;

    for (auto const hit: hits) {
      if (TDC_min  > hit.TDC().val ) TDC_min  = hit.TDC().val;
      if (TDC_max < hit.TDC().val ) TDC_max = hit.TDC().val; 
      if (P_min > hit.Plane().val)  P_min   = hit.Plane().val; 
      if (P_max < hit.Plane().val)  P_max   = hit.Plane().val;
      if (hit.View().val == daqchannelmap::X_VIEW) {
	if (X_max < hit.Cell().val)   X_max   = hit.Cell().val;
	if (X_min > hit.Cell().val)   X_min   = hit.Cell().val;
      }
      else {
	if (Y_max < hit.Cell().val)   Y_max   = hit.Cell().val;
	if (Y_min > hit.Cell().val)   Y_min   = hit.Cell().val;
      }

      if (hit.ADC().val > _ADCthr) ++N_sh;
      ADC_sum += hit.ADC().val;      
    }
    ADC_mean = ADC_sum/(N_hits+0.);

    int number =0;
    if (X_min<_xMin+_deltx) ++number;
    if (X_max>_xMax-_deltx) ++number;
    if (Y_min<_yMin+_delty) ++number;
    if (Y_max>_yMax-_delty) ++number;
    if (P_min<_zMin+_deltz) ++number;
    if (P_max>_zMax-_deltz) ++number;

    if (number>1) PassThrough = true;
    else PassThrough = false;
    _slicetree -> Fill();
  }
  return;
}


DEFINE_ART_MODULE(ddt::DataSliceEva)
