///
/// \file    HardwarePad.cxx
/// \brief   Drawing pad for X-Z or Y-Z projections of events
/// \author  messier@indiana.edu
/// \version $Id: HardwarePad.cxx,v 1.2 2011-10-21 20:37:50 gsdavies Exp $
///
#include "TBox.h"
#include "TH2F.h"
#include "TPad.h"
#include "TText.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "DAQChannelMap/HardwareDisplay.h"
#include "Geometry/Geometry.h"
#include "EventDisplay/HardwarePad.h"
#include "EventDisplay/RawDataDrawer.h"
#include "EventDisplayBase/EventHolder.h"
#include "EventDisplayBase/View2D.h"

// Numeric flags for the XZ or ZY views
static const int kX = 0;
static const int kY = 1;

namespace evd
{
  ///
  /// Create a pad showing a single X-Z or Y-Z projection of the detector
  /// \param nm  : Name of the pad
  /// \param ti  : Title of the pad
  /// \param x1  : Location of left  edge of pad (0-1)
  /// \param x2  : Location of right edge of pad (0-1)
  /// \param y1  : Location of bottom edge of pad (0-1)
  /// \param y2  : Location of top    edge of pad (0-1)
  /// \param opt : Options. Include 'x', 'y' for xz or yz projections
  ///
  HardwarePad::HardwarePad(const char* nm,
			   const char* ti,
			   double x1, double x2,
			   double y1, double y2,
			   const char* /*opt*/) :
    DrawingPad(nm, ti, x1, x2, y1, y2),
    fDetId(0),
    fHisto(0)
  {
    this->Pad()->SetTopMargin   (0.01);
    this->Pad()->SetBottomMargin(0.11);
    this->Pad()->SetLeftMargin  (0.11);
    this->Pad()->SetRightMargin (0.01);

    this->Pad()->SetBit(kCannotPick);
    this->Pad()->SetBit(TPad::kCannotMove);

    fView = new evdb::View2D();
  }

  //......................................................................
  
  void HardwarePad::Config(int detid)
  {
    if (fDetId == detid) return;
    fDetId = detid;
    
    if (fHisto) { delete fHisto; fHisto = 0; }

    daqchannelmap::HardwareDisplay hd;
    hd.SetupDet(detid);

    unsigned int nx = hd.fXmax;
    unsigned int ny = hd.fYmax;

    this->Pad()->cd();
    fHisto = new TH2F("fHisto","",nx,-0.5,nx-0.5,ny,-0.5,ny-0.5);
    fHisto->Draw("AB");    
  }

  //......................................................................
  
  HardwarePad::~HardwarePad() 
  {
    if (fHisto) { delete fHisto; fHisto = 0; }
    if (fView)  { delete fView;  fView  = 0; }
  }

  //......................................................................
  
  void HardwarePad::Draw(const char* /*opt*/) 
  {
    fView->Clear();
    
    //
    // Make sure we are configured for the right detector
    //
    art::ServiceHandle<geo::Geometry> geo;
    this->Config(geo->DetId());

    const art::Event* evt = evdb::EventHolder::Instance()->GetEvent();
    this->RawDataDraw()->RawDigitHardware(*evt, fView);

    fPad->Clear();
    fPad->cd();
    fHisto->Draw("AB");

    fView->Draw();
  }

} // end namespace evd
////////////////////////////////////////////////////////////////////////
