/// \file    RecoBaseDrawer.cxx
/// \brief   Class to aid in the rendering of RecoBase objects
/// \author  messier@indiana.edu
/// \version $Id: RecoBaseDrawer.cxx,v 1.37 2012-08-09 23:38:31 bckhouse Exp $
///
#include "EventDisplay/RecoBaseDrawer.h"

#include <cmath>
#include <map>
#include "TBox.h"
#include "TH1.h"
#include "TLine.h"
#include "TMarker.h"
#include "TPolyMarker.h"
#include "TPolyLine.h"
#include "TPolyLine3D.h"

#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/Handle.h" 
#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "art/Persistency/Common/Ptr.h"

#include "EventDisplayBase/Colors.h"
#include "EventDisplayBase/ColorScale.h"
#include "EventDisplayBase/View2D.h"
#include "EventDisplayBase/View3D.h"

#include "EventDisplay/RawDrawingOptions.h"
#include "EventDisplay/RecoDrawingOptions.h"
#include "EventDisplay/SliceNavigator.h"
#include "Geometry/Geometry.h"
#include "GeometryObjects/CellGeo.h"
#include "MichelEFilter/MichelECluster.h"
#include "RecoBase/CellHit.h"
#include "RecoBase/Cluster.h"
#include "RecoBase/HoughResult.h"
#include "RecoBase/Prong.h"
#include "RecoBase/Track.h"
#include "RecoBase/Vertex.h"
#include "Utilities/func/MathUtil.h"

namespace evd
{
  //....................................................................
  RecoBaseDrawer::RecoBaseDrawer() 
  {
  }

  //......................................................................

  RecoBaseDrawer::~RecoBaseDrawer() 
  {
 
  }

  //......................................................................

  ///
  /// Render CellHit objects on a 2D viewing canvas
  ///
  /// @param evt    : Event handle to get data objects from
  /// @param which  : Which collection of object to render
  /// @param xzview : Pointer to X-Z view to draw on
  /// @param yzview : Pointer to Y-Z view to draw on
  ///
  void RecoBaseDrawer::CellHit2D(const art::Event& evt,
				 evdb::View2D*     xzview,
				 evdb::View2D*     yzview) 
  {
    art::ServiceHandle<evd::RawDrawingOptions>  rawopt;

    if (rawopt->fWhichHits!=evd::RawDrawingOptions::kCAL_HITS)
      return;
    
    art::ServiceHandle<evd::SliceNavigator> nav;

    art::ServiceHandle<evd::RecoDrawingOptions> recoopt;
    art::ServiceHandle<geo::Geometry>           geo;
    
    std::vector<const rb::Cluster*> slicesToNavigate = nav->SlicesToNavigate(evt);
    std::vector<const rb::CellHit*> unfiltHits;
    for(int i=0, n = slicesToNavigate.size(); i < n; ++i ){
      for( int c = 0, nc = slicesToNavigate[i]->NCell(); c < nc; ++c)
	unfiltHits.push_back( slicesToNavigate[i]->Cell(c).get());
    }

   //std::cout << recoopt->fCellHitModules.size() << " modules found with cell hits." << std::endl;

    for (unsigned int imod=0; imod<recoopt->fCellHitModules.size(); ++imod) {
      const std::string& which    = recoopt->fCellHitModules[imod].c_str();
      const std::string& instance = recoopt->fCellHitInstances[imod];

      std::vector<const rb::CellHit*> cellhits;
      nav->GetProducts(evt, which, instance, cellhits);
      
      // Display all hits on the two 2D views provided
      for (unsigned int i=0; i<cellhits.size(); ++i) {
	const rb::CellHit* h = cellhits[i];
	evdb::View2D*  v = 0;
	
	// Select which view to draw this cell hit on
	if (h->View() == geo::kX) v = xzview;
	if (h->View() == geo::kY) v = yzview;
	if (v==0) continue;
	
	// Try to get the "best" charge measurement, ie. the one last in
	// the calibration chain
	float q = h->PE();
	float t = (h->TNS())/1000.0; // thisto is binned in usec
	
	// Render cell hits as markers with size and color encoding time
	// and charge
	double dz = geo->Plane(h->Plane())->Cell(h->Cell())->HalfD();
	double dv = geo->Plane(h->Plane())->Cell(h->Cell())->HalfW();
	double zpos=0;
	double vpos=0;
	
	double xyz[3];
	geo->Plane(h->Plane())->Cell(h->Cell())->GetCenter(xyz);
	zpos = xyz[2];
	if (h->View() == geo::kX) vpos = xyz[0];
	else                      vpos = xyz[1];
	
	double sf =   this->ScaleFactor(t, q);
	//double q0 = 400.0;
	int    c  = 1;
	art::ServiceHandle<evdb::Colors> colors;	
	switch (rawopt->fColor) {
	case evd::RawDrawingOptions::kCOLOR_BY_CHARGE:
	  c = colors->Scale("CalQ").GetColor(q);
	  break;
	case evd::RawDrawingOptions::kCOLOR_BY_TIME:
	  c = colors->Scale("CalT").GetColor(t);
	  //sf = sqrt(q/q0);
	  //if (sf>1.0) sf = 1.0;
	  break;
	default:
	  c = 1;
	}
	c = this->DimOutOfTime(t, t, c);
	if(nav->DimHits() == true) c = this->DimFiltered( h, unfiltHits, c );

	TBox& b1 = v->AddBox(zpos-sf*dz, vpos-sf*dv, zpos+sf*dz, vpos+sf*dv);
	b1.SetFillStyle(1001);
	b1.SetFillColor(c);
	b1.SetBit(kCannotPick);
      }
    } // loop on imod folders
  }

  //......................................................................

  ///
  /// Render CellHit objects on a 3D viewing canvas
  ///
  /// @param evt    : Event handle to get data objects from
  /// @param which  : Which collection of object to render
  /// @param view   : Pointer to 3D viewer
  ///
  void RecoBaseDrawer::CellHit3D(const art::Event& evt,
				 evdb::View3D*     view)
  {
    art::ServiceHandle<evd::RawDrawingOptions>  rawopt;
    if (rawopt->fWhichHits!=evd::RawDrawingOptions::kCAL_HITS) return;
    art::ServiceHandle<evd::SliceNavigator> nav;

    art::ServiceHandle<evd::RecoDrawingOptions> drawopt;
    art::ServiceHandle<geo::Geometry>           geo;

    for (unsigned int imod=0; imod<drawopt->fCellHitModules.size(); ++imod) {
      const char* which = drawopt->fCellHitModules[imod].c_str();
      const std::string& instance = drawopt->fCellHitInstances[imod];
    
      std::vector<const rb::CellHit*> cellhits;
      nav->GetProducts(evt, which, instance, cellhits);
  
      // Display all hits on the two 2D views provided
      for (unsigned int i=0; i<cellhits.size(); ++i) {
	const rb::CellHit* h = cellhits[i];
      
	// Try to get the "best" charge measurement, ie. the one last in
	// the calibration chain
	float q=h->PE();
	//       if (h->IsA()==rb::RecoHit::Class()) {
	// 	q = ((rb::RecoHit*)h)->GeV(); // or MIP or PECorr()
	//       }
      
	float t=h->TNS();
	//       if (h->IsA()==rb::RecoHit::Class()) {
	// 	t = ((rb::RecoHit*)h)->T();
	//       }
      
	// Render cell hits as markers with size and color encoding time
	// and charge
	static evdb::ColorScale tscale(0.0,30.0);
	int    c     = tscale(t);
	double q0    = 1.0;
	double sz    = 0.3*sqrt(q/q0);
	
	c = this->DimOutOfTime(t,t,c);
	
	TPolyLine3D& p = view->AddPolyLine3D(5,c,1,2);
	double x1, x2, y1, z1, z2;
	// double y2; // i get set but am not used
      
	double xyz[3];
	geo->Plane(h->Plane())->Cell(h->Cell())->GetCenter(xyz);
      
	//       if (h->IsA()==rb::RecoHit::Class()) {	
	// 	z1 = ((rb::RecoHit*)h)->Z();	
	// 	if (h->View() == geo::kX) {
	// 	  x1 = ((rb::RecoHit*)h)->V();
	// 	  y1 = xyz[1];
	// 	}
	// 	else {
	// 	  y1 = ((rb::RecoHit*)h)->V();
	// 	  x1 = xyz[1];
	// 	}
	//       }
	//       else {
	z1 = xyz[2];
	if (h->View() == geo::kX) {
	  x1 = xyz[0];
	  y1 = xyz[1];
	}
	else {
	  x1 = xyz[1];
	  y1 = xyz[0];
	}
	//       }
      
	x2 = x1+sz;
	//y2 = y1+sz; // i am not used anywhere
	z2 = z1+sz;
	x1 -= sz;
	y1 -= sz;
	z1 -= sz;
      
	p.SetPoint(0,x1-sz,y1,z1);
	p.SetPoint(1,x1+sz,y1,z1);
	p.SetPoint(2,x2,y1,z2);
	p.SetPoint(1,x1,y1,z2);
	p.SetPoint(0,x1,y1,z1);
      
      } // loop on i cell hits
    } // loop on imod folders
  }

  double RecoBaseDrawer::ScaleFactor(double /*t*/,double q)
{
  art::ServiceHandle<evd::RawDrawingOptions> drawopt;

  double sf = drawopt->fScaleFactor;
  double q0 = 3.0;
  if (drawopt->fRawDrawingOpt & evd::RawDrawingOptions::kSCALE_HITS_BY_CHARGE)
  {
    sf *= 0.3*sqrt(q/q0);
  }
  return sf;
}

  //......................................................................
  void RecoBaseDrawer::DrawCluster2D(std::vector<const rb::Cluster*>& clusts,
                                     evdb::View2D*                  xzview,
                                     evdb::View2D*                  yzview,
                                     std::vector<int>&              clustIndex)
  {
    art::ServiceHandle<geo::Geometry> geo;
    art::ServiceHandle<evd::RecoDrawingOptions> drawopt;

    bool drawAsMarkers = false;
    bool drawAsBoxes   = false;
    bool drawAsHulls   = false;
    if (drawopt->fClusterStyle & evd::RecoDrawingOptions::kCLUSTERS_AS_MARKERS) {
      drawAsMarkers = true;
    }
    if (drawopt->fClusterStyle & evd::RecoDrawingOptions::kCLUSTERS_AS_BOXES) {
      drawAsBoxes = true;
    }
    if (drawopt->fClusterStyle & evd::RecoDrawingOptions::kCLUSTERS_AS_HULLS) {
      drawAsHulls = true;
    }

    for (unsigned int iClust=0; iClust<clusts.size(); ++iClust) {
      const rb::Cluster* c = clusts[iClust];

      std::vector< int > clustIndex;
      bool drawThisCluster = clustIndex.empty();

      // if the indices have been provided, check whether or
      // not this cluster should be drawn
      for(int index = 0; index < (int)clustIndex.size(); index++){
        if( clustIndex[index] == -1  || clustIndex[index] == (int)iClust)
          drawThisCluster = true;
      }

      // Don't render "noise" clusters or clusters you don't want to draw
      if (c->IsNoise() || !drawThisCluster ) continue;

      if (drawAsMarkers) {
        // draw cluster with unique marker
        int color = this->GetClusterColor(iClust);
        color = this->DimOutOfTime(c->MinTNS(),c->MaxTNS(),color);

        if(xzview){
          TPolyMarker& pm = xzview->AddPolyMarker(c->NXCell(),
                                                  color,
                                                  kFullCircle, 0.8);
          for(unsigned int hitIdx = 0; hitIdx < c->NXCell(); ++hitIdx){
            const art::Ptr<rb::CellHit>& h = c->XCell(hitIdx);
            double xyz[3];
            geo->Plane(h->Plane())->Cell(h->Cell())->GetCenter(xyz);
            pm.SetPoint(hitIdx, xyz[2], xyz[0]);
          }
        } // end if xzview
        if(yzview){
          TPolyMarker& pm = yzview->AddPolyMarker(c->NYCell(),
                                                  color,
                                                  kFullCircle, 0.8);
          for(unsigned int hitIdx = 0; hitIdx < c->NYCell(); ++hitIdx){
            const art::Ptr<rb::CellHit>& h = c->YCell(hitIdx);
            double xyz[3];
            geo->Plane(h->Plane())->Cell(h->Cell())->GetCenter(xyz);
            pm.SetPoint(hitIdx, xyz[2], xyz[1]);
          }
        } // end if yzview
      }

      if(drawAsBoxes || drawAsHulls){
        std::vector<double> xzZpts, xzXpts;
        std::vector<double> yzZpts, yzYpts;

        if (drawAsBoxes) {
          this->GetClusterOutlines(c, geo::kX, xzXpts, xzZpts);
          this->GetClusterOutlines(c, geo::kY, yzYpts, yzZpts);
        }
        else if (drawAsHulls) {
          this->GetClusterOutlinesConvexHull(c, geo::kX, xzXpts, xzZpts);
          this->GetClusterOutlinesConvexHull(c, geo::kY, yzYpts, yzZpts);
        }

        bool isX   = !xzZpts.empty();
        bool isY   = !yzZpts.empty();
        int lcolor = this->GetClusterColor(iClust);
        int fcolor = (isX && isY) ? lcolor : 17; // fill color
        int width  = 1;  // line width
        int style  = 1;  // 1=solid line style

        lcolor = this->DimOutOfTime(c->MinTNS(), c->MaxTNS(), lcolor);
        fcolor = this->DimOutOfTime(c->MinTNS(), c->MaxTNS(), fcolor);

        if (isX && xzview) {
          TPolyLine& p1 = xzview->AddPolyLine(xzZpts.size(),
                                              lcolor,
                                              width,
                                              style);
          TPolyLine& p2 = xzview->AddPolyLine(xzZpts.size(),
                                              lcolor,
                                              width,
                                              style);
          p1.SetOption("f");
          p1.SetFillStyle(3003);
          p1.SetFillColor(fcolor);
          for (unsigned int i=0; i<xzZpts.size(); ++i) {
            p1.SetPoint(i, xzZpts[i], xzXpts[i]);
            p2.SetPoint(i, xzZpts[i], xzXpts[i]);
          } // loop on i points in ZX view
        } // if we have a cluster in the ZX view
        if (isY && yzview) {
          TPolyLine& p1 = yzview->AddPolyLine(yzZpts.size(),
                                              lcolor,
                                              width,
                                              style);
          TPolyLine& p2 = yzview->AddPolyLine(yzZpts.size(),
                                              lcolor,
                                              width,
                                              style);
          p1.SetOption("f");
          p1.SetFillStyle(3003);
          p1.SetFillColor(fcolor);
          for (unsigned int i=0; i<yzZpts.size(); ++i) {
            p1.SetPoint(i, yzZpts[i], yzYpts[i]);
            p2.SetPoint(i, yzZpts[i], yzYpts[i]);
          } // loop on i points in ZY view
        } // if we have a cluster in the ZY view
      }
    } // loop on iClust clusters
  }

  //......................................................................

  void RecoBaseDrawer::Cluster2D(const art::Event& evt,
				 evdb::View2D*     xzview,
				 evdb::View2D*     yzview) 
  {
    art::ServiceHandle<evd::RecoDrawingOptions> drawopt;
    if(drawopt->fClusterStyle == 0) return;

    art::ServiceHandle<evd::SliceNavigator> nav;

    for (unsigned int imod=0; imod<drawopt->fClusterModules.size(); ++imod) {
      const std::string& which = drawopt->fClusterModules[imod];
      const std::string& instance = drawopt->fClusterInstances[imod];
    
      std::vector<const rb::Cluster*> clust;
      nav->GetProducts(evt, which, instance, clust);

      DrawCluster2D(clust, xzview, yzview, drawopt->fClusterIndex[imod]);

    } // loop on imod folders
  }

  //......................................................................

  void RecoBaseDrawer::Cluster3D(const art::Event& /*evt*/,
				 evdb::View3D*     /*view*/) 
  {

  }

  //......................................................................

  void RecoBaseDrawer::OfflineChans2D(const art::Event& evt,
                                      evdb::View2D*     xzview,
                                      evdb::View2D*     yzview)
  {
    art::ServiceHandle<evd::RecoDrawingOptions> drawopt;
    if(drawopt->fOfflineChanOpt == evd::RecoDrawingOptions::kOFF) return;

    art::ServiceHandle<geo::Geometry> geom;
    art::ServiceHandle<evd::SliceNavigator> nav;

    for(unsigned int imod = 0; imod < drawopt->fOfflineChanModules.size(); ++imod){
      const char* which = drawopt->fOfflineChanModules[imod].c_str();
      const std::string& instance = drawopt->fOfflineChanInstances[imod];

      std::vector<const geo::OfflineChan*> chans;
      nav->GetProducts(evt, which, instance, chans);
      for(unsigned int chanIdx = 0; chanIdx < chans.size(); ++chanIdx){
	const geo::OfflineChan* c = chans[chanIdx];

        const geo::View_t view = geom->Plane(c->Plane())->View();

        evdb::View2D* viewobj = (view == geo::kX) ? xzview : yzview;
        if(!viewobj) continue;

        double xyz[3];
        geom->Plane(c->Plane())->Cell(c->Cell())->GetCenter(xyz);
	
	int color = GetClusterColor(imod);
	// color = this->DimOutOfTime(c->MinTNS(), c->MaxTNS(), color);

        viewobj->AddMarker(xyz[2], xyz[view],
                           color,
                           kFullCircle, 0.8);
      } // end for chanIdx
    } // end for imod
  }

  //......................................................................

  ///
  /// Make a set of points which outline a cluster
  ///
  /// @param c      : Reco base cluster to outline
  /// @param view   : Detector view to consider
  /// @param vzXpts : X or Y coordinates of points in XZ or YZ view that outline cluster
  /// @param vzZpts : Z coordinates of points in XZ ir YZ view that outline cluster
  ///
  void RecoBaseDrawer::GetClusterOutlines(const rb::Cluster* c,
                                          int viewint,
					  std::vector<double>& vzVpts,
					  std::vector<double>& vzZpts)
  {
    art::ServiceHandle<geo::Geometry> geo;

    geo::View_t view = geo::View_t(viewint);

    const int ncell = c->NCell(view);
    if(ncell == 0) return;

    // Map plane numbers to highest and lowest cells in the plane
    std::map<int,int> celllo, cellhi;
    // On first pass, initialize
    for (int j=0; j<ncell; ++j) {
      const int plane = c->Cell(view, j)->Plane();
      const int cell = c->Cell(view, j)->Cell();
      celllo[plane] = cell;
      cellhi[plane] = cell;
    }
    // Finalize on second pass
    for (int j=0; j<ncell; ++j) {
      const int plane = c->Cell(view, j)->Plane();
      const int cell = c->Cell(view, j)->Cell();
      if (cell < celllo[plane]) celllo[plane] = cell;
      if (cell > cellhi[plane]) cellhi[plane] = cell;
    }

    // Loop over planes and low cells to make lines along bottom
    // edge. Work from upstream edge to downstream edge
    std::map<int,int>::iterator itr   (celllo.begin());
    std::map<int,int>::iterator itrEnd(celllo.end());
    for (; itr!=itrEnd; ++itr) {
      const int plane = itr->first;
      const int cell  = itr->second;
      double pos[3], dpos[3];
      geo->CellInfo(plane, cell, &view, pos, dpos);

      vzZpts.push_back(pos[2]-dpos[2]);
      vzVpts.push_back(pos[view]-dpos[view]);

      vzZpts.push_back(pos[2]+dpos[2]);
      vzVpts.push_back(pos[view]-dpos[view]);
    }

    // Loop over planes and high cells to make lines along top
    // edge. Work from downstream edge toward upstream edge
    std::map<int,int>::reverse_iterator ritr   (cellhi.rbegin());
    std::map<int,int>::reverse_iterator ritrEnd(cellhi.rend());
    for (; ritr!=ritrEnd; ++ritr) {
      const int plane = ritr->first;
      const int cell  = ritr->second;
      double pos[3], dpos[3];
      geo->CellInfo(plane, cell, &view, pos, dpos);

      vzZpts.push_back(pos[2]+dpos[2]);
      vzVpts.push_back(pos[view]+dpos[view]);

      vzZpts.push_back(pos[2]-dpos[2]);
      vzVpts.push_back(pos[view]+dpos[view]);
    }

    // Add link to starting point to close the box
    vzZpts.push_back(vzZpts[0]);
    vzVpts.push_back(vzVpts[0]);
  }

  //......................................................................

  void RecoBaseDrawer::
  GetClusterOutlinesConvexHull(const rb::Cluster* c,
                               int viewint,
                               std::vector<double>& vzVpts,
                               std::vector<double>& vzZpts) const
  {
    // This is the "wrapping-paper" algorithm

    const geo::View_t view = geo::View_t(viewint);

    std::vector<double> vs;
    std::vector<double> zs;

    art::ServiceHandle<geo::Geometry> geom;

    for(unsigned int cellIdx = 0; cellIdx < c->NCell(view); ++cellIdx){
      const art::Ptr<rb::CellHit>& chit = c->Cell(view, cellIdx);
      const geo::CellGeo* geocell = geom->Plane(chit->Plane())->Cell(chit->Cell());
      double xyz[3];
      geocell->GetCenter(xyz);
      const double v0 = xyz[view];
      const double z0 = xyz[2];
      const double dv = geocell->HalfW();
      const double dz = geocell->HalfD();
      vs.push_back(v0-dv); zs.push_back(z0-dz);
      vs.push_back(v0-dv); zs.push_back(z0+dz);
      vs.push_back(v0+dv); zs.push_back(z0+dz);
      vs.push_back(v0+dv); zs.push_back(z0-dz);
    } // end for cellIdx

    const unsigned int N = vs.size();
    if(N == 0) return;

    unsigned int minIdx = 0;

    double minz = +1e10;
    for(unsigned int n = 0; n < N; ++n){
      if(zs[n] < minz){
        minz = zs[n];
        minIdx = n;
      }
    } // end for n

    vzVpts.push_back(vs[minIdx]);
    vzZpts.push_back(zs[minIdx]);

    // The way the "wrapping-paper" is pointing. We're working clockwise from the far left.
    double dv = 1;
    double dz = 0;
    unsigned int curIdx = minIdx;

    while(true){
      const double curV = vs[curIdx];
      const double curZ = zs[curIdx];

      double bestDot = -2;
      int bestIdx = -1;
      for(unsigned int n = 0; n < N; ++n){
        if(n == curIdx) continue;

        const double ourdV = vs[n]-curV;
        const double ourdZ = zs[n]-curZ;
        const double dot = (ourdV*dv+ourdZ*dz)/util::pythag(ourdV, ourdZ);;
        // Need to give the starting point a little bit of an advantage,
        // otherwise we can miss it out and end up in an infinite loop.
        if(dot > bestDot+((bestIdx == int(minIdx)) ? 1e-6 : 0) ||
           (dot > bestDot-1e-3 && n == minIdx)){
          bestDot = dot;
          bestIdx = n;
        }
      } // end for n

      assert(bestIdx >= 0);

      vzVpts.push_back(vs[bestIdx]);
      vzZpts.push_back(zs[bestIdx]);

      dv = vs[bestIdx]-vs[curIdx];
      dz = zs[bestIdx]-zs[curIdx];
      const double mag = util::pythag(dv, dz);
      dv /= mag;
      dz /= mag;
      curIdx = bestIdx;

      if(curIdx == minIdx) break;
    }

    double area = 0;
    for(unsigned int n = 1; n < vzZpts.size()-2; ++n){
      const double xprod = (vzZpts[n+1]-vzZpts[0])*(vzVpts[n]-vzVpts[0])-(vzVpts[n+1]-vzVpts[0])*(vzZpts[n]-vzZpts[0]);
      assert(xprod > -1e-6);
      area += xprod/2;
    }
  }

  //......................................................................

  void RecoBaseDrawer::
  DrawHoughResults2D(std::vector<const rb::HoughResult*>& htr,
		     evdb::View2D*                        xzv,
		     evdb::View2D*                        yzv)
  {
    art::ServiceHandle<evd::RawDrawingOptions>  rawopt;
    art::ServiceHandle<evd::RecoDrawingOptions> recoopt;
    art::ServiceHandle<geo::Geometry>           geom;
    art::ServiceHandle<evdb::Colors>            colors;

    bool colorByIndex = 
      (recoopt->fHoughOpt == evd::RecoDrawingOptions::kHOUGH_COLORS);
    
    for (unsigned int i=0; i<htr.size(); ++i) {
      double tave = 0.5*(htr[i]->fTNSlo+htr[i]->fTNShi)/1000.0;
      int c = colors->Scale("RawT").GetColor(tave);
      
      for (unsigned int j=0; j<htr[i]->fPeak.size(); ++j) {
        if (colorByIndex) c = this->GetClusterColor(j);
	c = this->DimOutOfTime(htr[i]->fTNSlo, htr[i]->fTNShi, c);
      
	//
	// Extract line segment parameters and intersect with detector sides
	//
	double m, b;
	htr[i]->SlopeIcept(j, &m, &b);

	double x1=0, x2=0;
	if (htr[i]->fView == geo::kX) {
	  x1 =  geom->DetHalfWidth();
	  x2 = -geom->DetHalfWidth();
	}
	else {
	  x1 =  geom->DetHalfHeight();
	  x2 = -geom->DetHalfHeight();
	}
      
	double z1, z2;
	z1 = (x1-b)/m;
	z2 = (x2-b)/m;
      
	if (z1<0.0)               { z1 = 0.0;               x1 = m*z1+b;}
	if (z1>geom->DetLength()) { z1 = geom->DetLength(); x1 = m*z1+b;}
	if (z2<0.0)               { z2 = 0.0;               x2 = m*z2+b;}
	if (z2>geom->DetLength()) { z2 = geom->DetLength(); x2 = m*z2+b;}
      
	if (xzv && htr[i]->fView==geo::kX) {
	  TLine& l = xzv->AddLine(z1, x1, z2, x2);
	  l.SetLineColor(c);
	  l.SetLineStyle(1);
	}

	if (yzv && htr[i]->fView==geo::kY) {
	  TLine& l = yzv->AddLine(z1, x1, z2, x2);
	  l.SetLineColor(c);
	  l.SetLineStyle(1);
	}
      }
    }
  }

  //......................................................................

  void RecoBaseDrawer::DrawVertices2D(std::vector<const rb::Vertex*>& vert,
				      evdb::View2D*     xzview,
				      evdb::View2D*     yzview, 
				      std::vector< int >& vertexIndex)
  {

    art::ServiceHandle<evd::RecoDrawingOptions> drawopt;
    art::ServiceHandle<geo::Geometry> geo;

    const int mstyle   = 28;  //set marker style (cross)
    const double msize = 2.0; //set marker size

    for (unsigned int i=0; i<vert.size(); ++i) {

      bool drawThisVertex = false;
      if( vertexIndex.empty() )
	drawThisVertex = true;
      
      for(int index = 0; index < (int)vertexIndex.size(); index++){
	if( vertexIndex[index] == -1 || vertexIndex[index] == (int)i)
	  drawThisVertex = true;
      }
      
      if(!drawThisVertex) continue;

      const double x = vert[i]->GetX();
      const double y = vert[i]->GetY();
      const double z = vert[i]->GetZ();
      const double t = vert[i]->GetT();
      
      int col = GetClusterColor(i);
      col = this->DimOutOfTime(t, t, col);
      
      if (xzview) xzview->AddMarker(z,x,col,mstyle,msize); 
      if (yzview) yzview->AddMarker(z,y,col,mstyle,msize);
    }
  }

  //......................................................................

  void RecoBaseDrawer::DrawProng2D(std::vector<const rb::Prong*>& prong,
				   evdb::View2D*                  xzview,
				   evdb::View2D*                  yzview,
				   std::vector< int >&            prongIndex) 
  {
    art::ServiceHandle<evd::RecoDrawingOptions> drawopt;
    art::ServiceHandle<geo::Geometry>           geo;
    art::ServiceHandle<evd::RawDrawingOptions>  rawopt;

    // Render prongs at 5 meters by default
    double ds_default = 500.0;

    for (unsigned int i=0; i<prong.size(); ++i) {


      bool drawThisProng = false;
      if( prongIndex.empty() )
	drawThisProng = true;

      for(int index = 0; index < (int)prongIndex.size(); index++){
	if( prongIndex[index] == -1 || prongIndex[index] == (int)i)
	  drawThisProng = true;
      }

      if (xzview && prong[i]->View() != geo::kY && drawThisProng) {
	const double x0 = prong[i]->Start().X();
	const double z0 = prong[i]->Start().Z();
	const double dx = prong[i]->Dir().X();
	const double dz = prong[i]->Dir().Z();
	double       ds = prong[i]->TotalLength();
	if (ds==0.0) ds = ds_default;

	int color = this->GetClusterColor(i);
	color = 
	  this->DimOutOfTime(prong[i]->MinTNS(), prong[i]->MaxTNS(),color);
	
	TLine& pzx = xzview->AddLine(z0, x0, z0+ds*dz, x0+ds*dx);
	pzx.SetLineColor(color);
	pzx.SetLineWidth(2);
	pzx.SetLineStyle(2);
      }
      if (yzview && prong[i]->View() != geo::kX && drawThisProng) {
	const double y0 = prong[i]->Start().Y();
	const double z0 = prong[i]->Start().Z();
	const double dy = prong[i]->Dir().Y();
	const double dz = prong[i]->Dir().Z();
	double       ds = prong[i]->TotalLength();
	if (ds==0.0) ds = ds_default;

	int color = this->GetClusterColor(i);
	color = this->DimOutOfTime(prong[i]->MinTNS(),
				   prong[i]->MaxTNS(),
				   color);
	
	TLine& pzy = yzview->AddLine(z0, y0, z0+ds*dz, y0+ds*dy);
	pzy.SetLineColor(color);
	pzy.SetLineWidth(2);
	pzy.SetLineStyle(2);
      }
    }
  }

  //......................................................................

  void RecoBaseDrawer::DrawTrack2D(std::vector<const rb::Track*>& trks,
				   evdb::View2D*                  xzview,
				   evdb::View2D*                  yzview,
				   std::vector< int >&            trackIndex) 
  {
    art::ServiceHandle<geo::Geometry> geom;
    art::ServiceHandle<evd::RecoDrawingOptions> drawopt;

    bool orthoHits = 
      (drawopt->fTrackOpt & evd::RecoDrawingOptions::kORTHO_HITS);
    
    const size_t trkMax = trks.size();
    for(size_t trkIdx = 0; trkIdx < trkMax; ++trkIdx){
      const rb::Track* trk = trks[trkIdx];

      bool drawThisTrack = false;
      if( trackIndex.empty() )
	drawThisTrack = true;

      for(int index = 0; index < (int)trackIndex.size(); index++){
	if( trackIndex[index] == -1 || trackIndex[index] == (int)trkIdx)
	  drawThisTrack = true;
      }
	
      if(trk->NTrajectoryPoints() < 2 || !drawThisTrack ) continue;
      
      int col = GetClusterColor(trkIdx);
      col = this->DimOutOfTime(trk->MinTNS(), trk->MaxTNS(), col);
      
      TPolyLine* px = 0;
      TPolyLine* py = 0; 
      if(xzview && trk->View() != geo::kY)
	px = &xzview->AddPolyLine(trk->NTrajectoryPoints(), col, 2, 2);
      if(yzview && trk->View() != geo::kX)
	py = &yzview->AddPolyLine(trk->NTrajectoryPoints(), col, 2, 2);

      for(size_t n = 0; n < trk->NTrajectoryPoints(); ++n){
	const double x = trk->TrajectoryPoint(n).X();
	const double y = trk->TrajectoryPoint(n).Y();
	const double z = trk->TrajectoryPoint(n).Z();
      
	if(px) px->SetPoint(n, z, x);
	if(py) py->SetPoint(n, z, y);
      } // end for n

      if (orthoHits) {
	const unsigned int cellMax = trk->NCell();
	for(unsigned int cellIdx = 0; cellIdx < cellMax; ++cellIdx){
	  art::Ptr<rb::CellHit> chit = trk->Cell(cellIdx);

	  const double w = trk->W(chit.get());

	  double xyz[3];
	  geom->Plane(chit->Plane())->Cell(chit->Cell())->GetCenter(xyz);
	  const double z = xyz[2];
	
	  if(chit->View() == geo::kX && yzview){
	    yzview->AddMarker(z, w, col, kFullCircle, .75);
	  }
	  if(chit->View() == geo::kY && xzview){
	    xzview->AddMarker(z, w, col, kFullCircle, .75);
	  }
	} // end for cellIdx
      } // end if
    } // end for trkIdx
  }

  //......................................................................

  void RecoBaseDrawer::DrawProng3D(std::vector<const rb::Prong*>& prong,
				   evdb::View3D*                  view,
				   std::vector< int >&            prongIndex) 
  {

    for (unsigned int i=0; i<prong.size(); ++i) {


      bool drawThisProng = false;
      if( prongIndex.empty() )
	drawThisProng = true;

      for(int index = 0; index < (int)prongIndex.size(); index++){
	if( prongIndex[index] == -1 || prongIndex[index] == (int)i)
	  drawThisProng = true;
      }
      
      if(!drawThisProng ) 
	continue;

      int color = this->GetClusterColor(i);
      color = this->DimOutOfTime(prong[i]->MinTNS(),
				 prong[i]->MaxTNS(),
				 color);
      
      TPolyLine3D& p = view->AddPolyLine3D(2,color,1,1);
    
      const double x0 = prong[i]->Start().X();
      const double y0 = prong[i]->Start().Y();
      const double z0 = prong[i]->Start().Z();

      const double dx = prong[i]->Dir().X();
      const double dy = prong[i]->Dir().Y();
      const double dz = prong[i]->Dir().Z();

      p.SetPoint(0, x0, y0, z0);
      // Prong line is always 5 m long
      p.SetPoint(1, x0+500*dx, y0+500*dy, z0+500*dz);
    }

  }

  //......................................................................

  void RecoBaseDrawer::DrawTrack3D(std::vector<const rb::Track*>& track,
				   evdb::View3D*                  view,
				   std::vector< int >&            trackIndex) 
  {

    for(unsigned int i = 0; i < track.size(); ++i){

      bool drawThisTrack = false;
      if( trackIndex.empty() )
	drawThisTrack = true;

      for(int index = 0; index < (int)trackIndex.size(); index++){
	if( trackIndex[index] == -1 || trackIndex[index] == (int)i)
	  drawThisTrack = true;
      }
      
      if(track[i]->NTrajectoryPoints() < 2 || !drawThisTrack) continue;

      int color = this->GetClusterColor(i);
      color = this->DimOutOfTime(track[i]->MinTNS(),
				 track[i]->MaxTNS(),
				 color);
      TPolyLine3D& p = view->AddPolyLine3D(2,color,1,1);
      for(unsigned int n = 0; n < track[i]->NTrajectoryPoints(); ++n){
	const double x = track[i]->TrajectoryPoint(n).X();
	const double y = track[i]->TrajectoryPoint(n).Y();
	const double z = track[i]->TrajectoryPoint(n).Z();
    
	p.SetPoint(n, x , y , z);
      }
    }
  }

  //......................................................................

  void RecoBaseDrawer::Vertex2D(art::Event const& evt,
				evdb::View2D*     xzview,
				evdb::View2D*     yzview) 
  {    
    art::ServiceHandle<evd::RecoDrawingOptions> drawopt;

    if(drawopt->fVertexOpt == evd::RecoDrawingOptions::kOFF) return;

    art::ServiceHandle<evd::SliceNavigator> nav;

    unsigned int imod;
    for (imod=0; imod<drawopt->fVertexModules.size(); ++imod) {
      std::string const& which = drawopt->fVertexModules[imod];
      std::string const& instance = drawopt->fVertexInstances[imod];
      
      std::vector<const rb::Vertex*> vert;
      nav->GetProducts(evt, which, instance, vert);
      std::vector<int> vertexIndex;
      if( drawopt->fVertexIndex.size() > imod )
	vertexIndex = drawopt->fVertexIndex[imod];
      
	this->DrawVertices2D(vert, xzview, yzview, vertexIndex);
    }
  }

  //......................................................................
  void RecoBaseDrawer::HoughResult2D(art::Event const& evt,
                                     evdb::View2D*     xzview,
                                     evdb::View2D*     yzview)
  {
    art::ServiceHandle<evd::RecoDrawingOptions> drawopt;
    art::ServiceHandle<evd::SliceNavigator> nav;
    unsigned int imod;

    if(drawopt->fHoughOpt == evd::RecoDrawingOptions::kOFF) return;

    for (imod=0; imod<drawopt->fHoughResultModules.size(); ++imod) {
      std::string const& which = drawopt->fHoughResultModules[imod];
      std::string const& instance = drawopt->fHoughResultInstances[imod];

      std::vector<const rb::HoughResult*> htr;
      nav->GetProducts(evt, which, instance, htr);
      this->DrawHoughResults2D(htr, xzview, yzview);
    }
  }

  //......................................................................

  void RecoBaseDrawer::Prong2D(art::Event const& evt,
			       evdb::View2D*     xzview,
			       evdb::View2D*     yzview) 
  {    
    art::ServiceHandle<evd::RecoDrawingOptions> drawopt;
    art::ServiceHandle<evd::SliceNavigator> nav;
    unsigned int imod;

    bool drawProng  =(drawopt->fProngOpt != 0);
    bool drawShowers=(drawopt->fShowerOpt != evd::RecoDrawingOptions::kOFF);

    if (drawProng || drawShowers) {
      for (imod=0; imod<drawopt->fProngModules.size(); ++imod) {
	std::string const& which = drawopt->fProngModules[imod];
        std::string const& instance = drawopt->fProngInstances[imod];

	std::vector<const rb::Prong*> prongs;
        nav->GetProducts(evt, which, instance, prongs);
	std::vector<int> prongIndex;
	if( drawopt->fProngIndex.size() > imod )
	  prongIndex = drawopt->fProngIndex[imod];

        if(drawopt->fProngOpt & evd::RecoDrawingOptions::kAS_CLUSTER){
          std::vector<const rb::Cluster*> clusts(prongs.begin(), prongs.end());
          DrawCluster2D(clusts, xzview, yzview, prongIndex);
        }

        if(drawopt->fProngOpt & evd::RecoDrawingOptions::kAS_PRONG){
          DrawProng2D(prongs, xzview, yzview, prongIndex);
        }
      }
    }
  }

  //......................................................................

  void RecoBaseDrawer::Track2D(const art::Event& evt,
			       evdb::View2D*     xzview,
			       evdb::View2D*     yzview)
  {
    art::ServiceHandle<evd::RecoDrawingOptions> drawopt;
    art::ServiceHandle<evd::SliceNavigator> nav;

    if(drawopt->fTrackOpt == 0) return;

    for(unsigned int imod = 0; imod < drawopt->fTrackModules.size(); ++imod){
      const std::string& which = drawopt->fTrackModules[imod];
      const std::string& instance = drawopt->fTrackInstances[imod];

      std::string whichMichel;
      for(unsigned int jmod = 0; jmod < drawopt->fClusterModules.size(); ++jmod){
        const std::string& clustLabel = drawopt->fClusterModules[jmod];
        bool foundMichel = clustLabel.find("michel") != std::string::npos;
        bool foundTrack  = clustLabel.find(which) != std::string::npos;

        if(foundMichel && foundTrack) whichMichel = clustLabel;
      }

      std::vector<const rb::Track*> tracks;
      if(whichMichel.empty())
        nav->GetProducts(evt, which, instance, tracks);
      else
        nav->GetProducts(evt, which, whichMichel, instance, tracks);

      std::vector<int> trackIndex;
      if(drawopt->fTrackIndex.size() > imod)
        trackIndex = drawopt->fTrackIndex[imod];

      if(drawopt->fTrackOpt & evd::RecoDrawingOptions::kAS_CLUSTER){
        std::vector<const rb::Cluster*> clusts(tracks.begin(), tracks.end());
        DrawCluster2D(clusts, xzview, yzview, trackIndex);
      }

      if(drawopt->fTrackOpt & evd::RecoDrawingOptions::kAS_PRONG){
        std::vector<const rb::Prong*> prongs(tracks.begin(), tracks.end());
        DrawProng2D(prongs, xzview, yzview, trackIndex);
      }

      if(drawopt->fTrackOpt & evd::RecoDrawingOptions::kAS_TRACK){
        DrawTrack2D(tracks, xzview, yzview, trackIndex);
      }
    }
  }

  //......................................................................

  void RecoBaseDrawer::Prong3D(art::Event const& evt,
			       evdb::View3D*     view) 
  {
    art::ServiceHandle<evd::RecoDrawingOptions> drawopt;
    art::ServiceHandle<evd::SliceNavigator> nav;

    bool drawProng  =(drawopt->fProngOpt != 0);
    bool drawShowers=(drawopt->fShowerOpt != evd::RecoDrawingOptions::kOFF);
    bool drawTracks =(drawopt->fTrackOpt != 0);

    if (drawProng || drawShowers) {
      for (unsigned int imod=0; imod<drawopt->fProngModules.size(); ++imod) {
	std::string const& which = drawopt->fProngModules[imod];
        std::string const& instance = drawopt->fProngInstances[imod];
      
	std::vector<const rb::Prong*> prong;
	nav->GetProducts(evt, which, instance, prong);

	std::vector<int> prongIndex;
        if( drawopt->fProngIndex.size() > imod )
          prongIndex = drawopt->fProngIndex[imod];

	DrawProng3D(prong, view, prongIndex);
      }
    }

    if (drawTracks) {
      for (unsigned int imod=0; imod<drawopt->fTrackModules.size(); ++imod) {
	std::string const& whichTrack = drawopt->fTrackModules[imod];
	std::string whichMichel = "";
        std::string const& instance = drawopt->fTrackInstances[imod];

	for (int jmod=0; jmod< (int)drawopt->fClusterModules.size(); ++jmod){
	  
	  std::string const& clustLabel = drawopt->fClusterModules[jmod];
	  std::size_t foundMichel = clustLabel.find("michel");
	  std::size_t foundTrack  = clustLabel.find(whichTrack);
	  
	  if( foundMichel!=std::string::npos && 
	      foundTrack!=std::string::npos)
	    whichMichel = clustLabel;
	}
	

	std::vector<const rb::Track*> track;
	if(whichMichel=="")
	  nav->GetProducts(evt, whichTrack, instance, track);
	else
	  nav->GetProducts(evt, whichTrack, whichMichel, instance, track);

	std::vector<int> trackIndex;
        if( drawopt->fTrackIndex.size() > imod )
          trackIndex = drawopt->fTrackIndex[imod];

	DrawTrack3D(track, view, trackIndex);
      }
    }
  }

  //......................................................................

  int RecoBaseDrawer::GetClusterColor(unsigned int idx) const
  {
    const int kNCol = 12;
    const int kCols[kNCol] = {
      //
      // Skip around the color wheel in RGB triplets
      //
      kRed-4,    kGreen-3, kBlue-7,
      kOrange+1, kTeal+1,  kViolet-1,
      kYellow+1, kCyan+1,  kMagenta+1,
      kSpring+1, kAzure+1, kPink+1
    };
    return kCols[idx%kNCol];
  }
  
  //......................................................................
  
  int RecoBaseDrawer::DimOutOfTime(double t1, double t2, int c) 
  {
    static const double ns2usec = 1e-3;
    t1 *= ns2usec;
    t2 *= ns2usec;

    art::ServiceHandle<evdb::Colors> colors;
    const evdb::ColorScale& cs = colors->Scale("RawT");
    if (cs.InBounds(t1) && cs.InBounds(t2)) return c;
    
    return -1;
  }
  
  //......................................................................

  int RecoBaseDrawer::DimFiltered(const rb::CellHit*              chit, 
				  std::vector<const rb::CellHit*> unfilt, 
				  int c)
  {
    if (std::find(unfilt.begin(),unfilt.end(),chit)==unfilt.end()) {
      return -1;
    };
    return c;
  }

  //......................................................................

  void RecoBaseDrawer::FillTQHisto(art::Event const& evt,
				   TH1F*             thisto,
				   TH1F*             qhisto) 
  {  
    art::ServiceHandle<evd::RecoDrawingOptions> drawopt;
    art::ServiceHandle<evd::SliceNavigator> nav;
 
    for (unsigned int imod=0; imod<drawopt->fCellHitModules.size(); ++imod) {
      std::string const& which = drawopt->fCellHitModules[imod].c_str();
      std::string const& instance = drawopt->fCellHitInstances[imod].c_str();

      std::vector<const rb::CellHit*> cellhits;
      nav->GetProducts(evt, which, instance, cellhits);

      for (unsigned int i=0; i<cellhits.size(); ++i) {
	const rb::CellHit* h = cellhits[i];

	// best available charge, time
	float q=h->PE();
	//       if (h->IsA()==rb::RecoHit::Class()) {
	//         q = ((rb::RecoHit*)h)->PECorr(); // or GeV or MIP?
	//       }
	float t=(h->TNS())/1000.0; // thisto is binned in usec
	//       if (h->IsA()==rb::RecoHit::Class()) {
	//         t = ((rb::RecoHit*)h)->T();
	//       }

	if (thisto!=0) thisto->Fill(t);
	if (qhisto!=0) qhisto->Fill(q);
      }
    }

  }

} // end namespace evd
////////////////////////////////////////////////////////////////////////
