////////////////////////////////////////////////////////////////////////
// \file   SliceButtonBar.cxx
// \brief  Controls for navigating between slices
// \author Christopher Backhouse - bckhouse@caltech.edu
////////////////////////////////////////////////////////////////////////

#include "EventDisplay/SliceButtonBar.h"

#include "TGButton.h"
#include "TGLabel.h"
#include "TGTextEntry.h"

#include "art/Framework/Services/Registry/ServiceHandle.h"

#include "EventDisplay/SliceNavigator.h"

namespace evd
{
  //---------------------------------------------------------------
  SliceButtonBar::SliceButtonBar(const TGWindow* p)
    : TGCompositeFrame(p, 60, 20, kSunkenFrame | kHorizontalFrame)
  {
    art::ServiceHandle<evd::SliceNavigator> nav;

    TGLabel* label = new TGLabel(this, "Slices: ");
    AddFrame(label, new TGLayoutHints(kLHintsCenterY | kLHintsLeft,
                                      0, 0, 5, 1));

    fPrevButton = new TGTextButton(this, "<< Previous", 50);
    fPrevButton->SetToolTipText("Go to previous slice");
    fPrevButton->Connect("Clicked()",
                         "evd::SliceButtonBar", this, "Previous()");
    AddFrame(fPrevButton, new TGLayoutHints(kLHintsTop | kLHintsLeft,
                                            2, 0, 2, 2));

    fNextButton = new TGTextButton(this, "Next >>", 50);
    fNextButton->SetToolTipText("Go to next slice");
    fNextButton->Connect("Clicked()", "evd::SliceButtonBar", this, "Next()");
    AddFrame(fNextButton, new TGLayoutHints(kLHintsTop | kLHintsLeft,
                                            2, 0, 2, 2));

    fEnableButton = new TGCheckButton(this, "Slice-by-slice", 0);
    AddFrame(fEnableButton, new TGLayoutHints(kLHintsCenterY | kLHintsLeft,
                                              2, 0, 2, 2));
    fEnableButton->Connect("Clicked()",
                           "evd::SliceButtonBar", this, "Enable()");
    fEnableButton->SetState(nav->FullEventMode() ? kButtonUp : kButtonDown);

    fNSlicesLabel = new TGLabel(this, " / 1 ");
    AddFrame(fNSlicesLabel, new TGLayoutHints(kLHintsCenterY | kLHintsRight,
                                              0, 0, 5, 1));

    fSliceTextEntry = new TGTextEntry(this, new TGTextBuffer(128));
    fSliceTextEntry->Connect("ReturnPressed()", "evd::SliceButtonBar",
                             this, "GoTo()");
    fSliceTextEntry->Resize(50, 20);
    fSliceTextEntry->SetText(TString::Format("%d", nav->CurrentSlice()));
    AddFrame(fSliceTextEntry,
             new TGLayoutHints(kLHintsTop | kLHintsRight, 2, 0, 2, 2));

    TGLabel* sliceLabel = new TGLabel(this, "[Slice]=");
    AddFrame(sliceLabel,
             new TGLayoutHints(kLHintsCenterY | kLHintsRight, 2, 0, 2, 2));

    // Set everything to the appropriate enabled state
    Enable();
  }

  //---------------------------------------------------------------
  SliceButtonBar::~SliceButtonBar()
  {
    delete fPrevButton;
    delete fNextButton;
    delete fEnableButton;
    delete fSliceTextEntry;
    delete fNSlicesLabel;
  }

  //---------------------------------------------------------------
  void SliceButtonBar::Enable()
  {
    art::ServiceHandle<evd::SliceNavigator> nav;
    nav->SetFullEventMode(fEnableButton->GetState() == kButtonUp);

    Update();
  }

  //---------------------------------------------------------------
  void SliceButtonBar::Previous()
  {
    art::ServiceHandle<evd::SliceNavigator> nav;
    nav->Previous();
  }

  //---------------------------------------------------------------
  void SliceButtonBar::Next()
  {
    art::ServiceHandle<evd::SliceNavigator> nav;
    nav->Next();
  }

  //---------------------------------------------------------------
  void SliceButtonBar::GoTo()
  {
    art::ServiceHandle<evd::SliceNavigator> nav;
    nav->SetCurrentSlice(atoi(fSliceTextEntry->GetText()));
  }

  //---------------------------------------------------------------
  void SliceButtonBar::Update()
  {
    art::ServiceHandle<evd::SliceNavigator> nav;

    fSliceTextEntry->SetText(TString::Format("%d", nav->CurrentSlice()));
    fNSlicesLabel->SetText(TString::Format(" / %d ", nav->MaxSlice()));

    const bool enable = fEnableButton->GetState() != kButtonUp;
    fSliceTextEntry->SetEnabled(enable);
    fPrevButton->SetEnabled(enable);
    fNextButton->SetEnabled(enable);

    if(!enable) fSliceTextEntry->SetText("N/A");
  }
}
