////////////////////////////////////////////////////////////////////////
///
/// \brief  GENIE Rock events generator
/// \author rhatcher@fnal.gov
/// \date
///
////////////////////////////////////////////////////////////////////////

#include <cassert>
#include <cstdlib>
#include <cmath>
#include <string>
#include <sstream>
#include <vector>
#include <map>
#include <unistd.h>

// ROOT includes
#include "TStopwatch.h"

// Framework includes
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/SubRun.h"
#include "fhiclcpp/ParameterSet.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "messagefacility/MessageLogger/MessageLogger.h"
#include "art/Framework/Core/EDProducer.h"
#include "art/Framework/Core/ModuleMacros.h"

// NOvA includes
#include "EventGeneratorBase/GENIE/GENIEHelper.h"
#include "Geometry/Geometry.h"
#include "RawData/DAQHeader.h"
#include "SimulationBase/GTruth.h"
#include "SimulationBase/MCFlux.h"
#include "SimulationBase/MCParticle.h"
#include "SimulationBase/MCTruth.h"
#include "SummaryData/POTSum.h"
#include "SummaryData/RunData.h"
#include "Utilities/AssociationUtil.h"


namespace evgen {
  /// A module to check the results from the Monte Carlo generator
  class GENIERockGen : public art::EDProducer {
  public:
    explicit GENIERockGen(fhicl::ParameterSet const &pset);
    virtual ~GENIERockGen();                        

    void produce(art::Event& evt);  
    void beginJob();  
    void beginRun(art::Run &run);
    void endSubRun(art::SubRun &sr);

  private:

    evgb::GENIEHelper  *fGENIEHelp;      ///< GENIEHelper object
    TStopwatch          fStopwatch;      ///keep track of how long it takes to run the job
    int                 fSpillCounter;
  };
};

namespace evgen{

  //___________________________________________________________________________
  GENIERockGen::GENIERockGen(fhicl::ParameterSet const &pset) 
    : fSpillCounter(0)
  {  
    fStopwatch.Start();

    produces< std::vector<simb::MCTruth> >();
    produces< std::vector<simb::MCFlux>  >();
    produces< std::vector<simb::GTruth>  >();
    produces< sumdata::POTSum, art::InSubRun  >();
    produces< sumdata::RunData, art::InRun    >();
    // Associate every truth with the flux it came from
    produces< art::Assns<simb::MCTruth, simb::MCFlux> >();
    produces< art::Assns<simb::MCTruth, simb::GTruth> >();

    art::ServiceHandle<geo::Geometry> geo;
    fGENIEHelp = new evgb::GENIEHelper(pset, 
				       geo->ROOTGeoManager(),
				       geo->ROOTFile(),
				       geo->TotalMass(pset.get< std::string>("TopVolume").c_str()));
  }

  //___________________________________________________________________________
  GENIERockGen::~GENIERockGen()
  {  
    fStopwatch.Stop();
    mf::LogInfo("GENIERockGen") << "Real time to produce file: " << fStopwatch.RealTime();
  }

  //___________________________________________________________________________
  void GENIERockGen::beginJob()
  {
  }

  //___________________________________________________________________________
  void GENIERockGen::beginRun(art::Run& run)
  {

    // grab the geometry object to see what geometry we are using
    art::ServiceHandle<geo::Geometry> geo;

    std::unique_ptr<sumdata::RunData> runcol(new sumdata::RunData(geo->DetId(),
								geo->FileBaseName()));

    run.put(std::move(runcol));

    // initialize the GENIEHelper here rather than in beginJob to avoid problems with
    // the Geometry reloading at a run boundary.  If we ever make more than one run
    // in a single job we will have to re-evaluate
    fGENIEHelp->Initialize();

    return;
  }

  //___________________________________________________________________________

  void GENIERockGen::endSubRun(art::SubRun &sr)
  {

    std::unique_ptr< sumdata::POTSum > p(new sumdata::POTSum);
    
    p->totpot = fGENIEHelp->TotalExposure();
    p->totgoodpot = fGENIEHelp->TotalExposure();
    p->totspills = fSpillCounter;
    p->goodspills = fSpillCounter;
    p->Print(std::cout);

    sr.put(std::move(p));
      

  }

  //___________________________________________________________________________
  void GENIERockGen::produce(art::Event& evt)
  {

    ///check if we are going to make events for this spill - could be 
    ///that we are using flux histograms and the poisson number for 
    ///how many interactions to make is 0
    if(fGENIEHelp->Stop()){
      throw cet::exception("NoNuInteraction")
        << "no neutrino interaction made for this spill\n"
        << __FILE__ << ":" << __LINE__ << "\n";
      return;
    }

    std::unique_ptr< std::vector<simb::MCTruth> > truthcol(new std::vector<simb::MCTruth>);
    std::unique_ptr< std::vector<simb::MCFlux>  > fluxcol (new std::vector<simb::MCFlux >);
    std::unique_ptr< std::vector<simb::GTruth>  > gtruthcol (new std::vector<simb::GTruth >);
    std::unique_ptr< art::Assns<simb::MCTruth, simb::GTruth> > tgtassn(new art::Assns<simb::MCTruth, simb::GTruth>);
    std::unique_ptr< art::Assns<simb::MCTruth, simb::MCFlux> > assns(new art::Assns<simb::MCTruth, simb::MCFlux>);

    art::ServiceHandle<geo::Geometry> geo;

    while(!fGENIEHelp->Stop()){
    
      simb::MCTruth truth;
      simb::MCFlux  flux;
      simb::GTruth  gTruth;

      ///GENIEHelper returns a false in the sample method if 
      ///either no neutrino was generated, or the interaction
      ///occurred beyond the detector's z extent - ie something we
      ///would never see anyway.
      if(fGENIEHelp->Sample(truth, flux, gTruth)){

 
	const simb::MCParticle nu = truth.GetNeutrino().Nu();

	if(fabs(nu.Vx())>geo->DetHalfWidth() 
	   || fabs(nu.Vy())>geo->DetHalfHeight() 
	   || nu.Vz()> geo->DetLength() 
	   || nu.Vz()<0 ){


	    truthcol->push_back(truth);
	    gtruthcol->push_back(gTruth);
	    fluxcol->push_back(flux);

	    util::CreateAssn(*this, evt, *truthcol, *fluxcol, *assns,
			     fluxcol->size()-1, fluxcol->size());
	    
	    util::CreateAssn(*this, evt, *truthcol, *gtruthcol, *tgtassn,
			     gtruthcol->size()-1, gtruthcol->size());
	}

      }///end if genie was able to make an event


    }///end event generation loop

    // put the collections in the event 
    evt.put(std::move(truthcol));
    evt.put(std::move(fluxcol));
    evt.put(std::move(gtruthcol));
    evt.put(std::move(assns));
    evt.put(std::move(tgtassn));

    return;
  }

}//namespace

namespace evgen { DEFINE_ART_MODULE(GENIERockGen); }
