// #include "EventGenerator/GENIE/NuMISpillTimeStructure.h"
#include "EventGenerator/GENIE/NuMISpillTimeStructure.h"
#include <cassert>
#include <iostream>
#include "TMath.h"
#include "TRandom.h"
using namespace evgen;

NuMISpillTimeStructure::NuMISpillTimeStructure() :
  //
  // These numbers specify the spill structure for NuMI
  //
  // * There are 84 RF buckets in a single batch
  // * There are 6 filled buckets in a single spill
  // * Within a batch, the first 81 buckets are filled
  //
  // * 53.103 MHz is the RF frequency of Main Injector at 120 GeV
  //
  // * The RF sine wave localizes the protons to the center of the
  //   bucket with a ~gaussian time profile of width ~750 ps (e-mail
  //   from Phil Adamson (May 28, 2015)
  //
  // * There is some variation bucket-to-bucket in the number of
  //   protons. This is estimated to be about +/- 15% with a gasussian
  //   spread based on a quick look at some of the data from the
  //   resistive wall monitors on the fast NuMI DAQ
  //
  // messier@indiana.edu (June 1, 2015)
  //
  fNbatch         (6),
  fNbucketPerBatch(84),
  fNfilledBucket  (81),
  fTbucket        (1e9/53.103E6),
  fBucketSigmaT   (0.750),
  fBucketSigmaPOT (0.15),
  fNbinPerBucket  (50),
  fNbin           (2*fNbinPerBucket+fNbucketPerBatch*fNbatch*fNbinPerBucket),
  fPOTbyBatch     (fNbatch),
  fSpillPOT       (fNbin),
  fSpillPOTI      (fNbin)
{ 
  fPOTbyBatch[0] = 1.9;
  fPOTbyBatch[1] = 1.9;
  fPOTbyBatch[2] = 1.0;
  fPOTbyBatch[3] = 1.0;
  fPOTbyBatch[4] = 1.0;
  fPOTbyBatch[5] = 1.0;

  for (unsigned int i=0; i<fSpillPOT. size(); ++i) fSpillPOT[i]  = 0;
  for (unsigned int i=0; i<fSpillPOTI.size(); ++i) fSpillPOTI[i] = 0;
}

//.....................................................................

double NuMISpillTimeStructure::BinToT(unsigned int i) const
{
  return fTbucket*((float)i-(float)fNbinPerBucket+0.5)/(float)fNbinPerBucket;
}

//.....................................................................

void NuMISpillTimeStructure::NewSpill()
{
  unsigned int i, j, k;
  unsigned int it, it0;
  double t, t0;
  double pot, fpot;

  //
  // Figure out the relative number of protons in each time bin
  //
  it = fNbinPerBucket;
  for (i=0; i<fNbatch; ++i) {
    pot = fPOTbyBatch[i];
    for (j=0; j<fNbucketPerBatch; ++j) {
      it0 = it+fNbinPerBucket/2;
      t0  = this->BinToT(it0);
      fpot = gRandom->Gaus(1.0, fBucketSigmaPOT);
      for (k=0; k<fNbinPerBucket; ++k) {
	if (j>fNfilledBucket) {
	  fSpillPOT[it] = 0.0;
	}
	else {
	  t = this->BinToT(it);
	  fSpillPOT[it] = pot*fpot*TMath::Gaus(t-t0,0,fBucketSigmaT);
	}
	++it;
	assert(it<=fNbin);
      }
    }
  }

  //
  // Now compute the integral and normalize it to 1 so we can sample
  // the timing distribution
  //
  fSpillPOTI[0] = fSpillPOT[i];
  for (i=1; i<fSpillPOTI.size(); ++i) {
    fSpillPOTI[i] = fSpillPOTI[i-1] + fSpillPOT[i];
  }
  for (i=0; i<fSpillPOTI.size(); ++i) {
    fSpillPOTI[i] /= fSpillPOTI[fSpillPOTI.size()-1];
  }
}

//......................................................................

double NuMISpillTimeStructure::GetTime() const
{
  unsigned int i;
  double r;
  r = gRandom->Uniform();
  for (i=0; i<fSpillPOTI.size(); ++i) {
    if (fSpillPOTI[i]>r) return this->BinToT(i);
  }
  return 0.0;
}

////////////////////////////////////////////////////////////////////////
