////////////////////////////////////////////////////////////////////////
/// \brief  Dumps a summary of info about events to the screen
/// \author Michael Baird - mbaird42@fnal.gov
/// \date   June 2011
////////////////////////////////////////////////////////////////////////

#include <vector>
#include <iostream>
#include <sstream>
#include <map>
#include <cmath>

// NOvASoft includes

#include "ChannelInfo/BadChanList.h"        // needed for bad channel mask
#include "CMap/CMap.h"                      // needed to get number of hits info
#include "DAQDataFormats/TriggerDefines.h"  // needed to print trigger ID
#include "Geometry/Geometry.h"
#include "RawData/RawDigit.h"
#include "RawData/RawTrigger.h"
#include "RecoBase/Cluster.h"
#include "RecoBase/HoughResult.h"
#include "RecoBase/Prong.h"
#include "RecoBase/Track.h"
#include "RecoBase/Vertex.h"
#include "SummaryData/SpillData.h"          // needed for getting spill data
#include "SimulationBase/MCTruth.h"         // needed to get MCTruth info

// ROOT includes
#include "TDatabasePDG.h"                   // needed to print particle names from pdg codes
#include "TTimeStamp.h"                     // needed for printing the date/time

// Framework includes
#include "art/Framework/Core/EDAnalyzer.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Services/Optional/TFileDirectory.h"
#include "art/Framework/Services/Optional/TFileService.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Persistency/Common/PtrVector.h"
#include "art/Persistency/Common/Ptr.h"
#include "fhiclcpp/ParameterSet.h"
#include "messagefacility/MessageLogger/MessageLogger.h"





// EventSummary header
namespace evtsum {
  class EventSummary : public art::EDAnalyzer {
    public:
      explicit EventSummary(fhicl::ParameterSet const& pset);
      explicit EventSummary(int i1, int i2, int i3, int i4, int i5, int i6,
			    std::string s1, std::string s2,
			    std::string s3, std::string s4,
			    std::string s5, std::string s6,
			    std::string s7, std::string s8,
			    std::string s9,
			    fhicl::ParameterSet const& pset);
			         
      ~EventSummary();                        
      
      void analyze(art::Event const & evt);

      void reconfigure(const fhicl::ParameterSet& pset);

      void PrintInfo(art::Event const & evt);

      void MakeOutput(art::Event const & evt, std::ostringstream & os);
      
    private:
      // configuration parameters

      int              fBasicInfoPrint;     ///< flag variable to tell module to print basic info
      int              fHitCountPrint;      ///< flag variable to tell module to print hit count info
      int              fSpillPrint;         ///< flag variable to tell module to print spill info
      int              fMCPrint;            ///< flag variable to tell module to print MC truth info
      int              fRecoPrint;          ///< flag variable to tell module to print reconstruction info
      int              fExtraRecoPrint;     ///< flag variable to tell module to print extended reconstruction info

      std::string      fRawDigitLabel;      ///< label for module that made RawDigit
      std::string      fRawTriggerLabel;    ///< label for module that made RawTrigger
      std::string      fSpillLabel;         ///< label for module that made SpillData
      std::string      fMCLabel;            ///< label for module that made MCTruth
      std::string      fHoughLabel;         ///< label for module that made HoughResults
      std::string      fVertexLabel;        ///< label for module that made vertices
      std::string      fClusterLabel;       ///< label for module that made clusters 
      std::string      fProngLabel;         ///< label for module that made prongs
      std::string      fTrackLabel;         ///< label for module that made tracks 

    };
}



namespace evtsum {

  EventSummary::EventSummary(fhicl::ParameterSet const& pset) 
    : EDAnalyzer(pset)
  {
    
    // constructor

    this->reconfigure(pset);
    
    // Be noisy to demonstrate what's happening
    mf::LogInfo("EventSummary") << " EventSummary::EventSummary()\n";

  }

  EventSummary::EventSummary(int i1, int i2, int i3, int i4, int i5, int i6,
			     std::string s1, std::string s2,
			     std::string s3, std::string s4,
			     std::string s5, std::string s6,
			     std::string s7, std::string s8,
			     std::string s9,
			     fhicl::ParameterSet const& pset)
    : EDAnalyzer(pset)
  {
    // alternate constructor
    fBasicInfoPrint   = i1;
    fHitCountPrint    = i2;
    fSpillPrint       = i3;
    fMCPrint          = i4;
    fRecoPrint        = i5;
    fExtraRecoPrint   = i6;
    
    // set to default values
    if(s1 == "") fRawDigitLabel   = "daq";
    else         fRawDigitLabel   = s1;

    if(s2 == "") fRawTriggerLabel = "daq";
    else         fRawTriggerLabel = s2;

    if(s3 == "") fSpillLabel      = "ifdbspillinfo";
    else         fSpillLabel      = s3;

    if(s4 == "") fMCLabel         = "generator";
    else         fMCLabel         = s4;

    if(s5 == "") fHoughLabel      = "MultiHoughT";
    else         fHoughLabel      = s5;

    if(s6 == "") fVertexLabel     = "HoughVertexFinder";
    else         fVertexLabel     = s6;

    if(s7 == "") fClusterLabel    = "slicer";
    else         fClusterLabel    = s7;    

    if(s8 == "") fProngLabel      = "KalmanTrack";
    else         fProngLabel      = s8;

    if(s9 == "") fTrackLabel      = "trackmerge";
    else         fTrackLabel      = s9;
  }
    
    


  //......................................................................
  EventSummary::~EventSummary()
  {
    //======================================================================
    // Clean up any memory allocated by your module
    //======================================================================
  }



  //......................................................................
  void EventSummary::reconfigure(const fhicl::ParameterSet& pset)
  { // get the input parameters from EventSummary.fcl
    fBasicInfoPrint   = pset.get< int    >("BasicInfoPrint");
    fHitCountPrint    = pset.get< int    >("HitCountPrint");
    fSpillPrint       = pset.get< int    >("SpillPrint");
    fMCPrint          = pset.get< int    >("MCPrint");
    fRecoPrint        = pset.get< int    >("RecoPrint");
    fExtraRecoPrint   = pset.get< int    >("ExtraRecoPrint");
    
    fRawDigitLabel    = pset.get< std::string >("RawDigitLabel");
    fRawTriggerLabel  = pset.get< std::string >("RawTriggerLabel");
    fSpillLabel       = pset.get< std::string >("SpillLabel");
    fMCLabel          = pset.get< std::string >("MCLabel");
    fHoughLabel       = pset.get< std::string >("HoughLabel");
    fVertexLabel      = pset.get< std::string >("VertexLabel");
    fClusterLabel     = pset.get< std::string >("ClusterLabel");
    fProngLabel       = pset.get< std::string >("ProngLabel");
    fTrackLabel       = pset.get< std::string >("TrackLabel");
  }


  //......................................................................
  void EventSummary::analyze(art::Event const & evt)
  { 
    
    this->PrintInfo(evt);

    return;

  }



  //......................................................................
  void EventSummary::PrintInfo(art::Event const & evt)
  {

    std::ostringstream OutputStream;

    this->MakeOutput(evt, OutputStream);

    std::cout << OutputStream.str();

    return;

  }



  //......................................................................
  void EventSummary::MakeOutput(art::Event const & evt, std::ostringstream & os)
  {

    // This module works in several steps:
    //
    // 1. read in raw digits
    // 2. disect the timestamp into date/time
    // 3. count hits per DCM
    // 4. count hits w/ bad channels masked in X,Y views
    // 5. get trigger and spill information
    // 6. count hits in NuMI or BNB spill windows
    // 7. get MC truth info
    // 8. get reconstruction info
    // 9. create string of event summary info





    //
    // read in the rawdigits
    //
    art::Handle< std::vector<rawdata::RawDigit> > digitcol;
    evt.getByLabel(fRawDigitLabel, digitcol);
    
    art::PtrVector<rawdata::RawDigit> digitlist;
    for(unsigned int i = 0; i < digitcol->size(); ++i) {
      art::Ptr<rawdata::RawDigit> rd(digitcol, i);
      digitlist.push_back(rd);
    }
    //
    // end reading in raw digits




    
    //
    // DISECTING the time from evt.time().value() into "human readable" format to display the date & time
    // taken from EventDisplay/HeaderDrawer.cxx
    //
    unsigned int year, month, day, dayofweek;
    unsigned int hour, minute, second;
    int          nano;

    // get the time stamp.  art::Timestamp::value() returns a TimeValue_t which is a typedef to unsigned long long.
    // The conventional use is for the upper 32 bits to have the seconds since 1970 epoch and the lower 32 bits to be
    // the number of microseconds with the current second.
    unsigned long long int tsval = evt.time().value();

    // taking it apart
    const unsigned long int mask32 = 0xFFFFFFFFUL;
    unsigned long int lup          = ( tsval >> 32 ) & mask32;
    unsigned long int llo          = tsval & mask32;
    TTimeStamp ts(lup, (int)llo);
    ts.GetDate(kTRUE,0,&year,&month,&day);
    ts.GetTime(kTRUE,0,&hour,&minute,&second);
    nano = ts.GetNanoSec();
    dayofweek = ts.GetDayOfWeek();

    double sec = (second + nano/1.0e9);

    // define some arrays to aid in printing
    static const char* days[] = {"","Mon","Tue","Wed","Thu","Fri","Sat","Sun"};
    static const char* months[] = {"","Jan","Feb","Mar","Apr","May","Jun","Jul","Aug","Sep","Oct","Nov","Dec"};
    //
    // end section for DISECTING the time





    //
    // counting hits by DCM (before bad channel mask is applied)
    // This is used to determine the number of DCM's reporting information for
    // each event.
    // (modeled after ChannelInfo/HotChannels.cxx)
    //
    art::ServiceHandle<geo::Geometry> geom;
    art::ServiceHandle<cmap::CMap>    cmap;

    const int detid = geom->DetId();

    const int nDCM = 200;               // a number big enough to include the number of DCM's in the far detector
    unsigned int DCMHits[nDCM] = {0};   // count of hits per DCM

    for (unsigned int i = 0; i < digitlist.size(); ++i) {
      // for each digit, get plane and cell and determine DCM ID with getDCMGlobal
      const uint32_t plane  = cmap->GetPlane(digitlist[i].get());
      const uint32_t cell   = cmap->GetCell(digitlist[i].get());

      const daqchannelmap::lchan LogicalChannel = cmap->Map()->encodeLChan(detid, plane, cell);
      const daqchannelmap::dchan DAQChannel     = cmap->Map()->encodeDChan(LogicalChannel);
      const uint32_t             DCM            = cmap->Map()->getDCM(DAQChannel);   // this is the DCM id number
      const uint32_t             diblock_number = cmap->Map()->getDiBlock(DAQChannel);

      /// Some unique number assigned for DCM in the detector
      const uint32_t DCMGlobal = diblock_number*12 + DCM;

      ++DCMHits[DCMGlobal];
    }

     // count of number of DCMs that reported hits
    int DCMCount = 0;
    for(int j = 1; j < nDCM; j++) {   
      if(DCMHits[j] != 0) DCMCount++;
    }
    //
    // end counting hits by DCM





    //
    // counting number of hits with bad channels masked
    // 
    unsigned int rdsnm = digitlist.size();     // rdsnm = "raw digit size no mask"

    // apply bad channel mask
    art::ServiceHandle<chaninfo::BadChanList> bcl;
    bcl->Apply(digitlist);

    // counting number of hits in x and y views
    // (modeled after util::XYCount)
    unsigned int nx = 0;
    unsigned int ny = 0;

    for (unsigned int i = 0; i < digitlist.size(); ++i) {
      unsigned int p = cmap->GetPlane(digitlist[i].get());
      geo::View_t  v = geom->Plane(p)->View();

      if(v == geo::kX) ++nx;
      else if(v == geo::kY) ++ny;
    }
    //
    // end counting number of hits




   
    //
    // get trigger info
    //
    unsigned int fTrigger       = 0;
    unsigned int fTriggerLength = 0;
    float        trigwin        = 0.0;

    art::Handle< std::vector<rawdata::RawTrigger> > trigv;
    evt.getByLabel(fRawTriggerLabel, trigv);
        
    fTrigger       = (*trigv)[0].fTriggerMask_TriggerType;
    fTriggerLength = (*trigv)[0].fTriggerRange_TriggerLength;
    
    // not sure what the units on fTriggerLength are (assuming it is in units of 500 ns)
    // so we will convert it to micro-sec.
    trigwin        = fTriggerLength*0.5;
    //
    // end getting trigger info



    
    
    //
    // get spill info
    //
    art::Handle< sumdata::SpillData > spill;
    bool spillFlag = 0;

    try {
      evt.getByLabel(fSpillLabel, spill);
      spillFlag = spill.isValid();
    }
    catch(...) {}


    //
    // count hits in NuMI or BNB spill windows
    //

    // define NuMI and BNB spill window times (in usec)
    float t1     = 0.0;
    float t2     = 0.0;
    int incount  = 0;

    // conversion factor for TDC to micro-sec
    static const double kUSEC_PER_TDC = (1.0e6/64.0e6);

    // get spill start time
    //unsigned int tdc0 = (*trigv)[0].TDCT0();
    
    // count hits in NuMI spill window
    if(fTrigger == 0) {
      // define the edges of the NuMI spill window (in micro-sec)
      t1 = 216.0;
      t2 = 228.0;
      for(unsigned int i = 0; i < digitlist.size(); ++i) {

	// time of hit with respect to start of spill in micro-seconds
	float hittime = (digitlist[i]->TDC())*kUSEC_PER_TDC;

	if(hittime >= t1 && hittime <= t2) incount++;	
      }   // end loop over i
    }   // end if(fTrigger == 0)

    // count hits in BNB spill window
    if(fTrigger == 1) {
      // define the edges of the BNB spill window in (micro-sec)
      t1 = 332.0;
      t2 = 334.0;
      for(unsigned int i = 0; i < digitlist.size(); ++i) {
	
	// time of hit with respect to start of spill in micro-seconds
	float hittime = (digitlist[i]->TDC())*kUSEC_PER_TDC;

	if(hittime >= t1 && hittime <= t2) incount++;	
      }   // end loop over i
    }   // end if(fTrigger == 1)
    //
    // end count hits in spill windows





    //
    // get the MC truth info
    //
    art::Handle< std::vector<simb::MCTruth> > mclist;
    bool MCFlag = 0;

    try {

      evt.getByLabel(fMCLabel, mclist);
      
      // some MC events are created with size zero (I'm not sure why) which causes
      // a seg fault so we will test to make sure this isn't the case before moving on      
      if(mclist->size() > 0) MCFlag  = 1;

    }   // end of try
    catch(...) {}  
    //
    // end get MC truth info





    //
    // get the reconstruction info
    //

    // get HoughResults
    art::Handle< std::vector< rb::HoughResult > > hr;
    bool HoughFlag = 0;
    unsigned int hrXsz = 0, hrYsz = 0;
    
    try {
      evt.getByLabel(fHoughLabel, hr);
      
      for(unsigned int i = 0; i < hr->size(); ++i) {
	if((*hr)[i].fView == geo::kX) hrXsz += (*hr)[i].fPeak.size();	
	if((*hr)[i].fView == geo::kY) hrYsz += (*hr)[i].fPeak.size();
      }

      HoughFlag = hr.isValid();
    }   // end of try
    catch(...) {}

    // get vertices
    art::Handle< std::vector< rb::Vertex > > vertices;
    bool VertexFlag = 0;

    try {
      evt.getByLabel(fVertexLabel, vertices);
      VertexFlag = vertices.isValid();
    }
    catch(...) {}

    // get cluster results
    art::Handle< std::vector< rb::Cluster > > cluster;
    bool ClusterFlag = 0;

    try {
      evt.getByLabel(fClusterLabel, cluster);
      ClusterFlag = cluster.isValid();
    }
    catch(...) {}

    // get prong results
    art::Handle< std::vector< rb::Prong > > prong;
    bool ProngFlag = 0;

    try {
      evt.getByLabel(fProngLabel, prong);
      ProngFlag = prong.isValid();
    }
    catch(...) {}

    // get track results
    art::Handle< std::vector< rb::Track > > track;
    bool TrackFlag = 0;

    try {
      evt.getByLabel(fTrackLabel, track);
      TrackFlag = track.isValid();
    }
    catch(...) {}

    //
    // end get reco info
   



    //
    // printing the EventSummary information
    //
    
    os << "\n\n||==================== Begin EventSummary output ====================||\n\n";

    // printing basic info
    if(fBasicInfoPrint != 0) {
      os << "Date/Time:        " << days[dayofweek] 
	 << " " << months[month] << " " << day << ", " << year << " / ";

      if(hour < 10) os << "0";
      os << hour << ":";

      if(minute < 10) os << "0";
      os << minute << ":";

      if(sec < 10.0) os << "0";
      os << sec << " UTC\n";

      os << "Event ID:         " << evt.id() << "\n";

      os << "Trigger ID:       " << fTrigger
         << " (" << daqdataformats::trigIDStrings[fTrigger] << ")\n";

      os << "Trigger window:   " << trigwin << " (micro-sec)\n\n";
    }   // end print basic info



    // printing hit count info
    if(fHitCountPrint != 0) {
  
      os << "number of DCMs reporting:                    " << DCMCount << "\n";
      os << "total number of hits (no masks applied):     " << rdsnm << "\n";
      os << "number of hits (with bad channels masked):   " << digitlist.size() << "\n";
      os << "number of XY hits (x-view,y-view):           (" << nx << "," << ny << ")\n";
        
    if(fTrigger == 0) os << "number of hits in NuMI spill window:         "  << incount << "\n";
    if(fTrigger == 1) os << "number of hits in BNB spill window:          "  << incount << "\n";    
    }   // end print hit count info
  


    // printing spill info
    if(fSpillPrint != 0) {
      os << "\nSpill data:";
      if(spillFlag)
	os << "\n===========\n"
	   << "time (nsec):         " << spill->spilltimensec << "\n"
	   << "delta time (nsec):   " << spill->deltaspilltimensec << "\n" 
 	   << "POT:                 " << spill->spillpot << " x 10^12" << "\n"
	   << "horn current:        " << spill->hornI << "\n"
	   << "beam position at target (X,Y):   (" << spill->posx << "," << spill->posy << ") (mm)\n"
	   << "beam width (x,y):                (" << spill->widthx << "," << spill->widthy << ") (mm)\n";
      else os << "      (no information under process label:   " << fSpillLabel << ")\n";
    }   // end print spill info



    // printing MC truth info
    if(fMCPrint != 0) {
      os << "\nMC truth info:";
      if(MCFlag == 1) {
	os << "\n==============\n" << "at t = " << nano/1e6 << " (msec)\n"
	   << "total number of MC reactions: " << mclist->size() << "\n";

	// loop over total number of MC reactions

	const TDatabasePDG* databasePDG = TDatabasePDG::Instance();

	for(unsigned int j = 0; j < mclist->size(); j++) {
	
	  os << "\nMC reaction number " << (j+1) << ":\n"
	     << "Origin:            ";

	  simb::Origin_t fOrigin = (*mclist)[j].Origin();
	  if(fOrigin == simb::kBeamNeutrino) os << "Beam Neutrino\n";
	  else if(fOrigin == simb::kCosmicRay) os << "Cosmic Ray\n";
	  else if(fOrigin == simb::kSuperNovaNeutrino) os << "Supernova Neutrino\n";
	  else if(fOrigin == simb::kSingleParticle) os << "Single Particle\n";
	  else os << "(Unknown)\n";

	  os << "\ninitial particles in reaction:\n";

          // print initial particles in reaction (status code = 0)
	  for(int i = 0; i < (*mclist)[j].NParticles(); i++) {

            if((*mclist)[j].GetParticle(i).StatusCode() == 0) {

              const TParticlePDG* p = databasePDG->GetParticle((*mclist)[j].GetParticle(i).PdgCode());
      	      std::string name = p ? p->GetName() : "(unknown)";

	      if(name == "(unknown)") os << name << " pdgcode: " << (*mclist)[j].GetParticle(i).PdgCode()
						 << " [" << (*mclist)[j].GetParticle(i).E() << " GeV]\n";
	      else os << name << " [" << (*mclist)[j].GetParticle(i).E() << " GeV]\n";

            } // end if
          
	  }  // end for i

          os << "\nfinal particles in reaction:\n";

          // print final particles in reaction (status code = 1)
          for(int k = 0; k < (*mclist)[j].NParticles(); k++) {

            if((*mclist)[j].GetParticle(k).StatusCode() == 1) {

              const TParticlePDG* p = databasePDG->GetParticle((*mclist)[j].GetParticle(k).PdgCode());
              std::string name = p ? p->GetName() : "(unknown)";

              if(name == "(unknown)") os << name << " pdgcode: " << (*mclist)[j].GetParticle(k).PdgCode()
					 << " [" << (*mclist)[j].GetParticle(k).E() << " GeV]\n";
	      else os << name << " [" << (*mclist)[j].GetParticle(k).E() << " GeV]\n";

            } // end if

          }  // end for k

	} // end for j

      }   // end if(MCFlag == 1)

      else os << "   (no information under process label:   " << fMCLabel << ")\n";

    }   // end printing MC truth info
  


    // print reconstruction info
    if(fRecoPrint != 0) {
      os <<      "\nReconstruction info:" << "\n====================\n\n";
      os <<      "number of hits in each view w/ BCM applied (x,y):   (" << nx << "," << ny << ")\n";

      // Hough Print
      if(HoughFlag == 1) os   <<    "HoughResults:   size = " << hr->size() << "\n";
      else os << "HoughResults:              (no information under process label:   " << fHoughLabel << ")\n";

      // vertices print
      if(VertexFlag == 1) os  <<    "Vertices:       size = " << vertices->size() << "\n";
      else os << "Vertices:                  (no information under process label:   " << fVertexLabel << ")\n";

      // slice print
      if(ClusterFlag == 1) os <<    "Clusters:       size = " << cluster->size() << "\n";
      else os << "Clusters:                  (no information under process label:   " << fClusterLabel << ")\n";
      
      // prong print
      if(ProngFlag == 1) os   <<    "Prongs:         size = " << prong->size() << "\n";
      else os << "Prongs:                    (no information under process label:   " << fProngLabel << ")\n";
      
      // track print
      if(TrackFlag == 1) os   <<    "Tracks:         size = " << track->size() << "\n";
      else os << "Tracks:                    (no information under process label:   " << fTrackLabel << ")\n";
     }   // end if fRecoPrint


    
    // print extra Reco info
    if(fExtraRecoPrint != 0) {
      // extra hough lines info
      if(HoughFlag == 1) {
	os <<   "\nHoughResults:   total number of Hough lines in each view (x,y):  (" << hrXsz << "," << hrYsz << ")\n";
	os <<   "Lines:   view, (slope, intercept) [cm]\n";     
	for(unsigned int i = 0; i < hr->size(); ++i) {
	  for(unsigned int j = 0; j < (*hr)[i].fPeak.size(); ++j) {
	    double m = 0.0, b = 0.0;
	    (*hr)[i].SlopeIcept(j, &m, &b);
	    if((*hr)[i].fView == geo::kX) os << "         X    (";
	    else if((*hr)[i].fView == geo::kY) os << "         Y    (";
	    else os << "         unknown view   (";
            os << m << ", " << b << ")\n";
	  }
	}
      }

      // extra vertex info
      if(VertexFlag == 1) {
	os <<   "\nVertices:     (x,y,z,t) [cm,ns]\n";
	for(unsigned int i = 0; i < vertices->size(); ++i) {
	  os << "              (" << (*vertices)[i].GetX() << ", " << (*vertices)[i].GetY() << ", "
	     << (*vertices)[i].GetZ() << ", " << (*vertices)[i].GetT() << ")\n";
	}
      }

      // extra cluster info
      if(ClusterFlag == 1) {
	os <<   "\nClusters:     view, # of hits, IsNoise\n";
	for(unsigned int i = 0; i < cluster->size(); ++i) {
	  os << "              ";
	  if((*cluster)[i].View() == geo::kX) os << "X   " << (*cluster)[i].NXCell();
	  if((*cluster)[i].View() == geo::kY) os << "Y   " << (*cluster)[i].NYCell();
          if((*cluster)[i].Is3D())            os << "3D  " << (*cluster)[i].NCell();
	  if((*cluster)[i].IsNoise()) os << "   Y\n";
	  else                        os << "   N\n";
	}
      }
      

      // extra prong info
      if(ProngFlag == 1) {
	os <<   "\nProngs:     view, # of hits, start (x,y,z) [cm], direction (x,y,z) [cm]\n";
	for(unsigned int i = 0; i < prong->size(); ++i) {
	  if((*prong)[i].View() == geo::kX) os << "X   " << (*prong)[i].NXCell();
	  if((*prong)[i].View() == geo::kY) os << "Y   " << (*prong)[i].NYCell();
          if((*prong)[i].Is3D())            os << "3D  " << (*prong)[i].NCell();
	  os << "   (" << (*prong)[i].Start().X() << ", " 
	               << (*prong)[i].Start().Y() << ", " 
	               << (*prong)[i].Start().Z() << ")"
	     << "   (" << (*prong)[i].Dir().X() << ", "
	               << (*prong)[i].Dir().Y()  << ", " 
	               << (*prong)[i].Dir().Z()  << ")\n";
	}
      }

      // extra track info
      if(TrackFlag == 1) {
	os <<   "\nTracks:     view, # of hits, length [cm], start (x,y,z) [cm], stop (x,y,z) [cm]\n";
	for(unsigned int i = 0; i < track->size(); ++i) {
	  if((*track)[i].View() == geo::kX) os << "            X   " << (*track)[i].NXCell();
	  if((*track)[i].View() == geo::kY) os << "            Y   " << (*track)[i].NYCell();
          if((*track)[i].Is3D())            os << "            3D  " << (*track)[i].NCell();
	  os << "   " << (*track)[i].TotalLength();
	  os << "   (" << (*track)[i].Start().X() << ", "
	               << (*track)[i].Start().Y() << ", "
                       << (*track)[i].Start().Z() << ")"
	     << "   (" << (*track)[i].Stop().X()  << ", "
                       << (*track)[i].Stop().Y()  << ", "
                       << (*track)[i].Stop().Z()  << ")\n";
	}
      }
	

    }
    os <<        "\n\n||===================== End EventSummary output =====================||\n\n\n";

    //
    // end printing EventSummary information
    
    return;
  }
} // end namespace
////////////////////////////////////////////////////////////////////////



#include "art/Framework/Core/ModuleMacros.h"
namespace evtsum
{
  DEFINE_ART_MODULE(EventSummary);
}
