////////////////////////////////////////////////////////////////////////
/// \brief   Class to filter truth information
///
/// \author  shanahan@fnal.gov
/// \date    
////////////////////////////////////////////////////////////////////////
#include <string>
#include <vector>

#include "TDatabasePDG.h"

#include "art/Framework/Core/EDProducer.h"
#include "art/Framework/Core/ModuleMacros.h" 
#include "art/Framework/Principal/Event.h" 
#include "art/Framework/Principal/Handle.h" 
#include "art/Framework/Services/Registry/ServiceHandle.h" 

#include "Simulation/FilterFlags.h"
#include "Simulation/FLSHitList.h"
#include "Simulation/ParticleNavigator.h"
#include "SimulationBase/MCNeutrino.h"
#include "SimulationBase/MCTruth.h"
#include "MCCheater/BackTracker.h"


namespace filter
{
  
  /// A class for filtering Truth hits and truth interactions
  class TruthFilter : public art::EDProducer {
    
  public:
    explicit TruthFilter(fhicl::ParameterSet const& pset);
    virtual ~TruthFilter(){};
    
    void produce(art::Event& evt);
    void beginJob();
    
  private:
    
    std::string fGenModuleLabel;   //< label name to find MC truth
    std::string fInputModuleLabel; //< label name to find MC Hits
    int         fRejectHadron;     //< reject all hits from hadrons?
    int         fRejectLepton;     //< reject all hits from primary lepton?
    int         fNCOnly;           //< accept only true NC interactions?
    int         fCCOnly;           //< accept only true CC interactions?
    int         fNuOnly;           //< reject nu-bar interactions?
    int         fNuBarOnly;        //< accept only nu-bar interactions?
    
    int         fNuE;              //< require (+1) or reject (-1) nu_e (and bar)?
    int         fNuMu;             //< require (+1) or reject (-1) nu_mu (and bar)?
    int         fNuTau;            //< require (+1) or reject (-1) nu_tau (and bar)?
    
    bool fDoRejectInteractions;    //< reject entire interactions?  Not directly configured by user.
    static bool fNotYetWarned;     //< for printing a warning
  };
    
}


namespace filter
{
  bool TruthFilter::fNotYetWarned=true;
  
  //-------------------------------------------
  TruthFilter::TruthFilter(fhicl::ParameterSet const& pset) :
    fGenModuleLabel      (pset.get< std::string >("GenModuleLabel")),	 
    fInputModuleLabel    (pset.get< std::string >("InputModuleLabel")),	 
    fRejectHadron        (pset.get< int         >("RejectLepton")),	 
    fRejectLepton        (pset.get< int         >("RejectHadron")),	 
    fNCOnly              (pset.get< int         >("NCOnly")),		 
    fCCOnly              (pset.get< int         >("CCOnly")),		 
    fNuOnly              (pset.get< int         >("NuOnly")),		 
    fNuBarOnly           (pset.get< int         >("NuBarOnly")),	 
    fNuE                 (pset.get< int         >("NuE")),		 
    fNuMu                (pset.get< int         >("NuMu")),		 
    fNuTau               (pset.get< int         >("NuTau")),             
    fDoRejectInteractions(false)
  {
    /*
      assert_jobc(fNuE>=-1&&fNuE<=1,
      "Choices for NuE are -1 (reject), 0 (no decision),+1 (require)\n");
      assert_jobc(fNuMu>=-1&&fNuMu<=1,
      "Choices for NuMu are -1 (reject), 0 (no decision),+1 (require)\n");
      assert_jobc(fNuTau>=-1&&fNuTau<=1,
      "Choices for NuTau are -1 (reject), 0 (no decision),+1 (require)\n");
    */
    
    int n=TMath::Abs(fNuE)+TMath::Abs(fNuMu)+TMath::Abs(fNuTau);
    if (fRejectLepton||fRejectHadron||fNCOnly||fCCOnly||fNuOnly||fNuBarOnly||n!=0)
      fDoRejectInteractions=true; else fDoRejectInteractions=false;
    
    //    assert_jobc(!(fRejectLepton&&fRejectHadron),"Cannot filter both Lepton and Hadron\n");
    
    produces< std::vector<sim::FilterFlags> >();  
  }
  
  //-------------------------------------------
  void TruthFilter::beginJob()
  {
  }
  
  //-------------------------------------------
  void TruthFilter::produce(art::Event& evt)
  {
    std::unique_ptr<std::vector<sim::FilterFlags> > fflags(new std::vector<sim::FilterFlags>);
    std::unique_ptr<std::vector<sim::FLSHitList> > fhitlist(new std::vector<sim::FLSHitList>);
    
    art::Handle< std::vector<sim::FLSHitList> >   hitlist;
    art::Handle< std::vector<simb::MCTruth> >     truthlist;
    
    evt.getByLabel(fInputModuleLabel, hitlist);
    evt.getByLabel(fGenModuleLabel,   truthlist);
    
    // want to define entire vector (on entry per interaction), to maintain
    // 1-to-1 correspondence with FLSHitList vector, etc.
    
    int NKeep=0; // keep track of kept interactions
    
    // loop over interactions
    for (size_t i = 0; i < truthlist->size(); ++i) {
      
      sim::FilterFlags flag;

      //get the truth information from the list of truths
      //taking the first event if there are two
      simb::MCTruth truth = (*truthlist)[0];

      //first, see if we can skip entire interaction
      if (fDoRejectInteractions 
	  && (truth.Origin() == simb::kBeamNeutrino ||
	      truth.Origin() == simb::kSuperNovaNeutrino)) {
	///get the neutrino information from the truth object
	const simb::MCNeutrino* nu = &(truthlist->at(i).GetNeutrino());
	
	if ((fNCOnly && nu->CCNC()==simb::kCC) ||
	    (fCCOnly && nu->CCNC()==simb::kNC)) flag.fRejectInter=true;
	
	if ((fNuOnly    && nu->Nu().PdgCode() < 0) ||
	    (fNuBarOnly && nu->Nu().PdgCode() > 0)) flag.fRejectInter=true;
	
	if ((fNuE ==  1 && TMath::Abs(nu->Nu().PdgCode()) != 12) ||
	    (fNuE == -1 && TMath::Abs(nu->Nu().PdgCode()) == 12)) 
	  flag.fRejectInter=true;
	
	if ((fNuMu ==  1 && TMath::Abs(nu->Nu().PdgCode()) != 14) ||
	    (fNuMu == -1 && TMath::Abs(nu->Nu().PdgCode()) == 14)) 
	  flag.fRejectInter=true;
	
	if ((fNuTau ==  1 && TMath::Abs(nu->Nu().PdgCode()) != 16) ||
	    (fNuTau == -1 && TMath::Abs(nu->Nu().PdgCode()) == 16)) 
	  flag.fRejectInter=true;
	
	// skip looping over hits if this whole interaction is rejected
	if (flag.fRejectInter) continue;
      }
      
      
      ++NKeep;//keep track of number of interactions not rejected

      art::ServiceHandle<cheat::BackTracker> bt;
      const sim::ParticleNavigator& pnav = bt->ParticleNavigator();
      
      sim::FLSHitList filteredHitList;
      // now loop over hits in this interaction
      for (unsigned int j=0; j< hitlist->at(i).fHits.size(); ++j) {
	
	int trk=hitlist->at(i).fHits[j].GetTrackID();
	
	// find ultimate parent track
	int motherTrk = pnav.EveId(trk);
        
	const unsigned int evePDG=TMath::Abs( pnav.Particle(motherTrk)->PdgCode() );
	
	bool IsLepton = (evePDG == 11) || (evePDG == 13) || (evePDG == 15);
	
	if ((fRejectHadron && !IsLepton) || (fRejectLepton && IsLepton))
	  flag.fRejectHits.insert(j);
	else
	  filteredHitList.fHits.push_back(hitlist->at(i).fHits[j]);
	
      }// for input hit j	
      
      fflags->push_back(flag);
      fhitlist->push_back(filteredHitList);
      
    }//for interaction i
    
    evt.put(std::move(fflags));
    evt.put(std::move(fhitlist));
    
    return;
    
  }

} // end namespace sim

////////////////////////////////////////////////////////////////////////
namespace filter
{ 
  DEFINE_ART_MODULE(TruthFilter);
}

