/////////////////////////////////////////////////////////////////////////
// \file    CoordinateTransformation.cxx
// \brief
// \version $Id: CoordinateTransformation.cxx,v 1.3 2012-11-09 17:22:59 jpaley Exp $
// \author  denis@fnal.gov
/////////////////////////////////////////////////////////////////////////

#include "GeometryObjects/CoordinateTransformation.h"
#include "NovaDAQConventions/DAQConventions.h"

#include <boost/numeric/ublas/io.hpp>

#include "messagefacility/MessageLogger/MessageLogger.h"

#include "boost/format.hpp"

#include <cstdio>
#include <iostream>
#include <iomanip>

namespace geo{

  //---------------------------------------------------------------------
  CoordinateTransformation::CoordinateTransformation():
    _detectorID(0),
    _beamType(0),
    _rotationMatrix(3,3),
    _rotationMatrixInverse(3,3),
    _simpleTranslation(3)
  {
  }

  //------------------------------------------------------------------------------
  CoordinateTransformation::CoordinateTransformation(uint32_t detector_id, uint32_t beam_type):
    _detectorID(detector_id),
    _beamType(beam_type),
    _rotationMatrix(3,3),
    _rotationMatrixInverse(3,3),
    _simpleTranslation(3)
  {
    setDetectorAndBeam(_detectorID, _beamType);
  }


  //------------------------------------------------------------------------------
  bool CoordinateTransformation::setDetectorAndBeam(uint32_t detector_id, uint32_t beam_type){

    /// Trying to set the detector parameters
    /// If unsuccessfull, set the detector and beam ID to zero
    if(!setCoordinateParameters(detector_id, beam_type)){
      _detectorID = 0;
      _beamType   = 0;
      return false;
    }// end of setting the detector parameters


    /// Have to reset the detector ID and beam type
    _detectorID = detector_id;
    _beamType   = beam_type;

    return true;
  }


  //------------------------------------------------------------------------------
  CoordinateTransformation::~CoordinateTransformation() {
    LOG_DEBUG("CoordinateTransformation")
      << "~CoordinateTransformation\n";
  }


  //------------------------------------------------------------------------------
  bool CoordinateTransformation::init() {

    double ca = cos(_alpha);
    double sa = sin(_alpha);
    double cb = cos(_beta);
    double sb = sin(_beta);

    _rotationMatrix(0, 0) =  cb;
    _rotationMatrix(0, 1) = -sb*sa;
    _rotationMatrix(0, 2) =  sb*ca;

    _rotationMatrix(1, 0) = 0.0;
    _rotationMatrix(1, 1) = ca;
    _rotationMatrix(1, 2) = sa;

    _rotationMatrix(2, 0) = -sb;
    _rotationMatrix(2, 1) = -cb*sa;
    _rotationMatrix(2, 2) =  cb*ca;


    _rotationMatrixInverse(0, 0) =  cb;
    _rotationMatrixInverse(0, 1) =  0.0;
    _rotationMatrixInverse(0, 2) = -sb;

    _rotationMatrixInverse(1, 0) = -sa*sb;
    _rotationMatrixInverse(1, 1) =  ca;
    _rotationMatrixInverse(1, 2) = -sa*cb;

    _rotationMatrixInverse(2, 0) =  ca*sb;
    _rotationMatrixInverse(2, 1) =  sa;
    _rotationMatrixInverse(2, 2) =  ca*cb;

    return true;
  }

  //------------------------------------------------------------------------------
  bool CoordinateTransformation::transformBeamToDetectorCoordinates
  (double* input_beam_coords,
   double* output_detector_coords) const {


    /// Check if the beam type and detector id are set
    if(!_beamType || !_detectorID) return false;

    /// Calculate the beam coordinates
    /// x`` = M^{-1} * (x-x0)
    for(int i = 0; i < 3; i++){
      double sum = 0.0;
      for(int j = 0; j < 3; j++)
	sum += _rotationMatrixInverse(i,j)*(input_beam_coords[j] - _simpleTranslation[j]);

      output_detector_coords[i] = sum;
    }// end of loop over coordinates

    return true;
  }// end of transformBeamToDetectorCoordinates

  //------------------------------------------------------------------------------
  bool CoordinateTransformation::transformDetectorToBeamCoordinates
  (double* input_detector_coords,
   double* output_beam_coords) const {


    /// Check if the beam type and detector id are set
    if(!_beamType || !_detectorID) return false;


    /// Calculate the detector coordinates
    /// x = x0 + M * x''
    for(int i = 0; i < 3; i++) {
      double sum = 0.0;
      for(int j = 0; j < 3; j++)
	sum += _rotationMatrix(i,j)*input_detector_coords[j];

      output_beam_coords[i] = _simpleTranslation[i] + sum;
    }// end of loop over coordinates

    return true;
  }// end of transformDetectorToBeamCoordinates


  //------------------------------------------------------------------------------
  bool CoordinateTransformation::print() {

    std::cout << "Coordinate system transformations for "
	      << "Beam Type = " << _beamType << " "
	      << "Detector = " << _detectorID << ":\n";

    std::cout << "Translation:\n";
    printf("    |%0.10E|\n", _simpleTranslation[0]);
    printf("S = |%0.10E|\n", _simpleTranslation[1]);
    printf("    |%0.10E|\n", _simpleTranslation[2]);



    std::cout << "Beam->Detector\n";
    printf("|x_det|      |  %0.10E    %0.10E    %0.10E   |   |  x_beam - %0.10E  |\n", _rotationMatrixInverse(0,0), _rotationMatrixInverse(0,1), _rotationMatrixInverse(0,2) , _simpleTranslation[0]);
    printf("|y_det|  =   |  %0.10E    %0.10E    %0.10E   | * |  y_beam - %0.10E  |\n", _rotationMatrixInverse(1,0), _rotationMatrixInverse(1,1), _rotationMatrixInverse(1,2) , _simpleTranslation[1]);
    printf("|z_det|      |  %0.10E    %0.10E    %0.10E   |   |  z_beam - %0.10E  |\n", _rotationMatrixInverse(2,0), _rotationMatrixInverse(2,1), _rotationMatrixInverse(2,2) , _simpleTranslation[2]);


    std::cout << "\n";
    std::cout << "Detector->Beam\n";
    printf("|x_beam|      | %0.10E |       |  %0.10E    %0.10E    %0.10E   |   |  x_det  |\n", _simpleTranslation[0], _rotationMatrix(0,0), _rotationMatrix(0,1), _rotationMatrix(0,2));
    printf("|y_beam|  =   | %0.10E |   +   |  %0.10E    %0.10E    %0.10E   | * |  y_det  |\n", _simpleTranslation[1], _rotationMatrix(1,0), _rotationMatrix(1,1), _rotationMatrix(1,2));
    printf("|z_beam|      | %0.10E |       |  %0.10E    %0.10E    %0.10E   |   |  z_det  |\n", _simpleTranslation[2], _rotationMatrix(2,0), _rotationMatrix(2,1), _rotationMatrix(2,2));

    /// Now calculating the angle at which point S sees the front surface of the detector

    /// First figuring out the detector UZ in beam coordinates
    double p0_det[3]  = {0,0,0};
    double p0_beam[3];
    transformDetectorToBeamCoordinates(p0_det, p0_beam);

    double p1_det[3]  = {0,0,1};
    double p1_beam[3];
    transformDetectorToBeamCoordinates(p1_det, p1_beam);

    /// Unit vector of detector's Uz in beam coordinates
    double dr_detector_uz[3] = {
      p1_beam[0] - p0_beam[0],
      p1_beam[1] - p0_beam[1],
      p1_beam[2] - p0_beam[2]
    };


    /// Distance from point S to detector zero
    double dr_from_pointS_to_detector_zero[3] = {
      _simpleTranslation[0],
      _simpleTranslation[1],
      _simpleTranslation[2] - pointS_Z
    };

    /// Now need to normalize it
    {
      double drmod_from_pointS_to_detector_zero = sqrt(
						       dr_from_pointS_to_detector_zero[0]*dr_from_pointS_to_detector_zero[0]
						       + dr_from_pointS_to_detector_zero[1]*dr_from_pointS_to_detector_zero[1]
						       + dr_from_pointS_to_detector_zero[2]*dr_from_pointS_to_detector_zero[2]
						       );
      for(int i=0; i<3; i++)
	dr_from_pointS_to_detector_zero[i] /= drmod_from_pointS_to_detector_zero;
    }// end of normalization

    /// cosine of the angle at which point-S sees the front surface of the detector
    double cos_detector_front_surface = dr_detector_uz[0] * dr_from_pointS_to_detector_zero[0]
      + dr_detector_uz[1] * dr_from_pointS_to_detector_zero[1]
      + dr_detector_uz[2] * dr_from_pointS_to_detector_zero[2];

    LOG_DEBUG("CoordinateTransformation")
      << "Cosine of the angle at which point S sees the front surface of the detector: "
      << cos_detector_front_surface
      << "\n";

    return true;
  }


  //------------------------------------------------------------------------------
  bool CoordinateTransformation::setCoordinateParameters(uint32_t detector_id,
							 uint32_t beam_type) {


    /// See if we support this beam type
    if(beam_type != BEAMTYPE_NUMI) {
      LOG_DEBUG("CoordinateTransformation")
	<< "From CoordinateTransformation::setCoordinateParameters. "
	<< "Beam type " << beam_type << " is not supported.\n";
      return false;
    }// end of checking the beam configuration


    /// See if we support this detector type
    if      (detector_id == novadaq::cnv::kNEARDET) {
      /// From NOvA-doc-5485
      _simpleTranslation[0] =  11.414  * CLHEP::m;
      _simpleTranslation[1] = -3.456   * CLHEP::m;
      _simpleTranslation[2] =  994.665 * CLHEP::m;

      _alpha = NuMiBeamLineAngle;
      _beta  = NOvAneardetNuMiAngle;
    }
    else if (detector_id == novadaq::cnv::kFARDET) {
      /// From NOvA-doc-5485
      _simpleTranslation[0] =  11.037  * CLHEP::km;
      _simpleTranslation[1] = -4.1626   * CLHEP::km;
      _simpleTranslation[2] =  810.42 * CLHEP::km;

      _alpha = - NuMiBeamLineAngle;
      _beta  =   NOvAneardetNuMiAngle;
    }
    else if (detector_id == novadaq::cnv::kNDOS) {
      /// From NOvA-doc-5485
      _simpleTranslation[0] = -0.29  * CLHEP::m;
      _simpleTranslation[1] =  92.21 * CLHEP::m;
      _simpleTranslation[2] =  841.76* CLHEP::m;

      _alpha = NuMiBeamLineAngle;
      _beta  = 0.0;
    }

    else {
      LOG_DEBUG("CoordinateTransformation")
	<< "From geo::CoordinateTransformation::setCoordinateParameters. "
	<< "Detector ID " << detector_id << " is not supported.\n";
      return false;
    }// end of checking the detector configuration

    return init();
  }

}
////////////////////////////////////////////////////////////////////////////
