/////////////////////////////////////////////////////////////////////////
// \file    PlaneGeo.cxx
// \brief
// \version $Id: PlaneGeo.cxx,v 1.9 2012-11-28 03:25:00 gsdavies Exp $
// \author  $Author: gsdavies $
/////////////////////////////////////////////////////////////////////////

#include "GeometryObjects/PlaneGeo.h"
#include "GeometryObjects/CellGeo.h"

#include <algorithm>
#include <cmath>
#include <cstdlib>
#include <iostream>

#include "TGeoNode.h"
#include "TGeoVolume.h"
#include "TGeoMatrix.h"

#include "cetlib/exception.h"

namespace geo
{
  //..............................................................
  static bool sort_hori(const CellGeo* c1, const CellGeo* c2)
  {
    double xyz1[3], xyz2[3];
    c1->GetCenter(xyz1);
    c2->GetCenter(xyz2);
    return xyz1[1]<xyz2[1];
  }
  static bool sort_vert(const CellGeo* c1, const CellGeo* c2)
  {
    double xyz1[3], xyz2[3];
    c1->GetCenter(xyz1);
    c2->GetCenter(xyz2);
    return xyz1[0]<xyz2[0];
  }

  //......................................................................

  PlaneGeo::PlaneGeo(std::vector<const TGeoNode*>& n, unsigned int depth)
  {
    // build a matrix to take us from the local to the world coordinates
    // in one step
    fGeoMatrix = new TGeoHMatrix(*n[0]->GetMatrix());
    for (unsigned int i=1; i<=depth; ++i) {
      fGeoMatrix->Multiply(n[i]->GetMatrix());
    }

    // Copy fGeoMatrix to Original
    fGeoMatrixOriginal = new TGeoHMatrix(*fGeoMatrix);

    // fGeoMatrix is still unmodified
    fIsOriginal = true;

    // Determine read-out orientation for the plane.  By convention, the
    // planes, cells, and fibers are constructed such that the z-axis
    // goes along the long direction with z increasing as one moves
    // toward the read out end. So, a good way to find the read out end
    // is to take a small step in the +z direction which moves you
    // toward the read out side of the detector.
    static const double step = 10.0;     // size of step in cm
    static const double tol  = 0.1*step; // testing tolerance
    static const double a0[3] = {0,0,0};
    static const double a1[3] = {0,0,step};
    double b0[3], b1[3];
    fGeoMatrix->LocalToMaster(a0,b0);
    fGeoMatrix->LocalToMaster(a1,b1);
    b1[0] -= b0[0]; b1[1] -= b0[1]; b1[2] -= b0[2];
    if (fabs(b1[0]-0.0)<tol && fabs(b1[1]-step)<tol) {
      fView    = kX;
      fReadout = kTop;
    }
    else if (fabs(b1[0]-step)<tol && fabs(b1[1]-0.)<tol) {
      fView    = kY;
      fReadout = kWest;
    }
    else if (fabs(b1[0]+step)<tol && fabs(b1[1]-0.)<tol) {
      fView    = kY;
      fReadout = kEast;
    }
    else {
      throw cet::exception("BAD_PLANE_VIEW")
        << "Could not determine view for plane\n"
        << __FILE__ << ":" << __LINE__ << "\n";
    }
    this->FindCells(n, depth);

    // Make sure horizontals are sorted top to bottom and verticals are
    // sorted east to west
    if (fView == kX) {
      std::sort(fCell.begin(), fCell.end(), sort_vert);
    }
    else if (fView == kY) {
      std::sort(fCell.begin(), fCell.end(), sort_hori);
    }
    else abort();
  }

  //......................................................................
  PlaneGeo::~PlaneGeo()
  {
    if(fGeoMatrix) delete fGeoMatrix;
    if(fGeoMatrixOriginal) delete fGeoMatrixOriginal;

    for(size_t c = 0; c < fCell.size(); ++c){
      if(fCell[c]) { delete fCell[c]; fCell[c] = 0; }
    }
    fCell.clear();
  }

  //......................................................................

  void PlaneGeo::FindCells(std::vector<const TGeoNode*>& n,
                           unsigned int depth)
  {
    if (strncmp(n[depth]->GetName(),"vCell",5)==0) {
      this->MakeCell(n,depth);
      return;
    }

    // Explore the next layer down
    unsigned int deeper = depth+1;
    if (deeper>=n.size()) {
      throw cet::exception("BAD_NODE")
        << "Exceeded maximum depth\n"
        << __FILE__ << ":" << __LINE__ << "\n";
    }
    const TGeoVolume* v = n[depth]->GetVolume();
    int nd = v->GetNdaughters();
    for (int i=0; i<nd; ++i) {
      n[deeper] = v->GetNode(i);
      this->FindCells(n, deeper);
    }
  }

  //......................................................................

  void PlaneGeo::MakeCell(std::vector<const TGeoNode*>& n,
                          unsigned int depth)
  {
    fCell.push_back(new CellGeo(n, depth));
  }

  //......................................................................

  void PlaneGeo::LocalToWorld(const double* plane, double* world) const
  {
    fGeoMatrix->LocalToMaster(plane, world);
  }

  //......................................................................

  void PlaneGeo::LocalToWorldVect(const double* plane, double* world) const
  {
    fGeoMatrix->LocalToMasterVect(plane, world);
  }

  //......................................................................

  void PlaneGeo::WorldToLocal(const double* world, double* plane) const
  {
    fGeoMatrix->MasterToLocal(world, plane);
  }

  //......................................................................

  /// Convert a vector from world frame to the local plane frame
  /// \param world : 3-D array. Vector in world coordinates; input.
  /// \param plane : 3-D array. Vector in plane coordinates; plane.
  void PlaneGeo::WorldToLocalVect(const double* world, double* plane) const
  {
    fGeoMatrix->MasterToLocalVect(world,plane);
  }

  //......................................................................

  void PlaneGeo::TranslatePlane(double dx, double dy, double dz, 
				double phi, double theta, double psi)
  {

    /// phi, theta and psi are euler angles
    TGeoRotation *rot = new TGeoRotation("alignmentRot",phi,theta,psi);
    /// dx, dy and dz are the translations we want
    TGeoTranslation *trans = new TGeoTranslation("",dx,dy,dz);
    TGeoCombiTrans *comb = new TGeoCombiTrans(*trans,*rot);

    this->TranslatePlane(comb);

  }

  //......................................................................

  void PlaneGeo::TranslatePlane(TGeoCombiTrans* comb)
  {

    /// Restart from original matrix
    fGeoMatrix = new TGeoHMatrix(*fGeoMatrixOriginal);

    /// Apply shift
    fGeoMatrix->MultiplyLeft(comb->MakeClone()); 

    /// Now apply shift to all cells
    for(unsigned int i=0; i<fCell.size(); i++){
      fCell.at(i)->TranslateCell(comb);
    }

    // fGeoMatrix is now modified
    fIsOriginal = false;

  }

  //......................................................................

  void PlaneGeo::RestoreOriginal()
  {

    /// Restore original matrix
    fGeoMatrix = new TGeoHMatrix(*fGeoMatrixOriginal);

    // fGeoMatrix is back to original
    fIsOriginal = true;

  }

} // end namespace geo
////////////////////////////////////////////////////////////////////////
