/***************************************************
\file    HMatrix.cxx
\brief   HMatrix producer. 

This code is designed to run over FD beam files and calculate the h-matrix chi2 variables, fills these variables in hme::HMPid object.
No histogram or tree filling is done in this producer code.

\version $Id: HMatrix.cxx,v 1.10 2012/11/08 00:01:58 hsahoo Exp $
\author  $Author: hsahoo $
*************************************/


#include "HMatrixE/HMPid.h"

// C/C++ includes
#include <vector>
#include <iostream>
#include <cmath>
#include <fstream>

// NOvASoft includes
#include "Calibrator/Calibrator.h"
#include "CMap/CMap.h"
#include "RecoBase/Track.h"
#include "RecoBase/CellHit.h"
#include "RecoBase/RecoHit.h"
#include "NovaDAQConventions/DAQConventions.h"

#include "SimulationBase/MCTruth.h"


// ROOT includes
#include "TH1.h"
#include "TH2.h"
#include "TTree.h"

// Framework includes
#include "art/Framework/Principal/Event.h" 
#include "art/Framework/Principal/Handle.h" 
#include "art/Framework/Services/Registry/ServiceHandle.h" 
#include "art/Framework/Services/Optional/TFileDirectory.h"
#include "art/Framework/Services/Optional/TFileService.h"
#include "cetlib/search_path.h"
#include "fhiclcpp/ParameterSet.h"  
#include "messagefacility/MessageLogger/MessageLogger.h" 
#include "art/Framework/Core/EDProducer.h"
#include "art/Framework/Core/ModuleMacros.h"

namespace hme {
  
  class HMatrix : public art::EDProducer {
    
  public:
    explicit HMatrix(const fhicl::ParameterSet& pset); // Required! explicit tag tells the compiler this is not a copy constructor      
    virtual ~HMatrix();                        
    
    // Optional, read/write access to event
    void reconfigure(const fhicl::ParameterSet& pset);    
    void beginJob();
    void produce(art::Event& evt); 

    double CalculateMIPPeak(art::Ptr<rb::Track> track);
    void SetUpEnergyArrays(std::vector<int>, std::vector<int>, std::vector<double>, int);
    std::vector<double> GetHMChisq();
    bool isInVertexFiducial(art::Ptr<rb::Track> track) const;

  private:
    int GetMCTruthM(const art::Event& evt, art::PtrVector<simb::MCTruth>& mctruth); 
    ///< Sample configurable module parameters
    int              fCounter;
    //cut counts
    double pass_cuts[11][11];//0 is numucc, 1 is nc, 2 is beamnue, 3 is nuecc
    
    static const int long_bins;
    static const int trans_bins;
    static const int comb_bins;

    // wait : HMatricies : please check that these have right dimensions
    double HM[21][21];
    double HTM[42][42];
    double HC[55][55];
    
    // dimensions of the HMatrix mean file    
    double  lmeans[21];
    double trmeans[42];
    double ltmeans[55];
    
    // dimensions of the HMatrix bins
    double lbin[21];
    double transm[42];
    double ltbin[55];


    double etot;
    double ttot; // for tran-only bins
    double esumlay[1000];
    double eplcell[1000][400];
    double Tesum; // total energy sum of hits used in test
    int nhits_tot;
    int nlay;
    int maxHcell[1000];
    std::vector<double> fhMeans;
    std::vector<double> fvMeans;
    std::vector<int> fmeans;
    float            fFloat;  
    double           fDouble; 
    double max_layers;
    double max_cells;

    std::string fTrackLabel;         ///< Read Tracks
    std::string fLongHM;             ///< Longitudinal HMatrix
    std::string fTransHM;            ///< Transvers HMatrix
    std::string fCombHM;             ///< Combined HMatrix
    std::string fLongMeans;
    std::string fTransMeans;
    std::string fCombMeans;
    
    
        
  };
}


namespace hme{
  
  HMatrix::HMatrix(const fhicl::ParameterSet& pset):
    fTrackLabel      (pset.get<std::string>("TrackLabel")    ),
    fLongHM          (pset.get<std::string>("LongHM")        ),
    fTransHM         (pset.get<std::string>("TransHM")       ),
    fCombHM          (pset.get<std::string>("CombHM")        ),
    fLongMeans       (pset.get<std::string>("LongMeans")     ),
    fTransMeans      (pset.get<std::string>("TransMeans")    ),
    fCombMeans       (pset.get<std::string>("CombMeans")     )
  {   
    reconfigure(pset);

    produces<std::vector<hme::HMPid> >();
  }
  
  void HMatrix::reconfigure(const fhicl::ParameterSet& pset)
  {
    fLongHM      = pset.get<std::string>("LongHM");
    fTransHM     = pset.get<std::string>("TransHM");
    fCombHM      = pset.get<std::string>("CombHM");
    fLongMeans   = pset.get<std::string>("LongMeans");
    fTransMeans  = pset.get<std::string>("TransMeans");
    fCombMeans   = pset.get<std::string>("CombMeans");
  }
  //......................................................................
  HMatrix::~HMatrix()
  {
  }

  /// TODO: Make me a fcl parameter!
  const int HMatrix::long_bins  = 21;
  const int HMatrix::trans_bins = 42;
  const int HMatrix::comb_bins  = 55;
  

  //......................................................................
  void HMatrix::beginJob()
  {

    max_layers = 1000;
    max_cells  = 400;
    //nlay       = 4;
    nlay       = 3; // nlay is changed to 3 on Aug 24


    //initilize the histograms that will hold the horizontal and vertical matrix means to zero.
    /// TODO: Remove hard-coded values
    for(int jj=0; jj<55; jj++){
      fhMeans.push_back(0);
      fvMeans.push_back(0);
    }
    
    //initialize all the matrix bins to zero    
    // wait : aug 17 : all matrix bins are initialized separately to zero.........

    for (int i=0; i<long_bins; ++i){
      for (int j=0; j<long_bins; ++j){
	HM[i][j] = 0.;
      }
    }

    for (int i=0; i<trans_bins; ++i){
      for (int j=0; j<trans_bins; ++j){
	HTM[i][j] = 0.;
      }
    }

    for (int i=0; i<comb_bins; ++i){
      for (int j=0; j<comb_bins; ++j){
	HC[i][j] = 0.;
      }
    }

    std::string LongHM, TransHM, CombHM;
    std::string LongMeans, TransMeans, CombMeans;

    cet::search_path sp("FW_SEARCH_PATH");
    sp.find_file(fLongHM,LongHM);
    sp.find_file(fTransHM,TransHM);
    sp.find_file(fCombHM,CombHM);

    sp.find_file(fLongMeans,  LongMeans);
    sp.find_file(fTransMeans, TransMeans);
    sp.find_file(fCombMeans,  CombMeans);

    ifstream longhmi;  
    ifstream transhmi;
    ifstream combhmi;
    
    ifstream longmeans;
    ifstream transmeans;
    ifstream combmeans;

    // ############################################################
    //std::cout << "START : ########## Reading the HMatrixI files" << std::endl;
    // The matrixes are read from input directory.
    // make sure SRT_PRIVATE_CONTEXT flag is set properely.

    longhmi.open(LongHM.c_str());
    
    Double_t ldat;
    if (!longhmi.is_open()){
      throw cet::exception("UnresolvedMatrix")
        << "Can not open Long HMatrix file " << LongHM << "!";      
    }
    for (int i=0; i<long_bins; ++i){
      for (int j=0; j<long_bins; ++j){
	longhmi >> ldat;
	HM[i][j] = ldat;
	//std::cout << HM[i][j] << " ";
      }
      //std::cout << "\n";
    }
    longhmi.close();    
    //std::cout << "\n----------------------------" << std::endl;


    transhmi.open(TransHM.c_str());
    Double_t tdat;
    if (!transhmi.is_open()){
      throw cet::exception("UnresolvedMatrix")
	<< "Can not open Trans HMatrix file " << TransHM << "!";
    }
    for (int i=0; i<trans_bins; ++i){
      for (int j=0; j<trans_bins; ++j){
	transhmi >> tdat;
	HTM[i][j] = tdat;
	//std::cout << HTM[i][j] << " ";
      }
      //std::cout << "\n";
    }
    transhmi.close();
    //std::cout << "\n----------------------------" << std::endl;


    combhmi.open(CombHM.c_str());
    Double_t cdat;
    if (!combhmi.is_open()){
      throw cet::exception("UnresolvedMatrix")
	<< "Can not open Combined HMatrix file " << CombHM << "!";
    }
    for (int i=0; i<comb_bins; ++i){
      for (int j=0; j<comb_bins; ++j){
	combhmi >> cdat;
	HC[i][j] = cdat;
	//std::cout << HC[i][j] << " ";
      }
      //std::cout << "\n";
    }
    combhmi.close();

    
    longmeans.open(LongMeans.c_str());    
    Double_t in_lmeans;
    if (!longmeans.is_open()){
      throw cet::exception("UnresolvedMatrix")
        << "Can not open Long HMatrix Means file " << LongMeans << "!";      
    }
    for (int i=0; i<long_bins; ++i){
      longmeans >> in_lmeans;
      lmeans[i] = in_lmeans;
      lbin[i] = 0;
      //std::cout << lmeans[i] << " ";
    }
    //std::cout << "\n";
    longmeans.close();
            
    transmeans.open(TransMeans.c_str());    
    Double_t in_trmeans;
    if (!transmeans.is_open()){
      throw cet::exception("UnresolvedMatrix")
        << "Can not open Trans HMatrix Means file " << TransMeans << "!";      
    }
    for (int i=0; i<trans_bins; ++i){
      transmeans >> in_trmeans;
      trmeans[i] = in_trmeans;
      transm[i] = 0;
      //std::cout << trmeans[i] << " ";
    }
    //std::cout << "\n";
    transmeans.close();
    
    
    combmeans.open(CombMeans.c_str());    
    Double_t in_combmeans;
    if (!combmeans.is_open()){
      throw cet::exception("UnresolvedMatrix")
        << "Can not open Comb HMatrix Means file " << CombMeans << "!";      
    }
    for (int i=0; i<comb_bins; ++i){
      combmeans >> in_combmeans;
      ltmeans[i] = in_combmeans;
      ltbin[i] = 0;
      //std::cout << ltmeans[i] << " ";
    }
    //std::cout << "\n";
    combmeans.close();

      
    for(int i=0; i<max_layers; i++){
      esumlay[i]=0;
      maxHcell[i]=0;
      
      for(int j=0; j<max_cells; j++){
	eplcell[i][j]=0;
      }
       
    }

    // initialize everything else to zero
    etot = 0.;  
    ttot = 0.; // for tran-only bins
    Tesum = 0.; // total energy sum of hits used in test	  
    nhits_tot=0; 
    fCounter=0;

  } // end of beginJob loop


  //-------------------------------------------------------------------------
  void HMatrix::produce(art::Event& evt)
  {

    // Reset here all matrix bins to zero for each event ...
    // all total energy, total transverse energy
    for ( int i=0; i<long_bins; ++i){
      lbin[i] = 0.0;
    }
    for ( int i=0; i<trans_bins; ++i){
      transm[i] = 0.0;
    }
    for ( int i=0; i<comb_bins; ++i){
      ltbin[i] = 0.0;
    }

    etot = 0.0;
    ttot = 0.0;


    // These are the cuts, they can be changed around to play with the cutflow if you like


    int ntracksl = 0;

    art::ServiceHandle<geo::Geometry> geo;
    art::ServiceHandle<calib::Calibrator> cal;
    
    int ifirst = 1;  // first plane of shower - use this to start binning

    // these vectors will hold the energy, plane and cell, for each hit in the event
    std::vector<double> energy_vec;
    std::vector<int> plane_vec;
    std::vector<int> cell_vec; 
    //this vector holds the first plane in the event 
    std::vector<double> min_plane_vec;

    /**********************
         Getting Slices ......
    *********************/

    art::Handle< std::vector<rb::Cluster> > slice_list;

    evt.getByLabel("slicer", slice_list);
    //int cluster_size = slice_list->size();

    //std::cout << "no of clusters " << slice_list->size() << std::endl;

    if (!slice_list->size()) return;

    // Find the slice having largest number of hits ........
    // skip the noise slice
    art::Ptr<rb::Cluster> largest_slice;
    unsigned int ncell_largest_slice = 0;

    for (unsigned int i=0; i<slice_list->size(); ++i){
      art::Ptr<rb::Cluster> fslice(slice_list,i);
      //std::cout << "slice " << i << "\t" << fslice->NCell() << std::endl; 

      if(fslice->IsNoise()==true) continue;

      if (fslice->NCell()>ncell_largest_slice){
	ncell_largest_slice = fslice->NCell();
	largest_slice =  fslice;
      }

      //std::cout << "slice_ " << i << "\t" << slice->NCell() << std::endl; 
    }

    //std::cout << "cell_hits in the largest_slice " << largest_slice->NCell() << std::endl;


    /**********************
         Getting Tracks ......
    *********************/
    art::Handle< std::vector<rb::Track> > track_list;
    
    // The module is designed to use any of the tracking algorithms as input :
    // kalmantrackmerge, discretetrackmerge, fuzzykvertex, trackmerge
    if (fTrackLabel=="fuzzykvertex"){
      evt.getByLabel(fTrackLabel, "Tracks3D", track_list); // fuzzykvertex, gives 3D tracks
      //std::cout << "Producer : fuzzykvertex is called : " << fTrackLabel.c_str() << std::endl;
    }else{
      evt.getByLabel(fTrackLabel,track_list);
      //std::cout << "Producer : tracking algorithm called : " << fTrackLabel.c_str() << std::endl;
    }
    std::cout << "Producer : tracking algorithm used : " << fTrackLabel.c_str() << std::endl;
    
    
    //int shower_number = (int)track_list->size();  
    //std::cout << "track.size() : shower " << track_list->size() << std::endl;
    
    if (!track_list->size()) return;
    // skip events having no tracks, otherwise Module failed due to an exception

    //the longest track
    art::Ptr<rb::Track> longest_track;
    double length_longest_track = 0;

    //ntracks = 0;

    for(unsigned int i=0; i<track_list->size(); i++){
      
      art::Ptr<rb::Track> track(track_list,i);
      
      if(track->IsNoise()==true)
	continue;
     

      //identify the longest track
      double length = track->TotalLength();



      if(length>length_longest_track){
	ntracksl++;
	longest_track = track;
	length_longest_track = length;
	if (ntracksl>1) ntracksl--;
      }

      //identify the total number of cells in the event
      const int ncells=track->NCell();

      // identify the total energy
      for(int j=0; j<ncells; j++){
	art::Ptr<rb::CellHit> cellhit=track->Cell(j);

	rb::RecoHit rhit(cal->MakeRecoHit(*cellhit, track->W(cellhit.get())));

	if(!rhit.IsCalibrated()){
	  std::cerr << "Not calibrated?!" << std::endl;
	  continue;
	}

	
	// wait : this is for test purpose HMatrix input.... all hits are includes..no longest track
	int plane     = cellhit->Plane();
	int cell      = cellhit->Cell();
	double recoPECorr = rhit.PECorr();
	
	energy_vec.push_back(recoPECorr);
	plane_vec.push_back(plane);
	cell_vec.push_back(cell);

      } // loop over all cell hits
      
    } // loop over track_list
    

    double MIP_PEAK = CalculateMIPPeak(longest_track);

    ifirst = longest_track->MinPlane();
    min_plane_vec.push_back(ifirst);
    
    //std::cout << "first plane " << ifirst << std::endl;
    // this is where the H-Matrix is setup
    // here, the plane_vec, cell_vec and energy_vec are filled with whatever shower candidate you like. 
    // Ifirst should indicate the first plane you'd like to test in that shower candidate

    SetUpEnergyArrays(plane_vec,cell_vec,energy_vec,ifirst);
    
    
    //calculate the HMatrix!!   
    // use the H matrix to calculate chi-squared for these electrons. 
    //the chisquared quantifies the transvers and longitudinal shape of the showers 
    // compared to a set of pre-made matricies created using single particle monte carlo electrons 

    std::vector<double> chisq_vec = GetHMChisq();

    double lchisq =  fabs(chisq_vec[0]); //longitudinal
    double cchisq =  fabs(chisq_vec[1]);//combined (a combination of transvers and longitidinal)
    double tchisq =  fabs(chisq_vec[2]); //transverse
    
    //std::cout << "lchisq_prod1 " << lchisq << "\t tchisq " << tchisq << "\t cchisq " << cchisq << std::endl;

    
    // fill the ntuple log10(chisq)
    double glchi2 = log10(lchisq);
    double gtchi2 = log10(tchisq);
    double gcchi2 = log10(cchisq);
    

    //std::cout << "log_lchisq_prod1 " << glchi2 << "\t tchisq " << gtchi2 << "\t cchisq " << gcchi2 << std::endl;

    std::unique_ptr<std::vector<hme::HMPid> > hmpcol(new std::vector<hme::HMPid>);

    hme::HMPid hmp;
    hmp.SetLongChi2(glchi2);
    hmp.SetTransChi2(gtchi2);
    hmp.SetCombChi2(gcchi2);
    hmp.SetMIPHits(MIP_PEAK);


    hmpcol->push_back(hmp);
    evt.put(std::move(hmpcol));

    return;
  } // end of analyze loop
  
  //#########################################################################
  double HMatrix::CalculateMIPPeak(art::Ptr<rb::Track> track){

    // the MIP Peak is the of cell hits with an energy in the region identified as the MIP Peak 
    // (ie the energy for a minimum ionizing cell hit). 
    // It is a number between 0 and 1, tending to be larger for muons as most hits are minimum ionizing

    art::ServiceHandle<calib::Calibrator> cal;
    int ncells=track->NCell();
    double cells_in_mip_peak=0;
    double cells_outside_mip_peak=0;

    for(int j=0; j<ncells; j++){
      art::Ptr<rb::CellHit> cellhit=track->Cell(j);

      rb::RecoHit rhit(cal->MakeRecoHit(*cellhit, track->W(cellhit.get())));

      if(!rhit.IsCalibrated()){
	std::cerr << "Not calibrated?!" << std::endl;
	continue;
      }

      double energy = rhit.GeV();
      double energy_mev=2*energy*1000;
      cells_outside_mip_peak++;

      if(4<energy_mev && energy_mev<16)
	cells_in_mip_peak++;      

    }

    cells_outside_mip_peak=cells_in_mip_peak/cells_outside_mip_peak;
    return cells_outside_mip_peak;

  }

  //-------------------------------------------------------------------------
  void HMatrix::SetUpEnergyArrays(std::vector<int> plane_vec, std::vector<int>cell_vec, std::vector<double> energy_vec, int ifirst){
    
    //this is where the H-Matrix is setup
    //here, the plane_vec, cell_vec and energy_vec are filled with whatever shower candidate you like. 
    //Ifirst should indicate the first plane you'd like to test in that shower candidate
    //setup up the total energy per layer array
    
    bool ffound=false;
    
    //initialize everything to zero to start
    for(int i=0; i<long_bins; i++){
      lbin[i]=0; 
    }
    for(int i=0; i<trans_bins; i++){
      transm[i]=0;
    }
    for(int i=0; i<comb_bins; i++){
      ltbin[i]=0;
    }

    etot = 0;
    ttot = 0;

    
    for(int i=0; i<max_layers; i++){
      esumlay[i]=0;
      maxHcell[i]=0;
      for(int j=0; j<max_cells; j++){
	eplcell[i][j]=0;
      }      
    }
    

    
    for (unsigned int i=0; i<energy_vec.size(); i++)
      {
	int plane     = 0;
	int cell      = 0;
	double energy = 0;
	
	plane  = plane_vec[i];
	energy = energy_vec[i];
	cell   = cell_vec[i];
	
	//  get energy sum in each layer
	// as well as the total energy sum
	if(plane<max_layers && cell<max_cells){
	  esumlay[plane] +=energy;
	  eplcell[plane][cell] += energy;
	  Tesum += energy;
	}
	
      }

    
    //       this finds layer of start of shower if you don't already know it

    for (unsigned int j=0; j<max_layers; ++j)
      {
	if (!ffound && esumlay[j]>0.005) // used .010 since that's mip value
	  {
	    ifirst = j;
	    ffound = true;
	  }
      }

    // try setting ifirst back one layer
    if (ifirst>0) ifirst = ifirst-1;
    
    //std::cout << "first plane " << ifirst << std::endl;

    // Find the cell in each layer with maximum energy deposited .....
    for (int i=ifirst; i<ifirst+50; ++i)
      {
	double emax = 0.; // initalize maximum energy variable per layer
	for (int j=0; j<max_cells; ++j)
	  {
	    
	    if (eplcell[i][j] > emax && i<max_layers)
	      {
		emax = eplcell[i][j];
		maxHcell[i] = j;
	      }
	  }
	
      }
    
    
    nhits_tot=0;
    //loop over hits
    for (unsigned int i=0; i<energy_vec.size(); ++i){ 
      
      int plane     = 0;
      int cell      = 0;
      double energy = 0;

      plane  = plane_vec[i];
      energy = energy_vec[i];
      cell   = cell_vec[i];

      //this sets up the bins to be compared against the pre-made hmatricies
      //the longitudinal bins consist of the energy in some number of longitidinal planes
      //the transverse bins then further split up the energy in the central planes by cell
      // it is now 3 layers per longitudinal bin .....
      // transverse inning starts from 3rd layer..... long_layer[2]


      //if (energy>0.0005 && plane<max_layers && cell<max_cells)
      if (plane<max_layers && cell<max_cells) // energy cut is removed ........
	{
	  nhits_tot++;
	  if (plane>ifirst-1 && plane<ifirst+21*nlay)
	    {
	      etot += energy;
	    }

	  if (plane>ifirst-1 && plane<ifirst+nlay)
	    {
	      //  first bin of longitudinal eflow
	      lbin[0] += energy;

	      //std::cout << "max cell " << maxHcell[plane] << std::endl;

	    } else if (plane>ifirst+nlay-1 && plane<ifirst+2*nlay)
	    {
	      lbin[1] += energy;

	    } else if (plane>ifirst+2*nlay-1 && plane<ifirst+3*nlay)
	    {
	      lbin[2] += energy;

	      //  start transverse bins here since 3 layers per bin
	      ttot += energy;
	      //std::cout << "max cell " << maxHcell[plane] << "\t " << ttot << std::endl;
	      if (cell==maxHcell[plane]-3) transm[0] += energy;
	      if (cell==maxHcell[plane]-2) transm[1] += energy;
	      if (cell==maxHcell[plane]-1) transm[2] += energy;
	      if (cell==maxHcell[plane]) transm[3] += energy;
	      if (cell==maxHcell[plane]+1) transm[4] += energy;
	      if (cell==maxHcell[plane]+2) transm[5] += energy;
	      if (cell==maxHcell[plane]+3) transm[6] += energy;

	    } else if (plane>ifirst+3*nlay-1 && plane<ifirst+4*nlay)
	    {
	      lbin[3] += energy;

	      ttot += energy;
	      if (cell==maxHcell[plane]-3) transm[7] += energy;
	      if (cell==maxHcell[plane]-2) transm[8] += energy;
	      if (cell==maxHcell[plane]-1) transm[9] += energy;
	      if (cell==maxHcell[plane]) transm[10] += energy;
	      if (cell==maxHcell[plane]+1) transm[11] += energy;
	      if (cell==maxHcell[plane]+2) transm[12] += energy;
	      if (cell==maxHcell[plane]+3) transm[13] += energy;

	    } else if (plane>ifirst+4*nlay-1 && plane<ifirst+5*nlay)
	    {
	      lbin[4] += energy;

	      ttot += energy;
	      if (cell==maxHcell[plane]-3) transm[14] += energy;
	      if (cell==maxHcell[plane]-2) transm[15] += energy;
	      if (cell==maxHcell[plane]-1) transm[16] += energy;
	      if (cell==maxHcell[plane]) transm[17] += energy;
	      if (cell==maxHcell[plane]+1) transm[18] += energy;
	      if (cell==maxHcell[plane]+2) transm[19] += energy;
	      if (cell==maxHcell[plane]+3) transm[20] += energy;

	    } else if (plane>ifirst+5*nlay-1 && plane<ifirst+6*nlay)
	    {
	      lbin[5] += energy;

	      ttot += energy;
	      if (cell==maxHcell[plane]-3) transm[21] += energy;
	      if (cell==maxHcell[plane]-2) transm[22] += energy;
	      if (cell==maxHcell[plane]-1) transm[23] += energy;
	      if (cell==maxHcell[plane]) transm[24] += energy;
	      if (cell==maxHcell[plane]+1) transm[25] += energy;
	      if (cell==maxHcell[plane]+2) transm[26] += energy;
	      if (cell==maxHcell[plane]+3) transm[27] += energy;

	    } else if (plane>ifirst+6*nlay-1 && plane<ifirst+7*nlay)
	    {
	      lbin[6] += energy;

	      ttot += energy;
	      if (cell==maxHcell[plane]-3) transm[28] += energy;
	      if (cell==maxHcell[plane]-2) transm[29] += energy;
	      if (cell==maxHcell[plane]-1) transm[30] += energy;
	      if (cell==maxHcell[plane]) transm[31] += energy;
	      if (cell==maxHcell[plane]+1) transm[32] += energy;
	      if (cell==maxHcell[plane]+2) transm[33] += energy;
	      if (cell==maxHcell[plane]+3) transm[34] += energy;

	    } else if (plane>ifirst+7*nlay-1 && plane<ifirst+8*nlay)
	    {
	      lbin[7] += energy;

	      ttot += energy;
	      if (cell==maxHcell[plane]-3) transm[35] += energy;
	      if (cell==maxHcell[plane]-2) transm[36] += energy;
	      if (cell==maxHcell[plane]-1) transm[37] += energy;
	      if (cell==maxHcell[plane]) transm[38] += energy;
	      if (cell==maxHcell[plane]+1) transm[39] += energy;
	      if (cell==maxHcell[plane]+2) transm[40] += energy;
	      if (cell==maxHcell[plane]+3) transm[41] += energy;

	    } else if (plane>ifirst+8*nlay-1 && plane<ifirst+9*nlay)
	    {
	      lbin[8] += energy;
	    } else if (plane>ifirst+9*nlay-1 && plane<ifirst+10*nlay)
	    {
	      lbin[9] += energy;
	    } else if (plane>ifirst+10*nlay-1 && plane<ifirst+11*nlay)
	    {
	      lbin[10] += energy;
	    } else if (plane>ifirst+11*nlay-1 && plane<ifirst+12*nlay)
	    {
	      lbin[11] += energy;
	    } else if (plane>ifirst+12*nlay-1 && plane<ifirst+13*nlay)
	    {
	      lbin[12] += energy;
	    } else if (plane>ifirst+13*nlay-1 && plane<ifirst+14*nlay)
	    {
	      lbin[13] += energy;
	    } else if (plane>ifirst+14*nlay-1 && plane<ifirst+15*nlay)
	    {
	      lbin[14] += energy;
	    } else if (plane>ifirst+15*nlay-1 && plane<ifirst+16*nlay)
	    {
	      lbin[15] += energy;
	    } else if (plane>ifirst+16*nlay-1 && plane<ifirst+17*nlay)
	    {
	      lbin[16] += energy;
	    } else if (plane>ifirst+17*nlay-1 && plane<ifirst+18*nlay)
	    {
	      lbin[17] += energy;
	    } else if (plane>ifirst+18*nlay-1 && plane<ifirst+19*nlay)
	    {
	      lbin[18] += energy;
	    } else if (plane>ifirst+19*nlay-1 && plane<ifirst+20*nlay)
	    {
	      lbin[19] += energy;
	    } else if (plane>ifirst+20*nlay-1 && plane<ifirst+21*nlay)
	    {
	      lbin[20] += energy;
	    }  // end of bin formation

	} // end threshold cut
	
    }  // end loop over Hits
      
      
    // get 55 combined long tran bins here-- these bins make the combined matrix
    // first 2 : long bins 0, 1
    // next 39 : trans bin 1 to 41
    // next 12, from lbin 8....

    ltbin[0] = lbin[0];
    ltbin[1] = lbin[1];
    ltbin[2] = transm[1];
    ltbin[3] = transm[2];
    ltbin[4] = transm[3];
    ltbin[5] = transm[4];
    ltbin[6] = transm[5];
    ltbin[7] = transm[7];
    ltbin[8] = transm[8];
    ltbin[9] = transm[9];
    ltbin[10] = transm[10];
    ltbin[11] = transm[11];
    ltbin[12] = transm[12];
    ltbin[13] = transm[13];
    ltbin[14] = transm[14];
    ltbin[15] = transm[15];
    ltbin[16] = transm[16];
    ltbin[17] = transm[17];
    ltbin[18] = transm[18];
    ltbin[19] = transm[19];
    ltbin[20] = transm[20];
    ltbin[21] = transm[21];
    ltbin[22] = transm[22];
    ltbin[23] = transm[23];
    ltbin[24] = transm[24];
    ltbin[25] = transm[25];
    ltbin[26] = transm[26];
    ltbin[27] = transm[27];
    ltbin[28] = transm[28];
    ltbin[29] = transm[29];
    ltbin[30] = transm[30];
    ltbin[31] = transm[31];
    ltbin[32] = transm[32];
    ltbin[33] = transm[33];
    ltbin[34] = transm[34];
    ltbin[35] = transm[35];
    ltbin[36] = transm[36];
    ltbin[37] = transm[37];
    ltbin[38] = transm[38];
    ltbin[39] = transm[39];
    ltbin[40] = transm[40];
    ltbin[41] = transm[41];
    ltbin[42] = lbin[8];
    ltbin[43] = lbin[9];
    ltbin[44] = lbin[10];
    ltbin[45] = lbin[11];
    ltbin[46] = lbin[12];
    ltbin[47] = lbin[13];
    ltbin[48] = lbin[14];
    ltbin[49] = lbin[15];
    ltbin[50] = lbin[16];
    ltbin[51] = lbin[17];
    ltbin[52] = lbin[18];
    ltbin[53] = lbin[19];
    ltbin[54] = lbin[20];
          
    // form E fractions in each layer for long-only bins
      
    fCounter = fCounter+1;
    
    for (int i=0; i<long_bins; ++i)
      {
	
	//std::cout << "lbin " << i << "\t " << lbin[i] << "\t " << etot << std::endl;


	lbin[i] = lbin[i]/etot;
	  
      }
     
      
    // form E fractions in each layer for combined LT
    //the bins are normalized by the total energy
    for (int i=0; i<comb_bins; ++i)
      {
	ltbin[i] = ltbin[i]/etot;
      }
      
      
    // form transverse E fractions in each cell
    //the bins are normalized by the total energy
    //if (ttot<.05) ttot = .05;

    if (ttot<.025) ttot = .025;
    for (int j=0; j<trans_bins; ++j)
      {
	//std::cout << "transbin " << j << "\t " << transm[j] << "\t " << ttot << std::endl;

	transm[j] = transm[j]/ttot;
		  
      }
      


  } // SetupEnergyArrays loop


  std::vector<double> HMatrix::GetHMChisq(){
    //this function calculates a chisqard by comparing the normalized energy per bin as set up 
    //in the SetUpEnergyArrays function against the hmatrix, created using single particle electrons
    // it also calculates the diagonal log chi-squared of the shower

    Double_t lchisq  = 0.;
    Double_t lchisqd = 0.;
    Double_t cchisq  = 0.;
    Double_t cchisqd = 0.;
    Double_t tchisq  = 0.;
    Double_t tchisqd = 0.;
    std::vector<double> chisq_vector;
    // wait


    /*****************************

    std::cout << "----------------------------" << std::endl;
    for ( int i=0; i<long_bins; ++i){
      std::cout << lmeans[i] << " "; 
    }
    std::cout << "\n ----------------------------" << std::endl;
    for ( int i=0; i<long_bins; ++i){
      for ( int j=0; j<long_bins; ++j){
	std::cout << HM[i][j] << " "; 
      }
      std::cout << " " << std::endl;
    }

    std::cout << "\n----------------------------" << std::endl;
    for ( int i=0; i<trans_bins; ++i){
      std::cout << trmeans[i] << " "; 
    }
    std::cout << "\n ----------------------------" << std::endl;

    for (int i=0; i<trans_bins; ++i){
      for (int j=0; j<trans_bins; ++j){
	std::cout << HTM[i][j] << " ";
      }
      std::cout << "\n" << std::endl;
    }

    std::cout << "\n ----------------------------" << std::endl;
    for ( int i=0; i<comb_bins; ++i){
      std::cout << ltmeans[i] << " " ;
    }
    std::cout << "\n ----------------------------" << std::endl;

    for (int i=0; i<comb_bins; ++i){
      for (int j=0; j<comb_bins; ++j){
	std::cout << HC[i][j] << " ";
      }
      std::cout << "\n" << std::endl;
    }

    std::cout << "----------------------------" << std::endl;
    for ( int i=0; i<long_bins; ++i){
      std::cout << lbin[i] << " "; 
    }
    std::cout << "\n ----------------------------" << std::endl;
    for ( int i=0; i<trans_bins; ++i){
      std::cout << transm[i] << " "; 
    }
    std::cout << "\n ----------------------------" << std::endl;
    for ( int i=0; i<comb_bins; ++i){
      std::cout << ltbin[i] << " "; 
    }
    std::cout << "\n ----------------------------" << std::endl;

    *****************************/

    // don't divide by nhits_tot, no energy cut on transverse chi2
    // check the size if number of bins in the HMatrix are changed ....
    for ( int i=0; i<long_bins; ++i)
      {
	//lchisqd += (lbin[i]-lmeans[i])*HM[i][i]*(lbin[i]-lmeans[i])/nhits_tot;
	lchisqd += (lbin[i]-lmeans[i])*HM[i][i]*(lbin[i]-lmeans[i]);
	for ( int j=0; j<long_bins; ++j)
	  {
	    //lchisq += (lbin[i]-lmeans[i])*HM[i][j]*(lbin[j]-lmeans[j])/nhits_tot;
	    lchisq += (lbin[i]-lmeans[i])*HM[i][j]*(lbin[j]-lmeans[j]);
		  
	  }
      }


    // wait : why there is an energy cut for transverse chi2 ............ now it is commented....
    for ( int i=0; i<trans_bins; ++i)
      {
	//if (ttot>0.05) tchisqd += (transm[i]-trmeans[i])*HTM[i][i]*(transm[i]-trmeans[i])/nhits_tot;
	//tchisqd += (transm[i]-trmeans[i])*HTM[i][i]*(transm[i]-trmeans[i])/nhits_tot;
	tchisqd += (transm[i]-trmeans[i])*HTM[i][i]*(transm[i]-trmeans[i]);
	for ( int j=0; j<trans_bins; ++j)
	  {
	    //if (ttot>0.05) tchisq += (transm[i]-trmeans[i])*HTM[i][j]*(transm[j]-trmeans[j])/nhits_tot;
	    //tchisq += (transm[i]-trmeans[i])*HTM[i][j]*(transm[j]-trmeans[j])/nhits_tot;
	    tchisq += (transm[i]-trmeans[i])*HTM[i][j]*(transm[j]-trmeans[j]);
	  }
      }


    for ( int i=0; i<comb_bins; ++i)
      {
	//cchisqd += (ltbin[i]-ltmeans[i])*HC[i][i]*(ltbin[i]-ltmeans[i])/nhits_tot;
	cchisqd += (ltbin[i]-ltmeans[i])*HC[i][i]*(ltbin[i]-ltmeans[i]);
	for ( int j=0; j<comb_bins; ++j)
	  {
	    //cchisq += (ltbin[i]-ltmeans[i])*HC[i][j]*(ltbin[j]-ltmeans[j])/nhits_tot;
	    cchisq += (ltbin[i]-ltmeans[i])*HC[i][j]*(ltbin[j]-ltmeans[j]);
	  }
      }

    chisq_vector.push_back(lchisq);
    chisq_vector.push_back(cchisq);
    chisq_vector.push_back(tchisq);
    chisq_vector.push_back(lchisqd);
    chisq_vector.push_back(cchisqd);							       
    chisq_vector.push_back(tchisqd);

    return chisq_vector;
  }

  //A function that grabs the MCTruth information
  int HMatrix::GetMCTruthM(const art::Event& evt,art::PtrVector<simb::MCTruth>& mctruth){
    mctruth.clear();
    art::PtrVector<simb::MCTruth> temp;
 

    // going to use getManyByType so that we don't have to 
    // worry about using a module label
    std::vector< art::Handle< std::vector<simb::MCTruth> > > tcol;

   
    evt.getManyByType(tcol);
    for(size_t mctc = 0; mctc < tcol.size(); ++mctc){
      for(unsigned int i = 0; i < tcol[mctc]->size(); ++i){
	art::Ptr<simb::MCTruth> t(tcol[mctc], i);
	temp.push_back(t);
      }
    }
    temp.swap(mctruth);
 

    return mctruth.size();
  }

  //------------------------------------------------------------------
  // a function that figures out if events are vertex fiducial. It is probably a outdated.
  bool HMatrix::isInVertexFiducial(art::Ptr<rb::Track> trk) const
  {
    // Start and end don't mean much, pick whichever is lowest in y.
    const TVector3 v1 = trk->Start();
    const TVector3 v2 = trk->Stop();
    //const TVector3 start = (v1.Y() < v2.Y()) ? v1 : v2; 
    const TVector3 start = v2;

    const double x = start.X();
    const double y = start.Y();
    const double z = start.Z();

    art::ServiceHandle<geo::Geometry> geom;
    switch(geom->DetId()){

      //case daqchannelmap::DAQChannelMap::NDOS_ID:
      //case daqchannelmap::DAQChannelMap::NEARDET_ID:

      //case daqchannelmap::NDOS_ID:
      //case daqchannelmap::NEARDET_ID:

    case novadaq::cnv::kNDOS:
    case novadaq::cnv::kNEARDET:



      if(fabs(x) > 110) return false; // Fiducial X
      if(fabs(y) > 180) return false; // Fiducial Y
      if(z < 20 || z > 1400) return false; // Fiducial Z
      
      if(v1.X()==v2.X()||v1.Y()==v2.Y()) return false; // dx or dy = 0 means bad fit
      break;
      //case daqchannelmap::DAQChannelMap::FARDET_ID:

      //case daqchannelmap::FARDET_ID:
    case novadaq::cnv::kFARDET:

      if(fabs(x) > 730 || fabs(y) > 730) return false;
      if(z < 50 || z > 6300) return false;
      
      if(v1.X()==v2.X()||v1.Y()==v2.Y()) return false;
      break;
    default:
      assert(0 && "Unknown detector");
    }

    return true;
  }


} // end of namespace hme




namespace hme{
  
  DEFINE_ART_MODULE(HMatrix);
  
}
