////////////////////////////////////////////////////////////////////////
// \file    MakeHM.cxx
// \brief   TODO
// \version $Id: MakeHM.cxx,v 1.10 2012-09-25 20:22:05 gsdavies Exp $
// \author  $Author: gsdavies $
////////////////////////////////////////////////////////////////////////

// C++ includes
#include <cmath>
#include <iostream>
#include <string>
#include <vector>

// ROOT includes
#include "TH1F.h"
#include "TH2F.h"
#include "TMatrixD.h"

// ART includes
#include "art/Framework/Principal/Event.h" 
#include "art/Framework/Services/Optional/TFileService.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"  
#include "art/Framework/Principal/Handle.h" 
#include "messagefacility/MessageLogger/MessageLogger.h" 
#include "art/Framework/Core/EDAnalyzer.h"
#include "art/Framework/Core/ModuleMacros.h" 

// NOvA includes
#include "Geometry/Geometry.h"
#include "MCCheater/SimHit.h"
#include "RawData/RawDigit.h"


namespace hme {
  
  class MakeHM : public art::EDAnalyzer {
    
  public:
    
    explicit MakeHM(fhicl::ParameterSet const &pset);  // the explicit tag tells the compiler that MakeHM is not an art::ParameterSet
    virtual ~MakeHM();
    
    // Optional, read/write access to event       
    void analyze(const art::Event& evt); 
    void beginJob();
    
  private:
    
    double         prod[6][6];
    double         tran[7][7];
    int            ntest;
    std::string    fSimHitListLabel;
    
    //  histograms
    TH1F*       fXRange;
    TH1F*       fYRange;
    TH1F*       fZRange;
    TH1F*       fnshptr;
    TH1F*       fcorx;
    TH1F*       fcory;
    TH2F*       fstsxy;
    TH2F*       fstexy;
    TH1F*       fe03;
    TH1F*       fe47;
    TH1F*       fe811;
    TH1F*       fe1215;
    TH1F*       fe1619;
    TH1F*       fe2023;
    TH1F*       fe2427;
    TH1F*       fe2831;
    TH1F*       fe3235;
    TH1F*       fe3639;
    TH1F*       fe4043;
    TH1F*       fe4447;
    TH1F*       fe4851;
    TH1F*       fe5255;
    TH1F*       fe5659;
    TH1F*       fetot;
    TH1F*       fttot;
    TH1F*       flchisq;
    TH1F*       flchisqd;
    TH1F*       ftchisq;
    TH1F*       ftchisqd;
    
  };
}


namespace hme 
{
  //........................................................................
  MakeHM::MakeHM(fhicl::ParameterSet const& pset) 
    : EDAnalyzer(pset),
      fSimHitListLabel   (pset.get< std::string >("SimHitListLabel") )
  {
    // Be noisy to demonstrate what's happening
    mf::LogInfo("HMatrixE") << " MakeHM::MakeHM()\n";
    // print this to keep track of version number
    mf::LogInfo("HMatrixE") << "          ****************************************\n"
			    << "          *     MakeHM Version Number 1          *\n"
			    << "          ****************************************\n";
  }
  
  //..............................................................
  MakeHM::~MakeHM()
  {
  }
  
  //......................................................................
  void MakeHM::beginJob()
  {
    //
    // Book histograms, ntuples, initialize counts etc., etc., ...
    //
    art::ServiceHandle<art::TFileService> tfs;
    
    fXRange  = tfs->make<TH1F>("fXRange", ";Range of SimHit X Coords", 120, -210., 210.);
    fYRange  = tfs->make<TH1F>("fYRange", ";Range of SimHit Y Coords", 120, -210., 210.);
    fZRange  = tfs->make<TH1F>("fZRange", ";Range of SimHit Z Coords", 100, 0., 1000.);
    fnshptr  = tfs->make<TH1F>("fnshptr", ";Number of SimHits per Track", 50, 0., 100.);
    fcorx    = tfs->make<TH1F>("fcorx", ";X Coord of xview plane", 120, -810., 810.);
    fcory    = tfs->make<TH1F>("fcory", ";Y Coord of yview plane", 120, -810., 810.);
    fe03     = tfs->make<TH1F>("fe03", ";E Frac Layers 0-3", 50, 0., 1.);
    fe47     = tfs->make<TH1F>("fe47", ";E Frac Layers 4-7", 50, 0., 1.);
    fe811    = tfs->make<TH1F>("fe811", ";E Frac Layers 8-11", 50, 0., 1.);
    fe1215   = tfs->make<TH1F>("fe1215", ";E Frac Layers 12-15", 50, 0., 1.);
    fe1619   = tfs->make<TH1F>("fe1619", ";E Frac Layers 16-19", 50, 0., 1.);
    fe2023   = tfs->make<TH1F>("fe2023", ";E Frac Layers 20-23", 50, 0., 1.);
    fe2427   = tfs->make<TH1F>("fe2427", ";E Frac Layers 24-27", 50, 0., 1.);
    fe2831   = tfs->make<TH1F>("fe2831", ";E Frac Layers 28-31", 50, 0., 1.);
    fe3235   = tfs->make<TH1F>("fe3235", ";E Frac Layers 32-35", 50, 0., 1.);
    fe3639   = tfs->make<TH1F>("fe3639", ";E Frac Layers 36-39", 50, 0., 1.);
    fe4043   = tfs->make<TH1F>("fe4043", ";E Frac Layers 40-43", 50, 0., 1.);
    fe4447   = tfs->make<TH1F>("fe4447", ";E Frac Layers 44-47", 50, 0., 1.);
    fe4851   = tfs->make<TH1F>("fe4851", ";E Frac Layers 48-51", 50, 0., 1.);
    fe5255   = tfs->make<TH1F>("fe5255", ";E Frac Layers 52-55", 50, 0., 1.);
    fe5659   = tfs->make<TH1F>("fe5659", ";E Frac Layers 56-59", 50, 0., 1.);
    fetot    = tfs->make<TH1F>("fetot", ";Total E all Layers", 50, 0., 5.);
    fttot    = tfs->make<TH1F>("fttot", ";Total E all Trans cells", 50, 0., 5.);
    flchisq  = tfs->make<TH1F>("flchisq", ";Log Long Chi-squared Distribution", 100, 0., 10.);
    flchisqd = tfs->make<TH1F>("flchisqd", ";Log Diagonal Long Chi-squared Distribution", 100, 0., 10.);
    ftchisq  = tfs->make<TH1F>("ftchisq", ";Log Trans Chi-squared Distribution", 100, 0., 10.);
    ftchisqd = tfs->make<TH1F>("ftchisqd", ";Log Diagonal Trans Chi-squared Distribution", 100, 0., 10.);
    
    ntest = 0;
    
    return;
  }
  
  //.....................................................................
  void MakeHM::analyze(const art::Event& evt)
  {
    
    mf::LogInfo("MakeHM") << "analyze() " << " New Event\n";
    
    art::ServiceHandle<geo::Geometry> geom;
    
    // increment counter for M matrix
    ntest++;
    
    //      art::PtrVector<cheat::SimHit> simhits;
    art::Handle<std::vector<cheat::SimHit> > simhits;
    evt.getByLabel(fSimHitListLabel, simhits);
    
    //  Run on single electrons, find averages for H-matrix 
    double etot  = 0.;
    double e03   = 0.;
    double e47   = 0.;
    double e811  = 0.;
    double e1215 = 0.;
    double e1619 = 0.;
    double e2023 = 0.;
    double e2427 = 0.;
    double e2831 = 0.;
    double e3235 = 0.;
    double e3639 = 0.;
    double e4043 = 0.;
    double e4447 = 0.;
    double e4851 = 0.;
    double e5255 = 0.;
    double e5659 = 0.;

    double elay[16];
    double etest[6]  = {0.,0.,0.,0.,0.,0.};
    double transm[7] = {0.,0.,0.,0.,0.,0.,0.};
    double ttot      = 0.;
    
    //  Loop over all hits used for making the HMatrix - this should be all
    //  SimHits when making HMatrix from single particles (electrons or photons)
    //  First, need to know where to start the binning for the HMatrix - can I
    //  first loop over the hits and sort them according to Plane number?
    for (unsigned int i=0; i<simhits->size(); ++i)
      {
	art::Ptr<cheat::SimHit> k(simhits, i);
      }
    
    for (unsigned int i=0; i<simhits->size(); ++i)  // loop over SimHits
      {
	
	art::Ptr<cheat::SimHit> h(simhits, i);
	
	etot += h->fTotE;
	elay[15] += h->fTotE;
	etest[5] += h->fTotE;
	if (h->Plane()<4)
          {
	    e03 += h->fTotE;
	    elay[0] += h->fTotE;
	    etest[0] += h->fTotE;
          } else if (h->Plane()>3 && h->Plane()<8)
          {
	    e47 += h->fTotE;
	    elay[1] += h->fTotE;
	    etest[0] += h->fTotE;
          } else if (h->Plane()>7 && h->Plane()<12)
          {
	    e811 += h->fTotE;
	    elay[2] += h->fTotE;
	    etest[0] += h->fTotE;
          } else if (h->Plane()>11 && h->Plane()<16)
          {
	    e1215 += h->fTotE;
	    elay[3] += h->fTotE;
	    etest[1] += h->fTotE;
	    if (h->Plane()==12 || h->Plane()==14)
              {
		ttot += h->fTotE;
		transm[6] += h->fTotE;
		if (h->Cell()==29) transm[0] += h->fTotE;
		if (h->Cell()==30) transm[1] += h->fTotE;
		if (h->Cell()==31) transm[2] += h->fTotE;
		if (h->Cell()==32) transm[3] += h->fTotE;
		if (h->Cell()==33) transm[4] += h->fTotE;
		if (h->Cell()==34) transm[5] += h->fTotE;
              }
	    if (h->Plane()==13 || h->Plane()==15)
              {
		ttot += h->fTotE;
		transm[6] += h->fTotE;
		if (h->Cell()==47) transm[0] += h->fTotE;
		if (h->Cell()==48) transm[1] += h->fTotE;
		if (h->Cell()==49) transm[2] += h->fTotE;
		if (h->Cell()==50) transm[3] += h->fTotE;
		if (h->Cell()==51) transm[4] += h->fTotE;
		if (h->Cell()==52) transm[5] += h->fTotE;
              }
          } 
	else if (h->Plane()>15 && h->Plane()<20)
          {
	    e1619 += h->fTotE;
	    elay[4] += h->fTotE;
	    etest[1] += h->fTotE;
	    if (h->Plane()==16 || h->Plane()==18)
              {
		ttot += h->fTotE;
		transm[6] += h->fTotE;
		if (h->Cell()==29) transm[0] += h->fTotE;
		if (h->Cell()==30) transm[1] += h->fTotE;
		if (h->Cell()==31) transm[2] += h->fTotE;
		if (h->Cell()==32) transm[3] += h->fTotE;
		if (h->Cell()==33) transm[4] += h->fTotE;
		if (h->Cell()==34) transm[5] += h->fTotE;
              }
	    if (h->Plane()==17 || h->Plane()==19)
              {
		ttot += h->fTotE;
		transm[6] += h->fTotE;
		if (h->Cell()==47) transm[0] += h->fTotE;
		if (h->Cell()==48) transm[1] += h->fTotE;
		if (h->Cell()==49) transm[2] += h->fTotE;
		if (h->Cell()==50) transm[3] += h->fTotE;
		if (h->Cell()==51) transm[4] += h->fTotE;
		if (h->Cell()==52) transm[5] += h->fTotE;
              }
          } 
	else if (h->Plane()>19 && h->Plane()<24)
          {
	    e2023 += h->fTotE;
	    elay[5] += h->fTotE;
	    etest[1] += h->fTotE;
          } 
	else if (h->Plane()>23 && h->Plane()<28)
          {
	    e2427 += h->fTotE;
	    elay[6] += h->fTotE;
	    etest[2] += h->fTotE;
          }
	else if (h->Plane()>27 && h->Plane()<32)
          {
	    e2831 += h->fTotE;
	    elay[7] += h->fTotE;
	    etest[2] += h->fTotE;
          } 
	else if (h->Plane()>31 && h->Plane()<36)
          {
	    e3235 += h->fTotE;
	    elay[8] += h->fTotE;
	    etest[2] += h->fTotE;
          } 
	else if (h->Plane()>35 && h->Plane()<40)
          {
	    e3639 += h->fTotE;
	    elay[9] += h->fTotE;
	    etest[3] += h->fTotE;
          } 
	else if (h->Plane()>39 && h->Plane()<44)
	  {
	    e4043 += h->fTotE;
	    elay[10] += h->fTotE;
	    etest[3] += h->fTotE;
          } 
	else if (h->Plane()>43 && h->Plane()<48)
          {
	    e4447 += h->fTotE;
	    elay[11] += h->fTotE;
	    etest[3] += h->fTotE;
          } 
	else if (h->Plane()>47 && h->Plane()<52)
          {
	    e4851 += h->fTotE;
	    elay[12] += h->fTotE;
	    etest[4] += h->fTotE;
          } 
	else if (h->Plane()>51 && h->Plane()<56)
          {
	    e5255 += h->fTotE;
	    elay[13] += h->fTotE;
	    etest[4] += h->fTotE;
          } 
	else if (h->Plane()>55 && h->Plane()<60)
	  {
	    e5659 += h->fTotE;
	    elay[14] += h->fTotE;
	    etest[4] += h->fTotE;
          }
      }  // end loop over SimHits
    
    fe03->Fill(etest[0]/etot);
    fe47->Fill(etest[1]/etot);
    fe811->Fill(etest[2]/etot);
    fe1215->Fill(etest[3]/etot);
    fe1619->Fill(etest[4]/etot);
    fe2023->Fill(transm[0]/ttot);
    fe2427->Fill(transm[1]/ttot);
    fe2831->Fill(transm[2]/ttot);
    fe3235->Fill(transm[3]/ttot);
    fe3639->Fill(transm[4]/ttot);
    fe4043->Fill(transm[5]/ttot);
    fe4447->Fill(e4447/etot);
    fe4851->Fill(e4851/etot);
    fe5255->Fill(e5255/etot);
    fe5659->Fill(e5659/etot);
    fetot->Fill(etot);
    fttot->Fill(ttot);
    
    //  form E fractions in each layer
    std::cout << " Long E fractions \n";
    for (unsigned int i=0; i<6; ++i)
      {
	if (i<5) etest[i] = etest[i]/etot;
	std::cout << "  " << etest[i];
      }
    std::cout << "\n";
    
    //  form transverse E fractions in each cell
    std::cout << " Trans E fractions \n";
    for (unsigned int j=0; j<7; ++j)
      {
	if (j<6) transm[j] = transm[j]/ttot;
	std::cout << "  " << transm[j];
      }
    std::cout << "\n";
    
    double trmeans[7] = {.007,.034,.382,.504,.046,.002,.385};
    double tmeans[6] = {.158,.398,.260,.116,.044,1.417};
    //double means[16] = {.008,.052,.098,.136,.136,.127,.109,.085,.066,.049,.038,.028,.020,.014,.010,1.417};
    
    for (unsigned int k=0; k<6; ++k)
      {
	for (unsigned int l=0; l<6; ++l)
          {
	    prod[k][l] += (etest[k]-tmeans[k])*(etest[l]-tmeans[l]);
          }
      }
    
    for (unsigned int k=0; k<7; ++k)
      {
	for (unsigned int l=0; l<7; ++l)
          {
	    tran[k][l] += (transm[k]-trmeans[k])*(transm[l]-trmeans[l]);
          }
      }
    
    //  get long matrix so far
    double Mij[6][6] = {{0}};
    for (unsigned int i=0; i<6; ++i)
      {
	for (unsigned int j=0; j<6; ++j)
          {
	    Mij[i][j] = (double) prod[i][j]/ntest;
          }
      }
    
    //  get Trans matrix so far
    double Trij[7][7] = {{0}};
    for (unsigned int i=0; i<7; ++i)
      {
	for (unsigned int j=0; j<7; ++j)
          {
	    Trij[i][j] = (double) tran[i][j]/ntest;
          }
      }
    
    double dmij[36] = {0};
    int ij = 0;
    std::cout << " Matrix Mij after " << ntest << " events\n";
    for (unsigned int i=0; i<6; ++i)
      {
	std::cout << "Row " << i;
	for (unsigned int j=0; j<6; ++j)
          {
	    dmij[ij] = Mij[i][j];
	    std::cout << "  " << Mij[i][j];
	    //                 std::cout << "  " << prod[i][j];
	    ij++;
          }
	std::cout << "\n";
      }
    
    // print out the transverse matrix
    double tmij[49] = {0};
    ij = 0;
    std::cout << " Matrix Trij after " << ntest << " events\n";
    for (unsigned int i=0; i<7; ++i)
      {
	std::cout << "Row " << i;
	for (unsigned int j=0; j<7; ++j)
          {
	    tmij[ij] = Trij[i][j];
	    std::cout << "  " << Trij[i][j];
	    //                 std::cout << "  " << tran[i][j];
	    ij++;
          }
	std::cout << "\n";
      }
    
    /*
    double testm[4];
    testm[0] = .001;
    testm[1] = .002;
    testm[2] = .003;
    testm[3] = .004;
    
    //  make a matrix in ROOT
    TMatrixD h(2,2,testm);
    */

    TMatrixD h(6,6,dmij);
    //         TArrayD data(36);
    std::cout << "Made Longitudinal TMatrixD with dmij array \n";
    
    TMatrixD ht(7,7,tmij);
    std::cout << "Made Transverse TMatrixD with tmij array \n";
    //          for (unsigned int k=0; k<36; ++k)
    //          {
    //              std::cout << "dmij array = " << dmij[k] << "\n";
    //          }
    
    //          h.SetMatrixArray(data);
    
    //  invert h to get H
    //          TMatrixD H(6,6);
    
    //          TMatrixD h;
    //          TArrayD data(36,dmij);
    //          h.SetMatrixArray(data);
    
    if (ntest > 998)
      {
	TMatrixD H = h.Invert();
	std::cout << "After Long Matrix Inversion \n";
	std::cout << TMatrixDRow(H,0)(0) << " " << TMatrixDRow(H,0)(1) << " " 
		  << TMatrixDRow(H,0)(2) << " " << TMatrixDRow(H,0)(3) << " " 
		  << TMatrixDRow(H,0)(4) << " " << TMatrixDRow(H,0)(5) << " " << "\n";
	std::cout << TMatrixDRow(H,1)(0) << " " << TMatrixDRow(H,1)(1) << " "
		  << TMatrixDRow(H,1)(2) << " " << TMatrixDRow(H,1)(3) << " "
		  << TMatrixDRow(H,1)(4) << " " << TMatrixDRow(H,1)(5) << " " << "\n";
	std::cout << TMatrixDRow(H,2)(0) << " " << TMatrixDRow(H,2)(1) << " "
		  << TMatrixDRow(H,2)(2) << " " << TMatrixDRow(H,2)(3) << " "
		  << TMatrixDRow(H,2)(4) << " " << TMatrixDRow(H,2)(5) << " " << "\n";
	std::cout << TMatrixDRow(H,3)(0) << " " << TMatrixDRow(H,3)(1) << " "
		  << TMatrixDRow(H,3)(2) << " " << TMatrixDRow(H,3)(3) << " "
		  << TMatrixDRow(H,3)(4) << " " << TMatrixDRow(H,3)(5) << " " << "\n";
	std::cout << TMatrixDRow(H,4)(0) << " " << TMatrixDRow(H,4)(1) << " "
		  << TMatrixDRow(H,4)(2) << " " << TMatrixDRow(H,4)(3) << " "
		  << TMatrixDRow(H,4)(4) << " " << TMatrixDRow(H,4)(5) << " " << "\n";
	std::cout << TMatrixDRow(H,5)(0) << " " << TMatrixDRow(H,5)(1) << " "
		  << TMatrixDRow(H,5)(2) << " " << TMatrixDRow(H,5)(3) << " "
		  << TMatrixDRow(H,5)(4) << " " << TMatrixDRow(H,5)(5) << " " << "\n";
	TMatrixD HT = ht.Invert();
	std::cout << "After Trans Matrix Inversion \n";
	std::cout << TMatrixDRow(HT,0)(0) << " " << TMatrixDRow(HT,0)(1) << " " 
		  << TMatrixDRow(HT,0)(2) << " " << TMatrixDRow(HT,0)(3) << " " 
		  << TMatrixDRow(HT,0)(4) << " " << TMatrixDRow(HT,0)(5) << " "
		  << TMatrixDRow(HT,0)(6) << "\n";
	std::cout << TMatrixDRow(HT,1)(0) << " " << TMatrixDRow(HT,1)(1) << " "
		  << TMatrixDRow(HT,1)(2) << " " << TMatrixDRow(HT,1)(3) << " "
		  << TMatrixDRow(HT,1)(4) << " " << TMatrixDRow(HT,1)(5) << " "
		  << TMatrixDRow(HT,1)(6) << "\n";
	std::cout << TMatrixDRow(HT,2)(0) << " " << TMatrixDRow(HT,2)(1) << " "
		  << TMatrixDRow(HT,2)(2) << " " << TMatrixDRow(HT,2)(3) << " "
		  << TMatrixDRow(HT,2)(4) << " " << TMatrixDRow(HT,2)(5) << " "
		  << TMatrixDRow(HT,2)(6) << "\n";
	std::cout << TMatrixDRow(HT,3)(0) << " " << TMatrixDRow(HT,3)(1) << " "
		  << TMatrixDRow(HT,3)(2) << " " << TMatrixDRow(HT,3)(3) << " "
		  << TMatrixDRow(HT,3)(4) << " " << TMatrixDRow(HT,3)(5) << " "
		  << TMatrixDRow(HT,3)(6) << "\n";
	std::cout << TMatrixDRow(HT,4)(0) << " " << TMatrixDRow(HT,4)(1) << " "
		  << TMatrixDRow(HT,4)(2) << " " << TMatrixDRow(HT,4)(3) << " "
		  << TMatrixDRow(HT,4)(4) << " " << TMatrixDRow(HT,4)(5) << " "
		  << TMatrixDRow(HT,4)(6) << "\n";
	std::cout << TMatrixDRow(HT,5)(0) << " " << TMatrixDRow(HT,5)(1) << " "
		  << TMatrixDRow(HT,5)(2) << " " << TMatrixDRow(HT,5)(3) << " "
		  << TMatrixDRow(HT,5)(4) << " " << TMatrixDRow(HT,5)(5) << " "
		  << TMatrixDRow(HT,5)(6) << "\n";
	std::cout << TMatrixDRow(HT,6)(0) << " " << TMatrixDRow(HT,6)(1) << " "
		  << TMatrixDRow(HT,6)(2) << " " << TMatrixDRow(HT,6)(3) << " "
		  << TMatrixDRow(HT,6)(4) << " " << TMatrixDRow(HT,6)(5) << " "
		  << TMatrixDRow(HT,6)(6) << "\n";
      }
    
    //  test chisquared for electrons
    double HM[6][6];
    HM[0][0]=797.5;  HM[0][1]=538.3;  HM[0][2]=598.1;  HM[0][3]=598.4;  HM[0][4]=819.3;  HM[0][5]=22.3;
    HM[1][0]=538.3;  HM[1][1]=638.1;  HM[1][2]=560.1;  HM[1][3]=618.5;  HM[1][4]=841.9;  HM[1][5]=-0.80;
    HM[2][0]=598.1;  HM[2][1]=560.1;  HM[2][2]=649.3;  HM[2][3]=550.7;  HM[2][4]=839.5;  HM[2][5]=9.5;
    HM[3][0]=598.4;  HM[3][1]=618.5;  HM[3][2]=550.7;  HM[3][3]=817.1;  HM[3][4]=686.4;  HM[3][5]=10.9;
    HM[4][0]=819.3;  HM[4][1]=841.9;  HM[4][2]=839.5;  HM[4][3]=686.4;  HM[4][4]=1707.8; HM[4][5]=-1.1;
    HM[5][0]=22.3;   HM[5][1]=-0.80;  HM[5][2]=9.5;    HM[5][3]=10.9;   HM[5][4]=-1.1;   HM[5][5]=399.9;
    
    double HTM[7][7];
    HTM[0][0]=10400.4; HTM[0][1]=2826.7;  HTM[0][2]=3226.8;  HTM[0][3]=3202.9;  HTM[0][4]=3302.2;  HTM[0][5]=3201.0;  HTM[0][6]=45.9;
    HTM[1][0]=2826.7;  HTM[1][1]=3069.7;  HTM[1][2]=2243.0;  HTM[1][3]=2274.2;  HTM[1][4]=2249.2;  HTM[1][5]=2387.9;  HTM[1][6]=-7.4;
    HTM[2][0]=3226.8;  HTM[2][1]=2243.0;  HTM[2][2]=2238.8;  HTM[2][3]=2162.5;  HTM[2][4]=2262.2;  HTM[2][5]=2318.1;  HTM[2][6]=-10.9;
    HTM[3][0]=3202.9;  HTM[3][1]=2274.2;  HTM[3][2]=2162.5;  HTM[3][3]=2216.7;  HTM[3][4]=2242.1;  HTM[3][5]=2339.4;  HTM[3][6]=-4.5;
    HTM[4][0]=3302.2;  HTM[4][1]=2249.2;  HTM[4][2]=2262.2;  HTM[4][3]=2242.1;  HTM[4][4]=2936.8;  HTM[4][5]=1904.6;  HTM[4][6]=-14.4;
    HTM[5][0]=3201.0;  HTM[5][1]=2387.9;  HTM[5][2]=2318.1;  HTM[5][3]=2339.4;  HTM[5][4]=1904.6;  HTM[5][5]=7707.0;  HTM[5][6]=31.0;
    HTM[6][0]=45.9;    HTM[6][1]=-7.4;    HTM[6][2]=-10.9;   HTM[6][3]=-4.5;    HTM[6][4]=-14.4;   HTM[6][5]=31.0;    HTM[6][6]=48.6;
    
    //  use the H matrix to calculate chi-squared for these electrons
    double lchisq = 0.;
    double lchisqd = 0.;
    double tchisq = 0.;
    double tchisqd = 0.;
    for (unsigned int i=0; i<6; ++i)
      {
	for (unsigned int j=0; j<6; ++j)
          {
	    lchisq += (etest[i]-tmeans[i])*HM[i][j]*(etest[j]-tmeans[j]);
	    if (i==j) lchisqd += (etest[i]-tmeans[i])*HM[i][j]*(etest[j]-tmeans[j]);
          }
      }
    for (unsigned int i=0; i<7; ++i)
      {
	for (unsigned int j=0; j<7; ++j)
          {
	    tchisq += (transm[i]-trmeans[i])*HTM[i][j]*(transm[j]-trmeans[j]);
	    if (i==j) tchisqd += (transm[i]-trmeans[i])*HTM[i][j]*(transm[j]-trmeans[j]);
          }
      }
    
    
    std::cout << " Long Chi-squared of this shower = " << lchisq << "\n";
    std::cout << " Diagonal Long Chi-squared of this shower = " << lchisqd << "\n";
    std::cout << " Trans Chi-squared of this shower = " << tchisq << "\n";
    std::cout << " Diagonal Trans Chi-squared of this shower = " << tchisqd << "\n";
    flchisq->Fill(log(lchisq));
    flchisqd->Fill(log(lchisqd));
    ftchisq->Fill(log(tchisq));
    ftchisqd->Fill(log(tchisqd));
    
    return;
  }

}  // namespace hmatrixe

 
namespace hme
{ 
   DEFINE_ART_MODULE(MakeHM);
}

