////////////////////////////////////////////////////////////////////////
/// \file  BackTracker.h
/// \brief back track the reconstruction to the simulation
///
/// \version $Id: BackTracker.h,v 1.37 2012-12-06 15:03:39 lein Exp $
/// \author  brebel@fnal.gov
////////////////////////////////////////////////////////////////////////


#ifndef CHEAT_BACKTRACKER_H
#define CHEAT_BACKTRACKER_H

#include <map>
#include <unordered_map>
#include <vector>

#include "GeometryObjects/OfflineChan.h"
#include "GeometryObjects/PlaneGeo.h"
#include "NovaDAQConventions/DAQConventions.h"
#include "RecoBase/CellHit.h"
#include "RecoBase/WeightedHit.h"
#include "RecoBase/Cluster.h"
#include "RecoBase/Track.h"
#include "Simulation/ParticleNavigator.h"
#include "Simulation/Particle.h"
#include "Simulation/FLSHit.h"
#include "Simulation/PhotonSignal.h"
#include "SimulationBase/MCTruth.h"

#include "art/Framework/Services/Registry/ActivityRegistry.h"
#include "art/Framework/Services/Registry/ServiceMacros.h"
#include "fhiclcpp/ParameterSet.h"


///code to link reconstructed objects back to the MC truth information
namespace cheat{

  typedef struct{
    int trackID;      ///< Geant4 supplied trackID
    float energyFrac; ///< fraction of hit energy from the particle with this trackID
    float energy;     ///< energy from the particle with this trackID
  } TrackIDE;

  typedef struct{
    art::Ptr<simb::MCTruth> neutrinoInt;   ///< Truth information about neutrino interaction
    double efficiency;                     ///< Efficiency (based on FLS energy) of neutrino interaction relative to slice
    double purity;                         ///< Purity (based on FLS energy) of neutrino interaction relative to slice
    unsigned int    truthColIndex;         ///< Position in column vector of MCTruths from the event
    double energySlice;                    ///< Sum of FLS hits from the neutrino contributing to hits included in the slice
    double energyTotal;                    ///< Sum of FLS hits from the neutrino contributing to any hit in the event
    int nSliceHits;                        ///< Number of hits from this neutrino in this slice
    int nTotalHits;                        ///< Total number of hits the neutrino left in the detector

  } NeutrinoEffPur;

  typedef struct{
    art::Ptr<simb::MCTruth> neutrinoInt;   ///< Truth information about neutrino interaction
    unsigned int    truthColIndex;         ///< Position in column vector of MCTruths from the event
  } NeutrinoWithIndex;

  typedef struct{
    int trackID;                           ///< Truth particle track ID
    double efficiency;                     ///< Efficiency of particle relative to track
    double purity;                         ///< Putity of particle relative to track
    int pdg;                               ///< Pdg of particle
  } ParticleEffPur;

  /// Does \a a have less energy than \a b?
  bool CompareByEnergy(const TrackIDE& a, const TrackIDE& b);

  class BackTracker
  {
  public:
    BackTracker(const fhicl::ParameterSet& pset,
                art::ActivityRegistry& reg);
    ~BackTracker();

    void reconfigure(const fhicl::ParameterSet& pset);

    /// This function rebuilds the various maps we need to answer backtracking queries.
    /// It is called automatically before each event is processed. For jobs involving
    /// Monte Carlo generation, this is too soon. So, ReadoutSim also calls this function
    /// when it knows all the relevant products have been produced. Requiring this kind
    /// of cooperation from other modules for a service to function properly is not
    /// recommended in general. In this case it seems like the least-bad option, but
    /// think very carefully before adding any other examples. "Do as I say, not as I do".
    void Rebuild(const art::Event& evt);

    /// \brief Is this a file with truth info in? (Is BackTracker going to be
    /// any use to you?)
    ///
    /// True if we succeeded in retrieving FLSHits, ie in general for MC
    /// files. Expected false in data, and mock data etc.
    bool HaveTruthInfo() const {return fHaveTruthInfo;}

    /// \brief Returns the PhotonSignals contributing the signal in the
    /// specified cell. WARNING: Use with extreme caution! Most users should use HitToPhotonSignal
    /// instead. Only use the cell version if you really want to know ALL the photon signals
    /// in the cell for the entire time window, NOT time-associated with the hit.
    ///
    /// Sorted from highest signal to lowest (which is probably not much use,
    /// since energy from one particle can be split over several signals).
    const std::vector<sim::PhotonSignal> CellToPhotonSignal(unsigned int const& plane,
							    unsigned int const& cell) const;

    /// \brief Returns the PhotonSignals contributing the signal in the
    /// specified hit.
    ///
    /// Sorted from highest signal to lowest (which is probably not much use,
    /// since energy from one particle can be split over several signals).
    const std::vector<sim::PhotonSignal> HitToPhotonSignal(const art::Ptr<rb::CellHit>& hit) const;

    /// \brief Returns the FLSHits contributing the signal in the specified
    /// cell. WARNING: Use with extreme caution! Most users should use HitToFLSHit
    /// instead. Only use the cell version if you really want to know ALL the FLSHits
    /// in the cell for the entire time window, NOT time-associated with the hit.
    ///
    /// The vector is sorted to return in order of highest contributing Geant4
    /// track to lowest.
    const std::vector<sim::FLSHit>    CellToFLSHit(unsigned int const& plane,
						   unsigned int const& cell) const;

    /// \brief All the FLSHits that contributed to this hit, sorted from most
    /// to least light
    const std::vector<sim::FLSHit> HitToFLSHit(const art::Ptr<rb::CellHit>& hit) const;

    /// \brief All the FLSHits that were created by the track id trackID,
    /// sorted from most to least light
    std::vector<sim::FLSHit> ParticleToFLSHit(const int& trackID) const;
 
    /// \brief All the FLSHits that were created by the track id trackID,
    /// which also have a particular pdg. Most users should use the other
    /// version of this function. However, sometimes when our truth is 
    /// condensed, the same mother track ID is assigned to FLS hits made 
    /// by non-important daughters, who can have different PDG values. An
    /// example of this is a muon track, whose delta rays often have the track
    /// id of the mother muon. By using this version of the function, you can 
    /// ensure you are looking only at the FLS hits deposited by the muon.
    std::vector<sim::FLSHit> ParticleToFLSHit(const int trackID, const int pdg) const;

    /// \brief The total energy deposited in this cell by all FLSHits with this
    /// trackId
    const double EnergyFromTrackId(const art::Ptr<rb::CellHit>& hit, int trackId) const;

    /// \brief Returns the sim::Particle object contributing the most light to
    /// the specified rb::CellHit
    const sim::Particle*              HitToParticle(art::Ptr<rb::CellHit> const& hit,
                                                    bool quiet = false) const;

    /// Is this hit not associated with any particles?
    bool                              IsNoise(const art::Ptr<rb::CellHit>& hit) const;

    /// Of this collection of hits, is it mostly noise or physics hits? True indicates noise.
    bool                              IsHitsNoise(std::vector< art::Ptr<rb::CellHit> > const& hits) const;

    /// \brief Returns a pointer to the sim::Particle object corresponding to
    /// the given TrackID
    const sim::Particle*              TrackIDToParticle(int const& id) const;
    const sim::Particle*              TrackIDToMotherParticle(int const& id) const;

    const art::Ptr<simb::MCTruth>&    TrackIDToMCTruth(int const& id) const;
    const art::Ptr<simb::MCTruth>&    ParticleToMCTruth(const sim::Particle* p) const;
    std::vector<const sim::Particle*> MCTruthToParticles(art::Ptr<simb::MCTruth> const& mct) const;

    /// \brief Returns the XYZ position of the energy deposition for a given
    /// hit.
    ///
    /// Note the return value is in cell coordinates, not detector
    TVector3 HitToXYZ(art::Ptr<rb::CellHit> const& hit) const;

    /// \brief Returns vector of sim::Particle objects contributing to the
    /// given collection of hits.
    ///
    /// Orders the returned vector with the particle contributing the most
    /// light in the collection first.
    const std::vector<const sim::Particle*> HitsToParticle(art::PtrVector<rb::CellHit> const& hits) const;

    /// \brief Returns vector of TrackIDE structs contributing to the given
    /// collection of hits.
    ///
    /// Orders the returned vector with the trackID contributing the most
    /// energy in the collection first.
    const std::vector<TrackIDE>       HitsToTrackIDE(art::PtrVector<rb::CellHit> const& hits) const;

    /// \brief A version of the above that follows the best practices
    const std::vector<TrackIDE>       HitsToTrackIDE(std::vector< art::Ptr<rb::CellHit> > const& hits) const;

    /// Convenience function. \ref HitsToTrackIDE but for a single hit.
    const std::vector<TrackIDE>       HitToTrackIDE(art::Ptr<rb::CellHit> const& hit) const;


    /// \brief Returns the fraction of hits in a collection that come from the
    /// specified Geant4 track ids.
    ///
    double HitCollectionPurity(const std::set<int>&                trackIDs,
                               const std::vector<rb::WeightedHit>& whits,
			       std::map<int, double>*              purMap = 0,
                               std::map<int, int>*                 parents = 0,
                               bool                                energyPur = false) const;
                    
    /// \brief Returns the fraction of hits in a collection that come from the
    /// specified Geant4 track ids.
    ///
    /// The optional \a purMap parameter returns the purities for each of the
    /// track IDs individually.
    ///
    /// The optional \a parents parameter maps particle IDs to the IDs they
    ///
    /// The optional \a energyPur parameter allows the purity to be calculated based on FLS energy instead of number of hits

    double HitCollectionPurity(const std::set<int>&               trackIDs,
                               const art::PtrVector<rb::CellHit>& hits,
                               std::map<int, double>*             purMap = 0,
                               std::map<int, int>*                parents = 0,
                               bool                               energyPur = false) const;

    /// \brief Returns the fraction of all energy in an event from a specific set
    /// of Geant4 track IDs that are represented in a collection of weighted hits
    ///
    /// \a desiredHits and 
    /// \a totalHits are also optional parameters that are the number of hits from the given set of 
    /// trackIds in the given slice and the total number of hits from these trackId's in the detector respectively

    const double HitCollectionEfficiency(std::set<int>                       trackIDs,
					 std::vector<rb::WeightedHit> const& whits,
					 art::PtrVector<rb::CellHit> const&  allhits,
					 geo::View_t                 const&  view,
                                         std::map<int, double>*              effMap = 0,
					 bool                                energyEff     = false,
					 double*                             desiredEnergy = NULL,
					 double*                             totalEnergy   = NULL,
					 int*                                desiredHits   = NULL,
					 int*                                totalHits     = NULL) const;




    /// \brief Returns the fraction of all hits in an event from a specific set
    /// of Geant4 track IDs that are represented in a collection of hits
    ///
    /// The optional \a effMap parameter returns the efficiencies for each of
    /// the track IDs individually.
    ///
    /// The optional \a energyPur parameter allows the efficiency to be calculated based on FLS energy instead of number of hits
    /// \a desiredHits and \a totalHits are also optional parameters that are the number of hits from the given set of trackIds 
    /// in the given slice and the total number of hits from these trackId's in the detector respectively
    const double HitCollectionEfficiency( std::set<int>                      trackIDs,
                                          art::PtrVector<rb::CellHit> const& hits,
                                          art::PtrVector<rb::CellHit> const& allhits,
                                          geo::View_t                 const& view,
                                          std::map<int, double>*             effMap = 0,
                                          bool                               energyEff     = false,
                                          double*                            desiredEnergy = NULL,
                                          double*                            totalEnergy   = NULL,
					  int*                               desiredHits   = NULL,
					  int*                               totalHits     = NULL) const;

    /// \brief Tool for NumuEAna that allows one to see if primary muon (or any track ID you feed the function)
    /// contributes to enough hits in the slice to pass the hit cut. It is currently set at 4 hits in each view.
    bool PassMuonCuts(int trackID, art::PtrVector<rb::CellHit> const& hits) const;

    /// \brief Tool for function SliceToNeutrinoInteractions, tells how to sort by putting structures with highest efficiency first.
    static bool sort_eff(const NeutrinoEffPur& a, const NeutrinoEffPur& b);

    /// \brief Tool for function SliceToNeutrinoInteractions, tells how to sort by putting structures with highest purity first.
    static bool sort_pur(const NeutrinoEffPur& a, const NeutrinoEffPur& b);

    /// \brief Given a collection of hits (often a slice), returns vector of structures of neutrino interactions,
    /// efficiency, and purity of that neutrino interaction relative to the slice. Efficiency is defined as FLS energy
    /// from neutrino interaction in slice / total FLS energy from neutrino interaction in event.
    /// This vector is sorted from the highest efficiency interaction to lowest.
    std::vector <NeutrinoEffPur> SliceToNeutrinoInteractions(art::PtrVector<rb::CellHit> const& sliceHits,
                                                             art::PtrVector<rb::CellHit> const& allHits,
                                                             bool                               sortPur = false) const;

    std::vector <NeutrinoEffPur> SliceToNeutrinoInteractions(art::Ptr<rb::Cluster>       const& sliceCluster,
                                                             art::PtrVector<rb::CellHit> const& allHits,
                                                             bool                               sortPur = false) const;

    /// \brief Given a collection of hits (often a slice), returns vector of structures of MCTruth,
    /// efficiency, and purity of that neutrino interaction relative to the slice. Efficiency is defined as FLS energy
    /// from neutrino interaction in slice / total FLS energy from neutrino interaction in event.
    /// This vector is sorted from the highest efficiency interaction to lowest.  This function returns all MCTruth,
    /// including those without neutrino interactions, i.e. cosmics.
    std::vector <NeutrinoEffPur> SliceToMCTruth(art::PtrVector<rb::CellHit>              const& sliceHits,
                                                             art::PtrVector<rb::CellHit> const& allHits,
                                                             bool                               sortPur = false) const;
    
    std::vector <NeutrinoEffPur> SliceToMCTruth(art::Ptr<rb::Cluster>                    const& sliceCluster,
                                                             art::PtrVector<rb::CellHit> const& allHits,
                                                             bool                               sortPur = false) const;
    /// \brief Given ALL the slices in an event, including the noise slice, returns a vector of vector of structures of MCTruth,
    /// efficiency, and purity of that MCTruth interaction relative to the slice. 
    /// The returned vector is the same size as the total number of slices, including the noise slice, which is treated the same
    /// as any other slice. It is sorted in the same order as the slice list.
    /// For each entry, there is a vector of NeutrinoEffPurs. This vector is the same size as the MCTruth list. It is sorted in
    /// the same order as the MCTruth list.
    /// Every slice/MCTruth pair shows up, even if the contribution is zero.
    /// Efficiency is defined as FLS energy from a MCTruth in a slice / total FLS energy from a MCTruth in event.
    /// This function is agnostic to MCTruth and treates them all the same, including those without neutrino interactions, i.e. cosmics.
    std::vector< std::vector<cheat::NeutrinoEffPur> >SlicesToMCTruthsTable(art::PtrVector<rb::Cluster> const& sliceList) const;

    /// \brief Given a collection of reconstructed tracks, this method finds the best match for each reco track to the
    /// input track id particles. 
    /// 
    /// For each reconstructed track, the true particle that best matches the track by
    /// purity from the input trkIDs is determined. Each input track ID is only allowed to match at most one input
    /// track and any track IDs specified in the input allowedDaughters are not counted against a tracks purity.
    /// The tracks are matched to particles in the order that they appear in the input vector. Each entry in the
    /// returned vector of ParticleEffPur corresponds to the track in the same position as the input track vector and
    /// holds the matched track id, efficiency, purity, and pdg of the best matched particle for the given track. If
    /// no track id is found to match the track, then the purity, efficiency, and pdg will be zero. You need to check
    /// that a match exists before trying to use the matched track id in other functions
    std::vector<ParticleEffPur>  TracksToParticles(std::vector<art::Ptr<rb::Track > >  const& tracks,
						   std::set<int>                       const& trkIDs, 
						   std::set<int>                       const& allowedDaughters, 
						   art::PtrVector<rb::CellHit>         const& allHits) const;
    

    /// \brief Function primarily for CAFMaker - returns a vector of all MCTruth interactions along with their truth vector position
    std::vector <NeutrinoWithIndex> allNeutrinoInteractions() const;

    std::vector <NeutrinoWithIndex> allMCTruth() const;

    /// \brief Get a reference to the ParticleNavigator
    const sim::ParticleNavigator&    ParticleNavigator() const { return fParticleNav; }

    /// \brief Get all G4 track ids present in the event.
    ///
    /// These do not include ids from EM shower secondaries, tertiaries, etc as
    /// they are not saved as particles.
    const std::set<int>               GetTrackIDList() const { return fTrackIDs; }

    /// \brief:
    //  Return a vector of clusters where the hits are separated by MCTruth.  A hit that
    //  had contributions from more than one MCTruth is clustered with the MCTruth that
    //  contributed the most energy.  Noise hits are put into the last cluster.
    //
    //  NOTE: It is not always the case that all MCTruths contribute hits to the event record,
    //  so some of the clusters returned might be empty!
    std::vector<rb::Cluster> ClusterByTruth(std::vector< art::Ptr< rb::CellHit > > const &hits);

    /// \brief:
    //  Given a list of hits (usually all hits) this returns a cluster of only the PRIMARY hits
    //  associated with a specific MCTruth.
    //
    // NOTE: A hit will only be added to this cluster if its primary energy contribution came from this MCTruth.
    rb::Cluster MCTruthToCluster(art::Ptr<simb::MCTruth>                const& truth,
				 std::vector< art::Ptr< rb::CellHit > > const& hits);



  private:
    /// \brief Internal helper, decides if a PhotonSignal and CellHit are close
    /// enough to match.
    bool CloseInTime(const rb::CellHit& chit, const sim::PhotonSignal& phot) const;

    /// \brief Internal helper, adds the contributions from each particle in
    /// this cell to the map
    void AccumulateHitContributions(art::Ptr<rb::CellHit> const& hit,
				    std::map<int, int>         & idToNPhot) const;

    bool fHaveTruthInfo; ///< Set by \ref Rebuild

    novadaq::cnv::DetId                    fDetID;                  ///< detector id

    std::unordered_map< geo::OfflineChan, std::list<sim::FLSHit> > fCellToFLSHit;    ///< map of FLSHits in each cell in event
    std::unordered_map< int,              std::list<sim::FLSHit> > fTrackIDToFLSHit; ///< map of G4 track ids to FLSHits
    std::unordered_map< geo::OfflineChan, std::vector<sim::PhotonSignal> > fCellToPhotonSignal; ///< map of PhotonSignals in each cell in event
    sim::ParticleNavigator                 fParticleNav;            ///< Particle navigator to map track ID to Particle
    std::set<int>                          fTrackIDs;               ///< G4 track ids for all particles in the events
    std::vector< art::Ptr<simb::MCTruth> > fMCTruthList;            ///< all the MCTruths for this event
    std::map<int, int>                     fTrackIDToMCTruthIndex;  ///< map of track id values to MCTruthList entry
    std::string                            fGeantModuleLabel;       ///< label for geant4 module
    std::string                            fPhotonModuleLabel;      ///< label for PhotonTransport module
    bool                                   fDoHitsCheck;            ///< label for DoHitsCheck


    /// \brief Minimum fraction of total energy a track ID must contribute to a
    /// cell for the particle to be counted.
    double                                 fMinContribFrac;
    int                                    fMinPhysicsHits;
    double                                 fMinPhysicsFrac;
  };
} // namespace

DECLARE_ART_SERVICE(cheat::BackTracker, LEGACY)
#endif // CHEAT_BACKTRACKER_H
