///////////////////////////////////////////////////////////////////////
// $Id: BackTracker_service.cc,v 1.7 2012-12-06 15:03:39 lein Exp $
//
//
// \file: BackTracker_service.cc
//
// brebel@fnal.gov
//
////////////////////////////////////////////////////////////////////////

#include "MCCheater/BackTracker.h"

#include <map>

// Framework includes
#include "art/Framework/Core/FindOneP.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "messagefacility/MessageLogger/MessageLogger.h"

#include "Simulation/FLSHitList.h"
#include "Geometry/Geometry.h"

// need a map of CellUniqueId to vector of FLSHits to make this work
// really should just have a CellHit/RawDigit to provide the Cell and Plane
// and then look through the event to gett the proper FLSHits
// have to know where the FLSHits are stored in the event if want to pass art::Event around

namespace cheat{

  //----------------------------------------------------------------------
  static bool sortFLSHit(sim::FLSHit const& a, sim::FLSHit const& b)
  {
    return a.GetEdep() > b.GetEdep();
  }

  //----------------------------------------------------------------------
  bool CompareByEnergy(const TrackIDE& a, const TrackIDE& b)
  {
    return a.energy < b.energy;
  }

  //----------------------------------------------------------------------
  BackTracker::BackTracker(const fhicl::ParameterSet& pset,
			   art::ActivityRegistry& reg)
    : fHaveTruthInfo(false)
  {
    reconfigure(pset);

    reg.sPreProcessEvent.watch(this, &BackTracker::Rebuild);
  }

  //----------------------------------------------------------------------
  BackTracker::~BackTracker()
  {
  }

  //----------------------------------------------------------------------
  void BackTracker::reconfigure(const fhicl::ParameterSet& pset)
  {
    fGeantModuleLabel  = pset.get<std::string>("GeantModuleLabel",  "geantgen");
    fPhotonModuleLabel = pset.get<std::string>("PhotonModuleLabel", "photrans");
    fDoHitsCheck       = pset.get<bool       >("DoHitsCheck", true);
    fMinContribFrac    = pset.get<double     >("MinContribFrac",    0.1       );
    fMinPhysicsHits    = pset.get<int        >("MinPhysicsHits",    4         );
    fMinPhysicsFrac    = pset.get<double     >("MinPhysicsFrac",    0.5       );
  }

  //----------------------------------------------------------------------
  void BackTracker::Rebuild(const art::Event& evt)
  {
    // Clear out anything remaining from previous calls to Rebuild
    fCellToFLSHit.clear();
    fTrackIDToFLSHit.clear();
    fCellToPhotonSignal.clear();
    fTrackIDs.clear();
    fParticleNav.clear();
    fMCTruthList.clear();

    fHaveTruthInfo = false;               // Haven't got anything yet
    fDetID = novadaq::cnv::kUNKNOWN_DET;  // Haven't found detector type yet

    // do nothing if this is data
    if(evt.isRealData()) return;

    // get detector for CloseInTime function
    art::ServiceHandle<geo::Geometry>  geom;
    fDetID = geom->DetId();
    
    // fill the map of OfflineChan to std::vector<FLSHit>
    // first get the FLSHitList from the event
    art::Handle< std::vector<sim::FLSHitList> > flslistHandle;
    evt.getByLabel(fGeantModuleLabel, flslistHandle);
    
    // Probably this is a MC generation job and this is the preEvent callback
    // which is too early. That's OK, we'll get called at the right time by
    // ReadoutSim in a moment. Or we're mock data or something.
    if(flslistHandle.failedToGet()) return;

    fHaveTruthInfo = true; // We got FLSHits, the rest should go well too

    // loop over the FLSHits and fill the std::vector for each 
    // channel
    for( size_t l = 0; l < flslistHandle->size(); ++l){
      const std::vector<sim::FLSHit> &hits = flslistHandle->at(l).fHits;
      for( size_t f = 0; f < hits.size(); ++f){
        const sim::FLSHit& hit = hits[f];
        const geo::OfflineChan chan(hit.GetPlaneID(), hit.GetCellID());
        fCellToFLSHit[chan].push_back(hit);
	fTrackIDToFLSHit[hit.GetTrackID()].push_back(hit);
      }// end loop over flshits
    }// end loop over flshit lists
    
    // loop over the map and sort the vectors of flshits
    for(auto itr = fCellToFLSHit.begin(); itr != fCellToFLSHit.end(); itr++){
      itr->second.sort(sortFLSHit);
    }

    // loop over the map and sort the vectors of flshits
    for(auto pfitr = fTrackIDToFLSHit.begin(); pfitr != fTrackIDToFLSHit.end(); pfitr++){
      pfitr->second.sort(sortFLSHit);
    }

    // Similarly, PhotonSignals
    art::Handle<std::vector<sim::PhotonSignal> > phots;
    evt.getByLabel(fPhotonModuleLabel, phots);

    if(!phots.failedToGet()){
      for(size_t i = 0; i < phots->size(); ++i){
        const sim::PhotonSignal& phot = (*phots)[i];
        const geo::OfflineChan chan(phot.Plane(), phot.Cell());
        fCellToPhotonSignal[chan].push_back(phot);
      } // end for i

      for(auto it = fCellToPhotonSignal.begin(); it != fCellToPhotonSignal.end(); ++it){
        // Sort in reverse order, ie from most photons to least
        std::sort(it->second.rbegin(), it->second.rend(), sim::CompareByNPhoton);
      }
    }
    else{
      mf::LogWarning("BackTracker") << "unable to find PhotonTransport "
                                    << "created in " << fPhotonModuleLabel << " "
                                    << "any attempt to get the PhotonTransport objects from "
                                    << "the backtracker will fail";
    }

    art::Handle< std::vector<sim::Particle> > pHandle;
    evt.getByLabel(fGeantModuleLabel, pHandle);
    if(pHandle.failedToGet()){
      mf::LogWarning("BackTracker") << "unable to find sim::Particles "
                                    << "created in " << fGeantModuleLabel << " "
                                    << "any attempt to get the Particles from "
                                    << "the backtracker will fail";
    }
    else{
      fMCTruthList.reserve(pHandle->size());
      art::FindOneP<simb::MCTruth> fo(pHandle, evt, fGeantModuleLabel);
      for(size_t p = 0; p < pHandle->size(); ++p){

        sim::Particle *part = new sim::Particle(pHandle->at(p));
        fParticleNav.Add(part);

        // get the simb::MCTruth associated to this vector<Particle>
        art::Ptr<simb::MCTruth> mct = fo.at(p);
        try{
          if(fMCTruthList.size() < 1) fMCTruthList.push_back(mct);
          else{
            // check that we are not adding a simb::MCTruth twice to the
            // collection we know that all the particles for a given
            // simb::MCTruth are put into the collection of particles at the
            // same time, so we can just check that the current art::Ptr has a
            // different id than the last one put
            if(mct != fMCTruthList.back()) fMCTruthList.push_back(mct);
          }
          // fill the track id to mctruth index map
          fTrackIDToMCTruthIndex[pHandle->at(p).TrackId()] = fMCTruthList.size() - 1;
        }
        catch(cet::exception &ex){
          mf::LogWarning("BackTracker") << "unable to find MCTruth from Particle list "
                                        << "created in " << fGeantModuleLabel << " "
                                        << "any attempt to get the MCTruth objects from "
                                        << "the backtracker will fail\n"
                                        << "message from caught exception:\n" << ex;
        }
      } // end loop over particles to get MCTruthList
    }


    // fill the set of track ids from the ParticleNavigator
    fTrackIDs.clear();
    for(auto& it: fParticleNav) fTrackIDs.insert(it.first);

    LOG_DEBUG("BackTracker") << fParticleNav;
  
  }

  //----------------------------------------------------------------------
  const std::vector<sim::PhotonSignal> BackTracker::CellToPhotonSignal(unsigned int const& plane,
								       unsigned int const& cell) const
  {
    auto it = fCellToPhotonSignal.find(geo::OfflineChan(plane, cell));

    if(it == fCellToPhotonSignal.end()){
      LOG_DEBUG("BackTracker") << "no collection of sim::PhotonSignal for "
			       << " plane = " << plane << " cell = " << cell
			       << " probably this was a noise hit, returning an"
			       << " empty vector";
      const std::vector<sim::PhotonSignal> empty;
      return empty;
    }
    return it->second;
  }

  //----------------------------------------------------------------------
  const std::vector<sim::PhotonSignal> BackTracker::HitToPhotonSignal(const art::Ptr<rb::CellHit>& hit) const
  {
    auto it = fCellToPhotonSignal.find(geo::OfflineChan(hit->Plane(), hit->Cell()));
    
    // Nothing to accumulate from this cell
    if(it == fCellToPhotonSignal.end()){
      LOG_DEBUG("BackTracker") << "no collection of sim::PhotonSignal for "
			       << " plane = " << hit->Plane() << " cell = " << hit->Cell()
			       << " probably this was a noise hit, returning an"
			       << " empty vector";
      const std::vector<sim::PhotonSignal> empty;
      return empty;
    }
    
    // Candidate photons, in that they're in the right cell.
    const std::vector<sim::PhotonSignal>& phots = it->second;

    std::vector<sim::PhotonSignal> hitPhots;
    
    // For every photon that's sufficiently in time with the hit,
    // accumulate its energy into the relevant trackID.
    for(unsigned int i = 0; i < phots.size(); ++i){
      const sim::PhotonSignal& phot = phots[i];
      if(CloseInTime(*hit.get(), phot)){
	hitPhots.push_back(phot);
      }
    }
    return hitPhots;
  }
  
  //----------------------------------------------------------------------
  const std::vector<sim::FLSHit> BackTracker::CellToFLSHit(unsigned int const& plane,
							   unsigned int const& cell) const
  {
    auto it = fCellToFLSHit.find(geo::OfflineChan(plane, cell));

    if(it == fCellToFLSHit.end()){
      LOG_DEBUG("BackTracker") << "no collection of sim::FLSHit for "
			       << " plane = " << plane << " cell = " << cell
			       << " probably this was a noise hit, returning an"
			       << " empty vector";
      const std::vector<sim::FLSHit> empty;
      return empty;
    }
    return std::vector<sim::FLSHit>(it->second.begin(), it->second.end());
  }

  //----------------------------------------------------------------------
  const std::vector<sim::FLSHit> BackTracker::HitToFLSHit(const art::Ptr<rb::CellHit>& hit) const
  {
    std::map<int, int> idToNPhot;
    AccumulateHitContributions(hit, idToNPhot);

    // Next, sort by number of photons. This turns out to be a bit fiddly.

    std::vector<std::pair<int, int> > nphotToId;
    nphotToId.reserve(idToNPhot.size());
    for(std::map<int, int>::iterator it = idToNPhot.begin(); it != idToNPhot.end(); ++it){
      nphotToId.push_back(std::make_pair(it->second, it->first));
    }

    if(nphotToId.empty()){
      LOG_DEBUG("BackTracker") << "No matching photons, this is probably a noise hit. Returning empty vector";
      return std::vector<sim::FLSHit>();
    }

    // Sort in reverse order (ie decreasing number of photons).
    // Tuples compare by first element first, so this works
    std::sort(nphotToId.rbegin(), nphotToId.rend());

    auto it = fCellToFLSHit.find(geo::OfflineChan(hit->Plane(), hit->Cell()));
    if ( it == fCellToFLSHit.end() ) {
      throw cet::exception("BackTracker")
        << "PhotonSignal with no matching FLSHit. That shouldn't be possible\n"
        << __FILE__ << ":" << __LINE__ << "\n";
    }

    std::vector<sim::FLSHit> ret;
    const std::list<sim::FLSHit>& flshits = it->second;

    ret.reserve(nphotToId.size()*flshits.size()); // Overestimate

    for(unsigned int idIdx = 0; idIdx < nphotToId.size(); ++idIdx){
      const int trackid = nphotToId[idIdx].second;
      for(const sim::FLSHit& fls: flshits){
	if(fls.GetTrackID() == trackid){
	  ret.push_back(fls);
	}
      } // end for flsIdx
    } // end for idIdx

    return ret;
  }

  //----------------------------------------------------------------------
  const double BackTracker::EnergyFromTrackId(const art::Ptr<rb::CellHit>& hit, int trackId) const
  {
    auto it = fCellToFLSHit.find(geo::OfflineChan(hit->Plane(), hit->Cell()));
    if(it == fCellToFLSHit.end()){
      mf::LogWarning("BackTracker") << "No FLSHits in this cell";
      return 0;
    }

    const std::list<sim::FLSHit>& flshits = it->second;

    double energy = 0;

    for(const sim::FLSHit& fls: flshits){
      if(fls.GetTrackID() == trackId){
	energy += fls.GetEdep();
      }
    } // end for flsIdx

    return energy;
  }

  //----------------------------------------------------------------------
  const sim::Particle* BackTracker::HitToParticle(const art::Ptr<rb::CellHit>& hit,
						  bool quiet) const
  {
    std::map<int, int> idToNPhot;
    AccumulateHitContributions(hit, idToNPhot);

    // Go through the map and find the particle ID that contributed the most light
    int bestId = -1;
    int mostPhot = -1;
    for(std::map<int, int>::iterator it = idToNPhot.begin(); it != idToNPhot.end(); ++it){
      if(it->second > mostPhot){
	mostPhot = it->second;
	bestId = it->first;
      }
    }

    if(bestId == -1){
      if(!quiet){
        mf::LogWarning("BackTracker") << "No photons in this cell are in time with the hit. "
				      << "Returning null pointer";
      }
      return 0;
    }

    return TrackIDToParticle(bestId);
  }

  //----------------------------------------------------------------------
  bool BackTracker::IsNoise(const art::Ptr<rb::CellHit>& hit) const
  {
    return (HitToParticle(hit, true) == 0);
  }

  //----------------------------------------------------------------------
  bool BackTracker::IsHitsNoise(std::vector< art::Ptr<rb::CellHit> > const& hits) const
  {
    int totalNoise = 0;
    int total = hits.size();

    for(size_t h = 0; h < hits.size(); ++h){
      art::Ptr<rb::CellHit> hit = hits[h];
      if (IsNoise(hit)) totalNoise += 1;
    }

    int totalPhysics = total - totalNoise;
    double physicsFrac = (double)totalPhysics / (double)total;

    // If at least fMinPhysicsHits physics hits in the collection, not a noise group. 
    // If at least fMinPhysicsFrac of the hits are physics, not a noise group. 
    // Otherwise, call it a noise group.

    return((totalPhysics < fMinPhysicsHits) && (physicsFrac < fMinPhysicsFrac));
  }


  //----------------------------------------------------------------------
  const sim::Particle* BackTracker::TrackIDToParticle(int const& id) const
  {
    sim::ParticleNavigator::const_iterator part_it = fParticleNav.find(abs(id));

    if(part_it == fParticleNav.end()){
      mf::LogWarning("BackTracker") << "can't find particle with track id "
				    << id << " in sim::ParticleNavigator"
				    << " returning null pointer";
      return 0;
    }

    return part_it->second;
  }

  //----------------------------------------------------------------------
  const sim::Particle* BackTracker::TrackIDToMotherParticle(int const& id) const
  {
    // get the mother id from the particle navigator
    // the EveId was adopted in the Rebuild method
 
    return this->TrackIDToParticle(fParticleNav.EveId(abs(id)));
  }

  //----------------------------------------------------------------------
  const art::Ptr<simb::MCTruth>& BackTracker::TrackIDToMCTruth(int const& id) const
  {
    // find the entry in the MCTruth collection for this track id
    size_t mct = fTrackIDToMCTruthIndex.find(abs(id))->second;
    
    if (mct > fMCTruthList.size() ) {
      throw cet::exception("BackTracker")
        << "attempting to find MCTruth index for "
        << "out of range value: " << mct
        << "/" << fMCTruthList.size() << "\n"
        << __FILE__ << ":" << __LINE__ << "\n";
    }
    return fMCTruthList[mct];
  }

  //----------------------------------------------------------------------
  const art::Ptr<simb::MCTruth>& BackTracker::ParticleToMCTruth(const sim::Particle* p) const
  {
    return this->TrackIDToMCTruth(p->TrackId());
  }

  //----------------------------------------------------------------------
  std::vector<const sim::Particle*> BackTracker::MCTruthToParticles(art::Ptr<simb::MCTruth> const& mct) const
  {
    std::vector<const sim::Particle*> ret;

    for(std::set<int>::const_iterator itr = fTrackIDs.begin(); itr != fTrackIDs.end(); itr++){
      if( this->TrackIDToMCTruth(*itr) == mct ) ret.push_back(fParticleNav.find(*itr)->second);
    }

    return ret;
  }

  //----------------------------------------------------------------------
  // this method orders the vector of particles from contributing most
  // to the collection to least
  const std::vector<const sim::Particle*> BackTracker::HitsToParticle(const art::PtrVector<rb::CellHit>& hits) const
  {
    std::map<int, int> idToNPhot;
    for(size_t h = 0; h < hits.size(); ++h){
      AccumulateHitContributions(hits[h], idToNPhot);
    }

    // Now "all" we need to do is sort by number of photons, and discard any IDs
    // that we can't find particles for. This turns out to be a bit fiddly.

    std::vector<std::pair<int, int> > nphotToId;
    for(std::map<int, int>::iterator it = idToNPhot.begin(); it != idToNPhot.end(); ++it){
      nphotToId.push_back(std::make_pair(it->second, it->first));
    }

    // Sort in reverse order (ie increasing number of photons).
    // Tuples compare by first element first, so this works
    std::sort(nphotToId.rbegin(), nphotToId.rend());

    std::vector<const sim::Particle*> ret;
    for(unsigned int n = 0; n < nphotToId.size(); ++n){
      const sim::Particle* part = TrackIDToParticle(nphotToId[n].second);
      if(part) ret.push_back(part);
    }

    return ret;
  }

  //----------------------------------------------------------------------
  // this method orders the vector of particles from contributing most
  // to the collection to least
  const std::vector<TrackIDE> BackTracker::HitsToTrackIDE(art::PtrVector<rb::CellHit> const& hits) const
  {
    std::vector< art::Ptr<rb::CellHit> > hitvec(hits.size());
    for(size_t h = 0; h < hits.size(); ++h) hitvec[h] = hits[h];
    return this->HitsToTrackIDE(hitvec);
  }
  
  const std::vector<TrackIDE> BackTracker::HitsToTrackIDE(std::vector< art::Ptr<rb::CellHit> > const& hits) const
  {
    std::map<int, double> sigtracks;

    double totalE = 0;
    for(size_t h = 0; h < hits.size(); ++h){
      const std::vector<sim::FLSHit> fls = HitToFLSHit(hits[h]);
      for(size_t f = 0; f < fls.size(); ++f){
	sigtracks[fls[f].GetTrackID()] += fls[f].GetEdep();
	totalE += fls[f].GetEdep();
      }
    } // end for h

    // Make trackIDE objects from our collected energies
    std::vector<TrackIDE> trackIDE;
    for(std::map<int, double>::iterator it = sigtracks.begin(); it != sigtracks.end(); ++it){
      TrackIDE ide;
      ide.trackID = it->first;
      ide.energy = it->second;
      ide.energyFrac = it->second/totalE;
      trackIDE.push_back(ide);
    }

    // Sort TrackIDE objects from most energy contributed to least
    std::sort(trackIDE.rbegin(), trackIDE.rend(), CompareByEnergy);

    return trackIDE;
  }

  //----------------------------------------------------------------------
  const std::vector<TrackIDE> BackTracker::HitToTrackIDE(art::Ptr<rb::CellHit> const& hit) const
  {
    std::vector< art::Ptr<rb::CellHit> > hits(1);
    hits[0] = hit;
    return HitsToTrackIDE(hits);
  }

  //----------------------------------------------------------------------
  TVector3 BackTracker::HitToXYZ(art::Ptr<rb::CellHit> const& hit) const
  {
    double x = 0.;
    double y = 0.;
    double z = 0.;
    double w = 0.;

    const std::vector<sim::FLSHit> fhits = this->HitToFLSHit(hit);
    
    // loop over the FLSHits for this cell and get the weighted central 
    // position of each track id in it.
    for(const sim::FLSHit& fhit: fhits){
      double weight = fhit.GetEdep();
      
      w += weight;
      x += weight * 0.5*(fhit.GetExitX() + fhit.GetEntryX());
      y += weight * 0.5*(fhit.GetExitY() + fhit.GetEntryY());
      z += weight * 0.5*(fhit.GetExitZ() + fhit.GetEntryZ());

    }// end loop over sim::FLSHits
	
    // if the sum of the weights is still 0, then return
    // the obviously stupid default values
    if(w < 1.e-5) return TVector3(-999, -999, -999);

    return TVector3(x/w, y/w, z/w);
  }

   //----------------------------------------------------------------------
  double BackTracker::
  HitCollectionPurity(const std::set<int>&                trackIDs,
                      const std::vector<rb::WeightedHit>& whits,
		      std::map<int, double>*              purMap,
                      std::map<int, int>*                 parents,
                      bool                                energyPur) const

  {
    // get the list of EveIDs that correspond to the hits in this collection
    // if the EveID shows up in the input list of trackIDs, then it counts
    double total   = whits.size();
    double desired = 0;

    double desiredE = 0;
    double totalE   = 0;

    if(purMap) purMap->clear();

    for(size_t h = 0; h < whits.size(); ++h){

      double totalEInCell   = 0;
      double desiredEInCell = 0;

      art::Ptr<rb::CellHit> hit = whits[h].hit;
      double weight = whits[h].weight;

      assert( weight <= 1);

      if(IsNoise(hit)) continue;

      std::vector<TrackIDE> tracks = HitToTrackIDE(hit);

      // The set of trackIDs (or parents) that contributed to this hit                                               
      std::set<int> tally;

      for(size_t t = 0; t < tracks.size(); ++t){
        const int trackID = tracks[t].trackID;
        if(trackIDs.find(trackID) != trackIDs.end()){
          tally.insert(trackID);
        }
        if(parents){
	  std::map<int, int>::iterator parentsIt = parents->find(trackID);
          if(parentsIt != parents->end()){
            tally.insert(parentsIt->second);
          }
        }
      } // end for t


      if (energyPur){
        for(size_t t = 0; t < tracks.size(); ++t){
          totalEInCell += tracks[t].energy * weight;
          if(trackIDs.find(tracks[t].trackID) != trackIDs.end()){
            desiredE += tracks[t].energy;
	    desiredEInCell += tracks[t].energy;
          }
        }
      }//End of loop for counting energy when doing purity based on FLS hits

      // Purity in a cell can sometimes be greater than one if the weight of
      // a hit is less than the fraction of energy from desired track IDs
      // in that cell.
      if( desiredEInCell >= totalEInCell )
        totalE += desiredEInCell;
      else
	totalE += totalEInCell;

      if(purMap){
        for(std::set<int>::iterator it = tally.begin(); it != tally.end(); ++it){
          ++(*purMap)[*it];
        }
      }

      if(!tally.empty()) ++desired;


    }// end loop over whits

    double purity = 0.;

    if((total  > 0) && !energyPur) purity = desired/total;
    if((totalE > 0) &&  energyPur) purity = desiredE/totalE;
    
    assert(purity >= 0);
    assert(purity <= 1);

    if(total > 0 && purMap){
      for(std::map<int, double>::iterator it = purMap->begin(); it != purMap->end(); ++it){
        it->second /= total;
        // This is a purity, it should be between 0-100%                                                               
        assert(it->second >= 0);
        assert(it->second <= 1);
      }
    }

    return purity;

  }
  

  //----------------------------------------------------------------------
  

  double BackTracker::
  HitCollectionPurity(const std::set<int>&               trackIDs,
                      const art::PtrVector<rb::CellHit>& hits,
                      std::map<int, double>*             purMap,
                      std::map<int, int>*                parents,
                      bool                               energyPur) const
  {
    // convert all cell hits to weighted hits with weight 1.
    std::vector<rb::WeightedHit> wHits;
    int nHits = hits.size();
    for(int i = 0; i < nHits; ++i){
      rb::WeightedHit newHit( hits[i], 1);
      wHits.push_back( newHit );
    }

    return HitCollectionPurity( trackIDs, wHits, purMap, parents, energyPur);

  }

  //----------------------------------------------------------------------
  const double BackTracker::HitCollectionEfficiency(std::set<int>                      trackIDs, 
						    art::PtrVector<rb::CellHit> const& hits,
						    art::PtrVector<rb::CellHit> const& allhits,
						    geo::View_t                 const& view,
						    std::map<int, double>*             effMap,
                                                    bool                               energyEff,
                                                    double*                            desiredEnergy,
                                                    double*                            totalEnergy,
						    int*                               desiredHits,
						    int*                               totHits) const
  {

    // convert cell hits to weighted hits with weight 1
    std::vector<rb::WeightedHit> wHits;
    int nHits = hits.size();
    for(int i = 0; i < nHits; ++i){
      rb::WeightedHit newHit( hits[i], 1);
      wHits.push_back( newHit );
    }

    return HitCollectionEfficiency( trackIDs, wHits,  allhits, view, 
				    effMap, energyEff, desiredEnergy, 
				    totalEnergy, desiredHits, totHits);
    
    
  }
  
  //----------------------------------------------------------------------
  const double BackTracker::HitCollectionEfficiency(std::set<int>                       trackIDs, 
						    std::vector<rb::WeightedHit> const& whits,
						    art::PtrVector<rb::CellHit> const&  allhits,
						    geo::View_t                 const&  view,
						    std::map<int, double>*              effMap,
                                                    bool                                energyEff,
                                                    double*                             desiredEnergy,
                                                    double*                             totalEnergy,
						    int*                                desiredHits,
                                                    int*                                totHits) const

  {

    // allhits should be a superset of hits
    assert(allhits.size() >= whits.size());

    // to check it's a superset more thoroughly, optional 
    if(fDoHitsCheck){
      for(unsigned int n = 0; n < whits.size(); ++n){
        assert(std::find(allhits.begin(), allhits.end(), whits[n].hit) != allhits.end());
      }
    }
    std::vector<art::Ptr<rb::CellHit> > hits_copy;
    for(int i = 0; i <  (int)whits.size(); i++)
      hits_copy.push_back( whits[i].hit);

    // Check for duplicates
    std::sort(hits_copy.begin(), hits_copy.end());
    for(unsigned int n = 0; n+1 < hits_copy.size(); ++n)
      assert(hits_copy[n] < hits_copy[n+1]);
    
    // get the list of EveIDs that correspond to the hits in this collection
    // and the energy associated with the desired trackID
    
    double desired = 0.;
    double total   = 0.;
    
    double desiredE = 0.;
    double totalE   = 0.;

    if(effMap) effMap->clear();
    std::map<int, double> totMap;

    for(size_t h = 0; h < whits.size(); ++h){

      double totalEInCell   = 0.;
      double desiredEInCell = 0.;

      art::Ptr<rb::CellHit> hit = whits[h].hit;
      double weight = whits[h].weight;
      assert( weight <= 1);
      
      if(IsNoise(hit)) continue;
      
      bool hitCounted = false;
      
      std::vector<TrackIDE> tracks = HitToTrackIDE(hit);
      
      assert(hit->View() == view || view == geo::kXorY);
      
      // don't double count if this hit has more than one of the
      // desired track IDs associated with it
      for(size_t t = 0; t < tracks.size(); ++t){
	totalEInCell += tracks[t].energy;
        if(trackIDs.find(tracks[t].trackID) != trackIDs.end() ){
	  desiredEInCell += tracks[t].energy;
	  if(tracks[t].energyFrac >= fMinContribFrac){
	    if(!hitCounted){
	      desired += 1.;   
	      hitCounted = true;
	    }
	    if(effMap) ++(*effMap)[tracks[t].trackID];
	  }
	}
      }// end loop over track IDEs

      if( (weight*totalEInCell) >= desiredEInCell)
	desiredE += desiredEInCell;
      else
	desiredE += (weight*totalEInCell);

    }// end loop over hits

    // now figure out how many hits in the whole collection are associated with this id
    for(size_t h = 0; h < allhits.size(); ++h){
      art::Ptr<rb::CellHit> hit = allhits[h];
      if(IsNoise(hit)) continue;

      std::vector<TrackIDE> tracks = HitToTrackIDE(hit);

      bool hitCounted = false;

      // check that we are looking at the appropriate view here
      // in the case of 3D objects we take all hits
      if(hit->View() != view && view != geo::kXorY) continue;
      
      for(size_t t = 0; t < tracks.size(); ++t){
	// don't worry about hits where the energy fraction for the chosen     
	// trackID is < fMinContribFrac                                     
	// also don't double count if this hit has more than one of the   
	// desired track IDs associated with it                           
	if(trackIDs.find(tracks[t].trackID) != trackIDs.end()){
	  totalE += tracks[t].energy;
	  if(tracks[t].energyFrac >= fMinContribFrac){
	    if(!hitCounted){
	      total += 1.;
	      hitCounted = true;
	    }
	    if(effMap) ++totMap[tracks[t].trackID];
	  }// end if energy fraction is above threshold
	}// end if the current track ID is in the list of ones to check
      }// end loop over tracks
    }// end loop over all hits

    double efficiency = 0.;
    if((total  > 0.) && !energyEff)     efficiency = desired/total;
    if((totalE > 0.) &&  energyEff)     efficiency = desiredE/totalE;

    if((total  > 0.) && desiredHits  )  *desiredHits   = desired;
    if((total  > 0.) && totHits      )  *totHits       = total;
    if((totalE > 0.) && desiredEnergy)  *desiredEnergy = desiredE;
    if((totalE > 0.) && totalEnergy  )  *totalEnergy   = totalE;

    if(effMap){
      for(std::map<int, double>::iterator it = effMap->begin(); it != effMap->end(); ++it){
        if(totMap[it->first] > 0) 
	  it->second /= totMap[it->first];
        // This is an efficiency. It should be between 0-100%                                                           
        assert(it->second >= 0);
        assert(it->second <= 1);
      }
    }
    
    return efficiency;
  }


  //----------------------------------------------------------------------
  bool BackTracker::PassMuonCuts(int                                trackID,
                                 art::PtrVector<rb::CellHit> const& hits) const
  {
    int nxcells = 0;
    int nycells = 0;

    for(size_t h = 0; h < hits.size(); ++h){

      art::Ptr<rb::CellHit> hit = hits[h];
      std::vector<TrackIDE> tracks = HitToTrackIDE(hit);
      for(size_t t = 0; t < tracks.size(); ++t){
        if(tracks[t].trackID == trackID){

          if (hit->View() == geo::kX){
            nxcells++;
          }// end of loop over x hits

          if (hit->View() == geo::kY){
            nycells++;
          }// end of loop over y hits
          
        }// end of loop over hits from the muon
      }// end of loop over tracks
    }// end loop over hits


    if ((nxcells >= 3) && (nycells >= 3)) return 1;
    else return 0;
  }
    

  //----------------------------------------------------------------------
  bool BackTracker::sort_eff(const NeutrinoEffPur& a, const NeutrinoEffPur& b){
    return a.efficiency>b.efficiency;
  }

  //----------------------------------------------------------------------
  bool BackTracker::sort_pur(const NeutrinoEffPur& a, const NeutrinoEffPur& b){
    return a.purity>b.purity;
  }

  //----------------------------------------------------------------------
  std::vector<NeutrinoEffPur> BackTracker::
  SliceToNeutrinoInteractions(art::PtrVector<rb::CellHit> const& sliceHits,
                              art::PtrVector<rb::CellHit> const& allHits,
                              bool                               sortPur) const
   {
    std::vector <NeutrinoEffPur> sliceNeutrinoInteractions;
    
    for (unsigned int s = 0; s < fMCTruthList.size(); s++){
      const art::Ptr<simb::MCTruth> neutrinoInteraction = fMCTruthList[s];
      if(!neutrinoInteraction->NeutrinoSet())
        continue;
      
      std::vector<const sim::Particle*> particleList = MCTruthToParticles(neutrinoInteraction);
      std::set<int> neutrinoTrackIDs;
      for (unsigned int i = 0; i < particleList.size(); ++i){
        neutrinoTrackIDs.insert(particleList[i]->TrackId());
      }

      double energySlice = -1.0;
      double energyTotal = -1.0;
      int    nSliceHits  = 0;
      int    nTotalHits  = 0;

      double sliceEff = HitCollectionEfficiency(neutrinoTrackIDs, sliceHits, allHits, geo::kXorY, 0, true, &energySlice, &energyTotal, &nSliceHits, &nTotalHits);
      double slicePur = HitCollectionPurity(neutrinoTrackIDs, sliceHits, 0, 0, true);

      if (sliceEff > 0){
        NeutrinoEffPur sliceNeutrinoInteraction = {neutrinoInteraction, sliceEff, slicePur, s, energySlice, energyTotal, nSliceHits, nTotalHits};
        sliceNeutrinoInteractions.push_back(sliceNeutrinoInteraction);
      }
    }//End of loop over neutrino interactions 

    if (!sortPur) std::sort(sliceNeutrinoInteractions.begin(),sliceNeutrinoInteractions.end(),sort_eff);
    else          std::sort(sliceNeutrinoInteractions.begin(),sliceNeutrinoInteractions.end(),sort_pur);

    return sliceNeutrinoInteractions;
  }
  
  //----------------------------------------------------------------------
  std::vector<NeutrinoEffPur> BackTracker::
  SliceToNeutrinoInteractions(art::Ptr<rb::Cluster>       const& sliceCluster,
                              art::PtrVector<rb::CellHit> const& allHits,
                              bool                               sortPur) const
  {
    art::PtrVector< rb::CellHit > sliceHits;
    sliceHits = sliceCluster->AllCells();
    
    return SliceToNeutrinoInteractions(sliceHits, allHits, sortPur);
  }
  
  //----------------------------------------------------------------------
  std::vector<NeutrinoEffPur> BackTracker::
  SliceToMCTruth(art::PtrVector<rb::CellHit> const& sliceHits,
                              art::PtrVector<rb::CellHit> const& allHits,
                              bool                               sortPur) const
  {
    std::vector <NeutrinoEffPur> sliceNeutrinoInteractions;
    
    for (unsigned int s = 0; s < fMCTruthList.size(); s++){
      const art::Ptr<simb::MCTruth> neutrinoInteraction = fMCTruthList[s];
      
      std::vector<const sim::Particle*> particleList = MCTruthToParticles(neutrinoInteraction);
      std::set<int> neutrinoTrackIDs;
      for (unsigned int i = 0; i < particleList.size(); ++i){
        neutrinoTrackIDs.insert(particleList[i]->TrackId());
      }
      
      double energySlice = -1.0;
      double energyTotal = -1.0;
      int    nSliceHits  = 0;
      int    nTotalHits  = 0;

      double sliceEff = HitCollectionEfficiency(neutrinoTrackIDs, sliceHits, allHits, geo::kXorY, 0, true, &energySlice, &energyTotal, &nSliceHits, &nTotalHits);
      double slicePur = HitCollectionPurity(neutrinoTrackIDs, sliceHits, 0, 0, true);
      
      if (sliceEff > 0){
        NeutrinoEffPur sliceNeutrinoInteraction = {neutrinoInteraction, sliceEff, slicePur, s, energySlice, energyTotal, nSliceHits, nTotalHits};
        sliceNeutrinoInteractions.push_back(sliceNeutrinoInteraction);
      }
    }//End of loop over neutrino interactions 
    
    if (!sortPur) std::sort(sliceNeutrinoInteractions.begin(),sliceNeutrinoInteractions.end(),sort_eff);
    else          std::sort(sliceNeutrinoInteractions.begin(),sliceNeutrinoInteractions.end(),sort_pur);
    
    return sliceNeutrinoInteractions;
  }
  
  //----------------------------------------------------------------------
  std::vector<NeutrinoEffPur> BackTracker::
  SliceToMCTruth(art::Ptr<rb::Cluster>       const& sliceCluster,
                              art::PtrVector<rb::CellHit> const& allHits,
                              bool                               sortPur) const
  {
    art::PtrVector< rb::CellHit > sliceHits;
    sliceHits = sliceCluster->AllCells();
    
    return SliceToMCTruth(sliceHits, allHits, sortPur);
  }

  //----------------------------------------------------------------------
  std::vector< std::vector<cheat::NeutrinoEffPur> > BackTracker::
  SlicesToMCTruthsTable(art::PtrVector<rb::Cluster> const& sliceList) const
  {

    unsigned int numSlices = sliceList.size();
    std::vector<NeutrinoWithIndex> truthList = allMCTruth();
    unsigned int numTruths = truthList.size();

    //This is what will eventually be returned
    std::vector< std::vector<cheat::NeutrinoEffPur> > table(numSlices,std::vector<cheat::NeutrinoEffPur>(numTruths));
    
    //Loop to fill NeutrinoEffPur's with proper truth information
    for (unsigned int s = 0; s < numSlices; s++){
      for (unsigned int t = 0; t < numTruths; t++){
        cheat::NeutrinoEffPur truthFill = {truthList[t].neutrinoInt,0.0,0.0,truthList[t].truthColIndex,0.0,0.0,0,0};
        table[s][t] = truthFill; 
      }
    }//End of defining basic container with truths filled in

    for (unsigned int s = 0; s < numSlices; s++){
      //Vector to hold the total energy contributions to slice from each MC Truth
      std::vector<double> sliceEnergies(numTruths,0.0);
      //Vector to hold the total number of hits in the slice from each MC Truth
      std::vector<int> sliceHits(numTruths,0);

      for (unsigned int h = 0; h < sliceList[s]->NCell(); h++){
        //Don't worry about noise hits - they can't count towards the energy an interaction deposited
        if (IsNoise(sliceList[s]->Cell(h))) continue;
        //Vector to hold the total energy contributions to hit from each MC Truth
        std::vector<double> hitEnergies(numTruths,0.0);

        const std::vector<TrackIDE> trackIDs = HitToTrackIDE(sliceList[s]->Cell(h));
        for (unsigned int id = 0; id < trackIDs.size(); id++){
          int trackID = trackIDs[id].trackID;
          //Gets the MCTruth index for the particle trackID
          int truthID = fTrackIDToMCTruthIndex.find(abs(trackID))->second;
          hitEnergies[truthID] += trackIDs[id].energy;
        }//End of loop over track ids for the hit   
        
        //Add hit values into the slice vectors
        for (unsigned int q = 0; q < numTruths; q++){  
          if (hitEnergies[q] > 0.0){
            sliceEnergies[q] += hitEnergies[q];
            sliceHits[q] += 1;
          }//Only do for MCTruths with nonzero contributions to hit
        }//End of loop adding hit values to slice vectors  
   
      }//End of loop over hits in slice

      double totalSliceE = std::accumulate(sliceEnergies.begin(),sliceEnergies.end(),0.0);
      for (unsigned int t = 0; t < numTruths; t++){

        double purity = 0.0;
        if (totalSliceE > 0){
          purity = (sliceEnergies[t])/totalSliceE;
        }

        if ((purity < 0) || (purity > 1)) {
	throw cet::exception("BackTrackerIllegalPurity")
          << "The purity value "<<purity<<" is not allowed. \n";
      }

        table[s][t].purity = purity;
        table[s][t].energySlice = sliceEnergies[t];
        table[s][t].nSliceHits = sliceHits[t];

        }//End of loop over MCTruths for the slice
    }//End of loop over slices

    for (unsigned int t = 0; t < numTruths; t++){

      double totalMCTruthE = 0.0;
      int totalMCTruthHits = 0;
      
      for (unsigned int s = 0; s < numSlices; s++){
        totalMCTruthE += table[s][t].energySlice;
        totalMCTruthHits += table[s][t].nSliceHits;
      }//Loop over slices to find summed values

      for (unsigned int s = 0; s < numSlices; s++){

        double efficiency = 0.0;
        if (totalMCTruthE > 0){
          efficiency = (table[s][t].energySlice)/totalMCTruthE;
        }

        if ((efficiency < 0) || (efficiency > 1)) {
	throw cet::exception("BackTrackerIllegalEfficiency")
          << "The efficiency value "<<efficiency<<" is not allowed. \n";
      }

        table[s][t].efficiency = efficiency;
        table[s][t].energyTotal = totalMCTruthE;
        table[s][t].nTotalHits = totalMCTruthHits;
      }//Loop over slices to insert final values
      
    }//End of final loop over MCTruths for summed values
    
    return table;
  }

  //----------------------------------------------------------------------
  std::vector<ParticleEffPur>  BackTracker::TracksToParticles(std::vector<art::Ptr<rb::Track > > const& tracks,
							      std::set<int>                      const& trackIDs, 
							      std::set<int>                      const& allowedDaughterIDs,
							      art::PtrVector<rb::CellHit>        const& allHits) const
  {
    std::vector<ParticleEffPur> matches;

    std::set<int> trkIDs = trackIDs; 
    std::set<int> allowedDaughters = allowedDaughterIDs;

    std::map<int, int> parents;
    for(std::set<int>::iterator it = trkIDs.begin(); it != trkIDs.end(); ++it){
      const sim::Particle* mother = TrackIDToParticle(*it);
      if(!mother) continue;
      for(int i = 0; i < mother->NumberDaughters(); ++i){
	const int id = mother->Daughter(i);
	assert(TrackIDToParticle(id));
	if(allowedDaughters.count(TrackIDToParticle(id)->PdgCode())){
	  parents[id] = *it;
	}
      } // end for i
    } // end for it

      // loop over all tracks
    for(unsigned int itrack = 0; itrack < tracks.size(); ++itrack){
      // get the purity of the track for all track IDs
      std::map<int, double> purs;
      HitCollectionPurity(trkIDs,tracks[itrack]->AllCells(),&purs,&parents);
      // find the most pure trackID for this track
      int bestid = -1;
      double bestpur = 0.0;
      for(std::set<int>::iterator it = trkIDs.begin(); it != trkIDs.end(); ++it){
	const double p = purs[*it];
	if(p > bestpur){
	  bestpur = p;
	  bestid = *it;
	}
      } // end of loop over track ids

      // update this tracks info for the matched true particle      
      double recoPur = bestpur;
      int recotrkid = bestid;
      int recopdg = 0;
      double recoEff = 0.0;
      if(bestpur > 0.0){
	// found a real particle that matched the track, update pdg and efficiency to match this
	recopdg = TrackIDToParticle(recotrkid)->PdgCode();
	std::set<int> ideff;
	ideff.insert(recotrkid);
	recoEff = HitCollectionEfficiency(ideff,tracks[itrack]->AllCells(),allHits,geo::kXorY);
	// remove this track id from the set of track ids so no other tracks can try to match to it
	trkIDs.erase(trkIDs.find(recotrkid));
	// remove any instance where this track id is the parent particle in the parents map
	std::map<int, int>::iterator parentsIt = parents.begin();
	while(parentsIt != parents.end()){
	  if(parentsIt->second == recotrkid){ parents.erase(parentsIt++); }
	  else{ ++parentsIt; }
	}
      }
      // record this tracks match
      ParticleEffPur match = {recotrkid, recoEff, recoPur, recopdg};
      matches.push_back(match);
    } // end of loop over tracks for matching to truth
    
    return matches;

  }

  //----------------------------------------------------------------------
  std::vector<NeutrinoWithIndex> BackTracker::allNeutrinoInteractions() const
   {
    std::vector <NeutrinoWithIndex> allNeutrinoInteractions;
    
    for (unsigned int s = 0; s < fMCTruthList.size(); s++){
      const art::Ptr<simb::MCTruth> neutrinoInteraction = fMCTruthList[s];
      if(!neutrinoInteraction->NeutrinoSet()) 
        continue;
      
      NeutrinoWithIndex neutrino = {neutrinoInteraction, s};
      allNeutrinoInteractions.push_back(neutrino);
    }

    return allNeutrinoInteractions;
  }

  //----------------------------------------------------------------------
  std::vector<NeutrinoWithIndex> BackTracker::allMCTruth() const
  {
    std::vector <NeutrinoWithIndex> allTruth;
    
    for (unsigned int s = 0; s < fMCTruthList.size(); s++){
      const art::Ptr<simb::MCTruth> neutrinoInteraction = fMCTruthList[s];
      NeutrinoWithIndex neutrino = {neutrinoInteraction, s};
      allTruth.push_back(neutrino);
    }
    
    return allTruth;
  }
  
  
  //----------------------------------------------------------------------
  bool BackTracker::CloseInTime(const rb::CellHit& chit, const sim::PhotonSignal& phot) const
  {
    //The constants used to determine this value change depending on detector
    //and hit timing reconstruction. See docdb 9092 to see how the current 
    //constants were determined (May 2013). To redo the study and see if new
    //constants are required, use MCCheckOut/CloseInTime_module.cc.

    // Empirically determined as an over-generous width of the timing reconstruction.
    double kWindowNs = 1000;
    // Empirically determined as the approximate mean bias of the timing reconstruction.
    // This is the amount the CellHit is expected to be earlier than the photons.
    double kOffsetNs = 350;

    if ((fDetID == novadaq::cnv::kFARDET)&&(chit.GoodTiming())){
      kWindowNs = 500;
      kOffsetNs = 10;
    }

    if ((fDetID == novadaq::cnv::kNEARDET)&&(chit.GoodTiming())){
      kWindowNs = 200;
      kOffsetNs = 10;
    }

    const double dt = chit.TNS()-phot.TimeMean();

    return fabs(dt+kOffsetNs) <= kWindowNs;
  }

  //----------------------------------------------------------------------
  void BackTracker::AccumulateHitContributions(art::Ptr<rb::CellHit> const& hit,
					       std::map<int, int>         & idToNPhot) const
  {
    auto it = fCellToPhotonSignal.find(geo::OfflineChan(hit->Plane(), hit->Cell()));
    // Nothing to accumulate from this cell
    if(it == fCellToPhotonSignal.end()) return;

    // Candidate photons, in that they're in the right cell.
    const std::vector<sim::PhotonSignal>& phots = it->second;

    // For every photon that's sufficiently in time with the hit,
    // accumulate its energy into the relevant trackID.
    for(unsigned int i = 0; i < phots.size(); ++i){
      const sim::PhotonSignal& phot = phots[i];
      if(CloseInTime(*hit.get(), phot)){
	idToNPhot[phot.TrackId()] += phot.NPhoton();
      }
    }
  }
  //----------------------------------------------------------------------

  std::vector<sim::FLSHit> BackTracker::ParticleToFLSHit(const int& trackID) const{
    std::unordered_map< int, std::list<sim::FLSHit> >::const_iterator it = fTrackIDToFLSHit.find(trackID);
    if( it == fTrackIDToFLSHit.end() ){
      std::vector<sim::FLSHit> empty;
      LOG_DEBUG("BackTracker") << "no collection of sim::FLSHit for "
			       <<"Track ID "<<trackID
                               << ". Returning empty vector";

      return empty;
    }
    else
      return std::vector<sim::FLSHit>(it->second.begin(), it->second.end());
  }


  //----------------------------------------------------------------------
  std::vector<sim::FLSHit> BackTracker::ParticleToFLSHit(const int trackID, const int pdg) const{

      const std::vector<sim::FLSHit>& flshits_with_same_track_id = ParticleToFLSHit(trackID);

      std::vector<sim::FLSHit> out_flshits;

      const int nflshits = flshits_with_same_track_id.size();

      /// Fill only FLSHits with a given PDG
      for(int i=0; i<nflshits; ++i){
          const sim::FLSHit& flshit= flshits_with_same_track_id[i];
          if(flshit.GetPDG() == pdg){
            out_flshits.push_back(flshit);
          }
      }

      return out_flshits;
  }

  //----------------------------------------------------------------------
  std::vector<rb::Cluster> BackTracker::ClusterByTruth(std::vector< art::Ptr< rb::CellHit > > const& hits)
  {
    std::vector<rb::Cluster> clusters(fMCTruthList.size()+1);

    for(unsigned int hiti = 0; hiti < hits.size(); ++hiti) {
      art::Ptr<rb::CellHit> hit = hits[hiti];

      // If the hit is noise, put it into the last cluster.
      if(this->IsNoise(hit)) {
	clusters[fMCTruthList.size()].Add(hit);
	continue;
      }

      const sim::Particle *part = this->HitToParticle(hit);
      unsigned int idx = fTrackIDToMCTruthIndex[part->TrackId()];
      clusters[idx].Add(hit);

    } // end loop over hits

    // Label the noise slice.
    clusters[fMCTruthList.size()].SetNoise(true);

    return clusters;
  }

  //----------------------------------------------------------------------
  rb::Cluster BackTracker::MCTruthToCluster(art::Ptr<simb::MCTruth>                const& mct,
					    std::vector< art::Ptr< rb::CellHit > > const& hits)
  {
    rb::Cluster cluster;

    // figure out where in the MCTruth list mct belongs
    unsigned int MCTidx;
    for(MCTidx = 0; MCTidx < fMCTruthList.size(); ++MCTidx) {
      if(fMCTruthList[MCTidx] == mct) break;
    }

    // loop over hits and add to the cluster
    for(unsigned int hiti = 0; hiti < hits.size(); ++hiti) {
      art::Ptr<rb::CellHit> hit = hits[hiti];

      // If the hit is noise, skip it.
      if(this->IsNoise(hit)) continue;

      const sim::Particle *part = this->HitToParticle(hit);
      unsigned int idx = fTrackIDToMCTruthIndex[part->TrackId()];
      if(idx == MCTidx) cluster.Add(hit);

    } // end loop over hits
    
    return cluster;
  }						       

} // namespace


namespace cheat{
  DEFINE_ART_SERVICE(BackTracker);
}
