////////////////////////////////////////////////////////////////////////
/// \brief  CheckBackTracking module
///
/// \author brebel@fnal.gov
///
////////////////////////////////////////////////////////////////////////

#include <string>
#include <vector>

// Framework includes
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "art/Framework/Core/EDAnalyzer.h"
#include "art/Framework/Principal/Event.h"
#include "fhiclcpp/ParameterSet.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Services/Optional/TFileService.h"
#include "art/Framework/Services/Optional/TFileDirectory.h"
#include "messagefacility/MessageLogger/MessageLogger.h"

// NOvASoft includes
#include "Geometry/Geometry.h"
#include "MCCheater/BackTracker.h"
#include "RecoBase/CellHit.h"
#include "Simulation/Particle.h"
#include "Simulation/ParticleHistory.h"
#include "Simulation/PhotonSignal.h"
#include "Simulation/FLSHit.h"

namespace cheat {
  class CheckBackTracking : public art::EDAnalyzer {
  public:
    explicit CheckBackTracking(fhicl::ParameterSet const& pset);
    virtual ~CheckBackTracking();

    void analyze(art::Event const& evt);

    void reconfigure(fhicl::ParameterSet const& pset);

 private:

    void CheckFLSHitConsistency(art::Ptr<rb::CellHit> const& hit);
    void CheckParticleConsistency(art::Ptr<rb::CellHit> const& hit);
    void CheckTrackIdToParticle();
    void CheckPurityAndEfficiency(std::vector< art::Ptr<rb::CellHit> > const& hits);
    
    art::ServiceHandle<cheat::BackTracker> fBT;             ///< BackTracker handle
    std::string                            fHitModuleLabel; ///< label for module creating rb::CellHit objects	   

  };
}

namespace cheat{

  //--------------------------------------------------------------------
  CheckBackTracking::CheckBackTracking(fhicl::ParameterSet const& pset)
    : EDAnalyzer(pset)
  {
    this->reconfigure(pset);
  }

  //--------------------------------------------------------------------
  CheckBackTracking::~CheckBackTracking()
  {
  }

  //--------------------------------------------------------------------
  void CheckBackTracking::reconfigure(fhicl::ParameterSet const& pset)
  {
    fHitModuleLabel = pset.get<std::string>("HitModuleLabel", "calhit");
  }

  //--------------------------------------------------------------------
  void CheckBackTracking::analyze(art::Event const& evt)
  {

    // grab the hits that have been reconstructed
    art::Handle< std::vector<rb::CellHit> > hitcol;
    evt.getByLabel(fHitModuleLabel, hitcol);
    
    // make a vector of them - we aren't writing anything out to a file
    // so no need for a art::PtrVector here
    std::vector< art::Ptr<rb::CellHit> > hits;
    art::fill_ptr_vector(hits, hitcol);

    // loop over the hits and figure out which particle contributed to each one
    std::vector< art::Ptr<rb::CellHit> >::iterator itr = hits.begin();

    while( itr != hits.end() ){

      // print the information for this hit
      mf::LogVerbatim("CheckBackTracking") << *(*itr).get();

      // check is this is a noise hit
      if (fBT->IsNoise((*itr))){
        mf::LogVerbatim("CheckBackTracking") << "This hit is a noise hit."
                                             << " No truth info will be printed.";
        itr++;
        continue;
      }

      this->CheckFLSHitConsistency(*itr);

      TVector3 xyz = fBT->HitToXYZ((*itr));  
      mf::LogVerbatim("CheckBackTracking") << "hit  weighted mean position is (" 
					   << xyz[0] << "," << xyz[1] << "," << xyz[2] << ")";
      
      std::vector<cheat::TrackIDE> trackides = fBT->HitToTrackIDE((*itr));
      for(size_t t = 0; t < trackides.size(); ++t){
	mf::LogVerbatim("CheckBackTracking") << "track id: " << trackides[t].trackID 
					     << " contributed " << trackides[t].energy << "/" 
					     << trackides[t].energyFrac 
					     << " to the current hit";
      }
      
      const sim::Particle* part = fBT->HitToParticle(*itr);
      if(part){
	mf::LogVerbatim("CheckBackTracking") << "largest contributing particle is: " << *part;
      }
      else{
	mf::LogVerbatim("CheckBackTracking") << "no particle corresponding to CellHit: " << *itr;
      }

      itr++;
    }// end loop over hits

    this->CheckTrackIdToParticle();
    this->CheckPurityAndEfficiency(hits);

    return;

  } // end analyze

  //----------------------------------------------------------------------------
  void CheckBackTracking::CheckFLSHitConsistency(art::Ptr<rb::CellHit> const& hit)
  {
    unsigned int plane = hit->Plane();
    unsigned int cell  = hit->Cell();

    const std::vector<sim::FLSHit> flshits   = fBT->CellToFLSHit(plane, cell); 
    const std::vector<sim::FLSHit> flshits2  = fBT->HitToFLSHit(hit); 

    for (size_t f = 0; f < flshits2.size(); ++f) {
      if ( std::find(flshits.begin(), flshits.end(), flshits2[f]) == flshits.end() ) {
	throw cet::exception("CheckBackTracking")
          << "FLSHits from Cell do not match FLSHits from Hit\n"
          << __FILE__ << ":" << __LINE__ << "\n";
      }
    }
    
    return;
  }

  //----------------------------------------------------------------------------
  void CheckBackTracking::CheckTrackIdToParticle()
  {
    // loop over the eveID values and calculate the purity and efficiency for each
    std::set<int>::iterator setitr = fBT->GetTrackIDList().begin();
    while( setitr != fBT->GetTrackIDList().end() ){

      sim::ParticleHistory phist(&(fBT->ParticleNavigator()), *setitr);
      
      const sim::Particle* pfromID  = fBT->TrackIDToParticle(*setitr);
      const sim::Particle* pfromNav = fBT->ParticleNavigator().find(*setitr)->second;
      if ( pfromID != pfromNav ) {
	throw cet::exception("CheckBackTracking")
          << "Same track ID returns different pointers: "
          << *pfromID << "\n"
          << *pfromNav << "\n"
          << __FILE__ << ":" << __LINE__ << "\n";
      }
      
      // attempt to recover the MCTruth object for this particle
      art::Ptr<simb::MCTruth> mct = fBT->TrackIDToMCTruth(*setitr);
      mf::LogVerbatim("CheckBackTracking") << "MCTruth for " << *setitr << " is " 
					   << *(mct.get());
      
      // check that you get the same art::Ptr<simb::MCTruth> from both accessors
      if ( mct.id() != fBT->ParticleToMCTruth(pfromID).id() ) {
	throw cet::exception("CheckBackTrack")
          << "Track ID and Particle return different MCTruth:\n"
          << *(mct.get()) << "\n"
          << *(fBT->ParticleToMCTruth(pfromID).get()) << "\n"
          << __FILE__ << ":" << __LINE__ << "\n";
      }
      
      // get the list of particles for this MCTruth
      std::vector<const sim::Particle*> parts = fBT->MCTruthToParticles(mct);
      if ( std::find(parts.begin(), parts.end(), pfromID) == parts.end() ) {
	throw cet::exception("CheckBackTracking")
          << "Cannot map particle to MCTruth and back again\n"
          << __FILE__ << ":" << __LINE__ << "\n";
      }
            
      // check the mother particle
      pfromNav = fBT->ParticleNavigator().find(pfromID->Mother())->second;
      pfromID  = fBT->TrackIDToMotherParticle(*setitr);
      if ( pfromID != pfromNav ) {
	throw cet::exception("CheckBackTracking")
          << "Same track ID returns different mothers: "
          << *pfromID << "\n"
          << *pfromNav << "\n\t Parentage is:\n\t"
          << phist << "\n"
          << __FILE__ << ":" << __LINE__ << "\n";
      }

    }
    
    return;
  }

  //----------------------------------------------------------------------------
  void CheckBackTracking::CheckPurityAndEfficiency(std::vector< art::Ptr<rb::CellHit> > const& hits)
  {
    art::PtrVector<rb::CellHit> hitPtrVec;
    for(size_t h = 0; h < hits.size(); ++h) hitPtrVec.push_back(hits[h]);

    // loop over the track ID values and calculate the purity and efficiency for each
    std::set<int>::iterator setitr = fBT->GetTrackIDList().begin();
    while( setitr != fBT->GetTrackIDList().end() ){

      std::set<int> id;
      id.insert(*setitr);
      mf::LogVerbatim ("CheckBackTracking") << "track ID: " << *setitr 
					    << " purity: " 
					    << fBT->HitCollectionPurity(id, hitPtrVec)
					    << " efficiency: "
					    << fBT->HitCollectionEfficiency(id, hitPtrVec, hitPtrVec, geo::kXorY);

      
      setitr++;
    }// end loop over track ids

    return;
  }

} // end namespace


namespace cheat{

  DEFINE_ART_MODULE(CheckBackTracking);

}

