///////////////////////////////////////////////////////////////////////////////
// Class:       TestTrackIds
// Module Type: analyzer
// File:        TestTrackIds_module.cc
///////////////////////////////////////////////////////////////////////////////

// ART includes
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Core/EDAnalyzer.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Services/Optional/TFileService.h"
#include "art/Framework/Services/Optional/TFileDirectory.h"
#include "messagefacility/MessageLogger/MessageLogger.h"

// NOvA includes
#include "MCCheater/BackTracker.h"
#include "SimulationBase/MCTruth.h"
#include "Simulation/FLSHit.h"
#include "Simulation/FLSHitList.h"
#include "Simulation/ParticleNavigator.h"

// ROOT includes
#include "TH1.h"

namespace cheat
{
  class TestTrackIds : public art::EDAnalyzer
  {
  public:
    explicit TestTrackIds(const fhicl::ParameterSet& p);
    virtual ~TestTrackIds();

    void analyze(const art::Event& evt);
    void beginJob();

  private:
    /// Function that actually aborts job if a particular test failed
    /// For now, equivalent to a single if condition
    void CrashIfFCLSaysSo() const;

    bool fCrashUponError; ///< Flag whether to abort if a particular test failed

    TH1F* fErrorMode; ///< Histogram breaking down how events failed

    std::map<std::string, double> fErrorModeMap; ///< Map of bin number to a specific error type

    static bool EssentiallyEqual(double a, double b, double precision = 0.5)
    { return a <= (b + precision) && a >= (b - precision); }

    static bool EssentiallyEqualVertex(TLorentzVector a, TLorentzVector b, double precision = 0.5)
    {
      return (EssentiallyEqual(a.X(), b.X(), precision) &&
              EssentiallyEqual(a.Y(), b.Y(), precision) &&
              EssentiallyEqual(a.Z(), b.Z(), precision));
    }

  };

  //----------------------------------------------------------------------
  TestTrackIds::TestTrackIds(const fhicl::ParameterSet& pset)
    : EDAnalyzer(pset),
      fCrashUponError(pset.get<bool>("CrashUponError"))
  {
  }

  //----------------------------------------------------------------------
  TestTrackIds::~TestTrackIds()
  {
  }

  //----------------------------------------------------------------------
  void TestTrackIds::beginJob()
  {
    art::ServiceHandle<art::TFileService> tfs;

    fErrorMode = tfs->make<TH1F>("fErrorMode", "Error Mode;;Events", 13, 0.5, 13.5);

    // Start at 1 to match ROOT bin convention
    // To avoid filling errors, find/replace the strings below,
    // instead of changing it directly and then searching
    fErrorModeMap["Real Data"] = 1.;
    fErrorModeMap["No MCTruth List"] = 2.;
    fErrorModeMap["No PhotonSignals"] = 3.;
    fErrorModeMap["No FLSHit match to Track"] = 4.;
    fErrorModeMap["No FLSHit List"] = 5.;
    fErrorModeMap["No Particle match to FLSHit"] = 6.;
    fErrorModeMap["Particle exists, not matched to FLSHit"] = 7.;
    fErrorModeMap["No Mother match to Track"] = 8.;
    fErrorModeMap["Mother exists, not matched to Track"] = 9.;
    fErrorModeMap["No Particle at Track"] = 10.;
    fErrorModeMap["Missing Daughter"] = 11.;
    fErrorModeMap["Bad Truth Match"] = 12.;
    fErrorModeMap["Might Be Cosmic"] = 13.;

    for(const auto& errorModePair : fErrorModeMap) {
      int bin = (int)errorModePair.second;
      std::string label = errorModePair.first;
      fErrorMode->GetXaxis()->SetBinLabel(bin, label.c_str());
    }
  }

  //----------------------------------------------------------------------
  void TestTrackIds::analyze(const art::Event& evt)
  {
    // Check that this is MC, stop if it isn't
    if(evt.isRealData()) {
      fErrorMode->Fill(fErrorModeMap["Real Data"]);

      CrashIfFCLSaysSo();

      return;
    }

    // Get the MC generator information out of the event
    bool isMCTruthListEmpty = false;
    art::Handle<std::vector<simb::MCTruth> > truths;
    try {
      evt.getByLabel("generator" /*fMCTruthModuleLabel*/, truths);
      if(truths->empty()) { isMCTruthListEmpty = true; }

      if(truths->size() > 1) {
        mf::LogVerbatim("TestTrackIds") << "Event " << evt.event() << " has "
                                        << truths->size() << " truth records." << std::endl;
      }
    }
    catch(...) { isMCTruthListEmpty = true; }
    if(isMCTruthListEmpty) {
      mf::LogError("TestTrackIds MCTruth") << "Error retrieving MCTruth list." << std::endl;
      fErrorMode->Fill(fErrorModeMap["No MCTruth List"]);

      CrashIfFCLSaysSo();

      return;
    }

    art::ServiceHandle<cheat::BackTracker> bt;

    // Invariant tested here is that every photon signal has a matching FLSHit stored
    bool isPhotonListError = false;
    art::Handle<std::vector<sim::PhotonSignal> > photlist;
    try {
      evt.getByLabel("photrans" /*fPhotonModuleLabel*/, photlist);
    }
    catch(...) {
      isPhotonListError = true;
      mf::LogWarning("TestTrackIds PhotonSignal") << "Error retrieving PhotonSignal list." << std::endl;
      fErrorMode->Fill(fErrorModeMap["No PhotonSignals"]);
    }

    if(!isPhotonListError) {
      std::vector<art::Ptr<sim::PhotonSignal> > phots;
      art::fill_ptr_vector(phots, photlist);

      // Loop over all photons
      // phot should be a sim::PhotonSignal
      for(const auto& phot : phots) {
        // Get the FLS hits from the photon
        const std::vector<sim::FLSHit> flslist = bt->CellToFLSHit(phot->Plane(),
                                                                  phot->Cell());

        // Loop over FLS hits
        // Check that the FLS hit and photon have the same track; one match is enough
        bool matched = false;
        for(const auto& flshit : flslist) {
          if(flshit.GetTrackID() == phot->TrackId()) {
            matched = true;
            break;
          }
        }

        // Log an error in the (unlikely) event the tracks don't match
        if(!matched) {
          mf::LogError("TestTrackIds") << "Event " << evt.event() 
                                       << ": PhotonSignal in plane " << phot->Plane() 
                                       << ", cell " << phot->Cell() 
                                       << ", TrackId " << phot->TrackId() 
                                       << " does not have any matching FLSHits." << std::endl;
          fErrorMode->Fill(fErrorModeMap["Track has no FLSHits"]);

          CrashIfFCLSaysSo();
        }
      } // end of loop over PhotonSignals
    } // end of conditional that there was no error retrieving the Photon List

    // Now get the sim::Particles from the event
    sim::ParticleNavigator pnav = bt->ParticleNavigator();

    // Used for a sanity checks of pnav
    const std::set<int> ids = bt->GetTrackIDList();

    // Invariant here is that every FLSHit's TrackId corresponds to a particle stored in the event
    bool isFLSHitListError = false;
    art::Handle<std::vector<sim::FLSHitList> > flshitlist;
    try {
      evt.getByLabel("geantgen" /*fGeantModuleLabel*/, flshitlist);
    }
    catch(...) {
      isFLSHitListError = true;
      mf::LogError("TestTrackIds FLSHitList") << "Error retrieving FLSHitList list." << std::endl;
      fErrorMode->Fill(fErrorModeMap["No FLSHit List"]);
    }

    if(!isFLSHitListError) {
      for(unsigned int i_List = 0, n_List = flshitlist->size(); i_List < n_List; ++i_List) {
        const std::vector<sim::FLSHit>& hits = flshitlist->at(i_List).fHits;

        for(const auto& hit : hits){
          const sim::ParticleNavigator::iterator part = pnav.find(hit.GetTrackID());
          if(part == pnav.end()){
            mf::LogError("TestTrackIds") << "Event " << evt.event() 
                                         << ": FLS hit in plane " << hit.GetPlaneID()
                                         << ", cell " << hit.GetCellID()
                                         << ", trackId " << hit.GetTrackID()
                                         << ", PDG code " << hit.GetPDG()
                                         << " has no matching sim::Particle" << std::endl;
            fErrorMode->Fill(fErrorModeMap["No Particle match to FLSHit"]);

            // Make sure it really isn't there
            if(ids.count(hit.GetTrackID()) != 0) {
              mf::LogWarning("TestTrackIds") << "However, the sim::Particle does exist." << std::endl;
              fErrorMode->Fill(fErrorModeMap["Particle exists, not matched to FLSHit"]);
            }

            CrashIfFCLSaysSo();
          }
        } // end of loop over FLSHits
      } // end of loop over lists in the FLSHitList
    } // end of conditional that there was no error retrieving the FLSHitList

    int badMatch = 0;
    std::vector<NeutrinoWithIndex> allNeutrinos = bt->allNeutrinoInteractions();
    
    // Invariant tested here is that every non-primary particle has its mother stored
    for(sim::ParticleNavigator::iterator it = pnav.begin(); it != pnav.end(); ++it) {
      const int id = it->first;

      if(pnav.IsPrimary(id)) {
        // Test that primary muons match vertex location of neutrino mother
        const sim::Particle* muon = bt->TrackIDToParticle(id);
        int pdg = muon->PdgCode();
        int mother = muon->Mother();
        TLorentzVector muonVert = muon->Position();
        double muonE = muon->E();

        const art::Ptr<simb::MCTruth>& motherTruth = bt->ParticleToMCTruth(muon);
        if(motherTruth->NeutrinoSet()) {
          TLorentzVector motherVert = motherTruth->GetNeutrino().Nu().Position();
          double lepE = motherTruth->GetNeutrino().Lepton().E();

          if(!EssentiallyEqualVertex(muonVert, motherVert)) {
            // Want to find neutrino in vector
            int matchedNeutrino = -1;
            int muonNeutrino = -1;

            for(unsigned int q = 0; q < allNeutrinos.size(); ++q) {
              TLorentzVector otherVert = allNeutrinos[q].neutrinoInt->GetNeutrino().Nu().Position();

              if(EssentiallyEqualVertex(otherVert, motherVert)) {
                matchedNeutrino = allNeutrinos[q].truthColIndex;
              } // Matches our neutrino truth

              if(EssentiallyEqualVertex(otherVert, muonVert)) {
                muonNeutrino = allNeutrinos[q].truthColIndex;
              } // Matches our muon vertex
            } // end of loop over vector of NeutrinoWithIndex

            // Sometimes primary particles travel away from the vertex somewhat and don't match any truth perfectly ... don't worry
            // Special cased particles: charged pions, photons, protons, neutrons, K longs, charged kaons
            // These were choosen because I saw examples of this behaviour. Others could be added.
            std::set<int> specialCases = {211, -211, 22, 2212, 130, -321, 321, 2112};
            if(!(specialCases.find(pdg) != specialCases.end() && muonNeutrino == -1)) {
              ++badMatch;
              mf::LogError("TestTrackIds") << " PDG: " << pdg << " Mother: " << mother << std::endl
                                           << "  Particle Vx: " << muon->Vx()
                                           << " Vy: " << muon->Vy()
                                           << " Vz: " << muon->Vz()
                                           << " Ptc E: " << muonE << std::endl
                                           << "  Neutrino Vx: " << motherTruth->GetNeutrino().Nu().Vx()
                                           << " Vy: " << motherTruth->GetNeutrino().Nu().Vy()
                                           << " Vz: " << motherTruth->GetNeutrino().Nu().Vz()
                                           << " Lep E: " << lepE << std::endl
                                           << "  Neutrino matched index: " << matchedNeutrino << std::endl
                                           << "  Ptc. vtx matched index: " << muonNeutrino << std::endl;

              fErrorMode->Fill(fErrorModeMap["Bad Truth Match"]);
              CrashIfFCLSaysSo();
            } // We are sad that they don't match - not just because particle traveled somewhat
          } // Vertex doesn't match between primary particle and neutrino truth
        }
        else {
          mf::LogWarning("Might Be Cosmic") << "The mother truth had no neutrino."
                                            << "This is a problem if the file is not cosmics." << std::endl;
          
          fErrorMode->Fill(fErrorModeMap["Might Be Cosmic"]);
          CrashIfFCLSaysSo();
        }

        continue;
      } // Skip the particle if it is a primary

      const int mom = it->second->Mother();
      const int pdg = it->second->PdgCode();
      const std::string proc = it->second->Process();

      if(pnav.find(mom) == pnav.end()) {
        mf::LogError("TestTrackIds") << "Event " << evt.event() 
                                     << ": Particle " << id 
                                     << ", PDG code " << pdg 
                                     << ", formed by process " << proc 
                                     << ", can't find mother, index " << mom << std::endl;
        fErrorMode->Fill(fErrorModeMap["No Mother match to Track"]);

        // Make sure it really isn't there
        if(ids.count(mom) != 0) {
          mf::LogWarning("TestTrackIds") << "However, the mother does exist." << std::endl;
          fErrorMode->Fill(fErrorModeMap["Mother exists, not matched to Track"]);

          CrashIfFCLSaysSo();
        }
      }
    } // end of loop over particles

    if(badMatch != 0) {
      mf::LogError("TestTrackIds") << " Number of bad matches of primary particles to simb truths is: " << badMatch << std::endl;
    }

    // Invaraint tested here is that all daughters of a particle should exist
    std::set<int> trackIDs = bt->GetTrackIDList();
    for(const auto& trackID : trackIDs) {
      const sim::Particle* mother = bt->TrackIDToParticle(trackID);

      if(!mother) {
        mf::LogError("TestTrackIds") << "Event " << evt.event() 
                                     << ", can't find particle corresponding to track id: " << trackID << std::endl; 
        fErrorMode->Fill(fErrorModeMap["No Particle at Track"]);

        CrashIfFCLSaysSo();

        continue;
      }

      for(int i_daughter = 0, n_daughter = mother->NumberDaughters(); i_daughter < n_daughter; ++i_daughter) {
        const int id = mother->Daughter(i_daughter);

        if(!bt->TrackIDToParticle(id)) {
          mf::LogError("TestTrackIds") << "Event " << evt.event() 
                                       << ", can't find the daughter of particle: " << trackID
                                       << ", and the daughter has track id: " << id << std::endl;
          fErrorMode->Fill(fErrorModeMap["Missing Daughter"]);

          CrashIfFCLSaysSo();
        }
      } // end of loop over the daughters of the current particle
     
    } // end of loop over all track ids

  } // end of analyze

  //----------------------------------------------------------------------
  void TestTrackIds::CrashIfFCLSaysSo() const
  {
    if(fCrashUponError) { abort(); }
    return;
  }

  DEFINE_ART_MODULE(TestTrackIds);
} // end of namespace cheat
