/// Macro to export histogram content of a ROOT file to an html page
/// Based on $ROOTSYS/bin/hadd.C
/// The script takes the file names of the two releases to be compared
/// Raphael Schroeter, May 2011

// C/C++ includes
#include <iostream>
#include <string.h>

// ROOT includes
#include "Riostream.h"
#include "TCanvas.h"
#include "TChain.h"
#include "TFile.h"
#include "TH1.h"
#include "TH2.h"
#include "TH3.h"
#include "TImage.h"
#include "TKey.h"
#include "TString.h"
#include "TSystem.h"

TString htmlpath; ///< Path to html page and local location of saved images
TString subDir;   ///< Name of current subdirectory

ofstream *histhtml; ///< Stream to send html code/text/commands to html file

double NormPOTPrevious, NormPOTCurrent; ///< Amount of POT used to generate input files

/// Write the initial portion of the html file and set up top level directory
void InitializeHtml(char* fileNew, char* fileOld);

/// Set up subdirectory structure of html file and local area
void WriteDirectoryStruct(TString inPath, TList* sourceList);

/// Get the amount of POT from both input files
void GetPOTInfo(TString inPath, TList* sourceList);

/// Recursively compare each histogram in the inPath directory
void CompareRootfile(TString inPath, TList* sourceList);

/// Write the final lines of the html file
void Finalize();

//.............................................................................
void HtmlGenieCompare(char* fileNew, char* fileOld)
{
  TList* fileList; // List of input files
  TString topPath = "/"; // Path to top directory of histogram locations/subdirectories

  // Open the files, add them to the TList
  fileList = new TList();
  fileList->Add(TFile::Open(fileNew));
  fileList->Add(TFile::Open(fileOld));

  // Get the base file names
  std::string baseFileNew(fileNew);
  std::string baseFileOld(fileOld);
  baseFileNew.erase(0, baseFileNew.rfind("/"));
  baseFileOld.erase(0, baseFileOld.rfind("/"));

  InitializeHtml((char*)baseFileNew.c_str(), (char*)baseFileOld.c_str());
  WriteDirectoryStruct(topPath, fileList);
  GetPOTInfo("potana", fileList);
  CompareRootfile(topPath, fileList);
  Finalize();

  return;
}

//.............................................................................
void InitializeHtml(char* fileNew, char* fileOld)
{
  // Set up html files for ease of image viewing
  gSystem->Exec("mkdir html_out/");

  char commandHelper[256];
  sprintf(commandHelper, "mkdir html_out/%s_comp/", fileNew);
  gSystem->Exec(commandHelper);

  sprintf(commandHelper, "html_out/%s_comp/", fileNew);
  htmlpath = commandHelper;
  histhtml = new ofstream(htmlpath + "index.html", ios::out);

  *histhtml << "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01 Transitional//EN\">\n" 
            << "<html>\n <head>\n"
            << "<title>" << fileNew << "</title>\n"
            << "<meta http-equiv=\"Content-Type\" content=\"text/html; charset=iso-8859-1\">\n"
            << "</head>\n<body>\n" << std::endl;

  *histhtml << "<a name=\"pagetop\" ></a>\n" 
            << "<h1>" << fileNew << "</h1>\n" << std::endl;

  *histhtml << "<table>" << std::endl;
  *histhtml << "<tr>" << std::endl;
  *histhtml << "<td>" << std::endl;
  *histhtml << "<h2>"<< fileNew <<"</font></h2>\n" << std::endl; 
  *histhtml << "</td>"<< std::endl;
  *histhtml << "<td>" << std::endl;
  *histhtml << "<h2><font color=\"#FF0000\">"<< fileOld <<"</font></h2>\n" << std::endl; 
  *histhtml << "</td>"<< std::endl;
  *histhtml << "</tr>" << std::endl;
  *histhtml << "</table>" << std::endl;

  *histhtml << "Skip to:\n" << std::endl;

  return;
}

//.............................................................................
void WriteDirectoryStruct(TString inPath, TList* sourceList)
{
  TFile* firstSource = (TFile*)sourceList->First();
  firstSource->cd(inPath);
  TDirectory* currentSourceDir = gDirectory;

  // Loop over all keys in this directory and write the subdirectory index on the html file
  TIter nextkey(currentSourceDir->GetListOfKeys());
  TKey* key;
  TKey* oldkey = 0;  

  while((key = (TKey*)nextkey())) {
    // Keep only the highest cycle number for each key
    if(oldkey && !strcmp(oldkey->GetName(), key->GetName())) { continue; }

    // Read object from first source file
    firstSource->cd(inPath);
    TObject* obj = key->ReadObj();

    if(obj->IsA()->InheritsFrom(TDirectory::Class())) { // It's a subdirectory
      subDir = obj->GetName();
      *histhtml << "<li><a href=\"#" << subDir << "\">" << subDir << "</a></li>\n" << std::endl;
    }

    delete obj;
  } // end of while loop over TKeys

  return;
}

//.............................................................................
void GetPOTInfo(TString inPath, TList* sourceList)
{
  // Name of the histogram to get POT information from
  std::string histPOT = "POTSubrun";

  TFile* firstSource = (TFile*)sourceList->First();

  if(firstSource->cd(inPath)) {
    TKey* keyPOTNew = (TKey*)gDirectory->GetListOfKeys()->FindObject(histPOT.c_str());

    if(keyPOTNew) {
      TH1* hPOTNew = (TH1*)keyPOTNew->ReadObj();
      NormPOTCurrent = hPOTNew->Integral();
      std::cout << "Total POT accounted in current release " << NormPOTCurrent << std::endl;
    }
    else {
      NormPOTCurrent = 1.;
    }

    TFile* nextSource = (TFile*)sourceList->After(firstSource);
    nextSource->cd(inPath);
    TKey* keyPOTOld = (TKey*)gDirectory->GetListOfKeys()->FindObject(histPOT.c_str());
    if(keyPOTOld) {
      TH1* hPOTOld = (TH1*)keyPOTOld->ReadObj();
      NormPOTPrevious = hPOTOld->Integral();
      std::cout << "Total POT accounted in previous release " << NormPOTPrevious << std::endl;
    }
    else {
      NormPOTPrevious = 1.;
    }
  }
  else {
    std::cout << "POT info not available!" << std::endl;
    NormPOTPrevious = 1.;
    NormPOTCurrent = 1.;
  }

  return;
}

//.............................................................................
void CompareRootfile(TString inPath, TList* sourceList)
{
  TFile* firstSource = (TFile*)sourceList->First();
  firstSource->cd(inPath);
  TDirectory* currentSourceDir = gDirectory;
  TString saveName;

  // Loop over all keys in this directory and save all histos found as PNG images
  TIter nextkey(currentSourceDir->GetListOfKeys());
  TKey* key;
  TKey* oldkey = 0;

  while((key = (TKey*)nextkey())) {
    // Keep only the highest cycle number for each key
    if(oldkey && !strcmp(oldkey->GetName(), key->GetName())) { continue; }

    // Read object from first source file
    firstSource->cd(inPath);
    TObject* obj = key->ReadObj();

    if(obj->IsA()->InheritsFrom(TH1::Class())) { // Could be TH1, TH2, or TH3
      std::cout << "Compare and export histograms " << obj->GetName() << std::endl;
      TH1* hNew = (TH1*)obj;

      // Loop over all source files and compare histograms 
      TFile* nextSource = (TFile*)sourceList->After(firstSource);
      while(nextSource) {
        // Make sure we are at the correct directory level
        nextSource->cd(inPath);

        // Get the object of the same name in the second file
        TKey* key2 = (TKey*)gDirectory->GetListOfKeys()->FindObject(hNew->GetName());

        if(key2) {
          TH1* hOld = (TH1*)key2->ReadObj();

          TCanvas* c = new TCanvas("c", "c", 700, 500);
          saveName = subDir + "/";

          TImage* img = TImage::Create();

          // If TH1, normalize by POT, draw both histograms in the same canvas and get ratio
          if(!(obj->IsA()->InheritsFrom(TH2::Class()) ||
               obj->IsA()->InheritsFrom(TH3::Class()))) {
            if(NormPOTPrevious > 0. && !strstr(hNew->GetName(), "Subrun")) {
              hOld->Scale(NormPOTCurrent/NormPOTPrevious);
            }

            // Zoom to a reasonable area for CosZ and CosNu plots
            if(strstr(hNew->GetName(), "CosZ") || strstr(hNew->GetName(), "CosNu")) {
              hNew->GetXaxis()->SetRangeUser(0.5, 1.);
              hOld->GetXaxis()->SetRangeUser(0.5, 1.);
            }

            // Normalize neutrino energy histograms by POT
            if(strstr(hNew->GetName(), "Enu")) {
              char yAxisLabel[256];
              sprintf(yAxisLabel, "Events / %.2E POT", NormPOTCurrent);
              hNew->GetYaxis()->SetTitle(yAxisLabel);
              hOld->GetYaxis()->SetTitle(yAxisLabel);
            }

            // When drawing, make sure both plots will be totally visible
            if(hOld->GetMaximum() > hNew->GetMaximum()) {
              hNew->SetMaximum(1.1*hOld->GetMaximum());
            }

            hNew->Sumw2();
            hOld->Sumw2();

            hNew->Draw("HIST");
            hOld->SetLineColor(2);
            hOld->Draw("HIST same");
            
            img->FromPad(c);
            img->WriteImage(htmlpath + saveName + obj->GetName() + ".png");

            TH1* hRatio = (TH1*)hNew->Clone("hRatio");
            hRatio->Divide(hOld);
            hRatio->SetMinimum(0.);
            hRatio->SetMaximum(5.);
            hRatio->GetYaxis()->SetTitle("");
            hRatio->Draw();

            img->FromPad(c);
            img->WriteImage(htmlpath + saveName + obj->GetName() + "_Ratio.png");

            *histhtml << "<p>" << subDir << "-"<< obj->GetName() << std::endl;
            *histhtml << "<table>" << std::endl;
            *histhtml << "<tr>" << std::endl;
            *histhtml << "<td>" << std::endl;
            *histhtml << "<p><p><img src=\"" << saveName +  obj->GetName() << ".png\">" << std::endl;
            *histhtml << "</td>"<< std::endl;
            *histhtml << "<td>" << std::endl;
            *histhtml << "<p><p><img src=\"" << saveName +  obj->GetName() << "_Ratio.png\">" << std::endl;
            *histhtml << "</td>"<< std::endl;
            *histhtml << "</tr>" << std::endl;
            *histhtml << "</table>" << std::endl;
            *histhtml << "<a href=\"#pagetop\" >Back to top</a>\n" << std::endl;
          } // End of condition if TObject is a TH1 and not TH2 or TH3
          else { // If TH2 or TH3, don't overlay histograms
            if(obj->IsA()->InheritsFrom(TH2::Class())) {
              hNew->Draw("colz");
            }
            else {
              hNew->Draw();
            }

            img->FromPad(c);
            img->WriteImage(htmlpath + saveName + obj->GetName() + ".png");

            *histhtml << "<p>" << subDir << "-"<< obj->GetName() << std::endl;
            *histhtml << "<table>" << std::endl;
            *histhtml << "<tr>" << std::endl;
            *histhtml << "<td>" << std::endl;
            *histhtml << "<p><p><img src=\"" << saveName +  obj->GetName() << ".png\">" << std::endl;
            *histhtml << "</td>"<< std::endl;

            hOld->SetMarkerColor(2);
            if(obj->IsA()->InheritsFrom(TH2::Class())) {
              hOld->Draw("colz");
            }
            else {
              hOld->Draw();
            }

            img->FromPad(c);
            img->WriteImage(htmlpath + saveName + obj->GetName() + "Previous.png");

            *histhtml << "<td>" << std::endl;
            *histhtml << "<p><p><img src=\"" << saveName +  obj->GetName() << "Previous.png\">" << std::endl;
            *histhtml << "</td>"<< std::endl;
            *histhtml << "</tr>" << std::endl;
            *histhtml << "</table>" << std::endl;

            *histhtml << "<a href=\"#pagetop\" >Back to top</a>\n" << std::endl;
          } // end of conditional if object is TH1 and (TH2 or TH3)

          delete c;
          delete img;
        } // end of conditional if key was found in "old file" by the same name as that found in "new file"
        else {
          std::cout << "Could not find " << hNew->GetName() << " in previous release." << std::endl;
        }

        // This should terminate the while loop if there are only two files in the TList
        nextSource = (TFile*)sourceList->After(nextSource);
      } // end of while loop on valid nextSource
    } // end of conditonal if object is TH1
    else if(obj->IsA()->InheritsFrom(TDirectory::Class())) { // It's a subdirectory
      subDir = obj->GetName();
      std::cout << "Found subdirectory " << subDir << std::endl;

      // Make local directories for the images
      gSystem->Exec("mkdir " + htmlpath + subDir);

      *histhtml << "<a name=\"" << subDir << "\">\n"
                << "<h2>" << subDir << "</h2>\n" << std::endl;

      // Create a new subdirectory
      TString newDir = obj->GetName();

      CompareRootfile(newDir, sourceList);
    } // end of conditional if object is a TDirectory
    else { // Object is of no type that this macro can handle
      std::cout << "Unknown object type, name: " << obj->GetName()
                << " title: " << obj->GetTitle() << std::endl;
    }

    delete obj;
  } // end of while loop over TKeys

  return;
}

//.............................................................................
void Finalize()
{
  *histhtml << "</body></html>" << std::endl;
  histhtml->close();

  return;
}
