///////////////////////////////////////////////////////////////////////////////
/// \brief  version 1
///
/// \authors Dan Pershey - Raphael Schroeter
//           rschroet@physics.harvard.edu
/// \date    12/01/25
///////////////////////////////////////////////////////////////////////////////

// C/C++ Includes
#include <cmath>
#include <map>
#include <vector>

// ART includes
#include "art/Framework/Core/EDAnalyzer.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Principal/SubRun.h"
#include "art/Framework/Services/Optional/TFileDirectory.h"
#include "art/Framework/Services/Optional/TFileService.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "art/Persistency/Common/Ptr.h"
#include "art/Persistency/Common/PtrVector.h"
#include "fhiclcpp/ParameterSet.h"
#include "messagefacility/MessageLogger/MessageLogger.h"

// NOvA includes
#include "Geometry/Geometry.h"
#include "GeometryObjects/CellGeo.h"
#include "MCCheater/BackTracker.h"
#include "Simulation/Particle.h"
#include "SimulationBase/MCNeutrino.h"
#include "SimulationBase/MCTruth.h"
#include "Simulation/FLSHitList.h"
#include "Simulation/ParticleNavigator.h"
#include "Utilities/func/MathUtil.h"

// ROOT Includes
#include "TH1.h"
#include "TH2.h"
#include "TH3.h"
#include "TParticle.h"
#include "TSystem.h"

namespace mcchk {
  /// Module to check the daughter leptons coming from neutrino interactions
  class LeptonAna : public art::EDAnalyzer {
    
  public:
    explicit LeptonAna(fhicl::ParameterSet const& pset);
    virtual ~LeptonAna();

    void analyze(art::Event const& evt);
    void beginRun(art::Run const& run);

  private:
    bool fDetOnly;

    // Truth information
    std::string fMCTruthModuleLabel; ///< Label MC Truth information module 

    std::map<int, std::string> fLeptons; ///< Map of outgoing lepton PDG and names

    /// Directional cosines between outgoing lepton and axes
    std::map<int, TH1F*> fLeptonCosX;   ///< Cosine of angle between lepton momentum and x axis
    std::map<int, TH1F*> fLeptonCosY;   ///< Cosine of angle between lepton momentum and y axis
    std::map<int, TH1F*> fLeptonCosZ;   ///< Cosine of angle between lepton momentum and z axis
    std::map<int, TH1F*> fLeptonCosNu;  ///< Cosine of angle between lepton momentum and nu momentum

    std::map<int, TH1F*> fLeptonE;      ///< Energy spectrum of daughter leptons
    std::map<int, TH1F*> fLeptonP;      ///< Momentum of daughter leptons
    std::map<int, TH1F*> fLeptonTrackL; ///< Track length of daughter leptons

    std::map<int, TH1F*> fLeptonEDiff;  ///< Difference between lepton true and deposited energies
    std::map<int, TH1F*> fLeptondEdx;   ///< Average dE/dx for lepton along its track/shower

    /// Energy deposition distribution of outgoing charged lepton hits
    TH2F* fEDepXZ; ///< X vs Z view of energy deposited by charged lepton
    TH2F* fEDepYZ; ///< Y vs Z view of energy deposited by charged lepton

    /// End Point Information for Contained CC Leptons
    TH2F* fEndPointXZ; ///< X vs Z view of contained charged lepton track end points
    TH2F* fEndPointYZ; ///< Y vs Z view of contained charged lepton track end points
    TH2F* fEndPointXE; ///< True Energy vs X view of contained charged lepton track end points
    TH2F* fEndPointYE; ///< True Energy vs Y view of contained charged lepton track end points
    TH2F* fEndPointZE; ///< True Energy vs Z view of contained charged lepton track end points

    /// Escape Point Information for Uncontained CC Leptons
    TH2F* fEscapeXZ; ///< X vs Z view of uncontained charged lepton detector escape point
    TH2F* fEscapeYZ; ///< Y vs Z view of uncontained charged lepton detector escape point
    TH2F* fEscapeXE; ///< True Energy vs X view of uncontained charged lepton detector escape point
    TH2F* fEscapeYE; ///< True Energy vs Y view of uncontained charged lepton detector escape point
    TH2F* fEscapeZE; ///< True Energy vs Z view of uncontained charged lepton detector escape point
  };
};

///////////////////////////////////////////////////////////////////////////////
namespace mcchk {
  static int msg1cnt = 0;

  //......................................................................
  LeptonAna::LeptonAna(fhicl::ParameterSet const& pset)
    : EDAnalyzer(pset),
      fDetOnly(pset.get<bool>("DetOnly")),
      fMCTruthModuleLabel(pset.get<std::string>("MCTruthModuleLabel")),
      fEDepXZ(0) // Flag to beginRun that histograms are uninitialized
  {
  }

  //......................................................................
  LeptonAna::~LeptonAna()
  {
  }

  //......................................................................
  void LeptonAna::beginRun(art::Run const& )
  {
    if(fEDepXZ) { return; }

    art::ServiceHandle<art::TFileService> tfs;
    art::ServiceHandle<geo::Geometry> geo;

    double x = 2.1*geo->DetHalfWidth();
    double y = 2.1*geo->DetHalfHeight();
    double z = 1.1*geo->DetLength();
    int xdiv = TMath::Nint(2*x/5.);
    int ydiv = TMath::Nint(2*y/5.);
    int zdiv = TMath::Nint(2*z/5.);

    fLeptons[11] = "Electron";
    fLeptons[13] = "Muon";

    char histoName[200];
    char histoTitle[200];
    char titleHelper[200];

    for(std::map<int, std::string>::iterator it = fLeptons.begin(); it != fLeptons.end(); ++it) {
      int pdg  = it->first;
      char name[it->second.length()];
      sprintf(name, "%s", it->second.c_str());

      // Make directional cosine histograms
      sprintf(titleHelper, "Cosine of Angle Between Outgoing");

      sprintf(histoName,  "f%sCosX", name);
      sprintf(histoTitle, "%s %s Momentum and X Axis;cos#theta_{x};Events", titleHelper, name);
      fLeptonCosX[pdg]   = tfs->make<TH1F>(histoName, histoTitle, 200, -1., 1.);

      sprintf(histoName,  "f%sCosY", name);
      sprintf(histoTitle, "%s %s Momentum and Y Axis;cos#theta_{y};Events", titleHelper, name);
      fLeptonCosY[pdg]   = tfs->make<TH1F>(histoName, histoTitle, 200, -1., 1.);

      sprintf(histoName,  "f%sCosZ", name);
      sprintf(histoTitle, "%s %s Momentum and Z Axis;cos#theta_{z};Events", titleHelper, name);
      fLeptonCosZ[pdg]   = tfs->make<TH1F>(histoName, histoTitle, 200, -1., 1.);

      sprintf(histoName,  "f%sCosNu", name);
      sprintf(histoTitle, "%s %s and Incoming Neutrino Momenta;cos#theta;Events", titleHelper, name);
      fLeptonCosNu[pdg]  = tfs->make<TH1F>(histoName, histoTitle, 200, -1., 1.);

      // Make energy, momentum, track length histograms
      sprintf(histoName,  "f%sE", name);
      sprintf(histoTitle, "True Energy of Outgoing %s;E (GeV);Events", name);
      fLeptonE[pdg]      = tfs->make<TH1F>(histoName, histoTitle, 200,  0., 10.);

      sprintf(histoName,  "f%sMomentum", name);
      sprintf(histoTitle, "Momentum of Outgoing %s;p (GeV);Events", name);
      fLeptonP[pdg]      = tfs->make<TH1F>(histoName, histoTitle, 500,  0., 50.);

      sprintf(histoName,  "f%sTrackLength", name);
      sprintf(histoTitle, "Track Length of Outgoing %s;L (cm);Events", name);
      fLeptonTrackL[pdg] = tfs->make<TH1F>(histoName, histoTitle, 1000, 0., 2000.);

      // Make histogram of difference between true and depostied energies
      sprintf(histoName,  "f%sEDiff", name);
      sprintf(histoTitle, "True - Deposited Energy of Outgoing %s;#Delta E (GeV);Events", name);
      fLeptonEDiff[pdg]  = tfs->make<TH1F>(histoName, histoTitle, 200,  0., 1.);

      // Make histogram for average energy loss of outgoing lepton
      sprintf(histoName,  "f%sdEdx", name);
      sprintf(histoTitle, "Average Energy Loss of Outgoing %s;dE/dx (MeV/cm);Events", name);
      fLeptondEdx[pdg]   = tfs->make<TH1F>(histoName, histoTitle, 400,  0., 10.);
    }

    // Make extra histograms for some of the maps

    // Directional cosines for all outgoing leptons
    sprintf(titleHelper, "Cosine of Angle Between Outgoing Lepton");

    sprintf(histoName,  "fLeptonCosX");
    sprintf(histoTitle, "%s Momentum and X Axis;cos#theta_{x};Events", titleHelper);
    fLeptonCosX[0]  = tfs->make<TH1F>(histoName, histoTitle, 200, -1., 1.);

    sprintf(histoName,  "fLeptonCosY");
    sprintf(histoTitle, "%s Momentum and Y Axis;cos#theta_{y};Events", titleHelper);
    fLeptonCosY[0]  = tfs->make<TH1F>(histoName, histoTitle, 200, -1., 1.);

    sprintf(histoName,  "fLeptonCosZ");
    sprintf(histoTitle, "%s Momentum and Z Axis;cos#theta_{z};Events", titleHelper);
    fLeptonCosZ[0]  = tfs->make<TH1F>(histoName, histoTitle, 200, -1., 1.);

    sprintf(histoName,  "fLeptonCosNu");
    sprintf(histoTitle, "%s and Incoming Neutrino Momenta;cos#theta;Events", titleHelper);
    fLeptonCosNu[0] = tfs->make<TH1F>(histoName, histoTitle, 200, -1., 1.);

    // Energy spectrum for NC events
    sprintf(titleHelper, "True Energy of Outgoing");

    sprintf(histoName,  "fnueE");
    sprintf(histoTitle, "%s #nu_{e};E (GeV);Events", titleHelper);
    fLeptonE[12] = tfs->make<TH1F>(histoName, histoTitle, 200,  0., 10.);

    sprintf(histoName,  "fnumuE");
    sprintf(histoTitle, "%s #nu_{#mu};E (GeV);Events", titleHelper);
    fLeptonE[14] = tfs->make<TH1F>(histoName, histoTitle, 200,  0., 10.);

    // Histogram of difference between true and deposited energies of all (outgoing) particles
    sprintf(histoName,  "fTotalEDiff");
    sprintf(histoTitle, "True - Deposited Energy of All Outgoing Particles;#Delta E (GeV);Events");
    fLeptonEDiff[0] = tfs->make<TH1F>(histoName, histoTitle, 200,  0., 1.);

    // Make histograms of outgoing lepton deposited energy distributions
    sprintf(titleHelper, "Energy Deposited by Outgoing Lepton");

    sprintf(histoName,  "fEDepXZ");
    sprintf(histoTitle, "%s;z (cm);x (cm)", titleHelper);
    fEDepXZ = tfs->make<TH2F>(histoName, histoTitle, zdiv, -0.02*z, z, xdiv, -x, x);

    sprintf(histoName,  "fEDepYZ");
    sprintf(histoTitle, "%s;z (cm);y (cm)", titleHelper);
    fEDepYZ = tfs->make<TH2F>(histoName, histoTitle, zdiv, -0.02*z, z, ydiv, -y, y);

    // Make histograms of endpoint of contained outgoing charged leptons
    sprintf(titleHelper, "View of Contained Charged Lepton Track End Points");

    sprintf(histoName,  "fEndPointXZ");
    sprintf(histoTitle, "X vs Z %s;z (cm);x (cm)", titleHelper);
    fEndPointXZ = tfs->make<TH2F>(histoName, histoTitle, zdiv, -0.02*z, z, xdiv, -x, x);

    sprintf(histoName,  "fEndPointYZ");
    sprintf(histoTitle, "Y vs Z %s;z (cm);y (cm)", titleHelper);
    fEndPointYZ = tfs->make<TH2F>(histoName, histoTitle, zdiv, -0.02*z, z, ydiv, -y, y);

    sprintf(histoName,  "fEndPointXE");
    sprintf(histoTitle, "True Lepton Energy vs X %s;x (cm);E (GeV)", titleHelper);
    fEndPointXE = tfs->make<TH2F>(histoName, histoTitle, xdiv, -x, x,200,0,10);

    sprintf(histoName,  "fEndPointYE");
    sprintf(histoTitle, "True Lepton Energy vs Y %s;x (cm);E (GeV)", titleHelper);
    fEndPointYE = tfs->make<TH2F>(histoName, histoTitle, ydiv, -y, y,200,0,10);

    sprintf(histoName,  "fEndPointZE");
    sprintf(histoTitle, "True Lepton Energy vs Z %s;x (cm);E (GeV)", titleHelper);
    fEndPointZE = tfs->make<TH2F>(histoName, histoTitle, zdiv, -0.02*z, z,200,0,10);

    // Make histograms of escape point of uncontained outgoing charged leptons
    sprintf(titleHelper, "View of Uncontained Charged Lepton Track Escape Points");

    sprintf(histoName,  "fEscapeXZ");
    sprintf(histoTitle, "X vs Z %s;z (cm);x (cm)", titleHelper);
    fEscapeXZ = tfs->make<TH2F>(histoName, histoTitle, zdiv, -0.02*z, z, xdiv, -x, x);

    sprintf(histoName,  "fEscapeYZ");
    sprintf(histoTitle, "Y vs Z %s;z (cm);y (cm)", titleHelper);
    fEscapeYZ = tfs->make<TH2F>(histoName, histoTitle, zdiv, -0.02*z, z, ydiv, -y, y);

    sprintf(histoName,  "fEscapeXE");
    sprintf(histoTitle, "True Lepton Energy vs X %s;x (cm);E (GeV)", titleHelper);
    fEscapeXE = tfs->make<TH2F>(histoName, histoTitle, xdiv, -x, x, 200, 0., 10.);

    sprintf(histoName,  "fEscapeYE");
    sprintf(histoTitle, "True Lepton Energy vs Y %s;x (cm);E (GeV)", titleHelper);
    fEscapeYE = tfs->make<TH2F>(histoName, histoTitle, ydiv, -y, y, 200, 0., 10.);

    sprintf(histoName,  "fEscapeZE");
    sprintf(histoTitle, "True Lepton Energy vs Z %s;x (cm);E (GeV)", titleHelper);
    fEscapeZE = tfs->make<TH2F>(histoName, histoTitle, zdiv, -0.02*z, z, 200, 0., 10.);
  }

  //......................................................................
  void LeptonAna::analyze(art::Event const& evt)
  {
    art::ServiceHandle<geo::Geometry> geo;

    double detHalfWidth  = geo->DetHalfWidth();
    double detHalfHeight = geo->DetHalfHeight();
    double detLength = geo->DetLength();

    // Pull the MC generator information out of the event
    bool isMCTruthListEmpty = false;
    art::Handle<std::vector<simb::MCTruth> > mclist;
    try {
      evt.getByLabel(fMCTruthModuleLabel, mclist);
      if(mclist->empty()) { isMCTruthListEmpty = true; }
    }
    catch(...) { isMCTruthListEmpty = true; }
    if(isMCTruthListEmpty) {
      mf::LogError("LeptonAna MCTruth Error") << "Error retrieving MCTruth list." << std::endl;
      return;
    }

    std::vector<art::Ptr<simb::MCTruth> > mcptrs;
    art::fill_ptr_vector(mcptrs, mclist);

    // Loop over each MCTruth object
    // mc should be an art::ptr<simb::MCTruth>
    for(const auto &mc : mcptrs) {
      // If there is no MCNeutrino, skip the event
      if(!mc->NeutrinoSet()) {
        if(msg1cnt < 5) {
          mf::LogError("LeptonAna MCNeutrino") << "There is no MCNeutrino in this event." << std::endl;
          ++msg1cnt;
        }

        return;
      }

      // Get the MCNeutrino object, incoming neutrino, and outgoing lepton
      const simb::MCNeutrino& mc_neutrino = mc->GetNeutrino();
      //const simb::MCParticle& nu  = mc_neutrino.Nu();
      const simb::MCParticle& lep = mc_neutrino.Lepton();

      // Skip non detector event if FCL says to skip them
      if(fDetOnly) {
        const simb::MCParticle& nu = mc_neutrino.Nu();
        if(std::abs(nu.Vx()) > detHalfWidth ||
           std::abs(nu.Vy()) > detHalfHeight ||
           nu.Vz() < 0. || nu.Vz() > detLength) {
          continue;
        }
      }

      // This will serve as the map key for many histograms
      int lepPDG = std::abs(lep.PdgCode());

      // Fill directional cosine histograms for:
      // All leptons
      fLeptonCosX[0] ->Fill(lep.Px()/lep.P());
      fLeptonCosY[0] ->Fill(lep.Py()/lep.P());
      fLeptonCosZ[0] ->Fill(lep.Pz()/lep.P());
      fLeptonCosNu[0]->Fill(std::cos(mc_neutrino.Theta()));
      // Specific leptons, i.e., e+/-, mu+/-, (anti-)numu, (anti-)nue
      if(fLeptonCosX.find(lepPDG) != fLeptonCosX.end()) {
        fLeptonCosX[lepPDG] ->Fill(lep.Px()/lep.P());
        fLeptonCosY[lepPDG] ->Fill(lep.Py()/lep.P());
        fLeptonCosZ[lepPDG] ->Fill(lep.Pz()/lep.P());
        fLeptonCosNu[lepPDG]->Fill(std::cos(mc_neutrino.Theta()));
      }

      // Fill energy histogram
      if(fLeptonE.find(lepPDG) != fLeptonE.end()) {
        fLeptonE[lepPDG]->Fill(lep.E());
      }

      // Fill momentum histogram
      if(fLeptonP.find(lepPDG) != fLeptonP.end()) {
        fLeptonP[lepPDG]->Fill(lep.P());
      }

      // BackTracker will match the list of particles to the MCTruth
      // Later, it will match FLSHits to the particles
      art::ServiceHandle<cheat::BackTracker> bt;

      // Create a ParticleNavigator for specialized looping
      const sim::ParticleNavigator& pnav = bt->ParticleNavigator();

      // Create and initialize a Particle pointer that will contain the outgoing lepton
      // This is similar to the MCParticle lep, but from a different point in the simulation
      const sim::Particle* primaryLep = pnav.Primary(0);
      int leptonTrackID = -1; // Track ID of the primary outgoing lepton

      // Loop over event primary particles
      for(int i_primary = 0, n_primary = pnav.NumberOfPrimaries(); i_primary < n_primary; ++i_primary) {
        // Save the primary matching the outgoing lepton from MCTruth

        if(std::abs(pnav.Primary(i_primary)->PdgCode()) == lepPDG) {
          primaryLep = pnav.Primary(i_primary);
          leptonTrackID = primaryLep->TrackId(); // Also save the track ID
          break;
        }
      }

      // If there are no particles/primaries according to ParticleNavigator, quit now
      if(leptonTrackID == -1) {
        if(msg1cnt < 5) {
          mf::LogWarning("LeptonAna ParticleNavigator")
            << "ParticleNavigator did not match MCTruth outgoing lepton." << std::endl
            << "Number of primaries: " << pnav.NumberOfPrimaries() << std::endl;
          ++msg1cnt;
        }

        return;
      }

      // Get the list of particles associated with the MCTruth object
      std::vector<const sim::Particle*> particles = bt->MCTruthToParticles(mc);

      float trueETotal = 0.;
      float depositedE = 0.;
      float depositedETotal = 0.;

      // Loop over all particles in the event
      for(const auto& particle : particles) {
        if(particle->Mother() >= 0) { // Only these particles are initialized
          // Sum the energies of all primaries
          if(pnav.IsPrimary(particle->TrackId())) {
            trueETotal += particle->E();
          }

          double xyz[3];
          double xyzworld[3];

          // Get the list of FLSHits associated with the current Particle
          const std::vector<sim::FLSHit>& hits = bt->ParticleToFLSHit(particle->TrackId());

          // Loop over all FLSHits
          for(const auto& hit : hits) {
            depositedETotal += hit.GetEdep(); // Sum the total energy deposited

            // Daughter particles may have a Geant4 Track ID equal to -leptonTrackID
            if(std::abs(hit.GetTrackID()) == leptonTrackID) {
              depositedE += hit.GetEdep(); // Sum the energy deposited by the lepton

              // Get the location of the FLSHit
              xyz[0] = 0.5*(hit.GetEntryX() + hit.GetExitX());
              xyz[1] = 0.5*(hit.GetEntryY() + hit.GetExitY());
              xyz[2] = 0.5*(hit.GetEntryZ() + hit.GetExitZ());
              geo->Plane(hit.GetPlaneID())->Cell(hit.GetCellID())->LocalToWorld(xyz,xyzworld);

              // Fill the FLSHit distribution histograms
              fEDepXZ->Fill(xyzworld[2], xyzworld[0], hit.GetEdep());
              fEDepYZ->Fill(xyzworld[2], xyzworld[1], hit.GetEdep());
            } // end of condition if current FLSHit track ID is the same as the primary lepton
          } // end of loop over FLSHits
        } // end of condition if current particle mother has a track ID greater than or equal to 0
      } // end of loop over particles

      // Find the last point in the track of the charged outgoing lepton in the detector
      float finalTrackPointX = 0.;
      float finalTrackPointY = 0.;
      float finalTrackPointZ = 0.;
      float trackLength = 0.;

      // If lepton end point is inside detector, store the end point for later and fill relevant histograms
      if(std::abs(primaryLep->EndX()) < detHalfWidth &&
         std::abs(primaryLep->EndY()) < detHalfHeight &&
         primaryLep->EndZ() < detLength && primaryLep->EndZ() > 0) {
        // Store the end of the lepton track
        finalTrackPointX = primaryLep->EndX();
        finalTrackPointY = primaryLep->EndY();
        finalTrackPointZ = primaryLep->EndZ();

        // Fill end point histograms
        fEndPointXZ->Fill(finalTrackPointZ, finalTrackPointX);
        fEndPointYZ->Fill(finalTrackPointZ, finalTrackPointY);
        // Fill end point vs lepton energy histograms
        fEndPointXE->Fill(finalTrackPointX, primaryLep->E());
        fEndPointYE->Fill(finalTrackPointY, primaryLep->E());
        fEndPointZE->Fill(finalTrackPointZ, primaryLep->E());
      }
      else {
        // Otherwise, find the last trajectory point inside the detector and fill escape point histograms
        for(unsigned int i_trajpt = 0, n_trajpt = primaryLep->NumberTrajectoryPoints();
            i_trajpt < n_trajpt; ++i_trajpt) {
          if(std::abs(primaryLep->Vx(i_trajpt)) < detHalfWidth &&
             std::abs(primaryLep->Vy(i_trajpt)) < detHalfHeight &&
             primaryLep->Vz(i_trajpt) < detLength &&
             primaryLep->Vz(i_trajpt) > 0.) {
            // Store the last trajectory point inside the detector
            finalTrackPointX = primaryLep->Vx(i_trajpt);
            finalTrackPointY = primaryLep->Vy(i_trajpt);
            finalTrackPointZ = primaryLep->Vz(i_trajpt);
          }
          else {
            break;
          }
        }

        fEscapeXZ->Fill(finalTrackPointZ, finalTrackPointX);
        fEscapeYZ->Fill(finalTrackPointZ, finalTrackPointY);
        fEscapeXE->Fill(finalTrackPointX, primaryLep->E());
        fEscapeYE->Fill(finalTrackPointY, primaryLep->E());
        fEscapeZE->Fill(finalTrackPointZ, primaryLep->E());
      }

      // Calculate the track length of the outgoing lepton
      trackLength = util::pythag(primaryLep->Vx()-finalTrackPointX,
                                 primaryLep->Vy()-finalTrackPointY,
                                 primaryLep->Vz()-finalTrackPointZ);

      // Fill difference between true and deposited energies for all particles
      fLeptonEDiff[0]->Fill(trueETotal - depositedETotal);

      if(fLeptons.find(lepPDG) != fLeptons.end()) {
        // Fill difference between true and deposited lepton energies
        fLeptonEDiff[lepPDG]->Fill(primaryLep->E() - depositedE);

        if(trackLength > 0.) {
          fLeptonTrackL[lepPDG]->Fill(trackLength); // Fill track length

          // Calculate and fill average dE/dx
          fLeptondEdx[lepPDG]->Fill(depositedE/trackLength);
        } // end of condition if track length is greater than 0
      } // end of condition that outgoing lepton is part of lepton map
    } // end of loop over MCTruth objects
  } // end of LeptonAna::analyze
} // end of namespace mcchk

///////////////////////////////////////////////////////////////////////////////
namespace mcchk {
  DEFINE_ART_MODULE(LeptonAna);
}
