///////////////////////////////////////////////////////////////////////////////
/// \brief
/// \author 
///////////////////////////////////////////////////////////////////////////////

// C/C++ includes
#include <cmath>
#include <vector>

// ART includes
#include "art/Framework/Core/EDAnalyzer.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Principal/SubRun.h"
#include "art/Framework/Services/Optional/TFileDirectory.h"
#include "art/Framework/Services/Optional/TFileService.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "fhiclcpp/ParameterSet.h"
#include "messagefacility/MessageLogger/MessageLogger.h"

// NOvA includes
#include "SummaryData/POTSum.h"

// ROOT includes
#include "TH1.h"

namespace mcchk {
  /// A module to check the results from the Monte Carlo generator
  class POTAna : public art::EDAnalyzer {

  public:
    explicit POTAna(fhicl::ParameterSet const& pset);
    virtual ~POTAna();

    void analyze(art::Event const& evt);
    void beginJob();
    void beginSubRun(art::SubRun const& sr);
    void endJob();
    void endSubRun(art::SubRun const& sr);

  private:
    std::string fPotLabel; ///< Module that produced the POTSum object

    bool fIsPOTSumObjectForSubRun = false; ///< Flag whether the SubRun has a POTSum object

    int fNeventsInSubrun; ///< Number of events in the subrun
    int fNevents;         ///< Total number of events

    std::vector<double> fPOTSpillSubrun;   ///< POT per spill per subrun
    std::vector<double> fPOTSubrun;        ///< POT per subrun
    std::vector<double> fNeventsPOTSubrun; ///< Events per POT per subrun
    std::vector<double> fNeventsSubrun;    ///< Total number of events per subrun

    sumdata::POTSum* fPOTSum;
  };
};

///////////////////////////////////////////////////////////////////////////////
namespace mcchk {
  //......................................................................
  POTAna::POTAna(fhicl::ParameterSet const& pset)
    : EDAnalyzer(pset),
      fPotLabel(pset.get<std::string>("PotLabel")),
      fIsPOTSumObjectForSubRun(false),
      fNeventsInSubrun(0),
      fNevents(0)
  {
  }

  //......................................................................
  POTAna::~POTAna()
  {
  }

  //......................................................................
  void POTAna::beginJob()
  {
    fPOTSum = new sumdata::POTSum();
  }

  //......................................................................
  void POTAna::analyze(art::Event const& evt)
  {
    ++fNeventsInSubrun;
    ++fNevents;
  }

  //......................................................................
  void POTAna::beginSubRun(art::SubRun const& sr)
  {
    // Reset variables for the next subrun
    fNeventsInSubrun = 0;
    fIsPOTSumObjectForSubRun = false;

    art::Handle<sumdata::POTSum> p;

    // Try to get the POTSum information from the SubRun
    try {
      sr.getByLabel(fPotLabel, p);

      // If successful, set this flag to true
      fIsPOTSumObjectForSubRun = true;

      // Pull some information about the SubRun, store it locally
      fPOTSum->totpot     = p->totpot;
      fPOTSum->totgoodpot = p->totgoodpot;
      fPOTSum->totspills  = p->totspills;
      fPOTSum->goodspills = p->goodspills;

      mf::LogInfo("POTAna") << "fPOTSum This sub run has "             << fPOTSum->totspills
                            << " total spills with an exposure of "    << fPOTSum->totpot     << " POT " << std::endl
                            << "with cuts on beam quality, there are " << fPOTSum->goodspills
                            << " good spills with an exposure of "     << fPOTSum->totgoodpot << std::endl;
    }
    catch(...) {
      mf::LogWarning("POTAna POTSum") << "Error trying to access POTSum information." << std::endl;
      fIsPOTSumObjectForSubRun = false; // If there was a problem, make absolutely sure this flag is false!
    }
  }

  //......................................................................
  void POTAna::endSubRun(art::SubRun const& sr)
  {
    mf::LogInfo("POTAna") << "Event Number "  << fNeventsInSubrun << std::endl;

    if(fIsPOTSumObjectForSubRun) {
      mf::LogInfo("POTAna") << "Subrun Number " << sr.subRun() << std::endl;
      mf::LogInfo("POTAna") << "Event per POT " << fNeventsInSubrun/fPOTSum->totpot << std::endl;

      fPOTSpillSubrun  .push_back(fPOTSum->totpot/fPOTSum->totspills);
      fPOTSubrun       .push_back(fPOTSum->totpot);
      fNeventsPOTSubrun.push_back(fNeventsInSubrun/fPOTSum->totpot);
    }

    // This plot does not depend on the existence of a POTSum object
    fNeventsSubrun.push_back(fNeventsInSubrun);
  }

  //......................................................................
  void POTAna::endJob()
  {
    art::ServiceHandle<art::TFileService> tfs;

    // Helpful for spacing/readability
    // Also helpful in case histograms are made in loops
    char HistoName[200];
    char HistoTitle[200];

    TH1F* hPOTSpillSubrun;
    TH1F* hPOTSubrun;
    TH1F* hNEventsPOTSubrun;
    TH1F* hNEventsSubrun;
    TH1F* hNEvents;

    unsigned int size_POTSpillSubrun    = fPOTSpillSubrun.size();
    unsigned int size_POTSubrun         = fPOTSubrun.size();
    unsigned int size_NEventsPOTSubrun  = fNeventsPOTSubrun.size();
    unsigned int size_NEventsSubrun     = fNeventsSubrun.size();

    sprintf(HistoName,  "POTSpillSubrun");
    sprintf(HistoTitle, "POT per Spill per SubRun;Subrun number;POT per Spill");
    hPOTSpillSubrun   = tfs->make<TH1F>(HistoName, HistoTitle,
                                        size_POTSpillSubrun + 1,   -0.5, size_POTSpillSubrun + 0.5);

    sprintf(HistoName,  "POTSubrun");
    sprintf(HistoTitle, "POT per SubRun;Subrun number;POT");
    hPOTSubrun        = tfs->make<TH1F>(HistoName, HistoTitle,
                                        size_POTSubrun + 1,        -0.5, size_POTSubrun + 0.5);

    sprintf(HistoName,  "NEventsPOTSubrun");
    sprintf(HistoTitle, "Events per POT per Subrun;Subrun number;Events per POT");
    hNEventsPOTSubrun = tfs->make<TH1F>(HistoName, HistoTitle,
                                        size_NEventsPOTSubrun + 1, -0.5, size_NEventsPOTSubrun + 0.5);

    sprintf(HistoName,  "NEventsSubrun");
    sprintf(HistoTitle, "Events per Subrun;Subrun number;Events");
    hNEventsSubrun     = tfs->make<TH1F>(HistoName, HistoTitle,
                                         size_NEventsSubrun + 1,   -0.5, size_NEventsSubrun + 0.5);

    sprintf(HistoName,  "NEvents");
    sprintf(HistoTitle, "Total Number of Events;;Events");
    hNEvents           = tfs->make<TH1F>(HistoName, HistoTitle, 1, -0.5, 0.5);
    hNEvents->SetBinContent(1, fNevents);

    for(unsigned int i = 0; i < size_POTSpillSubrun; ++i) {
      hPOTSpillSubrun->Fill(i, fPOTSpillSubrun[i]);
    }
    for(unsigned int i = 0; i < size_POTSubrun; ++i) {
      hPOTSubrun->Fill(i, fPOTSubrun[i]);
    }
    for(unsigned int i = 0; i < size_NEventsPOTSubrun; ++i) {
      hNEventsPOTSubrun->Fill(i, fNeventsPOTSubrun[i]);
    }
    for(unsigned int i = 0; i < size_NEventsSubrun; ++i) {
      hNEventsSubrun->Fill(i, fNeventsSubrun[i]);
    }
  } 
} // end of namespace mcchk

////////////////////////////////////////////////////////////////////////
namespace mcchk {
  DEFINE_ART_MODULE(POTAna);
}
