////////////////////////////////////////////////////////////////////////
// \file    RockAna.cxx
// \brief   Analyse monopole monte carlo events
// \version $Id: RockAna.cxx,v 1.2 2012-10-09 03:56:03 ricken Exp $
// \author  $Author: ricken $
////////////////////////////////////////////////////////////////////////
#include "Geometry/Geometry.h"
#include "GeometryObjects/CellGeo.h"
#include "MCCheater/BackTracker.h"
#include "RecoBase/CellHit.h"
#include "RecoBase/Cluster.h"
#include "SimulationBase/MCTruth.h"
#include "SimulationBase/MCNeutrino.h"
#include "Utilities/func/MathUtil.h"


#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "art/Framework/Services/Optional/TFileService.h"
#include "art/Framework/Services/Optional/TFileDirectory.h"
#include "fhiclcpp/ParameterSet.h"
#include "messagefacility/MessageLogger/MessageLogger.h"
#include "art/Framework/Core/EDAnalyzer.h"
#include "art/Persistency/Common/PtrVector.h"
#include "art/Persistency/Common/Ptr.h"
#include "art/Framework/Core/ModuleMacros.h"

#include "TLorentzVector.h"
#include "TTree.h"
#include "TH1F.h"

#include <iostream>
#include <string>

class TNtuple;
class TTree;

namespace mcchk {
  class RockAna : public art::EDAnalyzer
    {
    public:
      explicit RockAna(fhicl::ParameterSet const &pset);
      virtual ~RockAna();
                        
      void analyze(const art::Event &evt);
      void beginJob();
      void reconfigure(fhicl::ParameterSet const &pset);
      
    private:
    
      std::string fClusterInput;

      TTree*    Rock_MC;
      TTree*    Rock_Cheater;

      int nu_pdg;
      float vx,vy,vz,px,py,pz,Enu;

      float mcPE,cx,cy,cz,ct;
      int   cview; 


    };


};


namespace mcchk{

  RockAna::RockAna(fhicl::ParameterSet const& pset) 
    : EDAnalyzer(pset)
  {
    reconfigure(pset);
  }

  void RockAna::reconfigure(fhicl::ParameterSet const &pset)

  {
    fClusterInput    = pset.get< std::string >("ClusterInput") ;
  }



//......................................................................

  RockAna::~RockAna(){}
  

//......................................................................

  void RockAna::beginJob()
  {

    art::ServiceHandle<art::TFileService> tfs;

    Rock_MC =  tfs->make<TTree>("Rock_MC","Neutrino Info");
    Rock_MC -> Branch("pdg",&nu_pdg,"pdg/I");
    Rock_MC -> Branch("vx",&vx,"vx/F");
    Rock_MC -> Branch("vy",&vy,"vy/F");
    Rock_MC -> Branch("vz",&vz,"vz/F");
    Rock_MC -> Branch("px",&px,"px/F");
    Rock_MC -> Branch("py",&py,"py/F");
    Rock_MC -> Branch("pz",&pz,"pz/F");
    Rock_MC -> Branch("E", &Enu,"E/F");

    Rock_Cheater =  tfs->make<TTree>("Rock_Cheater","all hits");
    Rock_Cheater -> Branch("PE",&mcPE,"PE/F");
    Rock_Cheater -> Branch("x",&cx,"x/F");
    Rock_Cheater -> Branch("y",&cy,"y/F");
    Rock_Cheater -> Branch("z",&cz,"z/F");
    Rock_Cheater -> Branch("t",&ct,"t/F");
    Rock_Cheater -> Branch("view",&cview,"view/I");

  }

  void RockAna::analyze(const art::Event& evt)
  {
 
    art::ServiceHandle<geo::Geometry> geom;


    //Now, get the vertex position and energy of neutrino
    art::Handle<std::vector<simb::MCTruth> > mcTruths;
    try{ evt.getByLabel("generator",mcTruths); }
    catch(...){ std::cout<<"couldn't get the mcTruths"<<std::endl; return;}

    for (unsigned idx =0; idx!=mcTruths->size();++idx) {
      art::Ptr<simb::MCTruth> mc (mcTruths,idx);

      const simb::MCNeutrino& mc_neutrino = mc->GetNeutrino();
      const simb::MCParticle& nu = mc_neutrino.Nu();
      nu_pdg = nu.PdgCode();
      vx = nu.Vx();
      vy = nu.Vy();
      vz = nu.Vz();
      px = nu.Px();
      py = nu.Py();
      pz = nu.Pz();
      Enu = nu.P();

      Rock_MC->Fill();
    }

   //make a backtracker
    art::ServiceHandle<cheat::BackTracker> bt;
    //    const sim::ParticleNavigator& pnav = bt->ParticleNavigator();


    // now we collect all hits from the slice
    art::Handle< std::vector<rb::Cluster> > slices;
    try{ evt.getByLabel(fClusterInput, slices);}
    catch(...){ std::cout<<"couldn't get slices"<<std::endl; return;}
    const int sliceMax = slices->size();

    for(int sliceIdx = 0; sliceIdx < sliceMax; ++sliceIdx){
      const rb::Cluster& slice = (*slices)[sliceIdx];
      const int hitMax = slice.NCell();

      for (int hitIdx=0; hitIdx<hitMax; ++hitIdx)
	{
	  const art::Ptr<rb::CellHit> chit = slice.Cell(hitIdx);

	  unsigned short plane = chit->Plane();
	  unsigned short cell  = chit->Cell();
	  if (chit->View() == geo::kX) cview = 0;
	  else cview =1; 
	  try {
	    if (!(bt->IsNoise(chit))){
	      //Get the center position of each hit
	      double xyz[3];
	      geom->Plane(plane)->Cell(cell)->GetCenter(xyz);

	      cx = xyz[0];
	      cy = xyz[1];
	      cz = xyz[2];

	      mcPE=0;
	      ct =0;           
	      for (unsigned int i=0; i!=(bt->HitToPhotonSignal(chit)).size();++i){
		mcPE += (bt->HitToPhotonSignal(chit))[i].NPhoton();
		ct += (bt->HitToPhotonSignal(chit))[i].TimeMean();
	      }
	      ct = ct/((bt->HitToPhotonSignal(chit)).size()+0.0);


	      Rock_Cheater->Fill();
              
	    }	
	    

	  }

	
	  catch (...) {
	    continue;
	  }


	}


    }



   
    return;
  }

  DEFINE_ART_MODULE(RockAna)

}
