////////////////////////////////////////////////////////////////////////
/// \version 1
/// \brief
/// \author  Dan Pershey - pershdawg@gmail.com
////////////////////////////////////////////////////////////////////////
#include <cmath>
#include <iostream>
#include <string>
#include <vector>

#include "TH1.h"

#include "Geometry/Geometry.h"
#include "GeometryObjects/CellGeo.h"
#include "MCCheater/BackTracker.h"
#include "Simulation/FLSHit.h"
#include "Simulation/FLSHitList.h"
#include "Simulation/Particle.h"
#include "Simulation/ParticleNavigator.h"
#include "Utilities/func/MathUtil.h"

#include "art/Framework/Core/EDAnalyzer.h"
// Framework includes
#include "art/Framework/Core/EDAnalyzer.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h" 
#include "fhiclcpp/ParameterSet.h" 
#include "art/Framework/Principal/Handle.h" 
#include "art/Persistency/Common/Ptr.h" 
#include "art/Persistency/Common/PtrVector.h" 
#include "art/Framework/Services/Registry/ServiceHandle.h" 
#include "art/Framework/Services/Optional/TFileService.h" 
#include "art/Framework/Services/Optional/TFileDirectory.h" 
#include "messagefacility/MessageLogger/MessageLogger.h" 


namespace mcchk {
  /// check the results from the Monte Carlo generator
  class ShowerAnaCheck : public art::EDAnalyzer {
    
  public:
    explicit ShowerAnaCheck(fhicl::ParameterSet const& pset);
    virtual ~ShowerAnaCheck();

    void analyze(const art::Event& evt);
    void reconfigure(const fhicl::ParameterSet& pset);
    void beginJob();
    void EMShowerWidth(float& Event_distance, int MomsID, const sim::Particle* part, const std::vector<sim::FLSHit>& hits);
    void EMMoliereRadius(float& Moliere_radius, int MomsID, const sim::Particle* part, const std::vector<sim::FLSHit>& hits);
    float PointLineDistance(float x, float y, float z, float x1, float y1, float z1, float x2, float y2, float z2);

  private:

    std::string        fFLSHitListLabel;
    std::vector<float> Half_Width_of_Detector;
    std::vector<float> Half_Height_of_Detector;
    std::vector<float> Begin_of_Detector_in_Z; 
    std::vector<float> End_of_Detector_in_Z;
    std::vector<float> Energy_Percent;
    std::vector<float> Min_Energy;

    double XYZ[3];
    double XYZWorld[3];


    TH1F* fShowerMotherPDG;
    TH1F* fVx;
    TH1F* fVy;
    TH1F* fVz;
    TH1F* fTruthEnergy;
    TH1F* fCosTheta;
    TH1F* fNumberPerEvent;
    TH1F* fEnergyofShower;
    TH1F* fEventWidth;
    TH1F* fMoliereRadius;

  };
};


////////////////////////////////////////////////////////////////////////
namespace mcchk {
  
  ShowerAnaCheck::ShowerAnaCheck(fhicl::ParameterSet const& pset)
    : EDAnalyzer(pset)
  { 
    this->reconfigure(pset);
  }
  
  ////////////////////////////////////////////////////////////////////////////
  
  ShowerAnaCheck::~ShowerAnaCheck()  { }
  
  ////////////////////////////////////////////////////////////////////////////

  void ShowerAnaCheck::reconfigure(const fhicl::ParameterSet& pset)
  {
    fFLSHitListLabel           = pset.get< std::string >("FLSHitListLabel");
    Half_Width_of_Detector     = pset.get<std::vector<float> >("sHalf_Width_of_Detector");
    Half_Height_of_Detector    = pset.get<std::vector<float> >("sHalf_Height_of_Detector");
    Begin_of_Detector_in_Z     = pset.get<std::vector<float> >("sBegin_of_Detector_in_Z");
    End_of_Detector_in_Z       = pset.get<std::vector<float> >("sEnd_of_Detector_in_Z");
    Energy_Percent             = pset.get<std::vector<float> >("sEnergy_Percent");
    Min_Energy                 = pset.get<std::vector<float> >("sMin_Energy");
  }
  
  void ShowerAnaCheck::beginJob()
  {
    art::ServiceHandle<art::TFileService> tfs;
    
    fShowerMotherPDG = tfs->make<TH1F>("fFirstPDG","Plot of PDG Code of Shower Mother; PDG Code; Entries",120,0,120);
    fVx = tfs->make<TH1F>("fVx","X Coordinate of Shower Mother Vertex; X (cm); Entries",30,-150,150);
    fVy = tfs->make<TH1F>("fVy","Y Coordinate of Shower Mother Vertex; Y (cm); Entries",40,-200,200);
    fVz = tfs->make<TH1F>("fVz","Z Coordinate of Shower Mother Vertex; Z (cm); Entries",100,0,1000);
    fTruthEnergy = tfs->make<TH1F>("fTruthEnergy","MC Truth Energy of Shower Mother; Energy (GeV); Entries",40,0,8);
    fCosTheta = tfs->make<TH1F>("fCosTheta","Cosine of Angle Between the Shower Mother Momentum and Z Axis; radians;Entries",100,-1,1);
    fNumberPerEvent = tfs->make<TH1F>("fNumberPerEvent","Number of Particles in Each Event; Number in Event; Entries",15,0,15);
    fEnergyofShower = tfs->make<TH1F>("fEnergyofShower","Energy of All FLSHITS; Energy (GeV); Entries",100,0,12);
    fEventWidth = tfs->make<TH1F>("fEventWidth","Average Distance to FLSHit per Event; Distance (cm); Entries",50,0,15);
    fMoliereRadius = tfs->make<TH1F>("fMoliereRadius","Plot of the Moliere Radius of Each Event, 10-11 cm for NOvA; R_{Moliere}; Entries",25,0,20);

  }
  
  ///////////////////////////////////////////////////////////////////////////
  
  void ShowerAnaCheck::analyze(const art::Event& evt)
  {   
 
    mf::LogInfo("ShowerAnaCheck") << "Calculating Shower Parameters";  

    //  Get the FLS list information out of the event
    art::Handle< std::vector<sim::FLSHitList>  > hitlist;
    evt.getByLabel(fFLSHitListLabel, hitlist);   
    if (hitlist->empty()){
      std::cerr << "no flshit list for this event" << std::endl;
      return;
    }     

    art::Ptr<sim::FLSHitList> hlist(hitlist, 0);
    const std::vector<sim::FLSHit>& hits = hlist->fHits;

    int Number_of_showers_per_event = 0;

    art::ServiceHandle<cheat::BackTracker> bt;
    const sim::ParticleNavigator& pnav = bt->ParticleNavigator();

    // loop over particles, we test if each makes the cuts, then fill histograms
    for (sim::ParticleNavigator::const_iterator k = pnav.begin(); k != pnav.end(); ++k){
      
      const sim::Particle* part = (*k).second;
      
      // Yay! Cuts!
      bool isElectronorPhoton = 0;
      bool isVertexContained = 0;
      bool isEnergyContained = 0;
      bool isEnergeticEnough = 0;
      float Energy_of_shower = 0;
      for (unsigned int j=0; j<hits.size(); ++j) if(hits[j].GetTrackID() == part->TrackId()) Energy_of_shower += hits[j].GetEdep();
      // run a cut on particle type +-11 = electron/positron, 22 = photon
      if (abs(part->PdgCode()) == 11 || part->PdgCode() == 22) 
	isElectronorPhoton = 1;
      // run a cut on the vertex, require it to be in the interior of the detector
      if (abs(part->Vx()) < Half_Width_of_Detector[0]  && 
	  abs(part->Vy()) < Half_Height_of_Detector[0] && 
	  part->Vz() < End_of_Detector_in_Z[0] && 
	  part->Vz() > Begin_of_Detector_in_Z[0])
	isVertexContained = 1;
      // run a cut on how much energy is deposited by flshits, a measure of containment
      // the detector is 70% active, so if we have ~70% deposited, the shower is essentially contained
      if (Energy_of_shower/part->E() > Energy_Percent[0])
	isEnergyContained = 1;
      // cut on energy, don't want particles without enough energy to develope a shower
      if (part->E() > Min_Energy[0]) 
	isEnergeticEnough = 1;
      
      // If the particle passes the cuts, run this so that we can fill histograms
      if (isElectronorPhoton && isVertexContained && isEnergyContained && isEnergeticEnough){
	++ Number_of_showers_per_event;
	float Event_distance = 0;
	float Moliere_radius = 0;
	for (unsigned int j=0; j<hits.size(); ++j) 
	  if(hits[j].GetTrackID() == part->TrackId()) Energy_of_shower += hits[j].GetEdep();
	// Fill hists with quick MCParticle variables for MC validation
	fShowerMotherPDG->Fill(part->PdgCode());
	fVx->Fill(part->Vx());
	fVy->Fill(part->Vy());
	fVz->Fill(part->Vz());
	fTruthEnergy->Fill(part->E());
	fCosTheta->Fill(part->Pz()/part->P());
	fEnergyofShower->Fill(Energy_of_shower);
	// Call the shower width function to calculate the average distance to FLS Hits
	EMShowerWidth(Event_distance, part->TrackId(), part, hits);
	// Fill histogram with average distance for event
	fEventWidth->Fill(Event_distance);
	// Call the Moliere Radius function to calculate Moliere Radius
	EMMoliereRadius(Moliere_radius, part->TrackId(), part, hits);
	// Fill histogram with Moliere radius for event
	fMoliereRadius->Fill(Moliere_radius);
      }
      
    } // end loop over particles

    // After running through all particles, we want to fill the number of showers in the event here
    if (Number_of_showers_per_event > 0) fNumberPerEvent->Fill(Number_of_showers_per_event);
    
  } // end of main analyze function
  
  
  //////////////////////////////////////////////////////////////////////////// 
  
  
  /// When this function is run, it finds the average distance, weighted by energy, to flshits over the shower
  void ShowerAnaCheck::EMShowerWidth(float& Event_distance, int MomsID, const sim::Particle* part, const std::vector<sim::FLSHit>& hits){
    
    art::ServiceHandle<geo::Geometry> geom;
    
    float Energy_of_shower = 0;
    float X_cent = 0;
    float Y_cent = 0;
    float Z_cent = 0;
    
    // this for loop calculates the centroid of the shower, weights by energy
    for (unsigned int index1=0; index1<hits.size(); ++index1){

      if (hits[index1].GetTrackID() == MomsID){
	// here XYZ is the average of where the hit enters and where it exits in the cell's frame
	// XYZ is the global position
	XYZ[0] = 0.5*(hits[index1].GetEntryX() + hits[index1].GetExitX());
	XYZ[1] = 0.5*(hits[index1].GetEntryY() + hits[index1].GetExitY());
	XYZ[2] = 0.5*(hits[index1].GetEntryZ() + hits[index1].GetExitZ());
	geom->Plane(hits[index1].GetPlaneID())->Cell(hits[index1].GetCellID())->LocalToWorld(XYZ,XYZWorld);
	// recursively add the coordinates of the hit
	X_cent += hits[index1].GetEdep()*XYZWorld[0];
	Y_cent += hits[index1].GetEdep()*XYZWorld[1];
	Z_cent += hits[index1].GetEdep()*XYZWorld[2];
	Energy_of_shower += hits[index1].GetEdep();
      }

    }

    // normalize the centroid by the overall shower Energy
    X_cent = X_cent/Energy_of_shower;
    Y_cent = Y_cent/Energy_of_shower;
    Z_cent = Z_cent/Energy_of_shower;

    // here we find the average distance to the fls hits, weighted by energy
    for (unsigned int j=0; j<hits.size(); ++j)
      {
	if (hits[j].GetTrackID() == MomsID){
	  // here XYZ is the average of where the hit enters and where it exits in the cell's frame
	  // XYZ is the global position
	  XYZ[0] = 0.5*(hits[j].GetEntryX() + hits[j].GetExitX());
	  XYZ[1] = 0.5*(hits[j].GetEntryY() + hits[j].GetExitY());
	  XYZ[2] = 0.5*(hits[j].GetEntryZ() + hits[j].GetExitZ());
	  geom->Plane(hits[j].GetPlaneID())->Cell(hits[j].GetCellID())->LocalToWorld(XYZ,XYZWorld);
	  // Add the distance to each hit * energy in that hit / total energy = energy weighting scheme
	  Event_distance = Event_distance + hits[j].GetEdep()*PointLineDistance(XYZWorld[0],XYZWorld[1],XYZWorld[2],
									    part->Vx(),part->Vy(),part->Vz(),
									    X_cent,Y_cent,Z_cent)/Energy_of_shower;
	}
	
      } // end of FLSHit hits loop
    
  } // end of EMShowerWidth()
 

  ////////////////////////////////////////////////////////////////////////////

  /// This function calculates the Moliere Radius
  void ShowerAnaCheck::EMMoliereRadius(float& Moliere_radius, int MomsID, const sim::Particle* part, const std::vector<sim::FLSHit>& hits){

    art::ServiceHandle<geo::Geometry> geom;
    
    float Energy_of_shower = 0;
    float X_cent = 0;
    float Y_cent = 0;
    float Z_cent = 0;

    // Again, calculate the centroid of the showewr
    for (unsigned int index1=0; index1<hits.size(); ++index1){
      if (hits[index1].GetTrackID() == MomsID){
	// here XYZ is the average of where the hit enters and where it exits in the cell's frame
	// XYZ is the global position
	XYZ[0] = 0.5*(hits[index1].GetEntryX() + hits[index1].GetExitX());
	XYZ[1] = 0.5*(hits[index1].GetEntryY() + hits[index1].GetExitY());
	XYZ[2] = 0.5*(hits[index1].GetEntryZ() + hits[index1].GetExitZ());
    	geom->Plane(hits[index1].GetPlaneID())->Cell(hits[index1].GetCellID())->LocalToWorld(XYZ,XYZWorld);
	// recursively add the coordinates of the hit
	X_cent += hits[index1].GetEdep()*XYZWorld[0];
	Y_cent += hits[index1].GetEdep()*XYZWorld[1];
	Z_cent += hits[index1].GetEdep()*XYZWorld[2];
	Energy_of_shower += hits[index1].GetEdep();
      }
    }
    // Normalize the centroid of the shower by the overall energy
    X_cent = X_cent/Energy_of_shower;
    Y_cent = Y_cent/Energy_of_shower;
    Z_cent = Z_cent/Energy_of_shower;

    // Here we have a crude way to find the Moliere radius by counting the energy deposited in a cylinder
    // around the shower axis.  The first for loop slowly increases the radius, and the second goes over 
    // all hits and finds how much energy is deposited within the shell
    for (float Moliere_parameter = 0; Moliere_parameter < 100; Moliere_parameter += .1){
      float Energy_contained = 0;
      for (unsigned int j = 0; j<hits.size(); ++j){
	// here XYZ is the average of where the hit enters and where it exits in the cell's frame
	// XYZ is the global position
	XYZ[0] = 0.5*(hits[j].GetEntryX() + hits[j].GetExitX());
	XYZ[1] = 0.5*(hits[j].GetEntryY() + hits[j].GetExitY());
	XYZ[2] = 0.5*(hits[j].GetEntryZ() + hits[j].GetExitZ());
    	geom->Plane(hits[j].GetPlaneID())->Cell(hits[j].GetCellID())->LocalToWorld(XYZ,XYZWorld);
	// if the hit is within a cylinder of radius moliere_parameter, 
	if (hits[j].GetTrackID() == MomsID && PointLineDistance(XYZWorld[0],XYZWorld[1],XYZWorld[2],
							    part->Vx(),part->Vy(),part->Vz(),X_cent,Y_cent,Z_cent) < Moliere_parameter) 
	  Energy_contained += hits[j].GetEdep();
      }
      // Set Moliere_radius to the iterative constant above if <90% of energy is contained, else break
      if (Energy_contained/Energy_of_shower <= .9) Moliere_radius = Moliere_parameter;
      else break;
    }
    
  }

  ///////////////////////////////////////////////////////////////////////////

  /// This function calculates the distance between any point and a line.  
  /// The point is the first three arguments (x,y,z), then the line is the 
  /// line passing through the other two points, (x1,y1,z1) and (x2,y2,z2)
  float ShowerAnaCheck::PointLineDistance(float x, float y, float z, float x1, float y1, float z1, float x2, float y2, float z2){
    float Lineone_point[3];
    float Lineone_linetwo[3];
    float Distance_one = 0;
    float Distance_two = 0;
    float Cosine = 0;
    float Dot_product = 0;
    Lineone_point[0]=x-x1;
    Lineone_point[1]=y-y1;
    Lineone_point[2]=z-z1;
    Lineone_linetwo[0]=x2-x1;
    Lineone_linetwo[1]=y2-y1;
    Lineone_linetwo[2]=z2-z1;
    Distance_one = util::pythag(Lineone_point[0], Lineone_point[1], Lineone_point[2]);
    Distance_two = util::pythag(Lineone_linetwo[0], Lineone_linetwo[1], Lineone_linetwo[2]);
    Dot_product = Lineone_point[0]*Lineone_linetwo[0]+Lineone_point[1]*Lineone_linetwo[1]+Lineone_point[2]*Lineone_linetwo[2];
    Cosine = Dot_product/Distance_one/Distance_two;
    return Distance_one*sqrt(1-util::sqr(Cosine));
  }

} // end mcchk

////////////////////////////////////////////////////////////////////////
namespace mcchk
{
  DEFINE_ART_MODULE(ShowerAnaCheck);
}
