//////////////////////////////////////////////////////////////////////////////
/// \brief   Module to take Cosmics MCTruth object which holds list of particles
///          and make it produce one MCTruth per Mother particle
/// \author  Gavin Davies - gsdavies@iastate.edu
/// \date    Feb. 2013
////////////////////////////////////////////////////////////////////////////
#ifndef UTIL_TRUTHEXPAND_H
#define UTIL_TRUTHEXPAND_H

#include <iostream>
#include <string>
#include <vector>

// Framework includes
#include "art/Framework/Core/EDProducer.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h"
#include "fhiclcpp/ParameterSet.h"
#include "messagefacility/MessageLogger/MessageLogger.h"

// Nova includes
#include "RecoBase/CellHit.h"
#include "MCCheater/BackTracker.h"
#include "SimulationBase/MCTruth.h"
#include "Simulation/Particle.h"

namespace util {
  class TruthExpand : public art::EDProducer {
  public:
    explicit TruthExpand(fhicl::ParameterSet const& pset);
    virtual ~TruthExpand();
    void produce(art::Event& evt);
    void reconfigure(const fhicl::ParameterSet& p);
    
  private:    
    std::string  fHitInput;           // Input for rb::CellHit
  };
};

namespace util{
  
  //......................................................................
  TruthExpand::TruthExpand(fhicl::ParameterSet const& pset)
  {
    reconfigure(pset);
    produces< std::vector<simb::MCTruth> >();
  }
  
  //......................................................................
  TruthExpand::~TruthExpand()
  {
  }

  //......................................................................
  void TruthExpand::reconfigure(const fhicl::ParameterSet& pset)
  {
    fHitInput           = pset.get< std::string >("HitInput"    ) ;
  }

  //......................................................................
  /// Provides read-write access to the event data.

  void TruthExpand::produce(art::Event& evt)
  {
    art::ServiceHandle<cheat::BackTracker> bt;
    // Not to be run on Data of course! Do nothing then
    if( !bt->HaveTruthInfo() ) return;

    // Declare a container for MCTruth objects to be stored in the art::event
    std::unique_ptr< std::vector<simb::MCTruth> > mcts(new std::vector<simb::MCTruth>);

    if(!bt->HaveTruthInfo()){
      mf::LogWarning("TruthExpand") << "attempting to run MC truth check on "
				    << "things without truth, bail";
      return;
    }
    
    // get all hits in the event                                                      
    art::Handle< std::vector<rb::CellHit> > hithdl;
    evt.getByLabel(fHitInput, hithdl);

    art::PtrVector<rb::CellHit> allhits;
    for(size_t h = 0; h < hithdl->size(); ++h){
      art::Ptr<rb::CellHit> hit(hithdl, h);
      allhits.push_back(hit);
    }
    
    // Use BackTracker to get a vector of all particles matched to the hits
    // We then will only make one MCTruth for each Particle
    std::vector<const sim::Particle*> parts = bt->HitsToParticle(allhits);
    
    // check our vector of Particles isn't empty
    // If it is put out an empty truth to avoid exceptions
    if(parts.empty()) {
      simb::MCTruth truth;
      mcts->push_back(truth);
      evt.put(std::move(mcts));
      return;
    }

    /// iterate over the vector of Particles so that we can get at each one
    for(auto const& particle : parts){
      
      // make one truth object for each particle
      simb::MCTruth truth;

      // can't pass a const to MCTruth::Add so have to make a copy...
      // Only proceed if a Mother particle
      if((*particle).Mother() == 0){
	sim::Particle p = (*particle);
	
	LOG_DEBUG("TruthExpand") << "PDG: " << (*particle).PdgCode() << std::endl;
	
	// add one particle to one MCTruth
	truth.Add(p);	
	truth.SetOrigin(simb::kCosmicRay);
	mcts->push_back(truth);
      }
    }

    // put all MCTruths into the event
    evt.put(std::move(mcts));

  } // end produce

} // end namespace util

namespace util{

  DEFINE_ART_MODULE(TruthExpand);

}

#endif
