////////////////////////////////////////////////////////////////////////
/// \brief Module to extract and fill file metadata
/// \file  Metadata_module.cc
/// \author Adam Aurisano
///
/// Generated at Fri Apr 12 17:18:04 2013 by Adam Aurisano using artmod
/// from art v1_02_06.
////////////////////////////////////////////////////////////////////////

#include <algorithm>
#include <cctype>
#include <cstdlib>
#include <iterator>
#include <map>
#include <memory>
#include <set>
#include <sstream>
#include <string>
#include <vector>

#include "Geometry/Geometry.h"
#include "NovaDAQConventions/DAQConventions.h"
#include "Metadata/MetadataManager.h"

#include "art/Framework/Core/EDAnalyzer.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Services/System/FileCatalogMetadata.h"
#include "art/Framework/Core/FileBlock.h"
#include "art/Framework/Principal/Run.h"
#include "art/Framework/Principal/SubRun.h"
#include "art/Framework/Principal/Event.h"
#include "art/Persistency/Provenance/Timestamp.h"
#include "fhiclcpp/ParameterSet.h"
#include "messagefacility/MessageLogger/MessageLogger.h"

#include <IFDH_service.h>


namespace meta {
  
  class Metadata : public art::EDAnalyzer {
  public:
    explicit Metadata(const fhicl::ParameterSet& p);
    virtual ~Metadata();
    
    
    void beginRun(art::Run const & r);
    void beginSubRun(art::SubRun const & sr);
    void endRun(art::Run const & r);
    void analyze(const art::Event& e);
    void reconfigure(const fhicl::ParameterSet& p);
    void respondToOpenInputFile(art::FileBlock const & fb);
    void respondToCloseOutputFiles(art::FileBlock const & fb);
    
  private:
    void GetInputMetadata (std::string inMetadata);
    void FillMetadata();
    
    int         fNEvts;
    std::string fRunType;
    bool        fUseSubruns;
    bool        fUseFhclParent;

    std::set< int >                 fRuns;
    std::set< std::pair<int, int> > fSubRuns;
    std::vector< std::string >      fParents;
  };
};

namespace meta {
  Metadata::Metadata(const fhicl::ParameterSet& p) : 
    EDAnalyzer    (p),
    fNEvts        (0),
    fRunType      (p.get< std::string >("runType", "physics")),
    fUseSubruns   (p.get< bool        >("useSubruns", false) ),
    fUseFhclParent(p.get< bool        >("useFhclParent")     )
    
  {
    MetadataManager& fManager = MetadataManager::getInstance();
    
    fhicl::ParameterSet inPSet;
    bool foundPSet = p.get_if_present<fhicl::ParameterSet>("params", inPSet);
    if(foundPSet) fManager.DoParsing(inPSet);
    
    //push fcl (for simulation jobs into the parents vector
    std::string fclName = fManager.GetMetadata()["simulated.jobfcl"];
    if(fclName != "") fParents.push_back(fclName);
    
    if(fUseFhclParent) {
      art::ServiceHandle<IFDH> ifdhp;
      std::string oldMetadata;
      mf::LogInfo("Metadata") << "Parent File name is: " << fclName;
      try {
	oldMetadata = ifdhp->getMetadata(fclName);
      }
      catch(WebAPIException) {
	mf::LogError("Metadata") << "Web API Exception: ";
	oldMetadata = "";
      }
      mf::LogInfo("Metadata") << oldMetadata;
      GetInputMetadata(oldMetadata);  
    }
  }
  
  //.............................................................
  Metadata::~Metadata()
  {
  }
  
  //.............................................................
  void Metadata::reconfigure(const fhicl::ParameterSet& pset)
  {
  }

  void Metadata::GetInputMetadata(std::string inMetadata)
  {
    MetadataManager& fManager = MetadataManager::getInstance();
    
    std::istringstream metadataStream;
    metadataStream.str(inMetadata);
    while(metadataStream.good()){
      std::string line, key, value;
      std::getline(metadataStream, line);
      size_t pos = line.find(":");
      if(pos != std::string::npos){
	key = line.substr(0, pos);
	value = line.substr(pos+1);
	size_t dot_pos = key.find(".");
	  //if the key contains a dot, this is experiment specific: copy over
	if (dot_pos != std::string::npos) fManager.AddMetadata(key, value);
      }
    }  
    return;
  }
  
  //......................................................................
  void Metadata::beginRun(art::Run const & r)
  {
    int run_num = r.run();
    fRuns.insert( run_num );
    return;
  }
  
  //......................................................................
  void Metadata::beginSubRun(art::SubRun const & sr)
  {
    int run_num = sr.run();
    int sub_num = sr.subRun();
    fSubRuns.insert( std::make_pair( run_num, sub_num ) );
    return;
  }

  //.................................................................
  void Metadata::analyze(const art::Event & e)
  {
    //art::Timestamp evtTime = e.time();
    //if (fNEvts == 0)
    //{
    //  fStartTime = evtTime;
    //  fEndTime = evtTime;
    //}
    //else
    //{
    //  if (evtTime < fStartTime) fStartTime = evtTime;
    //  if (evtTime > fEndTime) fEndTime = evtTime;
    //}
    
    ++fNEvts;
    return;
  }
  
  //....................................................................
  void Metadata::respondToOpenInputFile(art::FileBlock const & fb)
  {
    //If we're forcing fcl parentage (like overlays) 
    //Don't get the metadata using the input file.
    if(fUseFhclParent) return; 
    
    std::string fileWPath = fb.fileName();
    if (fileWPath == "") return;
    std::string fileName;
    size_t pos = fileWPath.rfind("/");
    if (pos != std::string::npos) fileName = fileWPath.substr(pos+1);
    else fileName = fileWPath;
    
    fParents.push_back( fileName );
    
    //get old metadata from input file
    art::ServiceHandle<IFDH> ifdhp;
    std::string oldMetadata;
    try{
      oldMetadata = ifdhp->getMetadata(fileName);
    }
    catch(WebAPIException){
      oldMetadata = "";
    }
    GetInputMetadata(oldMetadata);
    
    return;
  }
  
  //..................................................................
  void Metadata::FillMetadata()
  {
    
    //if(fUseFhclParent) return;
    //fhicl file is already parsed - now add in additional metadata 
    MetadataManager& fManager = MetadataManager::getInstance();

    std::ostringstream value_stream;
    
    std::string detector("unknown");
    art::ServiceHandle<geo::Geometry>  geo;
    if(geo->DetId() == novadaq::cnv::kFARDET)       detector = "fd";
    else if(geo->DetId() == novadaq::cnv::kNEARDET) detector = "nd";
    else if(geo->DetId() == novadaq::cnv::kNDOS)    detector = "ndos";
    fManager.AddMetadata( "Nova.DetectorID", detector );
    
    //set release this stage of processing used (takes precedence over input files)
    fManager.AddMetadata( "Nova.Release",   std::getenv("SRT_BASE_RELEASE"));
    
    //value_stream.str("");
    //value_stream << fStartTime.value();
    //fManager.AddMetadata( "start_time", value_stream.str() );
    //value_stream.str("");
    //value_stream << fEndTime.value();
    //fManager.AddMetadata( "end_time", value_stream.str() );
    
    //set the pass number
    //by default is set to 1
    //if input files have a pass number, that supercedes the default
    //if NPASS exists as an environment variable, 
    //that supercedes value in input files
    char* npass = std::getenv("NPASS");
    if(npass != NULL){
      fManager.AddMetadata( "Nova.SubVersion", npass );
    }
    
    //Add in event count
    value_stream.str("");
    value_stream << fNEvts;
    fManager.AddMetadata( "event_count", value_stream.str() );
    
    //Add in runs list
    value_stream.str("");  
    if(fUseSubruns){
      value_stream << "[";
      for(auto it = fSubRuns.begin(); it != fSubRuns.end(); ++it){
	if(it != fSubRuns.begin()) value_stream << ",\n";
	value_stream << "[ " << it->first << " , " 
		     << it->second << " , " 
		     << "\"" << fRunType 
		     << "\"" << " ]"; 
      }
      value_stream << "]";
    }
    else{
      value_stream << "[";
      for( auto it = fRuns.begin(); it != fRuns.end(); ++it){
	if (it != fRuns.begin() ) value_stream << ",\n";
	value_stream << "[ " << *it << " , " 
		     << "\"" << fRunType << "\"" << " ]"; 
      }
      value_stream << "]";
    }
    fManager.AddMetadata( "runs", value_stream.str() );
    
    //Add in parents
    value_stream.str("");
    value_stream << "[";
    for(auto it = fParents.begin(); it != fParents.end(); ++it){
      if(it != fParents.begin() ) value_stream << ",\n";
      value_stream << "{ \"file_name\": \"" 
		   << *it << "\" }";
    }
    value_stream << "]";
    fManager.AddMetadata( "parents", value_stream.str() );
  }
  
  //..................................................................
  void Metadata::endRun(art::Run const & r)
  {
    FillMetadata();
  }
  
  void Metadata::respondToCloseOutputFiles(art::FileBlock const & fb)
  {
    //actually add the metadata to FileCatalogMetadata service here
    //this is the only time we actually write it out
    
    FillMetadata();
    
    MetadataManager& fManager = MetadataManager::getInstance();
    
    art::ServiceHandle<art::FileCatalogMetadata> md;
    for(auto& it : fManager.GetMetadata()){
      std::string key = it.first;
      std::string value = it.second;
      
      //remove "jobfcl" from the metadata when committing it     
      size_t found = key.find("jobfcl");
      if(found != std::string::npos) continue;
      
      md->addMetadata( key, value );      
    }
  }
  
  DEFINE_ART_MODULE(Metadata)
}
