///////////////////////////////////////////////////////////////////////                                         
/// \file  MichelEAna_module.cc                                                                                 
/// \brief A module to study the michel electron selection cuts and performance, 
///        and true michel electron spectra.
///                                                                                                            
/// \version $ Id: MichelEAna_module.cc, v 1.11 2012-10-26 22:19:03 ksachdev Exp $                              
/// \author  Kanika Sachdev (ksachdev@physics.umn.edu)                                                         
////////////////////////////////////////////////////////////////////////                                       
            
 

#include "SimulationBase/MCTruth.h"
#include "SimulationBase/MCNeutrino.h"
#include "Geometry/Geometry.h"
#include "MCCheater/BackTracker.h"
#include "RecoBase/CellHit.h"
#include "RecoBase/Track.h"
#include "RecoBase/Cluster.h"
#include "MichelEFilter/MichelECluster.h"


#include <art/Framework/Principal/Handle.h>
#include <art/Framework/Principal/Event.h>
#include "art/Framework/Core/FindManyP.h"
#include "art/Framework/Core/FindOneP.h"
#include <art/Framework/Core/ModuleMacros.h>
#include <art/Framework/Core/EDAnalyzer.h>
#include <art/Framework/Services/Registry/ServiceHandle.h>
#include <art/Framework/Services/Optional/TFileService.h>
#include <art/Framework/Services/Optional/TFileDirectory.h>
#include <fhiclcpp/ParameterSet.h>
#include <messagefacility/MessageLogger/MessageLogger.h>


#include <map>
#include <vector>
#include <cmath>
#include "TNtuple.h"

namespace me {

  class MichelEAna : public art::EDAnalyzer {

  public:
    explicit MichelEAna(fhicl::ParameterSet const& pset);
    virtual ~MichelEAna();

    void analyze(art::Event const& evt);
    void beginJob();
    void reconfigure(const fhicl::ParameterSet& p);

  private:

    TTree* fMichelTree;

    std::string fMCTruthLabel;
    std::string fTrackLabel;
    std::string fSliceLabel;
    std::string fMichelELabel;

    double fTrueE;
    double fTrueVisE;
    int    fTrueVisNCell;

    double fRecoE;
    double fRecoVisE;
    int    fRecoNCell;
    double fRecoDeltaT;
    double fRecoDeltaDist;
    bool   fIsCaptured;

    float fHitCollEff;
    float fHitCollPur;
    
    float fEvt;
    float fRun;
    float fSubrun;

    art::ServiceHandle<geo::Geometry> fgeom;

    void SetDefaults();

  };

}
namespace me {

  /*********************************************************************************************/
  ////////////////////////////////////////////////////////////////////////
  
  
  MichelEAna::MichelEAna(fhicl::ParameterSet const& p)
    : EDAnalyzer(p)
  {
    this->reconfigure(p);
  }
  

  /*********************************************************************************************/
  ////////////////////////////////////////////////////////////////////////
  
  MichelEAna::~MichelEAna() { 
  }


  /*********************************************************************************************/
  ////////////////////////////////////////////////////////////////////////
  

  void MichelEAna::reconfigure(const fhicl::ParameterSet& p){

    fTrackLabel    = p.get< std::string>("TrackLabel");
    fMCTruthLabel  = p.get< std::string>("MCTruthLabel");
    fSliceLabel    = p.get< std::string>("SliceLabel");
    fMichelELabel  = p.get< std::string>("MichelELabel");
    
  }
  
  
  /*********************************************************************************************/
  ////////////////////////////////////////////////////////////////////////

  void MichelEAna::beginJob(){
    
    art::ServiceHandle<art:: TFileService> tfs;

    fMichelTree    = tfs->make<TTree>("fMichelTree","fMichelTree");


    fMichelTree->Branch("trueE",           &fTrueE);
    fMichelTree->Branch("trueVisE",        &fTrueVisE);
    fMichelTree->Branch("trueVisNCell",    &fTrueVisNCell);

    fMichelTree->Branch("recoE",           &fRecoE);
    fMichelTree->Branch("reocVisE",        &fRecoVisE);
    fMichelTree->Branch("recoNCell",       &fRecoNCell);
    fMichelTree->Branch("recoDeltaT",      &fRecoDeltaT);
    fMichelTree->Branch("recoDeltaDist",   &fRecoDeltaDist);
    fMichelTree->Branch("iscap",           &fIsCaptured);

    fMichelTree->Branch("eff",             &fHitCollEff);
    fMichelTree->Branch("pur",             &fHitCollPur);

    fMichelTree->Branch("event",           &fEvt);
    fMichelTree->Branch("run",             &fRun);
    fMichelTree->Branch("Subrun",          &fSubrun);
  }
  
  /*********************************************************************************************/
  ////////////////////////////////////////////////////////////////////////
  
  void MichelEAna::analyze(art::Event const& evt){

    art::ServiceHandle<cheat::BackTracker>   bt;
    fEvt = evt.id().event();

    art::Handle< std::vector<rb::Cluster > > slicecol;
    evt.getByLabel( fSliceLabel, slicecol);
    art::PtrVector<rb::Cluster > slices;
    
    for(int i = 0; i < (int)slicecol->size(); i++){
      art::Ptr<rb::Cluster> sli(slicecol, i);
      slices.push_back(sli);
    }


    // A vector of all hits in the event is needed by backtracker

    int nSlices = slices.size();

    art::PtrVector< rb::CellHit>  allHits;
    for(int s = 0; s < nSlices; s++){
      for(int c = 0; c<(int)slices[s]->NCell(); c++)
        allHits.push_back( slices[s]->Cell(c) );
    }

    art::FindManyP<rb::Track> fmt(slicecol, evt, fTrackLabel);
    if( !fmt.isValid() )
      return;

    for( int isli = 0; isli < nSlices; isli++){

      if(slices[isli]->IsNoise()) 
	continue;

      std::vector< art::Ptr<rb::Track> > tracks = fmt.at(isli);

      if( tracks.empty() )
	continue;
      
      int ntrk = tracks.size();

      art::FindManyP<me::MichelECluster> fo(tracks, evt, fMichelELabel);   
      
      for(int itrk = 0; itrk < ntrk; ++itrk){
	SetDefaults();
	std::vector<art::Ptr<me::MichelECluster> > michelE   = fo.at(itrk);
	
	if( michelE.empty() )
	  continue;
	int nmichels = michelE.size();
	
	for(int ime = 0; ime < nmichels; ++ime){
	  
	  fRecoNCell = michelE[ime]->NCell();
	  fRecoVisE  = michelE[ime]->TotalGeV();
	  fRecoDeltaT = michelE[ime]->DeltaT();
	  fRecoDeltaDist = michelE[ime]->DistanceToTrackEnd();
	  
	  if( bt->HaveTruthInfo() ){
	    const std::vector< const sim::Particle * > particles = bt->HitsToParticle( michelE[ime]->AllCells() );
	    for( int ip = 0; ip < (int)particles.size(); ++ip){
	      const sim::Particle * part   = particles[ip];
	      const sim::Particle * mother = bt->TrackIDToParticle(part->Mother());

	      if( abs(part->PdgCode())   == 11  && 
		  abs(mother->PdgCode()) == 13 &&
		  (part->Process()       == "Decay" ||
		   part->Process()       == "muMinusCaptureAtRest"  )){
		// 'tis an electron!
		std::set< int > id;
		id.insert( part->TrackId());
		for( int ida = 0; ida < part->NumberDaughters(); ++ida)
		  id.insert( part->Daughter(ida) );
		
		fHitCollEff = bt->HitCollectionEfficiency( id, michelE[ime]->AllCells(), allHits, 
							   geo::kXorY, NULL, true, NULL, &fTrueVisE,
							   NULL, &fTrueVisNCell);
		fHitCollPur = bt->HitCollectionPurity( id, michelE[ime]->AllCells(), NULL, NULL, true);
		fTrueE = part->E(0);
		if(part->Process()       == "muMinusCaptureAtRest")
		  fIsCaptured = true;
	      }
	    }// end loop over particles
	  }// end if HaveTruthInfo
	  fMichelTree->Fill();
	}// end loop over michels

      }// end loop over tracks
    }// end loop over slices
    
    
    /*********************************************************************************************/
    ////////////////////////////////////////////////////////////////////////
    
  }// End of analyser
  
  void MichelEAna::SetDefaults(){

    fTrueE           = -5;
    fTrueVisE        = -5;
    fTrueVisNCell    = -5;

    fRecoE           = -5;
    fRecoVisE        = -5;
    fRecoNCell       = -5;
    fRecoDeltaT      = -5;
    fRecoDeltaDist   = -5;
    
    fHitCollEff      = 0;
    fHitCollPur      = 0;
    
    fIsCaptured      = false;
    
  }// End of SetDefaults
  
  DEFINE_ART_MODULE(MichelEAna);
  
}// End of namespace

/*********************************************************************************************/
////////////////////////////////////////////////////////////////////////


 


