////////////////////////////////////////////////////////////////////////
// Class:       MichelEFilter
// Module Type: filter
// File:        MichelEFilter.cxx
// \brief
// Generated at Wed Jan 25 09:37:57 2012 by Ji Liu using artmod
// from art v1_00_05.
////////////////////////////////////////////////////////////////////////

#include "MichelEFilter/MichelECluster.h"

// C/C++ includes
#include <list>
#include <vector>
#include <iostream>
#include <cmath>
#include <algorithm>

//ROOT includes
#include "TVector3.h"


//Framework includes
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "art/Persistency/Common/Ptr.h"
#include "Utilities/AssociationUtil.h"
#include "art/Framework/Core/EDFilter.h"
#include "art/Framework/Core/ModuleMacros.h"


// NOvASoft includes
#include "GeometryObjects/CellGeo.h"
#include "Geometry/Geometry.h"
#include "GeometryObjects/PlaneGeo.h"
#include "RecoBase/CellHit.h"
#include "RecoBase/Track.h"
#include "RecoBase/Cluster.h"
#include "Utilities/func/MathUtil.h"


/// Find Michel electron candidates
namespace me
{
  // -------------------------------------------------

  class MichelEFilter : public art::EDFilter {

  public:
    explicit MichelEFilter(fhicl::ParameterSet const & pset);
    virtual ~MichelEFilter();

    virtual bool filter(art::Event & evt);
    void reconfigure(const fhicl::ParameterSet& pset);
			
    void endJob();


  protected:
    double CellDis(const art::Ptr<rb::CellHit> cell, const TVector3 trkEnd) const;


  private:


    //You can change the value of following parameters in the MichelEFilter.fcl file

    double fTimeUpper1;  ///< First timing cut upbond: rule out cellhits later than (muon_track_meanTNS + fTimeUpper1) 
    double fTimeLower1;  ///< First timing cut lowerbond: rule out cellhits earlier than (muon_track_meanTNS + fTimeLower1)
    double fTimeUpper2;  ///< Second timing cut upbond: rule out cellhits later than (michel_candidate_TNS + fTimeUpper2) 
    double fTimeLower2;  ///< Second timing cut lowerbond: rule out cellhits later than (michel_candidate_TNS - fTimeLower2) 
    double fDistMuon;    ///< First distance cut: distance from the end of the muon track
    double fDistMichel;  ///< Second distance cut: distance from Michel e candidate
    bool   fPrint;       ///< if it is true: print out extral information about muon track and michel
    
    
    
    std::string fTrackLabel;     ///< Track module label 
    std::string fTrackInstLabel; ///< Track module instance label 
    std::string fSliceLabel;     ///< Slicer module label 
    std::string fCellHitLabel;   ///< CellHit module label 
    bool fApplyEventFilter;      ///< Boolean to determine if the filter should be applied. 

    int    N_event;
    int    N_track;
    int    N_stopper;
    int    N_bothend;
    int    N_oneend;

    /// Handle to geometry service 
    art::ServiceHandle<geo::Geometry> fGeom;

  };

  //..................................................................  	
  MichelEFilter::MichelEFilter(fhicl::ParameterSet const & pset):
    N_event(0),
    N_track(0),
    N_stopper(0),
    N_bothend(0),
    N_oneend(0)

  {
    produces<std::vector<me::MichelECluster> >();
    produces< art::Assns<me::MichelECluster, rb::Cluster> >();
    produces< art::Assns<me::MichelECluster, rb::Track> >();
    reconfigure(pset); 
  }

  //..................................................................  	
  MichelEFilter::~MichelEFilter()
  {
  }

  //......................................................................


  void MichelEFilter::reconfigure(const fhicl::ParameterSet& pset)
  {

    fTimeUpper1       = pset.get<float>("time1_u");
    fTimeLower1       = pset.get<float>("time1_l");
    fTimeUpper2       = pset.get<float>("time2_u");
    fTimeLower2       = pset.get<float>("time2_l");
    fDistMuon         = pset.get<float>("dis1");
    fDistMichel       = pset.get<float>("dis2");
    fPrint            = pset.get<bool>("print");
    fTrackLabel       = pset.get<std::string>("TrackLabel");
    fTrackInstLabel   = pset.get<std::string>("TrackInstLabel");
    fSliceLabel       = pset.get<std::string>("SlicerLabel");
    fCellHitLabel     = pset.get<std::string>("CellHit");
    fApplyEventFilter = pset.get<bool>("ApplyEventFilter");

  }	

  //..................................................................

  
  bool MichelEFilter::filter(art::Event & evt)
  {
    int event = evt.id().event();
    ++N_event;
    //Get Track
    art::Handle< std::vector<rb::Track> >track;
    try{ 
      evt.getByLabel(fTrackLabel, fTrackInstLabel, track);}
    catch(...) {
      std::cout<<"Could not get a track "<<std::endl;
    }


    art::Handle< std::vector<rb::Cluster> >slicelist;
    try{ 
      evt.getByLabel(fSliceLabel, slicelist);}
    catch(...) {
      std::cout<<"Could not get a slice "<<std::endl;
    }

    //Get CellHit
    art::Handle< std::vector<rb::CellHit> >cellhit;
    try{
      //  std::cout<<"CellHit"<<std::endl;
      evt.getByLabel(fCellHitLabel, cellhit);}
    catch(...) {
      std::cout<<"Could not get a cellhit "<<std::endl;
    }

    // sort the cell hits by time
    std::vector<art::Ptr<rb::CellHit>> cellHits;
    for(unsigned int k = 0; k < cellhit->size(); ++k){
      cellHits.push_back(art::Ptr<rb::CellHit>(cellhit, k));
    }
    rb::SortByTime(cellHits);


    std::unique_ptr<std::vector<me::MichelECluster> > MichelE(new std::vector<me::MichelECluster>);

    std::unique_ptr< art::Assns<me::MichelECluster, rb::Cluster> > sliceAssassin(new art::Assns<me::MichelECluster, rb::Cluster>);
    std::unique_ptr< art::Assns<me::MichelECluster, rb::Track>   > trackAssassin(new art::Assns<me::MichelECluster, rb::Track>);


    //Loop over all the tracks in this event
    for(unsigned int i = 0; i < track->size(); ++i) {
      art::Ptr<rb::Track> trk(track,i);
      if( !trk->Is3D() )
	continue;
      
      ++N_track;


      TVector3 trk_end[2];
      trk_end[0] = trk->Start();
      trk_end[1] = trk->Stop();

      //Loop both end of the track
      int N_me_end = 0;
      for(unsigned int k = 0; k < 2; ++k){
	if(fPrint){
	  std::cout<<"*************************************************************************************"<<std::endl;
	  std::cout<<"event: "<<event<<" track #:"<<i<<" endpoint #:"<<k
		   <<"\nTrack stop at: ("<<trk_end[k].X()<<", "<<trk_end[k].Y()
		   <<", "<<trk_end[k].Z()<<")"<<std::endl;
	  std::cout<<"track mean time:"<<trk->MeanTNS()<<std::endl;
	}

	me::MichelECluster mecluster;


	// cuts
	//CellHit Loop One -- Find the Michel e candidate which has the biggest PE in the timing and distance window
	std::list<art::Ptr<rb::CellHit> > melist;
	double cellpos[3] = {0,0,0};
	double MaxEPos[3] = {0,0,0};
	double dist    = 0;
	double MaxE    = 0;
	double MaxE_t  = 0;
	int sliceIndex = -5;
	art::Ptr<rb::Cluster> mathchedSlice;


	// get start and end cell hits that fall in time window
	std::vector<art::Ptr<rb::CellHit>>::iterator cellHitsIter, cellHitsBegin, cellHitsEnd;
        const double firstT = trk->MeanTNS()+fTimeLower1;
        const double lastT = trk->MeanTNS()+fTimeUpper1;

        auto timeCompare1 = [](art::Ptr<rb::CellHit> h, double t){return h->TNS() < t;};
        auto timeCompare2 = [](double t, art::Ptr<rb::CellHit> h){return h->TNS() > t;};

        cellHitsBegin = std::lower_bound(cellHits.begin(), cellHits.end(),
                                         firstT, timeCompare1);
        
        cellHitsEnd = std::upper_bound(cellHitsBegin, cellHits.end(),
                                       lastT, timeCompare2);
	
        
	int j=-1;
	for(cellHitsIter = cellHitsBegin; cellHitsIter != cellHitsEnd; ++cellHitsIter){

	  ++j;
          art::Ptr<rb::CellHit> c = *cellHitsIter;

	  // caculate the physics position of this CellHit
	  fGeom->Plane(c->Plane())->Cell(c->Cell())->GetCenter(cellpos,0.0);

	  // caculate the distance from this cellhit to the end the the muon track
	  dist = CellDis(c,trk_end[k]);

	  
	  // make the first distance cut - select only the cellhits that fall
	  // in the sphere around the end of the muon track
	  
	  if(dist < fDistMuon) {
	    if(fPrint){
	      std::cout<<"\n In CellHit Loop one and pass the first set of cuts. cellhit #:"
		       <<j<<std::endl;
	      std::cout<<"distance from the end of the track: "<<dist
		       <<"  cellhit TNS: "<<c->TNS()<<" DeltaT: "
		       << (c->TNS() - trk->MeanTNS())<<std::endl;
	    }
	    melist.push_back(c);
	    
	    // after those two cuts, select the cellhit that has the biggest
	    // PE as Michel candidate and record its information for the second set of cuts
	    
	    if(c->PE() > MaxE){
	      MaxE = c->PE();
	      MaxE_t = c->TNS();
	      MaxEPos[0] = cellpos[0];
	      MaxEPos[1] = cellpos[1];
	      MaxEPos[2] = cellpos[2];
	      
	      // find which slice this hit belongs to. We need that for making associations.
	      for(int sli = 0; sli < (int)slicelist->size(); sli++){
		art::Ptr<rb::Cluster> thisSlice(slicelist, sli);
		int nCells = thisSlice->NCell();
		for(int cell = 0; cell < nCells; cell++){
		  if( thisSlice->Cell(cell) == c ){
		    sliceIndex = sli;
		    mathchedSlice = thisSlice;
		    break;
		  }
		}// end loop over cells in thisSlice
		if( sliceIndex != -5)
		  break;
	      }// end loop over slices in this eventr
	      
	    }
	  }
	}//End of CellHit Loop One

	//CellHit Loop Two -- Using Michel e candidate to find Michel e associate CellHits
	double cellpos2[3]={0,0,0};
	double dist2;

	std::list<art::Ptr<rb::CellHit> >::iterator iter = melist.begin() ;
	while( iter != melist.end()){
	  if(fPrint)	
	    std::cout<<"\n In cellhit loop two and pass the second set of cuts!! me cellhit list size: "
		     <<melist.size()<<std::endl;

	  fGeom->Plane((*iter)->Plane())->Cell((*iter)->Cell())->GetCenter(cellpos2,0.0);
	  // calulate the distance from the cellhit to the michel candidate. 
	  // Only doing this in 1D because the cellhits position info are 2D.

	  dist2 = abs(cellpos2[2] - MaxEPos[2]);

	  // make the second timing cut - select only the cellhits that happen between
	  // michel_candidate_TNS+fTimeLower2 and michel_candidate_TNS+fTimeUpper2

	  if(((MaxE_t - fTimeLower2) < (*iter)->TNS()) && ((*iter)->TNS() < MaxE_t + fTimeUpper2)){

	    // make the second distance cut - select only the cellhits that fall in 
	    // the sphere around the michel candidate

	    if(dist2 < fDistMichel){
	      //Fill the Michel e cluster
	      if(fPrint)
		std::cout<<"michel e. dist: "<<dist2<<"  TNS:"<<(*iter)->TNS()<<std::endl;
	      mecluster.Add(*iter);
	    }
	  }
	  ++iter;
	}//End of CellHit Loop Two

	if(fPrint)
	  std::cout<<"\nCluster Loop!!  Num of michel:"<<mecluster.NCell()<<"  X view: "
		   <<mecluster.NXCell()<<" Y view:"<<mecluster.NYCell()<<std::endl;

	// Number Cut: only keep the Michel cluster that has more than one CellHits
	// and has both view CellHits
	if(mecluster.NCell()>1){
	  if(mecluster.NXCell() > 0 && mecluster.NYCell() > 0){  

	    if(fPrint){
	      std::cout<<"filling me object: "<<"\n PE: "<<mecluster.TotalPE()
		       <<" GeV: " <<mecluster.TotalGeV()<<std::endl;
	      std::cout<<"DeltaT: "<<MaxE_t - trk->MeanTNS()<<" N_ch: "
		       <<mecluster.NCell()<<std::endl;
	    }

	   
	    // Set MichelECluster parameters
	    
	    mecluster.SetParentTrack(trk);
	    MichelE->push_back(mecluster);
	    
	    util::CreateAssn( *this, evt, *MichelE, trk, *trackAssassin );
	    util::CreateAssn( *this, evt, *(MichelE.get()), mathchedSlice, *(sliceAssassin.get()) );
	    
	    ++N_stopper;
	    ++N_me_end;
	  }
	}		
      }//End of track ends loop

      if(N_me_end == 2){
	++N_bothend;
      }
      if(N_me_end == 1){
	++N_oneend;
      }
    }//End of track loop

    const bool ret = (!MichelE->empty() || !fApplyEventFilter) ;
    evt.put(std::move(MichelE));
    evt.put(std::move(trackAssassin));
    evt.put(std::move(sliceAssassin));


    return ret;

  }//End of filter


  //..................................................................  
  double MichelEFilter::CellDis(const art::Ptr<rb::CellHit> cell, const TVector3 trkEnd) const {

    double cellPos[3] = {0,0,0};
    fGeom->Plane(cell->Plane())->Cell(cell->Cell())->GetCenter(cellPos,0.0);
    if( cell->View() == geo::kY || trkEnd[0] == 0 ) 
      cellPos[0] = trkEnd[0];
    if( cell->View() == geo::kX || trkEnd[1] == 0)
      cellPos[1] = trkEnd[1]; 
    
    return util::pythag(cellPos[0]-trkEnd[0],
			cellPos[1]-trkEnd[1],
			cellPos[2]-trkEnd[2]);
  }

  //......................................................................
  void MichelEFilter::endJob() {
    std::cout<<"Total events : "<<N_event<<"\nTotal tracks: "<<N_track
	     <<"\nTotal stopper muon tracks: "<<N_stopper<<"\nBoth-end-Michel tracks: "
	     <<N_bothend<<"\none-end-michel track:"<<N_oneend<<std::endl;
  }




  DEFINE_ART_MODULE(me::MichelEFilter);


} // end namespace me


