/////////////////////////////////////////////////////////////////////////////
//  Michel Filter
//  Dan Pershey
//  pershey@caltech.edu
//
//
//  Based on Slice Geometry
//  This MEFilter seraches for spatically correlated noise hits around
//  slices passing a nue PID
//  The module calculates the minimum distance between each noise hit
//  and the slice in question.  Then, if for a high PE hit, this distance 
//  is less than some parameter (set to 20 cm) and within some time window 
//  of the slice (set to 200 ns to 10000 ns), the hit is classified as a 
//  Michel.  The module then further groups these hits into a cluster of 
//  hits. 
//////////////////////////////////////////////////////////////////////////////

// Framework includes
#include "art/Framework/Core/EDProducer.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Core/FindManyP.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Services/Optional/TFileService.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "art/Persistency/Common/PtrVector.h"
#include "art/Persistency/Common/Assns.h"
#include "cetlib/search_path.h"

// NOvASoft includes
#include "RecoBase/FilterList.h"
#include "Geometry/Geometry.h"
#include "GeometryObjects/CellGeo.h"
#include "GeometryObjects/PlaneGeo.h"
#include "MichelEFilter/SliceMEFCluster.h"
#include "RecoBase/CellHit.h"
#include "RecoBase/Cluster.h"
#include "Utilities/AssociationUtil.h"
#include "Utilities/func/MathUtil.h"

#include "TH1.h"
#include "TH2.h"
#include "TFile.h"
#include "TMath.h"

namespace me
{
  /* Stores Likelihood info discriminating between Michel electrons
     and Hadronic Fuzz */
  struct SlcMELikeHists
  {
    TH2F* fMEDeltaT_Dist;
    TH2F* fMECalE_NCells;

    TH2F* fHFDeltaT_Dist;
    TH2F* fHFCalE_NCells;

    TH2F* fRightMatch;
    TH2F* fWrongMatch;
  };

  class SliceMEF : public art::EDProducer
  {
  public:

    explicit SliceMEF(const fhicl::ParameterSet& pset);
    virtual ~SliceMEF();

    void produce(art::Event& evt);
    void beginJob();
    void reconfigure(const fhicl::ParameterSet& pset);

    std::vector<std::vector<art::Ptr<rb::CellHit> > > 
         Clusterer(std::vector< art::Ptr<rb::CellHit> > meHits);
    float HitToHitDistance(const art::Ptr<rb::CellHit>&, 
			   const art::Ptr<rb::CellHit>&);
    int nSlcMatches(const art::Ptr<rb::CellHit> &,
		    art::Handle< std::vector<rb::Cluster> >);

  protected:

    bool AssocWithSlc(double DeltaT, double Dist, int NMatches);
    double GetMID(double DeltaT, double Dist, double CalE, int NCells);
    /* Get from fhicl & store ME / HF Liklihood hists */
    std::string        fSlcMELikeFileName;
    SlcMELikeHists     fSlcMELikeHists;

    // parameters from fhicl file
    std::string        fSliceInput;
    std::string        fHitInput;
    double             DistanceCut;
    double             ParentSlcTimeCut;
    double             LowerTimeCut;
    double             UpperTimeCut;
    double             EnergyCut1;
    double             CellCut;
    double             EnergyCut2;

    art::ServiceHandle<geo::Geometry> fgeom;


  };
}


////////////////////////////////////////////////////////////////////////
namespace me
{

  
  //.......................................................................
  SliceMEF::SliceMEF(const fhicl::ParameterSet& pset) : 
    EDProducer()
  {
    reconfigure(pset);

    produces< std::vector<me::SliceMEFCluster> >();
    produces< art::Assns<me::SliceMEFCluster ,rb::Cluster> >();
  }
  
  //......................................................................
  SliceMEF::~SliceMEF()
  {
  }

  //......................................................................
  void SliceMEF::reconfigure(const fhicl::ParameterSet& pset)
  {
    fSlcMELikeFileName     = pset.get<std::string>("SlcMELikeFileName");

    fSliceInput            = pset.get< std::string >("SlicerInput");
    fHitInput              = pset.get< std::string >("HitInput");

    DistanceCut            = pset.get< double >("MinDistCut");
    ParentSlcTimeCut       = pset.get< double >("ParentSlcDelayCut");
    LowerTimeCut           = pset.get< double >("MinTimeDelayCut");
    UpperTimeCut           = pset.get< double >("MaxTimeDelayCut");
    EnergyCut1             = pset.get< double >("MinPECut1");
    CellCut                = pset.get< double >("CellCut");
    EnergyCut2             = pset.get< double >("MinPECut2");
  }

  /* Gets the ME Likelihood info into the SlcMELikeHists struct*/
  void SliceMEF::beginJob()
  {
    cet::search_path sp("FW_SEARCH_PATH");
    std::string SlcMELikeFileName;
    sp.find_file(fSlcMELikeFileName, SlcMELikeFileName);
    TFile* slcMELike = TFile::Open(SlcMELikeFileName.c_str());
    slcMELike->cd();
    fSlcMELikeHists.fMEDeltaT_Dist  = (TH2F*)gDirectory->Get("MEDeltaT_Dist");
    fSlcMELikeHists.fMECalE_NCells  = (TH2F*)gDirectory->Get("MECalE_NCells");

    fSlcMELikeHists.fHFDeltaT_Dist  = (TH2F*)gDirectory->Get("HFDeltaT_Dist");
    fSlcMELikeHists.fHFCalE_NCells  = (TH2F*)gDirectory->Get("HFCalE_NCells");

    fSlcMELikeHists.fRightMatch    = (TH2F*)gDirectory->Get("RightMatch");
    fSlcMELikeHists.fWrongMatch    = (TH2F*)gDirectory->Get("WrongMatch");
  }

  //......................................................................
  void SliceMEF::produce(art::Event& evt)
  {
  

    art::Handle<std::vector<rb::Cluster> > slices;
    evt.getByLabel(fSliceInput,slices);

    art::Handle<std::vector<rb::CellHit> > hits;
    evt.getByLabel(fHitInput,hits);

    double nuTime;    
    
    rb::Cluster slice;
    art::Ptr<rb::Cluster> slicePTR;

    std::unique_ptr< std::vector<me::SliceMEFCluster> > 
      SlcMEs(new std::vector<me::SliceMEFCluster>);
    std::unique_ptr< art::Assns<me::SliceMEFCluster, rb::Cluster> > 
      SlcAssns(new art::Assns<me::SliceMEFCluster, rb::Cluster>);
    
    
    int noiseIdx = 0;
    for (unsigned int sIdx = 0; sIdx < slices->size(); sIdx++){
      slice = slices->at(sIdx);
      if (slice.IsNoise()) noiseIdx = sIdx;
    }
    // Grab the Noise slice hits
    rb::Cluster noiseSlc = slices->at(noiseIdx);
    art::PtrVector<rb::CellHit > nHits = noiseSlc.AllCells();
    rb::SortByPlane(nHits);
      
    // Calculate the Physics slices in the nue slices
    for (unsigned int idx = 0; idx < slices->size(); idx++){
      
      std::vector< art::Ptr<rb::CellHit > > meHits;
      
      slice = slices->at(idx);
	
      if (slice.IsNoise()) continue;
      art::Ptr<rb::Cluster> slicePtr(slices,idx);
	
      // We'll nedd this later for checking delta T
      nuTime = slice.MeanTNS();

      art::PtrVector<rb::CellHit > pHits = slice.AllCells();
      rb::SortByPlane(pHits);
      
      if (pHits.size() == 0) continue;
          
      // And here we seperate out the noise hits into a vector
      // These are the hits our Michels will come from
      // Stored as a vector with the first index the X View hits
      // second index has the Y View hits.
      // Loop over x / y view cells noise
      for (unsigned int nIdx = 0; nIdx < (unsigned int)nHits.size(); nIdx++){

	// First check if the time delay is in an acceptable range	
	double nTime = nHits[nIdx]->TNS();
	double deltaT = nTime - nuTime;
	if (deltaT < LowerTimeCut || deltaT > UpperTimeCut)
	  continue;
	
	// Every calculation, set min_dist high
	float minDist = 10000;
	
	for (unsigned int pIdx = 0; pIdx < pHits.size(); pIdx++){
	  if (pHits[pIdx]->View() == nHits[nIdx]->View()){
	    double dist = HitToHitDistance(pHits[pIdx],nHits[nIdx]);
	    if (dist < minDist)
	      minDist = dist;
	  }
	}

	if (minDist < 0.01 && deltaT < ParentSlcTimeCut)
	  continue;
	
	if (minDist > DistanceCut)
	  continue;
	
	meHits.push_back(nHits[nIdx]);	    
      } // Loop over Noise Hits
      
      std::vector<std::vector<art::Ptr<rb::CellHit > > > meClusters
	= Clusterer(meHits);
      
      for (unsigned int cIdx = 0; cIdx < meClusters.size(); cIdx++){
	me::SliceMEFCluster meCluster;
	std::vector<art::Ptr<rb::CellHit> > curHits = meClusters[cIdx];
	int nMatches = 0;
	float minDist = 10000;
	int nRetriggers = 0;
	float ADCRetriggers = 0;
	float PERetriggers = 0;
	for (unsigned int m_idx = 0; m_idx < curHits.size(); m_idx++){
	  meCluster.Add(curHits[m_idx]);
	  int nMatch = nSlcMatches(curHits[m_idx],slices);
	  if (nMatch > nMatches)
	    nMatches = nMatch;
	  for (unsigned int pIdx = 0; pIdx < pHits.size(); pIdx++){
	    float dist = HitToHitDistance(pHits[pIdx],curHits[m_idx]);
	    if (dist < minDist) minDist = dist;
	  }
	  if (minDist == 0){
	    nRetriggers++;
	    ADCRetriggers += curHits[m_idx]->ADC();
	    PERetriggers += curHits[m_idx]->PE();
	  }
	}
	meCluster.SetNSlcMatches(nMatches);
	meCluster.SetDistToSlc(minDist);
	meCluster.SetDeltaT(curHits[0]->TNS() - slice.MeanTNS());
	meCluster.SetNRetriggeredCells(nRetriggers);
	meCluster.SetADCRetriggeredCells(ADCRetriggers);
	meCluster.SetPERetriggeredCells(PERetriggers);
	bool isAssocWithSlc = AssocWithSlc(meCluster.DeltaT(),
					   meCluster.DistToSlc(),
					   meCluster.NSlcMatches());
        meCluster.SetAssocWithSlc(isAssocWithSlc);
	double mid = GetMID(meCluster.DeltaT(),
			    meCluster.DistToSlc(),
			    meCluster.CalorimetricEnergy(),
			    meCluster.NCell());
	meCluster.SetMID(mid);
	SlcMEs->push_back(meCluster);
	util::CreateAssn(*this,evt,*SlcMEs,slicePtr,*(SlcAssns.get()));
      }
    } // End loop over slices
    //}// slices >= 2
    evt.put(std::move(SlcMEs));
    evt.put(std::move(SlcAssns));
  } // end produce
  
  bool SliceMEF::AssocWithSlc(double DeltaT, double Dist, int NMatches){
    bool passesMatch = false;
    if (NMatches == 1)
      return true;
    double RightMatchLike = 
      fSlcMELikeHists.fRightMatch->GetBinContent(fSlcMELikeHists.fRightMatch->FindBin(DeltaT, Dist));
    double WrongMatchLike = 
      fSlcMELikeHists.fWrongMatch->GetBinContent(fSlcMELikeHists.fWrongMatch->FindBin(DeltaT, Dist));
    if (RightMatchLike == 0)
      RightMatchLike = 1e-10;
    if (WrongMatchLike == 0)
      WrongMatchLike = 1e-10;
    if (RightMatchLike > WrongMatchLike)
      passesMatch = true;
    return passesMatch;
  }

  double SliceMEF::GetMID(double DeltaT, double Dist, double CalE, int NCells){
    double MELike = 
      fSlcMELikeHists.fMEDeltaT_Dist->GetBinContent(fSlcMELikeHists.fMEDeltaT_Dist->FindBin(DeltaT, Dist)) *
      fSlcMELikeHists.fMECalE_NCells->GetBinContent(fSlcMELikeHists.fMECalE_NCells->FindBin(CalE, NCells));
    if (MELike == 0)
      MELike = 1e-10;
    double HFLike = 
      fSlcMELikeHists.fHFDeltaT_Dist->GetBinContent(fSlcMELikeHists.fHFDeltaT_Dist->FindBin(DeltaT, Dist)) *
      fSlcMELikeHists.fHFCalE_NCells->GetBinContent(fSlcMELikeHists.fHFCalE_NCells->FindBin(CalE, NCells));
    if (HFLike == 0)
      HFLike = 1e-10;
    return TMath::Log(MELike) - TMath::Log(HFLike);
  }

  // This Piece of Code makes Clusters from a vector of hits
  std::vector<std::vector<art::Ptr<rb::CellHit> > > 
       SliceMEF::Clusterer(std::vector<art::Ptr<rb::CellHit> > meHits){

    std::vector<std::vector<art::Ptr<rb::CellHit> > > meClusters;
    bool neighborMat[meHits.size()][meHits.size()];
    for (unsigned int m_x = 0; m_x < meHits.size(); m_x++){
      for (unsigned int m_y = 0; m_y < meHits.size(); m_y++){
	bool time_cut = 0;
	if (fabs(meHits[m_x]->TNS() - meHits[m_y]->TNS()) < 200) time_cut = 1;
	if (time_cut)
	  neighborMat[m_x][m_y] = 1;
	else
	  neighborMat[m_x][m_y] = 0;
      }
    }
    bool matchMat[meHits.size()][meHits.size()];
    bool isMatch[meHits.size()];
    for (unsigned int m_x = 0; m_x < meHits.size(); m_x++){
      for (unsigned int m_y = 0; m_y < meHits.size(); m_y++){
	matchMat[m_x][m_y] = 0;
      }
      isMatch[m_x] = 0;
    }
    
    for (unsigned int m_idx = 0; m_idx < meHits.size(); m_idx++){
      if (!isMatch[m_idx]){
	matchMat[m_idx][m_idx] = 1;
	isMatch[m_idx] = 1;
      }
      for (unsigned int m_y = 0; m_y < meHits.size(); m_y++){
	if (matchMat[m_idx][m_y]){
	  for (unsigned int m_x = 0; m_x < meHits.size(); m_x++){
	    if (neighborMat[m_x][m_y] && !isMatch[m_x]){
	      isMatch[m_x] = 1;
	      matchMat[m_idx][m_x] = 1;
	    }
	  }
	}
      }
    }	

    for (unsigned int m_idx = 0; m_idx < meHits.size(); m_idx++){
      std::vector<art::Ptr<rb::CellHit > > curHits;
      for (unsigned int m_idx2 = 0; m_idx2 < meHits.size(); m_idx2++){
	if (matchMat[m_idx][m_idx2])
	  curHits.push_back(meHits[m_idx2]);
      }
      if (curHits.size() > 0)
	meClusters.push_back(curHits);
    }
    return meClusters;
  }
  // End code to seperate michel hits into clusters

  int SliceMEF::nSlcMatches(const art::Ptr<rb::CellHit >& hit,
			    art::Handle< std::vector< rb::Cluster > > slices)
  {
    int nMatches = 0;
    for (unsigned int sIdx = 0; sIdx < slices->size(); sIdx++){
      rb::Cluster slice = slices->at(sIdx);
      if (!slice.IsNoise()){
        float minDist = 10000;
        for (unsigned int h_idx = 0; h_idx < slice.NCell(hit->View()); h_idx++){
          float dist = HitToHitDistance(hit, slice.Cell(hit->View(), h_idx));
          if (dist < minDist) minDist = dist;
        }
	double DeltaT = hit->TNS() - slice.MeanTNS();
        if (minDist == 0 && DeltaT < ParentSlcTimeCut)
          continue;
        if (DeltaT < LowerTimeCut || DeltaT > UpperTimeCut)
          continue;
        if (minDist < DistanceCut)
          nMatches++;
      }
    }
    return nMatches;
  }

  //.......................................................
  float SliceMEF::HitToHitDistance(const art::Ptr<rb::CellHit >& hit1,
				   const art::Ptr<rb::CellHit >& hit2){
    // Function calculates the distance between two hits based on z
    // and x or y, whichever the cell measures
    float distance;
    // Only calculate if both hits are in the same view
    if (hit1->View() != hit2->View()) distance = 1000000;
    else{
      double xyz1[3], xyz2[3];
 


      unsigned short plane1_idx = hit1->Plane();
      unsigned short cell1_idx = hit1->Cell();
      const geo::PlaneGeo* plane1 = fgeom->Plane(plane1_idx);
      const geo::CellGeo* cell1 = plane1->Cell(cell1_idx);
      cell1->GetCenter(xyz1);

      unsigned short plane2_idx = hit2->Plane();
      unsigned short cell2_idx = hit2->Cell();
      const geo::PlaneGeo* plane2 = fgeom->Plane(plane2_idx);
      const geo::CellGeo* cell2 = plane2->Cell(cell2_idx);
      cell2->GetCenter(xyz2);

      // Agnostic to whether hits are in X or Y views because
      distance = util::pythag(xyz1[0]-xyz2[0], xyz1[1]-xyz2[1], xyz1[2]-xyz2[2]);
    }
    return distance;
  }

} // end namespace


////////////////////////////////////////////////////////////////////////
namespace me
{
  DEFINE_ART_MODULE(SliceMEF);
}

