////////////////////////////////////////////////////////////////////////                                                        
/// \file  MergeTruth.cxx                                                                                                        
/// \brief A module to produce truth information for Muon-Removed-Electron CC events.
///                                                                                                                              
/// \version $Id: MergeTruth.cxx,v 1.2 2012-09-20 22:01:37 greenc Exp $                                                       
/// \author  Kanika Sachdev (ksachdev@physics.umn.edu)                                                                                            
////////////////////////////////////////////////////////////////////////             

// C/C++ includes
#include <iostream>
#include <vector>

// Root includes
#include "TLorentzVector.h"

// Framework includes
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Persistency/Common/Ptr.h"
#include "art/Framework/Services/Optional/TFileService.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "messagefacility/MessageLogger/MessageLogger.h"
#include "art/Framework/Core/EDProducer.h"
#include "art/Framework/Core/ModuleMacros.h"

// NOvASoft includes

#include "Simulation/FLSHitList.h"
#include "Simulation/PhotonSignal.h"
#include "Simulation/Particle.h"
#include "SimulationBase/MCTruth.h"
#include "MCCheater/BackTracker.h"


namespace murem 
{
  class MergeTruth : public art::EDProducer 
  {
  public: 
    explicit MergeTruth(fhicl::ParameterSet const& pset);
    ~MergeTruth();
    
    void produce(art::Event& evt);
    
    void reconfigure(const fhicl::ParameterSet& pset);
    
  private: 
    
    std::string      fParticleLabel;
    std::string      fPhotonLabel;
    std::string      fMCTruthLabel;
    std::string      fMreTruthLabel;
    std::string      fMrePhotonLabel;
    std::string      fMrccInfoLabel;
  };
}

namespace murem
{

  /*********************************************************************************************/
  ////////////////////////////////////////////////////////////////////////

  MergeTruth::MergeTruth(fhicl::ParameterSet const& pset)
  {
    
    this->reconfigure(pset);
    
    produces< std::vector <sim::FLSHitList> >();
    produces< std::vector <sim::PhotonSignal> >();
    produces< std::vector <sim::Particle> >();
    //    produces< std::vector <simb::MCTruth> >();
    
    mf::LogInfo("MergeTruth") <<"MergeTruth::MergeTruth()\n";

  }

  /*********************************************************************************************/
  ////////////////////////////////////////////////////////////////////////

  MergeTruth::~MergeTruth()
  {
  
  }

  /*********************************************************************************************/
  ////////////////////////////////////////////////////////////////////////

  void MergeTruth::reconfigure(const fhicl::ParameterSet& pset)
  {
    fMreTruthLabel = pset.get<std::string>("MreTruthLabel");
    fMrccInfoLabel = pset.get<std::string>("MrccInfoLabel");
    fMrePhotonLabel= pset.get<std::string>("MrePhotonLabel");
    fMCTruthLabel  = pset.get<std::string>("MCTruthLabel");
    fPhotonLabel   = pset.get<std::string>("PhotonLabel");
    fParticleLabel = pset.get<std::string>("ParticleLabel");
  }

  /*********************************************************************************************/
  ////////////////////////////////////////////////////////////////////////

  void MergeTruth::produce(art::Event& evt)
  {

    std::unique_ptr< std::vector<sim::FLSHitList>   > flsHits(new std::vector<sim::FLSHitList>);
    std::unique_ptr< std::vector<sim::Particle>     > particleList(new std::vector<sim::Particle>);
    std::unique_ptr< std::vector<sim::PhotonSignal> > photonSignal(new std::vector<sim::PhotonSignal>);
    //std::unique_ptr< std::vector<simb::MCTruth>     > mcTruth(new std::vector<simb::MCTruth>);
     
    art::ServiceHandle<cheat::BackTracker> bt;
    
    // The following handle has no use but to indicate if the muon
    // was removed my truth. If not, there's not much sense in producing
    // merged truth. Exit asap- don't waste time!

    art::Handle< std::vector< sim::FLSHitList> > mrccFlsHandle;
    evt.getByLabel(fMrccInfoLabel, mrccFlsHandle);
    if( mrccFlsHandle.failedToGet() ) return;

    // Grab the Electron Truth Info saved to the
    // file by the MRE modeule. They live under 
    // label "mregen". Photons live under "elecphotrans"

    
    art::Handle< std::vector< simb::MCTruth> > elecTruthHandle;
    evt.getByLabel(fMreTruthLabel, elecTruthHandle);
    if( elecTruthHandle.failedToGet() ) return;

    art::Handle< std::vector<sim::FLSHitList> > elecFlsHandle;
    evt.getByLabel(fMreTruthLabel, elecFlsHandle);
    if( elecFlsHandle.failedToGet() ) return;

    art::Handle< std::vector<sim::Particle> > elecPlistHandle;
    evt.getByLabel(fMreTruthLabel, elecPlistHandle);
    if( elecPlistHandle.failedToGet() ) return;

    art::Handle< std::vector<sim::PhotonSignal> > elecSigHandle;
    evt.getByLabel(fMrePhotonLabel, elecSigHandle);
    if( elecSigHandle.failedToGet() ) return;

    // Get the original truth information.

    art::Handle< std::vector< simb::MCTruth> > truthHandle;
    evt.getByLabel(fMCTruthLabel, truthHandle);
    if( truthHandle.failedToGet() ) return;

    art::Handle< std::vector<sim::FLSHitList> > flslistHandle;
    evt.getByLabel(fParticleLabel, flslistHandle);
    if( flslistHandle.failedToGet() ) return;

    art::Handle< std::vector<sim::Particle> > plistHandle;
    evt.getByLabel(fParticleLabel, plistHandle);
    if( plistHandle.failedToGet() ) return;

    art::Handle< std::vector<sim::PhotonSignal> > photSigHandle;
    evt.getByLabel(fPhotonLabel, photSigHandle);
    if( photSigHandle.failedToGet() ) return;


    ////////////////////////////////////////////////////////////////

    // Grab the removed muon info and map  the muon trackID to electron

    art::Handle< std::vector<simb::MCParticle> > muInfoHandle;
    evt.getByLabel(fMrccInfoLabel, muInfoHandle);
    
    std::map< unsigned int, unsigned int> muToElec;
    std::set <int> muonID;

    int nMuInfoHandles = muInfoHandle->size();
    for( int m = 0; m < nMuInfoHandles; m++){
      simb::MCParticle muon = (*muInfoHandle)[m];
      muonID.insert(muon.TrackId());

      int nElecHandles = elecTruthHandle->size();
      for(int e = 0; e < nElecHandles; e++){
	simb::MCParticle elec = (*elecTruthHandle)[e].GetParticle(0);
	
	if( (elec.Vx(0) == muon.Vx(0)) && (elec.Vy(0) == muon.Vy(0)) && (elec.Vz(0) == muon.Vz(0)) && (elec.P(0) && muon.P(0))){
	  muToElec[ muon.TrackId()] = elec.TrackId();
	}
      }// end loop over electron MCTruths
    }// end loop over removed muons


    ////////////////////////////////////////////////////////////////

    int nFlsHandles = flslistHandle->size();
    for( int l = 0; l < nFlsHandles; ++l){
      int matchedElecID = 0;
      sim::FLSHitList newFLSList;
      
      const std::vector<sim::FLSHit> &hits = flslistHandle->at(l).fHits;
      
      for( int f = 0; f < (int)hits.size(); ++f){
        bool hitIsFromMuon = false;
	int hitID = hits[f].GetTrackID();
	int motherID = bt->TrackIDToMotherParticle(hitID)->TrackId();

	if( (muonID.find(motherID) != muonID.end())){
	  matchedElecID = muToElec[ motherID ];
	  hitIsFromMuon = true;
	}
	else if (muonID.find(hitID) != muonID.end()){
	  matchedElecID = muToElec[ hitID ];
	  hitIsFromMuon = true;
	}

     	if(!hitIsFromMuon)
	  newFLSList.fHits.push_back(hits[f]);
      }// end loop over flshits in this FLSHitList


      int nElecFlsHandles = elecFlsHandle->size();
      for( int e = 0; e < nElecFlsHandles; e++){
	int matchedElecList = -1;
	const std::vector<sim::FLSHit> &elechits = elecFlsHandle->at(e).fHits;	
	for(int f = 0; f < (int)elechits.size(); f++){
	  if( elechits[f].GetTrackID() == matchedElecID ){
	    matchedElecList = e;
	    break;
	  }
	  if(matchedElecList)
	    break;
	}//end loop over elec fls hits
	
	if(matchedElecList != -1){
	  const std::vector<sim::FLSHit> &elechits = elecFlsHandle->at(matchedElecList).fHits;	
	  int nElecFlsHits = elechits.size();
	  for(int el = 0; el < nElecFlsHits; el++)
	    newFLSList.fHits.push_back(elechits[el]);
	}
      }// end loop over elecflslists
      
      flsHits->push_back(newFLSList);
      
    }// end loop over flshit lists
    
    
    ////////////////////////////////////////////////////////////////
    
    int nPhots = photSigHandle->size();
    for( int l = 0; l < nPhots; ++l){
      sim::PhotonSignal phot = (*photSigHandle)[l];
      bool photIsFromMuon = false;
      int photID = phot.TrackId();
      int motherID = bt->TrackIDToMotherParticle( photID )->TrackId() ;

      if( (muonID.find(motherID) != muonID.end()) || (muonID.find(photID) != muonID.end()) ) 
	photIsFromMuon = true;
      
      if(!photIsFromMuon)
	photonSignal->push_back(phot);
   
    }// end loop over photons
    
    int nElecPhots = elecSigHandle->size();
    for( int e = 0; e < nElecPhots; e++){
      sim::PhotonSignal phot = (*elecSigHandle)[e];
      photonSignal->push_back(phot);
    }
    
    
    ////////////////////////////////////////////////////////////////

    
    for( size_t l = 0; l < plistHandle->size(); ++l){
      sim::Particle part = (*plistHandle)[l];
      bool partIsFromMuon = false;
      
      int partID = part.TrackId();
      int motherID = part.Mother();

      if( (muonID.find(motherID) != muonID.end()) || (muonID.find(partID) != muonID.end()) ) {
        partIsFromMuon = true;
      }

      if(!partIsFromMuon)
        particleList->push_back(part);
    }// end loop over photons

    for( int e = 0; e < (int)elecPlistHandle->size(); e++){
      sim::Particle elec = (*elecPlistHandle)[e];
      particleList->push_back(elec);
    }


    ////////////////////////////////////////////////////////////////

    // Write the objects to the events

    evt.put(std::move(flsHits));
    evt.put(std::move(photonSignal));
    evt.put(std::move(particleList));

    return;

  } // End of producer


  DEFINE_ART_MODULE(MergeTruth);


/*********************************************************************************************/
////////////////////////////////////////////////////////////////////////

}//end of namespace murem
