///////////////////////////////////////////////////////////////////////                                                      
/// \file  MuonRemoveAna_module.cc                                                                                          
/// \brief A module to assess the quality of muon removal
///                                                                                                                          
/// \version $ Id: MuonRemoveAna_module.cc, v 1.11 2012-10-26 22:19:03 ksachdev Exp $                                        
/// \author  Kanika Sachdev (ksachdev@physics.umn.edu)                                                                      
////////////////////////////////////////////////////////////////////////                                                    
 

#include "SimulationBase/MCTruth.h"
#include "SimulationBase/MCNeutrino.h"
#include "MuonRemove/art/TrackCleanUpAlg.h"
#include "MCCheater/BackTracker.h"
#include "RecoBase/CellHit.h"
#include "RecoBase/RecoHit.h"
#include "RecoBase/Track.h"
#include "RecoBase/Cluster.h"


#include <art/Framework/Principal/Handle.h>
#include <art/Framework/Principal/Event.h>
#include "art/Framework/Core/FindManyP.h"
#include <art/Framework/Core/ModuleMacros.h>
#include <art/Framework/Core/EDAnalyzer.h>
#include <art/Framework/Services/Registry/ServiceHandle.h>
#include <art/Framework/Services/Optional/TFileService.h>
#include <art/Framework/Services/Optional/TFileDirectory.h>
#include <fhiclcpp/ParameterSet.h>
#include <messagefacility/MessageLogger/MessageLogger.h>


#include <map>
#include <vector>
#include <cmath>
#include "TNtuple.h"
#include "TH1F.h"

namespace murem 
{

  class MuonRemoveAna : public art::EDAnalyzer 
  {

  public:
    explicit MuonRemoveAna(fhicl::ParameterSet const& pset);
    virtual ~MuonRemoveAna();

    void analyze(art::Event const& evt);
    void beginJob();
    void endJob();
    void reconfigure(const fhicl::ParameterSet& p);

  private:

    TrackCleanUpAlg*    fTrkCleanUpAlg;
    fhicl::ParameterSet fTrkCleanUpPSet;

    int   fEvtId;
    int   fIsCC;
    float fLength;
    float fNTracks;
    float fCosZ;
    float fNHits;
    float fPePerHit;
    float fRmsX;
    float fRmsY;
    float fRmsZ;
    float fMuFracRemaining;
    float fHadFracRemoved;
    float fTotMuonEnergy;
    float fTotHadEnergy;
    float fTrueSharedEnergy;
    float fRecoSharedEnergy;
    float fTrackEff;
    float fTrackPur;

    std::vector<float> fMuonHitMip;
    std::vector<float> fMuonHitGeV;

    TTree* fSliceTree;
    TTree* fMrccSliceTree;
    TH1F*  fdedxmu;
    TH1F*  fdedxsh;
    TH1F*  fpurl;
    TH1F*  fpurh;
    TH1F*  feffmuh;
    TH1F*  feffmul;
    TH1F*  feffhadl;
    TH1F*  feffhadh;

    std::string fMCTruthLabel;
    std::string fTrackLabel;
    std::string fMRTrackLabel;
    std::string fSliceLabel;
    std::string fMRSliceLabel;
    std::string fMrccLabel;

    art::ServiceHandle<geo::Geometry> fgeom;

  };

}
namespace murem 
{

  /*********************************************************************************************/
  ////////////////////////////////////////////////////////////////////////
  
  
  MuonRemoveAna::MuonRemoveAna(fhicl::ParameterSet const& p)
    : EDAnalyzer(p),
    fTrkCleanUpAlg(0),
    fTrkCleanUpPSet(p.get< fhicl::ParameterSet >("TrkCleanUpPSet") )
  {
    this->reconfigure(p);
  }
  

  /*********************************************************************************************/
  ////////////////////////////////////////////////////////////////////////
  
  MuonRemoveAna::~MuonRemoveAna() 
  { 
    if(fTrkCleanUpAlg) delete fTrkCleanUpAlg;
  }


  /*********************************************************************************************/
  ////////////////////////////////////////////////////////////////////////
  

  void MuonRemoveAna::reconfigure(const fhicl::ParameterSet& p)
  {
    fTrackLabel    = p.get< std::string>("TrackLabel");
    fMRTrackLabel  = p.get< std::string>("MRTrackLabel");
    fMCTruthLabel  = p.get< std::string>("fMCTruthLabel");
    fSliceLabel    = p.get< std::string>("SliceLabel");
    fMRSliceLabel  = p.get< std::string>("MRSliceLabel");
    fMrccLabel     = p.get< std::string>("MrccLabel");
  }
  
  
  /*********************************************************************************************/
  ////////////////////////////////////////////////////////////////////////

  void MuonRemoveAna::beginJob()
  {
    
    art::ServiceHandle<art:: TFileService> tfs;

    fSliceTree    = tfs->make<TTree>("fSliceTree","fSliceTree");

    fSliceTree->Branch("event",           &fEvtId);
    fSliceTree->Branch("isCC",            &fIsCC);
    fSliceTree->Branch("trackLength",     &fLength);
    fSliceTree->Branch("cosZ",            &fCosZ);
    fSliceTree->Branch("nTracks",         &fNTracks);
    fSliceTree->Branch("trackEff",        &fTrackEff);
    fSliceTree->Branch("trackPur",        &fTrackPur);
    fSliceTree->Branch("nHits",           &fNHits);
    fSliceTree->Branch("pePerHit",        &fPePerHit);
    fSliceTree->Branch("rmsX",            &fRmsX);
    fSliceTree->Branch("rmsY",            &fRmsY);
    fSliceTree->Branch("rmsZ",            &fRmsZ);
    fSliceTree->Branch("hadEnergy",       &fTotHadEnergy);
    fSliceTree->Branch("hadFracRemoved",  &fHadFracRemoved);
    fSliceTree->Branch("muEnergy",        &fTotMuonEnergy);
    fSliceTree->Branch("muFracRemaining", &fMuFracRemaining);
    fSliceTree->Branch("trueSharedEnergy",&fTrueSharedEnergy);
    fSliceTree->Branch("tecoSharedEnergy",&fRecoSharedEnergy);
    fSliceTree->Branch("muonHitMip",      &fMuonHitMip);
    fSliceTree->Branch("muonHitGev",      &fMuonHitGeV);


    fMrccSliceTree    = tfs->make<TTree>("fMrccSliceTree","fMrccSliceTree");

    fMrccSliceTree->Branch("event",           &fEvtId);
    fMrccSliceTree->Branch("isCC",            &fIsCC);
    fMrccSliceTree->Branch("trackLength",     &fLength);
    fMrccSliceTree->Branch("cosZ",            &fCosZ);
    fMrccSliceTree->Branch("nTracks",         &fNTracks);
    fMrccSliceTree->Branch("nHits",           &fNHits);
    fMrccSliceTree->Branch("pePerHit",        &fPePerHit);
    fMrccSliceTree->Branch("rmsX",            &fRmsX);
    fMrccSliceTree->Branch("rmsY",            &fRmsY);
    fMrccSliceTree->Branch("rmsZ",            &fRmsZ);
    
    fdedxmu = tfs->make<TH1F>("dedxmu", "dE/dx in a Plane; dE/dx per Plane [GeV/cm]; ", 200,0,0.005);
    fdedxsh = tfs->make<TH1F>("dedxsh", "dE/dx in a Plane; dE/dx per Plane [GeV/cm]; ", 200,0,0.005);

    fpurl = tfs->make<TH1F>("purl", "Purity Vs Lower Bound Value; dE/dx per Plane [GeV/cm]; Purity", 200,0,0.005);
    fpurh = tfs->make<TH1F>("purh", "Purity Vs Upper Bound Value; dE/dx per Plane [GeV/cm]; Purity", 200,0,0.005);

    feffmuh = tfs->make<TH1F>("effmuh", "Signal Efficiency Vs Upper Bound Value; dE/dx per Plane [GeV/cm]; Efficiency", 200,0,0.005);
    feffmul = tfs->make<TH1F>("effmul", "Signal Efficiency Vs Lower Bound Value; dE/dx per Plane [GeV/cm]; Efficiency", 200,0,0.005);

    feffhadl = tfs->make<TH1F>("effhadl", "Background Efficiency Vs Lower Bound Value; dE/dx per Plane [GeV/cm]; Efficiency", 200,0,0.005);
    feffhadh = tfs->make<TH1F>("effhadh", "Background Efficiency Vs Upper Bound Value; dE/dx per Plane [GeV/cm]; Efficiency", 200,0,0.005);
  }
  
  /*********************************************************************************************/
  ////////////////////////////////////////////////////////////////////////
  
  void MuonRemoveAna::analyze(art::Event const& evt)
  {
    
    fTrkCleanUpAlg = new TrackCleanUpAlg(fTrkCleanUpPSet);
    
    art::ServiceHandle<cheat::BackTracker>   bt;
    fEvtId = evt.id().event();

    fIsCC            = -5;
    fLength          = -5;
    fNTracks         = -5;
    fNHits           = -5;
    fPePerHit        = -5;
    fRmsX            = -5;
    fRmsY            = -5;
    fRmsZ            = -5;
    fMuFracRemaining = -5;
    fHadFracRemoved  = -5;
    fTotMuonEnergy   = -5;
    fTotHadEnergy    = -5;
    fTrueSharedEnergy= -5;
    fRecoSharedEnergy= -5;
    fTrackEff        = -5;
    fTrackPur        = -5;
    // Grab all slices 
    
    art::Handle< std::vector<rb::Cluster > > slicecol;
    evt.getByLabel( fSliceLabel, slicecol);
    art::PtrVector<rb::Cluster > slices;
    
    for(int i = 0; i < (int)slicecol->size(); i++){
      art::Ptr<rb::Cluster> sli(slicecol, i);
      slices.push_back(sli);
    }
    art::FindManyP<rb::Track> fmt(slicecol, evt, fTrackLabel);

    
    // Grab all mrcc slices                                                                                               
    
    art::Handle< std::vector<rb::Cluster > > mrccslicecol;
    evt.getByLabel( fMRSliceLabel, mrccslicecol);
    art::PtrVector<rb::Cluster > mrccslices;
    
    for(int i = 0; i < (int)mrccslicecol->size(); i++){
      art::Ptr<rb::Cluster> sli(mrccslicecol, i);
      mrccslices.push_back(sli);
    }
    //    art::FindManyP<rb::Track> fmtmrcc(mrccslicecol, evt, fMRTrackLabel);

    
    // Make a vector of all cellhits. It is needed for backtracker functions.

    art::PtrVector< rb::CellHit>  allHits;
    for(int s = 0; s<(int)slices.size(); s++){
      for(int c = 0; c<(int)slices[s]->NCell(); c++)
	allHits.push_back( slices[s]->Cell(c) );
    }


    // Make a vector of all mrcc cellhits. It is needed for backtracker functions.

    art::PtrVector< rb::CellHit>  mrccallHits;
    for(int s = 0; s<(int)mrccslices.size(); s++){
      for(int c = 0; c<(int)mrccslices[s]->NCell(); c++)
	mrccallHits.push_back( mrccslices[s]->Cell(c) );
    }

    // Get a list of removed muons.

    art::Handle< std::vector<simb::MCParticle> > removedMuonCol;
    evt.getByLabel( fMrccLabel, removedMuonCol);
    art::PtrVector<simb::MCParticle > removedMuons;

    for(int i = 0; i < (int)removedMuonCol->size(); i++){
      art::Ptr<simb::MCParticle> mu( removedMuonCol, i );
      removedMuons.push_back( mu );
    }

    // Fill slice level variables for mrcc slices.
    int nMrccSlices = mrccslices.size();
    for( int iSli = 0; iSli < nMrccSlices; iSli++){

      if(!mrccslices[iSli]->IsNoise()) {
	
	std::vector< cheat::NeutrinoEffPur > nuEffPur = 
	  bt->SliceToNeutrinoInteractions(mrccslices[iSli], mrccallHits);
	
	if (nuEffPur.size() > 0 && nuEffPur[0].purity>0.6){
	
	  const simb::MCNeutrino& mcNu = nuEffPur[0].neutrinoInt->GetNeutrino();
	  if ( abs(mcNu.Nu().PdgCode()) == 14 ){
	  
	    fIsCC  = mcNu.CCNC()==simb::kCC;
	    fNHits = mrccslices[iSli]->NCell();

	    //	    std::vector< art::Ptr<rb::Track> > mrcctracks = fmtmrcc.at(iSli);

	    //	    fNTracks = mrcctracks.size();
	    // rb::Track longestTrack;
	    // double maxLength = -1;
	    // double length;
	    
	    // int nMrccTracks = mrcctracks.size();
	    // for( int iTrk = 0; iTrk < nMrccTracks; iTrk++){
	    //   length = mrcctracks[iTrk]->TotalLength();
	    //   if( length > maxLength ){
	    // 	maxLength = length;
	    // 	longestTrack = *mrcctracks[iTrk];
	    //   }
	    // }

	    // fCosZ = longestTrack.Dir().Z();
	    // fLength = longestTrack.TotalLength();
	    
	  }// neutrino is muon type
	  
	}// can match to a neutrino interaction 
      }// non-noise slice 
    }// end loop over mrc slices
	
    
    fIsCC            = -5;
    fLength          = -5;
    fNTracks         = -5;
    fNHits           = -5;
    fPePerHit        = -5;
    fRmsX            = -5;
    fRmsY            = -5;
    fRmsZ            = -5;
    
 
    // Fill tree for original slices
    
    int nSlices = slices.size();
    for( int iSli = 0; iSli < nSlices; iSli++){

      if(!slices[iSli]->IsNoise()) {
	std::vector< cheat::NeutrinoEffPur > nuEffPur = bt->SliceToNeutrinoInteractions(slices[iSli], allHits);

	if (nuEffPur.size() > 0 && nuEffPur[0].purity > 0.6){

	  const simb::MCNeutrino& mcNu = nuEffPur[0].neutrinoInt->GetNeutrino();
	  if ( mcNu.Nu().PdgCode() == 14 ){
	    
	    fIsCC  = mcNu.CCNC()==simb::kCC;
	    fNHits = slices[iSli]->NCell();
	    std::vector< art::Ptr<rb::Track> > tracks = fmt.at(iSli);
	    
	    fNTracks = tracks.size();
	    rb::Track longestTrack;
	    double maxLength = -1;
	    double length;
	    
	    int nTracks = tracks.size();
	    for( int iTrk = 0; iTrk < nTracks; iTrk++){
	      length = tracks[iTrk]->TotalLength();
	      if( length > maxLength ){
		maxLength = length;
		longestTrack = *tracks[iTrk];
	      }
	    }
	    
	    fCosZ = longestTrack.Dir().Z();
	    fLength = longestTrack.TotalLength();

	    // Code to calculate the muon fraction remaining and hadronic fraction removed
	    // Also, to compare the true shared energy with what the TrackCleanUpAlg calls 
	    // shared energy, we need to find the track that was removed.

	    if (fIsCC){

	      std::vector< const sim::Particle*> particles = bt->MCTruthToParticles( nuEffPur[0].neutrinoInt );

	      int muMotherId;
	      int michelElID = -1;
	      std::set<int> primMuTrackIDsSet;
	      
	      for( int p = 0; p < (int)particles.size(); p++)
		if( abs(particles[p]->PdgCode()) == 13 &&
		    (particles[p]->Process() == "Primary" ||
		     particles[p]->Process() == "primary")){
		  muMotherId = particles[p]->TrackId();
		  primMuTrackIDsSet.insert(muMotherId);
		}
		  
	      for( int p = 0; p < (int)particles.size(); p++){
		if( particles[p]->Mother() == muMotherId ){

		  if( abs(particles[p]->PdgCode()) == 11 && 
		      (particles[p]->Process() == "muMinusCaptureAtRest" || 
		       particles[p]->Process() == "Decay")){
		    michelElID = particles[p]->TrackId();

		  }
		  else{
		    primMuTrackIDsSet.insert( particles[p]->TrackId() );
		  }
		}
	      }// end loop over particles associated with this MCTruth
	      

	      // Look for the track that was removed by MuonRemove

	      rb::Track muonTrack;
	      for( int t = 0; t< (int)tracks.size(); t++){
		if( !tracks[t]->Is3D() || tracks[t]->NCell()==0 )
		  continue;
		TVector3 trkStart=  tracks[t]->Start();
		TVector3 trkStop =  tracks[t]->Stop();

		for(int mu = 0; mu < (int)removedMuons.size(); mu++){

		  if( removedMuons[mu]->Position(0).Vect() == trkStart && 
		      removedMuons[mu]->EndPosition().Vect() == trkStop ){
		    muonTrack = *(tracks[t]);

		  }
		}
	      }// end loop over tracks                        


	      float totEnergy  = 0;
	      float muonEnergy = 0;
	      
	      int nMuCells = muonTrack.NCell();
	      
	      // vectors to store the indices for planes that have
	      // only muon energy in them, And those that have ha-
	      // -dronic contamination
	      std::vector<int> pureMuPlanes;
	      double totEInCell=0, muEInCell=0;

	      for( int hit = 0; hit < nMuCells; hit++ ){
		
		std::vector< cheat::TrackIDE > flshitsID = 
		  bt->HitToTrackIDE( muonTrack.Cell(hit) );

		for( int f = 0; f < (int)flshitsID.size(); f++){
		  totEnergy += flshitsID[f].energy;
		  totEInCell = flshitsID[f].energy;
		  
		  if( primMuTrackIDsSet.find( flshitsID[f].trackID) != primMuTrackIDsSet.end() ){
		    muonEnergy += flshitsID[f].energy;
		    muEInCell   = flshitsID[f].energy;
		  }
		    
		}//end loop over flh hits in cell

		if( muEInCell == totEInCell ){
		  int plane = muonTrack.Cell(hit)->Plane();
		  if( std::find( pureMuPlanes.begin(), pureMuPlanes.end(), plane ) == pureMuPlanes.end())
		    pureMuPlanes.push_back( plane);
		}
	      }// end loop over muon hits

	      fTrueSharedEnergy = totEnergy - muonEnergy;
	      if(muonTrack.NCell()>0){
		fRecoSharedEnergy = fTrkCleanUpAlg->ExtraEOnTrack( muonTrack, *slices[iSli] );
		std::map< unsigned int, double > dedxinplane = fTrkCleanUpAlg->DeDxInPlane( muonTrack );
		std::map< unsigned int, double >::iterator iter;
		for( iter = dedxinplane.begin(); iter != dedxinplane.end(); ++iter){
		  if( std::find( pureMuPlanes.begin(), pureMuPlanes.end(), iter->first) != pureMuPlanes.end() )
		    fdedxmu->Fill( iter->second );
		  else
		    fdedxsh->Fill( iter->second );
		}

	      }
	      float totMichelEnergy  = 0;
	      float totSliceEnergy   = 0;	  


	      fTotMuonEnergy   = 0;
	      fTotHadEnergy    = 0;
	      fMuFracRemaining = 0;
	      fHadFracRemoved  = 0;
	      
	      int nCells = slices[iSli]->NCell();
	      for( int c = 0; c < nCells; c++){
	      
		if(!bt->IsNoise( slices[iSli]->Cell(c) ) ){

		  muonEnergy = 0;
		  totEnergy  = 0;
		  float hadEnergy  = 0;
		  float michelE    = 0;
		  bool hasMuon     = false;
		  
		  std::vector< cheat::TrackIDE> flshitsID = 
		    bt->HitToTrackIDE( slices[iSli]->Cell(c) );

		  // Check if this is a muon hit                                                                            
		  
		  for( int f = 0; f < (int)flshitsID.size(); f++){
		    totEnergy += flshitsID[f].energy;
		    
		    const int testId = flshitsID[f].trackID;
		    if( primMuTrackIDsSet.find( testId ) != primMuTrackIDsSet.end() ){
		      muonEnergy += flshitsID[f].energy;
		      hasMuon = true;
		    }
		    
		    if( flshitsID[f].energy == michelElID)
		      michelE += flshitsID[f].energy;

		  }// end loop over flshits in this cellhit      
		 

		  totEnergy        = totEnergy - michelE;
		  totMichelEnergy += michelE;
		  totSliceEnergy  += totEnergy;

		  //		  if(hasMuon)
		  //		    fTrueSharedEnergy += totEnergy - muonEnergy;

		  if( totEnergy == muonEnergy ){
		    rb::RecoHit rHit = slices[iSli]->RecoHit(c);
		    if( rHit.IsCalibrated() ){
		      fMuonHitMip.push_back( rHit.MIP() );
		      fMuonHitGeV.push_back( rHit.GeV() );
		    }
		  }
		  // See if this cell has a hit in the mrcc-cellhits.                                                       
       
		  bool foundHit    = false;
		  rb::CellHit matchedHit;

		  int nMrccHits = mrccallHits.size();
		  for( int k = 0; k < nMrccHits; k++){
		    if( slices[iSli]->Cell(c)->Plane() == mrccallHits[k]->Plane() &&
			slices[iSli]->Cell(c)->Cell()  == mrccallHits[k]->Cell()  &&
			slices[iSli]->Cell(c)->TDC()   == mrccallHits[k]->TDC() ){
		      foundHit = true;
		      matchedHit = *(mrccallHits[k]);
		      break;
		    }
		  }// end loop over mrcc cellhits                
		    
		  float muadcRatio = 0;
		    
		  if(hasMuon){
		
		    if(foundHit)
		      muadcRatio = (float)( slices[iSli]->Cell(c)->ADC(0) 
					    - matchedHit.ADC(0) )/slices[iSli]->Cell(c)->ADC(0);
		    else
		      muadcRatio = 1; 
		  
		    fMuFracRemaining += (muonEnergy - (totEnergy * muadcRatio));
		    fTotMuonEnergy += muonEnergy;
		  }

		  hadEnergy = totEnergy - muonEnergy;
		
		  if(hadEnergy != 0){
		    float adcRatio = 0;
		    if(foundHit)
		      adcRatio = (float)matchedHit.ADC(0)/slices[iSli]->Cell(c)->ADC(0);
		    else
		      adcRatio = 0;
		    fHadFracRemoved += hadEnergy - (totEnergy * adcRatio);
		    fTotHadEnergy += hadEnergy;
		  }
		
		}// cellhit is not noise
	      }// end loop over slice cellhits

	      std::cout<<"truesharedE "<<fTrueSharedEnergy<<" recosharedE "<<fRecoSharedEnergy<<"\n";
	      if( totMichelEnergy/totSliceEnergy >= 0.9 || fTotMuonEnergy==0)
		continue;

	      if(fTotMuonEnergy != 0 )
		fMuFracRemaining = fMuFracRemaining/fTotMuonEnergy;
	    
	      if(fTotHadEnergy != 0 )
		fHadFracRemoved = fHadFracRemoved/fTotHadEnergy;
	    	    
	      std::cout<<" tot mu energy "<<fTotMuonEnergy<<" tot had energy "<<fTotHadEnergy<<std::endl;
	      std::cout<<" mu rem frac "<<fMuFracRemaining<<" had frac rem "<<fHadFracRemoved<<std::endl;

	      
	      fTrackEff = bt->HitCollectionEfficiency( primMuTrackIDsSet, muonTrack.AllCells(), allHits,  geo::kXorY);
	      fTrackPur = bt->HitCollectionPurity( primMuTrackIDsSet, muonTrack.AllCells(), NULL, NULL, true );

	    }// interaction is CC
	  
	    fSliceTree->Fill();

	  }// neutrino is muon type 
	}// can match to a neutrino interaction
      }// non-noise slices
    }// end loop over slices
    

      
  }// End of analyser
  
  
  /*********************************************************************************************/
  ////////////////////////////////////////////////////////////////////////
  
  void MuonRemoveAna::endJob()
  {
    int nbins = fdedxmu->GetNbinsX();
    float integralMu  = fdedxmu->Integral(1, nbins+1);
    float integralHad = fdedxsh->Integral(1, nbins+1);

    for( int  ibin = 1; ibin <= nbins; ++ibin){
      float leftIntegralMu   = fdedxmu->Integral(1, ibin);
      float rightIntegralMu  = fdedxmu->Integral(ibin, nbins+1);
      float leftIntegralHad  = fdedxsh->Integral(1, ibin);
      float rightIntegralHad = fdedxsh->Integral(ibin, nbins+1);

      
      float purHighCut    = 0;
      if( (leftIntegralMu + leftIntegralHad) != 0)
	purHighCut = 1 - ( leftIntegralHad/ (leftIntegralMu+leftIntegralHad) );

      float purLowCut     = 0;
      if( (rightIntegralMu + rightIntegralHad) != 0)
	purLowCut = 1 - ( rightIntegralHad/ rightIntegralMu+rightIntegralHad);

      float effHighCutMu  = leftIntegralMu / integralMu;
      float effLowCutMu   = rightIntegralMu / integralMu;
      float effHighCutHad = rightIntegralHad / integralHad;
      float effLowCutHad  = leftIntegralHad / integralHad;

      float x = fdedxmu->GetBinCenter(ibin);
      fpurh->Fill( x, purHighCut);
      fpurl->Fill( x, purLowCut);
      feffmuh->Fill( x, effHighCutMu);
      feffmul->Fill( x, effLowCutMu);
      feffhadh->Fill( x, effHighCutHad);
      feffhadl->Fill( x, effLowCutHad);
    }// end loop over bins
  }
  
  /*********************************************************************************************/
  ////////////////////////////////////////////////////////////////////////

  DEFINE_ART_MODULE(MuonRemoveAna);    

}// End of namespace


/*********************************************************************************************/
////////////////////////////////////////////////////////////////////////


 


