
///////////////////////////////////////////////////////////////////////////
// \file    MergeG4Collections.cxx
// \brief
// \authors janow@fnal.gov jmusser@indiana.edu
///////////////////////////////////////////////////////////////////////////  

#include <math.h>
#include <string>

#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "fhiclcpp/ParameterSet.h"
#include "messagefacility/MessageLogger/MessageLogger.h"
#include "art/Framework/Core/EDProducer.h"
#include "art/Framework/Core/ModuleMacros.h"


#include "DAQChannelMap/DAQChannelMap.h"
#include "SimulationBase/MCFlux.h"
#include "SimulationBase/MCTruth.h"
#include "SimulationBase/GTruth.h"


#include "Utilities/AssociationUtil.h"
#include "Simulation/FLSHit.h"
#include "Simulation/FLSHitList.h"
#include "Simulation/Particle.h"

#include "art/Framework/Core/FindOneP.h"

///group hits according to time and space
namespace MergeG4Collections {
  
  class MergeG4Collections : public art::EDProducer {
  public:
    explicit MergeG4Collections(fhicl::ParameterSet const &pset);
    virtual ~MergeG4Collections();

    void produce(art::Event& evt);

  protected:
    // internal methods

    // configuration settings
    std::string  fFirstGenCollection;     /// Label on the main genie collection
    std::string  fFirstGenProcessLabel;     /// Label on the main generator process
    std::string  fFirstG4Collection;     /// Label on the main G4 collection
    std::string  fSecondCollection;     /// Label on the File Mixer collection
    std::string  fThirdCollection;     /// Label on the MergeGen collection
    int fIsFirstCollectionCosmic; 

  };


  //----------------------------------------------------------------------
  MergeG4Collections::MergeG4Collections(fhicl::ParameterSet const &pset) :
    fFirstGenCollection         (pset.get< std::string >("FirstGenCollection")),
    fFirstGenProcessLabel       (pset.get< std::string >("FirstGenProcessLabel")),
    fFirstG4Collection          (pset.get< std::string >("FirstG4Collection")),
    fSecondCollection           (pset.get< std::string >("SecondCollection")),
    fThirdCollection            (pset.get< std::string >("ThirdCollection")),
    fIsFirstCollectionCosmic    (pset.get< int         >("IsFirstCollectionCosmic"))

     { 
       produces< std::vector<sim::FLSHitList >                >();
       produces< std::vector<sim::Particle>                   >();
       produces< art::Assns<sim::Particle, simb::MCTruth>     >();       
     }
  
  //----------------------------------------------------------------------
  MergeG4Collections::~MergeG4Collections()
  {
  }
  
  //----------------------------------------------------------------------
  void MergeG4Collections::produce(art::Event& evt)
  {

   std::unique_ptr< std::vector<simb::MCTruth> > truthcol (new std::vector<simb::MCTruth>);
   std::unique_ptr< art::Assns<sim::Particle, simb::MCTruth> > MCTruth_ParticleAssn(new art::Assns<sim::Particle, simb::MCTruth>);

   // Define the FLSHitList and Particle vectors.
   std::unique_ptr<std::vector<sim::FLSHitList>   >    finalFLSHitlistCol     (new std::vector<sim::FLSHitList>  );
   std::unique_ptr<std::vector<sim::Particle>     >    tempSecFileParticlelistCol (new std::vector<sim::Particle>    );
   std::unique_ptr<std::vector<sim::Particle>    >    finalParticlelistCol   (new std::vector<sim::Particle>);
   
   art::InputTag mclisttag1(fFirstGenCollection,"",fThirdCollection);
   art::InputTag mclisttag2(fFirstGenCollection,"",fFirstGenProcessLabel);
   
   // Pull the MC generator information out of the event
   art::Handle< std::vector<simb::MCTruth> > finalMCTruthlistCol;
   evt.getByLabel(mclisttag1, finalMCTruthlistCol);   
   art::Handle< std::vector<simb::MCTruth> > PrimFileMCTruthlistCol;
   evt.getByLabel(mclisttag2, PrimFileMCTruthlistCol);
   std::cout << " MCList sizes " << finalMCTruthlistCol->size() << " " << PrimFileMCTruthlistCol->size() << std::endl;

   //now get the G4 products
   art::Handle< std::vector<sim::FLSHitList> > PrimFileFLSHitlistCol;
   evt.getByLabel(fFirstG4Collection, PrimFileFLSHitlistCol);
   art::Handle< std::vector<sim::Particle> > PrimFileParticlelistCol;
   evt.getByLabel(fFirstG4Collection, PrimFileParticlelistCol);
   
   // push particles from primary collection onto merged collection.  Find largest TrackId
   // that will be used to seed trackID offsets in secondary file collections.
   // Particles are pushed onto finalParticlelistCol, which will hold combination of primary and secondary streams

   int IdOffset=0;   
   bool NewMCList = false;
   int MCListIndex = -1 ;
   art::FindOneP<simb::MCTruth> fo(PrimFileParticlelistCol, evt, fFirstG4Collection);
   for(unsigned int i = 0; i < PrimFileParticlelistCol->size(); ++i){
     sim::Particle part ((*PrimFileParticlelistCol)[i]); 
     if(part.Mother()!=0 )NewMCList=false;   
     if(part.Mother()==0 && !NewMCList ){   //particle with motherID==0 means new MCTruth 
       MCListIndex++; 
       NewMCList=true;
     }
     
     if(part.TrackId()>IdOffset)IdOffset=part.TrackId();
     for(int d=0; d<part.NumberDaughters(); d++){
       if(part.Daughter(d)>IdOffset)IdOffset=part.Daughter(d);
     }
     
     finalParticlelistCol->push_back(part);
     
     //here we find the MCTruth to associate with this particle. Since the primary file particle
     //list still has an intact MCTruth association, we can use a simple check on equality of 
     //a set of member variables to find this same MCTruth on the merged collection (finalMCTruthlistCol)
     
     art::Ptr<simb::MCTruth> mct = fo.at(i);   // this grabs associated MCTruth for this particle
     
     // For Cosmic Primary File, can't use the same matching procedure, based on GetNeutrino methods.
     // In Cosmic case, must assume 1 MCTruth per Particle
     if(! fIsFirstCollectionCosmic){
       bool found=false;
       for(unsigned int imc=0;imc<finalMCTruthlistCol->size();imc++){
	 art::Ptr<simb::MCTruth> mcptr(finalMCTruthlistCol,imc);
	 
	 if((mcptr->GetNeutrino().X()==mct->GetNeutrino().X()) && 
	    (mcptr->GetNeutrino().Y()==mct->GetNeutrino().Y()) && 
	    (mcptr->GetNeutrino().Nu().Vx()== mct->GetNeutrino().Nu().Vx()) &&
	    (mcptr->GetNeutrino().Nu().Vy()== mct->GetNeutrino().Nu().Vy()) &&
	    (mcptr->GetNeutrino().Nu().Vz()== mct->GetNeutrino().Nu().Vz())){
	    
	   // this is a check that the particle is on the MCTruth particle list for the MCTruth we've found
	   // only here to print warning in case of failure.
	   if(part.Mother()==0){
	     bool foundpart=false;
	     for(int ip=0;ip<mcptr->NParticles() ;ip++){
	       simb::MCParticle mcpart =  mcptr->GetParticle(ip);
	       float vpartsep = sqrt((part.Vx()-mcpart.Vx())*(part.Vx()-mcpart.Vx())+ 
				     (part.Vy()-mcpart.Vy())*(part.Vy()-mcpart.Vy())+
				     (part.Vz()-mcpart.Vz())*(part.Vz()-mcpart.Vz()));
	       if(vpartsep<0.01) foundpart=true;
	     }
	     if(!foundpart) std::cout << " !!! possible primary list particle and MCTruth position mismatch !!!  "<< std::endl;
	   }
	   // finally, make association
	   util::CreateAssn(*this, evt, *finalParticlelistCol, mcptr, *MCTruth_ParticleAssn, finalParticlelistCol->size()-1);
	   found=true;
	   break;
	 }
       }
       if(!found) std::cout << "MCTruth to primary collection particle not found " << std::endl;
     }
     else{ // here is case for cosmics association
       art::Ptr<simb::MCTruth> mcptr(finalMCTruthlistCol,MCListIndex);
       util::CreateAssn(*this, evt, *finalParticlelistCol, mcptr , *MCTruth_ParticleAssn, finalParticlelistCol->size()-1);
     }
   }

   std::cout << " MCListIndex after primary file processing " << MCListIndex << std::endl ;   
   if(PrimFileMCTruthlistCol->size()>0) MCListIndex=PrimFileMCTruthlistCol->size()-1;  // This sets initial MCListIndex for secondary files
   
   // Loop over neutrino interactions, and push fls hits from primary collection.
   // nothing else need by done on these, as trackIDs are unchanged
   for(unsigned int i = 0; i < PrimFileMCTruthlistCol->size(); ++i){     
     if(i<PrimFileFLSHitlistCol->size()){
       sim::FLSHitList mcfls ((*PrimFileFLSHitlistCol)[i]);
       finalFLSHitlistCol->push_back(mcfls);
     }
   }    
   
   // *** now onto secondary collections  ***
   art::Handle< std::vector<simb::MCTruth> > SecFileMCTruthlistCol;
   evt.getByLabel(fSecondCollection, SecFileMCTruthlistCol);
   
   art::Handle< std::vector<sim::FLSHitList> > SecFileFLSHitlistCol;
   evt.getByLabel(fSecondCollection, SecFileFLSHitlistCol);
   
   std::cout << " secondary MC, flshit list sizes " << SecFileMCTruthlistCol->size() << " " << SecFileFLSHitlistCol->size() << 
     " should be the same " << std::endl;
   if(SecFileMCTruthlistCol->size() != SecFileFLSHitlistCol->size() ) std::cout << " MC/FLSHit list size mismatch !!!" << std::endl;
   
   art::Handle< std::vector<sim::Particle> > SecFileParticlelistCol;
   evt.getByLabel(fSecondCollection, SecFileParticlelistCol);
  
   int nParticle=0;
   int trackIDList[100000];   // used as check of FLShit list TrackID consistency
   int LastPartId = 0;
   int *Offset = new int[1000];   // holds trackID offsets, one per MCTruth
   int OffsetIndex=1;  
   
   for(int i=0;i<1000;i++){
     Offset[i]=0;
   }
   Offset[0]=IdOffset; // trackID offset for first secondary MCTruth is the largest trackID from primary

   int MCindex = MCListIndex;    
   if(MCindex<0) MCindex=0;
   bool first=true;

   // here we loop over all secondary files particles, finding the event boundaries
   // the MCTruth/Particle matching algorithm is based on looking for a particle in the 
   // MCTruth particle list at the same
   // vertex location as the particle from SecFileParticlelistCol.  If we don't find one, and the particle's motherID==0
   // we move to the next MCTruth entry.

   for(unsigned int i = 0; i < SecFileParticlelistCol->size(); ++i){ 
     sim::Particle part ((*SecFileParticlelistCol)[i]);  // local copy of this particle     
     art::Ptr<simb::MCTruth> mcptr(finalMCTruthlistCol,MCindex);  // grab MCTruth at current index
     
     bool foundpart=false;  // now check to see if this new particle is consistent with being associated with this MCTruth
     for(int ip=0;ip<mcptr->NParticles() ;ip++){
       simb::MCParticle mcpart =  mcptr->GetParticle(ip);
       float vpartsep = sqrt((part.Vx()-mcpart.Vx())*(part.Vx()-mcpart.Vx())+ 
			     (part.Vy()-mcpart.Vy())*(part.Vy()-mcpart.Vy())+
			     (part.Vz()-mcpart.Vz())*(part.Vz()-mcpart.Vz()));
       if(vpartsep<0.1) foundpart=true;
     }
     
     if((LastPartId >= part.TrackId()) || (part.Mother()==0 && !foundpart ) ){  // make check on MCTruth/Particle consistency
       if(!first){
	 Offset[OffsetIndex]=Offset[OffsetIndex-1] + LastPartId;   //if here, new particle must be start of next event
	 OffsetIndex++;
       }
       if(MCindex+1 < (int)finalMCTruthlistCol->size())MCindex++;  
     }
     first=false;  
     LastPartId=part.TrackId();
     
     // construct new particle with offset trackID and add to particle collection
     // we must also offset daughters, and motherID (if motherID is not zero)

     int partmother = 0;
     if (part.Mother()!=0)partmother = part.Mother() + Offset[OffsetIndex-1]; 
     
     sim::Particle gpart(part.TrackId()+ Offset[OffsetIndex-1],
			 part.PdgCode(),
			 part.Process(),
			 partmother,
			 part.Mass(),
			 part.StatusCode());
     
     gpart.SetGvtx(part.Gvx(), part.Gvy(),part.Gvz(),part.Gvt());
     gpart.SetPolarization(part.Polarization());
     gpart.SetWeight(part.Weight());
     gpart.SetRescatter(part.Rescatter());
     for(int d=0; d<part.NumberDaughters(); d++){
       gpart.AddDaughter(part.Daughter(d) + Offset[OffsetIndex-1]);
     }
     for(unsigned int t=0; t<part.NumberTrajectoryPoints(); t++){
       gpart.AddTrajectoryPoint(part.Position(t),part.Momentum(t));
     }
     tempSecFileParticlelistCol->push_back(gpart);

     trackIDList[nParticle]=part.TrackId()+Offset[OffsetIndex-1];  // this array is used just to check consistency of results for FLSHit matching later
     nParticle++;
   }

   // loop over secondary mctruth list, which has same size as FLSHitlist list.
   for(unsigned int i = 0; i < SecFileMCTruthlistCol->size(); ++i){       
     if(i<SecFileFLSHitlistCol->size()){   
	 
       sim::FLSHitList mcfls ((*SecFileFLSHitlistCol)[i]); 
       sim::FLSHitList tmcfls;	
       sim::FLSHit fHit;

       // add flshits to final list offsetting trackIDs       
       const std::vector<sim::FLSHit>& hits = mcfls.fHits; 
       for (unsigned int j=0; j<hits.size(); ++j){
	 fHit.Clear();
	 fHit=hits[j];
	 fHit.SetTrackId(hits[j].GetTrackID() + Offset[i]);
	 tmcfls.fHits.push_back(fHit);

         // next, we check that this FLSHit trackID is on the particle trackID list
         // failure just results in warning message being printed	 
	 bool foundpart=false;  
	 for(int ip=0;ip<nParticle;ip++){
	   if(hits[j].GetTrackID() + Offset[i] == trackIDList[ip]){
	     foundpart=true;
	     break;
	   }
	 }
	 if(!foundpart){
	   std::cout << " !!!  FLSHit problem " << i << " " << j << " " << hits[j].GetTrackID() << " " << Offset[i] << std::endl;
	 }
       } 
       finalFLSHitlistCol->push_back(tmcfls);       
     }
     else std::cout << " FLSHit List length - MCList length mismatch ! " << std::endl;
   }	  
   
   // Create Particlelist:MCTruth association
    
   LastPartId=0; 
   int partcount=0;
   first=true;
   // set MCList index to first secondary event location
   if(MCListIndex+1<(int)finalMCTruthlistCol->size()){
     MCListIndex++;
   }
   else{
     std::cout << " !!! caught MCListIndex MCTruth size mismatch !!! " << std::endl;
   }

   //repeat loop over particle list
   for(unsigned int j = 0; j < tempSecFileParticlelistCol->size(); ++j){
     sim::Particle part ((*tempSecFileParticlelistCol)[j]);
     sim::Particle oldpart ((*SecFileParticlelistCol)[j]);

     art::Ptr<simb::MCTruth> mcptr(finalMCTruthlistCol,MCListIndex);  //grab MCTruth at index location

     // like above check that the particle can be found on this MCTruth particle list. If not increment MCTruth index
     bool foundpart=false;
     for(int ip=0;ip<mcptr->NParticles() ;ip++){
       simb::MCParticle mcpart =  mcptr->GetParticle(ip);
       float vpartsep = sqrt((part.Vx()-mcpart.Vx())*(part.Vx()-mcpart.Vx())+ 
			     (part.Vy()-mcpart.Vy())*(part.Vy()-mcpart.Vy())+
			     (part.Vz()-mcpart.Vz())*(part.Vz()-mcpart.Vz()));
       if(vpartsep<0.1) foundpart=true;
     }
  
     if((LastPartId >= oldpart.TrackId()) || (part.Mother()==0 && !foundpart && !first )){ //check whether to increment MCTruth
       if(MCListIndex +1 < (int)finalMCTruthlistCol->size()) 
	 {
	   MCListIndex++;
	 }
       else{
	 std::cout << " !!! caught MCListIndex MCTruth size mismatch !!! " << std::endl;
       } 
       partcount=0;
     }
     // check position if mother =0 with updated list index.  Should always find a match - this is just a check
     art::Ptr<simb::MCTruth> mcptr2(finalMCTruthlistCol,MCListIndex);
     foundpart=false;
     for(int ip=0;ip<mcptr2->NParticles() ;ip++){
       simb::MCParticle mcpart =  mcptr2->GetParticle(ip);
       float vpartsep = sqrt((part.Vx()-mcpart.Vx())*(part.Vx()-mcpart.Vx())+ 
			     (part.Vy()-mcpart.Vy())*(part.Vy()-mcpart.Vy())+
			     (part.Vz()-mcpart.Vz())*(part.Vz()-mcpart.Vz()));
       if(vpartsep<0.1) foundpart=true;
     }

     if(part.Mother()==0 && !foundpart ) std::cout << " !!! secondary particle and MCTruth particle list mismatch !!!  "<< std::endl;
     first=false;
     LastPartId=oldpart.TrackId();

     finalParticlelistCol->push_back(part);  //push this particle onto final particle list and make association
     util::CreateAssn(*this, evt, *finalParticlelistCol, mcptr2 , *MCTruth_ParticleAssn, finalParticlelistCol->size()-1);
     partcount ++;
   }

   // put the collections in the event
   std::cout <<  " mclistindex " << MCListIndex << " " << finalMCTruthlistCol->size() << std::endl;   
   
   evt.put(std::move(finalFLSHitlistCol));
   evt.put(std::move(finalParticlelistCol));
   evt.put(std::move(MCTruth_ParticleAssn));
   delete [] Offset ;
   
  }
  
     DEFINE_ART_MODULE(MergeG4Collections);
} //end namespace merge
  /////////////////////////////////////
