/////////////////////////////////////////////////////////
/// \brief  Sandbox to make objects for numu analysis
///         in CAF
/// \author Kirk Bays
////////////////////////////////////////////////////////
#include <string>

#include "art/Framework/Core/EDProducer.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Core/FindManyP.h"
#include "art/Framework/Services/Optional/TFileService.h"
#include "fhiclcpp/ParameterSet.h"

#include "NovaDAQConventions/DAQConventions.h"
#include "NumuSandbox/NumuSandFxs.h"
#include "NumuSandbox/NumuSandObj.h"
#include "Geometry/Geometry.h"
#include "Geometry/LiveGeometry.h"
#include "GeometryObjects/CellGeo.h"
#include "RecoBase/Cluster.h"
#include "RecoBase/Track.h"
#include "RecoBase/CellHit.h"
#include "ReMId/ReMId.h"
#include "Utilities/AssociationUtil.h"

#include "MCCheater/BackTracker.h"
#include "TFile.h"
#include "TMVA/Reader.h"

namespace numusand {

  class FillSandbox : public art::EDProducer {
  public:
    explicit FillSandbox(fhicl::ParameterSet const & pset);
    virtual ~FillSandbox();

    void produce(art::Event & evt);
    void beginRun(art::Run& run);

  private:

    numusand::NumuSandFxs*     fNumuSandFxs;
    std::string                fSlicerLabel;
    std::string                fTrackLabel;
    std::string                fPIDLabel;

  };
}


namespace numusand
{

  //----------------------------------------------------------------------
  FillSandbox::FillSandbox(fhicl::ParameterSet const & pset):
    fSlicerLabel (pset.get< std::string >("SlicerLabel")),
    fTrackLabel  (pset.get< std::string >("TrackLabel")),
    fPIDLabel    (pset.get< std::string >("PIDLabel"))
  {
    produces< std::vector<numusand::NumuSandObj> >();
    produces< art::Assns<numusand::NumuSandObj, rb::Cluster> >();
  }


  //----------------------------------------------------------------------
  FillSandbox::~FillSandbox()
  {
  }

  //-------------------------------------------------------------------
  void FillSandbox::beginRun(art::Run& run)
  {
    fNumuSandFxs = new NumuSandFxs();
  } //end beginRun

  //-------------------------------------------------------------------
  void FillSandbox::produce(art::Event& evt)
  {

    art::Handle<std::vector<rb::Cluster> > slicevec;
    evt.getByLabel(fSlicerLabel,slicevec);

    if(slicevec->empty()) {
      mf::LogWarning ("No Slices")<<"No Slices in the input file";
      return;
    }

    art::ServiceHandle<cheat::BackTracker> bt;
    art::ServiceHandle<geo::Geometry>  geom;
    art::ServiceHandle<geo::LiveGeometry>  livegeom;

    art::FindManyP<rb::Track> trackAssnList(slicevec, evt, fTrackLabel);
    std::unique_ptr< std::vector<numusand::NumuSandObj> > outputObjects(new std::vector<numusand::NumuSandObj>);
    std::unique_ptr< art::Assns<numusand::NumuSandObj, rb::Cluster> > assoc(new art::Assns<numusand::NumuSandObj, rb::Cluster>);

    for(size_t i = 0; i < slicevec->size(); ++i){

      art::Ptr<rb::Cluster> slice(slicevec,i);

      if(slice->IsNoise()) continue;

      numusand::NumuSandObj sandbox; // initialization in constructor

      float missE;

      fNumuSandFxs->getMissingE(slice, missE);

      sandbox.SetMissingE(missE);

      sandbox.SetNDropped(livegeom->NDropouts());
      sandbox.SetNBad(livegeom->NBadDCMBC());

      float xmin = 9999, ymin = 9999, zmin = 9999, xmax = -9999, ymax = -9999, zmax = -9999;
      float xmin2 = 9999,  ymin2 = 9999,  zmin2 = 9999,  xmint = 9999,  ymint = 9999,  zmint = 9999;
      float xmax2 = -9999, ymax2 = -9999, zmax2 = -9999, xmaxt = -9999, ymaxt = -9999, zmaxt = -9999;
      int ntx = 0, nty = 0;

      for(unsigned int j = 0; j < slice->NCell(); ++j) {
        const rb::CellHit* h = slice->Cell(j).get();
        double xyz[3];
        geom->Plane(h->Plane())->Cell(h->Cell())->GetCenter(xyz);
        const double x = xyz[0];
        const double y = xyz[1];
        const double z = xyz[2];

        if(h->View()==geo::kX) {
          if(x < xmin) { xmin2 = xmin; xmin = x;}
          else if(x < xmin2) xmin2 = x;
          if(z < zmin) { zmin2 = zmin; zmin = z;}
          else if(z < zmin2) zmin2 = z;
          if(x > xmax) { xmax2 = xmax; xmax = x;}
          else if(x > xmax2) xmax2 = x;
          if(z > zmax) { zmax2 = zmax; zmax = z;}
          else if (z > zmax2) zmax2 = zmax;

          if(h->PE() > 100) { // above threshold only TODO: add tracks and recohits, use pecorr
            ntx++;
            if(x < xmint) xmint = x;
            if(z < zmint) zmint = z;

            if(x > xmaxt) xmaxt = x;
            if(z > zmaxt) zmaxt = z;
          }
        }

        else if(h->View()==geo::kY) {
          if(y < ymin) { ymin2 = ymin; ymin = y;}
          else if (y < ymin2) ymin2 = y;
          if(z < zmin) { zmin2 = zmin; zmin = z;}
          else if (z < zmin2) zmin2 = z;

          if(y > ymax) { ymax2 = ymax; ymax = y;}
          else if(y > ymax2) ymax2 = y;
          if(z > zmax) { zmax2 = zmax; zmax = z;}
          else if(z > zmax2) zmax2 = z;

          if(h->PE() > 100) { // above threshold only TODO: add tracks and recohits, use pecorr
            nty++;
            if(y < ymint) ymint = y;
            if(z < zmint) zmint = z;

            if(y > ymaxt) ymaxt = y;
            if(z > zmaxt) zmaxt = z;
          }
        }

      }
      sandbox.SetXMin2(xmin2);
      sandbox.SetYMin2(ymin2);
      sandbox.SetZMin2(zmin2);
      sandbox.SetXMax2(xmax2);
      sandbox.SetYMax2(ymax2);
      sandbox.SetZMax2(zmax2);

      if(ntx > 0 && nty > 0) {
        sandbox.SetXMint(xmint);
        sandbox.SetYMint(ymint);
        sandbox.SetZMint(zmint);
        sandbox.SetXMaxt(xmaxt);
        sandbox.SetYMaxt(ymaxt);
        sandbox.SetZMaxt(zmaxt);
      }

      else { // -5 default can be confused as a real value; +=9999 too far from real values
        sandbox.SetXMint(-999);
        sandbox.SetYMint(-999);
        sandbox.SetZMint(-999);
        sandbox.SetXMaxt(-999);
        sandbox.SetYMaxt(-999);
        sandbox.SetZMaxt(-999);
      }


      if(!trackAssnList.isValid()) {
        std::cout << "NumuSandbox: No Kalman Tracks!" << std::endl;
        outputObjects->push_back(sandbox);
        util::CreateAssn(*this,evt,*(outputObjects.get()),slice,*(assoc.get()));
        continue;
      }

      const std::vector< art::Ptr<rb::Track> > sliceTracks = trackAssnList.at(i);

      art::FindOneP<remid::ReMId> remidVec(sliceTracks, evt, fPIDLabel);
      if(!remidVec.isValid()) {
        //        std::cout << "NumuSandbox: No ReMId information!" << std::endl;
        outputObjects->push_back(sandbox);
        util::CreateAssn(*this,evt,*(outputObjects.get()),slice,*(assoc.get()));
        continue;
      }


      int bestIdx = -1;  // for index in list of the highest PID track
      int pimuIdx = -5;  // for index in list of a pimu decay event
      int Idx2 = -1;
      int idx3d = -1;
      double bestpimuE = -999.;  // keep track of how much energy due to a non-primary muon was in a track
                         // this is for if a pi->mu muon contributes light to more than once track
                         // we only want to associate with the track that saw 'most' of it.
                         // note we assume all non-primary muons are from pi->mu decay.
      double bestval = -999., val2 = -999;
      int nmutrks = 0;

      for(size_t j = 0; j < sliceTracks.size(); ++j){
        const art::Ptr<rb::Track> track = sliceTracks[j];
        art::Ptr<remid::ReMId> trackRemid = remidVec.at(j);

        if(!track->Is3D()) {
          continue;
        }

        idx3d++;

        if(trackRemid->Value() > bestval) { // best track so far
          val2 = bestval;
          Idx2 = bestIdx;
          bestval = trackRemid->Value();
          bestIdx = j;
        }
        else if(trackRemid->Value() > val2) { // not best, but better second best track
          val2 = trackRemid->Value();
          Idx2 = j;
        }

        if(trackRemid->Value() > 0.6) nmutrks++;

        if(bt->HaveTruthInfo()) {
          art::PtrVector<rb::CellHit> trkhits = track->AllCells();
          const std::vector<const sim::Particle*> parts = bt->HitsToParticle(trkhits);
          const std::vector< cheat::TrackIDE > trids = bt->HitsToTrackIDE(trkhits);

          // check particle that contributes most to track for pi->mu decay
          if( !parts.empty() && !trids.empty() ) {
	    if(abs(parts[0]->PdgCode())==13 && parts[0]->Process()=="Decay") {
	      double tempE = trids[0].energyFrac * track->TotalGeV();
	      if(tempE > bestpimuE) {
		bestpimuE = tempE;
		pimuIdx = idx3d+1;
	      }
	    }
	  }

          // check particle that contributes second most; track can contain both pi and mu
          if(parts.size() > 1 && trids.size() > 1) {
            if(abs(parts[1]->PdgCode())==13 && parts[1]->Process()=="Decay") {
              double tempE = trids[1].energyFrac * track->TotalGeV();
              if(tempE > bestpimuE) {
                bestpimuE = tempE;
                pimuIdx = idx3d+1;
              }
            }
          }
        }
      } //end loop over tracks

      sandbox.SetNMuTrks(nmutrks);

      if(bt->HaveTruthInfo()) {  // no truth info case should correspond to unfilled

        if(bestIdx!=-1) {

          const art::Ptr<rb::Track> track1 = sliceTracks[bestIdx];
          const std::vector<const sim::Particle*> parts1 = bt->HitsToParticle(track1->AllCells());
          const std::vector< cheat::TrackIDE > trids1 = bt->HitsToTrackIDE(track1->AllCells());

          if(!parts1.empty()){
            sandbox.SetPDGBest(parts1[0]->PdgCode());
          }
        }

        if(Idx2!=-1) {

          const art::Ptr<rb::Track> track2 = sliceTracks[Idx2];
          const std::vector<const sim::Particle*> parts2 = bt->HitsToParticle(track2->AllCells());
          const std::vector< cheat::TrackIDE > trids2 = bt->HitsToTrackIDE(track2->AllCells());

          if(!parts2.empty()){
            sandbox.SetPDGSecondBest(parts2[0]->PdgCode());
          }
        }

        int nprotons = 0;
        if(pimuIdx == -5) pimuIdx = 0;  // differentiate unfilled with not found
        art::PtrVector<rb::CellHit> slchits = slice->AllCells();
        const std::vector<const sim::Particle*> parts = bt->HitsToParticle(slchits);
        const std::vector< cheat::TrackIDE > trids = bt->HitsToTrackIDE(slchits);
        if(!parts.empty() && !trids.empty()) {
          for(unsigned int i = 0; i < parts.size(); i++) {
            if(trids.size() > i) {
              if(parts[i]->PdgCode()==2212 && trids[i].energyFrac * slice->TotalGeV() > 0.05) nprotons++;
            }
          }
        }
        sandbox.SetNProtons(nprotons);
      }

      sandbox.SetPiMuDecay(pimuIdx);

      outputObjects->push_back(sandbox);
      util::CreateAssn(*this,evt,*(outputObjects.get()),slice,*(assoc.get()));
    } //end loop over slices

    evt.put(std::move(outputObjects));
    evt.put(std::move(assoc));

  } //end produce

} //end namespace


DEFINE_ART_MODULE(numusand::FillSandbox)
