#include "NumuSandbox/NumuSandFxs.h"

#include "Calibrator/Calibrator.h"
#include "Geometry/Geometry.h"
#include "Geometry/LiveGeometry.h"
#include "GeometryObjects/CellGeo.h"
#include "GeometryObjects/Geo.h"
#include "NovaDAQConventions/DAQConventions.h"
#include "RecoBase/RecoHit.h"
#include "MCCheater/BackTracker.h"
#include "RecoBase/Cluster.h"

#include "TMath.h"
#include "TVector3.h"

#include "art/Framework/Services/Registry/ServiceHandle.h"

#define PI 3.14159265

namespace numusand
{

  //-------------------------------------------------------
  NumuSandFxs::NumuSandFxs() //constructor
  {
    fKGeo = new trk::KalmanGeoHelper(true);
  }

  //-------------------------------------------------------
  NumuSandFxs::~NumuSandFxs() //destructor
  {
  }

  //-------------------------------------------------------
  // getAngle: function to get the angle of the lepton wrt the incoming neutrino.
  // currently the incoming neutrino direction estimate is hardcoded.

  float NumuSandFxs::getAngle(TVector3 trackdir)
  {
    art::ServiceHandle<geo::Geometry> geom;

    TVector3 beamdir = geom->NuMIBeamDirection();
    float    ret     = trackdir*beamdir;

    return ret;
  }

  //---------------------------------------------------------------------------

  float NumuSandFxs::fitslope(const std::vector< double > &l, const std::vector< double > &t, const std::vector< double > &q, double &chisq, double &inter, bool linflag) {

    assert(t.size()==l.size());
    assert(t.size()>=2);

    double ztStart[2] = {0.};
    double ztEnd[2]   = {0.};
    if(linflag==false) {
      chisq = geo::LinFit(t, l, q,
                          &ztStart[0], &ztStart[1],
                          &ztEnd[0], &ztEnd[1]);
    }
    else if(linflag==true) {
      chisq = geo::LinFitMinDperp(t, l, q,
                                  &ztStart[0], &ztStart[1],
                                  &ztEnd[0], &ztEnd[1]);
    }
    const double ztSlope = (ztEnd[1]-ztStart[1])/(ztEnd[0]-ztStart[0]);

    inter = ztStart[1]-(ztSlope*ztStart[0]);

    return ztSlope;
  }

  //----------------------------------------------------------------------------------------

  float NumuSandFxs::fitchisq(const std::vector< double > &t, const std::vector< double > &l, int flag, double c) {

    assert(t.size()==l.size());
    assert(t.size()>=2);

    double chi2 = 0.0;
    double tave = 0;
    double lave = 0;

    for (unsigned int i=1; i<t.size(); ++i) {
      tave += t[i];
      lave += l[i];
    }

    double norm = t.size();

    tave = tave / norm;
    lave = lave / norm;

    double tstart = tave - 500;
    double tend = tave + 500;
    double lstart, lend;

    assert(flag == 0 || flag == 1);
    if(flag==0) {
      lstart = lave - (c * 500);
      lend = lave + (c * 500);
    }
    else {
      lstart = lave + (c * 500);
      lend = lave - (c * 500);
    }

    for (unsigned int i=0; i<t.size(); ++i) {
      chi2 += geo::DsqrToLine(t[i],l[i],tstart,lstart,tend,lend);
    }
    return chi2;
  }

  //----------------------------------------------------------------------------------

  float NumuSandFxs::getScatt(const art::Ptr<rb::Track> track) {
    double tscattmax=0, tscatttot=0, tscatt;
    TVector3 traj1, traj2, dir, olddir;

    for(unsigned int p=0;p<track->NTrajectoryPoints()-1;p++) {
      traj1 = track->TrajectoryPoint(p);
      traj2 = track->TrajectoryPoint(p+1);
      dir = (traj2-traj1).Unit();
      if((dir-olddir).Mag()<1e-7) continue;
      if(p>0&&dir.Mag()<=1.0&&olddir.Mag()<=1.0) {
        tscatt = std::acos(dir.Dot(olddir)) * 180 / PI;
        if(tscatt > tscattmax) tscattmax = tscatt;
        tscatttot += tscatt;
      }
      olddir = dir;
    }
    return tscatttot;  // calculated both maximum and total scatter, total seems more useful, ignore max for now
  }

  //-----------------------------------------------------------------------------------

  void NumuSandFxs::getActivity(const art::Ptr<rb::Track> track, double &startact, double &endact) {
    double xyz[3];
    TVector3 tstart = track->Start();
    TVector3 tend = track->Stop();
    art::ServiceHandle<geo::Geometry>  geom;
    art::ServiceHandle<calib::Calibrator> cal;

    startact = 0;
    endact = 0;

    for(unsigned int hitIdx = 0; hitIdx < track->NCell(); ++hitIdx){
      art::Ptr<rb::CellHit> chit = track->Cell(hitIdx);
      rb::RecoHit rhit = track->RecoHit(hitIdx);
      geom->Plane(chit->Plane())->Cell(chit->Cell())->GetCenter(xyz);
      if(chit->View()==geo::kX) { // ignore y position of cell center
        if(sqrt((tstart.X()-xyz[0])*(tstart.X()-xyz[0]) + (tstart.Z()-xyz[2])*(tstart.Z()-xyz[2])) < 50) startact+=rhit.GeV();
        if(sqrt((tend.X()-xyz[0])*(tend.X()-xyz[0]) + (tend.Z()-xyz[2])*(tend.Z()-xyz[2])) < 50) endact+=rhit.GeV();
      }

      if(chit->View()==geo::kY) { // ignore x position of cell center
        if(sqrt((tstart.Y()-xyz[1])*(tstart.Y()-xyz[1]) + (tstart.Z()-xyz[2])*(tstart.Z()-xyz[2])) < 50) startact+=rhit.GeV();
        if(sqrt((tend.Y()-xyz[1])*(tend.Y()-xyz[1]) + (tend.Z()-xyz[2])*(tend.Z()-xyz[2])) < 50) endact+=rhit.GeV();
      }
    }
  }

  //--------------------------------------------------------------------------- start get missing energy

  bool NumuSandFxs::getMissingE(const art::Ptr<rb::Cluster> slice, float &missE)
  {

    art::ServiceHandle<cheat::BackTracker> bt;

    if(!bt->HaveTruthInfo()) return false; // No Truth Info!

    art::ServiceHandle<geo::Geometry>  geom;
    art::ServiceHandle<geo::LiveGeometry>  livegeom;

    art::PtrVector<rb::CellHit> allhits = slice->AllCells();

    const std::vector<const sim::Particle*> parts = bt->HitsToParticle(allhits);
    //    const std::vector< cheat::TrackIDE > trids = bt->HitsToTrackIDE(allhits);

    float sum=0;

    if(parts.size()==0) return false;

    for(unsigned int i = 0; i < parts.size(); i++) { // loop over particles
      if(!(abs(parts[i]->PdgCode())==11 || abs(parts[i]->PdgCode())==13 || parts[i]->PdgCode()==2212 ||  abs(parts[i]->PdgCode())==211 ||  parts[i]->PdgCode()==22)) continue; // not interesting particle
      if(parts[i]->E() < 0.001) continue;  // ignore things < 1 MeV
      if(geom->DetId() == novadaq::cnv::kFARDET) {
        if(livegeom->IsPointLive(parts[i]->Position().Vect())==false || livegeom->IsPointLive(parts[i]->EndPosition().Vect())==false) sum+=parts[i]->E(); // not contained
      }
      else if(geom->DetId() == novadaq::cnv::kNEARDET) {
        if( (livegeom->IsPointLive(parts[i]->Position().Vect())==false    && livegeom->InMuonCatcher(parts[i]->Position().Vect())==false) || 
            (livegeom->IsPointLive(parts[i]->EndPosition().Vect())==false && livegeom->InMuonCatcher(parts[i]->EndPosition().Vect())==false) ) sum+=parts[i]->E(); // not contained
      }
    }

    missE = sum;
    return true;

  }

} //end namespace
