////////////////////////////////////////////////////////////////////////////
/// \file    LEMPresel_module.cc
/// \brief   Speed up LEM processing, mostly of FD data
/// \author  Christopher Backhouse - bckhouse@caltech.edu
///////////////////////////////////////////////////////////////////////////

#include "art/Framework/Core/EDProducer.h"
#include "art/Framework/Core/FindManyP.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "art/Persistency/Common/Ptr.h"
#include "fhiclcpp/ParameterSet.h"

#include "CosRej/NueCosRej.h"
#include "NovaDAQConventions/DAQConventions.h"
#include "RecoBase/Cluster.h"
#include "RecoBase/FilterList.h"
#include "RunHistory/service/DetectorService.h"

namespace presel
{
  /// Speed up LEM processing, mostly of FD data
  class LEMPresel: public art::EDProducer
  {
  public:
    explicit LEMPresel(const fhicl::ParameterSet& pset);
    virtual ~LEMPresel();
    virtual void produce(art::Event& evt) override;

  protected:
    bool PassesContainment(const cosrej::NueCosRej& cr) const;

    std::string fSlicerLabel;
    std::string fNueCosRejLabel;

    double fMinCaloE;
    double fMaxCaloE;

    double fEastDist;
    double fWestDist;
    double fTopDist;
    double fBottomDist;
    double fFrontDist;
    double fBackDist;
  };

  //----------------------------------------------------------------------
  LEMPresel::LEMPresel(const fhicl::ParameterSet& pset):
    fSlicerLabel   (pset.get<std::string>("SlicerLabel")),
    fNueCosRejLabel(pset.get<std::string>("NueCosRejLabel")),
    fMinCaloE  (pset.get<double>("MinCaloE")),
    fMaxCaloE  (pset.get<double>("MaxCaloE")),
    fEastDist  (pset.get<double>("EastDist")),
    fWestDist  (pset.get<double>("WestDist")),
    fTopDist   (pset.get<double>("TopDist")),
    fBottomDist(pset.get<double>("BottomDist")),
    fFrontDist (pset.get<double>("FrontDist")),
    fBackDist  (pset.get<double>("BackDist"))
  {
    produces<rb::FilterList<rb::Cluster>>();
  }

  //----------------------------------------------------------------------
  LEMPresel::~LEMPresel()
  {
  }

  //----------------------------------------------------------------------
  void LEMPresel::produce(art::Event& evt)
  {
    std::unique_ptr< rb::FilterList<rb::Cluster> > filtcol(new rb::FilterList<rb::Cluster>);

    art::ServiceHandle<rh::DetectorService> det;
    if(det->DetId() != novadaq::cnv::kFARDET){
      evt.put(std::move(filtcol));
      return;
    }

    art::Handle< std::vector<rb::Cluster> > slices;
    evt.getByLabel(fSlicerLabel, slices);

    art::FindManyP<cosrej::NueCosRej> fmcr(slices, evt, fNueCosRejLabel);

    const int sliceMax = slices->size();
    for(int sliceIdx = 0; sliceIdx < sliceMax; ++sliceIdx){
      const rb::Cluster& slice = (*slices)[sliceIdx];
      if(slice.IsNoise()) continue;

      const double caloE = slice.CalorimetricEnergy();
      // Make sure to test calE even if NueCosRej is unavailable
      if(caloE < fMinCaloE || (fMaxCaloE > 0 && caloE > fMaxCaloE)){
        filtcol->Add(slices, sliceIdx);
        // Once we've filtered the slice there's no need to check anything else
        continue;
      }

      std::vector< art::Ptr<cosrej::NueCosRej> > cosrejs = fmcr.at(sliceIdx);
      if(cosrejs.empty()) continue;
      assert(cosrejs.size() == 1);

      if(!PassesContainment(*cosrejs[0])){
        filtcol->Add(slices, sliceIdx);
      }
    } // end for sliceIdx

    evt.put(std::move(filtcol));
  }

  //----------------------------------------------------------------------
  bool LEMPresel::PassesContainment(const cosrej::NueCosRej& cr) const
  {
    if(cr.StartDistToEast()   < fEastDist   ||
       cr.StopDistToEast()    < fEastDist  ) return false;

    if(cr.StartDistToWest()   < fWestDist   ||
       cr.StopDistToWest()    < fWestDist  ) return false;

    if(cr.StartDistToTop()    < fTopDist    ||
       cr.StopDistToTop()     < fTopDist   ) return false;

    if(cr.StartDistToBottom() < fBottomDist ||
       cr.StopDistToBottom()  < fBottomDist) return false;

    if(cr.StartDistToFront()  < fFrontDist  ||
       cr.StopDistToFront()   < fFrontDist ) return false;

    if(cr.StartDistToBack()   < fBackDist   ||
       cr.StopDistToBack()    < fBackDist  ) return false;

    return true;
  }

  DEFINE_ART_MODULE(LEMPresel)
} // end namespace presel
