////////////////////////////////////////////////////////////////////////////
/// \file    NuePresel.cxx
/// \brief   Preselection to filter events for nue analysis, pre-PID
/// \author  Ruth Toner
///////////////////////////////////////////////////////////////////////////
//Novasoft includes:
#include "Calibrator/Calibrator.h"
#include "Geometry/Geometry.h"
#include "NovaDAQConventions/DAQConventions.h"
#include "Preselection/NuePreselObj.h"
#include "RecoBase/CellHit.h"
#include "RecoBase/RecoHit.h"
#include "RecoBase/Cluster.h"
#include "RecoBase/Prong.h"
#include "RecoBase/Track.h"
#include "RecoBase/Shower.h"
#include "RecoBase/FilterList.h"
#include "RecoBase/PID.h"
#include "SimulationBase/MCTruth.h"
#include "SimulationBase/MCNeutrino.h"
#include "Utilities/AssociationUtil.h"

// Framework includes
#include "art/Framework/Core/FindManyP.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Services/Optional/TFileService.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "art/Persistency/Common/Ptr.h"
#include "fhiclcpp/ParameterSet.h"


namespace presel
{
  class NuePresel: public art::EDProducer
  {
  public:
    
    explicit NuePresel(const fhicl::ParameterSet& pset);
    virtual ~NuePresel();
    void beginRun(art::Run& run);
    virtual void produce(art::Event& evt);

  protected:
    
    std::string fSlicerLabel;
    std::string fRecoType;
    std::string fAssn3D;
    std::string fAssn2D;
    bool   fDoLengthCut;
    bool   fDoEnergyCut;
    bool   fDoPercMipCut;
    bool   fDoNCellCut;
    bool   fDoCPlanesCut;
    bool   fDoPlanesCut;
    double fNCellLow;
    double fNCellHigh;
    double fLengthLow;
    double fLengthHigh;
    double fPercMipLow;
    double fPercMipHigh;
    double fPlanesMax;
    double fCPlanesLow;
    double fCPlanesHigh;
    double fRecoELow;
    double fRecoEHigh;
    bool   fInFidCut;
    
    fhicl::ParameterSet fPSetNDOS, fPSetND, fPSetFD;
  };
  
} // end namespace presel

namespace presel
{
  //----------------------------------------------------------------------
  NuePresel::NuePresel(const fhicl::ParameterSet& pset): 
    fSlicerLabel  (pset.get<std::string>("SlicerLabel")), 
    fRecoType     (pset.get<std::string>("RecoType"   )), 
    fAssn3D       (pset.get<std::string>("Assn3D"     )),
    fAssn2D       (pset.get<std::string>("Assn2D"     )),
    fDoLengthCut  (0),
    fDoEnergyCut  (0),
    fDoPercMipCut (0),
    fDoNCellCut   (0),
    fDoCPlanesCut (0),
    fDoPlanesCut  (0),
    fNCellLow     (0),
    fNCellHigh    (0),
    fLengthLow    (0),
    fLengthHigh   (0),
    fPercMipLow   (0),
    fPercMipHigh  (0),
    fPlanesMax    (0),
    fCPlanesLow   (0),
    fCPlanesHigh  (0),
    fRecoELow     (0),
    fRecoEHigh    (0),
    fInFidCut     (0),
    fPSetNDOS     (pset.get<fhicl::ParameterSet>("ndos")),
    fPSetND       (pset.get<fhicl::ParameterSet>("nd")),
    fPSetFD       (pset.get<fhicl::ParameterSet>("fd"))
  {
    produces< rb::FilterList<rb::Cluster> >();
    produces< std::vector<presel::NuePreselObj> >();
    produces< std::vector<rb::PID> >();
    produces< art::Assns<rb::PID, rb::Cluster> >();
    produces< art::Assns<presel::NuePreselObj, rb::Cluster> >();
  }
  
  //----------------------------------------------------------------------
  NuePresel::~NuePresel()
  {
  }
  
  void NuePresel::beginRun(art::Run& run)
  {
    art::ServiceHandle<geo::Geometry> geom;
    
    fhicl::ParameterSet pset;
    
    switch(geom->DetId()){
    case novadaq::cnv::kNDOS:
      break;
    case novadaq::cnv::kNEARDET:
      pset = fPSetND;
      break;
    case novadaq::cnv::kFARDET:
      pset = fPSetFD;
      break;
    default:
      assert(0 && "Unknown detector");
    }
    
    fDoLengthCut  = pset.get< bool        >("DoLengthCut"  );
    fDoEnergyCut  = pset.get< bool        >("DoEnergyCut"  );
    fDoPercMipCut = pset.get< bool        >("DoPercMipCut" );
    fDoNCellCut   = pset.get< bool        >("DoNCellCut"   );
    fDoCPlanesCut = pset.get< bool        >("DoCPlanesCut" );
    fDoPlanesCut  = pset.get< bool        >("DoPlanesCut"  );
    fNCellLow     = pset.get< signed int  >("NCellLow"     );
    fNCellHigh    = pset.get< signed int  >("NCellHigh"    );
    fLengthLow    = pset.get< signed int  >("LengthLow"    );
    fLengthHigh   = pset.get< signed int  >("LengthHigh"   );
    fPercMipLow   = pset.get< double      >("PercMipLow"   );
    fPercMipHigh  = pset.get< double      >("PercMipHigh"  );
    fPlanesMax    = pset.get< signed int  >("PlanesMax"    );
    fCPlanesLow   = pset.get< signed int  >("CPlanesLow"   );
    fCPlanesHigh  = pset.get< signed int  >("CPlanesHigh"  );
    fRecoELow     = pset.get< signed int  >("RecoELow"     );
    fRecoEHigh    = pset.get< signed int  >("RecoEHigh"    );
    fInFidCut     = pset.get< bool        >("InFidCut"     );
  }

  //----------------------------------------------------------------------
  void NuePresel::produce(art::Event& evt)
  {
    std::unique_ptr< rb::FilterList<rb::Cluster> > filtcol(new rb::FilterList<rb::Cluster>);
    std::unique_ptr< std::vector<presel::NuePreselObj> > nueprecol(new std::vector<presel::NuePreselObj>);
    std::unique_ptr< std::vector<rb::PID> > pidcol(new std::vector<rb::PID>);
    std::unique_ptr< art::Assns<rb::PID, rb::Cluster> > assns(new art::Assns<rb::PID, rb::Cluster>);
    std::unique_ptr< art::Assns<presel::NuePreselObj, rb::Cluster> > nuepreassns(new art::Assns<presel::NuePreselObj, rb::Cluster>);

    //Set up Calibrator and Geometry services
    art::ServiceHandle<calib::Calibrator> cal;
    art::ServiceHandle<geo::Geometry> geom;

    //Grab event's slices
    art::Handle< std::vector<rb::Cluster> > slices;
    evt.getByLabel(fSlicerLabel, slices);

    // Object holding showers associated with slice
    art::FindManyP<rb::Shower> fmpShower3D(slices, evt, art::InputTag(fRecoType,fAssn3D));
    art::FindManyP<rb::Shower> fmpShower2D(slices, evt, art::InputTag(fRecoType,fAssn2D));

    // Object holding prongs associated with slice
    art::FindManyP<rb::Prong> fmpProng3D(slices, evt, art::InputTag(fRecoType,fAssn3D)); 
    art::FindManyP<rb::Prong> fmpProng2D(slices, evt, art::InputTag(fRecoType,fAssn2D)); 

    // Object holding tracks associated with slice
    art::FindManyP<rb::Track> fmpTrack3D(slices, evt, art::InputTag(fRecoType,fAssn3D)); 
    art::FindManyP<rb::Track> fmpTrack2D(slices, evt, art::InputTag(fRecoType,fAssn2D)); 

    const int sliceMax = slices->size();

    for(int sliceIdx = 0; sliceIdx < sliceMax; ++sliceIdx){
      const rb::Cluster& slice = (*slices)[sliceIdx];

      if(slice.IsNoise()) continue;

      presel::NuePreselObj nuepreobj; // nue preselection object initialisation

      double longestObject = 0;
      if(fDoLengthCut){
        //see if input is tracks
	if(fmpTrack3D.isValid()){
	  std::vector< art::Ptr<rb::Track> > tracks = fmpTrack3D.at(sliceIdx); 
          //if there is a defined 2D label, use that as well
          if(fmpTrack2D.isValid() && fAssn2D != ""){
	    std::vector< art::Ptr<rb::Track> > tracks2D = fmpTrack2D.at(sliceIdx); 
            for (unsigned int ip=0; ip<tracks2D.size(); ++ip) tracks.push_back(tracks2D[ip]);
          }
	
	  //Find the longest track in the slice:
	  for(unsigned int n = 0; n < tracks.size(); ++n){
	    const double L = tracks[n]->TotalLength();
	    if(L > longestObject) longestObject = L;
	  }
	}
        //else fall back to prongs
	else if(fmpProng3D.isValid()){
	  std::vector< art::Ptr<rb::Prong> > prongs = fmpProng3D.at(sliceIdx); 
          //if there is a defined 2D label, use that as well
          if(fmpProng2D.isValid() && fAssn2D != ""){
	    std::vector< art::Ptr<rb::Prong> > prongs2D = fmpProng2D.at(sliceIdx); 
            for (unsigned int ip=0; ip<prongs2D.size(); ++ip) prongs.push_back(prongs2D[ip]);
          }
	
	  //Find the longest prong in the slice:
	  for(unsigned int n = 0; n < prongs.size(); ++n){
	    const double L = prongs[n]->TotalLength();
	    if(L > longestObject) longestObject = L;
	  }
	}
        //else fall back to showers
	else if(fmpShower3D.isValid()){
	  std::vector< art::Ptr<rb::Shower> > showers = fmpShower3D.at(sliceIdx); 
          //if there is a defined 2D label, use that as well
          if(fmpShower2D.isValid() && fAssn2D != ""){
	    std::vector< art::Ptr<rb::Shower> > showers2D = fmpShower2D.at(sliceIdx); 
            for (unsigned int ip=0; ip<showers2D.size(); ++ip) showers.push_back(showers2D[ip]);
          }
	
	  //Find the longest shower in the slice:
	  for(unsigned int n = 0; n < showers.size(); ++n){
	    const double L = showers[n]->TotalLength();
	    if(L > longestObject) longestObject = L;
	  }
	}
      }

      const double nCells    = slice.NCell(); //Number of cells in slice
      const int lengthPlanes = slice.ExtentPlane(); //Number of planes in the slice

      //Find whether the slice is in the fiducial volume (based on center of mass):
      const TVector3 mean = slice.MeanXYZ();
      const bool inFid    = geom->isInsideFiducialVolume(mean.X(), mean.Y(), mean.Z());

      const double wx = (slice.NXCell() > 0) ? slice.W(slice.XCell(0).get()) : 0;
      const double wy = (slice.NYCell() > 0) ? slice.W(slice.YCell(0).get()) : 0;

      //Holds contiguous planes info:
      int planeNum=-1;
      std::vector<int> planelist;

      //Hit counters:
      double nTotHit = 0.0;
      double nMipHit = 0.0;
      double fMipFrac = -1;

      for(unsigned int hitIdx = 0; hitIdx < slice.NCell(); ++hitIdx){
        // This is slow because we have to keep recalculating the mean
        // positions. The manual version below is much better.
        // const rb::RecoHit rhit = slice.RecoHit(hitIdx);

        const art::Ptr<rb::CellHit>& chit = slice.Cell(hitIdx);

        //Get calbrated reco hit:
	const rb::RecoHit rhit(cal->MakeRecoHit(*chit, chit->View() == geo::kX ? wx : wy));

        //Make sure things have been calibrated:
        if(!rhit.IsCalibrated())
          continue;
        

        //Don't do this if we're not using contiguous planes cut
        if (fDoCPlanesCut){

          //Make a list of the planes which have hits in the event:
          planeNum = chit->Plane();
          if (std::find (planelist.begin(),planelist.end(),planeNum) == planelist.end()) {
            planelist.push_back(planeNum);
          }

        }

        //Count up total hits:
        nTotHit++;

        //Count up MIP hits:
        if ( (rhit.MIP() > 0.5) && (rhit.MIP() < 1.5) ) nMipHit++;
      }

      //Calculate fraction of MIP hits:
      if (nTotHit>0) fMipFrac = nMipHit/nTotHit;

      // Selected unless we fail any cuts below
      bool sel = true;

      //Do the Fiducial Cut:
      if(!inFid && fInFidCut) {
        sel = false;
      }

      //Do the NCell cut:
      if ( fDoNCellCut ){
        if (nCells <= fNCellLow || nCells > fNCellHigh) sel = false;
      }

      //Do the Longest reco object cut:
      if ( fDoLengthCut ){
        if (longestObject <= fLengthLow || longestObject > fLengthHigh) sel = false;
      }

      //Do the MIP Fraction cut:
      if ( fDoPercMipCut){
        if ( fMipFrac <= fPercMipLow || fMipFrac > fPercMipHigh ) sel = false;
      }

      //Do the Plane Length cut:
      if ( fDoPlanesCut ){
        if ( lengthPlanes > fPlanesMax ) sel = false;
      }

      //Do the Energy Cut:
      if (fDoEnergyCut){
        const double calE = slice.CalorimetricEnergy();
        if (calE <= fRecoELow || calE > fRecoEHigh) sel = false;
      }

      //----Find the maximum number of contiguous planes, if we're doing this
      //----cut and if it's still necessary:
      if (fDoCPlanesCut && sel){

        int nContig=1;
        int maxContig=-1;

        //Contiguous plane counter:
        if (planelist.size()>0) std::sort(planelist.begin(),planelist.end());
        for (unsigned int iplane=0; iplane<planelist.size(); iplane++){

          if (iplane!=0){
            if (planelist[iplane] == (planelist[iplane-1]+1)){
              nContig++;
              if (nContig>maxContig) maxContig=nContig;
            } else {
              nContig=0;
            }
          }
        }

        //Do the Contiguous Planes Cut
        if (maxContig <= fCPlanesLow || maxContig > fCPlanesHigh) sel = false;
      }

      if(!sel) filtcol->Add(slices, sliceIdx);
      
      // We're a nue preselection
      pidcol->push_back(rb::PID(12, sel));
      nuepreobj.SetPassNuePresel(sel);

      nueprecol->push_back(nuepreobj);

      util::CreateAssn(*this, evt, *pidcol, art::Ptr<rb::Cluster>(slices, sliceIdx), *assns);
      util::CreateAssn(*this, evt, *nueprecol, art::Ptr<rb::Cluster>(slices, sliceIdx), *nuepreassns);

    } // end for sliceIdx

    evt.put(std::move(pidcol));
    evt.put(std::move(nueprecol));
    evt.put(std::move(assns));
    evt.put(std::move(nuepreassns));
    evt.put(std::move(filtcol));
  }

}//end namespace presel

DEFINE_ART_MODULE(presel::NuePresel);
