////////////////////////////////////////////////////////////////////////////
// \file    NumiFiltering_module.cc
// \brief   Module to filter events for neutrino search.
//          Output files used for neutrino event scanning in Far Detector
//          Select Contained/Exiting/Entering tracks after cuts.
// \author  Gavin Davies - gsdavies@iastate.edu
///////////////////////////////////////////////////////////////////////////
#include <algorithm>
#include <cassert>
#include <string>

//ROOT includes
#include "TMath.h"
#include "TVector3.h"

//Novasoft includes
#include "CosRej/CosRejFxs.h"
#include "Geometry/Geometry.h"
#include "Geometry/LiveGeometry.h"
#include "GeometryObjects/CellGeo.h"
#include "MCCheater/BackTracker.h"
#include "NovaDAQConventions/DAQConventions.h"
#include "RecoBase/Cluster.h"
#include "RecoBase/FilterList.h"
#include "RecoBase/Track.h"

// Framework includes
#include "art/Framework/Core/EDFilter.h"
#include "art/Framework/Core/FindManyP.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Services/Optional/TFileService.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "fhiclcpp/ParameterSet.h"


namespace comi
{
  class NumiFiltering: public art::EDFilter
  {
  public:
    explicit NumiFiltering(fhicl::ParameterSet const &pset);
    ~NumiFiltering();
    bool beginRun(art::Run& run);
    bool filter(art::Event& evt);

  protected:

    std::string  fSlicerLabel;
    std::string  fTrackType;
    bool         fDoCleanup;
    bool         fDoTrackAngleCut;
    bool         fDoXYContainment;
    bool         fDoTrackContain;
    bool         fUseStrongCut;
    bool         fDoInTimeCut;
    bool         fDoUnContainedHitCut;
    unsigned int fNCellLow;
    unsigned int fMaxUncontainedHits;
    double       fMinTCut;
    double       fMaxTCut;
    bool         fUseWeakSel;

    bool         fSelectContained;
    bool         fSelectThroughGoing;
    bool         fSelectEntering;
    bool         fSelectExiting;

    float        fMinXFid;
    float        fMinYFid;
    float        fMinZ;
    float        fMaxXFid;
    float        fMaxYFid;
    float        fMaxZFid;
    float        fZCut;


    fhicl::ParameterSet fPSetNDOS, fPSetND, fPSetFD;
  private:

    cosrej::CosRejFxs*         fCosRejFxs;
  };

}//End namespace comi

namespace comi
{
  //----------------------------------------------------------------------
  NumiFiltering::NumiFiltering(fhicl::ParameterSet const &pset) : 
    fSlicerLabel        (pset.get< std::string  >("SlicerLabel")),
    fTrackType          (pset.get< std::string  >("TrackType")  ),
    fDoCleanup          (0),
    fDoTrackAngleCut    (0),
    fDoXYContainment    (0),
    fDoTrackContain     (0),
    fUseStrongCut       (0),
    fDoInTimeCut        (0),
    fDoUnContainedHitCut(0),
    fNCellLow           (0),
    fMaxUncontainedHits (0),
    fMinTCut            (0),
    fMaxTCut            (0),
    fUseWeakSel         (0),
    fSelectContained    (0),
    fSelectThroughGoing (0),
    fSelectEntering     (0),
    fSelectExiting      (0),
    fPSetNDOS           (pset.get< fhicl::ParameterSet >("ndos")),
    fPSetND             (pset.get< fhicl::ParameterSet >("nd")  ),
    fPSetFD             (pset.get< fhicl::ParameterSet >("fd")  )
  {
    produces< rb::FilterList<rb::Cluster> >();
  }

  //----------------------------------------------------------------------
  NumiFiltering::~NumiFiltering()
  {
  }

  //----------------------------------------------------------------------
  bool NumiFiltering::beginRun(art::Run& run)
  {
    art::ServiceHandle<geo::Geometry>  geom;

    fhicl::ParameterSet pset;
    
    switch(geom->DetId()){
    case novadaq::cnv::kNDOS:
      pset = fPSetNDOS;
      break;
    case novadaq::cnv::kNEARDET:
      pset = fPSetND;
      break;
    case novadaq::cnv::kFARDET:
      pset = fPSetFD;
      break;
    default:
      assert(0 && "Unknown detector");
    }

    fDoCleanup           = pset.get< bool         >("DoCleanup");
    fDoTrackAngleCut     = pset.get< bool         >("DoTrackAngleCut");
    fDoXYContainment     = pset.get< bool         >("DoXYContainment");
    fDoTrackContain      = pset.get< bool         >("DoTrackContain");
    fUseStrongCut        = pset.get< bool         >("UseStrongCut");
    fDoInTimeCut         = pset.get< bool         >("DoInTimeCut");
    fDoUnContainedHitCut = pset.get< bool         >("DoUnContainedHitCut");
    fNCellLow            = pset.get< unsigned int >("NCellLow");
    fMaxUncontainedHits  = pset.get< unsigned int >("MaxUncontainedHits");
    fMinTCut             = pset.get< double       >("MinTCut");
    fMaxTCut             = pset.get< double       >("MaxTCut");
    fUseWeakSel          = pset.get< bool         >("UseWeakSel");
    fSelectContained     = pset.get< bool         >("SelectContained");
    fSelectThroughGoing  = pset.get< bool         >("SelectThroughGoing");
    fSelectEntering      = pset.get< bool         >("SelectEntering");
    fSelectExiting       = pset.get< bool         >("SelectExiting");

    fMinXFid             = pset.get< float        >("MinXFid");
    fMaxXFid             = pset.get< float        >("MaxXFid");
    fMinYFid             = pset.get< float        >("MinYFid");
    fMaxYFid             = pset.get< float        >("MaxYFid");
    fMinZ                = pset.get< float        >("MinZ");
    fZCut                = pset.get< float        >("ZCut");

    return true;
  }

  //----------------------------------------------------------------------
  bool NumiFiltering::filter(art::Event& evt)
  {

    std::unique_ptr< rb::FilterList<rb::Cluster> > filtcol(new rb::FilterList<rb::Cluster>);

    bool selected = false;
    //Grab event's slices
    art::Handle< std::vector<rb::Cluster> > slices;
    evt.getByLabel(fSlicerLabel, slices);

    // Object holding tracks associated with slice
    art::FindManyP<rb::Track> fmpTrack(slices, evt, fTrackType); 

    const int sliceMax = slices->size();

    for(int sliceIdx = 0; sliceIdx < sliceMax; ++sliceIdx){
      const rb::Cluster& slice = (*slices)[sliceIdx];

      if(slice.IsNoise())
	{
	  filtcol->Add(slices, sliceIdx);
	  continue;
	}
      
      double ratio = 1, diry = 0, angle = 1;
      double minx  = -9999, miny = -9999, minz = -9999;
      double maxx  = 9999, maxy = 9999, maxz = 9999;

      TVector3 start,stop;
      
      double meant = 0;

      unsigned int nXplanes = 0, nYplanes = 0;
      bool   is3d  = false;

      unsigned int stot[6] = {0,0,0,0,0,0};

      const double nCells = slice.NCell(); //Number of cells in slice
      unsigned int PXhi  = 0;
      unsigned int PXlo  = UINT_MAX;
      unsigned int PYhi  = 0;
      unsigned int PYlo  = UINT_MAX;
      //loop over all hits
      
      art::ServiceHandle<cheat::BackTracker> bt;
      art::ServiceHandle<geo::Geometry>  geom;
      art::ServiceHandle<geo::LiveGeometry>  livegeom;
      
      double fMinZFid = 0.;
      double zfront1,zback1,zfront2,zback2;

      if(geom->DetId() == novadaq::cnv::kFARDET){
	int ndets = livegeom->InstrumentedDetEnds(zfront1,zback1,zfront2,zback2);
	// take the most downstream zfront and zback
	if(ndets > 0){
	  fMinZFid = zfront1 + fMinZ;
	}
      }
      else fMinZFid = fMinZ;

      
      fMaxZFid = zback1 - fZCut;
      
      //      std::cout << zback1 << " " << zfront1 << "\n";
      for(int view = geo::kX; view <= geo::kY; ++view) {
	const geo::View_t geoview = geo::View_t(view);
	
	for(unsigned int j = 0; j < slice.NCell(geoview); ++j) {

	  const rb::CellHit* h = slice.Cell(geoview,j).get();

	  // Get Hit position
	  double xyz[3];
          geom->Plane(h->Plane())->Cell(h->Cell())->GetCenter(xyz);
          const double x = xyz[0];
          const double y = xyz[1];
          const double z = xyz[2];

	  if(h->View() == geo::kX) {
	    if(h->Plane() > PXhi) PXhi = h->Plane();
	    if(h->Plane() < PXlo) PXlo = h->Plane();
	  }
	  if(h->View() == geo::kY) {
	    if(h->Plane() > PYhi) PYhi = h->Plane();
	    if(h->Plane() < PYlo) PYlo = h->Plane();
	  }

	  
	  // Total up number of cells > Max (X/Y/Z) and < Min (X/Y/Z)      
	  //TOP
	  if(y > fMaxYFid) {
	    stot[0]++;
	    // to avoid double-counting, stop counting for subsequent boundaries
	    continue; 
	  }
	  //BOTTOM
	  if(y < fMinYFid) {
	    stot[1]++;
	    continue;
	  }
	  //LEFT                       
	  if(x > fMaxXFid) {
	    stot[2]++;
	    continue;
	  }
	  //RIGHT
	  if(x < fMinXFid) {
	    stot[3]++;
	    continue;
	  }
	  //FRONT           
	  if(z < fMinZFid) {
	    stot[4]++;
	    continue;
	  }
	  //BACK
	  if(z > fMaxZFid) {
	    stot[5]++;
	    continue;
	  }
	  
	}
      }
      
      
      nXplanes = PXhi - PXlo+1;
      nYplanes = PYhi - PYlo+1;
      
      if(fmpTrack.isValid()){
        std::vector< art::Ptr<rb::Track> > tracks = fmpTrack.at(sliceIdx);
        if(tracks.size()>0) {
          const art::Ptr<rb::Track> track = tracks[0];
	  
	  meant = slice.MeanTNS();
	  
	  
	  
	  minx  = slice.MinXYZ().X();
	  miny  = slice.MinXYZ().Y();
	  minz  = slice.MinXYZ().Z();
	  
	  maxx  = slice.MaxXYZ().X();
	  maxy  = slice.MaxXYZ().Y();
	  maxz  = slice.MaxXYZ().Z();
	  
	  start = track->Start();
	  stop  = track->Stop();
	  
	  diry  = track->Dir().Y(); 
	  angle = fCosRejFxs->getAngle(track->Dir());
	  ratio = (track->NCell()*1.0) / (1.0*nCells);
	  
	  
	  is3d = track->Is3D();
	  

	}
      }
      
      // Selected unless we fail any cuts below
      bool sel = true;

      if(fDoInTimeCut){
	if(meant < fMinTCut || meant > fMaxTCut){
	  sel = false;
	}
      }
      if(sel == true) std::cout<< "True: Timing";
      //----------------Actually start the cut selections here----------------
      //Do the NCell cut
      if(fDoCleanup){
	if(nCells <= fNCellLow || !is3d) {
	  sel = false;
	}
      }
      if(sel == true) std::cout<< "True: Cleanup";
      //Do the XY slice.box containment cut + the track start/stop containment
      //Change this to use the fwddist/bckdist at a later date
      if(fDoXYContainment){	  
	if(minx < fMinXFid || miny < fMinYFid ||
	   maxx > fMaxXFid || maxy > fMaxYFid)
	  sel = false;
      }
      if(sel == true) std::cout<< "True: XYContainment";

      if(fDoUnContainedHitCut){
	const unsigned int smaxelement = *std::max_element(stot,stot+6);
	if(smaxelement > fMaxUncontainedHits)
	  sel = false;
      }
      if(sel == true) std::cout<< "True: BBC";

      if(fDoTrackContain){
	if(start.X() < fMinXFid || start.Y() < fMinYFid ||
	   stop.X()  > fMaxXFid || stop.Y()  > fMaxYFid ||
	   stop.X()  < fMinXFid || stop.Y()  < fMinYFid ||
	   start.X() > fMaxXFid || start.Y() > fMaxYFid) {
	  sel = false;
	}
      }
      if(sel == true) std::cout<< "True: Contain";

      //Do the Track Angle cut:
      if (fDoTrackAngleCut){
        if(fmpTrack.isValid()) {
	  
	  double cosmicvar = (angle*angle) - (diry*diry);
	  
	  if(fUseStrongCut == false) { // weak cut
	    if(ratio > 0.88) {
	      if(nCells > 50 && nCells <= 135 && 
		 cosmicvar < ((1.1/85)*nCells-1.65)) {
		sel = false;
	      }
	      
	      if(nCells > 135 && nCells <= 250 &&
		 cosmicvar < ((.6/115)*nCells-.6)) {
		sel = false;
	      }
	      
	      if(nCells > 250 && nCells <= 500 &&
		 cosmicvar < ((.1/250)*nCells+0.6)) {
		sel = false;
	      }
	      
	      if(nCells > 500 && cosmicvar > 0.8) {
		sel = false;
	      }
	    }
	  }
	  
	  if(fUseStrongCut == true) { // strong cut
	    if(nCells > 30 && nCells <= 135 && ratio > 0.8 && 
	       cosmicvar < ((1/100)*nCells-1.0)) { 
	      sel = false;
	    }
	    
	    if(nCells > 135 && nCells <= 250 && ratio > 0.8 &&
	       cosmicvar < ((.45/100)*nCells-.2575)) {
	      sel = false;
	    }
	    
	    if(nCells > 250 && nCells <= 500 && ratio > 0.83 &&
	       cosmicvar < ((.1/200)*nCells+0.6825)) {
	      sel = false;
	    }
	    
	    if(nCells > 500 && ratio > 0.8 && 
	       cosmicvar > 0.9) {
	      sel = false;
	    }
	  }
	}
      }
      if(sel == true) std::cout<< "True: AngVar";

      bool iscont = false, isexit = false;
      bool isenter = false, isthrough = false;

      if(fUseWeakSel == true){
	if(fSelectContained){
	  if(minz > fMinZFid && maxz < fMaxZFid) iscont = true;
	}
	if(fSelectThroughGoing){
	  if(minz < fMinZFid && maxz > fMaxZFid) isthrough = true;
	}
	if(fSelectEntering){
	  if(minz < fMinZFid && maxz < fMaxZFid) isexit = true;
	}
	if(fSelectExiting){
	  if(minz > fMinZFid && maxz > fMaxZFid) isenter = true;
	}	
		
	if(!(iscont || isexit || isenter || isthrough)) {
	  sel = false;
	}
      }
      /*
      std::cout << "MinZ: " << fMinZFid 
		<< " MaxZ: " << fMaxZFid
		<< " nXplanes: " << nXplanes
		<< " nYplanes: " << nYplanes
		<< " angle: " << angle
		<< "\n";
      */
      if(fUseWeakSel == false){
	if(fSelectContained){
	  //  std::cout << "Contained - start: " << start.Z() 
	  //	    << " stop: " << stop.Z()
	  //	    << "\n";
	  if(((start.Z() > fMinZFid && stop.Z() < fMaxZFid) ||
	      (stop.Z() > fMinZFid && start.Z() < fMaxZFid)) &&
	     nXplanes > 4 && nYplanes > 4 &&
	     (angle > 0.5 || angle < -0.5)) iscont = true;
	}
	
	if(fSelectThroughGoing){
	  if((start.Z() < fMinZFid && stop.Z() > fMaxZFid) ||
	     (stop.Z() < fMinZFid && start.Z() > fMaxZFid))
	    isthrough = true;
	}
	
	if(fSelectEntering){
	  if(((start.Z() < fMinZFid && stop.Z() < fMaxZFid) ||
	      (stop.Z() < fMinZFid && start.Z() < fMaxZFid)) &&
	     nXplanes > 4 && nYplanes > 4 && 
	     (angle > 0.95 || angle < -0.95)) isenter = true;
	}
	if(fSelectExiting){
	  if(sel==true)std::cout<< "True: Exiting check";
	  if(((start.Z() > fMinZFid && stop.Z() > start.Z()) ||
	      (stop.Z() > fMinZFid && start.Z() > stop.Z())) &&
	     nXplanes > 4 && nYplanes > 4 &&
	     (angle > 0.75 || angle < -0.75)) {
	    std::cout << "Exiting - start: " << start.Z() 
	              << " stop: " << stop.Z()
	    	        << " angle " << angle << "\n";
	    isexit = true;
	  }
	}
	std::cout << " IS EXIT " << isexit << " sel: " << sel << std::endl;
	if(!isexit) sel = false;
	/*
	if(sel ==true){
	  std::cout << "isexit: " << isexit << "\n";
	  if(!(iscont || isexit || isenter || isthrough)) {
	    sel = false;
	  }
	}
	*/
      }
      
      if(sel == true){
	std::cout << "Event: " << evt.id().event() << ", subrun: " << evt.subRun() << "\n";
	selected = sel;
      }
      
      if(sel == false){ 
	filtcol->Add(slices, sliceIdx);
      }
    } // end for sliceIdx
    
    evt.put(std::move(filtcol));

    return selected;
  }

  DEFINE_ART_MODULE(NumiFiltering)  
}//end namespace comi
