////////////////////////////////////////////////////////////////////           
/// \brief   A validation module for the QEEventFinder package    
/// \author  Nicholas Raddatz - raddatz@physics.umn.edu
/// \date    February 2013
//////////////////////////////////////////////////////////////////// 

// ROOT includes
#include "TH1.h"

// Framework includes
#include "art/Framework/Core/EDAnalyzer.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Services/Optional/TFileService.h"
#include "fhiclcpp/ParameterSet.h"
#include "art/Framework/Core/FindManyP.h"

//NOvA includes
#include "QEEventFinder/QePId.h"
#include "ReMId/ReMId.h"
#include "RecoBase/Track.h"
#include "NumuEnergy/NumuE.h"
#include "MCCheater/BackTracker.h"
#include "Utilities/AssociationUtil.h"


namespace qeef{
  class QeFinderVal : public art::EDAnalyzer{
  public:
    explicit QeFinderVal(fhicl::ParameterSet const& pset);
    virtual ~QeFinderVal();

    void beginJob();
    void analyze(art::Event const& evt);

  private:

    std::string fSliceLabel;
    std::string fTrackLabel;
    std::string fPIDLabel;
    std::string fEnergyLabel;
    std::string fQePIdLabel;

    // make plots of variables for all slices
    TH1F *fNtrkAll, *fModeAll, *fOffTrkEAll, *fEdiffAll, *fEdiffZAll, *fDedxAll, *fPdgAll, *fPidAll;
   
    // plots of qepid variables for 1 3d track slices
    TH1F *fNtrk1Trk, *fMode1Trk, *fOffTrkE1Trk, *fEdiff1Trk, *fEdiffZ1Trk, *fDedx1Trk, *fPdg1Trk, *fPid1Trk;
    
    // plots of qepid variables for 2 track and >=1 3d track slices
    TH1F *fNtrk2Trk, *fMode2Trk, *fOffTrkE2Trk, *fEdiff2Trk, *fEdiffZ2Trk, *fDedx2Trk, *fPdg2Trk, *fPid2Trk;

    // plots of qepid variables for slices that don't fall into the above to categories
    TH1F *fNtrk0Trk, *fMode0Trk, *fOffTrkE0Trk, *fEdiff0Trk, *fEdiffZ0Trk, *fDedx0Trk, *fPdg0Trk, *fPid0Trk;

    // plots of 1 and 2 track pid slcies that were true qe events
    TH1F *fNtrk1TrkTrueQE, *fMode1TrkTrueQE, *fOffTrkE1TrkTrueQE, *fEdiff1TrkTrueQE, *fEdiffZ1TrkTrueQE, *fDedx1TrkTrueQE, *fPdg1TrkTrueQE, *fPid1TrkTrueQE;
    TH1F *fNtrk2TrkTrueQE, *fMode2TrkTrueQE, *fOffTrkE2TrkTrueQE, *fEdiff2TrkTrueQE, *fEdiffZ2TrkTrueQE, *fDedx2TrkTrueQE, *fPdg2TrkTrueQE, *fPid2TrkTrueQE;

    // plots of 1 and 2 track pid slcies that were true nonqe events
    TH1F *fNtrk1TrkTrueNonQE, *fMode1TrkTrueNonQE, *fOffTrkE1TrkTrueNonQE, *fEdiff1TrkTrueNonQE, *fEdiffZ1TrkTrueNonQE, *fDedx1TrkTrueNonQE, *fPdg1TrkTrueNonQE, *fPid1TrkTrueNonQE;
    TH1F *fNtrk2TrkTrueNonQE, *fMode2TrkTrueNonQE, *fOffTrkE2TrkTrueNonQE, *fEdiff2TrkTrueNonQE, *fEdiffZ2TrkTrueNonQE, *fDedx2TrkTrueNonQE, *fPdg2TrkTrueNonQE, *fPid2TrkTrueNonQE;

  };
}

namespace qeef{

  //.......................................................................
  QeFinderVal::QeFinderVal(fhicl::ParameterSet const& pset)
    : EDAnalyzer(pset)
  {
    fSliceLabel    = (pset.get< std::string >("SliceLabel"));
    fTrackLabel    = (pset.get< std::string >("TrackLabel"));
    fPIDLabel      = (pset.get< std::string >("PIDLabel"));
    fEnergyLabel   = (pset.get< std::string >("EnergyLabel"));
    fQePIdLabel    = (pset.get< std::string >("QePIDLabel"));

  }

  //.......................................................................
  QeFinderVal::~QeFinderVal()
  {
  }

  //.......................................................................
  void QeFinderVal::beginJob()
  {

    art::ServiceHandle<art::TFileService> tfs;

    // qepid info for all slices
    fNtrkAll = tfs->make<TH1F>("NtrkAll",";Ntrk;Slices",5,-1.5,3.5);
    fModeAll = tfs->make<TH1F>("ModeAll",";Mode;Slices",3,-1.5,1.5);
    fOffTrkEAll = tfs->make<TH1F>("OffTrkEAll",";OffE;Slices",100,-0.05,1.05);
    fEdiffAll = tfs->make<TH1F>("EdiffAll",";Ediff;Slices",100,-1.0,1.0);
    fEdiffZAll = tfs->make<TH1F>("EdiffZAll",";EdiffZ;Slices",100,-15,5);
    fDedxAll = tfs->make<TH1F>("DedxAll",";Dedx;Slices",100,-1,6);
    fPdgAll = tfs->make<TH1F>("PdgAll",";Pdg;Slices",16,-0.5,15.5);
    fPidAll = tfs->make<TH1F>("PidAll",";PID;Slices",100,-5.5,1.5);


    // qepid info for slices with 1 and only 1 3d track
    fNtrk1Trk = tfs->make<TH1F>("Ntrk1Trk",";Ntrk;Slices",5,-1.5,3.5);
    fMode1Trk = tfs->make<TH1F>("Mode1Trk",";Mode;Slices",3,-1.5,1.5);
    fOffTrkE1Trk = tfs->make<TH1F>("OffTrkE1Trk",";OffE;Slices",100,-0.05,1.05);
    fEdiff1Trk = tfs->make<TH1F>("Ediff1Trk",";Ediff;Slices",100,-1.0,1.0);
    fEdiffZ1Trk = tfs->make<TH1F>("EdiffZ1Trk",";EdiffZ;Slices",100,-15,5);
    fDedx1Trk = tfs->make<TH1F>("Dedx1Trk",";Dedx;Slices",100,-1,6);
    fPdg1Trk = tfs->make<TH1F>("Pdg1Trk",";Pdg;Slices",16,-0.5,15.5);
    fPid1Trk = tfs->make<TH1F>("Pid1Trk",";PID;Slices",100,-5.5,1.5);

    // qepid info for slices with 2 total tracks, with at least one 3d track
    fNtrk2Trk = tfs->make<TH1F>("Ntrk2Trk",";Ntrk;Slices",5,-1.5,3.5);
    fMode2Trk = tfs->make<TH1F>("Mode2Trk",";Mode;Slices",3,-1.5,1.5);
    fOffTrkE2Trk = tfs->make<TH1F>("OffTrkE2Trk",";OffE;Slices",100,-0.05,1.05);
    fEdiff2Trk = tfs->make<TH1F>("Ediff2Trk",";Ediff;Slices",100,-1.0,1.0);
    fEdiffZ2Trk = tfs->make<TH1F>("EdiffZ2Trk",";EdiffZ;Slices",100,-15,5);
    fDedx2Trk = tfs->make<TH1F>("Dedx2Trk",";Dedx;Slices",100,-1,6);
    fPdg2Trk = tfs->make<TH1F>("Pdg2Trk",";Pdg;Slices",16,-0.5,15.5);
    fPid2Trk = tfs->make<TH1F>("Pid2Trk",";PID;Slices",100,-5.5,1.5);

    // qepid info slices that shouldn't have a qepid pid between 0 and 1
    fNtrk0Trk = tfs->make<TH1F>("Ntrk0Trk",";Ntrk;Slices",5,-1.5,3.5);
    fMode0Trk = tfs->make<TH1F>("Mode0Trk",";Mode;Slices",3,-1.5,1.5);
    fOffTrkE0Trk = tfs->make<TH1F>("OffTrkE0Trk",";OffE;Slices",100,-0.05,1.05);
    fEdiff0Trk = tfs->make<TH1F>("Ediff0Trk",";Ediff;Slices",100,-1.0,1.0);
    fEdiffZ0Trk = tfs->make<TH1F>("EdiffZ0Trk",";EdiffZ;Slices",100,-15,5);
    fDedx0Trk = tfs->make<TH1F>("Dedx0Trk",";Dedx;Slices",100,-1,6);
    fPdg0Trk = tfs->make<TH1F>("Pdg0Trk",";Pdg;Slices",16,-0.5,15.5);
    fPid0Trk = tfs->make<TH1F>("Pid0Trk",";PID;Slices",100,-5.5,1.5);

    // make plots for 1 and 2 track true qe events
    fNtrk1TrkTrueQE = tfs->make<TH1F>("Ntrk1TrkTrueQE",";Ntrk;Slices",5,-1.5,3.5);
    fMode1TrkTrueQE = tfs->make<TH1F>("Mode1TrkTrueQE",";Mode;Slices",3,-1.5,1.5);
    fOffTrkE1TrkTrueQE = tfs->make<TH1F>("OffTrkE1TrkTrueQE",";OffE;Slices",100,-0.05,1.05);
    fEdiff1TrkTrueQE = tfs->make<TH1F>("Ediff1TrkTrueQE",";Ediff;Slices",100,-1.0,1.0);
    fEdiffZ1TrkTrueQE = tfs->make<TH1F>("EdiffZ1TrkTrueQE",";EdiffZ;Slices",100,-15,5);
    fDedx1TrkTrueQE = tfs->make<TH1F>("Dedx1TrkTrueQE",";Dedx;Slices",100,-1,6);
    fPdg1TrkTrueQE = tfs->make<TH1F>("Pdg1TrkTrueQE",";Pdg;Slices",16,-0.5,15.5);
    fPid1TrkTrueQE = tfs->make<TH1F>("Pid1TrkTrueQE",";PID;Slices",100,-5.5,1.5);
    fNtrk2TrkTrueQE = tfs->make<TH1F>("Ntrk2TrkTrueQE",";Ntrk;Slices",5,-1.5,3.5);
    fMode2TrkTrueQE = tfs->make<TH1F>("Mode2TrkTrueQE",";Mode;Slices",3,-1.5,1.5);
    fOffTrkE2TrkTrueQE = tfs->make<TH1F>("OffTrkE2TrkTrueQE",";OffE;Slices",100,-0.05,1.05);
    fEdiff2TrkTrueQE = tfs->make<TH1F>("Ediff2TrkTrueQE",";Ediff;Slices",100,-1.0,1.0);
    fEdiffZ2TrkTrueQE = tfs->make<TH1F>("EdiffZ2TrkTrueQE",";EdiffZ;Slices",100,-15,5);
    fDedx2TrkTrueQE = tfs->make<TH1F>("Dedx2TrkTrueQE",";Dedx;Slices",100,-1,6);
    fPdg2TrkTrueQE = tfs->make<TH1F>("Pdg2TrkTrueQE",";Pdg;Slices",16,-0.5,15.5);
    fPid2TrkTrueQE = tfs->make<TH1F>("Pid2TrkTrueQE",";PID;Slices",100,-5.5,1.5);

    // make plots for 1 and 2 track true nonqe events
    fNtrk1TrkTrueNonQE = tfs->make<TH1F>("Ntrk1TrkTrueNonQE",";Ntrk;Slices",5,-1.5,3.5);
    fMode1TrkTrueNonQE = tfs->make<TH1F>("Mode1TrkTrueNonQE",";Mode;Slices",3,-1.5,1.5);
    fOffTrkE1TrkTrueNonQE = tfs->make<TH1F>("OffTrkE1TrkTrueNonQE",";OffE;Slices",100,-0.05,1.05);
    fEdiff1TrkTrueNonQE = tfs->make<TH1F>("Ediff1TrkTrueNonQE",";Ediff;Slices",100,-1.0,1.0);
    fEdiffZ1TrkTrueNonQE = tfs->make<TH1F>("EdiffZ1TrkTrueNonQE",";EdiffZ;Slices",100,-15,5);
    fDedx1TrkTrueNonQE = tfs->make<TH1F>("Dedx1TrkTrueNonQE",";Dedx;Slices",100,-1,6);
    fPdg1TrkTrueNonQE = tfs->make<TH1F>("Pdg1TrkTrueNonQE",";Pdg;Slices",16,-0.5,15.5);
    fPid1TrkTrueNonQE = tfs->make<TH1F>("Pid1TrkTrueNonQE",";PID;Slices",100,-5.5,1.5);
    fNtrk2TrkTrueNonQE = tfs->make<TH1F>("Ntrk2TrkTrueNonQE",";Ntrk;Slices",5,-1.5,3.5);
    fMode2TrkTrueNonQE = tfs->make<TH1F>("Mode2TrkTrueNonQE",";Mode;Slices",3,-1.5,1.5);
    fOffTrkE2TrkTrueNonQE = tfs->make<TH1F>("OffTrkE2TrkTrueNonQE",";OffE;Slices",100,-0.05,1.05);
    fEdiff2TrkTrueNonQE = tfs->make<TH1F>("Ediff2TrkTrueNonQE",";Ediff;Slices",100,-1.0,1.0);
    fEdiffZ2TrkTrueNonQE = tfs->make<TH1F>("EdiffZ2TrkTrueNonQE",";EdiffZ;Slices",100,-15,5);
    fDedx2TrkTrueNonQE = tfs->make<TH1F>("Dedx2TrkTrueNonQE",";Dedx;Slices",100,-1,6);
    fPdg2TrkTrueNonQE = tfs->make<TH1F>("Pdg2TrkTrueNonQE",";Pdg;Slices",16,-0.5,15.5);
    fPid2TrkTrueNonQE = tfs->make<TH1F>("Pid2TrkTrueNonQE",";PID;Slices",100,-5.5,1.5);

  }
  
  //.......................................................................
  void QeFinderVal::analyze(art::Event const& evt)
  {

        
    // get all of slices out of the event
    art::Handle<std::vector<rb::Cluster> > slices;
    evt.getByLabel(fSliceLabel,slices);

    // make a vector of all cellhits
    art::PtrVector<rb::CellHit> allHits;
    for(size_t isl = 0; isl < slices->size(); ++isl){
      art::Ptr<rb::Cluster> slice(slices,isl);
      for(unsigned int ihit = 0; ihit < slice->NCell(); ++ihit){
	allHits.push_back(slice->Cell(ihit));
      }
    }
    
    // set up associations between slices and qepids 
    art::FindManyP<qeef::QePId> qeefAss(slices,evt,fQePIdLabel);
    // set up associations between slices and tracks
    art::FindManyP<rb::Track> trkAss(slices,evt,fTrackLabel);
    // set up associations between slices and energy
    art::FindManyP<numue::NumuE> eAss(slices,evt,fEnergyLabel);

    // get a backtracker service handle
    art::ServiceHandle<cheat::BackTracker> bt;


    for(size_t isl = 0; isl < slices->size(); ++isl){

      // Get the slice
      art::Ptr<rb::Cluster> slice(slices,isl);
      // skip if it is noise
      if(slice->IsNoise()){ continue; }

      // get the true mctruth matching for this slice
      if(!bt->HaveTruthInfo()){
	mf::LogWarning("QeFinderVal") << "attempting to run MC truth check on "
				   << "things without truth, bail";
	return;
      }

      std::vector<cheat::NeutrinoEffPur> nus = bt->SliceToNeutrinoInteractions(slice->AllCells(),allHits);
      // Is this a true numu CCQE event
      bool isTrueQE = false;
      if(nus.size() > 0){
	if((nus[0].neutrinoInt->GetNeutrino().Nu().PdgCode() == 14 || 
	    nus[0].neutrinoInt->GetNeutrino().Nu().PdgCode() == -14) &&
	   nus[0].neutrinoInt->GetNeutrino().CCNC() == simb::kCC &&
	   nus[0].neutrinoInt->GetNeutrino().Mode() == simb::kQE){ isTrueQE = true; }
      }

      // get the pid that are associated with this slice
      const std::vector<art::Ptr<qeef::QePId> > qepids = qeefAss.at(isl);
      if(qepids.size() != 1){ std::cout<<"Missing a qe pid object on this slice"<<std::endl; }
      else{
	art::Ptr<qeef::QePId> qepid = qepids[0];
	fNtrkAll->Fill(qepid->Ntrk());
	fModeAll->Fill(qepid->Mode());
	fOffTrkEAll->Fill(qepid->OffTrkE());
	fEdiffAll->Fill(qepid->Ediff());
	fEdiffZAll->Fill(qepid->EdiffZ());
	fDedxAll->Fill(qepid->Dedx());
	fPdgAll->Fill(qepid->Pdg());
	fPidAll->Fill(qepid->Value());
	
	// get all of the tracks associated with this slice
	const std::vector<art::Ptr<rb::Track> > tracks = trkAss.at(isl);
	
	// count how many trks we have
	int n3d = 0;
	int ntrks = tracks.size();
	for(unsigned int itr = 0; itr < tracks.size(); ++itr){
	  if(tracks[itr]->View() == geo::kXorY){ ++n3d; }
	}
	
	// make plots for the individual categories
	if(n3d == 1 && ntrks ==1){
	  fNtrk1Trk->Fill(qepid->Ntrk());
	  fMode1Trk->Fill(qepid->Mode());
	  fOffTrkE1Trk->Fill(qepid->OffTrkE());
	  fEdiff1Trk->Fill(qepid->Ediff());
	  fEdiffZ1Trk->Fill(qepid->EdiffZ());
	  fDedx1Trk->Fill(qepid->Dedx());
	  fPdg1Trk->Fill(qepid->Pdg());
	  fPid1Trk->Fill(qepid->Value());
	  if(isTrueQE){
	    fNtrk1TrkTrueQE->Fill(qepid->Ntrk());
	    fMode1TrkTrueQE->Fill(qepid->Mode());
	    fOffTrkE1TrkTrueQE->Fill(qepid->OffTrkE());
	    fEdiff1TrkTrueQE->Fill(qepid->Ediff());
	    fEdiffZ1TrkTrueQE->Fill(qepid->EdiffZ());
	    fDedx1TrkTrueQE->Fill(qepid->Dedx());
	    fPdg1TrkTrueQE->Fill(qepid->Pdg());
	    fPid1TrkTrueQE->Fill(qepid->Value());
	  }
	  else{
	    fNtrk1TrkTrueNonQE->Fill(qepid->Ntrk());
	    fMode1TrkTrueNonQE->Fill(qepid->Mode());
	    fOffTrkE1TrkTrueNonQE->Fill(qepid->OffTrkE());
	    fEdiff1TrkTrueNonQE->Fill(qepid->Ediff());
	    fEdiffZ1TrkTrueNonQE->Fill(qepid->EdiffZ());
	    fDedx1TrkTrueNonQE->Fill(qepid->Dedx());
	    fPdg1TrkTrueNonQE->Fill(qepid->Pdg());
	    fPid1TrkTrueNonQE->Fill(qepid->Value());
	  }
	}
	else if(n3d >=1 && ntrks ==2){
	  fNtrk2Trk->Fill(qepid->Ntrk());
	  fMode2Trk->Fill(qepid->Mode());
	  fOffTrkE2Trk->Fill(qepid->OffTrkE());
	  fEdiff2Trk->Fill(qepid->Ediff());
	  fEdiffZ2Trk->Fill(qepid->EdiffZ());
	  fDedx2Trk->Fill(qepid->Dedx());
	  fPdg2Trk->Fill(qepid->Pdg());
	  fPid2Trk->Fill(qepid->Value());
	  if(isTrueQE){
	    fNtrk2TrkTrueQE->Fill(qepid->Ntrk());
	    fMode2TrkTrueQE->Fill(qepid->Mode());
	    fOffTrkE2TrkTrueQE->Fill(qepid->OffTrkE());
	    fEdiff2TrkTrueQE->Fill(qepid->Ediff());
	    fEdiffZ2TrkTrueQE->Fill(qepid->EdiffZ());
	    fDedx2TrkTrueQE->Fill(qepid->Dedx());
	    fPdg2TrkTrueQE->Fill(qepid->Pdg());
	    fPid2TrkTrueQE->Fill(qepid->Value());
	  }
	  else{
	    fNtrk2TrkTrueNonQE->Fill(qepid->Ntrk());
	    fMode2TrkTrueNonQE->Fill(qepid->Mode());
	    fOffTrkE2TrkTrueNonQE->Fill(qepid->OffTrkE());
	    fEdiff2TrkTrueNonQE->Fill(qepid->Ediff());
	    fEdiffZ2TrkTrueNonQE->Fill(qepid->EdiffZ());
	    fDedx2TrkTrueNonQE->Fill(qepid->Dedx());
	    fPdg2TrkTrueNonQE->Fill(qepid->Pdg());
	    fPid2TrkTrueNonQE->Fill(qepid->Value());
	  }
	}
	else{
	  fNtrk0Trk->Fill(qepid->Ntrk());
	  fMode0Trk->Fill(qepid->Mode());
	  fOffTrkE0Trk->Fill(qepid->OffTrkE());
	  fEdiff0Trk->Fill(qepid->Ediff());
	  fEdiffZ0Trk->Fill(qepid->EdiffZ());
	  fDedx0Trk->Fill(qepid->Dedx());
	  fPdg0Trk->Fill(qepid->Pdg());
	  fPid0Trk->Fill(qepid->Value());
	}

      }

 
    } // end loop over slices
    
  } // end of analyze

} // end of qeef namespace

namespace qeef
{
  DEFINE_ART_MODULE(QeFinderVal);
}
