////////////////////////////////////////////////////////////////////////
// \file    Track.cxx
// \brief   A track is a \ref Prong with full reconstructed trajectory.
// \version $Id: Track.cxx,v 1.22 2012-11-16 19:47:05 bckhouse Exp $
// \author  Christopher Backhouse - bckhouse@caltech.edu
////////////////////////////////////////////////////////////////////////
#include "RecoBase/Track.h"
#include "Geometry/Geometry.h"
#include "GeometryObjects/Geo.h"
#include "GeometryObjects/CellGeo.h"
#include "RecoBase/CellHit.h"

#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "messagefacility/MessageLogger/MessageLogger.h" 
#include "cetlib/exception.h"

namespace rb
{
  //......................................................................
  Track::Track(const art::PtrVector<rb::CellHit>& cells,
	       TVector3 start,
	       TVector3 dir,
	       int id)
    : rb::Prong(cells, start, dir, id)
  {
    AppendTrajectoryPoint(start);
  }

  //..................................................................
  Track::Track(const art::PtrVector<rb::CellHit>& cells,
	       geo::View_t view,
	       double v0,
	       double z0,
	       double dv,
	       double dz,
	       int id)
    : rb::Prong(cells, view, v0, z0, dv, dz, id)
  {
    AppendTrajectoryPoint(v0, z0);
  }

  //..................................................................
  Track::Track(const std::vector< art::Ptr<rb::CellHit> >& cells,
	       geo::View_t view,
	       double v0,
	       double z0,
	       double dv,
	       double dz,
	       int id)
    : rb::Prong(cells, view, v0, z0, dv, dz, id)
  {
    AppendTrajectoryPoint(v0, z0);
  }

  //......................................................................
  Track::Track(const rb::Cluster& clust,
	       TVector3 start,
	       TVector3 dir,
	       int id)
    : rb::Prong(clust, start, dir, id)
  {
    AppendTrajectoryPoint(start);
  }

  //......................................................................
  Track::Track(const rb::Cluster& clust,
	       double v0,
	       double z0,
	       double dv,
	       double dz,
	       int id)
    : rb::Prong(clust, v0, z0, dv, dz, id)
  {
    AppendTrajectoryPoint(v0, z0);
  }

  //......................................................................
  Track::Track(const rb::Prong& prong,
	       int id)
    : rb::Prong(prong)
  {
    SetID(id);
    AppendTrajectoryPoint(Start());
  }

  //......................................................................
  Track::~Track()
  {
  }

  //......................................................................
  void Track::AppendTrajectoryPoint(TVector3 pt)
  {
    fTraj.push_back(pt);
  }

  //......................................................................
  void Track::AppendTrajectoryPoint(double pv, double pz)
  {
    if(fView == geo::kXorY)
      throw cet::exception("Track") << "view is geo::kXorY, cannot add a 2D trajectroy point";

    if(fView == geo::kX)
      fTraj.push_back(TVector3(pv, 0, pz));
    else
      fTraj.push_back(TVector3(0, pv, pz));
  }

  //......................................................................
  void Track::PrependTrajectoryPoint(TVector3 pt)
  {
    fTraj.insert(fTraj.begin(), pt);

    fStart = pt;
  }

  //......................................................................
  void Track::PrependTrajectoryPoint(double pv, double pz)
  {
    if(fView == geo::kXorY)
      throw cet::exception("Track") << "view is geo::kXorY, cannot add a 2D trajectroy point";

    if(fView == geo::kX)
      PrependTrajectoryPoint(TVector3(pv, 0, pz));
    else
      PrependTrajectoryPoint(TVector3(0, pv, pz));
  }

  //......................................................................
  TVector3 Track::TrajectoryPoint(unsigned int i) const
  {
    if(i > fTraj.size())
      throw cet::exception("Track") << "attempt to get trajectory point " 
				    << i << " is out of range: " << fTraj.size();

    return fTraj[i];
  }

  //......................................................................
  void Track::SetStart(TVector3 start)
  {
    Prong::SetStart(start);

    if(fTraj.empty()) fTraj.resize(1);
    fTraj[0] = fStart;
  }

  //......................................................................
  void Track::SetStart(double v0, double z0)
  {
    Prong::SetStart(v0, z0);

    if(fTraj.empty()) fTraj.resize(1);
    fTraj[0] = fStart;
  }

  //......................................................................
  TVector3 Track::Stop() const
  {
    return fTraj.back();
  }

  //......................................................................
  // Get direction at the stop point
  TVector3 Track::StopDir() const
  {
    const int ntrajectory_points = NTrajectoryPoints();

    /// If number of trajectory points is less than 2,
    /// then can't figure out directions
    if(ntrajectory_points < 2){
        const TVector3 nullvector3;
        return nullvector3;
    }

    const TVector3 stop_point    = TrajectoryPoint(ntrajectory_points - 1);
    const TVector3 stop_point_m1 = TrajectoryPoint(ntrajectory_points - 2);

    const TVector3 diff          = stop_point - stop_point_m1;

    return diff.Unit();
  }

  //......................................................................
  double Track::TotalLength() const
  {
    if(fTraj.size() == 1) return 0;

    // We take the sum of the straight lines between the control points
    double dist = 0;

    const int N = fTraj.size();
    for(int n = 0; n < N-1; ++n){
      dist += (fTraj[n+1]-fTraj[n]).Mag();
    }

    return dist;
  }

  //......................................................................
  double Track::DistanceFromStart(double z) const
  {
    if(fTraj.size() == 1) return std::abs(z-fStart.Z());

    int i0, i1;
    FindNeighbouringPointIndices(z, i0, i1);

    // Distance along the track to the first of the neighbouring points
    double dist = 0;
    for(int i = 0; i < i0; ++i)
      dist += (fTraj[i+1]-fTraj[i]).Mag();

    // The extra accumulated between the points. Also correct for the cases
    // where we're off the start or end of the track.

    dist += (z-fTraj[i0].Z())/(fTraj[i1]-fTraj[i0]).Unit().Z();

    return dist;
  }

  //......................................................................
  double Track::DistanceFromEnd(double z) const
  {
    if(fTraj.size() == 1) return std::abs(z-fStart.Z());

    int i0, i1;
    FindNeighbouringPointIndices(z, i0, i1);

    // Distance from the end the track to the first of the neighbouring points
    double dist = 0;
    for(int i = fTraj.size()-1; i > i1; --i)
      dist += (fTraj[i]-fTraj[i-1]).Mag();

    // The extra accumulated between the points. Also correct for the cases
    // where we're off the start or end of the track.

    dist += (fTraj[i1].Z()-z)/(fTraj[i1]-fTraj[i0]).Unit().Z();

    return dist;
  }

  //......................................................................
  // Helper for Track::W()
  bool compZ(TVector3 v, double z)
  {
    return v.Z() < z;
  }

  //......................................................................
  double Track::W(const rb::CellHit* chit) const
  {
    if(fTraj.empty()) return 0;
    /// \todo CJB: do we want to allow the case with only one hit?
    if(fTraj.size() == 1) return Prong::W(chit);

    art::ServiceHandle<geo::Geometry> geom;
    double xyz[3];
    geom->Plane(chit->Plane())->Cell(chit->Cell())->GetCenter(xyz);

    InterpolateXY(xyz[2], xyz[0], xyz[1]);

    if(chit->View() == geo::kX) return xyz[1];
    return xyz[0];
    /*
    // Find the points before and after. This part does still assume that
    // searching along Z is the best way to find such points.
    TVector3 prev, next;
    FindNeighbouringPoints(xyz[2], prev, next);

    // Solve the problem of the closest approach of the trajectory to the line
    // made by the cell by projecting down to a 2D trajectory and 2D point for
    // the cell.
    TVector3 start2D = prev;
    TVector3 dir2D = next-prev;
    TVector3 pt2D(xyz);
    const int otherView = 1-chit->View();
    start2D[otherView] = dir2D[otherView] = pt2D[otherView] = 0;

    // Actually find the closest point
    TVector3 closest;
    geo::ClosestApproach(pt2D, start2D, dir2D, closest);

    // Need to solve for the W position of the line, the exact part that we
    // projected away above. So just figure out Z and evaluate the original
    // curve there.
    const double dt = (closest.Z()-prev.Z())/(next-prev).Z();

    return ((1-dt)*prev+dt*next)[otherView];
    */
  }

  //......................................................................
  void Track::InterpolateXY(double z, double& x, double& y) const
  {
    // The algorithm here is basically valid for z-directed tracks

    TVector3 a, b;
    FindNeighbouringPoints(z, a, b);

    InterpolatePts(z, a, b, x, y);
  }

  //......................................................................
  void Track::InterpolatePts(double z,
                             const TVector3& a, const TVector3& b,
                             double& x, double& y) const
  {
    const double denom = b.Z()-a.Z();
    const double dt = denom ? (z-a.Z())/denom : 0.5;

    x = (1-dt)*a.X()+dt*b.X();
    y = (1-dt)*a.Y()+dt*b.Y();
  }

  //......................................................................
  TVector3 Track::InterpolateDir(double z) const
  {
    // The algorithm here is basically valid for z-directed tracks

    TVector3 a, b;
    FindNeighbouringPoints(z, a, b);

    return (b-a).Unit();
  }

  //......................................................................
  void Track::ClearTrajectoryPoints()
  {
    fTraj.clear();
    fTraj.resize(0);
  }
  //.....................................................................
  void Track::RemoveTrajectoryPoint(unsigned int i)
  {
    if(i > fTraj.size())
      throw cet::exception("Track") << "attepmt to remove trajectory point " 
				    << i << " while there are only " << fTraj.size()
				    << " trajectory points";

    fTraj.erase(fTraj.begin()+i);
  }

  //......................................................................
  void Track::FindNeighbouringPointIndices(double z, int& i0, int& i1) const
  {
    if(fTraj.size() < 2)
      throw cet::exception("Track") << "attempting to find neighboring points with "
				    << "fewer than 2 trajectory points in track";

    // Track might be going upstream
    const int zDir = (fTraj.front().Z() > fTraj.back().Z()) ? -1 : 1;

    // If hit is off the end of the track extrapolate from the two closest points
    if( zDir*fTraj.front().Z() >= zDir*z ){
      i0 = 0;
      i1 = 1;
      return;
    }
    if( zDir*fTraj.back().Z() <= zDir*z ){
      i0 = fTraj.size()-2;
      i1 = fTraj.size()-1;
      return;
    }

    bool found = false;

    // We could do this more efficiently with std::[upper|lower]_bound, but the
    // presence of reversed tracks complicates that slightly.
    for(size_t n = 0; n < fTraj.size()-1; ++n){
      if( zDir*fTraj[n]  .Z() <= zDir*z && 
	  zDir*fTraj[n+1].Z() >= zDir*z ){
        // If we've already found a pair of neighbours then this new pair is
        // ambiguous which is bad. Probably it means the track reverses
        // direction wrt Z.
        if(found){
          // But, it might also mean the value passed in was right on a
          // trajectory point. In which case there isn't really a problem.
          // Either choice will end up just giving the value at this point.
          if(std::abs(fTraj[n].Z()-z) > 1e-3 && std::abs(fTraj[n+1].Z()-z) > 1e-3){
            LOG_DEBUG("Track")
              << "Choice of neighbouring trajectory points is ambiguous";
          }
        }
        i0 = n;
        i1 = n+1;
        found = true;
	break;
      }
    }

    if(!found)
      throw cet::exception("Track") << "unable to find neighboring point index";
  }

  //......................................................................
  void Track::FindNeighbouringPoints(double z,
				     TVector3& pt0, TVector3& pt1) const
  {
    int i0, i1;
    FindNeighbouringPointIndices(z, i0, i1);
    pt0 = fTraj[i0];
    pt1 = fTraj[i1];
  }

  bool comp_z(const TVector3& a, const TVector3& b)
  {
    return a.Z() < b.Z();
  }

  //......................................................................
  rb::Track Track::ZipWith(const rb::Track& trk) const
  {
    if(!Is2D() || !trk.Is2D())
      throw cet::exception("Track") << "attempting to zip a tracks with unknown dimensions; this track: "
				    << this->View() << " zipping track: " << trk.View();

    if(this->NTrajectoryPoints() < 2 || trk.NTrajectoryPoints() < 2)
      throw cet::exception("Track") << "too few points to attempt ZipWith: "
				    << this->NTrajectoryPoints() << " " 
				    << trk.NTrajectoryPoints();

    const rb::Track* xtrk = this;
    const rb::Track* ytrk = &trk;

    if(this->View() != geo::kX) std::swap(xtrk, ytrk);
    
    if(xtrk->View() != geo::kX || ytrk->View() != geo::kY)
      throw cet::exception("Track") << "views of 2D tracks are wrong";

    // Trajectory points upstream to downstream (which we always assume in the
    // output). User could reverse if they wanted.
    std::vector<TVector3> xpts = xtrk->fTraj;
    std::vector<TVector3> ypts = ytrk->fTraj;
    std::sort(xpts.begin(), xpts.end(), comp_z);
    std::sort(ypts.begin(), ypts.end(), comp_z);

    // The final 3D points
    std::vector<TVector3> pts;

    // check that the two sets of points line up with each other - ie that
    // there aren't more points at the start of one view than the other
    // if that happens, save those points for later and extrapolate based 
    // on the rest of the result
    std::vector<TVector3> extraPts;
    geo::View_t extraView = geo::kX;
    if(xpts.front().Z() < ypts.front().Z()){
      float diffZ = ypts.front().Z() - xpts.front().Z();
      // the maximum pathlength from one plane to the next is 8 cm
      while(diffZ > 8. && xpts.size() > 1){
	extraPts.push_back(xpts.front());
	xpts.erase(xpts.begin());
	diffZ = ypts.front().Z() - xpts.front().Z();
      }
    }
    else if(ypts.front().Z() < xpts.front().Z()){
      extraView = geo::kY;
      float diffZ = xpts.front().Z() - ypts.front().Z();
      // the maximum pathlength from one cell to the next is 8 cm
      while(diffZ > 8. && ypts.size() > 1){
	extraPts.push_back(ypts.front());
	ypts.erase(ypts.begin());
	diffZ = xpts.front().Z() - ypts.front().Z();
      }
    }

    // Go through the 2D points in Z order, and create a 3rd dimension for them
    // based on the closest two points in the other view. Equivalent to just
    // calling InterpolateXY for every point and then sorting, but this is
    // faster because InterpolateXY has to keep searching.
    double x = 0., y = 0.;
    // The point to take Z and one dimension from
    TVector3 pt;
    // The neighbours giving the other
    TVector3 nei1;
    TVector3 nei2;
    // Which view the neighbours contribute to
    geo::View_t view = geo::kXorY;
    const unsigned int X = xpts.size();
    const unsigned int Y = ypts.size();
    unsigned int curx = 0, cury = 0;

    while(curx < X && cury < Y){

      // If we ran out of one view, use the last two points to extrapolate
      if(curx == X){
        pt = ypts[cury];
        nei1 = xpts[X-2];
        nei2 = xpts[X-1];
        view = geo::kX;
      }
      else if(cury == Y){
        pt = xpts[curx];
        nei1 = ypts[Y-2];
        nei2 = ypts[Y-1];
        view = geo::kY;
      }
      // Both indices are OK, take the earlier one
      else if(xpts[curx].Z() < ypts[cury].Z()){
        pt = xpts[curx];
        // Before the start of the other view, must take first two indices
        if(cury == 0){
          nei1 = ypts[0];
          nei2 = ypts[1];
        }
        else{
          nei1 = ypts[cury-1];
          nei2 = ypts[cury];
        }
        view = geo::kY;
        ++curx;
      }
      else{
        pt = ypts[cury];
        if(curx == 0){
          nei1 = xpts[0];
          nei2 = xpts[1];
        }
        else{
          nei1 = xpts[curx-1];
          nei2 = xpts[curx];
        }
        view = geo::kX;
        ++cury;
      }

      InterpolatePts(pt.Z(), nei1, nei2, x, y);

      if(view == geo::kX)
        pts.push_back(TVector3(x, pt.Y(), pt.Z()));
      else
        pts.push_back(TVector3(pt.X(), y, pt.Z()));

    } // end while

    // may still have left over values in either the X or Y view	
    // unless there were exactly the same number in each view	
    // in this situation interpolation does not work so use
    // the vector from the previously found points to get the
    // value in the opposite view
    TVector3 vec;
    if(curx < X || cury < Y){
      vec = pts.back() - pts[pts.size() - 2];
//     LOG_VERBATIM("Track") << "\n\t" << pts.back().X() << " " << pts.back().Y() << " " << pts.back().Z()
// 			  << "\n\t" << (*ritr).X() << " " << (*ritr).Y() << " " << (*ritr).Z();

      
      while(curx < X){
	pt = xpts[curx];
	pt.SetY(pts.back().Y() + (pt.Z() - pts.back().Z())*vec.Unit().Y()/vec.Unit().Z());
	pts.push_back(pt);
	++curx;
      }
      while(cury < Y){
	pt = ypts[cury];
	pt.SetX(pts.back().X() + (pt.Z() - pts.back().Z())*vec.Unit().X()/vec.Unit().Z());
	pts.push_back(pt);
	++cury;
      }
    }// end if still have more to go in one view

    // check that we didn't accidentally add a repeat point on the end
    // if the difference between the last 2 points is less than 1 cm
    // just remove the last point.
    vec.SetXYZ(0., 0., 0.);
    while(vec.Mag() < 1){
      auto ritr = ++pts.rbegin();
      vec = pts.back() - *ritr;
      if(vec.Mag() < 1){
	pts.pop_back();
      }
    }

    // use the determined trajectory points to figure out
    // how to set the initial ones that didn't have information
    // from the other view
    // get the unit vector from the previous points
    vec = pts[0] - pts[1];
    for(auto &p : extraPts){
      if(     extraView == geo::kX) p.SetY(pts[0].Y() + (p.Z() - pts[0].Z())*vec.Unit().Y()/vec.Unit().Z());
      else if(extraView == geo::kY) p.SetX(pts[0].X() + (p.Z() - pts[0].Z())*vec.Unit().X()/vec.Unit().Z());
    }

    // the extra points are all at low z values by definition of this method,
    // make the track with them first, the direction of the track for these points is
    // based on the direction from the first two points in the pts vector
    rb::Track ret;
    if(extraPts.size() > 0){
      ret = Track(xtrk->AllCells(), extraPts[0], pts[1]-pts[0]);
//       LOG_VERBATIM("Track") << "track direction " << ret.Dir().X() << " " << ret.Dir().Y() << " " << ret.Dir().Z()
// 			 << "\nextra point: " << extraPts[0].X() << " " << extraPts[0].Y() << " " << extraPts[0].Z();
      for(size_t p = 1; p < extraPts.size(); ++p){
	ret.AppendTrajectoryPoint(extraPts[p]);
// 	LOG_VERBATIM("Track") << "extra point: " << extraPts[p].X() << " " << extraPts[p].Y() << " " << extraPts[p].Z();
      }
      for(auto &p : pts){
	ret.AppendTrajectoryPoint(p);
// 	LOG_VERBATIM("Track") << "point: " << p.X() << " " << p.Y() << " " << p.Z();
      }
    }
    else{
      ret = Track(xtrk->AllCells(), pts[0], pts[1]-pts[0]);
//       LOG_VERBATIM("Track") << "track direction " << ret.Dir().X() << " " << ret.Dir().Y() << " " << ret.Dir().Z()
// 			    << "\npoint: " << pts[0].X() << " " << pts[0].Y() << " " << pts[0].Z();
      for(size_t p = 1; p < pts.size(); ++p){
	ret.AppendTrajectoryPoint(pts[p]);
//  	LOG_VERBATIM("Track") << "point: " << pts[p].X() << " " << pts[p].Y() << " " << pts[p].Z();
      }
    }

    ret.Add(ytrk->AllCells());


    return ret;
  }

  //......................................................................
#if 0
  double Track::L(const rb::CellHit* chit) const
  {
    /// \todo CJB: do we want to allow the case with only one hit?
    if(fTraj.empty() || fTraj.size() == 1) return 0;

    // Find the two points we're interpolating using
    TVector3 a, b;
    FindNeighbouringPoints(chit, a, b);

    // Convert them into an origin and a direction
    double xyz[3] = {a.X(), a.Y(), a.Z()};
    double dxyz[3] = {b.X()-a.X(), b.Y()-a.Y(), b.Z()-a.Z()};

    // Find our cell, and make the geometry consider nothing else
    art::ServiceHandle<geo::Geometry> geom;
    const geo::CellGeo* cellgeo = geom->Plane(chit->Plane())->Cell(chit->Cell());
    const TGeoNode* cellnode = cellgeo->Node();

    // Looking at the implementation, this happens to be safe
    cellgeo->WorldToLocal(xyz, xyz);
    cellgeo->WorldToLocalVect(dxyz, dxyz);

    // This structure ensures we put the manager back as we found it
    struct GeoRestorer
    {
      GeoRestorer()
      {
  vol = gGeoManager->GetTopVolume();
  trk = gGeoManager->GetCurrentTrack();
      }
      ~GeoRestorer()
      {
  gGeoManager->SetTopVolume(vol);
  gGeoManager->SetCurrentTrack(trk);
      }
      TGeoVolume* vol;
      TVirtualGeoTrack* trk;
    } restorer;

    gGeoManager->SetTopVolume(cellnode->GetVolume());

    // Shoot a ray from our starting point. There are four possibilities:
    // the cell is in front of us, behind, we're in it, or we miss it.
    gGeoManager->InitTrack(xyz, dxyz);

    // TODO CJB: The assertions in here were firing, so they're commented out
    // TODO CJB: Understand the coordinate system. Is step really in m and everything else in cm?

    if(gGeoManager->FindNextBoundary()){
      // If we hit something and we're entering
      if(gGeoManager->IsStepEntering()){
        // Then if we keep going we must find the exit
        gGeoManager->Step(true, true);
        /*TGeoNode* node =*/ gGeoManager->FindNextBoundary();
        //      assert(node && gGeoManager->IsStepExiting());
        // Length of step from entry to exit
        return 100*gGeoManager->GetStep();
      }
      else{
        // If we're not entering, we must be exiting and hence inside the cell
        //      assert(gGeoManager->IsStepExiting());
        const double ell = gGeoManager->GetStep();
        // Reverse the direction vector
        for(int n = 0; n < 3; ++n) dxyz[n] *= -1;
        // Now we must be able to exit in the opposite direction too
        gGeoManager->InitTrack(xyz, dxyz);
        /*TGeoNode* node =*/ gGeoManager->FindNextBoundary();
        //      assert(node && gGeoManager->IsStepExiting());
        // Sum of the two routes to exit
        return 100*(ell+gGeoManager->GetStep());
      }
    }
    else{
      // We didn't hit anything first go. Try the other direction
      for(int n = 0; n < 3; ++n) dxyz[n] *= -1;
      gGeoManager->InitTrack(xyz, dxyz);
      // If we missed again, then this ray doesn't intersect the cell
      if(!gGeoManager->FindNextBoundary()) return 0;
      // Otherwise, we must enter and exit the cell, repeating the first case
      //    assert(gGeoManager->IsStepEntering());
      gGeoManager->Step(true, true);
      /*TGeoNode* node =*/ gGeoManager->FindNextBoundary();
      //    assert(node && gGeoManager->IsStepExiting());
      return 100*gGeoManager->GetStep();
    }
  }


#endif

  //-----------------------------------------------------------------------------
  std::map<unsigned int, TVector3> Track::PlaneDirMap() const
  {
    std::map<unsigned int, TVector3> planeDirMap;

    // loop over the planes in the track and get the points on either side
    // of the plane z
    art::ServiceHandle<geo::Geometry> geom;

    // looping over the cells this way does the X and Y
    // view separately, but the map orders by plane
    // number using the < operator, so the entries
    // in the map are linearly increasing, with X and Y
    // views mixed together.
    for(size_t c = 0; c < NCell(); ++c){
      unsigned int plane = Cell(c)->Plane();

      // check if we already have the direction at this plane
      // no need to overwrite it if we do.
      if(planeDirMap.count(plane) > 0) continue;

      // get the z position for this cell
      double xyz[3] = {0.};
      geom->Plane(plane)->Cell(Cell(c)->Cell())->GetCenter(xyz);

      // find the direction based on the points on either
      // side of this z location.
      // set the direction of travel as the direction
      // at this plane.
      planeDirMap[plane] = InterpolateDir(xyz[2]);

    }// end loop over cells

    return planeDirMap;
  }

} // end namespace rb
//////////////////////////////////////////////////////////////////////////////
