////////////////////////////////////////////////////////////////////////////////////
/// \file    Vertex.h
/// \brief   Vertex location in position and time
/// \version $Id: Vertex.h,v 1.8 2012-10-26 21:31:55 bckhouse Exp $
/// \author  eniner@indiana.edu
/////////////////////////////////////////////////////////////////////////////////////
#ifndef RB_VERTEX_H
#define RB_VERTEX_H

#include "art/Persistency/Common/PtrVector.h"

#include "TVector3.h"

namespace rb{ class Cluster; class Prong; class Track; }

namespace rb
{
  /// \brief A 3D position and time plus associated reconstructed objects
  ///
  /// A vertex may have any number of particles originating from it, in any
  /// degree of reconstruction, from simple clusters up to full tracks.
  /// The interface allows you to loop over all objects of any one type, or
  /// over the objects of a type plus its derived types (eg
  /// \ref ProngsAndTracks allows you to loop over all objects that have a
  /// defined direction.
  class Vertex
  {
  public:
    Vertex();
#ifndef __GCCXML__
    Vertex(TVector3 xyz, double t);
    Vertex(double x, double y, double z, double t);

    void Add(art::Ptr<rb::Cluster> c);
    void Add(art::Ptr<rb::Prong> p);
    void Add(art::Ptr<rb::Track> t);

    double GetX() const {return fX;}
    double GetY() const {return fY;}
    double GetZ() const {return fZ;}
    double GetT() const {return fT;}
    TVector3 GetXYZ() const;

    void SetX(double x){fX = x;}
    void SetY(double y){fY = y;}
    void SetZ(double z){fZ = z;}
    void SetT(double t){fT = t;}
    void SetXYZ(double x, double y, double z);
    void SetXXYZ(TVector3 xyz);

    /// \brief How many clusters are associated with this vertex?
    ///
    /// Not including prongs or tracks.
    unsigned int NClusters() const;
    /// Retrieve the cluster at index \a clustIdx (0-NClusters).
    art::Ptr<rb::Cluster> Cluster(unsigned int clustIdx) const;

    /// How many prongs (not including tracks) are associated with this vertex?
    unsigned int NProngs() const;
    /// Retrieve the prong at index \a prongIdx (0-NProngs).
    art::Ptr<rb::Prong> Prong(unsigned int prongIdx) const;

    /// How many tracks are associated with this vertex?
    unsigned int NTracks() const;
    /// Retrieve the track at index \a trkIdx (0-NTracks).
    art::Ptr<rb::Track> Track(unsigned int trkIdx) const;

    /// Total number of prongs and tracks
    unsigned int NProngsAndTracks() const;
    /// /brief Interface to prongs and the prong component of tracks
    ///
    /// Index is from 0 to \ref NProngsAndTracks
    const rb::Prong* ProngsAndTracks(unsigned int idx) const;

    /// Total number of reconstructed objects of all types
    unsigned int NClustersProngsAndTracks() const;
    /// \brief Retrieve the cluster component of any associated object
    ///
    /// Index is from 0 to \ref NClustersProngsAndTracks
    const rb::Cluster* ClustersProngsAndTracks(unsigned int idx) const;
#endif // __GCCXML__
  protected:
    double fX; ///< X coordinate (cm)
    double fY; ///< Y coordinate (cm)
    double fZ; ///< Z coordinate (cm)
    double fT; ///< Time         (ns)

    art::PtrVector<rb::Cluster> fClusters;
    art::PtrVector<rb::Prong> fProngs;
    art::PtrVector<rb::Track> fTracks;
  };

} // end namespace rb

#endif
////////////////////////////////////////////////////////////////////////
