////////////////////////////////////////////////////////////////////////
/// \brief   Example to use RecoJMShower 
///          for particle identification
/// \author  Jianming Bian
////////////////////////////////////////////////////////////////////////

#include <cmath>
#include <iostream>
#include <fstream>
#include <string>
#include <vector>

// ROOT includes
#include "TFile.h"
#include "TTree.h"

// NOvA includes
#include "Calibrator/Calibrator.h"
#include "RecoJMShower/JMShower.h"
#include "SummaryData/POTSum.h"

// ART includes
#include "art/Framework/Core/EDAnalyzer.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Principal/SubRun.h"
#include "art/Framework/Services/Optional/TFileService.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"


namespace jmshower {

  class RecoJMShowerAna : public art::EDAnalyzer
    {
    public:
      explicit RecoJMShowerAna(fhicl::ParameterSet const& pset); 
      ~RecoJMShowerAna();                        
      void analyze(const art::Event& evt);
      void beginJob();
      void beginSubRun(const art::SubRun& sr);
      void reconfigure(const fhicl::ParameterSet& pset);

      std::string   fPotLabel; ///< Module that produced the POTSum object

    private:

      TTree* fShower;
      double m_showerRun;
      double m_showerSubRun;
      double m_showerEvent;
      double m_showerId;


      int    m_showerTrueNuCCNC;
      int    m_showerTrueNuMode;
      int    m_showerTrueNuPdg;
      double m_showerTrueNuEnergy;
      int    m_showerTrueNuIsFid;
      double m_showerTrueNuP4[4];
      double m_showerTrueNuVtx[3];
      int    m_showerEvtTruePdg[20];
      double m_showerEvtTrueEnergy[20];

      double m_showerTrueDang;
      double m_showerTruePdg;
      double m_showerTrueMother;
      double m_showerTrueVtx[3];
      double m_showerTrueP4[4];
      double m_showerTruePartHit[10];
      double m_showerTruePartEnergy[10];

      double m_showerNPlane;
      double m_showerNTrk;
      double m_showerNXcell;
      double m_showerNYcell;
      double m_showerNMIPPlane;
      double m_showerContStartPlane;
      double m_showerDir[3];
      double m_showerStart[3];
      double m_showerStop[3];
      double m_showerPlaneEnergy[200];
      double m_showerPlaneDedx[200];
      double m_showerPlaneDeltaDedx[200];
      double m_showerTransCellDedx[20];
      double m_showerTransCellProbE[20];
      double m_showerTransCellProbPi0[20];

      double m_showerPlaneDedxprobE[200];
      double m_showerPlaneDedxprobG[200];
      double m_showerPlaneDedxprobMu[200];
      double m_showerPlaneDedxprobPi0[200];
      double m_showerDedxChisq[10];
      double m_showerRadiusChisq[10];
      double m_showerDedxNdof[10];
      double m_showerDedxProb[10];
      double m_showerTransDedxProb[10];
      double m_showerDedxLLL[10];
      double m_showerDedxLLL1[10];
      double m_showerDedxLLL2[10];
      double m_showerDedxLLL3[10];
      double m_showerDedxLLL4[10];
      double m_showerDedxLLL5[10];
      double m_showerDedxLLL6[10];
      double m_showerDedxLLT[10];
      double m_showerPlaneRadius[200];
      double m_showerADC;
      double m_showerEnergy0;
      double m_showerEnergy;
      double m_showerEnergySvc;
      double m_showerNCell;
      double m_showerRadius;
      double m_showerGap;
      double m_showerIsFid;
      double m_showerRminmax;

    };
}

namespace jmshower{
  //_________________________________________________________________________

  RecoJMShowerAna::RecoJMShowerAna(fhicl::ParameterSet const& pset)
    : EDAnalyzer(pset)
  {
    reconfigure(pset);
  }
  
  //......................................................................
  
  void RecoJMShowerAna::reconfigure(const fhicl::ParameterSet& pset)
  {
    fPotLabel       =pset.get< std::string>("PotLabel");
  }
  
  //......................................................................
  
  RecoJMShowerAna::~RecoJMShowerAna()
  {
  }
  
  //......................................................................
  
  void RecoJMShowerAna::beginJob()
  {
    std::cout<<"RecoJMShowerAna::beginJob()"<<std::endl;
    art::ServiceHandle<art::TFileService> tfs;
    fShower = tfs->make<TTree>("fShower","fShower");
    fShower->Branch("showerRun",&m_showerRun);
    fShower->Branch("showerSubRun",&m_showerSubRun);
    fShower->Branch("showerEvent",&m_showerEvent);
    fShower->Branch("showerTrueNuCCNC", &m_showerTrueNuCCNC);
    fShower->Branch("showerTrueNuMode", &m_showerTrueNuMode);
    fShower->Branch("showerTrueNuPdg", &m_showerTrueNuPdg);
    fShower->Branch("showerTrueNuEnergy", &m_showerTrueNuEnergy);
    fShower->Branch("showerTrueNuIsFid", &m_showerTrueNuIsFid);
    fShower->Branch("showerTrueNuP4[4]", m_showerTrueNuP4);
    fShower->Branch("showerTrueNuVtx[3]", m_showerTrueNuVtx);
    fShower->Branch("showerEvtTrueEnergy[20]", m_showerEvtTrueEnergy);
    fShower->Branch("showerEvtTruePdg[20]", m_showerEvtTruePdg);
    fShower->Branch("showerId",&m_showerId);
    fShower->Branch("showerNPlane", &m_showerNPlane); 
    fShower->Branch("showerNTrk", &m_showerNTrk); 
    fShower->Branch("showerNXcell", &m_showerNXcell); 
    fShower->Branch("showerNYcell", &m_showerNYcell); 
    fShower->Branch("showerDir[3]", m_showerDir);
    fShower->Branch("showerStart[3]", m_showerStart);
    fShower->Branch("showerStop[3]", m_showerStop);
    fShower->Branch("showerADC", &m_showerADC);
    fShower->Branch("showerEnergy0", &m_showerEnergy0);
    fShower->Branch("showerEnergy", &m_showerEnergy);
    fShower->Branch("showerNCell", &m_showerNCell);
    fShower->Branch("showerRadius", &m_showerRadius);
    fShower->Branch("showerGap", &m_showerGap);
    fShower->Branch("showerIsFid", &m_showerIsFid);
    fShower->Branch("showerNMIPPlane", &m_showerNMIPPlane);
    fShower->Branch("showerContStartPlane", &m_showerContStartPlane);
    fShower->Branch("showerRminmax", &m_showerRminmax);
    fShower->Branch("showerPlaneEnergy[200]", m_showerPlaneEnergy);
    fShower->Branch("showerPlaneDedx[200]", m_showerPlaneDedx);
    fShower->Branch("showerDeltaPlaneDedx[200]", m_showerPlaneDeltaDedx);
    fShower->Branch("showerDedxChisq[10]", m_showerDedxChisq);
    fShower->Branch("showerRadiusChisq[10]", m_showerRadiusChisq);
    fShower->Branch("showerDedxNdof[10]", m_showerDedxNdof);
    fShower->Branch("showerDedxProb[10]", m_showerDedxProb);
    fShower->Branch("showerTransDedxProb[10]", m_showerTransDedxProb);
    fShower->Branch("showerDedxLLL[10]", m_showerDedxLLL);
    fShower->Branch("showerDedxLLL1[10]", m_showerDedxLLL1);
    fShower->Branch("showerDedxLLL2[10]", m_showerDedxLLL2);
    fShower->Branch("showerDedxLLL3[10]", m_showerDedxLLL3);
    fShower->Branch("showerDedxLLL4[10]", m_showerDedxLLL4);
    fShower->Branch("showerDedxLLL5[10]", m_showerDedxLLL5);
    fShower->Branch("showerDedxLLL6[10]", m_showerDedxLLL6);
    fShower->Branch("showerDedxLLT[10]", m_showerDedxLLT);
    fShower->Branch("showerPlaneRadius[200]", m_showerPlaneRadius);
    fShower->Branch("showerPlaneDedxprobE[200]", m_showerPlaneDedxprobE);
    fShower->Branch("showerPlaneDedxprobG[200]", m_showerPlaneDedxprobG);
    fShower->Branch("showerPlaneDedxprobMu[200]", m_showerPlaneDedxprobMu);
    fShower->Branch("showerPlaneDedxprobPi0[200]", m_showerPlaneDedxprobPi0);
    fShower->Branch("showerTransCellDedx[20]", m_showerTransCellDedx);
    fShower->Branch("showerTransCellProbE[20]", m_showerTransCellProbE);
    fShower->Branch("showerTransCellProbPi0[20]", m_showerTransCellProbPi0);
    fShower->Branch("showerTrueDang", &m_showerTrueDang);
    fShower->Branch("showerTruePdg", &m_showerTruePdg);
    fShower->Branch("showerTrueMother", &m_showerTrueMother);
    fShower->Branch("showerTrueVtx[3]", m_showerTrueVtx);
    fShower->Branch("showerTrueP4[4]", m_showerTrueP4);
    fShower->Branch("showerTruePartHit[10]", m_showerTruePartHit);
    fShower->Branch("showerTruePartEnergy[10]", m_showerTruePartEnergy);
  }
  
  void RecoJMShowerAna::beginSubRun(art::SubRun const& sr) {
    art::Handle< sumdata::POTSum > p;
    sr.getByLabel(fPotLabel,p);
    std::cout<<"pot = "<<p->totgoodpot<<std::endl;
  } 
  //......................................................................
  void RecoJMShowerAna::analyze(const art::Event& evt)
  {
    //----------------------------------------------------
    //  Run, subrun and event information
    //----------------------------------------------------
    int run   = evt.run();
    int srun  = evt.subRun();
    int event = evt.id().event();
    art::Handle< std::vector<jmshower::JMShower> > showerhandle;
    evt.getByLabel("recojmshower", showerhandle);
    std::vector<art::Ptr<jmshower::JMShower> >  svcshowercol;
    for(unsigned int i = 0; i < showerhandle->size(); i++){
      art::Ptr<jmshower::JMShower> p(showerhandle, i);
      svcshowercol.push_back(p);
    }
    
    
    
    //**********************************************************
    //         Fill showers information
    //**********************************************************
    
    for(unsigned int i = 0; i < svcshowercol.size(); i++){
      // General information 
      double showerRadius    = svcshowercol[i]->Radius();
      double showerEnergy    = svcshowercol[i]->Energy();
      double showerEnergy0    = svcshowercol[i]->DepositEnergy();
      int  showerIsFid    = int(svcshowercol[i]->IsFiducial());
      if(showerEnergy>0.001)showerRadius = showerRadius/showerEnergy;
      m_showerNCell = svcshowercol[i]->NCell();
      m_showerNPlane = svcshowercol[i]->ExtentPlane();
      m_showerADC = svcshowercol[i]->TotalADC ();
      m_showerEnergy = showerEnergy;
      m_showerEnergy0 = showerEnergy0;
      m_showerDir[0] = (svcshowercol[i]->Dir())[0];
      m_showerDir[1] = (svcshowercol[i]->Dir())[1];
      m_showerDir[2] = (svcshowercol[i]->Dir())[2];
      m_showerStart[0]  = (svcshowercol[i]->Start())[0];
      m_showerStart[1]  = (svcshowercol[i]->Start())[1];
      m_showerStart[2]  = (svcshowercol[i]->Start())[2];
      m_showerStop[0]  = (svcshowercol[i]->Stop())[0];
      m_showerStop[1]  = (svcshowercol[i]->Stop())[1];
      m_showerStop[2]  = (svcshowercol[i]->Stop())[2];
      
      
      m_showerRadius = showerRadius;
      m_showerIsFid = showerIsFid;
      m_showerNMIPPlane = svcshowercol[i]->NMIPPlane();
      // Fill longitudinal and transverse information
      
      for( int itype = int(jmshower::kNull)+1; itype != int(jmshower::kLast); ++itype)
	{ //type: 0-electron, 1-photon, 2-muon, 3-pion-zero, 4-inc-had, 5-proton, 6-neutron, 7-charged-pion; 
	  m_showerDedxNdof[itype]=svcshowercol[i]->ExtentPlane();
	  m_showerDedxProb[itype] = svcshowercol[i]->DedxLongLL(itype);
	  m_showerTransDedxProb[itype] = svcshowercol[i]->DedxTransLL(itype);
	  m_showerDedxLLL[itype] = svcshowercol[i]->DedxLongLL(itype);
	  m_showerDedxLLT[itype] = svcshowercol[i]->DedxTransLL(itype);
	}
      
      m_showerRun= run; 
      m_showerSubRun= srun;
      m_showerEvent= event;
      m_showerId= i;
      m_showerNTrk= svcshowercol.size();
      fShower->Fill();
    }
  }
}// end namespace jmshower

namespace jmshower
{
  DEFINE_ART_MODULE(RecoJMShowerAna);
}
