#include "RunHistory/RunHistory.h"
#ifdef NOVACMAKE
#include "Database/cxx/include/Table.h"
#else
#include "Database/Table.h"
#endif
#include <libpq-fe.h>
#include <getopt.h>
#include <iostream>

bool printNActive = false;
bool printNAnalysis = false;
bool printPixelMasks = false;
bool printPixelThresholds = false;
bool printInstalledAPDs = false;
bool printNInstalledAPDs = false;
bool printAPDSettings = false;
bool printPOT = false;
bool printPartition = false;
bool printRunType = false;
bool printGoodDiBlockMask = false;
bool printDQSRInfo = false;

int run = -1;
std::string tstart = "";
std::string detStr="";

void PrintUsage()
{
  std::cout << "Usage: NovaRunInfo [options] -d|--detector [detectorName, eg FarDet] -r|--run [run number] OR -t|--time [time stamp, eg, 2013-10-18T15:00:00+5]" << std::endl;
  std::cout << "\noptions:" << std::endl;
  std::cout << "\t-a (--nactive)    : print number of active channels " << std::endl;
  std::cout << "\t-A (--nanalysis)    : print number of analysis channels " << std::endl;
  std::cout << "\t-p (--pixmasks)   : print individual FEB pixel masks" << std::endl;
  std::cout << "\t-q (--pixthresh)   : print individual FEB pixel masks" << std::endl;
  std::cout << "\t-i (--installed)  : print list of installed APDs" << std::endl;
  std::cout << "\t-I (--ninstalled) : print number of installed APDs" << std::endl;
  std::cout << "\t-s (--apdsettings): print APD settings" << std::endl;
  std::cout << "\t-P (--POT)        : print POT" << std::endl;
  std::cout << "\t-z (--partition)  : print partition number" << std::endl;
  std::cout << "\t-T (--runtype)    : print run type" << std::endl;
  std::cout << "\t-G (--gooddiblock) : print good diblock mask" << std::endl;
  std::cout << "\t-Q (--dqsubruninfo) : print Data Quality subrun info" << std::endl;
}

//------------------------------------------------------------

bool ParseCLArgs(int argc, char* argv[])
{

  struct option long_options[] = {
    {"help",   0, 0, 'h'},
    {"run",   0, 0, 'r'},
    {"time",   0, 0, 't'},
    {"detector",   0, 0, 'd'},
    {"nactive",0, 0, 'a'},
    {"nanalysis",0, 0, 'A'},
    {"pixelmasks",0, 0, 'p'},
    {"pixelthresh",0, 0, 'q'},
    {"installed",0, 0, 'i'},
    {"ninstalled",0, 0, 'I'},
    {"apdsettings",0,0,'s'},
    {"POT",0,0,'P'},
    {"partition",0,0,'z'},
    {"runtype",0,0,'T'},
    {"gooddiblock",0,0,'G'},
    {"dqsubruninfo",0,0,'Q'},
    {0,0,0,0}
  };
  
  struct tm ta;
  int tz;
  bool hasTZ = false;

  while (1) {
    int optindx;
    int c = getopt_long(argc,argv,"r:t:d:aAbpPiIhszTGqQ",long_options,&optindx);
        
    if (c==-1) break;
    
    switch(c) {
    case 'r':
      run = atoi(optarg);
      if (run < 10000) {
	std::cout << "Invalid run number." << std::endl;
	exit(0);
      }
      break;
    case 't':
      tstart = optarg;
      if (tstart.length() == 19)
	sscanf(tstart.c_str(),"%4d-%2d-%2dT%2d:%2d:%2d",
	       &ta.tm_year,&ta.tm_mon,&ta.tm_mday,
	       &ta.tm_hour,&ta.tm_min,&ta.tm_sec);
      else {
	sscanf(tstart.c_str(),"%4d-%2d-%2dT%2d:%2d:%2d%2d",
	       &ta.tm_year,&ta.tm_mon,&ta.tm_mday,
	       &ta.tm_hour,&ta.tm_min,&ta.tm_sec,&tz);
	hasTZ = true;
      }
      
      ta.tm_year -= 1900;
      ta.tm_mon -= 1;
      if (hasTZ) ta.tm_hour += tz;
      char buff[64];
      strftime(buff,64,"%Y-%m-%d %H:%M:%S",&ta);
      tstart=buff;
      break;
    case 'd':
      detStr = optarg;
      break;
    case 'h':
      PrintUsage();
      exit(0);
      break;
    case 'p':
      printPixelMasks = true;
      break;
    case 'q':
      printPixelThresholds = true;
      break;
    case 'a':
      printNActive = true;
      break;
    case 'A':
      printNAnalysis = true;
      break;
    case 'i':
      printInstalledAPDs = true;
      break;
    case 'I':
      printNInstalledAPDs = true;
      break;
    case 's':
      printAPDSettings = true;
      break;
    case 'P':
      printPOT = true;
      break;
    case 'G':
      printGoodDiBlockMask = true;
      break;
    case 'z':
      printPartition = true;
      break;
    case 'T':
      printRunType = true;
      break;
    case 'Q':
      printDQSRInfo = true;
      break;
    default:
      break;
    }
  }

  if (detStr == "") {
    PrintUsage();
    exit(0);
  }
  if (run == -1 && tstart=="") {
    PrintUsage();
    exit(0);
  }
  if (tstart != "" && run>0) {
    std::cerr << "Ambiguous command, please provide either run number OR timestamp!" << std::endl;
    PrintUsage();
    exit(0);
  }

  return true;
}

//------------------------------------------------------------

int main(int argc, char **argv)
{
  if (!ParseCLArgs(argc,argv)) {
    PrintUsage();
    return 1;
  }

  if (tstart != "") {
    nova::dbi::Table* rTable = new nova::dbi::Table();
    rTable->SetTableName("runs");
    rTable->SetDetector(detStr);
    rTable->SetTableType(nova::dbi::kGenericTable);
    rTable->SetIgnoreEnvVar(true);
    rTable->SetUser(getenv("NOVADBUSER"));
    rTable->SetVerbosity(0);
    if (detStr == "FarDet") {
      rTable->SetDBHost(getenv("NOVAFARDAQDBHOST"));
      rTable->SetDBName(getenv("NOVAFARDAQDBNAME"));
      rTable->SetDBPort(getenv("NOVAFARDAQDBPORT"));      
    }
    else if (detStr == "NearDet") {
      rTable->SetDBHost(getenv("NOVANEARDAQDBHOST"));
      rTable->SetDBName(getenv("NOVANEARDAQDBNAME"));
      rTable->SetDBPort(getenv("NOVANEARDAQDBPORT"));
    }

    if (!rTable->GetConnection()) {
      std::cerr << "Cannot get connection to database!" << std::endl;
      return 1;
    }
    PGresult* res;
    
    std::stringstream ss;
    ss << "select run,partition from " << detStr << ".runs where tstart<'" 
       << tstart << "' and tstop>'" << tstart << "' and tstop is not NULL";
    std::string query = ss.str();

    if (rTable->ExecuteSQL(query,res)) {
      if (PQresultStatus(res) != PGRES_TUPLES_OK) {
	std::cerr << "Query failed: " << query << std::endl;
	PQclear(res);
	rTable->CloseConnection();
	return 1;
      }
      
      int nRow = PQntuples(res);
      if (nRow == 0) {
	std::cerr << "No run found at " << tstart << std::endl;
	return 1;
      }
      else if (nRow > 1) {
	std::cerr << "Multiple runs found at " << tstart << ":\n";
	for (int i=0; i < nRow; i++) 
	  std::cerr << "Run " << PQgetvalue(res,i,0) << ", Partition " 
		    << PQgetvalue(res,i,1) << std::endl;
	return nRow;
      }
      else {
	run = atoi(PQgetvalue(res,0,0));
      }
    }
  }

  nova::dbi::RunHistory rh(run,detStr);
  rh.SetTimeQueries(false);
  rh.SetTimeParsing(false);
  if (printNAnalysis) {
    rh.SetIfAnyMasking(true);
    rh.SetAnalysisDiBlockMasking(true);
    rh.SetIfCustomMasking(false);
  }

  std::cout << "Run: " << rh.RunNumber() << std::endl;
  std::cout << "Number of subruns: " << rh.NSubruns() << std::endl;
  std::cout << "tStart = " << rh.TStartAsString() << std::endl;
  std::string tstop = rh.TStopAsString();
  if (tstop == "")
    tstop = "Unknown (the run either crashed or is in progress)";
  std::cout << "tStop = " << tstop << std::endl;

  if (printPOT)
    std::cout << "Recorded POT (estimate): " << rh.NPOT() << std::endl;

  if (printPartition)
    std::cout << "Partition: " << rh.Partition() << std::endl;

  if (printRunType)
    std::cout << "Run Type: " << rh.RunType() << std::endl;

  if (printNActive) {
    std::cout << "Number of Active Channels: " << rh.NActiveChannels();
    std::cout << "\n";
  }

  if (printNAnalysis) {
    std::cout << "Number of Analysis Channels: " << rh.NAnalysisChannels();
    std::cout << "\n";
  }

  if (printGoodDiBlockMask) {
    std::cout << "Good DiBlock Mask: 0x" << std::hex << rh.GoodDiBlockMask()
	      << std::endl;
    std::cout << std::dec;
  }

  if (printDQSRInfo) {
    std::cout << "Data Quality Subrun Information: " << std::endl;
    std::vector<nova::dbi::DQSubrunInfo> dqsr = rh.GetDQSubrunInfo();
    for (size_t i=0; i<dqsr.size(); ++i)
      std::cout << dqsr[i] << std::endl;
  }

  if (printPixelMasks) {
    if (rh.LoadPixelInfo()) {      
      for (int idb=0; idb<rh.NDiBlocks(); ++idb) {
	const nova::dbi::RunHistory::DiBlock& db = rh.GetDiBlock(idb,false);
	for (unsigned int idcm=0; idcm<db.dcm.size(); ++idcm) {
	  if (db.dcm[idcm].isActive) {
	    int64_t febmask = db.dcm[idcm].febMask;
	    for (int j=0; j<64; ++j) {
	      if ((febmask >> j)&0x1)
		std::cout << db.dcm[idcm].appname << " " << std::dec << j << " " << std::hex << (0xffffffff & db.dcm[idcm].feb[j].pixelMask) << std::endl;
	    }
	  }
	}
      }
      std::cout << std::dec;
      
    }
    else {
      std::cout << "Failed to load pixel masks!" << std::endl;
    }
  }

  if (printPixelThresholds) {
    if (rh.LoadPixelInfo()) {      
      for (int idb=0; idb<rh.NDiBlocks(); ++idb) {
	const nova::dbi::RunHistory::DiBlock& db = rh.GetDiBlock(idb,false);
	for (unsigned int idcm=0; idcm<db.dcm.size(); ++idcm) {
	  if (db.dcm[idcm].isActive) {
	    int64_t febmask = db.dcm[idcm].febMask;
	    for (int j=0; j<64; ++j) {
	      if ((febmask >> j)&0x1) {
		std::cout << db.dcm[idcm].appname << "-" << j << ": "; 
		for (int k=0; k<32; ++k) {
		  std::cout << db.dcm[idcm].feb[j].pixelThresh[k] << " ";
		}
		std::cout << std::endl;
	      }
	    }
	  }
	}
      }
    }
    else {
      std::cout << "Failed to load pixel thresholds!" << std::endl;
    }
  }

  if (printNInstalledAPDs) {
    if (rh.LoadHardwareInfo()) {
      std::cout << "Number of installed APDs: " << rh.NAPDInstalled()
		<< std::endl;      
    }
  }
  
  if (printInstalledAPDs) {
    if (rh.LoadHardwareInfo()) {
      std::cout << "Installed APD list: " << std::endl;
      for (int idb=0; idb<rh.NDiBlocks(); ++idb) {
	const nova::dbi::RunHistory::DiBlock& db = rh.GetDiBlock(idb,false);
	for (unsigned int idcm=0; idcm<db.dcm.size(); ++idcm) {
	  if (db.dcm[idcm].isActive) {
	    for (int j=0; j<64; ++j) {	    
	      if (db.dcm[idcm].feb[j].hasAPD) 
		std::cout << db.dcm[idcm].appname << "-" << j << std::endl;
	    }
	  }
	}
      }
    }
    else {
      std::cout << "Failed to load installation information!" << std::endl;
    }
  }

  if (printAPDSettings) {
    if (rh.LoadAPDSettings()) {
      std::cout << "APD Settings: " << std::endl;
      for (int idb=0; idb<rh.NDiBlocks(); ++idb) {
	const nova::dbi::RunHistory::DiBlock& db = rh.GetDiBlock(idb);
	for (unsigned int idcm=0; idcm<db.dcm.size(); ++idcm) {
	  if (db.dcm[idcm].isActive) {
	    for (int j=0; j<64; ++j) {
	      const nova::dbi::RunHistory::FEB& feb = db.dcm[idcm].feb[j];
	      if (feb.hasAPD && feb.gain < 1.) {
		std::cout << db.dcm[idcm].appname << "-" << j 
			  << ":" << std::endl;
		std::cout << "\tVoltage: " << feb.apdV 
			  << std::endl;
		std::cout << "\tGain: " << feb.gain
			  << std::endl;
		
		std::cout << "\tPixelGains: ";
		for(auto pgitr : feb.pixelGain)
		  std::cout << pgitr << " ";
		std::cout << std::endl;
		
	      }
	    }
	  }
	}
      }
    }
    else {
      std::cout << "Failed to load APD settings information!" << std::endl;
    }
  }

  
}


