////////////////////////////////////////////////////////////////////////
/// \file  DedxDistribution.h
/// \brief This is a helper class for ParticleIDAlg that provides a tidy structure
///  in which to hold the dE/dx histograms and provides access methods to the data.
///  This encapsulated all of the messy histogram stuff.
///
///
/// \version $Id: DedxDistribution.cxx,v 1.0 2013-10-04 10:00:00 absmith Exp $
/// \author  Alex Smith (smith@physics.umn.edu)
////////////////////////////////////////////////////////////////////////

#ifndef DedxDistribution_h
#define DedxDistribution_h

#include "fhiclcpp/ParameterSet.h"
// ROOT includes
#include "TFile.h"
#include "TH1D.h"

namespace slid {

    /// Number of XY regions into which detector is divided for dE/dx histograms.
    static const int kNumXYRegion = 4;
    /// Number of energy bins into which detector is divided for dE/dx histograms.
    static const int kNumEnergyBin = 11;
    /// Number of longitudinal planes considered for dE/dx histograms.
    static const int kNumLongitudinalPlane = 200;
    /// Number of "transverse" planes considered for dE/dx histograms ("planes" transverse to shower axis).
    static const int kNumTransversePlane = 20;
    /// bin boundaries of energy
    static const double kEnergyBins[] = {0.00, 0.25, 0.75, 1.25, 1.75, 2.25, 2.75, 3.25, 3.75, 4.25, 4.75, 5.25};

    /// \brief An enum used to give allowed particle types a visible name in the code. Note that for electron,
    ///  these types are broken down further by the process in which they reside.
    ///

    enum DedxParticleType {
        kELECTRON,
        kPHOTON,
        kMUON,
        kPI0,
        kPROTON,
        kNEUTRON,
        kPION,
        kELECTRONCCQE,
        kELECTRONCCRES,
        kELECTRONCCDIS,
        kELECTRONCCCOH,
        kDEDXPARTICLETYPESIZE
    };

    class DedxDistribution {
    public:
        explicit DedxDistribution();
        // Need a copy constructor to use in a std::vector.        

        DedxDistribution(const DedxDistribution& other) {
            // no members set in constructor, so nothing to copy/initialize
        }

        ~DedxDistribution();

        /// \brief Load the histogram files
        ///
        void Initialize(const fhicl::ParameterSet& pset, std::string fileNameFclTag);


        ///  \brief Longitudinal plane hit dE/dx.
        TH1D* GetLongitudinalPlaneHitDedxHist(int iXYRegion, int iEnergyBin, int iLongitudinalPlane);

        ///  \brief Transverse plane hit dE/dx ("planes" transverse to shower axis).
        TH1D* GetTransversePlaneCellDedxHist(int iXYRegion, int iEnergyBin, int iTransversePlane);

    private:

        void LoadHistFromFile(std::string filePath);

        /// \brief Load all of the histograms from the file into arrays of TH1D* objects.
        ///  The details of handling the histograms are taken care of within this method
        ///
        bool LoadDedxHistograms(TFile* fDistFile);

        std::string fHistFileName;
        //       TFile* fDistFile;

        ///  Longitudinal plane hit dE/dx
        TH1D *fLongitudinalPlaneHitDedx[kNumXYRegion][kNumEnergyBin][kNumLongitudinalPlane];
        ///  Transverse plane hit dE/dx
        TH1D *fTransversePlaneCellDedx[kNumXYRegion][kNumEnergyBin][kNumTransversePlane];


        /// Need to look up what this in in JM's code and give it a more descriptive name (expected hits?).
        unsigned fHtExpPlane[kNumXYRegion][kNumEnergyBin];
    };
} // end namespace slid

#endif // DedxDistribution_h

////////////////////////////////////////////////////////////////////////



