/* 
 * File:   LIDAlg.cxx
 * Author: smith
 * 
 * Created on November 14, 2013, 12:00 PM
 */



#include "ShowerLID/LIDAlg.h"
#include "Utilities/func/EnvExpand.h"
#include <cmath>

namespace slid {

  LIDAlg::LIDAlg(fhicl::ParameterSet const& pset,
		 NuEEnergyAlg* fNuEEnergyAlgIn,
		 bool withEnergy) :
    fNuEEnergyAlg(fNuEEnergyAlgIn),
    fWithEnergy(withEnergy),
    fMlp(0),
    fReader(0) {
    //
    //  read and set FCL parameters
    //
    fLibPath       = util::EnvExpansion(pset.get< std::string >("LibPath"));
    fUseTMVATraining = pset.get< bool >("UseTMVATraining");

    mf::LogInfo("LIDAlg") << " LIDAlg::LIDAlg() Alg = " << fAlgorithmName << "\n";

    Initialize();
  }

  LIDAlg::LIDAlg(const LIDAlg& orig) 
  {
  }

  LIDAlg::~LIDAlg() 
  {
    if (fReader) delete fReader;
    if (fMlp)    delete fMlp;
  }

  //    void LIDAlg::reconfigure(const fhicl::ParameterSet& pset) {
  //        fLibPath = pset.get< std::string >("LibraryPath");
  //    }

  void LIDAlg::Initialize() 
  {
    art::ServiceHandle<geo::Geometry> geom;
    int detectorID = geom->DetId();
    //
    // TMVA MVA Methods
    //
    std::string weightfile0(fLibPath);
    std::string weightfile2(fLibPath);
    switch (detectorID) {
    case novadaq::cnv::kFARDET:
      if (fUseTMVATraining){
        weightfile0 += "TMVAClassification_" + fMVAMethod + ".weights_FD_elec.xml";
        weightfile2 += "TMVAClassification_" + fMVAMethod + ".weights_FD_elec_withNue.xml";
      }
      else {
        weightfile0 += "weights_shape_fd_elec.txt";
        weightfile2 += "weights_shape_fd_elec_E.txt";
      }
      break;
    case novadaq::cnv::kNDOS:
      if (fUseTMVATraining){
        weightfile0 += "TMVAClassification_" + fMVAMethod + ".weights_NDOS_elec.xml";
        weightfile2 += "TMVAClassification_" + fMVAMethod + ".weights_NDOS_elec_withNue.xml";
      }
      else {
        weightfile0 += "weights_shape_fd_elec.txt";
        weightfile2 += "weights_shape_fd_elec_E.txt";
      }
      break;
    case novadaq::cnv::kNEARDET:
      if (fUseTMVATraining){
        weightfile0 += "TMVAClassification_" + fMVAMethod + ".weights_ND_elec.xml";
        weightfile2 += "TMVAClassification_" + fMVAMethod + ".weights_ND_elec_withNue.xml";
      }
      else {
        weightfile0 += "weights_shape_fd_elec.txt";
        weightfile2 += "weights_shape_fd_elec_E.txt";
      }
      break;
    default:
      throw cet::exception("BAD DETID") << __FILE__ << ":" << __LINE__
					<< " Bad detector id = " << detectorID;
    }


    //
    // mode 0 training variable set (base, without reco nue energy)
    //
    if (fUseTMVATraining){
      if (!fWithEnergy) {
        std::cout << "Defining TMVA::Reader object for mode 0..." << std::endl;
        fReader = new TMVA::Reader("!Color:!Silent");
        std::cout << "Adding variables to Reader for mode 0..." << std::endl;
        fReader->AddVariable("egLLL", &fShowerLID->fEGLLL);
        fReader->AddVariable("egLLT", &fShowerLID->fEGLLT);
        fReader->AddVariable("emuLLL", &fShowerLID->fEMuLLL);
        fReader->AddVariable("emuLLT", &fShowerLID->fEMuLLT);
        fReader->AddVariable("epi0LLL", &fShowerLID->fEPi0LLL);
        fReader->AddVariable("epi0LLT", &fShowerLID->fEPi0LLT);
        fReader->AddVariable("epLLL", &fShowerLID->fEPLLL);
        fReader->AddVariable("epLLT", &fShowerLID->fEPLLT);
        fReader->AddVariable("enLLL", &fShowerLID->fENLLL);
        fReader->AddVariable("enLLT", &fShowerLID->fENLLT);
        fReader->AddVariable("epiLLL", &fShowerLID->fEPiLLL);
        fReader->AddVariable("epiLLT", &fShowerLID->fEPiLLT);
        fReader->AddVariable("gap", &fShowerLID->fGap);
        fReader->AddVariable("pi0mass", &fShowerLID->fPi0mass);
        fReader->AddVariable("vtxgev", &fShowerLID->fVtxE);
        fReader->AddVariable("shE", &fShowerLID->fShwEFrac);
        fReader->AddSpectator("sbtype", &fType);
        //  Book the TMVA Algorithms
        //
        // Set the training weights file.  **NOTE** due to a bug in TMVA model using weights from file
        // does not work.
        //
        //  fWeightsFile = fXMLPath + "TMVAClassification_KNN.weights.xml";
        // Book Method
        fReader->BookMVA((fMVAMethod + TString(" method")), weightfile0);
      }
      else {
        std::cout << "Defining TMVA::Reader object for mode 2..." << std::endl;
        fReader = new TMVA::Reader("!Color:!Silent");
        std::cout << "Adding variables to Reader for mode 2..." << std::endl;
        fReader->AddVariable("egLLL", &(fShowerLID->fEGLLL));
        fReader->AddVariable("egLLT", &(fShowerLID->fEGLLT));
        fReader->AddVariable("emuLLL", &(fShowerLID->fEMuLLL));
        fReader->AddVariable("emuLLT", &(fShowerLID->fEMuLLT));
        fReader->AddVariable("epi0LLL", &(fShowerLID->fEPi0LLL));
        fReader->AddVariable("epi0LLT", &(fShowerLID->fEPi0LLT));
        fReader->AddVariable("epLLL", &(fShowerLID->fEPLLL));
        fReader->AddVariable("epLLT", &(fShowerLID->fEPLLT));
        fReader->AddVariable("enLLL", &(fShowerLID->fENLLL));
        fReader->AddVariable("enLLT", &(fShowerLID->fENLLT));
        fReader->AddVariable("epiLLL", &(fShowerLID->fEPiLLL));
        fReader->AddVariable("epiLLT", &(fShowerLID->fEPiLLT));
        fReader->AddVariable("gap", &(fShowerLID->fGap));
        fReader->AddVariable("pi0mass", &(fShowerLID->fPi0mass));
        fReader->AddVariable("vtxgev", &(fShowerLID->fVtxE));
        fReader->AddVariable("shE", &(fShowerLID->fShwEFrac));
        fReader->AddVariable("nueRecEnergy",&(fShowerLID->fNueEnergy));
        fReader->AddSpectator("sbtype", &fType);
        //  Book the TMVA Algorithms
        //
        // Set the training weights file.  **NOTE** due to a bug in TMVA model using weights from file
        // does not work.
        //
        //  fWeightsFile = fXMLPath + "TMVAClassification_KNN.weights.xml";
        // Book Method
        fReader->BookMVA((fMVAMethod + TString(" method")), weightfile2);
      }
    }
    else{

      //NOTE the varialbe shE really means showerEnergy/SliceEnergy.  I am not changing the name here because this is how it is listed in the training files. 
      //At some point the training should be updated so that this name makes more sense.

      double egLLL; 
      double egLLT;
      double emuLLL;
      double emuLLT;
      double epi0LLL;
      double epi0LLT;
      double epLLL;
      double epLLT;
      double enLLL;
      double enLLT;
      double epiLLL;
      double epiLLT;
      double vtxgev;
      double pi0mass;
      double shE;
      double nuE;
      double gap;
      double length;
      double costheta;
      int type;

      TTree *trainTree = new TTree("trainTree","m_trainTree");
      trainTree->Branch("egLLL", &egLLL, "egLLL/D");
      trainTree->Branch("egLLT", &egLLT, "egLLT/D");
      trainTree->Branch("emuLLL", &emuLLL, "emuLLL/D");
      trainTree->Branch("emuLLT", &emuLLT, "emuLLT/D");
      trainTree->Branch("epi0LLL", &epi0LLL, "epi0LLL/D");
      trainTree->Branch("epi0LLT", &epi0LLT, "epi0LLT/D");
      trainTree->Branch("epLLL", &epLLL, "epLLL/D");
      trainTree->Branch("epLLT", &epLLT, "epLLT/D");
      trainTree->Branch("enLLL", &enLLL, "enLLL/D");
      trainTree->Branch("enLLT", &enLLT, "enLLT/D");
      trainTree->Branch("epiLLL", &epiLLL, "epiLLL/D");
      trainTree->Branch("epiLLT", &epiLLT, "epiLLT/D");
      trainTree->Branch("gap",    &gap, "gap/D");
      trainTree->Branch("pi0mass", &pi0mass, "pi0mass/D");
      trainTree->Branch("vtxgev", &vtxgev, "vtxgev/D");
      trainTree->Branch("shE", &shE, "shE/D");
      trainTree->Branch("nuE", &nuE, "nuE/D");
      trainTree->Branch("costheta", &costheta, "costheta/D");
      trainTree->Branch("length", &length, "length/D");
      trainTree->Branch("type",   &type,   "type/I");

      if (!fWithEnergy) {
        fMlp = new TMultiLayerPerceptron("@egLLL,@egLLT,@emuLLL,@emuLLT,@epi0LLL,@epi0LLT,@epLLL,@epLLT,@enLLL,@enLLT,@epiLLL,@epiLLT,@pi0mass,@shE,@vtxgev,@gap,@costheta:22:12:6:type", trainTree);
        fMlp->LoadWeights(weightfile0.c_str());
      }
      else {
        fMlp = new TMultiLayerPerceptron("@egLLL,@egLLT,@emuLLL,@emuLLT,@epi0LLL,@epi0LLT,@epLLL,@epLLT,@enLLL,@enLLT,@epiLLL,@epiLLT,@pi0mass,@shE,@vtxgev,@gap,@costheta,@nuE:22:12:6:type", trainTree);
        fMlp->LoadWeights(weightfile2.c_str());
      }
      delete trainTree;
    }
    
  }

  double LIDAlg::CalcMVAResult( slid::ShowerLID & showerlid) 
  {
    double ann = -5.0;

    //        //  This is the correct way, except that there seems to be a bug in TMVA that makes in not work
    //          ann = fReader->EvaluateMVA(annInputs, (fMVAMethod + TString(" method")));


    //
    // Set the trained models.  **NOTE** due to a bug in TMVA model using weights from file
    // does not work.  As a workaround, loading the C++ code created by TMVA DOES work, so we use
    // that in order to move ahead.  Loading the XML weights file would be preferable however, 
    // so that recompilation is not required to pick up a new MVA model.
    //
    if (fUseTMVATraining){
      if (!fWithEnergy) {

        fTMVAModuleTMlpANNNoENu = new TMVAClassification_TMlpANN_NoENu();
        ann = fTMVAModuleTMlpANNNoENu->Value(0,
  					     showerlid.EGLLL(), 
  					     showerlid.EGLLT(), 
  					     showerlid.EMuLLL(), 
  					     showerlid.EMuLLT(),
  					     showerlid.EPi0LLL(),
  					     showerlid.EPi0LLT(), 
					     showerlid.EPLLL(), 
					     showerlid.EPLLT(),
					     showerlid.ENLLL(), 
					     showerlid.ENLLT(), 
					     showerlid.EPiLLL(),
					     showerlid.EPiLLT(),
					     showerlid.Gap(),
					     showerlid.Pi0mass(),
					     showerlid.ShowerEFrac(), 
					     showerlid.VertexEnergy() );
        delete fTMVAModuleTMlpANNNoENu;
      } 
	
      else if (fWithEnergy) {
        fTMVAModuleTMlpANNWithENu = new TMVAClassification_TMlpANN_WithENu();
        ann = fTMVAModuleTMlpANNWithENu->Value(0,
					       showerlid.EGLLL(), 
					       showerlid.EGLLT(), 
					       showerlid.EMuLLL(), 
					       showerlid.EMuLLT(),
					       showerlid.EPi0LLL(),
					       showerlid.EPi0LLT(), 
					       showerlid.EPLLL(), 
					       showerlid.EPLLT(),
					       showerlid.ENLLL(), 
					       showerlid.ENLLT(), 
					       showerlid.EPiLLL(),
					       showerlid.EPiLLT(),
					       showerlid.Gap(),
					       showerlid.Pi0mass(),
					       showerlid.ShowerEFrac(), 
					       showerlid.VertexEnergy(),
					       showerlid.NueEnergy());
        delete fTMVAModuleTMlpANNWithENu;
      } 
      else {
        throw cet::exception("MVAALgorithmNotDefined2")
	  << " not found.  See LIDBuilder_module.cc for list of defined algorithms.";
      }
    }
    else{
      if (!fWithEnergy) {
        Double_t params_shape[17];
        for(int i=0;i<17;i++){
          params_shape[i]=0;    
        }
        params_shape[0] = showerlid.EGLLL();
        params_shape[1] = showerlid.EGLLT();
        params_shape[2] = showerlid.EMuLLL();
        params_shape[3] = showerlid.EMuLLT();
        params_shape[4] = showerlid.EPi0LLL(); 
        params_shape[5] = showerlid.EPi0LLT(); 
        params_shape[6] = showerlid.EPLLL(); 
        params_shape[7] = showerlid.EPLLT(); 
        params_shape[8] = showerlid.ENLLL(); 
        params_shape[9] = showerlid.ENLLT(); 
        params_shape[10] = showerlid.EPiLLL(); 
        params_shape[11] = showerlid.EPiLLT(); 
        params_shape[12] = showerlid.Pi0mass(); 
        params_shape[13] = showerlid.ShowerEFrac(); 
        params_shape[14] = showerlid.VertexEnergy(); 
        params_shape[15] = showerlid.Gap(); 
        params_shape[16] = showerlid.CosTheta(); 

        ann = fMlp->Evaluate(0,params_shape);

      }
      else if (fWithEnergy){
        Double_t params_shape[18];
        for(int i=0;i<18;i++){
          params_shape[i]=0;    
        }
        params_shape[0] = showerlid.EGLLL();
        params_shape[1] = showerlid.EGLLT();
        params_shape[2] = showerlid.EMuLLL();
        params_shape[3] = showerlid.EMuLLT();
        params_shape[4] = showerlid.EPi0LLL(); 
        params_shape[5] = showerlid.EPi0LLT(); 
        params_shape[6] = showerlid.EPLLL(); 
        params_shape[7] = showerlid.EPLLT(); 
        params_shape[8] = showerlid.ENLLL(); 
        params_shape[9] = showerlid.ENLLT(); 
        params_shape[10] = showerlid.EPiLLL(); 
        params_shape[11] = showerlid.EPiLLT(); 
        params_shape[12] = showerlid.Pi0mass(); 
        params_shape[13] = showerlid.ShowerEFrac(); 
        params_shape[14] = showerlid.VertexEnergy(); 
        params_shape[15] = showerlid.Gap(); 
        params_shape[16] = showerlid.CosTheta(); 
        params_shape[17] = showerlid.NueEnergy();

        ann = fMlp->Evaluate(0,params_shape);

      }
      else {
        throw cet::exception("MVAALgorithmNotDefined2")
	  << " not found.  See LIDBuilder_module.cc for list of defined algorithms.";
      }
    }

    // protect against nan values
    try {
      if (std::isnan(ann)) throw -1;
    }
    catch (int e){
      if (e == -1){
        ann = -5;
        std::cout<<"!!!!!!!!!!!!!!!!!!!"<<std::endl;
        std::cout<<"  PID returned NaN value from calculation!"<<std::endl;
        std::cout<<"  EGLLL: "<<showerlid.EGLLL()<<" EGLLT: "<<showerlid.EGLLT()<<std::endl;
        std::cout<<"  EMuLLL: "<<showerlid.EMuLLL()<<" EMuLLT: "<<showerlid.EMuLLT()<<std::endl;
        std::cout<<"  ENLLL: "<<showerlid.ENLLL()<<" ENLLT: "<<showerlid.ENLLT()<<std::endl;
        std::cout<<"  EPLLL: "<<showerlid.EPLLL()<<" EPLLT: "<<showerlid.EPLLT()<<std::endl;
        std::cout<<"  EPiLLL: "<<showerlid.EPiLLL()<<" EPiLLT: "<<showerlid.EPiLLT()<<std::endl;
        std::cout<<"  EPiLLL: "<<showerlid.EPi0LLL()<<" EPi0LLT: "<<showerlid.EPi0LLT()<<std::endl;
        std::cout<<"  Pi0Mass: "<<showerlid.Pi0mass()<<" ShowerEFrac: "<<showerlid.ShowerEFrac()<<std::endl;
        std::cout<<"  Gap: "<<showerlid.Gap()<<" Costheta: "<<showerlid.CosTheta()<<std::endl;
        std::cout<<"!!!!!!!!!!!!!!!!!!!"<<std::endl;
      }
    }

    return ann;
  }

  void LIDAlg::FillTrainingTree(const slid::ShowerLID & showerlid) 
  {

  }

  void LIDAlg::DefineTrainingTree() 
  {
    //            //
    //            //  I think this in Jianming's output training TTree.  Should be packaged up
    //            // for readability
    //            //
    //            art::ServiceHandle<art::TFileService> tfs;
    //            //True neutrino information
    //            fEvent = tfs->make<TTree>("fEvent", "fEvent");
    ////            fEvent->Branch("evtTrueNuCCNC", &m_evtTrueNuCCNC);
    ////            fEvent->Branch("evtTrueNuMode", &m_evtTrueNuMode);
    ////            fEvent->Branch("evtTrueNuPdg", &m_evtTrueNuPdg);
    ////            fEvent->Branch("evtTrueNuEnergy", &m_evtTrueNuEnergy);
    ////            fEvent->Branch("evtTrueNuIsFid", &m_evtTrueNuIsFid);
    ////            fEvent->Branch("evtTrueNuP4[4]", m_evtTrueNuP4);
    ////            fEvent->Branch("evtTrueNuVtx[3]", m_evtTrueNuVtx);
    ////            fEvent->Branch("evtTrueEnergy[20]", m_evtTrueEnergy);
    ////            fEvent->Branch("evtTruePdg[20]", m_evtTruePdg);
    ////            fEvent->Branch("evtTrueMother[20]", m_evtTrueMother);
    //    
    //            //Reconstructed information
    //            fEvent->Branch("evtRun", &m_evtRun);
    //            fEvent->Branch("evtSRun", &m_evtSRun);
    //            fEvent->Branch("evtNSlice", &m_evtNSlice);
    //            fEvent->Branch("evtNSh", &m_evtNSh);
    //
    ////            fEvent->Branch("evtEtot", &m_evtEtot);
    ////            fEvent->Branch("evtSigOscW", &m_evtSigOscW);
    ////            fEvent->Branch("evtBkgOscW", &m_evtBkgOscW);
    ////            fEvent->Branch("evtIsMuon", &m_evtIsMuon);
    ////            fEvent->Branch("evtIsMuonSh1", &m_evtIsMuonSh1);
    ////           fEvent->Branch("evtIsInvPhoton", &m_evtIsInvPhoton);
    ////            fEvent->Branch("evtDetEnergy", &m_evtDetEnergy);
    ////            fEvent->Branch("evtNuEnergy", &m_evtNuEnergy);
    ////            fEvent->Branch("evtSh1ExclEnergy", &m_evtSh1ExclEnergy);
    ////           fEvent->Branch("evtSh1Length", &m_evtSh1Length);
    ////            fEvent->Branch("evtSh1Start[3]", m_evtSh1Start);
    ////            fEvent->Branch("evtSh1Stop[3]", m_evtSh1Stop);
    ////            fEvent->Branch("evtSh1Dir[3]", m_evtSh1Dir);
    //    
    ////            fEvent->Branch("evtVtx[3]", m_evtVtx);
    //
    //            fEvent->Branch("sh1EGLLL", &m_sh1EGLLL);
    //            fEvent->Branch("sh1EGLLT", &m_sh1EGLLT);
    //            fEvent->Branch("sh1EMuLLL", &m_sh1EMuLLL);
    //            fEvent->Branch("sh1EMuLLT", &m_sh1EMuLLT);
    //            fEvent->Branch("sh1EPi0LLL", &m_sh1EPi0LLL);
    //            fEvent->Branch("sh1EPi0LLT", &m_sh1EPi0LLT);
    //            fEvent->Branch("sh1EPLLL", &m_sh1EPLLL);
    //            fEvent->Branch("sh1EPLLT", &m_sh1EPLLT);
    //            fEvent->Branch("sh1ENLLL", &m_sh1ENLLL);
    //            fEvent->Branch("sh1ENLLT", &m_sh1ENLLT);
    //            fEvent->Branch("sh1EPiLLL", &m_sh1EPiLLL);
    //            fEvent->Branch("sh1EPiLLT", &m_sh1EPiLLT);
    //            fEvent->Branch("sh1VtxGeV", &m_sh1VtxGeV);
    //            fEvent->Branch("sh1Pi0Mass", &m_sh1Pi0Mass);
    //            fEvent->Branch("sh1ShE", &m_sh1ShE);
    //            fEvent->Branch("sh1Gap", &m_sh1Gap);
    //            fEvent->Branch("sh1Energy", &m_sh1Energy);
    //
    ////            fEvent->Branch("evtSh1ELLL", &m_evtSh1ELLL);
    ////            fEvent->Branch("evtSh1ELLT", &m_evtSh1ELLT);
    ////            fEvent->Branch("evtSh1MuLLL", &m_evtSh1MuLLL);
    ////            fEvent->Branch("evtSh1MuLLT", &m_evtSh1MuLLT);
    ////            fEvent->Branch("evtSh1GLLL", &m_evtSh1GLLL);
    ////            fEvent->Branch("evtSh1InvGLLL", &m_evtSh1InvGLLL);
    ////            fEvent->Branch("evtSh1DistToEdge", &m_evtSh1DistToEdge);
    ////            fEvent->Branch("evtSh1SliceEdge2GeV", &m_evtSh1SliceEdge2GeV);
    ////            fEvent->Branch("evtSh1SliceEdge5GeV", &m_evtSh1SliceEdge5GeV);
    ////            fEvent->Branch("evtSh1SliceEdge10GeV", &m_evtSh1SliceEdge10GeV);
    //
    //            
    //            fEvent->Branch("eIdANNNoNuE", &m_eIdANNNoNuE);
    //            fEvent->Branch("eIdANNWithNuE", &m_eIdANNWithNuE);

  }

  void LIDAlg::InitializeTrainingTree() 
  {
    //            //fEvent
    //            m_evtTrueNuCCNC = 0;
    //            m_evtTrueNuMode = 0;
    //            m_evtTrueNuPdg = 0;
    //            m_evtTrueNuEnergy = 0;
    //            m_evtTrueNuIsFid = 0;
    //            for (int i = 0; i < 4; i++) {
    //                m_evtTrueNuP4[i] = 0;
    //            }
    //            for (int i = 0; i < 3; i++) {
    //                m_evtTrueNuVtx[i] = 0;
    //            }
    //            for (int i = 0; i < 20; i++) {
    //                m_evtTruePdg[i] = 0;
    //            }
    //            for (int i = 0; i < 20; i++) {
    //                m_evtTrueMother[i] = 0;
    //            }
    //            for (int i = 0; i < 20; i++) {
    //                m_evtTrueEnergy[i] = 0;
    //            }
    //            m_evtRun = 0;
    //            m_evtSRun = 0;
    //            m_evtEvent = 0;
    //            m_evtNSlice = 0;
    //            m_evtNSh = 0;
    //            m_evt = 0;
    //            m_evtANN = -5;
    //            m_evtSh1DedxANN = -5;
    //            m_evtSh1DedxANNE = -5;
    //            m_evtEtot = 0;
    //            m_evtSigOscW = 0;
    //            m_evtBkgOscW = 0;
    //            m_evtIsMuon = 0;
    //            m_evtIsInvPhoton = 0;
    //            m_evtDetEnergy = 0;
    //            m_evtTrueNuEnergy = 0;
    //            m_evtNuEnergy = 0;
    //            m_evtSh1Energy = 0;
    //            m_evtSh1ExclEnergy = 0;
    //            m_evtSh1Length = 0;
    //    
    //            for (int i = 0; i < 3; i++) {
    //                m_evtVtx[i] = 0;
    //                m_evtSh1Start[i] = 0;
    //                m_evtSh1Stop[i] = 0;
    //                m_evtSh1Dir[i] = 0;
    //            }
    //    
    //            m_evtSh1EGLLL = 0;
    //            m_evtSh1EGLLT = 0;
    //            m_evtSh1EMuLLL = 0;
    //            m_evtSh1EMuLLT = 0;
    //            m_evtSh1EPi0LLL = 0;
    //            m_evtSh1EPi0LLT = 0;
    //            m_evtSh1EPLLL = 0;
    //            m_evtSh1EPLLT = 0;
    //            m_evtSh1ENLLL = 0;
    //            m_evtSh1ENLLT = 0;
    //            m_evtSh1EPiLLL = 0;
    //            m_evtSh1EPiLLT = 0;
    //            m_evtSh1VtxGeV = 0;
    //            m_evtSh1Pi0Mass = 0;
    //            m_evtSh1ShE = 0;
    //            m_evtSh1Gap = 0;
    //            m_evtSh1ELLL = 0;
    //            m_evtSh1ELLT = 0;
    //            m_evtSh1MuLLL = 0;
    //            m_evtSh1MuLLT = 0;
    //            m_evtSh1GLLL = 0;
    //            m_evtSh1InvGLLL = 0;
    //            m_evtSh1DistToEdge = 0;
    //            m_evtSh1SliceEdge2GeV = 0;
    //            m_evtSh1SliceEdge5GeV = 0;
    //            m_evtSh1SliceEdge10GeV = 0;
  }






} // end namespace slid
