

#include "messagefacility/MessageLogger/MessageLogger.h"
#include "art/Persistency/Common/Ptr.h"
#include "art/Persistency/Common/PtrVector.h"
#include "art/Framework/Principal/Event.h"

#include "RecoBase/Shower.h"
#include "RecoBase/CellHit.h"
#include "RecoBase/RecoHit.h"
#include "RecoBase/Vertex.h"
#include "ShowerLID/NuEEnergyAlg.h"
#include "Calibrator/Calibrator.h"
#include "GeometryObjects/CellGeo.h"
#include "CalibrationUtils/CalibUtil.h"

#include "TMath.h"
#include <map>

namespace slid{

  /*********************************************************************************************/
  ////////////////////////////////////////////////////////////////////////                                          

  NuEEnergyAlg::NuEEnergyAlg(fhicl::ParameterSet const& pset)
  {
    this->reconfigure(pset);
    mf::LogInfo("NuEEnergyAlg") << " NuEEnergyAlg::NuEEnergyAlg()\n";
    fMapIsFilled = false;
    fEvtId = art::EventID( 999,999,999);
  }
  
  /*********************************************************************************************/
  ////////////////////////////////////////////////////////////////////////       
  NuEEnergyAlg::~NuEEnergyAlg()
  {
    
  }
  
  /*********************************************************************************************/
  ////////////////////////////////////////////////////////////////////////                                              

  void NuEEnergyAlg::reconfigure(const fhicl::ParameterSet& pset)
  {
    fCorAbsCell                = pset.get< bool >("CorAbsCell");
    fUseStdCellE               = pset.get< bool >("UseStdCellE");
    fCorDataCell               = pset.get< bool >("CorDataCell");
    fUseUncalibHits            = pset.get< bool >("UseUncalibHits");
    fPEThreshold               = pset.get< float >("PEThreshold");
  }

  /*********************************************************************************************/
  ////////////////////////////////////////////////////////////////////////                                        

  double NuEEnergyAlg::CellEnergy( const rb::Shower *     shower,
				   std::vector< const rb::Shower *>  showercol,
				   const int  &           index,
				   const art::EventID  &  evtid,
				   bool*  isCalibrated)          

  {

    art::ServiceHandle<calib::Calibrator> cal;
    const rb::CellHit *  cHit = &(*(shower->Cell( index )));
    if( fEvtId != evtid )
      ClearVars( evtid );
    if( isCalibrated )
      *isCalibrated = true;
    geo::View_t view = cHit->View();
    double w = shower->W(cHit);

    //get distance to readout electronics
    double xyzd[4];
    calib::GetXYZD(cHit->OfflineChan(), w, xyzd);
    double readoutD = xyzd[3];

    int detid  = fGeom->DetId();

    rb::Prong prong  = *(shower);

    rb::RecoHit rhit = shower->RecoHit( index );
    double cellgev = 0;
 
    if(isCalibrated)
      *isCalibrated = rhit.IsCalibrated();
    
    if(rhit.IsCalibrated()){
      cellgev = rhit.GeV();
    }
    else if(fUseUncalibHits){

	double adc = double(shower->Cell( index )->ADC());
	double Att = 1;
	Att = CellADCToGeV( readoutD, view);
	cellgev =  adc/Att;
      }
    else 
      return cellgev;
  
      
    double weight = shower->Weight( index );
    double norm = 1;
    if( cellgev !=0 )
      norm = cHit->PE()/cellgev; 

    GetHitShowerMap( showercol );
    std::map< int, int > hitmap;
    if( weight != 1){
      for( auto & cmap : fHitShowerMap){
        if(cmap.first == shower->Cell( index )->Channel())
          hitmap = cmap.second;
      }// end loop over common cells

      norm = 0;

      for( auto & idxmap : hitmap ){
        double w1  = showercol[  idxmap.first ]->W(  showercol[ idxmap.first ]->Cell(idxmap.second).get() );

	rb::RecoHit rhit1 = cal->MakeRecoHit(*(showercol[ idxmap.first ]->Cell(idxmap.second)), w1); 
        double cellgev1 = 0;
	
        if(rhit1.IsCalibrated()){
          cellgev1 = rhit1.GeV();
        }
	else if( fUseUncalibHits ){
	  art::Ptr<rb::CellHit> thishit = showercol[ idxmap.first ]->Cell(idxmap.second);
          double adc1 = thishit->ADC();
	  double Att1 = 1;
	  double xyzd[4];
	  calib::GetXYZD(thishit->OfflineChan(), showercol[ idxmap.first ]->W(thishit.get()), xyzd);
	  double readoutD = xyzd[3];
          Att1 = CellADCToGeV( readoutD, view);
          cellgev1 =  adc1/Att1;
        }

        double ai = cellgev1/cHit->PE(); 
        norm += showercol[ idxmap.first ]->Weight(  idxmap.second )/ai;
      } 
    }

    if( fUseStdCellE ){
      if( rhit.IsCalibrated() )
	return rhit.GeV()*weight/norm;
      else{
	if(isCalibrated)
	*isCalibrated = false;
        return 0;
      }
    }
    return NonStdCellEnergy( cHit, detid, readoutD, w, norm, weight);

  }// end of CellEnergy


  /*********************************************************************************************/
  ////////////////////////////////////////////////////////////////////////                                        

  double NuEEnergyAlg::NonStdCellEnergy( const rb::CellHit *  cHit,
					 const int    & detid,
					 const double & readoutD,
					 const double & w,
					 const double & norm,
					 const double & weight)
  {
    art::ServiceHandle<calib::Calibrator> cal;
    geo::View_t view = cHit->View();
    double gev = 0; 

    if( detid != novadaq::cnv::kFARDET || !fCorAbsCell){
      gev = cHit->PE() * weight/norm;
    }

    else{
      // for FD, Attenuation effect calibrated for development                                                         
      // (Product S12.02.14), should not be changed to keep the                                                       
      // energy scale identical to the one used in ANN training!                                                       

      int adc = cHit->ADC(0);

      double Att = 1;
      if(view == geo::kX)
        Att = CellADCToGeV( readoutD, view);
      else if(view == geo::kY)
        Att = CellADCToGeV( readoutD, view);
      gev = adc*weight/Att;
    }
    
    gev = gev;
    return gev;
    
  }// end of NonStdCellEnergy


  /*********************************************************************************************/
  ////////////////////////////////////////////////////////////////////////                                        

  double NuEEnergyAlg::CellEDataMC( double cellEnergy, double w,
				    geo::View_t view)

  {
    double ecor = 1;    //S12-11-16                                                                                      
    int detid = fGeom->DetId();

    if( cellEnergy > 0 && w > 0 ){

      if( detid == novadaq::cnv::kFARDET ){

        if( view == geo::kX )
          ecor = 0.707772+0.000573526*w;
        
	else if( view == geo::kY )
          ecor = 0.707772+0.000573526*w;
        
      }
     
      else{
        
	if( view == geo::kX )
          ecor = 0.707772+0.000573526*w;
        
	else if( view == geo::kY )
          ecor = 0.707772+0.000573526*w;
        
      }
    }

    if(ecor<0.0001)ecor = 1.0;
    return ecor;

  }// end of CellEDataMC

  /*********************************************************************************************/
  ////////////////////////////////////////////////////////////////////////                                        


  double NuEEnergyAlg::CellADCToGeV( double readoutD, geo::View_t view)
  {
    double Att = 4.08050e+03;
    art::ServiceHandle<geo::Geometry> fGeom;
    int detid  = fGeom->DetId();

    if( detid == novadaq::cnv::kFARDET ){

      if( view==geo::kX )
        Att = 4.08574e+03+3.49097e+04*TMath::Exp(-2.67297e-03*readoutD)
          -1.62442e+04*TMath::Exp(-3.14363e-02*readoutD-(-3.93281e-05)
				  *readoutD*readoutD-1.49793e-08*readoutD*readoutD*readoutD);

      else if( view == geo::kY ) 
        Att = 4.09347e+03+3.50384e+04*TMath::Exp(-2.68025e-03*readoutD)
          -1.74280e+04*TMath::Exp(-3.08694e-02*readoutD-(-3.86870e-05)
				  *readoutD*readoutD-1.48737e-08*readoutD*readoutD*readoutD);
    }

    else{
   
      if( view == geo::kX )
      
	Att = 4.08574e+03+3.49097e+04*TMath::Exp(-2.67297e-03*readoutD)
          -1.62442e+04*TMath::Exp(-3.14363e-02*readoutD-(-3.93281e-05)
				  *readoutD*readoutD-1.49793e-08*readoutD*readoutD*readoutD);
      
      else if( view == geo::kY )
        Att = 4.09347e+03+3.50384e+04*TMath::Exp(-2.68025e-03*readoutD)
          -1.74280e+04*TMath::Exp(-3.08694e-02*readoutD-(-3.86870e-05)
				  *readoutD*readoutD-1.48737e-08*readoutD*readoutD*readoutD);
      
    }  
    return Att;

  }// end of CellADCToGeV

  /*********************************************************************************************/
  ////////////////////////////////////////////////////////////////////////                                        

  void NuEEnergyAlg::GetHitShowerMap( std::vector< const rb::Shower *>  showercol)
  {
    if( MapFilled( showercol) )
      return;

    //clear hit map in case it was used previously before filling
    fHitShowerMap.clear();

    // loop over all the showers and find the cellhits common
    // to 2 or more showers.
    int nSh = showercol.size();

    for( int iSh = 0; iSh < nSh; ++iSh){

      art::PtrVector< rb::CellHit > ihits = showercol[iSh]->AllCells();
      int nCells = ihits.size();
      // loop over cells in ith shower                                                                                            
      for( int iCell = 0; iCell < nCells; ++iCell ){

        // In the following loop, we will find which other                                                                       
	// showercol the ihits[iCell] belongs to.                                                                                 
        for( int jSh = iSh+1; jSh < nSh; ++jSh){
	  art::PtrVector< rb::CellHit > jhits = showercol[jSh]->AllCells();
          int mCells = jhits.size();
          // loop over cells in jth shower                                                                                        
          for(int  jCell = 0 ; jCell < mCells; ++jCell){

            if( jhits[ jCell] == ihits[ iCell] ){
              fHitShowerMap[ ihits[ iCell]->Channel() ][ iSh ] = iCell;
              fHitShowerMap[ ihits[ iCell]->Channel() ][ jSh ] = jCell;
            }
          }// end loop over jshower cells                                                                                         
        }// end j loop over showercol                                                                                            
      }// end loop over ishower cells                                                                                             
    }// end i loop over showercol                                           
    fShowerCol.clear();
    fShowerCol = showercol;
    fMapIsFilled = true;
    return;
  }

  /*********************************************************************************************/
  ////////////////////////////////////////////////////////////////////////                                        

  bool NuEEnergyAlg::MapFilled( std::vector< const rb::Shower *>  showercol)
  {
    
    if( !fMapIsFilled )
      return fMapIsFilled;
    // Check if this shower collection is the same as the one
    // we already have a map for (may be).

    if( fShowerCol.size() != showercol.size() ){
      fHitShowerMap.clear();
      fShowerCol.clear();
      fMapIsFilled = false;
      return fMapIsFilled;
    }
    
    for( int ish = 0, nsh = showercol.size(); ish < nsh ; ++ish){
      if( !(fShowerCol[ish]->NCell() == showercol[ish]->NCell() &&
	    fShowerCol[ish]->Dir()   == showercol[ish]->Dir()   &&
	    fShowerCol[ish]->Start() == showercol[ish]->Start() )){
	fHitShowerMap.clear();
	fShowerCol.clear();
	fMapIsFilled = false;
	return fMapIsFilled;
      }
    }

    return fMapIsFilled;
  }
  
  /*********************************************************************************************/
  ////////////////////////////////////////////////////////////////////////                                        
  
  double NuEEnergyAlg::ShowerDepEnergy( const rb::Shower *  shower,
					std::vector< const rb::Shower *>  showercol,
					const art::Event& evt)
  {
    double showerDepE = 0;
    int  ncells  = shower->NCell();
    art::EventID evtid   = evt.id();

    if( fEvtId != evtid )
      ClearVars( evtid );
    
    for( int i = 0; i < ncells; ++i){

      double cellE = CellEnergy( shower, showercol, i, evtid);
      showerDepE += cellE; 
    }
    
    return showerDepE;
    
  }// end of ShowerDepEnergy


 /*********************************************************************************************/
  ////////////////////////////////////////////////////////////////////////                              

  double NuEEnergyAlg::ECorrMC( double gev ){
    //EM shower energy correction                                                 
    double ecor = 5.68142e-01;    // with pigtail

    if( fCorAbsCell == false )
      ecor = 5.62343e-01;  //no mc correction    

    return ecor;
  }// end of CorrMC
  

  /*********************************************************************************************/
  ////////////////////////////////////////////////////////////////////////                    

  double NuEEnergyAlg::ShowerEnergy( const rb::Shower * shower,
				     std::vector< const rb::Shower *>  showercol,
				     const art::Event& evt)
  {
    double showerE = 0;
    double showerDepE = ShowerDepEnergy( shower, showercol, evt);
    double ecor = ECorrMC( showerDepE );
    //TODO: Average position is calculated in the way of RecoJMShower.  It is not clear whether it would be better
    //to calculate the average position as commented out below.  Results are slightly different, should be looked at.
    //For now doing it this way for consistency.
    double avgx = (shower->Start().X() + shower->Stop().X())/2.0;
    double avgy = (shower->Start().Y() + shower->Stop().Y())/2.0;
    //NOTE: Turn off energy weighting when calculating mean position.  On cosmics the low PE hits can skew the
    //average since attenuation is not taken into account.
    //double avgx = shower->MeanX(rb::kByClusterWeightOnly);
    //double avgy = shower->MeanY(rb::kByClusterWeightOnly);

    double poscor = 1.+6.28813e-05*avgx + 6.15165e-05*avgy; //with pigtail
    
    if( !fCorAbsCell )
      poscor = 1.0+1.04866e-04*avgx + 6.40668e-05*avgy;
    
    showerE = showerDepE/(ecor*poscor);

    return showerE;
    
  }// end of ShowerEnergy


  /*********************************************************************************************/
  ////////////////////////////////////////////////////////////////////////                                        

  double NuEEnergyAlg::HadronicDepEnergy( std::vector< const rb::Shower *>   showercol,
					  const rb::Cluster *  slice,
					  const int & elecShowerIndex,
					  const art::Event& evt)
  {
    art::ServiceHandle<calib::Calibrator> cal;
    if( fEvtId != evt.id() )
      ClearVars( evt.id() );
    
    double hadDepE = 0;
    int nShowers   = showercol.size();
    rb::Cluster excludedHits = *(slice);

    double meanx = slice->MeanX();
    double meany = slice->MeanY();

    for( int i = 0; i < nShowers; ++i ){
      const rb::Shower * shower  = showercol[i];
      int nhits = shower->NCell();

      for( int h = 0; h < nhits; ++h){
	excludedHits.RemoveHit( shower->Cell(h) );
      }
      if ( i == elecShowerIndex )
	continue;
      else{
	hadDepE += ShowerDepEnergy( showercol[i], showercol, evt );
      }
    }// end loop over showers

    int nExHits = excludedHits.NCell();
    for( int i = 0; i < nExHits; ++i){
      if( excludedHits.Cell(i)->PE() < fPEThreshold )
    	continue;
      rb::RecoHit rHit;
      if( excludedHits.Cell(i)->View() == geo::kX )
	rHit = rb::RecoHit(cal->MakeRecoHit(*(excludedHits.Cell(i)), meany));
      else
	rHit = rb::RecoHit(cal->MakeRecoHit(*(excludedHits.Cell(i)), meanx));

      if( rHit.IsCalibrated() ){
    	hadDepE += rHit.GeV();
      }
      else if( fUseUncalibHits ){
    	art::Ptr< rb::CellHit > cHit = excludedHits.Cell(i);
    	double adc = double( cHit->ADC(0) );
    	double Att = 1;
    	geo::View_t view = fGeom->Plane( cHit->Plane() )->View();

	double xyzd[4];
	if(view == geo::kX)
	  calib::GetXYZD(cHit->OfflineChan(), slice->MeanY(), xyzd); 
	else
	  calib::GetXYZD(cHit->OfflineChan(), slice->MeanX(), xyzd); 

	Att = CellADCToGeV( xyzd[3], view);
    	hadDepE += adc/Att;
      }

    }// end loop over orphan hits.
    return hadDepE;

  }// end of HadronicDepEnergy
  

  /*********************************************************************************************/
  ////////////////////////////////////////////////////////////////////////                                        

  double NuEEnergyAlg::HadronicEnergy( std::vector< const rb::Shower *>   showercol,
				       const rb::Cluster *  slice,
				       const int & elecShowerIndex,
				       const art::Event& evt)
  {
    double hadDepE = HadronicDepEnergy( showercol, slice, elecShowerIndex, evt);
    double hadE    = hadDepE/(4.00599e-01 
			      - 5.08688e+00 * TMath::Exp(-1.07930e+01*hadDepE-(-1.17127e+01)*pow(hadDepE,2)-1.01452e+01*pow(hadDepE,3))                     
			      +4.71995e+00  * TMath::Exp(-1.02536e+01*hadDepE-(-9.81034e+00)*pow(hadDepE,2)-4.22648e+00*pow(hadDepE,3)));
    return hadE;
  }// end of HadronicEnergy


  /*********************************************************************************************/
  ////////////////////////////////////////////////////////////////////////                                        

  double NuEEnergyAlg::NuEEnergy( std::vector< const rb::Shower *>  showercol,
				  const rb::Cluster * slice,
				  const int & elecShowerIndex,
				  const art::Event& evt)
  {
    double elecE = ShowerEnergy( showercol[elecShowerIndex], showercol, evt );
    double hadE  = HadronicEnergy( showercol, slice, elecShowerIndex, evt);
    double nuE   = ( elecE + hadE );
    
    return nuE;
    
  }// end of NuEEnergy

  /*********************************************************************************************/
  ////////////////////////////////////////////////////////////////////////                       
  
  double NuEEnergyAlg::VertexEnergy(const rb::Shower *  shower,
				    std::vector< const rb::Shower *>  showercol,
				    const rb::Vertex *  vertex,
				    const rb::Cluster * slice,
				    const art::Event&   evt)
  {
    art::ServiceHandle<geo::Geometry> geom;
    art::ServiceHandle<calib::Calibrator> cal;

    if( fEvtId != evt.id() )
      ClearVars( evt.id() );

    TVector3 trkstart = shower->Start();
    TVector3 trkstop  = shower->Stop();


    int startPlane, stopPlane, vtxPlane=0;

    geo::CellUniqueId cid;
 
    for(int trial = 0; trial < 4; ++trial){
      // Figure out what cell the vertex was in. Step 3cm in z in case of
      // failure. If that didn't work, it must be in plastic between cells,
      // move laterally in x or y.
      double offsetX = 0, offsetY = 0;
      if(trial == 1) offsetX = -2;
      if(trial == 2) offsetY = -2;
      if(trial == 3) offsetX = offsetY = -2;
      
      cid = geom->CellId(vertex->GetX()+offsetX, vertex->GetY()+offsetY, vertex->GetZ(),
                                                0, 0, 1, -1);
      if(cid)
	break;
    }
    
    if( cid )
      vtxPlane = geom->getPlaneID( cid );
    else 
      vtxPlane = slice->MinPlane();


    startPlane = shower->MinPlane();
    stopPlane = shower->MaxPlane();


    if( abs( stopPlane - vtxPlane) < abs( startPlane - vtxPlane)){
      int plane  = stopPlane;
      stopPlane  = startPlane;
      startPlane = plane;
    }
    int maxPlane = std::max(stopPlane, startPlane);
    int minPlane = std::min(stopPlane, startPlane);
    
    bool backtag = false;
    if(stopPlane < startPlane)
      backtag = true;

    rb::Cluster lesserSlice = slice->Exclude( shower );
    art::PtrVector<rb::CellHit> sliceHits = lesserSlice.AllCells();    

    int nsh   = sliceHits.size();
    int nShowers = showercol.size();

    for(int ic = 0; ic < nsh; ++ic){
      bool matched = false;
      for(int is = 0; is < nShowers; ++is){
	int nhits = showercol[is]->NCell();
	for( int ish = 0; ish < nhits; ++ish ){
	  if( *(sliceHits[ic]) == *(showercol[is]->Cell(ish)) ){
	    matched = true;
	    break;
	  }
	}
      }
      // hits that do not belong to any showers are quite
      // possibly noise hits and subjected to a PE cut.
      if( !matched && sliceHits[ic]->PE() <= fPEThreshold )
	lesserSlice.RemoveHit( sliceHits[ic]);
    }
    
    sliceHits = lesserSlice.AllCells();    
    nsh   = sliceHits.size();

    double vtxE = 0;

    for( int ish = 0; ish < nsh; ++ish){

      int dPlane = sliceHits[ish]->Plane() - startPlane;
      if( backtag )
	dPlane = -dPlane;
      int plane = sliceHits[ish]->Plane();
      int cell  = sliceHits[ish]->Cell();
      
      if( ((dPlane >= -8 && dPlane < 0) || 
	   (dPlane < 8 && plane <= maxPlane && plane >= minPlane) ) ){
      
	geo::View_t view = sliceHits[ish]->View();
	int detid = geom->DetId();

	double cellXYZ[3]={0,0,0};
	geom->Plane(plane)->Cell(cell)->GetCenter(cellXYZ);
	TVector3 trkdir = shower->Dir();

	double w;
	if( view == geo::kY)
	  w = trkstart.x()+(trkdir.x()/trkdir.z())*(cellXYZ[2]-trkstart.z());
	else
	  w = trkstart.y()+(trkdir.y()/trkdir.z())*(cellXYZ[2]-trkstart.z());

	if( fUseStdCellE ){
	  rb::RecoHit rhit(cal->MakeRecoHit(*(sliceHits[ish]), w));
	  if( rhit.IsCalibrated() )
	    vtxE += rhit.GeV();
	}
	else{
	  double readoutD;
	  double norm;

	  readoutD = geom->DistToElectronics( w, *geom->Plane(plane)->Cell( cell ));
	  norm = sliceHits[ish]->PE()/cal->GetPECorr(*(sliceHits[ish]), w ) ;

          double cellgev = 0;
	  rb::RecoHit rhit = cal->MakeRecoHit( *(sliceHits[ish]), w);
          
	  if(rhit.IsCalibrated()){
            cellgev = rhit.GeV();
          }
	  else if( fUseUncalibHits ){
            double adc = double(sliceHits[ish]->ADC());
            double Att = 1;
            Att = CellADCToGeV( readoutD, view);
            cellgev =  adc/Att;
          }
	  
	  norm =  sliceHits[ish]->PE()/cellgev;
	  
	  double weight = 1;
	  double e = NonStdCellEnergy( sliceHits[ish].get(), detid, readoutD, w, norm, weight);
	  vtxE += e;
	}
      }      
    }// end loop over slice hits

    return vtxE;
  }//End of VertexEnergy

  /*********************************************************************************************/
  ////////////////////////////////////////////////////////////////////////                       


  double NuEEnergyAlg::ShowerSumEnergy(std::vector< const rb::Shower *>  showercol,
				       const art::Event& evt)
  {
    double sumE = 0;

    for( auto const iShower: showercol){
      sumE += ShowerEnergy( iShower, showercol, evt);
    }// end loop over showers
    return sumE;

  }// End of ShowerSumEnergy
  
  /*********************************************************************************************/
  ////////////////////////////////////////////////////////////////////////                       


  void NuEEnergyAlg::ClearVars( const art::EventID& evtid ){
    
    fShowerCol.clear();
    fHitShowerMap.clear();
    fMapIsFilled = false;
    fEvtId = evtid;
    
  }// End of ClearVars
  /*********************************************************************************************/
  ////////////////////////////////////////////////////////////////////////                       


}// end of jmshower namespace
