
////////////////////////////////////////////////////////////////////////
/// \file  ParticleIDAlg.cxx
/// \brief Calculate dE/dx and log likelihoods for different particle hypotheses.
///     This information will be of use for identification of any particle that
///     uses rb::Showers as a starting point.
///
/// \todo Remove hardcoded physical constants.
///
///
/// \version $Id: ParticleIDAlg.cxx,v 1.11 2012-10-26 22:19:03 bckhouse Exp $
/// \author  Alex Smith (smith@physics.umn.edu)
////////////////////////////////////////////////////////////////////////


#include <art/Framework/Principal/Event.h>
#include "TLorentzVector.h"
#include "ShowerLID/ParticleIDAlg.h"
#include "Utilities/func/EnvExpand.h"
#include "GeometryObjects/Geo.h"

#include <iterator>
#include <algorithm>
#include <map>
#include <cmath>

namespace slid {

  //*********************************************************************************************
  ParticleIDAlg::ParticleIDAlg(fhicl::ParameterSet const& pset, 
			       novadaq::cnv::DetId detId,
                               NuEEnergyAlg* NuEEnergyAlgIn,
                               bool loadLibs) 
    : fClosestPi0Mass(-9999),
      fDedxDistributions(slid::kDEDXPARTICLETYPESIZE),
      fShower(0),
      fNuEEnergyAlg(NuEEnergyAlgIn),
      fAsym(1.0),
      fMipRangeHigh(0.005),
      fMipRangeLow(0.0005),
      fDetId(detId)
  {
    if (loadLibs){
      LoadDedxHistogramFiles(pset, detId);
    }
  }

  //*********************************************************************************************
  ParticleIDAlg::~ParticleIDAlg() 
  {
  }

  //*********************************************************************************************
  void ParticleIDAlg::reconfigure(const fhicl::ParameterSet & pset, 
				  novadaq::cnv::DetId detId) 
  {
  }
  
  //*********************************************************************************************
  TFile * LoadHistFromFile(TString filePath) 
  {
    std::cout << "LID Dedx File " << filePath << " is loaded." << std::endl;
    return new TFile(filePath);
  }

  /// \brief  Calculate transverse dE/dx log-likelihood for specific particle hypothesis
  //*********************************************************************************************
  double ParticleIDAlg::DedxTransLL(const slid::DedxParticleType partHyp, 
				    const rb::Shower* vShower,
				    const std::vector< const rb::Shower* > showercol,				    
				    const art::Event& evt) 
  {
    // Lazy setting of shower
    SetShower(vShower, showercol, evt);
    std::map< int, float>::iterator llIter = fPartTransLL.find(partHyp);
    if (llIter == fPartTransLL.end()) {
      double sumLogProb = 0;
      int transPlane = 0;
      for (int iTransPlane = 0; iTransPlane < kNumTransversePlane; iTransPlane++) {
        double dedx = CalcPlaneTransverseDedx(iTransPlane);
        if (dedx < -0.5) continue;
        double prob = 0;
        prob = CalcInterCellTransDedxProb(partHyp, iTransPlane);
        if (prob < 0.0001) prob = 0.0001;
        double logProb = log(prob);
        sumLogProb += logProb;
        transPlane++;
      }
      if (transPlane >0 ) sumLogProb = sumLogProb / (double)transPlane;
      if (sumLogProb > 999)sumLogProb = 999;
      if (sumLogProb<-999)sumLogProb = -999;

      fPartTransLL[partHyp] = sumLogProb;
      return sumLogProb;
    }
    return llIter->second;
  };

  /// \brief  Calculate longitudinal dE/dx log-likelihood for specific particle hypothesis
  //*********************************************************************************************
  double ParticleIDAlg::DedxLongLL(const slid::DedxParticleType partHyp, 
				   const rb::Shower* vShower,
				   const std::vector< const rb::Shower* > showercol,
				   const art::Event& evt) 
  {
    // Lazy setting of shower
    SetShower(vShower, showercol, evt);

    std::map< int, float>::iterator llIter = fPartLongLL.find(partHyp);
    if (llIter == fPartLongLL.end()) {
      double sumLogLikelihood = 0;
      unsigned int totplane = fPlaneHits.size();
      //unsigned int nonzeroplane = 0;
      unsigned int planecount = 0;

      if (vShower->Dir()[2] >= 0.0){ //forward going shower
        for (std::map<int, std::map<int, int> >::iterator itr = fPlaneHits.begin(); itr != fPlaneHits.end(); ++itr){
          double prob = 0;
          prob = CalcInterPlaneDedxProb(partHyp, planecount, itr->first);
          if (prob < 0.0001)prob = 0.0001;
          double lgprob = log(prob);
          //nonzeroplane++;
          sumLogLikelihood += lgprob;
          planecount++;
        }
      }
      else{ //backward going shower
        for (std::map<int, std::map<int, int> >::reverse_iterator itr = fPlaneHits.rbegin(); itr != fPlaneHits.rend(); ++itr){
          double prob = 0;
          prob = CalcInterPlaneDedxProb(partHyp, planecount, itr->first);
        
          if (prob < 0.0001)prob = 0.0001;
          double lgprob = log(prob);
          //nonzeroplane++;
          sumLogLikelihood += lgprob;
          planecount++;
        }
      }
      if (totplane > 0) sumLogLikelihood = sumLogLikelihood / (double) totplane;
      if (sumLogLikelihood > 999) sumLogLikelihood = 999;
      if (sumLogLikelihood<-999) sumLogLikelihood = -999;
      fPartLongLL[partHyp] = sumLogLikelihood;
      return sumLogLikelihood;
    }
    return llIter->second;
  }
  
  /// \brief  Calculate longitudinal dE/dx log-likelihood for specific particle hypothesis
  //*********************************************************************************************
  double ParticleIDAlg::DedxInverseLongLL(const slid::DedxParticleType partHyp, 
					  const rb::Shower* vShower,
					  const std::vector< const rb::Shower* > showercol,
					  const art::Event& evt) 
  {
    rb::Shower newShower = *vShower;
    newShower.SetStart( vShower->Stop());
    newShower.SetDir( -(vShower->Dir()) );
    return DedxLongLL(partHyp, &(newShower), showercol, evt);
  }

  /**
   * \brief Set rb::Shower object to be analyzed.  Calculate its total energy, which will be needed 
   * for calculations.
   * 
   * @param vShower
   * @param evt
   */
  //*********************************************************************************************
  void ParticleIDAlg::SetShower(const rb::Shower* vShower, 
				const std::vector< const rb::Shower* > showercol,
				const art::Event& evt)
  {
    //
    //  Only recalculate shower quantities if a different event or shower is presented or if 
    //  fShower is not already set
    //

    if ( fShower  != NULL && 
	 evt.id() == eventID &&
	 fShowerCol.size() == showercol.size() &&
	 std::equal( fShowerCol.begin(), fShowerCol.end(), showercol.begin()) &&
	 fShower->ID() == vShower->ID()) 
      return;

    //  Save to compare next one
    fShower    = vShower;
    fShowerCol = showercol; 
    eventID    = evt.id();

    //clear map of plane index to global plane
    fMapIdxToGlobalPlane.clear();

    //  Clear cache of longitudinal dedx values by longitudinal and tranverse plane
    fDedxLongByPlane.clear();
    fDedxTransByPlane.clear();

    fPartLongLL.clear();
    fPartTransLL.clear();

    fTransEnergy.clear();
    fTransPathLength.clear();
    fPi0Index = -5;

    fIxyz.SetXYZ(0.0,0.0,0.0);
    fAsym = 1.0;

    //  Calculate and save the total energy of the shower in a member field
    fShowerEnergyGev = fNuEEnergyAlg->ShowerEnergy(fShower, showercol, evt);
    if ((fShowerEnergyGev<1e-10)|| std::isnan(fShowerEnergyGev)) fShowerEnergyGev = 0.0;


    // Fill a  map where the first element is
    // plane number and the second is vector of global
    // cell indices in a shower of cells that belong to
    // that plane.
    fPlaneHits = CalcPlaneHits( vShower);

    // Fill a map where the first index is the
    // number of planes away from the track core, and the
    // second is a vector of global cell indices in a
    // shower that are that many planes from the track
    // core. It collapses the transverse profile of the
    // shower to a single plane.
    fPlaneCentroid = CalcPlaneCentroid( vShower );

    //
    // Determine detector XY region index
    //
    CalcDetectorXYRegionIndex();
    //
    //  Determine energy bins
    //
    CalculateEnergyBinIndex();

    //calculate shower asymmetry and ineria
    CalcAsymIneria();

  }

  //*********************************************************************************************
  std::map< int, std::map< int, int> > ParticleIDAlg::CalcPlaneHits(const rb::Shower* vShower)
  {
    std::map< int, std::map< int, int> > mapP;

    std::map<int, int> dummymap;

    int startPlane = vShower->MinPlane();
    int stopPlane  = vShower->MaxPlane();
    for (int i=startPlane; i<=stopPlane; ++i){
      mapP[i] = dummymap;
    }
    
    int ncells = vShower->NCell();
    for (int i = 0; i < ncells; ++i) {

      mapP[ vShower->Cell(i)->Plane() ][vShower->Cell(i)->Cell()] = i;
    }// end loop over cells

    //now map the plane index with the global plane number
    unsigned int planecount = 0;
    if (vShower->Dir()[2] >= 0.0){ //forward going shower
      for (std::map<int, std::map<int, int> >::iterator itr = mapP.begin(); itr != mapP.end(); ++itr){
        fMapIdxToGlobalPlane[planecount] = itr->first;
        planecount++;
      }
    }
    else{ //backward going shower
      for (std::map<int, std::map<int, int> >::reverse_iterator itr = mapP.rbegin(); itr != mapP.rend(); ++itr){
        fMapIdxToGlobalPlane[planecount] = itr->first;
        planecount++;
      }
    }

    return mapP;
  }// end of CalcPlaneHits

  //*********************************************************************************************
  std::map< int, double > ParticleIDAlg::CalcPlaneCentroid(const rb::Shower* vShower) 
  {

    std::map< int, double > mapT;

    //loop plane by plane to calculate weighted cell center and then transverse cells
    for(std::map<int, std::map<int, int> >::iterator itr=fPlaneHits.begin(); itr!=fPlaneHits.end(); ++itr){
      std::map<int, int> planecells = itr->second;
      //loop over hits in plane to determine energy weighted center
      double sumPos = 0.0;
      double planeGev = 0.0;
      for(std::map<int, int>::iterator itrC = planecells.begin(); itrC!=planecells.end(); ++itrC){
        double cellPos = itrC->first + 0.5;
        double cellE   = fNuEEnergyAlg->CellEnergy(vShower, fShowerCol, itrC->second, eventID );
        if ((cellE == -5) || std::isnan(cellE)) cellE = 0.0;
        sumPos += cellPos*cellE;
        planeGev += cellE;
      }
      double avgPos = -999.0; //non physical cell position as the default
      if (planeGev > 0.00001) avgPos = sumPos/planeGev;
      mapT[itr->first] = avgPos;
    }

    return mapT;
  }

  //*********************************************************************************************
  bool ParticleIDAlg::LoadDedxHistogramFiles(const fhicl::ParameterSet & pset, 
					     novadaq::cnv::DetId detectorID) 
  {

    std::string libPath = util::EnvExpansion(pset.get< std::string >("LibraryPath"));
    std::string thistname = pset.get< std::string >("HistFDDedxElectron");
    std::cout << "Loading LID library file from " << libPath  << std::endl;
    if (detectorID == novadaq::cnv::DetId::kFARDET) {

      /// FD dE/dx input histogram files
      std::cout<<"---- dE/dx histogram file for electron\n";
      fDedxDistributions[slid::kELECTRON].Initialize(pset, "HistFDDedxElectron");
      std::cout<<"---- dE/dx histogram file for photon\n";
      fDedxDistributions[slid::kPHOTON].Initialize(pset, "HistFDDedxPhoton");
      std::cout<<"---- dE/dx histogram file for muon\n";
      fDedxDistributions[slid::kMUON].Initialize(pset, "HistFDDedxMuon");
      std::cout<<"---- dE/dx histogram file for pi0\n";
      fDedxDistributions[slid::kPI0].Initialize(pset, "HistFDDedxPi0");
      std::cout<<"---- dE/dx histogram file for proton\n";
      fDedxDistributions[slid::kPROTON].Initialize(pset, "HistFDDedxProton");
      std::cout<<"---- dE/dx histogram file for neutron\n";
      fDedxDistributions[slid::kNEUTRON].Initialize(pset, "HistFDDedxNeutron");
      std::cout<<"---- dE/dx histogram file for charged pion\n";
      fDedxDistributions[slid::kPION].Initialize(pset, "HistFDDedxPion");
      std::cout<<"---- dE/dx histogram file for nue CCQE\n";
      fDedxDistributions[slid::kELECTRONCCQE].Initialize(pset, "HistFDDedxElectronCCQE");
      std::cout<<"---- dE/dx histogram file for nue CCRes\n";
      fDedxDistributions[slid::kELECTRONCCRES].Initialize(pset, "HistFDDedxElectronCCRES");
      std::cout<<"---- dE/dx histogram file for nue CC DIS\n";
      fDedxDistributions[slid::kELECTRONCCDIS].Initialize(pset, "HistFDDedxElectronCCDIS");
      std::cout<<"---- dE/dx histogram file for nue Coherent\n";
      fDedxDistributions[slid::kELECTRONCCCOH].Initialize(pset, "HistFDDedxElectronCCCOH");

    } else if (detectorID == novadaq::cnv::DetId::kNDOS) {

      /// NDOS dE/dx input histograms
      
      std::cout<<"---- dE/dx histogram file for electron\n";
      fDedxDistributions[slid::kELECTRON].Initialize(pset, "HistNDOSDedxElectron");
      std::cout<<"---- dE/dx histogram file for photon\n";
      fDedxDistributions[slid::kPHOTON].Initialize(pset, "HistNDOSDedxPhoton");
      std::cout<<"---- dE/dx histogram file for muon\n";
      fDedxDistributions[slid::kMUON].Initialize(pset, "HistNDOSDedxMuon");
      std::cout<<"---- dE/dx histogram file for pi0\n";
      fDedxDistributions[slid::kPI0].Initialize(pset, "HistNDOSDedxPi0");
      std::cout<<"---- dE/dx histogram file for proton\n";
      fDedxDistributions[slid::kPROTON].Initialize(pset, "HistNDOSDedxProton");
      std::cout<<"---- dE/dx histogram file for neutron\n";
      fDedxDistributions[slid::kNEUTRON].Initialize(pset, "HistNDOSDedxNeutron");
      std::cout<<"---- dE/dx histogram file for charged pion\n";
      fDedxDistributions[slid::kPION].Initialize(pset, "HistNDOSDedxPion");
      std::cout<<"---- dE/dx histogram file for nue CCQE\n";
      fDedxDistributions[slid::kELECTRONCCQE].Initialize(pset, "HistNDOSDedxElectronCCQE");
      std::cout<<"---- dE/dx histogram file for nue CCRes\n";
      fDedxDistributions[slid::kELECTRONCCRES].Initialize(pset, "HistNDOSDedxElectronCCRES");
      std::cout<<"---- dE/dx histogram file for nue CC DIS\n";
      fDedxDistributions[slid::kELECTRONCCDIS].Initialize(pset, "HistNDOSDedxElectronCCDIS");
      std::cout<<"---- dE/dx histogram file for nue Coherent\n";
      fDedxDistributions[slid::kELECTRONCCCOH].Initialize(pset, "HistNDOSDedxElectronCCCOH");

    } else if (detectorID == novadaq::cnv::DetId::kNEARDET) {

      /// ND dE/dx input histograms
      
      std::cout<<"---- dE/dx histogram file for electron\n";
      fDedxDistributions[slid::kELECTRON].Initialize(pset, "HistNDDedxElectron");
      std::cout<<"---- dE/dx histogram file for photon\n";
      fDedxDistributions[slid::kPHOTON].Initialize(pset, "HistNDDedxPhoton");
      std::cout<<"---- dE/dx histogram file for muon\n";
      fDedxDistributions[slid::kMUON].Initialize(pset, "HistNDDedxMuon");
      std::cout<<"---- dE/dx histogram file for pi0\n";
      fDedxDistributions[slid::kPI0].Initialize(pset, "HistNDDedxPi0");
      std::cout<<"---- dE/dx histogram file for proton\n";
      fDedxDistributions[slid::kPROTON].Initialize(pset, "HistNDDedxProton");
      std::cout<<"---- dE/dx histogram file for neutron\n";
      fDedxDistributions[slid::kNEUTRON].Initialize(pset, "HistNDDedxNeutron");
      std::cout<<"---- dE/dx histogram file for charged pion\n";
      fDedxDistributions[slid::kPION].Initialize(pset, "HistNDDedxPion");
      std::cout<<"---- dE/dx histogram file for nue CCQE\n";
      fDedxDistributions[slid::kELECTRONCCQE].Initialize(pset, "HistNDDedxElectronCCQE");
      std::cout<<"---- dE/dx histogram file for nue CC Res\n";
      fDedxDistributions[slid::kELECTRONCCRES].Initialize(pset, "HistNDDedxElectronCCRES");
      std::cout<<"---- dE/dx histogram file for nue CC DIS\n";
      fDedxDistributions[slid::kELECTRONCCDIS].Initialize(pset, "HistNDDedxElectronCCDIS");
      std::cout<<"---- dE/dx histogram file for nue Coherent\n";
      fDedxDistributions[slid::kELECTRONCCCOH].Initialize(pset, "HistNDDedxElectronCCCOH");
    } else {
      throw cet::exception("DetectorTypeNotDefined")
	<< "Detector type " << detectorID
	<< " not recognized.  See LIDBuilder_module.cc for list of defined algorithms.  "
	<< "Only novadaq::cnv::DetId::kNEARDET, kFARDET, kNDOS defined.";

    }
    return true;
  }

  //    bool ParticleIDAlg::LoadDedxHistograms() {
  //
  //        for (int iParticleType = 0; iParticleType < slid::kDEDXPARTICLETYPESIZE;
  //                iParticleType++) {
  //            fDedxDistributions[iParticleType].LoadDedxHistograms();
  //        }
  //        return true;
  //    }

  ///
  /// \brief Determine detector XY region
  ///

  //*********************************************************************************************
  void ParticleIDAlg::CalcDetectorXYRegionIndex() 
  {

    TVector3 pos((fShower->Start()[0] + fShower->Stop()[0]) / 2.0,
		 (fShower->Start()[1] + fShower->Stop()[1]) / 2.0,
		 (fShower->Start()[2] + fShower->Stop()[2]) / 2.0);
    iReg = 0;
    //only switch quadrants if this is the Far Detector, for ND whole detector is ~quadrant 0
    // (near the readout in both views)
    if (fDetId == novadaq::cnv::DetId::kFARDET){
      if (pos[0] >= 0 && pos[1] >= 0)iReg = 0;
      if (pos[0] >= 0 && pos[1] < 0)iReg = 1;
      if (pos[0] < 0 && pos[1] >= 0)iReg = 2;
      if (pos[0] <= 0 && pos[1] <= 0)iReg = 3;
    }
    return;
  }

  ///
  /// \brief  Calculate energy bin indices and energy values from the histograms
  ///  corresponding energy of shower to be used to interpolate between bins in energy 
  ///

  //*********************************************************************************************
  void ParticleIDAlg::CalculateEnergyBinIndex() 
  {
    //        int& igev0, int& igev1,
    //            double& e0, double& e1

    double energyBinBounds[slid::kNumEnergyBin];
    for (int ig = 0; ig < slid::kNumEnergyBin; ig++) {
      energyBinBounds[ig] = (slid::kEnergyBins[ig] + kEnergyBins[ig + 1]) / 2.;
    }

    igev0 = 0;
    igev1 = 0;
    if (fShowerEnergyGev > energyBinBounds[slid::kNumEnergyBin - 1]) {
      igev0 = slid::kNumEnergyBin - 1;
      igev1 = slid::kNumEnergyBin - 1;
    } else if (fShowerEnergyGev < energyBinBounds[0]) {
      igev0 = 0;
      igev1 = 0;
    } else {
      for (int ig = 0; ig < slid::kNumEnergyBin - 1; ig++) {
	if (fShowerEnergyGev > energyBinBounds[ig] && fShowerEnergyGev < energyBinBounds[ig + 1]) {
	  igev0 = ig;
	  igev1 = ig + 1;
	}
      }
    }
    if (igev0 > slid::kNumEnergyBin - 1) igev0 = slid::kNumEnergyBin - 1;
    if (igev1 > slid::kNumEnergyBin - 1) igev1 = slid::kNumEnergyBin - 1;
    e0 = fShowerEnergyGev - energyBinBounds[igev0];
    e1 = energyBinBounds[igev1] - fShowerEnergyGev;
  }

  //*********************************************************************************************
  double ParticleIDAlg::CalcTrkHitPath(unsigned int plane) 
  {

    TVector3 trkDirection = fShower->Dir();
    TVector3 trkstart = fShower->Start();
    TVector3 trkstop = fShower->Stop();
    double trklength = fabs((trkstart - trkstop).Mag());
    double cellDepth = 2 * fGeom->Plane(plane)->Cell(0)->HalfD();

    // if track is only one cell in length or short, return the calculated full track length
    if (fabs(trkDirection.z()) < 1e-6 || fPlaneHits.size() == 1) {
      return trklength;
    }
    // Calculate path length through cell for track with given direction
    double trkhitpath = cellDepth * (util::pythag(trkDirection.x(),trkDirection.y(),trkDirection.z()) / fabs(trkDirection.z()));
    return trkhitpath;
  }

  //*********************************************************************************************
  double ParticleIDAlg::CalcEnergyInLongitudinalPlane(int iPlane) {
    double energySum = 0;
    std::map<int, int> planecells = fPlaneHits[iPlane];
    for (std::map<int, int>::iterator itr=planecells.begin(); itr!= planecells.end(); ++itr) {
      double cellE = fNuEEnergyAlg->CellEnergy(fShower, fShowerCol, itr->second, eventID);
      if ((cellE == -5) || std::isnan(cellE)) cellE = 0.0;
      energySum += cellE;
    }
    double ecor = fNuEEnergyAlg->ECorrMC(energySum);
    return energySum/ecor;
  }

  //NOTE: The input agruement for this function is the actual plane number in the detector
  //*********************************************************************************************
  double ParticleIDAlg::CalcPlaneLongitudinalDedx(unsigned int iPlane) 
  {
    double dedx = 0.0;
    //
    //  Lazy caching-- only calculate if fDedxLongByPlane[iPlane] not yet set.
    //
    std::map< int, double>::iterator planeIter = fDedxLongByPlane.find(iPlane);
    //std::cout<<"Asked to calculate long dedx for plane "<<iPlane<<std::endl;
    if (planeIter == fDedxLongByPlane.end()) {
      double energyInPlane = CalcEnergyInLongitudinalPlane(iPlane);
      double pathLength = CalcTrkHitPath(iPlane);

      if (pathLength > 0.0) dedx = energyInPlane / pathLength;
      
      fDedxLongByPlane.insert(planeIter, std::pair< int, double>(iPlane, dedx));
      //std::cout<<"plane idx: "<<iPlane<<" path: "<<pathLength<<" ep: "<<energyInPlane<<" dedx: "<<dedx<<std::endl;
    }
    else dedx = planeIter->second;
    return dedx;
  }

  //note that this function is being implemented in the same fasion as RecoJMShower
  //once we verify it is doing the right thing there may be ways to simplify
  //*********************************************************************************************
  double ParticleIDAlg::CalcPlaneTransverseDedx(unsigned int iTransversePlane)
  {
    double dedx = -1.0;
    //
    //  Lazy caching-- only calculate if fDedxTransByPlane[iPlane] not yet set.
    //
    std::map< int, double>::iterator transPlaneIter = fDedxTransByPlane.find(iTransversePlane);
    if (transPlaneIter == fDedxTransByPlane.end()) {
      //loop over hits in longitudinal planes to calculate transverse plane quantities
      for(std::map<int, std::map<int, int> >::iterator itr=fPlaneHits.begin(); itr!= fPlaneHits.end(); ++itr){
        int cplane = itr->first;
        std::map<int, int> planecells = itr->second;
        if (planecells.size() == 0) continue;
        if (fPlaneCentroid[itr->first] == -999.0) continue;
        double avgPos = fPlaneCentroid[itr->first];
        int meanCell = (int)avgPos;
        double wm = avgPos - (double)meanCell;//energy splitting for the centroid cell
        double wp = 1-wm; //energy splitting for the centroid cell
        if(meanCell<0)meanCell=0;
        if(meanCell>(int)fGeom->Plane(cplane)->Ncells()-1)meanCell=(int)fGeom->Plane(cplane)->Ncells()-1;
        double dir = 1.;
        if(fabs(fShower->Dir()[2])>0.000001)dir=fabs(fShower->Dir()[2]);
        int ic = (int)iTransversePlane + 1;
        double dph = ((double)ic/dir)+wm;//upper limit for cells included in a transverse cell index
        double dpl = ((double)(ic-1)/dir)+wm;//lower limit for cells included in a transverse cell index
        //if the mean cell is inside the detector
        if(meanCell+(int)dph<(int)fGeom->Plane(cplane)->Ncells()){
          double pfHigh = dph-(int)dph; //fraction of last cell on high end of transverse plane
          double pfLow = (int)dpl+1-dpl; //fraction of last cell on low end of transverse plane
          int cellHigh = meanCell + (int)dph;
          int cellLow  = meanCell + (int)dpl;
          std::map<int, int> cellsInTransPlane;
          cellsInTransPlane.insert(planecells.lower_bound(cellLow),
                                   planecells.upper_bound(cellHigh));
          for (std::map<int, int>::iterator itrC = cellsInTransPlane.begin(); itrC != cellsInTransPlane.end(); ++itrC){
            double cellE = fNuEEnergyAlg->CellEnergy(fShower, fShowerCol, itrC->second, eventID);
            if ((cellE == -5) || std::isnan(cellE)) cellE = 0.0;
            if (itrC->first == cellLow)  fTransEnergy[iTransversePlane] += pfLow*cellE;
            else if (itrC->first == cellHigh) fTransEnergy[iTransversePlane] += pfHigh*cellE;
            else fTransEnergy[iTransversePlane] += cellE;
          }

          fTransPathLength[iTransversePlane] += CalcTrkHitPath(cplane);
        }
        double dmh = ((double)ic/dir)+wp;//upper limit for cells included in a transeverse cell index
        double dml = ((double)(ic-1)/dir)+wp;//lower limit for cells included in a transeverse cell index

        if(meanCell-(int)dmh>=0){
          double pfHigh = dmh-(int)dmh; //fraction of last cell on high end of transverse plane
          double pfLow = (int)dml+1-dml; //fraction of last cell on low end of transverse plane
          int cellHigh = meanCell - (int)dmh;
          int cellLow  = meanCell - (int)dml;
          std::map<int, int> cellsInTransPlane;
          cellsInTransPlane.insert(planecells.lower_bound(cellHigh),
                                   planecells.upper_bound(cellLow));
          for (std::map<int, int>::iterator itrC = cellsInTransPlane.begin(); itrC != cellsInTransPlane.end(); ++itrC){
            double cellE = fNuEEnergyAlg->CellEnergy(fShower, fShowerCol, itrC->second, eventID);
            if ((cellE == -5) || std::isnan(cellE)) cellE = 0.0;
            if (itrC->first == cellLow)  fTransEnergy[iTransversePlane] += pfLow*cellE;
            else if (itrC->first == cellHigh) fTransEnergy[iTransversePlane] += pfHigh*cellE;
            else fTransEnergy[iTransversePlane] += cellE;
          }
          fTransPathLength[iTransversePlane] += CalcTrkHitPath(cplane);
        }
      }
      fTransPathLength[iTransversePlane] /= 2.0;
      if(fTransPathLength[iTransversePlane]>0.00001)dedx = fTransEnergy[iTransversePlane]/fTransPathLength[iTransversePlane];
      fDedxTransByPlane[iTransversePlane] = dedx;
    }
    else dedx = transPlaneIter->second;


    return dedx;
  }

  //*********************************************************************************************
  double ParticleIDAlg::CalcPlaneLongDedxProb(DedxParticleType particleType, int iGev, unsigned int iPlane, unsigned int plane) 
  {
    double prob = 0;

    if (iReg < 0 || iReg > kNumXYRegion - 1) return 1;
    if (iGev >= kNumEnergyBin || iGev < 0) return 1;
    if (iPlane >= kNumLongitudinalPlane) return 1;
    double dedx = CalcPlaneLongitudinalDedx(plane);
    double normHist = double(fDedxDistributions[particleType].GetLongitudinalPlaneHitDedxHist(iReg, iGev, iPlane)->GetEntries());
    double dedxHistValue = double(fDedxDistributions[particleType].GetLongitudinalPlaneHitDedxHist(iReg, iGev, iPlane)->Interpolate(dedx));
    double nHistBins = fDedxDistributions[particleType].GetLongitudinalPlaneHitDedxHist(iReg, iGev, iPlane)->GetNbinsX();
    if (dedxHistValue < 0.0001) dedxHistValue = 0.0001;
    if (normHist >= 1) prob = dedxHistValue * nHistBins / normHist;
    return prob;
  }

  //*********************************************************************************************
  double ParticleIDAlg::CalcPlaneTransDedxProb(DedxParticleType particleType, int iGev, unsigned int iTransversePlane) 
  {
    double prob = 0;
    if (iReg < 0 || iReg > kNumXYRegion - 1)return 1;
    if (iGev >= kNumEnergyBin || iGev < 0)return 1;
    if (iTransversePlane >= kNumTransversePlane)return 1;
    double dedx = CalcPlaneTransverseDedx(iTransversePlane);
    double normHist = double(fDedxDistributions[particleType].GetTransversePlaneCellDedxHist(iReg, iGev, iTransversePlane)->GetEntries());
    double dedxHistValue = double(fDedxDistributions[particleType].GetTransversePlaneCellDedxHist(iReg, iGev, iTransversePlane)->Interpolate(dedx));
    double nHistBins = fDedxDistributions[particleType].GetTransversePlaneCellDedxHist(iReg, iGev, iTransversePlane)->GetNbinsX();
    if (dedxHistValue < 0.0001)dedxHistValue = 0.0001;
    if (normHist >= 1) prob = dedxHistValue * nHistBins / normHist;
    return prob;
  }

  //*********************************************************************************************
  double ParticleIDAlg::CalcInterPlaneDedxProb(const DedxParticleType type, unsigned int iLongPlane, unsigned int plane) 
  {

    
    double probPXPY = 0;
    double cos = fShower->Dir()[2];
    unsigned int plane0 = int( int(iLongPlane) / fabs(cos));
    unsigned int plane1 = int( int(iLongPlane) / fabs(cos)) + 1;
    double r0 = double(iLongPlane / fabs(cos)) - double(plane0);
    double r1 = 1 - r0;
    double probe0r0 = 1;
    double probe0r1 = 1;
    double probe1r0 = 1;
    double probe1r1 = 1;
    probe0r0 = CalcPlaneLongDedxProb(type, igev0, plane0, plane);
    probe0r1 = CalcPlaneLongDedxProb(type, igev0, plane1, plane);
    probe1r0 = CalcPlaneLongDedxProb(type, igev1, plane0, plane);
    probe1r1 = CalcPlaneLongDedxProb(type, igev1, plane1, plane);

    double probe0 = (r1 * probe0r0 + r0 * probe0r1);
    double probe1 = (r1 * probe1r0 + r0 * probe1r1);
    if (igev0 != igev1)probPXPY = (e1 * probe0 + e0 * probe1) / (e0 + e1);
    if (igev0 == igev1)probPXPY = probe0;
    return probPXPY;
  }

  //*********************************************************************************************
  double ParticleIDAlg::CalcInterCellTransDedxProb(const DedxParticleType type, unsigned int iTransPlane)
  {
    double probPXPY = 0;
    double probe0 = CalcPlaneTransDedxProb(type, igev0, iTransPlane);
    double probe1 = CalcPlaneTransDedxProb(type, igev1, iTransPlane);
    if (igev0 != igev1)probPXPY = (e1 * probe0 + e0 * probe1) / (e0 + e1);
    if (igev0 == igev1)probPXPY = probe0;
    return probPXPY;
  }

  //*********************************************************************************************
  double ParticleIDAlg::GetGapVertexToShowerStart(const rb::Shower* vShower, TVector3 evtVtx, const art::Event& evt)
  {

    //initialize event vertex to start of track and first plane in track
    //TVector3 evtVtx(vShower->Start()[0], vShower->Start()[1], vShower->Start()[2]);
    double vtxDoca = PointDoca(evtVtx, vShower->Start(), vShower->Stop());
    double vtxToStart = (evtVtx - vShower->Start()).Mag();
    double vtxIntToStart = sqrt(vtxToStart * vtxToStart - vtxDoca * vtxDoca);
    //(Jianming's comments) record the gap from vertex to the start of the track
    //if the vertex is inside a track record distance of closest approach
    //TODO: For fuzzyK there can be an internal gap, where the start of the track is misleading
    //Ruth has looked into this, will try to incorporate internal gaps into this algorithm
    double shGap = vtxIntToStart;
    if (vtxIntToStart >= 0 && vtxIntToStart < vtxDoca)shGap = vtxDoca;
    if (vtxIntToStart < 0)shGap = vtxDoca;
    return shGap;
  }

  //*********************************************************************************************
  double ParticleIDAlg::PointDoca(TVector3 vtx, TVector3 start, TVector3 stop) 
  {
    double d = 9999;
    double denorm = (start - stop).Mag();
    if (denorm < 0.00001) {
      d = (vtx - start).Mag();
    } else {
      d = ((vtx - start).Cross(vtx - stop)).Mag() / denorm;
    }
    return d;
  }

  //*********************************************************************************************
  double ParticleIDAlg::Pi0Mass(const std::vector<const rb::Shower*> showerCollection,
				unsigned int iShower, const art::Event& evt) 
  {
    // Load shower if this is a new one.
    fClosestPi0Mass = -999.;
    SetShower(showerCollection.at(iShower), showerCollection, evt);
    
    double thisShowerERaw = fNuEEnergyAlg->ShowerEnergy(showerCollection[iShower], showerCollection, evt);
    if ((thisShowerERaw < 1e-10) || std::isnan(thisShowerERaw)) thisShowerERaw = 0.0;
    TLorentzVector thisShower4Vec(showerCollection.at(iShower)->Dir()[0] * thisShowerERaw,
				  showerCollection.at(iShower)->Dir()[1] * thisShowerERaw,
				  showerCollection.at(iShower)->Dir()[2] * thisShowerERaw, thisShowerERaw);
    
    for (unsigned int j = 0; j < showerCollection.size(); j++) { // pi0 mass calculation
      if (j == iShower)continue;
      // NOT NECESSARY?                if (showerCollection.at(j).ISlice() != showerCollection.at(iShower).ISlice())continue;
      double showereraw = fNuEEnergyAlg->ShowerEnergy(showerCollection[j], showerCollection, evt);
      if ((showereraw < 1e-10) || std::isnan(showereraw)) showereraw = 0.0;
      TLorentzVector otherShower4Vec(showerCollection.at(j)->Dir()[0] * showereraw,
				     showerCollection.at(j)->Dir()[1] * showereraw,
				     showerCollection.at(j)->Dir()[2] * showereraw, showereraw);
      TLorentzVector p2g = thisShower4Vec + otherShower4Vec;
      double mgg = p2g.M();
      if (fabs(fClosestPi0Mass - 0.15) > fabs(mgg - 0.15)) {
	fClosestPi0Mass = mgg;
	fPi0Index = j;
      }
    }
    if (fClosestPi0Mass < 0)fClosestPi0Mass = 0.;
    return fClosestPi0Mass;
  }


  ///\brief Index in showerCollection of the shower for which the reconstructed 
  /// invariant mass was closest to pi0 mass for the shower at index iShower.
  //*********************************************************************************************
  int ParticleIDAlg::Pi0SecondPhoton(const std::vector< const rb::Shower* > showerCollection,
				     unsigned int iShower, const art::Event& evt)
  {
    if( fShower == NULL)
      Pi0Mass( showerCollection, iShower, evt);
    
    return fPi0Index;
          
  }// end of Pi0SecondPhoton

  //*********************************************************************************************
  double ParticleIDAlg::Radius(const std::vector<const rb::Shower*> showerCollection,
			       unsigned int iShower, const art::Event& evt)
  {
    SetShower(showerCollection.at(iShower), showerCollection, evt);
    double rad = 0;
    double gev = 0;
    double dist, di;
    for (unsigned int nh=0;nh<fShower->NCell();nh++) {
      art::Ptr<rb::CellHit>  cHit = fShower->Cell(nh);
      double cellEnergy = fNuEEnergyAlg->CellEnergy(fShower, fShowerCol, nh, eventID);;
      TVector3 start  = fShower->Start();
      TVector3 dir    = fShower->Dir();
      TVector3 xyz;
      fGeom->Plane(cHit->Plane())->Cell(cHit->Cell())->GetCenter(xyz);
      const TVector3 dxyz = (cHit->View() == geo::kX) ? TVector3(0, 1, 0): TVector3(1, 0, 0);
      dist = sqrt(geo::ClosestApproach(start, start+dir, xyz, xyz+dxyz, &di));
      rad += dist*cellEnergy;
      gev += cellEnergy;
    }
    if(gev>0.001)rad = rad/gev;
    return rad;
  }

  //*********************************************************************************************
  double ParticleIDAlg::PlaneLongDedx(unsigned int pIdx)
  {
    std::map<unsigned int, unsigned int>::iterator itr = fMapIdxToGlobalPlane.find(pIdx);
    double dedx = -1.0;
    if (itr != fMapIdxToGlobalPlane.end()) dedx = CalcPlaneLongitudinalDedx(itr->second);
    else  std::cout<<"Requested longitudinal plane index not found, returning dedx=-1"<<std::endl;

    return dedx;
  }

  //*********************************************************************************************
  double ParticleIDAlg::PlaneTransDedx(unsigned int tpIdx)
  {
    double dedx = -1.0;
    if (tpIdx < slid::kNumTransversePlane) dedx = CalcPlaneTransverseDedx(tpIdx);
    else std::cout<<"Requested transverse plane index exceeds valid range, maximum plane is: "<<slid::kNumTransversePlane<<", returning dedx=-1"<<std::endl;

    return dedx;
  }

  //*********************************************************************************************
  double ParticleIDAlg::PlaneRadius(unsigned int pIdx)
  {
    std::map<unsigned int, unsigned int>::iterator itr = fMapIdxToGlobalPlane.find(pIdx);
    double rad = -1.0;
    double gev = 0;
    double dist, di;
    if (itr != fMapIdxToGlobalPlane.end()){
      std::map<int, int> planecells = fPlaneHits[itr->second];
      rad = 0.0;
      for (std::map<int, int>::iterator it=planecells.begin(); it!= planecells.end(); ++it) {
        art::Ptr<rb::CellHit>  cHit = fShower->Cell(it->second);
        double cellEnergy = fNuEEnergyAlg->CellEnergy(fShower, fShowerCol, it->second, eventID);
        TVector3 start  = fShower->Start();
        TVector3 dir    = fShower->Dir();
        TVector3 xyz;
        fGeom->Plane(cHit->Plane())->Cell(cHit->Cell())->GetCenter(xyz);
        const TVector3 dxyz = (cHit->View() == geo::kX) ? TVector3(0, 1, 0): TVector3(1, 0, 0);
        dist = sqrt(geo::ClosestApproach(start, start+dir, xyz, xyz+dxyz, &di));
        rad += dist*cellEnergy;
        gev += cellEnergy;
      }
    }
    else  std::cout<<"Requested longitudinal plane index not found, returning dedx=-1"<<std::endl;

    if(gev>0.001)rad = rad/gev;
    return rad;
  }

  //*********************************************************************************************
  TVector3 ParticleIDAlg::PlaneHitXYZ(unsigned int pIdx)
  {
    TVector3 trkplanepos;
    std::map<unsigned int, unsigned int>::iterator itr = fMapIdxToGlobalPlane.find(pIdx);
    if (itr != fMapIdxToGlobalPlane.end()){
      Double_t cellXYZ[3]={0,0,0};
      art::ServiceHandle<geo::Geometry> geom;
      geom->Plane(itr->second)->Cell(0)->GetCenter(cellXYZ);
      TVector3 trkdir=fShower->Dir();
      TVector3 trkstart=fShower->Start();
      TVector3 trkstop=fShower->Stop();
      double trkplanex=99999,trkplaney=99999, trkplanez=0;
      if(fabs(trkdir.z())>0.0001){
        trkplanex = trkstart.x()+(cellXYZ[2]-trkstart.z())*(trkdir.x()/trkdir.z());
        trkplaney = trkstart.y()+(cellXYZ[2]-trkstart.z())*(trkdir.y()/trkdir.z());
      }
      else{
        trkplanex = (trkstart.x()+trkstop.x())/2.0; 
        trkplaney = (trkstart.y()+trkstop.y())/2.0;
      }
      trkplanez = cellXYZ[2];
      trkplanepos.SetX(trkplanex);
      trkplanepos.SetY(trkplaney);
      trkplanepos.SetZ(trkplanez);
    }

    return trkplanepos;
  }

  //*********************************************************************************************
  double ParticleIDAlg::PlaneCentroid(unsigned int pIdx)
  {
    double centroid = 9999.0;
    std::map<unsigned int, unsigned int>::iterator itr = fMapIdxToGlobalPlane.find(pIdx);
    if (itr != fMapIdxToGlobalPlane.end()){
      std::map<int, double>::iterator it = fPlaneCentroid.find(itr->second);
      if (it != fPlaneCentroid.end()) centroid = it->second;
    }

    return centroid;
  }

  //*********************************************************************************************
  void ParticleIDAlg::CalcAsymIneria()
  {
    art::ServiceHandle<geo::Geometry> geom;
    double I11 = 0;
    double I12 = 0;
    double I21 = 0;
    double I22 = 0;
  
    // Using 3-D nodes to calculate shower balance
    //     TVector3 core(shower.Start()[0],shower.Start()[1],shower.Start()[2]); // original point of shower frame
    TVector3 iz = fShower->Dir().Unit(); // z in shower frame
    TVector3 ix0(1.0,0.0,0);
    TVector3 iy0(0.0,1.0,0);
    TVector3 ix = iy0.Cross(iz).Unit(); // x axis in shower frame
    TVector3 iy = iz.Cross(ix).Unit(); // y axis in shower frame
  
    std::map<unsigned int, unsigned int>::iterator itr2 = fMapIdxToGlobalPlane.begin();
    for(std::map<unsigned int, unsigned int>::iterator itr = fMapIdxToGlobalPlane.begin(); itr != fMapIdxToGlobalPlane.end(); ++itr){
      itr2++;
      std::map<int, int> planecells1 = fPlaneHits[itr->second];
      if (itr2 == fMapIdxToGlobalPlane.end()) break;
      std::map<int, int> planecells2 = fPlaneHits[itr2->second];
      if(planecells1.size()==0)continue;
      unsigned int plane1 = itr->second;
      if(planecells2.size()==0)continue;
      unsigned int plane2 = itr2->second;
      geo::View_t view1 = geom->Plane(plane1)->View();
      geo::View_t view2 = geom->Plane(plane2)->View();
      if(view1==view2)continue;
      TVector3 core = PlaneHitXYZ(itr->first);
      TVector3 pcore = PlaneHitXYZ(itr2->first);
      std::vector<TVector3> planenodecol;
      std::vector<double> planenodeenergycol;
      planenodecol.clear();
      planenodeenergycol.clear();
      for (std::map<int, int>::iterator citr = planecells1.begin(); citr != planecells1.end(); ++citr) {
        unsigned int cell1 = citr->first;
        for(std::map<int, int>::iterator citr2 = planecells2.begin(); citr2 != planecells2.end(); ++citr2){
          unsigned int cell2 = citr2->first;
          TVector3 node = GetCellNodePos(plane1, cell1, plane2, cell2);// define 3-D node
          double nodez = (pcore - fShower->Start()).Mag();
          TVector3 nodetopcore = node - pcore;
          nodetopcore.SetZ(nodez);
          TVector3 nodetocore = node - core;
          nodetocore.SetZ(nodez);

          TVector3 xyz = node - core;//transform xy to shower frame 

          double p = fShower->Dir()[0];
          double q = fShower->Dir()[1];
          double r = fShower->Dir()[2];
          double x0 = xyz[0];
          double y0 = xyz[1];
          double z0 = xyz[2];
          TVector3 vt(q*z0-r*y0,r*x0-p*z0,p*y0-q*x0);

          double nodeenergy = fNuEEnergyAlg->CellEnergy(fShower, fShowerCol, citr->second, eventID) +
	    fNuEEnergyAlg->CellEnergy(fShower, fShowerCol, citr2->second, eventID);;

          I11 += nodeenergy*(nodetocore.Y()*nodetocore.Y());
          I12 += -nodeenergy*(nodetocore.X()*nodetocore.Y());
          I21 += -nodeenergy*(nodetocore.X()*nodetocore.Y());
          I22 += nodeenergy*(nodetocore.X()*nodetocore.X());
  
        }
      }
    }
    // calculate eiganvalues and diagonalize inertial tensor
    double Ixx = 0.5*(I11+I22+sqrt(pow(I11+I22,2)-4*(I11*I22-I12*I21)));
    double Iyy = 0.5*(I11+I22-sqrt(pow(I11+I22,2)-4*(I11*I22-I12*I21)));
    double asym = 1;
    if(fabs(Ixx+Iyy)>1e-10)asym=(Ixx-Iyy)/(Ixx+Iyy);

    TVector3 ixyz(Ixx,Iyy,0);
    fIxyz = ixyz;
    fAsym = asym;
  }

  //*********************************************************************************************
  TVector3 ParticleIDAlg::GetCellNodePos(unsigned int plane1, unsigned int cell1, unsigned int plane2, unsigned int cell2)
  {
    double cellXYZ1[3]={0,0,0};
    double cellXYZ2[3]={0,0,0};
    art::ServiceHandle<geo::Geometry> geom;
    geom->Plane(plane1)->Cell(cell1)->GetCenter(cellXYZ1);
    geom->Plane(plane2)->Cell(cell2)->GetCenter(cellXYZ2);
    geo::View_t view1 = geom->Plane(plane1)->View();
    geo::View_t view2 = geom->Plane(plane2)->View();
    TVector3 nullpos(-9999,-9999,-9999);
    if(view1==view2)return nullpos;
    double x=0,y=0,z=0;
    if(view1 == geo::kX && view2 == geo::kY){
      x = cellXYZ1[0];
      y = cellXYZ2[1];
    }else if(view1 == geo::kY && view2 == geo::kX){
      x = cellXYZ2[0];
      y = cellXYZ1[1];
    }
    z = (cellXYZ1[2]+cellXYZ2[2])/2.;
    TVector3 nodepos(x,y,z);
    return nodepos;
  }

  //*********************************************************************************************
  int ParticleIDAlg::GetMIPPlanes( const rb::Shower* shower,
				   const std::vector< const rb::Shower* > showercol,
				   const art::Event& evt)
  {
    if(shower == NULL){
      return 0;
    }
    SetShower( shower, showercol, evt);
    
    int nMipPlanes = 0;
    int nPlanes = 20;
    if( fPlaneHits.size() < 20 ) 
      nPlanes = fPlaneHits.size();
    if(nPlanes == 0)
      return 0;
   
    int count = 0;

    for( auto const & iPlaneHit : fPlaneHits ){
      if(count > nPlanes)
	break;
      ++count;

      double dedx = CalcPlaneLongitudinalDedx( iPlaneHit.first );

      if( dedx > fMipRangeHigh || dedx < fMipRangeLow )
	continue;

      nMipPlanes++;
    }
    return nMipPlanes;
  }// end of GetMIPPlanes

}// End of namespace slid
