////////////////////////////////////////////////////////////////////////
/// \brief   Build slid::LID objects to store electron ID, if asked for, 
///          otherwise, calculate LID info and make available.
///
/// \todo  
///
/// \author  Alex Smith smith@physics.umn.edu
////////////////////////////////////////////////////////////////////////

#ifndef SPIDALG_H
#define	SPIDALG_H

#include <vector>
#include <string>
#include "RecoBase/Cluster.h"
#include "RecoBase/Shower.h"

#include "TMVAClassification_TMlpANN_WithENu.h"
#include "TMVAClassification_TMlpANN_NoENu.h"
#include "ShowerLID/ShowerPID.h"

#include "ShowerLID/ParticleIDAlg.h"
#include "ShowerLID/NuEEnergyAlg.h"
#include "ShowerLID/DedxDistribution.h"
#include "messagefacility/MessageLogger/MessageLogger.h"
#include "TMVA/Reader.h"
#include "TMultiLayerPerceptron.h"


namespace slid {

  class SPIDAlg {
  public:

    /// Algorithm name with ENu

    explicit SPIDAlg(fhicl::ParameterSet const& pset, 
		    NuEEnergyAlg* fNuEEnergyAlgIn, 
		    int pidType);

    explicit SPIDAlg(const SPIDAlg& orig);
    virtual ~SPIDAlg();

    //        void reconfigure(const fhicl::ParameterSet& pset);

    /// \brief Initialize the MVA algorithm.  The parameter detectorID is an int corresponding to the 
    /// detector definitions in novadaq::cnv :  novadaq::cnv::kFARDET, novadaq::cnv::kNDOS,  novadaq::cnv::kNEARDET
    ///
    void Initialize();
    /// \brief Calculate value of predictors for MVA Algorithm using candidate rb::Shower.
    ///  A pointer to the object of ParticleIDAlg is also passed so that expensive 
    ///  calculations may not need to be done again if this particular shower is already loaded.
    ///

    double CalcMVAResult( slid::ShowerPID & showerspid);

    std::string Name() 
      {
	return fName;
      }

    std::string Description() 
      {
	return fDescription;
      }

    /// \brief Initialize tree to be used for training the algorithms
    void InitializeTrainingTree();

    /// \brief Define mapping of branches to class fields for tree to be used for training the algorithms
    void DefineTrainingTree();

    /// \brief Fill tree to be used for training the algorithms        
    void FillTrainingTree( const slid::ShowerPID & showerspid);
  private:

    ///  FCL Parameters for LIDAlg
	
    /// Name of MVA algorithm to use
    std::string fAlgorithmName;
	
    /// NuE energy calculations object
    NuEEnergyAlg* fNuEEnergyAlg;

    /// If true, the ANN is computed with event energy as an input variable
    int fPIDType;

    const std::string fMVAMethod = "TMlpANN";
    const std::string fName = fMVAMethod + "NoNuE";


    /// Name of algorithm
    const std::string fDescription = "TMVA ANN algorithm TMlpANN with no nu(E)";

    ///  Location of weight files.  Set from FCL file.
    std::string fLibPath; 

    /// Use TMVA training or RecoJMShower training
    bool fUseTMVATraining;


    /// FCL parameters for Nue energy calculation algorithm
    //        fhicl::ParameterSet fNuEEnergyAlgPSet;

    // JMTraining
    TMultiLayerPerceptron *fMlp;
    

    /// TMVA reader object used to store MVA model
    TMVA::Reader* fReader;

    
    /// XML weights file, but there seems to be a bug in TMVA that the module simply does 
    /// not work when loading from the weights for the TMlPANN algorithm.  Needs to be investigated further.
    TMVAClassification_TMlpANN_WithENu* fTMVAModuleTMlpANNWithENu; // Module with NuE
    TMVAClassification_TMlpANN_NoENu* fTMVAModuleTMlpANNNoENu; // Module without NuE


    /// event classification answer: 1 = signal, 0 = background
    int fType;
    //        ///
    //        ///  Model Predictors
    //        ///

    slid::ShowerPID* fShowerPID;


    float fVtxgev;
    float fShE;
    float fNueRecEnergy;
  };
} // end namespace slid
#endif	/* ANNTMVAALG_H */

