////////////////////////////////////////////////////////////////////////
/// \brief   Build and write out the ShowerPID objects and make shower PID
///           
/// \TODO    Jianming's oscillation weighting:  Need to find standard
//           infrastructure to do this and replace his
///           code, which I stripped out of this class
///
/// \author  Alex Smith smith@physics.umn.edu
////////////////////////////////////////////////////////////////////////

// ROOT includes
#include "TFile.h"
#include "TH1D.h"
#include "TLorentzVector.h"
#include "TMath.h"
#include "TMultiLayerPerceptron.h"
#include "TRandom.h"
#include "TTree.h"

// NOvA includes
#include "DAQChannelMap/DAQChannelMapConstants.h"
#include "Geometry/Geometry.h"
#include "RecoBase/Cluster.h"
#include "RecoBase/PID.h"
#include "RecoBase/FilterList.h"
#include "RecoBase/Vertex.h"
#include "Utilities/AssociationUtil.h"
#include "Calibrator/Calibrator.h"

// ART includes
#include "art/Framework/Core/EDProducer.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Services/Optional/TFileService.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Core/FindManyP.h"
#include "art/Framework/Core/FindMany.h"
#include "art/Persistency/Common/Assns.h"
#include "art/Persistency/Common/Ptr.h"
#include "messagefacility/MessageLogger/MessageLogger.h"


#include "fhiclcpp/ParameterSet.h"
#include "ShowerLID/DedxDistribution.h"
#include "ShowerLID/ParticleIDAlg.h"
#include "ShowerLID/NuEEnergyAlg.h"
#include "ShowerLID/SPIDAlg.h"
#include "ShowerLID/ShowerLID.h"
#include "Utilities/func/EnvExpand.h"

#include <cmath>

namespace slid {

  class SPIDBuilder : public art::EDProducer {
  public:
    explicit SPIDBuilder(const fhicl::ParameterSet& pset);
    ~SPIDBuilder();

    virtual void produce(art::Event& evt);

    virtual void reconfigure(const fhicl::ParameterSet& pset);

    //virtual void beginJob();
    virtual void beginRun(art::Run& run);
    //virtual void endJob();

  protected:
    std::string fShowerLabel;
    std::string fSlicerLabel;
    std::string fVertexLabel;
    bool fPIDOnlyForHighestEShower;
    bool fObeyPreselection;
    bool fSkipNoiseSlices;
    bool fTrainingMode;

    // future alg       const std::string kKNNWithENu = "knnwithenu";
    // future alg       const std::string kKNNNoENu = "knnnoenu";

    std::string fLibPath;
    int kIDNuE = 12; // TEMP, should find where stadard defs are.
    /// Particle ID alorithm (loglikelihoods and dE/dx)
    ParticleIDAlg* fParticleIDAlg;
    NuEEnergyAlg* fNuEEnergyAlg;

    /// MVAAlgorithm (Neural Net, KNN, etc. May include any algorithm that is
    /// a subclass of slid::SPIDAlg)
    slid::SPIDAlg* fMVAAlgEPi0;
    slid::SPIDAlg* fMVAAlgEPi0EL;
    /// FCL parameters for NuE Energy alorithm
    fhicl::ParameterSet fNuEEnergyAlgPSet;

    /// FCL parameters for particle ID alorithm (loglikelihoods and dE/dx)
    fhicl::ParameterSet fParticleIDAlgPSet;

    /// FCL parameters for SPID alorithm
    fhicl::ParameterSet fSPIDAlgPSet;

    /// determine if library load is necessary or not when moving to a new run
    bool libLoad = false;

  };
  bool CompareByE( const slid::ShowerPID& a, const slid::ShowerPID& b);
} // namespace




namespace slid {
  int Ncount0 = 0, Ncount1 = 0, Ncount2 = 0;
  int Nfill = 0, Nput = 0;
  //......................................................................

  SPIDBuilder::SPIDBuilder(fhicl::ParameterSet const& pset) 
    : fParticleIDAlg(0), fNuEEnergyAlg(0),
      fMVAAlgEPi0(0),
      fMVAAlgEPi0EL(0),
      fNuEEnergyAlgPSet(pset.get< fhicl::ParameterSet >("NuEEnergyAlgPSet")),
      fParticleIDAlgPSet(pset.get< fhicl::ParameterSet >("ParticleIDAlgPSet")),
      fSPIDAlgPSet(pset.get< fhicl::ParameterSet >("SPIDAlgPSet")) 
  {
    reconfigure(pset);
    // Define output structures produced.
    produces<std::vector<slid::ShowerPID> >();
    produces< art::Assns<slid::ShowerPID, rb::Shower> >();
    produces< art::Assns<slid::ShowerPID, rb::Cluster> >();
  }

  //......................................................................

  SPIDBuilder::~SPIDBuilder() 
  {
    if (fParticleIDAlg) delete fParticleIDAlg;
    if (fNuEEnergyAlg)  delete fNuEEnergyAlg;
    if (fMVAAlgEPi0)    delete fMVAAlgEPi0;
    if (fMVAAlgEPi0EL)  delete fMVAAlgEPi0EL;
  }

  //......................................................................

  void SPIDBuilder::reconfigure(const fhicl::ParameterSet& pset) 
  {
    //
    //  Read FCL parameters
    //
    fNuEEnergyAlgPSet = pset.get< fhicl::ParameterSet >("NuEEnergyAlgPSet");
    fParticleIDAlgPSet = pset.get< fhicl::ParameterSet >("ParticleIDAlgPSet");
    fSPIDAlgPSet = pset.get< fhicl::ParameterSet >("SPIDAlgPSet");
    fLibPath = util::EnvExpansion(pset.get< std::string >("LibraryPath"));
    fSlicerLabel = pset.get< std::string >("SlicerLabel");
    fShowerLabel = pset.get< std::string >("ShowerLabel");
    fVertexLabel = pset.get< std::string >("VertexLabel");
    fPIDOnlyForHighestEShower = pset.get<bool>("PIDOnlyForHighestEShower");
    fObeyPreselection = pset.get<bool>("ObeyPreselection");
    fSkipNoiseSlices = pset.get<bool>("SkipNoiseSlices");
    fTrainingMode = pset.get<bool>("TrainingMode");

  }

  void SPIDBuilder::beginRun(art::Run& run) 
  {
    art::ServiceHandle<geo::Geometry> fGeom;
    novadaq::cnv::DetId detId = fGeom->DetId();

    //NOTE: This code is here because the geometry service only becomes available to determine the
    //detector id at the begin run step.  However at the present moment the libraries being loaded
    //by the particle id and ann algorithms will not very from run to run and only depend on the detector id.
    //If there are more then one run in a file this loading process takes a lot of time, only do it once per job.
    if (!libLoad){
      if (fParticleIDAlg) delete fParticleIDAlg;
      if (fNuEEnergyAlg)  delete fNuEEnergyAlg;
      if (fMVAAlgEPi0)    delete fMVAAlgEPi0;
      if (fMVAAlgEPi0EL)  delete fMVAAlgEPi0EL;
      fNuEEnergyAlg     = new NuEEnergyAlg(fNuEEnergyAlgPSet);
      fParticleIDAlg    = new ParticleIDAlg(fParticleIDAlgPSet, detId, fNuEEnergyAlg);

      if (!fTrainingMode){
        fMVAAlgEPi0       = new SPIDAlg(fSPIDAlgPSet, fNuEEnergyAlg, 2);
        fMVAAlgEPi0EL     = new SPIDAlg(fSPIDAlgPSet, fNuEEnergyAlg, 3);
      }
      libLoad = true;
    }
  }


  //......................................................................

  void SPIDBuilder::produce(art::Event & evt) 
  {

    art::ServiceHandle<geo::Geometry> fGeom;

    LOG_DEBUG("SPIDBuilderDEBUG") << "In SPIDBuilder::produce" << '\n';

    std::unique_ptr< std::vector<slid::ShowerPID> >             showerpidcol(new std::vector<slid::ShowerPID>);
    std::unique_ptr< art::Assns<slid::ShowerPID, rb::Shower> >  assnSpidShower(new art::Assns<slid::ShowerPID, rb::Shower>);
    std::unique_ptr< art::Assns<slid::ShowerPID, rb::Cluster> > assnSpidSlice(new art::Assns<slid::ShowerPID, rb::Cluster>);

    art::Handle<std::vector<rb::Cluster> > slicecol;
    evt.getByLabel(fSlicerLabel, slicecol);
    art::PtrVector<rb::Cluster> slicelist;
    for (unsigned int i = 0; i < slicecol->size(); ++i) {
      art::Ptr<rb::Cluster> slice(slicecol, i);
      slicelist.push_back(slice);
    }
    //
    //  Grab list of rb::Shower* for calculations and reading and list of art::Ptr<rb::Shower>
    //  for writing the output data products.  Since these **should** be indexed the same, this
    //  should be OK.
    //
    art::FindManyP<rb::Shower> fmtP(slicecol, evt, fShowerLabel);
    art::FindMany<rb::Shower> fmt(slicecol, evt, fShowerLabel);
    art::FindManyP<rb::Vertex> fmv(slicecol, evt, fVertexLabel);

    // loop over all of the slices
    for (unsigned int iSlice = 0; iSlice < slicelist.size(); ++iSlice) {
      const size_t nextShwIdx = showerpidcol->size();
      //
      //  Skip noise slices if this option is selected in FCL
      //
      if (fSkipNoiseSlices && slicelist[iSlice]->IsNoise()) continue;
      //
      //  If there was nue preselection, skip slice if it failed if this option is selected in FCL
      //
      if (fObeyPreselection && rb::IsFiltered(evt, slicecol, iSlice)) continue;

      // get all of the showers in this slice
      if ((!fmt.isValid())||(!fmtP.isValid())) continue;
      const std::vector< const rb::Shower* > showerCollection = fmt.at(iSlice);
      const std::vector < art::Ptr < rb::Shower >> showerCollectionPtr = fmtP.at(iSlice);
      if ((showerCollection.empty())||(showerCollectionPtr.empty())) continue;

      //hold spid objects temporarily before sorting
      std::vector<slid::ShowerPID> tempspidcol;
      //keep track of shower index associationed with the spid object after sorting
      std::vector<std::pair<double,int> > associationlist;


      //get the vertex for the slice
      const std::vector<art::Ptr<rb::Vertex>> vertCol = fmv.at(iSlice);
      //currently only one vertex per slice
      TVector3 vert = vertCol[0]->GetXYZ();

      //
      // Get indices of highest and second-highest energy showers
      //
      double eShHighest = 0;
      double sliceShwE = 0;
      unsigned int iHighestEnergySh = -1;
      for (unsigned int jShower = 0; jShower < showerCollection.size(); ++jShower) {
	double thisESh = fNuEEnergyAlg->ShowerEnergy(showerCollection[jShower], showerCollection, evt);
        sliceShwE += thisESh;
	if (thisESh > eShHighest) {
	  eShHighest = thisESh;
	  iHighestEnergySh = jShower;
	}
      }

      TLorentzVector evtP4(0.0,0.0,0.0,0.0);
      //
      //  Loop over showers to fill PID info
      //
      std::map<int, int> showerLidToShower;
      unsigned int nShowers = showerCollection.size();
      for (unsigned int iShower = 0; iShower < nShowers; ++iShower) {
	
	
	slid::ShowerPID showerpid;

        double shwE = fNuEEnergyAlg->ShowerEnergy(showerCollection.at(iShower), showerCollection, evt);
        if (std::isnan(sliceShwE) || std::isnan(shwE) || std::isinf(sliceShwE) || std::isinf(shwE)) {
          mf::LogError("LIDBuilder")<< "sliceShwE/shwE is nan or inf";
          continue;
        }
        if (shwE < 1e-10) shwE = 0.0;
        showerpid.SetShowerEnergy(shwE);

        showerpid.SetDedx0(0);
        showerpid.SetDedx1(0);
        showerpid.SetDedx2(0);
        showerpid.SetDedx3(0);
        showerpid.SetDedx4(0);
        showerpid.SetDedx5(0);
        fParticleIDAlg->SetShower(showerCollection[iShower], showerCollection, evt);

	showerpid.SetELLL(fParticleIDAlg->DedxLongLL(slid::DedxParticleType::kELECTRON, showerCollection[iShower], showerCollection, evt));
       showerpid.SetELLT(fParticleIDAlg->DedxTransLL(slid::DedxParticleType::kELECTRON, showerCollection[iShower], showerCollection, evt));
	showerpid.SetGLLL(fParticleIDAlg->DedxLongLL(slid::DedxParticleType::kPHOTON, showerCollection[iShower], showerCollection, evt));
	showerpid.SetGLLT(fParticleIDAlg->DedxTransLL(slid::DedxParticleType::kPHOTON, showerCollection[iShower], showerCollection, evt));
	showerpid.SetMuLLL(fParticleIDAlg->DedxLongLL(slid::DedxParticleType::kMUON, showerCollection[iShower], showerCollection, evt));
	showerpid.SetMuLLT(fParticleIDAlg->DedxTransLL(slid::DedxParticleType::kMUON, showerCollection[iShower],showerCollection, evt));
	showerpid.SetPi0LLL(fParticleIDAlg->DedxLongLL(slid::DedxParticleType::kPI0, showerCollection[iShower],showerCollection, evt));
	showerpid.SetPi0LLT(fParticleIDAlg->DedxTransLL(slid::DedxParticleType::kPI0, showerCollection[iShower],showerCollection, evt));
	showerpid.SetPLLL(fParticleIDAlg->DedxLongLL(slid::DedxParticleType::kPROTON, showerCollection[iShower],showerCollection, evt));
	showerpid.SetPLLT(fParticleIDAlg->DedxTransLL(slid::DedxParticleType::kPROTON, showerCollection[iShower], showerCollection, evt));
	showerpid.SetNLLL(fParticleIDAlg->DedxLongLL(slid::DedxParticleType::kNEUTRON, showerCollection[iShower], showerCollection, evt));
	showerpid.SetNLLT(fParticleIDAlg->DedxTransLL(slid::DedxParticleType::kNEUTRON, showerCollection[iShower],showerCollection, evt));
	showerpid.SetPiLLL(fParticleIDAlg->DedxLongLL(slid::DedxParticleType::kPION, showerCollection[iShower],showerCollection, evt));
	showerpid.SetPiLLT(fParticleIDAlg->DedxTransLL(slid::DedxParticleType::kPION, showerCollection[iShower],showerCollection, evt));

        if(showerCollection[iShower]->ExtentPlane()>0)showerpid.SetDedx0(fParticleIDAlg->PlaneLongDedx(0));
        if(showerCollection[iShower]->ExtentPlane()>1)showerpid.SetDedx1(fParticleIDAlg->PlaneLongDedx(1));
        if(showerCollection[iShower]->ExtentPlane()>2)showerpid.SetDedx2(fParticleIDAlg->PlaneLongDedx(2));
        if(showerCollection[iShower]->ExtentPlane()>3)showerpid.SetDedx3(fParticleIDAlg->PlaneLongDedx(3));
        if(showerCollection[iShower]->ExtentPlane()>4)showerpid.SetDedx4(fParticleIDAlg->PlaneLongDedx(4));
        if(showerCollection[iShower]->ExtentPlane()>5)showerpid.SetDedx5(fParticleIDAlg->PlaneLongDedx(5));
//std::cout<<"SPID fParticleIDAlg dedx0 "<<fParticleIDAlg->PlaneLongDedx(0)<<std::endl;
//std::cout<<"SPID fParticleIDAlg ELLL, ELLT "<<showerpid.ELLL()<<" "<<showerpid.ELLT()<<std::endl;
//std::cout<<"SPID fParticleIDAlg GLLL, GLLT "<<showerpid.GLLL()<<" "<<showerpid.GLLT()<<std::endl;


        TLorentzVector showerptrk(showerCollection[iShower]->Dir()[0]*shwE,
                                  showerCollection[iShower]->Dir()[1]*shwE,
                                  showerCollection[iShower]->Dir()[2]*shwE,
                                  shwE);
        TVector3 nuDir = fGeom->NuMIBeamDirection();
        double theta = showerptrk.Angle(nuDir);
        showerpid.SetCosTheta(cos(theta));


        showerpid.SetVtxDoca(fParticleIDAlg->PointDoca(vertCol.at(0)->GetXYZ(),showerCollection[iShower]->Start(),showerCollection[iShower]->Stop()));


	if (!fPIDOnlyForHighestEShower || iShower == iHighestEnergySh){ 
          //in training mode just write -5, since the ann weights have not been determined
          if (fTrainingMode){
	    showerpid.SetAlgDescription(""); // not needed?
            showerpid.SetValEPi0(-5);
            showerpid.SetValEPi0EL(-5);
          }
          //standard running, calculate the ann
          else{
	    showerpid.SetValEPi0(fMVAAlgEPi0->CalcMVAResult(showerpid));
	    showerpid.SetValEPi0EL(fMVAAlgEPi0EL->CalcMVAResult(showerpid));
          }
	}

        showerpid.SetVtxDoca(fParticleIDAlg->PointDoca(vertCol.at(0)->GetXYZ(),showerCollection[iShower]->Start(),showerCollection[iShower]->Stop()));

        showerpid.fPartLongLL = fParticleIDAlg->fPartLongLL;
        showerpid.fPartTransLL = fParticleIDAlg->fPartTransLL;
        tempspidcol.push_back(showerpid);
        associationlist.push_back(std::make_pair(showerpid.ShowerEnergy(),iShower));
      }// end loop over showers
      
      std::sort( tempspidcol.begin(), tempspidcol.end(), CompareByE);
      std::sort(associationlist.rbegin(), associationlist.rend());
      
      for (unsigned int ispid=0; ispid<tempspidcol.size(); ++ispid){
	showerpidcol->push_back(tempspidcol[ispid]);
	util::CreateAssn(*this, evt, *(showerpidcol.get()), showerCollectionPtr[associationlist[ispid].second], *(assnSpidShower.get()));
      }

      //if we have added new showers, make event pid and associations to slice
      if (showerpidcol->size() > nextShwIdx){
        //create new associations for the most recent set of showers
	util::CreateAssn(*this, evt, *showerpidcol, slicelist[iSlice], *(assnSpidSlice.get()), nextShwIdx, showerpidcol->size());
      }
      
    } //end loop over slices
    
    evt.put(std::move(showerpidcol));
    evt.put(std::move(assnSpidShower));
    evt.put(std::move(assnSpidSlice));

  }// end of producer

  ////////////////////////////////////////////////////////////////////////

  bool CompareByE( const slid::ShowerPID& a, const slid::ShowerPID& b)

  {
    return a.ShowerEnergy() > b.ShowerEnergy();
  }
  
  DEFINE_ART_MODULE(SPIDBuilder);
}
