////////////////////////////////////////////////////////////////////////
/// \brief   Photon signal class
/// \author  jpaley@indiana.edu
/// \date
////////////////////////////////////////////////////////////////////////
#ifndef PHOTONSIGNAL_H
#define PHOTONSIGNAL_H

#include <iomanip>
#include <iosfwd>
#include <iostream>
#include <vector>

namespace sim 
{
  class PhotonSignal {
  public:
    PhotonSignal();
    PhotonSignal(double time, int npe);
    ~PhotonSignal();
    
    void  SetTimeMean(double t)             { fTimeMean = t;               }
    void  SetNPhoton(int npe)               { fNPhoton = npe;              }
    void  SetPlaneCell(int plane, int cell) { fCell = cell; fPlane = plane;}
    void  SetTrackId(int t)                 { fTrackId = t;                }
    void  SetPoissonLambda(double l)        { fPoissonLambda = l;          }
    //    void  Print();
    
    double         TimeMean() const { return fTimeMean; }
    int            NPhoton()  const { return fNPhoton;  }
    int            Cell()     const { return fCell;     }
    int            Plane()    const { return fPlane;    }
    int            TrackId()  const { return fTrackId;  }
    double         PoissonLambda() const { return fPoissonLambda; }

    friend std::ostream& operator<< (std::ostream& o, const PhotonSignal& ps);

    friend bool CompareByPlaneAndCell(const sim::PhotonSignal& a,
				      const sim::PhotonSignal& b);
  private:
    double         fTimeMean; ///< time of PEs at APD (ns)
    int            fNPhoton;  ///< number of PEs in this clump
    int            fCell;     ///< cell number
    int            fPlane;    ///< plane number
    int            fTrackId;  ///< track ID of parent

    /// \brief The mean of the distribution \ref fNPhoton was drawn from.
    ///
    /// Useful for calibration studies.
    double fPoissonLambda;
  };

  /// Helper for \ref SortByPlaneAndCell
  bool CompareByPlaneAndCell(const sim::PhotonSignal& a,
			     const sim::PhotonSignal& b);

  /// Sort \a c by plane and cell.
  /// ie guarantees hits on the same channel are adjacent.
  void SortByPlaneAndCell(std::vector<sim::PhotonSignal>& c);

  /// Helper for \ref SortByNPhoton
  bool CompareByNPhoton(const sim::PhotonSignal& a,
			const sim::PhotonSignal& b);

  /// Sort \a c by number of photons. Lowerst to highest.
  void SortByNPhoton(std::vector<sim::PhotonSignal>& c);

} // end namespace

#endif // PHOTONSIGNAL_H
//////////////////////////////////////////////////////////////////
