///////////////////////////////////////////////////////////////////////////
// \file    Slicer4D_module.cc
// \brief   This is based on a DBSCAN density based clustering algorithm
//          found in "A Density-Based Algorithm for Discovering Clusters in
//          Large Spatial Databases with Noise" by M.Ester et.al. (1996)
// \version $Id: Slicer4D_module.cc,v 1.0 2013-05-01 17:00:00 mbaird42 Exp $
// \authors mbaird42@fnal.gov, messier@indiana.edu
///////////////////////////////////////////////////////////////////////////

#include "art/Framework/Core/EDProducer.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Services/Optional/TFileDirectory.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"

#include "Calibrator/Calibrator.h"
#include "Geometry/Geometry.h"
#include "NovaDAQConventions/DAQConventions.h"
#include "RecoBase/CellHit.h"
#include "RecoBase/Cluster.h"
#include "Slicer/PointManager.h"

#include <cmath>
#include <string>



// Slicer4D header
namespace slicer {
  
  class Slicer4D : public art::EDProducer {
  public:
    explicit Slicer4D(fhicl::ParameterSet const &pset);
    virtual ~Slicer4D();

    void beginRun(art::Run& run);
    void produce(art::Event& evt);
    void endJob();

  protected:
    // internal methods
    bool GenerateCluster(unsigned int point, int ClID); 
    
    // configuration settings
    std::string  fCellHitInput;       ///< Read CellHits from this input
    bool         fDoNotSliceAnything; ///< If true, one big slice is written out

    double       fEpsilon;            ///< Minimum causal separation value for neighbors
    unsigned int fMinPts;             ///< minimum number of casual neighbors for a hit to be in a cluster
    bool         fUseLowPEPen;        ///< Add a causal distance penalty for PE (low PE is INCREASES the distance)
    double       fPEPen;              ///< PE value for which the low PE penalty will be 1.
    double       fDistPen;            ///< Distance (in cm) for which the distance penalty will be 1 (roughly one strong interaction length.)
    double       fOpVPlPen;           ///< Allowed plane gap for hits in opposite views for which the distance penalty will be 1.
    unsigned int fMinHitsPerView;     ///< Minimum number of hits per view to make a cluster

    fhicl::ParameterSet fPSetNDOS, fPSetND, fPSetFD;

    // global variables/objects
    PointManager fPtMan;              ///< Helper class for the 4D clustering
  };

}





namespace slicer
{
  //----------------------------------------------------------------------
  Slicer4D::Slicer4D(fhicl::ParameterSet const &pset) :
    fCellHitInput      (pset.get< std::string  >("CellHitInput")      ),
    fDoNotSliceAnything(pset.get< bool         >("DoNotSliceAnything")),
    fEpsilon           (0),
    fMinPts            (0),
    fUseLowPEPen       (0),
    fPEPen             (0),
    fDistPen           (0),
    fOpVPlPen          (0),
    fMinHitsPerView    (0),
    fPSetNDOS          (pset.get<fhicl::ParameterSet>("ndos")),
    fPSetND            (pset.get<fhicl::ParameterSet>("nd")),
    fPSetFD            (pset.get<fhicl::ParameterSet>("fd"))
  {
    produces< std::vector<rb::Cluster> >();
  }

  //----------------------------------------------------------------------
  Slicer4D::~Slicer4D()
  { }

  //----------------------------------------------------------------------
  void Slicer4D::beginRun(art::Run&)
  {
    art::ServiceHandle<geo::Geometry> geom;

    fhicl::ParameterSet pset;
    switch(geom->DetId()){
    case novadaq::cnv::kNDOS:
      pset = fPSetNDOS;
      break;
    case novadaq::cnv::kNEARDET:
      pset = fPSetND;
      break;
    case novadaq::cnv::kFARDET:
      pset = fPSetFD;
      break;
    default:
      assert(0 && "Unknown detector");
    }

    fEpsilon        = pset.get<double      >("Epsilon");
    fMinPts         = pset.get<unsigned int>("MinPts");
    fUseLowPEPen    = pset.get<bool        >("UseLowPEPen");
    fPEPen          = pset.get<double      >("PEPen");
    fDistPen        = pset.get<double      >("DistPen");
    fOpVPlPen       = pset.get<double      >("OppViewPlanePen");
    fMinHitsPerView = pset.get<unsigned int>("MinHitsPerView");
  }

  //----------------------------------------------------------------------
  void Slicer4D::endJob()
  {
    std::cout << "\n\n========== End Job ==========\n"
	      << "End Job: (eps, minpts, PEPen, DistPen, OpVPlPen, MinHitsPerView) "
	      << fEpsilon        << " "
	      << fMinPts         << " "
	      << fPEPen          << " "
	      << fDistPen        << " "
	      << fOpVPlPen       << " "
	      << fMinHitsPerView << "\n";
  }

  //----------------------------------------------------------------------
  void Slicer4D::produce(art::Event& evt)
  {
    // NOTE:  There is currently no requirement that a slice have hits in
    //        both views.

    art::ServiceHandle<calib::Calibrator> cal;

    // Get the cell hits
    art::Handle< std::vector<rb::CellHit> > hitcol;
    evt.getByLabel(fCellHitInput, hitcol);

    // Load the cell hits into a vector for easy use
    std::vector<art::Ptr<rb::CellHit > > hitlist;
    for(unsigned int i = 0; i < hitcol->size(); ++i){
      art::Ptr<rb::CellHit> hit(hitcol, i);
      hitlist.push_back(hit);  
    }
    
    // Sort the hits by time to improve the speed.
    // WARNING:  The rest of the code assumes that the hits are
    //           time sorted!
    rb::SortByTime(hitlist);



    // Initalize PointManager to help with the 4D clustering,
    // add the hit info to the point manager, and generate the
    // initial neighbor map.
    fPtMan.Initalize(fEpsilon, fPEPen, fDistPen, fOpVPlPen, fUseLowPEPen);
    
    fPtMan.Reset(hitlist.size());
    
    for(unsigned int i = 0; i < hitlist.size(); ++i) {
      fPtMan.AddHit(hitlist[i]->TNS(), cal->GetTimeRes(*hitlist[i]),
		    hitlist[i]->Plane(), hitlist[i]->Cell(), hitlist[i]->PE(),
		    hitlist[i]->View());
    }

    fPtMan.FindNeighbors();

    // OPTIONAL: Clean up the neighborhood (Get off my lawn you damn kids!)
    // fPtMan.CleanUpNeighborhood();

    std::unique_ptr< std::vector<rb::Cluster> >
      clustercol(new std::vector<rb::Cluster>);



    
    //
    // Use alternate slicing methods if requested
    //
    int CurrentClusterID = 1;

    if(fDoNotSliceAnything) {

      // We will only put one BIG cluster into the event.  I don't know
      // if anyone uses this feature, but I grandfathered it in from
      // the old slicer.

      rb::Cluster cluster;

      for(unsigned int i = 0; i < hitlist.size(); ++i)
	cluster.Add(hitlist[i]);
      
      // Add the cluster to clustercol
      clustercol->push_back(cluster);

    }
    // otherwise do the standard DBSCAN slicing
    else {
      for(unsigned int i = 0; i < hitlist.size(); ++i) {
	// If the point is unclassified, attempt to expand the cluster
	if(fPtMan.fClustIDs[i] == -1) {
	  if(this->GenerateCluster(i,CurrentClusterID))
	    CurrentClusterID++;
	} // end if point is unclassified
      } // end loop over all hits, i



      //
      // Construct the actual slices to be put into the event.
      //
      std::vector<rb::Cluster> clusters(CurrentClusterID);
      
      for(unsigned int i = 0; i < hitlist.size(); ++i)
	clusters[fPtMan.fClustIDs[i]].Add(hitlist[i]);
      
      // Move hits in non-noise clusters that are too small
      // to the noise cluster.
      for(unsigned int i = 1; i < clusters.size(); ++i) {

	if(clusters[i].NXCell() >= fMinHitsPerView &&
	   clusters[i].NYCell() >= fMinHitsPerView) continue;

	// This cluster is too small.  Put its hits into the noise
	// cluster and clear it.
	for(unsigned int j = 0; j < clusters[i].NCell(); ++j) {
	  art::Ptr<rb::CellHit> hit(clusters[i].Cell(j));
	  clusters[0].Add(hit);
	} // end for j (loop over hits)
	clusters[i].Clear();
      } // end for i (loop over clusters)

      // Tag the noise cluster as such.
      clusters[0].SetNoise(true);

      // Always put in the noise cluster even if it is empty.
      clustercol->push_back(clusters[0]);
      
      // Add all non-empty, non-noise clusters to clustercol
      for(unsigned int i = 1; i < clusters.size(); ++i) {

	if(clusters[i].NCell() == 0) continue;

	geo::View_t V = geo::kXorY;
	if(clusters[i].NXCell() == 0) V = geo::kY;
	if(clusters[i].NYCell() == 0) V = geo::kX;

	rb::Cluster tempcluster(clusters[i].AllCells(),V);
	clustercol->push_back(tempcluster);

      }

    } // end else { do the DBSCAN clustering }



    // put the slices into the event
    evt.put(std::move(clustercol));

  } // end Produce function

  //----------------------------------------------------------------------
  bool Slicer4D::GenerateCluster(unsigned int point, int ClID)
  {
    // In the original DBSCAN paper, this is the function they called
    // "ExpandCluster."  How this function works is most easily understood
    // by reading that paper (don't worry, it is pretty short...)

    // Get the list of neighbors for 'point.'
    std::vector<unsigned int> seeds = fPtMan.regionQuery(point);

    // If 'point' is not a core point, label it as noise and return.
    if(seeds.size() < fMinPts) {
      fPtMan.changeClustID(point, 0);
      return false;
    }
    else {
      // Assign 'point' and the neighbors of 'point' to the cluster ClID.
      for(unsigned int seedi = 0; seedi < seeds.size(); ++seedi)
	fPtMan.changeClustID(seeds[seedi],ClID);
      seeds.erase(seeds.begin());
    
      // Loop over the neighbors of 'point' to see if they are also core
      // points.
      while(seeds.size() > 0) {
	// Get the list of neighbors for the first seed point.
	std::vector<unsigned int> result = fPtMan.regionQuery(seeds[0]);
	
	// If the first seed point is also a core point, add its neighbors
	// to the list of seeds and assigne them to the cluster ClID.
	if(result.size() >= fMinPts) {
	  for(unsigned int resulti = 0; resulti < result.size(); ++resulti) {
	    // If the point is unassigned (-1) or previously labeled as
	    // noise (0), asign it ClID.
	    if(fPtMan.fClustIDs[result[resulti]] == -1 ||
	       fPtMan.fClustIDs[result[resulti]] == 0) {
	      // Only add this point to the list of seeds if it was
	      // previously unassigned.
	      if(fPtMan.fClustIDs[result[resulti]] == -1)
		seeds.push_back(result[resulti]);
	      fPtMan.changeClustID(result[resulti], ClID);
	    } // end if(resulti is -1 or 0)

	  } // end for resulti

	} // end if result.size() >= fMinPts
	seeds.erase(seeds.begin());

      } // end while
      return true;
    } // end else

  } // end GenerateCluster function

} // end namespace slicer
/////////////////////////////////////////////////////////////////////////

namespace slicer
{
  DEFINE_ART_MODULE(Slicer4D);
}
