//////////////////////////////////////////////////////////////////////////
/// \brief  Cosmic track selection code
/// \author Brian Rebel - brebel@fnal.gov
/// \date 
//////////////////////////////////////////////////////////////////////////
#include <vector>
#include <initializer_list>

#include "TVector3.h"

#include "messagefacility/MessageLogger/MessageLogger.h"

#include "TrackFit/CosmicTrackSelection.h"
#include "RecoBase/Track.h"
#include "RecoBase/Cluster.h"
#include "GeometryObjects/PlaneGeo.h"

namespace trk{

  //----------------------------------------------------------------------------
  CosmicTrackSelection::CosmicTrackSelection()
  {
    fhicl::ParameterSet pset;
    this->reconfigure(pset);

    return;
  }

  //----------------------------------------------------------------------------
  CosmicTrackSelection::~CosmicTrackSelection()
  {
  }

  //----------------------------------------------------------------------------
  void CosmicTrackSelection::reconfigure(fhicl::ParameterSet const& pset)
  {
    fMaxDeltaEndPlane   = pset.get< double >("MaxDeltaEndPlane",   3    );
    fMaxPlaneAsymmetry  = pset.get< double >("MaxPlaneAsymmetry",  0.1  );
    fMaxCellsPerPlane   = pset.get< double >("MaxCellsPerPlane",   6    );
    fBadStepSizeLimit   = pset.get< double >("BadStepSizeLimit",   3.   );
    fMaxCosTheta        = pset.get< double >("MaxCosTheta",        0.95 );
    fTrkLengthCut       = pset.get< double >("TrkLengthCut",       50.  );
    fMaxVtxDistFromEdge = pset.get< double >("MaxVtxDistFromEdge", 10.  );
    fMaxEndDistFromEdge = pset.get< double >("MaxEndDistFromEdge", 10.  );
    
    return;
  }

  //----------------------------------------------------------------------------
  bool CosmicTrackSelection::Stopper(TVector3 const& finalPoint)
  {
    // call it a stopper if the track ends at least 50 cm from an edge of
    // the detector
    // distEdge will be negative if the point is outside of the detector
    
    return (this->DistToEdge(finalPoint) > 50.);
  }

  //----------------------------------------------------------------------------
  // tracks that have huge changes in the step size from one trajectory
  // point to another are probably poorly reconstructed, and should be removed
  // from the analysis.  Usually the biggest steps are at either the start 
  // or end of the track in cases of really bad reconstruction
  bool CosmicTrackSelection::GoodSteps(rb::Track          const& track,
				       std::vector<float>      & steps,
				       float                   & median)
  {

    // first find the median step size
    size_t size = track.NTrajectoryPoints() - 1;

    // track with only two distances between trajectory points, 
    // probably not very useful for the analysis so return false
    if(size < 3) return false;

    std::vector<float> stepsize(size, 0.);
    float largestStep = -1.;
    for(size_t t = 1; t < size+1; ++t){
      TVector3 const& tp1 = track.TrajectoryPoint(t-1);
      TVector3 const& tp2 = track.TrajectoryPoint(t);
      stepsize[t-1] = (tp1 - tp2).Mag();
      largestStep = std::max(stepsize[t-1], largestStep);
    }

    // sort the vector of step sizes and find the median
    std::sort(stepsize.begin(), stepsize.end());

    median = 0.;
    if(size%2 == 0) 
      median = 0.5*(stepsize[size/2] + stepsize[(size/2) - 1]);
    else
      median = stepsize[size/2];

    steps.swap(stepsize);

    // the median represents the value of the step size 
    // distribution where half of the distribution is
    // below that value and half is above it.  If the 
    // largest step is more than twice the median, then 
    // it is very far outside the distribution and this
    // track was not well reconstructed
    return (largestStep < median*fBadStepSizeLimit);
  }

  //----------------------------------------------------------------------------
  bool CosmicTrackSelection::GoodReco(rb::Track const& track)
  {
    std::vector<float> steps;
    float median = 0.;

    return (std::abs(track.Stop().Z() - track.Start().Z()) >= fTrkLengthCut   &&
	    -track.Dir().Y() < fMaxCosTheta*track.Dir().Mag()                 &&
	    this->GoodSteps(track, steps, median) );    
  }

  //----------------------------------------------------------------------------
  bool CosmicTrackSelection::CleanSample(CosmicTrackInfo const& ri)
  {
    return (ri.stop                                                            && 
	    ri.goodVtx                                                         && 
	    std::max(ri.cellsPerPlaneX, ri.cellsPerPlaneY) < fMaxCellsPerPlane &&
	    std::abs(ri.xStartPlane - ri.yStartPlane)      < fMaxDeltaEndPlane &&
	    std::abs(ri.xEndPlane   - ri.yEndPlane)        < fMaxDeltaEndPlane &&
	    std::abs(ri.planesX - ri.planesY) < fMaxPlaneAsymmetry*(ri.planesX + ri.planesY));
  }

  //----------------------------------------------------------------------------
  bool CosmicTrackSelection::IsCalibTrack(rb::Track       const& track,
					  rb::Cluster     const& slice)
  {

    //Track must have at least two X cells and 2 Y cells:
    if(track.NXCell() < 2 || track.NYCell() < 2) { 
      return false; 
    }
	
    //Check that direction vector is well-defined:
    const TVector3 dr = track.Dir().Unit();
    const double   dx = std::abs(dr.X());
    const double   dy = std::abs(dr.Y());
    const double   dz = std::abs(dr.Z());

    //No undefined directions:
    if(std::isnan(dx) || 
       std::isnan(dy) || 
       std::isnan(dz)) return false;
    
    //At least one direction must be non-zero:
    if(dx == 0 && 
       dy == 0 && 
       dz == 0) {
      return false;
    }
    
    // Additional quality cuts. These remove tracks that often have
    // badly-reconstructed W positions.

    // Cut very steep tracks two ways. These can be misreconstructed, or
    // they can simply make it difficult for the W estimation to do a good
    // job.  OK to use the rb::Track::Dir() method here because the initial
    // direction of the track is what we want to cut on
    if(std::abs(track.Start().Z() - track.Stop().Z() )  <= 70 || 
       std::abs(track.Dir().Z()) <  0.2) return false;

    // Cut out actual reconstruction failures, where the track doesn't use
    // a significant amount of the slice hits.  If either view is not
    // complete enough discard the whole thing
    if(track.NXCell() < 0.8*slice.NXCell() || 
       track.NYCell() < 0.8*slice.NYCell()) return false;

    // Check that the average number of cells per plane is similar in 
    // each view.  If the value is very high in either view, or larger in 
    // one view than the other then you might have some form of FEB flash
    // or cosmic ray brem that will mess up the calibration
    std::set<unsigned short> xPlaneList;
    std::set<unsigned short> yPlaneList;
    for(size_t c = 0; c < track.NCell(); ++c){
      if     (track.Cell(c)->View() == geo::kX) xPlaneList.insert(track.Cell(c)->Plane());
      else if(track.Cell(c)->View() == geo::kY) yPlaneList.insert(track.Cell(c)->Plane());
    }// end loop over cells in track
    double xCells  = track.NXCell();
    double yCells  = track.NYCell();
    double xPlanes = 1. * xPlaneList.size();
    double yPlanes = 1. * yPlaneList.size();
    double cellsPerPlaneX = (xPlanes > 0) ? xCells/xPlanes : 999.;
    double cellsPerPlaneY = (yPlanes > 0) ? yCells/yPlanes : 999.;
    if( std::max(cellsPerPlaneX, cellsPerPlaneY) > fMaxCellsPerPlane ) return false;

    // remove tracks where there is an asymmetry between the number 
    // of planes hit in each view.  One class of asymmetry is where 
    // there are extra planes off either end of the track in one view.
    // Since there is no information in the other view for those planes
    // you can't be sure you got the length/containment correct
    double xStart = track.MaxPlane(geo::kX);
    double yStart = track.MaxPlane(geo::kY);
    double xEnd   = track.MinPlane(geo::kX);
    double yEnd   = track.MinPlane(geo::kY);
    if( std::abs(xStart - yStart)   > fMaxDeltaEndPlane ||
	std::abs(xEnd   - yEnd  )   > fMaxDeltaEndPlane ||
	std::abs(xPlanes - yPlanes) > fMaxPlaneAsymmetry*(xPlanes + yPlanes) ) return false;
    

    // Check that the track doesn't take one or two big steps
    // compared to the rest, that is usually a reconstruction failure.
    std::vector<float> steps;
    float median = 0.;
    if( !this->GoodSteps(track, steps, median) ) return false;

    // Remove tracks whose vertex position is well away from the edge of the
    // detector. Muon catcher is not counted as part of the detector proper in
    // ND so have to check for that as well.
    if( std::abs(this->DistToEdge(track.Start())) > fMaxVtxDistFromEdge) 
      return false;
    

    // Remove tracks whose end position is more than fMaxEndDistFromEdge
    // outside of the detector The minus sign is because LiveGeometry returns a
    // negative value if the point is outside of the active detector.  A track
    // can stop anywhere inside the detector and still be OK.
    if( this->DistToEdge(track.Stop()) < -fMaxEndDistFromEdge)
      return false;

    return true;
  }

  //----------------------------------------------------------------------------
  float CosmicTrackSelection::DistToEdge(TVector3 const& point)
  {
    return std::min(fLiveGeom->DistToEdgeXY(point), fLiveGeom->DistToEdgeZ(point));
  }

} // namespace trk
