//////////////////////////////////////////////////////////////////////////////
// \file    KalmanGeoHelper.cxx
// \brief   A class of useful track geometry functions for the KalmanTrack algorithm
// \version $Id: KalmanGeoHelper.cxx,v 1.1 2012-11-16 16:57:10 raddatz Exp $
// \author  Nick Raddatz
////////////////////////////////////////////////////////////////////////////
#include "TrackFit/KalmanGeoHelper.h"
#include "GeometryObjects/Geo.h"
#include "GeometryObjects/CellGeo.h"
#include "NovaDAQConventions/DAQConventions.h"
#include "TFile.h"

namespace trk
{
  //......................................................................
  KalmanGeoHelper::KalmanGeoHelper()
  {
    fBadChannels = true;
    SetupHelper();
  }

  //......................................................................
  KalmanGeoHelper::KalmanGeoHelper(bool badChan)
  {
    fBadChannels = badChan;
    SetupHelper();
  }

  //......................................................................
  KalmanGeoHelper::~KalmanGeoHelper()
  {
  }

  //......................................................................
  void KalmanGeoHelper::SetupHelper()
  {
    fnmaxplane = fgeom->NPlanes();
    if(fgeom->DetId() == novadaq::cnv::kFARDET){
      ndos = false;
      nd = false;
      fd = true;
    }
    else if(fgeom->DetId() == novadaq::cnv::kNEARDET){
      ndos = false;
      nd = true;
      fd = false;
    }
    else if(fgeom->DetId() == novadaq::cnv::kNDOS){
      ndos = true;
      nd = false;
      fd = false;
    }
    else{
      std::cerr<<"This geometry is not ndos, nd, or fd. This object doesn't know what to do"<<std::endl;
    }
    if(fd){
      fclipperLength = 1.5;
      fwlow = -700;
      fwhigh = 700;
      flengthmax = 6.0;
      effpar0 = TF1("par0","pol4",fwlow,fwhigh);
      effpar1 = TF1("par0","pol4",fwlow,fwhigh);
      effpar2 = TF1("par0","pol5",fwlow,fwhigh);
      effpar3 = TF1("par0","pol4",fwlow,fwhigh);

      // low threshold tuning
      effpar0.SetParameters(0.204074,-2.45217E-6,-3.61656E-7,4.38289E-11,4.3173E-13);
      effpar1.SetParameters(0.126228,3.77528E-4,7.85744E-7,-4.65944E-10,-1.0345E-12);
      effpar2.SetParameters(0.327222,-5.80372E-4,7.42265E-7,1.27296E-9,-1.13353E-12,-1.13839E-15);
      effpar3.SetParameters(1.35419E-2,-8.28103E-5,8.92098E-8,5.64673E-11,-6.05079E-14);

    }
    else if(nd){
      fclipperLength = 1.5;
      fwlow = -200;
      fwhigh = 200;
      flengthmax = 6.0;
      effpar0 = TF1("par0","pol4",fwlow,fwhigh);
      effpar1 = TF1("par0","pol4",fwlow,fwhigh);
      effpar2 = TF1("par0","pol5",fwlow,fwhigh);
      effpar3 = TF1("par0","pol4",fwlow,fwhigh);

      // low threshold tuning
      effpar0.SetParameters(0.139861,-5.05093E-5,2.66785E-7,-3.65295E-10,-1.46979E-11);
      effpar1.SetParameters(0.525903,2.00583E-4,-4.07681E-7,5.28495E-10,2.01945E-11);
      effpar2.SetParameters(0.0239177,-6.04344E-5,-1.08969E-7,-3.69116E-10,2.01116E-11,5.79678E-14);
      effpar3.SetParameters(-1.03056E-3,1.7984E-5,-3.04568E-7,-7.85076E-10,3.55217E-12);

    }
    else if(ndos){
      fclipperLength = 1.5;
    }
  }

  //......................................................................
  double KalmanGeoHelper::BestTrackPoint(double z, double v, double halfz, double halfv, std::vector<double> & bestPos,
			double z0,double v0, double dz, double dv)
  {
    // bestPos is a vector holding the best possible track hit position in the cell: bestPos[0] = z bestPos[1] = x/y.
    // output vector the minimum distance between the projected track and the best track hit position in the cell.
    // This function will take the initial state and z position and try to find the most compatible track hit point with the given hit
    
    double slope = dv/dz;
    // Get the projected track at the cell boundaries.
    double zlo = z - halfz;
    double zhi = z + halfz;
    double vlo = v - halfv;
    double vhi = v + halfv;
    double v1 = slope*(zlo-z0)+v0;
    double v2 = slope*(zhi-z0)+v0;
    double z1 = (vlo-v0)/slope+z0;
    double z2 = (vhi-v0)/slope+z0;
    double deltaz = 0;
    double deltav = 0;
    // Make a decision as to how to look for best track point as to avoid dividing by zero
    if(dz == 0){
      // either the z is in the z range or else it never will be, either case best position is z0, v
      bestPos[0] = z0;
      bestPos[1] = v;
      // Increment deltav accordingly if actually in the cell
      if(z >= zlo && z <= zhi){ deltav = vhi - vlo; }
    } // end of if dz == 0
    else{
      // Go through the various cases to see if the track projection intersects the cell hit or not
      if(slope>0){
	// check if track enters cell from left side
	if(v1 >= vlo && v1 <= vhi){
	  // check if track exits on the right side of the cell
	  if(v2 <= vhi){
	    bestPos[0] = z;
	    bestPos[1] = 0.5*v1+0.5*v2;
	    deltaz = zhi-zlo;
	    deltav = v2-v1;
	  }
	  else{
	    // track exits through the top of the cell
	    bestPos[0] = 0.5*zlo+0.5*z2;
	    bestPos[1] = 0.5*v1 + 0.5*vhi;
	    deltaz = z2-zlo;
	    deltav = vhi-v1;
	  }
	}
	// check if track enters cell from bottom side
	else if(v1 < vlo && v2 >= vlo){
	  // check if track exits on right side of cell
	  if(v2 <= vhi){
	    bestPos[0] = 0.5*z1+0.5*zhi;
	    bestPos[1] = 0.5*vlo+0.5*v2;
	    deltaz = zhi-z1;
	    deltav = v2-vlo;
	  }
	  else{
	    // track exits through top of the cell
	    bestPos[0] = 0.5*z1 + 0.5*z2;
	    bestPos[1] = v;
	    deltaz = z2-z1;
	    deltav = vhi-vlo;
	  }
	}
	else{
	  // track does not go through cell
	  // check if track is above or below cell
	  if(v1 < vlo){
	    // closest point is bottom right corner of the cell
	    bestPos[0] = zhi;
	    bestPos[1] = vlo;
	  }
	  else{
	    // closest point is top left corner of the cell
	    bestPos[0] = zlo;
	    bestPos[1] = vhi;
	  }
	}
      }
      else{
	// check if track enters cell from left side
	if(v1 >= vlo && v1<= vhi){
	  // check if the track exits on the right side of the cell
	  if(v2 >= vlo){
	    bestPos[0] = z;
	    bestPos[1] = 0.5*v1+0.5*v2;
	    deltaz = zhi-zlo;
	    deltav = v2-v1;
	  }
	  else{
	    // track exits through the bottom of the cell
	    bestPos[0] = 0.5*zlo+0.5*z1;
	    bestPos[1] = 0.5*v1+0.5*vlo;
	    deltaz = zlo-z1;
	    deltav = v1-vlo;
	  }
	}
	// check if the track enters the cell from the top side
	else if(v1 >= vhi && v2 <= vhi){
	  // check if the track exits on the right side of the cell
	  if(v2 >= vlo){
	    bestPos[0] = 0.5*z2+0.5*zhi;
	    bestPos[1] = 0.5*vhi+0.5*v2;
	    deltaz = z2-zhi;
	    deltav = vhi-v2;
	  }
	  else{
	    // track exits through the bottom of the cell
	    bestPos[0] = 0.5*z2+0.5*z1;
	    bestPos[1] = v;
	    deltaz = z2-z1;
	    deltav = vhi-vlo;
	  }
	}
	else{
	  // track does not go through the cell
	  // check if the track is above or below the cell
	  if(v1 > vhi){
	    // closest point is top right corner of the cell
	    bestPos[0] = zhi;
	    bestPos[1] = vhi;
	  }
	  else{
	    // closest point is the bottom left corner of the cell
	    bestPos[0] = zlo;
	    bestPos[1] = vlo;
	  }
	  // if slope is truly zero than pick midpoint of cell instead of corner points for best z position
	  if(slope == 0){ bestPos[0] = z; }
        }
      }
    } // end of else for if dz == 0
    return (deltaz*deltaz+deltav*deltav);
  }

  //......................................................................

  int   KalmanGeoHelper::CountMissedCellsOnLine(geo::View_t view,double vi, double zi, double vf, double zf, int maxCells) 
  {
    double fDetHalfWidth = fgeom->DetHalfWidth();
    double fDetHalfHeight = fgeom->DetHalfHeight();
    double fDetLength = fgeom->DetLength();
    int fplaneMax = fgeom->NPlanes();
    // if we are in the nd or ndos nead to worry about muon catcher which has different spacing between planes and distance
    if(nd || ndos){
      // take the ratio of number of planes to distance from the first 100 planes otherwise will get screwed up by the muon catcher
      fplaneMax = 100;
      double xyzp100[3];
      fgeom->Plane(100)->Cell(0)->GetCenter(xyzp100);
      fDetLength = xyzp100[2];
    }
    // Otherwise we would count when the ray was outside the detector in the  other view.
    TVector3 r1(vi,0,zi);
    TVector3 r2(vf,0,zf);
    if(view == geo::kY){
      r1.SetXYZ(0,vi,zi);
      r2.SetXYZ(0,vf,zf);
    }
    geo::ClampRayToDetectorVolume(&r1, &r2, &(*fgeom));
    int nummiss = 0;
    if(r1.Z() > r2.Z()) std::swap(r1, r2);

    double z1 = r1.Z();
    double z2 = r2.Z();

    const unsigned int planeMax = fgeom->NPlanes();

    // find the approximate start plane
    unsigned int iplane = 0;
    int pl = (int) z1/(fDetLength/((double)fplaneMax));
    if(pl < 0){iplane = 0; }
    else if (pl >= (int)planeMax){iplane = planeMax-1;}
    else iplane = pl;
    // check that our initial plane guess is correct
    double posp[3];
    fgeom->Plane(iplane)->Cell(0)->GetCenter(posp);
    double dpl = fgeom->Plane(iplane)->Cell(0)->HalfD();
    // ask if the current position is in this plane
    if(z1<(posp[2]-dpl)){
      bool foundInitPlane = false;
      // Did not get the right plane location. Check below if position is lower than the estimated plane center position
      // only need to worry about geusses that are too high in plane number otherwise accept the estimated guess as the start plane
      int currTestPlane = iplane-1;
      while(currTestPlane >= 0 && !foundInitPlane){
        fgeom->Plane(currTestPlane)->Cell(0)->GetCenter(posp);
        dpl = fgeom->Plane(currTestPlane)->Cell(0)->HalfD();
        if(z1>(posp[2]-dpl)){
          iplane = currTestPlane;
          foundInitPlane = true;
        }
        else{
          if(currTestPlane == 0){iplane = 0;}
          --currTestPlane;
        }
      }
    }
      
    const double v1 = r1[view];
    const double v2 = r2[view];
    
    const double denom = z2-z1;
    const double dvdz = denom ? (v2-v1)/denom : 0;

    if(denom == 0 && r2[view] < r1[view]){
      std::swap(r1, r2);
      z1 = r1.Z();
      z2 = r2.Z();
    }
    
    for(unsigned int plane = iplane; plane < planeMax; ++plane){
      const geo::PlaneGeo* geoplane = fgeom->Plane(plane);

      // Only look at the planes in the specified view
      if(geoplane->View() != view) continue;
      const geo::CellGeo* geocell0 = geoplane->Cell(0);
      // Assuming here perfect alignment and identical cells across plane
      double xyzp[3];
      geocell0->GetCenter(xyzp);
      const double zc = xyzp[2];
      const double dz = geocell0->HalfD();

      if(z1 > zc+dz) continue;
      // Plane positions increase monotonically
      if(z2 < zc-dz) break;

      // Position entering the cell or starting
      const double zin = std::max(zc-dz, z1);
      double vlo = v1+(zin-z1)*dvdz;
      // Position exiting or ending
      const double zout = std::min(zc+dz, z2);
      double vhi = v1+(zout-z1)*dvdz;

      // if slope is infinite set vlo and vhi as appropriately
      if(denom == 0){
	vlo = v1;
	vhi = v2;
      }

      // Ensure they're from low to high instead
      if(vlo > vhi) std::swap(vlo, vhi);

      //create an approximate map between position and cell number for cells in the xz view. 
      unsigned int icell = 0;
      int cell = (int) (vlo+fDetHalfWidth)/(2*fDetHalfWidth/((double) geoplane->Ncells()));
      int viewind = 0;
      if(view == geo::kY){
	cell = (int) (vlo+fDetHalfHeight)/(2*fDetHalfHeight/((double) geoplane->Ncells()));
	viewind = 1;
      }
      if(cell < 0){icell = 0; }
      else if (cell >= (int) geoplane->Ncells()){icell = geoplane->Ncells()-1;}
      else icell = cell;
      // check that our initial cell guess is correct
      double pos[3];
      geoplane->Cell(icell)->GetCenter(pos);
      double dp = geoplane->Cell(icell)->HalfW();
      if(vlo<(pos[viewind]-dp)){
        bool foundInitCell = false;
	// Did not get the right cell location. Check below if position is lower than the estimated cell center position
        int currTestCell = icell-1;
        while(currTestCell >= 0 && !foundInitCell){
          geoplane->Cell(currTestCell)->GetCenter(pos);
          dp = geoplane->Cell(currTestCell)->HalfW();
          // ask if the current position is in this cell
          if( vlo>(pos[viewind]-dp)){
            icell = currTestCell;
            foundInitCell = true;
          }
          else{
            if(currTestCell == 0){icell = 0;}
            --currTestCell;
          }
        }
      }

      for(unsigned int jcell = icell; jcell < geoplane->Ncells();++jcell){
        double xyz[3];
        geoplane->Cell(jcell)->GetCenter(xyz);
        const double vc = xyz[view];
        const double dv = geoplane->Cell(jcell)->HalfW();

        if(vlo > vc+dv) continue;
        // Cell positions increase monotonically
        if(vhi < vc-dv) break;
        if(fBadChannels){
          // check if this is a bad channel
          if(!fbadc->IsBad(plane,jcell) && !fbadc->IsLowEfficiency(plane,jcell)){ ++nummiss; }
        }
        else{ ++nummiss; }
        if(nummiss > maxCells) return nummiss;

      } // end for jcell
    } // end for plane
    return nummiss;
  } // end of CountMissedCellsOnLine function


  //......................................................................
  int   KalmanGeoHelper::CountMissedCellsOnLine(geo::View_t view,double vi, double zi, double vf, double zf,std::vector<geo::OfflineChan> &hitsonline, std::vector<double> &distance) 
  {
    double fDetHalfWidth = fgeom->DetHalfWidth();
    double fDetHalfHeight = fgeom->DetHalfHeight();
    double fDetLength = fgeom->DetLength();
    int fplaneMax = fgeom->NPlanes();
    // if we are in the nd or ndos nead to worry about muon catcher which has different spacing between planes and distance
    if(nd || ndos){
      // take the ratio of number of planes to distance from the first 100 planes otherwise muon catcher will screw this up
      fplaneMax = 100;
      double xyzp100[3];
      fgeom->Plane(100)->Cell(0)->GetCenter(xyzp100);
      fDetLength = xyzp100[2];
    }
    // Otherwise we would count when the ray was outside the detector in the  other view.
    TVector3 r1(vi,0,zi);
    TVector3 r2(vf,0,zf);
    if(view == geo::kY){
      r1.SetXYZ(0,vi,zi);
      r2.SetXYZ(0,vf,zf);
    }
    geo::ClampRayToDetectorVolume(&r1, &r2, &(*fgeom));
    int nummiss = 0;
    if(r1.Z() > r2.Z()) std::swap(r1, r2);

    double z1 = r1.Z();
    double z2 = r2.Z();

    const unsigned int planeMax = fgeom->NPlanes();

    // find the approximate start plane
    unsigned int iplane = 0;
    int pl = (int) z1/(fDetLength/((double)fplaneMax));
    if(pl < 0){iplane = 0; }
    else if (pl >= (int)planeMax){iplane = planeMax-1;}
    else iplane = pl;
    // check that our initial plane guess is correct
    double posp[3];
    fgeom->Plane(iplane)->Cell(0)->GetCenter(posp);
    double dpl = fgeom->Plane(iplane)->Cell(0)->HalfD();
    // ask if the current position is in this plane
    if(z1<(posp[2]-dpl)){
      bool foundInitPlane = false;
      // Did not get the right plane location. Check below if position is lower than the estimated plane center position
      // only need to worry about geusses that are too high in plane number otherwise accept the estimated guess as the start plane
      int currTestPlane = iplane-1;
      while(currTestPlane >= 0 && !foundInitPlane){
        fgeom->Plane(currTestPlane)->Cell(0)->GetCenter(posp);
        dpl = fgeom->Plane(currTestPlane)->Cell(0)->HalfD();
        if(z1>(posp[2]-dpl)){
          iplane = currTestPlane;
          foundInitPlane = true;
        }
        else{
          if(currTestPlane == 0){iplane = 0;}
          --currTestPlane;
        }
      }
    }

    const double v1 = r1[view];
    const double v2 = r2[view];
    
    const double denom = z2-z1;
    const double dvdz = denom ? (v2-v1)/denom : 0;

    if(denom == 0 && r2[view] < r1[view]){
      std::swap(r1, r2);
      z1 = r1.Z();
      z2 = r2.Z();
    }
 
    for(unsigned int plane = iplane; plane < planeMax; ++plane){
      const geo::PlaneGeo* geoplane = fgeom->Plane(plane);

      // Only look at the planes in the specified view
      if(geoplane->View() != view) continue;
      const geo::CellGeo* geocell0 = geoplane->Cell(0);
      // Assuming here perfect alignment and identical cells across plane
      double xyzp[3];
      geocell0->GetCenter(xyzp);
      const double zc = xyzp[2];
      const double dz = geocell0->HalfD();

      if(z1 > zc+dz) continue;
      // Plane positions increase monotonically
      if(z2 < zc-dz) break;

      // Position entering the cell or starting
      const double zin = std::max(zc-dz, z1);
      double vlo = v1+(zin-z1)*dvdz;
      // Position exiting or ending
      const double zout = std::min(zc+dz, z2);
      double vhi = v1+(zout-z1)*dvdz;

      // if slope is infinite set vlo and vhi as appropriately
      if(denom == 0){
	vlo = v1;
	vhi = v2;
      }

      // Ensure they're from low to high instead
      if(vlo > vhi) std::swap(vlo, vhi);

      //create an approximate map between position and cell number for cells in the xz view. 
      unsigned int icell = 0;
      int cell = (int) (vlo+fDetHalfWidth)/(2*fDetHalfWidth/((double) geoplane->Ncells()));
      int viewind = 0;
      if(view == geo::kY){
	cell = (int) (vlo+fDetHalfHeight)/(2*fDetHalfHeight/((double) geoplane->Ncells()));
	viewind = 1;
      }
      if(cell < 0){icell = 0; }
      else if(cell >= (int) geoplane->Ncells()){icell = geoplane->Ncells()-1;}
      else{ icell = cell; }
      // check that our initial cell guess is correct
      double pos[3];
      geoplane->Cell(icell)->GetCenter(pos);
      double dp = geoplane->Cell(icell)->HalfW();
      if(vlo<(pos[viewind]-dp)){
        bool foundInitCell = false;
	// Did not get the right cell location. Check below if position is lower than the estimated cell center position
        int currTestCell = icell-1;
        while(currTestCell >= 0 && !foundInitCell){
          geoplane->Cell(currTestCell)->GetCenter(pos);
          dp = geoplane->Cell(currTestCell)->HalfW();
          // ask if the current position is in this cell
          if( vlo>(pos[viewind]-dp)){
            icell = currTestCell;
            foundInitCell = true;
          }
          else{
            if(currTestCell == 0){icell = 0;}
            --currTestCell;
          }
        }
      }

      for(unsigned int jcell = icell; jcell < geoplane->Ncells();++jcell){
        double xyz[3];
        geoplane->Cell(jcell)->GetCenter(xyz);
        const double vc = xyz[view];
        const double dv = geoplane->Cell(jcell)->HalfW();

        if(vlo > vc+dv) continue;
        // Cell positions increase monotonically
        if(vhi < vc-dv) break;
	// get the distance in the cell
	std::vector<double> bestPos(2);
	double dist2 = BestTrackPoint(xyz[2],vc,dz,dv,bestPos,z1,v1,z2-z1,v2-v1);
	++nummiss;
	hitsonline.push_back(geo::OfflineChan(plane,jcell));
	distance.push_back(sqrt(dist2));
      } // end for jcell
    } // end for plane
    return nummiss;
  } // end of CountMissedCellsOnLine function 

  //......................................................................
  double   KalmanGeoHelper::CalcMissFrac(double dist, double w, geo::View_t view)
  {
    // if distance in cell is less clipper length, clipper hit return 1.0, unreliable hit efficiency 
    if(dist < fclipperLength){ return 1.0; }

    // define the length in cell to use
    double length = dist;
    if(length > flengthmax){ length = flengthmax; }
    // define what cell depth to use
    double depth = w;
    if(depth < fwlow){ depth = fwlow; }
    if(depth > fwhigh){ depth = fwhigh; }

    double effhit = Efficiency(depth,length);

    if(effhit > 1){ effhit = 1.0; }
    else if(effhit<0){ effhit = 0.0; }
    return 1.0-effhit;

  } // end of CalcMissFrac


  //......................................................................
  double   KalmanGeoHelper::Efficiency(double w, double length)
  {
    double eff = 0.0;

    if(fd){
      double llow = 3.4;
      double lhigh = 3.75;
      
      if( length < llow){
	eff = effpar0.Eval(w)*length + effpar1.Eval(w);
      }
      else if( length < lhigh){
	eff = effpar2.Eval(w)*length + (effpar0.Eval(w)-effpar2.Eval(w))*llow + effpar1.Eval(w);
      }
      else{
	eff = effpar3.Eval(w)*length + (effpar2.Eval(w)-effpar3.Eval(w))*lhigh + 
	  (effpar0.Eval(w)-effpar2.Eval(w))*llow + effpar1.Eval(w);
      }
    } 
    else if(nd){
      double llow = 3.0;
      double lhigh = 3.75;
      
      if( length < llow){
	eff = effpar0.Eval(w)*length + effpar1.Eval(w);
      }
      else if( length < lhigh){
	eff = effpar2.Eval(w)*length + (effpar0.Eval(w)-effpar2.Eval(w))*llow + effpar1.Eval(w);
      }
      else{
	eff = effpar3.Eval(w)*length + (effpar2.Eval(w)-effpar3.Eval(w))*lhigh + 
	  (effpar0.Eval(w)-effpar2.Eval(w))*llow + effpar1.Eval(w);
      }
    }
    else{ eff = 1.0; }
    // unless clipper, ndos shouldn't miss hits (need to check for bad channel definition of efficiency/badchannels outside of this function)

    return eff;
  }


  //......................................................................
  int  KalmanGeoHelper::MatchTrajectoryToPlane(TVector3 traj, int minplane, int maxplane)
  {
    int plane = minplane;
    int mintrialplane = minplane;
    int maxtrialplane = maxplane;
    if(minplane < 0){ mintrialplane = 0; }
    if(maxplane > fnmaxplane-1){ maxtrialplane = fnmaxplane-1; }

    // Try finding what plane/cell this point is in
    const geo::CellUniqueId cid = fgeom->CellId(traj.X(),traj.Y(),traj.Z(),1,1,0.1,2);
    
    // if found the cell unique id use it; otherwise backout plane from individual plane width checks
    if(cid){ fgeom->IdToPlane(cid, &plane); }
    else{
      for(int ip = maxtrialplane; ip >= mintrialplane; --ip){
	int trialPlane = ip;

	const geo::PlaneGeo* geoplane = fgeom->Plane(trialPlane);
	// get the estimated cell in this plane
	int mintrialcell = 0;
	int maxtrialcell = geoplane->Ncells();
	double v = traj.X();
	int trialCell = (int) (v+fgeom->DetHalfWidth())/(2*fgeom->DetHalfWidth()/((double) (geoplane->Ncells()-1)));
	int viewind = 0;
	if(geoplane->View() == geo::kY){
	  v = traj.Y();
	  trialCell = (int) (v+fgeom->DetHalfHeight())/(2*fgeom->DetHalfHeight()/((double) (geoplane->Ncells()-1)));
	  viewind = 1;
	}

	if(trialCell < 0){trialCell = 0; }
	else if (trialCell >= (int) geoplane->Ncells()){trialCell = geoplane->Ncells()-1;}

	// check that our initial cell guess is correct
	double trialCenter[3];
	geoplane->Cell(trialCell)->GetCenter(trialCenter);
	double dp = geoplane->Cell(trialCell)->HalfW();

	if(v < (trialCenter[viewind]-dp)){
	  // trial cell is too high, try iterating down to cell 0 until find right cell
	  maxtrialcell = trialCell;
	  for(int ic = maxtrialcell; ic >= mintrialcell; --ic){
	    fgeom->Plane(trialPlane)->Cell(ic)->GetCenter(trialCenter);
	    dp = geoplane->Cell(ic)->HalfW();
	    trialCell = ic;
	    if(v > (trialCenter[viewind]-dp)){ break; }
	  }
	}
	else if(v > (trialCenter[viewind]+dp)){
	  // trial cell is too low, try iterating up to Ncells in plane until find right cell
	  mintrialcell = trialCell;
	  for(int ic = mintrialcell; ic <= maxtrialcell; ++ic){
	    fgeom->Plane(trialPlane)->Cell(ic)->GetCenter(trialCenter);
	    dp = geoplane->Cell(ic)->HalfW();
	    trialCell = ic;
	    if(v > (trialCenter[viewind]-dp)){ break; }
	  }
	}

	fgeom->Plane(trialPlane)->Cell(trialCell)->GetCenter(trialCenter);
	double trialPlaneWidth = fgeom->Plane(trialPlane)->Cell(trialCell)->HalfD();
	// starting from max plane first so if the z is larger than the smallest z 
	// on the plane then this is the correct plane
	if(traj.Z() > (trialCenter[2]-trialPlaneWidth)){
	  plane = trialPlane;
	  break;
	}
      }
    }

    return plane;
  } // end of MatchTrajectoryToPlane

  //......................................................................
  int  KalmanGeoHelper::MatchTrajectoryToPlaneInView(TVector3 traj, geo::View_t view, int minplane, int maxplane)
  {

    assert(fgeom->Plane(minplane)->View() == view);
    assert(fgeom->Plane(maxplane)->View() == view);
    int plane = MatchTrajectoryToPlane(traj,minplane,maxplane);
    geo::View_t trialview = fgeom->Plane(plane)->View();
    if(view == geo::kXorY || trialview == view){ return plane; }
    else{
      int lowtrialplane = fgeom->NextPlaneOtherView(plane,-1);
      int hightrialplane = fgeom->NextPlaneOtherView(plane,+1);
      if(lowtrialplane == geo::kPLANE_NOT_FOUND){
	if(hightrialplane == geo::kPLANE_NOT_FOUND){ return minplane; }
	else{ return hightrialplane; }
      }
      else if(hightrialplane == geo::kPLANE_NOT_FOUND){ 
	if(lowtrialplane >= minplane){ return lowtrialplane; }
	else{ return minplane; }
      }
      else{
	// check whether the high or low plane is the closest in z
	double lowplanecenter[3];
	fgeom->Plane(lowtrialplane)->Cell(0)->GetCenter(lowplanecenter);
	double lowplanewidth = fgeom->Plane(lowtrialplane)->Cell(0)->HalfD();
	double highplanecenter[3];
	fgeom->Plane(hightrialplane)->Cell(0)->GetCenter(highplanecenter);
	double highplanewidth = fgeom->Plane(hightrialplane)->Cell(0)->HalfD();
	if(fabs(lowplanecenter[2]+lowplanewidth - traj.Z()) <=
	   fabs(highplanecenter[2]-highplanewidth - traj.Z())){ return lowtrialplane; }
	else{ return hightrialplane; }
      }
    }
    return minplane;

  } // end of MatchTrajectoryToPlaneInView

} // end of trk namespace
