//////////////////////////////////////////////////////////////////////////////
// \file    KalmanTrackAna.cxx
// \brief   Analysis module to make validation plots for reconstructed tracks
// \version $Id: KalmanTrackAna_module.cc,v 1.2 2012/12/10 21:28:53 raddatz Exp $
// \author  Nick Raddatz
////////////////////////////////////////////////////////////////////////////
#include <string>
#include <vector>

// ROOT includes
#include "TH1D.h"
#include "TH2D.h"
#include "TVector3.h"
#include "TTree.h"

// Framework includes
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Core/EDAnalyzer.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Services/Optional/TFileService.h"
#include "art/Framework/Core/FindManyP.h"
#include "messagefacility/MessageLogger/MessageLogger.h"

// NOvA includes
#include "Calibrator/Calibrator.h"
#include "Geometry/Geometry.h"
#include "GeometryObjects/CellGeo.h"
#include "RecoBase/CellHit.h"
#include "RecoBase/Cluster.h"
#include "RecoBase/Track.h"
#include "RecoBase/RecoHit.h"
#include "Simulation/Simulation.h"
#include "NovaDAQConventions/DAQConventions.h"
#include "MCCheater/BackTracker.h"
#include "RecoBase/FilterList.h"

namespace trk {
  // A module to analyze tracks produced by KalmanTrack module
  class KalmanTrackAna : public art::EDAnalyzer {
  public:
    
    explicit KalmanTrackAna(fhicl::ParameterSet const& pset);
    virtual ~KalmanTrackAna();
    void reconfigure(const fhicl::ParameterSet& pset);

    void beginJob();
    void analyze(art::Event const& evt);

  private:
    void CheckRecoTracks(std::vector< art::Ptr<rb::Track> > const& trkcol,
			 art::PtrVector<rb::CellHit> const& allhits);
    std::set<int> FindVisibleProngs(const art::PtrVector<rb::CellHit>& allhits,
				    std::set<int> ids) const;
    double TotalLengthInDetector(const sim::Particle* p,
				 geo::View_t view) const;
    double Containment(TVector3 start) const;

    void CheckAssociations(art::Ptr<rb::Cluster> slice,std::vector<art::Ptr<rb::Track> > tracks, bool filtered);


    std::string fSliceLabel;         ///<Where to find Slices to reconstruct
    std::string fTrackLabel;         ///<Where to find Tracks to reconstruct

    // Individual Track Level Histograms
    TH1D* fTrackLength;              ///< Length of reconstructed 3d tracks
    TH1D* fHitsPerTrack;             ///< Number of hits in each reconstructed 3d track
    TH1D* fTrackLength2d;            ///< Length of reconstructed 3d tracks
    TH1D* fHitsPerTrack2d;           ///< Number of hits in each reconstructed 3d track
    TH1D* fCosX;                     ///< Cosine theta of 3d tracks with respect to the X axis
    TH1D* fCosY;                     ///< Cosine theta of 3D tracks with respect ot the Y axis
    TH1D* fCosZ;                     ///< Cosine theta of 3d tracks with respect to the Z axis
    TH1D* fCosX2d;                   ///< Cosine theta of 2d tracks with respect to the X axis
    TH1D* fCosY2d;                   ///< Cosine theta of 2D tracks with respect ot the Y axis
    TH1D* fCosZ2d;                   ///< Cosine theta of 2d tracks with respect to the Z axis
    TH1D* fXPosStart;                ///< x start position of tracks
    TH1D* fYPosStart;                ///< y start position of tracks
    TH1D* fZPosStart;                ///< z start position of tracks
    TH1D* fXPosEnd;                  ///< x end position of tracks
    TH1D* fYPosEnd;                  ///< y end position of tracks
    TH1D* fZPosEnd;                  ///< z end position of tracks
    TH1D* fXPos;                     ///< position of x track hits
    TH1D* fYPos;                     ///< position of y track hits
    TH1D* fXRecoPos;                 ///< reconstructed x position from yz track hit
    TH1D* fYRecoPos;                 ///< reconstructed y position from xz track hit
    TH2D* fXvYStart;                 ///< Position map for track start position
    TH2D* fXvYEnd;                   ///< Position map for track end position
    TH2D* fXvYReco;                  ///< X vs Y positions of track hits
    TH2D* fZvXReco;                  ///< Z vs X of reconstructed track hit positions
    TH2D* fZvYReco;                  ///< Z vs Y of reconstructed track hit positions
    TH2D* fZvX;                      ///< Z vs X of track hit positions
    TH2D* fZvY;                      ///< Z vs Y of track hit positions

    // The following histograms are for CAF validation purposes
    TH1D* fCosXCAF;                  ///< Cosine theta of tracks with respect to the X axis for CAF comparison
    TH1D* fCosYCAF;                  ///< Cosine theta of tracks with respect ot the Y axis for CAF comparison
    TH1D* fCosZCAF;                  ///< Cosine theta of tracks with respect to the Z axis for CAF comparison
    TH1D* fViewCAF;                  ///< View stored in CAF
    TH1D* fLengthCAF;                ///< Length of reconstructed tracks stored in CAF
    TH1D* fHitsCAF;                  ///< Number of hits in each track stored in CAF
    TH1D* fXPosStartCAF;             ///< x start position of tracks in CAF
    TH1D* fYPosStartCAF;             ///< y start position of tracks in CAF
    TH1D* fZPosStartCAF;             ///< z start position of tracks in CAF
    TH1D* fXPosEndCAF;               ///< x end position of tracks in CAF
    TH1D* fYPosEndCAF;               ///< y end position of tracks in CAF
    TH1D* fZPosEndCAF;               ///< z end position of tracks in CAF

    // Event Level Histograms
    TH1D* fTracksPerEvent;           ///< number of 3d tracks found per event
    TH1D* f2DTracksPerEvent;         ///< number of 2d tracks per event
    TH1D* fTracksPerSlice;           ///< number of 3d tracks found per non noise slice
    TH1D* f2DTracksPerSlice;         ///< number of 2d tracks per non noise slice


    // Track Hit Level Histograms
    TH1D* fXHitsPerPlane;            ///< number of hits found in each x plane
    TH1D* fYHitsPerPlane;            ///< number of hits found in each y plane
    TH1D* fHitsPerPlane;             ///< number of hits found in each plane
    TH1D* fXHitsPerCell;             ///< number of hits found in each x cell
    TH1D* fYHitsPerCell;             ///< number of hits found in each y cell
    TH2D* fXPlaneVCell;              ///< Plane vs Cell detector map for x view
    TH2D* fYPlaneVCell;              ///< Plane vs Cell detector map for y view
    TH2D* fPlaneVCell;               ///< Plane vs Cell map for whole detector

    // Tree holding truth information for determining things like completeness, purity, etc... of muons
    TTree*    fTruthTree;

    int fpdg;
    float fE, fp, fvtxdst, feff, fpur, flen, fx, fy, fz, fdx, fdy, fdz, fenu;
    int fmode;
    int fccnc;
    float fweight;
    int fevent;
    int fpassMuon;
    int fnhits;
    int fview;
    float frecolen;

    struct Counts
    {
      Counts()
      {
	nHits[0] = nHits[1] = 0;
	nOnlyHits[0] = nOnlyHits[1] = 0;
      }
      
      int nHits[2]; // Indexed by view
      int nOnlyHits[2]; // Hit by only one particle
    };

  };

  //......................................................................

  KalmanTrackAna::KalmanTrackAna(fhicl::ParameterSet const& pset):
    EDAnalyzer(pset)
  {
    reconfigure(pset); 
  }

  //......................................................................

  KalmanTrackAna::~KalmanTrackAna()
  {
  }

  //.....................................................................
  void KalmanTrackAna::reconfigure(const fhicl::ParameterSet& pset)  {
    fSliceLabel        = (pset.get< std::string >("SliceLabel")         );
    fTrackLabel        = (pset.get< std::string >("TrackLabel")         );
  }

  //......................................................................

  void KalmanTrackAna::beginJob()
  {

    double maxLength = 3000.0;
    int  maxHits = 500;
    double xMax = 1200.0;
    double yMax = 1200.0;
    double zMax = 6500.0;
    double nCell = 384;
    double nPlane = 960;
    int xbins = 480;
    int ybins = 480;
    int zbins = 1300;

    art::ServiceHandle<art::TFileService> tfs;
    // Individual Track Level Histograms
    fTrackLength                = tfs->make<TH1D>("trackLength",";Track Length (cm);NTracks",600,0.0,maxLength);
    fHitsPerTrack               = tfs->make<TH1D>("hitsPerTrack",";NHits; NTracks", maxHits+1,-0.5,(double)maxHits+0.5);
    fTrackLength2d              = tfs->make<TH1D>("trackLength2d",";Track Length (cm);NTracks",600,0.0,maxLength);
    fHitsPerTrack2d             = tfs->make<TH1D>("hitsPerTrack2d",";NHits; NTracks", maxHits+1,-0.5,(double)maxHits+0.5);
    fCosX                       = tfs->make<TH1D>("cosX",";Cos #theta_{X}", 101,-1.0,1.0);
    fCosY                       = tfs->make<TH1D>("cosY",";Cos #theta_{Y}", 101,-1.0,1.0);
    fCosZ                       = tfs->make<TH1D>("cosZ",";Cos #theta_{Z}",100,-1.0,1.0);
    fCosX2d                     = tfs->make<TH1D>("cosX2d",";Cos #theta_{X}", 101,-1.0,1.0);
    fCosY2d                     = tfs->make<TH1D>("cosY2d",";Cos #theta_{Y}", 101,-1.0,1.0);
    fCosZ2d                     = tfs->make<TH1D>("cosZ2d",";Cos #theta_{Z}",100,-1.0,1.0);

    fXPosStart                  = tfs->make<TH1D>("xPosStart",";X Start Position (cm); NTracks",xbins,-xMax,xMax);
    fYPosStart                  = tfs->make<TH1D>("yPosStart",";Y Start Position (cm); NTracks",ybins,-yMax,yMax);
    fZPosStart                  = tfs->make<TH1D>("zPosStart",";Z Start Position (cm); NTracks",zbins,0.0,zMax);
    fXPosEnd                    = tfs->make<TH1D>("xPosEnd",";X End Position (cm); NTracks",xbins,-xMax,xMax);
    fYPosEnd                    = tfs->make<TH1D>("yPosEnd",";Y End Position (cm); NTracks",ybins,-yMax,yMax);
    fZPosEnd                    = tfs->make<TH1D>("zPosEnd",";Z End Position (cm); NTracks",zbins,0.0,zMax);
    fXvYStart                   = tfs->make<TH2D>("XvYStart",";X Start Position (cm); Y Start Position (cm)",xbins,-xMax,xMax,ybins,-yMax,yMax );
    fXvYEnd                     = tfs->make<TH2D>("XvYEnd",";X End Position (cm); Y End Position (cm)",xbins,-xMax,xMax,ybins,-yMax ,yMax);
    fXvYReco                    = tfs->make<TH2D>("XvYReco",";X Position (cm); Y Position (cm)",xbins,-xMax,xMax,ybins,-yMax,yMax);
    fZvXReco                    = tfs->make<TH2D>("ZvXReco",";Z Position (cm); X Position (cm)",zbins,-200.0,zMax,xbins,-xMax,xMax);
    fZvYReco                    = tfs->make<TH2D>("ZvYReco",";Z Position (cm); Y Position (cm)",zbins,-200.0,zMax,ybins,-yMax,yMax);
    fZvX                        = tfs->make<TH2D>("ZvX",";Z Position (cm); X Position (cm)",zbins,-200.0,zMax,xbins,-xMax,xMax);
    fZvY                        = tfs->make<TH2D>("ZvY",";Z Position (cm); Y Position (cm)",zbins,-200.0,zMax,ybins,-yMax,yMax);


    int ndirbins = 100;
    double dirmin = -1.1;
    double dirmax = 1.1;
    int xbinscaf = 100;
    int ybinscaf = 100;
    int zbinscaf = 100;
    double xmaxcaf = 1000.0;
    double ymaxcaf = 1000.0;
    double zmaxcaf = 6000.0;
    int nlenbins = 100;
    double nlenmin = -0.5;
    double nlenmax = 5500;
    int nhitsbins = 100;
    double nhitmin = -0.5;
    double nhitmax = 800.0;

    fCosXCAF                    = tfs->make<TH1D>("dirXCAF",";Cos #theta_{X}", ndirbins,dirmin,dirmax);
    fCosYCAF                    = tfs->make<TH1D>("dirYCAF",";Cos #theta_{Y}", ndirbins,dirmin,dirmax);
    fCosZCAF                    = tfs->make<TH1D>("dirZCAF",";Cos #theta_{Z}",ndirbins,dirmin,dirmax);
    fViewCAF                    = tfs->make<TH1D>("viewCAF",";view;",3,-0.5,2.5);
    fXPosStartCAF               = tfs->make<TH1D>("startXCAF",";X Start Position (cm); NTracks",xbinscaf,-xmaxcaf,xmaxcaf);
    fYPosStartCAF               = tfs->make<TH1D>("startYCAF",";Y Start Position (cm); NTracks",ybinscaf,-ymaxcaf,ymaxcaf);
    fZPosStartCAF               = tfs->make<TH1D>("startZCAF",";Z Start Position (cm); NTracks",zbinscaf,0.0,zmaxcaf);
    fXPosEndCAF                 = tfs->make<TH1D>("stopXCAF",";X End Position (cm); NTracks",xbinscaf,-xmaxcaf,xmaxcaf);
    fYPosEndCAF                 = tfs->make<TH1D>("stopYCAF",";Y End Position (cm); NTracks",ybinscaf,-ymaxcaf,ymaxcaf);
    fZPosEndCAF                 = tfs->make<TH1D>("stopZCAF",";Z End Position (cm); NTracks",zbinscaf,0.0,zmaxcaf);
    fLengthCAF                  = tfs->make<TH1D>("lenCAF",";Track Length (cm);NTracks",nlenbins,nlenmin,nlenmax);
    fHitsCAF                    = tfs->make<TH1D>("nhitCAF",";NHits; NTracks",nhitsbins,nhitmin,nhitmax);

    // Event Level Histograms
    fTracksPerEvent             = tfs->make<TH1D>("3dtracksPerEvent",";3D Tracks Per Event; NEvents",101,-0.5,100.5);
    f2DTracksPerEvent           = tfs->make<TH1D>("2dTracksPerEvent",";2D Tracks Per Event; NEvents",101,-0.5,100.5);
    fTracksPerSlice             = tfs->make<TH1D>("3dtracksPerSlice",";3D Tracks Per Event; NEvents",51,-0.5,50.5);
    f2DTracksPerSlice           = tfs->make<TH1D>("2dTracksPerSlice",";2D Tracks Per Event; NEvents",51,-0.5,50.5);

    //Hit Level Histograms
    fXHitsPerPlane              = tfs->make<TH1D>("xHitsPerPlane",";Plane Number; NHits",nPlane,-0.5,nPlane-0.5);
    fYHitsPerPlane              = tfs->make<TH1D>("yHitsPerPlane",";Plane Number; NHits",nPlane,-0.5,nPlane-0.5);
    fHitsPerPlane               = tfs->make<TH1D>("hitsPerPlane",";Plane Number; NHits",nPlane,-0.5,nPlane-0.5);
    fXHitsPerCell               = tfs->make<TH1D>("xHitsPerCell",";Cell Number; NHits",nCell,-0.5,nCell-0.5);
    fYHitsPerCell               = tfs->make<TH1D>("yHitsPerCell",";Cell Number; NHits",nCell,-0.5,nCell-0.5);
    fXPos                       = tfs->make<TH1D>("xPos",";X Position (cm); NTrackHits", xbins, -xMax, xMax);
    fYPos                       = tfs->make<TH1D>("yPos",";Y Position (cm); NTrackHits", ybins, -yMax, yMax);
    fXRecoPos                   = tfs->make<TH1D>("xRecoPos",";X Position (cm); NTrackHits", xbins, -xMax, xMax);
    fYRecoPos                   = tfs->make<TH1D>("yRecoPos",";Y Position (cm); NTrackHits", ybins, -yMax, yMax);
    fXPlaneVCell                = tfs->make<TH2D>("xPlaneVCell",";Plane Number; Cell Number",nPlane,-0.5,nPlane-0.5,nCell,-0.5,nCell-0.5);
    fYPlaneVCell                = tfs->make<TH2D>("yPlaneVCell",";Plane Number; Cell Number",nPlane,-0.5,nPlane-0.5,nCell,-0.5,nCell-0.5);
    fPlaneVCell                 = tfs->make<TH2D>("planeVCell",";Plane Number; Cell Number", nPlane,-0.5,nPlane-0.5,nCell,-0.5,nCell-0.5);   


    // Truth Info
    fTruthTree = tfs->make<TTree>("truthtree","");
    fTruthTree->Branch("pdg",&fpdg,"pdg/I");
    fTruthTree->Branch("E",&fE,"E/F");
    fTruthTree->Branch("p",&fp,"P/F");
    fTruthTree->Branch("vtxdst",&fvtxdst,"vtxdst/F");
    fTruthTree->Branch("eff",&feff,"eff/F");
    fTruthTree->Branch("pur",&fpur,"pur/F");
    fTruthTree->Branch("len",&flen,"len/F");
    fTruthTree->Branch("x",&fx,"x/F");
    fTruthTree->Branch("y",&fy,"y/F");
    fTruthTree->Branch("z",&fz,"z/F");
    fTruthTree->Branch("dx",&fdx,"dx/F");
    fTruthTree->Branch("dy",&fdy,"dy/F");
    fTruthTree->Branch("dz",&fdz,"dz/F");
    fTruthTree->Branch("Enu",&fenu,"Enu/F");
    fTruthTree->Branch("Mode",&fmode,"Mode/I");
    fTruthTree->Branch("weight",&fweight,"weight/F");
    fTruthTree->Branch("event",&fevent,"event/I");
    fTruthTree->Branch("passMuon",&fpassMuon,"passMuon/I");
    fTruthTree->Branch("nhits",&fnhits,"nhits/I");
    fTruthTree->Branch("view",&fview,"view/I");
    fTruthTree->Branch("recolen",&frecolen,"recolen/F");
    fTruthTree->Branch("ccnc",&fccnc,"ccnc/I");

  }

  //......................................................................

  ///
  /// The analysis method is intended to perform checks of the
  /// performance of the reconstruction for the purposes of
  /// optimization, testing, and validation.
  ///
  /// \param evt - Read-only access to the event data
  ///
  // checks to make include - 
  // 1. are the direction cosines of the track correct
  // 2. fraction of hits used in tracks
  // 3. number of particles leaving tracks vs number tracks found
  // 4. fraction of hits in a track that belong to a given particle

  void KalmanTrackAna::analyze(art::Event const& evt)
  {
    
    //get some information about the cell locations from the Geometry object.
    art::ServiceHandle<geo::Geometry> geom;

    // get a backtracker service handle
    art::ServiceHandle<cheat::BackTracker> bt;
    
    // get all the slices
    art::Handle< std::vector<rb::Cluster> > slicecol;
    evt.getByLabel(fSliceLabel,slicecol);

    art::PtrVector<rb::Cluster> slicelist;
    for(unsigned int i = 0; i<slicecol->size();++i){
      art::Ptr<rb::Cluster>slice(slicecol,i);
      slicelist.push_back(slice);
    }   

    // setup the associations between tracks and slices
    art::FindManyP<rb::Track> fmtrack(slicecol,evt,fTrackLabel);

    double n3dTracksPerEvt = 0;
    double n2dTracksPerEvt = 0;

    // loop over slices
    for(unsigned int islice = 0; islice < slicelist.size(); ++islice){
      
      std::vector<art::Ptr<rb::Track> > tracks;
      if(fmtrack.isValid()){
	tracks = fmtrack.at(islice);
      }

      bool filtered = rb::IsFiltered(evt,slicelist[islice]);

      CheckAssociations(slicelist[islice],tracks,filtered);

      if(slicelist[islice]->IsNoise()){ continue; }

      double n3dTracksPerSlice = 0;
      double n2dTracksPerSlice = 0;

      // get truth info for plots
      if(bt->HaveTruthInfo()){
	CheckRecoTracks(tracks,slicelist[islice]->AllCells());
      }

      //loop over tracks and make reco track histograms
      for(unsigned int i=0;i<tracks.size();i++){
	art::Ptr<rb::Track> track = tracks[i];
	// check if the track is 2d or 3d
	fViewCAF->Fill(track->View());
	fCosXCAF->Fill(track->Dir().X());
	fCosYCAF->Fill(track->Dir().Y());
	fCosZCAF->Fill(track->Dir().Z());
	fXPosStartCAF->Fill(track->Start().X());  
	fYPosStartCAF->Fill(track->Start().Y());  
	fZPosStartCAF ->Fill(track->Start().Z());     
	fXPosEndCAF->Fill(track->Stop().X());
	fYPosEndCAF->Fill(track->Stop().Y());
	fZPosEndCAF->Fill(track->Stop().Z());
	fLengthCAF->Fill(track->TotalLength());
	fHitsCAF->Fill(track->NCell());
	if(track->Is3D()){
	  ++n3dTracksPerEvt;
	  ++n3dTracksPerSlice;
	  TVector3 start(track->TrajectoryPoint(0));
	  fXPosStart->Fill(start.X());
	  fYPosStart->Fill(start.Y());
	  fZPosStart->Fill(start.Z());
	  fXvYStart->Fill(start.X(),start.Y());
	  TVector3 end(track->Stop());
	  fXPosEnd->Fill(end.X());
	  fYPosEnd->Fill(end.Y());
	  fZPosEnd->Fill(end.Z());
	  fXvYEnd->Fill(end.X(),end.Y());
	  TVector3 length = end-start;	
	  fHitsPerTrack->Fill(track->NCell());
	  fTrackLength->Fill(track->TotalLength());
	  double cosX = (end.X()-start.X())/length.Mag();
	  fCosX->Fill(cosX);
	  double cosY = (end.Y()-start.Y())/length.Mag();
	  fCosY->Fill(cosY);
	  double cosZ = (end.Z()-start.Z())/length.Mag();
	  fCosZ->Fill(cosZ);
	  for(unsigned int nx = 0; nx<track->NXCell();++nx){
	    art::Ptr<rb::CellHit> xcellhit = track->XCell(nx);
	    int cell = xcellhit->Cell();
	    int plane = xcellhit->Plane();
	    double xyz[3];
	    geom->Plane(plane)->Cell(cell)->GetCenter(xyz);
	    fZvX->Fill(xyz[2],xyz[0]);
	    fXPos->Fill(xyz[0]);
	    fXPlaneVCell->Fill(plane,cell);
	    fPlaneVCell->Fill(plane,cell);
	    fXHitsPerPlane->Fill(plane);
	    fHitsPerPlane->Fill(plane);
	    fXHitsPerCell->Fill(cell);
	    
	    //Get the recohit from this track cell hit
	    rb::RecoHit xReco = track->RecoHit(xcellhit);
	    fYRecoPos->Fill(xReco.Y());
	    fXvYReco->Fill(xReco.X(),xReco.Y());
	    fZvYReco->Fill(xReco.Z(),xReco.Y());
	  }
	  
	  for(unsigned int ny = 0; ny<track->NYCell();++ny){
	    art::Ptr<rb::CellHit> ycellhit = track->YCell(ny);
	    int cell = ycellhit->Cell();
	    int plane = ycellhit->Plane();
	    double xyz[3];
	    geom->Plane(plane)->Cell(cell)->GetCenter(xyz);
	    fZvY->Fill(xyz[2],xyz[1]);
	    fYPos->Fill(xyz[1]);
	    fYPlaneVCell->Fill(plane,cell);
	    fPlaneVCell->Fill(plane,cell);
	    fYHitsPerPlane->Fill(plane);
	    fHitsPerPlane->Fill(plane);
	    fYHitsPerCell->Fill(cell);
	    
	    //Get the recohit from this track cell hit
	    rb::RecoHit yReco = track->RecoHit(ycellhit);
	    fXRecoPos->Fill(yReco.X());
	    fXvYReco->Fill(yReco.X(),yReco.Y());
	    fZvXReco->Fill(yReco.Z(),yReco.X());
	  }
	  
	}
	else{
	  ++n2dTracksPerEvt;
	  ++n2dTracksPerSlice;
	  if(track->View() == geo::kX){
	    TVector3 start(track->TrajectoryPoint(0));
	    fXPosStart->Fill(start.X());
	    fZPosStart->Fill(start.Z());
	    TVector3 end(track->Stop());
	    fXPosEnd->Fill(end.X());
	    fZPosEnd->Fill(end.Z());
	    TVector3 length = end-start;	
	    fHitsPerTrack2d->Fill(track->NCell());
	    fTrackLength2d->Fill(track->TotalLength());
	    double cosX2d = (end.X()-start.X())/length.Mag();
	    fCosX2d->Fill(cosX2d);
	    double cosZ2d = (end.Z()-start.Z())/length.Mag();
	    fCosZ2d->Fill(cosZ2d);
	  }
	  else if(track->View() == geo::kY){
	    TVector3 start(track->TrajectoryPoint(0));
	    fYPosStart->Fill(start.Y());
	    fZPosStart->Fill(start.Z());
	    TVector3 end(track->Stop());
	    fYPosEnd->Fill(end.Y());
	    fZPosEnd->Fill(end.Z());
	    TVector3 length = end-start;	
	    fHitsPerTrack2d->Fill(track->NCell());
	    fTrackLength2d->Fill(track->TotalLength());
	    double cosY2d = (end.Y()-start.Y())/length.Mag();
	    fCosY2d->Fill(cosY2d);
	    double cosZ2d = (end.Z()-start.Z())/length.Mag();
	    fCosZ2d->Fill(cosZ2d);
	  }
	}
      }
      fTracksPerSlice->Fill(n3dTracksPerSlice);
      f2DTracksPerSlice->Fill(n2dTracksPerSlice);
    }
    
    fTracksPerEvent->Fill(n3dTracksPerEvt);
    f2DTracksPerEvent->Fill(n2dTracksPerEvt);
    
  }

  //-------------------------------------------------------------------
  void KalmanTrackAna::CheckRecoTracks(std::vector< art::Ptr<rb::Track> > const& trkcol,
				       art::PtrVector<rb::CellHit> const& allhits)
  {
    
    // grab the set of eve IDs for this plist
    art::ServiceHandle<cheat::BackTracker> bt;
    std::set<int> trackIDs = bt->GetTrackIDList();
    
    // Only keep the ones associated with not completely unreasonable tracks
    // remove for now, take care of this with PassMuonCuts function
    trackIDs = FindVisibleProngs(allhits, trackIDs);
    
    // Empty set means accept all PDGs. Otherwise, filter out the unwanted ones
    std::set<int> newTrackIDs;
    for(std::set<int>::iterator it = trackIDs.begin(); it != trackIDs.end(); ++it){
      const int pdg = bt->TrackIDToParticle(*it)->PdgCode();
      if(abs(pdg) != 13){ continue; }
      // make sure this is a primary particle
      if(bt->TrackIDToMotherParticle(*it)->TrackId() != *it){continue;}
      // check that its a particle we care about
      newTrackIDs.insert(*it);
    } // end for it
    trackIDs = newTrackIDs;
    
    std::map<int, int> parents;
    for(std::set<int>::iterator it = trackIDs.begin(); it != trackIDs.end(); ++it){
      const sim::Particle* mother = bt->TrackIDToParticle(*it);
      if(!mother) continue; // Can this happen?
      for(int i = 0; i < mother->NumberDaughters(); ++i){
	const int id = mother->Daughter(i);
	assert(bt->TrackIDToParticle(id));
	if(abs(bt->TrackIDToParticle(id)->PdgCode()) == 11){ parents[id] = *it; }
      } // end for i
    } // end for it
    
    // Indexed by what view they're in. X, Y, or 3D
    assert(geo::kXorY == 2); // Ensure enum still works how we think
    std::map<int, double> bestPur[3], bestEff[3], bestLength[3], bestHits[3];
    
    // Go through each reconstructed object, accumulating the best scores seen
    // for each trackID we're interested in.
    for(size_t c = 0; c < trkcol.size(); ++c){
      const rb::Track* trk = trkcol[c].get();
      const geo::View_t view = trk->View();
      
      // get the hits associated with this object
      const art::PtrVector<rb::CellHit> hits = trk->AllCells();
      
      // use the cheat::BackTracker to find purity and efficiency for these
      // hits for each particle ID
      std::map<int, double> purMap, effMap;
      bt->HitCollectionPurity(trackIDs, hits, &purMap, &parents);
      bt->HitCollectionEfficiency(trackIDs, hits, allhits, view, &effMap);
      
      for(std::set<int>::iterator it = trackIDs.begin(); it != trackIDs.end(); ++it){
	const double p = purMap[*it];
	const double e = effMap[*it];
	
	if(e > bestEff[view][*it]) {
	bestEff[view][*it] = e;
	bestPur[view][*it] = p;
	
	// If we were told to fill a track length histogram then downcast
	// should be safe.
	bestLength[view][*it] = trk->TotalLength();
	bestHits[view][*it] = trk->NCell();
	}
      } // end for it
    } // end for c
    
    // For every track. If it was matched in 3D, fill the histograms.
    // If there is no 3D match look for 2 2D matches and fill with weight 0.5
    // If required we can add a flag to only try 3D or only try 2D.
    for(std::set<int>::iterator it = trackIDs.begin(); it != trackIDs.end(); ++it){
      // Try 3D first
      for(int view = geo::kXorY; view >= geo::kX; --view){
	// If no 3D match, continue to two 2D options
	if(view == geo::kXorY && bestEff[view].find(*it) == bestEff[view].end()) continue;
	
	const double weight = (view == geo::kXorY) ? 1 : 0.5;
	
	const double trueLength = TotalLengthInDetector(bt->TrackIDToParticle(*it), geo::View_t(view));
	
	const sim::Particle* part = bt->TrackIDToParticle(*it);
	fE = part->E();
	fp = part->P();
	TVector3 truestart;
	truestart.SetXYZ(part->Vx(),part->Vy(),part->Vz());
	fx = part->Vx();
	fy = part->Vy();
	fz = part->Vz();
	fvtxdst = Containment(truestart);
	fpdg = part->PdgCode();
	fdx = part->Px()/part->P();
	fdy = part->Py()/part->P();
	fdz = part->Pz()/part->P();
	fnhits = bestHits[view][*it];
	feff = bestEff[view][*it];
	fpur = bestPur[view][*it];
	frecolen = bestLength[view][*it];
	flen = trueLength;
	fweight = weight;
	fpassMuon = bt->PassMuonCuts(*it,allhits);
	fview = view;
	fmode = -1;
	fenu = -1;

	// check if this is coming from a neutrino
	const art::Ptr<simb::MCTruth> mctruth = bt->ParticleToMCTruth(part);
	if(mctruth->NeutrinoSet()){
	  // get the neutrino if it exists
	  const simb::MCNeutrino neutrino = mctruth->GetNeutrino();
	  fmode = neutrino.Mode();
	  fenu = neutrino.Nu().E();
	  fccnc = neutrino.CCNC();
	}

	fTruthTree->Fill();
	
	// If we did find a 3D match, don't try 2D
	if(bestEff[geo::kXorY].find(*it) != bestEff[geo::kXorY].end()) break;
      } // end for view
    } // end for it
  }
  
  //-------------------------------------------------------------------
  std::set<int> KalmanTrackAna::FindVisibleProngs(const art::PtrVector<rb::CellHit>& allhits,
						  std::set<int> ids) const
  {
    std::map<int, Counts> counts;
    
    art::ServiceHandle<cheat::BackTracker> bt;
    
    // loop over all hits and check if this track id contributed to them
    for(unsigned int i = 0; i < allhits.size(); ++i){
      //get vector of track id's that contribute to this hit
      const art::Ptr<rb::CellHit> hit = allhits[i];
      const std::vector<cheat::TrackIDE> trackIDs = bt->HitToTrackIDE(hit);
      // loop over all the returned track ids
      int nTracks = 0; // number of tracks that contributed energy to the hit
      int loneID = -1; // Which ID it was if there was only one
      for(unsigned int j = 0; j < trackIDs.size(); ++j){
	const int trackID = trackIDs[j].trackID;
	if(trackID != sim::kNoiseId){
	  ++nTracks;
	  loneID = trackID;
	}

      // check if the current id matches the input track id
	if(ids.find(trackID) != ids.end()){
	  ++counts[trackID].nHits[hit->View()];
	}
    } // end for j
      if(nTracks == 1){
	++counts[loneID].nOnlyHits[hit->View()];
      }
    } // end for i
    
    const int kMinHits = 2; // minimum number of hits to be considered visible
    const int kMinOnlyHit = 1; // minimum number of hits in which it is only particle contributing energy to be visible
    
    std::set<int> ret;
    for(std::set<int>::iterator it = ids.begin(); it != ids.end(); ++it){
      const Counts c = counts[*it];
      // Check if this track passes the visible prong requirements
      if(c.nOnlyHits[geo::kX] >= kMinOnlyHit &&
	 c.nOnlyHits[geo::kY] >= kMinOnlyHit){
		
	if(c.nHits[geo::kX] >= kMinHits && c.nHits[geo::kY] >= kMinHits)
	  ret.insert(*it);
      }
    }
    return ret;
  }
  
  //-------------------------------------------------------------------
  double KalmanTrackAna::TotalLengthInDetector(const sim::Particle* p,
					       geo::View_t view) const
  {
    const unsigned int N = p->NumberTrajectoryPoints();
    
    if(N < 2){
      mf::LogWarning("RecoCheckAna") << "Particle has no trajectory points";
      return 0;
    }
    
    art::ServiceHandle<geo::Geometry> geom;
    
    double dist = 0;
    
    for(unsigned int n = 0; n < N-1; ++n){
      TVector3 p0 = p->Position(n).Vect();
      TVector3 p1 = p->Position(n+1).Vect();
      if(p0.Z() < 0 || p1.Z() < 0 ||
	 p0.Z() > geom->DetLength() || p1.Z() > geom->DetLength() ||
	 fabs(p0.X()) > geom->DetHalfWidth() || fabs(p1.X()) > geom->DetHalfWidth() ||
	 fabs(p0.Y()) > geom->DetHalfHeight() || fabs(p1.Y()) > geom->DetHalfHeight()) continue;
      
      if(view == geo::kX){
	p0.SetY(0);
	p1.SetY(0);
      }
      if(view == geo::kY){
	p0.SetX(0);
	p1.SetX(0);
      }
      dist += (p1-p0).Mag();
    }
    
    return dist;
  }
  
  //.....................................................................
  double KalmanTrackAna::Containment(TVector3 start) const
  {
    double dist = 0;
    double xmin,xmax,ymin,ymax,zmin,zmax;
    
    // define the boundaries of containment
    art::ServiceHandle<geo::Geometry>  geom;    
    
    double detHalfWidth   =  geom->DetHalfWidth();
    double detHalfHeight  =  geom->DetHalfHeight();
    double detLength =  geom->DetLength();
    
    // fd detector edges determined from event display for 28 block detector
    // get rid of offset between geom->DetHalfWidth()(Height()) and actual detector edges
    double voffset = 15.0;
    xmin = -detHalfWidth+voffset;
    xmax = detHalfWidth-voffset;
    ymin = -detHalfHeight+voffset;
    ymax = detHalfHeight-voffset;
    zmin = 0;
    zmax = detLength;
    
    TVector3 mins(xmin,ymin,zmin);
    TVector3 maxs(xmax,ymax,zmax);
    
    float mindist = 1e10;
    
    if(start.X() > xmin && start.X() < xmax && 
       start.Y() > ymin && start.Y() < ymax && 
       start.Z() > zmin && start.Z() < zmax){
      TVector3 mindiffs = mins - start;
      TVector3 maxdiffs = maxs - start;
      if(TMath::Abs(mindiffs.X()) < mindist){ mindist = TMath::Abs(mindiffs.X()); }
      if(TMath::Abs(mindiffs.Y()) < mindist){ mindist = TMath::Abs(mindiffs.Y()); }
      if(TMath::Abs(mindiffs.Z()) < mindist){ mindist = TMath::Abs(mindiffs.Z()); }
      if(TMath::Abs(maxdiffs.X()) < mindist){ mindist = TMath::Abs(maxdiffs.X()); }
      if(TMath::Abs(maxdiffs.Y()) < mindist){ mindist = TMath::Abs(maxdiffs.Y()); }
      if(TMath::Abs(maxdiffs.Z()) < mindist){ mindist = TMath::Abs(maxdiffs.Z()); }
      dist = mindist;
    }
    
    return dist;
  }
    
  //.....................................................................
  void KalmanTrackAna::CheckAssociations(art::Ptr<rb::Cluster> slice,std::vector<art::Ptr<rb::Track> > tracks, bool filtered)
  {
    if(filtered && tracks.size() > 0){
      mf::LogVerbatim("KalmanTrackAna")<<"Filtered slice has non-zero number of reconstructed tracks. This should never happen.";
      // Should never create tracks on slices that have been filtered
      abort();
    }
    if(slice->IsNoise() && tracks.size() > 0){
      mf::LogVerbatim("KalmanTrackAna")<<"Noise slice has non-zero number of reconstructed tracks. This should never happen.";
      // Should never create tracks on slices that have been filtered
      abort();
    }

    // check that reco'd track hits all belong to the slice hits
    art::PtrVector<rb::CellHit> slicehits = slice->AllCells();
    for(unsigned int itrack = 0; itrack < tracks.size(); ++itrack){
      // get all of the track hits for this track
      art::PtrVector<rb::CellHit> trackhits = tracks[itrack]->AllCells();
      // loop over the track hits
      for(unsigned int itrhit = 0; itrhit < trackhits.size(); ++itrhit){
	bool foundhit = false;
	// loop over all the slice hits
	for(unsigned int islhit = 0; islhit < slicehits.size(); ++islhit){
	  if(trackhits[itrhit] == slicehits[islhit]){ 
	    foundhit = true;
	    break;
	  }
	}
	// after looking at all the slice hits, did we find this track hit. If not, yell.
	if(!foundhit){
	  mf::LogVerbatim("KalmanTrackAna")<<"Track made from hits on a slice other than the slice the track is associated to. This should never happen.";
	  // Should never create tracks from hits on slices other than the slice its associated with
	  abort();
	}
      }
    }

  }

} // end namespace trk

DEFINE_ART_MODULE(trk::KalmanTrackAna);
