//////////////////////////////////////////////////////////////////////
/// \file    RLFit.cxx
/// \brief   Make a linear fit to a collection of raw hits
/// \version $Id: RLFit.cxx,v 1.9 2011-11-23 20:57:37 bckhouse Exp $
/// \author  messier@indiana.edu
/////////////////////////////////////////////////////////////////////
#include <cassert>
#include <cmath>

#include "art/Framework/Services/Registry/ServiceHandle.h"

#include "CMap/CMap.h"
#include "Geometry/Geometry.h"
#include "GeometryObjects/Geo.h"
#include "GeometryObjects/CellGeo.h"
#include "RawData/RawDigit.h"
#include "TrackFit/RLFit.h"
#include "Utilities/func/MathUtil.h"

namespace trk
{
  //.................................................................
  RLFit::RLFit(const std::vector< art::Ptr<rawdata::RawDigit> >& d,
	       unsigned int i1,
	       unsigned int i2) :
    fR(8),
    fT(8)
  {
    //
    // The schedule of the anealing. Units are cm^2. First iteration is
    // not used as weights come from the seeds.
    //
    fR[0] =  0.0;  fT[0] =   0.0;
    fR[1] = 36.0; fT[1] = 1000.0;
    fR[2] = 36.0; fT[2] =  500.0;
    fR[3] = 36.0; fT[3] =  250.0;
    fR[4] = 36.0; fT[4] =  100.0;
    fR[5] = 36.0; fT[5] =   50.0;
    fR[6] = 36.0; fT[6] =   25.0;
    fR[7] = 36.0; fT[7] =   10.0;
    //
    // The geometry box
    //
    art::ServiceHandle<geo::Geometry> geo;
    fBoxXlo = -geo->DetHalfWidth();
    fBoxXhi =  geo->DetHalfWidth();
    fBoxYlo = -geo->DetHalfHeight();
    fBoxYhi =  geo->DetHalfHeight();
    fBoxZlo =  0.0;
    fBoxZhi =  geo->DetLength();

    //
    // Pull out the hit information we need for a fit
    //
    art::ServiceHandle<cmap::CMap> cmap;
    for (unsigned int i=i1; i<=i2; ++i) {
      int plane = cmap->GetPlane(d[i].get());
      int cell  = cmap->GetCell(d[i].get());
      int view  = geo->Plane(plane)->View();
    
      double xyz[3];
      geo->Plane(plane)->Cell(cell)->GetCenter(xyz);
    
      if (view==geo::kX) {
	fXhit.push_back(d[i].get());
	fXx.push_back(xyz[0]);
	fZx.push_back(xyz[2]);
	fWx.push_back(0);
      }
      else if (view==geo::kY) {
	fYhit.push_back(d[i].get());
	fYy.push_back(xyz[1]);
	fZy.push_back(xyz[2]);
	fWy.push_back(0);
      }
    }
    this->SeedWeights(fXx, fZx, fYy, fZy, fWx);
    this->SeedWeights(fYy, fZy, fXx, fZx, fWy);
  }

  //......................................................................
  void RLFit::SeedWeights(const std::vector<double>& x, 
			  const std::vector<double>& zx,
			  const std::vector<double>& y,
			  const std::vector<double>& zy,
			  std::vector<double>& w)
  {
    for (unsigned int i=0; i<zx.size(); ++i) {
      double d;
      w[i] = 1.0E-9;
      //
      // Find closest hit in same view
      //
      for (unsigned int j=0; j<zx.size(); ++j) {
	if (i==j) continue;
	d = util::pythag(zx[i]-zx[j], x[i]-x[j]);
	if (d<30.0) w[i] += 1.0;
      }
      //
      // Find closest hit in orthogonal view
      //
      for (unsigned int j=0; j<zy.size(); ++j) {
	d = fabs(zx[i]-zy[j]);
	if (d<30.0) w[i] += 0.1;
      }
    }
  }

  //......................................................................
  void RLFit::Fit()
  {
    //
    // Make a line fit to the hits. To improve preformance in a noisy
    // environment, use a deterministic annealing algoritm
    //
    double x1  = 0;
    double x2  = 0;
    double zx1 = 0;
    double zx2 = 0;
    double y1  = 0;
    double y2  = 0;
    double zy1 = 0;
    double zy2 = 0;
    for (unsigned int i=0; i<fR.size(); ++i) {
      static const double eps = 1.0E-6;
      //
      // Calculate weights for this iteration. First iteration uses seed
      // weights
      //
      if (i>0) {
	for (unsigned int j=0; j<fWx.size(); ++j) {
	  double d = geo::DsqrToLine(fZx[j], fXx[j], zx1, x1, zx2, x2);
	
	  fWx[j] = (1.0+exp(-fR[i]/fT[i]))/(1.0+exp((d-fR[i])/fT[i]));
	  //
	  // Penalty for hits outside the detector in the other view
	  //
	  double y = fMzy*fZx[j] + fBzy;
	  if (y<fBoxYlo) fWx[j] *= exp(-(fBoxYlo-y)/fT[i]);
	  if (y>fBoxYhi) fWx[j] *= exp(-(y-fBoxYhi)/fT[i]);
	
	  if (fWx[j]<eps) fWx[j] = eps;
	}
	for (unsigned int j=0; j<fWy.size(); ++j) {
	  double d = geo::DsqrToLine(fZy[j], fYy[j], zy1, y1, zy2, y2);
	
	  fWy[j] = (1.0+exp(-fR[i]/fT[i]))/(1.0+exp((d-fR[i])/fT[i]));
	  //
	  // Penalty for hits outside the detector in the other view
	  //
	  double x = fMzx*fZy[j] + fBzx;
	  if (x<fBoxXlo) fWy[j] *= exp(-(fBoxXlo-x)/fT[i]);
	  if (x>fBoxXhi) fWy[j] *= exp(-(x-fBoxXhi)/fT[i]);
	
	  if (fWy[j]<eps) fWy[j] = eps;
	}
      }
      //
      // Perform the fits to the two views
      //
      geo::LinFitMinDperp(fZx, fXx, fWx, &zx1, &x1, &zx2, &x2);
      geo::LinFitMinDperp(fZy, fYy, fWy, &zy1, &y1, &zy2, &y2);
      //
      // Compute slopes and intercepts
      //
      double dzx = zx2-zx1;
      if (dzx<=0.0) dzx = 1.0E-9;
      fMzx = (x2-x1)/dzx;
      fBzx = (x1*zx2-x2*zx1)/dzx;
    
      double dzy = zy2-zy1;
      if (dzy<=0.0) dzy = 1.0E-9;
      fMzy = (y2-y1)/dzy;
      fBzy = (y1*zy2-y2*zy1)/dzy;
    }
    fChi2 = 0.0;
    fDOF  = 0.0;
    for (unsigned int j=0; j<fWx.size(); ++j) {
      double d = geo::DsqrToLine(fZx[j], fXx[j], zx1, x1, zx2, x2);
      fChi2 += fWx[j]*d;
      fDOF  += fWx[j];
    }
    for (unsigned int j=0; j<fWy.size(); ++j) {
      double d = geo::DsqrToLine(fZy[j], fYy[j], zy1, y1, zy2, y2);
      fChi2 += fWy[j]*d;
      fDOF  += fWy[j];
    }

    // Force north-going assumption on both views...
    if (zx1>zx2) { std::swap(x1, x2); std::swap(zx1, zx2); }
    if (zy1>zy2) { std::swap(y1, y2); std::swap(zy1, zy2); }
  
    //
    // Compute slopes and intercepts
    //
    double dzx = zx2-zx1;
    if (dzx<=0.0) dzx = 1.0E-9;
    fMzx = (x2-x1)/dzx;
    fBzx = (x1*zx2-x2*zx1)/dzx;
  
    double dzy = zy2-zy1;
    if (dzy<=0.0) dzy = 1.0E-9;
    fMzy = (y2-y1)/dzy;
    fBzy = (y1*zy2-y2*zy1)/dzy;

    //
    // Adjust z ranges to completely contain hits on the track. "On
    // track" is defined as having significant weight
    //
    fZ1 = fZx[0];
    fZ2 = fZx[0]+0.01;
    for (unsigned int i=0; i<fWx.size(); ++i) {
      if (fWx[i]>0.001) {
	if (fZx[i]<fZ1) fZ1 = fZx[i];
	if (fZx[i]>fZ2) fZ2 = fZx[i];
      }
    }
    for (unsigned int i=0; i<fWy.size(); ++i) {
      if (fWy[i]>0.001) {
	if (fZy[i]<fZ1) fZ1 = fZy[i];
	if (fZy[i]>fZ2) fZ2 = fZy[i];
      }
    }
    //
    // Use line fits to compute x and y locations at the start and end z
    // positions
    //
    fX1 = fMzx*fZ1 + fBzx;
    fX2 = fMzx*fZ2 + fBzx;

    fY1 = fMzy*fZ1 + fBzy;
    fY2 = fMzy*fZ2 + fBzy;
  }

  //......................................................................
  void RLFit::AssumeNorthGoing() 
  {
    if (fZ1>fZ2) {
      std::swap(fX1, fX2);
      std::swap(fY1, fY2);
      std::swap(fZ1, fZ2);
      this->UpdateSlopes();
    }
  }

  //......................................................................
  void RLFit::AssumeDownGoing() 
  {
    if (fY2>fY1) {
      std::swap(fX1, fX2);
      std::swap(fY1, fY2);
      std::swap(fZ1, fZ2);
      this->UpdateSlopes();
    }
  }

  //......................................................................
  void RLFit::UpdateSlopes() 
  {
    fMzx = (fX2-fX1)/(fZ2-fZ1);
    fBzx = (fX1*fZ2-fX2*fZ1)/(fZ2-fZ1);
  
    fMzy = (fY2-fY1)/(fZ2-fZ1);
    fBzy = (fY1*fZ2-fY2*fZ1)/(fZ2-fZ1);
  }

  //......................................................................
  double RLFit::CosThetaBeam()
  {
    this->AssumeNorthGoing();

    double dz = (fZ2-fZ1);
    double dx = (fX2-fX1);
    double dy = (fY2-fY1);
  
    return dz/sqrt(dx*dx+dy*dy+dz*dz);
  }

  //......................................................................
  double RLFit::PhiBeam()
  {
    this->AssumeNorthGoing();

    double dx = (fX2-fX1);
    double dy = (fY2-fY1);
  
    return atan2(dy,dx);
  }

  //......................................................................
  double RLFit::CosThetaCosmic()
  {
    this->AssumeDownGoing();

    double dz = (fZ2-fZ1);
    double dx = (fX2-fX1);
    double dy = (fY2-fY1);
  
    return -dy/sqrt(dx*dx+dy*dy+dz*dz);
  }

  //......................................................................
  double RLFit::PhiCosmic()
  {
    this->AssumeDownGoing();

    double dx = (fX2-fX1);
    double dz = (fZ2-fZ1);
  
    return atan2(-dx,-dz);
  }

  //......................................................................
  void RLFit::SetRTSchedule(const std::vector<double>& r, 
			    const std::vector<double>& t)
  {
    assert(r.size()==r.size());
    fR.resize(r.size());
    fT.resize(t.size());
    for (unsigned int i=0; i<r.size(); ++i) {
      fR[i] = r[i];
      fT[i] = t[i];
    }
  }

} // end namespace trk
////////////////////////////////////////////////////////////////////////
