//////////////////////////////////////////////////////////////////////////////
// \file    TrkAssn_module.cc
// \brief   Module to write associations between prongs, tracks, and vertices and the slices they belong too
// \version $Id: TrkAssn_module.cc,v 1.3 2013-01-30 21:28:53 raddatz Exp $
// \author  Nick Raddatz
////////////////////////////////////////////////////////////////////////////

#include <string>

// Framework includes
#include "art/Framework/Core/EDProducer.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h"
#include "fhiclcpp/ParameterSet.h"

// Nova includes
#include "RecoBase/Track.h"
#include "RecoBase/Cluster.h"
#include "RecoBase/CellHit.h"
#include "RecoBase/Prong.h"
#include "RecoBase/Vertex.h"
#include "Utilities/AssociationUtil.h"


namespace trk {
  class TrkAssn : public art::EDProducer {
  public:
    explicit TrkAssn(fhicl::ParameterSet const& pset);
    virtual ~TrkAssn();
    void produce(art::Event& evt);
    void reconfigure(const fhicl::ParameterSet& p);
    
  private:    
    std::string                            fTrackInput;        // Input for tracks
    std::string                            fSliceInput;        // Input for slices
    std::string                            fProngInput;        // Input for prongs
    std::string                            fVertexInput;       // Input for verticies
  };
};

namespace trk{

  
  //......................................................................
  TrkAssn::TrkAssn(fhicl::ParameterSet const& pset)
  {
    reconfigure(pset);
    produces< art::Assns<rb::Cluster, rb::Track> >();
    produces< art::Assns<rb::Cluster, rb::Prong> >();
    produces< art::Assns<rb::Cluster, rb::Vertex> >();
  }
  
  //......................................................................
  TrkAssn::~TrkAssn()
  {
  }

  //......................................................................
  void TrkAssn::reconfigure(const fhicl::ParameterSet& pset)
  {
    fTrackInput        = pset.get< std::string >("TrackInput") ;
    fSliceInput        = pset.get< std::string >("SliceInput") ;
    fProngInput        = pset.get< std::string >("ProngInput") ;
    fVertexInput       = pset.get< std::string >("VertexInput");
  }
  //......................................................................
  /// Reconstruction method for this module. Provides read-write access
  /// to the event data.

  void TrkAssn::produce(art::Event& evt)
  {
    // Declare a container for Track objects to be stored in the art::event
    std::unique_ptr< art::Assns<rb::Cluster,rb::Track> > assns(new art::Assns<rb::Cluster, rb::Track>);
    std::unique_ptr< art::Assns<rb::Cluster,rb::Prong> > passns(new art::Assns<rb::Cluster, rb::Prong>);
    std::unique_ptr< art::Assns<rb::Cluster,rb::Vertex> > vassns(new art::Assns<rb::Cluster, rb::Vertex>);

    // define a vector holding slices
    art::Handle<std::vector<rb::Cluster> > slicecol;
    evt.getByLabel(fSliceInput,slicecol);

    // define a vector holding tracks
    art::Handle<std::vector<rb::Track> > trkcol;
    evt.getByLabel(fTrackInput,trkcol);

    art::Handle<std::vector<rb::Prong> > pcol;
    evt.getByLabel(fProngInput,pcol);

    art::Handle<std::vector<rb::Vertex> > vcol;
    evt.getByLabel(fVertexInput,vcol);

    std::vector<bool> foundsl;
    for(unsigned int i = 0; i<trkcol->size();++i){
      foundsl.push_back(false);
    }

    std::vector<bool> foundslp;
    for(unsigned int i = 0; i<pcol->size();++i){
      foundslp.push_back(false);
    }

    std::vector<bool> foundslv;
    for(unsigned int i = 0; i<vcol->size();++i){
      foundslv.push_back(false);
    }

    // loop over the slices
    for(unsigned int isl = 0; isl<slicecol->size(); ++isl){
      // get the vector of cell hits for this slice
      art::Ptr<rb::Cluster> slice(slicecol,isl);
      art::PtrVector<rb::CellHit> slicehits = slice->AllCells();
      // loop over the tracks
      std::vector<art::Ptr<rb::Track> > slicetrks;
      for(unsigned int itrk = 0; itrk<trkcol->size(); ++itrk){
	art::Ptr<rb::Track> track(trkcol,itrk);
	if(foundsl[itrk]){ continue; }
	if(track->NCell() < 1){ continue; }
	// any hit on the track should only belong to one slice so take the first one
	art::Ptr<rb::CellHit> trkhit = track->Cell(0);
	// loop over the slice hits and see if this on is in the list
	bool foundtrkhit = false;
	for(unsigned int ihit = 0; ihit<slicehits.size(); ++ihit){
	  if(trkhit == slicehits[ihit]){
	    foundtrkhit = true;
	    break;
	  }
	}
	if(foundtrkhit){
	  foundsl[itrk] = true;
	  // add this track to the vector of tracks to be associated with this slice
	  slicetrks.push_back(track);
	}
      }
      // loop over the prongs
      std::vector<art::Ptr<rb::Prong> > sliceprongs;
      for(unsigned int ip = 0; ip<pcol->size(); ++ip){
	art::Ptr<rb::Prong> prong(pcol,ip);
	if(foundslp[ip]){ continue; }
	if(prong->NCell() < 1){ continue; }
	// any hit in the prong should only belong to one slice so take the first one
	art::Ptr<rb::CellHit> phit = prong->Cell(0);
	// loop over the slice hits and see if this on is in the list
	bool foundphit = false;
	for(unsigned int ihit = 0; ihit<slicehits.size(); ++ihit){
	  if(phit == slicehits[ihit]){
	    foundphit = true;
	    break;
	  }
	}
	if(foundphit){
	  foundslp[ip] = true;
	  // add this prong to the vector of prongs to be associated with this slice
	  sliceprongs.push_back(prong);
	}
      }
      // loop over the verticies
      std::vector<art::Ptr<rb::Vertex> > slicevertex;
      for(unsigned int iv = 0; iv<vcol->size(); ++iv){
	art::Ptr<rb::Vertex> vert(vcol,iv);
	if(foundslv[iv]){ continue; }
	// vertex time should fall inside a slice
	if ((vert->GetT() >= slice->MinTNS()) && (vert->GetT() <= slice->MaxTNS())){
	  foundslv[iv] = true;
	  // add this vertex to the vector of verticies to be associated with this slice
	  slicevertex.push_back(vert);
	}
      }
      if(slicetrks.size()>0){
	util::CreateAssn(*this,evt,slice,slicetrks,*(assns.get()));
      }
      if(sliceprongs.size()>0){
	util::CreateAssn(*this,evt,slice,sliceprongs,*(passns.get()));
      }
      if(slicevertex.size()>0){
	util::CreateAssn(*this,evt,slice,slicevertex,*(vassns.get()));
      }
    }

    evt.put(std::move(assns));
    evt.put(std::move(passns));
    evt.put(std::move(vassns));
  }

  //......................................................................
}

namespace trk{

  DEFINE_ART_MODULE(TrkAssn);

}

