///////////////////////////////////////////////////////////////////////
/// \file WindowTrackingAlg.cxx
/// \brief   Track finder for cosmic rays
/// \author  brebel@fnal.gov
/// \date
///////////////////////////////////////////////////////////////////////
#include <cmath>

// NOvASoft includes
#include "RecoBase/Track.h"
#include "GeometryObjects/CellGeo.h"
#include "GeometryObjects/Geo.h"
#include "Utilities/func/MathUtil.h"
#include "RecoBase/CellHit.h"
#include "RecoBase/Cluster.h"
#include "TrackFit/WindowTrackingAlg.h"

// ART includes
#include "fhiclcpp/ParameterSet.h"
#include "messagefacility/MessageLogger/MessageLogger.h"
#include "cetlib/exception.h"
#include "art/Framework/Services/Optional/TFileService.h"

#include "TStopwatch.h"

bool gt(unsigned short int a, unsigned short int b)
{
  return a > b;
}

bool lt(unsigned short int a, unsigned short int b)
{
  return a < b;
}

bool gtvec(std::pair<geo::View_t, TVector3> a, std::pair<geo::View_t, TVector3> b)
{
  return a.second.Z() > b.second.Z();
}
bool ltvec(std::pair<geo::View_t, TVector3> a, std::pair<geo::View_t, TVector3> b)
{
  return a.second.Z() < b.second.Z();
}

bool sortCellPtr(art::Ptr<rb::CellHit> const a, art::Ptr<rb::CellHit> const b)
{
  return a->Cell() < b->Cell();
}


namespace trk {

  //----------------------------------------------------------------------------
  // Constructor
  WindowTrackingAlg::WindowTrackingAlg(fhicl::ParameterSet const &pset)
    : TrackAlg(pset)
    , fCellHalfDepth(-1.)
  {
    fDHitGood           = pset.get<double      >("DHitGood"          );
    fMinViewPlanes   	= pset.get<unsigned int>("MinViewPlanes"     );
    fWindowSize      	= pset.get<unsigned int>("WindowSize"        );
    fMinHitsTryAgain 	= pset.get<unsigned int>("MinHitsToTryAgain" );
    fMinHitFraction  	= pset.get<double      >("MinHitFraction"    );
    fNTrajectory     	= pset.get<unsigned int>("NTrajectoryForDir" );
    fMaxPlaneSeparation = pset.get<int         >("MaxPlaneSeparation");
  }
  
  //----------------------------------------------------------------------------
  // Destructor
  WindowTrackingAlg::~WindowTrackingAlg() {}
  
  //----------------------------------------------------------------------------
  // MakeTrack
  std::vector<rb::Track> WindowTrackingAlg::MakeTrack( const rb::Cluster* slice ) 
  {
    if(fCellHalfDepth < 0.) fCellHalfDepth = fGeo->Plane(0)->Cell(0)->HalfD();


    //LOG_DEBUG("WindowTrackingAlg") << "fitting new slice";
    
    std::vector<rb::Track> tracks;
    
    // get the PtrVector of all cells in this slice
    art::PtrVector<rb::CellHit> allCells = slice->AllCells();
    allCells.sort();
    std::vector<art::Ptr<rb::CellHit> > allHitsFor;
    std::vector<art::Ptr<rb::CellHit> > allHitsRev;

    for(auto cell : allCells){
      allHitsFor.push_back(cell);
      allHitsRev.push_back(cell);
    }

    std::vector<art::Ptr<rb::CellHit> > removedHits;
    std::vector<art::Ptr<rb::CellHit> > removedHitsFor;
    std::vector<art::Ptr<rb::CellHit> > removedHitsRev;
    bool tryAnother = true;

    double chiSqrFor = 0.;
    double chiSqrRev = 0.;

    rb::Track track;

    while(tryAnother){
      tryAnother = false;

      rb::Track trackForward = this->Make3DTrack(allHitsFor, removedHitsFor, chiSqrFor, lt);
      rb::Track trackReverse = this->Make3DTrack(allHitsRev, removedHitsRev, chiSqrRev, gt);
            
      //LOG_DEBUG("WindowTrackingAlg") << "forward chi^2: "   << chiSqrFor << "/" << trackForward.NCell()
      //			<< "\nreverse chi^2: " << chiSqrRev << "/" << trackReverse.NCell();

      if(trackForward.NCell() > 0) chiSqrFor /= trackForward.NCell();
      if(trackReverse.NCell() > 0) chiSqrRev /= trackReverse.NCell();

      if(chiSqrFor < chiSqrRev){
	track = trackForward;
	removedHits.swap(removedHitsFor);
      }
      else{
	track = trackReverse;
	removedHits.swap(removedHitsRev);
      }

      removedHitsRev.clear();
      removedHitsFor.clear();

      // check to see if we made a legit track, it will have at least 1 cell
      if(track.NCell() > 1){
	this->SetTrackEndPoints(track);

	tracks.push_back(track);

	// Some debugging code commented out below, leaving it in place for future convenience
// 	if(track.Dir().Z() > 0){
// 	  // get the cells for this track and sort them
// 	  art::PtrVector<rb::CellHit> trCells = track.AllCells();
// 	  trCells.sort();

// 	  double xyzStart[3] = {0.};
// 	  double xyzEnd[3] = {0.};
// 	  fGeo->Plane(trCells[0]->Plane())->Cell(trCells[0]->Cell())->GetCenter(xyzStart);
// 	  fGeo->Plane(trCells[track.NCell()-1]->Plane())->Cell(trCells[track.NCell()-1]->Cell())->GetCenter(xyzEnd);

// 	  LOG_VERBATIM("WindowTrackingAlg") << "first cell " << xyzStart[2] //<< "," << *(trCells[0])
// 					    << "\nlast cell " << xyzEnd[2] //<< "," << *(trCells[track.NCell()-1])
// 					    << "\nstart: " << track.Start().X() << " " << track.Start().Y() << " " << track.Start().Z()
// 					    << "\nend: " << track.Stop().X() << " " << track.Stop().Y() << " " << track.Stop().Z() << std::endl;

// 	  for(size_t h = 0; h < track.NCell(); ++h)
// 	    LOG_VERBATIM("WindowTrackingAlg") << *(track.Cell(h));
// 	}

	if(removedHits.size() > fMinHitsTryAgain){
	  tryAnother = true;
	  //LOG_DEBUG("WindowTrackingAlg") << "trying to make another track from the " << removedHits.size() << " removed hits";
	  allHitsFor.clear();
	  allHitsRev.clear();
	  for(auto removed : removedHits){
	    //LOG_DEBUG("WindowTrackingAlg") << "removed " << *removed;
	    allHitsFor.push_back(removed);
	    allHitsRev.push_back(removed);
	  }
	  removedHits.clear();
	}// end if there are enough left over hits to look for another track
      }// end if the returned track was valid

    }// end loop to look for tracks in the slice

    // check to see if any of the tracks should be merged
    if(tracks.size() > 1){
      //LOG_VERBATIM("WindowTrackingAlg") << "there are " << tracks.size() << " tracks in the slice";
      TVector3 t1Start, t2Start, t1End, t2End, diff12, diff21;
      float t1ZDir = 0., t2ZDir = 0.;
      size_t trks = tracks.size();
      std::vector<bool> removed(trks, false);
      for(size_t tr1 = 0; tr1 < trks; ++tr1){
// 	int ctr = 0;
	if(removed[tr1]) continue;
	for(size_t tr2 = 0; tr2 < trks; ++tr2){
	  if(tr1 == tr2 || removed[tr2]) continue;

// 	  LOG_VERBATIM("WindowTrackingAlg") << "inner loop " << ctr;
// 	  ++ctr;
	  
	  // get the start and end points of each track
	  t1Start = tracks[tr1].Start();
	  t2Start = tracks[tr2].Start();
	  t1End   = tracks[tr1].Stop();
	  t2End   = tracks[tr2].Stop();

	  t1ZDir  = tracks[tr1].Dir().Z()/std::abs(tracks[tr1].Dir().Z());
	  t2ZDir  = tracks[tr2].Dir().Z()/std::abs(tracks[tr2].Dir().Z());

// 	  LOG_VERBATIM("WindowTrackingAlg") << "t1 start: " 
// 					    << t1Start.X() << " " << t1Start.Y() << " " << t1Start.Z() << " end: "
// 					    << t1End.X()   << " " << t1End.Y()   << " " << t1End.Z() << "\n"
// 					    << "t2 start: " 
// 					    << t2Start.X() << " " << t2Start.Y() << " " << t2Start.Z() << " end: "
// 					    << t2End.X()   << " " << t2End.Y()   << " " << t2End.Z();
	    
	  // now see if any of the end points match up
	  // that is they are within a cell of each other
	  // The total difference in the end points has to be 
	  // less than 2*fDHitGood and the z component of that
	  // should be less than fDHitGood
	  // also make sure they are going in the same z direction
	  if(t1ZDir*t2ZDir > 0.){
	    diff12 = t1Start - t2End;
	    diff21 = t2Start - t1End;
	    
	    if( (diff12.Mag() < 2.*fDHitGood && std::abs(diff12.Z()) < fDHitGood) ||
		(diff21.Mag() < 2.*fDHitGood && std::abs(diff21.Z()) < fDHitGood) ){
	      
	      // get the iterators in the correct order for the z direction of 
	      // the track
	      auto firstTr  = tracks[tr1];
	      auto secondTr = tracks[tr2];
	      if(t1Start.Z()*t1ZDir > t2Start.Z()*t2ZDir){
		firstTr  = tracks[tr2];
		secondTr = tracks[tr1];
	      }

	      // merge the tracks
	      std::vector<TVector3> mergedTrajectory;	      
	      // get a merged trajectory point for the intersection of the two tracks
	      // and set that to the be the first trajectory point in track 1, don't
	      // prepend last point from the trajectory of track 2
	      // the first two points and last two points in a track represent the edge 
	      // of and the midpoint of the final plane, so remove the last two points
	      // in firstTr and the first two points in secondTr from the merged track 
	      for(size_t tp = 0; tp < firstTr.Trajectory().size()-2; ++tp){
		mergedTrajectory.push_back(firstTr.TrajectoryPoint(tp));
// 		LOG_VERBATIM("WindowTrackingAlg") << mergedTrajectory.back().X() << " " 
// 						  << mergedTrajectory.back().Y() << " " 
// 						  << mergedTrajectory.back().Z();
	      }
	      mergedTrajectory.push_back(0.5*(secondTr.Trajectory().front() + firstTr.Trajectory().back()));
// 	      LOG_VERBATIM("WindowTrackingAlg") << "connection point " 
// 						<< mergedTrajectory.back().X() << " " 
// 						<< mergedTrajectory.back().Y() << " " 
// 						<< mergedTrajectory.back().Z();
	      for(size_t tp = 2; tp < secondTr.Trajectory().size(); ++tp){
		mergedTrajectory.push_back(secondTr.TrajectoryPoint(tp));
// 		LOG_VERBATIM("WindowTrackingAlg") << mergedTrajectory.back().X() << " " 
// 						  << mergedTrajectory.back().Y() << " " 
// 						  << mergedTrajectory.back().Z();
	      }

	      // now create the merged track with cells from both tracks
	      rb::Track mergedTrack(firstTr.AllCells(), mergedTrajectory.front(), firstTr.Dir());
	      for(auto c  : secondTr.AllCells()) mergedTrack.Add(c);	
	      for(size_t tp = 1; tp < mergedTrajectory.size(); ++tp) mergedTrack.AppendTrajectoryPoint(mergedTrajectory[tp]); 
	      tracks.push_back(mergedTrack);

// 	      LOG_VERBATIM("WindowTrackingAlg") << firstTr.Trajectory().size() << " " 
// 						<< secondTr.Trajectory().size() << " " 
// 						<< mergedTrajectory.size() << " trajectory points now: \n";
// 	      for(auto tp : tracks.back().Trajectory())
// 		LOG_VERBATIM("WindowTrackingAlg") << tp.X() << " " << tp.Y() << " " << tp.Z();		   

	      // remove both track 1 and 2
	      removed[tr1] = true;
	      removed[tr2] = true;
// 	      LOG_VERBATIM("WindowTrackingAlg") << "removing tracks " << tr1 << " " << tr2;
	    }// end if the tracks end points are close enough to merge
	  }// end if tracks are going in the same direction in Z
	}// end inner loop over tracks
      }// end outer loop over tracks

      // loop over the removed vector and remove those tracks from the vector
      size_t rmctr = 0;
      for(size_t r = 0; r < removed.size(); ++r){
	if(removed[r]){
	  tracks.erase(tracks.begin() + (r-rmctr));
	  ++rmctr;
	}
      }
//       LOG_VERBATIM("WindowTrackingAlg") << "there are " << tracks.size() << " tracks in the slice";
    }// end if more than one track in the slice

    return tracks;

  } // End of MakeTrack

  //......................................................................
  rb::Track WindowTrackingAlg::Make3DTrack(std::vector<art::Ptr<rb::CellHit> > & allCells,
					   std::vector<art::Ptr<rb::CellHit> > & removedHits,
					   double                              & chiSqr,
					   bool(*comp)(unsigned short int, unsigned short int))
  {
    // loop over all the hits for this slice and get the
    // x or y cells    
    std::map<unsigned short int, std::vector< art::Ptr<rb::CellHit> >, 
             bool(*)(unsigned short int, unsigned short int) >          xViewHits(comp);
    std::map<unsigned short int, std::vector< art::Ptr<rb::CellHit> >, 
             bool(*)(unsigned short int, unsigned short int) >          yViewHits(comp);
    std::map<unsigned short int, std::pair<float, float>,              
	     bool(*)(unsigned short int, unsigned short int) >          planeCoG(comp);

    double xyz[3] = {0.};
    double v      = 0.;

    // the allCells vector is already sorted properly.
    std::set<int> planes;
 
//     LOG_VERBATIM("WindowTrackingAlg") << "starting with these hits:";
    for(auto const chit : allCells){

      planes.insert(chit->Plane());

      fGeo->Plane(chit->Plane())->Cell(chit->Cell())->GetCenter(xyz);

//       LOG_VERBATIM("WindowTrackingAlg") << "\t" << chit->Plane() << " " << chit->Cell();
      
      if(chit->View() == geo::kX){
	xViewHits[chit->Plane()].push_back(chit);
	v = xyz[0];
      }
      if(chit->View() == geo::kY){
	yViewHits[chit->Plane()].push_back(chit);
	v = xyz[1];
      }

      if(planeCoG.find(chit->Plane()) != planeCoG.end()){
	planeCoG[chit->Plane()].first  += chit->ADC()*v;
	planeCoG[chit->Plane()].second += chit->ADC();
      }
      else
	planeCoG[chit->Plane()] = std::pair<float,float>(chit->ADC()*v,
							 chit->ADC());

    }// end loop over hits in this slice

    // loop over planeCoG and see what we have
//     for(auto pcog : planeCoG)
//       LOG_VERBATIM("WindowTrackingAlg") << pcog.first << " " << pcog.second.first/pcog.second.second;

    // remove planes that are too far separated from others in the slice
    auto prev = planes.begin();
    auto next = prev;
    bool removeThisPlane = false;

    for(auto itr = planes.begin(); itr != planes.end(); ++itr){
      next = itr; ++next;
      removeThisPlane = false;
      
      // is the first plane separated from the second by too much?
      if(itr == planes.begin() && next != planes.end()){
	if(std::abs(*itr - *next) > fMaxPlaneSeparation) removeThisPlane = true;
      }// end if on the first plane in the set
      // is the current plane separated by both the previous and next one 
      // too much
      else if(itr != planes.begin() && next != planes.end()){	
	prev = itr; --prev;
	if(std::abs(*itr - *prev) > fMaxPlaneSeparation && 
	   std::abs(*itr - *next) > fMaxPlaneSeparation) removeThisPlane = true;
      }
      else if(next == planes.end()){
	prev = itr; --prev; 
	if(std::abs(*itr - *prev) > fMaxPlaneSeparation) removeThisPlane = true;
      }

      if(removeThisPlane){
	xViewHits.erase(*itr);
	yViewHits.erase(*itr);
	planeCoG .erase(*itr);
// 	LOG_VERBATIM("WindowTrackingAlg") << "removing plane " << *itr << " from the fit; prev: "
// 					  << *prev << " next: " << *next;
      }

    }//end loop to look for big gaps between planes.

    rb::Track retTrack;

    if(this->ValidTrack(xViewHits) &&
       this->ValidTrack(yViewHits) ){

      double chiSqrX = 0.;
      double chiSqrY = 0.;

      rb::Track xTrack = FitView(xViewHits, planeCoG, removedHits, geo::kX, chiSqrX);
      rb::Track yTrack = FitView(yViewHits, planeCoG, removedHits, geo::kY, chiSqrY);            

      chiSqr = chiSqrX + chiSqrY;

      if(xTrack.View() != geo::kX || yTrack.View() != geo::kY){
	LOG_DEBUG("WindowTrackingAlg") << "no track found in at least one view, return empty track";
	return retTrack;
      }

      //LOG_DEBUG("WindowTrackingAlg") << " x track view: " << xTrack.View() 
      //	 			     << " y track view: " << yTrack.View();
      
      // try making a 3D track from these two 2D tracks
      try{
	retTrack = xTrack.ZipWith(yTrack);
      }
      catch(cet::exception &e){
	LOG_WARNING("WindowTrackingAlg") << "attempt to zip tracks failed with exception\n"
					 << e
					 << "\n return empty track";
	return retTrack;
      }

      TVector3 start(retTrack.Start());
      TVector3 end  (retTrack.Stop());
      
      this->CheckTrackDirectionInY(retTrack, start, end);
      
      // determine the direction of the track by fitting the first 
      // several trajectory points - just using the direction from the 
      // first point to the next can be misleading - need a bigger lever arm
      this->DetermineInitialDirection(retTrack);
      
      // restrict track to exist inside detector; if this isn't possible, 
      // the track is no good, so don't write
      if(geo::ClampRayToDetectorVolume(&start, &end, &(*fGeo))){
	//std::cout << "-------------------------------" << std::endl;
	//for(auto const& traj : retTrack.Trajectory()) traj.Print();
	return retTrack;
      }
      
    } // end if we can make a 3D track
    // see if we can make a short track - only worry about tracks that have at most fMinViewPlanes
    // in each view - if one view has 12 planes and the other 2, it is probably garbage anyway
    else if(xViewHits.size() < fMinViewPlanes && 
	    yViewHits.size() < fMinViewPlanes &&
	    xViewHits.size()  > 0             && 
	    yViewHits.size()  > 0 ){
//       LOG_VERBATIM("WindowTrackingAlg") << "attempting to fit a short track: "
//       					<< xViewHits.size() << " " << yViewHits.size();

      return this->ShortTrack(xViewHits, yViewHits, planeCoG, chiSqr);
    }
    else
      // can't make a track, so set the chiSqr value to something stupidly big 
      chiSqr = 1.e9;

    return retTrack;
  }
  
  //......................................................................
  rb::Track WindowTrackingAlg::FitView(std::map<unsigned short int, std::vector<art::Ptr<rb::CellHit> >,
				                bool(*)(unsigned short int, unsigned short int) >              & hits,
				       std::map<unsigned short int, std::pair<float, float>,
				                bool(*)(unsigned short int, unsigned short int) >         const& planeCoG,
				       std::vector<art::Ptr<rb::CellHit> >                                     & removedHits,
				       geo::View_t                                                               view,
				       double                                                                  & chiSqr)
  {
    // start off with an N plane window, then step out one plane at a time to add to the window
    size_t numPlanes = hits.size();

    // another vector that has all the hits for a given window
    std::vector< art::Ptr<rb::CellHit> > currWindow;
    std::vector< art::Ptr<rb::CellHit> > prevWindow;

    // get the iterator to the hits of the first plane in the window
    auto startItr = hits.begin();

    // get the iterator the hits of the last plane in the window
    auto endItr   = hits.begin();

    if(numPlanes > fWindowSize) std::advance(endItr, fWindowSize);
    else endItr = hits.end();

    bool dir = (startItr->first > endItr->first);

    // fit the initial window
    unsigned short int startPlane = startItr->second.front()->Plane();
    for(auto itr = startItr; itr != endItr; ++itr){
      for(size_t h = 0; h < itr->second.size(); ++h){
	currWindow.push_back(itr->second[h]);
// 	LOG_VERBATIM("WindowTrackAlg") << "adding to initial window: " << itr->second[h]->Plane() << " " << itr->second[h]->Cell();
      }
    }

    // this is the fit of the initial window, there is nothing in prevWindow
    // if all the hits are removed from currWindow, return an empty track
    this->FitWindow(prevWindow, currWindow, removedHits, planeCoG, view);
    if(currWindow.size() < 1){
      //LOG_DEBUG("WindowTrack") << "currWindow size is " << currWindow.size()
      //			  << " returning empty track";
      return rb::Track();
    }

    // put the current window's hits into the vector we'll use
    // to make the cluster
    std::vector< art::Ptr<rb::CellHit> > clHits(currWindow);

    // if there are no more planes to look at, just make the track
    // otherwise, endItr is already on the element past the initial
    // window
    if(endItr == hits.end() ){
      // make the track from the window
      return this->MakeViewTrack(clHits, view, chiSqr, dir, removedHits);
    }

    prevWindow.swap(currWindow);

    //LOG_DEBUG("WindowTrackingAlg") << "start sliding window";

    // slide the window until you run out of planes
    while(endItr != hits.end()){

      //LOG_DEBUG("WindowTrackingAlg") << "prev window size now " << prevWindow.size()
		// 		     << " curr window has size " << endItr->second.size()
// 				     << " starts with plane " << endItr->second.front()->Plane();
      
      // remove the first plane of hits from the previousWindow, but only if 
      // we added hits from the current window the last time
      auto itr = prevWindow.begin();
      while( itr != prevWindow.end() && currWindow.size() > 0){
	//LOG_DEBUG("WindowTrackingAlg") << (*itr)->Plane() << " " << (*itr)->Cell();
	if((*itr)->Plane() == startPlane) itr = prevWindow.erase(itr);
	else ++itr;
      }
      startPlane = prevWindow.front()->Plane();

      // put the current plane's hits to the the currWindow vector
      currWindow.swap(endItr->second);

      this->FitWindow(prevWindow, currWindow, removedHits, planeCoG, view);

      // add the surviving currWindow hits to the prevWindow and to the clHits
      for(size_t h = 0; h < currWindow.size(); ++h){
	prevWindow.push_back(currWindow[h]);
	clHits.push_back(currWindow[h]);
      }
    
      ++endItr;
    }// end loop over the windows

    // make the track using the hits that were found along the 
    // trajectory
    return this->MakeViewTrack(clHits, view, chiSqr, dir, removedHits);
  }

  //......................................................................
  rb::Track WindowTrackingAlg::MakeViewTrack(std::vector<art::Ptr<rb::CellHit> > const& trkHits,
					     geo::View_t                         const& view,
					     double                                   & chiSqr,
					     bool                                const& dir,
					     std::vector<art::Ptr<rb::CellHit> >      & removedHits)
  {
//     LOG_DEBUG("WindowTrackingAlg") << "view: " << view << " has " << trkHits.size()
//     			   << " cell hits";
    rb::Track retTrack;
    
    // make some trajectory points for this track - use the signal weighted
    // average point in each plane
    
    // first make a map of the planes to hits and then find the weighted average for
    // each plane, set the direction in plane number based on the input bool
    bool(*comp)(unsigned short int, unsigned short int) = (dir) ? gt : lt;
    std::map<unsigned short int, std::vector<art::Ptr<rb::CellHit> >, 
             bool(*)(unsigned short int, unsigned short int) > planesToHitLoc(comp);
  
    for(auto const hititr : trkHits)
      planesToHitLoc[hititr->Plane()].push_back(hititr);
  
    // check that you ended up with a reasonable track in terms 
    // of the number of hit planes compared to the number of planes crossed
    // set the chiSqr to something stupidly huge if that is not the case and
    // return an empty track
    if( !this->ValidTrack(planesToHitLoc) ){
      chiSqr = 1.e6;
      return retTrack;
    }

    std::vector<std::pair<double, double> > traj;
    double xyz[3] = {0.};
    double &w     = (view == geo::kX) ? xyz[0] : xyz[1];
    double avg    = 0.;
    chiSqr        = 0.;
    unsigned short int plane = 0;

    for(auto const mapitr : planesToHitLoc){      
      double weightedCells = 0.;
      double weights       = 0.;
      plane = mapitr.first;
      for(auto const cellitr : mapitr.second){
// 	LOG_VERBATIM("WindowTrackingAlg") << "makeviewtrack: " 
// 					  << cellitr->Plane() 
// 					  << " " << cellitr->Cell()
// 					  << " " << cellitr->ADC();
	fGeo->Plane(plane)->Cell(cellitr->Cell())->GetCenter(xyz);
	weightedCells += w*cellitr->ADC();
	weights       += 1.*cellitr->ADC();
	
      } //end loop over cell hits in the current plane
      
      // the z position is the same for all cells in a plane, and we are looping 
      // over all cells in each plane
      if(weights > 0.){
	avg = weightedCells/weights;
	
	for(auto const cellitr : mapitr.second){
	  fGeo->Plane(plane)->Cell(cellitr->Cell())->GetCenter(xyz);
	  chiSqr += std::pow(avg - w, 2.)/3.8; // cell height is 3.8 cm
	}

// 	LOG_DEBUG("WindowTrackingAlg") << "adding trajectory point in " 
// 	                              << view << " view: " << xyz[2] << " " << avg;

 	traj.push_back(std::pair<double, double>(avg, xyz[2]));
      }

    } // end loop over planesToHitLoc


    double dz = traj[1].second - traj[0].second;
    double dv = traj[1].first  - traj[0].first;

    //LOG_DEBUG("WindowTrackingAlg") << "there are " << traj.size() << " trajectory points in this view";

    retTrack = rb::Track(trkHits, view, traj[0].first, traj[0].second, dv, dz);
    for(size_t t = 1; t < traj.size(); ++t) 
      retTrack.AppendTrajectoryPoint(traj[t].first, traj[t].second);

    // look for bremstrahlung hits to add back in
    this->LookForBremsstrahlungHits(retTrack, removedHits);

    return retTrack;
  }

  //......................................................................
  // this method fits a straight line to the hits it is given and then drops
  // those that are too far off that line.
  unsigned int WindowTrackingAlg::FitWindow(std::vector<art::Ptr<rb::CellHit> >                        const& prevWindow,
					    std::vector<art::Ptr<rb::CellHit> >                             & currWindow, 
					    std::vector<art::Ptr<rb::CellHit> >                             & removedHits, 
					    std::map<unsigned short int, std::pair<float, float>,
					             bool(*)(unsigned short int, unsigned short int) > const& planeCoG,
					    geo::View_t                                                const& view)
  {
    size_t prevWinSize = prevWindow.size();

    double start[2] = {0.};
    double end[2]   = {0.};

    std::vector<double> x(prevWinSize + currWindow.size(), 0.);
    std::vector<double> y(prevWinSize + currWindow.size(), 0.);
    std::vector<double> w(prevWinSize + currWindow.size(), 0.);
    std::vector<size_t> planeBoundaries;
    double chiSqr   = 0.;

    // fill the vectors, first with the previous window, then the current one
    double xyz[3] = {0.};
    double &v     = (view == geo::kX) ? xyz[0] : xyz[1];
    for(size_t c = 0; c < prevWinSize; ++c){
      fGeo->Plane(prevWindow[c]->Plane())->Cell(prevWindow[c]->Cell())->GetCenter(xyz);
      x[c] = xyz[2];
      y[c] = v;
      w[c] = 2.;//prevWindow[c]->ADC();
//       LOG_VERBATIM("WindowTrackingAlg") << "previous window points: " 
// 				     << prevWindow[c]->Plane() << " " 
// 				     << prevWindow[c]->Cell()  << " " 
// 				     << x[c] << " " << y[c] << " " 
// 				     << w[c]; 

    }

    // fit the previous window to be able to figure out which cells
    // in the current window should go in the initial fit
    if(prevWinSize > 0)
      chiSqr = geo::LinFitMinDperp(x, y, w,
				   &start[0], &start[1],
				   &end[0],   &end[1]);

    unsigned int ndrop = 0;
    
    // variables to help identify outliers early on
    unsigned short int currPlane = currWindow.front()->Plane();
    unsigned short int prevPlane = currWindow.front()->Plane();
    float              cog       = 0;
    for(size_t c = 0; c < currWindow.size(); ++c){
      currPlane                = currWindow[c]->Plane();
      fGeo->Plane(currPlane)->Cell(currWindow[c]->Cell())->GetCenter(xyz);
      cog                      = planeCoG.find(currPlane)->second.first;
      cog                     /= planeCoG.find(currPlane)->second.second;
      x[c + prevWinSize]    = xyz[2];
      y[c + prevWinSize]    = v;
      w[c + prevWinSize]    = 1.;//currWindow[c]->ADC();

      if(currPlane != prevPlane) planeBoundaries.push_back(c);

      prevPlane = currPlane;

      // First pass, drop hits that are clear outliers (more than 2x tolerance
      // from anything else) in the current window. If it turns out they really 
      // are good hits (eg surrounded by bad channels) they will be added back 
      // in by the last step. But this helps prevent distant noise hits from 
      // throwing off the original fit.
      if(prevWinSize > 0){
	const double dSq = geo::DsqrToLine(x[c + prevWinSize],     
					   y[c + prevWinSize],
					   start[0], start[1],
					   end[0],   end[1]);

	if(dSq > util::sqr(fDHitGood*2)){
// 	  LOG_VERBATIM("WindowTrackingAlg") << "dropping hit at (z,v) = ("
// 					    << x[c + prevWinSize] << ","
// 					    << y[c + prevWinSize] 
// 					    << ") for being too separated from others: "
// 					    << dSq 
// 					    << " line start: " << start[0] << " " << start[1]
// 					    << " line end: "   << end[0]   << " " << end[1]
// 					    << " expected pos " 
// 					    << x[c + prevWinSize]*(end[1]-start[1])/(end[0]-start[0]) + (start[1] - start[0]*(end[1]-start[1])/(end[0]-start[0]));
	  
	  w[c + prevWinSize] = 0;
	  ++ndrop;
	}
      }// end if we have a previous window to check this hit against
      else{
	// set the weight to 0 of cells from the same plane
	if(std::abs(v - cog) > fDHitGood*4){
	  w[c + prevWinSize] = 0.;
	  ++ndrop;
	}
      }

//       LOG_VERBATIM("WindowTrackingAlg") << "current window points: " 
// 					<< currWindow[c]->Plane() << " " 
// 					<< currWindow[c]->Cell()  << " " 
// 					<< x[c + prevWinSize]     << " " 
// 					<< y[c + prevWinSize]     << " " 
// 					<< w[c + prevWinSize]     << " " 
// 					<< cog                    << " " 
// 					<< currWindow[c]->ADC();

    }
    
    // check if we have dropped all the hits from the current window
    // and have none in the previous window.  In that case, we are
    // on the first window of the given slice but have no way to do
    // a fit.  Remove all the hits from the current window
    if(ndrop == currWindow.size() && prevWinSize < 1){
      currWindow.clear();
      return ndrop;
    }

    // sometimes you get a plane in the initial window that has 
    // its hits well away from the others and that throws off the
    // fit.  Run through the initial window and drop each plane's
    // cells in turn.  If a plane causes the chi^2 to spike drop it
    if(prevWinSize < 1){

      // add the final boundary to the list
      planeBoundaries.push_back(w.size());

      size_t prevPB    = 0;
      double bestChi2  = 1.e6;
      double worstChi2 = 0.;
      std::pair<size_t, size_t> bestBounds(0, 0);
      for(auto const p : planeBoundaries){

	// set the weights for the plane to 0
	std::vector< std::pair<size_t, double> > prevWeights;
	for(auto pb = prevPB; pb < p; ++pb){
	  prevWeights.push_back(std::pair<size_t, double>(pb, w[pb]));
	  w[pb] = 0.;
	}

	// for(size_t c = prevWinSize; c < x.size(); ++c)
// 	  LOG_DEBUG("WindowTrackingAlg") << "look for outliers: " << p << " " 
// 					 << x[c + prevWinSize] << " " 
// 					 << y[c + prevWinSize] << " " 
// 					 << w[c + prevWinSize];
	
	try{
	  chiSqr = geo::LinFitMinDperp(x, y, w,
				       &start[0], &start[1],
				       &end[0],   &end[1]);
	}
	catch(cet::exception &e){

	  //LOG_DEBUG("WindowTrackingAlg") << "caught exception " << e
	  //			 << " and moving on to next plane";

	  // there were likely no non-zero weights, so just move on to the next plane
	  continue;
	}

	//LOG_DEBUG("WindowTrackingAlg") << "chisq = " << chiSqr;
	
	if(chiSqr < bestChi2){
	  bestChi2 = chiSqr;
	  bestBounds.first  = prevPB;
	  bestBounds.second = p;
	  //LOG_DEBUG("WindowTrackingAlg") << "best chisqr, bounds: " << bestBounds.first << " " << bestBounds.second;
	}
	if(chiSqr > worstChi2) worstChi2 = chiSqr;

	// set the weights back to 1 to evaluate the remaining planes
	for(auto pw : prevWeights) w[pw.first] = pw.second;

	// set the previous boundary to the current value
	prevPB = p;

      }// end loop over plane boundaries
      
      // set the worst plane contributions to 0 weight for now
      // but only do that if it adds 2 units of chi^2 or more
      // for each cell in the window
      if(worstChi2 - bestChi2 > 2.*w.size()) 
	for(auto b = bestBounds.first; b < bestBounds.second; ++b) w[b] = 0.;

    }// end if on the first window for a slice in the chosen view

    bool droppedAny = true;
    while(droppedAny){

      try{
	chiSqr = geo::LinFitMinDperp(x, y, w,
				     &start[0], &start[1],
				     &end[0],   &end[1]);
      }
      catch(cet::exception &e){
	
	// there were likely no non-zero weights, so stop trying to do a fit
	break;
      }

      //LOG_DEBUG("WindowTracking") << "end points from fit "
// 				  << start[0] << " " 
// 				  << start[1] << " " 
// 				  << end[0]   << " " 
// 				  << end[1]   << " chi^2 = "
// 				  << chiSqr;

      // loop over the hits and for the ones that are more than fDHitGood cm
      // off the line set the weight to be 0 so that they do not contribute to
      // future fits. Only drop the worst hit, and then iterate to get the next
      // worst and so on. Otherwise we can end up dropping lots of hits we
      // shouldn't due to one outlier.
      double dSqMax    = 0;
      int    dSqMaxIdx = -1;
      for(size_t n = prevWinSize; n < x.size(); ++n){
	// Already dropped
	if(w[n] == 0) continue;

	const double dSq = geo::DsqrToLine(x[n],  y[n],
					   start[0], start[1],
					   end[0],   end[1]);
	if(dSq > dSqMax){
	  dSqMax = dSq;
	  dSqMaxIdx = n;
	  //LOG_DEBUG("WindowTracking") << "highest dSq is at: "
	  //			 << x[n] << " " << y[n]
	  //			 << " " << dSq;
	}
      } // end for n

      droppedAny = false;
      if(dSqMax > util::sqr(fDHitGood)){
// 	LOG_VERBATIM("WindowTracking") << "dropping hit at ("
// 				       << x[dSqMaxIdx] << ", " 
// 				       << y[dSqMaxIdx] << ")"
// 				       << " dSq = " << dSqMax << " expected "
// 				       << x[dSqMaxIdx]*(end[1]-start[1])/(end[0]-start[0]) + (start[1] - start[0]*(end[1]-start[1])/(end[0]-start[0]));

	w[dSqMaxIdx] = 0.;
	droppedAny = true;
	++ndrop;
      }
    }// end while

    for(size_t n = prevWinSize; n < x.size(); ++n){
      // Already part of the fit
      if(w[n] != 0) continue;

      const double dSq = geo::DsqrToLine(x[n],  y[n],
					 start[0], start[1],
					 end[0],   end[1]);
      if(dSq < util::sqr(fDHitGood)){
	//LOG_DEBUG("WindowTracking") << "adding back hit at (" 
	//		       << x[n] << ", " << y[n]
	//		    << ")" << " dSq = " << dSq;
	
	w[n] = 1.;
	--ndrop;
      }
    } // end for n

    // remove any hits from the current window that have a weight of 0
    // erase all entries from w that are from the previous window
    auto witr = w.begin();
    if(prevWinSize > 0) 
      witr = w.erase(w.begin(), w.begin()+prevWinSize);
    auto citr = currWindow.begin();
    while( witr != w.end() && citr != currWindow.end()){
      if( (*witr) < 1. ){
//  	LOG_VERBATIM("WindowTrackingAlg") << "dropping hit at "
// 				       << (*citr)->Plane() << ","
// 				       << (*citr)->Cell();
	// add the hit to the removed hits vector
	removedHits.push_back(*citr);

	citr = currWindow.erase(citr);
	witr = w.erase(witr);
      }
      else{
	++witr;
	++citr;
      }
    }

    //LOG_DEBUG("WindowTracking") << "end points from fit "
    //			<< start[0] << " " << start[1] << " "
    //			<< end[0]   << " " << end[1];

    return ndrop;
  }

  //......................................................................
  void WindowTrackingAlg::LookForBremsstrahlungHits(rb::Track & track,
						    std::vector<art::Ptr<rb::CellHit> > & removedHits)
  {
    // map the z position of each trajectory point to the position in the view
    std::map<double, double> zToViewPos;

    for(auto tp : track.Trajectory()){
      if     (track.View() == geo::kX) zToViewPos[tp.Z()] = tp.X();
      else if(track.View() == geo::kY) zToViewPos[tp.Z()] = tp.Y();
    }// end loop to map z to trajectory in view

    // loop over the rejected cells to see if any of them are close enough
    // to the other cells in the plane to be added back to the track
    double xyz[3] = {0.};
    double &v     = (track.View() == geo::kX) ? xyz[0] : xyz[1];
    auto rhitr = removedHits.begin();
    bool useHit = false;
    while( rhitr != removedHits.end() ){
      useHit = false;
      auto c = *rhitr;
      fGeo->Plane(c->Plane())->Cell(c->Cell())->GetCenter(xyz);
      
      if(zToViewPos.count(xyz[2]) > 0){

	// if the cell is close enough to the trajectory at this location
	// add it to the track and remove it from the 
	if(std::abs(zToViewPos[xyz[2]] - v) < 8.*fDHitGood){
	  track.Add(c);
	  rhitr = removedHits.erase(rhitr);
	  useHit = true;
	}
      }// end if the z position is in the map

      // the erase method returns an iterator to the position
      // after the removed item, ie effectively incrementing the iterator
      // if we didn't remove the hit, we need to increment the iterator
      if(!useHit) ++rhitr;

    }// end loop over previously rejected hits    

    return;
  }


  //......................................................................
  void WindowTrackingAlg::DetermineInitialDirection(rb::Track & track)
  {

    size_t numTrajPts = (track.NTrajectoryPoints() > fNTrajectory) ? fNTrajectory : track.NTrajectoryPoints();

    // put the first n trajectory points into vectors
    std::vector<double> x(numTrajPts, 0.);
    std::vector<double> y(numTrajPts, 0.);
    std::vector<double> z(numTrajPts, 0.);
    std::vector<double> w(numTrajPts, 1.);
 
    TVector3 pos;
    for(size_t t = 0; t < numTrajPts; ++t){
      pos = track.TrajectoryPoint(t);
      x[t] = pos.X();
      y[t] = pos.Y();
      z[t] = pos.Z();      
    }

    double xzStart[2] = {0.};
    double xzEnd[2]   = {0.};
    double yzStart[2] = {0.};
    double yzEnd[2]   = {0.};

    geo::LinFitMinDperp(z, x, w, &xzStart[0], &xzStart[1], &xzEnd[0], &xzEnd[1]);
    geo::LinFitMinDperp(z, y, w, &yzStart[0], &yzStart[1], &yzEnd[0], &yzEnd[1]);

    if(xzStart[0] < z.front()) std::swap(xzStart, xzEnd);
    if(yzStart[0] < z.front()) std::swap(yzStart, yzEnd);

    TVector3 start(xzStart[1], yzStart[1], z.front());
    TVector3 end  (xzEnd[1],   yzEnd[1],   z.back());

    track.SetDir(end - start);
    
    return;
  }

  //......................................................................
  // method to make 3D tracks from views with only 1 or 2 hit planes in the slice
  rb::Track WindowTrackingAlg::ShortTrack(std::map<unsigned short int, std::vector<art::Ptr<rb::CellHit> >,
					           bool(*)(unsigned short int, unsigned short int) >              & xViewHits,
					  std::map<unsigned short int, std::vector<art::Ptr<rb::CellHit> >,
						   bool(*)(unsigned short int, unsigned short int) >              & yViewHits,
					  std::map<unsigned short int, std::pair<float, float>,
						   bool(*)(unsigned short int, unsigned short int) >         const& planeCoG,
					  double                                                                  & chiSqr)
  {

    // the assumption for these tracks is that they are very steep and leave lots of
    // hits in only at most 3 planes in a view.  Just use the two planes with the 
    // most hits to set the track direction and then add in the hits from the third 
    // plane if they fall along that line
    // make a copy of the input maps and then drop the planes with
    // the fewest hits in each view.  At this point there should 
    // be only two planes in each view
    auto cogCopy(planeCoG);    
    auto xHitsCopy(xViewHits);
    auto yHitsCopy(yViewHits);
    unsigned short int removedXPlane = 0;
    unsigned short int removedYPlane = 0;
    if(xViewHits.size() > 2) this->ShortTrackRemoveExtraPlane(xHitsCopy, cogCopy, removedXPlane);
    if(yViewHits.size() > 2) this->ShortTrackRemoveExtraPlane(yHitsCopy, cogCopy, removedYPlane);

    // make sure the two plane hit view has the planes separated by only 1 plane in the other view, and that 
    // the one hit plane view is between them
    int    firstXPlane  = xHitsCopy.cbegin()->first;
    int    secondXPlane = 0;
    int    firstYPlane  = yHitsCopy.cbegin()->first;
    int    secondYPlane = 0;
    float  firstXCoG    = cogCopy.find(firstXPlane)->second.first/cogCopy.find(firstXPlane)->second.second;
    float  firstYCoG    = cogCopy.find(firstYPlane)->second.first/cogCopy.find(firstYPlane)->second.second;
    float  secondXCoG   = 0.;
    float  secondYCoG   = 0.;
    double xyz[3]       = {0.};
    fGeo->Plane(xHitsCopy.cbegin()->first)->Cell((xHitsCopy.cbegin()->second.front())->Cell())->GetCenter(xyz);
    float  firstXZ      = xyz[2];
    fGeo->Plane(yHitsCopy.cbegin()->first)->Cell((yHitsCopy.cbegin()->second.front())->Cell())->GetCenter(xyz);
    float  firstYZ      = xyz[2];
    float  secondXZ     = xyz[2];
    float  secondYZ     = xyz[2];
    bool   twoXPlanes   = (xHitsCopy.size() == 2) ? true : false;
    bool   twoYPlanes   = (yHitsCopy.size() == 2) ? true : false;

    rb::Track retTrack;


    if(twoXPlanes){
      secondXPlane = xHitsCopy.crbegin()->first;
      secondXCoG   = cogCopy.find(secondXPlane)->second.first/cogCopy.find(secondXPlane)->second.second;
      fGeo->Plane(xHitsCopy.crbegin()->first)->Cell((xHitsCopy.crbegin()->second.front())->Cell())->GetCenter(xyz);
      secondXZ     = xyz[2];
    }
    if(twoYPlanes){
      secondYPlane = yHitsCopy.crbegin()->first;
      secondYCoG   = cogCopy.find(secondYPlane)->second.first/cogCopy.find(secondYPlane)->second.second;
      fGeo->Plane(yHitsCopy.crbegin()->first)->Cell((yHitsCopy.crbegin()->second.front())->Cell())->GetCenter(xyz);
      secondYZ     = xyz[2];
    }
//     LOG_VERBATIM("WindowTrackingAlg") << firstXZ << " " << firstXCoG << " " << secondXZ << " " << secondXCoG;
//     LOG_VERBATIM("WindowTrackingAlg") << firstYZ << " " << firstYCoG << " " << secondYZ << " " << secondYCoG;


    bool makeTrack = false;

    // check that the planes in each view are contiguous
    if(twoXPlanes && !twoYPlanes){
      if(std::abs(firstXPlane  - secondXPlane) < 3 &&
	 std::abs(firstXPlane  - firstYPlane)  < 2 &&
	 std::abs(secondXPlane - firstYPlane)  < 2)
	makeTrack = true;
    }
    else if(!twoXPlanes && twoYPlanes){
      if(std::abs(firstYPlane  - secondYPlane) < 3 &&
	 std::abs(firstYPlane  - firstXPlane)  < 2 &&
	 std::abs(secondYPlane - firstXPlane)  < 2)
	makeTrack = true;
    }
    else if(twoXPlanes && twoYPlanes){
      if(std::abs(firstYPlane  - secondYPlane) < 3 &&
	 std::abs(firstXPlane  - secondXPlane) < 3 &&
	 std::abs(firstYPlane  - firstXPlane ) < 2 &&
	 std::abs(secondYPlane - secondXPlane) < 2)
	makeTrack = true;
    }

    if(makeTrack){

//       LOG_VERBATIM("WindowTrackingAlg") << "making the short tracks in each view";

      rb::Track xTrack = this->ShortViewTrack(xHitsCopy, cogCopy, twoXPlanes, firstXPlane, secondXPlane, geo::kX);
      rb::Track yTrack = this->ShortViewTrack(yHitsCopy, cogCopy, twoYPlanes, firstYPlane, secondYPlane, geo::kY);

      // for a short track all the points are used, make the chiSqr 1
      chiSqr = 1.;

      // don't use ZipWith to make the 3D track - it is simple enough that
      // we can do it here
      // get the direction cosines in dx/dz and dy/dz
      TVector3 dxdz = xTrack.Dir();
      TVector3 dydz = yTrack.Dir();

      // figure out the trajectory points for each plane
      std::vector<std::pair<geo::View_t, TVector3> > trajPoints(4);
      trajPoints[0] = std::pair<geo::View_t, TVector3>(geo::kX, TVector3(firstXCoG,  0.,         firstXZ ));
      trajPoints[1] = std::pair<geo::View_t, TVector3>(geo::kY, TVector3(0.,         firstYCoG,  firstYZ ));
      trajPoints[2] = std::pair<geo::View_t, TVector3>(geo::kX, TVector3(secondXCoG, 0.,         secondXZ));
      trajPoints[3] = std::pair<geo::View_t, TVector3>(geo::kY, TVector3(0.,         secondYCoG, secondYZ));

      // remove the second point in a view if it doesn't exist
      if(!twoXPlanes) trajPoints.erase(trajPoints.begin()+2);
      if(!twoYPlanes) trajPoints.erase(trajPoints.begin()+trajPoints.size()-1);

      // check if we need to add in an extra plane in either view
      if(xViewHits.size() > 2) this->ShortTrackExtraPlane(xViewHits, planeCoG, trajPoints, xTrack, removedXPlane);
      if(yViewHits.size() > 2) this->ShortTrackExtraPlane(yViewHits, planeCoG, trajPoints, yTrack, removedYPlane);

      // now sort the list according to the z value of the planes
      if(cogCopy.cbegin()->first < cogCopy.crbegin()->first) 
	std::sort(trajPoints.begin(), trajPoints.end(), ltvec);
      else
	std::sort(trajPoints.begin(), trajPoints.end(), gtvec);

      double deltaZ = 0.;
      TVector3 neighbor;
      geo::View_t view = geo::kX;
      // loop over the trajectory points and set the x & y positions
      // be sure to get the closest neighboring trajectory point in the
      // opposite view to be able to set the 3D positions correctly
      // and use the direction cosines from the 2D tracks
      for(size_t tp = 0; tp < trajPoints.size(); ++tp){
	view = trajPoints[tp].first;
	if(tp < 1){
	  deltaZ = trajPoints[tp].second.Z() - trajPoints[tp+1].second.Z();
	  neighbor = trajPoints[tp+1].second;
	}
	else{       
	  deltaZ = trajPoints[tp].second.Z() - trajPoints[tp-1].second.Z();
	  neighbor = trajPoints[tp-1].second;
	}
	if     (view == geo::kX) trajPoints[tp].second.SetY(neighbor.Y() + deltaZ*dydz.Y()/dydz.Z());
	else if(view == geo::kY) trajPoints[tp].second.SetX(neighbor.X() + deltaZ*dxdz.X()/dxdz.Z());

	//trajPoints[tp].second.Print();
      } // end loop over trajectory points

      // make the 3D track
      rb::Track retTrack(xTrack.AllCells(), 
			 trajPoints.front().second, 
			 TVector3(trajPoints.back().second.X()-trajPoints.front().second.X(),
				  trajPoints.back().second.Y()-trajPoints.front().second.Y(),
				  trajPoints.back().second.Z()-trajPoints.front().second.Z())
			 );
      retTrack.Add(yTrack.AllCells());
      for(size_t tp = 1; tp < trajPoints.size(); ++tp) retTrack.AppendTrajectoryPoint(trajPoints[tp].second);

      
      this->CheckTrackDirectionInY(retTrack, trajPoints.front().second, trajPoints.back().second);

      // restrict track to exist inside detector; if this isn't possible, 
      // the track is no good, so don't write
      if(geo::ClampRayToDetectorVolume(&trajPoints.front().second, 
				       &trajPoints.back().second, 
				       &(*fGeo))) return retTrack;	
      //else LOG_DEBUG("WindowTrackAlg") << "Cannot clamp short track to detector";

    }// end if this works

    return retTrack;
      
  }

  //----------------------------------------------------------------------
  // method to determine which plane in a set of 3 (for a short track) has the least signal
  // and should be removed to determine the track
  void WindowTrackingAlg::ShortTrackRemoveExtraPlane(std::map<unsigned short int, std::vector<art::Ptr<rb::CellHit> >,
						        bool(*)(unsigned short int, unsigned short int) >       & viewHits,
                                                     std::map<unsigned short int, std::pair<float, float>,
						        bool(*)(unsigned short int, unsigned short int) >       & planeCoG,
						     unsigned short int                                         & removedPlane)
  {
    unsigned short int minPlane = 0;
      
    size_t minHits = 1e6;
    for(auto vh : viewHits){
      if(vh.second.size() < minHits){
	minPlane = vh.first;
	minHits  = vh.second.size();
      }
    }
    
    for(auto cogItr : planeCoG){
      if(cogItr.first == minPlane){
	removedPlane = cogItr.first;
	viewHits.erase(removedPlane);
	planeCoG.erase(removedPlane);
      }
    }

    return;
  }


  //----------------------------------------------------------------------
  rb::Track WindowTrackingAlg::ShortViewTrack(std::map<unsigned short int, std::vector<art::Ptr<rb::CellHit> >,
					               bool(*)(unsigned short int, unsigned short int) >       & viewHits,
					      std::map<unsigned short int, std::pair<float, float>,
					               bool(*)(unsigned short int, unsigned short int) >  const& planeCoG,
                                              bool                                                        const& twoPlanes,
					      int                                                         const& firstPlane,
					      int                                                         const& secondPlane,
					      geo::View_t                                                 const& view)

  {
      // make 2d tracks using the center of gravity for each plane
      std::vector<art::Ptr<rb::CellHit> > hits;
      for(auto itr : viewHits){
	for(auto ch : itr.second) hits.push_back(ch);
      }

      double xyz1[3] = {0.};
      double xyz2[3] = {0.};
      auto chVecItr = viewHits.cbegin()->second.cbegin();
      fGeo->Plane(firstPlane) ->Cell((*chVecItr)->Cell())->GetCenter(xyz1);
      float cog1 = planeCoG.find(firstPlane) ->second.first/planeCoG.find(firstPlane) ->second.second;
      float deltaW = 0.;
      float deltaZ = 2.*fCellHalfDepth;
      if(twoPlanes){
	chVecItr = viewHits.crbegin()->second.cbegin();
	fGeo->Plane(secondPlane)->Cell((*chVecItr)->Cell())->GetCenter(xyz2);
	deltaW = (planeCoG.find(secondPlane)->second.first/planeCoG.find(secondPlane)->second.second);
	deltaZ = xyz2[2]-xyz1[2];
      }
      else{
	// hits only on one plane, so use the center of the first and last cells in the plane to find the 
	// spread in the current view
	chVecItr = viewHits.cbegin()->second.cbegin();
	fGeo->Plane(firstPlane)->Cell((*chVecItr)->Cell())->GetCenter(xyz1);
	cog1 = (view == geo::kX) ? xyz1[0] : xyz1[1];
	fGeo->Plane(firstPlane)->Cell((viewHits.cbegin()->second.back())->Cell())->GetCenter(xyz2);
	deltaW  = (view == geo::kX) ? xyz2[0] : xyz2[1];
      }
      deltaW -= cog1;

      // be sure we don't just have one cell in the view
      if(deltaW == 0.) deltaW = fCellHalfDepth;

//       LOG_VERBATIM("WindowTrackingAlg") << view << " " << cog1 << " " << xyz1[2] << " " << deltaW << " " << deltaZ;

      rb::Track track(hits, view, cog1, xyz1[2], deltaW, deltaZ);
      track.AppendTrajectoryPoint(deltaW+cog1, deltaZ);

      return track;
  }

  //......................................................................
  // method to add back in an extra plane to a short track - short tracks 
  // are initially found using only the two planes with the most signal, it is 
  // possible there is another plane to add back in
  void WindowTrackingAlg::ShortTrackExtraPlane(std::map<unsigned short int, std::vector<art::Ptr<rb::CellHit> >,
					          bool(*)(unsigned short int, unsigned short int) >  const& viewHits,
                                               std::map<unsigned short int, std::pair<float, float>,
					          bool(*)(unsigned short int, unsigned short int) >  const& planeCoG,
					       std::vector<std::pair<geo::View_t, TVector3> >             & trajPoints,
					       rb::Track                                                  & viewTrack,
                                               unsigned short int                                    const& removedPlane)
  {
    
    // get the expected trajectory point for that view
    double xyz[3] = {0.};
    fGeo->Plane(removedPlane)->Cell(0)->GetCenter(xyz);
    TVector3 trajPoint;
    float dirZ      = viewTrack.Dir().Z();
    float dirV      = 0.;
    float cog       = planeCoG.find(removedPlane)->second.first/planeCoG.find(removedPlane)->second.second;
    float expectedV = 0.;
    float dV        = 0.;
    if(viewTrack.View() == geo::kX){
      trajPoint = trajPoints[2].second;
      dirV      = viewTrack.Dir().X();
      expectedV = trajPoint.X() + (dirV/dirZ)*std::abs(xyz[2] - trajPoint.Z());
      dV        = cog - trajPoint.X();

      // set the trajectory point for this plane in case we want to keep it
      trajPoint.SetXYZ(cog, 0., xyz[2]);
    }
    else{
      trajPoint = trajPoints[3].second;
      dirV      = viewTrack.Dir().Y();
      expectedV = trajPoint.Y() + (dirV/dirZ)*std::abs(xyz[2] - trajPoint.Z());
      dV        = cog - trajPoint.Y();

      // set the trajectory point for this plane in case we want to keep it
      trajPoint.SetXYZ(0., cog, xyz[2]);
    }
	
    // compare the expected trajectory location to the center of gravity for that plane, and if they are 
    // within 8*fDHitGood, ie a very steep track, add it back in
    if(std::abs(expectedV - cog) < 8*fDHitGood && dV*dirV > 0){
      trajPoints.push_back(std::pair<geo::View_t, TVector3>(viewTrack.View(), trajPoint));
      for(auto h : viewHits.at(removedPlane)) viewTrack.Add(h);
    }

    return;
  }


  //......................................................................
  // check that the track is going the correct direction for a 
  // cosmic ray, ie starts at high y and ends at low y.  The end 
  // points have to vary by at least fDHitGood to trigger a switch
  void WindowTrackingAlg::CheckTrackDirectionInY(rb::Track & track,
						 TVector3  & start,
						 TVector3  & end)
  {
    
    if(start.Y() < end.Y()){
      
      TVector3 pt2 = track.TrajectoryPoint(track.NTrajectoryPoints()-2);
      rb::Track ctrack(track.AllCells(), end, pt2-end);
      for(int p = track.NTrajectoryPoints()-2; p >= 0; --p)
	ctrack.AppendTrajectoryPoint(track.TrajectoryPoint(p));
      
      track = ctrack;
      
      start = ctrack.Start();
      end   = ctrack.Stop();
    }// end if we have to flip the direction on this track

    return;
  }

  //......................................................................
  // this method is for a single view, XZ or YZ
  bool WindowTrackingAlg::ValidTrack(std::map<unsigned short int, std::vector<art::Ptr<rb::CellHit> >,
			             bool(*)(unsigned short int, unsigned short int) >        const& planeToHit)
  {
    if(planeToHit.size() < fMinViewPlanes) return false;
  
    unsigned int const& startPlane = planeToHit.cbegin()->first;
    unsigned int const& endPlane   = planeToHit.crbegin()->first;				     
  
    // the 0.5 in the MinHitFraction accounts for the fact that
    // every other plane is X or Y as we are using the full extent
    // of the slice
    if(planeToHit.size() < fMinHitFraction*0.5*std::abs(1.*startPlane - 1.*endPlane) )
       return false;

    return true;
  } 

  //----------------------------------------------------------------------------
  void WindowTrackingAlg::SetTrackEndPoints(rb::Track & track)
  {
    // look for any trajectory points that have a Nan
//     std::cout << " checking out the track before endpoints set" << std::endl;
//     for(auto tp : track.Trajectory())
//       if(std::isnan(tp.X()) ||
// 	 std::isnan(tp.Y()) ||
// 	 std::isnan(tp.Z()) ) tp.Print();

    // extrapolate the track along the direction cosines
    // first get the cells from the first and last planes
    std::vector<art::Ptr<rb::CellHit> > startPlaneCells;
    std::vector<art::Ptr<rb::CellHit> > endPlaneCells;
    unsigned int minPlane = track.MinPlane();
    unsigned int maxPlane = track.MaxPlane();
    for(auto ch : track.AllCells() ){
      if(ch->Plane() == maxPlane) endPlaneCells.push_back(ch);
      if(ch->Plane() == minPlane) startPlaneCells.push_back(ch);
    }      

    std::sort(startPlaneCells.begin(), startPlaneCells.end(), sortCellPtr);
    std::sort(endPlaneCells.begin(),   endPlaneCells.end(),   sortCellPtr);
 
    // figure out the direction from the next to last point to the last point
    TVector3 endPt  = track.Trajectory().back();
    TVector3 endDir = endPt - track.Trajectory()[track.NTrajectoryPoints()-2];

    // figure out the direction from the second point to the first point
    TVector3 startPt  = track.Trajectory().front();
    TVector3 startDir = track.Trajectory()[1] - startPt;

    geo::View_t endPlaneView   = fGeo->Plane(maxPlane)->View();
    geo::View_t startPlaneView = fGeo->Plane(minPlane)->View();

    // determining which is the startPlane of the track and the endPlane
    // depends on the z direction.  Swap the min and max vectors if the
    // direction cosine in z is negative
    if(endDir.Z() < 0.){
      startPlaneCells.swap(endPlaneCells);
      std::swap(startPlaneView, endPlaneView);
    }

    // get the highest cell in the start plane and the lowest in the
    // end plane
    unsigned short int startCell = startPlaneCells.back()->Cell();
    unsigned short int endCell   = endPlaneCells.front()->Cell();
    // if the start or end plane is in the X view, then we need to 
    // determine the extrema for the cells based on the track 
    // trajectory
    if(startPlaneView == geo::kX && startDir.X() > 0.) startCell = startPlaneCells.front()->Cell();
    if(endPlaneView   == geo::kX && endDir.X()   > 0.) endCell   = endPlaneCells.back()->Cell();
      
    // point the track back from the starting trajectory point to 
    // a better location accounting for the potential difference between 
    // the center of gravity in a plane and its final cell in the plane
    double   xyz[3]    = {0.};
    fGeo->Plane(startPlaneCells.front()->Plane())->Cell(startCell)->GetCenter(xyz);
    this->FindEndPoint(startPlaneView, xyz, track.Start(), track.Dir(), true, track);

    // carry the track further in its direction of travel
    fGeo->Plane(endPlaneCells.front()->Plane())->Cell(endCell)->GetCenter(xyz);
    this->FindEndPoint(endPlaneView, xyz, track.Stop(), track.StopDir(), false, track);

    return;
  }

  //----------------------------------------------------------------------------
  void  WindowTrackingAlg::FindEndPoint(geo::View_t const& view, 
					double           * cellCenter,
					TVector3    const& endPoint,
					TVector3    const& epDir,
					bool        const& start,
					rb::Track        & track)
  {
    double deltaZ   = fGeo->Plane(0)->Cell(0)->HalfD();

    double  &v         = cellCenter[0];	    
    double   epV       = endPoint.X(); 
    double   epVDir    = epDir.X();
    int      signDirV  = 1.;

    std::string startEnd("start");
    if(!start) startEnd = "end";
    
    if(view == geo::kY){
      v      = cellCenter[1];	    
      epV    = endPoint.Y(); 
      epVDir = epDir.Y();
    }

//     LOG_VERBATIM("WindowTrackingAlg") << startEnd << " point: " 
// 				      << endPoint.X() << " " << endPoint.Y() << " " << endPoint.Z()
// 				      << " " << epDir.X() << " " << epDir.Y() << " " << epDir.Z()
// 				      << " cell at " 
// 				      << v << " " << cellCenter[2] << " " << deltaZ;

    // bail if the direction in the view is 0
    // the current end point will have to be good enough
    if(epVDir == 0.) return;

    double deltaV    = std::abs(v - epV);
    TVector3 dV(epDir.X()/epVDir, epDir.Y()/epVDir, epDir.Z()/epVDir); 
    // point the track backwards
    if     (epVDir > 0. &&  start) signDirV = -1;
    else if(epVDir < 0. && !start) signDirV = -1;
    TVector3 pt(endPoint.X() + deltaV*signDirV*dV.X(),
		endPoint.Y() + deltaV*signDirV*dV.Y(),
		endPoint.Z() + deltaV*signDirV*dV.Z());
        
    // only prepend if the point is different from the original starting point
    if(pt != endPoint){
      
      // if the new start point is more than half a cell width 
      // from the z position of the first cell in the start plane
      // just carry the track to the boundary of the plane
      if(std::abs(pt.Z() - cellCenter[2]) > deltaZ){
	signDirV = 1.;
	if     (epDir.Z() > 0. &&  start) signDirV = -1.;
	else if(epDir.Z() < 0. && !start) signDirV = -1.;
	pt.SetXYZ(endPoint.X() + deltaZ*signDirV*epDir.X()/epDir.Z(), 
		  endPoint.Y() + deltaZ*signDirV*epDir.Y()/epDir.Z(),
		  endPoint.Z() + deltaZ*signDirV);
      }

      // check that we don't have nans for the new coordinates
      if( !std::isnan(pt.X()) &&
	  !std::isnan(pt.Y()) &&
	  !std::isnan(pt.Z())){
	if(start) track.PrependTrajectoryPoint(pt);
	else      track.AppendTrajectoryPoint(pt);
	
// 	LOG_VERBATIM("WindowTrackingAlg") << "new " << startEnd << " point is: " 
// 					  << pt.X() << " " 
// 					  << pt.Y() << " " 
// 					  << pt.Z();

      }

    } // end if the new point and the previous endPoint are different

    return;
  }


}



