//////////////////////////////////////////////////////////////////////////////
/// \brief   Filter events based on their run/event numbers
/// \author  messier@indiana.edu
/// \date
//////////////////////////////////////////////////////////////////////////////

#include <cstdlib>
#include <iostream>
#include <fstream>
#include <string>
#include <vector>

#include "art/Framework/Core/EDFilter.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h"
#include "RecoBase/FilterList.h"
#include "RecoBase/Cluster.h"


namespace util {
  /// Select events based on a list of run and event numbers
  ///
  /// File format is expected to be a list of run/subrun/event triplets:
  /// 
  /// # run subrun event
  /// 10796  4     102012
  /// 10796  2     121093
  /// 10796  11    133092
  /// 
  /// with comments marked by a leading '#'. By default the module
  /// looks for these to be in a file called "event-list.txt",
  /// although the filename can be reconfigured. The module can be
  /// configured to run in one of two modes. The configuration
  /// include_runeventfilter sets Mode=0 which means that the events
  /// listed in the file are to be included in the output. The
  /// configuration exclude_runeventfilter sets Mode=1 which means
  /// that the events listed are to be excluded from the output
  /// file. See RunEventFilter.fcl.
  ///
  class RunEventFilter : public art::EDFilter 
  {
  public:
    explicit RunEventFilter(fhicl::ParameterSet const& p);
    virtual ~RunEventFilter();

    virtual void reconfigure(fhicl::ParameterSet const& p);
    
    virtual bool filter(art::Event& e);

  private:
    std::string               fEventFile;   ///< List of event numbers
    int                       fMode;        ///< 0=include, 1=exclude list
    bool                      fFilterSlice; ///< Apply filter at slice level
    std::string               fSlicerLabel; ///< Module that produced slices
    std::vector<unsigned int> fRun;         ///< List of run numbers
    std::vector<unsigned int> fSubrun;      ///< List of subrun numbers
    std::vector<unsigned int> fEvent;       ///< List of event numbers
    std::vector<unsigned int> fSlice;       ///< List of slice numbers
  };
}


////////////////////////////////////////////////////////////////////////
namespace util
{
  //......................................................................
  RunEventFilter::RunEventFilter(fhicl::ParameterSet const& p) 
  {
    this->reconfigure(p);
    
    if (fFilterSlice){
      produces< rb::FilterList<rb::Cluster> >();
    }
  }

  //......................................................................
  void RunEventFilter::reconfigure(fhicl::ParameterSet const& p)
  {
    fEventFile = p.get<std::string>("EventFile");
    fMode      = p.get<int>("Mode");
    fFilterSlice = p.get<bool>("FilterSlice");
    fSlicerLabel = p.get<std::string>("SlicerLabel");

    //
    // Reconfigure the run and event list given the new file name.
    //
    fRun.   clear();
    fSubrun.clear();
    fEvent. clear();
    fSlice. clear();
    std::ifstream ifs;
    ifs.open(fEventFile.c_str());
    if (ifs.is_open()) {
      while (ifs.good()) {
	std::string buff;
	std::getline(ifs, buff);
	if (buff[0]=='#') continue;
      
	int run, subrun, event, slice;
        if (fFilterSlice){
	  sscanf(buff.c_str(), "%d %d %d %d", &run, &subrun, &event, &slice);
          fSlice.push_back(slice);
        }
        else {
	  sscanf(buff.c_str(), "%d %d %d", &run, &subrun, &event);
        }

	fRun.   push_back(run);
	fSubrun.push_back(subrun);
	fEvent. push_back(event);
      }
      ifs.close();
    }
    else {
      std::cerr << __FILE__ << ":" << __LINE__ 
		<< " Failed to open file " << fEventFile << " for read" 
		<< std::endl;
    }
  }

  //......................................................................
  RunEventFilter::~RunEventFilter() 
  { 
  }

  //......................................................................
  bool RunEventFilter::filter(art::Event& evt) 
  {
    std::unique_ptr< rb::FilterList<rb::Cluster> > filtcol(new rb::FilterList<rb::Cluster>);

    //prepare slice index vector
    std::vector<unsigned int> slicelist;
    art::Handle< std::vector<rb::Cluster> > slices;
    if (fFilterSlice){
      evt.getByLabel(fSlicerLabel, slices);
      for(unsigned int i=0; i<slices->size(); ++i){
        slicelist.push_back(i);
      }
    }

    //
    // Look for the event in the list
    //
    bool inlist = false;
    for (unsigned register int i=0; i<fRun.size(); ++i) {
      if ((fRun[i]==evt.run()) &&
	  (fSubrun[i]==evt.subRun()) &&
	  (fEvent[i]==evt.id().event())) {
	inlist = true;
        if (fFilterSlice){
          slicelist.erase(slicelist.begin()+fSlice[i]);
        }
      }
    }

    //
    // Check if we should pass or fail the event
    //
    if (fMode==0) {
      //
      // Mode=0 means pass only the events in the list
      //
      if (fFilterSlice){
        for(unsigned int i=0; i<slicelist.size(); ++i){
          filtcol->Add(slices, slicelist[i]);
        }
        evt.put(std::move(filtcol));
      }

      if (inlist)  return true;
      else         return false;
    }
    else if (fMode==1) {
      //
      // Mode=1 means pass only the events not in the list
      //

      //don't filter the slices
      if (fFilterSlice){
        evt.put(std::move(filtcol));
      }

      if   (inlist) return false;
      else          return true;
    }
  
    //
    // We only reach this section of the code on an error condition
    //
    std::cerr << __FILE__ << ":" << __LINE__ 
	      << " Illegal mode = " << fMode << std::endl;
    abort();

    return false;
  }

} // end namespace util


////////////////////////////////////////////////////////////////////////
namespace util 
{
  DEFINE_ART_MODULE(RunEventFilter);
}
////////////////////////////////////////////////////////////////////////

