////////////////////////////////////////////////////////////////////////
/// \brief   DAQ2RawDigit validation
/// \author  Jarek Nowak
/// \date
////////////////////////////////////////////////////////////////////////

// C/C++ includes
#include <vector>
#include <iostream>
#include <map>
#include <cmath>

// ROOT includes
#include "TH1.h"

// Framework includes
#include "art/Framework/Core/EDAnalyzer.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "art/Framework/Services/Optional/TFileService.h"
#include "messagefacility/MessageLogger/MessageLogger.h"

// NOvASoft includes
#include "RawData/RawDigit.h"


namespace validation {
  class ValidateDAQ2RawDigit : public art::EDAnalyzer {
  public:
    explicit ValidateDAQ2RawDigit(fhicl::ParameterSet const& pset); 
    ~ValidateDAQ2RawDigit();                        
    
    void analyze(art::Event const& evt); 
    
    void reconfigure(const fhicl::ParameterSet& pset);
    
    
    void beginJob();
    void endJob();
    
  private:
    ///< Sample configurable module parameters
    std::string      fInputModuleLabel; ///> label for module that made the data products you need for this module 
    TH1F*            fNDigits;
    TH1F*            fNADC;
    TH1F*            fADC;
    
  };
}


////////////////////////////////////////////////////////////////////////

namespace validation{

  ValidateDAQ2RawDigit::ValidateDAQ2RawDigit(fhicl::ParameterSet const& pset) 
    : EDAnalyzer(pset)
  {
    this->reconfigure(pset);
    mf::LogInfo("Validation") << " Validation::ValidateDAQ2RawDigit()\n";
  }

  //......................................................................
  ValidateDAQ2RawDigit::~ValidateDAQ2RawDigit()
  {
  }

  //......................................................................
  void ValidateDAQ2RawDigit::reconfigure(const fhicl::ParameterSet& pset)
  {
    fInputModuleLabel = pset.get< std::string >("MyInputModuleLabel", "daq");
  }
      
  //......................................................................
  void ValidateDAQ2RawDigit::beginJob()
  {
    
    art::ServiceHandle<art::TFileService> tfs;

    fADC      = tfs->make<TH1F>("fADC",  "ADC; ADC; digits",    1000,0.,1000.);
    fNADC     = tfs->make<TH1F>("fNADC",    "NADC; NADC ; Events", 10,0.,10.);
    fNDigits  = tfs->make<TH1F>("fNDigits", "RawDigits; Number of raw digits; Events", 500,0.,5000.);
  }

  //......................................................................
  void ValidateDAQ2RawDigit::analyze(art::Event const & evt)
  {
    
    //  get rawdata::RawDigits from the event
    art::Handle< std::vector<rawdata::RawDigit> > digitcol;
    evt.getByLabel(fInputModuleLabel, digitcol);
    art::PtrVector<rawdata::RawDigit> digits;
    fNDigits->Fill(digitcol->size());
    std::cout<<digitcol->size()<<std::endl;
    for(unsigned int i = 0; i < digitcol->size(); ++i){
      art::Ptr<rawdata::RawDigit> dig(digitcol, i);
      digits.push_back(dig);
    }
    
    // make the collection of ints we want to store
    std::unique_ptr< std::vector<int> > adccol(new std::vector<int>);
    
    // fill a histogram with the size of each MCTruth
    for(unsigned int i = 0; i < digits.size(); ++i){
      
      fNADC->Fill(digits[i]->NADC());
      
      for(unsigned int adc = 0; adc < digits[i]->NADC(); ++adc){
	adccol->push_back( digits[i]->ADC(adc) );
	fADC->Fill( digits[i]->ADC(adc) );
      }
    }
    
    return;
  }
  
  void ValidateDAQ2RawDigit::endJob()
  {
    mf::LogInfo("Validation") << " Validation::ValidateDAQ2RawDigit() ---DONE\n";
    
  }

} //namespace validation


////////////////////////////////////////////////////////////////////////
namespace validation
{
  DEFINE_ART_MODULE(ValidateDAQ2RawDigit);
}
