////////////////////////////////////////////////////////////////////////
/// \file  FLSHitListAction.cxx
/// \brief Use Geant4's user "hooks" to maintain a list of FLSHits
///
/// \version $Id: FLSHitListAction.cxx,v 1.24 2012-09-20 21:45:42 greenc Exp $
/// \author  brebel@fnal.gov
////////////////////////////////////////////////////////////////////////
#include "g4nova/FLSHitListAction.h"
#include "g4nova/ParticleListAction.h"
#include "Simulation/FLSHit.h"
#include "Simulation/FLSHitList.h"

#include <vector>
#include <map>

// ROOT includes
#include "TGeoMaterial.h"
#include <TGeoManager.h>

// G4 includes
#include "Geant4/G4Event.hh"
#include "Geant4/G4Track.hh"
#include "Geant4/G4Step.hh"
#include "Geant4/G4EnergyLossForExtrapolator.hh"
#include "Geant4/globals.hh"

// ART includes
#include "messagefacility/MessageLogger/MessageLogger.h"
#include "cetlib/exception.h"

#include "G4Base/UserActionFactory.h"
USERACTIONREG3(g4n,FLSHitListAction,g4n::FLSHitListAction)

#include "GeometryObjects/CellGeo.h"
#include "GeometryObjects/PlaneGeo.h"

namespace g4n
{

  //-------------------------------------------------------------
  // Constructor.
  FLSHitListAction::FLSHitListAction() : 
    fFLSHit(nullptr),
    fFLSHitList(nullptr),
    fEnergyCut(0),
    fBirksConstant(0.0),
    fChouConstant(0.0),
    fIsRockVetoUsed(false),
    fEnergyExtrapolator(nullptr),
    fPerformFLSHitListSuppression(false)
  {
    // Create the particle list that we'll (re-)use during the course
    // of the Geant4 simulation.
    fFLSHit        = new sim::FLSHit();
    fFLSHitList    = new sim::FLSHitList();

  }

  //-------------------------------------------------------------
  // Constructor.
  FLSHitListAction::FLSHitListAction(const double min_energy, const bool is_rockveto_used) :
    fFLSHit(nullptr),
    fFLSHitList(nullptr),
    fEnergyCut(min_energy*GeV),
    fBirksConstant(0.0),
    fChouConstant(0.0),
    fIsRockVetoUsed(is_rockveto_used),
    fEnergyExtrapolator(nullptr),
    fPerformFLSHitListSuppression(false)
  {
    // Create the particle list that we'll (re-)use during the course
    // of the Geant4 simulation.
    fFLSHit        = new sim::FLSHit();
    fFLSHitList    = new sim::FLSHitList();

    // Define the EnergyLossExtrapolator class if rock veto is used
    if(is_rockveto_used)fEnergyExtrapolator = new G4EnergyLossForExtrapolator(0);

  }

  //-------------------------------------------------------------
  // Destructor.
  FLSHitListAction::~FLSHitListAction()
  {
    // Delete anything that we created with "new'.
    delete fFLSHitList;
    delete fFLSHit;
    if(fEnergyExtrapolator)delete fEnergyExtrapolator;
  }

  //-------------------------------------------------------------
  void FLSHitListAction::Config(fhicl::ParameterSet const& pset )
  {
    fEnergyCut                    = pset.get< double >("G4EnergyThreshold")*GeV;
    fBirksConstant                = pset.get< G4double >("BirksConstant")*cm/MeV;
    fChouConstant                 = pset.get< G4double >("ChouConstant")*cm2/MeV/MeV;
    fIsRockVetoUsed               = pset.get< bool   >("G4CheckRockVeto");
    fPerformFLSHitListSuppression = pset.get< bool   >("PerformFLSHitListSuppression");
  }

  //-------------------------------------------------------------
  void FLSHitListAction::PrintConfig(std::string const& /* opt */)
  {
    mf::LogInfo("FLSHitListAction")
      << "FLSHitListAction::PrintConfig \n"
      << "    EnergyCut          " << fEnergyCut                       << "\n"
      << "    BirksConstant      " << fBirksConstant*MeV/cm            << "\n"
      << "    ChouConstant       " << fChouConstant*MeV*MeV/cm2        << "\n"
      << "    IsRockVetoUsed     " << (fIsRockVetoUsed?"true":"false") << "\n";
  }

  //-------------------------------------------------------------
  void FLSHitListAction::BeginOfEventAction(const G4Event*)
  {
    // Clear any previous particle information.
    fFLSHitList->fHits.clear();
  }

  //-------------------------------------------------------------
  // Create our initial sim::FLSHit object and add it to the sim::FLSHitList.
  void FLSHitListAction::PreTrackingAction(const G4Track* track)
  {
    // Clear FLS hits and make some other initializations
    fIsParticleInsideDetectorBigBox = false;
    fIsRockVetoChecked = false;
    fFLSHit->Clear();
    fFLSHit->SetCellUniqueId(0);

    LOG_DEBUG("FLSHitListAction") << "pretracking step with track id: " 
				  << ParticleListAction::GetCurrentTrackID();
  }

  //-------------------------------------------------------------
  void FLSHitListAction::PostTrackingAction( const G4Track* /*track*/) 
  {
    LOG_DEBUG("FLSHitListAction") << "done tracking for this g4 track, now "
				  << fFLSHitList->fHits.size() << " hits";
  }

  //-------------------------------------------------------------
  // With every step, add to the particle's trajectory.
  void FLSHitListAction::SteppingAction(const G4Step* step)
  {
    //mf::LogInfo("FLSHitListAction") << "FLSHitListAction::SteppingAction";

    /// Get the pointer to the track
    G4Track *track = step->GetTrack();
  
    // Checking the RockVeto. This is to stop tracking low energy charged 
    // particles in the rock, since they are not going to produce any hits 
    // in the detector. Check it only for the first step.
    // The reason, this function appears in the Stepping and not in the PreTrackingAction,
    // where it is supposed to be is kind of stupid.
    // The input step and track pointers in the Actions must be declared as const.
    // Hence we cannot modify these objects. However step->GetTrack() returns non-const 
    // value (thank God) and thus we can modify it.
    if(fIsRockVetoUsed && !fIsRockVetoChecked){

      // Check the rock veto
      if(IsRockVetoed(track)){
	// Kill the track
	track->SetTrackStatus(fKillTrackAndSecondaries);
	return;
      }

      fIsRockVetoChecked = true;//! If it's already checked, don't need to check it on the next step
    }// end of rock veto


    // make sure the FLSHit object has been initialized
    if(fFLSHit == 0){
      mf::LogWarning("FLSHitListAction") << "WARNING - FLSHit object not initialized in FLSHitListAction";
      return;
    }

    const CLHEP::Hep3Vector &pos0 = step->GetPreStepPoint()->GetPosition(); // Start of the step
    const CLHEP::Hep3Vector &pos  = track->GetPosition();                   // End of the step
    const CLHEP::Hep3Vector &mom  = track->GetMomentum();

    LOG_DEBUG("FLSHitListAction") << "flshits momentum = "
				  << mom.x() << " " << mom.y() << " " << mom.z()
				  << " " << mom.mag();
  
    double tpos0[3] = {pos0.x()/cm, pos0.y()/cm, pos0.z()/cm}; ///< Start of the step
    double tpos1[3] = {pos.x()/cm , pos.y()/cm , pos.z()/cm};  ///< End of the step
    double tpos2[3] = {0.};                                    ///< Some temporary vector that we'll make use of

    // If it's a null step, don't use it. Otherwise it may induce an additional 
    // FLSHit, which is wrong
    if(tpos0[0]==tpos1[0] &&
       tpos0[1]==tpos1[1] &&
       tpos0[2]==tpos1[2])return;
  
    // If DetectorBigBox is used, we look if the particle is inside the box and 
    // going to be outside of the box. If so, stop tracking it.
    if(fGeo->isDetectorBigBoxUsed()){
      // tpos1 is the point after the step
      bool isParticleGoingToBeOutsideDetectorBigBox = fGeo->isInsideDetectorBigBox(tpos1[0], tpos1[1], tpos1[2]);

      if(fIsParticleInsideDetectorBigBox && !isParticleGoingToBeOutsideDetectorBigBox)
	track->SetTrackStatus(fStopAndKill);
      else
	fIsParticleInsideDetectorBigBox = isParticleGoingToBeOutsideDetectorBigBox;
    }// end of check if the detectorbigbox used


    //check that we are in the correct material to record a hit - ie scintillator
    std::string material = track->GetMaterial()->GetName();
    if(material.compare("Scintillator") != 0 ) {
      return;
    }

    //check if the current cell pointer is empty, that means
    //this is the first step for this track
    double dcos[3]  = {mom.x()/mom.mag(), mom.y()/mom.mag(), mom.z()/mom.mag()};
    geo::CellUniqueId cellid = fGeo->CellId(tpos1[0], tpos1[1], tpos1[2],
 					    dcos[0],  dcos[1],  dcos[2],
 					    step->GetStepLength()/cm);

    if(cellid == 0)   cellid = fGeo->CellId(tpos0[0], tpos0[1], tpos0[2],
					     dcos[0],  dcos[1],  dcos[2],
					     step->GetStepLength()/cm);

    LOG_DEBUG("FSHitListAction") << "attempting to get cell id pos is "
				  << tpos0[0] << ", " << tpos0[1] << ", " << tpos0[2] 
				  << " result is " << cellid;

    /// Get density for Birks' Law calculation
    TGeoManager *geomanager = fGeo->ROOTGeoManager();
    if ( ! geomanager ) {
       throw cet::exception("NoTGeoManager")
	 << "FLSHitListAction: no TGeoManager given by fGeo.  "
	 << "I need this to get the scintillator density.  Failing.\n"
	 << __FILE__ << ":" << __LINE__ << "\n";
       return;
    }
    TGeoMaterial *mat = geomanager->GetMaterial("Scintillator");
    if ( ! mat ) {
       throw cet::exception("NoTGeoMaterial")
	 << "FLSHitListAction: no TGeoMaterial names 'Scintillator' found.  "
	 << "I need this to get the scintillator density.  Failing.\n"
	 << __FILE__ << ":" << __LINE__ << "\n";
       return;
    }

    /// Setting Birks Constants
    /// \todo We should not be setting Birks Constant at each Geant4 step.
    /// 1. rho could be calculated in the very beginning (material is allways Scintillator) once
    /// 2. SetBirksConstant may probably also be done once somehow,
    ///    but for sure we could do it in PreTrackingAction
    {
      const double rho       = mat->GetDensity();
      track->GetMaterial()->GetIonisation()->SetBirksConstant( (fBirksConstant/rho));
    }// end of setting Birks Constants

    /// Getting Energy depositions
    const double edep      = step->GetTotalEnergyDeposit()/GeV;
    fNovaG4EmSaturation.SetChouConstant(fChouConstant);
    const double edepBirks = fNovaG4EmSaturation.VisibleEnergyDeposition(step)/GeV;

    /// Protect against geometry skew between G4/ROOT geometries
    /// if cellid returns 0 it means that the ROOT geometry couldnt
    /// find the same cell as the G4 Geometry - check the prestep point
    /// volume name against the current volume and if they are the
    /// same name then just reuse the current cell id
    if(cellid == 0 
       && step->GetPreStepPoint()->GetPhysicalVolume()->GetName() == track->GetVolume()->GetName()
       && fFLSHit->GetCellUniqueId() > 0
       ) cellid = fFLSHit->GetCellUniqueId();

    if(cellid == 0){
      /// oh dear, can't find which cell we are in with the ROOT geometry and
      /// this is a brand new track with the fFLSHit having been reset.
      /// add it to the flshit in the hopes that the next step will be in the same
      /// cell

      mf::LogWarning("CantFindCell") << "Can't find cell in ROOT Geometry, first energy deposition"
				     << " dE = " << step->GetTotalEnergyDeposit()/GeV
				     << " (GeV) for this track will be added to first located hit.";
	
      fFLSHit->AddEdep( edep );
      fFLSHit->AddEdepBirks( edepBirks );

      return;
    }// end of check if the cell is not found


    /// plane and cell number holders used below
    int ip = 0, ic = 0;
    /// Get local positions for the start of the step (tpos0) and store them in tpos2
    /// Also, get ip and ic
    fGeo->IdToCell(cellid, &ip, &ic)->WorldToLocal(tpos0, tpos2);

    //check to see if we are in a new cell with this step
    if( fFLSHit->GetCellUniqueId() != cellid ){

      // the first step in this volume - clear the hit
      fFLSHit->Clear();

      // Get the position and time the particle first enters
      // the volume, as well as the pdg code.  that information is
      // obtained from the G4Track object that corresponds to this step
      // account for the fact that we use cm, ns, GeV rather than the G4 defaults
      fFLSHit->SetPDG    ( track->GetDefinition()->GetPDGEncoding() );
      fFLSHit->SetTrackId( ParticleListAction::GetCurrentTrackID()  );

      //figure out the geometry information for the nova geometry
      fFLSHit->SetCellUniqueId(cellid);

      fFLSHit->SetPlane(ip);
      fFLSHit->SetCell(ic);
    }//end if first step in the volume

    fFLSHit->AddEdep( edep );
    fFLSHit->AddEdepBirks( edepBirks );
    /// Add position
    fFLSHit->AddPos(tpos2[0], tpos2[1], tpos2[2], (double)step->GetPreStepPoint()->GetGlobalTime()/ns, step->GetPreStepPoint()->GetKineticEnergy()  / GeV);

    /// maybe this is the last step in the volume - either the particle exits,
    /// disappears, or simply stops
    if(track->GetVolume()->GetName() != step->GetPostStepPoint()->GetPhysicalVolume()->GetName()
       || track->GetKineticEnergy() == 0.){

      //going into a new volume with the next step or no kinetic energy left for this particle
      if(fFLSHit->GetEdep() > 0.){
        /// Get local positions for the end of the step (tpos1) and store them in tpos2
	fGeo->IdToCell(fFLSHit->GetCellUniqueId(), &ip, &ic)->WorldToLocal(tpos1, tpos2);

        fFLSHit->AddPos(tpos2[0], tpos2[1], tpos2[2], (double)step->GetPostStepPoint()->GetGlobalTime()/ns, step->GetPostStepPoint()->GetKineticEnergy() / GeV);

	// Fill the FLSHitList
   	fFLSHitList->fHits.push_back( *fFLSHit );

        /// Zero out ID now since it's already recorded
        /// It needs to be zeroed out, because next step
        /// may arrive to the same cell
        /// and in that case we want a new FLSHit
        fFLSHit->SetCellUniqueId(0);
      }// end of check if the energy was deposited into the cell

    }//end if last hit in the volume


  }// end of FLSHitListAction::SteppingAction


  //------------------------------------------------------------------------------
  // There's one last thing to do: All the particles have their
  // parent IDs set (in PostTrackingAction), but we haven't set the
  // daughters yet.  That's done in this method.
  void FLSHitListAction::EndOfEventAction(const G4Event*)
  {

    /// Perform the suppression if requested
    if(fPerformFLSHitListSuppression){supressFLSHitList();}
  }


  //------------------------------------------------------------------------------
  // Suppress fFLSHitList
  bool FLSHitListAction::supressFLSHitList(){

    /// collect the hits coming from the same track ids in the same cells
    std::map< geo::CellUniqueId, std::vector<sim::FLSHit> >           cellToHits;
    std::map< geo::CellUniqueId, std::vector<sim::FLSHit> >::iterator cthitr;

    const int nflshits = fFLSHitList->fHits.size();
    /// Loop over all FLSHits and construct a map of vectors of hits with CellID as a key
    for(int h = 0; h < nflshits; ++h){
      const sim::FLSHit&      hit = fFLSHitList->fHits[h];
      const geo::CellUniqueId cid = hit.GetCellUniqueId();
      cthitr = cellToHits.find(cid);

      if( cthitr != cellToHits.end() )
	(*cthitr).second.push_back(hit);
      else{
	std::vector<sim::FLSHit> vec;
	vec.push_back(hit);
	cellToHits[cid] = vec;
      }
    }

    // clear the fFLSHitList as all the hits are now in the map
    fFLSHitList->fHits.clear();

    /// now loop over the map and collect hits from the
    /// same track id in the same cell.
    /// This loop is over all CellIDs
    for(cthitr = cellToHits.begin(); cthitr != cellToHits.end(); cthitr++){
      /// This will be a map of vector FLSHits with the same (CellID and TrackID)
      std::map< int, std::vector<sim::FLSHit> >           trackToHits;
      std::map< int, std::vector<sim::FLSHit> >::iterator tthitr;

      /// Loop over hits with the same Cell ID
      /// and fill the map
      for(size_t f = 0; f < (*cthitr).second.size(); ++f){
	tthitr = trackToHits.find((*cthitr).second.at(f).GetTrackID());
	if( tthitr != trackToHits.end() )
	  (*tthitr).second.push_back((*cthitr).second.at(f));
	else{
	  std::vector<sim::FLSHit> vec;
	  vec.push_back((*cthitr).second.at(f));
	  trackToHits[(*cthitr).second.at(f).GetTrackID()] = vec;
	}
      }// end of loop over hits with the same Cell ID and filling the map

      // now that all the hits from this cell are sorted by track id,
      // combine them into one hit for each id and add them to the fFLSHitList
      // weight the various values by the energy deposited
      for(tthitr = trackToHits.begin(); tthitr != trackToHits.end(); tthitr++){
	const int                       trackId = (*tthitr).first;
	const std::vector<sim::FLSHit>& vec     = (*tthitr).second;
	
        /// Construct the empty FLSHit
        sim::FLSHit out_hit;

        /// Set the CellID, plane etc from the first FLSHit (these entries are identical within this loop)
        out_hit.SetCellUniqueId(vec[0].GetCellUniqueId());
	out_hit.SetPlane       (vec[0].GetPlaneID());
	out_hit.SetCell        (vec[0].GetCellID());
	out_hit.SetPDG         (vec[0].GetPDG());
	out_hit.SetTrackId     (trackId);

        float  entry_x      = 0.0;
        float  entry_y      = 0.0;
        float  entry_z      = 0.0;
        double entry_t      = 0.0;
        float  entry_e      = 0.0;

        float  exit_x       = 0.0;
        float  exit_y       = 0.0;
        float  exit_z       = 0.0;
        double exit_t       = 0.0;
        float  exit_e       = 0.0;

        float  edep_total      = 0.0;
        float  edepbirks_total = 0.0;

        /// Loop over FLSHits with the same CellID and TrackID
	for(size_t f = 0; f < vec.size(); ++f){
          const sim::FLSHit& current_hit = vec[f];
          
          /// Energy deposition for the FLSHit
          const float edep = current_hit.GetEdep();

	  entry_x += edep * current_hit.GetEntryX();
	  entry_y += edep * current_hit.GetEntryY();
	  entry_z += edep * current_hit.GetEntryZ();
	  exit_x  += edep * current_hit.GetExitX();
	  exit_y  += edep * current_hit.GetExitY();
	  exit_z  += edep * current_hit.GetExitZ();

	  entry_t += edep * current_hit.GetEntryT();
	  exit_t  += edep * current_hit.GetExitT();

	  entry_e += current_hit.GetEntryEnergy();
	  exit_e  += current_hit.GetExitEnergy();

          edep_total      += edep;
	  edepbirks_total += current_hit.GetEdepBirks();
	}// end of the loop over FLSHits with the same Cell ID and TrackID

	entry_x /= edep_total;
	entry_y /= edep_total;
	entry_z /= edep_total;
	exit_x  /= edep_total;
	exit_y  /= edep_total;
	exit_z  /= edep_total;
	entry_t /= edep_total;
 	exit_t  /= edep_total;

        /// Set the averaged entry and exit positions to the new FLSHit
        out_hit.AddPos(entry_x, entry_y, entry_z, entry_t, entry_e);
        out_hit.AddPos(exit_x , exit_y , exit_z , exit_t , exit_e);

        /// Set the total edep, summed over all FLSHits with the same CellID and TrackID
        out_hit.AddEdep(edep_total);
        out_hit.AddEdepBirks(edepbirks_total);

	// add the hit to fFLSHitList
	fFLSHitList->fHits.push_back(out_hit);

      } // end loop over track id to hit map
    }// end loop over cell id to hit map

    return true;
  }

  //------------------------------------------------------------------------------
  // Checking if particle is in the rock and that it's not going to reach the detector.
  // If that's the case, veto it (kill the track and it's secondaries)
  bool FLSHitListAction::IsRockVetoed(const G4Track* track) {

    //G4double energy = track->GetKineticEnergy();
    // No veto'ing for uncharged particles

    if (! (track->GetDefinition()->GetPDGCharge()) ) return false;
 
    G4String  material = track->GetMaterial()->GetName();
    if(material != "Granite") return false;

    const CLHEP::Hep3Vector &pos = track->GetStep()->GetPreStepPoint()->GetPosition();         

    double tpos[3] = {pos.x()/cm, pos.y()/cm, pos.z()/cm};
    double tdir[3] = {track->GetMomentumDirection().getX(),
		      track->GetMomentumDirection().getY(),
		      track->GetMomentumDirection().getZ()};
        

    gGeoManager->InitTrack(tpos, tdir);

    gGeoManager->FindNextBoundary();

    // Calculate the energy in GeV (I think)
    double energyAfterStep = fEnergyExtrapolator->EnergyAfterStep(track->GetKineticEnergy()/MeV,
                                                                  gGeoManager->GetStep() / cm*mm,
                                                                  track->GetMaterial(),
                                                                  track->GetDefinition());

    return (energyAfterStep < fEnergyCut);
    
  }

  //------------------------------------------------------------------------------
  // For now, it's obsolete
  bool FLSHitListAction::ParticleProjection(G4Track* track) {
    
    art::ServiceHandle<geo::Geometry> fGeo;

    double x1, x2;
    double y1, y2;
    double z1, z2;
    fGeo->WorldBox(&x1, &x2, &y1, &y2, &z1, &z2);

    std::cout<<"World X "<<x1<<"  "<<x2<<"\n";
    std::cout<<"World Y "<<y1<<"  "<<y2<<"\n";
    std::cout<<"World Z "<<z1<<"  "<<z2<<"\n";

    fGeo->DetectorEnclosureBox(&x1, &x2, &y1, &y2, &z1, &z2);

    std::cout<<"Detector X "<<x1<<"  "<<x2<<"\n";
    std::cout<<"Detector Y "<<y1<<"  "<<y2<<"\n";
    std::cout<<"Detector Z "<<z1<<"  "<<z2<<"\n";

    return false;

    const CLHEP::Hep3Vector &pos = track->GetStep()->GetPreStepPoint()->GetPosition();

    double tpos[3] = {pos.x()/cm, pos.y()/cm, pos.z()/cm};
    double tdir[3] = {track->GetMomentumDirection().getX(),
		      track->GetMomentumDirection().getY(),
		      track->GetMomentumDirection().getZ()};

    gGeoManager->InitTrack(tpos, tdir);
    gGeoManager->FindNextBoundary();

    G4double steplength = gGeoManager->GetStep() / cm*mm;

    //Calculate energy in GeV (I think)
    double energyAfterStep = fEnergyExtrapolator->EnergyAfterStep(track->GetKineticEnergy()/MeV,
								  steplength,
								  track->GetMaterial(),
								  track->GetDefinition());

    return (energyAfterStep < fEnergyCut);
  }

}//end namespace
