////////////////////////////////////////////////////////////////////////
/// \file  FLSHitListAction.h
/// \brief Use Geant4's user "hooks" to maintain a list of particles generated by Geant4.
///
/// \version $Id: FLSHitListAction.h,v 1.11 2012-09-20 21:45:42 greenc Exp $
/// \author  brebel@fnal.gov
////////////////////////////////////////////////////////////////////////

/// This class implements the G4Base::UserAction interface in order to
/// accumulate a list of fls hits modeled by Geant4.
//
#pragma once

// NOvA includes
#include "G4Base/UserAction.h"
#include "Geometry/Geometry.h"

// G4 includes
#include "g4nova/NovaG4EmSaturation.hh"

//ART includes
#include "art/Framework/Services/Registry/ServiceHandle.h"

// Forward declarations.
class G4Event;
class G4Track;
class G4Step;
class G4EnergyLossForExtrapolator;

namespace sim{
    class FLSHit;
    class FLSHitList;
}

//namespace g4n{ class FLSHitListActionDebug;}


namespace g4n {

  ///list of energy deposits from Geant4
  class FLSHitListAction : public g4b::UserAction {

  public:
    // Standard constructors and destructors;
    FLSHitListAction();
    FLSHitListAction(const double min_energy, const bool is_rockveto_used);
    virtual ~FLSHitListAction();

    void Config(fhicl::ParameterSet const& pset);
    void PrintConfig(std::string const& opt);

    // UserActions method that we'll override, to obtain access to
    // Geant4's particle tracks and trajectories.
    void BeginOfEventAction(const G4Event*);
    void EndOfEventAction(const G4Event*);
    void PreTrackingAction(const G4Track*);
    void PostTrackingAction(const G4Track*);
    void SteppingAction(const G4Step*);
    bool IsRockVetoed(const G4Track*);
    bool ParticleProjection(G4Track*);

    //  Returns the current hit being saved in the list of
    //  hits.  
    const sim::FLSHit* GetFLSHit() { return fFLSHit; }

    //  Returns the FLSHitList accumulated during the current event.
    sim::FLSHitList* GetList() const { return fFLSHitList; }

  private:

    /// Suppress fFLSHitList
    bool supressFLSHitList();

  private:
    sim::FLSHit*                 fFLSHit;                 ///< The information for a single hit.
    sim::FLSHitList*             fFLSHitList;   	  ///< The accumulated information for hits in the event.
    G4double                     fEnergyCut;    	  ///< The minimum energy in GeV for a particle to       
                                                	  ///< be included in the list.                          
    G4double                     fBirksConstant;          ///< constant used in Birks suppression
    G4double                     fChouConstant;           ///< 2nd order constant used in Birks suppression
    bool                         fIsRockVetoChecked;      ///< Is rock veto already checked?
    bool                         fIsRockVetoUsed;         ///< Do we ise the RockVeto method?
    G4EnergyLossForExtrapolator* fEnergyExtrapolator;     ///< Used in the RockVeto method
    NovaG4EmSaturation           fNovaG4EmSaturation;     ///< Determines the visible energy (after Birks suppression) Modified to include Birks-Chou

    bool                         fIsParticleInsideDetectorBigBox;///< Is the particle inside the Big Box?

    art::ServiceHandle<geo::Geometry> fGeo;

    bool                         fPerformFLSHitListSuppression;///< Do we want to perform FLSHitList suppression?

  };

} // namespace g4n

