//
// ********************************************************************
// * License and Disclaimer                                           *
// *                                                                  *
// * The  Geant4 software  is  copyright of the Copyright Holders  of *
// * the Geant4 Collaboration.  It is provided  under  the terms  and *
// * conditions of the Geant4 Software License,  included in the file *
// * LICENSE and available at  http://cern.ch/geant4/license .  These *
// * include a list of copyright holders.                             *
// *                                                                  *
// * Neither the authors of this software system, nor their employing *
// * institutes,nor the agencies providing financial support for this *
// * work  make  any representation or  warranty, express or implied, *
// * regarding  this  software system or assume any liability for its *
// * use.  Please see the license in the file  LICENSE  and URL above *
// * for the full disclaimer and the limitation of liability.         *
// *                                                                  *
// * This  code  implementation is the result of  the  scientific and *
// * technical work of the GEANT4 collaboration.                      *
// * By using,  copying,  modifying or  distributing the software (or *
// * any work based  on the software)  you  agree  to acknowledge its *
// * use  in  resulting  scientific  publications,  and indicate your *
// * acceptance of all terms of the Geant4 Software license.          *
// ********************************************************************
//
// $Id: NovaG4MonopolePhysics.cc,v 1.3 2012-09-20 21:45:43 greenc Exp $
// GEANT4 tag $Name: not supported by cvs2svn $
//
//---------------------------------------------------------------------------
//
// ClassName:   NovaG4MonopolePhysics
//
// Author:      V.Ivanchenko 13.03.2005
//
// Modified:
//
//  12.07.10  S.Burdin (changed the magnetic and electric charge variables from integer to double)
//  2012-02-04:  R. Hatcher - incorporate into NOvA framework from base G4 v4.9.4.p02
//----------------------------------------------------------------------------
//
//
//....oooOO0OOooo........oooOO0OOooo........oooOO0OOooo........oooOO0OOooo......
//....oooOO0OOooo........oooOO0OOooo........oooOO0OOooo........oooOO0OOooo......

#include "g4nova/NovaG4MonopolePhysics.hh"
#include "g4nova/NovaG4MonopolePhysicsMessenger.hh"

#include "g4nova/NovaG4Monopole.hh"

#include "Geant4/G4ParticleDefinition.hh"
#include "Geant4/G4ProcessManager.hh"

#include "Geant4/G4StepLimiter.hh"
#include "Geant4/G4Transportation.hh"

#include "g4nova/NovaG4MonopoleTransportation.hh"

#include "Geant4/G4hMultipleScattering.hh"
#include "Geant4/G4mplIonisation.hh"
#include "Geant4/G4mplIonisationWithDeltaModel.hh"
#include "Geant4/G4hhIonisation.hh"
#include "Geant4/G4hIonisation.hh"

//....oooOO0OOooo........oooOO0OOooo........oooOO0OOooo........oooOO0OOooo......

namespace g4n {

NovaG4MonopolePhysics::NovaG4MonopolePhysics(const G4String& nam)
  : G4VPhysicsConstructor(nam)
{
  magCharge = 1.0;  // will * ( e / fine_structure_const * 0.5 )
  //  magCharge = -1.0;
  //  elCharge  = -50.0;
  elCharge  = 0.0;
  monopoleMass = 100.*GeV;
  theMessenger = new NovaG4MonopolePhysicsMessenger(this);
}

NovaG4MonopolePhysics::~NovaG4MonopolePhysics()
{
  delete theMessenger;
}

void NovaG4MonopolePhysics::ConstructParticle()
{
  mpl = NovaG4Monopole::MonopoleDefinition(monopoleMass, magCharge, elCharge);
}

void NovaG4MonopolePhysics::ConstructProcess()
{
  
  if(verboseLevel > 0) {
    G4cout << "NovaG4MonopolePhysics::ConstructProcess" << G4endl;
  }
  
  G4ProcessManager* pmanager = new G4ProcessManager(mpl);
  mpl->SetProcessManager(pmanager);
  
  // defined monopole parameters and binning

  G4double emax = mpl->GetPDGMass()*10;
  G4double magn = mpl->MagneticCharge();
  //  G4double emin = mpl->GetPDGMass()/20000.;
  // if(emin < keV) emin = keV;

  G4double emin = emax/1000000.;

  //  G4int nbin = G4int(std::log10(emin/eV));

  G4int nbin = 10000;
  /*
  if(nbin < 1) nbin = 1;
  nbin *= 10;
  */

  if(magn == 0.0) {
    pmanager->AddProcess( new G4Transportation(), -1, 0, 0);
  } else {
   pmanager->AddProcess( new NovaG4MonopoleTransportation(mpl), -1, 0, 0);
  }

  G4int idx = 1;
  if(mpl->GetPDGCharge() != 0.0) {
    //G4hMultipleScattering* hmsc = new G4hMultipleScattering();
    //pmanager->AddProcess(hmsc,  -1, idx, idx);
    //++idx;
    G4hIonisation* hhioni = new G4hIonisation();
    hhioni->SetDEDXBinning(nbin);
    hhioni->SetMinKinEnergy(emin);
    hhioni->SetMaxKinEnergy(emax);
    pmanager->AddProcess(hhioni,  -1, idx, idx);
    ++idx;
  }
  if(magn != 0.0) {
    G4mplIonisation* mplioni = new G4mplIonisation(magn);
    mplioni->SetDEDXBinning(nbin);
    mplioni->SetMinKinEnergy(emin);
    mplioni->SetMaxKinEnergy(emax);
    G4mplIonisationWithDeltaModel* mod = 
      new G4mplIonisationWithDeltaModel(magn,"PAI");
    mplioni->AddEmModel(0,mod,mod);
    pmanager->AddProcess(mplioni, -1, idx, idx);
    ++idx;
  }
  pmanager->AddProcess( new G4StepLimiter(),  -1, -1, idx);

}

void NovaG4MonopolePhysics::SetMagneticCharge(G4double val)
{
  magCharge = val;
}

void NovaG4MonopolePhysics::SetElectricCharge(G4double val)
{
  elCharge = val;
}

void NovaG4MonopolePhysics::SetMonopoleMass(G4double mass)
{
  monopoleMass = mass;
}


//....oooOO0OOooo........oooOO0OOooo........oooOO0OOooo........oooOO0OOooo......
} // end-of-namespace g4n

#include "G4Base/G4PhysicsProcessFactorySingleton.hh"
PHYSPROCREG3(g4n,NovaG4MonopolePhysics,g4n::NovaG4MonopolePhysics);
