////////////////////////////////////////////////////////////////////////
/// \brief Use Geant4's user "hooks" to maintain a list of particles generated by Geant4.
///
/// \author  seligman@nevis.columbia.edu, brebel@fnal.gov
////////////////////////////////////////////////////////////////////////
// NOvA Includes
#include "g4nova/ParticleListAction.h"
#include "G4Base/PrimaryParticleInformation.h"
#include "Simulation/Particle.h"
#include "Simulation/ParticleNavigator.h"
#include "Simulation/Simulation.h"


#include "G4Base/UserActionFactory.h"
USERACTIONREG3(g4n,ParticleListAction,g4n::ParticleListAction)

// G4 includes
#include "Geant4/G4Event.hh"
#include "Geant4/G4Track.hh"
#include "Geant4/G4ThreeVector.hh"
#include "Geant4/G4ParticleDefinition.hh"
#include "Geant4/G4PrimaryParticle.hh"
#include "Geant4/G4DynamicParticle.hh"
#include "Geant4/G4VUserPrimaryParticleInformation.hh"
#include "Geant4/G4Step.hh"
#include "Geant4/G4StepPoint.hh"
#include "Geant4/G4VProcess.hh"

// ROOT includes
#include <TLorentzVector.h>

// C/C++ includes
#include <algorithm>
#include <string>

// ART includes
#include "messagefacility/MessageLogger/MessageLogger.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"

namespace g4n{

  // Initialize static members.
  sim::Particle* ParticleListAction::fParticle       = 0;
  int            ParticleListAction::fCurrentTrackID = sim::kNoParticleId;
  int            ParticleListAction::fTrackIDOffset  = 0;
  

  //-------------------------------------------------------------
  // Constructor.
  ParticleListAction::ParticleListAction(double energythres, bool manyParticles) : 
     fEnergyCut          (energythres*GeV)
    ,fManyParticles      (manyParticles)
    ,fRecordFirstBrem    (false)
    ,fRecordFirstPair    (false)
    ,fStoreTrajOutsideBigBox(false)
    ,fisFirstBremRecorded(false)
    ,fisFirstPairRecorded(false)
  {
    // Create the particle list that we'll (re-)use during the course
    // of the Geant4 simulation.
    fParticleNav  = new sim::ParticleNavigator();

  }

  //-------------------------------------------------------------
  // Destructor.
  ParticleListAction::~ParticleListAction()
  {
    // Delete anything that we created with "new'.
    delete fParticleNav;
  }

  //-------------------------------------------------------------
  void ParticleListAction::Config(fhicl::ParameterSet const& pset )
  {
    fEnergyCut      = pset.get< double >("G4EnergyThreshold")*GeV;
    fManyParticles  = pset.get< bool   >("ManyParticles");
    fRecordFirstBrem= pset.get< bool   >("RecordFirstBrem");
    fRecordFirstPair= pset.get< bool   >("RecordFirstPair");
    fStoreTrajOutsideBigBox = pset.get< bool >("StoreTrajectoriesOutsideBigBox");
  }

  //-------------------------------------------------------------
  void ParticleListAction::PrintConfig(std::string const& /* opt */)
  {
    mf::LogInfo("ParticleListAction")
      << "ParticleListAction::PrintConfig \n"
      << "    EnergyCut          " << fEnergyCut                      << "\n"
      << "    ManyParticles      " << (fManyParticles?"true":"false") << "\n"
      << "    RecordFirstBrem          " <<  (fRecordFirstBrem?"true":"false")                      << "\n"
      << "    RecordFirstPair      " << (fRecordFirstPair?"true":"false") << "\n"
      << "    StoreTrajOutsideBigBox      " << (fStoreTrajOutsideBigBox?"true":"false") << "\n";
  }

  //-------------------------------------------------------------
  void ParticleListAction::BeginOfEventAction(const G4Event*)
  {
    // Clear any previous particle information.
    fParticle = 0;
    fParentIDMap.clear();
    fParticleNav->clear();
    fTrackIDToMCTruthIndex.clear();

    fCurrentTrackID = sim::kNoParticleId;

    /// If it is configured to record the first of these
    /// interactions, then set the variable to false.
    /// Otherwise, treat it as if we already recorded it (even though we did not)
    if(fRecordFirstBrem) fisFirstBremRecorded = false;
    else                 fisFirstBremRecorded = true;
    if(fRecordFirstPair) fisFirstPairRecorded = false;
    else                 fisFirstPairRecorded = true;
  }

  //-------------------------------------------------------------
  // figure out the ultimate parentage of the particle with track ID
  // trackid
  // assume that the current track id has already been added to
  // the fParentIDMap
  int ParticleListAction::GetParentage(int trackid) const
  {
    int parentid = sim::kNoParticleId;
    
    // search the fParentIDMap recursively until we have the parent id 
    // of the first EM particle that led to this one
    std::map<int,int>::const_iterator itr = fParentIDMap.find(trackid);
    while( itr != fParentIDMap.end() ){
      LOG_DEBUG("ParticleListAction") << "parentage for " << trackid
				      << " " << (*itr).second;
      // set the parentid to the current parent ID, when the loop ends
      // this id will be the first EM particle 
      parentid = (*itr).second;
      itr = fParentIDMap.find(parentid);
    }
    LOG_DEBUG("ParticleListAction") << "final parent ID " << parentid; 

    return parentid;
  }

  //-------------------------------------------------------------
  // Create our initial sim::Particle object and add it to the sim::ParticleList.
  void ParticleListAction::PreTrackingAction(const G4Track* track)
  {
    // get the track ID for this particle
    const G4int trackID = track->GetTrackID() + fTrackIDOffset;
    fCurrentTrackID = trackID;
    size_t mcTruthIndex = 0;

    // get the parent id from Geant for the current track:
    G4int parentID = track->GetParentID() + fTrackIDOffset;

    //RUN_ON_PARTICLELISTACTION_DEBUG(DEBUG_PRETRACKING_INITIAL, track);

    // Is there an MCTruth object associated with this G4Track?  We
    // have to go up a "chain" of information to find out:
    const G4ParticleDefinition* partdef = track->GetDefinition();
    const G4int                 pdg     = partdef->GetPDGEncoding();
    const G4DynamicParticle*    dp      = track->GetDynamicParticle();
    const G4PrimaryParticle*    pp      = dp->GetPrimaryParticle();

    LOG_DEBUG("ParticleListAction") << "preparing to track " << fCurrentTrackID 
					  << " pdg " << pdg
					  << " with parent " << parentID;

    std::string process_name;
    if ( pp != 0 ){
      const G4VUserPrimaryParticleInformation* gppi = pp->GetUserInformation();
      const g4b::PrimaryParticleInformation* ppi = dynamic_cast<const g4b::PrimaryParticleInformation*>(gppi);

      if ( ppi != 0 ){
	// If we've made it this far, a PrimaryParticleInformation
	// object exists and we are using a primary particle, set the 
	// process name accordingly.
	process_name = "Primary";

	// primary particles should still have parentID = 0, even 
	// if there are multiple MCTruths for this event
	parentID = 0;

	mcTruthIndex = ppi->MCTruthIndex();
      }

    } // Is there a G4PrimaryParticle?
    else{

      // figure out what process is making this track - skip it if it is 
      // one of pair production, compton scattering, photoelectric effect
      // bremstrahlung, annihilation, any ionization - who wants to save
      // a buttload of electrons that arent from a CC interaction?
      process_name = track->GetCreatorProcess()->GetProcessName();
      if( !fManyParticles && (process_name.find("conv")               != std::string::npos 
                              || process_name.find("LowEnConversion") != std::string::npos
                              ||(process_name.find("Pair")            != std::string::npos && fisFirstPairRecorded)
                              || process_name.find("compt")           != std::string::npos
                              || process_name.find("Compt")           != std::string::npos
                              ||(process_name.find("Brem")            != std::string::npos && fisFirstBremRecorded)
                              || process_name.find("phot")            != std::string::npos
                              || process_name.find("Photo")           != std::string::npos
                              || process_name.find("Ion")             != std::string::npos
                              || process_name.find("annihil")         != std::string::npos)){
        
	
	// find the ultimate parent of this particle that was not a secondary
	// of the EM shower
	// first add this track id and its parent to the fParentIDMap
	fParentIDMap[fCurrentTrackID] = parentID;

	fCurrentTrackID = this->GetParentage(fCurrentTrackID);

	LOG_DEBUG("ParticleListAction") << "current track ID " << fCurrentTrackID;

	// check that fCurrentTrackID is in the particle list - it is possible
	// that this particle's parent is a particle that did not get tracked.
	// An example is a parent that was made due to muMinusCaptureAtRest
	// and the daughter was made by the phot process.  The parent likely
	// isn't saved in the particle list because it is below the energy cut
	// which will put a bogus track id value into the sim::IDE object for 
	// the sim::SimChannel if we don't check it.
	if(fParticleNav->find(fCurrentTrackID) == fParticleNav->end() )
	  fCurrentTrackID = sim::kNoParticleId;

	// set fParticle to 0 as we are not stepping this particle 
	// and adding trajectory points to it
	fParticle = 0;

	LOG_DEBUG("ParticleListAction") << "killing TrackID: " << trackID << " bc EM daughter, "
					<< process_name << " " << pdg 
					<< ", use track id " << fCurrentTrackID;

	return;
      }// end if part of EM shower, but not primary particle

      // Check the energy of the particle.  If it falls below the energy
      // cut, don't add it to our list.
      const G4double energy = track->GetKineticEnergy();
      if ( energy < fEnergyCut ){
	fParticle = 0;
	LOG_DEBUG("ParticleListAction") << "killing TrackID: " << fCurrentTrackID << " energy/fEnergyCut";

	// do add the particle to the parent id map though
	// and set the current track id to be it's ultimate parent
	fParentIDMap[fCurrentTrackID] = parentID;

	fCurrentTrackID = this->GetParentage(fCurrentTrackID);
	return;
      }

      // check to see if the parent particle has been stored in the particle navigator
      // if not, then see if it is possible to walk up the fParentIDMap to find the
      // ultimate parent of this particle.  Use that ID as the parent ID for this
      // particle
      if( fParticleNav->find(parentID) == fParticleNav->end() ){
	// do add the particle to the parent id map
	// just in case it makes a daughter that we have to track as well
	fParentIDMap[fCurrentTrackID] = parentID;
	const int pid = this->GetParentage(parentID);
	
	// if we still can't find the parent in the particle navigator, 
	// we have to give up
	if( fParticleNav->find(pid) == fParticleNav->end() ){
	  mf::LogWarning("ParticleListAction") << "can't find parent id: "
					       << parentID 
					       << " in the particle navigator, or fParentIDMap."
					       << " Make " << parentID << " the mother ID for"
					       << " track ID " << fCurrentTrackID
					       << " in the hope that it will aid debugging.";
	}
	else
	  parentID = pid;
      }

      // Attempt to find the MCTruth index corresponding to the 
      // current particle.  If the fCurrentTrackID is not in the
      // map try the parent ID, if that is not there, throw an
      // exception
      if(fTrackIDToMCTruthIndex.count(fCurrentTrackID) > 0 ) 
	mcTruthIndex = fTrackIDToMCTruthIndex.at(fCurrentTrackID);
      else if(fTrackIDToMCTruthIndex.count(parentID) > 0 ) 
	mcTruthIndex = fTrackIDToMCTruthIndex.at(parentID);
      else
	throw cet::exception("ParticleListAction") << "Cannot find MCTruth index for track id "
						   << fCurrentTrackID << " or " << parentID;
	  
    }// end of if/else


    // protect against potentially empty process names.
    if     ( process_name.empty() )                         { process_name         = "unknown";}
    // if it's the brem photon, update the variable
    else if( process_name.find("Brem") != std::string::npos){ fisFirstBremRecorded = true;}
    else if( process_name.find("Pair") != std::string::npos){ fisFirstPairRecorded = true;}


    // Create the sim::Particle object
    fParticle = new sim::Particle(fCurrentTrackID, pdg, process_name, 
				  parentID, dp->GetMass()/GeV);
    //part.Print();

    // Polarization.
    const G4ThreeVector& polarization = track->GetPolarization();
    fParticle->SetPolarization( TVector3( polarization.x(),
					  polarization.y(),
					  polarization.z() ) );
  
    //RUN_ON_PARTICLELISTACTION_DEBUG(DEBUG_PRETRACKING_FINAL, track, parentID, trackID, process_name);

    fParticleNav->Add(fParticle);

    if(fTrackIDToMCTruthIndex.count(fCurrentTrackID) > 0)
      LOG_WARNING("ParticleListAction") << "attempting to put " << fCurrentTrackID
					<< " into fTrackIDToMCTruthIndex map "
					<< " particle is\n" << *fParticle;

    fTrackIDToMCTruthIndex[fCurrentTrackID] = mcTruthIndex;

    return;
  }

  //-------------------------------------------------------------
  void ParticleListAction::PostTrackingAction( const G4Track* track)
  {

    if( !fParticle ) return;

    // particle final position now saved in SteppingAction when 
    // filling the TrajectoryPoint, filled using the post step point.

    //RUN_ON_PARTICLELISTACTION_DEBUG(DEBUG_POSTTRACKING, track);

    const G4ThreeVector position = track->GetPosition();

    // If the final position wasn't inside the big box, then we wouldn't have
    // written it as a trajectory point. But we are interested in where
    // particles stop, so do it here.
    if(!fStoreTrajOutsideBigBox &&
       !fGeo->isInsideDetectorBigBox(position.x() / cm,
                                     position.y() / cm,
                                     position.z() / cm)){
      G4double time = track->GetGlobalTime();

      TLorentzVector fourPos( position.x() / cm,
			      position.y() / cm,
			      position.z() / cm,
			      time / ns);

      const G4ThreeVector momentum = track->GetMomentum();
      const G4double energy = track->GetTotalEnergy();
      TLorentzVector fourMom( momentum.x() / GeV,
			      momentum.y() / GeV,
			      momentum.z() / GeV,
			      energy / GeV);

      fParticle->AddTrajectoryPoint( fourPos, fourMom );
    }
  }

  //-------------------------------------------------------------
  // With every step, add to the particle's trajectory - not
  // done for nova - maybe we should?
  void ParticleListAction::SteppingAction(const G4Step* step){


     /*
     TRACE_ON_PARTICLELISTACTION_DEBUG("FROM ParticleListAction::SteppingAction.Stepping"
                                       <<" length = "<<step->GetStepLength()
                                      );
     */

    if ( !fParticle ) return;
    
    // For the most part, we just want to add the post-step
    // information to the particle's trajectory.  There's one
    // exception: In PreTrackingAction, the correct time information
    // is not available.  So add the correct vertex information here.
    
    if ( fParticle->NumberTrajectoryPoints() == 0 ){
      // Get the pre-step information from the G4Step.
      const G4StepPoint* preStepPoint = step->GetPreStepPoint();
      
      const G4ThreeVector position = preStepPoint->GetPosition();
      G4double time = preStepPoint->GetGlobalTime();
      
      // Remember that LArSoft uses cm, ns, GeV.
      TLorentzVector fourPos( position.x() / cm,
			      position.y() / cm,
			      position.z() / cm,
			      time / ns);
      
      const G4ThreeVector momentum = preStepPoint->GetMomentum();
      const G4double energy = preStepPoint->GetTotalEnergy();
      TLorentzVector fourMom( momentum.x() / GeV,
			      momentum.y() / GeV,
			      momentum.z() / GeV,
			      energy / GeV);

      // Add the first point in the trajectory.
      fParticle->AddTrajectoryPoint( fourPos, fourMom );
    }
    
    // Get the post-step information from the G4Step.
    const G4StepPoint* postStepPoint = step->GetPostStepPoint();
    
    const G4ThreeVector position = postStepPoint->GetPosition();
    G4double time = postStepPoint->GetGlobalTime();
    
    // Remember that LArSoft uses cm, ns, GeV.
    TLorentzVector fourPos( position.x() / cm,
			    position.y() / cm,
			    position.z() / cm,
			    time / ns );
    
    const G4ThreeVector momentum = postStepPoint->GetMomentum();
    const G4double energy = postStepPoint->GetTotalEnergy();
    TLorentzVector fourMom( momentum.x() / GeV,
			    momentum.y() / GeV,
			    momentum.z() / GeV,
			    energy / GeV );
    
    // Add another point in the trajectory.
    if(fStoreTrajOutsideBigBox ||
      
       fGeo->isInsideDetectorBigBox(position.x() / cm,
                                    position.y() / cm,
                                    position.z() / cm)){
      fParticle->AddTrajectoryPoint( fourPos, fourMom );
    }
  }

  //-------------------------------------------------------------
  std::vector<sim::Particle> ParticleListAction::GetList() const
  {
    // fill a vector<sim::Particle> using the sim::ParticleNavigator
    std::vector<sim::Particle> plist;

    // make sure to set the fTrackIDOffset with the highest G4 track id + 100
    // the 100 gives some cushion between lists
    int highestID = 0;
    for(sim::ParticleNavigator::iterator itr = fParticleNav->begin(); itr != fParticleNav->end(); ++itr){
      plist.push_back(*((*itr).second));
      if((*itr).first > highestID) highestID = (*itr).first;
    }

    fTrackIDOffset = highestID + 100;

    LOG_DEBUG("ParticleListAction") << *fParticleNav << "\ntrack id offset is now " << fTrackIDOffset;

    return plist;
  }

  //-------------------------------------------------------------
  std::map<int, size_t> ParticleListAction::TrackIDToMCTruthIndexMap() const
  {
    return fTrackIDToMCTruthIndex;
  }

  //-------------------------------------------------------------
  /// Utility class for the EndOfEventAction method: update the
  /// daughter relationships in the particle list.
  class UpdateDaughterInformation : public std::unary_function<sim::ParticleNavigator::value_type, void>
  {
  public:
    UpdateDaughterInformation()
      : particleNav(0)
    {}
    void SetParticleNav( sim::ParticleNavigator* p ) { particleNav = p; }
    void operator()( sim::ParticleNavigator::value_type& particleNavEntry )
    {
      // We're looking at this Particle in the list.
      sim::Particle* particle = particleNavEntry.second;

      // The parent ID of this particle.
      int parentID = particle->Mother();

      // If the parentID <= 0, this is a primary particle.
      if ( parentID <= 0 ) return;

      // If we get here, this particle is somebody's daughter.  Add
      // it to the list of daughter particles for that parent.

      // Get the parent particle from the list.
      sim::ParticleNavigator::iterator parentEntry = particleNav->find( parentID );

      if ( parentEntry == particleNav->end() ){
        // We have an "orphan": a particle whose parent isn't
        // recorded in the particle list.  This is not signficant;
        // it's possible for a particle not to be saved in the list
        // because it failed an energy cut, but for it to have a
        // daughter that passed the cut (e.g., a nuclear decay).
        return;
      }

      // Add the current particle to the daughter list of the
      // parent.
      sim::Particle* parent = (*parentEntry).second;
      parent->AddDaughter( particle->TrackId() );
    }
  private:
    sim::ParticleNavigator* particleNav;     
  };

  //-------------------------------------------------------------
  // There's one last thing to do: All the particles have their
  // parent IDs set (in PostTrackingAction), but we haven't set the
  // daughters yet.  That's done in this method.
  void ParticleListAction::EndOfEventAction(const G4Event*)
  {
      // Set up the utility class for the "for_each" algorithm.  (We only
      // need a separate set-up for the utility class because we need to
      // give it the pointer to the particle list.  We're using the STL
      // "for_each" instead of the C++ "for loop" because it's supposed
      // to be faster.
      UpdateDaughterInformation updateDaughterInformation;
      updateDaughterInformation.SetParticleNav( fParticleNav );

      // Update the daughter information for each particle in the list.
      std::for_each(fParticleNav->begin(), 
		    fParticleNav->end(), 
		    updateDaughterInformation);

      LOG_DEBUG("ParticleListAction") << *fParticleNav;
  }
  
} // end namespace
