////////////////////////////////////////////////////////////////////////
/// \brief Use Geant4's user "hooks" to kill particles in the rock
///
/// \author  rhatcher@fnal.gov
////////////////////////////////////////////////////////////////////////

// NOvA Includes
#include "g4nova/RockCutterAction.h"

#include "G4Base/PrimaryParticleInformation.h"
#include "G4Base/UserActionFactory.h"
USERACTIONREG3(g4n,RockCutterAction,g4n::RockCutterAction)

// G4 includes
#include "Geant4/G4Event.hh"
#include "Geant4/G4Track.hh"
#include "Geant4/G4ThreeVector.hh"
#include "Geant4/G4ParticleDefinition.hh"
#include "Geant4/G4PrimaryParticle.hh"
#include "Geant4/G4DynamicParticle.hh"
#include "Geant4/G4VUserPrimaryParticleInformation.hh"
#include "Geant4/G4Step.hh"
#include "Geant4/G4StepPoint.hh"
#include "Geant4/G4VProcess.hh"
#include "Geant4/G4VPhysicalVolume.hh"
#include "Geant4/G4VTouchable.hh"

// ROOT includes
#include <TLorentzVector.h>
#include <TString.h>

// C/C++ includes
#include <algorithm>
#include <string>

// ART includes
#include "messagefacility/MessageLogger/MessageLogger.h"

namespace g4n {

  // Initialize static members.
  
  //-------------------------------------------------------------
  // Constructor.
  RockCutterAction::RockCutterAction()
    : fVerbose(0)
    , fX0     (0)            
    , fY0     (0)            
    , fZ0     (728)          
    , fX_max  (304.8)
    , fX_min  (-1620)         
    , fY_max  (407.9)  
    , fY_min  (-178.9)     
    , fZ_max  (1800)            
    , fZ_min  (-840)          
 
  {

  }

  //-------------------------------------------------------------
  // Destructor.
  RockCutterAction::~RockCutterAction()
  {
    // Delete anything that we created with "new'.
  }

  //-------------------------------------------------------------
  void RockCutterAction::Config(fhicl::ParameterSet const& pset)
  {
    fVerbose            = pset.get< int    >("Verbose",0);   

    //Center of ND:
    fX0                 = pset.get< double >("X0",0);
    fY0                 = pset.get< double >("Y0",0);
    fZ0                 = pset.get< double >("Z0",728);

    //Grand Det Enclosure:
    fX_max              = pset.get< double >("X_max",304.8);
    fY_max              = pset.get< double >("Y_max",407.9);
    fZ_max              = pset.get< double >("Z_max",1800);
    fX_min              = pset.get< double >("X_min",-1620);
    fY_min              = pset.get< double >("Y_min",-178.9);
    fZ_min              = pset.get< double >("Z_min",-840);
    if(fVerbose>0) PrintConfig("");
  }

  //-------------------------------------------------------------
  void RockCutterAction::PrintConfig(std::string const& /* opt */)
  {
    mf::LogInfo("RockCutterAction")
      << "RockCutterAction::PrintConfig \n"
      << "    Verbose              " << fVerbose              << "\n"
      <<   "   fX0: "  <<fX0 << "   fY0: " <<fY0 << "   fZ0:" << fZ0 << "\n" 
      <<     "   fX_min: "  <<fX_min << "   fX_max: " << fX_max << "\n" 
      <<   "   fY_min: "  <<fY_min << "   fY_max: " << fY_max << "\n" 
      <<   "   fZ_min: "  <<fZ_min << "   fZ_max: " << fZ_max << "\n" ;  

  }

  //-------------------------------------------------------------
  G4ClassificationOfNewTrack 
  RockCutterAction::StackClassifyNewTrack(const G4Track* gtrack)
  {

    if(fVerbose>10) std::cout << " in RockCutter Action " << std::endl;
    //I have to do this to prevent duplicated trackIDes
    //BUT, TIDIUS CODING, DO NOT MODIFY OTHER G4TRACK INFO!
    G4Track *track = const_cast<G4Track*>(gtrack);

    const G4ThreeVector& g4pos = track->GetPosition();  // g4 in mm by default
    // assuming we want positions in cm ...
    const TVector3 startPos(g4pos.x()/cm,g4pos.y()/cm,g4pos.z()/cm);

    //Only do this to charged particles
    if (fabs(track->GetParticleDefinition()->GetPDGCharge())<0.1) {

      //Define grand enclosure: the box containing the tunnel (of course partially)
      //See if it is outside the grand detector enclosure:
      if (startPos.X()>fX_max ||
	  startPos.X()<fX_min ||
	  startPos.Y()>fY_max ||
	  startPos.Y()<fY_min ||
	  startPos.Z()>fZ_max ||
	  startPos.Z()<fZ_min )
	{

	  //OK, surgery is needed:

	  //Plan A: Fast

	  const double energy = (track->GetTotalEnergy())/GeV;
	  const double dEdx = 4.4732e-3;
	  const G4ThreeVector& dir = track->GetMomentumDirection();
	  const TVector3 direction(dir.x(),dir.y(),dir.z());
	  if (startPos.Dot(direction)>0){

	    //      std::cout<<"wrong direction ID:"<<track->GetTrackID()<<std::endl;
	    track->SetTrackStatus(fStopAndKill);
	  }
	  else if (Distance_Wall(startPos.X(),startPos.Y(),startPos.Z())*dEdx>energy){

	    //      std::cout<<"not fast enough ID:"<<track->GetTrackID()<<std::endl;
	    track->SetTrackStatus(fStopAndKill);
	  }

	}
    }

    return fUrgent;
  }


  double RockCutterAction::Distance_Wall(double x,double y,double z)
  {
    double distance = 0;
    double dx = std::min(fabs(x-fX_max),fabs(x-fX_min));
    double dy = std::min(fabs(y-fY_max),fabs(y-fY_min));
    double dz = std::min(fabs(z-fZ_max),fabs(z-fZ_min));

    if (x<fX_max && x>fX_min && y<fY_max && y>fY_min)
      distance = dz;

    else if (y<fY_max && y>-fY_min && z<fZ_max && z>fZ_min)
      distance = dx;

    else if (z<fZ_max && z>-fZ_min && x<fX_max && x>fX_min)
      distance = dy;

    else if (x<fX_max && x>fX_min)
      distance = sqrt(dy*dy+dz*dz);

    else if (y<fY_max && y>fY_min)
      distance = sqrt(dx*dx+dz*dz);

    else if (z<fZ_max && z>fZ_min)
      distance = sqrt(dx*dx+dy*dy);

    else distance = sqrt(dx*dx+dy*dy+dz*dz);

    return distance;

  }


  /*
  //-------------------------------------------------------------
  void RockCutterAction::BeginOfEventAction(const G4Event*)
  {

  }

  //-------------------------------------------------------------
  void RockCutterAction::EndOfEventAction(const G4Event*)
  {

  }

  //-------------------------------------------------------------
  void RockCutterAction::PreTrackingAction(const G4Track* track)
  {    
  }

  //-------------------------------------------------------------
  void RockCutterAction::PostTrackingAction( const G4Track* track)
  {

  }

  //-------------------------------------------------------------
  void RockCutterAction::SteppingAction(const G4Step* step)
  {

  }

  //-------------------------------------------------------------
  void RockCutterAction::StackNewStage()
  {

  }

  //-------------------------------------------------------------
  void RockCutterAction::StackPrepareNewEvent()
  {

  }
  */
  
} // end namespace
