///////////////////////////////////////////////////////////////////////////  
/// \brief   A producer that reconstruct shower energy, topological variables 
///          and PID information 
/// \author  Jianming Bian - bianjm@umn.physics.edu
////////////////////////////////////////////////////////////////////////
extern "C" {
#include <sys/types.h>
#include <sys/stat.h>
}

#include <fstream>
#include <cstdlib>
#include <vector>
#include <string>
#include <cmath>

// ROOT includes
#include "TFile.h"
#include "TH1D.h"
#include "TVector3.h"
#include "TTree.h"
#include "TMultiLayerPerceptron.h"

// ART includes
#include "art/Framework/Core/EDProducer.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "art/Framework/Services/Optional/TFileService.h"
#include "art/Framework/Core/FindManyP.h"
#include "art/Persistency/Common/Ptr.h"
#include "cetlib/search_path.h"

// NOvA includes
#include "Calibrator/Calibrator.h"
#include "CalibrationUtils/CalibUtil.h"
#include "Geometry/Geometry.h"
#include "GeometryObjects/CellGeo.h"
//#include "Geometry/func/CellUniqueId.h"
#include "RecoBase/CellHit.h"
#include "RecoBase/Cluster.h"
#include "RecoBase/FilterList.h"
#include "RecoBase/RecoHit.h"
#include "RecoBase/Track.h"
#include "RecoBase/Shower.h"
#include "RecoBase/Vertex.h"
#include "SimulationBase/MCParticle.h"
#include "SimulationBase/MCTruth.h"
#include "Simulation/FLSHitList.h"
#include "Utilities/func/EnvExpand.h"
#include "Utilities/func/MathUtil.h"
#include "Utilities/AssociationUtil.h"
#include "DAQChannelMap/DAQChannelMapConstants.h"
//#include "Geometry/OfflineChan.h"
#include "RecoJMShower/JMShower.h"


namespace jmshower {

  /// Store some RVP-like variables that could potentially factor into the EID
  struct RVPInfo 
  {
    double efrac_2slides; ///< highest energy fraction in 2 plane sliding window out of slice
    double efrac_4slides; ///< highest energy fraction in 4 plane sliding window out of slice
    double efrac_6slides; ///< highest energy fraction in 6 plane sliding window out of slice
    double efrac_20slides; ///< highest energy fraction in first 20 planes
    double efrac_2sigmaHOR; ///< energy fraction within 2sigma of slice center
    double efrac_4sigmaHOR; ///< energy fraction within 4sigma of slice center
 
    double eiso_2sigmaHOR; ///< energy isolation of 2sigma raod
    double eiso_4sigmaHOR; ///< energy isolation of 4sigma raod
    double eiso_6sigmaHOR; ///< energy isolation of 6sigma raod
  };

  class RecoJMShower : public art::EDProducer 
  {
  public:
    RecoJMShower(fhicl::ParameterSet const& pset);
    ~RecoJMShower();

    void preEvent(art::Event const& evt);
    void postBeginRun(art::Run const& run);
    void reconfigure(const fhicl::ParameterSet &pset);
    void produce(art::Event& evt);
    void beginRun(art::Run& run);

    //  Histograms for probability calculation
    TH1D *fHtPlaneDedxE[4][11][200];
    TH1D *fHtPlaneDedxG[4][11][200];
    TH1D *fHtPlaneDedxMu[4][11][200];
    TH1D *fHtPlaneDedxPi0[4][11][200];
    TH1D *fHtPlaneDedxHad[4][11][200];
    TH1D *fHtPlaneDedxP[4][11][200];
    TH1D *fHtPlaneDedxN[4][11][200];
    TH1D *fHtPlaneDedxPi[4][11][200];
    TH1D *fHtPlaneDedxEqe[4][11][200];
    TH1D *fHtPlaneDedxEres[4][11][200];
    TH1D *fHtPlaneDedxEdis[4][11][200];
    TH1D *fHtPlaneDedxEcoh[4][11][200];

    unsigned fHtExpPlaneE[4][11];
    unsigned fHtExpPlaneG[4][11];
    unsigned fHtExpPlaneMu[4][11];
    unsigned fHtExpPlanePi0[4][11];
    unsigned fHtExpPlaneHad[4][11];
    unsigned fHtExpPlaneP[4][11];
    unsigned fHtExpPlaneN[4][11];
    unsigned fHtExpPlanePi[4][11];
    unsigned fHtExpPlaneEqe[4][11];
    unsigned fHtExpPlaneEres[4][11];
    unsigned fHtExpPlaneEdis[4][11];
    unsigned fHtExpPlaneEcoh[4][11];



/*
    TH1D *fHtmipPlaneDedxE[11][200];
    TH1D *fHtmipPlaneDedxG[11][200];
    TH1D *fHtmipPlaneDedxMu[11][200];
    TH1D *fHtmipPlaneDedxPi0[11][200];
    TH1D *fHtmipPlaneDedxHad[11][200];
    TH1D *fHtmipPlaneDedxP[11][200];
    TH1D *fHtmipPlaneDedxN[11][200];
    TH1D *fHtmipPlaneDedxPi[11][200];
    TH1D *fHtmipPlaneDedxEqe[11][200];
    TH1D *fHtmipPlaneDedxEres[11][200];
    TH1D *fHtmipPlaneDedxEdis[11][200];
    TH1D *fHtmipPlaneDedxEcoh[11][200];

    TH1D *fHtshowerPlaneDedxE[11][200];
    TH1D *fHtshowerPlaneDedxG[11][200];
    TH1D *fHtshowerPlaneDedxMu[11][200];
    TH1D *fHtshowerPlaneDedxPi0[11][200];
    TH1D *fHtshowerPlaneDedxHad[11][200];
    TH1D *fHtshowerPlaneDedxP[11][200];
    TH1D *fHtshowerPlaneDedxN[11][200];
    TH1D *fHtshowerPlaneDedxPi[11][200];
    TH1D *fHtshowerPlaneDedxEqe[11][200];
    TH1D *fHtshowerPlaneDedxEres[11][200];
    TH1D *fHtshowerPlaneDedxEdis[11][200];
    TH1D *fHtshowerPlaneDedxEcoh[11][200];

    TH2F *fHt2dPlaneDedxE[11][200];
    TH2F *fHt2dPlaneDedxG[11][200];
    TH2F *fHt2dPlaneDedxMu[11][200];
    TH2F *fHt2dPlaneDedxPi0[11][200];
    TH2F *fHt2dPlaneDedxHad[11][200];
    TH2F *fHt2dPlaneDedxP[11][200];
    TH2F *fHt2dPlaneDedxN[11][200];
    TH2F *fHt2dPlaneDedxPi[11][200];
    TH2F *fHt2dPlaneDedxEqe[11][200];
    TH2F *fHt2dPlaneDedxEres[11][200];
    TH2F *fHt2dPlaneDedxEdis[11][200];
    TH2F *fHt2dPlaneDedxEcoh[11][200];
*/
 
    TH1D *fHtCellTransDedxE[4][11][20];
    TH1D *fHtCellTransDedxG[4][11][20];
    TH1D *fHtCellTransDedxMu[4][11][20];
    TH1D *fHtCellTransDedxPi0[4][11][20];
    TH1D *fHtCellTransDedxHad[4][11][20];
    TH1D *fHtCellTransDedxP[4][11][20];
    TH1D *fHtCellTransDedxN[4][11][20];
    TH1D *fHtCellTransDedxPi[4][11][20];
    TH1D *fHtCellTransDedxEqe[4][11][20];
    TH1D *fHtCellTransDedxEres[4][11][20];
    TH1D *fHtCellTransDedxEdis[4][11][20];
    TH1D *fHtCellTransDedxEcoh[4][11][20];



    std::string fFnameWeightsShape;
    std::string fFnameWeightsShapeElec;
    std::string fFnameWeightsShapeQE;
    std::string fFnameWeightsShapeRES;
    std::string fFnameWeightsShapeDIS;
    TMultiLayerPerceptron *mlp_shape;
    TMultiLayerPerceptron *mlp_shape_elec;
    TMultiLayerPerceptron *mlp_shape_E;
    TMultiLayerPerceptron *mlp_shape_NoCos;
    TMultiLayerPerceptron *mlp_shape_E_NoCos;
    TMultiLayerPerceptron *mlp_shape_QE;
    TMultiLayerPerceptron *mlp_shape_RES;
    TMultiLayerPerceptron *mlp_shape_DIS;

//  Energy deconvolution map 
    std::map<geo::OfflineChan, std::vector<double> > fHitEWeightMap;
    std::map<geo::OfflineChan, std::vector<double> > fHitDistMap;
    std::map<geo::OfflineChan, double> fHitEMap;
    std::vector<std::vector<jmshower::PCluster> > fShowerPClusterCol;
    std::vector<std::vector<jmshower::PCluster> > fTrackPClusterCol;
    std::vector<std::vector<jmshower::TCCluster> > fShowerTCClusterCol;
    std::vector<jmshower::JMShower> fExcludeShowerCol;



//  Geometry
    double GetPointDoca(TVector3 vtx, TVector3 start, TVector3 stop);
    TVector3 GetCellDistToPoint(unsigned int plane, unsigned int cell, TVector3 point);
    double GetCellDistToTrk(unsigned int plane, unsigned int cell, TVector3 start, TVector3 stop);

    TVector3 GetShwStop(rb::Prong shw);//Calculate the end point of a shower
    TVector3 GetShwStop(art::Ptr<rb::Prong>  shw);//Calculate the end point of a shower


    TVector3 GetTrkHitPos(unsigned int plane, unsigned int cell, art::Ptr<rb::Prong> trk);
    TVector3 GetTrkHitPos(unsigned int plane, unsigned int cell, rb::Prong trk);
    TVector3 GetTrkHitPos(unsigned int plane, unsigned int cell, TVector3 trkdir, TVector3 trkstart, TVector3 trkstop);
    TVector3 GetTrkPlanePos(unsigned int plane, rb::Prong trk);
    unsigned int GetTrkPlaneCell(unsigned int plane, art::Ptr<rb::Prong> trk);
    unsigned int GetTrkPlaneCell(unsigned int plane, rb::Prong trk);
    unsigned int GetTrkCPlaneCell(unsigned int plane, unsigned int i);
    TVector3 GetTrkPlaneDistToEdge(unsigned int plane, art::Ptr<rb::Prong> trk);
    TVector3 GetTrkPlaneDistToEdge(unsigned int plane, rb::Prong trk);
    bool IsFiducial(rb::Prong shower);
    double GetTrkDistToEdge(rb::Prong shower);
    double GetTrkHitPath(unsigned int plane, unsigned int cell, rb::Prong shower);
    unsigned int ContStartPlane(rb::Prong shower, unsigned int startp, int contp);
// Energy
    double DepositEnergy(rb::Cluster cluster, int i);
    double DepositEnergy(rb::Prong shower);
    double Energy(rb::Cluster cluster, int i);
    double Energy(rb::Prong shower);
    double ECorrMC(double gev);
    double CellADCToGeV(double d, geo::View_t view, int detid);
    double CellEDataMC(double gev, double w, geo::View_t view, int detid); // correct cell DATA/MC difference

    double PlaneDedxDataMC(double gev, int detid); // correct plane dedx  DATA/MC difference
    double CellTransDataMC(double gev, int detid); // correct transverse dedx DATA/MC difference

    double Radius(rb::Prong shower);
    double Radius(rb::Cluster cluster, rb::Prong shower);
    double RadiusV(rb::Cluster cluster);

// Shower shape
    double Gap(rb::Prong shower, TVector3 vertex);
    double GetPlaneDedx(rb::Prong shower, unsigned int plane);
    double GetPlaneDedxProb(rb::Prong shower, int ireg, int igev, unsigned int plane, double dedx, const int type);
//    double GetPlaneDedxProb(rb::Prong shower, int igev, unsigned int plane, double dedx, unsigned int nmip, const int type);
    double GetInterPlaneDedxProb(rb::Prong shower,  unsigned int plane, double dedx, const int type, TVector3 pos, int iReg, int igev0, int igev1, double e0, double e1);
//    double GetInterPlaneDedxProb(rb::Prong shower,  unsigned int plane, double dedx, const int type, unsigned int nmip);
    double GetDedxLongLL(rb::Prong shower, const int type);
    double GetDedxInvLongLL(rb::Prong shower, const int type);
    double GetDedxLongLL(rb::Prong shower, const int type, unsigned int startp,  int contp);

    double GetCellTransDedx(rb::Prong shower, unsigned int cell);
    double GetCellTransDedxProb(rb::Prong shower, int ireg, int igev, unsigned int cell, double dedx, const int type);
    double GetInterCellTransDedxProb(rb::Prong shower, unsigned int cell, double dedx, const int type, TVector3 pos, int iReg, int igev0, int igev1, double e0, double e1);
    double GetDedxTransLL(rb::Prong shower, const int type);
    bool IsMuon(jmshower::JMShower shower);
    bool IsInvPhoton(jmshower::JMShower shower);

    unsigned int NMIPPlane(rb::Prong shower);
    unsigned int NMIPPlane(rb::Prong shower, unsigned int startp);
    double CalcANN(unsigned int mode, jmshower::JMShower shower);

    //function to store rvp type variables that could potentially be used
    RVPInfo GetRVPStats(art::Handle<std::vector<rb::Cluster> > slices, unsigned int ic);
    TVector3 GetCellNodePos(unsigned int plane1, unsigned int cell1, unsigned int plane2, unsigned int cell2);

    std::string fLibPath;
    std::string pSliceDir;
    std::string pVertexLabel;
    std::string pInstLabel;
    std::string pTrackDir;
    double pCellHitPHThresh;
    int    pPIDHist;
    int    pRecoANN;
    bool   pRecoPID;
    bool   pRecoRVP;
    bool   pRecoNode;
    bool   pRecoPlaneR;
    bool   pRecoInvLL;
    bool   pRecluster;
    bool   pRecoVtx;
    bool   pVtxFit;
    bool   pBadChannel;
    bool   pCorAbsCell;
    bool   pCorPigtail;
    bool   pCorDataCell;
    bool   pCorDataDedx;
    bool   pUseUncalib;

    int pXEdgeId1;
    int pXEdgeId2;
    int pXEdgeId3;
    int pXEdgeId4;
    int pYEdgeId1;
    int pYEdgeId2;
    int pYEdgeId3;
    int pYEdgeId4;
    int pZEdgeId1;
    int pZEdgeId2;
    int pZEdgeId3;
    int pZEdgeId4;

    double pXEdge1;
    double pXEdge2;
    double pXEdge3;
    double pXEdge4;
    double pYEdge1;
    double pYEdge2;
    double pYEdge3;
    double pYEdge4;
    double pZEdge1;
    double pZEdge2;
    double pZEdge3;
    double pZEdge4;

// Exclude cluster
    TVector3 GetCentroid(rb::Cluster cluster);

  public:
    std::vector<jmshower::JMShower> RecoShowers(art::EDProducer const&    prod,
                                                art::Event& evt,
                                                std::unique_ptr< std::vector<jmshower::JMShower> >& recoshowercol,
                                                std::unique_ptr< art::Assns<jmshower::JMShower, rb::Prong> >&);

//    std::vector<jmshower::JMShower> RecoShowers(const art::Event& evt);
    std::vector<jmshower::JMShower> ExclShowers(const art::Event& evt);

  private:
    bool PreSel(const art::Event& evt);
    bool LoadDedxHistograms(int detid);
    bool LoadTreeWeights(int detid);

    int fDetid;
    bool isrealdata;
    bool fLoaded; ///< Have we done beginRun once yet?
    std::vector<int> fprongidx;
  };

} // end namespace jmshower

namespace jmshower 
{  
//------------------------------------------------------------
  PCluster::PCluster() : fP(9999)
  {
  }

  PCluster::~PCluster()
  {
  }

  TCCluster::TCCluster() : fTc(9999)
  {
  }

  TCCluster::~TCCluster()
  {
  }


//------------------------------------------------------------
  RecoJMShower::RecoJMShower(fhicl::ParameterSet const& pset):
    pSliceDir(pset.get<std::string>("SliceDir")),
    pTrackDir(pset.get<std::string>("TrackDir")),
    pCellHitPHThresh(50.),
    pPIDHist(1),
    pRecoANN(0),
    pRecoPID(true),
    pRecoRVP(true),
    pRecoNode(true),
    pRecoPlaneR(true),
    pRecoInvLL(true),
    pRecluster(true),
    pRecoVtx(true),
    pVtxFit(false),
    pBadChannel(false),
    pCorAbsCell(false),
    pCorPigtail(true),
    pCorDataCell(false),
    pCorDataDedx(false),
    pUseUncalib(false),
    pXEdge1(-774.7),
    pXEdge2(774.7),
    pXEdge3(-774.7),
    pXEdge4(-774.7),
    pYEdge1(-774.7),
    pYEdge2(774.7),
    pYEdge3(-774.7),
    pYEdge4(-774.7),
    pZEdge1(0.0),
    pZEdge2(5959.08),
    pZEdge3(0.0),
    pZEdge4(0.0),
    fLoaded(false)
  {
    reconfigure(pset);
    produces< std::vector<jmshower::JMShower>   >();
    produces< art::Assns<jmshower::JMShower, rb::Cluster> >();
    produces< art::Assns<jmshower::JMShower, rb::Prong> >();
  }
  
  //------------------------------------------------------------
  RecoJMShower::~RecoJMShower()  
  {
  }
  
  //------------------------------------------------------------
  void RecoJMShower::reconfigure(const fhicl::ParameterSet& pset)
  {
    fLibPath         = pset.get< std::string >("LibraryPath");
    pSliceDir        = pset.get< std::string >("SliceDir");
    pVertexLabel    =  pset.get< std::string >("VertexLabel");
    pTrackDir        = pset.get< std::string >("TrackDir");
    pInstLabel       = pset.get< std::string >("InstLabel");
    pCellHitPHThresh = pset.get< double      >("CellHitPHThresh");
    pPIDHist         = pset.get< int         >("PIDHist");
    pRecoANN         = pset.get< int         >("RecoANN");
    pRecoPID         = pset.get< bool        >("RecoPID");
    pRecoRVP         = pset.get< bool        >("RecoRVP");
    pRecoNode        = pset.get< bool        >("RecoNode");
    pRecoPlaneR      = pset.get< bool        >("RecoPlaneR");
    pRecoInvLL       = pset.get< bool        >("RecoInvLL");
    pRecluster       = pset.get< bool        >("Recluster");
    pRecoVtx         = pset.get< bool        >("RecoVtx");
    pVtxFit          = pset.get< bool        >("VtxFit");
    pBadChannel      = pset.get< bool        >("BadChannel");
    pCorAbsCell      = pset.get< bool        >("CorAbsCell");
    pCorPigtail      = pset.get< bool        >("CorPigtail");
    pCorDataCell     = pset.get< bool        >("CorDataCell");
    pCorDataDedx     = pset.get< bool        >("CorDataDedx");
    pUseUncalib      = pset.get< bool        >("UseUncalib");
    pXEdgeId1        = pset.get< int      >("XEdgeId1");
    pXEdgeId2        = pset.get< int      >("XEdgeId2");
    pXEdgeId3        = pset.get< int      >("XEdgeId3");
    pXEdgeId4        = pset.get< int      >("XEdgeId4");
    pYEdgeId1        = pset.get< int      >("YEdgeId1");
    pYEdgeId2        = pset.get< int      >("YEdgeId2");
    pYEdgeId3        = pset.get< int      >("YEdgeId3");
    pYEdgeId4        = pset.get< int      >("YEdgeId4");
    pZEdgeId1        = pset.get< int      >("ZEdgeId1");
    pZEdgeId2        = pset.get< int      >("ZEdgeId2");
    pZEdgeId3        = pset.get< int      >("ZEdgeId3");
    pZEdgeId4        = pset.get< int      >("ZEdgeId4");
    pXEdge1          = pset.get< double   >("XEdge1");
    pXEdge2          = pset.get< double   >("XEdge2");
    pXEdge3          = pset.get< double   >("XEdge3");
    pXEdge4          = pset.get< double   >("XEdge4");
    pYEdge1          = pset.get< double   >("YEdge1");
    pYEdge2          = pset.get< double   >("YEdge2");
    pYEdge3          = pset.get< double   >("YEdge3");
    pYEdge4          = pset.get< double   >("YEdge4");
    pZEdge1          = pset.get< double   >("ZEdge1");
    pZEdge2          = pset.get< double   >("ZEdge2");
    pZEdge3          = pset.get< double   >("ZEdge3");
    pZEdge4          = pset.get< double   >("ZEdge4");
  }
  
  //----------------------------------------------------------

  void RecoJMShower::beginRun(art::Run& /*run*/)
  {
    if(!fLoaded){
      art::ServiceHandle<geo::Geometry> geom;
      fDetid = geom->DetId();
      if(pRecoPID==true)LoadDedxHistograms(geom->DetId());
      LoadTreeWeights(geom->DetId());
      fLoaded = true;
    }


  }

  void RecoJMShower::postBeginRun(art::Run const& run)
  {
  }
  
  void RecoJMShower::preEvent(art::Event const& evt) {
  }
  
  //------------------------------------------------------------
  //   Load Probability Histograms and Training Tree Weight Files 
  //------------------------------------------------------------
  
  bool RecoJMShower::LoadDedxHistograms(int detid) 
  {

    std::string libpath = util::EnvExpansion(fLibPath);
    
    std::string fnameE(libpath);
    std::string fnameG(libpath);
    std::string fnameMu(libpath);
    std::string fnamePi0(libpath);
//    std::string fnameHad(libpath);
    std::string fnameP(libpath);
    std::string fnameN(libpath);
    std::string fnamePi(libpath);
    std::string fnameEqe(libpath);
    std::string fnameEres(libpath);
    std::string fnameEdis(libpath);
    std::string fnameEcoh(libpath);

    if(pPIDHist==1){
      switch (detid) {
      case novadaq::cnv::kFARDET:

        fnameE += "e10kdedxhist_fd_nue_cc.root";
        fnameG   += "g10kdedxhist_fd_nue_cc.root";
        fnameMu  += "mu10kdedxhist_fd_nue_cc.root";
        fnamePi0 += "pi010kdedxhist_fd_nue_cc.root";
 //       fnameHad += "had10kdedxhist_fd_nue_ccqe.root";
        fnameP += "protondedxhist_fd.root";
        fnameN += "neutrondedxhist_fd.root";
        fnamePi += "pidedxhist_fd.root";
        fnameEqe += "e10kdedxhist_fd_nue_ccqe.root";
        fnameEres += "e10kdedxhist_fd_nue_ccres.root";
        fnameEdis += "e10kdedxhist_fd_nue_ccdis.root";
        fnameEcoh += "e10kdedxhist_fd_nue_cccoh.root";

/*
        fnameE += "e10kdedxhist_fd_nue_cc.root";
        fnameG   += "g10kdedxhist_genie_fd_nhc.root";
        fnameMu  += "mu10kdedxhist_genie_fd_nhc.root";
        fnamePi0 += "pi010kdedxhist_genie_fd_nhc.root";
 //       fnameHad += "had10kdedxhist_fd_nue_ccqe.root";
        fnameP += "protondedxhist_fd.root";
        fnameN += "neutrondedxhist_fd.root";
        fnamePi += "pidedxhist_fd.root";
        fnameEqe += "e10kdedxhist_fd_nue_ccqe.root";
        fnameEres += "e10kdedxhist_fd_nue_ccres.root";
        fnameEdis += "e10kdedxhist_fd_nue_ccdis.root";
        fnameEcoh += "e10kdedxhist_fd_nue_cccoh.root";
*/
        break;
      case novadaq::cnv::kNDOS:    
        fnameE += "e10kdedxhist_fd_nue_cc.root";
        fnameG   += "g10kdedxhist_fd_nue_cc.root";
        fnameMu  += "mu10kdedxhist_fd_nue_cc.root";
        fnamePi0 += "pi010kdedxhist_fd_nue_cc.root";
 //       fnameHad += "had10kdedxhist_fd_nue_ccqe.root";
        fnameP += "protondedxhist_fd.root";
        fnameN += "neutrondedxhist_fd.root";
        fnamePi += "pidedxhist_fd.root";
        fnameEqe += "e10kdedxhist_fd_nue_ccqe.root";
        fnameEres += "e10kdedxhist_fd_nue_ccres.root";
        fnameEdis += "e10kdedxhist_fd_nue_ccdis.root";
        fnameEcoh += "e10kdedxhist_fd_nue_cccoh.root";
        break;
      case novadaq::cnv::kNEARDET: 

        fnameE += "e10kdedxhist_fd_nue_cc.root";
        fnameG   += "g10kdedxhist_fd_nue_cc.root";
        fnameMu  += "mu10kdedxhist_fd_nue_cc.root";
        fnamePi0 += "pi010kdedxhist_fd_nue_cc.root";
 //       fnameHad += "had10kdedxhist_fd_nue_ccqe.root";
        fnameP += "protondedxhist_fd.root";
        fnameN += "neutrondedxhist_fd.root";
        fnamePi += "pidedxhist_fd.root";
        fnameEqe += "e10kdedxhist_fd_nue_ccqe.root";
        fnameEres += "e10kdedxhist_fd_nue_ccres.root";
        fnameEdis += "e10kdedxhist_fd_nue_ccdis.root";
        fnameEcoh += "e10kdedxhist_fd_nue_cccoh.root";

        break;
      default:
        throw cet::exception("BAD DETID") << __FILE__ << ":" << __LINE__ 
  					<< " Bad detector id = " << detid;
      }
    }else {
      switch (detid) {
      case novadaq::cnv::kFARDET:

        fnameE += "e10kdedxhist_fd_nue_cc.root";
        fnameG   += "g10kdedxhist_fd_nue_cc.root";
        fnameMu  += "mu10kdedxhist_fd_nue_cc.root";
        fnamePi0 += "pi010kdedxhist_fd_nue_cc.root";
 //       fnameHad += "had10kdedxhist_fd_nue_ccqe.root";
        fnameP += "protondedxhist_fd.root";
        fnameN += "neutrondedxhist_fd.root";
        fnamePi += "pidedxhist_fd.root";
        fnameEqe += "e10kdedxhist_fd_nue_ccqe.root";
        fnameEres += "e10kdedxhist_fd_nue_ccres.root";
        fnameEdis += "e10kdedxhist_fd_nue_ccdis.root";
        fnameEcoh += "e10kdedxhist_fd_nue_cccoh.root";

        break;
      case novadaq::cnv::kNDOS:    

        fnameE += "e10kdedxhist_fd_nue_cc.root";
        fnameG   += "g10kdedxhist_fd_nue_cc.root";
        fnameMu  += "mu10kdedxhist_fd_nue_cc.root";
        fnamePi0 += "pi010kdedxhist_fd_nue_cc.root";
 //       fnameHad += "had10kdedxhist_fd_nue_ccqe.root";
        fnameP += "protondedxhist_fd.root";
        fnameN += "neutrondedxhist_fd.root";
        fnamePi += "pidedxhist_fd.root";
        fnameEqe += "e10kdedxhist_fd_nue_ccqe.root";
        fnameEres += "e10kdedxhist_fd_nue_ccres.root";
        fnameEdis += "e10kdedxhist_fd_nue_ccdis.root";
        fnameEcoh += "e10kdedxhist_fd_nue_cccoh.root";

        break;
      case novadaq::cnv::kNEARDET: 

        fnameE += "e10kdedxhist_fd_nue_cc.root";
        fnameG   += "g10kdedxhist_fd_nue_cc.root";
        fnameMu  += "mu10kdedxhist_fd_nue_cc.root";
        fnamePi0 += "pi010kdedxhist_fd_nue_cc.root";
 //       fnameHad += "had10kdedxhist_fd_nue_ccqe.root";
        fnameP += "protondedxhist_fd.root";
        fnameN += "neutrondedxhist_fd.root";
        fnamePi += "pidedxhist_fd.root";
        fnameEqe += "e10kdedxhist_fd_nue_ccqe.root";
        fnameEres += "e10kdedxhist_fd_nue_ccres.root";
        fnameEdis += "e10kdedxhist_fd_nue_ccdis.root";
        fnameEcoh += "e10kdedxhist_fd_nue_cccoh.root";
        break; 
      default:
        throw cet::exception("BAD DETID") << __FILE__ << ":" << __LINE__ 
  					<< " Bad detector id = " << detid;
      }
    }

/*  histograms stored in /nova/data/users/bianjm/share/development/RecoJMShower/histograms for the test versions

    std::string fFnameE;    
    std::string fFnameG;    
    std::string fFnameMu;    
    std::string fFnamePi0;    
    std::string fFnameHad;    
    std::string fFnameP;    
    std::string fFnameN;    
    std::string fFnamePi;    
    std::string fFnameEqe;    
    std::string fFnameEres;    
    std::string fFnameEdis;    
    std::string fFnameEcoh;    
// constructor decides if initialized value is a path or an environment variable
    cet::search_path sp("FW_SEARCH_PATH");
    sp.find_file(fnameE, fFnameE);
    sp.find_file(fnameG, fFnameG);
    sp.find_file(fnameMu, fFnameMu);
    sp.find_file(fnamePi0, fFnamePi0);
    sp.find_file(fnameHad, fFnameHad);
    sp.find_file(fnameP, fFnameP);
    sp.find_file(fnameN, fFnameN);
    sp.find_file(fnamePi, fFnamePi);
    sp.find_file(fnameEqe, fFnameEqe);
    sp.find_file(fnameEres, fFnameEres);
    sp.find_file(fnameEdis, fFnameEdis);
    sp.find_file(fnameEcoh, fFnameEcoh);
    
    struct stat sb;

    if (fFnameE.empty() || stat(fFnameE.c_str(), &sb)!=0){
      // Failed to resolve the file name
      mf::LogWarning("RecoJMShower") << "Histogram -- E " << fFnameE << " not found.";
      throw cet::exception("FILE_NOT_FOUND") << "Histogram file not found.";
    }

    if (fFnameEqe.empty() || stat(fFnameEqe.c_str(), &sb)!=0){
      // Failed to resolve the file name
      mf::LogWarning("RecoJMShower") << "Histogram -- Eqe " << fFnameEqe << " not found.";
      throw cet::exception("FILE_NOT_FOUND") << "Histogram file not found.";
    }
    if (fFnameEres.empty() || stat(fFnameEres.c_str(), &sb)!=0){
      // Failed to resolve the file name
      mf::LogWarning("RecoJMShower") << "Histogram -- Eres " << fFnameEres << " not found.";
      throw cet::exception("FILE_NOT_FOUND") << "Histogram file not found.";
    }
    if (fFnameEdis.empty() || stat(fFnameEdis.c_str(), &sb)!=0){
      // Failed to resolve the file name
      mf::LogWarning("RecoJMShower") << "Histogram -- Edis " << fFnameEdis << " not found.";
      throw cet::exception("FILE_NOT_FOUND") << "Histogram file not found.";
    }
    if (fFnameEcoh.empty() || stat(fFnameEcoh.c_str(), &sb)!=0){
      // Failed to resolve the file name
      mf::LogWarning("RecoJMShower") << "Histogram -- Ecoh " << fFnameEcoh << " not found.";
      throw cet::exception("FILE_NOT_FOUND") << "Histogram file not found.";
    }

    if (fFnameG.empty() || stat(fFnameG.c_str(), &sb)!=0){
      // Failed to resolve the file name
      mf::LogWarning("RecoJMShower") << "Histogram -- G " << fFnameG << " not found.";
      throw cet::exception("FILE_NOT_FOUND") << "Histogram file not found.";
    }
    if (fFnameMu.empty() || stat(fFnameMu.c_str(), &sb)!=0){
      // Failed to resolve the file name
      mf::LogWarning("RecoJMShower") << "Histogram -- Mu " << fFnameMu << " not found.";
      throw cet::exception("FILE_NOT_FOUND") << "Histogram file not found.";
    }
    if (fFnamePi0.empty() || stat(fFnamePi0.c_str(), &sb)!=0){
      // Failed to resolve the file name
      mf::LogWarning("RecoJMShower") << "Histogram -- Pi0 " << fFnamePi0 << " not found.";
      throw cet::exception("FILE_NOT_FOUND") << "Histogram file not found.";
    }

    if (fFnameHad.empty() || stat(fFnameHad.c_str(), &sb)!=0){
      // Failed to resolve the file name
      mf::LogWarning("RecoJMShower") << "Histogram -- Had " << fFnameHad << " not found.";
      throw cet::exception("FILE_NOT_FOUND") << "Histogram file not found.";
    }

    if (fFnameP.empty() || stat(fFnameP.c_str(), &sb)!=0){
      // Failed to resolve the file name
      mf::LogWarning("RecoJMShower") << "Histogram -- P " << fFnameP << " not found.";
      throw cet::exception("FILE_NOT_FOUND") << "Histogram file not found.";
    }

    if (fFnameN.empty() || stat(fFnameN.c_str(), &sb)!=0){
      // Failed to resolve the file name
      mf::LogWarning("RecoJMShower") << "Histogram -- N " << fFnameN << " not found.";
      throw cet::exception("FILE_NOT_FOUND") << "Histogram file not found.";
    }

    if (fFnamePi.empty() || stat(fFnamePi.c_str(), &sb)!=0){
      // Failed to resolve the file name
      mf::LogWarning("RecoJMShower") << "Histogram -- Pi " << fFnamePi << " not found.";
      throw cet::exception("FILE_NOT_FOUND") << "Histogram file not found.";
    }

*/
    TFile hfileE(fnameE.c_str()); 
    TFile hfileG(fnameG.c_str()); 
    TFile hfileMu(fnameMu.c_str()); 
    TFile hfilePi0(fnamePi0.c_str()); 
//    TFile hfileHad(fnameHad.c_str()); 
    TFile hfileP(fnameP.c_str()); 
    TFile hfileN(fnameN.c_str()); 
    TFile hfilePi(fnamePi.c_str()); 
    TFile hfileEqe(fnameEqe.c_str()); 
    TFile hfileEres(fnameEres.c_str()); 
    TFile hfileEdis(fnameEdis.c_str()); 
    TFile hfileEcoh(fnameEcoh.c_str()); 
    std::cout<<"PID Dedx File "<<fnameE.c_str()<<" is loaded."<<std::endl;
    std::cout<<"PID Dedx File "<<fnameG.c_str()<<" is loaded."<<std::endl;
    std::cout<<"PID Dedx File "<<fnameMu.c_str()<<" is loaded."<<std::endl;
    std::cout<<"PID Dedx File "<<fnamePi0.c_str()<<" is loaded."<<std::endl;
//    std::cout<<"PID Dedx File "<<fnameHad.c_str()<<" is loaded."<<std::endl;
    std::cout<<"PID Dedx File "<<fnameP.c_str()<<" is loaded."<<std::endl;
    std::cout<<"PID Dedx File "<<fnameN.c_str()<<" is loaded."<<std::endl;
    std::cout<<"PID Dedx File "<<fnamePi.c_str()<<" is loaded."<<std::endl;
    std::cout<<"PID Dedx File "<<fnameEqe.c_str()<<" is loaded."<<std::endl;
    std::cout<<"PID Dedx File "<<fnameEres.c_str()<<" is loaded."<<std::endl;
    std::cout<<"PID Dedx File "<<fnameEdis.c_str()<<" is loaded."<<std::endl;
    std::cout<<"PID Dedx File "<<fnameEcoh.c_str()<<" is loaded."<<std::endl;

    //The detector is divided into 4 regions in XY, these regions extend in z. 
    //Region 1 is the top left quadrant in XY when looking at the front of the detector
    //The training histograms are then divided into 11 energy regions from 0 to 5 GeV.
    //In the training each shower is put into a bin based on its reco energy
    //The first and last energy bin are 0-0.25 and 4.75-5 GeV.  All bins inbetween are 0.5 GeV wide.
    //For each region and each energy there is a de/dx histogram for each plane from 0 to 200
    //the histograms being loaded below are for longitudinal dedx.
    //These training histograms are also broken up by particle type (electron, muon, gamma, proton, pi0, pion, neutron)

    //NOTE: Jianming is exploring using the shower length as a training variable.
    //In order to do this the variables below are initialized and then store the
    //shower length at each energy in each region.  This is done by finding the 
    //last plane in which there was energy deposited.
    //These variables are not currently being used in the production version of the algorithm
    for(int r=0; r<4; r++){
      for(int i=0; i<11; i++){
        fHtExpPlaneE[r][i]=0;
        fHtExpPlaneG[r][i]=0;
        fHtExpPlaneMu[r][i]=0;
        fHtExpPlanePi0[r][i]=0;
        fHtExpPlaneHad[r][i]=0;
        fHtExpPlaneP[r][i]=0;
        fHtExpPlaneN[r][i]=0;
        fHtExpPlanePi[r][i]=0;
        fHtExpPlaneEqe[r][i]=0;
        fHtExpPlaneEres[r][i]=0;
        fHtExpPlaneEdis[r][i]=0;
        fHtExpPlaneEcoh[r][i]=0;
      }
    }

    for(int r=0; r<4; r++){
      for(int i=0; i<11; i++){
        for(int l=0; l<200; l++){
          char cht[200];
          sprintf(cht,"ht_%d_%d_%d",r,i,l);
          fHtPlaneDedxE[r][i][l] = (TH1D*)hfileE.Get(cht);
          fHtPlaneDedxG[r][i][l] = (TH1D*)hfileG.Get(cht);
          fHtPlaneDedxMu[r][i][l] = (TH1D*)hfileMu.Get(cht);
          fHtPlaneDedxPi0[r][i][l] = (TH1D*)hfilePi0.Get(cht);
          fHtPlaneDedxP[r][i][l] = (TH1D*)hfileP.Get(cht);
          fHtPlaneDedxN[r][i][l] = (TH1D*)hfileN.Get(cht);
          fHtPlaneDedxPi[r][i][l] = (TH1D*)hfilePi.Get(cht);
          fHtPlaneDedxEqe[r][i][l] = (TH1D*)hfileEqe.Get(cht);
          fHtPlaneDedxEres[r][i][l] = (TH1D*)hfileEres.Get(cht);
          fHtPlaneDedxEdis[r][i][l] = (TH1D*)hfileEdis.Get(cht);
          fHtPlaneDedxEcoh[r][i][l] = (TH1D*)hfileEcoh.Get(cht);

          
          if(fHtPlaneDedxE[r][i][l]->GetEntries()>0&&fHtExpPlaneE[r][i]==0){            
            if(fHtPlaneDedxE[r][i][l]->GetBinContent(1)/(double)fHtPlaneDedxE[r][i][l]->GetEntries()>0.997){
                    fHtExpPlaneE[r][i]=l;
            }
          }

          if(fHtPlaneDedxG[r][i][l]->GetEntries()>0&&fHtExpPlaneG[r][i]==0){
            if(fHtPlaneDedxG[r][i][l]->GetBinContent(1)/(double)fHtPlaneDedxG[r][i][l]->GetEntries()>0.997){      
                    fHtExpPlaneG[r][i]=l;
            }
          }

          if(fHtPlaneDedxMu[r][i][l]->GetEntries()>0&&fHtExpPlaneMu[r][i]==0){
            if(fHtPlaneDedxMu[r][i][l]->GetBinContent(1)/(double)fHtPlaneDedxMu[r][i][l]->GetEntries()>0.997){      
                    fHtExpPlaneMu[r][i]=l;
            }
          }

          if(fHtPlaneDedxPi0[r][i][l]->GetEntries()>0&&fHtExpPlanePi0[r][i]==0){
            if(fHtPlaneDedxPi0[r][i][l]->GetBinContent(1)/(double)fHtPlaneDedxPi0[r][i][l]->GetEntries()>0.997){      
                    fHtExpPlanePi0[r][i]=l;
            }
          }

          if(fHtPlaneDedxP[r][i][l]->GetEntries()>0&&fHtExpPlaneP[r][i]==0){
            if(fHtPlaneDedxP[r][i][l]->GetBinContent(1)/(double)fHtPlaneDedxP[r][i][l]->GetEntries()>0.997){      
                    fHtExpPlaneP[r][i]=l;
            }
          }

          if(fHtPlaneDedxN[r][i][l]->GetEntries()>0&&fHtExpPlaneN[r][i]==0){
            if(fHtPlaneDedxN[r][i][l]->GetBinContent(1)/(double)fHtPlaneDedxN[r][i][l]->GetEntries()>0.997){      
                    fHtExpPlaneN[r][i]=l;
            }
          }

          if(fHtPlaneDedxPi[r][i][l]->GetEntries()>0&&fHtExpPlanePi[r][i]==0){
            if(fHtPlaneDedxPi[r][i][l]->GetBinContent(1)/(double)fHtPlaneDedxPi[r][i][l]->GetEntries()>0.997){ 
                    fHtExpPlanePi[r][i]=l;
            }
          }

          if(fHtPlaneDedxEqe[r][i][l]->GetEntries()>0&&fHtExpPlaneEqe[r][i]==0){
            if(fHtPlaneDedxEqe[r][i][l]->GetBinContent(1)/(double)fHtPlaneDedxEqe[r][i][l]->GetEntries()>0.997){  
                    fHtExpPlaneEqe[r][i]=l;
            }
          }

          if(fHtPlaneDedxEres[r][i][l]->GetEntries()>0&&fHtExpPlaneEres[r][i]==0){
            if(fHtPlaneDedxEres[r][i][l]->GetBinContent(1)/(double)fHtPlaneDedxEres[r][i][l]->GetEntries()>0.997){                          
              fHtExpPlaneEres[r][i]=l;
            }
          }

          if(fHtPlaneDedxEdis[r][i][l]->GetEntries()>0&&fHtExpPlaneEdis[r][i]==0){
            if(fHtPlaneDedxEdis[r][i][l]->GetBinContent(1)/(double)fHtPlaneDedxEdis[r][i][l]->GetEntries()>0.997){
                    fHtExpPlaneEdis[r][i]=l;
            }
          }

          if(fHtPlaneDedxEcoh[r][i][l]->GetEntries()>0&&fHtExpPlaneEcoh[r][i]==0){
            if(fHtPlaneDedxEcoh[r][i][l]->GetBinContent(1)/(double)fHtPlaneDedxEcoh[r][i][l]->GetEntries()>0.997){                    
              fHtExpPlaneEcoh[r][i]=l;
            }
          }
        }
/*
        if(r==0){
          std::cout<<"igev "<<i<<" "<<std::endl;
          std::cout<<"exp plane E    "<<fHtExpPlaneE[r][i]<<std::endl;
          std::cout<<"exp plane G    "<<fHtExpPlaneG[r][i]<<std::endl;
          std::cout<<"exp plane Mu   "<<fHtExpPlaneMu[r][i]<<std::endl;
          std::cout<<"exp plane Pi0  "<<fHtExpPlanePi0[r][i]<<std::endl;
          std::cout<<"exp plane Had  "<<fHtExpPlaneHad[r][i]<<std::endl;
          std::cout<<"exp plane P    "<<fHtExpPlaneP[r][i]<<std::endl;
          std::cout<<"exp plane N    "<<fHtExpPlaneN[r][i]<<std::endl;
          std::cout<<"exp plane Pi   "<<fHtExpPlanePi[r][i]<<std::endl;
          std::cout<<"exp plane Eqe  "<<fHtExpPlaneEqe[r][i]<<std::endl;
          std::cout<<"exp plane Eres "<<fHtExpPlaneEres[r][i]<<std::endl;
          std::cout<<"exp plane Edis "<<fHtExpPlaneEdis[r][i]<<std::endl;
          std::cout<<"exp plane Ecoh "<<fHtExpPlaneEcoh[r][i]<<std::endl;
          std::cout<<""<<std::endl;
        }
*/
      }
    }




/*
    for(int i=0; i<11; i++){
      for(int l=0; l<200; l++){
        char cht[200];
        sprintf(cht,"ht_%d_%d",i,l);
        fHtPlaneDedxE[i][l] = (TH1D*)hfileE.Get(cht);
        fHtPlaneDedxG[i][l] = (TH1D*)hfileG.Get(cht);
        fHtPlaneDedxMu[i][l] = (TH1D*)hfileMu.Get(cht);
        fHtPlaneDedxPi0[i][l] = (TH1D*)hfilePi0.Get(cht);
//        fHtPlaneDedxHad[i][l] = (TH1D*)hfileHad.Get(cht);
        fHtPlaneDedxP[i][l] = (TH1D*)hfileP.Get(cht);
        fHtPlaneDedxN[i][l] = (TH1D*)hfileN.Get(cht);
        fHtPlaneDedxPi[i][l] = (TH1D*)hfilePi.Get(cht);
        fHtPlaneDedxEqe[i][l] = (TH1D*)hfileEqe.Get(cht);
        fHtPlaneDedxEres[i][l] = (TH1D*)hfileEres.Get(cht);
        fHtPlaneDedxEdis[i][l] = (TH1D*)hfileEdis.Get(cht);
        fHtPlaneDedxEcoh[i][l] = (TH1D*)hfileEcoh.Get(cht);

        char chtmip[200];
        sprintf(chtmip,"htmip_%d_%d",i,l);
        fHtmipPlaneDedxE[i][l] = (TH1D*)hfileE.Get(chtmip);
        fHtmipPlaneDedxG[i][l] = (TH1D*)hfileG.Get(chtmip);
        fHtmipPlaneDedxMu[i][l] = (TH1D*)hfileMu.Get(chtmip);
        fHtmipPlaneDedxPi0[i][l] = (TH1D*)hfilePi0.Get(chtmip);
//        fHtmipPlaneDedxHad[i][l] = (TH1D*)hfileHad.Get(chtmip);
        fHtmipPlaneDedxP[i][l] = (TH1D*)hfileP.Get(chtmip);
        fHtmipPlaneDedxN[i][l] = (TH1D*)hfileN.Get(chtmip);
        fHtmipPlaneDedxPi[i][l] = (TH1D*)hfilePi.Get(chtmip);
        fHtmipPlaneDedxEqe[i][l] = (TH1D*)hfileEqe.Get(chtmip);
        fHtmipPlaneDedxEres[i][l] = (TH1D*)hfileEres.Get(chtmip);
        fHtmipPlaneDedxEdis[i][l] = (TH1D*)hfileEdis.Get(chtmip);
        fHtmipPlaneDedxEcoh[i][l] = (TH1D*)hfileEcoh.Get(chtmip);

        char chtshower[200];
        sprintf(chtshower,"htshower_%d_%d",i,l);
        fHtshowerPlaneDedxE[i][l] = (TH1D*)hfileE.Get(chtshower);
        fHtshowerPlaneDedxG[i][l] = (TH1D*)hfileG.Get(chtshower);
        fHtshowerPlaneDedxMu[i][l] = (TH1D*)hfileMu.Get(chtshower);
        fHtshowerPlaneDedxPi0[i][l] = (TH1D*)hfilePi0.Get(chtshower);
//        fHtshowerPlaneDedxHad[i][l] = (TH1D*)hfileHad.Get(chtshower);
        fHtshowerPlaneDedxP[i][l] = (TH1D*)hfileP.Get(chtshower);
        fHtshowerPlaneDedxN[i][l] = (TH1D*)hfileN.Get(chtshower);
        fHtshowerPlaneDedxPi[i][l] = (TH1D*)hfilePi.Get(chtshower);
        fHtshowerPlaneDedxEqe[i][l] = (TH1D*)hfileEqe.Get(chtshower);
        fHtshowerPlaneDedxEres[i][l] = (TH1D*)hfileEres.Get(chtshower);
        fHtshowerPlaneDedxEdis[i][l] = (TH1D*)hfileEdis.Get(chtshower);
        fHtshowerPlaneDedxEcoh[i][l] = (TH1D*)hfileEcoh.Get(chtshower);
      }
    }
*/

    //now the histrograms for transverse dedx are also being loaded.
    for(int r=0; r<4; r++){
      for(int i=0; i<11; i++){
        for(int l=0; l<20; l++){
          char cht[200];
          sprintf(cht,"httrans_%d_%d_%d",r,i,l);
          fHtCellTransDedxE[r][i][l] = (TH1D*)hfileE.Get(cht);
          fHtCellTransDedxG[r][i][l] = (TH1D*)hfileG.Get(cht);
          fHtCellTransDedxMu[r][i][l] = (TH1D*)hfileMu.Get(cht);
          fHtCellTransDedxPi0[r][i][l] = (TH1D*)hfilePi0.Get(cht);
//          fHtCellTransDedxHad[r][i][l] = (TH1D*)hfileHad.Get(cht);
          fHtCellTransDedxP[r][i][l] = (TH1D*)hfileP.Get(cht);
          fHtCellTransDedxN[r][i][l] = (TH1D*)hfileN.Get(cht);
          fHtCellTransDedxPi[r][i][l] = (TH1D*)hfilePi.Get(cht);
          fHtCellTransDedxEqe[r][i][l] = (TH1D*)hfileEqe.Get(cht);
          fHtCellTransDedxEres[r][i][l] = (TH1D*)hfileEres.Get(cht);
          fHtCellTransDedxEdis[r][i][l] = (TH1D*)hfileEdis.Get(cht);
          fHtCellTransDedxEcoh[r][i][l] = (TH1D*)hfileEcoh.Get(cht);
        }
      }
    }

    hfileE.Close();
    hfileG.Close();
    hfileMu.Close();
    hfilePi0.Close();
//    hfileHad.Close();
    hfileP.Close();
    hfileN.Close();
    hfilePi.Close();
    hfileEqe.Close();
    hfileEres.Close();
    hfileEdis.Close();
    hfileEcoh.Close();
    return true;
  }

  bool RecoJMShower::LoadTreeWeights(int detid)
  {
    std::string libpath = util::EnvExpansion(fLibPath);
    std::string fnameWeightsShape(libpath);
    std::string fnameWeightsShapeElec(libpath);
    std::string fnameWeightsShapeE(libpath);
    std::string fnameWeightsShapeNoCos(libpath);
    std::string fnameWeightsShapeENoCos(libpath);


//    std::string fnameWeightsShapeQE(libpath);
//    std::string fnameWeightsShapeRES(libpath);
//    std::string fnameWeightsShapeDIS(libpath);
    switch (detid) {
      case novadaq::cnv::kFARDET:
        fnameWeightsShape += "weights_shape_fd_elec.txt";
        fnameWeightsShapeElec += "weights_shower_fd_elec.txt";
        fnameWeightsShapeE += "weights_shape_fd_elec_E.txt";
        fnameWeightsShapeNoCos += "weights_shape_fd_elec_NoCos.txt";
        fnameWeightsShapeENoCos += "weights_shape_fd_elec_E_NoCos.txt";
//        fnameWeightsShapeQE += "weights_shape_fd_shower_QE.txt";
//        fnameWeightsShapeRES += "weights_shape_fd_shower_RES.txt";
//        fnameWeightsShapeDIS += "weights_shape_fd_shower_DIS.txt";
      break;
      case novadaq::cnv::kNDOS:    
        fnameWeightsShape += "weights_shape_fd_elec.txt";
        fnameWeightsShapeElec += "weights_shower_fd_elec.txt";
        fnameWeightsShapeE += "weights_shape_fd_elec_E.txt";
        fnameWeightsShapeNoCos += "weights_shape_fd_elec_NoCos.txt";
        fnameWeightsShapeENoCos += "weights_shape_fd_elec_E_NoCos.txt";
//        fnameWeightsShapeQE += "weights_shape_fd_shower_QE.txt";
//        fnameWeightsShapeRES += "weights_shape_fd_shower_RES.txt";
//        fnameWeightsShapeDIS += "weights_shape_fd_shower_DIS.txt";
      break;
      case novadaq::cnv::kNEARDET: 
        fnameWeightsShape += "weights_shape_fd_elec.txt";
        fnameWeightsShapeElec += "weights_shower_fd_elec.txt";
        fnameWeightsShapeE += "weights_shape_fd_elec_E.txt";
        fnameWeightsShapeNoCos += "weights_shape_fd_elec_NoCos.txt";
        fnameWeightsShapeENoCos += "weights_shape_fd_elec_E_NoCos.txt";
//        fnameWeightsShapeQE += "weights_shape_fd_shower_QE.txt";
//        fnameWeightsShapeRES += "weights_shape_fd_shower_RES.txt";
//        fnameWeightsShapeDIS += "weights_shape_fd_shower_DIS.txt";
      break;
      default:
        throw cet::exception("BAD DETID") << __FILE__ << ":" << __LINE__
                                        << " Bad detector id = " << detid;
    }
/*  Histograms stored in /nova/data/users/bianjm/share/development/RecoJMShower/histograms for this test versions

    cet::search_path sp("FW_SEARCH_PATH");
    sp.find_file(fnameWeightsShape, fFnameWeightsShape);
    struct stat sb;
    if (fFnameWeightsShape.empty() || stat(fFnameWeightsShape.c_str(), &sb)!=0){
      // Failed to resolve the file name
      mf::LogWarning("RecoJMShower") << "Training Weight -- " << fFnameWeightsShape << " not found.";
      throw cet::exception("FILE_NOT_FOUND") << "ANN Weight file not found.";
    }
*/
//    std::cout<<"Trained ANN "<<fnameWeightsShape.c_str()<<" is loaded."<<std::endl;

    double egLLL; 
    double egLLT;
    double emuLLL;
    double emuLLT;
    double epi0LLL;
    double epi0LLT;
    double epLLL;
    double epLLT;
    double enLLL;
    double enLLT;
    double epiLLL;
    double epiLLT;
    double vtxgev;
    double pi0mass;
    double shE;
    double nuE;
    double gap;
    double length;
    double costheta;
    int type;

    TTree *trainTree_shape = new TTree("trainTree_shape","m_trainTree_shape");
    trainTree_shape->Branch("egLLL", &egLLL, "egLLL/D");
    trainTree_shape->Branch("egLLT", &egLLT, "egLLT/D");
    trainTree_shape->Branch("emuLLL", &emuLLL, "emuLLL/D");
    trainTree_shape->Branch("emuLLT", &emuLLT, "emuLLT/D");
    trainTree_shape->Branch("epi0LLL", &epi0LLL, "epi0LLL/D");
    trainTree_shape->Branch("epi0LLT", &epi0LLT, "epi0LLT/D");
    trainTree_shape->Branch("epLLL", &epLLL, "epLLL/D");
    trainTree_shape->Branch("epLLT", &epLLT, "epLLT/D");
    trainTree_shape->Branch("enLLL", &enLLL, "enLLL/D");
    trainTree_shape->Branch("enLLT", &enLLT, "enLLT/D");
    trainTree_shape->Branch("epiLLL", &epiLLL, "epiLLL/D");
    trainTree_shape->Branch("epiLLT", &epiLLT, "epiLLT/D");
    trainTree_shape->Branch("gap",    &gap, "gap/D");
    trainTree_shape->Branch("pi0mass", &pi0mass, "pi0mass/D");
    trainTree_shape->Branch("vtxgev", &vtxgev, "vtxgev/D");
    trainTree_shape->Branch("shE", &shE, "shE/D");
    trainTree_shape->Branch("nuE", &nuE, "nuE/D");
    trainTree_shape->Branch("costheta", &costheta, "costheta/D");
    trainTree_shape->Branch("length", &length, "length/D");
    trainTree_shape->Branch("type",   &type,   "type/I");



    mlp_shape = new TMultiLayerPerceptron("@egLLL,@egLLT,@emuLLL,@emuLLT,@epi0LLL,@epi0LLT,@epLLL,@epLLT,@enLLL,@enLLT,@epiLLL,@epiLLT,@pi0mass,@shE,@vtxgev,@gap,@costheta:22:12:6:type", trainTree_shape);
    mlp_shape->LoadWeights(fnameWeightsShape.c_str());

    mlp_shape_NoCos = new TMultiLayerPerceptron("@egLLL,@egLLT,@emuLLL,@emuLLT,@epi0LLL,@epi0LLT,@epLLL,@epLLT,@enLLL,@enLLT,@epiLLL,@epiLLT,@pi0mass,@shE,@vtxgev,@gap:22:12:6:type", trainTree_shape);
    mlp_shape_NoCos->LoadWeights(fnameWeightsShapeNoCos.c_str());


    mlp_shape_elec = new TMultiLayerPerceptron("@egLLL,@egLLT,@epi0LLL,@epi0LLT,@epLLL,@epLLT,@enLLL,@enLLT,@epiLLL,@epiLLT,@shE,@gap:18:12:6:type", trainTree_shape);

    mlp_shape_elec->LoadWeights(fnameWeightsShapeElec.c_str());

    mlp_shape_E = new TMultiLayerPerceptron("@egLLL,@egLLT,@emuLLL,@emuLLT,@epi0LLL,@epi0LLT,@epLLL,@epLLT,@enLLL,@enLLT,@epiLLL,@epiLLT,@pi0mass,@shE,@vtxgev,@gap,@costheta,@nuE:22:12:6:type", trainTree_shape);
    mlp_shape_E->LoadWeights(fnameWeightsShapeE.c_str());


    mlp_shape_E_NoCos = new TMultiLayerPerceptron("@egLLL,@egLLT,@emuLLL,@emuLLT,@epi0LLL,@epi0LLT,@epLLL,@epLLT,@enLLL,@enLLT,@epiLLL,@epiLLT,@pi0mass,@shE,@vtxgev,@gap,@nuE:22:12:6:type", trainTree_shape);
    mlp_shape_E_NoCos->LoadWeights(fnameWeightsShapeENoCos.c_str());


/*
    mlp_shape = new TMultiLayerPerceptron("@egLLL,@egLLT,@emuLLL,@emuLLT,@epi0LLL,@epi0LLT,@epLLL,@epLLT,@enLLL,@enLLT,@epiLLL,@epiLLT,@pi0mass,@shE,@gap,@costheta,@vtxgev,@length:25:16:9:type", trainTree_shape); 
    mlp_shape->LoadWeights(fnameWeightsShape.c_str());


    mlp_shape = new TMultiLayerPerceptron("egLLL,egLLT,epi0LLL,epi0LLT,epLLL,epLLT,enLLL,enLLT,epiLLL,epiLLT,pi0mass,vtxgev:18:12:6:type",trainTree_shape);
    mlp_shape->LoadWeights(fnameWeightsShape.c_str());
*/

/*
    mlp_shape_QE = new TMultiLayerPerceptron("@egLLL,@egLLT,@epi0LLL,@epi0LLT,@epLLL,@epLLT,@enLLL,@enLLT,@epiLLL,@epiLLT,@gap:18:12:6:type",trainTree_shape); 
    mlp_shape_RES = new TMultiLayerPerceptron("@egLLL,@egLLT,@epi0LLL,@epi0LLT,@epLLL,@epLLT,@enLLL,@enLLT,@epiLLL,@epiLLT,@gap:18:12:6:type",trainTree_shape); 
    mlp_shape_DIS = new TMultiLayerPerceptron("@egLLL,@egLLT,@epi0LLL,@epi0LLT,@epLLL,@epLLT,@enLLL,@enLLT,@epiLLL,@epiLLT,@gap:18:12:6:type",trainTree_shape); 
    mlp_shape_QE->LoadWeights(fnameWeightsShapeQE.c_str());
    mlp_shape_RES->LoadWeights(fnameWeightsShapeRES.c_str());
    mlp_shape_DIS->LoadWeights(fnameWeightsShapeDIS.c_str());
*/
    return true;
  }

  void RecoJMShower::produce(art::Event& evt)
  {
    isrealdata = evt.isRealData();
    fprongidx.clear();

    //containers to store shower objects in event and the associations to slices
    std::unique_ptr< std::vector<jmshower::JMShower> > recoshowercol(new std::vector<jmshower::JMShower> );
    std::unique_ptr< std::vector<jmshower::JMShower> > evtrecoshowercol(new std::vector<jmshower::JMShower> );
    std::unique_ptr< art::Assns<jmshower::JMShower, rb::Cluster> > assns(new art::Assns<jmshower::JMShower, rb::Cluster>);
    std::unique_ptr< art::Assns<jmshower::JMShower, rb::Prong> > prongassns(new art::Assns<jmshower::JMShower, rb::Prong>);
    // TODO: Prong-jmshower association is temporary. It should be deleted once ShowerLID validation is done.

    //pass the event to the function that does the work, creates a vector of shower objects
    //pass the event to the function that does the work, creates a vector of shower objects
//    std::vector<jmshower::JMShower> jmshowercol = RecoJMShower::RecoShowers(evt);
    std::vector<jmshower::JMShower> jmshowercol = RecoJMShower::RecoShowers(*this, evt, recoshowercol, prongassns);

    //get the slices contained in each event
    art::Handle<std::vector<rb::Cluster> > slicecol;
    evt.getByLabel(pSliceDir, slicecol);
    //loop over the event showers and put them in the event
    for(unsigned int i=0; i<jmshowercol.size();i++){
      evtrecoshowercol->push_back(jmshowercol[i]);
      //each shower has stored the slice it comes from, pull this out now to make proper association
      unsigned int iSlice = jmshowercol[i].ISlice();
      util::CreateAssn(*this,evt,*evtrecoshowercol,art::Ptr<rb::Cluster>(slicecol, iSlice),*assns);
    }

    evt.put(std::move(evtrecoshowercol));
    evt.put(std::move(assns));
    evt.put(std::move(prongassns));
    return;
  }


//  std::vector<jmshower::JMShower> RecoJMShower::RecoShowers(const art::Event& evt){
  std::vector<jmshower::JMShower> RecoJMShower::RecoShowers(art::EDProducer const&    prod,
                                                            art::Event& evt,
                                                            std::unique_ptr< std::vector<jmshower::JMShower> >&recoshowercol,
                                                            std::unique_ptr< art::Assns<jmshower::JMShower, rb::Prong> >& prongassns){

//    int run   = evt.run();
//    int srun  = evt.subRun();
//    int event = evt.id().event();
    //load services
    art::ServiceHandle<geo::Geometry> geom;
    art::ServiceHandle<calib::Calibrator> cal;

//Event level
    art::Handle<std::vector<rb::Cluster> > slicecol;
    evt.getByLabel(pSliceDir, slicecol);

    std::vector<jmshower::JMShower> evtshowercol;
    evtshowercol.clear();

     //get verticies associated with each slice
    art::FindManyP<rb::Vertex> fmv(slicecol, evt, pVertexLabel); // get vertices 
    art::FindManyP<rb::Prong> fmprong(slicecol, evt, art::InputTag( pTrackDir, pInstLabel)); // get prong
    if(pRecoVtx==true)if (!(fmv.isValid())) return evtshowercol;
    if (!(fmprong.isValid())) return evtshowercol;
    for(unsigned int ic = 0; ic < slicecol->size(); ++ic){
      if(rb::IsFiltered(evt,slicecol,ic)) continue;
      art::Ptr<rb::Cluster> clust(slicecol,ic);
      if(clust->IsNoise()) continue;
      std::vector<jmshower::JMShower> showercol;
      showercol.clear();

      std::vector<art::Ptr<rb::Prong> > prong2d3d  = fmprong.at(ic);
      std::vector<art::Ptr<rb::Prong> > track;
      track.clear();
      for(unsigned int iPro = 0; iPro < prong2d3d.size(); ++iPro){
         if( prong2d3d[iPro]->Is3D())track.push_back(prong2d3d[iPro]);
         fprongidx.push_back(iPro);
      }
      //if there are no 3D prongs or crazy # of prongs  skip this slice.
      //NOTE: This is the only form of Preselection in RecoJMShower, it must
      //have a 3D track to be able to proceed.
      if (track.size()==0||track.size()>9999){
        continue;
      }
  
      //map of cell hits to the slice they came from
      //map of cell hits to track they came from
      std::map<geo::OfflineChan, std::vector<int>> hittrkmap;
      std::map<geo::OfflineChan, std::vector<int>> hitshmap;
  
      //containers to slice and track info
      art::PtrVector< rb::CellHit > cellhitlist;
      std::vector<const rb::Cluster* >   cluster;
      std::vector<rb::Prong> vCluster;
      std::vector<rb::Prong> vXCluster;
      std::vector<rb::Prong> vYCluster;
      std::vector<bool> vNantag;// A tage to deal with crazy prongs
      std::vector<double> vtxGeV;
      std::vector<double> vtxCenterGeV;
      std::vector<double> trkGeV;
  
      double sliceGeV = 0;
      double sliceEtot = 0;
      double sliceEdge10GeV = 0;//slice energy close to detector edge - 10 cell/plane
      double sliceEdge5GeV = 0;//slice energy close to detector edge - 5 cell/plane
      double sliceEdge2GeV = 0;//slice energy close to detector edge - 2 cell/plane

      std::vector<int> shNCellToEdge;//Mininum distance from shower cells to detector edges
      std::vector<int> shClosetEdge;//Closet edge for a shower
      std::vector<double> shEdge20GeV;//shower energy close to detector edge - 20 cell/plane 
      std::vector<double> shEdge15GeV;//shower energy close to detector edge - 15 cell/plane 
      std::vector<double> shEdge10GeV;//shower energy close to detector edge - 10 cell/plane 
      std::vector<double> shEdge5GeV;//shower energy close to detector edge - 5 cell/plane
      std::vector<double> shEdge2GeV;//shower energy close to detector edge - 2 cell/plane

      double sliceExclGeV = 0;//slice energy close to detector edge - 2 cell/plane

      rb::Cluster excludeCluster;  //cluster of cells not in any tracks in slice
   
      //clear containers
      hittrkmap.clear();
      hitshmap.clear();
      fHitDistMap.clear();
      fHitEWeightMap.clear();
      fHitEMap.clear();
  
      fShowerPClusterCol.clear();  //longitudinal shower cluster
      fTrackPClusterCol.clear();  //longitudinal track cluster
      fShowerTCClusterCol.clear(); //transverse shower cluster
      fExcludeShowerCol.clear();  //shower objects for hits in slice not put into a track
  
      //clear containers
      cellhitlist.clear();
      cluster.clear();
      vCluster.clear();
      vXCluster.clear();
      vYCluster.clear();
      vNantag.clear();
      vtxGeV.clear();
      trkGeV.clear();
      shNCellToEdge.clear();
      shClosetEdge.clear();
      shEdge20GeV.clear();
      shEdge15GeV.clear();
      shEdge10GeV.clear();
      shEdge5GeV.clear();
      shEdge2GeV.clear();

      excludeCluster.Clear();
  
      //get the slices
  
      std::vector<TVector3> evtvtxcol; //hold vevent vertex
      std::vector<unsigned int> evtvtxplanecol; //plane associated with vertex
  
      //clear containers
      evtvtxcol.clear();
      evtvtxplanecol.clear();
  
      //initialize vectors to hold vertex info for each view
      Double_t cellXYZ1[3]={0,0,0};
      Double_t cellXYZ2[3]={0,0,0};
      geom->Plane(0)->Cell(0)->GetCenter(cellXYZ1);
      geom->Plane(1)->Cell(0)->GetCenter(cellXYZ2);
  
      //if this option is set, find the elastic arms vertex for each slice
      if(pRecoVtx==true){
        //NOTE: if there is no vertex, then there are no tracks, cannot do EID, so return
        //This is a form of "preselection"
          //if there was nue preselection, skip slice if it failed
          std::vector<art::Ptr<rb::Vertex> >  v = fmv.at(ic);
          //loop over verticies associated with slice
          for (unsigned int j=0; j<v.size(); ++j) {
            TVector3 evtVtx(0,0,0);
            unsigned int vtxPlane = 0;
            //store vertex coordinates
            evtVtx.SetX(v[j]->GetX());
            evtVtx.SetY(v[j]->GetY());
            evtVtx.SetZ(v[j]->GetZ());

            //find plane number closest to vertex
	    geo::CellUniqueId cid;

	    for(int trial = 0; trial < 4; ++trial){
	      // Figure out what cell the vertex was in. Step 3cm in z in case of
	      // failure. If that didn't work, it must be in plastic between cells,
	      // move laterally in x or y.

	      double offsetX = 0, offsetY = 0;
	      if(trial == 1) offsetX = -2;
	      if(trial == 2) offsetY = -2;
	      if(trial == 3) offsetX = offsetY = -2;

	      cid = geom->CellId(v[j]->GetX()+offsetX, v[j]->GetY()+offsetY, v[j]->GetZ(),
				 0, 0, 1, 3);
	      if(cid)
		break;
	    }

	    if( cid )
	      vtxPlane = geom->getPlaneID( cid );
	    else 
	      vtxPlane = clust->MinPlane();

            //store vertex, plane of vertex, and slice corresponding to vertex
            evtvtxcol.push_back(evtVtx);
            evtvtxplanecol.push_back(vtxPlane);
          }
        }
  
      //Clear hittrkmap  
        for(unsigned int i=0;i<clust->NXCell();i++){
          geo::OfflineChan key(clust->XCell(i)->Plane(),clust->XCell(i)->Cell());
          hittrkmap[key].clear(); //set match between hit and track to default
          hitshmap[key].clear(); //set match between hit and shower to default
        }
  
        //NOTE: When using FuzzyKVertex tracks cell hits can belong to more then one track
        //as this is written now the hit will be associated with the last track - has been solved by using a map 
        for(unsigned int i=0;i<track.size();i++){ //loop over tracks
          for(unsigned int ii=0;ii<track[i]->NCell();ii++){ //loop over cells in tracks
              geo::OfflineChan key(track[i]->Cell(ii)->Plane(),track[i]->Cell(ii)->Cell()); //make key value
              hittrkmap[key].push_back((int)i); //store track associated with cell
              hitshmap[key].push_back((int)i); //store shower associated with cell
          }
        }
  
        //TODO: This is setting the transverse coordinate of a view to the coordinate of the
        //first cell in that view.  This would be better if the average transverse coordinate was used.
        //This needs to be corrected but would also require retraining, needs to be looked at.
        const double wx = (clust->NXCell() > 0) ? clust->W(clust->XCell(0).get()) : 0;
        const double wy = (clust->NYCell() > 0) ? clust->W(clust->YCell(0).get()) : 0;
  
        //loop over x cells
        for(unsigned int i=0;i<clust->NXCell();i++){
           geo::OfflineChan key(clust->XCell(i)->Plane(),clust->XCell(i)->Cell());
          //only use cell hits on a track or cell hits with pe above threshold level set if fcl file
          if((clust->XCell(i)->PE())>pCellHitPHThresh||hittrkmap[key].size()>0){
            //make reco hits out of cells to sum energy
            const art::Ptr<rb::CellHit>& chit = clust->XCell(i);
            const rb::RecoHit rhit(cal->MakeRecoHit(*chit, wx ));

            double gev=0;
            geo::View_t view = geom->Plane(chit->Plane())->View();

            double pigtail = geom->GetPigtail(geom->Plane(chit->Plane())->Cell(chit->Cell())->Id());
            //double wxtoelec = wx - (pigtail - 92.6624);//W - (pigtail - avg. pigtail)
            //double wytoelec = wy - (pigtail - 92.6624);//W - (pigtail - avg. pigtail)

            double w1 = wx;
            if (view == geo::kY) w1 = wy;
            double xyzd1[4];
            calib::GetXYZD(chit->OfflineChan(), w1, xyzd1);
            double readDist = xyzd1[3];

            if(rhit.IsCalibrated()){
              gev = rhit.GeV();
            }else if(pUseUncalib == true){// use MC calibration if a cell is not calibrated
//              std::cout<<"slice, plane, cell, wx, gev "<<ic<<" "<<chit->Plane()<<" "<<chit->Cell()<<" "<<wx<<" Not calibrated?!" << std::endl;
              double adc = double(chit->ADC());
              double Att = 1;
              if(view == geo::kX){
                      Att = RecoJMShower::CellADCToGeV(readDist, view, fDetid); 
              }else if(view == geo::kY){
                      Att = RecoJMShower::CellADCToGeV(readDist, view, fDetid); 
              }
              gev = adc/Att;
            }
  
            double ecor = 1.;
            double disttoelec = geom->Plane(chit->Plane())->Cell(chit->Cell())->HalfL()+pigtail-wx; 
            if(evt.isRealData()==true&&pCorDataCell)ecor = RecoJMShower::CellEDataMC(gev, disttoelec,  view, fDetid); // correct Data/MC difference for cell energy
            gev = gev/ecor;
  
            sliceGeV += gev;
            if(abs((int)chit->Cell()-(int)pXEdgeId1)<10 ||
              abs((int)chit->Cell()-(int)pXEdgeId2)<10 ||
              abs((int)chit->Cell()-(int)pXEdgeId3)<10 ||
              abs((int)chit->Cell()-(int)pXEdgeId4)<10 ||
              abs((int)chit->Plane()-(int)pZEdgeId1)<10 ||  
              abs((int)chit->Plane()-(int)pZEdgeId2)<10 ||  
              abs((int)chit->Plane()-(int)pZEdgeId3)<10 ||  
              abs((int)chit->Plane()-(int)pZEdgeId4)<10){
              sliceEdge10GeV += gev; 
            }
  
            if(abs((int)chit->Cell()-(int)pXEdgeId1)<5 ||
              abs((int)chit->Cell()-(int)pXEdgeId2)<5 ||
              abs((int)chit->Cell()-(int)pXEdgeId3)<5 ||
              abs((int)chit->Cell()-(int)pXEdgeId4)<5 ||
              abs((int)chit->Plane()-(int)pZEdgeId1)<5 ||  
              abs((int)chit->Plane()-(int)pZEdgeId2)<5 ||  
              abs((int)chit->Plane()-(int)pZEdgeId3)<5 ||  
              abs((int)chit->Plane()-(int)pZEdgeId4)<5){
              sliceEdge5GeV += gev; 
            }
  
            if(fabs((int)chit->Cell()-(int)pXEdgeId1)<2 ||
              abs((int)chit->Cell()-(int)pXEdgeId2)<2 ||
              abs((int)chit->Cell()-(int)pXEdgeId3)<2 ||
              abs((int)chit->Cell()-(int)pXEdgeId4)<2 ||
              abs((int)chit->Plane()-(int)pZEdgeId1)<2 || 
              abs((int)chit->Plane()-(int)pZEdgeId2)<2 || 
              abs((int)chit->Plane()-(int)pZEdgeId3)<2 || 
              abs((int)chit->Plane()-(int)pZEdgeId4)<2){
              sliceEdge2GeV += gev;  
            }
            cellhitlist.push_back(clust->XCell(i));
          }
        }
  
        //loop over y cells
        for(unsigned int i=0;i<clust->NYCell();i++){
          geo::OfflineChan key(clust->YCell(i)->Plane(),clust->YCell(i)->Cell());
          //use hits above pe threshold or on a track
          if((clust->YCell(i)->PE())>pCellHitPHThresh||hittrkmap[key].size()>0){
            //make reco hits out of cells
            const art::Ptr<rb::CellHit>& chit = clust->YCell(i);
            const rb::RecoHit rhit(cal->MakeRecoHit(*chit, wy ));
            double gev=0;
            geo::View_t view = geom->Plane(chit->Plane())->View();
            double pigtail = geom->GetPigtail(geom->Plane(chit->Plane())->Cell(chit->Cell())->Id());
            //double wxtoelec = wx - (pigtail - 92.6624);//W - (pigtail - avg. pigtail)
            //double wytoelec = wy - (pigtail - 92.6624);//W - (pigtail - avg. pigtail)

            double w1 = wx;
            if (view == geo::kY) w1 = wy;
            double xyzd1[4];
            calib::GetXYZD(chit->OfflineChan(), w1, xyzd1);
            double readDist = xyzd1[3];

            if(rhit.IsCalibrated()){
              gev = rhit.GeV();
            }else if(pUseUncalib == true){// use MC calibration if a cell is not calibrated
              double adc = double(chit->ADC());
              double Att = 1;
              if(view == geo::kX){
                      Att = RecoJMShower::CellADCToGeV(readDist, view, novadaq::cnv::kFARDET); 
              }else if(view == geo::kY){
                      Att = RecoJMShower::CellADCToGeV(readDist, view, novadaq::cnv::kFARDET); 
              }
              gev = adc/Att;
            }
  
            double ecor = 1.;
            double disttoelec = geom->Plane(chit->Plane())->Cell(chit->Cell())->HalfL()+pigtail-wy;           
            if(evt.isRealData()==true&&pCorDataCell)ecor = RecoJMShower::CellEDataMC(gev, disttoelec,  view, fDetid); // correct Data/MC difference for cell energy
            gev = gev/ecor;
  
            sliceGeV += gev;
  
            if(abs((int)chit->Cell()-(int)pYEdgeId1)<10 || 
              abs((int)chit->Cell()-(int)pYEdgeId2)<10 ||
              abs((int)chit->Cell()-(int)pYEdgeId3)<10 ||
              abs((int)chit->Cell()-(int)pYEdgeId4)<10 ||
              abs((int)chit->Plane()-(int)pZEdgeId1)<10 || 
              abs((int)chit->Plane()-(int)pZEdgeId2)<10 || 
              abs((int)chit->Plane()-(int)pZEdgeId3)<10 || 
              abs((int)chit->Plane()-(int)pZEdgeId4)<10){
              sliceEdge10GeV += gev; 
            }
            if(abs((int)chit->Cell()-(int)pYEdgeId1)<5 ||
              abs((int)chit->Cell()-(int)pYEdgeId2)<5 ||
              abs((int)chit->Cell()-(int)pYEdgeId3)<5 ||
              abs((int)chit->Cell()-(int)pYEdgeId4)<5 ||
              abs((int)chit->Plane()-(int)pZEdgeId1)<5 ||
              abs((int)chit->Plane()-(int)pZEdgeId2)<5 ||
              abs((int)chit->Plane()-(int)pZEdgeId3)<5 ||
              abs((int)chit->Plane()-(int)pZEdgeId4)<5){
              sliceEdge5GeV += gev;
            }
            if(abs((int)chit->Cell()-(int)pYEdgeId1)<2 ||
              abs((int)chit->Cell()-(int)pYEdgeId2)<2 ||
              abs((int)chit->Cell()-(int)pYEdgeId3)<2 ||
              abs((int)chit->Cell()-(int)pYEdgeId4)<2 ||
              abs((int)chit->Plane()-(int)pZEdgeId1)<2 ||
              abs((int)chit->Plane()-(int)pZEdgeId2)<2 ||
              abs((int)chit->Plane()-(int)pZEdgeId3)<2 ||
              abs((int)chit->Plane()-(int)pZEdgeId4)<2){
              sliceEdge2GeV += gev;
            }
  
            cellhitlist.push_back(clust->YCell(i));
          }
        }
  
  
      //declare and clear some containers for shower variables
      std::vector<TVector3> startcol,endcol,dircol;
      std::vector<TVector3> shstartcol,shendcol;
      std::vector<unsigned int> startplanecol,stopplanecol;
      std::vector<unsigned int> shstartplanecol,shstopplanecol;
      std::vector<bool> backtagcol;
      std::vector<double> gapcol;
      std::vector<double> vtxdocacol;
      startcol.clear();
      endcol.clear();
      dircol.clear();
      startplanecol.clear();
      stopplanecol.clear();
      backtagcol.clear();
      gapcol.clear();
      vtxdocacol.clear();
      shstartcol.clear();
      shendcol.clear();
      shstartplanecol.clear();
      shstopplanecol.clear();
  
  
      //loop over tracks
      //this loop stores infomation about the start and stop of each track
      //and the distance to the vertex
      for(unsigned int i=0;i<track.size();i++){
        //initialize cell container
        art::PtrVector<rb::CellHit> trackcells;
        trackcells.clear();
        //copy the track into a new track object
        rb::Prong Clust(trackcells,track[i]->Start(),track[i]->Dir());
  //      Clust.AppendTrajectoryPoint(GetShwStop(track[i]));
  
        //push this copy into vector that can be manipulated
  
        TVector3 tStart = track[i]->Start();
        bool nantag = false;// reset track start/stop if nan vale is found
        if(std::isnan(tStart.x())||fabs(tStart.x())>10000){tStart.SetX(track[i]->MinX());nantag = true;}
        if(std::isnan(tStart.y())||fabs(tStart.y())>10000){tStart.SetY(track[i]->MinY());nantag = true;}
        if(std::isnan(tStart.z())||fabs(tStart.z())>10000){tStart.SetZ(track[i]->MinZ());nantag = true;}
        TVector3 tStop = RecoJMShower::GetShwStop(track[i]);
        if(std::isnan(tStop.x()||fabs(tStop.x())>10000)){tStop.SetX(track[i]->MaxX());nantag = true;}
        if(std::isnan(tStop.y()||fabs(tStop.y())>10000)){tStop.SetY(track[i]->MaxY());nantag = true;}
        if(std::isnan(tStop.z()||fabs(tStop.z())>10000)){tStop.SetZ(track[i]->MaxZ());nantag = true;}
        TVector3 tVect = (tStop-tStart).Unit();
        if(nantag == true){
  //        Clust.ClearTrajectoryPoints();
          Clust.SetStart(tStart);
  //        Clust.AppendTrajectoryPoint(tStop);
          Clust.SetDir(tVect);
        }
  
        vCluster.push_back(Clust);
        vXCluster.push_back(Clust);
        vYCluster.push_back(Clust);
        vNantag.push_back(nantag);
  
        vtxGeV.push_back(0); //initialize vertex energy
        vtxCenterGeV.push_back(0); //initialize vertex energy
        trkGeV.push_back(0);  //initialize track energy
        shNCellToEdge.push_back(9999);//initialize mininum distance from shower cells to detector edges
        shClosetEdge.push_back(-1);//initialize Closet edge for a shower 
        shEdge20GeV.push_back(0);//initialize shower energy close to detector edge - 20 cell/plane 
        shEdge15GeV.push_back(0);//initialize shower energy close to detector edge - 15 cell/plane 
        shEdge10GeV.push_back(0);//initialize shower energy close to detector edge - 10 cell/plane 
        shEdge5GeV.push_back(0);//initialize shower energy close to detector edge - 5 cell/plane
        shEdge2GeV.push_back(0);//initialize shower energy close to detector edge - 2 cell/plane
  

      //initialize event vertex to start of track and first plane in track
        TVector3 evtVtx(track[i]->Start()[0],track[i]->Start()[1],track[i]->Start()[2]);
        geo::OfflineChan key(track[i]->Cell(0)->Plane(),track[i]->Cell(0)->Cell());
        //if the vertex is being used by the EID (Which is the default option and what the training is based on)
        //then find the vertex in the slice the track comes from
        if(pRecoVtx==true&&pVtxFit==true){
          for(unsigned iv=0;iv<evtvtxcol.size();iv++){
  //        double doca = RecoJMShower::GetPointDoca(evtvtxcol[iv], track[i]->Start(), RecoJMShower::GetShwStop(track[i]));
  //        std::cout<<"dist,doca "<<(evtVtx-evtvtxcol[iv]).Mag()<<" "<<doca<<" "<<std::endl;
  //        if((evtVtx-evtvtxcol[iv]).Mag()<60.&&doca<20.){
            evtVtx = evtvtxcol[iv]; //get the vertex for that slice
            break;
  //        }
          }
        }
  
        //get the distance of closest approach to vertex
        double vtxDoca = RecoJMShower::GetPointDoca(evtVtx, track[i]->Start(), RecoJMShower::GetShwStop(track[i]));
        double vtxToStart = (evtVtx-track[i]->Start()).Mag();
        double vtxToStop = (evtVtx-RecoJMShower::GetShwStop(track[i])).Mag();
        double vtxIntToStart = sqrt(vtxToStart*vtxToStart-vtxDoca*vtxDoca);
        double vtxIntToStop = sqrt(vtxToStop*vtxToStop-vtxDoca*vtxDoca);
        TVector3 vtxStart = track[i]->Start();
        TVector3 vtxStop = RecoJMShower::GetShwStop(track[i]);
  //      unsigned int vtxStopPlane = track[i]->MaxPlane(); 
  
        //record track start and end planes
        TVector3 shStart = track[i]->Start();
        TVector3 shStop = RecoJMShower::GetShwStop(track[i]);
  
        unsigned int shStartPlane = track[i]->MinPlane();
        unsigned int shStopPlane = track[i]->MaxPlane();
        double vtxTrkLength = (vtxStop-vtxStart).Mag();
        bool showerVtxBackTag = false;
        //flip start and stop for backward tracks if needed
        if(vtxToStart>vtxToStop) {
          showerVtxBackTag = true;
          vtxStart = RecoJMShower::GetShwStop(track[i]);
          vtxStop = track[i]->Start();
          double tmp = vtxIntToStart;
          vtxIntToStart = vtxIntToStop;
          vtxIntToStop = tmp;
          shStart = RecoJMShower::GetShwStop(track[i]);
          shStop = track[i]->Start();
        }
        //if(abs((int)track[i]->MinPlane()-(int)vtxPlane)>abs((int)track[i]->MaxPlane()-(int)vtxPlane)){
        if(track[i]->Dir()[2] < 0.0){
  //        vtxStopPlane = track[i]->MinPlane();
          shStartPlane = track[i]->MaxPlane();
          shStopPlane = track[i]->MinPlane();
        }
        if(vtxIntToStop<vtxTrkLength){
          vtxIntToStart = -vtxIntToStart;
        }
  
        TVector3 trkVtxVect = (vtxStop-evtVtx).Unit();
        TVector3 shVect = (shStop-shStart).Unit();
  
        //Set start and direction 
        if(pVtxFit==true){
          bool vtxnantag = false;// reset track start/stop if nan vale is found
          if(std::isnan(tStart.x())||fabs(tStart.x())>10000){tStart.SetX(track[i]->MinX());vtxnantag = true;}
          if(std::isnan(tStart.y())||fabs(tStart.y())>10000){tStart.SetY(track[i]->MinY());vtxnantag = true;}
          if(std::isnan(tStart.z())||fabs(tStart.z())>10000){tStart.SetZ(track[i]->MinZ());vtxnantag = true;}
          if(std::isnan(tStop.x()||fabs(tStop.x())>10000)){tStop.SetX(track[i]->MaxX());vtxnantag = true;}
          if(std::isnan(tStop.y()||fabs(tStop.y())>10000)){tStop.SetY(track[i]->MaxY());vtxnantag = true;}
          if(std::isnan(tStop.z()||fabs(tStop.z())>10000)){tStop.SetZ(track[i]->MaxZ());vtxnantag = true;}
          if(vtxnantag==true) shVect = (shStop-shStart).Unit();
          vCluster[i].SetStart(shStart);
          vCluster[i].SetDir(shVect);
          vXCluster[i].SetStart(shStart);
          vXCluster[i].SetDir(shVect);
          vYCluster[i].SetStart(shStart);
          vYCluster[i].SetDir(shVect);

        }
  //  reverse shower using vertex
        //put track start and end points into vectors
        startcol.push_back(shStart);
        endcol.push_back(shStop);
        dircol.push_back(shVect);
        startplanecol.push_back(shStartPlane);
        stopplanecol.push_back(shStopPlane);
  
        backtagcol.push_back(showerVtxBackTag);
        //record the gap from vertex to the start of the track
        //if the vertex is inside a track record distance of closest approach
        //TODO: For fuzzyK there can be an internal gap, where the start of the track is misleading
        //Ruth has looked into this, will try to incorporate internal gaps into this algorithm
        double shGap = vtxIntToStart;
        if(vtxIntToStart>=0&&vtxIntToStart<vtxDoca)shGap = vtxDoca;
        if(vtxIntToStart<0)shGap = vtxDoca;
   
        //push gap and track start/stop variables into vectors
        gapcol.push_back(shGap);
//        vtxdocacol.push_back(vtxDoca);
        shstartcol.push_back(shStart);
        shendcol.push_back(shStop);
        shstartplanecol.push_back(shStartPlane);
        shstopplanecol.push_back(shStopPlane);
      }
        
  //*********************************************
  //  Cell energy deconvolution and calibration
  //*********************************************
  
      //now looping over all the cell hits in slice to do energy calibration 
      for (unsigned int i=0;i<cellhitlist.size();i++) {
        //get iterator for ith cell
        std::vector<art::Ptr<rb::CellHit> >::iterator itr = cellhitlist.begin() + i;
        //turn cell into an offlinechan so it can be found in the maps
        geo::OfflineChan key((*itr)->Plane(),(*itr)->Cell());
  //      double cellL = 2*geom->Plane((*itr)->Plane())->Cell((*itr)->Cell())->HalfL();// longest
  //      double cellD = 2*geom->Plane((*itr)->Plane())->Cell((*itr)->Cell())->HalfD();// longitudinal
        //store the cell width
        double cellW = 2*geom->Plane((*itr)->Plane())->Cell((*itr)->Cell())->HalfW();
  //      std::cout<<"cell L,D,W "<<cellL<<" "<<cellD<<" "<<cellW<<std::endl;
        //initialize weights
        double thiteweighttotal = 0;
        double thiteweighttotalatt = 0;
        double thiteweight[100];
  //      double trkattcal[20]; 
  
        //now loop over tracks to calculate vertex energy
        for(unsigned int itrk=0;itrk<track.size();itrk++){
          if(vNantag[itrk]==true)continue;
          geo::View_t view = geom->Plane((*itr)->Plane())->View();
          bool trktag=false;// on orignal tracks ?
          for(unsigned int imap=0;imap<hittrkmap[key].size();imap++){
            if(hittrkmap[key][imap]==(int)itrk)trktag=true;
          }
          if(trktag==true){//store x, y view clusters for track quality check.
            if(view == geo::kX)vXCluster[itrk].Add((*itr));
            if(view == geo::kY)vYCluster[itrk].Add((*itr));
          }


          //TVector3 trkdir=track[itrk]->Dir();
          //TVector3 trkstart=track[itrk]->Start();
          //TVector3 trkstop=RecoJMShower::GetShwStop(track[itrk]);
  
          //get the track start, stop and direction
          TVector3 trkdir=dircol[itrk];
          TVector3 trkstart=startcol[itrk];
          TVector3 trkstop=endcol[itrk];
  
          //clear that tracks hit weight
          thiteweight[itrk]=0;
          //trkattcal[itrk] = 0;
  
          //get the difference between the plane of the cell hit and the first plane of the track
          int plane = (int)(*itr)->Plane()-(int)startplanecol[itrk];
          if(startplanecol[itrk]>stopplanecol[itrk])plane = -plane; //reverse if track is backwards
          if(itrk==0){
                  for(unsigned int icc=0;icc<track[itrk]->NCell();icc++){
                    geo::OfflineChan tt(track[itrk]->Cell(icc)->Plane(),track[itrk]->Cell(icc)->Cell());
                  }
          }
          //vertex region is 8 planes in front and back of the start of a track
          //first looking if cell hit is in backwards region
          double celltrkdist= RecoJMShower::GetCellDistToTrk((*itr)->Plane(),(*itr)->Cell(),shstartcol[itrk],shendcol[itrk]);

          if(plane>=-8&&plane<0){
            double cellgev = 0;
            geo::View_t view = geom->Plane((*itr)->Plane())->View();
            TVector3 trkhitpos = RecoJMShower::GetTrkHitPos((*itr)->Plane(),(*itr)->Cell(),track[itrk]);
            double w1 = track[itrk]->W((*itr).get());
            double xyzd1[4];
            calib::GetXYZD((*itr)->OfflineChan(), w1, xyzd1);
            double readDist = xyzd1[3];
            //compute an attenuation correction based on hit distance to readout.  This is done for the FD
            //Correction factor determined in MC by comparing simulated calibration to truth
  
            if(fDetid!=novadaq::cnv::kFARDET||pCorAbsCell==false){
               // for NDOS or ND
               double witrk = track[itrk]->W((*itr).get());
  //To consider extra fiber of each cell in MC, for data dont apply this because cell are calibrated one by one.
               double pigtail = geom->GetPigtail(geom->Plane((*itr)->Plane())->Cell((*itr)->Cell())->Id());
               if(evt.isRealData()==false&&pCorPigtail==true)witrk = witrk - (pigtail - 92.6624);//W - (pigtail - avg. pigtail)
               const rb::RecoHit rhit = cal->MakeRecoHit(*(*itr), witrk);
               if(rhit.IsCalibrated()){
                 cellgev = rhit.GeV();
               }else if(pUseUncalib == true){
                 double adc = double((*itr)->ADC());
                 double Att = 1;
                 if(view == geo::kX){
                   Att = RecoJMShower::CellADCToGeV(readDist, view, fDetid);
                 }else if(view == geo::kY){
                   Att = RecoJMShower::CellADCToGeV(readDist, view, fDetid);
                 }
                 cellgev =  adc/Att;
               }
            }
            else{
              double adc = double((*itr)->ADC());
              double Att = 1;
              //Factor is different for each detector view
              if(view == geo::kX){
                      Att = RecoJMShower::CellADCToGeV(readDist, view, fDetid); 
              }else if(view == geo::kY){
                      Att = RecoJMShower::CellADCToGeV(readDist, view, fDetid); 
              }
              cellgev = adc/Att;
            }
            double ecor = 1.0;
            if(evt.isRealData()==true&&pCorDataCell)ecor = RecoJMShower::CellEDataMC(cellgev, readDist,  view, fDetid); // correct Data/MC difference for cell energy
            vtxGeV[itrk]+=cellgev/ecor;
            if(celltrkdist<19.5*cellW)vtxCenterGeV[itrk]+=cellgev/ecor;
          }

          int shnplane = abs((int)stopplanecol[i]-(int)startplanecol[i])+1;

          if(plane>=-1*shnplane&&plane<2*shnplane&&celltrkdist<19.5*cellW){// Calculate minimum distance from shower cells to detector edges, extrapolating the shower to -L from start point and +L from stop point. Cells within 20 cells to the core are considered.
            double cellgev = 0;
            geo::View_t view = geom->Plane((*itr)->Plane())->View();
            TVector3 trkhitpos = RecoJMShower::GetTrkHitPos((*itr)->Plane(),(*itr)->Cell(),track[itrk]);
            //compute an attenuation correction based on hit distance to readout.  This is done for the FD
            //Correction factor determined in MC by comparing simulated calibration to truth
            double w1 = track[itrk]->W((*itr).get());
            double xyzd1[4];
            calib::GetXYZD((*itr)->OfflineChan(), w1, xyzd1);
            double readDist = xyzd1[3];

            if(fDetid!=novadaq::cnv::kFARDET||pCorAbsCell==false){
                double witrk = track[itrk]->W((*itr).get());
  //To consider extra fiber of each cell in MC, for data dont apply this because cell are calibrated one by one.
                double pigtail = geom->GetPigtail(geom->Plane((*itr)->Plane())->Cell((*itr)->Cell())->Id());
                if(evt.isRealData()==false&&pCorPigtail==true)witrk = witrk - (pigtail - 92.6624);//W - (pigtail - avg. pigtail)
               const rb::RecoHit rhit = cal->MakeRecoHit(*(*itr), witrk);
               if(rhit.IsCalibrated()){
                 cellgev = rhit.GeV();
               }else if(pUseUncalib == true){
                 double adc = double((*itr)->ADC());
                 double Att = 1;
                 if(view == geo::kX){
                   Att = RecoJMShower::CellADCToGeV(readDist, view, fDetid);
                 }else if(view == geo::kY){
                   Att = RecoJMShower::CellADCToGeV(readDist, view, fDetid);
                 }
                 cellgev =  adc/Att;
               }
            }
            else{
              double adc = double((*itr)->ADC());
              double Att = 1;
              //Factor is different for each detector view
              if(view == geo::kX){
                      Att = RecoJMShower::CellADCToGeV(readDist, view, fDetid); 
              }else if(view == geo::kY){
                      Att = RecoJMShower::CellADCToGeV(readDist, view, fDetid); 
              }
              cellgev = adc/Att;
            }
            double ecor = 1.0;
            if(evt.isRealData()==true&&pCorDataCell)ecor = RecoJMShower::CellEDataMC(cellgev, readDist,  view, fDetid); // correct Data/MC difference for cell energy

            int dx1 = abs((int)(*itr)->Cell()-(int)pXEdgeId1);
            int dx2 = abs((int)(*itr)->Cell()-(int)pXEdgeId2);
            int dy1 = abs((int)(*itr)->Cell()-(int)pYEdgeId1);
            int dy2 = abs((int)(*itr)->Cell()-(int)pYEdgeId2);
            int dz1 = abs((int)(*itr)->Plane()-(int)pZEdgeId1);
            int dz2 = abs((int)(*itr)->Plane()-(int)pZEdgeId2);
            int cmd = 9999;
            int ce = -1;
            if(cmd>dx1){cmd=dx1;ce=1;}
            if(cmd>dx2){cmd=dx2;ce=2;}
            if(cmd>dy1){cmd=dy1;ce=3;}
            if(cmd>dy2){cmd=dy2;ce=4;}
            if(cmd>dz1){cmd=dz1;ce=5;}
            if(cmd>dz2){cmd=dz2;ce=6;}
            if(shNCellToEdge[itrk]>cmd){
              shNCellToEdge[itrk]=cmd;
              shClosetEdge[itrk]=ce;
            }
            if(cmd<2)shEdge2GeV[itrk] += cellgev/ecor;
            if(cmd<5)shEdge5GeV[itrk] += cellgev/ecor;
            if(cmd<10)shEdge10GeV[itrk] += cellgev/ecor;
            if(cmd<15)shEdge15GeV[itrk] += cellgev/ecor;
            if(cmd<20)shEdge20GeV[itrk] += cellgev/ecor;
          }
 
  
          //if the hit occurs before(after) the start(end) of the track continue
          if((*itr)->Plane()<std::min(startplanecol[itrk],stopplanecol[itrk])||(*itr)->Plane()>std::max(startplanecol[itrk],stopplanecol[itrk])){fHitDistMap[key].push_back(0);continue;}//for vtx fit
          //get the position of hit with respect to track. (distance to readout,transverse coord, transverse dist from hit to track)
          TVector3 trkhitpos = RecoJMShower::GetTrkHitPos((*itr)->Plane(),(*itr)->Cell(),track[itrk]);

          double w1 = track[itrk]->W((*itr).get());
          double xyzd1[4];
          calib::GetXYZD((*itr)->OfflineChan(), w1, xyzd1);
          double readDist = xyzd1[3];
  //        double celltrkdist= trkhitpos.z();
  //        double celltrkdist= RecoJMShower::GetCellDistToTrk((*itr)->Plane(),(*itr)->Cell(),track[itrk].Start(),RecoJMShower::GetShwStop(track[itrk]));
          //get the perpendicular distance from the cell to the track
//          double celltrkdist= RecoJMShower::GetCellDistToTrk((*itr)->Plane(),(*itr)->Cell(),shstartcol[itrk],shendcol[itrk]);//orig
           /*
          if((startplanecol[itrk]<std::min(shstartplanecol[itrk],shstopplanecol[itrk])||startplanecol[itrk]>std::max(shstartplanecol[itrk],shstopplanecol[itrk]))&&((*itr)->Plane()>=std::min(startplanecol[itrk],shstartplanecol[itrk])&&(*itr)->Plane()<=std::max(startplanecol[itrk],shstartplanecol[itrk]))&&(startcol[itrk]-shstartcol[itrk]).Mag()>cellW&&pVtxFit==true){
            celltrkdist= RecoJMShower::GetCellDistToTrk((*itr)->Plane(),(*itr)->Cell(),startcol[itrk],shstartcol[itrk]);
          }
          */
  
          // double celltrkdistorig = RecoJMShower::GetCellDistToTrk((*itr)->Plane(),(*itr)->Cell(),track[itrk]->Start(),RecoJMShower::GetShwStop(track[itrk]));
          //record the perp distance from the cell track inside the map
          fHitDistMap[key].push_back(celltrkdist);
  
          if(plane>=8&&celltrkdist>19.5*cellW&&trktag==false)continue;
          if(plane<8&&celltrkdist>1.5*cellW&&trktag==false){
            double cellgev = 0;
//            geo::View_t view = geom->Plane((*itr)->Plane())->View();
//            TVector3 trkhitpos = RecoJMShower::GetTrkHitPos((*itr)->Plane(),(*itr)->Cell(),track[itrk]);
            if(fDetid!=novadaq::cnv::kFARDET||pCorAbsCell==false){
	      // for NDOS or ND
	      double witrk = track[itrk]->W((*itr).get());
	      //To consider extra fiber of each cell in MC, for data dont apply this because cell are calibrated one by one.
	      double pigtail = geom->GetPigtail(geom->Plane((*itr)->Plane())->Cell((*itr)->Cell())->Id());
	      if(evt.isRealData()==false&&pCorPigtail==true)witrk = witrk - (pigtail - 92.6624);//W - (pigtail - avg. pigtail)
	      const rb::RecoHit rhit = cal->MakeRecoHit(*(*itr), witrk);
	      if(rhit.IsCalibrated()){
		cellgev = rhit.GeV();
	      }else if(pUseUncalib == true){
		double adc = double((*itr)->ADC());
		double Att = 1;
		if(view == geo::kX){
		  Att = RecoJMShower::CellADCToGeV(readDist, view, fDetid);
		}else if(view == geo::kY){
		  Att = RecoJMShower::CellADCToGeV(readDist, view, fDetid);
		}
		cellgev =  adc/Att;
	      }        
            }else{      
              double adc = double((*itr)->ADC());
              double Att = 1;
              if(view == geo::kX){
                Att = RecoJMShower::CellADCToGeV(readDist, view, fDetid); 
              }else if(view == geo::kY){
                Att = RecoJMShower::CellADCToGeV(readDist, view, fDetid); 
              }
              cellgev = adc/Att;
            }
            double ecor = 1.0;
            if(evt.isRealData()==true&&pCorDataCell)ecor = RecoJMShower::CellEDataMC(cellgev, readDist,  view, fDetid); // correct Data/MC difference for cell energy
            vtxGeV[itrk]+=cellgev/ecor;
            if(celltrkdist<19.5*cellW)vtxCenterGeV[itrk]+=cellgev/ecor;
            continue;
          }
          if(pRecluster==false&&trktag==false)continue;
  //        if(plane<8&&celltrkdist>19.5*cellW)continue;
          vCluster[itrk].Add((*itr));
          bool isinsh = false;//Add this cell to cell-shower map
          for(unsigned int ish=0;ish<hitshmap[key].size();ish++){
            if(hitshmap[key][ish]==(int)itrk)isinsh=true;
          }
          if(isinsh==false) hitshmap[key].push_back((int)itrk);
  
          double witrk = track[itrk]->W((*itr).get());
  //To conder extra fiber of each cell in MC, for data dont apply this because cell are calibrated one by one.
          double pigtail = geom->GetPigtail(geom->Plane((*itr)->Plane())->Cell((*itr)->Cell())->Id());
          if(evt.isRealData()==false&&pCorPigtail==true)witrk = witrk - (pigtail - 92.6624);//W - (pigtail - avg. pigtail)
          double peattitrk = 0;
          const rb::RecoHit rhit = cal->MakeRecoHit(*(*itr), witrk);
          if(rhit.IsCalibrated()){
            peattitrk = rhit.GeV(); 
          }else if(pUseUncalib == true){
            double adc = double((*itr)->ADC());
            double Att = 1;
            if(view == geo::kX){
              Att = RecoJMShower::CellADCToGeV(readDist, view, fDetid);
            }else if(view == geo::kY){
              Att = RecoJMShower::CellADCToGeV(readDist, view, fDetid);
            }
            peattitrk =  adc/Att;
          } 
//          const double peattitrk = cal->GetPECorr(*(*itr), witrk);
  //        trkattcal[itrk] = peattitrk;
          const double ai = peattitrk/(*itr)->PE();
          thiteweight[itrk] = track[itrk]->TotalPE()*exp(-3.28769e-01*celltrkdist);//Shower lateral profile constant from a fit to (Transverse dE/dx)/(shower Energy) as a function of distance to shower core for Monte Carlo electron.
  //        thiteweight[itrk] = track[itrk]->TotalPE()*exp(-2.1*trkhitpos.z()/8.5);
          thiteweighttotal += thiteweight[itrk];
          thiteweighttotalatt += thiteweight[itrk]/ai;
        }
  
        for(unsigned int itrk=0;itrk<track.size();itrk++){
          if(vNantag[itrk]==true){
            fHitEWeightMap[key].push_back(0);
            fHitEMap[key]+=0;
            continue;
          }
  //        const double gev = trueEnergy[key]*(thiteweight[itrk]/thiteweighttotalatt);
          if(thiteweighttotal>0){
            double gev = 0; 
            geo::View_t view = geom->Plane((*itr)->Plane())->View();
            TVector3 trkhitpos = RecoJMShower::GetTrkHitPos((*itr)->Plane(),(*itr)->Cell(),track[itrk]);
            double w1 = track[itrk]->W((*itr).get());
            double xyzd1[4];
            calib::GetXYZD((*itr)->OfflineChan(), w1, xyzd1);
            double readDist = xyzd1[3];
  
  //          if(fDetid==novadaq::cnv::kFARDET||fDetid==novadaq::cnv::kNDOS||fDetid==novadaq::cnv::kNEARDET)
//            double witrk = track[itrk]->W((*itr).get());
            if(fDetid!=novadaq::cnv::kFARDET||pCorAbsCell==false){
              gev =  (*itr)->PE()*(thiteweight[itrk]/thiteweighttotalatt);
//              const rb::RecoHit rhit = cal->MakeRecoHit((*itr).get(), witrk);
//              if(rhit.IsCalibrated()){
//                gev = rhit.GeV();
//              }else if(pUseUncalib == true){
//                double adc = double((*itr)->ADC());
//                double Att = 1;
//                if(view == geo::kX){
//                  Att = RecoJMShower::CellADCToGeV(trkhitpos[0], view, fDetid);
//                }else if(view == geo::kY){
//                  Att = RecoJMShower::CellADCToGeV(trkhitpos[0], view, fDetid);
//                }
//                gev =  adc/Att;
//              }               // for NDOS or ND
            }else{
               // for FD, Attenuation effect calibrated for development (Product S12.02.14), 
  	       // should not be changed to keep the energy scale identical to the one used in ANN training! 
  	      double adc = double((*itr)->ADC());
  	      double Att = 1;
  	      if(view == geo::kX){
                  Att = RecoJMShower::CellADCToGeV(readDist, view, fDetid); 
  	      }else if(view == geo::kY){
                  Att = RecoJMShower::CellADCToGeV(readDist, view, fDetid); 
  	      }
              gev = adc*(thiteweight[itrk]/thiteweighttotal)/Att; 
  	    }

            double ecor = 1.0;
            if(evt.isRealData()==true&&pCorDataCell)ecor = RecoJMShower::CellEDataMC(gev, readDist,  view, fDetid); // correct Data/MC difference for cell energy
            gev = gev/ecor;
  	    fHitEWeightMap[key].push_back(gev);
  	    fHitEMap[key]+=gev;
  	  }else{
  	    fHitEWeightMap[key].push_back(0);
  	    fHitEMap[key]+=0;
  	  }
        }
      }
  
      for(unsigned int itrk=0;itrk<track.size();itrk++){// calculate orignal prong energy
        if(vNantag[itrk]==true)continue;
        for(unsigned int i=0;i<track[itrk]->NCell();i++){
          art::Ptr< rb::CellHit > trkcell = track[itrk]->Cell(i);
  	  double gev = 0;
          geo::View_t view = geom->Plane(trkcell->Plane())->View();
          TVector3 trkhitpos = RecoJMShower::GetTrkHitPos(trkcell->Plane(),trkcell->Cell(),track[itrk]);
          double w1 = track[itrk]->W(trkcell.get());
          double xyzd1[4];
          calib::GetXYZD(trkcell->OfflineChan(), w1, xyzd1);
          double readDist = xyzd1[3];
  	if(fDetid==novadaq::cnv::kFARDET||fDetid==novadaq::cnv::kNDOS||fDetid==novadaq::cnv::kNEARDET||pCorAbsCell==false){
  	  //      if(fDetid!=novadaq::cnv::kFARDET)
  	  // for NDOS or ND
	  double witrk = track[itrk]->W(trkcell.get());
  //To con  sider extra fiber of each cell in MC, for data dont apply this because cell are calibrated one by one.
	  double pigtail = geom->GetPigtail(geom->Plane(trkcell->Plane())->Cell(trkcell->Cell())->Id());
	  if(evt.isRealData()==false&&pCorPigtail==true)witrk = witrk - (pigtail - 92.6624);//W - (pigtail - avg. pigtail)
            const rb::RecoHit rhit = cal->MakeRecoHit(*trkcell, witrk);
            if(rhit.IsCalibrated()){
              gev = rhit.GeV();
            }else if(pUseUncalib == true){
              double adc = double(trkcell->ADC());
              double Att = 1;
              if(view == geo::kX){
                Att = RecoJMShower::CellADCToGeV(readDist, view, fDetid);
              }else if(view == geo::kY){
                Att = RecoJMShower::CellADCToGeV(readDist, view, fDetid);
              }
              gev =  adc/Att;
            }
  	}else{
  		// for FD, Attenuation effect calibrated for development (Product S12.02.14),
  		// should not be changed to keep the energy scale identical to the one used in ANN training!
            double adc = double(trkcell->ADC());
            double Att = 1;
            if(view == geo::kX){
                Att = RecoJMShower::CellADCToGeV(readDist, view, fDetid);
            }else if(view == geo::kY){
                Att = RecoJMShower::CellADCToGeV(readDist, view, fDetid);
           }
            gev = adc/Att;
          }
          double ecor = 1.0;
          if(evt.isRealData()==true&&pCorDataCell)ecor = RecoJMShower::CellEDataMC(gev, readDist, view, fDetid); // correct Data/MC difference for cell energy
          gev = gev/ecor;
          trkGeV[itrk]+= gev;
        }
      }
  
        //loop over x cells
        for(unsigned int i=0;i<clust->NXCell();i++){
           geo::OfflineChan key(clust->XCell(i)->Plane(),clust->XCell(i)->Cell());
          //only use cell hits that not in any showers and with pe above threshold level set if fcl file
          if((clust->XCell(i)->PE())>pCellHitPHThresh&&hitshmap[key].size()==0){
            //make reco hits out of cells to sum energy
            const art::Ptr<rb::CellHit>& chit = clust->XCell(i);
            const rb::RecoHit rhit(cal->MakeRecoHit(*chit, wx ));

            double gev=0;
            geo::View_t view = geom->Plane(chit->Plane())->View();
            double pigtail = geom->GetPigtail(geom->Plane(chit->Plane())->Cell(chit->Cell())->Id());
            //double wxtoelec = wx - (pigtail - 92.6624);//W - (pigtail - avg. pigtail)
            //double wytoelec = wy - (pigtail - 92.6624);//W - (pigtail - avg. pigtail)
            double w1 = wx;
            if (view == geo::kY) w1 = wy;
            double xyzd1[4];
            calib::GetXYZD(chit->OfflineChan(), w1, xyzd1);
            double readDist = xyzd1[3];
            if(rhit.IsCalibrated()){
              gev = rhit.GeV();
            }else if(pUseUncalib == true){// use MC calibration if a cell is not calibrated
//              std::cout<<"slice, plane, cell, wx, gev "<<ic<<" "<<chit->Plane()<<" "<<chit->Cell()<<" "<<wx<<" Not calibrated?!" << std::endl;
              double adc = double(chit->ADC());
              double Att = 1;
              if(view == geo::kX){
                Att = RecoJMShower::CellADCToGeV(readDist, view, fDetid);
              }else if(view == geo::kY){
                Att = RecoJMShower::CellADCToGeV(readDist, view, fDetid);
              }
              gev = adc/Att;
            }
  
            double ecor = 1.;
            double disttoelec = geom->Plane(chit->Plane())->Cell(chit->Cell())->HalfL()+pigtail-wx;
            if(evt.isRealData()==true&&pCorDataCell)ecor = RecoJMShower::CellEDataMC(gev, disttoelec,  view, fDetid); // correct Data/MC difference for cell energy
            gev = gev/ecor;
            sliceExclGeV += gev;
          }
        }

        //loop over y cells
        for(unsigned int i=0;i<clust->NYCell();i++){
          geo::OfflineChan key(clust->YCell(i)->Plane(),clust->YCell(i)->Cell());
          //use hits above pe threshold and are not on any showers
          if((clust->YCell(i)->PE())>pCellHitPHThresh&&hitshmap[key].size()==0){
            //make reco hits out of cells
            const art::Ptr<rb::CellHit>& chit = clust->YCell(i);
            const rb::RecoHit rhit(cal->MakeRecoHit(*chit, wy ));

            double gev=0;
            geo::View_t view = geom->Plane(chit->Plane())->View();
            double pigtail = geom->GetPigtail(geom->Plane(chit->Plane())->Cell(chit->Cell())->Id());
            //double wxtoelec = wx - (pigtail - 92.6624);//W - (pigtail - avg. pigtail)
            //double wytoelec = wy - (pigtail - 92.6624);//W - (pigtail - avg. pigtail)
            double w1 = wx;
            if (view == geo::kY) w1 = wy;
            double xyzd1[4];
            calib::GetXYZD(chit->OfflineChan(), w1, xyzd1);
            double readDist = xyzd1[3];
            if(rhit.IsCalibrated()){
              gev = rhit.GeV();
            }else if(pUseUncalib == true){// use MC calibration if a cell is not calibrated
//              std::cout<<"slice, plane, cell, wy, gev "<<ic<<" "<<chit->Plane()<<" "<<chit->Cell()<<" "<<wy<<" Not calibrated?!" << std::endl;
              double adc = double(chit->ADC());
              double Att = 1;
              if(view == geo::kX){
                Att = RecoJMShower::CellADCToGeV(readDist, view, fDetid);
              }else if(view == geo::kY){
                Att = RecoJMShower::CellADCToGeV(readDist, view, fDetid);
              }
              gev = adc/Att;
            }
            double ecor = 1.;
            double disttoelec = geom->Plane(chit->Plane())->Cell(chit->Cell())->HalfL()+pigtail-wy;
            if(evt.isRealData()==true&&pCorDataCell)ecor = RecoJMShower::CellEDataMC(gev, disttoelec,  view, fDetid); // correct Data/MC difference for cell energy
            gev = gev/ecor;
  
            sliceExclGeV += gev;
          }
        }
  
  //After clustering, initial pclusters 
     for(unsigned int i = 0; i < vCluster.size(); i++){
       std::vector<PCluster> showerpcluster;
       std::vector<PCluster> trackpcluster;
       showerpcluster.clear();
       trackpcluster.clear();
       if(vNantag[i]==true){ 
         fShowerPClusterCol.push_back(showerpcluster);
         fTrackPClusterCol.push_back(trackpcluster);
         continue;
       }
       if(startplanecol[i]<=stopplanecol[i]){// deal with backward
         for(int ip = (int)startplanecol[i];ip<=(int)stopplanecol[i];ip++){
           PCluster pcluster;
           pcluster.SetPlane(ip);
           showerpcluster.push_back(pcluster);
           trackpcluster.push_back(pcluster);
         }
       }else{
         for(int ip = (int)startplanecol[i];ip>=(int)stopplanecol[i];ip--){
           PCluster pcluster;
           pcluster.SetPlane(ip);
           showerpcluster.push_back(pcluster);
           trackpcluster.push_back(pcluster);
         }
       } 
       for (unsigned int nh=0;nh<vCluster[i].NCell();nh++) {
         const art::Ptr<rb::CellHit>  cellhit = vCluster[i].Cell(nh);
         int plane = (int)cellhit->Plane()-(int)startplanecol[i];
         if(startplanecol[i]>stopplanecol[i])plane=-plane;
         showerpcluster[plane].Add((cellhit));
         geo::OfflineChan key(cellhit->Plane(),cellhit->Cell());
         bool trktag=false;// on orignal tracks ?
         for(unsigned int imap=0;imap<hittrkmap[key].size();imap++){
           if(hittrkmap[key][imap]==(int)i)trktag=true;
         }
         if(trktag==true)trackpcluster[plane].Add((cellhit));
       }
       fShowerPClusterCol.push_back(showerpcluster);
       fTrackPClusterCol.push_back(trackpcluster);
     }
     for(unsigned int i = 0; i < vCluster.size(); i++){//calculate vertex doca to each shower
       double startx=-9999,starty=-9999,startz=-9999;
       double stopx=-9999,stopy=-9999,stopz=-9999;
       bool startxtag=false,startytag=false,startztag=false;
       bool stopxtag=false,stopytag=false,stopztag=false;
       if(vNantag[i]==true){
         vtxdocacol.push_back(9999);
         continue;
       }
       for(int plane = 0; plane < (int)fShowerPClusterCol[i].size(); plane++ ){ 
         unsigned int geoplane = fShowerPClusterCol[i][plane].Plane();
         geo::View_t view = geom->Plane(geoplane)->View();
         unsigned int planecell = RecoJMShower::GetTrkCPlaneCell((unsigned int)plane, i);
         if(planecell==9999)continue;
         Double_t cellXYZ[3]={0,0,0};
         geom->Plane(geoplane)->Cell(planecell)->GetCenter(cellXYZ);
         if(startztag==false){startz = cellXYZ[2];startztag=true;} 
         if(view==geo::kX&&startxtag==false){startx = cellXYZ[0];startxtag=true;}
         if(view==geo::kY&&startytag==false){starty = cellXYZ[1];startytag=true;}
         if(startxtag==true&&startytag==true&&startztag==true)break;
       }

       for(int plane = (int)fShowerPClusterCol[i].size()-1;plane >= 0; plane-- ){ 
         unsigned int geoplane = fShowerPClusterCol[i][plane].Plane();
         geo::View_t view = geom->Plane(geoplane)->View();
         unsigned int planecell = RecoJMShower::GetTrkCPlaneCell((unsigned int)plane, i);
         if(planecell==9999)continue;
         Double_t cellXYZ[3]={0,0,0};
         geom->Plane(geoplane)->Cell(planecell)->GetCenter(cellXYZ);
         if(stopztag==false){stopz = cellXYZ[2];stopztag=true;} 
         if(view==geo::kX&&stopxtag==false){stopx = cellXYZ[0];stopxtag=true;}
         if(view==geo::kY&&stopytag==false){stopy = cellXYZ[1];stopytag=true;}
         if(stopxtag==true&&stopytag==true&&stopztag==true)break;
       }
         double vtxDoca = 9999;
         if(!(startx<-9000||starty<-9000||startz<-9000||stopx<-9000||stopy<-9000||stopz<-9000)&&evtvtxcol.size()>0){
           TVector3 start(startx,starty,startz);
           TVector3 stop(stopx,stopy,stopz);
           vtxDoca = RecoJMShower::GetPointDoca(evtvtxcol[0], start, stop);
         }
       vtxdocacol.push_back(vtxDoca);
     }
     
  //Initial tcclusters 
  /*
     for(unsigned int i = 0; i < vCluster.size(); i++){
       std::vector<TCCluster> showertccluster;
       showertccluster.clear();
       if(vNantag[i]==true){
         fShowerTCClusterCol.push_back(showertccluster);
         continue;
       }
       for(int itc = 0; itc<=20;itc++){
         TCCluster tccluster;
         tccluster.SetTCell(itc);
         showertccluster.push_back(tccluster);
       }
  
       for (unsigned int nh=0;nh<vCluster[i].NCell();nh++) {
         const art::Ptr<rb::CellHit>  cellhit = vCluster[i].Cell(nh);
         unsigned transcellId = cellhit->Cell();
         unsigned int planecell = RecoJMShower::GetTrkPlaneCell(cellhit->Plane(), vCluster[i]);
         int plane = (int)cellhit->Plane()-(int)startplanecol[i];
         if(startplanecol[i]>stopplanecol[i])plane=-plane;
  //       unsigned int planecell = RecoJMShower::GetTrkCPlaneCell((unsigned int)plane, i);
  //       std::cout<<"GetTrkPlaneCell, GetTrkCPlaneCell "<<planecell1<<" "<<planecell<<std::endl;
         if(planecell==9999)continue;
         int itc = fabs(int(planecell) - int(transcellId));
         if(itc>20)continue;
         showertccluster[itc].Add(cellhit);
       }
       fShowerTCClusterCol.push_back(showertccluster);
     }
  */
  
  //fill cells to showercol
      for(unsigned int i = 0; i < vCluster.size(); i++){
          if(vNantag[i]==true)continue;
  //      jmshower::JMShower shower(vCluster[i], startcol[i], dircol[i]);
        jmshower::JMShower shower(vCluster[i], i, vCluster[i].TotalLength());

        if(vCluster[i].NCell()==0){
  
          std::cout<<"Zero Cell Track! i, NCell, NTrk  "<<i<<" "<<track[i]->NCell()<<" "<<track.size()<<std::endl;
          std::cout<<"Zero Cell Shower! i, NCell, NSh "<<i<<" "<<vCluster[i].NCell()<<" "<<vCluster.size()<<std::endl;
          continue;
        }
  //      shower.fPClusterCol.clear();
  //      shower.SetWindowSize(shower.ExtentPlane());
        shower.SetWindowSize(fShowerPClusterCol[i].size());
        shower.SetStartPlane(startplanecol[i]);
        shower.SetStopPlane(stopplanecol[i]);
        shower.SetNPlane(abs((int)stopplanecol[i]-(int)startplanecol[i])+1);
        shower.SetXNPlane(vXCluster[i].ExtentPlane());
        shower.SetYNPlane(vYCluster[i].ExtentPlane());
        shower.SetXMinPlane(vXCluster[i].MinPlane());
        shower.SetYMinPlane(vYCluster[i].MinPlane());
        shower.SetXMaxPlane(vXCluster[i].MaxPlane());
        shower.SetYMaxPlane(vYCluster[i].MaxPlane());

        shower.SetHitColSize(shower.NCell());
        geo::OfflineChan key(vCluster[i].Cell(0)->Plane(),vCluster[i].Cell(0)->Cell());
        shower.SetISlice(ic);
        shower.SetSliceGeV(sliceGeV);
        shower.SetSliceEtot(sliceEtot);
        shower.SetSliceEdge10GeV(sliceEdge10GeV);
        shower.SetSliceEdge5GeV(sliceEdge5GeV);
        shower.SetSliceEdge2GeV(sliceEdge2GeV);

        shower.SetNCellToEdge(shNCellToEdge[i]);
        shower.SetClosetEdge(shClosetEdge[i]);
        shower.SetEdge20GeV(shEdge20GeV[i]);
        shower.SetEdge15GeV(shEdge15GeV[i]);
        shower.SetEdge10GeV(shEdge10GeV[i]);
        shower.SetEdge5GeV(shEdge5GeV[i]);
        shower.SetEdge2GeV(shEdge2GeV[i]);

        shower.SetSliceExclGeV(sliceExclGeV);
        for (unsigned int nh=0;nh<shower.NCell();nh++) {
          art::Ptr<rb::CellHit>  cellhit = shower.Cell(nh);
          geo::OfflineChan key(cellhit->Plane(),cellhit->Cell());
          double cellEnergy = fHitEWeightMap[key][i];
          shower.SetHitDeconvE(nh, cellEnergy);
          shower.SetHitDist(nh, fHitDistMap[key][i]);
          shower.SetHitSumE(nh, fHitEMap[key]);
        }
  
  //Set ID
        shower.SetID(i);
  //Set Start and end point
  //      shower.AppendTrajectoryPoint(endcol[i]);
        shower.SetStop(endcol[i]);
        shower.SetStart(startcol[i]);
        shower.SetGap(gapcol[i]);
// determine the gap between vertex and the start of the continuous plane
        unsigned int newstartp = RecoJMShower::ContStartPlane(shower, 0, 3);
        unsigned int contstartp = fShowerPClusterCol[i][newstartp].Plane();
        double contGap = 0;
        if(evtvtxplanecol.size()>0&&contstartp!=0){
          double cellD = 2*geom->Plane(0)->Cell(0)->HalfD();
          if(fabs(shower.Dir().Z())>0.00001)contGap=(double)abs((int)contstartp-(int)evtvtxplanecol[0])*cellD/fabs(shower.Dir().Z());   
        }        
        contGap=std::max(gapcol[i],contGap);
        shower.SetContGap(contGap);
        shower.SetVtxDoca(vtxdocacol[i]);
        shower.SetBackTag(backtagcol[i]);
  
  //Set Geometry 
        shower.SetIsFiducial(RecoJMShower::IsFiducial(shower));
        shower.SetDistToEdge(RecoJMShower::GetTrkDistToEdge(shower));
  //Set Energy 
        double trkcor = RecoJMShower::ECorrMC(trkGeV[i]);
        if(trkcor<0.0001||trkcor>1000)trkcor = 1;
  
        shower.SetDepositEnergy(RecoJMShower::DepositEnergy(shower, shower.ID()));
        shower.SetEnergy(RecoJMShower::Energy(shower));// with position correction
        sliceEtot+=RecoJMShower::Energy(shower);
   //      shower.SetEnergy(RecoJMShower::Energy(shower, shower.ID())); // wihout position correction
        shower.SetRadius(RecoJMShower::Radius(shower));
  
  
        TVector3 pos((shower.Start()[0]+RecoJMShower::GetShwStop(shower)[0])/2.0,(shower.Start()[1]+RecoJMShower::GetShwStop(shower)[1])/2.0,(shower.Start()[2]+RecoJMShower::GetShwStop(shower)[2])/2.0);
        int iReg = 0;
        if(pos[0]>=0&&pos[1]>=0)iReg = 0;
        if(pos[0]>=0&&pos[1]<0)iReg = 1;
        if(pos[0]<0&&pos[1]>=0)iReg = 2;
        if(pos[0]<=0&&pos[1]<=0)iReg = 3;
  
        double gev = RecoJMShower::Energy(shower);
        double elower[11] = {0.00, 0.25, 0.75, 1.25, 1.75, 2.25, 2.75, 3.25, 3.75, 4.25, 4.75};
        double eupper[11] = {0.25, 0.75, 1.25, 1.75, 2.25, 2.75, 3.25, 3.75, 4.25, 4.75, 5.25};
        double epoint[11];
        for(int ig=0;ig<11;ig++){
          epoint[ig] = (elower[ig]+eupper[ig])/2.;
        }
        int igev0 = 0;
        int igev1 = 0;
        if(gev>epoint[10]){igev0=10;igev1=10;}
        else if(gev<epoint[0]){igev0=0;igev1=0;}
        else{
          for(int ig=0;ig<10;ig++){
            if(gev>epoint[ig]&&gev<epoint[ig+1]){igev0=ig;igev1=ig+1;}
          }
        }
        if(igev0>10)igev0=10;
        if(igev1>10)igev1=10;
        double e0 = gev-epoint[igev0];
        double e1 = epoint[igev1] - gev;
  
  
        for(unsigned int plane = 0; plane<fShowerPClusterCol[i].size();plane++){
          double dedx = RecoJMShower::GetPlaneDedx(shower, plane);
  
  /*
          double path = RecoJMShower::GetTrkHitPath(fShowerPClusterCol[i][plane].Plane(), 0, shower);
          double ep = RecoJMShower::Energy(fShowerPClusterCol[i][plane], shower.ID());
          if(plane==0&&dedx<0.000001){
            unsigned int lastplane = fShowerPClusterCol[i].size()-1;
            std::cout<<"Ncell, adc, pe "<<fShowerPClusterCol[i][plane].NCell()<<" "<<fShowerPClusterCol[i][plane].TotalADC()<<" "<<fShowerPClusterCol[i][plane].TotalPE()<<" "<<std::endl;
            std::cout<<"gev,ep,path "<<fShowerPClusterCol[i][plane].TotalGeV()<<" "<<ep<<" "<<path<<std::endl;
            std::cout<<"last Ncell, adc, pe "<<fShowerPClusterCol[i][lastplane].NCell()<<" "<<fShowerPClusterCol[i][lastplane].TotalADC()<<" "<<fShowerPClusterCol[i][lastplane].TotalPE()<<" "<<std::endl;
            std::cout<<"lastgev,ep,path "<<fShowerPClusterCol[i][lastplane].TotalGeV()<<" "<<ep<<" "<<path<<std::endl;
          }
  */
  
          shower.SetPlaneDedx(plane, dedx);

          if(pRecoPlaneR==true){ 
//          double planerad = RecoJMShower::Radius(fShowerPClusterCol[i][plane],shower);//Use jmshower 
            double planerad = RecoJMShower::Radius(fTrackPClusterCol[i][plane],shower);//Use FuzzyKvertex Radius
            shower.SetPlaneRadius(plane, planerad);
          }
          TVector3 hitxyz = RecoJMShower::GetTrkPlanePos(fShowerPClusterCol[i][plane].Plane(), shower);
          shower.SetPlaneHitXYZ(plane, hitxyz);
          if(pRecoPID==true){
            for( int itype = int(jmshower::kNull)+1; itype != int(jmshower::kLast); ++itype){
              double prob = RecoJMShower::GetInterPlaneDedxProb(shower, plane, dedx, itype, pos, iReg, igev0, igev1, e0, e1);
              shower.SetPlaneProb(plane, prob, itype);
            }
          }
        }
        for(unsigned int tcell = 0; tcell<20; tcell++){
          shower.SetCellTransDedx(tcell, RecoJMShower::GetCellTransDedx(shower, tcell));
        }
        double sumdr = 0;
        double sum10dr = 0;
        for(unsigned int plane = 0; plane<fShowerPClusterCol[i].size();plane++){
          if((int)plane>=(int)fShowerPClusterCol[i].size()-1-(int)plane)break;
          double r1=shower.PlaneRadius(plane);
          double r2=shower.PlaneRadius(fShowerPClusterCol[i].size()-1-plane);
          double dr=r2-r1;
          if(r1>1e-10&&r2>1e-10){
            sumdr+=dr;
            if(plane<10)sum10dr+=dr;
          }
        } 
        shower.SetSumDRadius(sumdr); 
        shower.SetSum10DRadius(sum10dr); 
    //Virtual nodes and moments calculation
       std::vector<TVector3> nodetocorecol;
       std::vector<geo::OfflineChan> nodexidcol;
       std::vector<geo::OfflineChan> nodeyidcol;
       std::vector<double> nodetocoreenergycol;
       nodetocorecol.clear();
       nodetocoreenergycol.clear();
       nodexidcol.clear();
       nodeyidcol.clear();
  
       if(pRecoNode==true&&shower.TotalLength()<600.){
  
         double sumE = 0;
         TVector3 sumENodeToCore(0,0,0);
         double I11 = 0;
         double I12 = 0;
         double I21 = 0;
         double I22 = 0;
  
    // Using 3-D nodes to calculate shower balance
  //     TVector3 core(shower.Start()[0],shower.Start()[1],shower.Start()[2]); // original point of shower frame
         TVector3 iz = shower.Dir().Unit(); // z in shower frame
         TVector3 ix0(1.0,0.0,0);
         TVector3 iy0(0.0,1.0,0);
         TVector3 ix = iy0.Cross(iz).Unit(); // x axis in shower frame
         TVector3 iy = iz.Cross(ix).Unit(); // y axis in shower frame
  
         for(unsigned int plane = 0; plane<fShowerPClusterCol[i].size()-1;plane++){
           if(fShowerPClusterCol[i][plane].NCell()==0)continue;
           unsigned int plane1 = fShowerPClusterCol[i][plane].Plane();
           if(fShowerPClusterCol[i][plane+1].NCell()==0)continue;
           unsigned int plane2 = fShowerPClusterCol[i][plane+1].Plane();
           geo::View_t view1 = geom->Plane(plane1)->View();
           geo::View_t view2 = geom->Plane(plane2)->View();
           if(view1==view2)continue;
           TVector3 core = RecoJMShower::GetTrkPlanePos(plane1, shower);
           TVector3 pcore = RecoJMShower::GetTrkPlanePos(plane1, shower);
           std::vector<TVector3> planenodecol;
           std::vector<double> planenodeenergycol;
           planenodecol.clear();
           planenodeenergycol.clear();
           for (unsigned itp1 = 0;itp1<fShowerPClusterCol[i][plane].NCell();itp1++) {
             unsigned int cell1 = fShowerPClusterCol[i][plane].Cell(itp1)->Cell();
             for(unsigned itp2 = 0;itp2<fShowerPClusterCol[i][plane+1].NCell();itp2++){
               unsigned int cell2 = fShowerPClusterCol[i][plane+1].Cell(itp2)->Cell();
               TVector3 node = RecoJMShower::GetCellNodePos(plane1, cell1, plane2, cell2);// define 3-D node
               double nodez = (pcore - shower.Start()).Mag();
               TVector3 nodetopcore = node - pcore;
               nodetopcore.SetZ(nodez);
               TVector3 nodetocore = node - core;
               nodetocore.SetZ(nodez);
   
  
               TVector3 xyz = node - core;//transform xy to shower frame 
  
    //           double xp = ix[0]*xyz[0]+iy[0]*xyz[1]+iz[0]*xyz[2];// rotate to shower frame
    //           double yp = ix[1]*xyz[0]+iy[1]*xyz[1]+iz[1]*xyz[2];
    //           double zp = ix[2]*xyz[0]+iy[2]*xyz[1]+iz[2]*xyz[2];
    //           double rxy = sqrt(xp*xp+yp*yp);
    //           double rxypcore = sqrt(nodetopcore[0]*nodetopcore[0]+nodetopcore[1]*nodetopcore[1]);           
  
               double p = shower.Dir()[0];
               double q = shower.Dir()[1];
               double r = shower.Dir()[2];
               double x0 = xyz[0];
               double y0 = xyz[1];
               double z0 = xyz[2];
               TVector3 vt(q*z0-r*y0,r*x0-p*z0,p*y0-q*x0);
  //             double angx = vt.Angle(ix);
  //             double angy = vt.Angle(iy);
  //             double angz = vt.Angle(iz);
  
  //             double R = sqrt((q*z0-r*y0)*(q*z0-r*y0)+(r*x0-p*z0)*(r*x0-p*z0)+(p*y0-q*x0)*(p*y0-q*x0));// Rotate xy plane 
  //             TVector3 izcross =  ix.Cross(vt);
  //             double angzzcross = izcross.Angle(iz);
  //             if(fabs(angzzcross-3.14159265)<0.01)angx = 2*3.14159265-angx;
  //             double xrp = R*cos(angx);
  //             double yrp = R*sin(angx);
  //             double zrp = sqrt(xyz.Mag()*xyz.Mag()-R*R);
    /*
               std::cout<<"angle z zp "<<angzzcross<<std::endl;
               std::cout<<"angx,angy,angz "<<angx<<" "<<angy<<" "<<angz<<std::endl;
               std::cout<<"rxy, rxypcore, R "<<rxy<<" "<<rxypcore<<" "<<R<<std::endl;
               std::cout<<"Shower Dir          "<<shower.Dir()[0]<<" "<<shower.Dir()[1]<<" "<<shower.Dir()[2]<<std::endl;
               std::cout<<"ix Dir          "<<ix[0]<<" "<<ix[1]<<" "<<ix[2]<<std::endl;
               std::cout<<"iy Dir          "<<iy[0]<<" "<<iy[1]<<" "<<iy[2]<<std::endl;
               std::cout<<"iz Dir          "<<iz[0]<<" "<<iz[1]<<" "<<iz[2]<<std::endl;
               std::cout<<"xyz in shower Rotat "<<xrp<<" "<<yrp<<" "<<zrp<<std::endl;
               std::cout<<"xyz in shower       "<<xp<<" "<<yp<<" "<<zp<<std::endl;
               std::cout<<"xyz in shower pcore "<<nodetopcore[0]<<" "<<nodetopcore[1]<<" "<<nodez<<std::endl;
    */
  
    //         TVector3 nodetocore(xp,yp,zp); 
  //             TVector3 nodetocore(xrp,yrp,zrp);
               double rcn = nodetocore.Perp();
               geo::OfflineChan key1(plane1,cell1);
               geo::OfflineChan key2(plane2,cell2);
               double nodeenergy = fHitEWeightMap[key1][i]+fHitEWeightMap[key2][i];
               rcn = rcn*nodeenergy;
               sumE += nodeenergy;
               sumENodeToCore+=nodeenergy*nodetocore;
               I11 += nodeenergy*(nodetocore.Y()*nodetocore.Y());
               I12 += -nodeenergy*(nodetocore.X()*nodetocore.Y());
               I21 += -nodeenergy*(nodetocore.X()*nodetocore.Y());
               I22 += nodeenergy*(nodetocore.X()*nodetocore.X());
  
               planenodecol.push_back(node);
               planenodeenergycol.push_back(nodeenergy);
  
               nodetocorecol.push_back(nodetocore);
               nodetocoreenergycol.push_back(nodeenergy);
               if(view1 == geo::kX){
                 nodexidcol.push_back(key1);
                 nodeyidcol.push_back(key2);
               }else{
                 nodexidcol.push_back(key2);
                 nodeyidcol.push_back(key1);
               }
             }
           }
         }
    // calculate eiganvalues and diagonalize inertial tensor
         double Ixx = 0.5*(I11+I22+sqrt(pow(I11+I22,2)-4*(I11*I22-I12*I21)));
         double Iyy = 0.5*(I11+I22-sqrt(pow(I11+I22,2)-4*(I11*I22-I12*I21)));
         double asym = 1;
         if(fabs(Ixx+Iyy)>1e-10)asym=(Ixx-Iyy)/(Ixx+Iyy);
         double alpha = 3.14159265/2.;
         if(fabs(I12)>1e-10)alpha = atan((Ixx-I11)/I12);//rotate nodes to principal axes frame
         for(int inode=0;inode<(int)nodetocorecol.size();inode++){
           nodetocorecol[inode].RotateZ(-alpha);
         }
  
         shower.SetNodes(nodetocorecol, nodetocoreenergycol,nodexidcol,nodeyidcol);
         TVector3 ixyz(Ixx,Iyy,0);
         shower.SetIxyz(ixyz);
         shower.SetAsym(asym);
       }else{
         shower.SetNodes(nodetocorecol, nodetocoreenergycol,nodexidcol,nodeyidcol);
         TVector3 ixyz(0,0,0);
         shower.SetIxyz(ixyz);
         shower.SetAsym(0);
       }
  
  
  
        if(pRecoPID==true){
          for( int itype = int(jmshower::kNull)+1; itype != int(jmshower::kLast); ++itype)
          {
  //          std::cout<<"in RecoJMShower itype, LLL "<<itype<<" "<<RecoJMShower::GetDedxLongLL(shower, itype)<<std::endl;
  //          std::cout<<"in RecoJMShower itype, LLT "<<itype<<" "<<RecoJMShower::GetDedxTransLL(shower, itype)<<std::endl;
  //          std::cout<<"itype Proton = "<<int(jmshower::kProton)<<std::endl;
            shower.SetDedxLongLL(RecoJMShower::GetDedxLongLL(shower, itype), itype);
            if(pRecoInvLL==true){
//            if(itype==int(jmshower::kGamma))shower.SetDedxInvLongLL(RecoJMShower::GetDedxInvLongLL(shower, itype), itype);
              shower.SetDedxInvLongLL(RecoJMShower::GetDedxInvLongLL(shower, itype), itype);
            }
            shower.SetDedxTransLL(RecoJMShower::GetDedxTransLL(shower, itype), itype);
          }
        }
  
        shower.SetNMIPPlane(RecoJMShower::NMIPPlane(shower));
        shower.SetCentroid(RecoJMShower::GetCentroid(shower));
        shower.SetVtxGeV(vtxGeV[i]);
        shower.SetVtxCenterGeV(vtxCenterGeV[i]);
        shower.SetTrkGeV(trkGeV[i]/trkcor);
  //      for(unsigned int plane=0;plane<fShowerPClusterCol[i].size();plane++){
  //        shower.fPClusterCol.push_back(fShowerPClusterCol[i][plane]);
  //      }
  
  //      for(int itc = 0; itc<=20;itc++){
  //        shower.fTCClusterCol.push_back(fShowerTCClusterCol[i][itc]);
  //      }
  
  /*    
        if(pRecoPID==true)shower.SetIsMuon(RecoJMShower::IsMuon(shower));
        if(pRecoANN==1){
          shower.SetANN(0, RecoJMShower::CalcANN(0, shower));
          std::cout<<"pRecoANN, ANN = "<<pRecoANN<<" "<<RecoJMShower::CalcANN(0, shower)<<std::endl;
        }
        //std::cout<<"Testing ... "<<_efrac_2slides<<" "<<_efrac_4slides<<" "<<_efrac_6slides<<" "<<_efrac_2sigmaHOR<<" "<<_efrac_4sigmaHOR<<std::endl;
  */
  
  
        //fill in the rvp variables that could potentially be used
        if(pRecoRVP==true){
          RVPInfo rvpstats = this->GetRVPStats(slicecol, ic);
          shower.SetEfrac_2slides(rvpstats.efrac_2slides);
          shower.SetEfrac_4slides(rvpstats.efrac_4slides);
          shower.SetEfrac_6slides(rvpstats.efrac_6slides);
          shower.SetEfrac_20slides(rvpstats.efrac_20slides);
          shower.SetEfrac_2sigmaHOR(rvpstats.efrac_2sigmaHOR);
          shower.SetEfrac_4sigmaHOR(rvpstats.efrac_4sigmaHOR);
          shower.SetEiso_2sigmaHOR(rvpstats.eiso_2sigmaHOR);
          shower.SetEiso_4sigmaHOR(rvpstats.eiso_4sigmaHOR);
          shower.SetEiso_6sigmaHOR(rvpstats.eiso_6sigmaHOR);
        }
        showercol.push_back(shower);
      }
  
      std::vector<TLorentzVector> showerpGam;// Set shower momentum
      showerpGam.clear();
  
      for(unsigned int i = 0; i < showercol.size(); i++){
        double showereraw = showercol[i].Energy();
        TLorentzVector showerptrk(showercol[i].Dir()[0]*showereraw,showercol[i].Dir()[1]*showereraw,showercol[i].Dir()[2]*showereraw,showereraw);
        showerpGam.push_back(showerptrk);
      }
  
      for(unsigned int i = 0; i < showercol.size(); i++){
        double minpi0mgg = -9999;
        int minj = -1;
        double shexcle = 0;
        for(unsigned int j=0;j< showercol.size();j++){ // pi0 mass calculation
          if(j==i)continue;
          if(showercol[j].ISlice()!=showercol[i].ISlice())continue;
          TLorentzVector p2g = showerpGam[i] + showerpGam[j];
          double mgg = p2g.M();
          if( fabs(minpi0mgg-0.15) > fabs(mgg-0.15)){
            minj = (int)j;
            minpi0mgg = mgg;
          }
          shexcle+=showercol[j].DepositEnergy();
        }
        shexcle+=sliceExclGeV;
        if(minpi0mgg<0)minpi0mgg=0.;
        showercol[i].SetPi0Mgg(minpi0mgg);
        showercol[i].SetPi0G2ID(minj);
        showercol[i].SetSliceEtot(sliceEtot);
        showercol[i].SetExclEnergy(shexcle);
        double elecE = showercol[i].Energy();
        double hadE = showercol[i].ExclEnergy();
  //Nue Energy = electron energy + hadron energy 
        double nuE = (elecE+hadE/(4.00599e-01
          -5.08688e+00*TMath::Exp(-1.07930e+01*hadE-(-1.17127e+01)*pow(hadE,2)-1.01452e+01*pow(hadE,3))                             +4.71995e+00*TMath::Exp(-1.02536e+01*hadE-(-9.81034e+00)*pow(hadE,2)-4.22648e+00*pow(hadE,3))));
  //      double nuE = (showercol[i].Energy() + 0.282525 + 1.0766*(hadE))/1.015;
        showercol[i].SetNueEnergy(nuE);
      }
  
  
  //   TVector3 nuDir0(-0.845e-3, 52.552e-3, 0.998618);
      TVector3 nuDir = geom->NuMIBeamDirection();
      for(unsigned int i = 0; i < showercol.size(); i++){
        double Theta = showerpGam[i].Angle(nuDir);
        showercol[i].SetTheta(Theta);
      }
  
      for(unsigned int i=0;i < showercol.size(); i++){
        if(pRecoPID==true){
          showercol[i].SetIsMuon(RecoJMShower::IsMuon(showercol[i]));
          showercol[i].SetIsInvPhoton(RecoJMShower::IsInvPhoton(showercol[i]));
        }
        if(pRecoANN==1){
          showercol[i].SetANN(0, RecoJMShower::CalcANN(0, showercol[i]));
          showercol[i].SetANN(2, RecoJMShower::CalcANN(2, showercol[i]));
          showercol[i].SetANN(3, RecoJMShower::CalcANN(3, showercol[i]));
          showercol[i].SetANN(4, RecoJMShower::CalcANN(4, showercol[i]));
  //        showercol[i].SetANN(1, RecoJMShower::CalcANN(1, showercol[i]));
  //        std::cout<<"pRecoANN, ANN = "<<pRecoANN<<" "<<RecoJMShower::CalcANN(0, showercol[i])<<std::endl;
        }
      }

      for(unsigned int i = 0;i<showercol.size();i++){
        recoshowercol->push_back(showercol[i]);
        evtshowercol.push_back(showercol[i]);
//ad  a protection for reconstruction fail
        if(i>=fprongidx.size())continue;
        if((unsigned int)fprongidx[i]>=track.size())continue;
        util::CreateAssn(prod,evt, *recoshowercol, track[fprongidx[i]] ,*prongassns);
      }
    }



    return evtshowercol;
  }

  std::vector<jmshower::JMShower> RecoJMShower::ExclShowers(const art::Event& evt){
    return fExcludeShowerCol;
  }

  bool RecoJMShower::PreSel(const art::Event& evt){
    bool tagPreSel = true;
    art::Handle< std::vector<rb::Prong> > trackhandle;
    std::vector<art::Ptr<rb::Prong> > track;
    evt.getByLabel(pTrackDir, trackhandle);

    for(unsigned int i = 0; i < trackhandle->size(); i++){
      art::Ptr<rb::Prong> p(trackhandle, i);
      track.push_back(p);
    }

    for(unsigned int itrk=0;itrk<track.size();itrk++){
      if(track[itrk]->ExtentPlane()>200)tagPreSel = false;  
    }
    return tagPreSel;
  }


//******************************************************
// Geoometry
//******************************************************

  double RecoJMShower::GetPointDoca(TVector3 vtx, TVector3 start, TVector3 stop){
    double d = 9999;
    double denorm = (start-stop).Mag();
    if(denorm<0.00001){d = (vtx-start).Mag();}else{
      d = ((vtx-start).Cross(vtx-stop)).Mag()/denorm;
      TVector3 dir = stop-start;
    }
    return d;
  }


  TVector3 RecoJMShower::GetCellDistToPoint(unsigned int plane, unsigned int cell, TVector3 point){
    Double_t cellXYZ[3]={0,0,0};
    art::ServiceHandle<geo::Geometry> geom;
    geom->Plane(plane)->Cell(cell)->GetCenter(cellXYZ);
    geo::View_t view = geom->Plane(plane)->View();
    double celldistx = -99;
    double celldisty = -99;
    double celldistz = -99;
    if(view == geo::kX){
      celldistx = fabs(cellXYZ[0]-point[0]);
      celldisty = 0; 
      celldistz = fabs(cellXYZ[2]-point[2]);
    }else if(view == geo::kY){
      celldistx = 0;
      celldisty = fabs(cellXYZ[1]-point[1]);
      celldistz = fabs(cellXYZ[2]-point[2]);
    }
    TVector3 celldist(celldistx,celldisty,celldistz);
    return celldist;
  }

  //******************************************************************************************************************
  double RecoJMShower::GetCellDistToTrk(unsigned int plane, unsigned int cell, TVector3 start, TVector3 stop){
    Double_t cellXYZ[3]={0,0,0};
    art::ServiceHandle<geo::Geometry> geom;
    geom->Plane(plane)->Cell(cell)->GetCenter(cellXYZ);
    geo::View_t view = geom->Plane(plane)->View();

    double x1 = start[0];
    double y1 = start[1];
    double z1 = start[2];
    double p1 = (stop[0] - start[0])/(stop - start).Mag();
    double q1 = (stop[1] - start[1])/(stop - start).Mag();
    double r1 = (stop[2] - start[2])/(stop - start).Mag();

    double x2 = cellXYZ[0];
    double y2 = cellXYZ[1];
    double z2 = cellXYZ[2];

    double p2 = 1;
    double q2 = 0;
    double r2 = 0;
    if(view == geo::kX){
      p2 = 0;
      q2 = 1;
      r2 = 0;
    }else if(view == geo::kY){
      p2 = 1;
      q2 = 0;
      r2 = 0;
    }
    double ma = x2-x1;
    double mb = y2-y1;
    double mc = z2-z1;
    double md = p1;
    double me = q1;
    double mf = r1;
    double mg = p2;
    double mh = q2;
    double mi = r2;

    TMatrixD dDeNorm1(2,2);
    TMatrixD dDeNorm2(2,2);
    TMatrixD dDeNorm3(2,2);
    dDeNorm1[0][0] = p1;
    dDeNorm1[0][1] = q1;
    dDeNorm1[1][0] = p2;
    dDeNorm1[1][1] = q2;
    double det1 = dDeNorm1[0][0]*dDeNorm1[1][1] - dDeNorm1[0][1]*dDeNorm1[1][0];
    dDeNorm2[0][0] = q1;
    dDeNorm2[0][1] = r1;
    dDeNorm2[1][0] = q2;
    dDeNorm2[1][1] = r2;
    double det2 = dDeNorm2[0][0]*dDeNorm2[1][1] - dDeNorm2[0][1]*dDeNorm2[1][0];
    dDeNorm3[0][0] = r1;
    dDeNorm3[0][1] = p1;
    dDeNorm3[1][0] = r2;
    dDeNorm3[1][1] = p2;
    double det3 = dDeNorm3[0][0]*dDeNorm3[1][1] - dDeNorm3[0][1]*dDeNorm3[1][0];
    double det0 = ma*me*mi + mb*mf*mg + mc*md*mh - mc*me*mg - mb*md*mi - ma*mf*mh;
    double celldist = fabs(det0)/sqrt(det1*det1+det2*det2+det3*det3);
    return celldist;
  }

//****************************************************************************************************************


  TVector3 RecoJMShower::GetShwStop(rb::Prong shw){
    TVector3 shwdir=shw.Dir();
    TVector3 shwstart=shw.Start();
    TVector3 shwend  = shwstart + shw.TotalLength()*shwdir;
    return shwend;
  }

  TVector3 RecoJMShower::GetShwStop(art::Ptr<rb::Prong> shw){
    TVector3 shwdir=shw->Dir();
    TVector3 shwstart=shw->Start();
    TVector3 shwend  = shwstart + shw->TotalLength()*shwdir;
    return shwend;
  }


//given a plane, cell and track, return distance of cell with respect to track
  TVector3 RecoJMShower::GetTrkHitPos(unsigned int plane, unsigned int cell, art::Ptr<rb::Prong> trk){
    Double_t cellXYZ[3]={0,0,0}; //declare container to hold cell coordinate
    art::ServiceHandle<geo::Geometry> geom;
    geom->Plane(plane)->Cell(cell)->GetCenter(cellXYZ); //get cell center

    //get track direction, start, end
    TVector3 trkdir=trk->Dir();
    TVector3 trkstart=trk->Start();
    TVector3 trkstop=RecoJMShower::GetShwStop(trk);
//    double cellL=2*geom->Plane(plane)->Cell(cell)->HalfL();
    geo::View_t view = geom->Plane(plane)->View();
    double celldist = 0;
    double trkx = 0;
    double trky = 0;
    //calculate the tracks x position at the cell's z coordinate
    trkx = trkstart.x()+(trkdir.x()/trkdir.z())*(cellXYZ[2]-trkstart.z());
    //calculate the tracks y position at the cell's z coordinate
    trky = trkstart.y()+(trkdir.y()/trkdir.z())*(cellXYZ[2]-trkstart.z());
//    double trkdtr0 = 0;
    double trkdtr = 0;
    if(view == geo::kX){
      //calculate distance from cell x coordinate to track x coordninate at same z position
      //NOTE: Perpendicular distance to track may be a better metric as this will be incorrect for steep tracks
      celldist=fabs((cellXYZ[0]-trkstart.x())-((cellXYZ[2]-trkstart.z())*(trkdir.x()/trkdir.z())));
//      trkdtr0 = geom->Plane(plane)->Cell(cell)->DistToReadOut(trky); //distance to readout for cell
      trkdtr = geom->DistToElectronics(trky, *geom->Plane(plane)->Cell(cell));

    }else if(view == geo::kY){
      //calculate distance from cell x coordinate to track x coordninate at same z position
      celldist=fabs((cellXYZ[1]-trkstart.y())-((cellXYZ[2]-trkstart.z())*(trkdir.y()/trkdir.z())));
//      trkdtr0 = geom->Plane(plane)->Cell(cell)->DistToReadOut(trkx); //distance to readout for cell
      trkdtr = geom->DistToElectronics(trkx, *geom->Plane(plane)->Cell(cell));
    }
    if(trkdtr<0)trkdtr=0;
    //return results
    TVector3 def(-1000,-1000,-1000);
    TVector3 trkhitposxcell(trkdtr,trky,celldist);
    TVector3 trkhitposycell(trkdtr,trkx,celldist);
    if(view == geo::kX){return trkhitposxcell;}
    if(view == geo::kY){return trkhitposycell;}
    return def;
  }

//*********************************************************************************************************************
  TVector3 RecoJMShower::GetTrkHitPos(unsigned int plane, unsigned int cell, rb::Prong trk){
    Double_t cellXYZ[3]={0,0,0}; 
    art::ServiceHandle<geo::Geometry> geom;
    geom->Plane(plane)->Cell(cell)->GetCenter(cellXYZ);
    TVector3 trkdir=trk.Dir();
    TVector3 trkstart=trk.Start();
    TVector3 trkstop=RecoJMShower::GetShwStop(trk);
//    double cellL=2*geom->Plane(plane)->Cell(cell)->HalfL();
    geo::View_t view = geom->Plane(plane)->View();
    double celldist = 0;
    double trkx = 0;
    double trky = 0;
    trkx = trkstart.x()+(trkdir.x()/trkdir.z())*(cellXYZ[2]-trkstart.z());
    trky = trkstart.y()+(trkdir.y()/trkdir.z())*(cellXYZ[2]-trkstart.z());

//    double trkdtr0 = 0;
    double trkdtr = 0;
    if(view == geo::kX){
      celldist=fabs((cellXYZ[0]-trkstart.x())-((cellXYZ[2]-trkstart.z())*(trkdir.x()/trkdir.z())));
//      trkdtr = geom->Plane(plane)->Cell(cell)->DistToReadOut(trky);
      trkdtr = geom->DistToElectronics(trky, *geom->Plane(plane)->Cell(cell));
    }else if(view == geo::kY){
      celldist=fabs((cellXYZ[1]-trkstart.y())-((cellXYZ[2]-trkstart.z())*(trkdir.y()/trkdir.z())));
//      trkdtr = geom->Plane(plane)->Cell(cell)->DistToReadOut(trkx);
      trkdtr = geom->DistToElectronics(trkx, *geom->Plane(plane)->Cell(cell));
    }
    if(trkdtr<0)trkdtr=0;
    TVector3 def(-1000,-1000,-1000);
    TVector3 trkhitposxcell(trkdtr,trky,celldist);
    TVector3 trkhitposycell(trkdtr,trkx,celldist);
    if(view == geo::kX){return trkhitposxcell;}
    if(view == geo::kY){return trkhitposycell;}
    return def;
  }

  TVector3 RecoJMShower::GetTrkHitPos(unsigned int plane, unsigned int cell, TVector3 trkdir, TVector3 trkstart, TVector3 trkstop){
    Double_t cellXYZ[3]={0,0,0}; // XYZ[0] = x, XYZ[1] = y, XYZ[2] = z
    art::ServiceHandle<geo::Geometry> geom;
    geom->Plane(plane)->Cell(cell)->GetCenter(cellXYZ);
//    double cellL=2*geom->Plane(plane)->Cell(cell)->HalfL();
    geo::View_t view = geom->Plane(plane)->View();
    double celldist = 0;
    double trkx = 0;
    double trky = 0;
    trkx = trkstart.x()+(trkdir.x()/trkdir.z())*(cellXYZ[2]-trkstart.z());
    trky = trkstart.y()+(trkdir.y()/trkdir.z())*(cellXYZ[2]-trkstart.z());
//    double trkdtr0 = 0;
    double trkdtr = 0;
    if(view == geo::kX){
      celldist=fabs((cellXYZ[0]-trkstart.x())-((cellXYZ[2]-trkstart.z())*(trkdir.x()/trkdir.z())));
//      trkdtr0 = geom->Plane(plane)->Cell(cell)->DistToReadOut(trky);
      trkdtr = geom->DistToElectronics(trky, *geom->Plane(plane)->Cell(cell));
    }else if(view == geo::kY){
      celldist=fabs((cellXYZ[1]-trkstart.y())-((cellXYZ[2]-trkstart.z())*(trkdir.y()/trkdir.z())));
//      trkdtr0 = geom->Plane(plane)->Cell(cell)->DistToReadOut(trkx);
      trkdtr = geom->DistToElectronics(trkx, *geom->Plane(plane)->Cell(cell));
    }
    if(trkdtr<0)trkdtr=0;
    TVector3 def(-1000,-1000,-1000);
    TVector3 trkhitposxcell(trkdtr,trky,celldist);
    TVector3 trkhitposycell(trkdtr,trkx,celldist);
    if(view == geo::kX){return trkhitposxcell;}
    if(view == geo::kY){return trkhitposycell;}
    return def;
  }


  TVector3 RecoJMShower::GetTrkPlanePos(unsigned int plane, rb::Prong trk){
    Double_t cellXYZ[3]={0,0,0};
    art::ServiceHandle<geo::Geometry> geom;
    geom->Plane(plane)->Cell(0)->GetCenter(cellXYZ);
    TVector3 trkdir=trk.Dir();
    TVector3 trkstart=trk.Start();
    TVector3 trkstop=RecoJMShower::GetShwStop(trk);
    double trkplanex=99999,trkplaney=99999, trkplanez=0;
    if(fabs(trkdir.z())>0.0001){
      trkplanex = trkstart.x()+(cellXYZ[2]-trkstart.z())*(trkdir.x()/trkdir.z());
      trkplaney = trkstart.y()+(cellXYZ[2]-trkstart.z())*(trkdir.y()/trkdir.z());
    }else{
      trkplanex = (trkstart.x()+trkstop.x())/2.0; 
      trkplaney = (trkstart.y()+trkstop.y())/2.0;
    }
    trkplanez = cellXYZ[2];
    TVector3 trkplanepos(trkplanex, trkplaney, trkplanez);
    return trkplanepos;;
  }

/*
  unsigned int RecoJMShower::GetTrkPlaneCell(unsigned int plane, art::Ptr<rb::Prong> trk){
    Double_t cellXYZ[3]={0,0,0};
    art::ServiceHandle<geo::Geometry> geom;
    geom->Plane(plane)->Cell(0)->GetCenter(cellXYZ);
    double cellW = 2*geom->Plane(plane)->Cell(0)->HalfW();
    TVector3 trkdir=trk->Dir();
    TVector3 trkstart=trk->Start();
    TVector3 trkstop=RecoJMShower::GetShwStop(trk);
    double trkplanex=99999,trkplaney=99999, trkplanez=0;
    if(fabs(trkdir.z())>0.0001){
      trkplanex = trkstart.x()+(cellXYZ[2]-trkstart.z())*(trkdir.x()/trkdir.z());
      trkplaney = trkstart.y()+(cellXYZ[2]-trkstart.z())*(trkdir.y()/trkdir.z());
    }
    trkplanez = cellXYZ[2];
    geo::View_t view = geom->Plane(plane)->View();
    int trkcellId = -9999;
    if(view == geo::kX)trkcellId = int((trkplanex+cellW/2-cellXYZ[0])/cellW);
    if(view == geo::kY)trkcellId = int((trkplaney+cellW/2-cellXYZ[1])/cellW);
    return unsigned(trkcellId);
  }
*/


  unsigned int RecoJMShower::GetTrkPlaneCell(unsigned int plane, art::Ptr<rb::Prong> trk){
    Double_t cellXYZ[3]={0,0,0};
    Double_t cellXYZ1[3];
    Double_t cellXYZmax[3];
    art::ServiceHandle<geo::Geometry> geom;
    geom->Plane(plane)->Cell(0)->GetCenter(cellXYZ);
    geom->Plane(plane)->Cell(geom->Plane(plane)->Ncells()-1)->GetCenter(cellXYZ1);
    geom->Plane(plane)->Cell(geom->Plane(plane)->Ncells()-1)->GetCenter(cellXYZmax);
    double cellW = 2*geom->Plane(plane)->Cell(0)->HalfW();
    double cellwx = (cellXYZ1[0] - cellXYZ[0] + cellW)/geom->Plane(plane)->Ncells();
    double cellwy = (cellXYZ1[1] - cellXYZ[1] + cellW)/geom->Plane(plane)->Ncells();
    TVector3 trkdir=trk->Dir();
    TVector3 trkstart=trk->Start();
    TVector3 trkstop=RecoJMShower::GetShwStop(trk);
    double trkplanex=99999,trkplaney=99999;//, trkplanez=0;
//    double trkplanex0=99999,trkplaney0=99999;
//    double trkplanex1=99999,trkplaney1=99999;
    if(fabs(trkdir.z())>0.0001){
      trkplanex = trkstart.x()+(cellXYZ[2]-trkstart.z())*(trkdir.x()/trkdir.z());
      trkplaney = trkstart.y()+(cellXYZ[2]-trkstart.z())*(trkdir.y()/trkdir.z());
//      trkplanex0 = cellXYZ[0];
//      trkplaney0 = cellXYZ[1];
//      trkplanex1 = cellXYZmax[0];
//      trkplaney1 = cellXYZmax[1];
    }else{
      trkplanex = (trkstart.x()+ trkstop.x())/2.0;
      trkplaney = (trkstart.y()+ trkstop.y())/2.0;
    }
//    trkplanez = cellXYZ[2];
    geo::View_t view = geom->Plane(plane)->View();

/*
    if(view == geo::kX){
     std::cout<<"x view plane"<<std::endl;
    std::cout<<"x1-x0,cellW "<<" "<<cellXYZ1[0] - cellXYZ[0]<<" "<<cellW<<std::endl;
    }else if(view == geo::kY){
     std::cout<<"y view plane"<<std::endl;
    std::cout<<"y1-y0,cellW "<<" "<<cellXYZ1[1] - cellXYZ[1]<<" "<<cellW<<std::endl;
    }
*/

    int trkcellId = -9999;
//    int trkcellId0 = -9999;
//    int trkcellId1 = -9999;
//    int trkcellIdmax = geom->Plane(plane)->Ncells()-1;

    if(view == geo::kX)trkcellId = int((trkplanex+cellwx/2-cellXYZ[0])/cellwx);
    if(view == geo::kY)trkcellId = int((trkplaney+cellwy/2-cellXYZ[1])/cellwy);

//    if(view == geo::kX)trkcellId0 = int((trkplanex0+cellwx/2-cellXYZ[0])/cellwx);
//    if(view == geo::kY)trkcellId0 = int((trkplaney0+cellwy/2-cellXYZ[1])/cellwy);
//
//    if(view == geo::kX)trkcellId1 = int((trkplanex1+cellwx/2-cellXYZ[0])/cellwx);
//    if(view == geo::kY)trkcellId1 = int((trkplaney1+cellwy/2-cellXYZ[1])/cellwy);
    return unsigned(trkcellId);
  }

  unsigned int RecoJMShower::GetTrkPlaneCell(unsigned int plane, rb::Prong trk){
    Double_t cellXYZ[3]={0,0,0};
    Double_t cellXYZ1[3];
    Double_t cellXYZmax[3];
    art::ServiceHandle<geo::Geometry> geom;
    geom->Plane(plane)->Cell(0)->GetCenter(cellXYZ);
    geom->Plane(plane)->Cell(geom->Plane(plane)->Ncells()-1)->GetCenter(cellXYZ1);
    geom->Plane(plane)->Cell(geom->Plane(plane)->Ncells()-1)->GetCenter(cellXYZmax);
    double cellW = 2*geom->Plane(plane)->Cell(0)->HalfW();
    double cellwx = (cellXYZ1[0] - cellXYZ[0] + cellW)/geom->Plane(plane)->Ncells();
    double cellwy = (cellXYZ1[1] - cellXYZ[1] + cellW)/geom->Plane(plane)->Ncells();
    TVector3 trkdir=trk.Dir();
    TVector3 trkstart=trk.Start();
    TVector3 trkstop=RecoJMShower::GetShwStop(trk);
    double trkplanex=99999,trkplaney=99999;//, trkplanez=0;
//    double trkplanex0=99999,trkplaney0=99999;
//    double trkplanex1=99999,trkplaney1=99999;
    if(fabs(trkdir.z())>0.0001){
      trkplanex = trkstart.x()+(cellXYZ[2]-trkstart.z())*(trkdir.x()/trkdir.z());
      trkplaney = trkstart.y()+(cellXYZ[2]-trkstart.z())*(trkdir.y()/trkdir.z());
//      trkplanex0 = cellXYZ[0];
//      trkplaney0 = cellXYZ[1];
//      trkplanex1 = cellXYZmax[0];
//      trkplaney1 = cellXYZmax[1];
    }else{
      trkplanex = (trkstart.x()+trkstop.x())/2.0;
      trkplaney = (trkstart.y()+trkstop.y())/2.0;
    }
//    trkplanez = cellXYZ[2];
    geo::View_t view = geom->Plane(plane)->View();

/*
    if(view == geo::kX){
     std::cout<<"x view plane"<<std::endl;
    std::cout<<"x1-x0,cellW "<<" "<<cellXYZ1[0] - cellXYZ[0]<<" "<<cellW<<std::endl;
    }else if(view == geo::kY){
     std::cout<<"y view plane"<<std::endl;
    std::cout<<"y1-y0,cellW "<<" "<<cellXYZ1[1] - cellXYZ[1]<<" "<<cellW<<std::endl;
    }
*/

    int trkcellId = -9999;
//    int trkcellId0 = -9999;
//    int trkcellId1 = -9999;
//    int trkcellIdmax = geom->Plane(plane)->Ncells()-1;

    if(view == geo::kX)trkcellId = int((trkplanex+cellwx/2-cellXYZ[0])/cellwx);
    if(view == geo::kY)trkcellId = int((trkplaney+cellwy/2-cellXYZ[1])/cellwy);

//    if(view == geo::kX)trkcellId0 = int((trkplanex0+cellwx/2-cellXYZ[0])/cellwx);
//    if(view == geo::kY)trkcellId0 = int((trkplaney0+cellwy/2-cellXYZ[1])/cellwy);
//
//    if(view == geo::kX)trkcellId1 = int((trkplanex1+cellwx/2-cellXYZ[0])/cellwx);
//    if(view == geo::kY)trkcellId1 = int((trkplaney1+cellwy/2-cellXYZ[1])/cellwy);
    return unsigned(trkcellId);
  }

  unsigned int RecoJMShower::GetTrkCPlaneCell(unsigned int plane, unsigned int i){//calculate centroid of a plane
    art::ServiceHandle<geo::Geometry> geom;
    double sumid = 0;
    double gev = 0;
    for (unsigned int nh=0;nh<fShowerPClusterCol[i][plane].NCell();nh++) {
      art::Ptr<rb::CellHit>  cellhit = fShowerPClusterCol[i][plane].Cell(nh);
      geo::OfflineChan key(cellhit->Plane(),cellhit->Cell());
      double cellEnergy = fHitEWeightMap[key][i];
      double cellPos = (double)cellhit->Cell() + 0.5;
      sumid += cellEnergy*cellPos;
      gev += cellEnergy;
     }
    if(gev<0.00001)return 9999; 
    double avgid = sumid/gev;
    int cid = (int)avgid;
    if(cid<0)cid=0;
    if(cid>(int)geom->Plane(plane)->Ncells()-1)cid=(int)geom->Plane(plane)->Ncells()-1;
    return (unsigned int)cid;
  }

  TVector3 RecoJMShower::GetTrkPlaneDistToEdge(unsigned int plane, art::Ptr<rb::Prong> trk){
    Double_t cellXYZ[3]={0,0,0};
    art::ServiceHandle<geo::Geometry> geom;
    geom->Plane(plane)->Cell(0)->GetCenter(cellXYZ);
    TVector3 trkdir=trk->Dir();
    TVector3 trkstart=trk->Start();
    TVector3 trkstop=RecoJMShower::GetShwStop(trk);
    double trkplanex=99999,trkplaney=99999, trkplanez=0;
    if(fabs(trkdir.z())>0.0001){
      trkplanex = trkstart.x()+(cellXYZ[2]-trkstart.z())*(trkdir.x()/trkdir.z());
      trkplaney = trkstart.y()+(cellXYZ[2]-trkstart.z())*(trkdir.y()/trkdir.z());
    }
    trkplanez = cellXYZ[2];

    double xe[4]={9999,9999,9999,9999};
    double ye[4]={9999,9999,9999,9999};
    double ze[4]={9999,9999,9999,9999};

    xe[0] = fabs(pXEdge1-trkplanex);
    xe[1] = fabs(pXEdge2-trkplanex);
    xe[2] = fabs(pXEdge3-trkplanex);
    xe[3] = fabs(pXEdge4-trkplanex);

    ye[0] = fabs(pYEdge1-trkplaney);
    ye[1] = fabs(pYEdge2-trkplaney);
    ye[2] = fabs(pYEdge3-trkplaney);
    ye[3] = fabs(pYEdge4-trkplaney);

    ze[0] = fabs(pZEdge1-trkplanez);
    ze[1] = fabs(pZEdge2-trkplanez);
    ze[2] = fabs(pZEdge3-trkplanez);
    ze[3] = fabs(pZEdge4-trkplanez);

    double xtoedge = 9999.; 
    double ytoedge = 9999.; 
    double ztoedge = 9999.; 

    for(int ii=0;ii<4;ii++){
      if(xtoedge>xe[ii])xtoedge=xe[ii];
      if(ytoedge>ye[ii])ytoedge=ye[ii];
      if(ztoedge>ze[ii])ztoedge=ze[ii];
    }

//    double xtoedge0 = fabs(geom->DetHalfWidth()-fabs(trkplanex));
//    double ytoedge0 = fabs(geom->DetHalfHeight()-fabs(trkplaney));
//    double ztoedge0 = std::min(fabs(geom->DetLength()-fabs(trkplanez)),trkplanez);
//   std::cout<<"Det W/2,H/2,L "<<geom->DetHalfWidth()<<" "<<geom->DetHalfHeight()<<" "<<geom->DetLength()<<std::endl;
//    double cellL = 2*geom->Plane(plane)->Cell(0)->HalfL();
//    double cellW = 2*geom->Plane(plane)->Cell(0)->HalfW();
//    double cellD = 2*geom->Plane(plane)->Cell(0)->HalfD();
//    geo::View_t view = geom->Plane(plane)->View();
/*
    if(view == geo::kX){
     std::cout<<"x view plane"<<std::endl;
    }else if(view == geo::kY){
     std::cout<<"y view plane"<<std::endl;
    }
   std::cout<<"cell L, cellW, cell D "<<cellL<<" "<<cellW<<" "<<cellD<<std::endl;
*/
//   std::cout<<"To Edge X,Y,Z "<<xtoedge<<" "<<ytoedge<<" "<<ztoedge<<std::endl;
    TVector3 toedge(xtoedge, ytoedge, ztoedge);
    return toedge;
  }


  TVector3 RecoJMShower::GetTrkPlaneDistToEdge(unsigned int plane, rb::Prong trk){
    Double_t cellXYZ[3]={0,0,0};
    art::ServiceHandle<geo::Geometry> geom;
    geom->Plane(plane)->Cell(0)->GetCenter(cellXYZ);
    TVector3 trkdir=trk.Dir();
    TVector3 trkstart=trk.Start();
    TVector3 trkstop=RecoJMShower::GetShwStop(trk);
    double trkplanex=99999,trkplaney=99999, trkplanez=0;
    if(fabs(trkdir.z())>0.0001){
      trkplanex = trkstart.x()+(cellXYZ[2]-trkstart.z())*(trkdir.x()/trkdir.z());
      trkplaney = trkstart.y()+(cellXYZ[2]-trkstart.z())*(trkdir.y()/trkdir.z());
    }
    trkplanez = cellXYZ[2];

    double xe[4]={9999,9999,9999,9999};
    double ye[4]={9999,9999,9999,9999};
    double ze[4]={9999,9999,9999,9999};

    xe[0] = fabs(pXEdge1-trkplanex);
    xe[1] = fabs(pXEdge2-trkplanex);
    xe[2] = fabs(pXEdge3-trkplanex);
    xe[3] = fabs(pXEdge4-trkplanex);

    ye[0] = fabs(pYEdge1-trkplaney);
    ye[1] = fabs(pYEdge2-trkplaney);
    ye[2] = fabs(pYEdge3-trkplaney);
    ye[3] = fabs(pYEdge4-trkplaney);

    ze[0] = fabs(pZEdge1-trkplanez);
    ze[1] = fabs(pZEdge2-trkplanez);
    ze[2] = fabs(pZEdge3-trkplanez);
    ze[3] = fabs(pZEdge4-trkplanez);

    double xtoedge = 9999.; 
    double ytoedge = 9999.; 
    double ztoedge = 9999.; 

    for(int ii=0;ii<4;ii++){
      if(xtoedge>xe[ii])xtoedge=xe[ii];
      if(ytoedge>ye[ii])ytoedge=ye[ii];
      if(ztoedge>ze[ii])ztoedge=ze[ii];
    }

//    double xtoedge0 = fabs(geom->DetHalfWidth()-fabs(trkplanex));
//    double ytoedge0 = fabs(geom->DetHalfHeight()-fabs(trkplaney));
//    double ztoedge0 = std::min(fabs(geom->DetLength()-fabs(trkplanez)),trkplanez);

//    double xtoedge = geom->DetHalfWidth()-fabs(trkplanex);
//    double ytoedge = geom->DetHalfHeight()-fabs(trkplaney);
//    double ztoedge = geom->DetLength()-fabs(trkplanez);
//   std::cout<<"Det W/2,H/2,L "<<geom->DetHalfWidth()<<" "<<geom->DetHalfHeight()<<" "<<geom->DetLength()<<std::endl;
//    double cellL = 2*geom->Plane(plane)->Cell(0)->HalfL();
//    double cellW = 2*geom->Plane(plane)->Cell(0)->HalfW();
//    double cellD = 2*geom->Plane(plane)->Cell(0)->HalfD();
//    geo::View_t view = geom->Plane(plane)->View();
/*
    if(view == geo::kX){
     std::cout<<"x view plane"<<std::endl;
    }else if(view == geo::kY){
     std::cout<<"y view plane"<<std::endl;
    }
   std::cout<<"cell L, cellW, cell D "<<cellL<<" "<<cellW<<" "<<cellD<<std::endl;
*/
    TVector3 toedge(xtoedge, ytoedge, ztoedge);
    return toedge;
  }


  bool RecoJMShower::IsFiducial(rb::Prong shower){
    int i = shower.ID();
    bool isfiducial = true;
    for(unsigned int plane=0;plane<fShowerPClusterCol[i].size();plane++){
      double planerad = RecoJMShower::Radius(fShowerPClusterCol[i][plane],shower);
      TVector3 disttoedge = RecoJMShower::GetTrkPlaneDistToEdge(fShowerPClusterCol[i][plane].Plane(), shower);
      if(disttoedge.x()>2*planerad&&disttoedge.y()>2*planerad&&disttoedge.z()>2*planerad)continue;
      isfiducial = false;
    }
    return isfiducial;
  }

  double RecoJMShower::GetTrkDistToEdge(rb::Prong shower){
    int i = shower.ID();
    double mindist = 999999.; 
    for(unsigned int plane=0;plane<fShowerPClusterCol[i].size();plane++){
      TVector3 disttoedge = RecoJMShower::GetTrkPlaneDistToEdge(fShowerPClusterCol[i][plane].Plane(), shower);
      double minxyz=999999.;
      for(int ii=0;ii<3;ii++){
        if(fabs(disttoedge[ii])<minxyz)minxyz=fabs(disttoedge[ii]);
      }
      if(minxyz<mindist)mindist=minxyz;
    }
    return mindist;
  }


  double RecoJMShower::GetTrkHitPath(unsigned int plane, unsigned int cell, rb::Prong shower){
    int i = shower.ID();
    Double_t cellXYZ[3]={0,0,0}; // xyz[0] = x, xyz[1] = y, xyz[2] = z
    art::ServiceHandle<geo::Geometry> geom;
    geom->Plane(plane)->Cell(cell)->GetCenter(cellXYZ);
    TVector3 trkdir=shower.Dir();
    TVector3 trkstart=shower.Start();
    TVector3 trkstop=RecoJMShower::GetShwStop(shower);
    double trklength = fabs((trkstart - trkstop).Mag());
    double cellD=2*geom->Plane(plane)->Cell(cell)->HalfD();
    double cellL=2*geom->Plane(plane)->Cell(cell)->HalfL();
    if(cellL<cellD||cellL<0.0001){
      std::cout<<"Geometry error! "<<std::endl; 
      return 9999;
    }
    if(fabs(trkdir.z())< 1e-6||fShowerPClusterCol[i].size()==1){
//      std::cout<<"Small Z Dir! "<<std::endl;
//      std::cout<<"N Plane = "<<shower.ExtentPlane()<<std::endl;
//      std::cout<<"Path    = "<<trklength<<std::endl;
      return trklength;
    }
    double trkhitpath=cellD*(sqrt(trkdir.x()*trkdir.x()+trkdir.y()*trkdir.y()+trkdir.z()*trkdir.z())/fabs(trkdir.z()));
    return trkhitpath;
  }


  unsigned int RecoJMShower::ContStartPlane(rb::Prong shower, unsigned int startp, int contp){
    int i = shower.ID();
    if(int(fShowerPClusterCol[i].size()-startp)<contp)return startp;
    unsigned int newstartp = startp;
    for(unsigned int plane= startp; plane<fShowerPClusterCol[i].size();plane++){
      int starttag = 1;
      for(int ic=0;ic<contp;ic++){
        if(plane+ic>fShowerPClusterCol[i].size()-1){
          starttag=0;
          break;
        }
//        double path = RecoJMShower::GetTrkHitPath(fShowerPClusterCol[i][plane+ic].Plane(), 0, shower);
//        double ep = RecoJMShower::Energy(fShowerPClusterCol[i][plane+ic],shower.ID());
//        double dedx = ep/path;
        double dedx = RecoJMShower::GetPlaneDedx(shower, plane+ic);
        if(dedx<0.0005){
          starttag=0;
          break;
        }
      }
      if(starttag==1){
        newstartp = plane;
        break;
      }
    }
    return newstartp;
  } 



//******************************************************
//   Energy 
//******************************************************

  double RecoJMShower::DepositEnergy(rb::Cluster cluster, int i){
    double gev = 0;
    for (unsigned int nh=0;nh<cluster.NCell();nh++) {
      art::Ptr<rb::CellHit>  cellhit = cluster.Cell(nh);
      geo::OfflineChan key(cellhit->Plane(),cellhit->Cell());
      double cellEnergy = fHitEWeightMap[key][i];
      gev += cellEnergy;
    }
    return gev;
  }

  double RecoJMShower::DepositEnergy(rb::Prong shower){
    double gev = 0;
    int i = shower.ID();
    for (unsigned int nh=0;nh<shower.NCell();nh++) {
      art::Ptr<rb::CellHit>  cellhit = shower.Cell(nh);
      geo::OfflineChan key(cellhit->Plane(),cellhit->Cell());
      double cellEnergy = fHitEWeightMap[key][i];
      gev += cellEnergy;
    }
    return gev;
  }

  double RecoJMShower::Energy(rb::Cluster cluster, int i){
    double gev = 0;
    for (unsigned int nh=0;nh<cluster.NCell();nh++) {
      art::Ptr<rb::CellHit>  cellhit = cluster.Cell(nh);
      geo::OfflineChan key(cellhit->Plane(),cellhit->Cell());
      double cellEnergy = fHitEWeightMap[key][i];
      gev += cellEnergy;
    }
    double ecor = RecoJMShower::ECorrMC(gev);
    gev = gev/ecor;
    return gev;
  }

  double RecoJMShower::Energy(rb::Prong shower){
    double gev = 0;
    int i = shower.ID();
    for (unsigned int nh=0;nh<shower.NCell();nh++) {
      art::Ptr<rb::CellHit>  cellhit = shower.Cell(nh);
      geo::OfflineChan key(cellhit->Plane(),cellhit->Cell());
      double cellEnergy = fHitEWeightMap[key][i];
      gev += cellEnergy;
    }
    double ecor = RecoJMShower::ECorrMC(gev);
    double avgx = (shower.Start()[0]+RecoJMShower::GetShwStop(shower)[0])/2.; 
    double avgy = (shower.Start()[1]+RecoJMShower::GetShwStop(shower)[1])/2.; 
//    double poscor = (9.9899e-01+7.46983e-05*avgx+7.34449e-05*avgy);//S12-06-17 
//    double poscor=1.+8.18868e-05*avgx+8.20582e-05*avgy;//without pigtail 
    double poscor=1.+6.28813e-05*avgx+6.15165e-05*avgy; //with pigtail
    if(pCorAbsCell==false) poscor = 1.0+1.04866e-04*avgx+6.40668e-05*avgy;
    gev = gev/(ecor*poscor);
    //std::cout<<"JMShower Energy calc, depE: "<<showerDepE<<" ecor "<<ecor<<" avgx "<<avgx<<" avgy "<<avgy<<" poscor "<<poscor<<" showerE "<<gev<<std::endl;
//    gev = gev/ecor;
    return gev;
  }

  double RecoJMShower::ECorrMC(double gev){//EM shower energy correction
//    double ecor = 5.63616e-01;    //without pigtail
    double ecor = 5.68142e-01;    //with pigtail
    if(pCorAbsCell==false)ecor = 5.62343e-01;  //no mc cor
//    double ecor = 9.05891e-01*(5.53444e-01+9.92889e-02*gev-4.30034e-02*gev*gev-5.55264e-01*exp(-1.61264e+00*gev-1.07062e+01*gev*gev));
    return ecor;
  }

  double RecoJMShower::CellADCToGeV(double d, geo::View_t view, int detid){
    double Att = 4.08050e+03;
    if(detid==novadaq::cnv::kFARDET){
      if(view==geo::kX){
        Att = 4.08574e+03+3.49097e+04*TMath::Exp(-2.67297e-03*d)
             -1.62442e+04*TMath::Exp(-3.14363e-02*d-(-3.93281e-05)*d*d-1.49793e-08*d*d*d); 
      }else if(view == geo::kY){
        Att = 4.09347e+03+3.50384e+04*TMath::Exp(-2.68025e-03*d)
             -1.74280e+04*TMath::Exp(-3.08694e-02*d-(-3.86870e-05)*d*d-1.48737e-08*d*d*d);
      }
    }else{
        if(view==geo::kX){
          Att = 4.08574e+03+3.49097e+04*TMath::Exp(-2.67297e-03*d)
               -1.62442e+04*TMath::Exp(-3.14363e-02*d-(-3.93281e-05)*d*d-1.49793e-08*d*d*d); 
        }else if(view == geo::kY){
          Att = 4.09347e+03+3.50384e+04*TMath::Exp(-2.68025e-03*d)
               -1.74280e+04*TMath::Exp(-3.08694e-02*d-(-3.86870e-05)*d*d-1.48737e-08*d*d*d);
        }
    }
    return Att;
  }


  double RecoJMShower::CellEDataMC(double gev, double w, geo::View_t view, int detid){//
    double ecor = 1;    //S12-11-16
    if(gev>0&&w>0){
      if(detid==novadaq::cnv::kFARDET){
        if(view == geo::kX){
          ecor =  1.60823+0.000519042*w-2.47044e-07*w*w; 
        }else if(view == geo::kY){
          ecor = 1.3538+0.00051901*w-1.86311e-07*w*w;
        }
      }else{
        if(view == geo::kX){
           ecor =  1.60823+0.000519042*w-2.47044e-07*w*w;
        }else if(view == geo::kY){
           ecor = 1.3538+0.00051901*w-1.86311e-07*w*w;
        }
      }
    }
    if(ecor<0.0001)ecor = 1.0;
    return ecor;
  }

  double RecoJMShower::PlaneDedxDataMC(double gev, int detid){//
    double ecor = 1;    
    if(gev>0){
      if(detid==novadaq::cnv::kFARDET){
        ecor = 1.0/1.23933;
      }else{
        ecor = 1.0/1.23933;
      }
    }
    if(ecor<0.0001)ecor = 1.0;
    return ecor;
  }

  double RecoJMShower::CellTransDataMC(double gev, int detid){//
    double ecor = 1;    
    if(gev>0){
      if(detid==novadaq::cnv::kFARDET){
        ecor = 1.0/1.336;
      }else{
        ecor = 1.0/1.336;
      }
    }
    if(ecor<0.0001)ecor = 1.0;
    return ecor;
  }

  double RecoJMShower::Radius(rb::Prong shower){
    int i = shower.ID();
    double rad = 0;
    double gev = 0;
    for (unsigned int nh=0;nh<shower.NCell();nh++) {
      art::Ptr<rb::CellHit>  cellhit = shower.Cell(nh);
      geo::OfflineChan key(cellhit->Plane(),cellhit->Cell());
      double cellEnergy = fHitEWeightMap[key][i];
//      double celldist = (RecoJMShower::GetTrkHitPos(cellhit->Plane(),cellhit->Cell(),shower.Dir(),shower.Start(),RecoJMShower::GetShwStop(shower))).z();
      double celldist= RecoJMShower::GetCellDistToTrk(cellhit->Plane(),cellhit->Cell(),shower.Start(),RecoJMShower::GetShwStop(shower));
      rad += fHitEWeightMap[key][i]*celldist;
       gev += cellEnergy;
     }
    if(gev>0.001)rad = rad/gev;
    return rad;
  }

  double RecoJMShower::Radius(rb::Cluster cluster, rb::Prong shower){
    int i = shower.ID();
    double rad = 0;
    double gev = 0;
    for (unsigned int nh=0;nh<cluster.NCell();nh++) {
      art::Ptr<rb::CellHit>  cellhit = cluster.Cell(nh);
      geo::OfflineChan key(cellhit->Plane(),cellhit->Cell());
      double cellEnergy = fHitEWeightMap[key][i];
//      double celldist = (RecoJMShower::GetTrkHitPos(cellhit->Plane(),cellhit->Cell(),shower.Dir(),shower.Start(),RecoJMShower::GetShwStop(shower))).z();
      double celldist= RecoJMShower::GetCellDistToTrk(cellhit->Plane(),cellhit->Cell(),shower.Start(),RecoJMShower::GetShwStop(shower));
       rad += fHitEWeightMap[key][i]*celldist;
       gev += cellEnergy;
     }
    if(gev>0.001)rad = rad/gev;
    return rad;
  }

  double RecoJMShower::RadiusV(rb::Cluster cluster){
    double rad = 0;
    double gev = 0;
    TVector3 clustercentroid = RecoJMShower::GetCentroid(cluster); 
    for (unsigned int nh=0;nh<cluster.NCell();nh++) {
      art::Ptr<rb::CellHit>  cellhit = cluster.Cell(nh);
      geo::OfflineChan key(cellhit->Plane(),cellhit->Cell());
      double cellEnergy = cellhit->PE();
      TVector3 celldist = RecoJMShower::GetCellDistToPoint(cellhit->Plane(), cellhit->Cell(),clustercentroid); 
      rad += cellEnergy*sqrt(celldist[0]*celldist[0]+celldist[1]*celldist[1]+celldist[2]*celldist[2]);
      gev += cellEnergy;
     }
    if(gev>0.001)rad = rad/gev;
    return rad;
  }

  TVector3 RecoJMShower::GetCentroid(rb::Cluster cluster){
    double xadc = 0;
    double yadc = 0;
    double zadc = 0;
    double xadccellx = 0; 
    double yadccelly = 0; 
    double zadccellz = 0; 
    double cx = 0;
    double cy = 0;
    double cz = 0;

    for(unsigned int nh=0; nh<cluster.NXCell();nh++){
      art::Ptr<rb::CellHit>  cellhit = cluster.XCell(nh);
      Double_t cellXYZ[3]={0,0,0}; 
      art::ServiceHandle<geo::Geometry> geom;
      geom->Plane(cellhit->Plane())->Cell(cellhit->Cell())->GetCenter(cellXYZ);
      double cellADC = double(cellhit->ADC());
      xadc += cellADC; 
      zadc += cellADC; 
      xadccellx += cellADC*cellXYZ[0]; 
      zadccellz += cellADC*cellXYZ[2]; 
    }

    for(unsigned int nh=0; nh<cluster.NYCell();nh++){
      art::Ptr<rb::CellHit>  cellhit = cluster.YCell(nh);
      Double_t cellXYZ[3]={0,0,0};
      art::ServiceHandle<geo::Geometry> geom;
      geom->Plane(cellhit->Plane())->Cell(cellhit->Cell())->GetCenter(cellXYZ);
      double cellADC = double(cellhit->ADC());
      yadc += cellADC;
      zadc += cellADC;
      yadccelly += cellADC*cellXYZ[1];
      zadccellz += cellADC*cellXYZ[2];
    }
    if(xadc>0)cx = xadccellx/xadc;  
    if(yadc>0)cy = yadccelly/yadc;  
    if(zadc>0)cz = zadccellz/zadc;
    TVector3 centroid(cx, cy, cz);
    return centroid;
  }


  double RecoJMShower::Gap(rb::Prong shower, TVector3 vertex){
    TVector3 showerStart=shower.Start();
    double gap = (vertex-showerStart).Mag();
    return gap;  
  }

  double RecoJMShower::GetPlaneDedx(rb::Prong shower, unsigned int plane){
    int i = shower.ID(); 
    if(fShowerPClusterCol[i].size()<1)
    {
      std::cout<<"JMShower: Nothing in plane "<<i<<std::endl;
      return 9999;
    }
    double path = RecoJMShower::GetTrkHitPath(fShowerPClusterCol[i][plane].Plane(), 0, shower);
    double ep = RecoJMShower::Energy(fShowerPClusterCol[i][plane], shower.ID());
//    TVector3 hitpos = RecoJMShower::GetTrkPlanePos(fShowerPClusterCol[i][plane].Plane(), shower);
//    double poscor = (0.927377+6.46394e-05*hitpos[0])*(0.993889+3.83504e-05*hitpos[1]);/from plane
//    double poscor = (9.9899+7.46983e-05*avgx+7.34449e-05*avgy);//from shower 
//    double dedx = (ep/poscor)/path;
    double dedx = ep/path;
    if(isrealdata==true&&pCorDataDedx)dedx=dedx/RecoJMShower::PlaneDedxDataMC(dedx, fDetid);
    //std::cout<<"shower: "<<i<<" plane idx: "<<plane<<" plane: "<<fShowerPClusterCol[i][plane].Plane()<<" path: "<<path<<" ep: "<<ep<<" dedx: "<<dedx<<std::endl;
    return dedx;
  }


  double RecoJMShower::GetPlaneDedxProb(rb::Prong shower, int ireg, int igev, unsigned int plane, double dedx, int  type){
//    int i = shower.ID();
    double p = 0;
    if(ireg<0||ireg>3)return 1;
    if(igev>=11||igev<0)return 1;
    if(plane>=200)return 1;

    if(int(jmshower::kElectron)==type){
      double ntot = double(fHtPlaneDedxE[ireg][igev][plane]->GetEntries());
      double n = double(fHtPlaneDedxE[ireg][igev][plane]->Interpolate(dedx));
      double nbin = fHtPlaneDedxE[ireg][igev][plane]->GetNbinsX();
      if(n<0.0001)n=0.0001;
      if(ntot<1){
        p = 0;
      }else{
        p = n*nbin/ntot;
      }
      //std::cout<<"type "<<type<<" "<<ireg<<" "<<igev<<" "<<plane<<" "<<dedx<<" "<<ntot<<" "<<n<<" "<<nbin<<" "<<p<<std::endl;
    }
    if(int(jmshower::kGamma)==type){
      double ntot = double(fHtPlaneDedxG[ireg][igev][plane]->GetEntries());
      double n = double(fHtPlaneDedxG[ireg][igev][plane]->Interpolate(dedx));
      double nbin = fHtPlaneDedxG[ireg][igev][plane]->GetNbinsX();
      if(n<0.0001)n=0.0001;
      if(ntot<1){
        p = 0;
      }else{
        p = n*nbin/ntot;
      }
    }
    if(int(jmshower::kMuon)==type){
      double ntot = double(fHtPlaneDedxMu[ireg][igev][plane]->GetEntries());
      double n = double(fHtPlaneDedxMu[ireg][igev][plane]->Interpolate(dedx));
      double nbin = fHtPlaneDedxMu[ireg][igev][plane]->GetNbinsX();
      if(n<0.0001)n=0.0001;
      if(ntot<1){
        p = 0;
      }else{
        p = n*nbin/ntot;
      }
    }
    if(int(jmshower::kPi0)==type){
      double ntot = double(fHtPlaneDedxPi0[ireg][igev][plane]->GetEntries());
      double n = double(fHtPlaneDedxPi0[ireg][igev][plane]->Interpolate(dedx));
      double nbin = fHtPlaneDedxPi0[ireg][igev][plane]->GetNbinsX();
      if(n<0.0001)n=0.0001;
      if(ntot<1){
        p = 0;
      }else{
        p = n*nbin/ntot;
      }
    }
/*
    if(int(jmshower::kHad)==type){
      double ntot = double(fHtPlaneDedxHad[ireg][igev][plane]->GetEntries());
      double n = double(fHtPlaneDedxHad[ireg][igev][plane]->Interpolate(dedx));
      double nbin = fHtPlaneDedxHad[ireg][igev][plane]->GetNbinsX();
      if(n<0.0001)n=0.0001;
      if(ntot<1){
        p = 0;
      }else{
        p = n*nbin/ntot;
      }
    }
*/

    if(int(jmshower::kProton)==type){
      double ntot = double(fHtPlaneDedxP[ireg][igev][plane]->GetEntries());
      double n = double(fHtPlaneDedxP[ireg][igev][plane]->Interpolate(dedx));
      double nbin = fHtPlaneDedxP[ireg][igev][plane]->GetNbinsX();
      if(n<0.0001)n=0.0001;
      if(ntot<1){
        p = 0;
      }else{
        p = n*nbin/ntot;
      }
    }

    if(int(jmshower::kNeutron)==type){
      double ntot = double(fHtPlaneDedxN[ireg][igev][plane]->GetEntries());
      double n = double(fHtPlaneDedxN[ireg][igev][plane]->Interpolate(dedx));
      double nbin = fHtPlaneDedxN[ireg][igev][plane]->GetNbinsX();
      if(n<0.0001)n=0.0001;
      if(ntot<1){
        p = 0;
      }else{
        p = n*nbin/ntot;
      }
    }

    if(int(jmshower::kPion)==type){
      double ntot = double(fHtPlaneDedxPi[ireg][igev][plane]->GetEntries());
      double n = double(fHtPlaneDedxPi[ireg][igev][plane]->Interpolate(dedx));
      double nbin = fHtPlaneDedxPi[ireg][igev][plane]->GetNbinsX();
      if(n<0.0001)n=0.0001;
      if(ntot<1){
        p = 0;
      }else{
        p = n*nbin/ntot;
      }
    }
    if(int(jmshower::kElectronQE)==type){
      double ntot = double(fHtPlaneDedxEqe[ireg][igev][plane]->GetEntries());
      double n = double(fHtPlaneDedxEqe[ireg][igev][plane]->Interpolate(dedx));
      double nbin = fHtPlaneDedxEqe[ireg][igev][plane]->GetNbinsX();
      if(n<0.0001)n=0.0001;
      if(ntot<1){
        p = 0;
      }else{
        p = n*nbin/ntot;
      }
    }

    if(int(jmshower::kElectronRES)==type){
      double ntot = double(fHtPlaneDedxEres[ireg][igev][plane]->GetEntries());
      double n = double(fHtPlaneDedxEres[ireg][igev][plane]->Interpolate(dedx));
      double nbin = fHtPlaneDedxEres[ireg][igev][plane]->GetNbinsX();
      if(n<0.0001)n=0.0001;
      if(ntot<1){
        p = 0;
      }else{
        p = n*nbin/ntot;
      }
    }

    if(int(jmshower::kElectronDIS)==type){
      double ntot = double(fHtPlaneDedxEdis[ireg][igev][plane]->GetEntries());
      double n = double(fHtPlaneDedxEdis[ireg][igev][plane]->Interpolate(dedx));
      double nbin = fHtPlaneDedxEdis[ireg][igev][plane]->GetNbinsX();
      if(n<0.0001)n=0.0001;
      if(ntot<1){
        p = 0;
      }else{
        p = n*nbin/ntot;
      }
    }

    if(int(jmshower::kElectronCOH)==type){
      double ntot = double(fHtPlaneDedxEcoh[ireg][igev][plane]->GetEntries());
      double n = double(fHtPlaneDedxEcoh[ireg][igev][plane]->Interpolate(dedx));
      double nbin = fHtPlaneDedxEcoh[ireg][igev][plane]->GetNbinsX();
      if(n<0.0001)n=0.0001;
      if(ntot<1){
        p = 0;
      }else{
        p = n*nbin/ntot;
      }
    }

    return p;
  }

/*
  double RecoJMShower::GetPlaneDedxProb(rb::Prong shower, int igev, unsigned int plane, double dedx, unsigned int nmip, const int type){
//    int i = shower.ID();
    double p = 0;
    if(igev>=11||igev<0)return 1;
    if(plane>=200)return 1;
//    double dedx1 = RecoJMShower::GetPlaneDedx(shower, plane+1);
    if(int(jmshower::kElectron)==type){
      if(plane<nmip){
        double ntot = double(fHtmipPlaneDedxE[igev][plane]->GetEntries());
        double ntotcomb = double(fHtmipPlaneDedxE[igev][plane]->GetEntries()+fHtshowerPlaneDedxE[igev][plane]->GetEntries());
        double n = double(fHtmipPlaneDedxE[igev][plane]->Interpolate(dedx));
        double nbin = fHtmipPlaneDedxE[igev][plane]->GetNbinsX();
//        double nbincomb = fHtmipPlaneDedxE[igev][plane]->GetNbinsX()+fHtshowerPlaneDedxE[igev][plane]->GetNbinsX();
        if(n<0.0001)n=0.0001;
        if(ntotcomb<100){
          if(igev>1)p = RecoJMShower::GetPlaneDedxProb(shower, igev-1, plane, dedx, nmip, type); 
          if(igev<=1)p = RecoJMShower::GetPlaneDedxProb(shower, igev+1, plane, dedx, nmip, type); 
        }else{
          p = (n*nbin/ntot)*(ntot/ntotcomb);
        }
      }
      if(plane>=nmip){
//        double ntot = double(fHtshowerPlaneDedxE[igev][plane]->GetEntries());
        double ntotcomb = double(fHtshowerPlaneDedxE[igev][plane]->GetEntries()+fHtmipPlaneDedxE[igev][plane]->GetEntries());
        double n = double(fHtshowerPlaneDedxE[igev][plane]->Interpolate(dedx));
        double nbin = fHtshowerPlaneDedxE[igev][plane]->GetNbinsX();
//        double nbincomb = fHtshowerPlaneDedxE[igev][plane]->GetNbinsX()+fHtmipPlaneDedxE[igev][plane]->GetNbinsX();
        if(n<0.0001)n=0.0001;
        if(ntotcomb<100){
          if(igev>1)p = RecoJMShower::GetPlaneDedxProb(shower, igev-1, plane, dedx, nmip, type);             
          if(igev<=1)p = RecoJMShower::GetPlaneDedxProb(shower, igev+1, plane, dedx, nmip, type);
        }else{
          p = n*nbin/ntotcomb;
        }
      }
    }


    if(int(jmshower::kGamma)==type){
      if(plane<nmip){
//        double ntot = double(fHtmipPlaneDedxG[igev][plane]->GetEntries());
        double ntotcomb = double(fHtmipPlaneDedxG[igev][plane]->GetEntries()+fHtshowerPlaneDedxG[igev][plane]->GetEntries());
        double n = double(fHtmipPlaneDedxG[igev][plane]->Interpolate(dedx));
        double nbin = double(fHtmipPlaneDedxG[igev][plane]->GetNbinsX());
//        double nbincomb = double(fHtmipPlaneDedxG[igev][plane]->GetNbinsX()+fHtshowerPlaneDedxG[igev][plane]->GetNbinsX());
        if(n<0.0001)n=0.0001;
        if(ntotcomb<100){
          if(igev>1)p = RecoJMShower::GetPlaneDedxProb(shower, igev-1, plane, dedx, nmip, type);             
          if(igev<=1)p = RecoJMShower::GetPlaneDedxProb(shower, igev+1, plane, dedx, nmip, type);
        }else{
          p = n*nbin/ntotcomb;
        }
      }
      if(plane>=nmip){
//        double ntot = double(fHtshowerPlaneDedxG[igev][plane]->GetEntries());
        double ntotcomb = double(fHtshowerPlaneDedxG[igev][plane]->GetEntries()+fHtmipPlaneDedxG[igev][plane]->GetEntries());
        double n = double(fHtshowerPlaneDedxG[igev][plane]->Interpolate(dedx));
        double nbin = double(fHtshowerPlaneDedxG[igev][plane]->GetNbinsX());
//        double nbincomb = double(fHtshowerPlaneDedxG[igev][plane]->GetNbinsX()+fHtmipPlaneDedxG[igev][plane]->GetNbinsX());
        if(n<0.0001)n=0.0001;
        if(ntotcomb<100){
          if(igev>1)p = RecoJMShower::GetPlaneDedxProb(shower, igev-1, plane, dedx, nmip, type);             
          if(igev<=1)p = RecoJMShower::GetPlaneDedxProb(shower, igev+1, plane, dedx, nmip, type);
        }else{
          p = n*nbin/ntotcomb;
        }
      }
    }


    if(int(jmshower::kMuon)==type){
      double ntot = double(fHtPlaneDedxMu[igev][plane]->GetEntries());
      double n = double(fHtPlaneDedxMu[igev][plane]->Interpolate(dedx));
      double nbin = fHtPlaneDedxMu[igev][plane]->GetNbinsX();
      if(n<0.0001)n=0.0001;
      if(ntot<1&&igev>0&&igev<10){
        p = 0;
      }else{
        p = n*nbin/ntot;
      }
    }

    if(int(jmshower::kPi0)==type){
      if(plane<nmip){
//        double ntot = double(fHtmipPlaneDedxPi0[igev][plane]->GetEntries());
        double ntotcomb = double(fHtmipPlaneDedxPi0[igev][plane]->GetEntries()+fHtshowerPlaneDedxPi0[igev][plane]->GetEntries());
        double n = double(fHtmipPlaneDedxPi0[igev][plane]->Interpolate(dedx));
        double nbin = double(fHtmipPlaneDedxPi0[igev][plane]->GetNbinsX());
//        double nbincomb = double(fHtmipPlaneDedxPi0[igev][plane]->GetNbinsX()+fHtshowerPlaneDedxPi0[igev][plane]->GetNbinsX());
        if(n<0.0001)n=0.0001;
        if(ntotcomb<100){
          if(igev>1)p = RecoJMShower::GetPlaneDedxProb(shower, igev-1, plane, dedx, nmip, type);             
          if(igev<=1)p = RecoJMShower::GetPlaneDedxProb(shower, igev+1, plane, dedx, nmip, type);
        }else{
          p = n*nbin/ntotcomb;
        }
      }
      if(plane>=nmip){
//        double ntot = double(fHtshowerPlaneDedxPi0[igev][plane]->GetEntries());
        double ntotcomb = double(fHtshowerPlaneDedxPi0[igev][plane]->GetEntries()+fHtmipPlaneDedxPi0[igev][plane]->GetEntries());
        double n = double(fHtshowerPlaneDedxPi0[igev][plane]->Interpolate(dedx));
        double nbin = double(fHtshowerPlaneDedxPi0[igev][plane]->GetNbinsX());
//        double nbincomb = double(fHtshowerPlaneDedxPi0[igev][plane]->GetNbinsX()+fHtmipPlaneDedxPi0[igev][plane]->GetNbinsX());
        if(n<0.0001)n=0.0001;
        if(ntotcomb<100){
          if(igev>1)p = RecoJMShower::GetPlaneDedxProb(shower, igev-1, plane, dedx, nmip, type);             
          if(igev<=1)p = RecoJMShower::GetPlaneDedxProb(shower, igev+1, plane, dedx, nmip, type);
        }else{
          p = n*nbin/ntotcomb;
        }
      }
    }


    if(int(jmshower::kHad)==type){
      if(plane<nmip){
//        double ntot = double(fHtmipPlaneDedxHad[igev][plane]->GetEntries());
        double ntotcomb = double(fHtmipPlaneDedxHad[igev][plane]->GetEntries()+fHtshowerPlaneDedxHad[igev][plane]->GetEntries());
        double n = double(fHtmipPlaneDedxHad[igev][plane]->Interpolate(dedx));
        double nbin = double(fHtmipPlaneDedxHad[igev][plane]->GetNbinsX());
//        double nbincomb = double(fHtmipPlaneDedxHad[igev][plane]->GetNbinsX()+fHtshowerPlaneDedxHad[igev][plane]->GetNbinsX());
        if(n<0.0001)n=0.0001;
        if(ntotcomb<100){
          if(igev>1)p = RecoJMShower::GetPlaneDedxProb(shower, igev-1, plane, dedx, nmip, type);             
          if(igev<=1)p = RecoJMShower::GetPlaneDedxProb(shower, igev+1, plane, dedx, nmip, type);
        }else{
          p = n*nbin/ntotcomb;
        }
      }
      if(plane>=nmip){
//        double ntot = double(fHtshowerPlaneDedxHad[igev][plane]->GetEntries());
        double ntotcomb = double(fHtshowerPlaneDedxHad[igev][plane]->GetEntries()+fHtmipPlaneDedxHad[igev][plane]->GetEntries());
        double n = double(fHtshowerPlaneDedxHad[igev][plane]->Interpolate(dedx));
        double nbin = double(fHtshowerPlaneDedxHad[igev][plane]->GetNbinsX());
//        double nbincomb = double(fHtshowerPlaneDedxHad[igev][plane]->GetNbinsX()+fHtmipPlaneDedxHad[igev][plane]->GetNbinsX());
        if(n<0.0001)n=0.0001;
        if(ntotcomb<100){
          if(igev>1)p = RecoJMShower::GetPlaneDedxProb(shower, igev-1, plane, dedx, nmip, type);             
          if(igev<=1)p = RecoJMShower::GetPlaneDedxProb(shower, igev+1, plane, dedx, nmip, type);
        }else{
          p = n*nbin/ntotcomb;
        }
      }
    }


    if(int(jmshower::kProton)==type){
      if(plane<nmip){
//        double ntot = double(fHtmipPlaneDedxP[igev][plane]->GetEntries());
        double ntotcomb = double(fHtmipPlaneDedxP[igev][plane]->GetEntries()+fHtshowerPlaneDedxP[igev][plane]->GetEntries());
        double n = double(fHtmipPlaneDedxP[igev][plane]->Interpolate(dedx));
        double nbin = double(fHtmipPlaneDedxP[igev][plane]->GetNbinsX());
//        double nbincomb = double(fHtmipPlaneDedxP[igev][plane]->GetNbinsX()+fHtshowerPlaneDedxP[igev][plane]->GetNbinsX());
        if(n<0.0001)n=0.0001;
        if(ntotcomb<100){
          if(igev>1)p = RecoJMShower::GetPlaneDedxProb(shower, igev-1, plane, dedx, nmip, type);             
          if(igev<=1)p =  RecoJMShower::GetPlaneDedxProb(shower, igev+1, plane, dedx, nmip, type);
        }else{
          p = n*nbin/ntotcomb;
        }
      }
      if(plane>=nmip){
//        double ntot = double(fHtshowerPlaneDedxP[igev][plane]->GetEntries());
        double ntotcomb = double(fHtshowerPlaneDedxP[igev][plane]->GetEntries()+fHtmipPlaneDedxP[igev][plane]->GetEntries());
        double n = double(fHtshowerPlaneDedxP[igev][plane]->Interpolate(dedx));
        double nbin = double(fHtshowerPlaneDedxP[igev][plane]->GetNbinsX());
//        double nbincomb = double(fHtshowerPlaneDedxP[igev][plane]->GetNbinsX()+fHtmipPlaneDedxP[igev][plane]->GetNbinsX());
        if(n<0.0001)n=0.0001;
        if(ntotcomb<100){
          if(igev>1)p = RecoJMShower::GetPlaneDedxProb(shower, igev-1, plane, dedx, nmip, type);             
          if(igev<=1)p = RecoJMShower::GetPlaneDedxProb(shower, igev+1, plane, dedx, nmip, type);
        }else{
          p = n*nbin/ntotcomb;
        }
      }
    }



    if(int(jmshower::kNeutron)==type){
      if(plane<nmip){
//        double ntot = double(fHtmipPlaneDedxN[igev][plane]->GetEntries());
        double ntotcomb = double(fHtmipPlaneDedxN[igev][plane]->GetEntries()+fHtshowerPlaneDedxN[igev][plane]->GetEntries());
        double n = double(fHtmipPlaneDedxN[igev][plane]->Interpolate(dedx));
        double nbin = double(fHtmipPlaneDedxN[igev][plane]->GetNbinsX());
//        double nbincomb = double(fHtmipPlaneDedxN[igev][plane]->GetNbinsX()+fHtshowerPlaneDedxN[igev][plane]->GetNbinsX());
        if(n<0.0001)n=0.0001;
        if(ntotcomb<100){
          if(igev>1)p = RecoJMShower::GetPlaneDedxProb(shower, igev-1, plane, dedx, nmip, type);             
          if(igev<=1)p = RecoJMShower::GetPlaneDedxProb(shower, igev+1, plane, dedx, nmip, type);
        }else{
          p = n*nbin/ntotcomb;
        }
      }
      if(plane>=nmip){
//        double ntot = double(fHtshowerPlaneDedxN[igev][plane]->GetEntries());
        double ntotcomb = double(fHtshowerPlaneDedxN[igev][plane]->GetEntries()+fHtmipPlaneDedxN[igev][plane]->GetEntries());
        double n = double(fHtshowerPlaneDedxN[igev][plane]->Interpolate(dedx));
        double nbin = double(fHtshowerPlaneDedxN[igev][plane]->GetNbinsX());
//        double nbincomb = double(fHtshowerPlaneDedxN[igev][plane]->GetNbinsX()+fHtmipPlaneDedxN[igev][plane]->GetNbinsX());
        if(n<0.0001)n=0.0001;
        if(ntotcomb<100){
          if(igev>1)p = RecoJMShower::GetPlaneDedxProb(shower, igev-1, plane, dedx, nmip, type);             
          if(igev<=1)p = RecoJMShower::GetPlaneDedxProb(shower, igev+1, plane, dedx, nmip, type);
        }else{
          p = n*nbin/ntotcomb;
        }
      }
    }

    if(int(jmshower::kPion)==type){
      if(plane<nmip){
//        double ntot = double(fHtmipPlaneDedxPi[igev][plane]->GetEntries());
        double ntotcomb = double(fHtmipPlaneDedxPi[igev][plane]->GetEntries()+fHtshowerPlaneDedxPi[igev][plane]->GetEntries());
        double n = double(fHtmipPlaneDedxPi[igev][plane]->Interpolate(dedx));
        double nbin = double(fHtmipPlaneDedxPi[igev][plane]->GetNbinsX());
//        double nbincomb = double(fHtmipPlaneDedxPi[igev][plane]->GetNbinsX()+fHtshowerPlaneDedxPi[igev][plane]->GetNbinsX());
        if(n<0.0001)n=0.0001;
        if(ntotcomb<100){
          if(igev>1)p = RecoJMShower::GetPlaneDedxProb(shower, igev-1, plane, dedx, nmip, type);             
          if(igev<=1)p = RecoJMShower::GetPlaneDedxProb(shower, igev+1, plane, dedx, nmip, type);
        }else{
          p = n*nbin/ntotcomb;
        }
      }
      if(plane>=nmip){
//        double ntot = double(fHtshowerPlaneDedxPi[igev][plane]->GetEntries());
        double ntotcomb = double(fHtshowerPlaneDedxPi[igev][plane]->GetEntries()+fHtmipPlaneDedxPi[igev][plane]->GetEntries());
        double n = double(fHtshowerPlaneDedxPi[igev][plane]->Interpolate(dedx));
        double nbin = double(fHtshowerPlaneDedxPi[igev][plane]->GetNbinsX());
//        double nbincomb = double(fHtshowerPlaneDedxPi[igev][plane]->GetNbinsX()+fHtmipPlaneDedxPi[igev][plane]->GetNbinsX());
        if(n<0.0001)n=0.0001;
        if(ntotcomb<100){
          if(igev>1)p = RecoJMShower::GetPlaneDedxProb(shower, igev-1, plane, dedx, nmip, type);             
          if(igev<=1)p =  RecoJMShower::GetPlaneDedxProb(shower, igev+1, plane, dedx, nmip, type);
        }else{
          p = n*nbin/ntotcomb;
        }
      }
    }



    if(int(jmshower::kElectronQE)==type){
      if(plane<nmip){
        double ntotcomb = double(fHtmipPlaneDedxEqe[igev][plane]->GetEntries()+fHtshowerPlaneDedxEqe[igev][plane]->GetEntries());
        double n = double(fHtmipPlaneDedxEqe[igev][plane]->Interpolate(dedx));
        double nbin = double(fHtmipPlaneDedxEqe[igev][plane]->GetNbinsX());
        if(n<0.0001)n=0.0001;
        if(ntotcomb<100){
          if(igev>1)p = RecoJMShower::GetPlaneDedxProb(shower, igev-1, plane, dedx, nmip, type);             
          if(igev<=1)p = RecoJMShower::GetPlaneDedxProb(shower, igev+1, plane, dedx, nmip, type);
        }else{
          p = n*nbin/ntotcomb;
        }
      }
      if(plane>=nmip){
        double ntotcomb = double(fHtshowerPlaneDedxEqe[igev][plane]->GetEntries()+fHtmipPlaneDedxEqe[igev][plane]->GetEntries());
        double n = double(fHtshowerPlaneDedxEqe[igev][plane]->Interpolate(dedx));
        double nbin = double(fHtshowerPlaneDedxEqe[igev][plane]->GetNbinsX());
        if(n<0.0001)n=0.0001;
        if(ntotcomb<100){
          if(igev>1)p = RecoJMShower::GetPlaneDedxProb(shower, igev-1, plane, dedx, nmip, type);             
          if(igev<=1)p =  RecoJMShower::GetPlaneDedxProb(shower, igev+1, plane, dedx, nmip, type);
        }else{
          p = n*nbin/ntotcomb;
        }
      }
    }


    if(int(jmshower::kElectronRES)==type){
      if(plane<nmip){
        double ntotcomb = double(fHtmipPlaneDedxEres[igev][plane]->GetEntries()+fHtshowerPlaneDedxEres[igev][plane]->GetEntries());
        double n = double(fHtmipPlaneDedxEres[igev][plane]->Interpolate(dedx));
        double nbin = double(fHtmipPlaneDedxEres[igev][plane]->GetNbinsX());
        if(n<0.0001)n=0.0001;
        if(ntotcomb<100){
          if(igev>1)p = RecoJMShower::GetPlaneDedxProb(shower, igev-1, plane, dedx, nmip, type);             
          if(igev<=1)p = RecoJMShower::GetPlaneDedxProb(shower, igev+1, plane, dedx, nmip, type);
        }else{
          p = n*nbin/ntotcomb;
        }
      }
      if(plane>=nmip){
        double ntotcomb = double(fHtshowerPlaneDedxEres[igev][plane]->GetEntries()+fHtmipPlaneDedxEres[igev][plane]->GetEntries());
        double n = double(fHtshowerPlaneDedxEres[igev][plane]->Interpolate(dedx));
        double nbin = double(fHtshowerPlaneDedxEres[igev][plane]->GetNbinsX());
        if(n<0.0001)n=0.0001;
        if(ntotcomb<100){
          if(igev>1)p = RecoJMShower::GetPlaneDedxProb(shower, igev-1, plane, dedx, nmip, type);             
          if(igev<=1)p =  RecoJMShower::GetPlaneDedxProb(shower, igev+1, plane, dedx, nmip, type);
        }else{
          p = n*nbin/ntotcomb;
        }
      }
    }


    if(int(jmshower::kElectronDIS)==type){
      if(plane<nmip){
        double ntotcomb = double(fHtmipPlaneDedxEdis[igev][plane]->GetEntries()+fHtshowerPlaneDedxEdis[igev][plane]->GetEntries());
        double n = double(fHtmipPlaneDedxEdis[igev][plane]->Interpolate(dedx));
        double nbin = double(fHtmipPlaneDedxEdis[igev][plane]->GetNbinsX());
        if(n<0.0001)n=0.0001;
        if(ntotcomb<100){
          if(igev>1)p = RecoJMShower::GetPlaneDedxProb(shower, igev-1, plane, dedx, nmip, type);             
          if(igev<=1)p = RecoJMShower::GetPlaneDedxProb(shower, igev+1, plane, dedx, nmip, type);
        }else{
          p = n*nbin/ntotcomb;
        }
      }
      if(plane>=nmip){
        double ntotcomb = double(fHtshowerPlaneDedxEdis[igev][plane]->GetEntries()+fHtmipPlaneDedxEdis[igev][plane]->GetEntries());
        double n = double(fHtshowerPlaneDedxEdis[igev][plane]->Interpolate(dedx));
        double nbin = double(fHtshowerPlaneDedxEdis[igev][plane]->GetNbinsX());
        if(n<0.0001)n=0.0001;
        if(ntotcomb<100){
          if(igev>1)p = RecoJMShower::GetPlaneDedxProb(shower, igev-1, plane, dedx, nmip, type);             
          if(igev<=1)p =  RecoJMShower::GetPlaneDedxProb(shower, igev+1, plane, dedx, nmip, type);
        }else{
          p = n*nbin/ntotcomb;
        }
      }
    }


    if(int(jmshower::kElectronCOH)==type){
      if(plane<nmip){
        double ntotcomb = double(fHtmipPlaneDedxEcoh[igev][plane]->GetEntries()+fHtshowerPlaneDedxEcoh[igev][plane]->GetEntries());
        double n = double(fHtmipPlaneDedxEcoh[igev][plane]->Interpolate(dedx));
        double nbin = double(fHtmipPlaneDedxEcoh[igev][plane]->GetNbinsX());
        if(n<0.0001)n=0.0001;
        if(ntotcomb<100){
          if(igev>1)p = RecoJMShower::GetPlaneDedxProb(shower, igev-1, plane, dedx, nmip, type);             
          if(igev<=1)p = RecoJMShower::GetPlaneDedxProb(shower, igev+1, plane, dedx, nmip, type);
        }else{
          p = n*nbin/ntotcomb;
        }
      }
      if(plane>=nmip){
        double ntotcomb = double(fHtshowerPlaneDedxEcoh[igev][plane]->GetEntries()+fHtmipPlaneDedxEcoh[igev][plane]->GetEntries());
        double n = double(fHtshowerPlaneDedxEcoh[igev][plane]->Interpolate(dedx));
        double nbin = double(fHtshowerPlaneDedxEcoh[igev][plane]->GetNbinsX());
        if(n<0.0001)n=0.0001;
        if(ntotcomb<100){
          if(igev>1)p = RecoJMShower::GetPlaneDedxProb(shower, igev-1, plane, dedx, nmip, type);             
          if(igev<=1)p =  RecoJMShower::GetPlaneDedxProb(shower, igev+1, plane, dedx, nmip, type);
        }else{
          p = n*nbin/ntotcomb;
        }
      }
    }


    return p;
  }
*/

  double RecoJMShower::GetInterPlaneDedxProb(rb::Prong shower,  unsigned int plane, double dedx, const int type, TVector3 pos, int iReg, int igev0, int igev1, double e0, double e1){

    double probPXPY = 0;
    bool interPos = false; 

//    int i = shower.ID();
    art::ServiceHandle<geo::Geometry> geom;
//    TVector3 pos = RecoJMShower::GetTrkPlanePos(fShowerPClusterCol[i][plane].Plane(), shower);
//    std::cout<<"in getinterplane pos X,Y "<<pos[0]<<" "<<pos[1]<<std::endl;

//    TVector3 pos((shower.Start()[0]+RecoJMShower::GetShwStop(shower)[0])/2.0,(shower.Start()[1]+RecoJMShower::GetShwStop(shower)[1])/2.0,(shower.Start()[2]+RecoJMShower::GetShwStop(shower)[2])/2.0);


    double cos = shower.Dir()[2];

    if(interPos==true){
      double X0 = -geom->DetHalfWidth()/2.0;
      double X1 = geom->DetHalfWidth()/2.0;
      double Y0 = -geom->DetHalfHeight()/2.0;
      double Y1 = geom->DetHalfHeight()/2.0;

      double probreg[4]={1,1,1,1};
      for(int ireg=0;ireg<4;ireg++){
        unsigned int plane0 = int(int(plane)/fabs(cos));
        unsigned int plane1 = int(int(plane)/fabs(cos)) + 1;
        double r0 = double(plane/fabs(cos))-double(plane0);
        double r1 = 1-r0;
        double probe0r0 = 1;
        double probe0r1 = 1;
        double probe1r0 = 1;
        double probe1r1 = 1;

        probe0r0 = RecoJMShower::GetPlaneDedxProb(shower, ireg, igev0, plane0, dedx,  type);
        probe0r1 = RecoJMShower::GetPlaneDedxProb(shower, ireg, igev0, plane1, dedx,  type);
        probe1r0 = RecoJMShower::GetPlaneDedxProb(shower, ireg, igev1, plane0, dedx,  type);
        probe1r1 = RecoJMShower::GetPlaneDedxProb(shower, ireg, igev1, plane1, dedx,  type);
//        std::cout<<"ireg, prob probe0r0 "<<ireg<<" "<<probe0r0<<std::endl;
//        std::cout<<"ireg, prob probe0r1 "<<ireg<<" "<<probe0r1<<std::endl;
//        std::cout<<"ireg, prob probe1r0 "<<ireg<<" "<<probe1r0<<std::endl;
//        std::cout<<"ireg, prob probe1r0 "<<ireg<<" "<<probe1r0<<std::endl;
        double probe0 = (r1*probe0r0+r0*probe0r1);
        double probe1 = (r1*probe1r0+r0*probe1r1);
        if(igev0!=igev1)probreg[ireg] = (e1*probe0+e0*probe1)/(e0+e1);
        if(igev0==igev1)probreg[ireg] = probe0;
//        std::cout<<"ireg, probreg "<<ireg<<" "<<probreg[ireg]<<std::endl;
      }
      double PX = pos[0];
      double PY = pos[1];
      double probX1Y1 = probreg[0];
      double probX1Y0 = probreg[1];
      double probX0Y1 = probreg[2];
      double probX0Y0 = probreg[3];
//      std::cout<<"prob X1Y1, X1Y0, X0Y1, X0Y0 "<<probX1Y1<<" "<<probX1Y0<<" "<<probX0Y1<<" "<<probX0Y0<<std::endl;
      double probX1PY = ((PY-Y0)/(Y1-Y0))*(probX1Y1-probX1Y0)+probX1Y0;
      double probX0PY = ((PY-Y0)/(Y1-Y0))*(probX0Y1-probX0Y0)+probX0Y0; 
      probPXPY = ((PX-X0)/(X1-X0))*(probX1PY-probX0PY)+probX0PY;
    }
//    std::cout<<"prob X1PY, X0PY, PXPY"<<probX1PY<<" "<<probX0PY<<" "<<probPXPY<<" "<<std::endl;
//   std::cout<<"gev,e0,e1,r0,r0: "<<gev<<" "<<e0<<" "<<e1<<" "<<r0<<" "<<r1<<std::endl;
//   std::cout<<"prob,pe0r0,pe0r1,pe1r0,pe1r1: "<<prob<<" "<<probe0r0<<" "<<probe0r1<<" "<<probe1r0<<" "<<probe1r1<<std::endl;
    if(interPos==false||probPXPY<0){
      unsigned int plane0 = int(int(plane)/fabs(cos));
      unsigned int plane1 = int(int(plane)/fabs(cos)) + 1;
      double r0 = double(plane/fabs(cos))-double(plane0);
      double r1 = 1-r0;
      double probe0r0 = 1;
      double probe0r1 = 1;
      double probe1r0 = 1;
      double probe1r1 = 1;
      probe0r0 = RecoJMShower::GetPlaneDedxProb(shower, iReg, igev0, plane0, dedx,  type);
      probe0r1 = RecoJMShower::GetPlaneDedxProb(shower, iReg, igev0, plane1, dedx,  type);
      probe1r0 = RecoJMShower::GetPlaneDedxProb(shower, iReg, igev1, plane0, dedx,  type);
      probe1r1 = RecoJMShower::GetPlaneDedxProb(shower, iReg, igev1, plane1, dedx,  type);
      double probe0 = (r1*probe0r0+r0*probe0r1);
      double probe1 = (r1*probe1r0+r0*probe1r1);
      if(igev0!=igev1)probPXPY = (e1*probe0+e0*probe1)/(e0+e1);
      if(igev0==igev1)probPXPY = probe0;
    }

    return probPXPY;
  }
/*
  double RecoJMShower::GetInterPlaneDedxProb(rb::Prong shower,  unsigned int plane, double dedx, const int type, unsigned int nmip){
    int i = shower.ID();
    double gev = RecoJMShower::Energy(shower);
    double cos = shower.Dir()[2];
    double elower[11] = {0.00, 0.25, 0.75, 1.25, 1.75, 2.25, 2.75, 3.25, 3.75, 4.25, 4.75};
    double eupper[11] = {0.25, 0.75, 1.25, 1.75, 2.25, 2.75, 3.25, 3.75, 4.25, 4.75, 5.25};
    double epoint[11];
    for(int ig=0;ig<11;ig++){
      epoint[ig] = (elower[ig]+eupper[ig])/2.;
    }
    int igev0 = 0;
    int igev1 = 0;
    if(gev>epoint[10]){igev0=10;igev1=10;}
    else if(gev<epoint[0]){igev0=0;igev1=0;}
    else{
      for(int ig=0;ig<10;ig++){
        if(gev>epoint[ig]&&gev<epoint[ig+1]){igev0=ig;igev1=ig+1;}
      }
    }
    if(igev0>10)igev0=10;
    if(igev1>10)igev1=10;
    unsigned int plane0 = int(int(plane)/fabs(cos));
    unsigned int plane1 = int(int(plane)/fabs(cos)) + 1;
    double r0 = double(plane/fabs(cos))-double(plane0);
    double r1 = 1-r0;
    double e0 = gev-epoint[igev0];
    double e1 = epoint[igev1] - gev;
    double probe0r0 = 1;
    double probe0r1 = 1;
    double probe1r0 = 1;
    double probe1r1 = 1;

    if(plane1<=10){
      probe0r0 = RecoJMShower::GetPlaneDedxProb(shower, igev0, plane0, dedx,  nmip,  type);
      probe0r1 = RecoJMShower::GetPlaneDedxProb(shower, igev0, plane1, dedx,  nmip,  type);
      probe1r0 = RecoJMShower::GetPlaneDedxProb(shower, igev1, plane0, dedx,  nmip,  type);
      probe1r1 = RecoJMShower::GetPlaneDedxProb(shower, igev1, plane1, dedx,  nmip,  type);

//      if(int(jmshower::kElectron)==type&&shower.ID()==0&&fShowerPClusterCol[i].size()>20&&igev0>=3){
//        std::cout<<"e probe0r0, type, tot, mip "<<type<<" "<<totprobe0r0<<" "<<probe0r0<<std::endl;
//        std::cout<<"e probe0r1, type, tot, mip "<<type<<" "<<totprobe0r1<<" "<<probe0r1<<std::endl;
//        std::cout<<"e probe1r0, type, tot, mip "<<type<<" "<<totprobe1r0<<" "<<probe1r0<<std::endl;
//        std::cout<<"e probe1r1, type, tot, mip "<<type<<" "<<totprobe1r1<<" "<<probe1r1<<std::endl;
//        double probeE = RecoJMShower::GetPlaneDedxProb(shower, igev0, plane0, dedx,  nmip,  0);
//        double probePi0 = RecoJMShower::GetPlaneDedxProb(shower, igev0, plane0, dedx,  nmip,  3);
//        double totprobeE = RecoJMShower::GetPlaneDedxProb(shower, igev0, plane0, dedx,  0);
//        double totprobePi0 = RecoJMShower::GetPlaneDedxProb(shower, igev0, plane0, dedx,   3);
//        std::cout<<"=================================  plane, igev0, nmip "<<plane0<<" "<<igev0<<" "<<nmip<<std::endl;
//        std::cout<<"dedx, mip e, pi0, delta === "<<dedx<<" "<<probeE<<" "<<probePi0<<" "<<probeE-probePi0<<std::endl;
//        std::cout<<"dedx, tot e, pi0, delta === "<<dedx<<" "<<totprobeE<<" "<<totprobePi0<<" "<<totprobeE-totprobePi0<<std::endl;
//      }
    }
    else{
      probe0r0 = RecoJMShower::GetPlaneDedxProb(shower, igev0, plane0, dedx,  type);
      probe0r1 = RecoJMShower::GetPlaneDedxProb(shower, igev0, plane1, dedx,  type);
      probe1r0 = RecoJMShower::GetPlaneDedxProb(shower, igev1, plane0, dedx,  type);
      probe1r1 = RecoJMShower::GetPlaneDedxProb(shower, igev1, plane1, dedx,  type);
    }

    double probe0 = (r1*probe0r0+r0*probe0r1);
    double probe1 = (r1*probe1r0+r0*probe1r1);
    double prob = 1;
    if(igev0!=igev1)prob = (e1*probe0+e0*probe1)/(e0+e1);
    if(igev0==igev1)prob = probe0;
//   std::cout<<"gev,e0,e1,r0,r0: "<<gev<<" "<<e0<<" "<<e1<<" "<<r0<<" "<<r1<<std::endl;
//   std::cout<<"prob,pe0r0,pe0r1,pe1r0,pe1r1: "<<prob<<" "<<probe0r0<<" "<<probe0r1<<" "<<probe1r0<<" "<<probe1r1<<std::endl;
    return prob;
  }
*/

  double RecoJMShower::GetDedxLongLL(rb::Prong shower, const int type){
    int i = shower.ID();
    TVector3 pos((shower.Start()[0]+RecoJMShower::GetShwStop(shower)[0])/2.0,(shower.Start()[1]+RecoJMShower::GetShwStop(shower)[1])/2.0,(shower.Start()[2]+RecoJMShower::GetShwStop(shower)[2])/2.0);
    int iReg = 0;
    if(pos[0]>=0&&pos[1]>=0)iReg = 0;
    if(pos[0]>=0&&pos[1]<0)iReg = 1;
    if(pos[0]<0&&pos[1]>=0)iReg = 2;
    if(pos[0]<=0&&pos[1]<=0)iReg = 3;

    double gev = RecoJMShower::Energy(shower);
    double elower[11] = {0.00, 0.25, 0.75, 1.25, 1.75, 2.25, 2.75, 3.25, 3.75, 4.25, 4.75};
    double eupper[11] = {0.25, 0.75, 1.25, 1.75, 2.25, 2.75, 3.25, 3.75, 4.25, 4.75, 5.25};
    double epoint[11];
    for(int ig=0;ig<11;ig++){
      epoint[ig] = (elower[ig]+eupper[ig])/2.;
    }
    int igev0 = 0;
    int igev1 = 0;
    if(gev>epoint[10]){igev0=10;igev1=10;}
    else if(gev<epoint[0]){igev0=0;igev1=0;}
    else{
      for(int ig=0;ig<10;ig++){
        if(gev>epoint[ig]&&gev<epoint[ig+1]){igev0=ig;igev1=ig+1;}
      }
    }
    if(igev0>10)igev0=10;
    if(igev1>10)igev1=10;
    double e0 = gev-epoint[igev0];
    double e1 = epoint[igev1] - gev;


    double sumlgprob = 0;
//    unsigned int totplane = 90; 
    unsigned int totplane = fShowerPClusterCol[i].size();
    unsigned int nonzeroplane = 0; 

/*
    unsigned int shnp = fShowerPClusterCol[i].size(); 
    if(int(jmshower::kElectron)==type)totplane=std::max(shnp,fHtExpPlaneE[iReg][igev1]);
    if(int(jmshower::kGamma)==type)totplane=std::max(shnp,fHtExpPlaneG[iReg][igev1]);
    if(int(jmshower::kMuon)==type)totplane=std::max(shnp,fHtExpPlaneMu[iReg][igev1]);
    if(int(jmshower::kPi0)==type)totplane=std::max(shnp,fHtExpPlanePi0[iReg][igev1]);
    if(int(jmshower::kHad)==type)totplane=std::max(shnp,fHtExpPlaneHad[iReg][igev1]);
    if(int(jmshower::kProton)==type)totplane=std::max(shnp,fHtExpPlaneP[iReg][igev1]);
    if(int(jmshower::kNeutron)==type)totplane=std::max(shnp,fHtExpPlaneN[iReg][igev1]);
    if(int(jmshower::kPion)==type)totplane=std::max(shnp,fHtExpPlanePi[iReg][igev1]);
    if(int(jmshower::kElectronQE)==type)totplane=std::max(shnp,fHtExpPlaneEqe[iReg][igev1]);
    if(int(jmshower::kElectronRES)==type)totplane=std::max(shnp,fHtExpPlaneEres[iReg][igev1]);
    if(int(jmshower::kElectronDIS)==type)totplane=std::max(shnp,fHtExpPlaneEdis[iReg][igev1]);
    if(int(jmshower::kElectronCOH)==type)totplane=std::max(shnp,fHtExpPlaneEcoh[iReg][igev1]);
*/

    
    //std::cout<<"type, energy, cos, np, totplane "<<type<<" "<<gev<<" "<<shower.Dir()[2]<<" "<<shnp<<" "<<totplane<<std::endl;

    for(unsigned int plane=0;plane<totplane;plane++){
//      TVector3 disttoedge = RecoJMShower::GetTrkPlaneDistToEdge(fShowerPClusterCol[i][plane].Plane(), shower);
//      double planerad = RecoJMShower::Radius(fShowerPClusterCol[i][plane], shower);
//      if(disttoedge.x()<2*planerad)continue;
//      if(disttoedge.y()<2*planerad)continue;
//      if(disttoedge.z()<2*planerad)continue;
//      double path = RecoJMShower::GetTrkHitPath(fShowerPClusterCol[i][plane].Plane(), 0, shower);
//      double ep = RecoJMShower::Energy(fShowerPClusterCol[i][plane],shower.ID());
//      double dedx = ep/path;
      double dedx = 0;
      if(plane<fShowerPClusterCol[i].size()){
        dedx = RecoJMShower::GetPlaneDedx(shower, plane);
      }
      double prob = 0;
//      unsigned int nmip = RecoJMShower::NMIPPlane(shower);
      prob = RecoJMShower::GetInterPlaneDedxProb(shower, plane, dedx, type, pos, iReg, igev0, igev1, e0, e1);
//      prob = RecoJMShower::GetPlaneDedxProb(shower, i, hiteweightmap, 2,plane, dedx, type);
      if(prob<0.0001)prob=0.0001;
      double lgprob = log(prob);
      if(pBadChannel==true){ //Only consider non-zero plane
        if(dedx>0.0000001){
          nonzeroplane++;
          sumlgprob+=lgprob;
        }
      }else{
        nonzeroplane++;
        sumlgprob+=lgprob;
      }
      //std::cout<<"type: "<<type<<" "<<plane<<" "<<prob<<" "<<sumlgprob<<std::endl;
    }
    if(pBadChannel==true){
      if(nonzeroplane>0)sumlgprob = sumlgprob/(double)nonzeroplane;
    }else{
      if(totplane>0)sumlgprob = sumlgprob/(double)totplane;
    }
//    std::cout<<"sumlgprob, totplane "<<sumlgprob<<" "<<totplane<<std::endl;
//    sumlgprob = sumlgprob/double(fShowerPClusterCol[i].size());
    if(sumlgprob>999)sumlgprob=999;
    if(sumlgprob<-999)sumlgprob=-999;
    //std::cout<<"DedxLongLL  type: "<<type<<" ll "<<sumlgprob<<std::endl;
    return sumlgprob;
  }

  double RecoJMShower::GetDedxInvLongLL(rb::Prong shower, const int type){
    int i = shower.ID();
    TVector3 pos((shower.Start()[0]+RecoJMShower::GetShwStop(shower)[0])/2.0,(shower.Start()[1]+RecoJMShower::GetShwStop(shower)[1])/2.0,(shower.Start()[2]+RecoJMShower::GetShwStop(shower)[2])/2.0);
    int iReg = 0;
    if(pos[0]>=0&&pos[1]>=0)iReg = 0;
    if(pos[0]>=0&&pos[1]<0)iReg = 1;
    if(pos[0]<0&&pos[1]>=0)iReg = 2;
    if(pos[0]<=0&&pos[1]<=0)iReg = 3;

    double gev = RecoJMShower::Energy(shower);
    double elower[11] = {0.00, 0.25, 0.75, 1.25, 1.75, 2.25, 2.75, 3.25, 3.75, 4.25, 4.75};
    double eupper[11] = {0.25, 0.75, 1.25, 1.75, 2.25, 2.75, 3.25, 3.75, 4.25, 4.75, 5.25};
    double epoint[11];
    for(int ig=0;ig<11;ig++){
      epoint[ig] = (elower[ig]+eupper[ig])/2.;
    }
    int igev0 = 0;
    int igev1 = 0;
    if(gev>epoint[10]){igev0=10;igev1=10;}
    else if(gev<epoint[0]){igev0=0;igev1=0;}
    else{
      for(int ig=0;ig<10;ig++){
        if(gev>epoint[ig]&&gev<epoint[ig+1]){igev0=ig;igev1=ig+1;}
      }
    }
    if(igev0>10)igev0=10;
    if(igev1>10)igev1=10;
    double e0 = gev-epoint[igev0];
    double e1 = epoint[igev1] - gev;


    double sumlgprob = 0;
//    unsigned int totplane = 90; 
    unsigned int totplane = fShowerPClusterCol[i].size();
    unsigned int nonzeroplane = 0; 

/*
    unsigned int shnp = fShowerPClusterCol[i].size(); 
    if(int(jmshower::kElectron)==type)totplane=std::max(shnp,fHtExpPlaneE[iReg][igev1]);
    if(int(jmshower::kGamma)==type)totplane=std::max(shnp,fHtExpPlaneG[iReg][igev1]);
    if(int(jmshower::kMuon)==type)totplane=std::max(shnp,fHtExpPlaneMu[iReg][igev1]);
    if(int(jmshower::kPi0)==type)totplane=std::max(shnp,fHtExpPlanePi0[iReg][igev1]);
    if(int(jmshower::kHad)==type)totplane=std::max(shnp,fHtExpPlaneHad[iReg][igev1]);
    if(int(jmshower::kProton)==type)totplane=std::max(shnp,fHtExpPlaneP[iReg][igev1]);
    if(int(jmshower::kNeutron)==type)totplane=std::max(shnp,fHtExpPlaneN[iReg][igev1]);
    if(int(jmshower::kPion)==type)totplane=std::max(shnp,fHtExpPlanePi[iReg][igev1]);
    if(int(jmshower::kElectronQE)==type)totplane=std::max(shnp,fHtExpPlaneEqe[iReg][igev1]);
    if(int(jmshower::kElectronRES)==type)totplane=std::max(shnp,fHtExpPlaneEres[iReg][igev1]);
    if(int(jmshower::kElectronDIS)==type)totplane=std::max(shnp,fHtExpPlaneEdis[iReg][igev1]);
    if(int(jmshower::kElectronCOH)==type)totplane=std::max(shnp,fHtExpPlaneEcoh[iReg][igev1]);
*/

    
    for(unsigned int plane=0;plane<totplane;plane++){
//      TVector3 disttoedge = RecoJMShower::GetTrkPlaneDistToEdge(fShowerPClusterCol[i][plane].Plane(), shower);
//      double planerad = RecoJMShower::Radius(fShowerPClusterCol[i][plane], shower);
//      if(disttoedge.x()<2*planerad)continue;
//      if(disttoedge.y()<2*planerad)continue;
//      if(disttoedge.z()<2*planerad)continue;
//      double path = RecoJMShower::GetTrkHitPath(fShowerPClusterCol[i][plane].Plane(), 0, shower);
//      double ep = RecoJMShower::Energy(fShowerPClusterCol[i][plane],shower.ID());
//      double dedx = ep/path;
      double dedx = 0;
      if(totplane-plane<fShowerPClusterCol[i].size()){
        dedx = RecoJMShower::GetPlaneDedx(shower, totplane-plane);
      }
      double prob = 0;
//      unsigned int nmip = RecoJMShower::NMIPPlane(shower);
      prob = RecoJMShower::GetInterPlaneDedxProb(shower, plane, dedx, type, pos, iReg, igev0, igev1, e0, e1);
//      prob = RecoJMShower::GetPlaneDedxProb(shower, i, hiteweightmap, 2,plane, dedx, type);
      if(prob<0.0001)prob=0.0001;
      double lgprob = log(prob);
      if(pBadChannel==true){ //Only consider non-zero plane
        if(dedx>0.0000001){
          nonzeroplane++;
          sumlgprob+=lgprob;
        }
      }else{
        nonzeroplane++;
        sumlgprob+=lgprob;
      }
    }
    if(pBadChannel==true){
      if(nonzeroplane>0)sumlgprob = sumlgprob/(double)nonzeroplane;
    }else{
      if(totplane>0)sumlgprob = sumlgprob/(double)totplane;
    }
//    sumlgprob = sumlgprob/double(fShowerPClusterCol[i].size());
    if(sumlgprob>999)sumlgprob=999;
    if(sumlgprob<-999)sumlgprob=-999;
    return sumlgprob;
  }

/*
  double RecoJMShower::GetDedxLongLL(rb::Prong shower, const int type, unsigned int startp, int contp,  art::Event const& evt){
    int i = shower.ID();
    double sumlgprob = 0;
    if(int(fShowerPClusterCol[i].size()-startp)<contp)return -999;
    unsigned int newstartp = RecoJMShower::ContStartPlane(shower, startp, contp);

    for(unsigned int plane= newstartp; plane<fShowerPClusterCol[i].size();plane++){
      TVector3 disttoedge = RecoJMShower::GetTrkPlaneDistToEdge(fShowerPClusterCol[i][plane].Plane(), shower);
//      double planerad = RecoJMShower::Radius(fShowerPClusterCol[i][plane], shower);
//      if(disttoedge.x()<2*planerad)continue;
//      if(disttoedge.y()<2*planerad)continue;
//      if(disttoedge.z()<2*planerad)continue;
//      double path = RecoJMShower::GetTrkHitPath(fShowerPClusterCol[i][plane].Plane(), 0, shower);
//      double ep = RecoJMShower::Energy(fShowerPClusterCol[i][plane],shower.ID());
//      double dedx = ep/path;
      double dedx = RecoJMShower::GetPlaneDedx(shower, plane);
      double prob = 0;
//      unsigned int nmip = RecoJMShower::NMIPPlane(shower, newstartp);
      if(newstartp==startp)prob = RecoJMShower::GetInterPlaneDedxProb(shower, plane, dedx, type);
      if(newstartp>startp)prob = RecoJMShower::GetInterPlaneDedxProb(shower, plane-newstartp, dedx, type);
//      prob = RecoJMShower::GetPlaneDedxProb(shower, i, hiteweightmap, 2,plane, dedx, type);
      if(prob<0.0001)prob = 0.0001;
      double lgprob = log(prob);
      sumlgprob+=lgprob;
//     std::cout<<"i, plane, dedx, prob "<<i<<" "<<plane<<" "<<dedx<<" "<<prob<<std::endl;
    }
    sumlgprob = sumlgprob/double(fShowerPClusterCol[i].size() - newstartp);
    return sumlgprob;
    if(sumlgprob>999)sumlgprob=999;
    if(sumlgprob<-999)sumlgprob=-999;
  }
*/

/*
  double RecoJMShower::GetCellTransDedx(rb::Prong shower, unsigned int cell){
    int i = shower.ID();
    double totpath = 0 ;
    for(unsigned int plane=0;plane<fShowerPClusterCol[i].size();plane++){
      double path = RecoJMShower::GetTrkHitPath(fShowerPClusterCol[i][plane].Plane(), 0, shower);
      totpath+=path;
    }
    double ep =   RecoJMShower::DepositEnergy(fShowerTCClusterCol[i][cell], i);
    double dedx = ep/totpath;
    if(isrealdata==true&&pCorDataDedx)dedx=dedx/RecoJMShower::CellTransDataMC(dedx, fDetid);
    return dedx;
  }
*/

  double RecoJMShower::GetCellTransDedx(rb::Prong shower, unsigned int cell){// calculate transverse cell dE/dx based on centroid
//if(cell!=0)return 0;
    int i = shower.ID();
    double totpath = 0 ;
    double ep = 0;  
    for(unsigned int plane=0;plane<fShowerPClusterCol[i].size();plane++){
      unsigned int absplane = fShowerPClusterCol[i][plane].Plane();
      art::ServiceHandle<geo::Geometry> geom;
      double sumPos = 0;
      double gev = 0;
      for (unsigned int nh=0;nh<fShowerPClusterCol[i][plane].NCell();nh++) {
        art::Ptr<rb::CellHit>  cellhit = fShowerPClusterCol[i][plane].Cell(nh);
        geo::OfflineChan key(cellhit->Plane(),cellhit->Cell());
        double cellEnergy = fHitEWeightMap[key][i];
        double cellPos = (double)cellhit->Cell() + 0.5;
        sumPos += cellEnergy*cellPos;
        gev += cellEnergy;
      }
      if(gev<0.00001)continue;
      double avgPos = sumPos/gev;
      int cid = (int)avgPos;// calculate centroid 
      double wm = avgPos - (double)cid;//energy splitting for the centroid cell
      double wp = 1-wm; //energy splitting for the centroid cell
      if(cid<0)cid=0;
      if(cid>(int)geom->Plane(absplane)->Ncells()-1)cid=(int)geom->Plane(absplane)->Ncells()-1;
      double dir = 1.;
      if(fabs(shower.Dir()[2])>0.000001)dir=fabs(shower.Dir()[2]);
//      double dx=shower.Dir()[0];// Validate dir
//      double dy=shower.Dir()[1];
//      double dz=shower.Dir()[2];
//      if(geom->Plane(absplane)->View()==geo::kX&&dx*dx+dz*dz>1e-10)dir=fabs(dz/sqrt(dx*dx+dz*dz));
//      if(geom->Plane(absplane)->View()==geo::kY&&dy*dy+dz*dz>1e-10)dir=fabs(dz/sqrt(dy*dy+dz*dz));
      int ic = (int)cell + 1;
      double dph = ((double)ic/dir)+wm;//upper limit for cells included in a transeverse cell index
      double dpl = ((double)(ic-1)/dir)+wm;//lower limit for cells included in a transeverse cell index
      if(cid+(int)dph<(int)geom->Plane(plane)->Ncells()){
        double eps1 = dph-(int)dph;
        double eps2 = (int)dpl+1-dpl;
        geo::OfflineChan keye1(absplane, cid+(int)dph);
        geo::OfflineChan keye2(absplane, cid+(int)dpl);
        double e1 = 0; 
        double e2 = 0; 
        double e = 0;
        if((int)fHitEWeightMap[keye1].size()>i)e1 = eps1*fHitEWeightMap[keye1][i];
        if((int)fHitEWeightMap[keye2].size()>i)e2 = eps2*fHitEWeightMap[keye2][i];
        int nc = 0;
        if((int)dph-(int)dpl>1){
          for(int ii=(int)dph-1;ii>=(int)dpl+1;ii--){
            geo::OfflineChan keye(absplane,cid+ii);
            if((int)fHitEWeightMap[keye].size()>i)e+=fHitEWeightMap[keye][i];
            nc++;
          }
        }
        double path = RecoJMShower::GetTrkHitPath(fShowerPClusterCol[i][plane].Plane(), 0, shower);
        totpath+=path;
        ep+=e1+e2+e;
// Simple calculation by using cellE/path
//        double ecid = 0;
//        geo::OfflineChan keycid(absplane,cid);
//        if((int)fHitEWeightMap[keycid].size()>i)ecid = fHitEWeightMap[keycid][i];
//         double dir3 = 1;
//         if(fabs(shower.Dir()[2])>0.000001)dir3=shower.Dir()[2];
//        ep+=ecid/fabs(dir3)/2.;
      }
      double dmh = ((double)ic/dir)+wp;//upper limit for cells included in a transeverse cell index
      double dml = ((double)(ic-1)/dir)+wp;//lower limit for cells included in a transeverse cell index

      if(cid-(int)dmh>=0){
        double eps1 = dmh-(int)dmh;
        double eps2 = (int)dml+1-dml;
        geo::OfflineChan keye1(absplane, cid-(int)dmh);
        geo::OfflineChan keye2(absplane, cid-(int)dml);
        double e1 = 0;
        double e2 = 0;
        double e = 0;
        if((int)fHitEWeightMap[keye1].size()>i)e1 = eps1*fHitEWeightMap[keye1][i];
        if((int)fHitEWeightMap[keye2].size()>i)e2 = eps2*fHitEWeightMap[keye2][i];
        int nc = 0;
        if((int)dmh-(int)dml>1){
          for(int ii=(int)dmh-1;ii>=(int)dml+1;ii--){
            geo::OfflineChan keye(absplane,cid-ii);
            if((int)fHitEWeightMap[keye].size()>i)e+=fHitEWeightMap[keye][i];
            nc++;
          }
        }
        ep+=e1+e2+e;

// Simple calculation by using cellE/path
//      double ecid = 0;
//      geo::OfflineChan keycid(absplane,cid);
//      if(fHitEWeightMap[keycid].size()>i)ecid = fHitEWeightMap[keycid][i];
//      double dir3 = 1;
//      if(fabs(shower.Dir()[2])>0.000001)dir3=shower.Dir()[2];
//      ep+=ecid/fabs(dir3)/2.;

        double path = RecoJMShower::GetTrkHitPath(fShowerPClusterCol[i][plane].Plane(), 0, shower);
        totpath+=path;
      }
    }
//    double ep =   RecoJMShower::DepositEnergy(fShowerTCClusterCol[i][cell], i);
    double dedx = -1;
    if(totpath>0.00001)dedx = ep/(totpath/2.);
    if(isrealdata==true&&pCorDataDedx)dedx=dedx/RecoJMShower::CellTransDataMC(dedx, fDetid);
    //std::cout<<"jm trans "<<cell<<" p "<<totpath/2.<<" ep "<<ep<<" dedx "<<dedx<<std::endl;
    return dedx;
  }

  double RecoJMShower::GetCellTransDedxProb(rb::Prong shower, int ireg, int igev, unsigned int cell, double dedx, const int type){
    if(ireg<0||ireg>3)return 1;
    if(igev>=11||igev<0)return 1;
    if(cell>=20)return 1;

//    if(dedx<=0.00003)return 1;
    double p = 0;
    if(int(jmshower::kElectron)==type){
      double ntot = double(fHtCellTransDedxE[ireg][igev][cell]->GetEntries());
      double n = double(fHtCellTransDedxE[ireg][igev][cell]->Interpolate(dedx));
//      if(dedx<=0.00003)n = double(fHtCellTransDedxE[ireg][igev][cell]->GetBinContent(1));
      double nbin = fHtCellTransDedxE[ireg][igev][cell]->GetNbinsX();      
      if(n<0.0001)n=0.0001;
      if(ntot<1){
        p = 0;
      }else{
        p = n*nbin/ntot;
      }
    }
    if(int(jmshower::kGamma)==type){
      double ntot = double(fHtCellTransDedxG[ireg][igev][cell]->GetEntries());
      double n = double(fHtCellTransDedxG[ireg][igev][cell]->Interpolate(dedx));
//      if(dedx<=0.00003)n = double(fHtCellTransDedxG[ireg][igev][cell]->GetBinContent(1));
      double nbin = fHtCellTransDedxG[ireg][igev][cell]->GetNbinsX();      
      if(n<0.0001)n=0.0001;
      if(ntot<1){
        p = 0;
      }else{
        p = n*nbin/ntot;
      }
    }

    if(int(jmshower::kMuon)==type){
      double ntot = double(fHtCellTransDedxMu[ireg][igev][cell]->GetEntries());
      double n = double(fHtCellTransDedxMu[ireg][igev][cell]->Interpolate(dedx));
 //     if(dedx<=0.00003)n = double(fHtCellTransDedxMu[ireg][igev][cell]->GetBinContent(1));
      double nbin = fHtCellTransDedxMu[ireg][igev][cell]->GetNbinsX();      
      if(n<0.0001)n=0.0001;
      if(ntot<1){
        p = 0;
      }else{
        p = n*nbin/ntot;
      }
    }

    if(int(jmshower::kPi0)==type){
      double ntot = double(fHtCellTransDedxPi0[ireg][igev][cell]->GetEntries());
      double n = double(fHtCellTransDedxPi0[ireg][igev][cell]->Interpolate(dedx));
//      if(dedx<=0.00003)n = double(fHtCellTransDedxPi0[ireg][igev][cell]->GetBinContent(1));
//      double n = double(fHtCellTransDedxPi0[ireg][igev][cell]->GetBinContent(fHtCellTransDedxPi0[ireg][igev][cell]->FindBin(dedx)));
//      int nnbin = fHtCellTransDedxPi0[ireg][igev][cell]->FindBin(dedx);
      double nbin = fHtCellTransDedxPi0[ireg][igev][cell]->GetNbinsX();
      if(n<0.0001)n=0.0001;
      if(ntot<1){
        p = 0;
      }else{
        p = n*nbin/ntot;
      }
    }
/*
    if(int(jmshower::kHad)==type){
      double ntot = double(fHtCellTransDedxHad[ireg][igev][cell]->GetEntries());
      double n = double(fHtCellTransDedxHad[ireg][igev][cell]->Interpolate(dedx));
//      if(dedx<=0.00003)n = double(fHtCellTransDedxHad[ireg][igev][cell]->GetBinContent(1));
//      double n = double(fHtCellTransDedxHad[ireg][igev][cell]->GetBinContent(fHtCellTransDedxHad[ireg][igev][cell]->FindBin(dedx)));
//      int nnbin = fHtCellTransDedxHad[ireg][igev][cell]->FindBin(dedx);
      double nbin = fHtCellTransDedxHad[ireg][igev][cell]->GetNbinsX();
      if(n<0.0001)n=0.0001;
      if(ntot<1){
        p = 0;
      }else{
        p = n*nbin/ntot;
      }
    }
*/



    if(int(jmshower::kProton)==type){
      double ntot = double(fHtCellTransDedxP[ireg][igev][cell]->GetEntries());
      double n = double(fHtCellTransDedxP[ireg][igev][cell]->Interpolate(dedx));
//      if(dedx<=0.00003)n = double(fHtCellTransDedxP[ireg][igev][cell]->GetBinContent(1));
//      double n = double(fHtCellTransDedxP[ireg][igev][cell]->GetBinContent(fHtCellTransDedxP[ireg][igev][cell]->FindBin(dedx)));
//      int nnbin = fHtCellTransDedxP[ireg][igev][cell]->FindBin(dedx);
      double nbin = fHtCellTransDedxP[ireg][igev][cell]->GetNbinsX();
      if(n<0.0001)n=0.0001;
      if(ntot<1){
        p = 0;
      }else{
        p = n*nbin/ntot;
      }
    }


    if(int(jmshower::kNeutron)==type){
      double ntot = double(fHtCellTransDedxN[ireg][igev][cell]->GetEntries());
      double n = double(fHtCellTransDedxN[ireg][igev][cell]->Interpolate(dedx));
//      if(dedx<=0.00003)n = double(fHtCellTransDedxN[ireg][igev][cell]->GetBinContent(1));
//      double n = double(fHtCellTransDedxN[ireg][igev][cell]->GetBinContent(fHtCellTransDedxN[ireg][igev][cell]->FindBin(dedx)));
//      int nnbin = fHtCellTransDedxN[ireg][igev][cell]->FindBin(dedx);
      double nbin = fHtCellTransDedxN[ireg][igev][cell]->GetNbinsX();
      if(n<0.0001)n=0.0001;
      if(ntot<1){
       p = 0;
      }else{
        p = n*nbin/ntot;
      }
    }

    if(int(jmshower::kPion)==type){
      double ntot = double(fHtCellTransDedxPi[ireg][igev][cell]->GetEntries());
      double n = double(fHtCellTransDedxPi[ireg][igev][cell]->Interpolate(dedx));
//      if(dedx<=0.00003)n = double(fHtCellTransDedxPi[ireg][igev][cell]->GetBinContent(1));
//      double n = double(fHtCellTransDedxPi[ireg][igev][cell]->GetBinContent(fHtCellTransDedxPi[ireg][igev][cell]->FindBin(dedx)));
//      int nnbin = fHtCellTransDedxPi[ireg][igev][cell]->FindBin(dedx);
      double nbin = fHtCellTransDedxPi[ireg][igev][cell]->GetNbinsX();
      if(n<0.0001)n=0.0001;
      if(ntot<1){
       p = 0;
      }else{
        p = n*nbin/ntot;
      }
    }

    if(int(jmshower::kElectronQE)==type){
      double ntot = double(fHtCellTransDedxEqe[ireg][igev][cell]->GetEntries());
      double n = double(fHtCellTransDedxEqe[ireg][igev][cell]->Interpolate(dedx));
      double nbin = fHtCellTransDedxEqe[ireg][igev][cell]->GetNbinsX();
      if(n<0.0001)n=0.0001;
      if(ntot<1){
       p = 0;

      }else{
        p = n*nbin/ntot;
      }
    }

    if(int(jmshower::kElectronRES)==type){
      double ntot = double(fHtCellTransDedxEres[ireg][igev][cell]->GetEntries());
      double n = double(fHtCellTransDedxEres[ireg][igev][cell]->Interpolate(dedx));
      double nbin = fHtCellTransDedxEres[ireg][igev][cell]->GetNbinsX();
      if(n<0.0001)n=0.0001;
      if(ntot<1){
       p = 0;
      }else{
        p = n*nbin/ntot;
      }
    }

    if(int(jmshower::kElectronDIS)==type){
      double ntot = double(fHtCellTransDedxEdis[ireg][igev][cell]->GetEntries());
      double n = double(fHtCellTransDedxEdis[ireg][igev][cell]->Interpolate(dedx));
      double nbin = fHtCellTransDedxEdis[ireg][igev][cell]->GetNbinsX();
      if(n<0.0001)n=0.0001;
      if(ntot<1){
       p = 0;
      }else{
        p = n*nbin/ntot;
      }
    }

    if(int(jmshower::kElectronCOH)==type){
      double ntot = double(fHtCellTransDedxEcoh[ireg][igev][cell]->GetEntries());
      double n = double(fHtCellTransDedxEcoh[ireg][igev][cell]->Interpolate(dedx));
      double nbin = fHtCellTransDedxEcoh[ireg][igev][cell]->GetNbinsX();
      if(n<0.0001)n=0.0001;
      if(ntot<1){
       p = 0;
      }else{
        p = n*nbin/ntot;
      }
    }

    return p;
  }

  double RecoJMShower::GetInterCellTransDedxProb(rb::Prong shower, unsigned int cell, double dedx, const int type, TVector3 pos, int iReg, int igev0, int igev1, double e0, double e1){
/*
    TVector3 pos((shower.Start()[0]+RecoJMShower::GetShwStop(shower)[0])/2.0,(shower.Start()[1]+RecoJMShower::GetShwStop(shower)[1])/2.0,(shower.Start()[2]+RecoJMShower::GetShwStop(shower)[2])/2.0);
    int ireg = 0;
    if(pos[0]>=0&&pos[1]>=0)ireg = 0;
    if(pos[0]>=0&&pos[1]<0)ireg = 1;
    if(pos[0]<0&&pos[1]>=0)ireg = 2;
    if(pos[0]<=0&&pos[1]<=0)ireg = 3;
*/
    art::ServiceHandle<geo::Geometry> geom;

    double probPXPY = 0;
    bool interPos = false;


    if(interPos==true){
      double X0 = -geom->DetHalfWidth()/2.0;
      double X1 = geom->DetHalfWidth()/2.0;
      double Y0 = -geom->DetHalfHeight()/2.0;
      double Y1 = geom->DetHalfHeight()/2.0;

      double probreg[4]={1,1,1,1};
      for(int ireg=0;ireg<4;ireg++){
        double probe0 = RecoJMShower::GetCellTransDedxProb(shower, ireg, igev0, cell, dedx, type);
        double probe1 = RecoJMShower::GetCellTransDedxProb(shower, ireg, igev1, cell, dedx, type);
        if(igev0!=igev1)probreg[ireg] = (e1*probe0+e0*probe1)/(e0+e1);
        if(igev0==igev1)probreg[ireg] = probe0;
      }

      double PX = pos[0];
      double PY = pos[1];
      double probX1Y1 = probreg[0];
      double probX1Y0 = probreg[1];
      double probX0Y1 = probreg[2];
      double probX0Y0 = probreg[3];
      double probX1PY = ((PY-Y0)/(Y1-Y0))*(probX1Y1-probX1Y0)+probX1Y0;
      double probX0PY = ((PY-Y0)/(Y1-Y0))*(probX0Y1-probX0Y0)+probX0Y0; 
      probPXPY = ((PX-X0)/(X1-X0))*(probX1PY-probX0PY)+probX0PY;
    }

    if(interPos==false||probPXPY<0){
      double probe0 = RecoJMShower::GetCellTransDedxProb(shower, iReg, igev0, cell, dedx, type);
      double probe1 = RecoJMShower::GetCellTransDedxProb(shower, iReg, igev1, cell, dedx, type);
      if(igev0!=igev1)probPXPY = (e1*probe0+e0*probe1)/(e0+e1);
      if(igev0==igev1)probPXPY = probe0;
    }

    return probPXPY;
  }

  double RecoJMShower::GetDedxTransLL(rb::Prong shower, const int type){
//    int i = shower.ID();
    double transcelldedx[20];
    for(int itd=0;itd<20;itd++){
      transcelldedx[itd] = 0;
    }
    for(unsigned itp = 0;itp<20;itp++){
      transcelldedx[itp] =  RecoJMShower::GetCellTransDedx(shower, itp);
    }
    double sumlgprob = 0;

    TVector3 pos((shower.Start()[0]+RecoJMShower::GetShwStop(shower)[0])/2.0,(shower.Start()[1]+RecoJMShower::GetShwStop(shower)[1])/2.0,(shower.Start()[2]+RecoJMShower::GetShwStop(shower)[2])/2.0);
    int iReg = 0;
    if(pos[0]>=0&&pos[1]>=0)iReg = 0;
    if(pos[0]>=0&&pos[1]<0)iReg = 1;
    if(pos[0]<0&&pos[1]>=0)iReg = 2;
    if(pos[0]<=0&&pos[1]<=0)iReg = 3;


    double gev = RecoJMShower::Energy(shower);
    double elower[11] = {0.00, 0.25, 0.75, 1.25, 1.75, 2.25, 2.75, 3.25, 3.75, 4.25, 4.75};
    double eupper[11] = {0.25, 0.75, 1.25, 1.75, 2.25, 2.75, 3.25, 3.75, 4.25, 4.75, 5.25};
    double epoint[11];
    for(int ig=0;ig<11;ig++){
      epoint[ig] = (elower[ig]+eupper[ig])/2.;
    }
    int igev0 = 0;
    int igev1 = 0;
    if(gev>epoint[10]){igev0=10;igev1=10;}
    else if(gev<epoint[0]){igev0=0;igev1=0;}
    else{
      for(int ig=0;ig<10;ig++){
        if(gev>epoint[ig]&&gev<epoint[ig+1]){igev0=ig;igev1=ig+1;}
      }
    }
    if(igev0>10)igev0=10;
    if(igev1>10)igev1=10;

    double e0 = gev-epoint[igev0];
    double e1 = epoint[igev1] - gev;

    int ntcell = 0;
    for(int itd=0;itd<20;itd++){
      if(transcelldedx[itd]<-0.5)continue;
      double prob = 0;
      prob = RecoJMShower::GetInterCellTransDedxProb(shower, itd, transcelldedx[itd], type, pos, iReg, igev0, igev1, e0, e1);

      if(prob<0.0001)prob=0.0001;
      double lgprob = log(prob);
      sumlgprob+=lgprob;
      ntcell++;
    }
    if(ntcell>0)sumlgprob = sumlgprob/(double)ntcell;
//    if(ntcell>0)sumlgprob = sumlgprob/20.;
    if(sumlgprob>999)sumlgprob=999;
    if(sumlgprob<-999)sumlgprob=-999;
    //std::cout<<"trans ll prob "<<sumlgprob<<std::endl;
    return sumlgprob;
  }


  unsigned int RecoJMShower::NMIPPlane(rb::Prong shower){
    int i = shower.ID();
    unsigned int nmipplane = 0;

    unsigned int extp = 20;
    if(fShowerPClusterCol[i].size()<20) extp = fShowerPClusterCol[i].size();
    for(unsigned int plane=0;plane<extp;plane++){
      if(RecoJMShower::GetPlaneDedx(shower, plane)>0.005||RecoJMShower::GetPlaneDedx(shower, plane)<0.0005)continue;
      nmipplane++;
    }


/*
    unsigned int extp = 15;
    if(fShowerPClusterCol[i].size()<15)extp = fShowerPClusterCol[i].size();
    for(unsigned int plane=0;plane<extp; plane++){
      if(plane>=extp/2)break;
      if(RecoJMShower::GetPlaneDedx(shower, plane)>0.005||RecoJMShower::GetPlaneDedx(shower, plane)<0.0005)continue;;
      nmipplane++;
    }
*/
/*
    unsigned int mipstartp = 0;
    for(unsigned int plane=0;plane<fShowerPClusterCol[i].size();plane++){
      if(RecoJMShower::GetPlaneDedx(shower, plane)<0.005&&RecoJMShower::GetPlaneDedx(shower, plane)>0.0005){
        mipstartp = plane;
        break;
      }
    }

    for(unsigned int plane = mipstartp;plane<fShowerPClusterCol[i].size();plane++){
      if(RecoJMShower::GetPlaneDedx(shower, plane)>0.005)break;
      if(RecoJMShower::GetPlaneDedx(shower, plane)<0.0005)continue;
      nmipplane++;
    }
*/
    return nmipplane;
  }


  unsigned int RecoJMShower::NMIPPlane(rb::Prong shower, unsigned int startp){
    int i = shower.ID();
    unsigned int nmipplane = startp;
    for(unsigned int plane = startp; plane<fShowerPClusterCol[i].size();plane++){
      if(RecoJMShower::GetPlaneDedx(shower, plane)>0.005)break;
      nmipplane++;
    }
    return nmipplane - startp;
  }

  bool RecoJMShower::IsMuon(jmshower::JMShower shower){
    double ellmax = -9999;
    double elllmax = -9999;
    double elltmax = -9999;
    for(int itype = int(jmshower::kNull)+1; itype != int(jmshower::kLast); ++itype){
      if(itype == int(jmshower::kElectron)||itype == int(jmshower::kElectronQE)||itype == int(jmshower::kElectronRES)||itype == int(jmshower::kElectronDIS)||itype == int(jmshower::kElectronCOH)){
        if(ellmax<shower.DedxLongLL(itype)+shower.DedxTransLL(itype)){
          ellmax = shower.DedxLongLL(itype)+shower.DedxTransLL(itype);
          elllmax = shower.DedxLongLL(itype);
          elltmax = shower.DedxTransLL(itype);
        }
      }
    }
    double emuLLL = elllmax - shower.DedxLongLL(int(jmshower::kMuon));
    double emuLLT = elltmax - shower.DedxTransLL(int(jmshower::kMuon));
    if(emuLLL>-0.2&&emuLLT>-99999){
      return kFALSE;
    }else{
      return kTRUE;
    }
  }

  bool RecoJMShower::IsInvPhoton(jmshower::JMShower shower){
    if(shower.DedxLongLL(int(jmshower::kGamma))<shower.DedxInvLongLL(int(jmshower::kGamma))){
      return kTRUE;
    }else{
      return kFALSE;
    }
  }

  double RecoJMShower::CalcANN(unsigned int mode, jmshower::JMShower shower){
//    if(mode!=0&&mode!=1&&mode!=2&&&&mode!=3&&&&mode!=4) return -99;
    double ellmax = -9999;
    double elllmax = -9999;
    double elltmax = -9999;

    for(int itype = int(jmshower::kNull)+1; itype != int(jmshower::kLast); ++itype){
      if(itype == int(jmshower::kElectron)||itype == int(jmshower::kElectronQE)||itype == int(jmshower::kElectronRES)||itype == int(jmshower::kElectronDIS)||itype == int(jmshower::kElectronCOH)){
        if(ellmax<shower.DedxLongLL(itype)+shower.DedxTransLL(itype)){
          ellmax = shower.DedxLongLL(itype)+shower.DedxTransLL(itype);
          elllmax = shower.DedxLongLL(itype);
          elltmax = shower.DedxTransLL(itype);
        }
      }
    }

    double egLLL = elllmax - shower.DedxLongLL(int(jmshower::kGamma)); 
    double egLLT = elltmax - shower.DedxTransLL(int(jmshower::kGamma));
    double emuLLL = elllmax - shower.DedxLongLL(int(jmshower::kMuon));
    double emuLLT = elltmax - shower.DedxTransLL(int(jmshower::kMuon));
    double epi0LLL = elllmax - shower.DedxLongLL(int(jmshower::kPi0)); 
    double epi0LLT = elltmax - shower.DedxTransLL(int(jmshower::kPi0)); 
    double epLLL = elllmax - shower.DedxLongLL(int(jmshower::kProton));
    double epLLT = elltmax - shower.DedxTransLL(int(jmshower::kProton));
    double enLLL = elllmax - shower.DedxLongLL(int(jmshower::kNeutron));
    double enLLT = elltmax - shower.DedxTransLL(int(jmshower::kNeutron));
    double epiLLL = elllmax - shower.DedxLongLL(int(jmshower::kPion));
    double epiLLT = elltmax - shower.DedxTransLL(int(jmshower::kPion));
    double vtxgev = shower.VtxGeV();
    double pi0mass = std::max(shower.Pi0Mgg(),0.);
    double shE = 0;
    if(shower.SliceEtot()>1e-10)shE=shower.Energy()/shower.SliceEtot();//Use calibrated cell energy on showers.
//    if(shower.SliceGeV()>1e-10)shE=shower.Energy()/shower.SliceGeV();// Use total cell energy, calibrated by the slice mean position
    double nuE = shower.NueEnergy(); 
//    double nuE = (shower.Energy() + 0.282525 + 1.0766*(shower.SliceEtot()-shower.Energy()))/1.015;
//    double nuE = (shower.Energy() + 0.282525 + 1.0766*(shower.SliceGeV()-shower.DepositEnergy()))/1.015;
    double gap = shower.Gap();
    double  costheta = cos(shower.Theta());
//    double length = 0;
//    if(shower.Energy()>1e-10)length = shower.TotalLength()/shower.Energy();

    Double_t params_shape[17];
    for(int i=0;i<17;i++){
      params_shape[i]=0;    
    }
    params_shape[0] = egLLL;
    params_shape[1] = egLLT;
    params_shape[2] = emuLLL;
    params_shape[3] = emuLLT;
    params_shape[4] = epi0LLL; 
    params_shape[5] = epi0LLT; 
    params_shape[6] = epLLL; 
    params_shape[7] = epLLT; 
    params_shape[8] = enLLL; 
    params_shape[9] = enLLT; 
    params_shape[10] = epiLLL; 
    params_shape[11] = epiLLT; 
    params_shape[12] = pi0mass; 
    params_shape[13] = shE; 
    params_shape[14] = vtxgev; 
    params_shape[15] = gap; 
    params_shape[16] = costheta; 


    Double_t params_shape_NoCos[16];
    for(int i=0;i<16;i++){
      params_shape_NoCos[i]=0;    
    }
    params_shape_NoCos[0] = egLLL;
    params_shape_NoCos[1] = egLLT;
    params_shape_NoCos[2] = emuLLL;
    params_shape_NoCos[3] = emuLLT;
    params_shape_NoCos[4] = epi0LLL; 
    params_shape_NoCos[5] = epi0LLT; 
    params_shape_NoCos[6] = epLLL; 
    params_shape_NoCos[7] = epLLT; 
    params_shape_NoCos[8] = enLLL; 
    params_shape_NoCos[9] = enLLT; 
    params_shape_NoCos[10] = epiLLL; 
    params_shape_NoCos[11] = epiLLT; 
    params_shape_NoCos[12] = pi0mass; 
    params_shape_NoCos[13] = shE; 
    params_shape_NoCos[14] = vtxgev; 
    params_shape_NoCos[15] = gap; 

    Double_t params_elec[12];
    for(int i=0;i<12;i++){
      params_elec[i]=0;    
    }
    params_elec[0] = egLLL;
    params_elec[1] = egLLT;
    params_elec[2] = epi0LLL; 
    params_elec[3] = epi0LLT; 
    params_elec[4] = epLLL; 
    params_elec[5] = epLLT; 
    params_elec[6] = enLLL; 
    params_elec[7] = enLLT; 
    params_elec[8] = epiLLL; 
    params_elec[9] = epiLLT; 
    params_elec[10] = shE; 
    params_elec[11] = gap; 

    Double_t params_shape_E[18];
    for(int i=0;i<18;i++){
      params_shape_E[i]=0;    
    }
    params_shape_E[0] = egLLL;
    params_shape_E[1] = egLLT;
    params_shape_E[2] = emuLLL;
    params_shape_E[3] = emuLLT;
    params_shape_E[4] = epi0LLL; 
    params_shape_E[5] = epi0LLT; 
    params_shape_E[6] = epLLL; 
    params_shape_E[7] = epLLT; 
    params_shape_E[8] = enLLL; 
    params_shape_E[9] = enLLT; 
    params_shape_E[10] = epiLLL; 
    params_shape_E[11] = epiLLT; 
    params_shape_E[12] = pi0mass; 
    params_shape_E[13] = shE; 
    params_shape_E[14] = vtxgev; 
    params_shape_E[15] = gap; 
    params_shape_E[16] = costheta; 
    params_shape_E[17] = nuE; 


    Double_t params_shape_E_NoCos[17];
    for(int i=0;i<17;i++){
      params_shape_E_NoCos[i]=0;    
    }
    params_shape_E_NoCos[0] = egLLL;
    params_shape_E_NoCos[1] = egLLT;
    params_shape_E_NoCos[2] = emuLLL;
    params_shape_E_NoCos[3] = emuLLT;
    params_shape_E_NoCos[4] = epi0LLL; 
    params_shape_E_NoCos[5] = epi0LLT; 
    params_shape_E_NoCos[6] = epLLL; 
    params_shape_E_NoCos[7] = epLLT; 
    params_shape_E_NoCos[8] = enLLL; 
    params_shape_E_NoCos[9] = enLLT; 
    params_shape_E_NoCos[10] = epiLLL; 
    params_shape_E_NoCos[11] = epiLLT; 
    params_shape_E_NoCos[12] = pi0mass; 
    params_shape_E_NoCos[13] = shE; 
    params_shape_E_NoCos[14] = vtxgev; 
    params_shape_E_NoCos[15] = gap; 
    params_shape_E_NoCos[16] = nuE; 

/*

    Double_t params_shape[18];
    for(int i=0;i<18;i++){
      params_shape[i]=0;    
    }
    params_shape[0] = egLLL;
    params_shape[1] = egLLT;
    params_shape[2] = emuLLL;
    params_shape[3] = emuLLT;
    params_shape[4] = epi0LLL; 
    params_shape[5] = epi0LLT; 
    params_shape[6] = epLLL; 
    params_shape[7] = epLLT; 
    params_shape[8] = enLLL; 
    params_shape[9] = enLLT; 
    params_shape[10] = epiLLL; 
    params_shape[11] = epiLLT; 
    params_shape[12] = pi0mass; 
    params_shape[13] = shE; 
    params_shape[14] = gap; 
    params_shape[15] = costheta; 
    params_shape[16] = vtxgev; 
    params_shape[17] = length; 

*/

/*
    Double_t params_shape[11];
    for(int i=0;i<11;i++){
      params_shape[i]=0;    
    }
    params_shape[0] = egLLL;
    params_shape[1] = egLLT;
    params_shape[2] = epi0LLL; 
    params_shape[3] = epi0LLT; 
    params_shape[4] = epLLL; 
    params_shape[5] = epLLT; 
    params_shape[6] = enLLL; 
    params_shape[7] = enLLT; 
    params_shape[8] = epiLLL; 
    params_shape[9] = epiLLT; 
    params_shape[10] = pi0mass; 
    params_shape[11] = vtxgev; 
*/
    if(mode==0){
      double ann = mlp_shape->Evaluate(0,params_shape);
//      std::cout<<"mode, ann "<<mode<<" "<<ann<<std::endl;
      return ann;
    }

    if(mode==1){
      double ann = mlp_shape_elec->Evaluate(0,params_elec);
//      std::cout<<"mode, ann "<<mode<<" "<<ann<<std::endl;
      return ann;
    }

    if(mode==2){
      double ann = mlp_shape_E->Evaluate(0,params_shape_E);
//      std::cout<<"mode, ann "<<mode<<" "<<ann<<std::endl;
      return ann;
    }

    if(mode==3){
      double ann = mlp_shape_NoCos->Evaluate(0,params_shape_NoCos);
      return ann;
    } 

    if(mode==4){
      double ann = mlp_shape_E_NoCos->Evaluate(0,params_shape_E_NoCos);
      return ann;
    }


    double ann0 = mlp_shape->Evaluate(0,params_shape);
    return ann0; 
/*
    double ann[3] = {-99,-99,-99};
    ann[0] = mlp_shape_QE->Evaluate(0,params_shape);
    ann[1] = mlp_shape_RES->Evaluate(0,params_shape);
    ann[2] = mlp_shape_DIS->Evaluate(0,params_shape);

    double annmax=-99;
    for(int ii=0;ii<3;ii++){
      if(annmax<ann[ii]){
        annmax = ann[ii];
      }
    }
*/
  }

//*******************************************************************************
  //function to compute and store rvp like variables for slice.  this could be used in future eid versions
  RVPInfo RecoJMShower::GetRVPStats(art::Handle<std::vector<rb::Cluster> > slices, unsigned int ic)
  {
    art::ServiceHandle<calib::Calibrator> cal;

    const rb::Cluster& slice = (*slices)[ic];

    RVPInfo rvp;
    rvp.efrac_2slides=1.;
    rvp.efrac_4slides=1.;
    rvp.efrac_6slides=1.;
    rvp.efrac_20slides=1.;
    rvp.efrac_2sigmaHOR=1.;
    rvp.efrac_4sigmaHOR=1.;
    rvp.eiso_2sigmaHOR=0.;
    rvp.eiso_4sigmaHOR=0.;
    rvp.eiso_6sigmaHOR=0.;

    const int lengthPlanes = slice.ExtentPlane();
    double recoEnergy=0.;
    for( unsigned int icell=0; icell<slice.NCell(); ++ icell){
     const art::Ptr<rb::CellHit>& chit = slice.Cell(icell);
     const TVector3 mean = slice.MeanXYZ();
     const double wx = (slice.NXCell() > 0) ? slice.W(slice.XCell(0).get()) : 0;
     const double wy = (slice.NYCell() > 0) ? slice.W(slice.YCell(0).get()) : 0;

     const rb::RecoHit rhit(cal->MakeRecoHit(*chit, chit->View() == geo::kX ? wx : wy));

     if( !rhit.IsCalibrated() ) continue;
     recoEnergy += rhit.GeV();
    }//loop all cells

    //calculate per plane cell energy
    std::vector<double> PlaneSumEnergy;
    PlaneSumEnergy.clear();
    double totalE_road20=0.;
    double totalE_road40=0.;
    double totalE_road60=0.;

    for( int iplane=0; iplane<lengthPlanes; ++iplane ){
      double cellEnergy=0.;//sum of cells energy in each plane
      double cellX=0.;//energetic cell X in each plane
      double cellY=0.;//energetic cell Y in each plane
      double Ecell=0.;
      double EXcell=0.;
      double EYcell=0.;

      std::vector<double> Xxview;
      std::vector<double> Exview;
      std::vector<double> Yyview;
      std::vector<double> Eyview;
      Xxview.clear();
      Exview.clear();
      Yyview.clear();
      Eyview.clear();

      if( slice.NCell()>0 ){
        for( unsigned int icell=0; icell<slice.NCell(); ++ icell){
          const art::Ptr<rb::CellHit>& chit = slice.Cell(icell);
          const TVector3 mean = slice.MeanXYZ();
          const double wx = (slice.NXCell() > 0) ? slice.W(slice.XCell(0).get()) : 0;
          const double wy = (slice.NYCell() > 0) ? slice.W(slice.YCell(0).get()) : 0;
          const rb::RecoHit rhit(cal->MakeRecoHit(*chit, chit->View() == geo::kX ? wx : wy));

          if( !rhit.IsCalibrated() ) continue;

          int Dplane=chit->Plane()-slice.MinPlane();

          if( Dplane==iplane ){
            cellEnergy += rhit.GeV();
            Ecell += rhit.GeV();
            EXcell += rhit.GeV()*rhit.X();
            EYcell += rhit.GeV()*rhit.Y();

            if( chit->View() == geo::kX ){
              Xxview.push_back(rhit.X());
              Exview.push_back(rhit.GeV());
            }//in X view
            else {
              Yyview.push_back(rhit.Y());
              Eyview.push_back(rhit.GeV());
            }//in Y view
          }
        }//loop cells
      }//has cells 
      if( Ecell>0. ){
        cellX = EXcell/Ecell;
        cellY = EYcell/Ecell;
      }

      PlaneSumEnergy.push_back(cellEnergy);
      
      int Nxviews=Xxview.size();
      int Nyviews=Yyview.size();
      if( Nxviews>0 ){
        for( int ixv=0; ixv<Nxviews; ++ixv ){
          if( fabs(Xxview[ixv]-cellX) < 2.*2. ) totalE_road20+=Exview[ixv];
          if( fabs(Xxview[ixv]-cellX) < 2.*4. ) totalE_road40+=Exview[ixv];
          if( fabs(Xxview[ixv]-cellX) < 2.*6. ) totalE_road60+=Exview[ixv];
        }
      }//in X view
      if( Nyviews>0 ){
        for( int iyv=0; iyv<Nyviews; ++iyv ){
          if( fabs(Yyview[iyv]-cellY) < 2.*2. ) totalE_road20+=Eyview[iyv];
          if( fabs(Yyview[iyv]-cellY) < 2.*4. ) totalE_road40+=Eyview[iyv];
          if( fabs(Yyview[iyv]-cellY) < 2.*6. ) totalE_road60+=Eyview[iyv];
        }
      }//in Y view

/*
      if( slice.NCell()>0 ){
        for( unsigned int icell=0; icell<slice.NCell(); ++ icell){
          const art::Ptr<rb::CellHit>& chit = slice.Cell(icell);
          const TVector3 mean = slice.MeanXYZ();
          const double wx = (slice.NXCell() > 0) ? slice.W(slice.XCell(0).get()) : 0;
          const double wy = (slice.NYCell() > 0) ? slice.W(slice.YCell(0).get()) : 0;
          const rb::RecoHit rhit = cal->MakeRecoHit(*chit,
                                                    chit->View() == geo::kX ? wx : wy);
          if( !rhit.IsCalibrated() ) continue;
          int Dplane=chit->Plane()-slice.MinPlane();
          if( Dplane==iplane ){
            double dx=fabs(rhit.X()-cellX);
            double dy=fabs(rhit.Y()-cellY);
            if( chit->View()==geo::kX ){
              if( dx<2.*2. ) totalE_road20+=rhit.GeV();
              if( dx<2.*4. ) totalE_road40+=rhit.GeV();
            }//in X view
            else{
              if( dy<2.*2. ) totalE_road20+=rhit.GeV();
              if( dy<2.*4. ) totalE_road40+=rhit.GeV();
            }//in Y view
          }
        }//loop all cells
      }//has cells
    */

    }//loop planes

    double efrac_plane2=-1.;
    double efrac_plane4=-1.;
    double efrac_plane6=-1.;
    double Eplane20=0.;
    int nplanes=PlaneSumEnergy.size();
    if (recoEnergy > 0.0){
      for( int i=0; i<nplanes; ++i ){
        if( i<20 ) Eplane20 += PlaneSumEnergy[i];
        //maximal energy fraction in 2 sliding window
        if( i<nplanes-2 ){
          double e_frac2=(PlaneSumEnergy[i]+PlaneSumEnergy[i+1])/recoEnergy;
          if( efrac_plane2<e_frac2 ) efrac_plane2=e_frac2; 
        }
        else if( nplanes<=2 ) efrac_plane2=1.;
        //maximal energy fraction in 4 sliding window
        if( i<nplanes-4 ){
          double e_frac4=(PlaneSumEnergy[i]+PlaneSumEnergy[i+1]+PlaneSumEnergy[i+2]+PlaneSumEnergy[i+3])/recoEnergy;
          if( efrac_plane4<e_frac4 ) efrac_plane4=e_frac4;
        }
        else if( nplanes<=4 ) efrac_plane4=1.;
        //maximal energy fraction in 6 sliding window
        if( i<nplanes-6 ){
          double e_frac6=(PlaneSumEnergy[i]+PlaneSumEnergy[i+1]+PlaneSumEnergy[i+2]+PlaneSumEnergy[i+3]+PlaneSumEnergy[i+4]+PlaneSumEnergy[i+5])/recoEnergy;
          if( efrac_plane6<e_frac6 ) efrac_plane6=e_frac6; 
        }
        else if( nplanes<=6 ) efrac_plane6=1.;
      }
    }
    if( recoEnergy>0.0 ){
      rvp.efrac_2sigmaHOR=totalE_road20/recoEnergy;
      rvp.efrac_4sigmaHOR=totalE_road40/recoEnergy;
      rvp.efrac_20slides=Eplane20/recoEnergy;
      rvp.efrac_2slides=efrac_plane2;
      rvp.efrac_4slides=efrac_plane4;
      rvp.efrac_6slides=efrac_plane6;
    }
    else{
      rvp.efrac_2sigmaHOR=-1.0;
      rvp.efrac_4sigmaHOR=-1.0;
      rvp.efrac_20slides=-1.0;
      rvp.efrac_2slides=efrac_plane2;
      rvp.efrac_4slides=efrac_plane4;
      rvp.efrac_6slides=efrac_plane6;
    }

    if( totalE_road20>0. ) rvp.eiso_2sigmaHOR=(recoEnergy-totalE_road20)/totalE_road20;
    if( totalE_road40>0. ) rvp.eiso_4sigmaHOR=(recoEnergy-totalE_road40)/totalE_road40;
    if( totalE_road60>0. ) rvp.eiso_6sigmaHOR=(recoEnergy-totalE_road60)/totalE_road60;
    return rvp;
  }

  TVector3 RecoJMShower::GetCellNodePos(unsigned int plane1, unsigned int cell1, unsigned int plane2, unsigned int cell2){
    double cellXYZ1[3]={0,0,0};
    double cellXYZ2[3]={0,0,0};
    art::ServiceHandle<geo::Geometry> geom;
    geom->Plane(plane1)->Cell(cell1)->GetCenter(cellXYZ1);
    geom->Plane(plane2)->Cell(cell2)->GetCenter(cellXYZ2);
    geo::View_t view1 = geom->Plane(plane1)->View();
    geo::View_t view2 = geom->Plane(plane2)->View();
    TVector3 nullpos(-9999,-9999,-9999);
    if(view1==view2)return nullpos;
    double x=0,y=0,z=0;
    if(view1 == geo::kX && view2 == geo::kY){
      x = cellXYZ1[0];
      y = cellXYZ2[1];
    }else if(view1 == geo::kY && view2 == geo::kX){
      x = cellXYZ2[0];
      y = cellXYZ1[1];
    }
    z = (cellXYZ1[2]+cellXYZ2[2])/2.;
    TVector3 nodepos(x,y,z);
    return nodepos;
  }


} // end namespace jmshower

///////////////////////////////////////////////////////////////////////
namespace jmshower
{
  DEFINE_ART_MODULE(RecoJMShower);
}
