////////////////////////////////////////////////////////////////////////
// \file    FuzzyKValidate.cxx
// \brief   Analysis module for performance of FuzzyK
// \version $Id: FuzzyKValidate.cxx,v 1.7 2012-12-12 16:24:35 edniner Exp $
// \author  $Author: edniner $
////////////////////////////////////////////////////////////////////////

// C/C++ includes
#include <cmath>
#include <iostream>
#include <map>
#include <string>
#include <vector>

// ROOT includes
#include "TH1F.h"
#include "TH2F.h"
#include "TTree.h"

// Framework includes
#include "art/Framework/Core/EDAnalyzer.h"
#include "art/Framework/Core/FindManyP.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Services/Optional/TFileDirectory.h"
#include "art/Framework/Services/Optional/TFileService.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "art/Persistency/Common/Ptr.h"
#include "art/Persistency/Common/PtrVector.h"
#include "fhiclcpp/ParameterSet.h"
#include "messagefacility/MessageLogger/MessageLogger.h"

// NOvASoft includes
#include "Simulation/Particle.h"
#include "SimulationBase/MCNeutrino.h"
#include "SimulationBase/MCParticle.h"
#include "SimulationBase/MCTruth.h"

#include "Calibrator/Calibrator.h"
#include "MCCheater/BackTracker.h"
#include "RecoBase/CellHit.h"
#include "RecoBase/RecoHit.h"
#include "RecoBase/Prong.h"
#include "RecoBase/Vertex.h"
#include "Utilities/func/MathUtil.h"
#include "Utilities/AssociationUtil.h"


namespace fuzz {
  class FuzzyKValidate : public art::EDAnalyzer {
    public:
      explicit FuzzyKValidate(fhicl::ParameterSet const& pset);    
      ~FuzzyKValidate();                        
      
      void analyze(const art::Event& evt); 
      void reconfigure(const fhicl::ParameterSet& pset);
      void beginJob();
      void ClearVectors();
      float GeVCalc(art::Ptr<rb::Prong> clus, geo::View_t v); 
      std::vector<int> getFaveIdsByEfficiency(std::vector<cheat::NeutrinoWithIndex>& allNus, 
                                              std::vector< std::vector<cheat::NeutrinoEffPur> >& sliceTruthTable);
      std::set<int> FindVisibleProngs3D(const art::PtrVector<rb::CellHit>& allhits, std::set<int> ids);
      
    private:

      /// MC information
      std::string      fSliceLabel;
      std::string      fCellHitLabel;
      std::string      fTrackLabel;
      std::string      fVertexLabel;
      std::string      fAssn3DLabel;
      std::string      fAssn2DLabel;
      int              fCompMCNeutrino;
      
      TTree* fOutTree;
      //event summary info
      int run;
      int subrun;
      int event;

      //reco Prong info and direction
      int nt3D; //number 3D Prongs
      int nt2D; //number 2D Prongs
      std::vector<int> view; //Prong view
      std::vector<float> sX; //start position x
      std::vector<float> sY;//start position y
      std::vector<float> sZ;//start position z
      std::vector<float> dX; //direction x
      std::vector<float> dY;//direction y
      std::vector<float> dZ;//direction z
      std::vector<int>   ncells; //num cells total
      std::vector<int>   ncellsX;//x cells
      std::vector<int>   ncellsY;//y cells
      std::vector<float> gevTot; //total GeV
      std::vector<float> gevX;//x GeV
      std::vector<float> gevY;//y GeV
      std::vector<float> totGeV;
      std::vector<float> len;  //Prong length
      //Prong comparisons these are only for 3D Prongs
      //below vectors are for true visible primaries that were matched to a reco Prong
      std::vector<int> matchedPDG;//pdg of visible particle being matched
      std::vector<int> matchedProngIndx;//index of reco Prong being matched
      std::vector<float> matchedVET;//visible energy of true paricle in entire slice
      std::vector<float> matchedVER;//visible energy of true particle captured in Prong
      std::vector<float> matchedET;//total energy of true particle
      std::vector<float> matchedXStartT;//x start of true particle
      std::vector<float> matchedYStartT;//y start of true particle
      std::vector<float> matchedZStartT;//z start of true particle
      std::vector<float> matchedPXT;//x momentum of true particle
      std::vector<float> matchedPYT;//y momentum of true particle
      std::vector<float> matchedPZT;//z momentum of true particle
      std::vector<float> matchedAngleDiff;//angle difference between reco Prong and truth
      std::vector<float> matchedEff;//efficiency of match
      std::vector<float> matchedPur;//purity of match
      std::vector<int> matchedMother;//pdg of mother particle
      //the below are for true visible primaries that were NOT matched to a reco Prong
      std::vector<int> unmatchedMother;//mother pdg code
      std::vector<int> unmatchedPDG;//unmatched primary pdg code
      std::vector<float> unmatchedVET;//visible energy of unmatched in the slice
      std::vector<float> unmatchedET;//total energy of unmatched
      std::vector<float> unmatchedXStartT;//x start of true particle
      std::vector<float> unmatchedYStartT;//y start of true particle
      std::vector<float> unmatchedZStartT;//z start of true particle
      std::vector<float> unmatchedPXT;//x momentum of true particle
      std::vector<float> unmatchedPYT;//y momentum of true particle
      std::vector<float> unmatchedPZT;//z momentum of true particle
      //below vectors are for reco Prongs not matched as first choice to a visible primary.  The matches made are too the best option
      std::vector<int> danglerPDG;//pdg of visible particle being matched
      std::vector<int> danglerProngIndx;//index of reco Prong being matched
      std::vector<float> danglerVET;//visible energy of true paricle in entire slice
      std::vector<float> danglerVER;//visible energy of true particle captured in Prong
      std::vector<float> danglerET;//total energy of true particle
      std::vector<float> danglerXStartT;//x start of true particle
      std::vector<float> danglerYStartT;//y start of true particle
      std::vector<float> danglerZStartT;//z start of true particle
      std::vector<float> danglerPXT;//x momentum of true particle
      std::vector<float> danglerPYT;//y momentum of true particle
      std::vector<float> danglerPZT;//z momentum of true particle
      std::vector<float> danglerAngleDiff;//angle difference between reco Prong and truth
      std::vector<float> danglerEff;//efficiency of match
      std::vector<float> danglerPur;//purity of match
      std::vector<int> danglerMother;//pdg of mother particle
      int matchedProngs;//number of successfully matched primaries
      int unmatchedProngs;//number of unmatched primaries
      int danglerProngs;//number of reco Prongs not matched


      //overall optimization score for the event
      double eventScore;


      //Prong histograms
      TH1F* fHistoNt;
      TH1F* fHisto3D;
      TH1F* fHisto2D;
      TH2F* fHisto3Dvs2D;
      TH1F* fHistoLen;
      TH1F* fHistoNCell;
      TH1F* fHistoSX;
      TH1F* fHistoSY;
      TH1F* fHistoSZ;
      TH1F* fHistoDX;
      TH1F* fHistoDY;
      TH1F* fHistoDZ;
      TH1F* fHistoGeV;
      TH2F* fHistoGeVYvX;
      TH1F* fHistoNCells;
      TH1F* fHistoNCellsX;
      TH1F* fHistoNCellsY;
      TH2F* fHistoNCellsXY;
      //Prong comparisons
      TH1F* fHistoEff;
      TH1F* fHistoPur;
      TH1F* fHistoAngle;
      TH1F* fHistoMatched;
      TH1F* fHistoUnMatched;
      TH1F* fHistoLeft;
      TH1F* fHistoLeptonEff;
      TH1F* fHistoLeptonPur;
      TH1F* fHistoLeptonAngle;
      TH2F* fHistoUvL;
      TH1F* fHistoUnVisE;
      TH1F* fHistoMVisE;
      TH2F* fHistoEffvEvis;
      TH2F* fHistoPurvEvis;
      TH2F* fHistoAnglevEvis;

      //efficiency vs purity for all visible primaries matched to their best reco Prong
      TH2F* fHistoEffvPur;
      TH2F* fHistoDEffvPur;
      TH1F* fHistoMatchMetric; //store value of match quality between reco Prong and visible particle

      //store event optimization score
      TH1F* fHistoEventScore;

      //neutrino information
      float Vx;//start x
      float Vy;//start y
      float Vz;//start z
      float NuEff;//efficiency of neutrino for slice
      float NuPur;//purity of neutrino for slice
      float NuEVSl;//visible energy of neutrino in slice
      float NuEVTot;//visible energy of neutrino in all slices
      float x;
      float y;
      float q2;
      float W2;
      int NuPdg;//neutrino pdg
      int ccnc;//is cc or nc
      int mode;//interaction mode
      float NuE;//total energy of neutrino

      float rVx;//start x
      float rVy;//start y
      float rVz;//start z
      
      



    };
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////

namespace fuzz
{
  //.......................................................................
  FuzzyKValidate::FuzzyKValidate(fhicl::ParameterSet const& pset) 
    : EDAnalyzer(pset)
  {
    this->reconfigure(pset);
  }

  //......................................................................
  FuzzyKValidate::~FuzzyKValidate()
  {
    //======================================================================
    // Clean up any memory allocated by your module
    //======================================================================
  }

  //......................................................................
  void FuzzyKValidate::reconfigure(const fhicl::ParameterSet& pset)
  {

    fSliceLabel        = pset.get< std::string >("SliceLabel");
    fCellHitLabel      = pset.get< std::string >("CellHitLabel");
    fVertexLabel       = pset.get< std::string >("VertexLabel");
    fTrackLabel        = pset.get< std::string >("TrackLabel");
    fAssn3DLabel       = pset.get< std::string >("Assn3DLabel");
    fAssn2DLabel       = pset.get< std::string >("Assn2DLabel");
    fCompMCNeutrino    = pset.get< int         >("CompMCNeutrino");
    
  }
      
  //......................................................................
  void FuzzyKValidate::beginJob()
  {

    art::ServiceHandle<art::TFileService> tfs;

    // define the tree
    fOutTree = tfs->make<TTree>("FuzzyKValidate","FuzzyK Validation");

    //store event summary information
    fOutTree->Branch("run", &run, "run/I");
    fOutTree->Branch("subrun", &subrun, "subrun/I");
    fOutTree->Branch("event", &event, "event/I");

    //store Prong info
    fOutTree->Branch("nt3D", &nt3D);
    fOutTree->Branch("nt2D", &nt2D);
    fOutTree->Branch("view", &view);
    fOutTree->Branch("sX",   &sX);
    fOutTree->Branch("sY",   &sY);
    fOutTree->Branch("sZ",   &sZ);
    fOutTree->Branch("dX",   &dX);
    fOutTree->Branch("dY",   &dY);
    fOutTree->Branch("dZ",   &dZ);
    fOutTree->Branch("ncells", &ncells);
    fOutTree->Branch("ncellsX", &ncellsX);
    fOutTree->Branch("ncellsY", &ncellsY);
    fOutTree->Branch("gevTot", &gevTot);
    fOutTree->Branch("gevX", &gevX);
    fOutTree->Branch("gevY", &gevY);
    fOutTree->Branch("totGeV", &totGeV);
    fOutTree->Branch("len", &len);
    //store Prong comparisons
    fOutTree->Branch("matchedPDG", &matchedPDG);
    fOutTree->Branch("matchedProngIndx", &matchedProngIndx);
    fOutTree->Branch("matchedVET", &matchedVET);
    fOutTree->Branch("matchedVER", &matchedVER);
    fOutTree->Branch("matchedET", &matchedET);
    fOutTree->Branch("matchedXStartT", &matchedXStartT);
    fOutTree->Branch("matchedYStartT", &matchedYStartT);
    fOutTree->Branch("matchedZStartT", &matchedZStartT);
    fOutTree->Branch("matchedPXT", &matchedPXT);
    fOutTree->Branch("matchedPYT", &matchedPYT);
    fOutTree->Branch("matchedPZT", &matchedPZT);
    fOutTree->Branch("matchedAngleDiff", &matchedAngleDiff);
    fOutTree->Branch("matchedEff", &matchedEff);
    fOutTree->Branch("matchedPur", &matchedPur);
    fOutTree->Branch("matchedMother", &matchedMother);
    fOutTree->Branch("unmatchedMother", &unmatchedMother);
    fOutTree->Branch("unmatchedPDG", &unmatchedPDG);
    fOutTree->Branch("unmatchedVET", &unmatchedVET);
    fOutTree->Branch("unmatchedXStartT", &unmatchedXStartT);
    fOutTree->Branch("unmatchedYStartT", &unmatchedYStartT);
    fOutTree->Branch("unmatchedZStartT", &unmatchedZStartT);
    fOutTree->Branch("unmatchedPXT", &unmatchedPXT);
    fOutTree->Branch("unmatchedPYT", &unmatchedPYT);
    fOutTree->Branch("unmatchedPZT", &unmatchedPZT);
    fOutTree->Branch("matchedProngs", &matchedProngs);
    fOutTree->Branch("unmatchedProngs", &unmatchedProngs);
    fOutTree->Branch("danglerProngs", &danglerProngs);
    fOutTree->Branch("danglerPDG", &danglerPDG);
    fOutTree->Branch("danglerProngIndx", &danglerProngIndx);
    fOutTree->Branch("danglerVET", &danglerVET);
    fOutTree->Branch("danglerVER", &danglerVER);
    fOutTree->Branch("danglerET", &danglerET);
    fOutTree->Branch("danglerXStartT", &danglerXStartT);
    fOutTree->Branch("danglerYStartT", &danglerYStartT);
    fOutTree->Branch("danglerZStartT", &danglerZStartT);
    fOutTree->Branch("danglerPXT", &danglerPXT);
    fOutTree->Branch("danglerPYT", &danglerPYT);
    fOutTree->Branch("danglerPZT", &danglerPZT);
    fOutTree->Branch("danglerAngleDiff", &danglerAngleDiff);
    fOutTree->Branch("danglerEff", &danglerEff);
    fOutTree->Branch("danglerPur", &danglerPur);
    fOutTree->Branch("danglerMother", &danglerMother);

    //store overall event score for optimization
    //score = (1 - E_dangle/E_total)sqrt(sum[ (E_i/E_tot)^2*(eff^2 + pur^2)*0.5] ) 
    //the sum is over all the visible particles 
    //E_i is the energy of the true particle used in the match and E_tot is the total visible energy in the slice
    //a visible primary or a reco Prong with no match goes into the score as a 0,0
    //The factor (1- E_dangle/E_total) is a weight factor that accounts for unmatched reco Prongs.  
    //E_dangle/E_total is the fraction of fls energy put into reco Prongs that were not associated with true particles
    //this score will have a value between 0 and 1, the higher the score the better the event performance
    fOutTree->Branch("eventScore", &eventScore);


    //Prong histograms
    fHistoNt = tfs->make<TH1F>("fHistoNt","Number of Prongs (3D+2D) per vertex;Prongs",17,-0.5,16.5);
    fHisto3D = tfs->make<TH1F>("fHisto3D","Number of 3D Prongs per vertex;3D Prongs",9,-0.5,8.5);
    fHisto2D = tfs->make<TH1F>("fHisto2D","Number of 2D Prongs per vertex;2D Prongs",9,-0.5,8.5);
    fHisto3Dvs2D = tfs->make<TH2F>("fHisto3Dvs2D","3D Prongs vs 2D Prongs;2D Prongs;3D Prongs",9,-0.5,8.5,9,-0.5,8.5);
    fHistoLen = tfs->make<TH1F>("fHistoLen","Prong length;length(cm)",350,0.0,7000.0);
    fHistoSX  = tfs->make<TH1F>("fHistoSX","Start X coordinate ;x(cm)",160,-800.0,800.0);
    fHistoSY  = tfs->make<TH1F>("fHistoSY","Start Y coordinate ;y(cm)",160,-800.0,800.0);
    fHistoSZ  = tfs->make<TH1F>("fHistoSZ","Start Z coordinate ;z(cm)",120,0.0,6000.0);
    fHistoDX  = tfs->make<TH1F>("fHistoDX","X direction ;x dir",101,-1.01,1.01);
    fHistoDY  = tfs->make<TH1F>("fHistoDY","Y direction ;y dir",101,-1.01,1.01);
    fHistoDZ  = tfs->make<TH1F>("fHistoDZ","Z direction ;z dir",101,-1.01,1.01);
    fHistoNCells = tfs->make<TH1F>("fHistoNCells","Number of cells in Prong; cells",200,0.0,400.0);
    fHistoNCellsX = tfs->make<TH1F>("fHistoNCellsX","Number of cells in Prong X view; x cells",150,0.0,300.0);
    fHistoNCellsY = tfs->make<TH1F>("fHistoNCellsY","Number of cells in Prong Y view; y cells",150,0.0,300.0);
    fHistoNCellsXY = tfs->make<TH2F>("fHistoNCellsXY","Y cells vs. X cells;x cells;y cells"
                                      ,150,0.0,300.0,150,0.0,300.0);
    fHistoGeV = tfs->make<TH1F>("fHistoGeV","GeV of Prong ;GeV",500,0.0,50.0);
    fHistoGeVYvX = tfs->make<TH2F>("fHistoGeVYvX","GeVY vs GeVX ;GeVX;GeVY",
                                      250,0.0,25.0,250,0.0,25.0);
    fHistoEff = tfs->make<TH1F>("fHistoEff","Efficiency of matched visible particles;eff",101,0.0,1.01);
    fHistoPur = tfs->make<TH1F>("fHistoPur","Purity of matched visible particles;purity",101,0.0,1.01);
    fHistoAngle = tfs->make<TH1F>("fHistoAngle","Angle between true particle and reco Prong;theta",91,0.0,182.0);
    fHistoMatched = tfs->make<TH1F>("fHistoMatched","Number of visible primaries matched to reco Prongs"
                                    ,13,-0.5,12.5);
    fHistoUnMatched = tfs->make<TH1F>("fHistoUnMatched","Number of visible primaries unmatched to reco Prongs"
                                      ,13,-0.5,12.5);
    fHistoLeft = tfs->make<TH1F>("fHistoLeft","Number of reco Prongs not matched",9,-0.5,8.5);
    fHistoLeptonEff = tfs->make<TH1F>("fHistoLeptonEff","Completeness of matched primary leptons;eff",101,0.0,1.01);
    fHistoLeptonPur = tfs->make<TH1F>("fHistoLeptonPur","Purity of matched primary leptons;purity",101,0.0,1.01);
    fHistoLeptonAngle = tfs->make<TH1F>("fHistoLeptonAngle","Angle between primary lepton and reco Prong",91,0.0,182.0);
    fHistoUvL = tfs->make<TH2F>("fHistoUvL","Unmatched primaries vs Unmatched reco Prongs;reco Prongs;primaries"
                                ,9,-0.5,8.5,9,-0.5,8.5);
    fHistoUnVisE = tfs->make<TH1F>("fHistoUnVisE","Visible energy of unmatched primaries;GeV",250,0.0,5.0);
    fHistoMVisE = tfs->make<TH1F>("fHistoMVisE","Visible energy of matched primaries;GeV",250,0.0,5.0);
    fHistoEffvEvis = tfs->make<TH2F>("fHistoEffvEvis","Completeness vs Visible energy for match;visE(GeV);eff"
                                     ,250,0.0,5.0,101,0.0,1.01);
    fHistoPurvEvis = tfs->make<TH2F>("fHistoPurvEvis","Purity vs Visible energy for match;visE(GeV);purity"
                                     ,250,0.0,5.0,101,0.0,1.01);
    fHistoAnglevEvis = tfs->make<TH2F>("fHistoAnglevEvis","Angle diff vs Visible energy for match;visE(GeV);theta"
                                       ,250,0.0,5.0,91,0.0,182.0);
    fHistoEffvPur = tfs->make<TH2F>("fHistoEffvPur","Completeness vs Purity for visible primaries;purity;eff"
                                    ,101,0.0,1.01,101,0.0,1.01);

    fHistoDEffvPur = tfs->make<TH2F>("fHistoDEffvPur","Completeness vs Purity for dangling reco Prongs;purity;eff"
                                    ,101,0.0,1.01,101,0.0,1.01);

    fHistoMatchMetric = tfs->make<TH1F>("fHistoMatchMetric","|E_truepart - E_recoProng|/E_truepart",200,0.0,100.0);

    fHistoEventScore = tfs->make<TH1F>("fHistoEventScore","Event Score for optimization;score;events",101,0.0,1.01);

    //store neutrino info
    fOutTree->Branch("Vx", &Vx);//neutrino position info
    fOutTree->Branch("Vy", &Vy);
    fOutTree->Branch("Vz", &Vz);
    fOutTree->Branch("NuEff", &NuEff);//info on quality of neutrino in slice
    fOutTree->Branch("NuPur", &NuPur);
    fOutTree->Branch("NuEVSl", &NuEVSl);
    fOutTree->Branch("NuEVTot", &NuEVTot);
    fOutTree->Branch("NuPdg", &NuPdg);
    fOutTree->Branch("ccnc", &ccnc);
    fOutTree->Branch("mode", &mode);
    fOutTree->Branch("NuE", &NuE);
    fOutTree->Branch("y", &y);
    fOutTree->Branch("x", &x);
    fOutTree->Branch("q2", &q2);
    fOutTree->Branch("W2", &W2);

    fOutTree->Branch("rVx", &rVx);
    fOutTree->Branch("rVy", &rVy);
    fOutTree->Branch("rVz", &rVz);

    
  }

  //......................................................................
  void FuzzyKValidate::analyze(const art::Event& evt)
  {
    art::ServiceHandle<cheat::BackTracker> bt;

    art::Handle<std::vector<rb::Cluster> > sHandle;
    evt.getByLabel(fSliceLabel, sHandle);
    if(sHandle.failedToGet()){
      std::cout<<"No slices found, skipping event" << std::endl;
    }
    art::PtrVector<rb::Cluster> slices;
    for(unsigned int i=0; i<sHandle->size(); ++i){
      slices.push_back(art::Ptr<rb::Cluster>(sHandle, i));
    }

    //get all the hits to calculate neutrino efficiency/purity for slices
    art::Handle<std::vector<rb::CellHit>> htHandle;
    evt.getByLabel(fCellHitLabel, htHandle);
    art::PtrVector<rb::CellHit> allHits;
    for(unsigned int i=0; i<htHandle->size(); ++i){
      art::Ptr<rb::CellHit> hit(htHandle,i);
      allHits.push_back(hit);
    }

    //get associations between slices and vertices
    art::FindManyP<rb::Vertex> fmVert(sHandle, evt, fVertexLabel);
    //try to get associations between Prongs and slices
    art::FindManyP<rb::Prong> fmSlProng3D(sHandle, evt, art::InputTag(fTrackLabel, fAssn3DLabel));
    art::FindManyP<rb::Prong> fmSlProng2D(sHandle, evt, art::InputTag(fTrackLabel, fAssn2DLabel));


    //if these are events with neutrino information to compare too, get it
    std::vector<int> bestNuId;
    std::vector<cheat::NeutrinoWithIndex> nus;
    std::vector<std::vector<cheat::NeutrinoEffPur>> sEffPur;
    if(fCompMCNeutrino == 1){
      nus = bt->allMCTruth();
      //std::cout<<allNus.size()<<std::endl;
      sEffPur = bt->SlicesToMCTruthsTable(slices);
      //std::cout<<"Done matching"<<std::endl;
    
      bestNuId = getFaveIdsByEfficiency(nus, sEffPur); 
    }
    

    //loop over slices
    for( unsigned int i=0; i<slices.size(); ++i){
      this->ClearVectors();
      //skip noise slice
      if (slices[i]->IsNoise()) continue;
      //if we are comparing to truth, make sure there is a neutrino with this slice
      if(fCompMCNeutrino == 1){
        if (bestNuId[i] == -1) continue;
      }
      std::vector<art::Ptr<rb::Prong>> fk3D;
      std::vector<art::Ptr<rb::Prong>> fk2D;
      //if there are associations between Prongs and slices use those
      if(fmSlProng3D.isValid()){
        fk3D = fmSlProng3D.at(i);
        fk2D = fmSlProng2D.at(i);
      }
      //else see if there are Prongs connected to vertex
      else if(fmVert.isValid()){
        std::vector<art::Ptr<rb::Vertex>> v = fmVert.at(i);
        //should only be one vertex
        if (v.size() != 1) continue;
        art::FindManyP<rb::Prong> fmVertProng3D(v, evt, art::InputTag(fTrackLabel, fAssn3DLabel));
        art::FindManyP<rb::Prong> fmVertProng2D(v, evt, art::InputTag(fTrackLabel, fAssn2DLabel));
        //skip if there are no prongs
        if(!(fmVertProng3D.isValid())) continue;
        //only one vertex per slice so only need index 0
        fk3D = fmVertProng3D.at(0);
        fk2D = fmVertProng2D.at(0);

        rVx = v[0]->GetX();
        rVy = v[0]->GetY();
        rVz = v[0]->GetZ();
      }
      else{
        continue;
      }

      run = evt.run();
      subrun = evt.subRun();
      event = evt.id().event();

      //fill Prong statistics
      nt3D = fk3D.size();
      nt2D = fk2D.size();
      for(unsigned int it=0; it<fk3D.size(); ++it){
        view.push_back(fk3D[it]->View());
        sX.push_back(fk3D[it]->Start().X());
        sY.push_back(fk3D[it]->Start().Y());
        sZ.push_back(fk3D[it]->Start().Z());
        dX.push_back(fk3D[it]->Dir().X());
        dY.push_back(fk3D[it]->Dir().Y());
        dZ.push_back(fk3D[it]->Dir().Z());
        ncells.push_back(fk3D[it]->NCell());
        ncellsX.push_back(fk3D[it]->NXCell());
        ncellsY.push_back(fk3D[it]->NYCell());
        gevTot.push_back(fk3D[it]->TotalGeV());
        totGeV.push_back(fk3D[it]->TotalGeV());
        float gevCX = GeVCalc(fk3D[it], geo::kX);
        float gevCY = GeVCalc(fk3D[it], geo::kY);
        gevX.push_back(gevCX);
        gevY.push_back(gevCY);
        len.push_back(fk3D[it]->TotalLength());
        
        fHistoLen->Fill(len.back());
        fHistoSX->Fill(sX.back());
        fHistoSY->Fill(sY.back());
        fHistoSZ->Fill(sZ.back());
        fHistoDX->Fill(dX.back());
        fHistoDY->Fill(dY.back());
        fHistoDZ->Fill(dZ.back());
        fHistoGeV->Fill(gevTot.back());
        fHistoGeVYvX->Fill(gevX.back(),gevY.back());
        fHistoNCells->Fill(ncells.back());
        fHistoNCellsX->Fill(ncellsX.back());
        fHistoNCellsY->Fill(ncellsY.back());
        fHistoNCellsXY->Fill(ncellsX.back(),ncellsY.back());
      }
      fHistoNt->Fill(nt3D+nt2D);
      fHisto3D->Fill(nt3D);
      fHisto2D->Fill(nt2D);
      fHisto3Dvs2D->Fill(nt2D,nt3D);
      //now fill 2D Prong info
      for(unsigned int it=0; it<fk2D.size(); ++it){
        view.push_back(fk2D[it]->View());
        sX.push_back(fk2D[it]->Start().X());
        sY.push_back(fk2D[it]->Start().Y());
        sZ.push_back(fk2D[it]->Start().Z());
        dX.push_back(fk2D[it]->Dir().X());
        dY.push_back(fk2D[it]->Dir().Y());
        dZ.push_back(fk2D[it]->Dir().Z());
        ncells.push_back(fk2D[it]->NCell());
        ncellsX.push_back(fk2D[it]->NXCell());
        ncellsY.push_back(fk2D[it]->NYCell());
        gevTot.push_back(fk2D[it]->TotalGeV());
        totGeV.push_back(fk2D[it]->TotalGeV());
        float gevCX = GeVCalc(fk2D[it], geo::kX);
        float gevCY = GeVCalc(fk2D[it], geo::kY);
        gevX.push_back(gevCX);
        gevY.push_back(gevCY);
        len.push_back(fk2D[it]->TotalLength());
        fHistoLen->Fill(len.back());
        if (fk2D[it]->View() == geo::kX){
          fHistoSX->Fill(sX.back());
          fHistoDX->Fill(dX.back());
          fHistoNCellsX->Fill(ncellsX.back());
        }
        else if(fk2D[it]->View() == geo::kY){
          fHistoSY->Fill(sY.back());
          fHistoDY->Fill(dY.back());
          fHistoNCellsY->Fill(ncellsY.back());
        }
        fHistoNCells->Fill(ncells.back());
        fHistoSZ->Fill(sZ.back());
        fHistoDZ->Fill(dZ.back());
        fHistoGeV->Fill(gevTot.back());
      }

      //if we do not want to compare to truth, dump into to tree and move to next slice
      if(fCompMCNeutrino != 1){
        fOutTree->Fill();
        continue;
      }

      //now access true neutrino info for further validation
      art::Ptr<simb::MCTruth> nuTru = sEffPur[i][bestNuId[i]].neutrinoInt;

      NuEff = sEffPur[i][bestNuId[i]].efficiency;
      NuPur = sEffPur[i][bestNuId[i]].purity;
      NuEVSl = sEffPur[i][bestNuId[i]].energySlice;
      NuEVTot = sEffPur[i][bestNuId[i]].energyTotal;

      //if this a neutrino object, store this information, if this is a cosmic,
      //then skip
      if (nuTru->NeutrinoSet()){
        NuPdg = nuTru->GetNeutrino().Nu().PdgCode();
        ccnc = nuTru->GetNeutrino().CCNC();
        mode = nuTru->GetNeutrino().Mode();
        NuE = nuTru->GetNeutrino().Nu().Momentum().E();

        q2 = nuTru->GetNeutrino().QSqr();
        x = nuTru->GetNeutrino().X();
        y = nuTru->GetNeutrino().Y();
        W2 = nuTru->GetNeutrino().W()*nuTru->GetNeutrino().W();

        //now get vertex info and make comparisons
        Vx = nuTru->GetNeutrino().Nu().Vx();
        Vy = nuTru->GetNeutrino().Nu().Vy();
        Vz = nuTru->GetNeutrino().Nu().Vz();
      }

      
      //now go about finding list of primary visible daughters sorted from largest to smallest visible energy
      std::vector<cheat::TrackIDE> sliceProngs = bt->HitsToTrackIDE(slices[i]->AllCells());
      std::set<int> sT;
      for(unsigned int j=0; j<sliceProngs.size(); ++j){
        sT.insert(sliceProngs[j].trackID);
      }
      std::set<int>  visIds = FindVisibleProngs3D(slices[i]->AllCells(), sT);

     //std::cout<<"There are: "<<visIds.size()<<" visible particles."<<std::endl;

      std::vector<const sim::Particle*> particles = bt->MCTruthToParticles(nuTru);
      //std::cout<<"Checking against the list of: "<<particles.size()<<" particles."<<std::endl;

      //now keep particles that meet the visible criteria and are primary. 
      //NOTE: testing using all visible partilces and not just the primaries.  The variable names will
      //be changed to reflect this if this change works.  For now visPrimaries really means visParticles 
      std::vector<const sim::Particle*> visPrimaries;
      for(std::set<int>::iterator itr = visIds.begin(); itr!=visIds.end(); itr++){
        //std::cout<<"Looking for primary that matches visible Prong id: "<<(*itr)<<std::endl;
        for(unsigned int k=0; k<particles.size(); ++k){
          //std::cout<<"Checking particle: "<<k<<" with id: "<<particles[k]->ProngId()<<" and pdg: "<<particles[k]->PdgCode()<<std::endl;
          if (visIds.find(particles[k]->TrackId()) == visIds.end()) continue;
          if (*(visIds.find(particles[k]->TrackId())) == (*itr)){
            visPrimaries.push_back(particles[k]);
            break;
          }
          /*if ((*(visIds.find(particles[k]->TrackId())) == (*itr)) && (particles[k]->Mother() == 0)){
            visPrimaries.push_back(particles[k]);
            std::cout<<"Found a visible primary: "<<particles[k]->PdgCode()<<" with track id "<<particles[k]->TrackId()<<std::endl;
            break;
          }
          //in the case that these are pi0 daughters make sure they get counted
          if ((particles[k]->Mother() != 0) && (*(visIds.find(particles[k]->TrackId())) == (*itr))){
            const sim::Particle* mTest = bt->TrackIDToMotherParticle(particles[k]->TrackId());
            if ((mTest->PdgCode() == 111) && (mTest->Mother() == 0)){
              visPrimaries.push_back(particles[k]);
              std::cout<<"Found a visible primary: "<<particles[k]->PdgCode()<<std::endl;
              break;
            }
          }*/
        }
      }

      //initialize score for slice
      double score = 0.0;

      //now loop through visible primaries, starting with highest energy.
      //for each one find the 3D track for which it is most efficient.
      //visible primaries are allowed to match to the same reco track if that is what is desired
      int matched = 0;
      int unmatched = 0;
      std::set<int> usedProngs;
      std::map<int, double> purMap, effMap;
      std::map<int, int> parents;
      //std::cout<<"There are: "<<visPrimaries.size()<<" primary particles.  Making matches."<<std::endl;
      for(unsigned int j=0; j<visPrimaries.size(); ++j){
        //std::cout<<"Matching particle: "<<j<<std::endl;
        std::set<int> temp;
        temp.insert(visPrimaries[j]->TrackId());
        double eff, pur, matchMetric;
        double maxEff = -1.0;
        double maxPur = -1.0;
        int maxTr = -1;
        double dE, tE, dEMax, tEMax;
        dEMax = -1.0;
        tEMax = -1.0;
        //match a truth particle to reco track with metric |E_truepart - E_recotrack|/E_truepart
        //a good match should have the smallest metric possible
        double maxMatchMetric = 9999.0;
        for(unsigned int k=0; k<fk3D.size(); ++k){
          //if (usedTracks.find(k) != usedTracks.end()) continue;
          eff = bt->HitCollectionEfficiency(temp, fk3D[k]->AllCells(), slices[i]->AllCells(), geo::kXorY, &effMap,true,&dE,&tE);
          pur = bt->HitCollectionPurity(temp, fk3D[k]->AllCells(), &purMap, &parents,true);
          if (pur > 0.0) matchMetric = fabs(tE - dE/pur)/tE;
          else matchMetric = 9999.0;
          //std::cout<<"***************With reco track: "<<k<<" efficiency: "<<eff<<" and metric: "<<matchMetric<<std::endl;
          if(eff > maxEff){
          //if(matchMetric < maxMatchMetric){
            maxMatchMetric = matchMetric;
            maxEff = eff;
            maxPur = pur;
            maxTr = k;
            dEMax = dE;
            tEMax = tE;
          }
          //if efficieny is the same, tiebreak on purity
          else if ((eff == maxEff) && (pur > maxPur)){
            maxMatchMetric = matchMetric;
            maxEff = eff;
            maxPur = pur;
            maxTr = k;
            dEMax = dE;
            tEMax = tE;
          }
        }
        if (maxEff > 0.0){
        //if (maxMatchMetric < 9999.0){
          usedProngs.insert(maxTr);
          const sim::Particle* mTest = bt->TrackIDToMotherParticle(visPrimaries[j]->TrackId());
          matchedMother.push_back(mTest->PdgCode());
          matched++;
          matchedPDG.push_back(visPrimaries[j]->PdgCode());
          matchedProngIndx.push_back(maxTr);
          matchedVET.push_back(tEMax);
          matchedVER.push_back(dEMax);
          matchedET.push_back(visPrimaries[j]->E());
          matchedXStartT.push_back(visPrimaries[j]->Vx());
          matchedYStartT.push_back(visPrimaries[j]->Vy());
          matchedZStartT.push_back(visPrimaries[j]->Vz());
          matchedPXT.push_back(visPrimaries[j]->Px());
          matchedPYT.push_back(visPrimaries[j]->Py());
          matchedPZT.push_back(visPrimaries[j]->Pz());
          float costheta = (dX[maxTr]*matchedPXT.back() + dY[maxTr]*matchedPYT.back() + dZ[maxTr]*matchedPZT.back())/
                           (util::pythag(dX[maxTr],dY[maxTr],dZ[maxTr])*util::pythag(matchedPXT.back(),matchedPYT.back()
                           ,matchedPZT.back()));
          matchedAngleDiff.push_back(acos(costheta)*180.0/M_PI);
          matchedEff.push_back(maxEff);
          matchedPur.push_back(maxPur);

          fHistoEff->Fill(maxEff);
          fHistoPur->Fill(maxPur);
          fHistoAngle->Fill(matchedAngleDiff.back());
          if ((abs(matchedPDG.back()) == 13)){
            if (matchedMother.back() != 111){
              fHistoLeptonEff->Fill(maxEff);
              fHistoLeptonPur->Fill(maxPur);
              fHistoLeptonAngle->Fill(matchedAngleDiff.back());
            }
          }
          fHistoMVisE->Fill(tEMax);
          fHistoEffvEvis->Fill(tEMax,maxEff);
          fHistoPurvEvis->Fill(tEMax,maxPur);
          fHistoAnglevEvis->Fill(tEMax,matchedAngleDiff.back());
          fHistoEffvPur->Fill(maxPur,maxEff);
         
          score += (tEMax*NuEff/NuEVSl)*(tEMax*NuEff/NuEVSl)*(maxEff*maxEff + maxPur*maxPur)*0.5;
          fHistoMatchMetric->Fill(maxMatchMetric);

          //std::cout<<"Made a match for particle: "<<j<<" to reco track: "<<maxTr<<" with eff: "<<maxEff<<" and pur: "<< maxPur<<" score is now: "<<score<<std::endl;
          //std::cout<<"The true particle had total energy: "<<tEMax<<std::endl;
        }
        else{
          //std::cout<<"No match made for particle: "<<j<<std::endl;
          unmatched++;
          const sim::Particle* mTest = bt->TrackIDToMotherParticle(visPrimaries[j]->TrackId());
          unmatchedMother.push_back(mTest->PdgCode());
          unmatchedPDG.push_back(visPrimaries[j]->PdgCode());
          unmatchedVET.push_back(tEMax);
          unmatchedET.push_back(visPrimaries[j]->E());
          unmatchedXStartT.push_back(visPrimaries[j]->Vx());
          unmatchedYStartT.push_back(visPrimaries[j]->Vy());
          unmatchedZStartT.push_back(visPrimaries[j]->Vz());
          unmatchedPXT.push_back(visPrimaries[j]->Px());
          unmatchedPYT.push_back(visPrimaries[j]->Py());
          unmatchedPZT.push_back(visPrimaries[j]->Pz());
          fHistoUnVisE->Fill(tEMax);
          fHistoEffvPur->Fill(0.0,0.0);
        }
      }

      //now if there are unmatched  3D tracks, look at them and get the info for the most efficient visible primary
      int dangler = 0.0;
      std::set<art::Ptr<rb::CellHit>> dCells; //store energy from cells from all danglers to calculate energy weight
      for (unsigned int j=0; j<fk3D.size(); j++){
        if (usedProngs.find(j) != usedProngs.end()) continue;
        //std::cout<<"Reco track: "<<j<<" with "<<fk3D[j]->NCell()<<" cells has no match, looking now."<<std::endl;
        for (unsigned int k=0; k<fk3D[j]->NCell(); k++){
          dCells.insert(fk3D[j]->Cell(k));
        }
        //std::cout<<"There are now: "<<dCells.size()<<" cells in dangler."<<std::endl;
        double eff, pur, matchMetric;
        double maxEff = -1.0;
        double maxPur = -1.0;
        double maxMatchMetric = 9999.0;
        int maxPr = -1;
        double dE, tE, dEMax, tEMax;
        dEMax = -1.0;
        tEMax = -1.0;
        for (unsigned int k=0; k<visPrimaries.size(); k++){
          std::set<int> temp;
          temp.insert(visPrimaries[k]->TrackId());

          eff = bt->HitCollectionEfficiency(temp, fk3D[j]->AllCells(), slices[i]->AllCells(), geo::kXorY, &effMap,true,&dE,&tE);
          pur = bt->HitCollectionPurity(temp, fk3D[j]->AllCells(), &purMap, &parents,true);
          if (pur > 0.0) matchMetric = fabs(tE - dE/pur)/tE;
          else matchMetric = 9999.0;
          //std::cout<<"**********This track matches primary: "<<k<<" with efficiency: "<<eff<<" and metric: "<<matchMetric<<std::endl;
          if(eff > maxEff){
          //if(matchMetric < maxMatchMetric){
            maxMatchMetric = matchMetric;
            maxEff = eff;
            maxPur = pur;
            maxPr = k;
            dEMax = dE;
            tEMax = tE;
          }
          else if ((eff == maxEff) && (pur > maxPur)){
            maxMatchMetric = matchMetric;
            maxEff = eff;
            maxPur = pur;
            maxPr = k;
            dEMax = dE;
            tEMax = tE;
          }
        }
        if (maxEff > 0.0){
        //if (maxMatchMetric < 9999.0){
          std::set<int> temp;
          temp.insert(visPrimaries[maxPr]->TrackId());
          const sim::Particle* mTest = bt->TrackIDToMotherParticle(visPrimaries[maxPr]->TrackId());
          danglerMother.push_back(mTest->PdgCode());
          dangler++;
          danglerPDG.push_back(visPrimaries[maxPr]->PdgCode());
          danglerProngIndx.push_back(j);
          danglerVET.push_back(tEMax);
          danglerVER.push_back(dEMax);
          danglerET.push_back(visPrimaries[maxPr]->E());
          danglerXStartT.push_back(visPrimaries[maxPr]->Vx());
          danglerYStartT.push_back(visPrimaries[maxPr]->Vy());
          danglerZStartT.push_back(visPrimaries[maxPr]->Vz());
          danglerPXT.push_back(visPrimaries[maxPr]->Px());
          danglerPYT.push_back(visPrimaries[maxPr]->Py());
          danglerPZT.push_back(visPrimaries[maxPr]->Pz());
          float costheta = (dX[j]*matchedPXT.back() + dY[j]*matchedPYT.back() + dZ[j]*matchedPZT.back())/
                           (util::pythag(dX[j],dY[j],dZ[j])*util::pythag(matchedPXT.back(),matchedPYT.back()
                           ,matchedPZT.back()));
          danglerAngleDiff.push_back(acos(costheta)*180.0/M_PI);
          danglerEff.push_back(maxEff);
          danglerPur.push_back(maxPur);

          fHistoDEffvPur->Fill(maxPur,maxEff);
         
          //score += (tEMax*NuEff/NuEVSl)*(tEMax*NuEff/NuEVSl)*(maxEff*maxEff + maxPur*maxPur)*0.5;
          //std::cout<<"Made a match for reco track: "<<j<<" to particle: "<<maxPr<<" with eff: "<<maxEff<<" and pur: "<< maxPur<<" score is now: "<<score<<std::endl;
          fHistoMatchMetric->Fill(maxMatchMetric);

        }
        else {
          //std::cout<<"No match made for reco track: "<<j<<std::endl;
          maxPur = 0.0;
          maxEff = 0.0;
          danglerMother.push_back(-99);
          dangler++;
          danglerPDG.push_back(-99);
          danglerProngIndx.push_back(j);
          danglerVET.push_back(-1.0);
          danglerVER.push_back(-1.0);
          danglerET.push_back(-1.0);
          danglerXStartT.push_back(-999.0);
          danglerYStartT.push_back(-999.0);
          danglerZStartT.push_back(-999.0);
          danglerPXT.push_back(-999.0);
          danglerPYT.push_back(-999.0);
          danglerPZT.push_back(-999.0);
          danglerAngleDiff.push_back(-1.0);
          danglerEff.push_back(maxEff);
          danglerPur.push_back(maxPur);

          fHistoDEffvPur->Fill(maxPur,maxEff);
        }
      }


      matchedProngs = matched;
      unmatchedProngs = unmatched;
      danglerProngs = dangler;
      fHistoMatched->Fill(matched);
      fHistoUnMatched->Fill(unmatched);
      fHistoLeft->Fill(dangler);
      fHistoUvL->Fill(dangler,unmatched);

      double dEnergy = 0.0;
      for(std::set<art::Ptr<rb::CellHit>>::iterator dItr = dCells.begin(); dItr!=dCells.end(); dItr++){
        std::vector<sim::FLSHit> tHits = bt->HitToFLSHit((*dItr));
        for (unsigned int f=0; f<tHits.size(); ++f){
          dEnergy += tHits[f].GetEdep();
        }
      }
      score = sqrt(score);
      //add weight factor to penalize for unmatched reco tracks
      //std::cout<<dEnergy<<" "<<NuEVSl/NuEff<<std::endl;
      score = (1.0 - dEnergy*NuEff/NuEVSl)*score;
      eventScore = score;
      fHistoEventScore->Fill(score);

      //std::cout<<"matched: "<<matched<<" unmatched: "<<unmatched<<" dangler: "<<dangler<<" score: "<<score<<std::endl;
      
      //fill tree
      fOutTree->Fill();
    }//end loop over slices


    
  }

  //.........................................................................
  void FuzzyKValidate::ClearVectors()
  {
    view.clear();
    sX.clear();
    sY.clear();
    sZ.clear();
    dX.clear();
    dY.clear();
    dZ.clear();
    ncells.clear();
    ncellsX.clear();
    ncellsY.clear();
    gevTot.clear();
    totGeV.clear();
    gevX.clear();
    gevY.clear();
    len.clear();
    matchedPDG.clear();
    matchedProngIndx.clear();
    matchedVET.clear();
    matchedVER.clear();
    matchedET.clear();
    matchedXStartT.clear();
    matchedYStartT.clear();
    matchedZStartT.clear();
    matchedPXT.clear();
    matchedPYT.clear();
    matchedPZT.clear();
    matchedAngleDiff.clear();
    matchedEff.clear();
    matchedPur.clear();
    matchedMother.clear();
    unmatchedMother.clear();
    unmatchedPDG.clear();
    unmatchedVET.clear();
    unmatchedET.clear();
    unmatchedXStartT.clear();
    unmatchedYStartT.clear();
    unmatchedZStartT.clear();
    unmatchedPXT.clear();
    unmatchedPYT.clear();
    unmatchedPZT.clear();
    matchedProngs = 0;
    unmatchedProngs = 0;
    ccnc = -1;
    NuE = -1.0;
    NuEff = -1.0;
    NuPur = -1.0;
    NuEVSl = -1.0;
    NuEVTot = -1.0;
    Vx = -999.0;
    Vy = -999.0;
    Vz = -999.0;
    NuPdg = 0;
    mode = -1;
  }
  
  //.........................................................................
  std::vector<int> FuzzyKValidate::getFaveIdsByEfficiency(std::vector<cheat::NeutrinoWithIndex>& allNus, 
                                                          std::vector< std::vector<cheat::NeutrinoEffPur> >& sliceTruthTable)
  {
    // double check that slice truth table is safe
    for(auto slc = sliceTruthTable.begin(); slc != sliceTruthTable.end(); ++slc)
    {
      assert(slc->size() == allNus.size() && "Slice/truth table must be properly shaped");
    }
    
    // We'll make an array with an entry for every subevt to store the unique identifier of 
    // the favourite neutrino within the subevt, with the correct size and all entries initialized to -1
    std::vector<int> faveIds(sliceTruthTable.size(), -1);
    
    // if no neutrinos, there are no favorites 
    if(allNus.size()  == 0) 
      return faveIds;
    
    
    // List of best slice for each nu by slicer efficiency 
    std::vector<int> bestSlices(allNus.size());
    //First we find the best slice for each neutrino
    for(size_t iNu = 0; iNu < allNus.size(); ++iNu)
    {
      
      float maxValue =  0;  // we will use total slicer efficiency  
      int bestSlice   = -1;
      for(size_t iSlc = 0; iSlc < sliceTruthTable.size(); ++iSlc)
      {
        
        if(!sliceTruthTable[iSlc][iNu].neutrinoInt.isNull())
        {
          if(sliceTruthTable[iSlc][iNu].efficiency > maxValue)
          {
            maxValue = sliceTruthTable[iSlc][iNu].efficiency;
            bestSlice = iSlc;
          }
        }
      }
      
      bestSlices[iNu] = bestSlice;
    }
    
    // List of best nu for each slice
    
    // now we find the best neutrino for each slice.  
    for(size_t iSlc = 0; iSlc < sliceTruthTable.size(); ++iSlc)
    {
      
      float maxValue =  0;  // we will use slicer efficiency 
      int bestNu   = -1;
      for(size_t iNu = 0; iNu < sliceTruthTable[iSlc].size(); ++iNu)
      {
        
        if(!sliceTruthTable[iSlc][iNu].neutrinoInt.isNull())
        {
          if(sliceTruthTable[iSlc][iNu].efficiency > maxValue)
          {
            maxValue = sliceTruthTable[iSlc][iNu].efficiency;
            bestNu = iNu;
          }
        }
      }
      if(bestSlices[bestNu] == (int)iSlc)
      {
        faveIds[iSlc] = bestNu;
      }
    }
    
    
    
    return faveIds;
    
  }   

  //.........................................................................
  float FuzzyKValidate::GeVCalc(art::Ptr<rb::Prong> clus, geo::View_t v) 
  {
    art::ServiceHandle<calib::Calibrator> cal;
    int ncells = clus->NCell(v);
    float gevTot = 0.0;
    for(int j = 0; j < ncells; j++){
      rb::RecoHit rhit = clus->RecoHit(v, j);
      if(!rhit.IsCalibrated()){
        std::cerr << "Not calibrated?!" << std::endl;
        continue;
      }
      gevTot += rhit.GeV()*clus->Weight(v, j);
    }  
    return gevTot;
  }

  //.........................................................................
  struct Counts
  {
    Counts()
    {
      nHits[0] = nHits[1] = 0;
      nOnlyHits[0] = nOnlyHits[1] = 0;
    }
    int nHits[2]; // Indexed by view
    int nOnlyHits[2]; // Hit by only one particle
  };
  
  //......................................................................
  std::set<int> FuzzyKValidate::FindVisibleProngs3D(const art::PtrVector<rb::CellHit>& allhits,
                                                    std::set<int> ids) {

    std::map<int, Counts> counts;
    art::ServiceHandle<cheat::BackTracker> bt;

    // loop over all hits and check if this track id contributed to them
    //std::cout<<"Finding visible prongs, preparing to loop over hits"<<std::endl;
    for(unsigned int i = 0; i < allhits.size(); ++i){

      //get vector of track id's that contribute to this hit
      const art::Ptr<rb::CellHit> hit = allhits[i];
      //std::cout<<"Getting hit "<<i<<std::endl;
      const std::vector<cheat::TrackIDE> ProngIDs = bt->HitToTrackIDE(hit);

      // loop over all the returned track ids
      int nProngs = 0; // number of Prongs that contributed energy to the hit
      int loneID = -1; // Which ID it was if there was only one

      for(unsigned int j = 0; j < ProngIDs.size(); ++j){

        const int ProngID = ProngIDs[j].trackID;
        const float eFrac = ProngIDs[j].energyFrac;
        //detector energy resolution is around 10%
        //to be a visible track we are looking for at least 1 hit in each view 
        //where a true particle deposited 20% (2 sigma) of the energy in that cell
        //this is a very minimum definition but leaves the most room for optimization
        if(eFrac >= 0.2){
          ++nProngs;
          loneID = ProngID;

	  // check if the current id matches the input Prong id
          if(ids.find(ProngID) != ids.end()){
            ++counts[ProngID].nHits[hit->View()];
          }
        }
      } // end for j

      if(nProngs == 1){
        ++counts[loneID].nOnlyHits[hit->View()];
      }

    } // end for i

    const int kMinHits = 1;    // minimum number of hits to be considered visible
    const int kMinOnlyHit = 0; // minimum number of hits in which it is only particle contributing energy to be visible
    std::set<int> ret;

    for(std::set<int>::iterator it = ids.begin(); it != ids.end(); ++it){

      const Counts c = counts[*it];
      if((c.nOnlyHits[geo::kX] >= kMinOnlyHit) && (c.nOnlyHits[geo::kY] >= kMinOnlyHit)){
        if((c.nHits[geo::kX] >= kMinHits) && (c.nHits[geo::kY] >= kMinHits)){
          //std::cout<<"For track: "<<(*it)<<" xcells: "<<c.nHits[geo::kX]<<" ycells "<<c.nHits[geo::kY]<<std::endl;
          ret.insert(*it);
        }
      }
      
    }
    return ret;
  }  // end FindVisibleProngs


} // end namespace numuCC
////////////////////////////////////////////////////////////////////////

namespace fuzz
{
  DEFINE_ART_MODULE(FuzzyKValidate);
}
////////////////////////////////////////////////////////////////////////
